////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EBasic.cpp: implementation of the CEBasic class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <stdarg.h>

//////////////////////////////////////////////////////////////////////
// CEObjRef
//////////////////////////////////////////////////////////////////////
CEObjRef::CEObjRef(CEObject* ref) : CEObject(true) { m_ref=ref; }
bool CEObjRef::contain_vector(void * obj) {
	if(m_ref!=NULL) { if((m_ref == obj) || (m_ref->contain_vector(obj))) return true; }
	return false;
}
CEObject* CEObjRef::copy() const { return new CEObjRef(m_ref); }

//
CDebug* CEBasic::dbg = new CDebug("CEBasic");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CEBasic::CEBasic() { }
CEBasic::~CEBasic() { }

//////////////////////////////////////////////////////////////////////
// Library functions
//////////////////////////////////////////////////////////////////////

// DeleteObj: used to short 'delete x' statements before 'return'
void CEBasic::DeleteObj(int n, CEObject* obj1, ...) {	//'...' is CEObject* obj2, obj3, obj4,...
	//CScriptFuncInfo* CScriptFuncInfo::createFuncInfo(const char* f_title, void* p_func, int args_num, int min_args_number, ...) {
	va_list vl; va_start(vl,obj1);
	if(obj1!=NULL)delete obj1;
	for(int i=1;i<n;i++){ //iterate left 'n-1' objects
		CEObject* obj=va_arg(vl,CEObject*); if(obj!=NULL)delete obj;
	}
	va_end(vl);
}

CExpression* CEBasic::Congugate(CExpression* src, bool del_src) { // destroys the source
	if (src==NULL)return NULL;
	CExpression* res=src->conjugated();
	if (res!=NULL) {
		if ((del_src)&&(res!=src)) delete src;
	}  else {
		res=new CFuncConjugated(src); if(del_src)delete src;
	}
	return res;
}

/*
CExpression* CEBasic::Congugate(CExpression* src_const) { // unaffects the source
	if (src_const==NULL)return NULL;
	CExpression* res=src_const->conjugated();
	if (res==NULL) { res=new CFuncConjugated(src_const); }
	return res;
}
*/

/*
*	destroys the source
*	applies attributes to the whole object
*/
CESum* CEBasic::convertToSum(CExpression* src) {
	if (src == NULL) return NULL;
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return (CESum*)src;

	CESum* res = new CESum();
	src->copyAttributesTo(res, true);
	src->clear_all_attributes();
	res->addItem(src);
	return res;
}

/*
*	destroys the source
*	applies attributes to the whole object
*/
CEProduct* CEBasic::convertToProduct(CExpression* src) {
	if (src == NULL) return NULL;
	if (src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return (CEProduct*)src;

	CEProduct* res = new CEProduct();
	src->copyAttributesTo(res, true);
	src->clear_all_attributes();
	res->addItem(src);
	return res;
}

CExpression* CEBasic::inverseLevelAdv(CExpression* src) { // x->Prod{x^{-1}}, destroys the source
	if (src==NULL) return NULL;
	if (src->getLevel()) {
		src->inverseLevel();
		CEProduct* pr=new CEProduct(); pr->addItem(src);
		src=pr;
	} else {
		src->inverseLevel();
	}
	return src;
}

bool CEBasic::isZero(CExpression* src) {
	if (src == NULL) return false;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) return false;
	
	CENumber* val = (CENumber*)src;
	bool res = ((val->getNum() == 0)
		&& (val->getNumIm() == 0)
		&& (val->getDen() != 0)
		&& (val->getLevel()));
	return res;
}

bool CEBasic::isZeroProd(CExpression* src) { // parces products
	if (src == NULL) return false;
	if (src->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) return CEBasic::isZero(src);
	if (src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		CEProduct* pr = (CEProduct*)src;
		long s = pr->size();
		for(long i=0; i<s; i++) { CExpression* item=pr->GetAt(i); if(item==NULL)continue; if(item->getLevel() && isZeroProd(item)) return true; }
		return false;
	}
	return false;
}

bool CEBasic::isZeroSeq(CExpression* src) { // parces sequences
	if (src == NULL) return false;
	if (src->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) return CEBasic::isZero(src);
	if (src->ClassName() == CSequence::CLASS_NAME) {
		CSequence* seq = (CSequence*)src;
		long s = seq->size();
		for(long i=0; i<s; i++) { if (!isZeroSeq(seq->GetAt(i))) return false; }
		return true;
	}
	return false;
}

bool CEBasic::isUnit(CExpression* src) {
	if (src == NULL) return false;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) return false;
	
	CENumber* val = (CENumber*)src;
	bool res = ((val->getNum() == 1)
		&& (val->getNumIm() == 0)
		&& (val->getDen() == 1)
		&& (val->getFlag()));
	return res;
}

bool CEBasic::isExpression(CEObject* src) {
	if (src == NULL) return false;
	bool res = ((src->GetObjectKind() & CEObject::KIND_EXPRESSION) == CEObject::KIND_EXPRESSION);
	return res;
}

// reduces & checks zero
bool CEBasic::tendsZero(CExpression* src) {
	if (src == NULL) return false;
	if (CEBasic::isZero(src)) return true;
	
	src = CExpression::reduce(src->clone());
	bool res = CEBasic::isZero(src);
	delete src;
	return res;
}

bool CEBasic::equalsSignless(CExpression* a, CExpression* b) {
	bool a_Msign = (!a->getFlag()); if (a_Msign) a->inverseFlag();
	bool b_Msign = (!b->getFlag()); if (b_Msign) b->inverseFlag();
	bool res = a->equals(b);
	if (a_Msign) a->inverseFlag();
	if (b_Msign) b->inverseFlag();
	return res;
}

bool CEBasic::equalsToSymbol(CExpression* a, const char* s) {
	bool res=false;
	if ((a->ClassName() == CESymbol::CLASS_NAME) && (a->getFlag()) && (a->getLevel())) {
		res=CEUtil::equals(((CESymbol*)a)->get_name(),s);
	}
	return res;
}

long CEBasic::CalcComplexity(CExpression* src) { return CRCollectSymbols::CountObjects(src); }

CExpression* CEBasic::GetSqrt2() { return CEPower::sqrt(CENumber::GetTwo()); }
CExpression* CEBasic::m_refPi = NULL; // cache for GetPiSym()
CExpression* CEBasic::GetPiSym() { // const result, never NULL
	if (m_refPi==NULL) {
		m_refPi = CEProcessor::getDataStorage()->gete(CConst::GL_PI);
		if (m_refPi==NULL) { m_refPi=new CESymbol("Pi"); CEProcessor::AddToDeleteList(m_refPi); }
	}
	return m_refPi;
}

void CEBasic::appendLineBreak(CEString* str, int mode) {
	if (str == NULL) return;

	switch (mode) {
	case STORE_WR_MATHEMATICA:
		str->AddString("\\[IndentingNewLine]\n");
		break;
	case STORE_TEX:
		str->AddString("\\newline\n");
		break;
	case STORE_PLAIN_TEXT:
	default:
		str->AddString("\n");
	}
}

char* CEBasic::GetSymbolLine(int len, char sym) {
	if (len < 0) return NULL;
	char* res = (char*)CMemoryManager::mmalloc(len+1);
	int i;
	for(i=0; i<len; i++) { *(res + i) = sym; }
	*(res + len) = '\0';
	return res;
}

CEString* CEBasic::GetSymbolString(int len, char sym) {
	char* str = GetSymbolLine(len, sym);
	CEString* res = new CEString(str);
	CMemoryManager::mfree(str);
	return res;
}

// returns "NULL" for NULL object
CEString* CEBasic::ObjToString(CESerializable* obj, int mode) {
	CEString* res = NULL;
	if (obj != NULL) res = obj->toString(mode);
	else res = new CEString("NULL");
	return res;
}

CExpression* CEBasic::m_TempIndex = NULL;
CExpression* CEBasic::InverseTensorIndex(CExpression* src_const, CExpression* ind, bool to_cov) { // returns NULL if fails
	if (m_TempIndex==NULL) { m_TempIndex=CESymbol::getUnique(); CEProcessor::AddToDeleteList(m_TempIndex); }

	if (src_const==NULL) return NULL;
	if (!CRContainCheck::search(src_const,ind)) return NULL;

	CExpression* val=CExpression::reduce(src_const->clone()->multiply(new CEMetric(ind,m_TempIndex,to_cov,to_cov)));
	if (CRContainCheck::search(val,ind)) { delete val; return NULL; }

	val=CEReplasement::Replace(val,m_TempIndex,ind);
	if (CRContainCheck::search(val,m_TempIndex)) { delete val; return NULL; }

	return val;
}

/*
bool CEBasic::LeaveSeries(CEProduct* src, CExpression* var, int order_to) {
	if ((src == NULL) || (var == NULL)) { dbg->log(0, "{Error} LeaveSeries(): NULL call"); return false; }
	if (order_to != 0) { dbg->log(0, "LeaveSeries(): unsupported call"); return false; }

	// unimplemented
	dbg->log(0, "{ERROR} LeaveSeries(): not implemented code");

	// should be implemented using something like
	//virtual CExpression* GetSeriesToOrder(CExpression* var, int order);

	bool chd = false;

	return chd;
}
*/
