////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EComplexValue.cpp: implementation of the CEComplexValue class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEComplexValue::CEComplexValue(double _re, double _im) : CEValue()
{
	m_RE = _re;
	m_IM = _im;
}

CEComplexValue::~CEComplexValue()
{
}

DWORD CEComplexValue::get_val_type()
{
	return CEValue::VTYPE_COMPLEX;
}

unsigned long CEComplexValue::props_long() {
	unsigned long res = CExpression::props_long();
	res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}

void CEComplexValue::collectProperties(CEProperties* props, unsigned set) {
	CEValue::collectProperties(props, set);
	props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

double CEComplexValue::get_RE() {
	return m_RE;
}

double CEComplexValue::get_IM() {
	return m_IM;
}

void CEComplexValue::normalize()
{
	if (!getFlag()) {
		m_RE = -m_RE;
		m_IM = -m_IM;
		inverseFlag();
	}
	if (!getLevel()) {
		CEValue* inv_res = inverse();
		if (inv_res != this) {printf("ERROR inline inversing of complex value\n");}
	}
}

CExpression* CEComplexValue::add(CExpression * other)
{
	if (other == NULL) return NULL;
	if (other->getProperty(CEProperties::PROP_MASK_IS_NUMERIC_VALUE)) {
		DWORD val_type = ((CEValue*)other)->get_val_type();
		if (val_type & CEValue::VTYPE_COMPLEX) {
			CEComplexValue* cv = (CEComplexValue*)other;
			normalize();
			cv->normalize();
			m_RE += cv->get_RE();
			m_IM += cv->get_IM();
			delete cv;
			return this;
		}
		else if (val_type & CEValue::VTYPE_REAL) {
			CERealValue* rv = (CERealValue*)other;
			normalize();
			m_RE += rv->to_double();
			delete rv;
			return this;
		}
	}

	//This should not occur
	dbg->log(1, "Unexpected case");
	return CExpression::add(other);
}

CExpression* CEComplexValue::multiply(CExpression * other)
{
	if (other == NULL) return NULL;
	if (other->getProperty(CEProperties::PROP_MASK_IS_NUMERIC_VALUE)) {
		DWORD val_type = ((CEValue*)other)->get_val_type();
		if (val_type & CEValue::VTYPE_COMPLEX) {
			CEComplexValue* cv = (CEComplexValue*)other;
			normalize();
			cv->normalize();
			double re1 = m_RE;
			double im1 = m_IM;
			double re2 = cv->get_RE();
			double im2 = cv->get_IM();
			m_RE = re1*re2-im1*im2;
			m_IM = re1*im2+im1*re2;
			delete cv;
			return this;
		}
		else if (val_type & CEValue::VTYPE_REAL) {
			CERealValue* rv = (CERealValue*)other;
			normalize();
			double f = rv->to_double();
			m_RE *= f;
			m_IM *= f;
			delete rv;
			return this;
		}
	}

	//This should not occur
	dbg->log(1, "Unexpected case");
	return CExpression::multiply(other);
}

CEValue* CEComplexValue::inverse(bool do_coinverse_level)
{
	double f = m_RE*m_RE+m_IM*m_IM;
	if (f == 0) {
		dbg->log(0, "Error: divizion by zero during inversing");
		return NULL;
	}
	m_RE = m_RE/f;
	m_IM = -m_IM/f;
	return this;
}

CEString* CEComplexValue::toString(int mode)
{
	char* data_re = (char*)CMemoryManager::mmalloc(32);
	char* data_im = (char*)CMemoryManager::mmalloc(32);
	sprintf(data_re, "%g", m_RE);
	sprintf(data_im, "%g", m_IM);
	
	CEString* res = new CEString(data_re);
	if (m_IM >= 0) res->AddString("+");
	res->AddString(data_im);
	res->AddString("i");
	res->insertString("\\((", 0);
	if (!getFlag()) res->insertString("- ", 0);
	res->AddString(")\\)");
	CMemoryManager::mfree(data_re);
	CMemoryManager::mfree(data_im);
	return res;
}

CExpression* CEComplexValue::clone() const
{
	CEComplexValue* res = new CEComplexValue(m_RE, m_IM);
	copyAttributesTo(res);
	return res;
}

/*CEValue* CEComplexValue::Sqrt()
{
	dbg->log(0, "Error: SQRT(COMPLEX) is not implemented.");
	CEComplexValue* res = NULL;
	//if (!getLevel()) res->inverseLevel();
	return res;
}

CEValue* CEComplexValue::Sqr()
{
	CEComplexValue* res = new CEComplexValue(m_RE*m_RE-m_IM*m_IM, 2*m_RE*m_IM);
	if (!getLevel()) res->inverseLevel();
	return res;
}

void CEComplexValue::Abs()
{
	normalize();
	//if (!getFlag()) apply_sign();
	m_value = fabs(m_value);
}*/

bool CEComplexValue::less_than(CExpression * other)
{
	if (other->getProperty(CEProperties::PROP_MASK_IS_NUMERIC_VALUE)) {
		CEValue* ov = (CEValue*)other;
		DWORD val_type = ov->get_val_type();
		if (val_type & CEValue::VTYPE_COMPLEX) {
			if (get_RE() < ((CEComplexValue*)ov)->get_RE()) return true;
			//else
			return (get_IM() < ((CEComplexValue*)ov)->get_IM());
		}
		else return (get_RE() < ov->to_double());
	}
	return CExpression::less_than(other);
}

CEValue* CEComplexValue::Sqrtx(bool do_clone)
{
	if (do_clone) return ((CEValue*)clone())->Sqrtx();
	normalize();
	dbg->log(0, "Warning: SQRT(COMPLEX) is not implemented.");

	m_RE = sqrt(sqrt(m_RE*m_RE+m_IM*m_IM));
	m_IM = 0;
	return this;
}

CEValue* CEComplexValue::Absx(bool do_clone)
{
	if (do_clone) return ((CEValue*)clone())->Absx();
	normalize();
	m_RE = sqrt(m_RE*m_RE+m_IM*m_IM);
	m_IM = 0;
	return this;
}

CEValue* CEComplexValue::Sqrx(bool do_clone)
{
	if (do_clone) return ((CEValue*)clone())->Sqrx();
	normalize();
	double r = m_RE;
	double i = m_IM;
	m_RE = r*r-i*i;
	m_IM = 2*r*i;
	return this;;
}

CEValue* CEComplexValue::multiply(double val)
{
	normalize();
	m_RE *= val;
	m_IM *= val;
	return this;
}

double CEComplexValue::to_double()
{
	double res = m_RE;
	if (!getFlag()) res = -res;
	if (!getLevel()) {
		double f = m_RE*m_RE+m_IM*m_IM;
		res = res/f;
	}
	return res;
}


/*
*	      ?
*/
bool CEComplexValue::equals(CExpression * other)
{
	if (other->getProperty(CEProperties::PROP_MASK_IS_NUMERIC_VALUE)) {
		if (((CEValue*)other)->get_val_type() & CEValue::VTYPE_COMPLEX) {
			if (getFlag() != other->getFlag()) return false;
			if (getLevel() != other->getLevel()) return false;
			CEComplexValue* cv = (CEComplexValue*)other;

			if (m_RE != cv->get_RE()) return false;
			if (m_IM != cv->get_IM()) return false;
			return true;
		}
	}
	return false;
}
