////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EReplasement.cpp: implementation of the CEReplasement class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const char* CEReplasement::CLASS_NAME = "CEReplasement";
const DWORD CEReplasement::REPL_AVOID_NULL_DENOMINATORS	= 0x01;
const DWORD CEReplasement::REPL_NOT_IN_SUBSUMS			= 0x02;
const DWORD CEReplasement::REPL_NOT_IN_SUMS				= 0x04;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEReplasement::CEReplasement(const CExpression* a, const CExpression* b, bool do_exchange) : CERule()
{
	if ((a == NULL) || (b == NULL)) { dbg->log(0, "Error: CEReplasement(NULL parameter) creation"); CEException::halt(); }
	m_a = a->clone();
	m_b = b->clone();
	m_do_exchange = do_exchange;
	m_operation_flags = 0;
}

CEReplasement::CEReplasement(const CExpression* a, const CExpression* b,
							 bool do_exchange, DWORD _operation_flags) : CERule()
{
	m_a = a->clone();
	m_b = b->clone();
	m_do_exchange = do_exchange;
	m_operation_flags = _operation_flags;
}

CEReplasement::~CEReplasement()
{
	delete m_a;
	delete m_b;
}

DWORD CEReplasement::GetOperationFlags() { return m_operation_flags; }
void CEReplasement::SetOperationFlags(DWORD _operation_flags) { m_operation_flags=_operation_flags; }

const char* CEReplasement::ClassName() const { return CLASS_NAME; }
CEString* CEReplasement::toString(int mode) {
	CEString* res  = new CEString("\\({");
	res->AddString(a()->toString(mode));
	if (exchange()) res->AddString("<->");
	else res->AddString("-->");
	res->AddString(b()->toString(mode));
	if (m_operation_flags != 0) { res->AddString(" #"); res->add_int(m_operation_flags); }
	res->AddString("}\\)");
	return res;
}

CERule* CEReplasement::GetRuleCopy() const { return new CEReplasement(m_a, m_b, m_do_exchange, m_operation_flags); }

bool CEReplasement::contain_vector(void * obj) {
	if ((m_a == obj) || (m_b == obj)) return true;
	if (m_a->contain_vector(obj)) return true;
	if (m_b->contain_vector(obj)) return true;
	return false;
}

CExpression* CEReplasement::a()
{
	return m_a;
}

CExpression* CEReplasement::b()
{
	return m_b;
}

bool CEReplasement::exchange()
{
	return m_do_exchange;
}

/*
*	! deletes previous value.
*/
void CEReplasement::set_a(CExpression *_a, bool do_copy)
{
	if (_a == NULL) return;

	if (_a == m_a) {
		//if (do_copy) m_a = m_a->clone()	???
	}
	else {
		if (m_a != NULL) delete m_a;
		m_a = _a;
		if (do_copy) m_a = m_a->clone();
	}
}

/*
*	! deletes previous value.
*/
void CEReplasement::set_b(CExpression *_b, bool do_copy)
{
	if (_b == NULL) return;

	if (_b == m_b) {
		//if (do_copy) m_b = m_b->clone()	???
	}
	else {
		if (m_b != NULL) delete m_b;
		m_b = _b;
		if (do_copy) m_b = m_b->clone();
	}
}

/*
*	Important: a-value should be both '+'-signed and '+'-leveled.
*	So and b-value in case of crossing replacements.
*	
*	In the other case the following procedures should be updated.
*/
CExpression* CEReplasement::ApplyTo(CExpression *src, bool del_src)
{
	CExpression* res = do_replace(src, del_src);
	if (res == NULL) res = src;
	return res;
}

void CEReplasement::ResetTimesApplied() { m_times_applied=0; }
DWORD CEReplasement::GetTimesApplied() { return m_times_applied; }

CExpression* CEReplasement::do_replace(CExpression *src, bool do_delete_src)
{
	if (src == NULL) return NULL;

	if (m_operation_flags & REPL_NOT_IN_SUMS) {
		if(src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			// do not apply rule here, skip sub-items
			return src;
		}
	}

	CExpression* res = NULL;
	bool sign = src->getFlag();
	bool lev = src->getLevel();
	bool skip_subitems = false;

	if (!sign) src->setFlag();
	if (!lev) src->setLevel(true);

	if (src->equals(a())) {							//	1. checking for forward exchange
		bool repl_OK = true;

		if ((m_operation_flags & REPL_AVOID_NULL_DENOMINATORS) && (!lev)) {
			if (CEBasic::isZero(b())) repl_OK = false; skip_subitems = true;
		}

		if (repl_OK) res = b()->clone();
	}
	else if ((exchange()) && (src->equals(b()))) {	//	2. checking for backward exchange
		bool repl_OK = true;

		if ((m_operation_flags & REPL_AVOID_NULL_DENOMINATORS) && (!lev)) {
			if (CEBasic::isZero(a())) repl_OK = false; skip_subitems = true;
		}

		if (repl_OK) res = a()->clone();
	}

	if (res != NULL) {
		m_times_applied++;
		if (do_delete_src) {
			delete src;
		}
		else {
			// restore source
			if (!sign) src->unsetFlag();
			if (!lev) src->setLevel(lev);
		}
		if (!sign) res->inverseFlag();
		if (!lev) res->inverseLevel();
		return res;
	}

	// restore source
	if (!sign) src->unsetFlag();
	if (!lev) src->setLevel(lev);

	if (skip_subitems) return src;

	// cahce source expression for possible replacement cancel
	CExpression* src_cpy = NULL;
	if ((m_operation_flags & REPL_AVOID_NULL_DENOMINATORS) && (!lev)) { src_cpy = src->clone(); }

	//	3. Applying to sub-items.
	DWORD flag_old = 0;
	if ((m_operation_flags & REPL_NOT_IN_SUBSUMS) && ((m_operation_flags & REPL_NOT_IN_SUMS) == 0)) {
		flag_old=m_operation_flags; m_operation_flags |= REPL_NOT_IN_SUMS;
	}
	res = src->applyRule(this);
	if (flag_old != 0) m_operation_flags=flag_old; // restore flag
	if ((m_operation_flags & REPL_AVOID_NULL_DENOMINATORS) && (!lev) && (res != NULL)) {
		// check result for zero
		CExpression* res_cpy = res->clone();
		res_cpy->inverseLevel(); // we had a denominator here
		res_cpy = CExpression::arrge(res_cpy);
		if (CEBasic::isZero(res_cpy)) { // cancel the replacement
			if (res != src) delete res;
			res = src_cpy; src_cpy = NULL;
		}
		delete res_cpy;
	}
	if (src_cpy != NULL) delete src_cpy;

	if ((res != src) && (res != NULL)) {
		if (do_delete_src) delete src;
		return res;
	}
	
	//	source object not changed
	return src;
}

//////////////////////////////////////////////////////////////////////
// Static Entry
//////////////////////////////////////////////////////////////////////
CExpression* CEReplasement::Replace(CExpression* source, CExpression* a, CExpression* b, bool do_exchange) {
	if (source == NULL) return source;
	if ((a == NULL) || (b == NULL)) {
		dbg->log(0, "<Warning> CEReplasement::Replace() NULL parameter, operation skipped. Source is", source);
		return source;
	}
	
	CEReplasement r(a, b, do_exchange);
	return r.ApplyTo(source);
}
