////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// ESymbol.cpp: implementation of the CESymbol class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// CENull
//////////////////////////////////////////////////////////////////////
CENull::CENull() : CExpression() {}
const char* CENull::CLASS_NAME = "CENull";
const char* CENull::ClassName() const { return CLASS_NAME; }
bool CENull::IsNull(CExpression* src) { return ((src!=NULL)?(src->ClassName()==CENull::CLASS_NAME):false); }
CExpression* CENull::clone() const { CENull* res=new CENull(); copyAttributesTo(res); return res; }
bool CENull::equals(CExpression *other) {
	if (other==NULL) return false;
	if (!attributes_equals_to(other)) return false;
	return (other->ClassName()==CENull::CLASS_NAME);
}
CEString* CENull::toString(int mode) { CEString* res=new CEString("Null"); if(!getFlag())res->insertString("- ",0); return res; }

const char* CENull::CLASS_TAG_NAME = "e_null";
const char* CENull::get_TAG_Name() { return CLASS_TAG_NAME; }
long CENull::store_stream(ofstream* p_ofs, DWORD settings) { return 0; }
CESerializable* CENull::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) { return new CENull(); }

//////////////////////////////////////////////////////////////////////
// CEInfinity
//////////////////////////////////////////////////////////////////////
CEInfinity::CEInfinity() : CExpression() {}
const char* CEInfinity::CLASS_NAME = "CEInfinity";
const char* CEInfinity::ClassName() const { return CLASS_NAME; }
CExpression* CEInfinity::clone() const { CEInfinity* res=new CEInfinity(); copyAttributesTo(res); return res; }
bool CEInfinity::equals(CExpression *other) {
	if (other==NULL) return false;
	if (!attributes_equals_to(other)) return false;
	return (other->ClassName()==CEInfinity::CLASS_NAME); /* or do not unify infinities?*/
}
CEString* CEInfinity::toString(int mode) { CEString* res=new CEString("Infinity"); if(!getFlag())res->insertString("- ",0); return res; }

const char* CEInfinity::CLASS_TAG_NAME = "infinity";
const char* CEInfinity::get_TAG_Name() { return CLASS_TAG_NAME; }
long CEInfinity::store_stream(ofstream* p_ofs, DWORD settings) { return 0; }
CESerializable* CEInfinity::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) { return new CEInfinity(); }

//////////////////////////////////////////////////////////////////////
// CESymbol
//////////////////////////////////////////////////////////////////////
const char* CESymbol::CLASS_NAME = "CESymbol";
const char*	CESymbol::CLASS_TAG_NAME = "s";
long CESymbol::unique_num = 0;
bool CESymbol::m_UseCharMap = false;
bool CESymbol::m_NoCharMapForUniques = true;

// skips spaces in name (NO!!!)
CESymbol::CESymbol(const char* name):CExpression() {
	unsigned len = 1;
	unsigned i;
	for(i=0; *(name + i) != '\0'; i++) {
		len++; // if (*(name + i) != ' ') len++;
	}
	bool use_CM = m_UseCharMap;
	if (use_CM && m_NoCharMapForUniques) { if (*name == '$') use_CM = false; } // is unique
	if (use_CM) {
		m_name = (char*)CLSymbolDataBank::GetCharOrigin(name, true);
		if (m_name == NULL) { CEException::halt("<CESymbol> Can't create string in CharMap"); }
		m_CharFromMap = true;
	} else {
		// fast copy (TODO: copy as long should be 4 times faster)
		m_name = (char*)CMemoryManager::mmalloc(len);
		for (i=0; i<len; i++) {
			*(m_name+i)=(char)*(name+i); // if (*(name + i) != ' ') { *(m_name + i) = *(name + i); }
		}
		m_CharFromMap = false;
	}
	m_len = len;
	m_is_commutative = true;
	m_bSelfConjugated = true;
}

// skips spaces in name (NO!!!)
CESymbol::CESymbol(const char* name, bool _is_commutative, bool _bSelfConjugated)
:CExpression() {
	unsigned len = 1;
	unsigned i;
	for (i=0; *(name + i) != '\0'; i++) {
		len++; //if (*(name + i) != ' ') len++;
	}
	bool use_CM = m_UseCharMap;
	if (use_CM && m_NoCharMapForUniques) { if (*name == '$') use_CM = false; } // is unique
	if (use_CM) {
		m_name = (char*)CLSymbolDataBank::GetCharOrigin(name, true);
		if (m_name == NULL) { CEException::halt("<CESymbol> Can't create string in CharMap"); }
		m_CharFromMap = true;
	} else {
		m_name = (char*)CMemoryManager::mmalloc(len);
		for (i = 0; i<len; i++) {
			*(m_name+i)=(char)*(name+i); //if (*(name + i) != ' ') { *(m_name + i) = *(name + i); }
		}
		m_CharFromMap = false;
	}
	m_len = len;
	m_is_commutative = _is_commutative;
	m_bSelfConjugated = _bSelfConjugated;
}

// clone-constructor
CESymbol::CESymbol(bool is_char_from_bank, const char* name, unsigned len, bool _is_commutative, bool _bSelfConjugated)
:CExpression() {
	bool use_CM = m_UseCharMap;
	if ((use_CM) && (m_NoCharMapForUniques)) { if (*name == '$') use_CM = false; } // is unique
	if (use_CM) {
		if (is_char_from_bank) {
			m_name = (char*)name;
			if (m_name == NULL) { CEException::halt("<CESymbol> NULL symbol name"); }
		} else {
			m_name = (char*)CLSymbolDataBank::GetCharOrigin(name, true);
			if (m_name == NULL) { CEException::halt("<CESymbol> Can't create string in CharMap"); }
		}
		m_CharFromMap = true;
	} else {
		// fast copy (TODO: copy as long should be 4 times faster)
		m_name = (char*)CMemoryManager::mmalloc(len);
		unsigned i;
		for (i=0; i<len; i++) {
			*(m_name + i) = (char)*(name + i);
			//if (*(name + i) != ' ') { *(m_name + i) = *(name + i); }
		}
		m_CharFromMap = false;
	}

	m_len = len;
	m_is_commutative = _is_commutative;
	m_bSelfConjugated = _bSelfConjugated;
}

CESymbol::~CESymbol() {
	if (!m_CharFromMap) CMemoryManager::mfree(m_name);
}

bool CESymbol::isCharFromMap() { return m_CharFromMap; }

void CESymbol::collectProperties(CEProperties* props, unsigned set) {
	CExpression::collectProperties(props, set);
	if (!m_is_commutative) props->setCommutative(false);
	if (!m_bSelfConjugated) props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

unsigned long CESymbol::props_long() {
	unsigned long res = CExpression::props_long();
	if(!m_is_commutative) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_COMMUTATIVE,false);
	if(!m_bSelfConjugated) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}

/*
* The value of symbol is checked out in it's container.
*/
CExpression* CESymbol::applyRule(CEAbstractRule* rule) { return this; }

CExpression* CESymbol::evaluate(int level) { return this; }
CExpression* CESymbol::Simplify(int level) { return this; }
CExpression* CESymbol::Arrange(int level) { return this; }

CEString* CESymbol::toString(int mode) {
	CEString* res = NULL;
	switch (mode) {
	case STORE_PLAIN_TEXT:
	case STORE_WR_MATHEMATICA:
		if (getFlag()) { res = new CEString(m_name); } else { res = new CEString("- "); res->AddString(m_name); }
		break;
	case STORE_CPP:
	case STORE_FORTRAN_SYMBOL:
		if (getFlag()) { res = new CEString(m_name); } else { res = new CEString("-"); res->AddString(m_name); }
		break;
	case STORE_TEX:
		res = new CEString(getFlag()?"":"- ");
		if (*m_name == '$') {
			res->AddString("U");
			res->AddString(m_name+1);
		} else {
			//[old]res->AddString(m_name);

			long bLastSpec=false;
			for(long i=0;;i++) { //parse for '\_' or '\&'
				char c=*(m_name+i); if(c=='\0')break;
				if(c=='\\'){ bLastSpec=true; continue; }
				//if(bLastSpec){} //TODO convert '\\&' --> '\\^' etc. her
				res->addChar(c); bLastSpec=false;
			}
		}
		break;
	default:
		if (getFlag()) { res = new CEString(m_name); } else { res = new CEString("-"); res->AddString(m_name); }
	}
	return res;
}

void CESymbol::toStream(ostream* os, int mode) { CEException::halt(); }
const char* CESymbol::ClassName() const { return CESymbol::CLASS_NAME; }
const char* CESymbol::get_TAG_Name() { return CLASS_TAG_NAME; }
bool CESymbol::contain_vector(void * obj) { return (m_name == obj); }
CExpression* CESymbol::numeric() { return CEProcessor::find_numeric(this); }
const char* CESymbol::get_name() { return m_name; }
long CESymbol::get_name_length() { return m_len; }
bool CESymbol::isSymbolCommutative() { return m_is_commutative; }
bool CESymbol::isSymbolSelfConjugated() { return m_bSelfConjugated; }

bool CESymbol::equals(CExpression *other) {
	if (other == NULL) return false;

	bool res = false;
	if (CEUtil::equals(ClassName(), other->ClassName())) {
		if (!attributes_equals_to(other)) return false;

		CESymbol* s = (CESymbol*) other;
		if (m_CharFromMap && (s->isCharFromMap())) {
			res = ( (get_name() == s->get_name()) &&
				CEUtil::LEQU(isSymbolCommutative(), s->isSymbolCommutative()) );
		} else {
			res = ( CEUtil::equals(get_name(), s->get_name()) &&
				CEUtil::LEQU(isSymbolCommutative(), s->isSymbolCommutative()) );
		}
	}
	return res;
}


CExpression* CESymbol::clone() const {
	//CExpression* res = new CESymbol(m_name, m_is_commutative);
	CExpression* res = new CESymbol(m_CharFromMap, m_name, m_len, m_is_commutative, m_bSelfConjugated);
	copyAttributesTo(res);
	return res;
}

/*
* All usual symbols are supposed to be self-conjugated.
*/
CExpression* CESymbol::conjugated(int mode) {
	CExpression* res=((m_bSelfConjugated)?clone():(new CFuncConjugated(this)));
	return res;
}

CESymbol* CESymbol::getUnique() {
	if (unique_num==LONG_MAX) CEException::halt("Unique symbols limit exceded!");
	unique_num++;
	char* data = (char*)CMemoryManager::mmalloc(16);
	*data = '$';
	CEUtil::addACBNumber(data, unique_num, 1);
	//sprintf(data, "$%ld", unique_num);
	CESymbol* res = new CESymbol(data);
	CMemoryManager::mfree(data);
	return res;
}

CESymbol* CESymbol::getUnique(bool commutative) {
	if (unique_num==LONG_MAX) CEException::halt("Unique symbols limit exceded!");
	unique_num++;
	char* data = (char*)CMemoryManager::mmalloc(16);
	*data = '$';
	CEUtil::addACBNumber(data, unique_num, 1);
	//sprintf(data, "$%ld", unique_num);
	CESymbol* res = new CESymbol(data, commutative);
	CMemoryManager::mfree(data);
	return res;
}

bool CESymbol::isSymbolUnique(CESymbol* sym) {
	if (sym == NULL) return false;
	const char* base = sym->get_name();
	if (*base != '$') return false;
	// check also following integer value (to be implemented)
	return true;
}

bool CESymbol::less_than(CExpression *other) {
	if (other == NULL) return false;

	// compare to CESymbol, CEPower, CESProduct, CEVector, CEMetric, CEAntisymmetric, CEAntiSymmProd, CEFuncBondarev
	// a < a^n < a.b < a_nu < a_mu_nu < a_abcd
	const char* o_cn = other->ClassName();
	if (o_cn == CEPower::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CEPower* o = (CEPower*)other;
		if (less_than(o->basis())) return true;
		if (!equals(o->basis())) return false;
		return true;
	}

	if (o_cn == CESymbol::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CESymbol* s = (CESymbol*)other;
		return CEUtil::a_less_b(get_name(), s->get_name());
	}

	if (o_cn == CESProduct::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CESProduct* o = (CESProduct*)other;
		if (less_than(o->get_a())) return true;
		if (!equals(o->get_a())) return false;
		return true;
	}

	if ((o_cn==CEVector::CLASS_NAME)||(o_cn==CEMetric::CLASS_NAME)||(o_cn==CEAntisymmetric::CLASS_NAME)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		IETensor* o = IETensor::GetIETensor(other);
		if (less_than(o->base())) return true;
		if (!equals(o->base())) return false;
		return true;
	}
	
	if (o_cn==CEAntiSymmProd::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	if (o_cn==CEFuncBondarev::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	return CExpression::less_than(other);
}

CESerializable* CESymbol::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	long end_pos = CESupport::fing_item_end_pos(data, CLASS_TAG_NAME, pos);
	char* sym_str = CEUtil::copyStringFromTo(data, pos, end_pos-1);

	bool _is_comm = true; bool bSelfConj=true;
	if (params != NULL) {
		void* comm_flag = params->get("c");
		if (comm_flag != NULL) {
			char* comm_str = ((CEString*)comm_flag)->getData();
			if (*comm_str == 'n') _is_comm = false;
		}
		void* conj_flag = params->get("j");
		if (conj_flag != NULL) {
			char* conj_str = ((CEString*)conj_flag)->getData();
			if (*conj_str == 'n') bSelfConj=false;
		}
	}

	CESymbol* res = new CESymbol(sym_str, _is_comm, bSelfConj);
	CMemoryManager::mfree(sym_str);
	return res;
}

void* CESymbol::getStorageFlags() { // returns CEMap* object or NULL
	void* p_sup = CExpression::getStorageFlags();
	CEMap* res = (p_sup != NULL)?((CEMap*)p_sup):NULL;
	if (!m_is_commutative) { if(res==NULL)res = CEMap::createNewMap(); res->put_ptr("c", (void*)"n"); }
	if (!m_bSelfConjugated) { if(res==NULL)res = CEMap::createNewMap(); res->put_ptr("j", (void*)"n"); }
	return res;
}

long CESymbol::store_stream(ofstream *p_ofs, DWORD settings) {
	*p_ofs << m_name;
	return CEUtil::length(m_name);
}
