////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EPower.cpp: implementation of the CEPower class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const char* CEPower::CLASS_NAME = "CEPower";
const char* CEPower::CLASS_TAG_NAME = "pow";
CDebug* CEPower::dbg = new CDebug("CEPower");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEPower::CEPower(CExpression* basis, CExpression* power):CExpression()
{
	m_basis = basis->clone();
	m_power = power->clone();
}

CEPower::CEPower(CExpression* basis, long power):CExpression()
{
	m_basis = basis->clone();
	m_power = new CENumber(power);
}

CEPower::CEPower(const char* basis, long power):CExpression()
{
	m_basis = new CESymbol(basis);
	m_power = new CENumber(power);
}

CEPower::~CEPower()
{
	delete m_basis;
	delete m_power;
}

void CEPower::collectProperties(CEProperties* props, unsigned set) {
	if ((m_basis->props_long() & CEProperties::PROP_MASK_SELF_CONJUGATED)==0) props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

unsigned long CEPower::props_long() {
	unsigned long res = CExpression::props_long();
	if ((m_basis->props_long() & CEProperties::PROP_MASK_SELF_CONJUGATED)==0) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}

CExpression* CEPower::Simplify(int level)
{
	return NULL;
}

CExpression* CEPower::Arrange(int level) {
	m_basis = CExpression::arrge(m_basis);
	m_power = CExpression::arrge(m_power);
	arrange_basis_sign();

	if (level == AL_EXTRACT) {
		if (CEBasic::isZero(m_power)) {
			//0^0=1 ==> this check is first
			CExpression* res=CENumber::GetUnit(); copyAttributesTo(res, true); return res;
		} else if (CEBasic::isZero(m_basis)) {
			//0^0=1
			if (m_power->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
				if (((CENumber*)m_power)->is_positive()) {
					// power > 0
					CExpression* res=CENumber::GetZero(); copyAttributesTo(res,true); return res;
				} else if (((CENumber*)m_power)->is_negative()) {
					// power < 0
					CExpression* res=CENumber::GetZero(); copyAttributesTo(res,true); res->inverseLevel(); return res;
				}//else zero or complex power
			}//else symbolic power
		} else if (CEBasic::isUnit(m_basis)) {
			CExpression* res = CENumber::GetUnit(); copyAttributesTo(res, true); return res;
		} else if (CEBasic::isUnit(m_power)) {
			CExpression* res=m_basis->clone(); copyAttributesTo(res, true); return res;
		}
		
		// expand (a*b*c)^N
		CExpression* res=ExpandABCpN(); if(res!=NULL){ return CExpression::arrge(res); } // repeat arrange

		// expand {a^b}^c ==> a^{b*c}
		if(ExpandApBpC()){ return CExpression::arrge(this, false); } // repeat arrange

		// number^x calculation
		if (m_basis->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
			CExpression* res = to_product(basis(), power(),true);
			if (res != NULL) {
				copyAttributesTo(res, true);
				return CExpression::arrge(res); // repeat arrange
			}
		}
	}
	else {
		// wrong check: the sign of power should be also inspected
		//if (CEBasic::isZero(basis())) { setPower(1); }
	}
	return this;
}
CExpression* CEPower::ExpandABCpN() { //Expand (a*b*c)^N, NULL if unavailable
	if (!m_basis->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)){ return NULL; }
	CEProduct* seq = (CEProduct*)m_basis;
	CEProduct* res = new CEProduct();
	long s = seq->size();
	for (long i=0; i<s; i++) {
		CExpression* item = seq->GetAt(i);
		CExpression* new_item = NULL;
		if (item->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) new_item=to_product(item, m_power);
		if (new_item == NULL) new_item = new CEPower(item, m_power);
		res->addItem(new_item);
	}
	copyAttributesTo(res, true);
	return res;
}
bool CEPower::ExpandApBpC() { //Expands (a^b)^c => a^(b*c) inline, 'false' if unavailable
	if(m_basis->ClassName()==CEPower::CLASS_NAME){
		//Nt: no sign flag may be applied to basis here
		if ((m_basis->getFlag()) && (m_basis->getLevel())) {
			//Nt: base is -L, we should set 'this' to -L, but -L may be contained by product only
			CEPower* p_bas=(CEPower*)m_basis;
			CExpression* p_bas_base=p_bas->basis()->clone();
			CExpression* p_bas_pow=p_bas->power()->clone();
			delete m_basis;
			m_basis=p_bas_base;
			m_power=p_bas_pow->multiply(m_power);
			return true;
		}
	}
	return false;
}

CExpression* CEPower::evaluate(int level)
{
	CExpression* _old = NULL;
	CExpression* _new = NULL;
	_old = m_basis;
	_new = _old->evaluate();
	if (_old != _new) {
		if (_new != NULL) {
			delete m_basis;
			m_basis = _new;
		}
		else dbg->log(0, "<Warning><NULL as evaluate()'s result.>", this);
	}
	_old = m_power;
	_new = _old->evaluate();
	if (_old != _new) {
		if (_new != NULL) {
			delete m_power;
			m_power = _new;
		}
		else dbg->log(0, "<Warning><NULL as evaluate()'s result.>", this);
	}
	return this;
}

CExpression* CEPower::applyRule(CEAbstractRule* rule)
{
	CERule* r = (CERule*)rule;
	setBasis(r->ApplyTo(basis(), false), false);
	setPower(r->ApplyTo(power(), false), false);

	// quick arrange level
	//arrange_basis_level();
	if (!(basis()->getLevel())) { basis()->inverseLevel(); power()->inverseFlag(); }

	return this;
}

CEString* CEPower::toString(int mode) {
	CEString* res = NULL;

	unsigned long power_props = m_power->props_long();
	bool single_power = ((power_props & CEProperties::PROP_MASK_SINGLE_ITEM) != 0);
	if((single_power) && ((power_props & CEProperties::PROP_MASK_IS_NUMBER)==CEProperties::PROP_MASK_IS_NUMBER)) {
		if( !(((CENumber*)m_power)->is_integer()) ) single_power=false; // fraction/complex numbers are not SINGLE_ITEMs
	}

	switch(mode) {
	case STORE_WR_MATHEMATICA:
		res = new CEString((getFlag()?"":"- "));
		StoreInBrackets(res,m_basis,mode);
		res->AddString("\\^");
		if (m_power->getFlag() && single_power) {
			res->AddString(m_power->toString(mode));
		} else { //No visible brackets required for complex power
			res->AddString("\\("); res->AddString(m_power->toString(mode)); res->AddString("\\)");
		}
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

void CEPower::toStream(ostream* os, int mode) {
}

void CEPower::setBasis(CExpression* basis, bool do_clone) {
	if ((m_basis != basis) && (basis != NULL)) {
		CExpression* t=((do_clone)?basis->clone():basis); delete m_basis; m_basis=t;
	}
}

void CEPower::setPower(CExpression* power, bool do_clone) {
	if ((m_power != power) && (power != NULL)) {
		CExpression* t=((do_clone)?power->clone():power); delete m_power; m_power=t;
	}
}

void CEPower::setPower(long power) {
	delete m_power;
	m_power = new CENumber(power);
}

const char* CEPower::ClassName() const { return CEPower::CLASS_NAME; }

CExpression* CEPower::clone() const
{
	CExpression* res = new CEPower(m_basis, m_power);
	copyAttributesTo(res);
	return res;
}

CExpression* CEPower::expand(bool recursive) {
	dbg->log(10, "expand", this);

	if (recursive) {
		if(getLevel()){ setBasis(m_basis->expand(recursive), false); }
		else { m_basis->inverseLevel(); setBasis(m_basis->expand(recursive),false); m_basis->inverseLevel(); }
		setPower(m_power->expand(recursive), false);
	}
	CExpression* arr_res = Arrange(AL_EXTRACT);
	if (arr_res!=this) {
		//Repeat expand here (example: ((a+b)*c)^x is (a+b)^x*c^x here)
		if (recursive) { // further expand result
			CExpression* res_new = arr_res->expand(recursive);
			if ((res_new != arr_res) && (res_new != NULL)) { delete arr_res; arr_res=res_new; }
		}
		return arr_res;
	}

	//Old: expand (a*b*c)^N -->moved to Arrange(AL_EXTRACT)
	//Old: number^x calculation -->moved to Arrange(AL_EXTRACT)
	//Old: expand {a^b}^c ==> a^{b*c} -->moved to Arrange(AL_EXTRACT)

	// expand (sum)^N
	if ((basis()->getProperty(CEProperties::PROP_MASK_IS_SUM)) && (getLevel())) {
		CExpression* res = to_product(basis(), power());
		if (res != NULL) {
			copyAttributesTo(res, true);

			// further expand result (one of items may be product)
			res = CExpression::reduce(res);
			return res;
		}
	}

	// nothing more to expand
	return this;
}

bool CEPower::contain_vector(void * obj)
{
	CExpression* temp = basis();
	if ((temp == obj) || (temp->contain_vector(obj)))
		return true;
	temp = power();
	if ((temp == obj) || (temp->contain_vector(obj)))
		return true;
	return false;
}

/*
*	does not affects data
*/
CExpression* CEPower::to_product(CExpression *_base, CExpression *_power, bool allowLevelInverse) {
	CExpression* res = NULL;
	if (_power->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
		long n=((CENumber*)_power)->getLong();
		if (n!=LONG_MIN) {
			if (n>=0) {
				if (n>1) {
					// Skip n==0 and n==1 cases (applied at Arrange())
					res = _base->clone();
					for(int i=1; i<n; i++) { res = res->multiply(_base->clone()); }
				}
			} else if (allowLevelInverse) { //n<=-1 here
				// Do not skip n==-1 case
				long nAbs=-n;
				res=_base->clone();
				for(int i=1; i<nAbs; i++) { res = res->multiply(_base->clone()); }
				res->inverseLevel();
				CEProduct* pr_res=new CEProduct();
				pr_res->addItem(res);
				res=pr_res;
				// Note: we may not return object with different level -- it could be not a product-item
			}
		}
	}
	return res;
}

CExpression* CEPower::numeric()
{
	CExpression* res = NULL;
	if (power()->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
		CENumber* _pow = (CENumber*)power();
		long s = _pow->getNum();
		if ((_pow->getDen() == 1) && (_pow->getNumIm() == 0) && (s>1)) {
			CExpression* bas_num = basis()->numeric();
			res = bas_num->clone();
			int i;
			for(i=1; i<s; i++) {
				res = res->multiply(bas_num->clone());
			}
			delete bas_num;
		}
	}
	copyAttributesTo(res, true);
	return res;
}

bool CEPower::equals(CExpression *other)
{
	if (other == NULL) return false;
	if (!CEUtil::equals(other->ClassName(), ClassName())) return false;

	CEPower* o = (CEPower*)other;

	if ((basis()->equals(o->basis())) &&
		(power()->equals(o->power())) &&
		(attributes_equals_to(o))) return true;

	return false;
}

bool CEPower::less_than(CExpression* other) {
	if (other == NULL) return false;
	const char* o_cn = other->ClassName();

	// compare to CESymbol, CEPower, CESProduct, CEVector, CEMetric, CEAntisymmetric, CEAntiSymmProd, CEFuncBondarev
	// a < a^n < a.b < a_nu < a_mu_nu < a_abcd
	if (o_cn == CEPower::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CEPower* o = (CEPower*)other;
		if (basis()->less_than(o->basis())) return true;
		if (!basis()->equals(o->basis())) return false;
		if (power()->less_than(o->power())) return true;
		return false;
	}

	if (o_cn == CESymbol::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		if (basis()->less_than(other)) return true;
		return false;
	}

	if (o_cn == CESProduct::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CESProduct* o = (CESProduct*)other;
		if (basis()->less_than(o->get_a())) return true;
		if (!basis()->equals(o->get_a())) return false;
		return true;
	}

	if ((o_cn==CEVector::CLASS_NAME)||(o_cn==CEMetric::CLASS_NAME)||(o_cn==CEAntisymmetric::CLASS_NAME)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		IETensor* o = IETensor::GetIETensor(other);
		if (basis()->less_than(o->base())) return true;
		if (!basis()->equals(o->base())) return false;
		return true;
	}

	if (o_cn==CEAntiSymmProd::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	if (o_cn==CEFuncBondarev::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	return CExpression::less_than(other);
}

CExpression* CEPower::square(CExpression *source) { // deletes source
	CExpression* res = new CEPower(source, (long)2);
	delete source;
	return res;
}

CExpression* CEPower::sqrt(CExpression* source) { // deletes source
	CExpression* half = CENumber::GetHalf();
	CExpression* res = new CEPower(source, half);
	delete source; delete half;
	return res;
}

CExpression* CEPower::as_product()
{
	CExpression* res = to_product(basis(), power());
	if (res!=NULL) copyAttributesTo(res, true);
	return res;
}

void CEPower::arrange_basis_sign() {
	if (basis()->getFlag()) return;

	//basis < 0
	if (m_power->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
		if(((CENumber*)m_power)->is_integer()) {
			long nPow=((CENumber*)m_power)->getLong();
			if(nPow!=LONG_MIN){ basis()->setFlag(); if(!CLMath::is_even(((nPow>=0)?nPow:(-nPow))))inverseFlag(); }
		}
	}
}

/*
void CEPower::arrange_basis_level() {
	if (!(basis()->getLevel())) { basis()->inverseLevel(); power()->inverseFlag(); }
}
*/

CESerializable* CEPower::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	long skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) {
		//Error
		dbg->log(0, "Error restoring CEPower");
		return NULL;
	}
	CESerializable* item_1 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;
	
	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) {
		//Error
		dbg->log(0, "Error restoring CEPower");
		delete item_1;
		return NULL;
	}
	CESerializable* item_2 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos > 0) {
		//Error
		dbg->log(0, "Error restoring CEPower");
		delete item_1;
		delete item_2;
		return NULL;
	}
	
	CEPower* res = new CEPower((CExpression*)item_1, (CExpression*)item_2);
	delete item_1; delete item_2;
	return res;
}

const char* CEPower::get_TAG_Name()
{
	return CLASS_TAG_NAME;
}

long CEPower::store_stream(ofstream *p_ofs, DWORD settings)
{
	long res = CESupport::store_stream(basis(), p_ofs, settings);
	res += CESupport::store_stream(power(), p_ofs, settings);
	return res;
}

long CEPower::getLongPower()
{
	CExpression* pow_expr = power();
	if (pow_expr->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
		CENumber* num_pow = (CENumber*)pow_expr;
		if ((num_pow->getLevel()) && (num_pow->getDen() == 1) && (num_pow->getNumIm() == 0)) {
			long res =  num_pow->getNum(); if (!num_pow->getFlag()) res=-res;
			return res;
		}
		//if ((num_pow->getDen() != 1) || (n->getNumIm() != 0)) return LONG_MIN;
		//double gap = 0.00001;
		//long lPower = (long)(num_pow->to_double()+gap);
		//return lPower;
	}

	//return error code
	return LONG_MIN;
}

bool CEPower::isPowerInteger() {
	CExpression* pw = power();
	if (pw->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
		CENumber* n = (CENumber*)pw;
		if ((n->getLevel()) && (n->getDen() == 1) && (n->getNumIm() == 0)) return true;
	}
	return false;
}
