////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EAntiSymmProd.cpp: implementation of the CEAntiSymmProd class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const char* CEAntiSymmProd::CLASS_NAME = "CEAntiSymmProd";
const char* CEAntiSymmProd::CLASS_TAG_NAME = "asym_pr";
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEAntiSymmProd::CEAntiSymmProd(const CExpression* v1, const CExpression* v2, const CExpression* v3, const CExpression* v4)
:CEScalar()
{
	m_v[0] = v1->clone();
	m_v[1] = v2->clone();
	m_v[2] = v3->clone();
	m_v[3] = v4->clone();
}

CEAntiSymmProd::~CEAntiSymmProd() {
	for(int i=0; i<4; i++) { if (m_v[i] != NULL) delete m_v[i]; }
}

void CEAntiSymmProd::collectProperties(CEProperties* props, unsigned set) {
	CEScalar::collectProperties(props,set);
	for(int i=0; i<4; i++) {
		if (m_v[i] == NULL) continue;
		if ((m_v[i]->props_long() & CEProperties::PROP_MASK_SELF_CONJUGATED)==0) props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
	}
}

unsigned long CEAntiSymmProd::props_long() {
	unsigned long res=CEScalar::props_long();
	for(int i=0; i<4; i++) {
		if (m_v[i] == NULL) continue;
		if ((m_v[i]->props_long() & CEProperties::PROP_MASK_SELF_CONJUGATED)==0) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	}
	return res;
}

CExpression* CEAntiSymmProd::numeric() {
	dbg->log(0, "Internal Error: CEAntiSymmProd::numeric() not implemented");
	CEException::halt();
	return NULL;
}

long CEAntiSymmProd::store_stream(ofstream* p_ofs, DWORD settings) {
	long res = CESupport::store_stream(m_v[0], p_ofs, settings);
	res += CESupport::store_stream(m_v[1], p_ofs, settings);
	res += CESupport::store_stream(m_v[2], p_ofs, settings);
	res += CESupport::store_stream(m_v[3], p_ofs, settings);
	return res;
}

const char* CEAntiSymmProd::get_TAG_Name() { return CLASS_TAG_NAME; }

CESerializable* CEAntiSymmProd::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	long skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntiSymmProd"); return NULL; }
	CESerializable* item_1 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;
	
	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntiSymmProd"); delete item_1; return NULL; }
	CESerializable* item_2 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntiSymmProd"); delete item_1; delete item_2; return NULL; }
	CESerializable* item_3 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntiSymmProd"); delete item_1; delete item_2; delete item_3; return NULL; }
	CESerializable* item_4 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos > 0) { dbg->log(0, "Error restoring CEAntiSymmProd"); delete item_1; delete item_2; delete item_3; delete item_4;  return NULL; }

	CEAntiSymmProd* res = new CEAntiSymmProd((CExpression*)item_1, (CExpression*)item_2, (CExpression*)item_3, (CExpression*)item_4);
	delete item_1; delete item_2; delete item_3; delete item_4;
	return res;
}

const char* CEAntiSymmProd::ClassName() const { return CLASS_NAME; }

CExpression* CEAntiSymmProd::get_v(int i) {
	if ((i<0) || (i>3)) return NULL;
	return m_v[i];
}

bool CEAntiSymmProd::equals(CExpression* other) {
	if (other == NULL) return false;
	if (!CEUtil::equals(other->ClassName(), CLASS_NAME)) return false;
	CEAntiSymmProd* o = (CEAntiSymmProd*)other;

	if ((get_v(0)->equals(o->get_v(0))) &&
		(get_v(1)->equals(o->get_v(1))) &&
		(get_v(2)->equals(o->get_v(2))) &&
		(get_v(3)->equals(o->get_v(3))) &&
		(attributes_equals_to(o))) return true;

	return false;
}

bool CEAntiSymmProd::less_than(CExpression *other) {
	if (other == NULL) return false;

	// compare to CESymbol, CEPower, CESProduct, CEVector, CEMetric, CEAntisymmetric, CEAntiSymmProd, CEFuncBondarev
	// a < a^n < a.b < a_nu < a_mu_nu < a_abcd
	const char* o_cn = other->ClassName();
	if (o_cn == CEPower::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return false;
	}

	if (o_cn == CESymbol::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return false;
	}

	if (o_cn == CESProduct::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return false;
	}

	if ((o_cn==CEVector::CLASS_NAME)||(o_cn==CEMetric::CLASS_NAME)||(o_cn==CEAntisymmetric::CLASS_NAME)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return false;
	}

	if (o_cn==CEAntiSymmProd::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CEAntiSymmProd* o = (CEAntiSymmProd*)other;
		if (get_v(0)->less_than(o->get_v(0))) return true;
		if (!get_v(0)->equals(o->get_v(0))) return false;
		if (get_v(1)->less_than(o->get_v(1))) return true;
		if (!get_v(1)->equals(o->get_v(1))) return false;
		if (get_v(2)->less_than(o->get_v(2))) return true;
		if (!get_v(2)->equals(o->get_v(2))) return false;
		if (get_v(3)->less_than(o->get_v(3))) return true;
		return false;
	}

	if (o_cn==CEFuncBondarev::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	return CExpression::less_than(other);
}

bool CEAntiSymmProd::contain_vector(void * obj) {
	if (CETensor::contain_vector(obj)) return true;
	if ((m_v[0] == obj) || (m_v[0]->contain_vector(obj))) return true;
	if ((m_v[1] == obj) || (m_v[1]->contain_vector(obj))) return true;
	if ((m_v[2] == obj) || (m_v[2]->contain_vector(obj))) return true;
	if ((m_v[3] == obj) || (m_v[3]->contain_vector(obj))) return true;
	return false;
}

CExpression* CEAntiSymmProd::expand(bool recursive) {
	CExpression* arr_res = Arrange(AL_EXTRACT);
	if (arr_res != this) {
		CExpression* arr_res2 = arr_res->expand();
		if (arr_res2 != arr_res) { delete arr_res; arr_res=arr_res2; }
		return arr_res;
	}
	
	int i_sum = -1;
	int i;
	int sums_no = 0;
	for(i=0; i<4; i++) {
		if (m_v[i]->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			sums_no++;
			if (i_sum<0) i_sum = i;
		}
	}

	if (i_sum >= 0) {
		CESum* res = new CESum();
		CESum* e_sum = (CESum*)m_v[i_sum];
		long e_size = e_sum->size();
		
		long ix;
		for (ix=0; ix<e_size; ix++) {
			CExpression* e_item = e_sum->GetAt(ix);
			
			switch(i_sum) {
			case 0: res->addItem(new CEAntiSymmProd(e_item,m_v[1],m_v[2],m_v[3])); break;
			case 1: res->addItem(new CEAntiSymmProd(m_v[0],e_item,m_v[2],m_v[3])); break;
			case 2: res->addItem(new CEAntiSymmProd(m_v[0],m_v[1],e_item,m_v[3])); break;
			case 3: res->addItem(new CEAntiSymmProd(m_v[0],m_v[1],m_v[2],e_item)); break;
			default:
				CEException::halt("CEAntiSymmProd::expand(): Internal Error");
			}
		}
		
		if (!e_sum->getFlag()) {
			res->inverseFlag();
			dbg->log(0, "Warning: not arranged CEAntiSymmProd expanding", this);
		}
		
		copyAttributesTo(res, true);
		
		if (sums_no>1) {
			CExpression* res2 = res->expand();
			if (res2 != res) delete res;
			return res2;
		}
		
		return res;
	}

	return this;
}

CExpression* CEAntiSymmProd::Arrange(int level) {
	m_v[0] = CExpression::arrge(m_v[0]);
	m_v[1] = CExpression::arrge(m_v[1]);
	m_v[2] = CExpression::arrge(m_v[2]);
	m_v[3] = CExpression::arrge(m_v[3]);

	if (!m_v[0]->getFlag()) { inverseFlag(); m_v[0]->inverseFlag(); }
	if (!m_v[1]->getFlag()) { inverseFlag(); m_v[1]->inverseFlag(); }
	if (!m_v[2]->getFlag()) { inverseFlag(); m_v[2]->inverseFlag(); }
	if (!m_v[3]->getFlag()) { inverseFlag(); m_v[3]->inverseFlag(); }

	// sort parameters
	int i;
	for(i=0; i<3;) {
		if (m_v[i+1]->less_than(m_v[i])) {
			CExpression* temp = m_v[i];
			m_v[i] = m_v[i+1];
			m_v[i+1] = temp;
			inverseFlag();
			if(i>0) { i--; } // loop back one step
			else { i++; } // loop next
		} else {
			i++; // loop next
		}
	}

	if (level == AL_EXTRACT) {
		if (CEBasic::isZero(m_v[0]) ||
			CEBasic::isZero(m_v[1]) ||
			CEBasic::isZero(m_v[2]) ||
			CEBasic::isZero(m_v[3]) ) {
			CExpression* res = CENumber::GetZero();
			copyAttributesTo(res, true);
			return res;
		}

		int i1;
		int i2;
		for(i1=0; i1<4; i1++) {
			CExpression* ei = m_v[i1];
			for(i2=(i1+1); i2<4; i2++) {
				if (ei->equals(m_v[i2])) {
					CExpression* res = CENumber::GetZero();
					copyAttributesTo(res, true);
					return res;
				}
			}
		}
	}
	return this;
}

CExpression* CEAntiSymmProd::applyRule(CEAbstractRule* rule) {
	CERule* r = (CERule*)rule;
	m_v[0] = r->ApplyTo(m_v[0]);
	m_v[1] = r->ApplyTo(m_v[1]);
	m_v[2] = r->ApplyTo(m_v[2]);
	m_v[3] = r->ApplyTo(m_v[3]);
	return this;
}

CExpression* CEAntiSymmProd::clone() const {
	CExpression* res = new CEAntiSymmProd(m_v[0], m_v[1], m_v[2], m_v[3]);
	copyAttributesTo(res);
	return res;
}

CEString* CEAntiSymmProd::toString(int mode) {
	CEString* res = NULL;
	switch (mode)
	{
	case STORE_CPP:
	case STORE_FORTRAN_SYMBOL:
		res = new CEString((getFlag()?"":"-"));
		res->AddString("E");
		res->AddString(m_v[0]->toString(mode)); //res->AddString("_");
		res->AddString(m_v[1]->toString(mode)); //res->AddString("_");
		res->AddString(m_v[2]->toString(mode)); //res->AddString("_");
		res->AddString(m_v[3]->toString(mode));
		break;
	case STORE_WR_MATHEMATICA:
		res = new CEString((getFlag()?"E\\+\\(":"- E\\+\\("));
		res->AddString(m_v[0]->toString(mode)); res->AddString(",");
		res->AddString(m_v[1]->toString(mode)); res->AddString(",");
		res->AddString(m_v[2]->toString(mode)); res->AddString(",");
		res->AddString(m_v[3]->toString(mode));
		res->AddString("\\)");
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

bool CEAntiSymmProd::InvolveAsymmProd(CExpression* src) {
	if (src == NULL) return false;

	bool res = false;
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		res = InvolveAsymmProdImpl((CESum*)src);
	} else if (src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		res = InvolveAsymmProdImpl((CEProduct*)src);
	}
	return res;
}

bool CEAntiSymmProd::InvolveAsymmProdImpl(CESum* src) {
	//dbg->log(2, "CEAntiSymmProd::InvolveAsymmProdImpl (SUM)", src);
	long s = src->size();
	long i;
	bool res = false;
	for(i=0; i<s; i++) {
		CExpression* item = src->GetAt(i);
		bool chgd = false;
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			chgd = InvolveAsymmProdImpl((CESum*)item);
		} else if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			chgd = InvolveAsymmProdImpl((CEProduct*)item);
		}
		if (chgd) res=true;
	}
	return res;
}

bool CEAntiSymmProd::InvolveAsymmProdImpl(CEProduct* src) {
	//dbg->log(2, "CEAntiSymmProd::InvolveAsymmProdImpl", src);
	long s = src->size();
	long i;

	// 1. search e-object
	long e_pos = -1;
	long e_num = 0;
	bool chgd=false;
	for(i=0; i<s; i++) {
		CExpression* item = src->GetAt(i);

		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			if (InvolveAsymmProdImpl((CESum*)item)) chgd=true;
		} else if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			if (InvolveAsymmProdImpl((CEProduct*)item)) chgd=true;
		} else if (CEUtil::equals(item->ClassName(), CEAntisymmetric::CLASS_NAME)) {
			if (item->getLevel()) {
				e_num++;
				if (e_pos<0) e_pos=i;
			}
		}
	}

	if (e_pos<0) return chgd;

	//dbg->log(2, e_pos, "CEAntiSymmProd::InvolveAsymmProdImpl (e_pos = %ld)", src);
	// search for e-coupled vectors
	int j;
	long vec_pos[4];
	vec_pos[0]=-1; vec_pos[1]=-1; vec_pos[2]=-1; vec_pos[3]=-1;
	CEAntisymmetric* asym = (CEAntisymmetric*)src->GetAt(e_pos);
	for(j=0; j<4; j++) {
		CTIndex* tindex = asym->GetTIndex(j+1);
		if (tindex == NULL) CEException::halt("CEAntiSymmProd::InvolveAsymmProdImpl(): Internal Error");
		for(i=0; i<s; i++) {
			if (i == e_pos) continue;
			CExpression* item = src->GetAt(i);
			if (CEUtil::equals(item->ClassName(), CEVector::CLASS_NAME)) {
				CEVector* vec = (CEVector*)item;
				if ((vec->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) &&
					(CTIndex::CheckCoupled(tindex,vec->GetTIndex(1)))) {
					if (vec->getLevel()) {
						vec_pos[j]=i;
						break;
					}
				}
			}
		}
		//dbg->log(2, vec_pos[j], "CEAntiSymmProd::InvolveAsymmProdImpl (vec_pos[j] = %ld)", index);
		if (vec_pos[j] < 0) return false;
	}

	// all vectors found
	if(!asym->getFlag()) { asym->inverseFlag(); src->inverseFlag(); }

	CEVector* v0 = (CEVector*)src->GetAt(vec_pos[0]);
	//dbg->log(2, "CEAntiSymmProd::InvolveAsymmProdImpl (v0)", v0);
	CEVector* v1 = (CEVector*)src->GetAt(vec_pos[1]);
	//dbg->log(2, "CEAntiSymmProd::InvolveAsymmProdImpl (v1)", v1);
	CEVector* v2 = (CEVector*)src->GetAt(vec_pos[2]);
	//dbg->log(2, "CEAntiSymmProd::InvolveAsymmProdImpl (v2)", v2);
	CEVector* v3 = (CEVector*)src->GetAt(vec_pos[3]);
	//dbg->log(2, "CEAntiSymmProd::InvolveAsymmProdImpl (v3)", v3);
	
	if(!v0->getFlag()) { v0->inverseFlag(); src->inverseFlag(); }
	if(!v1->getFlag()) { v1->inverseFlag(); src->inverseFlag(); }
	if(!v2->getFlag()) { v2->inverseFlag(); src->inverseFlag(); }
	if(!v3->getFlag()) { v3->inverseFlag(); src->inverseFlag(); }

	CEAntiSymmProd* a_pr = new CEAntiSymmProd(v0->base(), v1->base(), v2->base(), v3->base());
	
	src->SetAt(e_pos, a_pr); delete a_pr;
	
	CExpression* unit = CENumber::GetUnit();
	src->SetAt(vec_pos[0], unit);
	src->SetAt(vec_pos[1], unit);
	src->SetAt(vec_pos[2], unit);
	src->SetAt(vec_pos[3], unit);
	delete unit;

	if (e_num>1) { InvolveAsymmProdImpl(src); } //loop again
	//dbg->log(2, "CEAntiSymmProd::InvolveAsymmProdImpl result", src);
	return true;
}
