////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EAntisymmetric.cpp: implementation of the CEAntisymmetric class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

const char* CEAntisymmetric::CLASS_NAME	= "CEAntisymmetric";
CExpression* CEAntisymmetric::m_AntisymmetricBase = NULL;
const CExpression* CEAntisymmetric::GetAntisymmetricBase() {
	if (m_AntisymmetricBase==NULL) {
		m_AntisymmetricBase = CEProcessor::getDataStorage()->gete(CConst::DS_ANTISYMMETRIC_BASE);
		// Note: m_AntisymmetricBase will be deleted with DataStorage
		if (m_AntisymmetricBase==NULL) { m_AntisymmetricBase = new CESymbol("E"); CEProcessor::AddToDeleteList(m_AntisymmetricBase); }
	}
	return m_AntisymmetricBase;
}

CEAntisymmetric::CEAntisymmetric(const CExpression* index1, bool covariance1,
		const CExpression* index2, bool covariance2,
		const CExpression* index3, bool covariance3,
		const CExpression* index4, bool covariance4)
		:CETensor(GetAntisymmetricBase())
{
	//Note: currently 'nDIM' MUST be equal for all indices! --> elsewhere change 'E[.i..j.] p^i p^j -> 0' and other checks
	m_indices->add(CTIndex::CreateVectorIndex(index1, covariance1));
	m_indices->add(CTIndex::CreateVectorIndex(index2, covariance2));
	m_indices->add(CTIndex::CreateVectorIndex(index3, covariance3));
	m_indices->add(CTIndex::CreateVectorIndex(index4, covariance4));
}

CEAntisymmetric::~CEAntisymmetric()
{

}

CExpression* CEAntisymmetric::unite_with(IETensor *other, bool other_is_at_right, bool bCalledByOther)
{
	CExpression* o_e = other->ThisExpr();
	const char* _ocn = o_e->ClassName();
	if (_ocn == CLASS_NAME) {
		if (rank() != other->rank()) return NULL;
		return unite((CEAntisymmetric*)o_e);
	}
	else if ((_ocn == CEMetric::CLASS_NAME) && (!bCalledByOther))
		return other->unite_with(this, !other_is_at_right, true);
	return NULL;
}

const char* CEAntisymmetric::ClassName() const { return CLASS_NAME; }

/*
* There is exist an exact formula for non-summarizing product antisymmetric tensors,
* But we do not use it.
*/
CExpression* CEAntisymmetric::unite(CEAntisymmetric *other)
{
	int r = rank();
	if (r != 4) return NULL;
	if ((!getLevel()) && (!other->getLevel())) return NULL;
	
	//searching for similar indices,
	//moving them to first places
	//& Using simplified formula
	int i;
	for(i=1; i<=r; i++) {
		CExpression* ind = index(i);
		int j;
		for(j=1; j<=r; j++) {
			CExpression* t = other->index(j);
			if (ind->equals(t)) {
				move_index(i, 1);
				other->move_index(j, 1);
				CExpression* unite_res = unite_formula(other);
				if (unite_res != NULL) return unite_res;
			}
		}
	}
	//int i;

	//elsewhere use common formula
	//result is MINUS det(g_xx)
	CSequence* arr = new CSequence();
	for(i=1; i<=r; i++) {
		CExpression* o_index = other->index(i);
		bool o_cov = other->covariance(i);
		CSequence* col = new CSequence();
		int j;
		for(j=1; j<=r; j++) {
			col->addItem(new CEMetric(index(j), o_index, covariance(j), o_cov));
		}
		arr->addItem(col);
	}
	CExpression* res = CLMath::det(arr);
	delete arr;
	res->inverseFlag();

	if (!getFlag()) res->inverseFlag();
	if (!other->getFlag()) res->inverseFlag();
	return res;
}

/*
* 1<=any_pos<=rank();
*/
void CEAntisymmetric::move_index(int from_pos, int dest_pos)
{
	int r = rank();
	if ((from_pos > r) || (dest_pos > r) || (from_pos < 1) || (dest_pos < 1)) return;
	if (from_pos > dest_pos) {
		int i;
		for(i=from_pos; i>dest_pos; i--) {
			CExpression* ind1 = index(i)->clone();
			bool cov1 = covariance(i);
			CExpression* ind0 = index(i-1)->clone();
			bool cov0 = covariance(i-1);
			changeIndex(i, ind0, cov0);
			changeIndex(i-1, ind1, cov1);
			delete ind1;
			delete ind0;
			inverseFlag();
		}
	}
	else {
		int i;
		for(i=from_pos; i<dest_pos; i++) {
			CExpression* ind1 = index(i)->clone();
			bool cov1 = covariance(i);
			CExpression* ind0 = index(i+1)->clone();
			bool cov0 = covariance(i+1);
			changeIndex(i, ind0, cov0);
			changeIndex(i+1, ind1, cov1);
			delete ind1;
			delete ind0;
			inverseFlag();
		}
	}
}

/*
* The rank of both is supposed to be = 4.
*/
CExpression* CEAntisymmetric::unite_formula(CEAntisymmetric *other)
{
	if (!index(1)->equals(other->index(1))) {
		printf("CEAntisymmetric::unite_fotmula(e(a...)*e(a...)) : invalid call.\n");
		return NULL;
	}
	if (covariance(1) == other->covariance(1))
		dbg->log(0,"<Warning><covariance == other->covariance in CEAntisymmetric::unite_formula");
	
	CESum* res = new CESum();
	bool sign = false;
	int i;
	for(i=2; i<=4; i++) {
		int j;
		for(j=2; j<=4; j++) if (j != i) {
			int k;
			for(k=2; k<=4; k++) if ((k != i) && (k != j)) {
				CEProduct* pr = new CEProduct();
				pr->addItem(new CEMetric(index(2), other->index(i), covariance(2), other->covariance(i)));
				pr->addItem(new CEMetric(index(3), other->index(j), covariance(3), other->covariance(j)));
				pr->addItem(new CEMetric(index(4), other->index(k), covariance(4), other->covariance(k)));
				pr->putFlag(sign);
				res->addItem(pr);
				//delete pr;
			}
			sign = !sign;
		}
		sign = !sign;
	}
	if (!getFlag()) res->inverseFlag();
	if (!other->getFlag()) res->inverseFlag();
	return res;
}

CExpression* CEAntisymmetric::clone() const {
	CExpression* res = new CEAntisymmetric(
		GetIndex(1), covariance(1),
		GetIndex(2), covariance(2),
		GetIndex(3), covariance(3),
		GetIndex(4), covariance(4));
	copyAttributesTo(res);
	return res;
}

CExpression* CEAntisymmetric::Arrange(int level)
{
	// sorting indices?

	if (level == AL_EXTRACT) {
		// searching for similar indices.
		int s = rank();
		int i;
		for(i=1; i<s; i++) {
			int j;
			for(j=i+1; j<=s; j++) {
				if (index(i)->equals(index(j))) {
					if (covariance(i) == covariance(j))
						printf("<Warning><(CEAntisymmetric): index(i) == index(j) & covariance(i) == covariance(j)>\n");
					return CENumber::GetZero();
				}
			}
		}
	}
	return this;
}

bool CEAntisymmetric::less_than(CExpression *other) {
	if (other == NULL) return false;

	// compare to CESymbol, CEPower, CESProduct, CEVector, CEMetric, CEAntisymmetric, CEAntiSymmProd, CEFuncBondarev
	// a < a^n < a.b < a_nu < a_mu_nu < a_abcd
	const char* o_cn = other->ClassName();
	if (o_cn == CEPower::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		if (base()->less_than(((CEPower*)other)->basis())) return true;
		return false;
	}

	if (o_cn == CESymbol::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		if (base()->less_than(other)) return true;
		return false;
	}

	if (o_cn == CESProduct::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		if (base()->less_than(((CESProduct*)other)->get_a())) return true;
		return false;
	}

	if ((o_cn==CEVector::CLASS_NAME)||(o_cn==CEMetric::CLASS_NAME)||(o_cn==CEAntisymmetric::CLASS_NAME)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return TLessThen(IETensor::GetIETensor(other), o_cn);
	}

	if (o_cn==CEAntiSymmProd::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	if (o_cn==CEFuncBondarev::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	return CExpression::less_than(other);
}

/*
*	Searches for ..ppe..->0 sequences.
*/
bool CEAntisymmetric::parse_product_for_zero(CEProduct *seq)
{
	long s = seq->size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* item = seq->GetAt(i);
		if (CEUtil::equals(CEAntisymmetric::CLASS_NAME, item->ClassName())) {
			if (parse_product_for_zero(seq, (CEAntisymmetric*)item)) return true;
		}
	}
	return false;
}

bool CEAntisymmetric::parse_product_for_zero(CEProduct *seq, CEAntisymmetric* asym)
{
	long s = seq->size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* item = seq->GetAt(i);
		if (item->getProperty(CEProperties::PROP_MASK_IS_TENSOR)) {
			if (CEUtil::equals(CEVector::CLASS_NAME, item->ClassName())) {
				CEVector* vec = (CEVector*)item;
				if (asym->contain_index(vec->index())) {
					long j;
					for (j=i+1; j<s; j++) {
						item = seq->GetAt2(j);
						if (item->getProperty(CEProperties::PROP_MASK_IS_TENSOR)) {
							if (CEUtil::equals(CEVector::CLASS_NAME, item->ClassName())) {
								CEVector* vec_2 = (CEVector*)item;
								if (vec->base()->equals(vec_2->base())) {
									if (asym->contain_index(vec_2->index())) {
										return true;
									}
								}
							}
						}
					}
				}
			}
		}
	}
	return false;
}

const char* CEAntisymmetric::CLASS_TAG_NAME = "asymm";
const char* CEAntisymmetric::get_TAG_Name() { return CLASS_TAG_NAME; }

long CEAntisymmetric::store_stream(ofstream* p_ofs, DWORD settings) {
	if (rank() != 4) { dbg->log(0, "ERROR: saving non-4 rank asymetic tesor is not implemented"); return 0; }
	
	long res = CESupport::store_stream(index(1), p_ofs, settings);
	CENumber cov_no1(covariance(1)?1:0);
	res += CESupport::store_stream(&cov_no1, p_ofs, settings);

	res += CESupport::store_stream(index(2), p_ofs, settings);
	CENumber cov_no2(covariance(2)?1:0);
	res += CESupport::store_stream(&cov_no2, p_ofs, settings);

	res += CESupport::store_stream(index(3), p_ofs, settings);
	CENumber cov_no3(covariance(3)?1:0);
	res += CESupport::store_stream(&cov_no3, p_ofs, settings);

	res += CESupport::store_stream(index(4), p_ofs, settings);
	CENumber cov_no4(covariance(4)?1:0);
	res += CESupport::store_stream(&cov_no4, p_ofs, settings);
	return res;
}

CESerializable* CEAntisymmetric::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	CEPtrArray items(true); // for auto-delete items
	long skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* item_1 = CESupport::read_buffer(data, pos, settings);
	items.add(item_1);
	if (!CEBasic::isExpression(item_1)) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* cov_1 = CESupport::read_buffer(data, pos, settings);
	items.add(cov_1);
	if ((cov_1 == NULL) || (!CEUtil::equals(cov_1->ClassName(), CENumber::CLASS_NAME))) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* item_2 = CESupport::read_buffer(data, pos, settings);
	items.add(item_2);
	if (!CEBasic::isExpression(item_2)) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* cov_2 = CESupport::read_buffer(data, pos, settings);
	items.add(cov_2);
	if ((cov_2 == NULL) || (!CEUtil::equals(cov_2->ClassName(), CENumber::CLASS_NAME))) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* item_3 = CESupport::read_buffer(data, pos, settings);
	items.add(item_3);
	if (!CEBasic::isExpression(item_3)) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* cov_3 = CESupport::read_buffer(data, pos, settings);
	items.add(cov_3);
	if ((cov_3 == NULL) || (!CEUtil::equals(cov_3->ClassName(), CENumber::CLASS_NAME))) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* item_4 = CESupport::read_buffer(data, pos, settings);
	items.add(item_4);
	if (!CEBasic::isExpression(item_4)) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	CESerializable* cov_4 = CESupport::read_buffer(data, pos, settings);
	items.add(cov_4);
	if ((cov_4 == NULL) || (!CEUtil::equals(cov_4->ClassName(), CENumber::CLASS_NAME))) { dbg->log(0, "Error restoring CEAntisymmetric"); return NULL; }
	pos = skip_pos;

	CEAntisymmetric* res = new CEAntisymmetric((CExpression*)item_1, (((CENumber*)cov_1)->getLong() != 0),
		(CExpression*)item_2, (((CENumber*)cov_2)->getLong() != 0),
		(CExpression*)item_3, (((CENumber*)cov_3)->getLong() != 0),
		(CExpression*)item_4, (((CENumber*)cov_4)->getLong() != 0)
		);
	return res;
}
