////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EGammaSpecial.cpp: implementation of the CEGammaSpecial class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CExpression* CEGammaSpecial::m_symGammaSpecialBase = NULL;
const CExpression* CEGammaSpecial::GetGammaSpecialBase() { // never returns NULL
	if (m_symGammaSpecialBase==NULL) {
		m_symGammaSpecialBase=CEProcessor::getDataStorage()->gete(CConst::DS_GAMMA_SPECIAL_BASE);
		if (m_symGammaSpecialBase==NULL) { m_symGammaSpecialBase=new CESymbol("GammaX"); CEProcessor::AddToDeleteList(m_symGammaSpecialBase); }
	}
	return m_symGammaSpecialBase;
}
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

const char* CEGammaSpecial::CLASS_NAME	= "CEGammaSpecial";

CEGammaSpecial::CEGammaSpecial(int _value) :CEScalar(CEGammaSpecial::GetGammaSpecialBase())
{
	m_value = _value;
	if (m_value > 0) m_value = 1;
	else if (m_value < 0) m_value = -1;
}

CEGammaSpecial::~CEGammaSpecial() {}

void CEGammaSpecial::collectProperties(CEProperties* props, unsigned set)
{
	CETensor::collectProperties(props, set);
	props->setCommutative(false);
	props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

unsigned long CEGammaSpecial::props_long()
{
	unsigned long res = CETensor::props_long();
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_COMMUTATIVE, false);
	res = CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}

const char* CEGammaSpecial::ClassName() const { return CLASS_NAME; }
int CEGammaSpecial::getGammaValue() { return m_value; }
void CEGammaSpecial::setGammaValue(int _value) { m_value = _value; }
void CEGammaSpecial::inverseGammaValue() { m_value = -m_value; }

CExpression* CEGammaSpecial::conjugated(int mode) {
	CExpression* res = clone();
	if (m_value == 0) res->inverseFlag(); // -gamma_{5}
	else ((CEGammaSpecial*)res)->inverseGammaValue(); //gamma_{-X}
	return res;
}

CEString* CEGammaSpecial::toString(int mode) {
	CEString* res = NULL;
	switch (mode)
	{
	case STORE_WR_MATHEMATICA:
		if (getFlag()) res = new CEString("\\(");
		else res = new CEString("- \\(");
		// m_value: 0 for gamma_{5}, 1 for gamma_{+} and -1 for gamma_{-}
		if (m_value == 0) {
			res->AddString("\\[Gamma]\\+5");
		} else if (m_value > 0) {
			res->AddString("\\[Gamma]\\++");
		} else {
			res->AddString("\\[Gamma]\\+-");
		}
		res->AddString("\\)");
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

CExpression* CEGammaSpecial::unite_with(IETensor *other, bool other_is_at_right, bool bCalledByOther)
{
	CExpression* res = NULL;
	CExpression* o_e = other->ThisExpr();
	if (CEUtil::equals(o_e->ClassName(), ClassName())) {
		int other_value = ((CEGammaSpecial*)o_e)->getGammaValue();

		if ((m_value == 0) && (other_value == 0)) { // gamma_5*gamma_5 ==> 1
			res = CENumber::GetUnit();
		}
		else if (m_value == 0) { // gamma_5*gamma_{+/-} ==> -gamma_{+/-}
			res = new CEGammaSpecial(other_value); // do not clone other: extra sign will be applied further
			if (other_value < 0) res->inverseFlag();
		}
		else if (other_value == 0) { // gamma_{+/-}*gamma_{5} ==> -gamma_{+/-}
			res = new CEGammaSpecial(m_value); // do not clone other: extra sign will be applied further
			if (m_value < 0) res->inverseFlag();
		}
		else if (m_value == other_value) { // gamma_{+}*gamma_{+} or gamma_{-}*gamma_{-}  ==> gamma_{+/-}
			res = new CEGammaSpecial(m_value); // do not clone other: extra sign will be applied further
		}
		else { // gamma_{+}*gamma_{-} or gamma_{-}*gamma_{+} ==> 0
			res = CENumber::GetZero();
		}

		if (!getFlag()) res->inverseFlag();
		if (!o_e->getFlag()) res->inverseFlag();
	}
	//no calls to other->unite_with() -- no other couplings possible
	return res;
}

CExpression* CEGammaSpecial::clone() const
{
	CExpression* res = new CEGammaSpecial(m_value);
	copyAttributesTo(res, true);
	return res;
}

bool CEGammaSpecial::equals(CExpression *other) {
	if (other == NULL) return false;
	if (!CEUtil::equals(other->ClassName(), CLASS_NAME)) return false;
	if (m_value != ((CEGammaSpecial*)other)->getGammaValue()) return false;
	if (props_long() != other->props_long()) {
		// how can it occur?
		dbg->log(0, "Test point: different CEGammaSpecial properties", this); dbg->log(2, "Test point: other:", other);
		return false;
	}
	if (!attributes_equals_to(other)) return false;
	return true;
}

bool CEGammaSpecial::less_than(CExpression* other) {
	if (other == NULL) return false;
	if (other->ClassName() == CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CEGammaSpecial* gsOther = (CEGammaSpecial*)other;
		if (m_value != gsOther->getGammaValue()) return (m_value < gsOther->getGammaValue());
		return false;
	}
	return CExpression::less_than(other);
}

CExpression* CEGammaSpecial::exchange_subst(IETensor* other, bool other_is_at_right, bool bCalledByOther)
{
	//Note: CETensor::exchange_subst() must be called LAST!

	/*
	*	Old: try unition
	*	if(!bCalledByOther){ CExpression* res=unite_with(other,other_is_at_right); if(res!=NULL)return res; }
	*/

	CExpression* eOther = other->ThisExpr();

	//1. Commutation to Spinor
	if (eOther->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) {
		return NULL;
	}

	//2. Commutation to CEGamma & CESGammaProduct
	const char* cnOther=eOther->ClassName();
	if ((cnOther==CEGamma::CLASS_NAME) || //==CEGamma::CLASS_NAME
		(cnOther==CESGammaProduct::CLASS_NAME))
	{
		CExpression* res = NULL;
		if (m_value == 0) { // Gamma_5 is anticommutative
			if(other_is_at_right){ res=eOther->clone()->multiply(clone()); }else{ res=clone()->multiply(eOther->clone()); }
			res->inverseFlag();
		} else { // Gamma_+/- changes it's value
			CEGammaSpecial* cpy=(CEGammaSpecial*)clone(); cpy->inverseGammaValue();
			if(other_is_at_right){ res=eOther->clone()->multiply(cpy); }else{ res = cpy->multiply(eOther->clone()); }
		}
		return res;
	}

	return CETensor::exchange_subst(other,other_is_at_right,bCalledByOther);
}

/*
*	Library functions
*/
CExpression* CEGammaSpecial::expandGammaPM(CExpression* src) {
	CERule* r = new CESocketRule((void*)CEGammaSpecial::expandGammaPM_ruleImpl);
	src = r->ApplyTo(src);
	delete r;
	return src;
}

CExpression* CEGammaSpecial::expandGammaPM_ruleImpl(CExpression* src, CEPtrArray* data_arr) {
	if (!(CEUtil::equals(src->ClassName(), CEGammaSpecial::CLASS_NAME))) return src;

	CEGammaSpecial* gX = (CEGammaSpecial*)src;
	int gX_value = gX->getGammaValue();
	if (gX_value == 0) return src;

	// gamma_{+/-} = 1/2(1 +/- gamma_{5})
	CExpression* g5 = new CEGammaSpecial(0);
	if (gX_value < 0) g5->inverseFlag();
	CExpression* res = (CENumber::GetUnit())->add(g5);
	res = (CENumber::GetHalf())->multiply(res);
	src->copyAttributesTo(res, true);
	return res;
}
