////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EGellMann.cpp: implementation of the CEGellMann and related classes.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CExpression* CEGellMann::m_GellMannBase = NULL;
const CExpression* CEGellMann::GetGellMannBase() {
	if(m_GellMannBase==NULL){
		//[old] m_GellMannBase = CEProcessor::getDataStorage()->gete(CConst::GL_LAMBDA); // Note: 'm_GellMannBase' will be deleted with DataStorage
		//
		if (m_GellMannBase==NULL) { m_GellMannBase=new CESymbol("t"); CEProcessor::AddToDeleteList(m_GellMannBase); }
	}
	return m_GellMannBase;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
const char* CEGellMann::CLASS_NAME = "CEGellMann";
CEGellMann::CEGellMann(const CExpression* _index)
: CETensor(GetGellMannBase())
{
	//
	//	Note: here we ignore color indices!
	//
	if(_index==NULL)CEException::halt("CEGellMann::CEGellMann(NULL index)");
	m_indices->add(CTIndex::CreateMatrixIndex(_index,8));
}
CEGellMann::~CEGellMann() {}
const char* CEGellMann::ClassName() const { return CEGellMann::CLASS_NAME; }

CExpression* CEGellMann::clone() const {
	CExpression* res=new CEGellMann(GetIndex(1));
	copyAttributesTo(res,false);
	return res;
}

unsigned long CEGellMann::props_long() {
	unsigned long res = CETensor::props_long();
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_COMMUTATIVE, false);
	return res;
}

void CEGellMann::collectProperties(CEProperties* props, unsigned set) {
	CETensor::collectProperties(props, set);
	props->setCommutative(false);
}

CExpression* CEGellMann::expand(bool recursive) {
	//return Arrange(AL_EXTRACT);
	return this;
}

CExpression* CEGellMann::Arrange(int level) {
	//if (level == AL_EXTRACT) { CExpression* res = ...; copyAttributesTo(res, true); return res; }
	return this;
}

CExpression* CEGellMann::unite_with(IETensor* other, bool other_is_at_right, bool bCalledByOther) {
	if(other==NULL)return NULL;

	CExpression* o_e = other->ThisExpr();
	if(o_e->ClassName() == CEGellMann::CLASS_NAME) {
		if (getLevel() == o_e->getLevel()) {
			//t^a t^b --> 1/6 \delta^ab + 1/2 h^abc t^c
			//t^a t^a --> 4/3
			//
			if (CTIndex::CheckCoupled(GetTIndex(1),other->GetTIndex(1))) {		//t^a t^a
				CExpression* res = CENumber::GetFraction(4,3);
				copyAttributesTo(res,false); if(!o_e->getFlag())res->inverseFlag();
				return res;
			}
		}
		return NULL;
	}

	if(!bCalledByOther){ return other->unite_with(this, !other_is_at_right, true); }
	return NULL;
}

CExpression* CEGellMann::exchange_subst(IETensor* other, bool other_is_at_right, bool bCalledByOther) {
	if(other==NULL)return NULL;

	//Note: CETensor::exchange_subst() must be called LAST!

	//1. Commutate to Gell-Mann matrix
	CExpression* eOther = other->ThisExpr();
	const char* cnOther = eOther->ClassName();
	if(cnOther == CEGellMann::CLASS_NAME) { //==CESGammaProduct::CLASS_NAME
		if((!getLevel()) || (!eOther->getLevel()))return NULL;

		//t^a t^b --> t^b t^a + i f^abc t^c
		//don't commutate
		return NULL;
	}

	//2. Commutate to any spinor-space object except of the spinor itself
	//Note: don't commutate to spinor until separated color vectors are involved
	//bool bOComm = other->getProperty(CEProperties::PROP_MASK_COMMUTATIVE);	//always 'false' due to previous CETensor:: call
	//if((!bOComm) &&
	if((cnOther==CEGamma::CLASS_NAME) || (cnOther==CEGammaSpecial::CLASS_NAME) || (cnOther==CESGammaProduct::CLASS_NAME)) {
		CExpression* res = (other_is_at_right ? (eOther->clone()->multiply(clone())) : (clone()->multiply(eOther->clone())) );
		return res;
	}

	return CETensor::exchange_subst(other, other_is_at_right, bCalledByOther);
}

//
///////////////////////////////////////////////////// CEGellMannH: h^{abc} = d^{abc} + i f^{abc}
//

CExpression* CEGellMannH::m_GellMannHBase = NULL;
const CExpression* CEGellMannH::GetGellMannHBase() {
	if(m_GellMannHBase==NULL){ m_GellMannHBase=new CESymbol("h"); CEProcessor::AddToDeleteList(m_GellMannHBase); }
	return m_GellMannHBase;
}
CEGellMannH::CEGellMannH(const CExpression* index1, const CExpression* index2, const CExpression* index3)
: CETensor(GetGellMannHBase())
{
	if((index1==NULL) || (index2==NULL) || (index3==NULL))CEException::halt("$CEGellMannH(NULL index)");
	m_indices->add(CTIndex::CreateMatrixIndex(index1,8));
	m_indices->add(CTIndex::CreateMatrixIndex(index2,8));
	m_indices->add(CTIndex::CreateMatrixIndex(index3,8));
}
CEGellMannH::~CEGellMannH() {
}
const char* CEGellMannH::CLASS_NAME = "CEGellMannH";
const char* CEGellMannH::ClassName() const { return CEGellMannH::CLASS_NAME; }

CExpression* CEGellMannH::clone() const {
	CExpression* res=new CEGellMannH(GetIndex(1),GetIndex(2),GetIndex(3)); copyAttributesTo(res,false); return res;
}

unsigned long CEGellMannH::props_long() {
	unsigned long res = CETensor::props_long();
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_SELF_CONJUGATED, false);
	return res;
}

void CEGellMannH::collectProperties(CEProperties* props, unsigned set) {
	CETensor::collectProperties(props, set);
	props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

CExpression* CEGellMannH::Arrange(int level) {
	CExpression* i1=index(1); CExpression* i2=index(2); CExpression* i3=index(3);
	if((i1!=NULL) && (i2!=NULL) && (i3!=NULL)){
		//conversion to numbers here
		if (level == AL_EXTRACT) {
			if((i1->equals(i3)) || (i1->equals(i2)) || (i2->equals(i3))) {
				CExpression* res = CENumber::GetZero(); copyAttributesTo(res, true); return res;
			}
		}

		//Arrange index sequence
		if(i3->less_than(i1)) {
			if(i2->less_than(i3)) { RotateIndices(2); }else{ RotateIndices(1); }
		}else {
			if(i2->less_than(i1)) { RotateIndices(2); }
		}
		//Note: iX-objects may already be deleted
	}
	return this;
}

CExpression* CEGellMannH::unite_with(IETensor* other, bool other_is_at_right, bool bCalledByOther) {
	if(other==NULL)return NULL;
	CExpression* o_e = other->ThisExpr();
	if((o_e->ClassName() == CEGellMannH::CLASS_NAME) && (getLevel() == o_e->getLevel())) {
		//h^{abc} h^{aij} --> ...
		//
		//if (CTIndex::CheckCoupled(GetTIndex(1),other->GetTIndex(1))) {
		//	CExpression* res = CENumber::GetFraction(4,3);
		//	copyAttributesTo(res,false); if(!o_e->getFlag())res->inverseFlag();
		//	return res;
		//}
		//

		//[temporary, to be removed to single procedure with 2h, 4h,... - couplings]
		if (CTIndex::CheckCoupled(GetTIndex(1),other->GetTIndex(1))) {
			if ((CTIndex::CheckCoupled(GetTIndex(2),other->GetTIndex(2))) &&
				(CTIndex::CheckCoupled(GetTIndex(3),other->GetTIndex(3)))) {
				CExpression* res=CENumber::GetFraction(32,3); res->inverseFlag(); //-32/3
				copyAttributesTo(res,true); if(!o_e->getFlag())res->inverseFlag();
				return res;
			}
			if ((CTIndex::CheckCoupled(GetTIndex(2),other->GetTIndex(3))) &&
				(CTIndex::CheckCoupled(GetTIndex(3),other->GetTIndex(2)))) {
				CExpression* res=CENumber::GetFraction(112,3); //112/3
				copyAttributesTo(res,true); if(!o_e->getFlag())res->inverseFlag();
				return res;
			}
		}
		return NULL;
	}
	if(!bCalledByOther){ return other->unite_with(this, !other_is_at_right, true); }
	return NULL;
}

void CEGellMannH::RotateIndices(int nTimes) { //nTimes: 0,1,2, abc -> cab -> bca
	if(nTimes<=0)return;
	if(nTimes>2){ nTimes=nTimes%3; }
	for(int i=0; i<nTimes; i++){
		CExpression* i1=index(1); CExpression* i2=index(2); CExpression* i3=index(3);
		if((i1==NULL) || (i2==NULL) || (i3==NULL))break;
		CExpression* i1Cpy=i1->clone(); if(i1Cpy==NULL)break;
		setIndex(1,i3); setIndex(3,i2); setIndex(2,i1Cpy);
		delete i1Cpy;
	}
}

//
///////////////////////////////////////////////////// CEGellMannTr: Tr[t^a t^b ...]
//
CDebug* CEGellMannTr::dbg = new CDebug("CEGellMannTr");

//
// static entries
//

//	check if the object is G-M operator (contains t^a-matrices)
bool CEGellMannTr::IsGellMannOperator(CExpression* src) {
	CRContainCheck r(CEGellMann::CLASS_NAME); r.ApplyTo(src,false);
	return r.getSearchResult();
}

//	extract & evaluate trace
//	Nt: Trace[non-GM-objects] = non-GM-objects*Tr[1] = non-GM-objects*3
//	returns cloned result
CExpression* CEGellMannTr::TraceCalc(CExpression* src, bool do_delete_src) {
	dbg->log(35, "TraceCalc", src);

	CEGellMannTr* f_tr = new CEGellMannTr(src); if(do_delete_src)delete src;
	CExpression* res = f_tr->evaluate(); if(res!=f_tr)delete f_tr;

	//res = CExpression::reduce(res); //expand here?
	dbg->log(55, "TraceCalc finished", res);
	return res;
}

//	extract & factorize trace, don't evaluate
//	returns cloned result
CExpression* CEGellMannTr::TraceFactorize(CExpression* src, bool do_expand_src, bool do_delete_src) {
	dbg->log(35, "TraceFactorize", src);
	CEGellMannTr* f_tr = new CEGellMannTr(src); if(do_delete_src)delete src;

	CExpression* res = f_tr->ToNormalForm(do_expand_src);
	if(res==NULL){ res = f_tr; }else if(res!=f_tr){ delete f_tr; }

	//res = CExpression::reduce(res); //expand here?
	dbg->log(55, "TraceFactorize finished", res);
	return res;
}

//
CEGellMannTr::CEGellMannTr(const CExpression* src) : CEAbstractFunction("Tr", src) {}
CEGellMannTr::~CEGellMannTr() {}
const char* CEGellMannTr::CLASS_NAME = "CEGellMannTr";
const char* CEGellMannTr::ClassName() const { return CLASS_NAME; }
CExpression* CEGellMannTr::clone() const {
	CEGellMannTr* res = new CEGellMannTr(dataConst()); copyAttributesTo(res); return res;
}
unsigned long CEGellMannTr::props_long() {
	unsigned long res = CEAbstractFunction::props_long();
	res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}
void CEGellMannTr::collectProperties(CEProperties* props, unsigned set) {
	CEAbstractFunction::collectProperties(props, set);
	props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}
CExpression* CEGellMannTr::Arrange(int level) { // To extract Tr[0]->0
	if (level==AL_EXTRACT) {
		CExpression* src = data();
		if ((src!=NULL) && (CEBasic::isZeroProd(src))) {
			CExpression* res=CENumber::GetZero(); copyAttributesTo(res); if(!res->getFlag())res->inverseFlag(); return res;
		}
	}
	return this;
}

CExpression* CEGellMannTr::evaluate(int level) { // evaluate trace
	dbg->log(105, "evaluate", this);

	// 1. Evaluate normal-formed trace
	if(IsNormalForm()){
		CExpression* src = data();
		if(!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			// Unit
			if (CEBasic::isUnit(src)) {
				CExpression* res=CENumber::GetNumber(3);	//TODO: involve Dim[colors] or N[colors]
				copyAttributesTo(res,true);
				dbg->log(105, "evaluate result", res);
				return res;
			}

			// Single t^a
			if (src->ClassName()==CEGellMann::CLASS_NAME) {
				CExpression* res=CENumber::GetZero();
				copyAttributesTo(res,true);
				dbg->log(105, "evaluate result", res);
				return res;
			}
			CEException::halt("IError in CEGellMannTr::evaluate"); return this;
		}

		// Product{t^a t^b t^c ...}
		CExpression* res = CEGellMannTr::CalcTrace((CEProduct*)src);
		if(res!=NULL){ copyAttributesTo(res,true); return res; }
		return this;
	}

	// 2. Convert trace to normal form & evaluate it
	CExpression* valNorm = ToNormalForm(true); if(valNorm==NULL)return this;
	//if(!valNorm->IsNormalForm()) { CEException::halt("IError in CEGellMannTr::evaluate: can't turn object to normal form"); if(valNorm!=this)delete valNorm; return this; }
	CExpression* res = valNorm->evaluate(level); if(res=NULL)return valNorm;
	if((valNorm!=res) && (valNorm!=this)){ delete valNorm; }
	dbg->log(105, "evaluate result", res);
	return res;
}

bool CEGellMannTr::IsNormalForm() {	// normal form is Tr[t^a t^b ... t^m], no other objects
	CExpression* src = data(); if(src==NULL)return false;

	if(!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->ClassName()==CEGellMann::CLASS_NAME) return true;
		if (CEBasic::isUnit(src)) return true;
		return false;
	}

	CEProduct* pr=(CEProduct*)src; long s=pr->size(); if(s==0)return false;
	for(long i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i); if(item==NULL)return false;
		if(item->ClassName() != CEGellMann::CLASS_NAME) return false;
	}
	return true;
}

CExpression* CEGellMannTr::ToNormalForm(bool bExpandSrc) {	// to Sum{ const*Tr[t^a t^b ... t^m] }, returns this or cloned, NULL if failed
	if(IsNormalForm())return this;

	CExpression* src = data(); bool bSrcCloned=false;
	if(src==NULL)return NULL;								//error

	// 1. expand source
	if(bExpandSrc){
		src = src->clone(); bSrcCloned=true;
		src = CExpression::reduce(src);
	}

	// 2. expand sum
	if(src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		if(!src->getLevel())return NULL;					//error
		CESum* src_sum = (CESum*)src; long s=src_sum->size();
		CESum* res = new CESum();
		for(long i=0; i<s; i++) {
			CEGellMannTr* item_tr = new CEGellMannTr(src_sum->GetAt(i)); CExpression* item = NULL;
			if (item_tr->IsNormalForm()) {
				item = item_tr;
			}else{
				item = item_tr->ToNormalForm(false); if(item!=item_tr){ delete item_tr; }
				if(item==NULL){
					dbg->log(0, "Inernal Error: CEGellMannTr::ToNormalForm() failed", this);
					delete res; if(bSrcCloned)delete src;
					return NULL;
				}
			}
			res->addItem(item);
		}
		src->copyAttributesTo(res,true); copyAttributesTo(res,true);
		if(bSrcCloned){ delete src; }
		return res;
	}

	// 3. turn 'src' to product + expand product
	if(!src->getLevel()){ return NULL; }					//error
	if(!bSrcCloned){ src=src->clone(); bSrcCloned=true; }	//clone src if not cloned before
	CEProduct* res = CEBasic::convertToProduct(src);		//'src' is now destroyed
	CEProduct* res_GM = new CEProduct();
	for(long i=0; i<res->size(); i++) {						//extract t^a from 'res' to 'res_GM'
		CExpression* item = res->GetAt(i);
		if(item->ClassName() == CEGellMann::CLASS_NAME) {
			res_GM->addItem(res->RemoveAtX(i)); i--;
		}
	}
	if(res_GM->size()==0)res_GM->addItem(CENumber::GetUnit());
	CEGellMannTr* tr_GM = new CEGellMannTr((res_GM->size()>1) ? ((CExpression*)res_GM) : ((CExpression*)res_GM->GetAt(0)));
	delete res_GM;
	//if (!tr_GM->IsNormalForm()) {}						//extra check - not required
	res->addItem(tr_GM);
	copyAttributesTo(res,true);
	return res;												// don't delete 'src' here!
}

//
//	Product{t^a t^b t^c ...}
//	// 'pr' is {t^a t^b ... t^m'} unaffected
//
CExpression* CEGellMannTr::CalcTrace(CEProduct* pr) {
	if(pr==NULL)return NULL;

	// extra checks; these cases must not occur here, but ...
	if(!pr->getLevel())return NULL;
	int n = pr->size();
	for(int i=0; i<n; i++) {
		CExpression* item = pr->GetAt(i); if(item==NULL)return NULL;
		if(item->ClassName() != CEGellMann::CLASS_NAME) return NULL;
		if((!item->getFlag()) || (!item->getLevel()))return NULL;
	}

	CExpression* res = CalcTraceImpl(pr);
	if(res != NULL) {
		if(!pr->getFlag())res->inverseFlag();				//Nt: pt's flag is ignored in 'CalcTraceImpl'
		res = CExpression::reduce(res);						//Inline sub-products
		return res;
	}
	return NULL;
}

//
//	'pr' is {t^a t^b ... t^m'} or {}, unaffected
//	Used for recursive calls
//	Nt: "!level" is allowed in 'pr' here
//	Nt: "!sign" is allowed but ignored here
//
CExpression* CEGellMannTr::CalcTraceImpl(CEProduct* pr) {
	if(pr==NULL)return NULL;
	int n = pr->size();

	if(n==0){
		CExpression* res=CENumber::GetNumber(3);			//TODO: involve Dim[colors] or N[colors]
		pr->copyAttributesTo(res,true); return res;
	}

	if(n==1){
		CExpression* res=CENumber::GetZero();
		pr->copyAttributesTo(res,true); return res;
	}

	CExpression* prCpyE = pr->clone();
	if((prCpyE==NULL) || (!prCpyE->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) || (((CEProduct*)prCpyE)->size()!=n)) { dbg->log(0,"Clone error",pr); CEException::halt("IError in CEGellMannTr::CalcTraceImpl"); return NULL; }
	CEProduct* prCpy = (CEProduct*)prCpyE;

	CEGellMann* tA=(CEGellMann*)prCpy->GetAt(n-2); CTIndex* tindA=tA->GetTIndex(1);
	CEGellMann* tB=(CEGellMann*)prCpy->GetAt(n-1); CTIndex* tindB=tB->GetTIndex(1);
	CESymbol* indC = CESymbol::getUnique();					//TODO: avoid extra symbol booking for every trace
															//	or implement symbol releasing
	//t^a t^b --> 1/6 \delta^ab + 1/2 h^abc t^c
	CEMetric* dAB = new CEMetric(tindA,tindB);				//we presume 'covariance' as unused
	CEGellMannH* hABC = new CEGellMannH(tindA->index(),tindB->index(),indC);
	CEGellMann* tC = new CEGellMann(indC);
	delete indC;

	prCpy->RemoveAt(n-1); prCpy->SetAtX(n-2,tC);
	CExpression* item2 = CEGellMannTr::CalcTraceImpl(prCpy);
	prCpy->RemoveAt(n-2);
	CExpression* item1 = CEGellMannTr::CalcTraceImpl(prCpy);
	delete prCpy;

	if((item1==NULL) || (item2==NULL)){ delete dAB; delete hABC; return NULL; }
	CExpression* res = (CENumber::GetFraction(1,6)->multiply(dAB)->multiply(item1))
		->add(CENumber::GetFraction(1,2)->multiply(hABC)->multiply(item2));
	return res;
}
