////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// ESProduct.cpp: implementation of the CESProduct class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const char* CESProduct::CLASS_NAME		= "CESProduct";
const char* CESProduct::CLASS_TAG_NAME	= "spr";// or "cpl"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CESProduct::CESProduct(const CExpression* a, const CExpression* b):CEScalar()
{
	m_a = a->clone();
	m_b = b->clone();
}

CESProduct::~CESProduct()
{
	delete m_a;
	delete m_b;
}

/*
* The product should first be arranged.
* (Neither m_a no n_b may contain sign.).
*/
CEString* CESProduct::toString(int mode)
{
	CEString* res = NULL;
	switch (mode)
	{
	case STORE_CPP:
	case STORE_FORTRAN_SYMBOL:
		res = new CEString((getFlag()?"":"-")); res->AddString(m_a->toString(mode)); *res<<"_"; res->AddString(m_b->toString(mode));
		break;
	case STORE_WR_MATHEMATICA:
		if (m_a->equals(m_b)) {
			res = new CEString((getFlag()?"":"- ")); StoreInBrackets(res, m_a, mode); *res<<"\\^2";
		}
		else {
			res = new CEString((getFlag()?"\\(":"- \\("));
			StoreInBrackets(res, m_a, mode); *res<<" . "; StoreInBrackets(res, m_b, mode); *res<<"\\)";
		}
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

CExpression* CESProduct::clone() const
{
	CExpression* res = new CESProduct(m_a, m_b);
	copyAttributesTo(res);
	return res;
}

void CESProduct::collectProperties(CEProperties* props, unsigned set) {
	CETensor::collectProperties(props, set);
	if ((m_a->props_long()&CEProperties::PROP_MASK_SELF_CONJUGATED)==0) props->setCommutative(false);
	if ((m_b->props_long()&CEProperties::PROP_MASK_SELF_CONJUGATED)==0) props->setCommutative(false);
	if ((m_a->props_long()&CEProperties::PROP_MASK_SELF_CONJUGATED)==0) props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
	if ((m_b->props_long()&CEProperties::PROP_MASK_SELF_CONJUGATED)==0) props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

unsigned long CESProduct::props_long() {
	unsigned long res = CETensor::props_long();
	if ((m_a->props_long()&CEProperties::PROP_MASK_COMMUTATIVE)==0) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_COMMUTATIVE,false);
	if ((m_b->props_long()&CEProperties::PROP_MASK_COMMUTATIVE)==0) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_COMMUTATIVE,false);
	if ((m_a->props_long()&CEProperties::PROP_MASK_SELF_CONJUGATED)==0) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	if ((m_b->props_long()&CEProperties::PROP_MASK_SELF_CONJUGATED)==0) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}

CExpression* CESProduct::applyRule(CEAbstractRule *rule)
{
	CERule* r = (CERule*)rule;
	m_a = r->ApplyTo(m_a);
	m_b = r->ApplyTo(m_b);
	return this;

	//dbg->log(25, "CESProduct::applyRule finished", this);
}

CExpression* CESProduct::Arrange(int level)
{
	m_a = CExpression::arrge(m_a);
	m_b = CExpression::arrge(m_b);
	if (m_b->less_than(m_a)) {
		CExpression* temp = m_a;
		m_a = m_b;
		m_b = temp;
	}

	if (!m_a->getFlag()) { inverseFlag(); m_a->inverseFlag(); }
	if (!m_b->getFlag()) { inverseFlag(); m_b->inverseFlag(); }

	if (level == AL_EXTRACT) {
		if (CEBasic::isZero(m_a) || CEBasic::isZero(m_b)) {
			CExpression* res = CENumber::GetZero();
			copyAttributesTo(res, true);
			return res;
		}
		//if ( (CEBasic::isZero(m_a)) || (CEBasic::isZero(m_b)) ) {
		//	CExpression* res = CENumber::GetZero();
		//	copyAttributesTo(res, true);
		//	return res;
		//}
	}
	return this;
}

bool CESProduct::contain_vector(void * obj)
{
	if (CETensor::contain_vector(obj)) return true;
	if ((m_a == obj) || (m_a->contain_vector(obj))) return true;
	if ((m_b == obj) || (m_b->contain_vector(obj))) return true;
	return false;
}

bool CESProduct::equals(CExpression *other)
{
	if (other == NULL) return false;
	if (!CEUtil::equals(other->ClassName(), CLASS_NAME)) return false;

	CESProduct* o = (CESProduct*)other;

	if ((get_a()->equals(o->get_a())) &&
		(get_b()->equals(o->get_b())) &&
		(attributes_equals_to(o))) return true;

	return false;
}

bool CESProduct::less_than(CExpression *other) {
	if (other == NULL) return false;

	// compare to CESymbol, CEPower, CESProduct, CEVector, CEMetric, CEAntisymmetric, CEAntiSymmProd, CEFuncBondarev
	// a < a^n < a.b < a_nu < a_mu_nu < a_abcd
	const char* o_cn = other->ClassName();
	if (o_cn == CEPower::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CEPower* o = (CEPower*)other;
		if (get_a()->less_than(o->basis())) return true;
		return false;
	}

	if (o_cn == CESymbol::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		if (get_a()->less_than(other)) return true;
		return false;
	}

	if (o_cn == CESProduct::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CESProduct* o = (CESProduct*)other;
		if (get_a()->less_than(o->get_a())) return true;
		if (!get_a()->equals(o->get_a())) return false;
		if (get_b()->less_than(o->get_b())) return true;
		return false;
	}

	if ((o_cn==CEVector::CLASS_NAME)||(o_cn==CEMetric::CLASS_NAME)||(o_cn==CEAntisymmetric::CLASS_NAME)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		IETensor* o = IETensor::GetIETensor(other);
		if (get_a()->less_than(o->base())) return true;
		if (!get_a()->equals(o->base())) return false;
		return true;
	}

	if (o_cn==CEAntiSymmProd::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	if (o_cn==CEFuncBondarev::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	return CExpression::less_than(other);
}

CExpression* CESProduct::expand(bool recursive)
{
	Arrange(AL_INLINE);

	if (m_a->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* res = new CESum();
		CESum* a = (CESum*)m_a;
		bool a_sign = a->getFlag(); if (!a_sign) dbg->log(0, "Warning: expanding of non-arranged CESProduct", this);
		long s_a = a->size();
		long i;
		if (m_b->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			CESum* b = (CESum*)m_b;
			bool b_sign = b->getFlag(); if (!b_sign) dbg->log(0, "Warning: expanding of non-arranged CESProduct", this);
			long s_b = b->size();
			for (i=0; i<s_a; i++) {
				long j;
				for (j=0; j<s_b; j++) {
					res->addItem(new CESProduct(a->GetAt(i), b->GetAt(j)));
				}
			}
			if (!b_sign) res->inverseFlag();
		}
		else {
			for (i=0; i<s_a; i++) {
				res->addItem(new CESProduct(a->GetAt(i), m_b));
			}
		}
		if (!a_sign) res->inverseFlag();
		copyAttributesTo(res, true);
		CExpression* r = res->expand();
		if (r != res) delete res;
		return r;
	}
	else if (m_b->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* res = new CESum();
		CESum* b = (CESum*)m_b;
		bool b_sign = b->getFlag(); if (!b_sign) dbg->log(0, "Warning: not arranged CESProduct expanding", this);
		long s_b = b->size();
		long j;
		for (j=0; j<s_b; j++) {
			res->addItem(new CESProduct(m_a, b->GetAt(j)));
		}
		if (!b_sign) res->inverseFlag();
		copyAttributesTo(res, true);
		CExpression* r = res->expand();
		if (r != res) delete res;
		return r;
	}
	if (m_a->equals(m_b)) {
		CExpression* res = CEVector::get_predefined_square(m_a);
		if (res != NULL) {
			copyAttributesTo(res, true);
			return res;
		}
	}
	return this;
}

const char* CESProduct::ClassName() const
{
	return CESProduct::CLASS_NAME;
}

CExpression* CESProduct::get_a()
{
	return m_a;
}

CExpression* CESProduct::get_b()
{
	return m_b;
}

CESerializable* CESProduct::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	long skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) {
		//Error
		dbg->log(0, "Error restoring CESProduct");
		return NULL;
	}
	CESerializable* item_1 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;
	
	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) {
		//Error
		dbg->log(0, "Error restoring CESProduct");
		delete item_1;
		return NULL;
	}
	CESerializable* item_2 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos > 0) {
		//Error
		dbg->log(0, "Error restoring CESProduct");
		delete item_1;
		delete item_2;
		return NULL;
	}
	
	CESProduct* res = new CESProduct((CExpression*)item_1, (CExpression*)item_2);
	delete item_1; delete item_2;
	return res;
}

const char* CESProduct::get_TAG_Name()
{
	return CLASS_TAG_NAME;
}

long CESProduct::store_stream(ofstream *p_ofs, DWORD settings)
{
	long res = CESupport::store_stream(m_a, p_ofs, settings);
	res += CESupport::store_stream(m_b, p_ofs, settings);
	return res;
}

CExpression* CESProduct::numeric()
{
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) return this;

	CENumVector* stub_a = kin->get_vector_stub(m_a);
	if (stub_a == NULL) return this;
	CENumVector* stub_b = kin->get_vector_stub(m_b);
	if (stub_b == NULL) return this;

	CExpression* res = stub_a->couple(stub_b, false);
	copyAttributesTo(res, true);
	return res;
}
