////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// ESpinor.cpp: implementation of the CESpinor class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CESpinor::CESpinor(const CExpression* base, bool _is_conjugated)
:CETensor(base)
{
	//m_base = base->clone();
	m_conjugated = _is_conjugated;
}

CESpinor::~CESpinor()
{
	//delete m_base;
}

void CESpinor::collectProperties(CEProperties *props, unsigned int set)
{
	CETensor::collectProperties(props, set);
	props->setSpinor(true);
	props->setCommutative(false);
	props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

unsigned long CESpinor::props_long()
{
	unsigned long res = CETensor::props_long();
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_IS_SPINOR, true);
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_COMMUTATIVE, false);
	res = CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}

const char* CESpinor::CLASS_NAME = "CESpinor";
const char* CESpinor::ClassName() const { return CLASS_NAME; }

CEString* CESpinor::toString(int mode)
{
	CEString* res = NULL;
	switch (mode)
	{
	case STORE_WR_MATHEMATICA:
		res = new CEString(((getFlag())?"u":"- u"));
		if (m_conjugated) res->AddString("\\&_(");
		else res->AddString("(");
		res->AddString(m_base->toString(mode));
		res->AddString(")");
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

CExpression* CESpinor::applyRule(CEAbstractRule *rule)
{
	CERule* r = (CERule*)rule;
	CExpression* item = base();
	CExpression* item_repl = r->ApplyTo(item, false);
	if ((item_repl != item) && (item_repl != NULL)) {
		setBase(item_repl);
		delete item_repl;
	}
	return this;
}

/*CExpression* CESpinor::base()
{
	return m_base;
}*/

bool CESpinor::contain_vector(void *obj)
{
	if (CETensor::contain_vector(obj)) return true;
	return false;
}

CExpression* CESpinor::conjugated(int mode) {
	CExpression* res = new CESpinor(base(), !m_conjugated);
	copyAttributesTo(res, true);
	return res;
}

/*bool CESpinor::setBase(CExpression *_new)
{
	bool res = false;
	if ((m_base != _new) && (_new != NULL)) {
		delete m_base;
		m_base = _new->clone();
		res = true;
	}
	return res;
}*/

CExpression* CESpinor::unite_with(IETensor * other, bool other_is_at_right, bool bCalledByOther) {
	if (other == NULL) { dbg->log(0, "<Warning><CESpinor::unite_with : NULL parameter.>"); return NULL; }

	CExpression* o_e = other->ThisExpr();
	if (o_e->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) return NULL;

	// uncomment next lines for automatic spin summation
	//if (o_e->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) {
	//	return couple_with_spinor((CESpinor*)o_e, other_is_at_right);
	//}

	// Dirac equation involving
	// u(p)p->um
	// u(-p)p->u(-m)
	if (CEUtil::equals(o_e->ClassName(), CESGammaProduct::CLASS_NAME)) {
		CESGammaProduct* o = (CESGammaProduct*)o_e;
		if (is_conjugated() != other_is_at_right) return NULL;

		bool p_flag = true;
		if (base()->getFlag()) {
			if (!base()->equals(o->getBaseExpression())) return NULL;
		}
		else {
			base()->inverseFlag();
			if (!base()->equals(o->getBaseExpression())) { base()->inverseFlag(); return NULL; }
			base()->inverseFlag();
			p_flag = false;
		}

		// get mass form physics data
		CExpression* mass_sym = NULL;
		CKinematics* kin = CEProcessor::kinematics();
		if (kin != NULL) {
			CExpression* mom_abs = CLSimplifyMisc::TruncateSign(base());
			mass_sym = kin->get_mass(mom_abs);
			delete mom_abs;
			if (mass_sym != NULL) mass_sym = mass_sym->clone();
		}
		if (mass_sym == NULL) {
			dbg->log(0, "Warning: mass symbol not found for spinor", this);
			mass_sym = CESymbol::getUnique();
			dbg->log(0, "New (UNIQUE) symbol for mass involved", mass_sym);
		}

		CExpression* res = NULL;
		if (other_is_at_right) res = clone()->multiply(mass_sym);
		else res = mass_sym->multiply(clone());
		if (!o_e->getFlag()) res->inverseFlag();
		if (!p_flag) res->inverseFlag();
		return res;
	}
	return NULL;
}

CExpression* CESpinor::couple_with_spinor(CESpinor* other, bool other_is_at_right) {
	
	// compare spinors
	if (!base()->equals(other->base())) return NULL;
	if ((!(is_conjugated())) == (!(other->is_conjugated()))) return NULL;

	// \bar{u} u -- spinor normalization -- not implemented
	if ((!is_conjugated()) == (!other_is_at_right)) return NULL;

	// process POLARIZATION_SUM{ u\bar{u} } further
	// find out perticle info, return NULL for unknown spinor
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) return NULL;
	
	CExpression* mom_abs = CLSimplifyMisc::TruncateSign(base());
	CEParticle* part = kin->get_particle(mom_abs); delete mom_abs;
	if (part == NULL) return NULL;
	if (!part->check_PK(CEParticle::PARTICLE_FERMION, true)) { dbg->log(0, "Warning: Spinor for non-fermion particle spotted", part); throw(new CEException(CEException::INT_ERROR)); }

	// RETRIEVE MASS SYMBOL AND POLARIZATION STATE FROM KINEMATICS
	// LET PARTICLE OBJECT DEFINE THE DENSITY MATRIX
	CEPtrArray* DM_vars = new CEPtrArray(); DM_vars->add(base()->clone());
	CExpression* res = part->getDensityMatrix(DM_vars); delete DM_vars;
	if (res == NULL) { dbg->log(0, "Warning: No density matrix found for particle", part); return NULL; }

	// return density matrix
	if (!getFlag()) res->inverseFlag();
	if (!other->getFlag()) res->inverseFlag();
	return res;
}

CExpression* CESpinor::clone() const {
	CExpression* res = new CESpinor(GetBase(), is_conjugated());
	copyAttributesTo(res);
	return res;
}
