////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// ETensor.cpp: implementation of the CETensor class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CETensor::dbg = new CDebug("CETensor");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CETensor::CETensor(const CExpression* base):CExpression()
{
	if (base != NULL) {
		m_base = base->clone();
	} else {
		m_base = new CESymbol("X"); // default base
	}
	m_indices = new CEPtrArray();
}

CETensor::~CETensor()
{
	delete m_indices;
	delete m_base;
}

void CETensor::collectProperties(CEProperties* props, unsigned set) {
	CExpression::collectProperties(props, set);
	if (m_base != NULL) {
		unsigned long lPropsI=m_base->props_long();
		if ((lPropsI & CEProperties::PROP_MASK_COMMUTATIVE)==0) props->setCommutative(false);
		if ((lPropsI & CEProperties::PROP_MASK_SELF_CONJUGATED)==0) props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
	}
	props->setTensor();
}

unsigned long CETensor::props_long() {
	unsigned long res = CExpression::props_long();
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_IS_TENSOR, true);
	if (m_base != NULL) {
		unsigned long lPropsI=m_base->props_long();
		if ((lPropsI & CEProperties::PROP_MASK_COMMUTATIVE)==0) res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_COMMUTATIVE, false);
		if ((lPropsI & CEProperties::PROP_MASK_SELF_CONJUGATED)==0) res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	}
	return res;
}

bool CETensor::equals(CExpression* other) {
	//if (!other->getProperty(CEProperties::PROP_MASK_IS_TENSOR)) return false;
	if (!CEUtil::equals(ClassName(), other->ClassName())) return false;
	if (!attributes_equals_to(other)) return false;
	CETensor* o = (CETensor*)other;

	if (!m_base->equals(o->base())) return false;
	int r = rank();
	if (r != o->rank()) return false;
	for(int i=1; i<=r; i++) {
		CTIndex* ti1=GetTIndex(i); CTIndex* ti2=o->GetTIndex(i);
		if(ti1!=NULL){ if(!ti1->equalsIndex(ti2))return false; }else{ if(ti2!=NULL)return false; }
	}
	return true;
}

bool CETensor::less_than(CExpression* other) {
	if (other == NULL) return false;
	const char* o_cn = other->ClassName();
	if (o_cn == ClassName()) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return TLessThen(IETensor::GetIETensor(other), o_cn);
	}
	return CExpression::less_than(other);
}

bool CETensor::TLessThen(IETensor* o, const char* o_cn) {
	// for internal use
	// [Update] Warning: this method violates 'less_than' rule "first compare sign & CLASS_NAME, than others"!
	// [Update] ToDo: fix it!
	//
	if (base()->less_than(o->base())) return true;
	if (!base()->equals(o->base())) return false;
	int s = rank(); int s2 = o->rank();
	for(int i=1; ((i<=s) && (i <= s2)); i++) {
		if (!CEUtil::LEQU(covariance(i), o->covariance(i))) return covariance(i); // cov is < than !cov
		CExpression* ind1 = index(i); CExpression* ind2 = o->index(i);
		if (ind1->less_than(ind2)) return true;
		if (!ind1->equals(ind2)) return false;
	}
	if (s != s2) return (s<s2);
	if (ClassName() != o_cn) return CEUtil::a_less_b(ClassName(),o_cn);
	return false;
}

CExpression* CETensor::index(int i) {
	CTIndex* tindex=(CTIndex*)m_indices->GetAt(i-1);
	return ((tindex != NULL) ? tindex->index() : NULL);
}
const CExpression* CETensor::GetIndex(int i) const {
	const CTIndex* tindex=(const CTIndex*)m_indices->GetAtConst(i-1);
	return ((tindex != NULL) ? tindex->GetIndex() : NULL);
}
CTIndex* CETensor::GetTIndex(int i) {
	return (CTIndex*)m_indices->GetAt(i-1);
	//CTIndex* res = (CTIndex*)m_indices->GetAt(i-1);
	//if(res==NULL)dbg->log(0,"<Warning><CETensor: Invalid GetTIndex() parameter.>");
	//return res;
}
const CTIndex* CETensor::GetTIndexConst(int i) const { return ((m_indices!=NULL)?(const CTIndex*)m_indices->GetAtConst(i-1) : NULL); }

bool CETensor::covariance(int i) const {
	CTIndex* tindex = (CTIndex*)m_indices->GetAtConst(i-1);
	if (tindex != NULL) return tindex->covariance();
	return false;
}

bool CETensor::setIndex(int i, CExpression* index) {
	CTIndex* tindex = (CTIndex*)m_indices->GetAt(i-1);
	if (tindex != NULL) return tindex->setIndex(index);
	dbg->log(0, "<Warning><CETensor: Invalid setIndex request parameter.>");
	return false;
}

bool CETensor::liftIndex(int i, bool to_covariance)
{
	CTIndex* tindex = (CTIndex*)m_indices->GetAt(i-1);
	if (tindex != NULL) return tindex->liftIndex(to_covariance);
	log("<Warning><CETensor: Invalid liftIndex request parameter.>");
	return false;
}

bool CETensor::changeIndex(int i, CExpression *index, bool covariance)
{
	CTIndex* tindex = (CTIndex*)m_indices->GetAt(i-1);
	if (tindex != NULL) return tindex->changeIndex(index, covariance);
	log("<Warning><CETensor: Invalid changeIndex request parameter.>");
	return false;
}

CExpression* CETensor::ThisExpr() { return (CExpression*)this; }
CExpression* CETensor::base() { return m_base; }
const CExpression* CETensor::GetBase() const { return m_base; }

CExpression* CETensor::unite_with(IETensor *other, bool other_is_at_right, bool bCalledByOther) {
	if(!bCalledByOther){ return other->unite_with(this, !other_is_at_right, true); }
	return NULL;
}

/*
*	To be called AFTER child commutation check (calls other->exchange_subst())
*/
CExpression* CETensor::exchange_subst(IETensor* other, bool other_is_at_right, bool bCalledByOther) {
	CExpression* res = NULL;
	//commutatives
	CExpression* eOther=other->ThisExpr();
	if ((eOther->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) || (getProperty(CEProperties::PROP_MASK_COMMUTATIVE))) {
		if(other_is_at_right){ res=eOther->clone()->multiply(clone()); }else{ res=clone()->multiply(eOther->clone()); }
	}
	//call other->exchange_subst()
	if ((res==NULL) && (!bCalledByOther)) {
		res=other->exchange_subst(this, !other_is_at_right, true);
	}
	return res;
}

void* CETensor::GetInterface(long interface_ID) {
	if (interface_ID == CConst::dwIETensor) return this;
	return CExpression::GetInterface(interface_ID);
}

int CETensor::rank() const { return m_indices->size(); }

CEString* CETensor::toString(int mode)
{
	CEString* res = NULL;
	bool last_covariance;
	int i;
	switch (mode)
	{
	case STORE_WR_MATHEMATICA:
		//Nt: sign is saved later

		//Note: see '\(..\)' brackets politics in InternalNotes[1.1].
		if(IsSimpleItem(m_base)){ //additional '\(..\)'-brackets are required for simple items
			res=new CEString("\\("); res->AddString(m_base->toString(mode)); res->AddString("\\)");
		}else{ //branch is similar to 'StoreInBrackets()'
			res=new CEString("\\(("); res->AddString(m_base->toString(mode)); res->AddString(")\\)");
		}

		for(i=0; i<rank(); i++) {
			if (i==0) {
				last_covariance = covariance(1);
				res->AddString((last_covariance)?"\\_":"\\&");
				res->AddString("\\(");
				res->AddString(index(1)->toString(mode));
				res->AddString("\\)");
			}
			else {
				bool cov = covariance(i+1);
				if (cov == last_covariance) {
					long ins_pos = res->length()-2;//insert before bracket closed
					res->insertString(index(i+1)->toString(mode), ins_pos);
					res->insertString("\\ ", ins_pos);
				}
				else {
					res->insertString("\\(", 0);
					res->AddString("\\)");
					res->AddString((cov)?"\\_":"\\&");
					res->AddString("\\(");
					res->AddString(index(i+1)->toString(mode));
					res->AddString("\\)");
					last_covariance = cov;
				}
			}
		}

		if (!getFlag()) {
			res->insertString("- \\(", 0);
			res->AddString("\\)");
		}
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

/*
* Warning: changing tensor's base break down tensor's structure
* (for example, we cann't change metric's base)
* Maybe, we should not allow base changing.
*/
CExpression* CETensor::applyRule(CEAbstractRule *rule)
{
	CERule* r = (CERule*)rule;
	CExpression* item = base();
	CExpression* item_repl = r->ApplyTo(item, false);
	if ((item_repl != item) && (item_repl != NULL)) {
		setBase(item_repl);
		delete item_repl;
	}
	
	int s = rank();
	int i;
	for(i=1; i<=s; i++) {
		item = index(i);
		item_repl = r->ApplyTo(item, false);
		if ((item_repl != item) && (item_repl != NULL)) {
			setIndex(i, item_repl);
			delete item_repl;
		}
	}
	return this;
}

bool CETensor::setBase(CExpression *_new)
{
	bool res = false;
	if ((m_base != _new) && (_new != NULL)) {
		delete m_base;
		m_base = _new->clone();
		res = true;
	}
	return res;
}

bool CETensor::contain_vector(void * obj)
{
	CExpression* b = base();
	if ((b == obj) || (b->contain_vector(obj))) return true;

	int s = rank();
	int i;
	for(i=0; i<s; i++) {
		CTIndex* ind = (CTIndex*)m_indices->GetAt(i);
		if ((ind == obj) || (ind->contain_vector(obj))) return true;
	}
	return false;
}

void CETensor::exchange_indices(int i, int j)
{
	CTIndex* tindex_1 = (CTIndex*)m_indices->GetAt(i-1);
	CTIndex* tindex_2 = (CTIndex*)m_indices->GetAt(j-1);
	m_indices->SetAt(i-1, tindex_2);
	m_indices->SetAt(j-1, tindex_1);
}

bool CETensor::contain_index(CExpression *ind)
{
	if (ind == NULL) return false;
	int s = rank();
	int i;
	for(i=0; i<s; i++) {
		CTIndex* t_ind = (CTIndex*)m_indices->GetAt(i);
		if (ind->equals(t_ind->index())) return true;
	}
	return false;
}

/*
*	Changes the UNIQUE-SYMBOL tensors indices in the expression to another uniques
*	Does not changes indices, listed in "protected_syms" array
*	Does not afects "protected_syms" list
*	Warning #1: does not changes non-unique indices
*		==> to be used carefully
*	Warning #2: changes also spare vector indices in whole expression
*		==> involve common vector indices as "protected_syms" list
*/
CExpression* CETensor::ChangeVolatileIndices(CExpression* src, CEPtrArray* protected_syms) {
	if (src == NULL) return src;

	CRCollectSymbols* r_coll = new CRCollectSymbols(
		CRCollectSymbols::COLLECT_UNIQUES | CRCollectSymbols::COLLECT_TENSOR_INDICES_ONLY,
		protected_syms);
	src = r_coll->ApplyTo(src); // nothing changes to 'm_expr_val' here
	CEPtrArray* unique_indices = r_coll->getCollectedSymbols();

	int s = unique_indices->size();
	if (s == 0) { delete r_coll; return src; } // nothing to change

	// create unique substitutes
	CEReplList* r_list = new CEReplList();
	int i;
	for(i=0; i<s; i++) {
		CExpression* unique_new = CESymbol::getUnique();
		r_list->add((CExpression*)unique_indices->GetAt(i), unique_new);
		delete unique_new;
	}
	delete r_coll; // unique_indices is no longer required
	
	src = r_list->ApplyTo(src);
	delete r_list;
	return src;
}

bool CETensor::m_IsNDimensionSpace = false; // true for N-dim, false for 4-space
void CETensor::setNDimensionSpace(bool set_N) { m_IsNDimensionSpace=set_N; }
CExpression* CETensor::GetNDimExpr(int i) {
	CExpression* res=NULL;
	CTIndex* ti=GetTIndex(i); if(ti!=NULL){ res=ti->GetNDimExpr(); }
	if(res==NULL){ CEException::halt("CETensor::GetNDimExpr == NULL"); res=new CENull(); }
	return res;
}

CExpression* CETensor::m_refSymDimensionDelta=NULL; // to cache getDimensionDelta() calls
CExpression* CETensor::getDimensionDelta() { // returns not-copied expr or NULL for 4-dim space
	CExpression* res = NULL;
	if (CETensor::isNDimensionSpace()) {
		if (m_refSymDimensionDelta==NULL) {
			m_refSymDimensionDelta=CEProcessor::getDataStorage()->gete(CConst::GL_ETA);
			if (m_refSymDimensionDelta==NULL) { m_refSymDimensionDelta=new CESymbol("DimensionDelta"); CEProcessor::AddToDeleteList(m_refSymDimensionDelta); }
		}
		res=m_refSymDimensionDelta;
	}
	return res;
}
