////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EVector.cpp: implementation of the CEVector class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

const char* CEVector::CLASS_NAME	= "CEVector";

CEVector::CEVector(const CExpression* base, const CExpression* index, bool covariance):CETensor(base)
{
	//m_id = NULL;
	m_indices->add(CTIndex::CreateVectorIndex(index, covariance));
}

CEVector::~CEVector()
{
	//CMemoryManager::mfree(m_id);
}

CEString* CEVector::toString(int mode)
{
	CEString* res = NULL;
	CTIndex* tindex = GetTIndex(1); if(tindex==NULL)CEException::halt();
	switch (mode)
	{
	case STORE_WR_MATHEMATICA:
		//Note: see '\(..\)' brackets politics in InternalNotes[1.1].
		if(IsSimpleItem(m_base)){ //additional '\(..\)'-brackets are required for simple items
			res=new CEString((getFlag()?"\\(":"- \\(")); res->AddString(m_base->toString(mode)); res->AddString("\\)");
		}else{ //branch is similar to 'StoreInBrackets()'
			res=new CEString((getFlag()?"\\((":"- \\((")); res->AddString(m_base->toString(mode)); res->AddString(")\\)");
		}

		res->AddString((tindex->covariance())?"\\_":"\\&");
		if (tindex->index()->getFlag()) res->AddString(tindex->index()->toString(mode));
		else {
			dbg->log(0,"<Warning><Negative vector index>");
			*res<<"\\("; res->AddString(tindex->index()->toString(mode)); *res<<"\\)";
		}		
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

const char* CEVector::ClassName() const
{
	return CLASS_NAME;
}

bool CEVector::covariance() const { return CETensor::covariance(1); }
CExpression* CEVector::index() { return CETensor::index(1); }

CExpression* CEVector::clone() const {
	CExpression* res = new CEVector(GetBase(), GetIndex(1), covariance());
	copyAttributesTo(res);
	return res;
}

CEVector* CEVector::CreateUnique(CExpression* ind, bool cov, bool commutative) {
	if (ind == NULL) return NULL;
	CESymbol* tmp_base = CESymbol::getUnique(commutative);
	CEVector* res = new CEVector(tmp_base, ind, cov);
	delete tmp_base;
	return res;
}

CExpression* CEVector::unite_with(IETensor *other, bool other_is_at_right, bool bCalledByOther)
{
	if (other == NULL) { dbg->log(0, "<Warning><CEVector::unite_with : NULL parameter.>"); return NULL; }

	// do not unite uncommutatives
	if (!getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) {
		// the tensor base is probably uncommutative
		return NULL;
	}


	CExpression* res = NULL;
	CExpression* o_e = other->ThisExpr();
	if (other->rank() == 1) {
		if(!CTIndex::CheckCoupled(GetTIndex(1),other->GetTIndex(1)))return NULL;

		if (o_e->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) {
			const char* o_cn = o_e->ClassName();
			if (o_cn == CEVector::CLASS_NAME) {
				if (base()->equals(other->base()))
					res = square(base());
				else
					res = new CESProduct(base(), other->base());
				if (!getFlag()) res->inverseFlag();
				if (!o_e->getFlag()) res->inverseFlag();
			} else if ((o_cn == CEFuncBondarev::CLASS_NAME) && (!bCalledByOther)) {
				res = other->unite_with(this, !other_is_at_right, true);
			}
		}
		else if(!bCalledByOther) { res=other->unite_with(this, !other_is_at_right, true); }
	}
	else if(!bCalledByOther) { res=other->unite_with(this, !other_is_at_right, true); }
	return res;
}

CExpression* CEVector::square(CExpression *base) {
	if (base == NULL) { CEException::halt(dbg, "<Error><CEVector::square() : NULL data>"); }

	// NOTE: base may have sign
	CExpression* res = get_predefined_square(base);
	if (res == NULL) res = new CESProduct(base, base);
	return res;
}

CExpression* CEVector::module(CExpression *base) {
	if (!base->getFlag()) dbg->log(0, "CEVector: module() for signed value WARNING");
	return CEProcessor::kinematics()->get_mass(base);
}

char* CEVector::getVectorDataId(CExpression *base) {
	char* res = base->toLine();
	return res;
}

char* CEVector::id() { return base()->toLine(); }

CExpression* CEVector::Arrange(int level) {
	if (!m_base->getFlag()) {
		m_base->inverseFlag();
		inverseFlag();
	}
	if (level & AL_EXTRACT) {
		if (CEBasic::isZero(m_base)) {
			CExpression* res = CENumber::GetZero();
			copyAttributesTo(res, true);
			return res;
		}
	}
	return this;
}

CExpression* CEVector::expand(bool recursive) {
	Arrange(AL_INLINE);

	//TODO: move it to CETensor::expand()?
	CExpression* b = base();
	if (b->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* _base = (CESum*)b;
		bool base_sign = _base->getFlag();
		CESum* res = new CESum();
		CEVector* temp = NULL;
		long s = _base->size();
		long i;
		for(i=0; i<s; i++) {
			temp = new CEVector(_base->GetAt(i), index(), covariance());
			if (!base_sign) temp->inverseFlag();
			res->addItem(temp);
		}
		copyAttributesTo(res, true);
		return res;
	} else if (CEBasic::isZero(b)) {
		CExpression* res = CENumber::GetZero();
		copyAttributesTo(res, true);
		return res;
	}
	return this;
}

bool CEVector::less_than(CExpression *other) {
	if (other == NULL) return false;

	// compare to CESymbol, CEPower, CESProduct, CEVector, CEMetric, CEAntisymmetric, CEAntiSymmProd, CEFuncBondarev
	// a < a^n < a.b < a_nu < a_mu_nu < a_abcd
	const char* o_cn = other->ClassName();
	if (o_cn == CEPower::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CEPower* o = (CEPower*)other;
		if (base()->less_than(o->basis())) return true;
		return false;
	}

	if (o_cn == CESymbol::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		if (base()->less_than(other)) return true;
		return false;
	}

	if (o_cn == CESProduct::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		CESProduct* o = (CESProduct*)other;
		if (base()->less_than(o->get_a())) return true;
		return false;
	}

	if ((o_cn==CEVector::CLASS_NAME)||(o_cn==CEMetric::CLASS_NAME)||(o_cn==CEAntisymmetric::CLASS_NAME)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return TLessThen(IETensor::GetIETensor(other), o_cn);
	}

	if (o_cn==CEAntiSymmProd::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	if (o_cn==CEFuncBondarev::CLASS_NAME) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		return true;
	}

	return CExpression::less_than(other);
}

const char* CEVector::CLASS_TAG_NAME = "vec";
const char* CEVector::get_TAG_Name() { return CLASS_TAG_NAME; }

long CEVector::store_stream(ofstream* p_ofs, DWORD settings) {
	long res = CESupport::store_stream(base(), p_ofs, settings);
	res += CESupport::store_stream(index(), p_ofs, settings);
	CENumber cov_no(covariance()?1:0);
	res += CESupport::store_stream(&cov_no, p_ofs, settings);
	return res;
}

CESerializable* CEVector::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	long skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEVector"); return NULL; }
	CESerializable* item_1 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;
	
	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEVector"); delete item_1; return NULL; }
	CESerializable* item_2 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { dbg->log(0, "Error restoring CEVector"); delete item_1; delete item_2; return NULL; }
	CESerializable* item_3 = CESupport::read_buffer(data, pos, settings);
	pos = skip_pos;

	// no mere data expected
	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos > 0) { dbg->log(0, "Error restoring CEVector"); delete item_1; delete item_2; delete item_3; return NULL; }

	if ( (!CEBasic::isExpression(item_1)) || (!CEBasic::isExpression(item_2)) ||
		(!CEUtil::equals(item_3->ClassName(), CENumber::CLASS_NAME)) ) {
		dbg->log(0, "Error restoring CEVector: invalid data loaded"); delete item_1; delete item_2; delete item_3; return NULL;
	}
	int cov = ((CENumber*)item_3)->getLong();
	CEVector* res = new CEVector((CExpression*)item_1, (CExpression*)item_2, (cov != 0));
	delete (CExpression*)item_1; delete (CExpression*)item_2; delete (CENumber*)item_3;
	return res;
}

/*********************************************
***		Vector squares cache
*********************************************/
CECache* CEVector::m_VecSqrCache=NULL;
void CEVector::InitVecSquareCache() { // init cache, to be cleared at kinematics reinstall
	if (m_VecSqrCache == NULL) {
		m_VecSqrCache = new CECache(false);
		CEProcessor::AddToDeleteList(CEVector::m_VecSqrCache);
	} else { // reinit
		m_VecSqrCache->clear_all();
	}
}
void CEVector::ClearVecSquareCache() { if(m_VecSqrCache!=NULL)m_VecSqrCache->clear_all(); }
void CEVector::AddToVecSquareCache(CExpression *base, CExpression *val) { if(m_VecSqrCache==NULL)InitVecSquareCache(); m_VecSqrCache->put_value(base,val); }
void CEVector::RemoveVecSquareCached(CExpression *base) { if(m_VecSqrCache!=NULL)m_VecSqrCache->remove_val(base); }
CExpression* CEVector::GetVecSquareCached(CExpression *base) { if(m_VecSqrCache==NULL)InitVecSquareCache(); return m_VecSqrCache->get_val(base); }

// main entry
CExpression* CEVector::get_predefined_square(CExpression *base) { // returns cloned value or NULL
	if (base == NULL) { CEException::halt(dbg, "<Error><CEVector::get_predefined_square() : NULL data>"); }
	if (base->getProperty(CEProperties::PROP_MASK_IS_SUM)) return NULL;
	if (!base->getFlag()) {
		base->inverseFlag();
		CExpression* res = get_predefined_square(base);
		base->inverseFlag();
		return res;
	}
	if(m_VecSqrCache==NULL)InitVecSquareCache();
	// search in cache first
	CExpression* res = CEVector::GetVecSquareCached(base);
	if (res != NULL) return res->clone();

	// search in current kinematics
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) return NULL;
	res = kin->GetVectorSquare(base);

	//add to cache
	if (res != NULL) CEVector::AddToVecSquareCache(base, res);
	return res;
}
