////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// TIndex.cpp: implementation of the CTIndex class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTIndex::CTIndex(int nDim, bool bMinkowskySpace, const CExpression* index, bool covariance, bool bUseCovariance)
: CEObject(true) {
	m_index=index->clone();
	m_covariance=covariance; m_bUseCovariance=bUseCovariance; if(!m_bUseCovariance)m_covariance=true;
	m_nDim=nDim; m_bMinkowskySpace=bMinkowskySpace;
}

CTIndex::~CTIndex() { delete m_index; }

bool CTIndex::setIndex(CExpression *index)
{
	if (m_index != index) {
		delete m_index;
		m_index = index->clone();
		return true;
	}
	return false;
}

bool CTIndex::liftIndex(bool to_covariance) {
	if(m_bUseCovariance) {
		if(m_covariance != to_covariance) { m_covariance=to_covariance; return true; }
		return false;
	}
	return true;
}

bool CTIndex::changeIndex(CExpression *index, bool covariance)
{
	bool res=setIndex(index);
	if(liftIndex(covariance)) res=true;
	return res;
}

CTIndex* CTIndex::clone() const { return new CTIndex(m_nDim, m_bMinkowskySpace, m_index, m_covariance, m_bUseCovariance); }

const char* CTIndex::CLASS_NAME = "CTIndex";
const char* CTIndex::ClassName() const { return CLASS_NAME; }
bool CTIndex::contain_vector(void * obj)
{
	if ((m_index == obj) || (m_index->contain_vector(obj)))
		return true;
	return false;
}

char* CTIndex::toLine(int mode) {
	char* res = NULL;
	char* res_tmp = NULL;
	if (m_covariance) res = CEUtil::copyString("<+");
	else res = CEUtil::copyString("<-");

	bool no_data = true;
	if (m_index != NULL) {
		CEString* ind_str = m_index->toString(mode);
		if (ind_str != NULL) {
			char* ind_line = ind_str->getData();
			if (ind_line != NULL) {
				res_tmp = CEUtil::concat(res, ind_line); CMemoryManager::mfree(res); res = res_tmp;
				no_data = false;
			}
			delete ind_str;
		}
	}

	if (no_data) {
		res_tmp = CEUtil::concat(res, "NULL"); CMemoryManager::mfree(res); res = res_tmp;
	}
	
	res_tmp = CEUtil::concat(res, ">"); CMemoryManager::mfree(res); res = res_tmp;
	return res;
}

bool CTIndex::equalsIndex(CTIndex* other) {
	if (other == NULL) return false;
	if (covariance() != other->covariance()) return false;
	CExpression* ind1 = index();
	CExpression* ind2 = other->index();
	if (ind1 == ind2) return true;
	if ((ind1 == NULL) || (ind2 == NULL)) return false;
	if (!ind1->equals(ind2)) return false;
	if (GetNDimID() != other->GetNDimID()) return false;
	if (!CEUtil::LEQU(IsMinkowskySpace(),other->IsMinkowskySpace())) return false;
	return true;
}

CExpression* CTIndex::GetNDimExpr() { // returns copied expr
	int nDim=GetNDimID();
	CExpression* res;
	if(nDim == CETensor_NDIM_4orN) {
		res = CENumber::GetFour();
		if (CETensor::isNDimensionSpace()) {
			CExpression* del=CETensor::getDimensionDelta(); if(del==NULL)CEException::halt(CETensor::dbg,"NULL dim-delta");
			res=res->add(del->clone());
		}
	} else {
		res = CENumber::GetNumber(nDim);
	}
	return res;
}

//[main check] if 'index1' is coupled to 'index2'
bool CTIndex::CheckCoupled(CTIndex* index1, CTIndex* index2) {
	if ((index1==NULL) || (index2==NULL))return false;
	CExpression* eIndex1=index1->index();
	if ((eIndex1==NULL) || (!eIndex1->equals(index2->index())))return false;
	if ((index1->GetNDimID() != index2->GetNDimID()) ||
		(!CEUtil::LEQU(index1->IsMinkowskySpace(),index2->IsMinkowskySpace()))) return false;

	// Compare 'covariance' flag if required
	if(index1->IsCovarianceUsed()){
		if(index1->covariance() == index2->covariance()){ //fast & unsafe check
			CETensor::dbg->log(0, "<Warning in CTIndex::isCoupledTo><covariance(1)==covariance(2) for similar indices>",index1,index2);
			return false;
		}
	}
	return true;
}
