////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EGellMann.h: interface for the CEGellMann and related classes.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EGellMann_H__INCLUDED_)
#define AFX_EGellMann_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class CEGellMann : public CETensor  
{
public:
	static CExpression* m_GellMannBase;
	static const CExpression* GetGellMannBase();

public:
	CEGellMann(const CExpression* _index);
	virtual ~CEGellMann();
	static const char* CLASS_NAME;
	virtual const char* ClassName() const;
	//equals, less_than, toString: used from CETensor
	virtual CExpression* clone() const;

	virtual CExpression* expand(bool recursive = true);
	virtual CExpression* Arrange(int level);
	virtual CExpression* unite_with(IETensor* other, bool other_is_at_right=true, bool bCalledByOther=false);
	virtual CExpression* exchange_subst(IETensor* other, bool other_is_at_right, bool bCalledByOther=false);

public:
	virtual unsigned long props_long();
protected:
	virtual void collectProperties(CEProperties* props, unsigned set = CEProperties::PROP_MASK_ALL);
};

///////////////////////////////////////////////////// CEGellMannH: h^{abc} = d^{abc} + i f^{abc}

class CEGellMannH : public CETensor  
{
public:
	static CExpression* m_GellMannHBase;
	static const CExpression* GetGellMannHBase();

public:
	CEGellMannH(const CExpression* index1, const CExpression* index2, const CExpression* index3);
	virtual ~CEGellMannH();
	static const char* CLASS_NAME;
	virtual const char* ClassName() const;
	//equals, less_than, toString: used from CETensor
	virtual CExpression* clone() const;
public:
	virtual unsigned long props_long();
protected:
	virtual void collectProperties(CEProperties* props, unsigned set = CEProperties::PROP_MASK_ALL);
public:
	virtual CExpression* Arrange(int level);
	virtual CExpression* unite_with(IETensor* other, bool other_is_at_right=true, bool bCalledByOther=false);

private:
	void RotateIndices(int nTimes); //nTimes: 0,1,2, abc -> cab -> bca
};

///////////////////////////////////////////////////// CEGellMannTr: Tr[t^a t^b ...]

class CEGellMannTr : public CEAbstractFunction
{
public:
	//
	//	static entries
	//
	//	check if the object is G-M operator (contains t^a-matrices)
	static bool IsGellMannOperator(CExpression* src);
	//
	//	extract & evaluate trace
	//	Nt: Trace[non-GM-objects] = non-GM-objects*Tr[1] = non-GM-objects*3
	//	returns cloned result, 'src' is unaffected if !do_delete_src
	static CExpression* TraceCalc(CExpression* src, bool do_delete_src=true);
	//
	//	extract & factorize trace, don't evaluate
	//	returns cloned result, 'src' is unaffected if !do_delete_src
	static CExpression* TraceFactorize(CExpression* src, bool do_expand_src=true, bool do_delete_src=true);

public:
	CEGellMannTr(const CExpression* src);
	virtual ~CEGellMannTr();
	static CDebug* dbg;
	static const char* CLASS_NAME;
	virtual const char* ClassName() const;
	virtual CExpression* clone() const;
	virtual unsigned long props_long();
protected:
	virtual void collectProperties(CEProperties* props, unsigned set = CEProperties::PROP_MASK_ALL);
public:
	virtual CExpression* Arrange(int level);			// To extract Tr[0]->0
	//equals, less_than, toString						// used from CEAbstractFunction
	//expand											// not used
	virtual CExpression* evaluate(int level = 0);		// evaluate trace

private:
	bool IsNormalForm();								// normal form is Tr[t^a t^b ... t^m], no other objects
	CExpression* ToNormalForm(bool bExpandSrc);			// to Sum{ const*Tr[t^a t^b ... t^m] }, returns this or cloned, NULL if failed
	static CExpression* CalcTrace(CEProduct* pr);		// 'pr' is 't^a t^b ... t^m', unaffected
	static CExpression* CalcTraceImpl(CEProduct* pr);	// 'pr' is {t^a t^b ... t^m'} or {}, unaffected

	////////////////
public:
	//static CExpression* Spur(CExpression* val, bool do_delete = true);

	// Tr[] operations:
	//bool IsOfNormalForm(); // Tr[... G..G-hat.G...G5/P/M], G5/P/M is last, thr only G and G-hat are other non-commutatives
	//CExpression* ToNormalForm(); // this, cloned result(SUM of Tr) or NULL(if failed)
	//int CalcGammaNum(); // Number of Gamma/G-hat objects (but no G5/+/-), -1 if unexpected object spoted (or g-5 at wrong place)
	//bool SwapToGamma(int i); // Makes Gamma-i first in Gamma-sequence, false if failed (unexpected non-commutative object spotted)
	//int FindGammaCoupleForIndex(CExpression* ind, bool cov); // Logical no is returned (0..CalcGammaNum()-1)
	//bool CoupleGammaToMetric(CEMetric* metr); // metr is not affected, try every gamma
	//CEPtrArray* GetTIndicesList(); //Array of CTIndex*/NULL objects of Gamma-indices (NULL for Gamma-X or coupled internally), to be deleted by user
	//int GetLastGammaXID(); // 0/1/-1 ot INT_MIN
	//CExpression* ExpandLastGammaX(bool toGpm); // this or sum(clone(G+)+clone(G-)) or ZERO(for 1-sized source)
	//CExpression* RetrieveSrcSeq(bool _StraightOrder, bool _NoLastGX, bool _NoFirstG); // cloned product w/o(or with) last gamma-5 and first Gamma-i
	//static CExpression* TruncateZeroTraces(CExpression* src, bool del_src=true);
	//static CExpression* TruncateZeroTracesRuleProc(CExpression* src, CEPtrArray* data_arr);
private:
	//CExpression* SpCommon(CEProduct* src);

	//CExpression* Sp(CEProduct* seq);
	//CExpression* Sp(CEPtrArray* data);
	//static CExpression* Sp_calculation(CEPtrArray *data, bool do_fill_cache = false);
};

#endif // !defined(AFX_EGellMann_H__INCLUDED_)
