////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EPtrArray.cpp: implementation of the CEPtrArray class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#define ALL_SIZE_DEFAULT	8
#define ALL_INCR_DEFAULT	8
const char* CEPtrArray::CLASS_NAME = "CEPtrArray";
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEPtrArray::CEPtrArray(bool allow_NULLs, long sz_incr) : CEObject(false)
{
	if (sz_incr<0) sz_incr = ALL_INCR_DEFAULT;
	m_allow_NULLs = allow_NULLs;
	m_iterator = 0;

	m_size = 0;
	m_allocated_size = sz_incr;
	m_data = (CEObject**)CMemoryManager::mmalloc(m_allocated_size*sizeof(CEObject*));
	m_alloc_increment = sz_incr;
}

// used for cloning
CEPtrArray::CEPtrArray(long _size, long _allocate_size, long _alloc_increment,
		   long _iterator, CEObject** _buf, bool allow_NULLs) : CEObject(false)
{
	m_allow_NULLs = allow_NULLs;
	m_iterator = _iterator;

	m_size = _size;
	m_allocated_size = _allocate_size;
	m_data = (CEObject**)CMemoryManager::mmalloc(m_allocated_size*sizeof(CEObject*));
	m_alloc_increment = _alloc_increment;

	if (_buf != NULL) {
		// copy objects
		long i;
		for (i=0; i<_size; i++) {
			CEObject* item = _buf[i];
			if (item != NULL) item = item->copy();
			m_data[i] = item;
		}
	}
}

CEPtrArray::~CEPtrArray()
{
	long s = m_size;
	long i;
	for (i=0; i<s; i++) {
		CEObject* obj = m_data[i];
		if (obj != NULL) delete obj;
	}
	CMemoryManager::mfree(m_data);
}
void CEPtrArray::SetAllowNULLs(bool allow_NULLs) {
	m_allow_NULLs = allow_NULLs;
	if(!m_allow_NULLs){ //check previous values
		for (long i=0; i<m_size; i++) {
			if(m_data[i]==NULL){ CEException::halt("CEPtrArray::SetAllowNULLs(false) for NULL-containing array"); }
		}
	}
}

const char* CEPtrArray::ClassName() const {return CLASS_NAME;}
CEObject* CEPtrArray::copy() const { return clone(); }
bool CEPtrArray::obj_equals(CEObject* other) {
	if (other == NULL) return false;
	if (other == this) return true;
	if (!CEUtil::equals(ClassName(), other->ClassName())) return false;
	CEPtrArray* o_arr = (CEPtrArray*)other;

	// compare arrays by items
	long s = size();
	if (o_arr->size() != s) return false;
	long i;
	for (i=0; i<s; i++) {
		CEObject* item_1 = GetAt(i);
		CEObject* item_2 = o_arr->GetAt(i);
		if (item_1 != item_2) {
			if (item_1 == NULL) return false; //(cause item_2 != NULL)
			if (!item_1->obj_equals(item_2)) return false;
		}
	}
	return true;
}

char* CEPtrArray::toLine(int mode) {
	char* res = NULL;
	long s = m_size;
	res = CEUtil::copyString("{");
	long i;
	for (i=0; i<s; i++) {
		CEObject* p = m_data[i];
		if (p != NULL) {
			char* str = p->toLine(mode);
			char* res_new = CEUtil::concat(res, str);
			CMemoryManager::mfree(str);
			CMemoryManager::mfree(res);
			res = res_new;
		}
		else {
			char* res_new = CEUtil::concat(res, "NULL");
			CMemoryManager::mfree(res);
			res = res_new;
		}

		if (i<(s-1)) {
			char* res_new = CEUtil::concat(res, "; ");
			CMemoryManager::mfree(res);
			res = res_new;
		}
	}

	char* res_new = CEUtil::concat(res, "}");
	CMemoryManager::mfree(res);
	res = res_new;
	return res;
}

bool CEPtrArray::contain_vector(void * obj) {
	long s = size();

	//direct search
	long i;
	for (i=0; i<s; i++) {
		CEObject* o = m_data[i];
		if (o == obj) return true;
	}

	// scan sub-items
	for (i=0; i<s; i++) {
		CEObject* o = m_data[i];
		if (o != NULL) {if (o->contain_vector(obj)) return true;}
	}
	return false;
}

void CEPtrArray::incAllocatedSize() {
	// update increment
	if (m_allocated_size > (m_alloc_increment<<2)) {
		long old_val = m_alloc_increment;
		m_alloc_increment *=2; // may overflow long value
		if (m_alloc_increment>m_allocated_size) m_alloc_increment=old_val;
		if (m_alloc_increment < ALL_INCR_DEFAULT) m_alloc_increment=ALL_INCR_DEFAULT;
	}

	m_allocated_size += m_alloc_increment;
	m_data = (CEObject**)CMemoryManager::mrealloc(m_data, m_allocated_size*sizeof(CEObject*));
}

void CEPtrArray::decAllocatedSize() {
	// decrease size up to m_size

	if (m_size < m_allocated_size) {
		m_allocated_size = m_size;
		if (m_allocated_size < ALL_SIZE_DEFAULT) m_allocated_size = ALL_SIZE_DEFAULT;
		m_data = (CEObject**)CMemoryManager::mrealloc(m_data, m_allocated_size*sizeof(CEObject*));
		m_alloc_increment = m_allocated_size>>2;
		if (m_alloc_increment < ALL_INCR_DEFAULT) m_alloc_increment = ALL_INCR_DEFAULT;
	}
}

bool CEPtrArray::checkDecSizeProfit() {
	return (((m_size<<1) + m_alloc_increment + ALL_SIZE_DEFAULT) < m_allocated_size);
}

//////////////////////////////////////////////////////////////////////
// Basic operations
//////////////////////////////////////////////////////////////////////
long CEPtrArray::size() const { return m_size; }
CEObject* CEPtrArray::GetAt(long i) { if(i<m_size)return m_data[i]; return NULL; }
CEObject* CEPtrArray::GetAt2(long i) { if(i<m_size)return m_data[i]; return NULL; }
//CEObject* CEPtrArray::GetCopyAt(long i) const { return (((i<m_size) && (m_data[i]!=NULL))? (m_data[i]->copy()) : NULL); }
const CEObject* CEPtrArray::GetAtConst(long i) const { if(i<m_size)return m_data[i]; return NULL; }

CEObject* CEPtrArray::get(long i) {
	if ((i>=0) && (i < m_size)) return m_data[i];
	return NULL;
}

void CEPtrArray::add(CEObject *obj) {
	if (m_size >= m_allocated_size) incAllocatedSize();
	m_data[m_size] = obj;
	m_size++;

	if ((!m_allow_NULLs) && (obj == NULL)) { CEException::halt("CEPtrArray::add(NULL) forbidden call"); }
}

void CEPtrArray::InsertAt(long i, CEObject *obj) {
	if ((i<0) || (i>m_size)) { CEBasic::dbg->log(0, i, "CEPtrArray: InsertAt wrong pos: %ld", this); return; }
	if (i == m_size) { add(obj); return; }

	if (m_size >= m_allocated_size) incAllocatedSize();

	if (i<m_size) CMemoryManager::mmemmove(m_data+(i+1), //*sizeof(CEObject*),
		m_data+i, //*sizeof(CEObject*),
		(m_size-i)*sizeof(CEObject*));
	m_size++;
	m_data[i] = obj;

	if ((!m_allow_NULLs) && (obj == NULL)) {
		printf("CEPtrArray::add(NULL) forbidden call\n"); throw(new CEException(CEException::INT_ERROR));
	}
}

void CEPtrArray::MoveFromTo(long from, long to) {
	if ((from < 0) || (to < 0) || (from >= m_size) || (to >= m_size)) return;// throw()
	if (from == to) return;

	CEObject* obj = m_data[from];
	if (from > to) { // move (to..from-1) right
		CMemoryManager::mmemmove(m_data + (to+1), //*sizeof(CEObject*),
			m_data + to, //*sizeof(CEObject*),
			(from-to)*sizeof(CEObject*));
	} else { // move (from+1..to) left
		CMemoryManager::mmemmove(m_data+from, //*sizeof(CEObject*),
			m_data+(from+1), //*sizeof(CEObject*),
			(to-from)*sizeof(CEObject*));
	}
	m_data[to] = obj;
}

void CEPtrArray::SetAt(long i, CEObject *obj, bool delete_old) {
	if (i<0) return;
	if ((obj == NULL) && (!m_allow_NULLs)) CEException::halt("CEPtrArray::SetAt(NULL): forbidden call");

	if (i>=m_size) { // fill up with NULLs
		if (i==m_size) { add(obj); return; }
		else{
			if (m_allow_NULLs) { long j; for(j=m_size; j<=i; j++) add(NULL); }
			else{ CEBasic::dbg->log(0, "CEPtrArray::SetAt() Warning: invalid index, can't fill with NULLs"); CEException::halt(); }
		}
	}

	if(delete_old){ CEObject* old=m_data[i]; m_data[i]=obj; if((old!=obj)&&(old!=NULL))delete old; }else{ m_data[i]=obj; }
}

void CEPtrArray::RemoveAt(long i, bool delete_old) {
	if ((i<0) || (i>=m_size)) return;
	if (delete_old) { CEObject* old = m_data[i]; if (old != NULL) delete old; }

	m_size--;
	if (i < m_size) { // move {i+1..size} left
		CMemoryManager::mmemmove(m_data+i, //*sizeof(CEObject*),
			m_data+(i+1), //*sizeof(CEObject*),
			(m_size-i)*sizeof(CEObject*));
	}

#ifdef ARR_ADJUST_SIZE // decrease allocated size here
	// FAST:
	// if (((m_size<<1) + m_alloc_increment + ALL_SIZE_DEFAULT) < m_allocated_size) decAllocatedSize();
	// SLOW:
	if (checkDecSizeProfit()) decAllocatedSize();
#endif //ARR_ADJUST_SIZE
}

CEObject* CEPtrArray::ExtractAt(long i) {
	if ((i<0) || (i>=m_size)) return NULL;
	CEObject* res = m_data[i];

	m_size--;
	if (i < m_size) { // move {i+1..size} left
		CMemoryManager::mmemmove(m_data+i, //*sizeof(CEObject*),
			m_data+(i+1), //*sizeof(CEObject*),
			(m_size-i)*sizeof(CEObject*));
	}

#ifdef ARR_ADJUST_SIZE // decrease allocated size here
	// FAST:
	// if (((m_size<<1) + m_alloc_increment + ALL_SIZE_DEFAULT) < m_allocated_size) decAllocatedSize();
	// SLOW:
	if (checkDecSizeProfit()) decAllocatedSize();
#endif //ARR_ADJUST_SIZE
	return res;
}

//////////////////////////////////////////////////////////////////////
// Iterator operations
//////////////////////////////////////////////////////////////////////
bool CEPtrArray::has_next() { return (m_iterator < m_size); }
void CEPtrArray::reset_iterator() { m_iterator = 0; }
CEObject* CEPtrArray::first() { m_iterator = 0; return next(); }
CEObject* CEPtrArray::last() { if (m_size==0) return NULL; return get(m_size-1); }
CEObject* CEPtrArray::next() {
	if (m_iterator < m_size) { CEObject* res = m_data[m_iterator]; m_iterator++; return res; }
	return NULL;
}

//////////////////////////////////////////////////////////////////////
void CEPtrArray::clear_all() {
	long s = m_size;
	long i;
	for (i=0; i<s; i++) {
		CEObject* item = m_data[i];
		if (item != NULL) delete item;
	}
	m_size = 0;

#ifdef ARR_ADJUST_SIZE // decrease allocated size here
	decAllocatedSize();
#endif //ARR_ADJUST_SIZE
}

CEPtrArray* CEPtrArray::clone() const {
	CEPtrArray* res = new CEPtrArray(m_size, m_allocated_size, m_alloc_increment,
		m_iterator, m_data, m_allow_NULLs);
	// Note: object will be copied
	return res;
}

CEPtrArray* CEPtrArray::get_copy(const CEPtrArray * other, bool do_create) {
	if (other != NULL) return other->clone();
	if (do_create) return new CEPtrArray();
	return NULL;
}

void CEPtrArray::add_array(CEPtrArray *other) {
	if (other == NULL) return;
	
	long s = other->size();
	long i;
	for (i=0; i<s; i++) { add(other->GetAt(i)); }
	// TODO: make it faster using CMemoryManager::mmemcpy()
	other->clearall_without_delete_objects();
	delete other;
}

void CEPtrArray::clearall_without_delete_objects() {
	m_size = 0;

#ifdef ARR_ADJUST_SIZE // decrease allocated size here
	decAllocatedSize();
#endif //ARR_ADJUST_SIZE
}
