////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// RMaxPowerToSym.cpp: implementation of the CRMaxPowerToSym class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRMaxPowerToSym::CRMaxPowerToSym(const CExpression* _base) : CERule() {
	if (_base == NULL) { dbg->log(0, "ERROR: new CRMaxPowerToSym(NULL) call"); CEException::halt(); }
	m_base = _base->clone();
	if (!_base->getFlag()) { dbg->log(0, "ERROR CRMaxPowerToSym(signed item)"); CEException::halt(); }
	m_repl = NULL;
	m_mode = 1; // 1: search, 2: replace;
	m_max_power = -1; // no info
}

CRMaxPowerToSym::CRMaxPowerToSym(const CExpression* _base, int _max_power, int _mode, const CExpression* _repl) : CERule() {
	m_base = _base->clone();
	if (!_base->getFlag()) { dbg->log(0, "ERROR CRMaxPowerToSym(2)(signed item)"); CEException::halt(); }
	m_repl = (_repl != NULL)?(_repl->clone()):NULL;
	m_mode = _mode;
	m_max_power = _max_power;
}

CRMaxPowerToSym::~CRMaxPowerToSym() {
	delete m_base;
	if (m_repl != NULL) delete m_repl;
}

const char* CRMaxPowerToSym::CLASS_NAME = "CRMaxPowerToSym";
const char* CRMaxPowerToSym::ClassName() const { return CLASS_NAME; }
CERule* CRMaxPowerToSym::GetRuleCopy() const { return new CRMaxPowerToSym(m_base, m_max_power, m_mode, m_repl); }
int CRMaxPowerToSym::GetMaxPower() { return m_max_power; }

void CRMaxPowerToSym::setReplMode(CExpression* _repl) {
	if (_repl == NULL) { dbg->log(0, "ERROR: CRMaxPowerToSym::setReplMode(NULL) call"); CEException::halt(); }
	if (m_repl != NULL) delete m_repl;
	m_repl = _repl->clone();
	m_mode = 2; // 1: search, 2: replace;
}

bool CRMaxPowerToSym::contain_vector(void * obj) {
	if ((m_base == obj) || (m_repl == obj)) return true;
	if (m_base != NULL) { if (m_base->contain_vector(obj)) return true; }
	if (m_repl != NULL) { if (m_repl->contain_vector(obj)) return true; }
	return false;
}

CEString* CRMaxPowerToSym::toString(int mode) {
	CEString* res  = new CEString("\\({ ");
	res->AddString(m_base->toString(mode));
	res->AddString(", "); res->add_int(m_max_power);
	res->AddString(", "); res->add_int(m_mode);
	if (m_repl != NULL) { res->AddString(", "); res->AddString(m_repl->toString(mode)); }
	res->AddString(" }\\)");
	return res;
}

CExpression* CRMaxPowerToSym::ApplyTo(CExpression* src, bool del_src) {
	if (src == NULL) return NULL;

	// m_mode: 1: search, 2: replace;
	if (m_mode == 1) {
		if (m_max_power == 1) return src;

		// check EQUAL/OPPOSITE/INVERSE
		bool equ = false;
		if (src->getLevel()) {
			if (src->getFlag()) { if (m_base->equals(src)) equ=true; }
			else {
				m_base->inverseFlag();
				if (m_base->equals(src)) equ=true;
				m_base->inverseFlag();
			}
		} else {
			m_base->inverseLevel();
			if (src->getFlag()) { if (m_base->equals(src)) equ=true; }
			else {
				m_base->inverseFlag();
				if (m_base->equals(src)) equ=true;
				m_base->inverseFlag();
			}
			m_base->inverseLevel();
		}
		if (equ) { m_max_power = 1; return src; }

		bool apply_to_subitems = true;
		if (CEUtil::equals(src->ClassName(), CEPower::CLASS_NAME)) {
			CExpression* b = ((CEPower*)src)->basis();
			bool b_equ = false;
			if (b->getFlag()) {
				b_equ = m_base->equals(b);
			} else {
				m_base->inverseFlag(); if (m_base->equals(src)){b_equ=true;} m_base->inverseFlag();
				if (b_equ) {
					if (((CEPower*)src)->isPowerInteger()) { dbg->log(0, "ERROR: CRMaxPowerToSym::ApplyTo(NOT-ARRANGED source), rule skipped"); }
					m_max_power=1; return src;
				}
			}
			if (b_equ) {
				long pow = ((CEPower*)src)->getLongPower();
				if (pow == LONG_MIN) { m_max_power=1; return src; }

				if (m_max_power < 0) { // not defined yet
					m_max_power = pow;
				} else {
					m_max_power = CLMath::greatest_common_divisor(m_max_power, pow);
				}
				apply_to_subitems = false;
			}
		}
		
		if (apply_to_subitems) {
			// Applying to sub-items.
			CExpression* sub_res = src->applyRule(this);
			if (sub_res != src) { dbg->log(0, "CRMaxPowerToSym: internal error."); CEException::halt(); }
		}
		return src;
	}
	
	if (m_mode == 2) {
		if (CEUtil::equals(src->ClassName(), CEPower::CLASS_NAME)) {
			CExpression* b = ((CEPower*)src)->basis();
			bool b_equ = false;
			if (b->getFlag()) {
				b_equ = m_base->equals(b);
			} else {
				m_base->inverseFlag(); if (m_base->equals(src)){b_equ=true;} m_base->inverseFlag();
				if (b_equ) { dbg->log(0, "ERROR: CRMaxPowerToSym::ApplyTo(NOT-ARRANGED source)"); CEException::halt(); }
			}
			if (b_equ) {
				long pow = ((CEPower*)src)->getLongPower();
				if (pow == LONG_MIN) { dbg->log(0, "ERROR: CRMaxPowerToSym::ApplyTo(NON-NUMERIC powers spotted)"); CEException::halt(); }
				long pow_new = pow/m_max_power;
				if (pow_new < 1) { dbg->log(0, "ERROR: CRMaxPowerToSym::ApplyTo() internal ERROR, resulting power < 1"); CEException::halt(); }
				
				CExpression* res = (pow_new > 1)?(new CEPower(m_repl, pow_new)):(m_repl->clone());
				src->copyAttributesTo(res, true);
				return res;
			}
		}
		
		// Applying to sub-items.
		CExpression* res = src->applyRule(this);
		if ((res != src) && (res != NULL)) {
			if (del_src) delete src;
			return res;
		}
		return src;
	}

	dbg->log(0, "CRMaxPowerToSym: internal error, UNKNOWN MODE.");
	return src;
}

/*
*	Note: base_sym is supposed to be a SYMBOL with SIMPLE string (e.g. with no brackets)
*/
CExpression* CRMaxPowerToSym::MaxPowerToSym(CExpression* src, CExpression* base_sym, int* pow_found,
											CEPtrArray* arrNewSyms, // to add replacment symbol at seq[size()] (may be NULL)
											const char* pow_suffix, bool del_src) {
	if ((src == NULL) || (base_sym == NULL)) { dbg->log(0, "CRMaxPowerToSym::MaxPowerToSym(invalid parameters)"); CEException::halt(); }
	if (!base_sym->getFlag()) {
		base_sym = base_sym->clone();
		base_sym->setFlag();
		CExpression* res = MaxPowerToSym(src, base_sym, pow_found, arrNewSyms, pow_suffix, del_src);
		delete base_sym;
		return res;
	}

	if (!base_sym->getLevel()) {
		base_sym = base_sym->clone();
		base_sym->setLevel(true);
		CExpression* res = MaxPowerToSym(src, base_sym, pow_found, arrNewSyms, pow_suffix, del_src);
		delete base_sym;
		return res;
	}

	CRMaxPowerToSym* r = new CRMaxPowerToSym(base_sym);
	src = r->ApplyTo(src); // src may not be changed here

	CExpression* res = NULL;
	int max_pow = r->GetMaxPower();
	*pow_found = max_pow;
	if (max_pow > 1) {
		CEString* repl_str = base_sym->toString(STORE_PLAIN_TEXT);
		if (pow_suffix != NULL) { repl_str->AddString(pow_suffix); }
		repl_str->add_int(max_pow);

		CESymbol repl_sym(repl_str->getData()); delete repl_str;
		if(arrNewSyms!=NULL)arrNewSyms->add(repl_sym.clone());

		r->setReplMode(&repl_sym);
		res = r->ApplyTo(src, del_src);
	} else {
		res = src;
	}

	delete r;
	return src;
}
