////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// FSp.cpp: implementation of the CFSp class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CFSp::dbg = new CDebug("CFSp");
CEPtrArray* CFSp::m_spur_cache = NULL;
int CFSp::m_sc_order_max = 0;

DWORD CFSp::m_Settings = 0;
const DWORD CFSp::SP_CHECK_SUCCESS	= 0x01; // forbid Sp[..] in final result

bool CFSp::m_TracesBondarev	= false;
CExpression* CFSp::m_gP=NULL; // gamma+
CExpression* CFSp::m_gM=NULL; // gamma-
CExpression* CFSp::m_g5=NULL; // gamma5

//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// CSpurCacheCart: utility
//////////////////////////////////////////////////////////////////////
class CSpurCacheCart : public CEObject
{
	int m_rank;
	CExpression* m_spur_val;
	CSequence* m_param_list;
public:
	CSpurCacheCart(CExpression* spur_val, CEPtrArray* param_list);
	~CSpurCacheCart();
	static const char* CLASS_NAME;
	virtual const char* ClassName() const;
	virtual bool contain_vector(void * obj);

	CExpression* get_spur_value(CEPtrArray* param_list);
	int get_rank();
};

/*
*	"param_list" contains CTIndex* objects.
*	Both "spur_val" and "param_list" are cloned.
*/
CSpurCacheCart::CSpurCacheCart(CExpression* spur_val, CEPtrArray* param_list)
: CEObject(true)
{
	CFSp::dbg->log(17, "cache value storing", spur_val);

	m_rank = param_list->size();
	CExpression* sv_cpy = spur_val->clone();
	m_param_list = new CSequence();
	bool do_post_reduce = false;

	CEReplList* rule = new CEReplList();
	int i;
	for(i=0; i<m_rank; i++) {
		CTIndex* ti = (CTIndex*)param_list->GetAt(i);
		CESymbol* par_alt = CESymbol::getUnique();
		CExpression* par = ti->index();
		if (ti->covariance()) {
			rule->add(par, par_alt, false);
		}
		else {
			//	We should put down the index
			sv_cpy = sv_cpy->multiply(new CEMetric(par_alt, par, true, true));
			do_post_reduce = true;
			//	Also we change the flowing index
			//	   ,  ()    (   ?).
			CESymbol* sym_flow_alt = CESymbol::getUnique();
			rule->add(par, sym_flow_alt, false);
			delete sym_flow_alt;
		}
		m_param_list->addItem(par_alt);
	}
	if (do_post_reduce) sv_cpy = CExpression::reduce(sv_cpy);
	m_spur_val = sv_cpy->applyRule(rule);
	if (m_spur_val != sv_cpy) delete sv_cpy;

	CFSp::dbg->log(17, "cache value stored", m_spur_val);
	delete rule;
}

CSpurCacheCart::~CSpurCacheCart()
{
	delete m_spur_val;
	delete m_param_list;
}

const char* CSpurCacheCart::CLASS_NAME = "CSpurCacheCart";
const char* CSpurCacheCart::ClassName() const { return CLASS_NAME; }
bool CSpurCacheCart::contain_vector(void * obj) {
	if ((m_spur_val == obj) || (m_param_list == obj)) return true;
	if (m_spur_val != NULL) { if (m_spur_val->contain_vector(obj)) return true; }
	if (m_param_list != NULL) { if (m_param_list->contain_vector(obj)) return true; }
	return false;
}

/*
*	"param_list" contains CTIndex* objects.
*/
CExpression* CSpurCacheCart::get_spur_value(CEPtrArray* param_list) {
	if (param_list->size() != m_rank) return NULL;

	CExpression* sv_cpy = m_spur_val->clone();
	bool do_post_reduce = false;
	CEReplList* rule = new CEReplList();
	int i;
	for (i=0; i<m_rank; i++) {
		CExpression* par = m_param_list->GetAt(i);
		CTIndex* ti = (CTIndex*)param_list->GetAt(i);
		CExpression* par_req = ti->index();
		if (ti->covariance()) {
			rule->add(par, par_req, false);
		}
		else {
			//	We should raise up the index
			sv_cpy = sv_cpy->multiply(new CEMetric(par_req, par, false, false));
			do_post_reduce = true;
			//	We do not change the internal index, it should escape after reducing.
		}
	}
	if (do_post_reduce) sv_cpy = CExpression::reduce(sv_cpy);
	CExpression* res = sv_cpy->applyRule(rule);
	if (res != sv_cpy) delete sv_cpy;
	delete rule;

	CFSp::dbg->log(27, "cache value retrieved", res);
	return res;
}

int CSpurCacheCart::get_rank() {
	return m_rank;
}
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Static entry
//////////////////////////////////////////////////////////////////////
CExpression* CFSp::Spur(CExpression *val, bool do_delete)
{
	CFSp* f_sp = new CFSp(val);
	if (do_delete) delete val;
	CExpression* res = f_sp->evaluate();
	if (res != f_sp) delete f_sp;

	//reduction & tensor coupling moved into evaluate()
	//res = CExpression::reduce(res);
	return res;
}

//////////////////////////////////////////////////////////////////////
// CFSp: Construction/Destruction
//////////////////////////////////////////////////////////////////////
CFSp::CFSp(const CExpression* data):CEAbstractFunction("Sp", data) {
	if (m_spur_cache == NULL) {
		m_spur_cache = new CEPtrArray();
		CEProcessor::AddToDeleteList(m_spur_cache);
		m_sc_order_max = 0;
	}
}

CFSp::~CFSp() {}

unsigned long CFSp::props_long() {
	unsigned long res = CEAbstractFunction::props_long();
	res=CEProperties::set_property_to(res,CEProperties::PROP_MASK_SELF_CONJUGATED,false);
	return res;
}

void CFSp::collectProperties(CEProperties* props, unsigned set) {
	CEAbstractFunction::collectProperties(props, set);
	props->setProperty(CEProperties::PROP_MASK_SELF_CONJUGATED, false);
}

CExpression* CFSp::clone() const {
	CFSp* res = new CFSp(dataConst());
	copyAttributesTo(res);
	return res;
}

CExpression* CFSp::Arrange(int level) { // To extract Sp[0]->0
	if (level==AL_EXTRACT) {
		CExpression* src = data();
		if ((src!=NULL) && (CEBasic::isZeroProd(src))) {
			CExpression* res=CENumber::GetZero(); copyAttributesTo(res); if(!res->getFlag())res->inverseFlag(); return res;
		}
	}
	return this;
}

const char* CFSp::CLASS_NAME = "CFSp";
const char* CFSp::ClassName() const { return CFSp::CLASS_NAME; }

//////////////////////////////////////////////////////////////////////
// Tr[] operations
//////////////////////////////////////////////////////////////////////
int CFSp::CalcGammaNum() { // Number of Gamma/G-hat objects (but no G5/+/-), -1 if unexpected object spoted
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return -1;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) return 0;
		const char* cn = src->ClassName();
		if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME)) return 1;
		if (cn == CEGammaSpecial::CLASS_NAME) return 0;
		return -1;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	long res=0;
	for (long i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) return -1;
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return -1;
		if (!item->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) {
			const char* cn = item->ClassName();
			if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME)) {
				res++;
			} else {
				if ((cn==CEGammaSpecial::CLASS_NAME) && (i<(s-1))) return -1;
			}
		}
	}
	return res;
}

bool CFSp::SwapToGamma(int i) { // Makes Gamma-i first in Gamma-sequence, false if failed (unexpected non-commutative object spotted)
	if (i<0) return false;
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (i!=0) return false;
		const char* cn = src->ClassName();
		if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME)) return true;
		return false;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size(); if (s==0) return false;
	long i_pos = -1;
	int next_no=0;
	// if (i==0) just look for the first Gamma to check it exists
	long j;
	for (j=0; j<s; j++) {
		CExpression* item = pr->GetAt(j);
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return false;
		const char* cn = item->ClassName();
		if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME)) {
			if (i==next_no) { if (i==0) { return true; } i_pos = j; break; }
			next_no++;
		} else if (cn==CEGammaSpecial::CLASS_NAME) {
			if (j < (s-1)) return false;
		} else if (!item->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) return false;

	}

	// Swap items
	// Extract gX (if exists)
	CExpression* gX = pr->GetAt(s-1);
	if (gX->ClassName() == CEGammaSpecial::CLASS_NAME) {
		gX=gX->clone(); pr->RemoveAt(s-1); s--;
	} else {
		gX=NULL;
	}

	CExpression* part1 = pr->extract_items(0, i_pos-1); if(part1==NULL){CEException::halt("CFSp::SwapToGamma failed (point #4)");}
	pr->addItem(part1);
	if (gX != NULL) pr->addItem(gX);

	CExpression* res = CExpression::reduce(pr, false);
	if (res != pr) {
		dbg->log(0, "(CFSp::SwapToGamma) Internal ERROR: reduced product != self", pr, res);
		delete res;
		return false;
	}
	return true;
}

int CFSp::FindGammaCoupleForIndex(CExpression* ind, bool cov) { // Logical no is returned (0..CalcGammaNum()-1)
	if (ind == NULL) return -1;
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return -1;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->ClassName() != CEGamma::CLASS_NAME) return -1;
		CExpression* ind2 = ((CEGamma*)src)->index();
		if (!ind->equals(ind2)) return -1;
		if (CEUtil::LEQU(cov, ((CEGamma*)src)->covariance())) { dbg->log(0,"SIMILAR COVARIANCE WARNING. FindGammaCoupleForIndex()", this); }
		return 0;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	long res=0;
	for (long i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) return -1;
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return -1;
		if (!item->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) {
			const char* cn = item->ClassName();
			if (cn == CEGamma::CLASS_NAME) {
				CExpression* ind2 = ((CEGamma*)item)->index();
				if (ind->equals(ind2)) {
					if (CEUtil::LEQU(cov, ((CEGamma*)item)->covariance())) { dbg->log(0,"SIMILAR COVARIANCE WARNING. FindGammaCoupleForIndex()", this); }
					return res;
				}
				res++;
			} else if (cn == CESGammaProduct::CLASS_NAME) {
				res++;
			} else {
				if ((cn==CEGammaSpecial::CLASS_NAME) && (i<(s-1))) return -1;
			}
		}
	}
	return -1;
}

bool CFSp::CoupleGammaToMetric(CEMetric* metr) { // metr is not affected, try every gamma
	if (metr == NULL) return false;
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->ClassName() != CEGamma::CLASS_NAME) return false;
		CExpression* sub_res = metr->unite_with((CEGamma*)src);
		if (sub_res == NULL) return false;
		set_param(0, sub_res); delete sub_res; return true;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	for (long i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return false;
		if (item->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) continue;
		const char* cn = item->ClassName();
		if (cn == CEGamma::CLASS_NAME) {
			CExpression* sub_res = metr->unite_with((CEGamma*)item);
			if (sub_res != NULL) { pr->SetAtX(i, sub_res); return true; }
		} else if (cn != CESGammaProduct::CLASS_NAME) {
			if ((cn==CEGammaSpecial::CLASS_NAME) && (i<(s-1))) return false;
		}
	}
	return false;
}

// Array of CTIndex*/NULL objects of Gamma-indices (NULL for Gamma-X or coupled internally), to be deleted by user
CEPtrArray* CFSp::GetTIndicesList() {
	CExpression* src = data();
	CEPtrArray* res = new CEPtrArray(true);
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return res;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		const char* cn = src->ClassName();
		if (cn == CEGamma::CLASS_NAME) { res->add(CTIndex::CreateVectorIndex(((CEGamma*)src)->index(),((CEGamma*)src)->covariance())); return res; }
		if (cn == CESGammaProduct::CLASS_NAME) { res->add(NULL); return res; }
		return res;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	for (long i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		const char* cn = item->ClassName();
		if (cn == CEGamma::CLASS_NAME) {
			res->add(CTIndex::CreateVectorIndex(((CEGamma*)item)->index(),((CEGamma*)item)->covariance()));
		} else if (cn == CESGammaProduct::CLASS_NAME) {
			res->add(NULL);
		}
	}
	return res;
}

bool CFSp::IsOfNormalForm() { // Tr[... G..G-hat.G...G5/P/M], G5/P/M is last, thr only G and G-hat are other non-commutatives
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) return true;
		const char* cn = src->ClassName();
		if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME)  || (cn == CEGammaSpecial::CLASS_NAME)) return true;
		return false;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	if (s==0) return false;
	for (long i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return false;
		if (item->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) continue;
		const char* cn = item->ClassName();
		if ((cn != CEGamma::CLASS_NAME) && (cn != CESGammaProduct::CLASS_NAME)) {
			if ((cn==CEGammaSpecial::CLASS_NAME) && (i<(s-1))) return false;
		}
	}
	return true;
}


CExpression* CFSp::ToNormalForm() { // this, cloned result(SUM of Tr) or NULL(if failed)
	if (IsOfNormalForm()) return this;

	CExpression* src = data();

	// 1. expand source
	src = src->clone();
	//CExpression* src = CExpression::reduce(src);
	// to be expanded in CLGammaSeqSimplify::simplify()

	// 2. simplify gamma sequences
	src = CLGammaSeqSimplify::simplify(src);

	if (!src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		set_param(0,src); delete src;
		if (IsOfNormalForm()) return this;
		dbg->log(0, "Inernal WARNING: CFSp::ToNormalForm() failed", this);
		return NULL;
	}

	// 3. Expand Tr[SUM]
	CESum* src_sum = (CESum*)src;
	long s=src_sum->size();
	CESum* res = new CESum();
	for(long i=0; i<s; i++) {
		CFSp* sp_I = new CFSp(src_sum->GetAt(i));
		if (!sp_I->IsOfNormalForm()) { dbg->log(0, "Inernal WARNING: CFSp::ToNormalForm() failed", this, src); delete sp_I; delete src; delete res; return NULL; }
		res->addItem(sp_I);
	}
	src_sum->copyAttributesTo(res, true); copyAttributesTo(res, true);
	delete src;
	return res;
}

int CFSp::GetLastGammaXID() { // 0/1/-1 ot INT_MIN
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return INT_MIN;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) return INT_MIN;
		const char* cn = src->ClassName();
		if (cn != CEGammaSpecial::CLASS_NAME) return INT_MIN;
		int gVal = ((CEGammaSpecial*)src)->getGammaValue();
		if (gVal>0) return 1; if (gVal<0) return -1; return 0;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	if (s < 1) return INT_MIN;
	CExpression* last = pr->GetAt(s-1);
	if (last->ClassName()!=CEGammaSpecial::CLASS_NAME) return INT_MIN;
	int gVal = ((CEGammaSpecial*)last)->getGammaValue();
	if (gVal>0) return 1; if (gVal<0) return -1; return 0;
}

CExpression* CFSp::ExpandLastGammaX(bool toGpm) { // this or sum(clone(G+)+clone(G-)) or ZERO(for 1-sized source)
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return NULL;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->getProperty(CEProperties::PROP_MASK_COMMUTATIVE)) return NULL;
		const char* cn = src->ClassName();
		if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME) || (cn == CEGammaSpecial::CLASS_NAME)) {
			// Pre-calc 1-sized Trace
			CExpression* res = CENumber::GetZero(); copyAttributesTo(res, true); return res;
		}
		return NULL;
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	if (s < 1) return NULL;
	CExpression* last = pr->GetAt(s-1);
	CExpression* res = NULL; CExpression* tmp=NULL;
	if (last->ClassName() != CEGammaSpecial::CLASS_NAME) {
		tmp = CExpression::reduce(pr->clone()->multiply(new CEGammaSpecial(1)));
		res = new CFSp(tmp); delete tmp;
		tmp = CExpression::reduce(pr->clone()->multiply(new CEGammaSpecial(-1)));
		res = res->add(new CFSp(tmp)); delete tmp;
		copyAttributesTo(res, true); return res;
	}
	int gVal = ((CEGammaSpecial*)last)->getGammaValue();
	if (gVal != 0) return this;

	CEProduct* pr_copy = (CEProduct*)pr->clone();
	pr_copy->SetAtX(s-1,new CEGammaSpecial(1));
	res = new CFSp(pr_copy);
	pr_copy->SetAtX(s-1,new CEGammaSpecial(-1));
	res = res->minus(new CFSp(pr_copy)); delete pr_copy;
	copyAttributesTo(res, true); return res;
}

CExpression* CFSp::RetrieveSrcSeq(bool _StraightOrder, bool _NoLastGX, bool _NoFirstG) { // cloned product w/o(or with) last gamma-5 and first Gamma-i
	CExpression* src = data();
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		const char* cn = src->ClassName();
		if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME)) { if (_NoFirstG) return NULL; }
		if (cn == CEGammaSpecial::CLASS_NAME) { if (_NoLastGX) return NULL; }
		return src->clone();
	}
	CEProduct* pr = (CEProduct*)src;
	long s = pr->size();
	if (s==0) return NULL;
	CEProduct* res = new CEProduct();
	bool _GammaFirstSpotted = false;
	for (long i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		const char* cn = item->ClassName();
		if ((cn == CEGamma::CLASS_NAME) || (cn == CESGammaProduct::CLASS_NAME)) {
			if ((_NoFirstG) && (!_GammaFirstSpotted)) { _GammaFirstSpotted=true; continue; }
		} else if (cn==CEGammaSpecial::CLASS_NAME) {
			if ((_NoLastGX) && (i == (s-1))) continue;
		}
		if (_StraightOrder) { res->addItem(item->clone()); } else { res->InsertAt(0, item); }
	}
	if (res->size() == 0) { delete res; return NULL; }
	if (!src->getFlag()) res->inverseFlag();
	return res;
}

//////////////////////////////////////////////////////////////////////
// CFSp: trace calculation engine
//////////////////////////////////////////////////////////////////////
CExpression* CFSp::evaluate(int level) {
	//
	// 00. sub-evaluate parameter
	if(CEAbstractFunction::evaluate(level) != this){ CEException::halt("IError in CFSp::evaluate"); }

	//dbg->watch_periods(true);
	// 0. init
	if ((m_gP==NULL) || (m_gM==NULL) || (m_g5==NULL)) {
		m_gP = CEProcessor::getDataStorage()->gete(CConst::DV_GAMMA_P);
		m_gM = CEProcessor::getDataStorage()->gete(CConst::DV_GAMMA_M);
		m_g5 = CEProcessor::getDataStorage()->gete(CConst::DV_GAMMA5);
		if ((m_gP==NULL) || (m_gM==NULL) || (m_g5==NULL)) {
			dbg->log(0, "INTERNAL ERROR, DV_GAMMA_P,DV_GAMMA_M or DV_GAMMA5 NULL data");
			m_gP=new CEGammaSpecial(1); CEProcessor::AddToDeleteList(m_gP);
			m_gM=new CEGammaSpecial(-1); CEProcessor::AddToDeleteList(m_gM);
			m_g5=new CEGammaSpecial(0); CEProcessor::AddToDeleteList(m_g5);
		}
	}

	// 1. expand source
	CExpression* _data = CExpression::reduce(data()->clone());
	dbg->log(10, "evaluate(...) :", _data);

	// 2. simplify gamma sequences
	_data = CLGammaSeqSimplify::simplify(_data);
	dbg->log(45, "source simplified", _data);

	CExpression* res = NULL;

	// 3. parse the source
	unsigned long _src_props = _data->props_long();
	if ((_src_props & CEProperties::PROP_MASK_IS_PRODUCT) != 0) {
		res = SpCommon((CEProduct*)_data);
		if (res == NULL) res = this;
	}
	else if ((_src_props & CEProperties::PROP_MASK_IS_SUM) != 0) {
		CESum* sum = (CESum*)_data;
		dbg->log(15, "sum parsing:", sum);
		long size = sum->size();
		long i;
		for (i=0; i<size; i++) {
			CExpression* _func = new CFSp(sum->GetAt(i));
			CExpression* _fres = _func->evaluate(level);
			if (_fres != _func) {
				delete _func;
			} else {
				if (m_Settings & SP_CHECK_SUCCESS) {
					dbg->log(0, "WARNING: SpCommon() evaluation failed", _func);
					dbg->set_level(200);
					CLGammaSeqSimplify::dbg->set_level(200);
					dbg->log(2, "=== Sp()evaluate repeation with high debug level ===");
					delete _func;
					_func = new CFSp(sum->GetAt(i));
					_fres = _func->evaluate(level);
					if (_fres != _func) delete _func;
					CEException::halt();
				}
			}
			sum->SetAtX(i, _fres);
		}
		dbg->log(20, "... sum parsed:", sum);
		res = sum;
	}
	else if ((_src_props & CEProperties::PROP_MASK_IS_TENSOR) != 0) {
		// for example, Sp(p^)
		CEProduct* pr = CEBasic::convertToProduct(_data->clone());
		res = SpCommon(pr);
		if (res != pr) delete pr;
		//dbg->log(0, "evaluate(): Warning: Unexpected tensor case", _data);
		/*CETensor* _t = IETensor::GetIETensor(_data);
		if ((_t->rank() == 1) && (!_props->isCommutative())) res = CENumber::GetZero();*/
	}
	if (res == NULL) {
		// non-tensor object, Sp(C)=4*C, nevermind the space dimension
		res = _data->clone()->multiply(CENumber::GetFour());
	}
	if (!getFlag()) res->inverseFlag();
	if (res != _data) delete _data;

	//dbg->log(7, "evaluate() ...finished", res);
	//dbg->logperiod(0, "spur evaluation");

	// 4. reduce result
	//CEProduct::dbg->set_level(100); CETensor::dbg->set_level(100);
	res = CExpression::reduce(res, (res != this));
	//CEProduct::dbg->set_level(5); CETensor::dbg->set_level(5);
	dbg->log(25, "evaluate()'s result reduced", res);
	//dbg->logperiod(0, "result reduction");
	//dbg->watch_periods(false);

	return res;
}

CExpression* CFSp::SpCommon(CEProduct *seq) {
	if (m_TracesBondarev) return SpBondarev(seq);
	return Sp(seq);
}

// Note: the only Gamma_x may be NON-COMUTATIVE here
// does not deletes source
CExpression* CFSp::Sp(CEProduct *seq) {
	dbg->log(25, "Sp(CEProduct *seq)", seq);

	// Gamma5/+/- objects should be arranged at CLGammaSeqSimplify::simplify() stage
	//CDiracAlgebra::arrangeGammaSpecial(seq); seq->Arrange();

	///////////////////////////////////////////////////////////////////////////////////////
	/// Gamma_{5/+/-} processing
	///////////////////////////////////////////////////////////////////////////////////////
	/// 1. expand Gamma_{+/-} objects if they exists
	if ( (CRContainCheck::search(seq, m_gP)) || (CRContainCheck::search(seq, m_gM)) ) {
		CExpression* src_new = CEGammaSpecial::expandGammaPM(seq->clone());
		src_new = CExpression::reduce(src_new);
		if ( (CRContainCheck::search(src_new, m_gP)) || (CRContainCheck::search(src_new, m_gM)) ) { // Error
			dbg->log(0, "<Error><CFSp::Sp() : Unexpected Gamma_{+/-} found after its' expanding>", src_new);
			dbg->log(0, "<Error><CFSp::Sp()>  source was", seq);
			throw(new CEException(CEException::INT_ERROR));
		}
		
		CExpression* _func = new CFSp(src_new); delete src_new;
		CExpression* _fres = _func->evaluate(); if (_fres != _func) delete _func;
		return _fres;
	}
	
	long s = seq->size();
	long i;
	///////////////////////////////////////////////////////////////////////////////////////
	/// Note: Gamma_{5/+/-} may only ba at s-1 position after reducing
	/// Check it
	///////////////////////////////////////////////////////////////////////////////////////
	for(i=0; i<(s-1); i++) {
		if (CEUtil::equals(seq->GetAt(i)->ClassName(), CEGammaSpecial::CLASS_NAME)) {
			dbg->log(0, "Internal Error: Gamma_5 at unexpected position after simplification", seq); CEException::halt();
		}
	}

	// A. extract all the momenta coupled to gamma matrices
	CDiracAlgebra::expandGammaProducts(seq);
	seq->Arrange(AL_INLINE);
	dbg->log(30, "Sp(CEProduct *seq) after expandGammaProducts", seq);

	long f = seq->getFirstMatrixPos();
	s = seq->size();

	/// 2. reduce gamma_{5} - expressions
	if ((s>0) && (CEUtil::equals(seq->GetAt(s-1)->ClassName(), CEGammaSpecial::CLASS_NAME))) {
		if (f<0) { dbg->log(0, "Error: gamma_5 spotted in commutative expression"); throw(new CEException(CEException::INT_ERROR)); }

		CEGammaSpecial* gX = (CEGammaSpecial*)seq->GetAt(s-1);
		int gX_value = gX->getGammaValue();
		
		if (gX_value == 0) {
			// check result for zero
			if ((s-f) < 5) return CENumber::GetZero();

			// reduce gamma number before Gamma_{5}
			CExpression* g1 = seq->GetAt(s-4);
			CExpression* g2 = seq->GetAt(s-3);
			CExpression* g3 = seq->GetAt(s-2);
			if ((CEUtil::equals(g1->ClassName(), CEGamma::CLASS_NAME))
				&& (CEUtil::equals(g2->ClassName(), CEGamma::CLASS_NAME))
				&& (CEUtil::equals(g3->ClassName(), CEGamma::CLASS_NAME)))
			{
				bool res_sign = gX->getFlag();
				if (!g1->getFlag()) res_sign = !res_sign;
				if (!g2->getFlag()) res_sign = !res_sign;
				if (!g3->getFlag()) res_sign = !res_sign;

				CEProduct* pr_temp = (CEProduct*)seq->clone();
				CExpression* subst = CDiracAlgebra::get_3Gamma_Gamma5_subst(
					((CEGamma*)g1)->index(), ((CEGamma*)g1)->covariance(),
					((CEGamma*)g2)->index(), ((CEGamma*)g2)->covariance(),
					((CEGamma*)g3)->index(), ((CEGamma*)g3)->covariance());
				pr_temp->SetAt(s-4, subst); delete subst;
				pr_temp->RemoveAt(s-1);
				pr_temp->RemoveAt(s-2);
				pr_temp->RemoveAt(s-3);
				dbg->log(25, "pr_temp", pr_temp);
				CExpression* _func = new CFSp(pr_temp);
				CExpression* _fres = _func->evaluate();//(level)
				delete pr_temp;
				if (_fres != _func) delete _func;

				if (!res_sign) {
					dbg->log(0, "Warning: sign found at product item after reducing", seq);
					_fres->inverseFlag();
				}
				return _fres;
			}
			
			dbg->log(0, "<Error><CFSp::Sp() : Unexpected product element after arranging.>");
			throw(new CEException(CEException::INT_ERROR));
		} else {
			dbg->log(0, "<Error><CFSp::Sp() : Unexpected Gamma_{+/-} spotted.>");
			throw(new CEException(CEException::INT_ERROR));
		}
		//return this;
	}
	
	// 3. calculate trace for gamma-sequence
	CEProduct* res = (CEProduct*)seq->clone();
	CEPtrArray* data = new CEPtrArray();
	if (f >= 0) {
		for (i=f; i<s; i++) {
			CExpression* item = seq->GetAt(i);
			if (CEUtil::equals(item->ClassName(), CEGamma::CLASS_NAME)) {
				CEGamma* gamma = (CEGamma*)item;
				data->add(CTIndex::CreateVectorIndex(gamma->index(), gamma->covariance()));
				res->RemoveAt(f);
			}
			else {
				log("<Warning><Unexpected non-commutative item found.>");
				//data->clear();
				delete data;
				delete res;
				return this;
			}
		}
	}
	res->addItem(Sp(data));
	delete data;
	dbg->log(15, "Sp(CEProduct *seq) result:", res);
	return res;
}

CExpression* CFSp::Sp(CEPtrArray *data)
{
	long s = data->size();
	if (s == 0) return CENumber::GetFour(); // Sp(1)=4, nevermind the space dimension
	if (s%2 != 0) return CENumber::GetZero();

	//	get predefined value
	if (s <= m_sc_order_max) {
		long spur_cache_size = m_spur_cache->size();
		long i;
		for (i=0; i<spur_cache_size; i++) {
			CSpurCacheCart* scc = (CSpurCacheCart*)m_spur_cache->GetAt(i);
			if (scc->get_rank() == s) return scc->get_spur_value(data);
		}
		//error.
		dbg->log(0, "Cache error.");
	}

	return Sp_calculation(data, false);
}

/*
*	Wraning: if not all the indices are different do not fill the cache.
*/
CExpression* CFSp::Sp_calculation(CEPtrArray *data, bool do_fill_cache)
{
	long s = data->size();
	if (s == 0) return CENumber::GetFour(); // Sp(1)=4, nevermind the space dimension
	if (s%2 != 0) return CENumber::GetZero();

	//	get predefined value
	if (s <= m_sc_order_max) {
		long spur_cache_size = m_spur_cache->size();
		long sc_counter;
		for (sc_counter=0; sc_counter<spur_cache_size; sc_counter++) {
			CSpurCacheCart* scc = (CSpurCacheCart*)m_spur_cache->GetAt(sc_counter);
			if (scc->get_rank() == s) return scc->get_spur_value(data);
		}
		//error.
		dbg->log(0, "INTERNAL ERROR. Cache error.");
	}

	CESum* res = new CESum();
	CTIndex* ind1 = (CTIndex*)data->GetAt(0);
	bool curr_sign = true;
	long i;
	for (i=1; i<s; i++) {
		CTIndex* ind2 = (CTIndex*)data->GetAt(i);
		CEProduct* prod = new CEProduct();
		prod->addItem(new CEMetric(ind1->index(), ind2->index(),
			ind1->covariance(), ind2->covariance()));
		CEPtrArray* arr = new CEPtrArray();
		long j;
		for (j=1; j<s; j++) {
			if (j != i) arr->add(((CTIndex*)data->GetAt(j))->clone());
		}
		prod->addItem(Sp_calculation(arr, do_fill_cache));
		//arr->clear();
		delete arr;
		if (!curr_sign) prod->unsetFlag();
		res->addItem(prod);
		curr_sign = (!curr_sign);
	}

	//dbg->log(5, "spur (not reduced)", res);
	CExpression* res_reduced = CExpression::reduce(res);
	//dbg->log(5, "spur", res_reduced);

	if (do_fill_cache) {
	//if (s>7) {
		//	store value into cache
		m_spur_cache->add(new CSpurCacheCart(res_reduced, data));
	}

	return res_reduced;
}

// Note: the only Gamma_x may be NON-COMUTATIVE here
// does not deletes source
// source is expeted to bo Num-leveled
CExpression* CFSp::SpBondarev(CExpression* src) {
	dbg->log(25, "SpBondarev(expr)", src);
	if (src==NULL) return NULL;
	if (!src->getLevel()) { dbg->log(0, "INTERNAL ERROR: CFSp::SpX(Denominator item)", src); CEException::halt(); return src; }
	long s;long i;

	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* src_sum=(CESum*)src;
		s=src_sum->size();
		for(i=0;i<s;i++) {
			CExpression* item = src_sum->GetAt(i);
			CExpression* item_res = SpBondarev(item);
			if (item_res != item) { src_sum->SetAtX(i,item_res); }
		}
		return src;
	}

	CEProduct* src_pr=NULL;
	if (src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		src_pr=(CEProduct*)src;
	} else {
		src_pr=CEBasic::convertToProduct(src->clone());
	}
	s=src_pr->size();
	CExpression* res=NULL;

	// drop out commutatives
	long f = src_pr->getFirstMatrixPos();
	if (f<0) {
		if (s==0) {
			dbg->log(0, "(internal error) WARNING: 0-sized product in SpBondarev()", src);
			if(src_pr != src) delete src_pr;
			return CENumber::GetZero();
		}
		res = src_pr;
		if(res == src) res=res->clone();
		res=res->multiply(CENumber::GetFour());
		return res;
	}

	if ((s-f) < 2) {
		if(src_pr != src) delete src_pr;
		return CENumber::GetZero();
	}

	bool CalcTrUsual=false; //Non-Bondarev procedure usage
	int gX_value=0; // Value of the last GammaX object
	bool gX_exist=false;
	///////////////////////////////////////////////////////////////////////////////////////
	/// Note: Gamma_{5/+/-} may only ba at s-1 position after reducing, Check it //TODO: remove next after debug
	for(i=0; i<(s-1); i++) {
		CExpression* item = src_pr->GetAt(i);
		if (item->ClassName()==CEGammaSpecial::CLASS_NAME) { dbg->log(0, "Internal Error in SpBondarev(), wrong source structure", src_pr); CEException::halt(); }
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) { dbg->log(0, "Internal Error in SpBondarev(), wrong source structure", src_pr); CEException::halt(); }
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) { dbg->log(0, "Internal Error in SpBondarev(), wrong source structure", src_pr); CEException::halt(); }
	}

	// Gamma5/+/- objects should be already arranged at CLGammaSeqSimplify::simplify() stage (& be at last pos)
	// if last is gamma5 or non-GammaX uncommutative object --> expand them to G+ and G-
	CExpression* item_last = src_pr->GetAt(s-1);
	CEGammaSpecial* gX = NULL;
	if (item_last->ClassName() == CEGammaSpecial::CLASS_NAME) {
		// GammaX item is last
		gX_exist=true;
		gX = (CEGammaSpecial*)item_last;
		gX_value = gX->getGammaValue();
		if (gX_value == 0) {
			if ((s-f) <= 5) { // gggg{g5}
				// Non-bondarev procedure is better
				// TODO: uncomment next lines after FX_mu*E^mu reduction will be implemented
				//CExpression* res = Sp(src_pr);
				//if (res != src_pr) { if (src_pr != src) delete src_pr; }
				//return res;
			}
			
			// check result for zero
			if ((s-f) < 5) { if (src_pr!=src) delete src_pr; return CENumber::GetZero(); }
				
			CESum* res_sum = new CESum();
			if (src_pr==src) { src_pr=(CEProduct*)src_pr->clone();}
			CEProduct* sub_prod = (CEProduct*)src_pr->clone();
			
			sub_prod->SetAtX(s-1,new CEGammaSpecial(1));
			res_sum->addItem(sub_prod);
			sub_prod=src_pr;
			sub_prod->SetAtX(s-1,new CEGammaSpecial(-1)); sub_prod->inverseFlag();
			res_sum->addItem(sub_prod);
			
			// self-check
			//if (CRContainCheck::search(res_sum, m_g5)) { CEException::halt("self check failed in SpBondarev()"); } //TODO: remove after debug
			res = SpBondarev(res_sum); // sign is already included into res_sum
			if (res != res_sum) delete res_sum;
			return res;
		}
	} else {
		// Non-GammaX item is last
		// self-check
		//if (CRContainCheck::search(res_sum, m_g5)) { CEException::halt("self check failed in SpBondarev()"); } //TODO: remove after debug

		if ((s-f) <= 4) { // gggg
			// Non-bondarev procedure is better
			CalcTrUsual=true;
		} else {
			CESum* res_sum = new CESum();
			if (src_pr==src) { src_pr=(CEProduct*)src_pr->clone();}
			CEProduct* sub_prod = (CEProduct*)src_pr->clone();
			sub_prod->addItem(new CEGammaSpecial(1));
			res_sum->addItem(sub_prod);
			sub_prod = src_pr;
			sub_prod->addItem(new CEGammaSpecial(-1));
			res_sum->addItem(sub_prod);
			
			res = SpBondarev(res_sum); // sign is already included into res_sum
			if (res != res_sum) delete res_sum;
			return res;
		}
	}
	
	// now the last item is : if (!CalcTrUsual) --> (G+ or G-) else (CalcTrUsual)-->(G5 or Non-GX) object
	// We DELETE last gX object now!
	if (!CalcTrUsual) {
		if ((gX_value == 0) || (gX==NULL)) { CEException::halt("SpBondarev() internal ERROR, gX(P,M) is expected"); }

		if ((s-f) == 3) { // (gX-->1/2) (see below) and run Non-bondarev procedure
			CalcTrUsual=true;
			// A. turn CEGammaSpecial --> 1/2
			// gX is now == src_pr->GetAt(s-1);
			CExpression* gX_subst= CENumber::GetHalf(); gX->copyAttributesTo(gX_subst,true);
			src_pr->SetAtX(s-1,gX_subst);
			src_pr->Arrange(AL_INLINE);
		} else {
			// A. delete CEGammaSpecial
			src_pr->RemoveAt(s-1);
		}
	} else {
		if (gX_exist) { CEException::halt("SpBondarev() internal ERROR (2)"); }
	}

	// B. extract all the momenta coupled to gamma matrices
	CDiracAlgebra::expandGammaProducts(src_pr);
	src_pr->Arrange(AL_INLINE);
	dbg->log(30, "SpBondarev() after expandGammaProducts", src_pr);
	f = src_pr->getFirstMatrixPos();
	s = src_pr->size();

	// C. calculate trace for gamma-sequence
	CEPtrArray* data = new CEPtrArray();
	if (f >= 0) {
		for (i=f; i<s;) {
			CExpression* item = src_pr->GetAt(i);
			if (CEUtil::equals(item->ClassName(), CEGamma::CLASS_NAME)) {
				CEGamma* gamma = (CEGamma*)item;
				if (!gamma->getFlag()) { dbg->log(0, "Int. WARNING: non-arranged source",src_pr); src_pr->inverseFlag(); gamma->inverseFlag(); }
				if (!gamma->getLevel()) { dbg->log(0, "Int. ERROR: non-arranged source",src_pr); CEException::halt(); }
				data->add(CTIndex::CreateVectorIndex(gamma->index(), gamma->covariance()));
				src_pr->RemoveAt(i); s--;
			} else {
				dbg->log(0, "Internal Error: Unexpected non-commutative item found.", src_pr, item); CEException::halt();
			}
		}
	}

	CExpression* dataTr = NULL;
	if (CalcTrUsual) {
		dataTr = Sp(data);
	} else {
		dataTr = SpBondarev(data, (gX_value>0));
	}
	src_pr->addItem(dataTr);
	delete data;
	dbg->log(15, "SpBondarev() result:", src_pr);
	return src_pr;
}

CExpression* CFSp::SpBondarev(CEPtrArray* data, bool _gPlus) {
	long s = data->size();
	if (s == 0) return CENumber::GetFour(); // Sp(1)=4, nevermind the space dimension
	if (s%2 != 0) return CENumber::GetZero();

	// s>=2 now
	int d=(_gPlus?4:0);

	CExpression* val1=CENumber::GetUnit();
	CExpression* val2=NULL;
	CExpression* val3=CENumber::GetUnit();
	CExpression* val4=NULL;
	CExpression* tmp=NULL;
	for(long i=0;i<(s-2);i+=2) {
		CTIndex* indi = (CTIndex*)data->GetAt(i);
		CTIndex* indi2 = (CTIndex*)data->GetAt(i+1);

		CExpression* val1_new=NULL;
		if (val1!=NULL) {
			tmp = val1->clone()->multiply(new CEFuncBondarev(1+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			val1_new=tmp;
		}
		if (val4!=NULL) {
			tmp = val4->clone()->multiply(new CEFuncBondarev(4+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			if (val1_new==NULL) val1_new=tmp;
			else val1_new=val1_new->add(tmp);
		}

		CExpression* val2_new=NULL;
		if (val2!=NULL) {
			tmp = val2->clone()->multiply(new CEFuncBondarev(1+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			val2_new=tmp;
		}
		if (val3!=NULL) {
			tmp = val3->clone()->multiply(new CEFuncBondarev(4+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			if (val2_new==NULL) val2_new=tmp;
			else val2_new=val2_new->add(tmp);
		}

		CExpression* val3_new=NULL;
		if (val2!=NULL) {
			tmp = val2->multiply(new CEFuncBondarev(2+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			val3_new=tmp;
		}
		if (val3!=NULL) {
			tmp = val3->multiply(new CEFuncBondarev(3+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			if (val3_new==NULL) val3_new=tmp;
			else val3_new=val3_new->add(tmp);
		}

		CExpression* val4_new=NULL;
		if (val1!=NULL) {
			tmp = val1->multiply(new CEFuncBondarev(2+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			val4_new=tmp;
		}
		if (val4!=NULL) {
			tmp = val4->multiply(new CEFuncBondarev(3+d,NULL,NULL,indi->index(),indi->covariance(),indi2->index(),indi2->covariance()));
			if (val4_new==NULL) val4_new=tmp;
			else val4_new=val4_new->add(tmp);
		}
		val1=val1_new;val2=val2_new;val3=val3_new;val4=val4_new;
	}

	CESum* res = new CESum();
	CTIndex* indL0 = (CTIndex*)data->GetAt(s-2); // last
	CTIndex* indL1 = (CTIndex*)data->GetAt(s-1); // last-1
	if (val1!=NULL) {
		tmp=val1->multiply(new CEFuncBondarev(1+d,NULL,NULL,indL0->index(),indL0->covariance(),indL1->index(),indL1->covariance()));
		res->addItem(tmp);
	}
	if (val2!=NULL) {
		tmp=val2->multiply(new CEFuncBondarev(2+d,NULL,NULL,indL0->index(),indL0->covariance(),indL1->index(),indL1->covariance()));
		res->addItem(tmp);
	}
	if (val3!=NULL) {
		tmp=val3->multiply(new CEFuncBondarev(3+d,NULL,NULL,indL0->index(),indL0->covariance(),indL1->index(),indL1->covariance()));
		res->addItem(tmp);
	}
	if (val4!=NULL) {
		tmp=val4->multiply(new CEFuncBondarev(4+d,NULL,NULL,indL0->index(),indL0->covariance(),indL1->index(),indL1->covariance()));
		res->addItem(tmp);
	}
	
	dbg->log(55, "SpBondarev(not reduced)", res);
	CExpression* res_reduced = CExpression::reduce(res);
	dbg->log(55, "SpBondarev(reduced)", res_reduced);
	return res_reduced;
}

CExpression* CFSp::TruncateZeroTraces(CExpression* src, bool del_src) {
	dbg->log(75, "TruncateZeroTraces()", src);
	CESocketRule r((void*)CFSp::TruncateZeroTracesRuleProc);
	CExpression* res = r.ApplyTo(src, del_src);
	if (res==NULL) res=src;
	dbg->log(75, "TruncateZeroTraces() result", res);
	return res;
}

CExpression* CFSp::TruncateZeroTracesRuleProc(CExpression* src, CEPtrArray* data_arr) {
	if (src == NULL) return src;
	if (src->ClassName() == CFSp::CLASS_NAME) {
		long gNum = ((CFSp*)src)->CalcGammaNum();
		if ((gNum>0) && (gNum%2!=0)) return CENumber::GetZero();
	}
	return src;
}

//////////////////////////////////////////////////////////////////////
// Utilities
//////////////////////////////////////////////////////////////////////
void CFSp::fill_cache_upto(int order)
{
	if (m_spur_cache == NULL) {
		m_spur_cache = new CEPtrArray();
		CEProcessor::AddToDeleteList(m_spur_cache);
		m_sc_order_max = 0;
	}

	if (order%2 != 0) order--;
	if (order <= m_sc_order_max) return;
	
	//	calculate spur for unique indices
	CEPtrArray* ind_alt = new CEPtrArray();
	int i;
	for(i=0; i<order; i++) {
		CESymbol* uniq_sym = CESymbol::getUnique();
		CTIndex* ti = CTIndex::CreateVectorIndex(uniq_sym, true);
		delete uniq_sym;
		ind_alt->add(ti);
	}
	delete Sp_calculation(ind_alt, true);
	delete ind_alt;

	m_sc_order_max = order;
}
