////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// FeynDiagramFactory.cpp: implementation of the CFeynDiagramFactory class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//	flag settings
const int CFeynDiagramFactory::COMPOSE_SETTINGS_BIND_TO_NEXT	= 0x01;
const int CFeynDiagramFactory::COMPOSE_SETTINGS_BIND_TO_FIRST	= 0x02;
const int CFeynDiagramFactory::COMPOSE_SETTINGS_VERTEX_ORDER	= 0x04;
const int CFeynDiagramFactory::COMPOSE_SETTINGS_NO_UNBINDED		= 0x08;

CDebug* CFeynDiagramFactory::dbg = new CDebug("CFeynDiagramFactory");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CFeynDiagramFactory::CFeynDiagramFactory(CDiagramPhysics* _phys, int _order, CKinematics* _kin, int alg_setts)
: CELibrary()
{
	m_phys = _phys;
	m_order = _order;
	m_kin = _kin;
	m_diagrams = NULL;

	m_COMPOSE_SETTINGS = alg_setts;
	if (m_COMPOSE_SETTINGS == 0) {
		// set defaults
		//m_COMPOSE_SETTINGS = COMPOSE_SETTINGS_BIND_TO_NEXT | COMPOSE_SETTINGS_VERTEX_ORDER;
		m_COMPOSE_SETTINGS = COMPOSE_SETTINGS_VERTEX_ORDER | COMPOSE_SETTINGS_BIND_TO_NEXT | COMPOSE_SETTINGS_NO_UNBINDED;
		//m_COMPOSE_SETTINGS = COMPOSE_SETTINGS_BIND_TO_FIRST | COMPOSE_SETTINGS_NO_UNBINDED;
	}
}

CFeynDiagramFactory::~CFeynDiagramFactory() {
	if (m_diagrams != NULL) delete m_diagrams;
}

bool CFeynDiagramFactory::check_setting(int setting) {
	return ((m_COMPOSE_SETTINGS & setting) != 0);
	/*
	// 1. get enumerated settings
	int enum_set = (setting & COMPOSE_SETTINGS_MASK);
	// 2. get flag settings
	int flag_set = (setting & (!COMPOSE_SETTINGS_MASK));
	if (enum_set != 0) {
		// check enumerated settings
		int enum_set_curr = (m_COMPOSE_SETTINGS & COMPOSE_SETTINGS_MASK);
		if (enum_set != enum_set_curr) return false;
	}
	if (flag_set != 0) {
		// check flag settings
		int flag_set_curr = (m_COMPOSE_SETTINGS & (!COMPOSE_SETTINGS_MASK));
		if ((flag_set_curr & flag_set) != flag_set) return false;
	}
	return true;*/
}

CDiagramSet* CFeynDiagramFactory::ComposeAndFilterDiagrams(int order, int alg_setts) {
	CDiagramSet* diags = CFeynDiagramFactory::compose_diagrams(order, alg_setts);
	if (diags != NULL) CLDiagTopologyInspector::FilterDiagrams(diags,
		CLDiagTopologyInspector::NON_EULER_CIRCUIT |
		CLDiagTopologyInspector::LOOPS_IN_EXTERNAL_LINES |
		CLDiagTopologyInspector::ODD_LINEAR_FERMION_LOOP);
	return diags;
}

CDiagramSet* CFeynDiagramFactory::compose_diagrams(int order, int alg_setts) {
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if (phys == NULL) {
		dbg->log(0, "Error: no physics installed");
		return NULL;
	}
	return compose_diagrams(phys, order, alg_setts);
}

CDiagramSet* CFeynDiagramFactory::compose_diagrams(CDiagramPhysics* phys, int order, int alg_setts)
{
	dbg->log(5, order, "compose_diagrams started (order == %ld).");
	if (phys == NULL) return NULL;
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) return NULL;

	CFeynDiagramFactory* factory = new CFeynDiagramFactory(phys, order, kin, alg_setts);
	CDiagramSet* res = factory->compose_diagrams_impl();
	dbg->log(5, res->get_diagrams_number(), "diagrams composed (total %ld diagrams).");

	bool mom_OK = res->ArrangeMomenta();
	dbg->log(20, (mom_OK)?1:0, "Diagrams momenta arranged (OK == %ld).");
	return res;
}

/*
	//	IMPLEMENTATION NOTES.
	//
	//	QUESTION: do create external lines immideatelly?
	//	It will lead to problems with double-external lines
	//	(if the particle do not interact with others)
	//	BUT this case can be added optionally here

	//	1. create external lines
	//	2. loop
	//		1. loop on every vertex type
	//			1. attach new vertex
	//			2. for every tale of new vertex
	//				1. try to bind it to all of the spare lines
	//				2. or create new internal line
	//
	//
	//
*/

CDiagramSet* CFeynDiagramFactory::compose_diagrams_impl() {
	if (m_diagrams != NULL) delete m_diagrams;
	m_diagrams = new CDiagramSet(m_phys);

	///////////
	CFeynDiagram* blanc = CFeynDiagram::create_blanc(m_phys, true, m_kin);
	develop_diagram(blanc);
	delete blanc;
	///////////

	CDiagramSet* res = m_diagrams; m_diagrams = NULL;
	return res;
}

/*
*	diag will not be deleted or affected
*/
void CFeynDiagramFactory::develop_diagram(CFeynDiagram* diag) {
	//if (m_diagrams->get_diagrams_number() > 1) return; // create the first diagrams only (debug)

	int v_num = diag->get_vertices_number();
	dbg->log(25, v_num, "develop_diagram (%ld vertices).");

	// 1. check number of vertices

	// this should not occur: no else vertex should be created if (v_num == m_order) 
	if (v_num > m_order) { dbg->log(0, "Warning: unexpected case during diagram creation"); return; }

	// validate diagram
	if (v_num == m_order) {
		if (diag->is_valid(m_order)) {
			dbg->log(10, "Diagram created:", diag);

			// add diagram copy to result array
			m_diagrams->add_diagram(diag->clone());
			return;
		}
		else { return; }
	}

	// 2. add new vertex to diagram
	CEPtrArray* vi_arr = m_phys->getVerticesInfo();
	int number_of_vertices = vi_arr->size();
	dbg->log(25, number_of_vertices, "number of vertices in physics : %ld.");

	int prev_v_ID = 0;
	if (check_setting(COMPOSE_SETTINGS_VERTEX_ORDER)) {
		// new vertex ID should be >= previous vertex ID
		if (v_num >= 1) {
			prev_v_ID = diag->get_vertex(v_num-1)->get_vertex_info()->get_vertex_ID();
		}
	}

	int vertex_ID;
	for(vertex_ID = prev_v_ID; vertex_ID<number_of_vertices; vertex_ID++) {
		CDiagramVertexInfo* vi = (CDiagramVertexInfo*)vi_arr->GetAt(vertex_ID);
		dbg->log(25, "add vertex of kind: ", vi);

		// can this vertex be binded?
		//if (!can_vertex_be_binded(diag, vi)) continue;

		// bind new vertex to diagram clone
		CDVertex* v = new CDVertex(vi, NULL);

		CFeynDiagram* d_new = diag->clone();
		d_new->add_vertex(v);
		develop_last_vertex(d_new);
		delete d_new;
	}
}

/*
*	diag will not be deleted or affected
*/
void CFeynDiagramFactory::develop_last_vertex(CFeynDiagram* diag)
{
	//if (m_diagrams->get_diagrams_number() > 1) return; // create the first diagrams only (debug)
	dbg->log(25, "develop_last_vertex.", diag);

	// 1. retrieve diagram info
	int v_num = diag->get_vertices_number();
	int l_num = diag->get_lines_number();
	if (v_num == 0) return;// Error

	CDVertex* v = diag->get_vertex(v_num-1);
	CDiagramVertexInfo* vi = v->get_vertex_info();
	int number_of_tales = vi->get_tales_num();

	CEPtrArray* v_links = v->get_links();
	int next_tale_num = v_links->size();

	dbg->log(25, next_tale_num, "Num tales at last vertex: %ld.");

	// 2. check number of bonds at last vertex
	if (next_tale_num >= number_of_tales) { ////////////////// vertex is developed here
		// avoid non-connected loops in diagrams
		if (!v->has_binded_tale_line()) return;

		// vertex developed
		// further diagram develop (validate or add new vertex)
		develop_diagram(diag);
		return;
	}

	//	3. bind vertex to all the spare diagram lines
	//	Note: if vertex is the same kind to previous one,
	//	Start to bind it after the line, previous vertex was started.
	//	(for first vertex' line check only) (!!!)
	//	It is required to avoid similar diagrams to be created.
	bool skip_lines = false;
	CDElement* prev_first_line = NULL;

	if (check_setting(COMPOSE_SETTINGS_BIND_TO_NEXT)) {
		if (next_tale_num == 0) {// for first line only
			// prev vertex
			if (v_num >= 2) {
				CDVertex* v_prev = diag->get_vertex(v_num-2);
				int prev_v_ID = v_prev->get_vertex_info()->get_vertex_ID();
				int curr_v_ID = vi->get_vertex_ID();
				if (prev_v_ID == curr_v_ID) {
					// first line previous line is binded to
					prev_first_line = v_prev->get_link(next_tale_num);
					if (prev_first_line != NULL) {
						skip_lines = true;
					} else {
						// Error
						printf("Error: binding vertex until previous is binded properly.\n");
						throw(new CEException(CEException::INT_ERROR));
					}
					dbg->log(30, "start creating lines after", prev_first_line);
				}
			}
		}
	}

	int pk = vi->get_particle_kind(next_tale_num);

	//	Do bind vertex to only init/final line with this pk to avoid crossings.
	bool linked_to_initial = false;
	bool linked_to_final = false;
	int i;
	for(i=0; i<l_num; i++) {
		CDLine* l = diag->get_line(i);

		//	check up: is it the line to start after?
		if (skip_lines) {
			if (l == prev_first_line) skip_lines = false; // start from the next line.
			continue;
		}

		if (l->is_binded()) continue;

		// comparison of PK
		int line_PK = l->get_particle_kind();
		bool line_is_external = false;
		bool line_is_final = false;
		if (l->is_external()) {
			line_is_external = true;
			if (!(((CDExtLine*)l)->is_initial())) line_is_final = true;
		}
		if (!((line_is_external) && (line_is_final))) line_PK = -line_PK;

		// if particle is neutral boson check up any sign: -photon == photon
		int pk_ABS = CLMath::abs(pk); int line_PK_ABS = CLMath::abs(line_PK);
		if (pk_ABS == line_PK_ABS) {
			if ((pk_ABS & CEParticle::PARTICLE_NEUTRAL_BOSON) == CEParticle::PARTICLE_NEUTRAL_BOSON) {
				// make line_PK same to pk
				line_PK = pk;
			}
		}
		if (line_PK != pk) continue;
		// </comparison of PK> END
		
		if (line_is_external) {
			// avoiding crossings
			if (line_is_final) {
				if (linked_to_final) continue;
				else linked_to_final = true;
			} else {
				if (linked_to_initial) continue;
				else linked_to_initial = true;
			}
		}
		else {
			// check up on self-loops
			if (v == ((CDIntLine*)l)->get_other_link(NULL)) continue;
			// "get_other_link(NULL)" gives the same:
			// if (v == ((CDIntLine*)l)->get_other_link(NULL)) break;
		}
		
		dbg->log(40, "Bind vertex to line: ", l);
		// clone diagram and bind current vertex to this line
		CFeynDiagram* d_new = diag->clone();
		// all the integer values should remains the same
		CDVertex* v_new = d_new->get_vertex(v_num-1);
		CDLine* l_new = d_new->get_line(i);
		// bind v_new and l_new
		v_new->get_links()->add(l_new);
		l_new->get_links()->add(v_new);
		// further proceed this vertex
		develop_last_vertex(d_new);
		delete d_new;

		if (check_setting(COMPOSE_SETTINGS_BIND_TO_FIRST)) {
			// first time every vertex bind to first spotted line only
			if (v->get_first_binded_link_number() < 0) return;
			//if (next_tale_num == 0) return; // wrong: vertex could start new internal lines
		}
	}
	if (skip_lines) { printf("Internal error creating diagram.\n"); throw(new CEException(CEException::INT_ERROR)); }

	// 4. bind vertex to new internal line
	if (check_setting(COMPOSE_SETTINGS_NO_UNBINDED)) {
		// if this tale is last to bind...
		if ((next_tale_num+1) == number_of_tales) {
			// ...check up if all previous lines are created unbinded
			bool all_are_unbinded = true;
			int no_unbinded_i;
			for(no_unbinded_i=0; no_unbinded_i<next_tale_num; no_unbinded_i++) {
				CDElement* el = v->get_link(no_unbinded_i);
				if (el->is_binded()) { all_are_unbinded = false; break; }
			}
			if (all_are_unbinded) return;
		}
	}

	CDIntLine* int_l_new = CDIntLine::createIntLine(pk, NULL);
	diag->add_line(int_l_new);
	v->get_links()->add(int_l_new);
	int_l_new->get_links()->add(v);
	dbg->log(40, "New line binded to vertex: ", int_l_new);
	// further proceed this vertex
	develop_last_vertex(diag);
}

/*
*	find first spare line with 'pk'- particle kind
*/
int CFeynDiagramFactory::find_first_unbinded_line(CFeynDiagram* diag, int part_kind) {
	CEPtrArray* list = diag->get_lines_list();
	long s = list->size();
	long i;
	for(i=0; i<s; i++) {
		CDLine* l = (CDLine*)list->GetAt(i);
		if ((l->get_particle_kind() == part_kind) && (!l->is_binded()))
			return i;
	}
	return -1;
}

/*
*	defines if vertex can be binded to current diagram structure
*/
bool CFeynDiagramFactory::can_vertex_be_binded(CFeynDiagram* diag, CDiagramVertexInfo* vi) {
	int number_of_tales = vi->get_tales_num();
	
	int tale_num;
	for(tale_num=0; tale_num<number_of_tales; tale_num++) {
		int pk = vi->get_particle_kind(tale_num);
		
		int line_no = find_first_unbinded_line(diag, pk);
		if (line_no >= 0) return true;
	}
	return false;
}

////////////////
