////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// DElement.cpp: implementation of the CDElement class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CDElement::dbg = new CDebug("CDElement");
const char* CDElement::CLASS_NAME = "CDElement";

const int CDElement::EL_VERTEX			= 0x0010;
const int CDElement::EL_LINE			= 0x0020;
const int CDElement::EL_INT_LINE		= 0x0021;
const int CDElement::EL_EXT_LINE		= 0x0022;
const int CDElement::EL_VECTOR			= 0x0040;
const int CDElement::EL_SPINOR			= 0x0080;
const int CDElement::EL_LN_INITIAL		= 0x0100;
const int CDElement::EL_COUNTER_TERM	= 0x0200;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDElement::CDElement(const CDElementViewInfo* view_info) : CESerializable()
{
	m_DE_kind = 0; // undefined
	m_links = new CEPtrArray();
	m_pos_info = (view_info != NULL)?view_info->clone():NULL;
	m_temporary_mark = NULL;
}

CDElement::~CDElement()
{
	m_links->clearall_without_delete_objects();
	delete m_links;
	if (m_temporary_mark != NULL) delete m_temporary_mark;
	if (m_pos_info != NULL) delete m_pos_info;
}

const char* CDElement::ClassName() const {return CLASS_NAME;}
bool CDElement::contain_vector(void * obj) {
	if ((m_pos_info == obj) ||
		(m_temporary_mark == obj)) return true;
	if (m_pos_info != NULL) { if (m_pos_info->contain_vector(obj)) return true; }
	if (m_temporary_mark != NULL) { if (m_temporary_mark->contain_vector(obj)) return true; }
	return false;
}

CEPtrArray* CDElement::get_links() {return m_links;}
int CDElement::get_links_number() {return m_links->size();}
CDElement* CDElement::get_link(int i) {
	int s = get_links_number();
	if ((i<0) || (i>=s)) return NULL;
	return (CDElement*)m_links->GetAt(i);
}

int CDElement::DE_kind() {return m_DE_kind;}
bool CDElement::checkDEKind(int _mask) { return ((m_DE_kind & _mask) == _mask); }

void CDElement::setViewInfo(CDElementViewInfo* view_info) {
	if (m_pos_info != view_info) {
		if (m_pos_info != NULL) delete m_pos_info;
		m_pos_info = view_info;
	}
}

CEObject* CDElement::copy() const { return clone(); /*m_recent_copy = res;*/ }

/*
*	Note: links are allowed to refs_old objects only.
*	Links to source objects can be added simply.
*	If links may lead to other arrays (more than this two) there are two ways:
*	1. Enhance this method (let refs arrays be contain other arrays, etc.)
*	2. Group all refs into single array,
*	that will be deleted afterwards (without objects deleting).
*/
bool CDElement::copy_links(CEPtrArray *target, CEPtrArray *source, CEPtrArray *refs_new, CEPtrArray *refs_old)
{
	if ((target == NULL) || (source == NULL) || (refs_new == NULL) || (refs_old == NULL)) {
		printf("Error copying arrays links: NULL data.\n"); return false;}

	int arr_size = source->size();
	int refs_size = refs_old->size();

	if ((arr_size != target->size()) || (refs_size != refs_new->size())) {
		printf("Error copying arrays links: invalid sizes.\n"); return false;}

	int i;
	for(i=0; i<arr_size; i++) {
		CEPtrArray* links_old = ((CDElement*)source->GetAt(i))->get_links();
		CEPtrArray* links_new = ((CDElement*)target->GetAt(i))->get_links();
		if (links_new->size() > 0) {printf("Error copying arrays links: target not empty.\n"); return false;}

		int links_old_size = links_old->size();
		int j;
		for(j=0; j<links_old_size; j++) {
			// find link target
			CEObject* link_curr = links_old->GetAt(j);
			// all link should direct the line DE
			int target_num = -1;
			int k;
			for(k=0; k<refs_size; k++) {
				if (refs_old->GetAt(k) == link_curr) {target_num=k; break;}
			}

			if (target_num < 0) {printf("Error copying arrays links: invalid link.\n"); return false;}

			// add link to DE with same number
			links_new->add(refs_new->GetAt(target_num));
		}
	}
	return true;
}

int CDElement::get_first_binded_link_number() {
	int s = get_links_number();
	int i;
	for(i=0; i<s; i++) {
		if (get_link(i)->is_binded()) return i;
	}
	return -1;
}
