////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// DExtLine.cpp: implementation of the CDExtLine class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDExtLine::CDExtLine(CEParticle* _part, bool _is_initial, const CDElementViewInfo* view_info) : CDLine(view_info)
{
	m_particle = _part;
	m_DE_kind |= CDElement::EL_EXT_LINE;
	if (_is_initial) m_DE_kind |= CDElement::EL_LN_INITIAL;
	m_PK = m_particle->get_PK(); // cache basic data
}

CDExtLine::~CDExtLine()
{
}

const char* CDExtLine::CLASS_NAME = "CDExtLine";
const char* CDExtLine::ClassName() const { return CLASS_NAME; }

const char* CDExtLine::CLASS_TAG_NAME = "diagEL";
const char* CDExtLine::get_TAG_Name() { return CLASS_TAG_NAME; }

long CDExtLine::store_stream(ofstream* p_ofs, DWORD settings) {
	long res = 0;
	// Note: Links (from CDElement) should be saved separately

	// particle is stored by it's ID in kinematics
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) { dbg->log(0, "CDExtLine::store_stream() ERROR: NULL kinematics"); return 0; }
	int pID = kin->get_particle_ID(m_particle);
	if (pID == 0) { dbg->log(0, "CDExtLine::store_stream() ERROR: NO PARTICLE FOUND in kinematics"); return 0; }

	res += CESupport::storeXMLLong(pID, p_ofs, settings);
	res += CESupport::storeXMLLong((is_initial()?1:0), p_ofs, settings);
	res += CESupport::storeXMLLong(m_PK, p_ofs, settings); // for pID check

	return res;
}

CESerializable* CDExtLine::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	// Note: Links (from CDElement) should be restored separately

	long pID = 0;
	bool read_OK = CESupport::readXMLLong(data, &pos, &pID, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	long isIn = -1;
	read_OK = CESupport::readXMLLong(data, &pos, &isIn, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	long pk = 0;
	read_OK = CESupport::readXMLLong(data, &pos, &pk, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	read_OK = CESupport::checkAllRead(data, &pos, settings, CLASS_NAME);
	if (!read_OK) return NULL;

	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) { dbg->log(0, "CDExtLine::restore_from_buffer() ERROR: NULL kinematics"); CEException::halt(); }
	CEParticle* part = kin->get_particle_by_ID(pID);
	if (part == 0) { dbg->log(0, "CDExtLine::restore_from_buffer() ERROR: NO PARTICLE FOUND in kinematics"); CEException::halt(); }
	if (part->get_PK() != pk) { dbg->log(0, "CDExtLine::restore_from_buffer() ERROR: diagram loading at different kinematics"); CEException::halt(); }
	
	CDExtLine* res = new CDExtLine(part, (isIn>0), NULL);
	return res;
}

long CDExtLine::get_particle_kind() const { return m_PK; }

void CDExtLine::set_particle_kind(long pk) {
	if (m_PK != pk) { dbg->log(0, "Internal WARNING: illegal call CDExtLine::set_particle_kind() is ignored."); }
}

/*
*	returns ID of particle in kinematics
*/
int CDExtLine::get_particle_ID() {
	return m_particle->get_ID();
}

void CDExtLine::set_momentum(CExpression* _momentum, CDElement* vertex_from) { // error
	printf("Error : CDExtLine::set_momentum() call.\n"); throw(new CEException(CEException::INT_ERROR));
}

bool CDExtLine::is_binded()
{
	return (get_links_number() >= 1);
}

CExpression* CDExtLine::get_momentum(CDElement* vertex_to) {
	if (m_particle == NULL) { dbg->log(0, "<Error> NULL particle."); throw(new CEException(CEException::INT_ERROR)); }
	CExpression* res = m_particle->get_momentum_symbol();
	if (res == NULL) return NULL; // Error?
	if (vertex_to == NULL) return res->clone();

	// find out vertex in links list
	CEPtrArray* links = get_links();
	int s = links->size();
	if (s != 1) {
		// Error
		printf("Error retrieving EXTERNAL line momentum.\n"); throw(new CEException(CEException::INT_ERROR));
	}
	
	CDElement* v = (CDElement*)links->GetAt(0);
	if (vertex_to != v) return NULL; // unknown vertex parameter
	
	res = res->clone();
	if(!is_initial()) res->inverseFlag();
	return res;
}

CExpression* CDExtLine::get_internal_momentum_val() {
	if (m_particle == NULL) { dbg->log(0, "<Error> NULL particle."); throw(new CEException(CEException::INT_ERROR)); }
	return m_particle->get_momentum_symbol();
}

CDElement* CDExtLine::clone() const {
	return new CDExtLine(m_particle, is_initial(), getViewInfoConst());
}

bool CDExtLine::is_initial() const {
	if (m_DE_kind & CDElement::EL_LN_INITIAL) return true;
	return false;
}

CEString* CDExtLine::toString(int mode) {
	CEString* res = NULL;

	res  = new CEString("{CDExtLine: ");
	res->add_int(get_particle_kind(),true);
	res->AddString(", ");
	int links_num = get_links_number();
	res->add_int(links_num);
	res->AddString((links_num==1)?" link, ":" links, ");
	CDElementViewInfo* vi = getViewInfo();
	if (vi == NULL) res->AddString("No View Info");
	else res->AddString(vi->toString(mode));
	res->AddString("}");

	return res;
}

CEString* CDExtLine::draw(int settings)
{
	//dbg->log(80, "CDExtLine::draw");

	CEString* res = NULL;

	CDElementViewInfo* pos_inf_1 = NULL;
	CDElementViewInfo* pos_inf_2 = NULL;
	CPos2D* pos_1 = NULL;
	CPos2D* pos_2 = NULL;

	// if position exist draw line
	pos_inf_1 = getViewInfo();
	if (pos_inf_1 != NULL) pos_1 = pos_inf_1->get_pos();

	CEPtrArray* links = get_links();
	if (links->size() == 1) {
		pos_inf_2 = ((CDElement*)links->GetAt(0))->getViewInfo();
		if (pos_inf_2 != NULL) pos_2 = pos_inf_2->get_pos();
	}

	if ((pos_1 != NULL) && (pos_2 != NULL)) {
		int line_PK = get_particle_kind();
		if (!is_initial()) line_PK = -line_PK; // see p.31 for details.
		if (settings & CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES) {
			//settings|=CFeynDiagram::DRAW_CAPS_PARTICLE_SIGNS;
			if (CEParticle::is_bosonic_PK(line_PK)) settings|=CFeynDiagram::DRAW_CAPS_PARTICLE_SIGNS;
		}
		CEString* line_cap = GetDrawCaption(settings);
		res = CLDrawDiagView::draw_diagram_line(line_PK, pos_1, pos_2, settings, line_cap, false);
		if (line_cap != NULL) delete line_cap;
	}
	else {
		dbg->log(0, "CDExtLine::draw WARNING : no legal position found");
		dbg->log(3, "CDElementViewInfo #1:", pos_inf_1);
		dbg->log(3, "CDElementViewInfo #2:", pos_inf_2);
	}

	//dbg->log(80, "CDExtLine::draw result: ", res);
	return res;
}

CExpression* CDExtLine::get_vector_index() {

	CExpression* res = NULL;
	if (m_particle->check_PK(CEParticle::PARTICLE_VECTOR_BOSON, true)) {
		res = ((CPAbstractBoson*)m_particle)->get_vector_index();
	}
	return res;
}

/*bool CDExtLine::is_similar_to(CDExtLine *other)
{
	if (other == NULL) return false;

	if (is_initial() == other->is_initial()) {
		if (get_particle_kind() == other->get_particle_kind())
			return true;
	}
	return false;
}*/
