////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// DIntLine.cpp: implementation of the CDIntLine class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const char* CDIntLine::CLASS_NAME = "CDIntLine";

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CDIntLine::CDIntLine(long part_kind, const CExpression* _momentum, bool _is_CT_line, const CDElementViewInfo* view_info)
: CDLine(view_info)
{
	m_DE_kind |= CDElement::EL_INT_LINE;
	m_particle_kind = part_kind;

	m_momentum = NULL; // NULL means no momentum assigned.
	if (_momentum != NULL) m_momentum = _momentum->clone();

	m_is_CT_line = _is_CT_line;
	if (m_is_CT_line) m_DE_kind |= CDElement::EL_COUNTER_TERM;
}

CDIntLine::CDIntLine(): CDLine(NULL) { CEException::halt("CDIntLine() illegal call"); }

CDIntLine::~CDIntLine()
{
	if (m_momentum != NULL) delete m_momentum;
}

CDIntLine* CDIntLine::createIntLine(int part_kind, CExpression* _momentum) {
	return new CDIntLine(part_kind, _momentum, false, NULL);
}

CDIntLine* CDIntLine::createCounterTermLine(int part_kind, CExpression* _momentum) {
	return new CDIntLine(part_kind, _momentum, true, NULL);
}

const char* CDIntLine::CLASS_TAG_NAME = "diagIL";
const char* CDIntLine::get_TAG_Name() { return CLASS_TAG_NAME; }

long CDIntLine::store_stream(ofstream* p_ofs, DWORD settings) {
	// Note: Links (from CDElement) should be saved separately
	long res = 0;

	if (m_momentum != NULL) {
		res += CESupport::store_stream(m_momentum, p_ofs, settings);
	} else {
		CExpression* m = CENumber::GetZero();
		res += CESupport::store_stream(m, p_ofs, settings);
		delete m;
	}

	res += CESupport::storeXMLLong(m_particle_kind, p_ofs, settings);
	res += CESupport::storeXMLLong((m_is_CT_line?1:0), p_ofs, settings);

	return res;
}

CESerializable* CDIntLine::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	// Note: Links (from CDElement) should be restored separately
	
	CExpression* m = CESupport::readXMLExpr(data, &pos, settings, CLASS_NAME, true);
	if (m == NULL) return NULL;
	long pk = 0;
	bool read_OK = CESupport::readXMLLong(data, &pos, &pk, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	long isCT = -1;
	read_OK = CESupport::readXMLLong(data, &pos, &isCT, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	read_OK = CESupport::checkAllRead(data, &pos, settings, CLASS_NAME);
	if (!read_OK) return NULL;

	if (CEBasic::isZero(m)) { delete m; m=NULL; }

	CDIntLine* res = new CDIntLine(pk, m, (isCT>0), NULL);
	delete m;
	return res;
}

CEString* CDIntLine::toString(int mode)
{
	CEString* res = NULL;

	res  = new CEString("{CDIntLine:");
	res->add_int(get_particle_kind(),true);
	res->AddString(", ");
	CExpression* mom_val = get_internal_momentum_val();
	if (mom_val != NULL) res->AddString(get_internal_momentum_val()->toString(mode));
	else res->AddString("NULL");
	res->AddString(", ");
	int links_num = get_links_number();
	res->add_int(links_num);
	res->AddString((links_num==1)?" link, ":" links, ");
	CDElementViewInfo* vi = getViewInfo();
	if (vi == NULL) res->AddString("No View Info");
	else res->AddString(vi->toString(mode));
	res->AddString("}");
	return res;
}

CEString* CDIntLine::draw(int settings) {
	//dbg->log(30, "CDIntLine::draw");

	CEString* res = NULL;

	CDElementViewInfo* pos_inf_1 = NULL;
	CDElementViewInfo* pos_inf_2 = NULL;
	CPos2D* pos_1 = NULL;
	CPos2D* pos_2 = NULL;

	// if positions exist draw line
	CEPtrArray* links = get_links();
	if (links->size() == 2) {
		pos_inf_1 = ((CDElement*)links->GetAt(0))->getViewInfo();
		if (pos_inf_1 != NULL) pos_1 = pos_inf_1->get_pos();

		pos_inf_2 = ((CDElement*)links->GetAt(1))->getViewInfo();
		if (pos_inf_2 != NULL) pos_2 = pos_inf_2->get_pos();
	}
	
	if ((pos_1 != NULL) && (pos_2 != NULL)) {
		CDElementViewInfo* vi_curr = getViewInfo();
		CEString* line_cap = GetDrawCaption(settings);
		bool _is_CT = ((m_DE_kind & CDElement::EL_COUNTER_TERM) != 0);
		res = CLDrawDiagView::draw_diagram_line(get_particle_kind(),
			pos_1, pos_2, settings, line_cap,
			vi_curr->get_repetition_ID(), vi_curr->get_repetition_number(), _is_CT);
		if (line_cap != NULL) delete line_cap;
	}
	else {
		dbg->log(20, "CDIntLine::draw WARNING : no legal position found");
		dbg->log(20, "CDElementViewInfo #1:", pos_inf_1);
		dbg->log(20, "CDElementViewInfo #2:", pos_inf_2);
	}

	//dbg->log(30, "CDIntLine::draw result", res);
	return res;
}

CDElement* CDIntLine::clone() const
{
	return new CDIntLine(get_particle_kind(), m_momentum, m_is_CT_line, getViewInfoConst());
}

const char* CDIntLine::ClassName() const {return CLASS_NAME;}
long CDIntLine::get_particle_kind() const {return m_particle_kind;}
void CDIntLine::set_particle_kind(long pk) {
	if ((pk<0) && (CEParticle::is_neutral_bosonic_PK(pk))) pk=-pk;
	m_particle_kind=pk;
}

bool CDIntLine::contain_vector(void * obj) {
	if (CDElement::contain_vector(obj)) return true;
	if (m_momentum == obj) return true;
	if (m_momentum != NULL) { if (m_momentum->contain_vector(obj)) return true; }
	return false;
}

void CDIntLine::ArrangeElement() {
	if ((m_particle_kind < 0) && (is_binded())) {
		CEPtrArray* v_links = get_links();

		// swap vertex links
		CDElement* v1 = (CDElement*)v_links->GetAt(0);
		v_links->RemoveAt(0, false);
		v_links->add(v1);

		m_particle_kind = -m_particle_kind;
	}
}

bool CDIntLine::is_binded()
{
	return (get_links()->size() >= 2);
}

/*
*	Do not check whether "link" realy exist.
*	if (link == NULL) returns first link met.
*/
CDElement* CDIntLine::get_other_link(CDElement* link) {
	CEPtrArray* l_links = get_links();
	int s = l_links->size();
	if (s > 2) { printf("Internal line with more than 2 links.\n"); throw(new CEException(CEException::INT_ERROR)); }

	int i;
	for(i=0; i<s; i++) {
		CEObject* vert = l_links->GetAt(i);
		if (vert != link) return (CDElement*)vert;
		else if (i==1) {
			// both ends of line in single vertex
			// (i==1) check means  that link is met in second time
			return (CDElement*)vert;
		}
	}
	return NULL;
}

CDElement* CDIntLine::get_vertex_to() {
	if (get_links_number() != 2) {
		dbg->log(0, "Error: :get_vertex_to() at non-binded line."); return NULL;
	}

	long pk = get_particle_kind();
	CDElement* res = get_link( (pk>=0)?1:0 );
	return res;
}

CDElement* CDIntLine::get_vertex_from() {
	if (get_links_number() != 2) {
		dbg->log(0, "Error: :get_vertex_from() at non-binded line."); return NULL;
	}

	long pk = get_particle_kind();
	CDElement* res = get_link( (pk>=0)?0:1 );
	return res;
}

///////////////////////////////////////////////////////////////////////////////////////
//	Returns the momentum of line goes to vertex.
//	For (vertex != NULL) returns CLONED value.
//	For "NULL" vertex also returns CLONED internal momentum value
///////////////////////////////////////////////////////////////////////////////////////
CExpression* CDIntLine::get_momentum(CDElement* vertex_to) {
	if (m_momentum == NULL) return NULL;
	if (vertex_to == NULL) return m_momentum->clone();

	// find out vertex in links list
	CEPtrArray* links = get_links();
	int s = links->size();
	if (s != 2) {
		// Error
		printf("Error retrieving line momentum.\n"); throw(new CEException(CEException::INT_ERROR));
	}

	CDElement* v1 = (CDElement*)links->GetAt(0);
	CDElement* v2 = (CDElement*)links->GetAt(1);
	//if (v1 == v2) return NULL;
	if ((vertex_to != v1) && (vertex_to != v2)) return NULL;

	CExpression* res = m_momentum->clone();
	if (vertex_to == v1) res->inverseFlag();
	res->Arrange(AL_INLINE);
	return res;
}

///////////////////////////////////////////////////////////////////////////////////////
//	used for cloning.
///////////////////////////////////////////////////////////////////////////////////////
CExpression* CDIntLine::get_internal_momentum_val() { return m_momentum; }

void CDIntLine::applyRuleToMomentum(CERule* r) {  // Used to change volatile variables in momentum
	if (r == NULL) return;
	m_momentum = r->ApplyTo(m_momentum, true);
}

///////////////////////////////////////////////////////////////////////////////////////
//	Treats momentum as OUTCOMING from "vertex_from" vertex.
//	If (vertex_from == NULL) the momentum can only be cleared (if _momentum -> NULL).
//	Does not affect _momentum object.
///////////////////////////////////////////////////////////////////////////////////////
void CDIntLine::set_momentum(CExpression* _momentum, CDElement* vertex_from) {
	if (_momentum == NULL) {
		if (m_momentum != NULL) delete m_momentum;
		m_momentum = NULL;
		return;
	}
	if (vertex_from == NULL) return;

	// find out vertex in links list
	CEPtrArray* links = get_links();
	int s = links->size();
	if (s != 2) {
		// Error
		printf("Error setting line momentum.\n"); throw(new CEException(CEException::INT_ERROR));
	}

	CDElement* v1 = (CDElement*)links->GetAt(0);
	CDElement* v2 = (CDElement*)links->GetAt(1);
	//if (v1 == v2) return;
	if ((vertex_from != v1) && (vertex_from != v2)) return; // Error. Skip this call.
	if (m_momentum == _momentum) return; // Error. Skip this call.

	if (m_momentum != NULL) delete m_momentum;
	m_momentum = _momentum->clone();
	if (vertex_from == v2) m_momentum->inverseFlag();
}


/*CDIntLine::CDIntLine(CDVertex* _v_to):CDLine()
{
	m_v_to = _v_to;
	m_v_from = NULL;
	//m_is_binded = false;
}

CDIntLine::CDIntLine(CDVertex* _v_to, CDVertex* _v_from, bool _is_binded):CDLine()
{
	m_v_to = _v_to;
	m_v_from = _v_from;
	m_is_binded = _is_binded;
}*/

/*void CDIntLine::bind(CDVertex * v)
{
	m_v_from = v;
	m_is_binded = true;
}*/

/*CDVertex* CDIntLine::vertex_to()
{
	return m_v_to;
}

CDVertex* CDIntLine::vertex_from()
{
	return m_v_from;
}*/

/*CDVertex* CDIntLine::get_other_vertex(CDVertex * v)
{
	if (v == vertex_to()) return vertex_from();
	if (v == vertex_from()) return vertex_to();
	//Error.
	printf("Error (CDIntLine::get_other_vertex) : undefined vector request.\n");
	CDElement::dbg->log(1, "Error (CDIntLine::get_other_vertex) : undefined vector request", this);
	return NULL;
}*/

/*void CDIntLine::set_momentum(CDVertex* v_to, CExpression* val)
{
	dbg->log(35, "set momentum", val);
	//dbg->log(15, "old momentum", m_momentum);
	if (m_momentum == val) return;
	if (m_momentum != NULL) delete m_momentum;

	if (val != NULL) {
		m_momentum = val->clone();
		
		if (v_to != NULL) {
			if (v_to == vertex_from()) {
				m_momentum->inverseFlag();
			}
			else if (v_to != vertex_to()) {
				//Error: invalid vertex parameter.
				CDElement::dbg->log(1, "Error (CDIntLine::set_momentum) : invalid vertex parameter.", this);
			}
		}

		//do m_momentum->Arrange() here?
		m_momentum = CExpression::arrge(m_momentum);
	}
	else m_momentum = NULL;
	//dbg->log(15, "new momentum", m_momentum);
}
*/
