////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// DVertex.cpp: implementation of the CDVertex class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDVertex::CDVertex(const CDiagramVertexInfo* vi, const CDElementViewInfo* view_info) : CDElement(view_info)
{
	m_vertex_info = (vi != NULL)?vi->clone():NULL;
	
	m_DE_kind |= CDElement::EL_VERTEX;
	if (vi != NULL) {
		if (vi->check_vert_type(CDiagramVertexInfo::VERTEX_COUNTERTERM)) {
			m_DE_kind |= CDElement::EL_COUNTER_TERM;
		}
	}
}

CDVertex::~CDVertex()
{
	if (m_vertex_info != NULL) delete m_vertex_info;
}

const char* CDVertex::CLASS_NAME = "CDVertex";
const char* CDVertex::ClassName() const { return CLASS_NAME; }
bool CDVertex::contain_vector(void * obj) {
	if (CDElement::contain_vector(obj)) return true;
	if (m_vertex_info == obj) return true;
	if (m_vertex_info != NULL) { if (m_vertex_info->contain_vector(obj)) return true; }
	return false;
}

const char* CDVertex::CLASS_TAG_NAME = "diagV";
const char* CDVertex::get_TAG_Name() { return CLASS_TAG_NAME; }

long CDVertex::store_stream(ofstream* p_ofs, DWORD settings) {
	long res = 0;

	// Note: Links (from CDElement) should be saved separately

	// VertexInfo
	if (m_vertex_info != NULL) { res += CESupport::store_stream(m_vertex_info, p_ofs, settings); }

	return res;
}

CESerializable* CDVertex::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {

	// Note: Links (from CDElement) should be restored separately
	
	CESerializable* vi = CESupport::readXMLObj(data, &pos, CDiagramVertexInfo::CLASS_NAME, settings, CLASS_NAME);
	if (vi == NULL) { return NULL; }

	bool read_OK = CESupport::checkAllRead(data, &pos, settings, CLASS_NAME);
	if (!read_OK) { delete vi; return NULL; }

	CDVertex* res = new CDVertex((CDiagramVertexInfo*)vi, NULL);
	delete vi;
	return res;
}

CDiagramVertexInfo* CDVertex::get_vertex_info() { return m_vertex_info; }

void CDVertex::set_vertex_info(CDiagramVertexInfo* vi_const) {
	CDiagramVertexInfo* vi_old=m_vertex_info;
	m_vertex_info = (vi_const != NULL)?vi_const->clone():NULL;
	if (vi_old!=NULL) delete vi_old;
}

bool CDVertex::is_binded()
{
	int links_num = get_links()->size();
	int links_num_required = get_vertex_info()->get_tales_num();
	return (links_num == links_num_required);
}

/*
*	means that vertex is connected to any other vertex
*/
bool CDVertex::has_binded_tale_line()
{
	CEPtrArray* lines = get_links();
	int s = lines->size();
	int i;
	for(i=0; i<s; i++) {
		if (((CDElement*)lines->GetAt(i))->is_binded()) return true;
	}
	return false;
}

CDElement* CDVertex::clone() const
{
	return new CDVertex(m_vertex_info, getViewInfoConst());
}

CEString* CDVertex::toString(int mode)
{
	CEString* res = NULL;
	res  = new CEString("{CDVertex: binded="); res->add(is_binded());
	res->AddString(", id:"); res->add_int(m_vertex_info->get_vertex_ID());

	res->AddString("("); 
	int s = m_vertex_info->get_tales_num(); int i;
	for(i=0; i<s; i++) {
		long pk = m_vertex_info->get_particle_kind(i); res->add_int(pk,true); if ((i+1)<s) res->AddString(", ");
	}
	res->AddString("), ");
	CDElementViewInfo* vi = getViewInfo();
	if (vi == NULL) res->AddString("No View Info");
	else res->AddString(vi->toString(mode));
	res->AddString("}");
	
	return res;
}

CEString* CDVertex::draw(int settings)
{
	CEString* res = NULL;
	if ((m_DE_kind & CDElement::EL_COUNTER_TERM) != 0) {
		// if positions exist draw cross (counterterm mark)
		CDElementViewInfo* pos_inf = getViewInfo();
		if (pos_inf != NULL) {
			CPos2D* pos = pos_inf->get_pos();
			int xc = pos->get_X();
			int yc = pos->get_Y();
			int mark_sz = 5;
			
			res = new CEString("\\Line(");
			res->add_int(xc+mark_sz); res->AddString(", "); res->add_int(yc+mark_sz);
			res->AddString(")(");
			res->add_int(xc-mark_sz); res->AddString(", "); res->add_int(yc-mark_sz);
			res->AddString(")");
			
			res->AddString("\\Line(");
			res->add_int(xc+mark_sz); res->AddString(", "); res->add_int(yc-mark_sz);
			res->AddString(")(");
			res->add_int(xc-mark_sz); res->AddString(", "); res->add_int(yc+mark_sz);
			res->AddString(")");
		} else { dbg->log(0, "WARNING: drawing diagram with ndefined CT-vertex position"); }
	}
	return res;
}

CDVertex* CDVertex::get_vertex_linked(int n) {
	if ((n<0) || (n>=get_links_number())) return NULL;
	CDElement* line = (CDElement*)get_links()->GetAt(n);

	// find other link of line
	CEPtrArray* l_links = line->get_links();
	int s = l_links->size();
	if (s != 2) return NULL;// can be external line
	int i;
	for(i=0; i<s; i++) {
		CEObject* vert = l_links->GetAt(i);
		if (vert != this) return (CDVertex*)vert;
		else if (i == (s-1)) return (CDVertex*)vert;//both ends of line in single vertex
	}

	//Invalid vertex data
	printf("CDVertex::get_vertex_linked : Invalid vertex data\n");
	return NULL;
}

//////////////////////////////////////////////////////////////////////////
//	returns -1 if no momenta can be arranged.
//	returns 0 if all the momenta are arranged.
//	returns 1 if momenta arranged successfully.
//////////////////////////////////////////////////////////////////////////
int CDVertex::ArrangeMomenta(CKinematics* kin) {
	dbg->log(50, "CDVertex::ArrangeMomenta()");
	CDLine* unarranged_line = NULL;

	// find out unarranged line
	int s = get_links_number();
	int i;
	for(i=0; i<s; i++) {
		CDLine* l = (CDLine*)get_links()->GetAt(i);
		if (l->get_internal_momentum_val() == NULL) {
			if (unarranged_line != NULL) {
				dbg->log(60, "CDVertex::ArrangeMomenta() : return -1.");
				return -1; // two or more unarranged lines at vertex
			}
			unarranged_line = l;
		}
	}
	if (unarranged_line == NULL) {
		dbg->log(60, "CDVertex::ArrangeMomenta() : return 0.");
		return 0; // all the momenta are arranged
	}
	if (s < 2) {
		dbg->log(60, "CDVertex::ArrangeMomenta() : return -1.");
		return -1; // invalid vertex (vertex is supposed to be binded)
	}

	dbg->log(70, "Line to be arranged: ", unarranged_line);

	// involve momentum conservation
	CExpression* res_momentum = NULL;
	for (i=0; i<s; i++) {
		CDLine* l = (CDLine*)get_links()->GetAt(i);
		CExpression* l_mom = l->get_momentum(this);
		if (l_mom != NULL) {// or if(l_mom != unarranged_line)
			if (res_momentum == NULL) res_momentum = l_mom;
			else res_momentum = res_momentum->add(l_mom);
		}
	}

	if (res_momentum == NULL) {
		// internal error
		printf("Internal ERROR.\n"); throw(new CEException(CEException::INT_ERROR));
	}

	res_momentum = CExpression::reduce(res_momentum);
	dbg->log(80, "Momentum composed: ", res_momentum);

	// simplify momentum
	if (kin != NULL) res_momentum = kin->get_simpliest(res_momentum);
	dbg->log(80, "Momentum simplified: ", res_momentum);

	// res_momentum is sum of "INCOMING" lines -- set as "OUTGOING" momentum.
	unarranged_line->set_momentum(res_momentum, this);
	delete res_momentum;
	dbg->log(60, "CDVertex::ArrangeMomenta() : return 1.");
	return 1;
}

/*
*	Debug function: used to vertex self-check
*	throws the exception is attached lines order does not fits the VErtexInfo structure.
*	returns true otherwise
*	Note: does not matter if vertex is completely binded
*	All the binded lines should have the same PK (or -PK for neutral bosons) as in VertexInfo.
*/
bool CDVertex::check_VI_validity() {
	const char* VERTEX_DATA_ERROR_STRING = "Invalid Vertex Data";
	if (m_vertex_info == NULL) {
		dbg->log(0, "Error: Vertex with NULL Vertex Info structure", this);
		throw(new CEException(VERTEX_DATA_ERROR_STRING));
	}

	int vi_s = m_vertex_info->get_tales_num();
	int links_num = get_links()->size();
	if (links_num != vi_s) {
		dbg->log(0, links_num, vi_s, "<Error> links number (%ld) does not fits  VetrtexInfo data (%ld)");
		dbg->log(0, "Vertex is", this);
		throw(new CEException(VERTEX_DATA_ERROR_STRING));
	}

	int i;
	for(i=0; i<links_num; i++) {
		long pk_VI = m_vertex_info->get_particle_kind(i);
		CDLine* l = (CDLine*)get_link(i);
		long pk_line = l->get_particle_kind();

		// inverse line PK if line is outgoing
		// (Note: all lines in Vertex Info corresponds to incoming particles)
		if (l->is_external()) {
			if ( !((CDExtLine*)l)->is_initial() ) pk_line = -pk_line;
		} else {
			if (l->get_link(1) != this) pk_line = -pk_line;  // i.e. outgoing line
		}

		if (pk_line == pk_VI) continue;

		// compare absolute values for NEUTRAL BOSONS
		long pk_VI_abs = CLMath::abs(pk_VI);
		long pk_line_abs = CLMath::abs(pk_line);
		if ((pk_VI_abs & (long)CEParticle::PARTICLE_NEUTRAL_BOSON) == (long)CEParticle::PARTICLE_NEUTRAL_BOSON) {
			if (pk_VI_abs == pk_line_abs) continue;
		}

		// declare internal error
		dbg->log(0, "Error: Vertex Data do not meet the Vertex Info structure", this);
		dbg->log(0, pk_VI, "pk_VI: %lx, VI:", m_vertex_info);
		dbg->log(0, l->get_particle_kind(), "pk_line: %lx, line:", l);
		throw(new CEException(VERTEX_DATA_ERROR_STRING));
	}
	return true;
}

long CDVertex::GetLogicalPKForAttachedLine(CDElement* l) {
	if ((l->DE_kind() & CDElement::EL_LINE) == 0) { dbg->log(0, "Error: GetLogicalPKForAttachedLine() illegal call", l); CEException::halt(); }
	CDLine* ln = (CDLine*)l;
	long res = ln->get_particle_kind();

	// inverse line PK if line is outgoing (this(link0)->other(link1))
	// (Note: all lines in Vertex Info corresponds to INCOMING particles)
	if (ln->is_external()) {
		if ( !((CDExtLine*)ln)->is_initial() ) res = -res;
	} else {
		// check line
		if ((ln->get_link(0) != this) && (ln->get_link(1) != this)) { dbg->log(0, "<Error> GetLogicalPKForAttachedLine(): Not binded line spotted (invalid digaram structure?)."); }
		if ((ln->get_link(0) == this) && (ln->get_link(1) == this)) { dbg->log(0, "<Error> GetLogicalPKForAttachedLine(): Not expected self-loop spotted"); }

		if (ln->get_link(1) != this) res = -res; // i.e. outgoing line (this(link0)->other(link1))
	}
	return res;
}

void CDVertex::SetLogicalPKForAttachedLine(CDElement* l, long pk_incoming) {
	if ((l->DE_kind() & CDElement::EL_LINE) == 0) { dbg->log(0, "Error: SetLogicalPKForAttachedLine() illegal call", l); CEException::halt(); }
	CDLine* ln = (CDLine*)l;
	long res = pk_incoming;

	// inverse line PK if line is outgoing (this(link0)->other(link1))
	// (Note: all lines in Vertex Info corresponds to INCOMING particles)
	if (ln->is_external()) {
		if ( !((CDExtLine*)ln)->is_initial() ) res = -res;
	} else {
		// check line
		if ((ln->get_link(0) != this) && (ln->get_link(1) != this)) { dbg->log(0, "<Error> GetLogicalPKForAttachedLine(): Not binded line spotted (invalid digaram structure?)."); }
		//if ((ln->get_link(0) == this) && (ln->get_link(1) == this)) { dbg->log(0, "<Error> GetLogicalPKForAttachedLine(): Not expected self-loop spotted"); }

		if (ln->get_link(1) != this) res = -res; // i.e. outgoing line (this(link0)->other(link1))
	}
	ln->set_particle_kind(res);
}

/*
*	sorts links order according to VertexInfo data
*/
bool CDVertex::sortLinksOrder() {
	int vi_s = m_vertex_info->get_tales_num();
	int links_num = get_links()->size();
	if (links_num != vi_s) { dbg->log(0, links_num, vi_s, "<Error> links number (%ld) does not fits  VetrtexInfo data (%ld). Arrangement skipped."); return false; }
	
	int i;
	for(i=0; i<links_num; i++) {
		long pk_VI = m_vertex_info->get_particle_kind(i);
		long pk_VI_abs = CLMath::abs(pk_VI);
		bool bIsClassPK=((pk_VI_abs & CEParticle::PARTICLE_CLASS_FLAG)==CEParticle::PARTICLE_CLASS_FLAG);

		int appr_link_no = -1;
		// find out appropriate link
		int j;
		for(j=i; j<links_num; j++) {
			bool line_fits = false;
			CDLine* l = (CDLine*)get_link(j);
			long pk_line = GetLogicalPKForAttachedLine(l); long pk_line_abs = CLMath::abs(pk_line);
			if (bIsClassPK) {
				if ( ((pk_VI>=0) && (pk_line>=0)) || ((pk_VI<0) && (pk_line<0)) ) {
					if (CEParticle::FitClassPk(pk_line_abs,pk_VI_abs)) line_fits=true;
				} else { // compare absolute values for NEUTRAL BOSONS (???)
					if ((pk_VI_abs & (long)CEParticle::PARTICLE_NEUTRAL_BOSON) == (long)CEParticle::PARTICLE_NEUTRAL_BOSON) {
						if (CEParticle::FitClassPk(pk_line_abs,pk_VI_abs)) line_fits=true;
					}
				}
			} else {
				if (pk_line == pk_VI) line_fits = true;
				else { // compare absolute values for NEUTRAL BOSONS (???)
					if ((pk_line_abs & (long)CEParticle::PARTICLE_NEUTRAL_BOSON) == (long)CEParticle::PARTICLE_NEUTRAL_BOSON) {
						if (pk_VI_abs == pk_line_abs) line_fits = true;
					}
				}
			}

			if (line_fits) { appr_link_no = j; break; }
		}

		if (appr_link_no < 0) {
			// can't sort vertex
			dbg->log(0, "<Error> Vertex can't be arranged according to VertexInfo", this);
			dbg->log(2, vi_s, "VI (%ld tales): ", m_vertex_info);
			dbg->log(2, i, pk_VI, "vertex links (halted on index #%ld, pk_VI: %lx): ", get_links());
			for(j=0; j<links_num; j++) {
				CDLine* l = (CDLine*)get_link(j);
				//dbg->log(2, GetLogicalPKForAttachedLine(l), l->get_particle_kind(), "logical PK is %lx instead of %ld for line", l);
				long pkL=GetLogicalPKForAttachedLine(l);
				if (pkL>=0) {
					dbg->log(2, pkL, "logical PK is 0x%lx for line", l);
				} else {
					dbg->log(2, -pkL, "logical PK is -0x%lx for line", l);
				}
				//long pkL_abs=(pkL>0)?pkL:-pkL; dbg->log(2, pkL_abs, "logical PK is %lx for line", l);
			}
			return false;
		}

		if (appr_link_no == i) continue; // order is correct

		// move appropriate link ahead
		CEPtrArray* links_array = get_links();
		links_array->MoveFromTo(j, i);
	}
	return true;
}

/*
*	sorts links order according to external order
*	Note: wrong order will halt the program!
*	To avoid halting, the back-up list should be kept.
*/
bool CDVertex::sortLinksOrder(CIntArray* order) {
	if (order==NULL) return false;
	int links_num=order->GetSize();
	if (links_num!=m_vertex_info->get_tales_num()) return false;
	CEPtrArray* links_array = get_links(); if(links_array==NULL)return false;
	if (links_num!=links_array->size()) return false;
	int i;

	// order check
	for(i=0; i<links_num; i++) {
		int old_link_no = order->GetAt(i); if ((old_link_no<0) || (old_link_no>=links_num)) { dbg->log(0,"CDVertex::sortLinksOrder(INVALID order)",order); return false; }
	}

	CEPtrArray links_old; // array of non-cloned links at old order
	// fill links_array with NULLs to trac same link repeation
	for(i=0; i<links_num; i++) { links_old.add(links_array->GetAt(i)); links_array->SetAt(i,new CENull(),false); }

	for(i=0; i<links_num; i++) {
		long pk_VI = m_vertex_info->get_particle_kind(i);
		long pk_VI_abs = CLMath::abs(pk_VI);
		//bool bIsClassPK=((pk_VI_abs & CEParticle::PARTICLE_CLASS_FLAG)==CEParticle::PARTICLE_CLASS_FLAG);
		int old_link_no = order->GetAt(i);
		CEObject* link=links_old.GetAt(old_link_no);
		if ((link==NULL) || (link->ClassName()==CENull::CLASS_NAME)) { dbg->log(0,"CDVertex::sortLinksOrder(INVALID order)",order); CEException::halt(); return false; }
		links_array->SetAt(i,link,true); //delete temporary CENULL
		links_old.SetAt(old_link_no,new CENull(),false); // mark link as used
	}
	return true;
}
