////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// DiagramPhysics.cpp: implementation of the CDiagramPhysics class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CDiagramPhysics::dbg = new CDebug("CDiagramPhysics");
CDiagramPhysics* CDiagramPhysics::m_physics = NULL;

const int CDiagramPhysics::SYMBOL_STYLE_TEX	= 1;
const int CDiagramPhysics::SYMBOL_STYLE_NB	= 2;

const char* CDiagramPhysics::m_SymGeV2PbConst = "GEV2PB";

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CDiagramPhysics::CDiagramPhysics() : CEObject(true) {
	// Note: set_mass_sym() function is noe virtual. Do not call it here.
	m_vertices_info = new CEPtrArray();
	m_DR_mass_symbol = NULL;
	m_arrClassPKTitles.clear_all(); m_iIdClassPKTitles.ToSize(0);
}

CDiagramPhysics::~CDiagramPhysics()
{
	delete m_vertices_info;
	if (m_DR_mass_symbol != NULL) { delete m_DR_mass_symbol; m_DR_mass_symbol = NULL; }
}

const char* CDiagramPhysics::CLASS_NAME = "CDiagramPhysics";
const char* CDiagramPhysics::ClassName() const { return CLASS_NAME; }
bool CDiagramPhysics::contain_vector(void * obj) {
	if ((m_vertices_info == obj) || (m_DR_mass_symbol == obj)) return true;
	if (m_vertices_info != NULL) { if (m_vertices_info->contain_vector(obj)) return true; }
	if (m_DR_mass_symbol != NULL) { if (m_DR_mass_symbol->contain_vector(obj)) return true; }
	return false;
}

int CDiagramPhysics::getDiagramLineType(long pk_abs) { return 0; /* use default selection procedure */ }
long CDiagramPhysics::CollectVerticesForClass(long vnumClassVertex, CIntArray* idAltVerts) { return 0; }
bool CDiagramPhysics::AreCTPropagatorsVertices() { return true; }

CDiagramPhysics* CDiagramPhysics::get_curr_physics()
{
	return m_physics;
}

CExpression* CDiagramPhysics::getClassesHamiltonianE() { return NULL; } // NULL if not available
CExpression* CDiagramPhysics::getClassesHamiltonianRen() { return NULL; }; // NULL if not available

void CDiagramPhysics::install_diagram_physics(CDiagramPhysics* phys, bool delete_old) {
	if (m_physics != phys) {
		if ((delete_old) && (m_physics != NULL)) { delete m_physics; }
		m_physics = phys;
	}
	if (m_physics != NULL) m_physics->init_physics();
}

CEPtrArray* CDiagramPhysics::getVerticesInfo()
{
	return m_vertices_info;
}

CExpression* CDiagramPhysics::GetGeV2PbFactor() { return new CESymbol(CDiagramPhysics::m_SymGeV2PbConst); }

int CDiagramPhysics::getCharge3(long pk) { return INT_MIN; }
const char* CDiagramPhysics::get_particle_symbolA(long _kind, int sym_style) {
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if (phys==NULL) return "Error";
	return phys->get_particle_symbol(_kind, sym_style);
}

/*void CDiagramPhysics::add_new_vertex_info(CDiagramVertexInfo* vi) {
	if (vi != NULL) m_vertices_info->add(vi);
}*/

void CDiagramPhysics::init() {
	m_physics = NULL;
}

void CDiagramPhysics::finalize() {
	if (m_physics != NULL) delete m_physics; m_physics = NULL;
}

void CDiagramPhysics::set_DR_mass_PK(long pk) {
	if (pk < 0) pk = -pk;

	if (pk != 0) {
		bool is_pk_valid = validate_part_kind(pk);
		if (!is_pk_valid) {
			dbg->log(0, pk, "<ERROR> set_DR_mass_PK(&ld): invalid PK value. PK=0 is set");
			pk = 0;
		}
	}

	// delete old value
	if (m_DR_mass_symbol != NULL) { delete m_DR_mass_symbol; m_DR_mass_symbol = NULL; }

	// set new value
	m_DR_mass_PK = pk;
	if (pk != 0) {
		m_DR_mass_symbol = get_mass_sym(pk);
		if (m_DR_mass_symbol != NULL) {
			m_DR_mass_symbol = m_DR_mass_symbol->clone();
		} else {
			dbg->log(0, pk, "Warning: no mass symbol found for valid pk: %ld");
			m_DR_mass_symbol = new CESymbol("\\[Mu]");
		}
	} else {
		m_DR_mass_symbol = new CESymbol("\\[Mu]");
	}
}

CExpression* CDiagramPhysics::get_DR_mass_sym() {
	if (m_DR_mass_symbol==NULL) set_DR_mass_PK(0);
	return m_DR_mass_symbol;
}

long CDiagramPhysics::GetPDGParticleCode(long pk) { //default implementation should fit most cases
	long pk_abs = ((pk>=0)?pk:(-pk));
	long res = 0;

	if (pk_abs == CEParticle::PARTICLE_ELECTRON) res = 11;
	else if (pk_abs == CEParticle::PARTICLE_MUON) res = 13;
	else if (pk_abs == CEParticle::PARTICLE_TAU_LEPTON) res = 15;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_ELECTRON) res = 12;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_MUON) res = 14;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_TAU) res = 16;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_D) res = 1;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_U) res = 2;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_S) res = 3;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_C) res = 4;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_B) res = 5;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_T) res = 6;
	else if (pk_abs == CEParticle::PARTICLE_PHOTON) res = 22;
	else if (pk_abs == CEParticle::PARTICLE_Z_BOSON) res = 23;
	else if (pk_abs == CEParticle::PARTICLE_W_BOSON) res = 24;
	else if (pk_abs == CEParticle::PARTICLE_HIGGS) res = 25;
	//else if (pk_abs == CEParticle::PARTICLE_SCALAR_PHI) res = "\\phi";
	//else if (pk_abs == CEParticle::PARTICLE_SCALAR_CHI) res = "\\chi";
	//else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UA) res = "G_{\\gamma}";
	//else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UZ) res = "G_z";
	//else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWp) res = "G^{+}_{W}";
	//else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWm) res = "G^{-}_{W}";
	//else if (pk_abs == CEParticle::PARTICLE_FPGHOST) res = "u^x";
	else if (pk_abs == CEParticle::PARTICLE_PROTON) res = 2212;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRON) res = 2112;
	else if (pk_abs == CEParticle::PARTICLE_GLUON) res = 21;

	if(pk<0)res=-res;
	return res;
}

const char* CDiagramPhysics::GetClassPKTitle(long pkClass) {
	const char* res=NULL;
	long s=m_arrClassPKTitles.size();
	if (s != m_iIdClassPKTitles.GetSize()) { /* Error. Reset arrays. */ m_arrClassPKTitles.clear_all(); m_iIdClassPKTitles.ToSize(0); }
	long i;
	for(i=0;i<s;i++) {
		if (m_iIdClassPKTitles.GetAt(i)==pkClass) {
			CEString* str=(CEString*)m_arrClassPKTitles.GetAt(i);
			if (str!=NULL) { res=str->getData(); } else { res="NULL"; }
			break;
		}
	}
	if (res==NULL) {
		CEString* str=new CEString(); str->add_int(pkClass,true);
		res=str->getData();
		m_arrClassPKTitles.add(str);
		m_iIdClassPKTitles.Add(pkClass);
	}
	return res;
}
