////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// DiargamSet.cpp: implementation of the CDiagramSet class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CDiagramSet::dbg = new CDebug("CDiagramSet");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDiagramSet::CDiagramSet(CDiagramPhysics* phys) : CESerializable()
{
	m_kinHashID=0;
	m_physics = phys;
	m_diagram_list = new CEPtrArray();
	m_drawing_flags = CFeynDiagram::SET_LOOP_120 | CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES |
		CFeynDiagram::DRAW_CAPS_LINE_INFO | CFeynDiagram::DRAW_CAPS_SMALL;
}

CDiagramSet::CDiagramSet(CDiagramPhysics* phys, CEPtrArray* diagram_list, int _drawing_flags, DWORD _kinHashID) : CESerializable()
{
	//printf("new CDiagramSet\n");
	m_kinHashID=_kinHashID;
	m_physics = phys;
	m_diagram_list = CEPtrArray::get_copy(diagram_list, true);
	m_drawing_flags = _drawing_flags;
	//printf("new CDiagramSet OK\n");
}

CDiagramSet::~CDiagramSet()
{
	delete m_diagram_list;
}

const char* CDiagramSet::CLASS_NAME = "CDiagramSet";
const char* CDiagramSet::ClassName() const { return CLASS_NAME; }

int CDiagramSet::GetDrawingFlags() { return m_drawing_flags; }
void CDiagramSet::SetDrawingFlags(int _drawing_flags) { m_drawing_flags = _drawing_flags; }

CDiagramSet* CDiagramSet::getEmptyFrame() { return new CDiagramSet(m_physics, NULL, m_drawing_flags, 0); }

CDiagramSet* CDiagramSet::clone() const
{
	return new CDiagramSet(m_physics, m_diagram_list, m_drawing_flags, m_kinHashID);
}

CEObject* CDiagramSet::copy() const { return clone(); }

const char* CDiagramSet::CLASS_TAG_NAME = "DiagSet";
const char* CDiagramSet::get_TAG_Name() { return CLASS_TAG_NAME; }

long CDiagramSet::store_stream(ofstream* p_ofs, DWORD settings) {
	long res = 0;

	// save number to further check
	long s = m_diagram_list->size();
	res += CESupport::storeXMLLong(s, p_ofs, settings);

	for(long i=0; i<s; i++) {
		CFeynDiagram* diag = (CFeynDiagram*)m_diagram_list->GetAt(i);
		res += CESupport::store_stream(diag, p_ofs, settings);
		res += CESupport::storeLineBreak(p_ofs);
	}

	res += CESupport::storeXMLLong(m_drawing_flags, p_ofs, settings);
	res += CESupport::storeXMLLong(m_kinHashID, p_ofs, settings); // store kinenatics ID

	// store current physics ID (optional!)
	CDiagramPhysics* phys = m_physics; //CDiagramPhysics::get_curr_physics();
	if (phys!=NULL) {
		const char* _sPhysCN = phys->ClassName(); DWORD _dwPhysID = phys->getPhysicsID();
		if (_sPhysCN!=NULL) {
			CESymbol sPhysCN(_sPhysCN);
			res += CESupport::store_stream(&sPhysCN, p_ofs, settings);
			res += CESupport::storeXMLLong((long)_dwPhysID, p_ofs, settings);
		}
	}
	return res;
}

CESerializable* CDiagramSet::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	long num = 0;
	bool read_OK = CESupport::readXMLLong(data, &pos, &num, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	
	CEPtrArray* diagram_list = new CEPtrArray(false, num);
	for(long i=0; i<num; i++) {
		CESerializable* diag = CESupport::readXMLObj(data, &pos, CFeynDiagram::CLASS_NAME, settings, CLASS_NAME);
		if (diag == NULL) { delete diagram_list; return NULL; }
		diagram_list->add(diag);
	}

	long dr_flags = 0;
	read_OK = CESupport::readXMLLong(data, &pos, &dr_flags, settings, CLASS_NAME);
	if (!read_OK) return NULL;

	long _kinHashID=0;
	read_OK = CESupport::readXMLLong(data, &pos, &_kinHashID, settings, CLASS_NAME, true);
	if(read_OK) {
		CKinematics* kin = CEProcessor::kinematics();
		if ((kin==NULL) || (_kinHashID != kin->GetHashCode())) {
			dbg->log(0,"WARNING diagrams loading: Diagrams are loaded at DIFFERENT kinematics and can't be used.");
		}
	}

	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	CEObject* objPhysClName = CESupport::readXMLObj(data, &pos, CESymbol::CLASS_NAME, settings, CLASS_NAME, true);
	if (objPhysClName != NULL) {
		if (CEUtil::equals(CESymbol::CLASS_NAME, objPhysClName->ClassName())) { // extra check
			CESymbol* sPhysClName = (CESymbol*)objPhysClName;
			const char* _sPhysCN = sPhysClName->get_name();
			long lPhysID = 0;
			read_OK = CESupport::readXMLLong(data, &pos, &lPhysID, settings, CLASS_NAME);
			if ((read_OK) && (_sPhysCN!=NULL)) {
				if (phys != NULL) {
					if ((!CEUtil::equals(_sPhysCN,phys->ClassName())) || (lPhysID != phys->getPhysicsID())) {
						dbg->log(0, "ERROR: Loading diagrams for physics model different to current");
						CEString mess; mess.AddString(_sPhysCN); mess.AddString("-"); mess.add_int(lPhysID);
						dbg->log(0, "Loaded for model: ", mess.getData());
						mess.Reset(); mess.AddString(phys->ClassName()); mess.AddString("-"); mess.add_int(phys->getPhysicsID());
						dbg->log(0, "Current model: ", mess.getData());
						CEException::halt("Error loading diagrams");
					}
				} else {
					dbg->log(0, "ERROR: Loading diagrams for physics model different to current.");
					CEString mess; mess.AddString("Expected physics: "); mess.AddString(_sPhysCN); mess.AddString("-"); mess.add_int(lPhysID);
					dbg->log(0, "Loaded for model: ", mess.getData());
					dbg->log(0, "Current model: NULL");
					CEException::halt("Error loading diagrams");
				}
			}
		}
		delete objPhysClName;
	}

	//TODO: add PHYSICS settings STORE/CHECK
	CDiagramSet* res = new CDiagramSet(phys, diagram_list, dr_flags, _kinHashID);
	delete diagram_list;
	return res;
}

long CDiagramSet::GetKinHashID() { return m_kinHashID; }

void CDiagramSet::add_diagram(CFeynDiagram* diag) {
	if(diag!=NULL) {
		if ((m_kinHashID!=diag->GetKinHashID()) && (m_kinHashID!=0)) { dbg->log(0,"ERROR: CDiagramSet::add_diagram(DIAGRAM FOR DIFFERENT KINEMATICS)"); CEException::halt(); return; }
		if (m_kinHashID==0) m_kinHashID=diag->GetKinHashID();
		m_diagram_list->add(diag);
	}
}

void CDiagramSet::add_diagrams(CDiagramSet* diags) {
	if (diags==NULL) return;
	if ((m_kinHashID!=diags->GetKinHashID()) && (m_kinHashID!=0)) { dbg->log(0,"ERROR: CDiagramSet::add_diagrams(DIAGRAMS FOR DIFFERENT KINEMATICS)"); CEException::halt(); return; }
	if (diags->m_physics != m_physics) { dbg->log(0, "add_diagrams ERROR: can't combine diagrams for different physics models"); CEException::halt(); return; }
	if (m_kinHashID==0) m_kinHashID=diags->GetKinHashID();
	long s=diags->get_diagrams_number();
	for(long i=0; i<s; i++) { CFeynDiagram* d = diags->get_diagram(i); if(d!=NULL)add_diagram(d->clone()); }
}

long CDiagramSet::get_diagrams_number() { return m_diagram_list->size(); }

CFeynDiagram* CDiagramSet::get_diagram(long i) {
	if ((i<0) || (i>=get_diagrams_number())) return NULL;
	return (CFeynDiagram*)m_diagram_list->GetAt(i);
}

void CDiagramSet::remove_diagram(long i) {
	if ((i<0) || (i>=get_diagrams_number())) return;
	m_diagram_list->RemoveAt(i, true);
}

CEString* CDiagramSet::toString(int mode) {
	CEString* res = NULL;
	switch (mode) {
	case STORE_WR_MATHEMATICA:
		{
			long N=m_diagram_list->size();
			res=new CEString("\\({CDiagramSet(N="); res->add_int(N);
			res->AddString(", kinHash="); res->add_int(m_kinHashID);
			res->AddString(")} ");
			for(long i=0; i<N; i++) {
				if (i>0) res->AddString(", ");
				res->AddString("\\(");
				res->AddString(get_diagram(i)->toString(mode));
				res->AddString("\\)");
			}
			res->AddString("\\)");
		}
		break;
	case STORE_TEX:
		// draw diagram with default settings
		res = draw(GetDrawingFlags());
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

CEString* CDiagramSet::draw(int settings)
{
	int s = m_diagram_list->size();

	unsigned page_width = 400;
	unsigned page_height = 600;

	unsigned diag_width = 100;
	unsigned diag_height = 100;

	int diag_per_width = CLMath::Trunc(((double)page_width)/diag_width);
	int diag_per_height = CLMath::Trunc(((double)page_height)/diag_height);
	
	CEString* res = new CEString("");
	//res->AddString("\\newpage \n");
	res->AddString("Total ");
	res->add_int(s);
	res->AddString(" diagrams.\n\n");

	int diag_no = 0;
	for(;(diag_no < s);) {
		unsigned next_page_raw_num = diag_per_height;
		int diags_left = s-diag_no;
		if (diags_left < diag_per_width*diag_per_height) {
			unsigned diags_in_last_raw = diags_left%diag_per_width;
			next_page_raw_num = (diags_left-diags_in_last_raw)/diag_per_width;
			if (diags_in_last_raw > 0) next_page_raw_num++;
		}
		unsigned next_page_height = diag_height*next_page_raw_num;

		//res  = new CEString("\\begin{figure}\n\\begin{picture}(");
		res->AddString("\\begin{picture}(");
		res->add_int(page_width); res->AddString(", "); res->add_int(next_page_height);
		res->AddString(")\n");

		int i;
		for(i=0; ((i<diag_per_height) && (diag_no < s)); i++) {
			unsigned y_offset = (next_page_raw_num-i-1)*diag_height;

			int j;
			for(j=0; ((j<diag_per_width) && (diag_no < s)); j++) {
				CFeynDiagram* diag = (CFeynDiagram*)m_diagram_list->GetAt(diag_no); diag_no++;

				res->AddString("\\SetOffset(");
				res->add_int(j*diag_width); res->AddString(", "); res->add_int(y_offset);
				res->AddString(")\n");

				// scale diagram to make borders
				CFeynDiagram* diag_copy = diag->clone();
				CLDiagLocateView::scale_diagram(diag_copy, 0.7, NULL);
				CEString* diag_str = diag_copy->draw(diag_width, diag_height, settings | CFeynDiagram::DRAW_NO_ENVIRONMENT);
				delete diag_copy;
				if (diag_str == NULL) { diag_str = new CEString("\nError\n"); }

				if (settings & CFeynDiagram::DRAW_DIAGRAM_NUMBERS) {
					diag_str->AddString("\\Text(");
					diag_str->add_int(diag_width*7/20); diag_str->AddString(", "); diag_str->add_int(1); diag_str->AddString(")[c]{");
					diag_str->AddString("\\tiny "); // \small, \footnotesize, \scriptsize, \tiny
					diag_str->add_int(diag_no-1);
					diag_str->AddString(".}");
				}
				res->AddString(diag_str); res->AddString("\n");
			}
		}

		//res->AddString("\\end{picture}\n\\end{figure}\n\n");
		res->AddString("\\end{picture}\n\n");
	}

	return res;
}

bool CDiagramSet::LocateDiagramsView(int Algorithm_ID)
{
	dbg->log(20, Algorithm_ID, GetDrawingFlags(), "LocateDiagramsView. Algorithm == %ld. Drawing flags: %x");
	dbg->log(20, get_diagrams_number(), "Total %ld diagrams to locate.");

	if (Algorithm_ID < 0) Algorithm_ID = CLDiagLocateView::ALGORITHM_DEFAULT;
	bool res = true;
	long s = get_diagrams_number();
	long i;
	for (i=0; i<s ; i++) {
		CFeynDiagram* diag = (CFeynDiagram*)m_diagram_list->GetAt(i);
		bool locate_OK = CLDiagLocateView::LocateDiagramElements(diag, Algorithm_ID, GetDrawingFlags());
		if (!locate_OK) res = false;
	}

	if(res){ dbg->log(20, "Diagrams view locate : successfull."); }else{ dbg->log(0, "Diagrams view locate : failed."); }
	return res;
}

bool CDiagramSet::ArrangeMomenta() {
	dbg->log(20, get_diagrams_number(), "ArrangeMomenta (Total %ld diagrams).");
	CKinematics* kin = CEProcessor::kinematics();
	if (kin==NULL) { dbg->log(0,"ERROR: CDiagramSet::ArrangeMomenta(NULL kinematics) call"); CEException::halt(); return false; }
	if ((m_kinHashID!=kin->GetHashCode())&&(m_kinHashID!=0)) { dbg->log(0,"CDiagramSet::ArrangeMomenta() ERROR: Can't operate with diagram for different kinematics"); CEException::halt(); return false; }

	bool res = true;
	long s = get_diagrams_number();
	long i;
	for (i=0; i<s ; i++) {
		CFeynDiagram* diag = (CFeynDiagram*)m_diagram_list->GetAt(i);
		bool diag_OK = diag->ArrangeMomenta(kin);
		if (!diag_OK) res = false;
	}

	dbg->log(30, (res)?1:0, "Momenta arranged (res == %ld).");
	return res;
}

CMatrixElement* CDiagramSet::RetrieveMatrixElement() {
	dbg->log(20, get_diagrams_number(), "RetrieveMatrixElement (Total %ld diagrams).");

	CKinematics* kin = CEProcessor::kinematics();
	if (kin==NULL) { dbg->log(0,"ERROR: CDiagramSet::RetrieveMatrixElement(NULL kinematics) call"); CEException::halt(); return false; }
	if ((m_kinHashID!=kin->GetHashCode())&&(m_kinHashID!=0)) { dbg->log(0,"CDiagramSet::RetrieveMatrixElement() ERROR: Can't operate with diagram for different kinematics"); CEException::halt(); return false; }
	CMatrixElement* res = NULL;
	long s = get_diagrams_number();

	if (s>0) {
		kin->reset_volatile_variables(true);
		long i;
		for (i=0; i<s ; i++) {
			CFeynDiagram* diag = (CFeynDiagram*)m_diagram_list->GetAt(i);
			dbg->log(70, i, "Diagram no. %ld processing");
			CMatrixElement* sub_res = diag->RetrieveMatrixElement();
			if (sub_res != NULL) {
				if (res == NULL) res = sub_res;
				else {
					bool bAddOK=res->addME(sub_res);
					if (!bAddOK) {
						//dbg->log(0,"ERROR: Retrieving matrix element from incompartible diagrams.");
						dbg->log(0,"Please, split incompartible diagrams before this procedure.");
						dbg->log(0,i,"ERROR: Diagram to be separated: #%ld.");
						CEException::halt(NULL,__FILE__,__LINE__);
					}
				}
			}
			else {
				// declare error.
				dbg->log(0, i, "ERROR:  NULL Matrix Element retrieved from diagram (no. %ld): ", diag);
			}
		}
		kin->reset_volatile_variables(false);
	} else {
		// create NULL matrix element (expr:0, others: NULL)
		CENumber zero((long)0);
		res = new CMatrixElement(&zero, NULL, NULL);
	}

	dbg->log(30, "Matrix Element retrieved:", res);
	return res;
}
