////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// FeynDiagram.cpp: implementation of the CFeynDiagram class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const int CFeynDiagram::DRAW_NO_ENVIRONMENT			= 0x0001;
const int CFeynDiagram::DRAW_IN_DEFAULT_BOX			= 0x0002;

const int CFeynDiagram::DRAW_CAPS_SMALL				= 0x0008;

const int CFeynDiagram::SET_LOOP_90					= 0x0010;
const int CFeynDiagram::SET_LOOP_120				= 0x0020;

const int CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES	= 0x0040;
const int CFeynDiagram::DRAW_CAPS_LINE_INFO			= 0x0080;
const int CFeynDiagram::DRAW_CAPS_MOMENTA			= 0x0100;
const int CFeynDiagram::DRAW_DIAGRAM_NUMBERS		= 0x0200; // no of diagram in DiagramSet
const int CFeynDiagram::DRAW_SWAP_TALES				= 0x0400;
const int CFeynDiagram::DRAW_DONT_SWAP_TALES		= 0x0800;
const int CFeynDiagram::DRAW_CAPS_PARTICLE_SIGNS	= 0x1000;

const char* CFeynDiagram::CLASS_NAME = "CFeynDiagram";
CDebug* CFeynDiagram::dbg = new CDebug("CFeynDiagram");

CFeynDiagram* CFeynDiagram::create_blanc(CDiagramPhysics* _physics,
										 bool create_EXT_lines,
										 CKinematics* kin)
{
	if (_physics == NULL) return NULL;

	// Unimplemented case: we need to fill _kinHashID afterwards.
	if (kin==NULL) { dbg->log(0,"ERROR: unexpected CFeynDiagram::create_blanc(,,NULL kinematics) call"); CEException::halt(); }

	CEPtrArray* lines_list = NULL;
	long _kinHashID=0; if(kin!=NULL)_kinHashID=kin->GetHashCode();
	if (create_EXT_lines) {
		if (kin == NULL) return NULL;
		lines_list = new CEPtrArray();

		// create EXT line for each particle
		int num_init = kin->num_init_particles();
		int i;
		for(i=0; i<num_init; i++) {
			CEParticle* part = kin->pin(i);
			lines_list->add(new CDExtLine(part, true, NULL));
		}
		
		int num_final = kin->num_final_particles();
		for (i=0; i<num_final; i++) {
			CEParticle* part = kin->pfin(i);
			lines_list->add(new CDExtLine(part, false, NULL));
		}
	}
	CFeynDiagram* res = new CFeynDiagram(_physics, NULL, lines_list, NULL, NULL, _kinHashID, -1);
	delete lines_list;
	return res;
}

CFeynDiagram::CFeynDiagram(CDiagramPhysics* _physics,
		CEPtrArray* _vertices_list,
		CEPtrArray* _lines_list,
		CEPtrArray* _virt_momenta_list,
		CExpression* _factor_expr,
		long _kinHashID,
		int locate_algorithm) : CESerializable()
{
	m_kinHashID=_kinHashID;
	m_physics = _physics;
	m_locate_algorithm = locate_algorithm;
	m_virt_momenta_list = CEPtrArray::get_copy(_virt_momenta_list, true);
	m_ext_vector_indices_list = NULL;

	/*m_draw_box_size = draw_box_size;
	if (m_draw_box_size == NULL) {
		// default draw box (100 X 100)
		m_draw_box_size = new CPos2D(100, 100, 100, 100);
	}*/

	m_vertices_list = CEPtrArray::get_copy(_vertices_list, true);
	m_lines_list = CEPtrArray::get_copy(_lines_list, true);
	m_factor_expr = (_factor_expr == NULL)?NULL:(_factor_expr->clone());
}

CFeynDiagram::~CFeynDiagram()
{
	delete m_vertices_list;
	delete m_lines_list;
	delete m_virt_momenta_list;
	if (m_ext_vector_indices_list != NULL) delete m_ext_vector_indices_list;
	if (m_factor_expr != NULL) delete m_factor_expr;
	//delete m_draw_box_size;
}

bool CFeynDiagram::contain_vector(void * obj) {
	if ((m_vertices_list == obj) ||
		(m_lines_list == obj) ||
		(m_virt_momenta_list == obj) ||
		(m_ext_vector_indices_list == obj) ||
		(m_factor_expr == obj)) return true;
	if (m_vertices_list != NULL) { if(m_vertices_list->contain_vector(obj)) return true; }
	if (m_lines_list != NULL) { if(m_lines_list->contain_vector(obj)) return true; }
	if (m_virt_momenta_list != NULL) { if(m_virt_momenta_list->contain_vector(obj)) return true; }
	if (m_ext_vector_indices_list != NULL) { if(m_ext_vector_indices_list->contain_vector(obj)) return true; }
	if (m_factor_expr != NULL) { if(m_factor_expr->contain_vector(obj)) return true; }
	return false;
}

const char* CFeynDiagram::ClassName() const { return CLASS_NAME; }
int CFeynDiagram::get_locate_algorithm() { return m_locate_algorithm; }
void CFeynDiagram::set_locate_algorithm(int locate_algorithm) { m_locate_algorithm = locate_algorithm; }
CEPtrArray* CFeynDiagram::get_lines_list() {return m_lines_list;}
CEPtrArray* CFeynDiagram::get_vertices_list() {return m_vertices_list;}
int CFeynDiagram::get_lines_number() {return m_lines_list->size();}
int CFeynDiagram::get_vertices_number() {return m_vertices_list->size();}
CDLine* CFeynDiagram::get_line(int n) {
	if ((n < 0) || (n >= m_lines_list->size())) return NULL;
	return (CDLine*)m_lines_list->GetAt(n);
}
CDVertex* CFeynDiagram::get_vertex(int n) {
	if ((n < 0) || (n >= m_vertices_list->size())) return NULL;
	return (CDVertex*)m_vertices_list->GetAt(n);
}
void CFeynDiagram::add_line(CDLine* l) {m_lines_list->add(l);}
void CFeynDiagram::add_vertex(CDVertex* v) {m_vertices_list->add(v);}

int CFeynDiagram::get_vertex_pos(CDVertex* v) {
	if (v == NULL) return -1;
	int s = get_vertices_number();
	int i;
	for(i=0; i<s; i++) { if (get_vertex(i) == v) return i; }
	return -1;
}

int CFeynDiagram::get_line_pos(CDLine* l) {
	if (l == NULL) return -1;
	int s = get_lines_number();
	int i;
	for(i=0; i<s; i++) { if (get_line(i) == l) return i; }
	return -1;
}

CExpression* CFeynDiagram::get_factor_expr() { return m_factor_expr; }
void CFeynDiagram::set_factor_expr(CExpression* val) { // does not affects patrameter
	if (m_factor_expr == val) return;
	CExpression* old_expr=m_factor_expr;
	m_factor_expr = ((val!=NULL)?val->clone():NULL);
	if (old_expr != NULL) delete old_expr;
}

long CFeynDiagram::GetKinHashID() { return m_kinHashID; }

CEObject* CFeynDiagram::copy() const { return clone(); }
CFeynDiagram* CFeynDiagram::clone() const {
	CFeynDiagram* res = new CFeynDiagram(m_physics,
		m_vertices_list, m_lines_list,
		m_virt_momenta_list,
		m_factor_expr,
		m_kinHashID,
		m_locate_algorithm);

	// links are volatile and we clone them manually
	CEPtrArray* _lines = res->get_lines_list();
	CEPtrArray* _vertices = res->get_vertices_list();
	bool copy_OK = CDElement::copy_links(_vertices, m_vertices_list, _lines, m_lines_list);
	if (!copy_OK) {throw(new CEException(CEException::INT_ERROR));}
	copy_OK = CDElement::copy_links(_lines, m_lines_list, _vertices, m_vertices_list);
	if (!copy_OK) {throw(new CEException(CEException::INT_ERROR));}

	return res;
}

const char* CFeynDiagram::CLASS_TAG_NAME = "diag";
const char* CFeynDiagram::get_TAG_Name() { return CLASS_TAG_NAME; }

long CFeynDiagram::store_stream(ofstream* p_ofs, DWORD settings) {
	long res = 0;

	// arrays of CDElement objects
	//CEPtrArray* m_vertices_list;
	//CEPtrArray* m_lines_list;
	int v_num = m_vertices_list->size();
	int l_num = m_lines_list->size();
	int i;
	// store vertices & lines without links
	res += CESupport::storeXMLLong(v_num, p_ofs, settings);
	for(i=0; i<v_num; i++) {
		CDElement* v = (CDElement*)m_vertices_list->GetAt(i);
		res += CESupport::store_stream(v, p_ofs, settings);
	}
	res += CESupport::storeXMLLong(l_num, p_ofs, settings);
	for(i=0; i<l_num; i++) {
		CDElement* l = (CDElement*)m_lines_list->GetAt(i);
		res += CESupport::store_stream(l, p_ofs, settings);
	}

	// store links data
	for(i=0; i<v_num; i++) {
		CDElement* v = (CDElement*)m_vertices_list->GetAt(i);

		// store links as numbers
		CEPtrArray* links = v->get_links();
		int links_num = links->size();
		res += CESupport::storeXMLLong(links_num, p_ofs, settings);
		int links_i;
		for (links_i=0; links_i<links_num; links_i++) {
			CEObject* link_curr = links->GetAt(links_i);
			int ref_no = -1;
			for(int tmp_i=0; tmp_i<l_num; tmp_i++) { if (m_lines_list->GetAt(tmp_i) == link_curr) { ref_no=tmp_i; break; } }
			if (ref_no < 0) { dbg->log(0, "CFeynDiagram::store_stream() ERROR, INVALID diagram structure", this); CEException::halt(); }
			res += CESupport::storeXMLLong(ref_no, p_ofs, settings);
		}
	}
	for(i=0; i<l_num; i++) {
		CDElement* l = (CDElement*)m_lines_list->GetAt(i);

		// store links as numbers
		CEPtrArray* links = l->get_links();
		int links_num = links->size();
		res += CESupport::storeXMLLong(links_num, p_ofs, settings);
		int links_i;
		for (links_i=0; links_i<links_num; links_i++) {
			CEObject* link_curr = links->GetAt(links_i);
			int ref_no = -1;
			for(int tmp_i=0; tmp_i<v_num; tmp_i++) { if (m_vertices_list->GetAt(tmp_i) == link_curr) { ref_no=tmp_i; break; } }
			if (ref_no < 0) { dbg->log(0, "CFeynDiagram::store_stream() ERROR, INVALID diagram structure", this); CEException::halt(); }
			res += CESupport::storeXMLLong(ref_no, p_ofs, settings);
		}
	}

	// virtual momenta list
	int s = m_virt_momenta_list->size();
	res += CESupport::storeXMLLong(s, p_ofs, settings);
	for(i=0; i<s; i++) { res += CESupport::store_stream((CExpression*)m_virt_momenta_list->GetAt(i), p_ofs, settings); }

	//m_ext_vector_indices_list: volatile

	// locate info
	// Note: currently position is NOT stored with diagrams ==>
	//		==> don't save 'm_locate_algorithm' for proper call of 'LocateDiagramsView' on restored graphs
	res += CESupport::storeXMLLong(-1, p_ofs, settings);
	//[wrong]res += CESupport::storeXMLLong(m_locate_algorithm, p_ofs, settings);

	if (m_factor_expr != NULL) {  // NULL means "*1" factor
		res += CESupport::store_stream(m_factor_expr, p_ofs, settings);
	} else {
		CENumber unit((long)1); res += CESupport::store_stream(&unit, p_ofs, settings);
	}

	res += CESupport::storeXMLLong(m_kinHashID, p_ofs, settings);

	return res;
}

CESerializable* CFeynDiagram::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {

	long i;
	long v_num = 0;
	bool read_OK = CESupport::readXMLLong(data, &pos, &v_num, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	CEPtrArray* v_list = new CEPtrArray();
	for(i=0; i<v_num; i++) {
		CESerializable* v = CESupport::readXMLObj(data, &pos, CDVertex::CLASS_NAME, settings, CLASS_NAME);
		if (v == NULL) { delete v_list; return NULL; }
		v_list->add(v);
	}

	long l_num = 0;
	read_OK = CESupport::readXMLLong(data, &pos, &l_num, settings, CLASS_NAME);
	if (!read_OK) return NULL;
	CEPtrArray* l_list = new CEPtrArray();
	for(i=0; i<l_num; i++) {
		CESerializable* l = CESupport::readXMLObj(data, &pos, NULL, settings, CLASS_NAME);
		if (l == NULL) { delete v_list; delete l_list; return NULL; }
		const char* l_CN = l->ClassName();
		if (!(
			(CEUtil::equals(l_CN, CDIntLine::CLASS_NAME)) || (CEUtil::equals(l_CN, CDExtLine::CLASS_NAME))
			)) {
			dbg->log(0, "CFeynDiagram::restore() ERROR, LINE object expected, read object: ", l);
			delete v_list; delete l_list; CEException::halt();
		}
		l_list->add(l);
	}

	for(i=0; i<v_num; i++) {
		CEPtrArray* links = ((CDVertex*)v_list->GetAt(i))->get_links();
		long links_num = 0;
		read_OK = CESupport::readXMLLong(data, &pos, &links_num, settings, CLASS_NAME);
		if (!read_OK) { delete v_list; delete l_list; return NULL; }
		long j;
		for(j=0; j<links_num; j++) {
			long link_no = 0;
			read_OK = CESupport::readXMLLong(data, &pos, &link_no, settings, CLASS_NAME);
			if (!read_OK) { delete v_list; delete l_list; return NULL; }
			if (link_no >= l_num) { dbg->log(0, "CFeynDiagram::restore() ERROR, Invalid links structure loaded"); delete v_list; delete l_list; CEException::halt(); }
			links->add(l_list->GetAt(link_no));
		}
	}
	for(i=0; i<l_num; i++) {
		CEPtrArray* links = ((CDLine*)l_list->GetAt(i))->get_links();
		long links_num = 0;
		read_OK = CESupport::readXMLLong(data, &pos, &links_num, settings, CLASS_NAME);
		if (!read_OK) { delete v_list; delete l_list; return NULL; }
		long j;
		for(j=0; j<links_num; j++) {
			long link_no = 0;
			read_OK = CESupport::readXMLLong(data, &pos, &link_no, settings, CLASS_NAME);
			if (!read_OK) { delete v_list; delete l_list; return NULL; }
			if (link_no >= v_num) { dbg->log(0, "CFeynDiagram::restore() ERROR, Invalid links structure loaded"); delete v_list; delete l_list; CEException::halt(); }
			links->add(v_list->GetAt(link_no));
		}
	}

	// virtual momenta list
	long virt_size = 0;
	read_OK = CESupport::readXMLLong(data, &pos, &virt_size, settings, CLASS_NAME);
	if (!read_OK) { delete v_list; delete l_list; return NULL; }
	CEPtrArray* virt_mom_list = new CEPtrArray();
	CEReplList* repl = new CEReplList();
	for(i=0; i<virt_size; i++) {
		CExpression* virt_mom = CESupport::readXMLExpr(data, &pos, settings, CLASS_NAME);
		if (virt_mom == NULL) { delete virt_mom_list; delete v_list; delete l_list; return NULL; }
		
		// Note: virtual momenta are UNIQUE symbols, but UNIQUE numration is new now
		if (!CEUtil::equals(virt_mom->ClassName(), CESymbol::CLASS_NAME)) {
			dbg->log(0, "CFeynDiagram::restore() ERROR: non-SYMBOL virtual momenta read", virt_mom);
		}
		CExpression* vmom_new = CESymbol::getUnique();
		repl->add(virt_mom, vmom_new, false);
		delete virt_mom;
		virt_mom_list->add(vmom_new);
	}

	// replace UNIQUES in internal momenta
	for(i=0; i<l_num; i++) {
		CDLine* l = (CDLine*)l_list->GetAt(i);
		if (l->is_external()) continue;
		((CDIntLine*)l)->applyRuleToMomentum(repl);
	}
	delete repl;

	// locate info
	long _locate_algorithm = -1;
	read_OK = CESupport::readXMLLong(data, &pos, &_locate_algorithm, settings, CLASS_NAME);
	if (!read_OK) { delete virt_mom_list; delete v_list; delete l_list; return NULL; }

	// Note: now factor is saved obligatory
	// Keep 'readXMLExpr(...true)' for some time for old versions support
	CExpression* _factor = CESupport::readXMLExpr(data, &pos, settings, CLASS_NAME, true);
	// Let NULL be a zero factor
	if (_factor!=NULL) { if (CEBasic::isUnit(_factor)) { delete _factor; _factor=NULL; } }

	// Note: now _kinHashID is saved obligatory
	// Keep 'readXMLLong(...true)' for some time for old versions support
	long _kinHashID = 0;
	read_OK = CESupport::readXMLLong(data, &pos, &_kinHashID, settings, CLASS_NAME, true);

	read_OK = CESupport::checkAllRead(data, &pos, settings, CLASS_NAME);
	if (!read_OK) {
		if (_factor != NULL) delete _factor;
		delete virt_mom_list; delete v_list; delete l_list; return NULL;
	}
	
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();

	//TODO: add PHYSICS settings STORE/CHECK
	CFeynDiagram* res = new CFeynDiagram(phys,
		v_list,
		l_list,
		virt_mom_list,
		_factor,
		_kinHashID,
		_locate_algorithm);
	if (_factor != NULL) delete _factor;
	delete virt_mom_list;

	// links are volatile and we clone them manually
	CEPtrArray* _lines = res->get_lines_list();
	CEPtrArray* _vertices = res->get_vertices_list();
	bool copy_OK = CDElement::copy_links(_vertices, v_list, _lines, l_list);
	if (!copy_OK) { CEException::halt(NULL,__FILE__,__LINE__); }
	copy_OK = CDElement::copy_links(_lines, l_list, _vertices, v_list);
	if (!copy_OK) { CEException::halt(NULL,__FILE__,__LINE__); }
	
	delete v_list; delete l_list; 
	return res;
}

/*
*	in_order < 0 means check up validity in any order
*	note: order may be greather than vertices number due to counter terms presence
*/
bool CFeynDiagram::is_valid(int in_order) {

	int v_num = get_vertices_number();
	bool calc_order = (in_order >= 0);
	int diag_order = 0;
	//int CT_num = 0; // number of counterterms
	if (calc_order) { if (v_num > in_order) return false; }

	// check all the elements to be binded
	int l_num = get_lines_number();
	int i;
	for(i=0; i<l_num; i++) {
		CDLine* l = get_line(i);
		if (!l->is_binded()) return false;
		if (calc_order) { if (((l->DE_kind()) & CDElement::EL_COUNTER_TERM) != 0) diag_order+=2; }
	}
	for(i=0; i<v_num; i++) {
		CDVertex* v = get_vertex(i);
		if (!v->is_binded()) return false;
		if (calc_order) { diag_order += v->get_vertex_info()->get_vertex_order(); }
		//if (calc_order) { if (((v->DE_kind()) & CDElement::EL_COUNTER_TERM) != 0) CT_num++; }
	}

	// check order
	//if (calc_order) { if ((v_num + 2*CT_num) != in_order) return false; }
	if (calc_order) { if (diag_order != in_order) return false; }

	// check also all the target links to be in the same diagram?
	return true;
}

CEString* CFeynDiagram::toString(int mode) {
	CEString* res = NULL;
	int i = 0;
	int s = 0;
	switch (mode) {
	case STORE_WR_MATHEMATICA:
		res  = new CEString("\\({CFeynDiagram:");
		if (m_factor_expr != NULL) {
			res->AddString(" factor = ");
			res->AddString(m_factor_expr->toString(mode));
			res->AddString(",");
		}
		res->AddString(" lines num =");
		s = get_lines_number();
		res->add_int(s);
		for (i=0; i<s; i++) {
			res->AddString(", ");
			res->add_int(i);
			res->AddString(": ");
			res->AddString(get_line(i)->toString(mode));
			addBindingInfo(res, get_line(i));
		}

		res->AddString(" , vertices num =");
		s = get_vertices_number();
		res->add_int(s);
		for (i=0; i<s; i++) {
			res->AddString(", ");
			res->add_int(i);
			res->AddString(": ");
			res->AddString(get_vertex(i)->toString(mode));
			addBindingInfo(res, get_vertex(i));
		}

		res->AddString(" , locate algorithm ="); res->add_int(get_locate_algorithm());
		res->AddString(" , is binded = "); res->add_int(is_valid());
		res->AddString(" , kin-hashID = "); res->add_int(m_kinHashID);
		res->AddString("}\\)");
		break;
	case STORE_TEX:
		// draw diagram in default box
		res = draw(100, 100, DRAW_IN_DEFAULT_BOX |
			DRAW_CAPS_PARTICLE_TITLES | DRAW_CAPS_LINE_INFO | DRAW_CAPS_SMALL);
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

void CFeynDiagram::addBindingInfo(CEString* str, CDElement* el) {
	if ((str == NULL) || (el == NULL)) return;
	CEPtrArray* obj_arr = NULL;
	int el_kind = el->DE_kind();
	if (el_kind & CDElement::EL_LINE) { obj_arr = get_vertices_list(); }
	else { obj_arr = get_lines_list(); }
	CEPtrArray* links_arr = el->get_links();
	if ((obj_arr == NULL) || (links_arr == NULL)) { str->AddString("{?}"); return; }
	int s = links_arr->size();
	int obj_s = obj_arr->size();
	if (el_kind & CDElement::EL_LINE) str->AddString("{linked vertices#: ");
	else str->AddString("{linked lines#: ");
	int i;
	for(i=0; i<s; i++) {
		CEObject* item = links_arr->GetAt(i);
		bool link_found = false;
		int j;
		for(j=0; j<obj_s; j++) {
			if (item == obj_arr->GetAt(j)) {
				str->add_int(j);
				link_found = true; break;
			}
		}
		if (!link_found) str->AddString("?");
		if (i<(s-1)) str->AddString(", ");
	}
	str->AddString("}");
}

//////////////////////////////////////////////////////////////////////
// Storing diagram to TeX string in terms of AxoDraw package.
//////////////////////////////////////////////////////////////////////
CEString* CFeynDiagram::draw(unsigned width, unsigned height, int settings) {
	CFeynDiagram* old_diag_drawing = CLDrawDiagView::SetCurrentDiagram(this);

	// check up parameters
	unsigned size_min = 30;
	if (height < size_min) height = size_min;
	if (width < size_min) width = size_min;

	if (get_locate_algorithm() < 0) {
		// try to locate diagram elements
		if (!CLDiagLocateView::LocateDiagramElements(this, CLDiagLocateView::ALGORITHM_DEFAULT, 0))
			CLDrawDiagView::SetCurrentDiagram(old_diag_drawing);
			return NULL;
	}

	// assign integer positions to all the diagram elements
	set_diagram_size(width, height);

	CEString* res = NULL;

	if (settings & DRAW_NO_ENVIRONMENT) {
		res  = new CEString("\n");
	} else {
		//res  = new CEString("\\begin{figure}\n\\begin{picture}(");
		res  = new CEString("\\begin{picture}(");
		res->add_int(width); res->AddString(", "); res->add_int(height);
		res->AddString(")\n");
	}

	int s = get_lines_number();
	int i;
	for(i=0; i<s; i++) {
		CDLine* l = get_line(i); if (l==NULL) continue;
		CEString* el_str = l->draw(settings);
		if (el_str != NULL) {
			res->AddString(el_str); res->AddString("\n");
		}
	}


	s = get_vertices_number();
	for (i=0; i<s; i++) {
		CEString* el_str = get_vertex(i)->draw(settings);
		if (el_str != NULL) {
			res->AddString(el_str); res->AddString("\n");
		}
	}

	// print factor
	if (m_factor_expr != NULL) {
		res->AddString("\\Text(");
		res->add_int(3); res->AddString(", "); res->add_int(height*7/20); res->AddString(")[c]{");
		res->AddString("\\tiny "); // \small, \footnotesize, \scriptsize, \tiny
		res->AddString("$");
		res->AddString(m_factor_expr->toString(STORE_TEX));
		res->AddString("$}");
	}

	if ((settings & DRAW_NO_ENVIRONMENT) == 0) {
		//res->AddString("\\end{picture}\n\\end{figure}\n\n");
		res->AddString("\\end{picture}\n\n");
	}

	CLDrawDiagView::SetCurrentDiagram(old_diag_drawing);
	return res;
}

void CFeynDiagram::clear_temporary_data(bool do_delete) {

	int s = get_lines_number();
	int i;
	for(i=0; i<s; i++) {
		CDElement* item = get_line(i);
		CESerializable* tmp_val = item->m_temporary_mark;
		if ((tmp_val != NULL) && (do_delete)) delete tmp_val;
		item->m_temporary_mark = NULL;
	}
	
	s = get_vertices_number();
	for (i=0; i<s; i++) {
		CDElement* item = get_vertex(i);
		CESerializable* tmp_val = item->m_temporary_mark;
		if ((tmp_val != NULL) && (do_delete)) delete tmp_val;
		item->m_temporary_mark = NULL;
	}
}

void CFeynDiagram::clear_locate_data(bool _lines, bool _vertices) {
	int s; int i;
	if (_lines) {
		s = get_lines_number();
		for(i=0; i<s; i++) {
			get_line(i)->setViewInfo(NULL);
		}
	}

	if (_vertices) {
		s = get_vertices_number();
		for (i=0; i<s; i++) {
			get_vertex(i)->setViewInfo(NULL);
		}
	}

	if ((_lines)&&(_vertices)) set_locate_algorithm(-1);
}

/*
*	Assigns integer positions to all the diagram elements.
*/
void CFeynDiagram::set_diagram_size(unsigned width, unsigned height) {

	// check up parameters
	unsigned size_min = 30;
	if (height < size_min) height = size_min;
	if (width < size_min) width = size_min;

	int s = get_vertices_number();
	int i;
	for(i=0; i<s; i++) {
		CDElementViewInfo* pos_info = get_vertex(i)->getViewInfo();
		if (pos_info != NULL) {
			CPos2D* pos = pos_info->get_pos();
			if (pos != NULL) {
				pos->set_box_size(width, height);
			}
		}
	}
	
	s = get_lines_number();
	for(i=0; i<s; i++) {
		CDElementViewInfo* pos_info = get_line(i)->getViewInfo();
		if (pos_info != NULL) {
			CPos2D* pos = pos_info->get_pos();
			if (pos != NULL) {
				pos->set_box_size(width, height);
			}
		}
	}
}

bool CFeynDiagram::ArrangeMomenta(CKinematics* kin) {
	dbg->log(35, "ArrangeMomenta.");
	if (kin==NULL) { dbg->log(0,"ERROR: CFeynDiagram::ArrangeMomenta(NULL) call"); CEException::halt(); return false; }
	if ((m_kinHashID != kin->GetHashCode())&&(m_kinHashID!=0)) { dbg->log(0,"WARNING: CFeynDiagram::ArrangeMomenta(DIFFERENT KINEMATICS) call"); return false; }
	//dbg->log(60, "diagram:", this);

	for(;;) {
		int lines_arranged_this_turn = 0;
		int lines_unarranged_left = 0;

		int v_s = get_vertices_number();
		int i;
		for(i=0; i<v_s; i++) {
			int vert_res = get_vertex(i)->ArrangeMomenta(kin);
			if (vert_res > 0) lines_arranged_this_turn++;
			else if (vert_res < 0) lines_unarranged_left++;
		}
		
		if (lines_unarranged_left == 0) break;
		
		if (lines_arranged_this_turn == 0) {
			// no more lines can be arranged
			// involve new virtual momentum
			dbg->log(45, "ArrangeMomenta: involve virtual momentum.");

			// find out useful unarranged line
			CDLine* l = NULL;
			int l_pk_abs = 0;
			int l_s = get_lines_number();
			int j;
			for(j=0; j<l_s; j++) {
				CDLine* l_curr = get_line(j);
				if (l_curr->get_internal_momentum_val() == NULL) {
					if (l_curr->is_external()) {
						printf("Error: external line with NULL momenta.\n"); 
						return false;
					}

					// it is better to assign new symbol for virtual PHOTON line.
					bool is_best = false;
					int pk_curr_abs = CLMath::abs(l_curr->get_particle_kind());
					if (pk_curr_abs == (int)CEParticle::PARTICLE_PHOTON) is_best = true;
					if (is_best) { l = l_curr; l_pk_abs = pk_curr_abs; break; }

					// the choise of bosonic line is better
					if (CEParticle::is_vec_bosonic_PK(pk_curr_abs)) {
						bool is_better = false;
						if (CEParticle::is_vec_bosonic_PK(l_pk_abs)) {
							if (pk_curr_abs < l_pk_abs) is_better = true;
						} else {
							is_better = true;
						}
						if (is_better) { l = l_curr; l_pk_abs = pk_curr_abs; }
					}
					
					if (l == NULL) { l = l_curr; l_pk_abs = pk_curr_abs; }
				}
			}
			
			if (l == NULL) { printf("Internal ERROR arranging momenta.\n"); throw(new CEException(CEException::INT_ERROR)); }
			
			CEPtrArray* l_links = l->get_links();
			if (l_links->size() != 2) { printf("Internal ERROR arranging momenta.\n"); throw(new CEException(CEException::INT_ERROR)); }

			// create new virtual momentum symbol
			CESymbol* virt_mom = CESymbol::getUnique();
			l->set_momentum(virt_mom, (CDElement*)l_links->GetAt(0));
			dbg->log(45, "new virt momentum symbol assigned to line :", l);
			// add symbol to virtual momenta list
			m_virt_momenta_list->add(virt_mom);
		}
	}

	dbg->log(35, "Momenta arranged OK.");
	return true;
}

/*
*	Note:
*	Do not use this function directly. Use CDiagramSet::RetrieveMatrixElement() instead.
*	If use:
*	kin->reset_volatile_variables(true); should be calles before matrix element retrieving.
*	kin->reset_volatile_variables(false); should be calles after matrix element retrieving.
*/
CMatrixElement* CFeynDiagram::RetrieveMatrixElement() {
	dbg->log(15, "RetrieveMatrixElement from diagram:", this);
	if (m_physics == NULL) return NULL;
	CKinematics* kin=CEProcessor::kinematics();
	if (kin==NULL) { dbg->log(0,"ERROR. CFeynDiagram::RetrieveMatrixElement() with NO kinematics declared."); return NULL; }
	if ((m_kinHashID!=kin->GetHashCode())&&(m_kinHashID!=0)) { dbg->log(0,"ERROR. Can't use diagram when DIFFERENT kinematics is declared."); CEException::halt(); return NULL; }
	CDiagCorrRules* rules = m_physics->getCorrespondenceRules();
	if (rules == NULL) return NULL;
	dbg->log(45, "Correspondence rules found:", rules);
	// init vector index list
	if (m_ext_vector_indices_list != NULL) delete m_ext_vector_indices_list;
	m_ext_vector_indices_list = new CEPtrArray();

	// clear temporary data
	clear_temporary_data(false); // please, clear temporary data immediately after use.
	
	CExpression* expr_me = (m_factor_expr == NULL)?NULL:(m_factor_expr->clone());
	// 1. add to ME all the fermion tales.
	int ls = get_lines_number();
	int i;
	for(i=0; i<ls; i++) {
		CDLine* l_curr = get_line(i);
		// find out fermion tales
		if (!l_curr->is_external()) continue;
		int pk = l_curr->get_particle_kind();
		if (!(CEParticle::is_fermion_PK(pk))) continue; // if not a fermion
		// check up if the line starts a fermion tale
		CDExtLine* l_ext = (CDExtLine*)l_curr;
		bool is_line_start = CEUtil::LXOR(l_ext->is_initial(), (pk > 0));
		if (!is_line_start) continue;

		CExpression* me_fact = RetrieveFermionTaleFactor(l_curr);
		if (me_fact != NULL) {
			if(expr_me==NULL){ expr_me=me_fact; }else{ expr_me=expr_me->multiply(me_fact); }
		} else { dbg->log(0, "Error retrieving ME."); throw(new CEException(CEException::INT_ERROR)); }
	}
	// all the fermion lines added to ME are now marked
	dbg->log(35, "External-started fermion lines added to ME.");

	// 2. add contributions of fermion loops (of non-marked still fermion lines)
	for(i=0; i<ls; i++) {
		CDLine* l_curr = get_line(i);
		// find out non-marked internal fermion line
		if (l_curr->is_external()) continue;
		int pk = l_curr->get_particle_kind();
		if (!(CEParticle::is_fermion_PK(pk))) continue; // if not a fermion
		if (l_curr->m_temporary_mark != NULL) continue;

		CExpression* me_fact = RetrieveFermionTaleFactor(l_curr);
		if (me_fact != NULL) {
			//Note: The (-1) factor for internal loop is NOT invloved in RetrieveFermionTaleFactor().
			//We have to involve it HERE:
			me_fact=me_fact->multiply(CENumber::GetMinusUnit());

			if(expr_me==NULL){ expr_me=me_fact; }else{ expr_me=expr_me->multiply(me_fact); }
		} else { dbg->log(0, "Error retrieving ME."); throw(new CEException(CEException::INT_ERROR)); }
	}
	// all the fermion lines should be marked now marked
	// all the fermion vertices should also be marked now marked
	// all the bosonic lines attached to fermion vertices
	// should be marked now with CSequence of vector indices
	dbg->log(35, "Fermion loops added to ME.");

	// 3. add non-fermion vertices (non-marked yet).
	int vs = get_vertices_number();
	for(i=0; i<vs; i++) {
		CDVertex* v = get_vertex(i);
		if (v->m_temporary_mark != NULL) continue;

		CExpression* me_fact = RetrieveVertexFactor(v);
		if (me_fact != NULL) {
			if (expr_me == NULL) expr_me = me_fact;
			else expr_me = expr_me->multiply(me_fact);
		} else { dbg->log(0, "Error retrieving ME."); throw(0); }
	}
	// all the vertices should be marked now
	// all the bosonic lines should be marked now with CSequence of their vector indices
	dbg->log(35, "Pure bosonic vertices added to ME.");

	// 4. add contributions of non-fermion lines
	CENumber unit((long)1);
	for(i=0; i<ls; i++) {
		CDLine* l_curr = get_line(i);
		int pk = l_curr->get_particle_kind();
		//int PK_abs = (pk >= 0)?pk:(-pk);
		if (CEParticle::is_fermion_PK(pk)) continue; // if a fermion

		CExpression* me_fact = RetrieveBosonicLineFactor(l_curr);
		if (me_fact != NULL) {
			if (!me_fact->equals(&unit)) {
				if (expr_me == NULL) expr_me = me_fact;
				else expr_me = expr_me->multiply(me_fact);
			} else {
				delete me_fact;
			}
		} else { dbg->log(0, "Error retrieving ME."); throw(new CEException(CEException::INT_ERROR)); }
	}
	dbg->log(35, "Bosonic lines added to ME.");
	clear_temporary_data(true);
	if (expr_me == NULL) { dbg->log(0, "Error retrieving ME (NULL result)."); throw(new CEException(CEException::INT_ERROR)); }

	dbg->log(15, "Matrix element before arranging:", expr_me);
	// inline sub-products
	expr_me = CExpression::arrge(expr_me);

	if (CRContainCheck::search(expr_me,CEInfinity::CLASS_NAME)) {
		expr_me=CRArrangeInfinity::ArrageInfinityToUnit(expr_me);
		if (CRContainCheck::search(expr_me,CEInfinity::CLASS_NAME)) {
			dbg->log(0, "WARNING: Matrix element contains INFINITY object:", expr_me, this);
		} else {
			dbg->log(15, "Matrix element after Infinity arrangement:", expr_me);
		}
	}

	CMatrixElement* res = new CMatrixElement(expr_me, m_virt_momenta_list, m_ext_vector_indices_list);
	delete expr_me;
	delete m_ext_vector_indices_list; m_ext_vector_indices_list = NULL;
	return res;
}

/*
*	print out fermion tale
*	if l is internal line the loop is expected
*	Note: The (-1) factor for internal loop is NOT invloved here! Simply +Sp{...} for loops.
*/
CExpression* CFeynDiagram::RetrieveFermionTaleFactor(CDLine* l) {
	dbg->log(45, "RetrieveFermionTaleFactor:", l);

	bool started_from_external = l->is_external();
	CDLine* line_start = l;

	CExpression* res = NULL;

	CDVertex* v_to = NULL;
	CDVertex* v_from = NULL;
	if (started_from_external) {
		v_from = (CDVertex*)l->get_link(0); // the only vertex (vertex "from")
	} else {
		v_to = (CDVertex*)((CDIntLine*)l)->get_vertex_to();
		v_from = (CDVertex*)((CDIntLine*)l)->get_vertex_from();
	}
	
	for(;;) {
		// add fermion line factor
		CExpression* l_fact = NULL;
		if (v_to != NULL) l_fact = RetrieveFermionLineFactor(l, v_to, true);
		else l_fact = RetrieveFermionLineFactor(l, v_from, false);
		if (l_fact == NULL) {
			dbg->log(0, "Error retrieving fermion line factor.");
			if (res != NULL) delete res;
			return NULL;
		}
		if (res == NULL) res = l_fact;
		else res = res->multiply(l_fact);

		// check up end of fermion tale
		if (l->is_external()) {
			if (v_to != NULL) {
				if (!started_from_external) { dbg->log(0, "Error retrieving ME: unexpected end of fermion line."); if (res != NULL) delete res; return NULL; }
				break;
			}
		}

		// add fermion vertex factor (for "v_from" only)
		if (v_from != NULL) {
			CExpression* v_fact = RetrieveVertexFactor(v_from);
			if (v_fact == NULL) {
				dbg->log(0, "Error retrieving fermion vertex factor.");
				if (res != NULL) delete res;
				return NULL;
			}
			if (res == NULL) res = v_fact;
			else res = res->multiply(v_fact);
		}
		else { dbg->log(0, "Retrieve fermion tale factor ERROR: NULL vertex spotted."); throw(new CEException(CEException::INT_ERROR)); }

		// find next fermion line.
		CDLine* line_prev = l; l = NULL;
		int links_num = v_from->get_links_number();
		int j;
		for(j=0; j<links_num; j++) {
			CDLine* line_j = (CDLine*)v_from->get_link(j);
			if (line_j != line_prev) {
				if (CEParticle::is_fermion_PK(line_j->get_particle_kind())) {
					l = line_j; break;
				}
			}
		}
		if (l == NULL) { dbg->log(0, "Retrieve fermion tale factor ERROR: invalid diagram structure."); throw(new CEException(CEException::INT_ERROR)); }

		// check up fermion loop complition
		if(l == line_start) {
			if (started_from_external) { dbg->log(0, "Retrieve ME ERROR: Invalid diagram structure."); throw(new CEException(CEException::INT_ERROR)); }
			CExpression* sp_res = new CFSp(res); delete res;
			res = sp_res;
			break;
		}

		// find next v_from vertex.
		v_to = v_from;
		if (l->is_external()) {
			v_from = NULL;
		} else {
			v_from = (CDVertex*)(((CDIntLine*)l)->get_other_link(v_to));
			if (v_from == NULL) { dbg->log(0, "Retrieve fermion tale factor ERROR: Unbinded diagram."); throw(new CEException(CEException::INT_ERROR)); }
		}
	}

	dbg->log(55, "RetrieveFermionTaleFactor result:", res);
	return res;
}

/*
*	v_to is "v_from" for starting external line.
*	Note: The (-1) factor for internal loop is NOT invloved here! Simply +Sp{...} for loops.
*/
CExpression* CFeynDiagram::RetrieveFermionLineFactor(CDLine* l, CDVertex* v, bool v_is_to) {
	dbg->log(60, "RetrieveFermionLineFactor:", l);
	if (l->m_temporary_mark != NULL) { dbg->log(0, "Retrieving ME ERROR: algorithm error."); throw(new CEException(CEException::INT_ERROR)); }
	CDiagCorrRules* rules = m_physics->getCorrespondenceRules();

	int pk = l->get_particle_kind();
	int pk_abs = CLMath::abs(pk);
	CExpression* mom = l->get_momentum(v); // momentum incoming to vertex "v".
	if (!v_is_to) mom->inverseFlag(); // momentum for outgoing particle (v is "v_from").
	CExpression* mass = NULL; // defined in a different manner

	int line_type = 0;
	if (l->is_external()) {
		//mass = NULL; // mass may be specific for external particle
		mass = m_physics->get_mass_sym(pk_abs); // for internal particle

		line_type |= CDiagCorrRules::CORR_RULE_EXT_LINE;

		// check up for starting line (conjugated spinor and outgoing particle)
		if (!v_is_to) {
			line_type |= CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED;
		}
	} else {
		mass = m_physics->get_mass_sym(pk_abs); // for internal particle

		line_type |= CDiagCorrRules::CORR_RULE_INT_LINE;
		if (l->checkDEKind(CDElement::EL_COUNTER_TERM)) line_type |= CDiagCorrRules::CORR_RULE_COUNTER_TERM;
	}
	
	l->m_temporary_mark = CENumber::GetUnit(); // mark any cloned value;
	CExpression* res = rules->GetLineCorrespondence(line_type, pk_abs, mom, mass, NULL);
	if(res==NULL) { dbg->log(0,line_type,pk_abs,"ERROR: rules->GetLineCorrespondence(linetype=%ld, pkabs=%ld)==NULL",mom); }
	delete mom;
	dbg->log(60, "RetrieveFermionLineFactor result:", res);
	return res;
}

CExpression* CFeynDiagram::RetrieveVertexFactor(CDVertex* v) {
	dbg->log(60, "RetrieveVertexFactor:", v);
	if (!v->is_binded()) { dbg->log(0, "Retrieving ME ERROR: unbinded diagram."); throw(new CEException(CEException::INT_ERROR)); }
	if (v->m_temporary_mark != NULL) { dbg->log(0, "Retrieving ME ERROR: algorithm error."); throw(new CEException(CEException::INT_ERROR)); }
	if (!v->check_VI_validity()) { dbg->log(0, "Retrieving ME ERROR: invalid vertex.", v); throw(new CEException(CEException::INT_ERROR)); }
	CDiagCorrRules* rules = m_physics->getCorrespondenceRules();
	CDiagramVertexInfo* vi = v->get_vertex_info();
	bool vec_momenta_required = rules->NeedVecMomentaForVertexCorr(vi->get_vertex_ID());
	bool all_momenta_required = rules->NeedAllMomentaForVertexCorr(vi->get_vertex_ID());

	/*
	*	[Note] Index usage:
	*		- vector index for every vector particle
	*		- t_a-index for CEParticle::PARTICLE_GLUON
	*		- momentum symbol if ( (vec_momenta_required && (is vector particle)) || all_momenta_required)
	*/

	// create vector index for every bosonic line
	CSequence* temp_syms = new CSequence(); // indices (& momenta if required)
	int s = vi->get_tales_num();
	int i;
	for(i=0; i<s; i++) {
		int pk_i_abs = CLMath::abs(vi->get_particle_kind(i));
		if (!CEParticle::is_vec_bosonic_PK(pk_i_abs)) {
			if (all_momenta_required) {
				CDLine* l = (CDLine*)v->get_link(i);
				CExpression* mom_cpy = l->get_momentum(v); // incoming momentum
				temp_syms->addItem(mom_cpy);
			}
			continue; // if not a boson
		}

		CDLine* l = (CDLine*)v->get_link(i);
		if (vec_momenta_required || all_momenta_required) {
			CExpression* mom_cpy = l->get_momentum(v); // incoming momentum
			temp_syms->addItem(mom_cpy);
		}

		// involve vector index
		CExpression* new_ind = NULL; // new vector index
		if (l->is_external()) {
			// try retrieve vector index from particle data (it may be predefined)
			new_ind = ((CDExtLine*)l)->get_vector_index();
			if (new_ind != NULL) new_ind = new_ind->clone();
		}
		if (new_ind == NULL) new_ind = CESymbol::getUnique();

		if (l->m_temporary_mark == NULL) l->m_temporary_mark = new CSequence();
		CSequence* l_indices = (CSequence*)l->m_temporary_mark;
		l_indices->addItem(new_ind->clone());
		temp_syms->addItem(new_ind);

		// involve gluon index if required
		if ((pk_i_abs & CEParticle::PARTICLE_GLUON) == CEParticle::PARTICLE_GLUON) {
			CExpression* gm_index = NULL;
			if(l->is_external()){
				//Note: no external gluon line is expected!
				CEException::halt("Unexpected external gluon line spotted");
				// try retrieve vector index from particle data (it may be predefined)
				//gm_index = ((CDExtLine*)l)->[...]();
				if(gm_index!=NULL)gm_index=gm_index->clone();
			}
			if(gm_index==NULL)gm_index=CESymbol::getUnique();
			if(l->m_temporary_mark==NULL)l->m_temporary_mark=new CSequence();
			CSequence* l_indices = (CSequence*)l->m_temporary_mark;
			l_indices->addItem(gm_index->clone());
			temp_syms->addItem(gm_index);
		}
	}

	v->m_temporary_mark = CENumber::GetUnit(); // mark any cloned value;
	CExpression* res = rules->GetVertexCorrespondence(vi->get_vertex_ID(), temp_syms);
	delete temp_syms;
	dbg->log(60, "RetrieveVertexFactor result:", res);
	return res;
}

CExpression* CFeynDiagram::RetrieveBosonicLineFactor(CDLine* l) {
	dbg->log(60, "RetrieveBosonicLineFactor:", l);
	if (!l->is_binded()) { dbg->log(0, "Retrieving ME algorithm ERROR: unbinded line spotted."); throw(new CEException(CEException::INT_ERROR)); }
	CDiagCorrRules* rules = m_physics->getCorrespondenceRules();

	CSequence* indices = NULL;
	if (l->m_temporary_mark != NULL) {
		indices = (CSequence*)l->m_temporary_mark;
		if (indices->size() == 0) indices = NULL;
	}
	if ((indices == NULL) && (CEParticle::is_vec_bosonic_PK(l->get_particle_kind()))) { CEException::halt(dbg, "Retrieving ME algorithm ERROR: no vector bosonic indices available."); }

	CExpression* res = NULL;
	if (l->is_external()) {
		// add index to external lines indices list
		if (indices != NULL) {
			int pid = ((CDExtLine*)l)->get_particle_ID();
			if (indices->size() != 1) { dbg->log(0, pid, "Retrieving ME algorithm ERROR: No valid index found for external boson (pid == %ld)."); throw(new CEException(CEException::INT_ERROR)); }
			m_ext_vector_indices_list->add(new CExprContainer(indices->GetAt(0), pid));
		}

		/*
		// return cloned unit
		return CENumber::GetUnit();
		*/

		int pk_abs = CLMath::abs(l->get_particle_kind());
		CExpression* mom = l->get_momentum(l->get_link(1)); // momentum incoming to vertex.
		CExpression* mass = m_physics->get_mass_sym(pk_abs); // mass for internal line
		int line_type = CDiagCorrRules::CORR_RULE_EXT_LINE;
		res = rules->GetLineCorrespondence(line_type, pk_abs, mom, mass, indices);
		delete mom;
	}
	else {
		int pk_abs = CLMath::abs(l->get_particle_kind());
		CExpression* mom = l->get_momentum(l->get_link(1)); // momentum incoming to vertex.
		CExpression* mass = m_physics->get_mass_sym(pk_abs); // mass for internal line

		int line_type = CDiagCorrRules::CORR_RULE_INT_LINE;
		if (l->checkDEKind(CDElement::EL_COUNTER_TERM)) line_type |= CDiagCorrRules::CORR_RULE_COUNTER_TERM;

		/*
		*	[Note] Index usage:
		*		- momentum, mass, indices{vec1[,glu1...],vec2[,glu2...]}
		*/
		res = rules->GetLineCorrespondence(line_type, pk_abs, mom, mass, indices);
		delete mom;
	}
	dbg->log(60, "RetrieveBosonicLineFactor result:", res);
	return res;
}

/*
*	Compares all the elemets equivalence
*	Uses Arrange library data to locate unsorted similar elements in diagram
*/
bool CFeynDiagram::IsEqualTo(CFeynDiagram* other, CELibrary* arr_lib, CELibrary* arr_lib_other, bool ignore_factor) {
	dbg->log(45, ignore_factor, "IsEqualTo(ignore_factor: %ld)");
	dbg->log(245, "IsEqualTo() diagrams", this, other);

	if (other == this) return true;
	if (other == NULL) return false;
	if (m_kinHashID!=other->GetKinHashID()) return false;

	// 0.1 compare physics ID
	if (m_physics != other->m_physics) return false;

	// 0.2 compare factors
	if (!ignore_factor) {
		CExpression* dfact_1 = get_factor_expr(); CExpression* dfact_2 = other->get_factor_expr();
		if (dfact_1 != dfact_2) {
			if ((dfact_1 == NULL) || (dfact_2 == NULL)) return false;
			if (!dfact_1->equals(dfact_2)) return false;
		}
	}

	if (arr_lib != NULL) { if (arr_lib->ClassName()!=CLDiagArrange::CLASS_NAME) dbg->log(0,"CFeynDiagram::IsEqualTo invalid parameter #2. CLDiagArrange object expected."); CEException::halt(); }
	if (arr_lib_other != NULL) { if (arr_lib_other->ClassName()!=CLDiagArrange::CLASS_NAME) dbg->log(0,"CFeynDiagram::IsEqualTo invalid parameter #3. CLDiagArrange object expected."); CEException::halt(); }
	CLDiagArrange* lib = (CLDiagArrange*)arr_lib;
	CLDiagArrange* lib_other = (CLDiagArrange*)arr_lib_other;
	bool libs_OK = ((lib != NULL) && (lib_other != NULL));

	// 1. compare vertices
	int s = get_vertices_number();
	if (s != other->get_vertices_number()) return false;
	int i;
	for(i=0; i<s; i++) {
		CDVertex* v1 = get_vertex(i);
		CDVertex* v2 = other->get_vertex(i);

		bool v_equal = CLDiagArrange::AreVerticesSimilar(v1, this, v2, other);
		if (v_equal) continue;
		if (!libs_OK) return false; // false

		// check next for unsorted similar vertices
		int eq_s1 = lib->GetEqualsVerticesNum(i);
		int eq_s2 = lib_other->GetEqualsVerticesNum(i);
		if (eq_s1 != eq_s2) return false;

		bool similar_found = false;
		int j;
		for(j=i+1; j<(i+eq_s1); j++) {
			CDVertex* v2_next = other->get_vertex(j);
			if (CLDiagArrange::AreVerticesSimilar(v1, this, v2_next, other)) {
				similar_found = true; break;
			}
		}
		if (!similar_found) return false;
		i+=(eq_s1-1);
	}

	// 2. compare lines
	s = get_lines_number();
	if (s != other->get_lines_number()) return false;
	for (i=0; i<s; i++) {
		CDLine* l1 = get_line(i);
		CDLine* l2 = other->get_line(i);

		bool v_equal = CLDiagArrange::AreLinesSimilar(l1, this, l2, other);
		if (v_equal) continue;
		if (!libs_OK) return false; // false

		// check next for unsorted similar vertices
		int eq_s1 = lib->GetEqualsLinesNum(i);
		int eq_s2 = lib_other->GetEqualsLinesNum(i);
		if (eq_s1 != eq_s2) return false;

		bool similar_found = false;
		int j;
		for(j=i+1; j<(i+eq_s1); j++) {
			CDLine* l2_next = other->get_line(j);
			if (CLDiagArrange::AreLinesSimilar(l1, this, l2_next, other)) {
				similar_found = true; break;
			}
		}
		if (!similar_found) return false;
		i+=(eq_s1-1);
	}

	dbg->log(45, "IsEqualTo() result: TRUE");
	return true;
}

int CFeynDiagram::GetLoopsNumber() {
	if (m_virt_momenta_list == NULL) return -1; // this may not occur
	return m_virt_momenta_list->size();
}

int CFeynDiagram::GetCTNumber() { // counter-terms number
	int res=0; int i,s;
	s=get_vertices_number();
	for(i=0; i<s; i++) {
		CDVertex* v=get_vertex(i); if(v==NULL)continue;
		CDiagramVertexInfo* vi=v->get_vertex_info(); if(vi==NULL)continue;
		if (vi->check_vert_type(CDiagramVertexInfo::VERTEX_COUNTERTERM)) res++;
	}
	s=get_lines_number();
	for(i=0; i<s; i++) {
		CDLine* l=get_line(i); if(l==NULL)continue;
		if((l->DE_kind()&CDElement::EL_COUNTER_TERM)==CDElement::EL_COUNTER_TERM)res++;
	}
	return res;
}

/////////////	old

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
/*
CFeynDiagram::CFeynDiagram(CEPtrArray* _photons,
						   CEPtrArray* _in_fermions,
						   CEPtrArray* _out_fermions,
						   CEPtrArray* _internal_fermions,
						   CEPtrArray* _internal_bosons,
						   CEPtrArray* _vertices,
						   CEPtrArray* _crossing_info)
						   :CESerializable()
{
	m_matrix_element = NULL;
	m_me_integration_variables = NULL;

	//m_x_temp = 0;
	//m_y_temp = 0;
	//m_x_num = 0;
	//m_y_num = 0;

	m_photons = CEPtrArray::get_copy(_photons, true);
	m_in_fermions = CEPtrArray::get_copy(_in_fermions, true);
	m_out_fermions = CEPtrArray::get_copy(_out_fermions, true);
	m_internal_fermions = CEPtrArray::get_copy(_internal_fermions, true);
	m_internal_bosons = CEPtrArray::get_copy(_internal_bosons, true);
	m_vertices = CEPtrArray::get_copy(_vertices, true);

	if (_crossing_info == NULL) m_crossing_info = NULL;
	else m_crossing_info = CEPtrArray::get_copy(_crossing_info, false);

//	dbg->log(30, m_photons->size()
//		+ m_in_fermions->size()
//		+ m_out_fermions->size()
//		+ m_internal_fermions->size()
//		+ m_internal_bosons->size()
//		+ m_vertices->size(), "Diagram created, total elements: %d");
}

CFeynDiagram::~CFeynDiagram()
{
	dbg->log(35, "deleting diagram");
	delete m_photons;
	//dbg->log(35, "deleting 1");
	delete m_in_fermions;
	//dbg->log(35, "deleting 2");
	delete m_out_fermions;
	//dbg->log(35, "deleting 3");
	delete m_internal_fermions;
	//dbg->log(35, "deleting 4");
	delete m_internal_bosons;
	delete m_vertices;

	//dbg->log(35, "deleting me");
	
	if (m_matrix_element != NULL) delete m_matrix_element;
	if (m_me_integration_variables != NULL) delete m_me_integration_variables;
	if (m_crossing_info != NULL) delete m_crossing_info;
	//dbg->log(35, "deleted");
}

CEPtrArray* CFeynDiagram::photons()
{
	return m_photons;
}

CEPtrArray* CFeynDiagram::in_fermions()
{
	return m_in_fermions;
}

CEPtrArray* CFeynDiagram::out_fermions()
{
	return m_out_fermions;
}

CEPtrArray* CFeynDiagram::internal_fermions()
{
	return m_internal_fermions;
}

CEPtrArray* CFeynDiagram::internal_bosons()
{
	return m_internal_bosons;
}

CEPtrArray* CFeynDiagram::vertices()
{
	return m_vertices;
}

CFeynDiagram* CFeynDiagram::clone() const
{
	dbg->log(30, "cloning");

	//dbg->log(30, "before cloning", photons()->get(0));

	CFeynDiagram* res = new CFeynDiagram(m_photons,
		m_in_fermions,
		m_out_fermions,
		m_internal_fermions,
		m_internal_bosons,
		m_vertices,
		m_crossing_info);

	//re-assigning links.
	//dbg->log(30, "before aranging links", res->photons()->get(0));

	re_assign_links(res->photons(), photons());
	re_assign_links(res->in_fermions(), in_fermions());
	re_assign_links(res->out_fermions(), out_fermions());
	re_assign_links(res->internal_fermions(), internal_fermions());
	re_assign_links(res->internal_bosons(), internal_bosons());
	re_assign_links(res->vertices(), vertices());

	//dbg->log(30, "after aranging", res->photons()->get(0));

	dbg->log(30, "cloned", res);
	return res;
}

CEObject* CFeynDiagram::copy() const
{
	return clone();
}

bool CFeynDiagram::is_valid()
{
	int i;
	for(i=0; i<photons()->size(); i++) {
		if (!((CDLine*)photons()->get(i))->is_binded())
			return false;
	}

	for (i=0; i<in_fermions()->size(); i++) {
		if (!((CDLine*)in_fermions()->get(i))->is_binded())
			return false;
	}

	for (i=0; i<out_fermions()->size(); i++) {
		if (!((CDLine*)out_fermions()->get(i))->is_binded())
			return false;
	}

	for (i=0; i<internal_fermions()->size(); i++) {
		if (!((CDLine*)internal_fermions()->get(i))->is_binded())
			return false;
	}

	for (i=0; i<internal_bosons()->size(); i++) {
		if (!((CDLine*)internal_bosons()->get(i))->is_binded())
			return false;
	}
	return true;
}

void CFeynDiagram::re_assign_links(CEPtrArray* elements, CEPtrArray* old_elements)
{
	int s = elements->size();
	//dbg->log(30, s, "arranging links for array of size %d");
	if (s != old_elements->size()) {
		printf("Error (CFeynDiagram) non-equivalent arrays recieved.");
		return;
	}
	int i;
	for(i=0; i<s; i++) {
		CDElement* el = (CDElement*)elements->get(i);
		el->arange_links((CDElement*)old_elements->get(i));
	}
}

void CFeynDiagram::pre_draw(unsigned width, unsigned height)
{
	if ((height < 50) || (width < 80)) return;

	unsigned x_gap = 10;

	int num_init = num_initial_lines();
	int num_final = num_final_lines();
	
	unsigned x_half = width/2;
	unsigned y_half = height/2;

	unsigned yi_pos;
	unsigned yi_step;
	if (num_init < 2) {
		yi_pos = y_half;
		yi_step = y_half;//not used.
	}
	else {
		yi_pos = y_half/2;
		yi_step = y_half/(num_init-1);
	}

	unsigned yf_pos;
	unsigned yf_step;
	if (num_final < 2) {
		yf_pos = y_half;
		yf_step = y_half;//not used.
	}
	else {
		yf_pos = y_half/2;
		yf_step = y_half/(num_final-1);
	}

	int i;
	for(i=0; i<out_fermions()->size(); i++) {
		CDExtLine* ext_line = (CDExtLine*)out_fermions()->get(i);
		if (ext_line->is_initial()) {
			ext_line->set_pos(x_gap, yi_pos);
			yi_pos += yi_step;
		}
		else {
			ext_line->set_pos(width-x_gap, yf_pos);
			yf_pos += yf_step;
		}
	}

	for (i=0; i<in_fermions()->size(); i++) {
		CDExtLine* ext_line = (CDExtLine*)in_fermions()->get(i);
		if (ext_line->is_initial()) {
			ext_line->set_pos(x_gap, yi_pos);
			yi_pos += yi_step;
		}
		else {
			ext_line->set_pos(width-x_gap, yf_pos);
			yf_pos += yf_step;
		}
	}

	for (i=0; i<photons()->size(); i++) {
		CDExtLine* ext_line = (CDExtLine*)photons()->get(i);
		if (ext_line->is_initial()) {
			ext_line->set_pos(x_gap, yi_pos);
			yi_pos += yi_step;
		}
		else {
			ext_line->set_pos(width-x_gap, yf_pos);
			yf_pos += yf_step;
		}
	}

	int v_size = vertices()->size();

	for (i=0; i<v_size; i++) {
		((CDVertex*)vertices()->get(i))->clear_pos();
	}

	bool loop = true;
	int num_links = 4;
	while (loop) {
		loop = false;
		num_links--;
		for (i=0; i<v_size; i++) {
			CDVertex* v = (CDVertex*)vertices()->get(i);
			if (!v->pos_exist()) {
				if (!calc_vertex_pos(v, num_links, x_half, y_half)) loop = true;
			}
		}
	}
}

int CFeynDiagram::order()
{
	return vertices()->size();
}

int CFeynDiagram::num_initial_lines()
{
	return num_externals(true);
}

int CFeynDiagram::num_final_lines()
{
	return num_externals(false);
}

int CFeynDiagram::num_externals(bool initial)
{
	int res = 0;
	int i;
	for(i=0; i<out_fermions()->size(); i++) {
		if (initial == ((CDExtLine*)out_fermions()->get(i))->is_initial())
			res++;
	}

	for (i=0; i<in_fermions()->size(); i++) {
		if (initial == ((CDExtLine*)in_fermions()->get(i))->is_initial())
			res++;
	}

	for (i=0; i<photons()->size(); i++) {
		if (initial == ((CDExtLine*)photons()->get(i))->is_initial())
			res++;
	}
	return res;
}

bool CFeynDiagram::calc_vertex_pos(CDVertex * v, int num_links_required, unsigned x_center, unsigned y_center)
{
	if (v == NULL) return false;
	if ((num_links_required > 3) || (num_links_required < 0)) return false;

	v->m_x_temp = 0;
	v->m_y_temp = 0;
	v->m_x_num = 0;
	v->m_y_num = 0;
	v->m_pos_defined = 0;
	v->m_have_external_link = false;
	v->m_all_links_are_external = true;

	link_vertex_pos_to(v, (CDLine*)v->line_to());
	link_vertex_pos_to(v, (CDLine*)v->line_from());
	link_vertex_pos_to(v, (CDLine*)v->line_boson());

	int num_pos_defined = v->m_pos_defined;
	unsigned x_pos = v->m_x_temp;
	unsigned y_pos = v->m_y_temp;
	unsigned x_num = v->m_x_num;
	unsigned y_num = v->m_y_num;

	if (num_pos_defined < num_links_required) return false;

	switch (num_pos_defined)
	{
	case 0:
		break;
	case 1:
		if (v->m_have_external_link) {
			x_pos = (x_center+x_pos)/2;
		}
		else {
			y_pos = (y_center+y_pos)/2;
		}
		break;
	case 2:
		if (v->m_all_links_are_external) {
			x_pos += x_center;
			x_num++;
		}
	default:
		x_pos = x_pos/x_num;
		y_pos = y_pos/y_num;
	}

	int k = 3;
	while ((pos_is_used(x_pos, y_pos)) && (k<10)) {
		y_pos += y_center/k;
		k++;
	}

	v->set_pos(x_pos, y_pos);
	return true;
	//return v->calc_vertex_pos(num_links_required, x_center, y_center);
}

CEString* CFeynDiagram::toString(int mode)
{
	CEString* res = NULL;
	
	switch (mode)
	{
	case STORE_WR_MATHEMATICA:
		res  = new CEString("\\({CFeynDiagram: order=");
		res->add_int(order());
		res->AddString(" , init=");
		res->add_int(num_initial_lines());
		res->AddString(" , final=");
		res->add_int(num_final_lines());
		res->AddString(" , internal bosons=");
		res->add_int(internal_bosons()->size());
		res->AddString(" , internal fermions=");
		res->add_int(internal_fermions()->size());
		res->AddString("}\\)");
		break;
	case STORE_TEX:
		res = draw(100, 100);
		break;
	default:
		res = new CEString("<Not supported data format.>");
	}
	return res;
}

CEString* CFeynDiagram::draw(unsigned int w, unsigned int h)
{
	CEString* res = NULL;
	pre_draw(w, h);

	//res  = new CEString("\\begin{figure}\n\\begin{picture}(");
	res  = new CEString("\\begin{picture}(");

	res->add_int(w);
	res->AddString(", ");
	res->add_int(h);
	res->AddString(")\n");

	int i;
	for(i=0; i<internal_fermions()->size(); i++) {
		draw_line(res, (CDLine*)internal_fermions()->get(i), false);
	}

	for (i=0; i<out_fermions()->size(); i++) {
		draw_line(res, (CDLine*)out_fermions()->get(i), false);
	}

	for (i=0; i<in_fermions()->size(); i++) {
		draw_line(res, (CDLine*)in_fermions()->get(i), false);
	}

	for (i=0; i<internal_bosons()->size(); i++) {
		draw_line(res, (CDLine*)internal_bosons()->get(i), true);
	}

	for (i=0; i<photons()->size(); i++) {
		draw_line(res, (CDLine*)photons()->get(i), true);
	}

	//res->AddString("\\end{picture}\n\\end{figure}\n\n");
	res->AddString("\\end{picture}\n\n");

	return res;
}

CEString* CFeynDiagram::analyze()
{
	CEString* res = NULL;
	
	res  = new CEString("CFeynDiagram analysis.\norder=");
	res->add_int(order());
	res->AddString(" , init=");
	res->add_int(num_initial_lines());
	res->AddString(" , final=");
	res->add_int(num_final_lines());
	res->AddString(" , internal bosons=");
	res->add_int(internal_bosons()->size());
	res->AddString(" , internal fermions=");
	res->add_int(internal_fermions()->size());

	res->AddString("\n  ph: ");
	int i;
	for(i=0; i<photons()->size(); i++) {
		CDExtLine* l = (CDExtLine*)photons()->get(i);
		res->add_int(i);
		res->AddString("{to: ");
		add_element_info(res, l->get_vertex());
		res->AddString("} ");
	}

	res->AddString("\n  if: ");
	for (i=0; i<in_fermions()->size(); i++) {
		CDExtLine* l = (CDExtLine*)in_fermions()->get(i);
		res->add_int(i);
		res->AddString("{to: ");
		add_element_info(res, l->get_vertex());
		res->AddString("} ");
	}

	res->AddString("\n  ff: ");
	for (i=0; i<out_fermions()->size(); i++) {
		CDExtLine* l = (CDExtLine*)out_fermions()->get(i);
		res->add_int(i);
		res->AddString("{to: ");
		add_element_info(res, l->get_vertex());
		res->AddString("} ");
	}

	res->AddString("\nintf: ");
	for (i=0; i<internal_fermions()->size(); i++) {
		CDIntLine* l = (CDIntLine*)internal_fermions()->get(i);

		res->add_int(i);
		res->AddString("{from: ");
		add_element_info(res, l->vertex_from());
		res->AddString("; to: ");
		add_element_info(res, l->vertex_to());
		res->AddString("} ");
	}
	
	res->AddString("\nintb: ");
	for (i=0; i<internal_bosons()->size(); i++) {
		CDIntLine* l = (CDIntLine*)internal_bosons()->get(i);

		res->add_int(i);
		res->AddString("{from: ");
		add_element_info(res, l->vertex_from());
		res->AddString("; to: ");
		add_element_info(res, l->vertex_to());
		res->AddString("} ");
	}

	res->AddString("\n   v: ");
	for (i=0; i<vertices()->size(); i++) {
		CDVertex* v = (CDVertex*)vertices()->get(i);

		res->add_int(i);
		res->AddString("{from: ");
		add_element_info(res, v->line_from());
		res->AddString("; to: ");
		add_element_info(res, v->line_to());
		res->AddString("; bos: ");
		add_element_info(res, v->line_boson());
		res->AddString("} ");
	}
	res->AddString("\nAnalysis complete.\n");

	return res;
}

void CFeynDiagram::add_element_info(CEString * res, CDElement * el)
{
	if (el == NULL) {
		res->AddString("null");
		return;
	}

	bool element_found = false;
	int i;
	for(i=0; i<internal_fermions()->size(); i++) {
		if (internal_fermions()->get(i) == el) {
			res->AddString("(intf, ");
			res->add_int(i);
			res->AddString(")");
			element_found = true;
		}
	}

	for (i=0; i<internal_bosons()->size(); i++) {
		if (internal_bosons()->get(i) == el) {
			res->AddString("(intb, ");
			res->add_int(i);
			res->AddString(")");
			element_found = true;
		}
	}

	for (i=0; i<in_fermions()->size(); i++) {
		if (in_fermions()->get(i) == el) {
			res->AddString("(if, ");
			res->add_int(i);
			res->AddString(")");
			element_found = true;
		}
	}

	for (i=0; i<out_fermions()->size(); i++) {
		if (out_fermions()->get(i) == el) {
			res->AddString("(ff, ");
			res->add_int(i);
			res->AddString(")");
			element_found = true;
		}
	}

	for (i=0; i<photons()->size(); i++) {
		if (photons()->get(i) == el) {
			res->AddString("(ph, ");
			res->add_int(i);
			res->AddString(")");
			element_found = true;
		}
	}

	for (i=0; i<vertices()->size(); i++) {
		if (vertices()->get(i) == el) {
			res->AddString("(v, ");
			res->add_int(i);
			res->AddString(")");
			element_found = true;
		}
	}

	if (!element_found) {
		res->AddString("??? ");
	}
}

bool CFeynDiagram::link_vertex_pos_to(CDVertex* v, CDLine * l)
{
	if (l == NULL) return false;
	
	if (l->is_external()) {
		CDExtLine* ext_l = (CDExtLine*)l;
		if (ext_l->pos_exist()) {
			v->m_x_temp += ext_l->pos_x();
			v->m_x_num++;
			if (v->m_have_external_link) {
				//v->y_pos += ext_l->pos_y();
			}
			else {
				v->m_y_temp = ext_l->pos_y();
				v->m_y_num = 1;
				v->m_have_external_link = true;
			}
			v->m_pos_defined++;
		}
	}
	else {
		CDIntLine* int_l = (CDIntLine*)l;
		CDVertex* v_other = int_l->get_other_vertex(v);
		if (v_other != NULL) {
			if (v_other->pos_exist()) {
				v->m_x_temp += v_other->pos_x();
				v->m_x_num++;
				if (!v->m_have_external_link) {
					v->m_y_temp += v_other->pos_y();
					v->m_y_num++;
					v->m_all_links_are_external = false;
				}
				v->m_pos_defined++;
			}
		}
		else {
			dbg->log(10, "Can't find other vertex for", v);
			dbg->log(10, "at line", l);
		}
	}
	return true;
}

bool CFeynDiagram::pos_is_used(unsigned x, unsigned y)
{
	int i;
	for(i=0; i<vertices()->size(); i++) {
		CDVertex* v = (CDVertex*)vertices()->get(i);
		if (v->pos_exist()) {
			if ((v->pos_x() == x) && (v->pos_y() == y)) return true;
		}
	}
	return false;
}

void CFeynDiagram::draw_line(CEString* res, CDLine *l, bool is_boson_line)
{
	if (l->is_external()) {
		CDExtLine* l_ext = (CDExtLine*)l;

		if (is_boson_line) res->AddString("\\Photon(");
		else res->AddString("\\ArrowLine(");

		res->add_int((l_ext->get_vertex())->pos_x());
		res->AddString(", ");
		res->add_int(l_ext->get_vertex()->pos_y());
		res->AddString(")(");
		res->add_int(l_ext->pos_x());
		res->AddString(", ");
		res->add_int(l_ext->pos_y());

		if (is_boson_line) res->AddString("){1}{4}\n");
		else res->AddString(")\n");
	}
	else {
		CDIntLine* l_int = (CDIntLine*)l;
		CDVertex* v_from = l_int->vertex_from();
		CDVertex* v_to = l_int->vertex_to();

		unsigned x1 = v_from->pos_x();
		unsigned x2 = v_to->pos_x();
		unsigned y1 = v_from->pos_y();
		unsigned y2 = v_to->pos_y();

		bool draw_arc = false;
		if (is_boson_line) {
			if ((v_from->line_from() == v_to->line_to()) |
				(v_to->line_from() == v_from->line_to()))
			{
				res->AddString("\\PhotonArc(");
				draw_arc = true;
			}
			else res->AddString("\\Photon(");
		}
		else res->AddString("\\ArrowLine(");

		if (draw_arc) {
			res->add_int((x1+x2)/2);
			res->AddString(", ");
			res->add_int((y1+y2)/2);
			res->AddString(")(");
			unsigned r = (unsigned)(CLMath::length(x1, x2, y1, y2)/2);
			res->add_int(r);
			res->AddString(", ");
			unsigned a = (unsigned)CLMath::angle(x1-x2, y1-y2, true);
			if (a>180) a-=180;
			res->add_int(a-180);
			res->AddString(", ");
			res->add_int(a);
			res->AddString("){1}{4.5}\n");
		}
		else {
			res->add_int(x1);
			res->AddString(", ");
			res->add_int(y1);
			res->AddString(")(");
			res->add_int(x2);
			res->AddString(", ");
			res->add_int(y2);
			if (is_boson_line) res->AddString("){1}{4}\n");
			else res->AddString(")\n");
		}
	}
}

CExpression* CFeynDiagram::get_me()
{
	if (m_matrix_element != NULL) return m_matrix_element->clone();

	CEProduct* res = new CEProduct();

	bool k_sym_is_used = false;

	int i;
	for(i=0; i<out_fermions()->size(); i++) {
		CDExtLine* f_out = (CDExtLine*)out_fermions()->get(i);
		CExpression* mom_out = f_out->get_momentum(NULL);
		CDVertex* v = f_out->get_vertex();

		res->addItem(f_out->get_me_factor());

		bool loop = true;
		while (loop) {
			res->addItem(v->get_me_factor());

			CDLine* l_fermion = (CDLine*)v->line_from();
			CExpression* mom_in = l_fermion->get_momentum(v);

			CDLine* l_boson = (CDLine*)v->line_boson();
			CExpression* mom_boson_in = l_boson->get_momentum(v);

			if (mom_in != NULL) {
				if (mom_boson_in != NULL) {
					//do check out conservation laws and/or simplify last momentum.
					//dbg->log(20, "3 monenta defined.");
//					CDLine* l_to_temp = (CDLine*)v->line_to();
//					if (!l_to_temp->is_external()) {
//						delete mom_out;
//						mom_out = mom_in->add(mom_boson_in);
//						CDIntLine* lf = (CDIntLine*)l_to_temp;
//						lf->set_momentum(lf->vertex_to(), mom_out);
//						delete mom_out;
//					}
//					else {
//						delete mom_boson_in;
//						delete mom_in;
//						delete mom_out;
//					}
					delete mom_boson_in;
					delete mom_in;
					delete mom_out;
				}
				else {
					mom_boson_in = CEProcessor::kinematics()
						->get_simpliest(mom_out->minus(mom_in));
					//l_boson should occur internal
					if (!l_boson->is_external()) {
						((CDIntLine*)l_boson)->set_momentum(v, mom_boson_in);
					}
					delete mom_boson_in;
				}

				loop = false;
			}
			else {
				if (mom_boson_in == NULL) {
					if (k_sym_is_used) mom_boson_in = CESymbol::getUnique();
					else {
						mom_boson_in = new CESymbol("k");
						k_sym_is_used = true;
					}
					
					//l_boson should occur internal
					if (!l_boson->is_external()) {
						((CDIntLine*)l_boson)->set_momentum(v, mom_boson_in);
					}
				}

				mom_in = CEProcessor::kinematics()
					->get_simpliest(mom_out->minus(mom_boson_in));

				//l_fermion should occur internal
				if (!l_fermion->is_external()) {
					CDIntLine* l_in = (CDIntLine*)l_fermion;
					l_in->set_momentum(v, mom_in);
					v = l_in->vertex_from();
				}
				else loop = false;//This should not occur.

				mom_out = mom_in;
			}

			res->addItem(l_fermion->get_me_factor());
		}
	}

	for (i=0; i<internal_bosons()->size(); i++) {
		CDVPhoton* vph = (CDVPhoton*)internal_bosons()->get(i);
		res->addItem(vph->get_me_factor());
	}

	m_matrix_element = res->clone();
	return res;
}

CEPtrArray* CFeynDiagram::get_crossing_info()
{
	return m_crossing_info;
}

bool CFeynDiagram::is_crossed_graph() {
	if (m_crossing_info == NULL) return true;
	if (m_crossing_info->size() == 0) return true;
	return false;
}

//
//	if crossing a<->b is already assigned, the rule former is to be removed (a->b->a).
//
void CFeynDiagram::add_crossing_info(CEParticle *part_a, CEParticle *part_b)
{
	if (m_crossing_info == NULL) {
		m_crossing_info = new CEPtrArray();
		m_crossing_info->add(new CRParticeReplasement(part_a, part_b));
	}
	else {
		int s = m_crossing_info->size();
		bool rule_already_exist = false;
		int i;
		for(i=0; i<s; i++) {
			CRParticeReplasement* item = (CRParticeReplasement*)m_crossing_info->get(i);
			if ((item->part_a() == part_a) && (item->part_b() == part_b)) {
				rule_already_exist = true;
				m_crossing_info->RemoveAt(i);
				i = s;
				if (s == 1) {
					delete m_crossing_info;
					m_crossing_info = NULL;
				}
			}
		}
		if (!rule_already_exist) {
			m_crossing_info->add(new CRParticeReplasement(part_a, part_b));
		}
	}
}
*/
