////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LDiagArrange.cpp: implementation of the CLDiagArrange class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLDiagArrange::dbg = new CDebug("CLDiagArrange");

const int CLDiagArrange::CMP_A_LESS_THAN_B		= 0x01;
const int CLDiagArrange::CMP_A_GREATER_THAN_B	= 0x02;
const int CLDiagArrange::CMP_A_EQUAL_B			= 0x04;
const int CLDiagArrange::CMP_AB_UNDEFINED		= 0x08;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CLDiagArrange::CLDiagArrange() : CELibrary() {
	m_diag = NULL; m_arrange_OK = false;
	m_lines = NULL; m_verts = NULL;
}

CLDiagArrange::~CLDiagArrange() { }

const char* CLDiagArrange::CLASS_NAME = "CLDiagArrange";
const char* CLDiagArrange::ClassName() const { return CLASS_NAME; }

//////////////////////////////////////////////////////////////////////
// Static entry
//////////////////////////////////////////////////////////////////////
void CLDiagArrange::DiagArrange(CFeynDiagram* diag) {
	if (diag == NULL) return;
	if (!diag->is_valid()) {
		dbg->log(0, "Warning: Not valid diagram arranging. Arranging skipped.");
		return;
	}

	CLDiagArrange* lib = new CLDiagArrange();
	bool arrange_OK = lib->DiagArrangeImpl(diag);
	delete lib;

	if (!arrange_OK) dbg->log(0, "Error arranging diagram");
}

/*
*	The same, but returns library instance just worked
*	The library object should be deleted manually.
*/
CLDiagArrange* CLDiagArrange::DiagArrangeGetLib(CFeynDiagram* diag) {
	if (diag == NULL) return NULL;
	if (!diag->is_valid()) {
		dbg->log(0, "Warning: Not valid diagram arranging. Arranging skipped.");
		return NULL;
	}

	CLDiagArrange* lib = new CLDiagArrange();
	bool arrange_OK = lib->DiagArrangeImpl(diag);

	//if (!arrange_OK) dbg->log(0, "Error arranging diagram");
	return lib;
}

//////////////////////////////////////////////////////////////////////
// Implementation
//////////////////////////////////////////////////////////////////////
bool CLDiagArrange::DiagArrangeImpl(CFeynDiagram* diag) {

	///// init
	m_diag = diag;
	m_arrange_OK = false;

	///// PK arrange
	DiagArrangeLinesPK(diag);

	///// Elements sequence sorting
	DiagArrangeElementsOrder();

	return m_arrange_OK;
}

int CLDiagArrange::IsDiagArrangedOK() { return m_arrange_OK; }

/*
*	Swaps internal line links if line's PK is <0.
*/
void CLDiagArrange::DiagArrangeLinesPK(CFeynDiagram* diag) {
	dbg->log(40, "DiagArrangeLinesPK");
	int s = diag->get_lines_number();
	int i;
	for(i=0; i<s; i++) {
		CDLine* l = diag->get_line(i);
		if (!l->is_external()) ((CDIntLine*)l)->ArrangeElement();
	}
}

/*
*	Uniform order for arranged diagram is the following.
*	Lines.
*		1. External lines in the fixed order.
*		2. Internal lines (?).
*	Vertices.
*		1. Vertex connected to external lines according to their order.
*		2. Other vertices (?).
*		Vertex connected to first arranged vertex
*/
void CLDiagArrange::DiagArrangeElementsOrder() {
	dbg->log(40, "DiagArrangeElementsOrder");
	// init
	m_lines = m_diag->get_lines_list();
	m_verts = m_diag->get_vertices_list();
	m_lines_num = m_lines->size();
	m_verts_num = m_verts->size();
	m_lines_sorted = 0;
	m_verts_sorted = 0;

	// 1. Leave external lines in their initial order
	int i;
	for(i=0; i<m_lines_num; i++) {
		if (!( m_diag->get_line(i)->is_external() )) { m_lines_sorted=i; break; }
	}

	// 2. Check for no external lines remains after the first internal
	for(i=m_lines_sorted+1; i<m_lines_num; i++) {
		if ( m_diag->get_line(i)->is_external() ) {
			dbg->log(0, "Inbvalid diagram structire: external lines follows the internal.");
			throw(new CEException(CEException::INT_ERROR));
		}
	}

	// 3. Move externals-connected vertices ahead
	// Will be done in sorting.
	
	// 4. sort other elements
	bool last_loop_successful = true;
	for(;;) {
		if ((m_lines_sorted >= m_lines_num) && (m_verts_sorted >= m_verts_num)) {
			// arrangment complete
			m_arrange_OK = true;
			break;
		}

		if (!last_loop_successful) {
			// nothing to do more
			dbg->log(0, "Warning: Unarrageble diagram. Arranging stopped.");
			break;
		}
		last_loop_successful = false;

		// is the only line/vertex left -- set it as arranged
		if ((m_lines_sorted+1) == m_lines_num) {
			m_lines_sorted = m_lines_num;
			last_loop_successful = true;
			continue;
		}
		if ((m_verts_sorted+1) == m_verts_num) {
			m_verts_sorted = m_verts_num;
			last_loop_successful = true;
			continue;
		}

		// find out the first of unsorted vertices
		// if there is at least one vertex unsorted
		if (m_verts_sorted < m_verts_num) {
			int min_pos = m_verts_sorted;
			bool min_is_defined = true;
			int a;
			for(a=m_verts_sorted+1; a<m_verts_num; a++) {
				int cmp_res = VertexABeforeB(a, min_pos);
				dbg->log(60, cmp_res, "Vertices comparison result: %ld");
				if (cmp_res == CMP_A_LESS_THAN_B) { min_pos = a; min_is_defined = true; }
				else if (cmp_res == CMP_AB_UNDEFINED) { min_is_defined = false; }
			}
			if (min_is_defined) {
				// move vertex to first unsorted position
				if (min_pos != m_verts_sorted) m_verts->MoveFromTo(min_pos, m_verts_sorted);
				dbg->log(60, m_verts_sorted, "Last vertex sorted: %ld");
				m_verts_sorted++;
				last_loop_successful = true;
				continue;
			}
		}

		// find out the first of unsorted lines
		// if there is at least one line unsorted
		if (m_lines_sorted < m_lines_num) {
			int min_pos = m_lines_sorted;
			bool min_is_defined = true;
			int a;
			for(a=m_lines_sorted+1; a<m_lines_num; a++) {
				int cmp_res = LineABeforeB(a, min_pos);
				dbg->log(60, cmp_res, "Lines comparison result: %ld");
				if (cmp_res == CMP_A_LESS_THAN_B) { min_pos = a; min_is_defined = true; }
				else if (cmp_res == CMP_AB_UNDEFINED) { min_is_defined = false; }
			}
			if (min_is_defined) {
				// move vertex to first unsorted position
				m_lines->MoveFromTo(min_pos, m_lines_sorted);
				m_lines_sorted++;
				last_loop_successful = true;
				continue;
			}
		}
		////////////////////////////////////////////////////////////
	}

	dbg->log(40, m_arrange_OK, "DiagArrangeElementsOrder completed (OK = %ld)");
}

/*
*	yes (A less than B): CMP_A_LESS_THAN_B
*	no (B less than A): CMP_A_GREATER_THAN_B
*	equal: CMP_A_EQUAL_B
*	undefined: CMP_AB_UNDEFINED
*/
int CLDiagArrange::LineABeforeB(int a, int b) {
	dbg->log(40, a, "LineABeforeB (a = %ld)");
	dbg->log(80, b, "LineABeforeB (b = %ld)");
	if ((a < 0) || (b < 0)) { dbg->log(0, "Internal Error: LineABeforeB() invalid parameters"); return CMP_AB_UNDEFINED; }
	if ((a < m_lines_sorted) || (b < m_lines_sorted)) {
		if (a == b) return CMP_A_EQUAL_B;
		return ((a<b)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
	}

	if (a == b) { dbg->log(0, "Warning: comparison Line to self"); return CMP_AB_UNDEFINED; }

	CDLine* line_A = m_diag->get_line(a);
	CDLine* line_B = m_diag->get_line(b);

	int min_link_after = -1; // min value to compare (used when all the previous are equal)
	for(;;) {
		int link_no_A = getLineMinSortedLink(line_A, min_link_after);
		int link_no_B = getLineMinSortedLink(line_B, min_link_after);

		if (link_no_A < 0) {
			if (link_no_B < 0) break; // no more sorted links
			else return CMP_A_GREATER_THAN_B; // B before A
		}
		if (link_no_B < 0) return CMP_A_LESS_THAN_B; // A before B

		if (link_no_A == link_no_B) {
			// the two conections to vertex is better than one
			int cnt_A = getNumLinksToSortedVert(line_A, link_no_A);
			int cnt_B = getNumLinksToSortedVert(line_B, link_no_B);
			if ((cnt_A == 0) || (cnt_B == 0)) { dbg->log(0, "Internal Error during LineABeforeB()"); throw(new CEException(CEException::INT_ERROR)); }
			if (cnt_A != cnt_B) return ((cnt_A>cnt_B)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
			min_link_after = link_no_A; // compare other sorted links
			continue;
		}
		return ((link_no_A<link_no_B)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
	}

	if ( (areAllLinksSorted(line_A)) && (areAllLinksSorted(line_B)) ) {
		// if all the lines links are equal --> compare PK
		if (areLinesParallel(line_A, line_B)) {
			// the lower line starts from the lower vertex
			if ((isLineDirect(line_A)) && (!isLineDirect(line_B))) return CMP_A_LESS_THAN_B;
			if ((!isLineDirect(line_A)) && (isLineDirect(line_B))) return CMP_A_GREATER_THAN_B;

			int pk_A = line_A->get_particle_kind();
			int pk_B = line_B->get_particle_kind();

			if (pk_A == pk_B) return CMP_A_EQUAL_B;
			return ((pk_A<pk_B)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);// the PKs are absolute here
			// elsewhere PK are equal
		}
		
		// if two lines with equal PK have the same links -- use their order in vertex' data
		// Note: if both vertices are equal ==>
		// compare lines if only the order in both vertices should be the same
	}
	return CMP_AB_UNDEFINED;
}

/*
*	1 means true; -1 means false; 0 means undefined answer
*/
int CLDiagArrange::VertexABeforeB(int a, int b) {
	dbg->log(40, a, "VertexABeforeB (a = %ld)");
	dbg->log(80, b, "VertexABeforeB (b = %ld)");

	if ((a < 0) || (b < 0)) { dbg->log(0, "Internal Error: VertexABeforeB() invalid parameters"); return CMP_AB_UNDEFINED; }
	if ((a < m_verts_sorted) || (b < m_verts_sorted)) {
		if (a == b) return CMP_A_EQUAL_B;
		return ((a<b)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
	}

	if (a == b) { dbg->log(0, "Warning: comparison Vertex to self"); return CMP_AB_UNDEFINED; }

	CDVertex* vert_A = m_diag->get_vertex(a);
	CDVertex* vert_B = m_diag->get_vertex(b);

	int min_link_after = -1;
	for(;;) {
		int link_no_A = getVertMinSortedLink(vert_A, min_link_after);
		int link_no_B = getVertMinSortedLink(vert_B, min_link_after);

		if (link_no_A < 0) {
			if (link_no_B < 0) break; // no more sorted links
			else return CMP_A_GREATER_THAN_B; // B before A
		}
		if (link_no_B < 0) return CMP_A_LESS_THAN_B; // A before B

		if (link_no_A == link_no_B) {
			// the more conections to sorted line is better
			int cnt_A = getNumLinksToSortedLine(vert_A, link_no_A);
			int cnt_B = getNumLinksToSortedLine(vert_B, link_no_B);;
			if ((cnt_A == 0) || (cnt_B == 0)) { dbg->log(0, "Internal Error during VertexABeforeB()"); throw(new CEException(CEException::INT_ERROR)); }
			if (cnt_A != cnt_B) return ((cnt_A>cnt_B)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
			min_link_after = link_no_A; // compare other sorted links
			continue;
		}
		return ((link_no_A<link_no_B)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
	}

	// compare other vertices bind to this
	// consider al the vertices bind to this (same as for line comparison)
	min_link_after = -1; // min value to compare (used when all the previous are equal)
	for(;;) {
		int link_no_A = getMinVertexLinked(vert_A, min_link_after);
		int link_no_B = getMinVertexLinked(vert_B, min_link_after);

		if (link_no_A < 0) {
			if (link_no_B < 0) break; // no more sorted links
			else return CMP_A_GREATER_THAN_B; // B before A
		}
		if (link_no_B < 0) return CMP_A_LESS_THAN_B; // A before B

		if (link_no_A == link_no_B) {
			// the two conections to vertex is better than one
			int cnt_A = getNumLinksToOtherVertex(vert_A, link_no_A);
			int cnt_B = getNumLinksToOtherVertex(vert_B, link_no_B);
			if ((cnt_A == 0) || (cnt_B == 0)) { dbg->log(0, "Internal Error during VertexABeforeB()"); throw(new CEException(CEException::INT_ERROR)); }
			if (cnt_A != cnt_B) return ((cnt_A>cnt_B)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
			min_link_after = link_no_A; // compare other sorted links
			continue;
		}
		return ((link_no_A<link_no_B)?CMP_A_LESS_THAN_B:CMP_A_GREATER_THAN_B);
	}

	// check if vertices are similar
	if ( (areAllLinksSorted(vert_A)) && (areAllLinksSorted(vert_B)) ) {
		if (areVerticesParallel(vert_A, vert_B)) return CMP_A_EQUAL_B;
	}

	return CMP_AB_UNDEFINED;
}

int CLDiagArrange::getLineMinSortedLink(CDLine* l, int min_after) {
	dbg->log(80, min_after, "getLineMinSortedLink (min_after = %ld)", l);

	int res = -1;
	int s = l->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		int val = m_diag->get_vertex_pos((CDVertex*)l->get_link(i));
		if ((val < m_verts_sorted) && (val > min_after)) {
			if ((res < 0) || (val < res)) res = val;
		}
	}
	return res;
}

int CLDiagArrange::getVertMinSortedLink(CDVertex* v, int min_after) {
	dbg->log(80, min_after, "getVertMinSortedLink (min_after = %ld)", v);

	int res = -1;
	int s = v->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		int val = m_diag->get_line_pos((CDLine*)v->get_link(i));
		if ((val < m_lines_sorted) && (val > min_after)) {
			if ((res < 0) || (val < res)) res = val;
		}
	}
	return res;
}

/*
*	calc how many links of line points to vertex
*/
int CLDiagArrange::getNumLinksToSortedVert(CDLine* l, int v) {
	dbg->log(80, v, "getNumLinksToSortedVert (v = %ld)", l);

	if (v >= m_verts_sorted) { throw(new CEException(CEException::INT_ERROR));}
	int res = 0;
	int s = l->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		int val = m_diag->get_vertex_pos((CDVertex*)l->get_link(i));
		if (val == v) res++;
	}
	return res;
}

/*
*	calc how many links of vertex points to line
*/
int CLDiagArrange::getNumLinksToSortedLine(CDVertex* v, int l) {
	dbg->log(80, l, "getNumLinksToSortedLine (l = %ld)", v);

	if (l >= m_lines_sorted) { throw(new CEException(CEException::INT_ERROR));}
	int res = 0;
	int s = v->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		int val = m_diag->get_line_pos((CDLine*)v->get_link(i));
		if (val == l) res++;
	}
	return res;
}

// note: this vertex supposed to be not sorted
int CLDiagArrange::getMinVertexLinked(CDVertex* v, int min_after) {
	dbg->log(80, min_after, "getMinVertexLinked (min_after = %ld)", v);

	if (m_diag->get_vertex_pos(v) < m_verts_sorted) {
		dbg->log(0, "Warning: trying to sort already sorted vertex (Internal Error).");
		return m_diag->get_vertex_pos(v);
	}
	
	int res = -1;
	int s = v->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		int val = getLineMinSortedLink((CDLine*)v->get_link(i), min_after);
		if ((val < m_lines_sorted) && (val > min_after)) {
			if ((res < 0) || (val < res)) res = val;
		}
	}
	return res;
}

// note: this vertex supposed to be not sorted
int CLDiagArrange::getNumLinksToOtherVertex(CDVertex* v, int v_other) {
	dbg->log(80, v_other, "getNumLinksToOtherVertex (v_other = %ld)", v);

	if (m_diag->get_vertex_pos(v) == v_other) {
		dbg->log(0, "Warning: calc links to itself (Internal Error).");
		return 1;
	}
	if (v_other >= m_lines_sorted) { throw(new CEException(CEException::INT_ERROR));}
	
	int res = 0;
	int s = v->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		int val = getNumLinksToSortedVert((CDLine*)v->get_link(i), v_other);
		res += val;
	}
	return res;
}

bool CLDiagArrange::areAllLinksSorted(CDVertex* v) {
	dbg->log(80, "areAllLinksSorted", v);

	int s = v->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		if (m_diag->get_line_pos((CDLine*)v->get_link(i)) >= m_lines_sorted) return false;
	}
	return true;
}

bool CLDiagArrange::areAllLinksSorted(CDLine* l) {
	dbg->log(80, "areAllLinksSorted", l);

	int s = l->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		if (m_diag->get_vertex_pos((CDVertex*)l->get_link(i)) >= m_verts_sorted) return false;
	}
	return true;
}

bool CLDiagArrange::areLinesParallel(CDLine* a, CDLine* b) {
	dbg->log(40, "areLinesParallel()");

	if ((a->get_link(0) == b->get_link(0)) && (a->get_link(1) == b->get_link(1))) return true;
	if ((a->get_link(0) == b->get_link(1)) && (a->get_link(1) == b->get_link(0))) return true;
	return false;
}

/*
*	check if verices have similar links (in similar order)
*/
bool CLDiagArrange::areVerticesParallel(CDVertex* a, CDVertex* b) {
	dbg->log(80, "areVerticesParallel()");

	if (a == b) return true;
	int s = a->get_links_number();
	if (s != b->get_links_number()) return false;
	int i;
	for(i=0; i<s; i++) {
		int pos_1 = m_diag->get_line_pos((CDLine*)a->get_link(i));
		int pos_2 = m_diag->get_line_pos((CDLine*)b->get_link(i));
		if (pos_1 != pos_2) return false;
	}
	return true;
}

/*
*	i.e. starts from vertex with lower position
*	returns (pk>0) for external lines
*	false for internal with unsorted vertices
*	and true for self-loops (equal links)
*/
bool CLDiagArrange::isLineDirect(CDLine* l) {
	dbg->log(80, "isLineDirect", l);

	int s = l->get_links_number();
	if (l->is_external()) { return (l->get_particle_kind() > 0); }
	if (s != 2) { dbg->log(0, s, "Internal Line with unexpected number of links: %ld"); throw(new CEException(CEException::INT_ERROR)); }

	int v1 = m_diag->get_vertex_pos((CDVertex*)l->get_link(0));
	int v2 = m_diag->get_vertex_pos((CDVertex*)l->get_link(1));
	if ((v1 >= m_verts_sorted) || (v2 >= m_verts_sorted)) return false;
	
	return (v1 <= v2);
}

// "1" means no similar vertices repeated, "0" means error
int CLDiagArrange::GetEqualsVerticesNum(int first_pos) {
	dbg->log(60, first_pos, "GetEqualsVerticesNum (first_pos = %ld)");

	int s = m_diag->get_vertices_number();
	if (first_pos >= s) { dbg->log(0, "Error: {GetEqualsVerticesNum} vertex does not exist"); return 0;}
	int res = 1;
	int i;
	for(i=first_pos+1; i<s; i++) {
		int cmp_res = VertexABeforeB(first_pos, i);
		dbg->log(60, cmp_res, "Vertices comparison result: %ld");
		if (cmp_res != CMP_A_EQUAL_B) break;
		res++;
	}
	return res;
}

// "1" means no similar lines repeated, "0" means error
int CLDiagArrange::GetEqualsLinesNum(int first_pos) {
	dbg->log(60, first_pos, "GetEqualsLinesNum (first_pos = %ld)");

	int s = m_diag->get_lines_number();
	if (first_pos >= s) { dbg->log(0, "Error: {GetEqualsLinesNum} line does not exist"); return 0; }
	int res = 1;
	int i;
	for(i=first_pos+1; i<s; i++) {
		int cmp_res = LineABeforeB(first_pos, i);
		dbg->log(60, cmp_res, "Lines comparison result: %ld");
		if (cmp_res != CMP_A_EQUAL_B) break;
		res++;
	}
	return res;
}

/*
*	Vertices' links should point to objects with similar position
*/
bool CLDiagArrange::AreVerticesSimilar(CDVertex* v1, CFeynDiagram* d1,
									   CDVertex* v2, CFeynDiagram* d2) {
	dbg->log(40, "AreVerticesSimilar()");

	if ((v1 == NULL) || (v2 == NULL) || (d1 == NULL) || (d2 == NULL)) { dbg->log(0, "Error: {AreVerticesSimilar} NULL parameter"); return false; }

	CDiagramVertexInfo* vi1 = v1->get_vertex_info();
	if (vi1 == NULL) { dbg->log(0, "ERROR: NULL Vertex Info structure"); throw(new CEException(CEException::INT_ERROR)); }
	if (!vi1->isEqualTo(v2->get_vertex_info())) return false;
	
	int s = v1->get_links_number();
	if (s != v2->get_links_number()) return false;
	
	int i;
	for(i=0; i<s; i++) {
		int pos_1 = d1->get_line_pos((CDLine*)v1->get_link(i));
		int pos_2 = d2->get_line_pos((CDLine*)v2->get_link(i));
		if (pos_1 != pos_2) return false;
	}

	dbg->log(80, "AreVerticesSimilar() result: TRUE");
	return true;
}

/*
*	Lines' links should point to objects with similar position
*/
bool CLDiagArrange::AreLinesSimilar(CDLine* l1, CFeynDiagram* d1,
									CDLine* l2, CFeynDiagram* d2) {
	dbg->log(40, "AreLinesSimilar()");

	if ((l1 == NULL) || (l2 == NULL) || (d1 == NULL) || (d2 == NULL)) { dbg->log(0, "Error: {AreLinesSimilar} NULL parameter"); return false; }
	if (l1->get_particle_kind() != l2->get_particle_kind()) return false;
	
	int s = l1->get_links_number();
	if (s != l2->get_links_number()) return false;
	
	int i;
	for(i=0; i<s; i++) {
		int pos_1 = d1->get_vertex_pos((CDVertex*)l1->get_link(i));
		int pos_2 = d2->get_vertex_pos((CDVertex*)l2->get_link(i));
		if (pos_1 != pos_2) return false;
	}

	dbg->log(80, "AreLinesSimilar() result: TRUE");
	return true;
}
