////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LDiagLocateOrdinary.cpp: implementation of the CLDiagLocateOrdinary class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const int CLDiagLocateOrdinary::CALC_MEDIUM_NEGLECT_FINALS	= 0x0001;

const int CLDiagLocateOrdinary::LOCATE_START_FROM_CENTER		= 0x0010;
const int CLDiagLocateOrdinary::LOCATE_START_FROM_HALF_CENTER	= 0x0020;
const int CLDiagLocateOrdinary::LOCATE_START_FROM_LINKS_POS		= 0x0040;

const int CLDiagLocateOrdinary::LOCATE_FINALS_NEGLECT_POS		= 0x0001;
const int CLDiagLocateOrdinary::LOCATE_FINALS_USE_X_POS_ONLY	= 0x0002;
const int CLDiagLocateOrdinary::LOCATE_FINALS_USE_POS			= 0x0004;
const int CLDiagLocateOrdinary::LOCATE_FINALS_ANY				= 0x0007;

const int CLDiagLocateOrdinary::LOCATE_TRY_OTHER_SETTINGS		= 0x1000;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLDiagLocateOrdinary::CLDiagLocateOrdinary(CFeynDiagram* diag, CKinematics* kin) : CLDiagLocateView(diag, kin)
{
	m_vertex_series = new CEPtrArray();
	//m_locate_settings = LOCATE_START_FROM_CENTER | LOCATE_FINALS_USE_X_POS_ONLY;
	m_locate_settings = LOCATE_START_FROM_CENTER | LOCATE_FINALS_USE_POS | LOCATE_TRY_OTHER_SETTINGS;
}

CLDiagLocateOrdinary::~CLDiagLocateOrdinary()
{
	// delete all the series in m_vertex_series
	// without deleting objects in series
	int s = m_vertex_series->size();
	int i;
	for(i=0; i<s; i++) {
		CEPtrArray* item = (CEPtrArray*)m_vertex_series->GetAt(0);
		item->clearall_without_delete_objects();
		m_vertex_series->RemoveAt(0, true);
	}
	delete m_vertex_series;
}

bool CLDiagLocateOrdinary::do_locate_vertices() {
	if (m_diag == NULL) return false;
	//if (m_kin == NULL) return false;

	dbg->log(30, "CLDiagLocateOrdinary::do_locate_vertices involved.");

	// external lines should already be allocated.

	int ser_num;
	for(ser_num=0;; ser_num++) {
		CEPtrArray* series = create_next_vertex_series();
		if (series == NULL) break;

		// allocate vertices from series to center
		// and add series to m_vertex_series array
		allocate_vertex_series(series);

		// arange locations of vertices
		arrange_locations_of_series(ser_num);//(ser_num == (m_vertex_series->size()-1))
	}

	if (m_locate_settings & LOCATE_TRY_OTHER_SETTINGS) {
		// if some vertices are too close try it reallocate
		int closes_num = CalcCloseVerticesNum();
		if (closes_num>0) {
			int old_setts = m_locate_settings;
			m_locate_settings &= (!LOCATE_TRY_OTHER_SETTINGS); // avoid recursion
			int old_setts_with_no_recursion = m_locate_settings;

			// clear LOCATE_FINALS_ (ANY) flag
			m_locate_settings &= (!LOCATE_FINALS_ANY);

			bool good_config_found = false;
			int setts_count;
			for(setts_count=0; ((setts_count<3) && (!good_config_found)); setts_count++) {
				if (setts_count == 0) {
					// 1. LOCATE_FINALS_USE_POS
					if (old_setts & LOCATE_FINALS_USE_POS) continue;
					m_locate_settings |= LOCATE_FINALS_USE_POS;
				} else if (setts_count == 1) {
					// 2. LOCATE_FINALS_USE_X_POS_ONLY
					if (old_setts & LOCATE_FINALS_USE_X_POS_ONLY) continue;
					m_locate_settings |= LOCATE_FINALS_USE_X_POS_ONLY;
				} else {
					// 3. LOCATE_FINALS_NEGLECT_POS
					if (old_setts & LOCATE_FINALS_NEGLECT_POS) continue;
					m_locate_settings |= LOCATE_FINALS_NEGLECT_POS;
				}

				m_diag->clear_locate_data(); // clear current data
				localize_external_lines();
				bool loc_vert_OK = do_locate_vertices();

				if (loc_vert_OK) {
					int closes_num_new = CalcCloseVerticesNum();
					if (closes_num_new == 0) good_config_found = true;
				}
			}
			
			if (!good_config_found) {
				// return back
				m_diag->clear_locate_data();
				m_locate_settings = old_setts_with_no_recursion;
				localize_external_lines();
				do_locate_vertices();
			}
			m_locate_settings = old_setts;
		}
	}

	dbg->log(30, "All vertex series allocated:", m_vertex_series);
	return true;
}

/*
*	DEBUG IT! fails if caled twice. What's wrong?
*/
bool CLDiagLocateOrdinary::do_update_vertices() {
	if (m_diag == NULL) return false;
	dbg->log(30, "CLDiagLocateOrdinary::do_update_vertices() involved.");
	// ALL items should already be allocated.

	int i;
	for(i=0; i<m_v_num; i++) {
		CDVertex* v = m_diag->get_vertex(i); if(v==NULL) return false;
		//CPos2D* pos = m_center_pos->clone();
		CPos2D* pos = calc_best_vertex_pos(v); if (pos==NULL) continue;
		CDElementViewInfo* pos_info = v->getViewInfo(); if (pos_info==NULL)return false;
		CPos2D* pos_v = pos_info->get_pos();
		if (pos_v!=NULL) pos_v->set_XY(pos->get_double_X(),pos->get_double_Y());
		delete pos;
	}
	return true;
}

/*
*	NULL means all the vertices are allocated or
*	there are invalid bonds in diagram (some vertices unlinked to externals)
*/
CEPtrArray* CLDiagLocateOrdinary::create_next_vertex_series()
{
	// find oput vertices with maximal number of allocated neighbours
	// OR (different way, perhaps better) with minimal number of unallocated ones
	dbg->log(30, m_v_num, "CLDiagLocateOrdinary::create_next_vertex_series involved (total %ld vertices).");

	int AL_num = -1;
	CEPtrArray* series = new CEPtrArray();
	//CFeynDiagram* m_diag;

	int i;
	for(i=0; i<m_v_num; i++) {
		CDVertex* v = m_diag->get_vertex(i);
		CDElementViewInfo* pos_info = v->getViewInfo();
		if (pos_info == NULL) {
			// define number of localized links

			int AL_curr = 0;
			int unAL_curr = 0;
//			CEPtrArray* v_links = v->get_links();
//			int links_size = v_links->size();

			int links_size = v->get_links_number();
			int j;
			for(j=0; j<links_size; j++) {
				CDElementViewInfo* link_pos_info = get_link_pos_info(v, j);
				
				if (link_pos_info != NULL) {AL_curr++;}
				else {unAL_curr++;}
			}

			// use value number of allocated links (AL_curr)
			if (AL_curr > 0) {
				// vertices without ony link will not be located
				if (AL_curr == AL_num) {
					series->add(v);
				}
				else if (AL_curr > AL_num) {
					series->clearall_without_delete_objects();
					//delete series; series = new CEPtrArray();
					series->add(v);
					AL_num = AL_curr;
				}
			}
		}
	}

	int series_zize = series->size();
	if (series_zize == 0) {delete series; return NULL;}
	dbg->log(30, series_zize, "creates series of size %ld.");
	return series;
}

CDElementViewInfo* CLDiagLocateOrdinary::get_link_pos_info(CDVertex* v, int link_no)
{
	dbg->log(55, link_no, "get_link_pos_info, (link_no = %ld) vertex: ", v);
	if ((v == NULL) || (link_no < 0) || (link_no >= v->get_links_number())) return NULL;
	CDElementViewInfo* res = NULL;
	
	// vertex can be linked to external line or to other vertex
	CDLine* line_linked = (CDLine*)(v->get_links()->GetAt(link_no));
	if (line_linked->is_external()) {
		res = line_linked->getViewInfo();
	}
	else {
		CDVertex* v_linked = v->get_vertex_linked(link_no);
		if (v_linked != NULL) {
			res = v_linked->getViewInfo();
		}
		else {
			// Invalid diagram structure
			printf("CLDiagLocateOrdinary Error: Invalid diagram structure.\n");
			// Try to continue.
		}
	}

	dbg->log(55, "get_link_pos_info, result: ", res);
	return res;
}

/*
*	Set initial location for all the vertices in series.
*	And add series to series list.
*/
void CLDiagLocateOrdinary::allocate_vertex_series(CEPtrArray* series)
{
	dbg->log(20, "allocate_vertex_series : ", series);
	CEPtrArray* init_locations = new CEPtrArray();
	int s = series->size();
	int i;
	for(i=0; i<s; i++) {
		CDElement* el = (CDElement*)series->GetAt(i);
		
		//LOCATE_START_FROM_CENTER; LOCATE_START_FROM_HALF_CENTER; LOCATE_START_FROM_LINKS_POS;
		CPos2D* start_pos = NULL;

		if (m_locate_settings & (LOCATE_START_FROM_LINKS_POS | LOCATE_START_FROM_HALF_CENTER)) {
			// calc links average position
			start_pos = calc_medium_pos((CDVertex*)el, 0);
			if (start_pos == NULL) { // No links (may it occur?)
				start_pos = m_center_pos->clone();
			}

			if (m_locate_settings & LOCATE_START_FROM_HALF_CENTER) {
				start_pos->AddPos(m_center_pos);
				start_pos->MultiplyFactor(0.5);
			}
		}
		else {
			// LOCATE_START_FROM_CENTER is default, do not check it
			start_pos = m_center_pos->clone();
		}

		init_locations->add(new CDElementViewInfo(start_pos));
	}

	for(i=0; i<s; i++) {
		CDElement* el = (CDElement*)series->GetAt(i);
		el->setViewInfo((CDElementViewInfo*)init_locations->GetAt(i));
	}

	init_locations->clearall_without_delete_objects(); delete init_locations;
	m_vertex_series->add(series);
}

void CLDiagLocateOrdinary::arrange_locations_of_series(int num)
{
	dbg->log(25, num, "arrange_locations_of_series no. %ld.");
	if (num < 0) return;

	int series_array_size = m_vertex_series->size();
	if (num >= series_array_size) {
		// Error.
		printf("CLDiagLocateOrdinary::arrange_locations_of_series() ERROR: invalid argument.\n"); return;
	}

	CEPtrArray* series = (CEPtrArray*)m_vertex_series->GetAt(num);
	CEPtrArray* vertex_best_positions = new CEPtrArray();
	int s = series->size();
	int i;
	for(i=0; i<s; i++) {
		//CPos2D* move_vec = calc_vertex_move_vector((CDVertex*)series->GetAt(i));
		CPos2D* best_pos = calc_best_vertex_pos((CDVertex*)series->GetAt(i));
		if (best_pos == NULL) {
			// Error. Do not terminate calculation.
			best_pos = new CPos2D(0, 0);
		}
		vertex_best_positions->add(best_pos);
	}

	dbg->log(40, "vertex_best_positions array filled: ", vertex_best_positions);

	for(i=0; i<s; i++) {
		CPos2D* best_pos = (CPos2D*)vertex_best_positions->GetAt(i);
		CDElement* v = (CDElement*)series->GetAt(i);
		//v->getViewInfo()->get_pos()->AddPos(move_vec);
		//delete move_vec;
		v->setViewInfo(new CDElementViewInfo(best_pos));
	}

	vertex_best_positions->clearall_without_delete_objects();
	delete vertex_best_positions;
	
	dbg->log(40, "Vertex series allocated:", series);
	arrange_locations_of_series(num-1);
}

/*
*	NULL means undefined (no localized links found).
*/
CPos2D* CLDiagLocateOrdinary::calc_vertex_move_vector(CDVertex* v) {
	dbg->log(30, "calc_vertex_move_vector :", v);

	CPos2D* res = calc_medium_pos(v, CALC_MEDIUM_NEGLECT_FINALS);

	dbg->log(35, "calc_vertex_move_vector result :", res);
	return res;
}

CPos2D* CLDiagLocateOrdinary::calc_best_vertex_pos(CDVertex* v) {
	dbg->log(30, "calc_best_vertex_pos :", v);
	
	CPos2D* res = NULL;

	int links_involved_X = 0;
	int links_involved_Y = 0;
	
	int s = v->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		CDElementViewInfo* link_pos_info = get_link_pos_info(v, i);
		
		if (link_pos_info != NULL) {
			CPos2D* pos = link_pos_info->get_pos(); if(pos==NULL)continue; bool del_pos=false;

			bool is_final = false;
			CDLine* line_linked = (CDLine*)(v->get_link(i)); if(line_linked==NULL)continue;
			if (line_linked->is_external()) {
				if (!(((CDExtLine*)line_linked)->is_initial())) is_final = true;
			}

			if ((is_final) && ((m_locate_settings & LOCATE_FINALS_USE_POS) == 0)) {
				if (m_locate_settings & LOCATE_FINALS_NEGLECT_POS) continue;

				// LOCATE_FINALS_USE_X_POS_ONLY is default option, do not check it
				
				// if position belongs to final particle use only it's X-pos.
				// (for correct crossing visualization)
				pos = new CPos2D(pos->get_double_X(), 0); del_pos=true;
				links_involved_Y--;
			}
			
			if (res == NULL) {
				res = ((del_pos)?pos:pos->clone());
			} else {
				res->AddPos(pos); if (del_pos) delete pos;
			}
			links_involved_X++; links_involved_Y++;
		}
	}

	// check up Y-position.
	// final external particle (unused above) could be it's ontly link
	if ((res != NULL) && (links_involved_Y <= 0)) {
		// calc Y-pos again with no dropping of finals
		delete res;
		res = calc_medium_pos(v);
	} else {
		// do divide result to number of links_involved
		if ((res != NULL) && (links_involved_X>0) && (links_involved_Y>0)) {
			if (links_involved_X==1) { CPos2D poxA(m_center_pos->get_double_X(),0); res->AddPos(&poxA); links_involved_X++; }
			if (links_involved_Y==1) { CPos2D poxA(0,m_center_pos->get_double_Y()); res->AddPos(&poxA); links_involved_Y++; }
			double f_X = 1./((double)links_involved_X);
			double f_Y = 1./((double)links_involved_Y);
			res->MultiplyFactors(f_X, f_Y);
		}
	}

	dbg->log(35, "calc_best_vertex_pos result :", res);
	return res;
}

CPos2D* CLDiagLocateOrdinary::calc_medium_pos(CDVertex* v, int setts) {
	dbg->log(35, "calc_medium_pos for vector:", v);
	CPos2D* res = NULL;

	int links_involved = 0;
	int s = v->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		CDElementViewInfo* link_pos_info = get_link_pos_info(v, i);
		
		if (link_pos_info != NULL) {
			CPos2D* link_pos_copy = link_pos_info->get_pos()->clone();

			if (setts & CALC_MEDIUM_NEGLECT_FINALS) {
				CDLine* line_linked = (CDLine*)(v->get_links()->GetAt(i));
				if (line_linked->is_external()) {
					if (!(((CDExtLine*)line_linked)->is_initial())) {
						// if position belongs to final particle use only it's X-pos.
						// (for correct crossing visualization)
						double x_new = link_pos_copy->get_double_X();
						//double old_y = link_pos_copy->get_double_Y();
						double y_new = m_center_pos->get_double_Y();
						delete link_pos_copy;
						link_pos_copy = new CPos2D(x_new, y_new);
						// y is taken to center
					}
				}
			}

			if (res == NULL) res = link_pos_copy;
			else { res->AddPos(link_pos_copy); delete link_pos_copy; }
			links_involved++;
		}
	}

	// do divide result to number of links_involved
	if ((res != NULL)&&(links_involved>0)) {
		double f = 1./((double)links_involved);
		res->MultiplyFactor(f);
	}

	dbg->log(35, "calc_medium_pos result :", res);
	return res;
}

/*
*	returns the number of close vertices pairs
*/
int CLDiagLocateOrdinary::CalcCloseVerticesNum() {
	dbg->log(30, "CLDiagLocateOrdinary: CalcCloseVerticesNum() check call.");

	int res = 0;
	int i;
	for(i=0; i<m_v_num; i++) {
		CDVertex* v1 = m_diag->get_vertex(i);
		CDElementViewInfo* pos_info_1 = v1->getViewInfo();
		if (pos_info_1 == NULL) continue;
		CPos2D* pos_1 = pos_info_1->get_pos();
		if (pos_1 == NULL) continue;
		
		int j;
		for(j=i+1; j<m_v_num; j++) {
			CDVertex* v2 = m_diag->get_vertex(j);
			CDElementViewInfo* pos_info_2 = v2->getViewInfo();
			if (pos_info_2 == NULL) continue;
			CPos2D* pos_2 = pos_info_2->get_pos();
			if (pos_2 == NULL) continue;

			// compare positions
			double min_gap = 0.002;
			double x_gap = CLMath::Abs(pos_1->get_double_X()-pos_2->get_double_X());
			if (x_gap >= min_gap) continue;
			double y_gap = CLMath::Abs(pos_1->get_double_Y()-pos_2->get_double_Y());
			if (y_gap >= min_gap) continue;
			res++;
		}
	}
	dbg->log(30, res, "CLDiagLocateOrdinary: CalcCloseVerticesNum() result %ld.");
	return res;
}
