////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LDiagLocateView.cpp: implementation of the CLDiagLocateView class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLDiagLocateView::dbg = new CDebug("CLDiagLocateView");

//////////////////////////////////////////////////////////////////////
// Static Entry
//////////////////////////////////////////////////////////////////////
const int CLDiagLocateView::ALGORITHM_DEFAULT = 1;

bool CLDiagLocateView::LocateDiagramElements(CFeynDiagram* diag, int AlgorithmID, int DrawFlags)
{
	dbg->log(10, AlgorithmID, DrawFlags, "LocateDiagramElements (Algorithm: %ld, Flags: %x)", diag);
	if (diag == NULL) return false;
	if (AlgorithmID < 0) AlgorithmID = ALGORITHM_DEFAULT;
	int AlgorithmOld = diag->get_locate_algorithm();
	if (AlgorithmOld == AlgorithmID) return true;
	if (AlgorithmOld >= 0) {
		// clear old positions
		diag->clear_locate_data();
	}

	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) return false;

	CLDiagLocateView* lib = NULL;
	if (AlgorithmID == ALGORITHM_DEFAULT) {
		lib = new CLDiagLocateOrdinary(diag, kin);
	}

	if (lib == NULL) {
		printf("Locate Diagram Elements ERROR: No algorithm found\n");
		return false;
	}
	
	//	localize procedure
	//	Note: diagrams with/without 'bSwapAllowed' may look DIFFERENT even if the order of tales occurs similar.
	//	See also InternalNotes[2.1].
	bool swap_finals = ((DrawFlags & CFeynDiagram::DRAW_SWAP_TALES) != 0);
	bool res = false;
	if(swap_finals) {
		//TODO: debug it
		//	1. drop out flag 'swap_finals' due to check 'if (kin->num_final_particles()>4) {'???
		//	2. 'TODO' in do_update_vertices() ???
		//	3. 2xdo_update_vertices calls ???
		//	4. Why not simply 'do_locate_vertices()' after last localize_external_lines() ???
		// Test on:
		//>SetKinematics(2,3,PROTON,"p\_1","e\_1",ELECTRON,"k\_1","g\_1",PROTON,"p\_2","e\_2",ELECTRON,"k\_2","g\_2",PHOTON,"k","g"); 
		//>SetDiagramPhysics(PHYS_QED|PHYS_PROTON_STRUCTFUNC); 
		//>DrawDiagrams(ComposeDiagrams(5),texfile,DD_DEFAULT or DD_DEFAULT|DD_DONT_SWAP_TALES);

		lib->localize_external_lines(true,true,true,true);
		res = lib->do_locate_vertices();

		long itNo=0; //iterations
		for(;itNo<10;itNo++) {
			int ln = lib->localize_final_lines_by(true);
			diag->clear_locate_data(false,true);
			res = lib->do_locate_vertices();
			if (ln==0) break;
		}
		lib->localize_final_lines_by(true);
		lib->localize_external_lines(false,true,false,false);
		//if (itNo==10) { lib->localize_final_lines_by(true); }

		//Note. The loop diagrams are much better NOW, before last 'do_update_vertices' calls.
		//TODO: make loop diagrams better?
		lib->do_update_vertices();
		if (kin->num_final_particles()>4) {
			lib->scale_vertex_points();
		} else {
			lib->do_update_vertices();
		}
	} else {
		lib->localize_external_lines(false,true,true,false);
		res = lib->do_locate_vertices();
	}
	if (res) { res = lib->do_locate_internal_lines(); }
	delete lib;

	diag->set_locate_algorithm(AlgorithmID);
	dbg->log(10, AlgorithmID, "LocateDiagramElements processed SUCCESSFULL (Algorithm == %ld):", diag);
	return res;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLDiagLocateView::CLDiagLocateView(CFeynDiagram* diag, CKinematics* kin) : CELibrary()
{
	m_diag = diag;
	m_kin = kin;

	if ((m_diag == NULL) || (m_kin == NULL)) { printf("Error initializing CLDiagLocateView.\n"); throw(new CEException(CEException::INT_ERROR)); }

	m_l_num = m_diag->get_lines_number();
	m_v_num = m_diag->get_vertices_number();
	
	m_center_pos = new CPos2D(0.5, 0.5);
}

CLDiagLocateView::~CLDiagLocateView()
{
	delete m_center_pos;
}

//////////////////////////////////////////////////////////////////////
//	Set external lines positions.
//////////////////////////////////////////////////////////////////////
void CLDiagLocateView::localize_external_lines(bool bFinalsToPoint, bool bInits, bool bFinals, bool bInitsToMargins) {

	// determine gap between external tales
	// the gap will be uniform for internal & external lines
	int i_num = m_kin->num_init_particles();
	int f_num = m_kin->num_final_particles();

	int max_num = (f_num>i_num)?f_num:i_num;
	if (max_num < 2) {printf("CLDiagDrawingPrepare ERROR: 1->1 process.\n"); throw(new CEException(CEException::INT_ERROR)); }

	double border = 0;//m_border_widht
	double f_ext_y_del = (((double)1.-border*2))/(max_num-1);
	double i_ext_y_del = f_ext_y_del;
	double i_y_curr = border;
	double f_y_curr = border;
	if (f_num>i_num) {
		if ((bInitsToMargins)&&(i_num==2)) {
			i_ext_y_del = (((double)1.-border*2));
		} else {
			i_y_curr = i_ext_y_del*(max_num-i_num)/2;// the same is ((double)1.-ext_del*(i_num-1))/2;
		}
	} else {
		f_y_curr = f_ext_y_del*(max_num-f_num)/2;// the same is ((double)1.-ext_del*(f_num-1))/2;
	}
	double i_x_curr = border;
	double f_x_curr = (1.-border);

	// m_border_widht is required for ensure that
	// "i_y_curr += ext_y_del"-operations will not exceed "1."
	// 
	int i;
	for(i=0; i<m_l_num; i++) {
		CDLine* l = m_diag->get_line(i);
		if (l->is_external()) {
			if (l->DE_kind() & CDElement::EL_LN_INITIAL) {
				if (bInits) {
					l->setViewInfo(new CDElementViewInfo(i_x_curr, i_y_curr));
					i_y_curr += i_ext_y_del;
				}
			}
			else {
				if (bFinals) {
					if(bFinalsToPoint) {
						l->setViewInfo(new CDElementViewInfo(f_x_curr, border+f_ext_y_del/2));
					} else {
						l->setViewInfo(new CDElementViewInfo(f_x_curr, f_y_curr));
						f_y_curr += f_ext_y_del;
					}
				}
			}
		}
	}
}

int CLDiagLocateView::localize_final_lines_by(bool by_vertices) {
	int f_num = m_kin->num_final_particles();
	int res=0;

	int max_num = f_num;
	if (max_num < 2) {printf("CLDiagDrawingPrepare ERROR: 1->1 process.\n"); throw(new CEException(CEException::INT_ERROR)); }

	double border = 0;//m_border_widht
	double ext_y_del = (((double)1.-border*2))/(max_num-1);
	double f_y_curr = border;
	double f_x_curr = (1.-border);
	int i; int f_no;

	CIntArray f_move(f_num); f_move.ToSize(f_num); f_move.FillAllSeq(0);
	CIntArray f_poss(f_num); f_poss.ToSize(f_num); f_poss.FillAllSeq(0);
	CEDoubleArray f_Yvals(f_num); f_Yvals.ToSize(f_num); f_Yvals.FillAll(0);
	CEDoubleArray f_Yvals2(f_num); f_Yvals2.ToSize(f_num); f_Yvals2.FillAll(0);
	CEDoubleArray f_YLines(f_num); f_YLines.ToSize(f_num); f_YLines.FillAll(0);
	bool use_Y2=!by_vertices;

	f_no=0;
	for(i=0; i<m_l_num; i++) { // fill y-data
		CDLine* l = m_diag->get_line(i);
		if (!l->is_external()) continue;
		if (l->DE_kind() & CDElement::EL_LN_INITIAL) continue;
		CDExtLine* ext_line = (CDExtLine*)l; f_no++;
		double y=0;
		//CDElementViewInfo* vi=l->getViewInfo(); if(vi!=NULL){ CPos2D* pos=vi->get_pos(); if(pos!=NULL)y=pos->get_double_Y(); }
		//if (y<0)y=0;
		if (ext_line->get_links_number() != 1) continue;
		CDVertex* v = (CDVertex*)ext_line->get_link(0); if(v==NULL)continue;
		CDElementViewInfo* vi = v->getViewInfo(); if(vi==NULL)continue;
		CPos2D* pos = vi->get_pos(); if(pos==NULL)continue;
		y=pos->get_double_Y();

		if (use_Y2) {
			double y2=0; int y_num=0;
			int s = v->get_links_number();
			int i;
			for(i=0; i<s; i++) {
				CDLine* line_linked = (CDLine*)(v->get_links()->GetAt(i)); if(line_linked==NULL)continue;
				//if (line_linked->is_external()) { if (!(((CDExtLine*)line_linked)->is_initial())) continue; }
				CDElementViewInfo* link_pos_info = CLDiagLocateOrdinary::get_link_pos_info(v, i); if(link_pos_info==NULL)continue;
				pos = link_pos_info->get_pos(); if(pos==NULL)continue;
				y2+=pos->get_double_Y(); y_num++;
			}
			if (y_num>0) y2=y2/y_num;
			f_Yvals2.SetAt(f_no-1,y2);
		}
		f_Yvals.SetAt(f_no-1,y);
	}
	if( f_no!=f_num ) dbg->log(0,"WARNING in CLDiagLocateView::localize_final_lines_by_vertices()");

	char* dbg_str=NULL; bool log_arrays = (dbg->get_level() >= 100);
	if (log_arrays) {
		dbg_str = (char*)CMemoryManager::mmalloc(10000); *dbg_str='\0';//DEBUG
		f_Yvals.SprintfTo(dbg_str,10000); printf("f_Yvals:  "); printf(dbg_str); printf("\n");//DEBUG
		if (use_Y2) {
			f_Yvals2.SprintfTo(dbg_str,10000); printf("f_Yvals2: "); printf(dbg_str); printf("\n");//DEBUG
		}
	}

	for(i=1; i<f_num; i++) { // sort lines
		double y = f_Yvals.GetAt(i); int pos = f_move.GetAt(i);
		if (use_Y2) { double y2=f_Yvals2.GetAt(i); if ((y2>0) && (y2<y)) y=y2; }
		bool bImoved=false;
		for(int j=i-1; j>=0; j--) {
			double yJ=f_Yvals.GetAt(j);
			// to swap if (y<yJ)
			if (use_Y2) { double y2J=f_Yvals2.GetAt(j); if ((y2J>0)&&(y2J>yJ)) yJ=y2J; }
			if (y >= yJ) break;
			f_Yvals.SetAt(j+1,yJ); f_move.SetAt(j+1,f_move.GetAt(j));
			f_Yvals.SetAt(j,y); f_move.SetAt(j,pos);
			bImoved=true;
		}
		//if (bImoved) res++;
	}
	for(i=0; i<f_num; i++) { f_poss.SetAt(f_move.GetAt(i),i); }

	if (log_arrays) {
		f_poss.SprintfTo(dbg_str,10000); printf("f_poss: "); printf(dbg_str); printf("\n");//DEBUG
	}

	// check order
	double y_prev=-1;
	for(i=0; i<f_num; i++) {
		double y=f_YLines.GetAt(f_move.GetAt(i));
		if (y_prev > y) { res++; }
		y_prev=y;
	}

	f_no=0;
	for(i=0; i<m_l_num; i++) { // set values
		CDLine* l = m_diag->get_line(i);
		if (!l->is_external()) continue;
		if (l->DE_kind() & CDElement::EL_LN_INITIAL) continue;
		
		double y = border+ext_y_del*f_poss.GetAt(f_no);
		l->setViewInfo(new CDElementViewInfo(f_x_curr, y));
		f_no++;
		if (log_arrays) {
			f_Yvals.SetAt(f_no-1,l->getViewInfo()->get_pos()->get_double_Y());//DEBUG
		}
	}
	if( f_no!=f_num ) dbg->log(0,"WARNING2 in CLDiagLocateView::localize_final_lines_by_vertices()");

	if (log_arrays) {
		f_Yvals.SprintfTo(dbg_str,10000); printf("vals: "); printf(dbg_str); printf("\n");//DEBUG
		CMemoryManager::mfree(dbg_str);//DEBUG
		//printf("res: %ld\n", res);//DEBUG
	}
	return res;
}

//////////////////////////////////////////////////////////////////////
//	Locate internal lines according to vertex positions.
//	Enumerate lines with similar vertices
//	for further drawing.
//////////////////////////////////////////////////////////////////////
bool CLDiagLocateView::do_locate_internal_lines() {
	if (m_diag == NULL) return false;

	int s = m_diag->get_lines_number();

	// create CDElementViewInfo for each line
	int i;
	for(i=0; i<s; i++) {
		CDElement* l = m_diag->get_line(i);
		if (l->getViewInfo() == NULL) {
			CEPtrArray* links = l->get_links();
			int links_size = links->size();
			if (links_size != 2) {
				// Error
				// external lines should be localized earlier
				printf("CLDiagLocateView::do_locate_internal_lines ERROR: unexpected links number.\n");
				return false;
			}

			l->setViewInfo(new CDElementViewInfo(NULL));
		}
	}

	// search for lines with similar pairs of links.
	for(i=0; i<s; i++) {
		CDElement* l1 = m_diag->get_line(i);
		CEPtrArray* links1 = l1->get_links();
		if (links1->size() != 2) continue;

		CEPtrArray* repeted_lines = NULL;

		CDElement* v11 = (CDElement*)links1->GetAt(0);
		CDElement* v12 = (CDElement*)links1->GetAt(1);

		int j;
		for(j=i+1; j<s; j++) {
			CDElement* l2 = m_diag->get_line(j);
			CEPtrArray* links2 = l2->get_links();
			if (links2->size() != 2) continue;
			
			CDElement* v21 = (CDElement*)links2->GetAt(0);
			CDElement* v22 = (CDElement*)links2->GetAt(1);

			// compare vertices and fill in repeated lines array
			if (((v11 == v21) && (v12 == v22)) || ((v11 == v22) && (v12 == v21))) {
				if (repeted_lines == NULL) {
					repeted_lines = new CEPtrArray();
					repeted_lines->add(l1);
					repeted_lines->add(l2);
				}
				else {
					repeted_lines->add(l2);
				}
			}
		}

		// if line 'l1' is repeated arrange repetition info
		if (repeted_lines != NULL) {
			int repetition_factor = repeted_lines->size();

			int rep_ID;
			for(rep_ID = 1; rep_ID<=repetition_factor; rep_ID++) {
				CDElement* l = (CDElement*)repeted_lines->GetAt(rep_ID-1);

				CDElement* vcurr_1 = (CDElement*)(l->get_links()->GetAt(0));

				int rep_ID_adv = rep_ID;
				if (v11 != vcurr_1) rep_ID_adv = -rep_ID_adv;

				l->getViewInfo()->set_repetition(rep_ID_adv, repetition_factor);
			}

			repeted_lines->clearall_without_delete_objects();
			delete repeted_lines;
		}
	}
	return true;
}

void CLDiagLocateView::scale_vertex_points() {
	if (m_diag == NULL) return;

	int s = m_diag->get_vertices_number();
	CIntArray f_move(s); f_move.ToSize(s); f_move.FillAllSeq(0);
	CIntArray f_poss(s); f_poss.ToSize(s); f_poss.FillAllSeq(0);
	CEDoubleArray f_Yvals(s); f_Yvals.ToSize(s); f_Yvals.FillAll(0);
	int i;
	for(i=0; i<s; i++) {
		double x=0;
		CDElementViewInfo* pos_info = m_diag->get_vertex(i)->getViewInfo();
		if (pos_info != NULL) { CPos2D* pos = pos_info->get_pos(); if (pos != NULL) x=pos->get_double_X(); }
		f_Yvals.SetAt(i,x);
	}

	for(i=1; i<s; i++) { // sort lines
		double y = f_Yvals.GetAt(i); int pos = f_move.GetAt(i);
		for(int j=i-1; j>=0; j--) {
			double yJ=f_Yvals.GetAt(j); // to swap if (y<yJ)
			if (y >= yJ) break;
			f_Yvals.SetAt(j+1,yJ); f_move.SetAt(j+1,f_move.GetAt(j));
			f_Yvals.SetAt(j,y); f_move.SetAt(j,pos);
		}
	}
	for(i=0; i<s; i++) { f_poss.SetAt(f_move.GetAt(i),i); }

	double x0=0.2; double x1=0.9; double y0=0; double y1=1; double x_del=((s>1)?((x1-x0)/(s-1)):(x1-x0));

	for(i=0; i<s; i++) {
		CDElementViewInfo* pos_info = m_diag->get_vertex(i)->getViewInfo(); if (pos_info == NULL) continue;
		CPos2D* pos = pos_info->get_pos(); if (pos == NULL) continue;
		pos->set_XY(x0+x_del*f_poss.GetAt(i),pos->get_double_Y());
	}
}

//////////////////////////////////////////////////////////////////////
//	Scaling util.
//////////////////////////////////////////////////////////////////////
// Scale positions of all the diagram elements.
// scale_center == NULL means center at (0,0) point
void CLDiagLocateView::scale_diagram(CFeynDiagram* diag, double scale, CPos2D* scale_center) {
	if (diag == NULL) return;

	int s = diag->get_vertices_number();
	int i;
	for(i=0; i<s; i++) {
		CDElementViewInfo* pos_info = diag->get_vertex(i)->getViewInfo();
		if (pos_info != NULL) {
			CPos2D* pos_old = pos_info->get_pos();
			if (pos_old != NULL) {
				CPos2D* pos_new = pos_old->clone();
				if (scale_center != NULL) pos_new->SubtrPos(scale_center);
				pos_new->MultiplyFactor(scale);
				if (scale_center != NULL) pos_new->AddPos(scale_center);

				pos_info->set_pos(pos_new);
			}
		}
	}
	
	s = diag->get_lines_number();
	for(i=0; i<s; i++) {
		CDElementViewInfo* pos_info = diag->get_line(i)->getViewInfo();
		if (pos_info != NULL) {
			CPos2D* pos_old = pos_info->get_pos();
			if (pos_old != NULL) {
				CPos2D* pos_new = pos_old->clone();
				if (scale_center != NULL) pos_new->SubtrPos(scale_center);
				pos_new->MultiplyFactor(scale);
				if (scale_center != NULL) pos_new->AddPos(scale_center);

				pos_info->set_pos(pos_new);
			}
		}
	}
}
