////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LDiagTopologyInspector.cpp: implementation of the CLDiagTopologyInspector class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLDiagTopologyInspector::dbg = new CDebug("CLDiagTopologyInspector");
const DWORD CLDiagTopologyInspector::NON_EULER_CIRCUIT			= 0x01;
const DWORD CLDiagTopologyInspector::LOOPS_IN_EXTERNAL_LINES	= 0x02;
const DWORD CLDiagTopologyInspector::ODD_LINEAR_FERMION_LOOP	= 0x04;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CLDiagTopologyInspector::CLDiagTopologyInspector() : CELibrary() {}
CLDiagTopologyInspector::~CLDiagTopologyInspector() {}

void CLDiagTopologyInspector::FilterDiagrams(CDiagramSet* diags, DWORD diag_type) {
	CEException::halt("Error: Non-stable code invoked (CLDiagTopologyInspector::FilterDiagrams)"); //TODO: remove this after examine
	if (diags == NULL) return;
	if (diag_type == 0) return;
	
	//////////////////////////////////////////////////////////////////
	int i;
	for(i=0; i<diags->get_diagrams_number(); i++) {
		CFeynDiagram* diag = diags->get_diagram(i);
		if (CheckDiagram(diag, diag_type)) {
			diags->remove_diagram(i); i--;
		}
	}
	
	FilterSimilarDiagrams(diags,false,true);
}

bool CLDiagTopologyInspector::CheckDiagram(CFeynDiagram* diag, DWORD diag_type) {
	dbg->log(25, diag_type, "CheckDiagram, type == %ld, diagram:", diag);
	if (diag == NULL) return false; //Error
	if (diag_type == 0) return false;

	bool check_performed = false;
	if (diag_type & NON_EULER_CIRCUIT) {
		bool res = CheckNonEulerGraph(diag);
		if (res) return true; // no further check required
		check_performed = true;
	}

	if (diag_type & LOOPS_IN_EXTERNAL_LINES) {
		bool res = CheckLoopsInExtLines(diag);
		if (res) return true; // no further check required
		check_performed = true;
	}

	if (diag_type & ODD_LINEAR_FERMION_LOOP) {
		bool res = CheckOddLinearFermionLoop(diag);
		if (res) return true; // no further check required
		check_performed = true;
	}

	if (!check_performed) {
		dbg->log(0, diag_type, "ERROR: Unknown topology request: %ld.\n");
		return false;
	}
	return false;
}

/*
*	supported flags: NON_EULER_CIRCUIT only
*/
bool CLDiagTopologyInspector::CheckOpSeq(CMEOperatorSeq* seq, DWORD diag_type) {
	dbg->log(25, diag_type, "CheckOpSeq, type == %ld, seq:", seq);
	if (seq == NULL) return false; //Error
	if (diag_type == 0) return false;

	DWORD checked_flags = 0;
	if (diag_type & NON_EULER_CIRCUIT) {
		bool res = CheckNonEulerOpSeq(seq);
		if (res) return true; // no further check required
		checked_flags |= NON_EULER_CIRCUIT;
	}

	if (checked_flags == 0) { dbg->log(0, diag_type, "ERROR: Unsupported topology request for Operator Sequence: %ld.\n"); }
	if (checked_flags != diag_type) { dbg->log(0, (diag_type^checked_flags), "ERROR: Unsupported (partial) topology request for Operator Sequence: %ld.\n"); }
	return false;
}

bool CLDiagTopologyInspector::CheckNonEulerGraph(CFeynDiagram* diag) {
	
	// delete old whatever it can be
	// it must be cleared after the previous use
	diag->clear_temporary_data(true);
	
	CDLine* l = diag->get_line(0);
	MarkDiagramElementAndLinks(l);

	// check up unmarked items
	bool res = false;
	int s = diag->get_lines_number();
	int i;
	for(i=0; i<s; i++) {
		if (diag->get_line(i)->m_temporary_mark == NULL) { res = true; break; }
	}
	if (!res) {
		s = diag->get_vertices_number();
		for(i=0; i<s; i++) {
			if (diag->get_vertex(i)->m_temporary_mark == NULL) { res = true; break; }
		}
	}
	diag->clear_temporary_data(false);

	return res;
}

/*
*	Note: non-interacting particle (i-f-line) is not suspected here,
*	and does not treated as unbinded
*	The check is only connected to point bonds.
*/
bool CLDiagTopologyInspector::CheckNonEulerOpSeq(CMEOperatorSeq* seq) {

	int s = seq->getMaxPointID()+1;
	int* _marks = (int*)CMemoryManager::mmalloc(s*sizeof(int));
	int i;
	for(i=0; i<s; i++) { _marks[i]=0; } // mark all as unrouted
	
	// mark points, strrt from point#0, than mark all connected points
	MarkOpSeqPoints(seq, 0, _marks);

	bool res = false;
	for (i=0; i<s; i++) { if (_marks[i]<=0) { res = true; break; } }
	CMemoryManager::mfree(_marks);
	return res;
}

void CLDiagTopologyInspector::MarkDiagramElementAndLinks(CDElement* el) {
	if (el == NULL) return;

	// if element is not marked, mark it and it's neighbours
	if (el->m_temporary_mark != NULL) return;

	// mark (with itself for instance) -- do not delete marks source further.
	el->m_temporary_mark = el;

	int s = el->get_links_number();
	int i;
	for(i=0; i<s; i++) {
		MarkDiagramElementAndLinks(el->get_link(i));
	}
}

void CLDiagTopologyInspector::MarkOpSeqPoints(CMEOperatorSeq* seq, int point_ID, int* _marks) {
	if (_marks[point_ID] > 0) return;
	_marks[point_ID] = 1;

	long s = seq->size();
	long i;
	for(i=0; i<s; i++) {
		int next_pos = seq->GetNextPFPosAtPoint(point_ID, i, 0);
		if (next_pos < 0) break;
		i = next_pos;
		
		CEParticleFunction* pf = seq->GetPFAtPos(next_pos);
		if (!pf->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)pf;
		if (fo->isCoupled()) {
			CEFieldOperator* fo_cpl = seq->GetCoupledFO(fo);
			int next_pID = fo_cpl->getPointID();
			MarkOpSeqPoints(seq, next_pID, _marks);
		}
	}
}

/*
*	Note: complicated loops should also be extracted here (NOT IMPLEMENTED YET!).
*/
bool CLDiagTopologyInspector::CheckLoopsInExtLines(CFeynDiagram* diag) {	

	int s = diag->get_lines_number();
	int i;
	for(i=0; i<s; i++) {
		CDLine* l = diag->get_line(i);
		if (l->is_external()) {
			if ( CheckLoopsInExtLine((CDExtLine*)l) ) return true;
		}
	}
	return false;
}

/*
*	search for loops at this line
*	Note: complicated loops should also be searched here (NOT IMPLEMENTED YET!).
*/
bool CLDiagTopologyInspector::CheckLoopsInExtLine(CDExtLine* ext_l) {
	CDVertex* v = (CDVertex*)ext_l->get_link(0);
	if (v == NULL) return false;

	int s = v->get_links_number(); int i;
	if (s == 4) {
		CDElement* v1=NULL; CDElement* v2=NULL; CDElement* v3=NULL;
		for(i=0; i<s; i++) {
			CDLine* l = (CDLine*)v->get_link(i); if(l==ext_l)continue;
			if (l->is_external()) return false;
			CDElement* vX=((CDIntLine*)l)->get_other_link(v); if(vX==NULL)return false; /*unbinded*/
			if (v1 == v) continue;
			if (v1==NULL){ v1=vX; }else if(v2==NULL){ v2=vX; }else{ v3=vX; }
		}
		if (v1==NULL) return false; // v2 & v3 ==NULL too (self-looped ext. line with no connection)
		if (v2==NULL) return true; // v3 ==NULL too (self-looped ext. line with 1 connection -> drop it)
		if (v3==NULL) {
			/* the 4-tales vertex is allowed at v1 (as well as 2 tales)*/
			if ((v1==v2) && (v1->get_links_number() == 3)) return true;
		} else {
			/* the 4-tales vertex is allowed at v1 (as well as 2 tales)*/
			if ((v1==v2) && (v1==v3) && (v1->get_links_number() == 4)) return true;
		}
		return false;
	} else if (s == 3) {
		CDElement* v1=NULL; CDElement* v2=NULL;
		for(i=0; i<s; i++) {
			CDLine* l = (CDLine*)v->get_link(i); if(l==ext_l)continue;
			if (l->is_external()) return false;
			CDElement* vX=((CDIntLine*)l)->get_other_link(v); if(vX==NULL)return false; /*unbinded*/
			if(vX==v)return false; // old value: if(vX==v)return true;
			if (v1 == NULL) { v1=vX; } else { v2=vX; }
		}
		if ((v1==v2) && (v1->get_links_number() == 3)) return true; /* the 4-tales vertex is allowed at v1 (as well as 2 tales)*/
		return false;
	} else if (s==2) {
		// 2-tales vertex at external line: Ext-Ext (leave here) or counter-term (drop)
		for(i=0; i<s; i++) {
			CDLine* l = (CDLine*)v->get_link(i); if(l==ext_l)continue;
			if (!l->is_external()) return true;
		}
		return false;
	}

	return false;
}

bool CLDiagTopologyInspector::CheckOddLinearFermionLoop(CFeynDiagram* diag) {
	dbg->log(35, "CheckOddLinearFermionLoop in diagram:", diag);
	bool res = false;

	// delete old whatever it can be
	// it must be cleared after the previous use
	diag->clear_temporary_data(true);

	// first mark all fermion lines started from external lines
	int s = diag->get_lines_number();
	int i;
	for(i=0; i<s; i++) {
		CDLine* l = diag->get_line(i);
		if (!CEParticle::is_fermion_PK(l->get_particle_kind())) continue;
		if (l->m_temporary_mark == NULL) {
			// check up line if it is loop
			if (CheckOddFermionLoop(l, l, 1)) { res = true; break; }
		}
	}

	diag->clear_temporary_data(false);
	dbg->log(35, (res)?1:0, "CheckOddLinearFermionLoop result = %ld");
	return res;
}

bool CLDiagTopologyInspector::CheckOddFermionLoop(CDLine* l, CDLine* l_start, int lines_num) {
	dbg->log(35, lines_num, "CheckOddFermionLoop, lines_num == %ld.");
	l->m_temporary_mark = l; // mark with itself
	if (l->is_external()) return false;
	if (l->get_links_number() != 2) return false;

	// get second vertex if (pk>0) and first if (pk<0)
	int next_vert_num = (l->get_particle_kind()>0)?1:0;
	CDVertex* v = (CDVertex*)l->get_link(next_vert_num);
	if (!v->is_binded()) return false;
	CDiagramVertexInfo* vi = v->get_vertex_info();
	int s = vi->get_tales_num();
	if (s != 3) return false;
	int i;
	for(i=0; i<s; i++) {
		if (!CEParticle::is_fermion_PK(vi->get_particle_kind(i))) continue;
		CDLine* l_next = (CDLine*)v->get_link(i);
		if (l_next != l) {
			if (l_next == l_start) {
				// fermion loop found, number of lines == lines_num
				dbg->log(35, lines_num, "Fermion loop of size %ld found.");
				return (lines_num%2 == 1);// true if odd
			}
			if (l_next->m_temporary_mark != NULL) return false;
			return CheckOddFermionLoop(l_next, l_start, lines_num+1);
		}
	}
	return false;
}

/*
*	Arranges diagrams & uses arrange library data to identify non-sorting similar diagrams.
*/
void CLDiagTopologyInspector::FilterSimilarDiagrams(CDiagramSet* diags, bool arrange_factors, bool use_arrange_libs) {
	if ((use_arrange_libs) || (!arrange_factors)) {
		CEException::halt("Error: Non-stable code invoked (CLDiagTopologyInspector::FilterSimilarDiagrams)"); //TODO: remove this after examine
	}
	dbg->log(15, "FilterSimilarDiagrams()");
	if (diags==NULL) return;
	long i;
	long s=diags->get_diagrams_number();
	dbg->log(25, s, "FilterSimilarDiagrams() size before filtration: %ld");

	// 1. arrange diagrams
	CEPtrArray lib_insts(true); // array of library instances
	if (use_arrange_libs) {
		CDiagramSet* unsorted_list = diags->getEmptyFrame();
		for(i=0; i<s; i++) {
			CFeynDiagram* d = diags->get_diagram(i);
			CLDiagArrange* lib = CLDiagArrange::DiagArrangeGetLib(d);
			if (!lib->IsDiagArrangedOK()) { unsorted_list->add_diagram(d->clone()); }
			lib_insts.add(lib);
		}
		if (unsorted_list->get_diagrams_number() > 0) {
			dbg->log(0, unsorted_list->get_diagrams_number(), "Warning: Total %ld unsorted diagrams");
			dbg->log(0, unsorted_list->get_diagrams_number(), "Unsorted diagrams list (total %ld diagrams)", unsorted_list, CDebug::LOG_TEX);
		}
		delete unsorted_list;
	}
	
	// compare diagram structure to all the followed diagrams and drop similar ones
	CDiagramSet* dropped_list = diags->getEmptyFrame();
	for(i=0; i<diags->get_diagrams_number(); i++) {
		CFeynDiagram* di = diags->get_diagram(i);
		bool bRemoveI=false;
		CLDiagArrange* li = ((use_arrange_libs)?((CLDiagArrange*)lib_insts.GetAt(i)):NULL);
		for(long j=i+1; j<diags->get_diagrams_number(); j++) {
			CFeynDiagram* dj = diags->get_diagram(j);
			CLDiagArrange* lj = ((use_arrange_libs)?((CLDiagArrange*)lib_insts.GetAt(j)):NULL);

			// comparison
			bool IJ_are_equal = di->IsEqualTo(dj, li, lj, arrange_factors);

			// if they are equal
			if (IJ_are_equal) {
				if (arrange_factors) {
					CExpression* fI=di->get_factor_expr(); fI=((fI!=NULL)?fI->clone():CENumber::GetUnit());
					CExpression* fJ=dj->get_factor_expr(); fJ=((fJ!=NULL)?fJ->clone():CENumber::GetUnit());
					fI=CExpression::reduce(fI->add(fJ));
					if (CEBasic::isZero(fI)) {
						bRemoveI=true;
					} else { // modify fI's factor
						di->set_factor_expr(((CEBasic::isUnit(fI))?NULL:fI));
					}
					delete fI;
				}
				dropped_list->add_diagram(dj->clone()); diags->remove_diagram(j);
				if(use_arrange_libs) lib_insts.RemoveAt(j, true); // delete arrange library ("lj")
				j--; // examine next diagram
				if(bRemoveI)break;
			}
		}

		if (bRemoveI) {
			dropped_list->add_diagram(di->clone()); diags->remove_diagram(i);
			if(use_arrange_libs) lib_insts.RemoveAt(i, true); // delete arrange library ("li")
			i--; // examine next diagram
		}
	}

	// dropped list is currently unused
	if (dropped_list->get_diagrams_number() > 0) {
		dbg->log(0, dropped_list->get_diagrams_number(), "FilterSimilarDiagrams: %ld diagrams combined");
		dbg->log(95, "Dropped diagrams list", dropped_list, CDebug::LOG_TEX);
	}
	delete dropped_list;
	dbg->log(25, diags->get_diagrams_number(), "FilterSimilarDiagrams() size after filtration: %ld");
}

void CLDiagTopologyInspector::DropSimilarDiagrams(CDiagramSet* diags1, CDiagramSet* diags2, CDiagramSet* dropped_list) {
	if ((diags1==NULL)||(diags2==NULL)) return;
	dbg->log(15, diags1->get_diagrams_number(), diags2->get_diagrams_number(), "DropSimilarDiagrams(%ld diagrams, %ld diagrams)");
	long i;

	bool use_arrange_libs=false;
	CEPtrArray lib_insts1(true); // array of library instances
	CEPtrArray lib_insts2(true); // array of library instances
	if (use_arrange_libs) {
		CDiagramSet* unsorted_list1 = diags1->getEmptyFrame();
		for(i=0; i<diags1->get_diagrams_number(); i++) {
			CFeynDiagram* d = diags1->get_diagram(i); if(d==NULL)CEException::halt("Unexpected NULL at CLDiagTopologyInspector::DropSimilarDiagrams");
			CLDiagArrange* lib = CLDiagArrange::DiagArrangeGetLib(d);
			if (!lib->IsDiagArrangedOK()) { unsorted_list1->add_diagram(d->clone()); }
			lib_insts1.add(lib);
		}
		if (unsorted_list1->get_diagrams_number() > 0) {
			dbg->log(0, unsorted_list1->get_diagrams_number(), "Warning: Total %ld unsorted diagrams");
			dbg->log(2, unsorted_list1->get_diagrams_number(), "Unsorted diagrams list (total %ld diagrams)", unsorted_list1, CDebug::LOG_TEX);
		}
		delete unsorted_list1;
		CDiagramSet* unsorted_list2 = diags2->getEmptyFrame();
		for(i=0; i<diags2->get_diagrams_number(); i++) {
			CFeynDiagram* d = diags2->get_diagram(i); if(d==NULL)CEException::halt("Unexpected NULL at CLDiagTopologyInspector::DropSimilarDiagrams");
			CLDiagArrange* lib = CLDiagArrange::DiagArrangeGetLib(d);
			if (!lib->IsDiagArrangedOK()) { unsorted_list2->add_diagram(d->clone()); }
			lib_insts2.add(lib);
		}
		if (unsorted_list2->get_diagrams_number() > 0) {
			dbg->log(0, unsorted_list2->get_diagrams_number(), "Warning: Total %ld unsorted diagrams");
			dbg->log(2, unsorted_list2->get_diagrams_number(), "Unsorted diagrams list (total %ld diagrams)", unsorted_list1, CDebug::LOG_TEX);
		}
		delete unsorted_list2;
	}

	long n_dropped=0;
	// compare diagram structure and drop similar ones
	for(i=0; i<diags1->get_diagrams_number(); i++) {
		CFeynDiagram* d1 = diags1->get_diagram(i); if(d1==NULL)continue;
		CLDiagArrange* l1 = ((use_arrange_libs)?((CLDiagArrange*)lib_insts1.GetAt(i)):NULL);
		for(long j=0; j<diags2->get_diagrams_number(); j++) {
			CFeynDiagram* d2 = diags2->get_diagram(j); if(d2==NULL)continue;
			CLDiagArrange* l2 = ((use_arrange_libs)?((CLDiagArrange*)lib_insts2.GetAt(j)):NULL);
			//bool chgDbg=((i==0) && (j==3)); if (chgDbg) { CFeynDiagram::dbg->set_level(500); CLDiagArrange::dbg->set_level(500); }
			if (d1->IsEqualTo(d2,l1,l2,false)) { // if equal is found
				if (dropped_list != NULL) { dropped_list->add_diagram(d1->clone()); }
				diags2->remove_diagram(j); diags1->remove_diagram(i); n_dropped++;
				if(use_arrange_libs) lib_insts1.RemoveAt(i, true); // delete arrange library ("lj")
				if(use_arrange_libs) lib_insts2.RemoveAt(j, true); // delete arrange library ("lj")
				i--; // examine next diagram from diags1
				break;
			}
			//if (chgDbg) { CFeynDiagram::dbg->set_level(5); CLDiagArrange::dbg->set_level(5); }
		}
	}
	dbg->log(15,n_dropped,diags1->get_diagrams_number(),diags2->get_diagrams_number(), "DropSimilarDiagrams: %ld similar, %ld and %ld are different");
}
