////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LDrawDiagView.cpp: implementation of the CLDrawDiagView class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const int CLDrawDiagView::LINE_FERMION		= 0x01;
const int CLDrawDiagView::LINE_BOSON		= 0x02;
const int CLDrawDiagView::LINE_SCALAR		= 0x04;
const int CLDrawDiagView::LINE_GLUON		= 0x08;
const int CLDrawDiagView::LINE_NO_ARROW		= 0x10;

CFeynDiagram* CLDrawDiagView::m_current_diag = NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLDrawDiagView::CLDrawDiagView() {}
CLDrawDiagView::~CLDrawDiagView() {}

//////////////////////////////////////////////////////////////////////
// note: no diagrams may cloned/deleted here
CFeynDiagram* CLDrawDiagView::GetCurrentDiagram() { return m_current_diag; }
// returns old link (old link should be restored after diagram is drawed)
CFeynDiagram* CLDrawDiagView::SetCurrentDiagram(CFeynDiagram* _diag) {
	CFeynDiagram* old = m_current_diag; m_current_diag = _diag; return old;
}

//////////////////////////////////////////////////////////////////////
int CLDrawDiagView::get_line_kind(int particle_kind) {
	bool inv_sign = false;
	int pk_abs;
	if (particle_kind < 0) { pk_abs = -particle_kind; inv_sign = true; }
	else { pk_abs = particle_kind; }

	int res = 0;
	CDiagramPhysics* phys=CDiagramPhysics::get_curr_physics();
	if (phys!=NULL) res=phys->getDiagramLineType(pk_abs);

	if (res==0) {
		// use default selection procedure
		if (CEParticle::is_vec_bosonic_PK(pk_abs)) {
			if(pk_abs==CEParticle::PARTICLE_GLUON){
				res = LINE_GLUON;
			}else{
				res = LINE_BOSON;
			}
		}
		else if (CEParticle::is_fermion_PK(pk_abs)) res = LINE_FERMION;
		else if (CEParticle::is_scalar_PK(pk_abs)) res = LINE_SCALAR;
		else {
			CFeynDiagram::dbg->log(0, pk_abs, "<CLDrawDiagView> Unexpected particle kind spotted: %ld"); throw(new CEException(CEException::INT_ERROR));
		}
		if (CEParticle::is_neutral_bosonic_PK(pk_abs)) res|=LINE_NO_ARROW;
	}

	if (inv_sign) res = -res;
	return res;
}

void CLDrawDiagView::add_line_caption(CEString* str, CEString* _line_caption, int particle_kind,
									  int x1, int y1, int x2, int y2,
									  int settings, int line_width) {
	if (_line_caption == NULL) return;

	int xc = (x1+x2)/2;
	int yc = (y1+y2)/2;

	int text_gap = 7;
	if (settings & CFeynDiagram::DRAW_CAPS_SMALL) text_gap = 4;
	text_gap+=line_width;

	int xdiff = (x1-x2)/2; int xdiff_abs=((xdiff>=0)?xdiff:-xdiff);
	int ydiff = (y1-y2)/2; int ydiff_abs=((ydiff>=0)?ydiff:-ydiff);
	int dr = CLMath::Round(CLMath::length(0, xdiff, 0, ydiff));
	int xd;int yd;
	if (dr>0) {
		xd = (ydiff*text_gap)/dr; yd = -(xdiff*text_gap)/dr;
	} else {
		xd=4; yd=0;
	}

	// we want dy to be < 0, dx > 0
	bool is_below = false;
	if ((yd > 0) && (yd > xd)) { xd = -xd; yd = -yd; is_below=true; } // make index below the line
	if ((xd < 0) && (xd < yd)) { xd = -xd; yd = -yd; } // make index at right of the line

	/*
	if (is_below) {
		xc = (2*xc+x1)/3;
		yc = (2*yc+y1)/3;
	} else {
		xc = (2*xc+x1)/3;
		yc = (2*yc+y1)/3;
	}
	*/
	if (xdiff_abs == ydiff_abs) {
		xc = (2*xc+x1)/3;
		yc = (2*yc+y1)/3;
	} else if (xdiff_abs == 0) {
		xc = (5*xc+x1)/6;
		yc = (5*yc+y1)/6;
	}

	int x = xc + xd;
	int y = yc + yd;

	int align_par = 0;
	if (x > (xc+3)) { align_par = +1; x-=2; }
	if (x < (xc-3)) { align_par = -1; x+=2; }

	str->AddString("\\Text("); str->add_int(x);
	str->AddString(", "); str->add_int(y);
	switch(align_par) {
	case 1:
		str->AddString(")[l]{"); break;
	case -1:
		str->AddString(")[r]{"); break;
	default:
		str->AddString(")[c]{");
	}
	if (settings & CFeynDiagram::DRAW_CAPS_SMALL) str->AddString("\\tiny ");
	// \small, \footnotesize, \scriptsize, \tiny
	str->AddString("$"); str->AddString(_line_caption, false); str->AddString("$}");
}

void CLDrawDiagView::add_counterterm_mark(CEString* str,
										  int x1, int y1, int x2, int y2, int settings) {
	int xc = (x1+x2)/2;
	int yc = (y1+y2)/2;

	int mark_size = 5;
	if (settings & CFeynDiagram::DRAW_CAPS_SMALL) mark_size = 4;
	int xdiff = (x1-x2)/2;
	int ydiff = (y1-y2)/2;
	int dr = CLMath::Round(CLMath::length(0, xdiff, 0, ydiff));
	int nx = (xdiff*mark_size)/dr;
	int ny = (ydiff*mark_size)/dr;
	// {nx, ny} & {-nx, -ny} are parallel vectors
	// {ny, -nx} & {-ny, nx} are normal vectors

	int v1x = nx+ny; int v1y = ny-nx;
	int v2x = nx-ny; int v2y = ny+nx;
	
	str->AddString("\\Line(");
	str->add_int(xc+v1x); str->AddString(", "); str->add_int(yc+v1y);
	str->AddString(")(");
	str->add_int(xc-v1x); str->AddString(", "); str->add_int(yc-v1y);
	str->AddString(")");

	str->AddString("\\Line(");
	str->add_int(xc+v2x); str->AddString(", "); str->add_int(yc+v2y);
	str->AddString(")(");
	str->add_int(xc-v2x); str->AddString(", "); str->add_int(yc-v2y);
	str->AddString(")");
}

void CLDrawDiagView::add_line_caption_arc(CEString* str, CEString* _line_caption, int particle_kind,
										  int xc, int yc, int r, double f1, double f2,
										  int settings, int line_width) {
	if (_line_caption == NULL) return;

	//printf("\n%ld, %ld, %ld, %f, %f", xc, yc, r, f1, f2);
	if (f2 < f1) f2+=360;
	double ang_cent = (f2+f1)/2.;

	int text_gap = 6;
	if (settings & CFeynDiagram::DRAW_CAPS_SMALL) text_gap = 4;
	text_gap+=line_width;

	double xd = CLMath::Cos(ang_cent, true)*(r+text_gap);
	double yd = CLMath::Sin(ang_cent, true)*(r+text_gap);
	//printf(" --> %f, %f\n", xd, yd);

	int x = xc + CLMath::Round(xd);
	int y = yc + CLMath::Round(yd);

	int align_par = 0;
	if (x > (xc+3)) { align_par = +1; x-=2; }
	if (x < (xc-3)) { align_par = -1; x+=2; }

	str->AddString("\\Text("); str->add_int(x);
	str->AddString(", "); str->add_int(y);
	switch(align_par) {
	case 1:
		str->AddString(")[l]{"); break;
	case -1:
		str->AddString(")[r]{"); break;
	default:
		str->AddString(")[c]{");
	}
	if (settings & CFeynDiagram::DRAW_CAPS_SMALL) str->AddString("\\tiny ");
	// \small, \footnotesize, \scriptsize, \tiny
	str->AddString("$"); str->AddString(_line_caption, false); str->AddString("$}");
}

void CLDrawDiagView::add_counterterm_mark_arc(CEString* str,
											  int xc, int yc, int r,
											  double f1, double f2, int settings) {
	if (f2 < f1) f2+=360;
	double ang_cent = (f2+f1)/2.;

	double xd = CLMath::Cos(ang_cent, true)*(r);
	double yd = CLMath::Sin(ang_cent, true)*(r);
	xc = xc + CLMath::Round(xd);
	yc = yc + CLMath::Round(yd);

	int mark_size = 5;
	if (settings & CFeynDiagram::DRAW_CAPS_SMALL) mark_size = 4;

	xd = xd*mark_size/r;
	yd = yd*mark_size/r;
	int nx = CLMath::Round(xd);
	int ny = CLMath::Round(yd);
	// {nx, ny} & {-nx, -ny} are parallel vectors
	// {ny, -nx} & {-ny, nx} are normal vectors

	int v1x = nx+ny; int v1y = ny-nx;
	int v2x = nx-ny; int v2y = ny+nx;
	
	str->AddString("\\Line(");
	str->add_int(xc+v1x); str->AddString(", "); str->add_int(yc+v1y);
	str->AddString(")(");
	str->add_int(xc-v1x); str->AddString(", "); str->add_int(yc-v1y);
	str->AddString(")");

	str->AddString("\\Line(");
	str->add_int(xc+v2x); str->AddString(", "); str->add_int(yc+v2y);
	str->AddString(")(");
	str->add_int(xc-v2x); str->AddString(", "); str->add_int(yc-v2y);
	str->AddString(")");
}

/*
*	Does not affect _line_caption
*/
CEString* CLDrawDiagView::draw_diagram_line(int particle_kind,
											CPos2D *pos_from, CPos2D *pos_to,
											int settings, CEString* _line_caption,
											bool _draw_CT_line)
{
	return draw_diagram_line(particle_kind, pos_from, pos_to, settings, _line_caption, 1, 1, _draw_CT_line);
}

/*
*	Does not affect _line_caption
*/
CEString* CLDrawDiagView::draw_diagram_line(int particle_kind,
											CPos2D *pos_from, CPos2D *pos_to,
											int settings, CEString* _line_caption,
											int repetition_no, int multiplicity,
											bool _draw_CT_line)
{
	int line_kind = get_line_kind(particle_kind);
	if ((pos_from == NULL) || (pos_to == NULL)) return NULL;

	// settings check
	// Note SET_LOOP_120 setting is default
	if ((settings & CFeynDiagram::SET_LOOP_120) && (settings & CFeynDiagram::SET_LOOP_90)) {
		settings ^= CFeynDiagram::SET_LOOP_120;	// set off SET_LOOP_120 setting.
	}
	else if (((settings & CFeynDiagram::SET_LOOP_90) == 0) && ((settings & CFeynDiagram::SET_LOOP_120) == 0)) {
		settings |= CFeynDiagram::SET_LOOP_120; // set default
	}

	bool draw_backward = false;
	if (line_kind < 0) {draw_backward = true; line_kind = -line_kind;}
	
	if ((line_kind & (LINE_BOSON|LINE_GLUON|LINE_FERMION|LINE_SCALAR))==0) { CFeynDiagram::dbg->log(0,"ERROR line drawing: unknown line type"); return NULL; }

	CPos2D* pos_1 = pos_from;
	CPos2D* pos_2 = pos_to;
	if (draw_backward) {
		// swap positions
		pos_1 = pos_to;
		pos_2 = pos_from;

		// and swap repetition number
		repetition_no = -repetition_no;
		//repetition_no = multiplicity-repetition_no+1;
	}

	bool draw_arc = false;
	if (multiplicity > 1) {
		draw_arc = true;
		if (((multiplicity+1) == (2*repetition_no)) ||
			((multiplicity+1) == (-2*repetition_no))) draw_arc = false;
	}

	int x1 = pos_1->get_X();
	int y1 = pos_1->get_Y();
	int x2 = pos_2->get_X();
	int y2 = pos_2->get_Y();
	int lLen = CLMath::LengthApprox(x1,y1,x2,y2);

	// check for similar coordinates
	if ((x1 == x2) &&  (y1 == y2)) {
		// one can draw round mark or move one of the positions
		x1 += 2;
	}

	CEString* res = new CEString("");
	///////////////////////////////////////////////	set line settings (width)
	//add_line_settings(res, particle_kind);

	if (!draw_arc) {
		int iLnWidth=1;
		if ((line_kind & LINE_BOSON)==LINE_BOSON) {
			int nWiggs=lLen/10; if(nWiggs<2)nWiggs=2; if(nWiggs>7)nWiggs=7;
			res->AddString("\\Photon(");
			res->add_int(x1); res->AddString(", "); res->add_int(y1);
			res->AddString(")(");
			res->add_int(x2); res->AddString(", "); res->add_int(y2);
			res->AddString("){1}{"); res->add_int(nWiggs); res->AddString("}");
		} else if ((line_kind & LINE_GLUON)==LINE_GLUON) {
			int nWiggs=lLen/10; if(nWiggs<2)nWiggs=2; if(nWiggs>7)nWiggs=7;
			res->AddString("\\Gluon(");
			res->add_int(x1); res->AddString(", "); res->add_int(y1);
			res->AddString(")(");
			res->add_int(x2); res->AddString(", "); res->add_int(y2);
			res->AddString("){3}{"); res->add_int(nWiggs); res->AddString("}");
			iLnWidth=3;
		} else if ((line_kind & LINE_FERMION)==LINE_FERMION) {
			res->AddString("\\ArrowLine(");
			res->add_int(x1); res->AddString(", "); res->add_int(y1);
			res->AddString(")(");
			res->add_int(x2); res->AddString(", "); res->add_int(y2);
			res->AddString(")");
		} else if ((line_kind & LINE_SCALAR)==LINE_SCALAR) {
			if ((line_kind & LINE_NO_ARROW)==LINE_NO_ARROW) { res->AddString("\\DashLine("); }
			else { res->AddString("\\DashArrowLine("); }
			res->add_int(x1); res->AddString(", "); res->add_int(y1);
			res->AddString(")(");
			res->add_int(x2); res->AddString(", "); res->add_int(y2);
			res->AddString(")");
			res->AddString("{1}"); // dashsize
		}

		// caption
		add_line_caption(res, _line_caption, particle_kind, x1, y1, x2, y2, settings, iLnWidth);

		if (_draw_CT_line) add_counterterm_mark(res, x1, y1, x2, y2, settings);
	}
	else {	// draw arc
		//printf("repetition_no = %ld, ", repetition_no);
		// DIAGRAM DESIGN PARAMETER
		if (repetition_no < 0) { repetition_no = multiplicity + repetition_no + 1;}

		// define if arc is clockwise or counter-clockwise
		bool counter_clockwise = true;
		if ((repetition_no*2) > (multiplicity+1)) {
			counter_clockwise = false;
		}

		// DIAGRAM DESIGN PARAMETERS
		double arc_start_angle = -1;
		if (settings & CFeynDiagram::SET_LOOP_90) {
			// 1. 90^o between two lines
			arc_start_angle = ((double)repetition_no*2.-1)*90/multiplicity;
		} else if (settings & CFeynDiagram::SET_LOOP_120) {
			// 2. 120^o between two lines
			// repetition_no supposed to be > 0 here
			if (multiplicity == 2) {
				arc_start_angle = (repetition_no == 1)?30.:150.;
			} else if (multiplicity == 3) {
				arc_start_angle = (repetition_no == 1)?15.:165.;
			} else {
				arc_start_angle = ((double)repetition_no*2.-1)*90/multiplicity;
			}
		} else {
			printf("Error loop drawing.\n"); throw(new CEException(CEException::INT_ERROR));
		}

		//printf("arc_start_angle = %f\n", arc_start_angle);
		double cos_a = CLMath::Cos(arc_start_angle, true);
		if (cos_a == 0) {printf("Drawing line internal ERROR\n"); throw(new CEException(CEException::INT_ERROR));}
		double sin_a = CLMath::Sin(arc_start_angle, true);
		double tg_a = sin_a/cos_a;
		//printf("tg_a = %f\n", tg_a);

		double BA_angle = CLMath::angle(x1-x2, y1-y2, true);
		//printf("A (%ld, %ld), B(%ld, %ld)\n", x1, y1, x2, y2);
		//printf("Angle (BA, Ox), %f\n", BA_angle);

		// calculate center pos
		int xc = CLMath::Round((double)((double)(x1+x2)+tg_a*(y1-y2))/2.);
		int yc = CLMath::Round((double)((double)(y1+y2)+tg_a*(x2-x1))/2.);
		//printf("C(%ld, %ld)\n", xc, yc);

		// calculate radius
		int r = CLMath::Round(CLMath::length(x1, x2, y1, y2) /(CLMath::Abs(cos_a) * 2.));
		//printf("r = %ld\n", r);

		// calculate arc angles
		int f1;
		int f2;
		if (counter_clockwise) {
			f1 = CLMath::Round(BA_angle + arc_start_angle);
			f2 = CLMath::Round(BA_angle - arc_start_angle + 180.);
		} else {
			//f1 = CLMath::Round(BA_angle - arc_start_angle);
			//f2 = CLMath::Round(BA_angle + arc_start_angle - 180.);
			f1 = CLMath::Round(BA_angle + arc_start_angle - 180.);
			f2 = CLMath::Round(BA_angle - arc_start_angle);
		}
		//printf("angular limits: %ld, %ld.\n", f1, f2);

		//if ((f1 == (-174)) || (f2 == -84)) throw(new CEException(CEException::INT_ERROR));

		if ((line_kind & LINE_BOSON)==LINE_BOSON) {
			res->AddString("\\PhotonArc");
			if (!counter_clockwise) { // swap angles
				int tmp = f1; f1 = f2; f2 = tmp; counter_clockwise = true;
			}
		} else if ((line_kind & LINE_GLUON)==LINE_GLUON) {
			res->AddString("\\GlueArc");
			if (!counter_clockwise) { // swap angles
				int tmp = f1; f1 = f2; f2 = tmp; counter_clockwise = true;
			}
		} else if ((line_kind & LINE_FERMION)==LINE_FERMION) {
			if (counter_clockwise) res->AddString("\\ArrowArc");
			else res->AddString("\\ArrowArcn");
		} else if ((line_kind & LINE_SCALAR)==LINE_SCALAR) {
			if ((line_kind & LINE_NO_ARROW)==LINE_NO_ARROW) {
				res->AddString("\\DashCArc");
				if (!counter_clockwise) { // swap angles
					int tmp = f1; f1 = f2; f2 = tmp; counter_clockwise = true;
				}
			} else {
				if (counter_clockwise) res->AddString("\\DashArrowArc");
				else res->AddString("\\DashArrowArcn");
			}
		}

		// center
		res->AddString("(");
		res->add_int(xc); res->AddString(", "); res->add_int(yc);
		res->AddString(")(");

		// radius and angles
		res->add_int(r); res->AddString(", ");
		res->add_int(f1); res->AddString(", "); res->add_int(f2);
		res->AddString(")");

		int iLnWidth=1;
		if ((line_kind & LINE_BOSON)==LINE_BOSON) {
			int nWiggs=lLen/10; if(nWiggs<2)nWiggs=2; if(nWiggs>7)nWiggs=7;
			res->AddString("{1}{"); res->add_int(nWiggs); res->AddString(".5}");
		} else if ((line_kind & LINE_GLUON)==LINE_GLUON) {
			int nWiggs=lLen/10; if(nWiggs<2)nWiggs=2; if(nWiggs>7)nWiggs=7;
			res->AddString("{3}{"); res->add_int(nWiggs); res->AddString(".5}");
			iLnWidth=3;
		} else if ((line_kind & LINE_FERMION)==LINE_FERMION) {
		} else if ((line_kind & LINE_SCALAR)==LINE_SCALAR) {
			res->AddString("{1}"); // dashsize
		}


		// caption
		if (counter_clockwise) {
			add_line_caption_arc(res, _line_caption, particle_kind, xc, yc, r, f1, f2, settings, iLnWidth);
			if (_draw_CT_line) add_counterterm_mark_arc(res, xc, yc, r, f1, f2, settings);
		}
		else {
			add_line_caption_arc(res, _line_caption, particle_kind, xc, yc, r, f2, f1, settings, iLnWidth);
			if (_draw_CT_line) add_counterterm_mark_arc(res, xc, yc, r, f2, f1, settings);
		}
	}

	///////////////////////////////////////////////	restore line settings (width)
	//add_settings_restore(res, particle_kind);
	return res;
}
