////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EFieldOperator.cpp: implementation of the CEFieldOperator class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CEFieldOperator::dbg = new CDebug("CEFieldOperator");
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEFieldOperator::CEFieldOperator(long _particle_kind, bool _is_conjugated,
								 int _vertex_ID, unsigned long _point_ID)
: CEParticleFunction(_particle_kind, _vertex_ID, _point_ID)
{
	// truncate PK sign here
	// PK < 0 means conjugation for field,
	// it should be checked in specific field construction
	if (_particle_kind < 0) {
		_is_conjugated = !_is_conjugated; // PK < 0 : inverse conjugation flag
		m_particle_kind = -_particle_kind; // update PK value
	}
	m_is_conjugated = _is_conjugated;

	m_coupling_no = 0;
	m_is_CT_coupling = false;
}

// constructor for cloning, no parameters check are required
CEFieldOperator::CEFieldOperator(long _particle_kind, bool _is_conjugated,
								 int _vertex_ID, unsigned long _point_ID,
								 int _coupling_no, bool _is_CT_coupling)
: CEParticleFunction(_particle_kind, _vertex_ID, _point_ID)
{
	m_is_conjugated = _is_conjugated;
	m_coupling_no = _coupling_no;
	if (m_coupling_no > 0) m_is_CT_coupling = _is_CT_coupling;
	else m_is_CT_coupling = false;
}

CEFieldOperator::~CEFieldOperator() {}
const char* CEFieldOperator::CLASS_NAME = "CEFieldOperator";
const char* CEFieldOperator::ClassName() const { return CLASS_NAME; }

void CEFieldOperator::collectProperties(CEProperties* props, unsigned set) {
	CEParticleFunction::collectProperties(props, set);
	props->setProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR, true);
}

unsigned long CEFieldOperator::props_long() {
	unsigned long res = CEParticleFunction::props_long();
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_IS_FIELD_OPERATOR, true);
	return res;
}

bool CEFieldOperator::equals(CExpression *other) {
	if (!CEParticleFunction::equals(other)) return false;

	// specific comparison
	CEFieldOperator* f_other = (CEFieldOperator*)other;
	bool res = ((getCouplingNumber() == f_other->getCouplingNumber()) &&
		(isConjugated() == f_other->isConjugated()) &&
		(CEUtil::LEQU(isCounterTermCoupling(), f_other->isCounterTermCoupling())));
	return res;
}

CEString* CEFieldOperator::toString(int mode) {
	CEString* res = new CEString("UnknownFieldOperator");
	if (!getFlag()) res->insertString("- ");
	return res;
}

long CEFieldOperator::getLogicalPK() { return (m_is_conjugated)?(-m_particle_kind):(m_particle_kind); }
bool CEFieldOperator::isConjugated() { return m_is_conjugated; }
int CEFieldOperator::getCouplingNumber() { return m_coupling_no; }
bool CEFieldOperator::isCoupled() { return (m_coupling_no > 0); }
bool CEFieldOperator::isCounterTermCoupling() { return m_is_CT_coupling; }

CExpression* CEFieldOperator::conjugated(int mode) {
	CEException::halt("Illegal Field Operator conjugation involved.");
	return NULL;
}

bool CEFieldOperator::MayBeCoupledWith(CEFieldOperator* fo) {
	if (fo == NULL) { dbg->log(0, "<Warning> MayBeCoupledWith(NULL) call"); return false; }
	if (isCoupled() || fo->isCoupled()) return false;

	// compare PK
	if (getParticleKind() != fo->getParticleKind()) return false;

	// avoid self-loops
	if (getPointID() == fo->getPointID()) return false;
	return true;
}

// sets false if no for uncoupled field nomatter the parameter value
void CEFieldOperator::setCounterTermCoupling(bool _CT_coupling) {
	if (m_coupling_no > 0) m_is_CT_coupling = _CT_coupling;
	else m_is_CT_coupling = false;
}

/*
*	Note: setCouplingNo(0); will clears the coupling information
*/
void CEFieldOperator::setCouplingNo(int _coupling_no) {
	//if (isCoupled()) { dbg->log(0, "<error> coupling of already coupled operator", this); CEException::halt(); }
	m_coupling_no = _coupling_no;
	if (m_coupling_no == 0) m_is_CT_coupling = false;
}
