////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// OpBosonField.cpp: implementation of the COpBosonField class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

COpBosonField::COpBosonField(long _particle_kind,
							 CExpression* _vector_index,
							 bool _vector_index_cov,
							 int _vertex_ID,
							 unsigned long _point_ID)
:CEFieldOperator(_particle_kind, false, _vertex_ID, _point_ID)
{
	if (!CEParticle::is_bosonic_PK(m_particle_kind)) { CEException::halt("Bosonic operator construction for non-bosonic field"); }

	if (m_is_conjugated) { // neutrals should be !conjugated
		if ((m_particle_kind & CEParticle::PARTICLE_NEUTRAL_BOSON) == CEParticle::PARTICLE_NEUTRAL_BOSON) {
			dbg->log(0, m_particle_kind, "Warning: conjugation flag is set for neutral particle (pk = %ld). Flag is cleared.");
			m_is_conjugated = false;
		}
	}

	// Note: NULL vector index means scalar particle wave function
	if (_vector_index != NULL) {
		m_vector_index = CTIndex::CreateVectorIndex(_vector_index, _vector_index_cov);
	} else { // Error
		m_vector_index = NULL;
		if (CEParticle::is_vec_bosonic_PK(m_particle_kind)) CEException::halt("NULL vector index for bosonic field operator");
	}
}

// protected constructor for cloning usage
COpBosonField::COpBosonField(long _particle_kind,
							 bool _is_conjugated,
							 CTIndex* _vector_index,
							 int _vertex_ID,
							 unsigned long _point_ID,
							 int _coupling_no,
							 bool _is_CT_coupling)
:CEFieldOperator(_particle_kind, _is_conjugated, _vertex_ID, _point_ID, _coupling_no, _is_CT_coupling) {
	if (!CEParticle::is_bosonic_PK(m_particle_kind)) { CEException::halt("Bosonic operator construction for non-bosonic field"); }
	if (m_is_conjugated) { // neutrals should be !conjugated
		if ((m_particle_kind & CEParticle::PARTICLE_NEUTRAL_BOSON) == CEParticle::PARTICLE_NEUTRAL_BOSON) {
			dbg->log(0, m_particle_kind, "Warning: conjugation flag is set for neutral particle (pk = %ld). Flag is cleared.");
			m_is_conjugated = false;
		}
	}

	// Note: NULL vector index means scalar particle wave function
	if (_vector_index != NULL) {
		m_vector_index = _vector_index->clone();
	} else { // Error
		m_vector_index = NULL;
		if (CEParticle::is_vec_bosonic_PK(m_particle_kind)) CEException::halt("NULL vector index for bosonic field operator");
	}
}

COpBosonField::~COpBosonField() {
	if (m_vector_index != NULL) delete m_vector_index;
}

CExpression* COpBosonField::clone() const {
	CExpression* res = new COpBosonField(m_particle_kind,
		m_is_conjugated,
		m_vector_index,
		m_vertex_ID,
		m_point_ID,
		m_coupling_no,
		m_is_CT_coupling);
	copyAttributesTo(res);
	return res;
}

const char* COpBosonField::CLASS_NAME = "COpBosonField";
const char* COpBosonField::ClassName() const { return CLASS_NAME; }
bool COpBosonField::contain_vector(void * obj) {
	if (m_vector_index == obj) return true;
	if (m_vector_index != NULL) { if (m_vector_index->contain_vector(obj)) return true; }
	return false;
}

void COpBosonField::collectProperties(CEProperties* props, unsigned set) {
	CEFieldOperator::collectProperties(props, set);
}

unsigned long COpBosonField::props_long() {
	unsigned long res = CEFieldOperator::props_long();
	return res;
}

CExpression* COpBosonField::applyRule(CEAbstractRule* rule) {
	CERule* r = (CERule*)rule;

	if (m_vector_index != NULL) {
		CExpression* item = m_vector_index->index();
		CExpression* item_repl = r->ApplyTo(item, false);
		if ((item_repl != item) && (item_repl != NULL)) {
			m_vector_index->setIndex(item_repl); delete item_repl;
		}
	}
	return this;
}

bool COpBosonField::equals(CExpression *other) {
	if (!CEFieldOperator::equals(other)) return false;

	// specific comparison (PK, point ID, CN, iC & iCT are already compared)
	COpBosonField* f_other = (COpBosonField*)other;

	// compare index
	CTIndex* ind1 = getVectorIndex();
	CTIndex* ind2 = f_other->getVectorIndex();

	if ((ind1 == NULL) && (ind2 == NULL)) return true;
	if ((ind1 == NULL) || (ind2 == NULL)) return false;
	bool res = ind1->equalsIndex(ind2);
	return res;
}

#ifdef PF_LESS_THAN
bool COpBosonField::less_than(CExpression* other) {
	if (other == NULL) return false;
	if (other->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		const char* o_cn = other->ClassName();
		// compare to COpFermionField, COpBosonField, CWaveFuncFermion, CWaveFuncBoson
		// A < Psi < S, Wf < Field
		if (o_cn == COpFermionField::CLASS_NAME) {
			bool is_Scalar = (getVectorIndex() == NULL);
			return (!is_Scalar);
		}
		if (o_cn == COpBosonField::CLASS_NAME) {
			COpBosonField* o = (COpBosonField*)other;
			CTIndex* i_b = o->getVectorIndex();
			if (!CEUtil::LEQU((m_vector_index==NULL), (i_b==NULL))) return (m_vector_index != NULL); // A<S
			const char* sym_a = CDiagramPhysics::get_particle_symbolA(m_particle_kind, CDiagramPhysics::SYMBOL_STYLE_NB);
			const char* sym_b = CDiagramPhysics::get_particle_symbolA(o->getParticleKind(), CDiagramPhysics::SYMBOL_STYLE_NB);
			if (CEUtil::a_less_b(sym_a,sym_b)) return true;
			if (!CEUtil::equals(sym_a,sym_b)) return false;
			if (!CEUtil::LEQU(m_is_conjugated, o->isConjugated())) return (!m_is_conjugated); // W+ < W-
			//if (!CEUtil::LEQU(m_is_conjugated, o->isConjugated())) return (m_is_conjugated); // W- < W+
			if (m_point_ID != o->getPointID()) return ((long)m_point_ID < o->getPointID());
			if (m_coupling_no != o->getCouplingNumber()) return (m_coupling_no<o->getCouplingNumber());

			if ((m_vector_index != NULL) && (i_b != NULL)) {
				if (!CEUtil::LEQU(m_vector_index->covariance(), i_b->covariance())) return (i_b->covariance()); // \\& < \\_
				CExpression* ia = m_vector_index->index(); CExpression* ib = i_b->index();
				if (ia->less_than(ib)) return true;
			}
			return false;
		}
		if (o_cn == CWaveFuncFermion::CLASS_NAME) {
			bool is_Scalar = (getVectorIndex() == NULL);
			return (!is_Scalar);
		}
		if (o_cn == CWaveFuncBoson::CLASS_NAME) {
			bool is_Scalar = (getVectorIndex() == NULL);
			bool is_Scalar_b = (((CWaveFuncBoson*)other)->getVectorIndex() == NULL);
			if (!CEUtil::LEQU(is_Scalar, is_Scalar_b)) return (is_Scalar_b); // A<S
			return false;
		}
	}
	return CExpression::less_than(other);
}
#endif //PF_LESS_THAN

CEString* COpBosonField::toString(int mode) {
	CEString* res = NULL;
	// base symbol
	if (m_vector_index != NULL) {
		res = new CEString("A");
	} else {
		res = new CEString("S");
	}

	res->AddString("\\_\\(");
	res->AddString(CDiagramPhysics::get_particle_symbolA(m_particle_kind, CDiagramPhysics::SYMBOL_STYLE_NB));
	if ((m_particle_kind & CEParticle::PARTICLE_NEUTRAL_BOSON) != CEParticle::PARTICLE_NEUTRAL_BOSON) {
		// charged boson
		if (m_is_conjugated) { res->AddString("-"); }
		else { res->AddString("+"); }
	}
	res->AddString("\\)");

	res->insertString("\\(");
	res->AddString("\\)\\+");
	res->add_int(m_point_ID);

	if (m_coupling_no > 0) {
		res->insertString("\\(");
		res->AddString("\\)\\&");
		if (m_is_CT_coupling) {
			res->AddString("\\(");
			res->add_int(m_coupling_no);
			res->AddString("*\\)");
		} else {
			res->add_int(m_coupling_no);
		}
	}

	// vector index
	if (m_vector_index != NULL) {
		bool cov = m_vector_index->covariance();
		CExpression* ind = m_vector_index->index();
		if (ind == NULL) CEException::halt("NULL expression for vector index");

		res->insertString("\\(", 0);
		res->AddString("\\)");
		res->AddString((cov)?"\\_":"\\&");
		res->AddString("\\(");
		res->AddString(ind->toString(mode));
		res->AddString("\\)");
	}

	// sign
	if (getFlag()) res->insertString("\\(");
	else res->insertString("- \\(");
	res->AddString("\\)");
	return res;
}

CTIndex* COpBosonField::getVectorIndex() { return m_vector_index; }
bool COpBosonField::isVectorBoson() { return (m_vector_index != NULL); }

bool COpBosonField::MayBeCoupledWith(CEFieldOperator* fo) {
	bool res_common = CEFieldOperator::MayBeCoupledWith(fo);
	if (!res_common) return false;

	if ((m_particle_kind & CEParticle::PARTICLE_NEUTRAL_BOSON) == CEParticle::PARTICLE_NEUTRAL_BOSON) {
		return true; // nothing to compare here
	}

	// PK is already compared
	// compare conjugation flags
	if (CEUtil::LEQU(isConjugated(), fo->isConjugated())) return false;
	return true;
}

bool COpBosonField::CanBeCoupledExternal(CEParticle* part, bool is_initial, bool bCheckClassesOnly) {
	if (part == NULL) return false;
	if (isCoupled()) return false;

	long pk = getParticleKind(); // always positive
	long part_PK = part->get_PK(); long part_PK_abs=CLMath::abs(part_PK);
	if (bCheckClassesOnly) {
		if ((pk & CEParticle::PARTICLE_CLASS_FLAG)==CEParticle::PARTICLE_CLASS_FLAG) {
			if (!CEParticle::FitClassPk(part_PK_abs,pk)) return false;
		} else {
			if (part_PK_abs != pk) return false;
		}
	} else {
		if (part_PK_abs != pk) return false;
	}
	if ((pk & CEParticle::PARTICLE_NEUTRAL_BOSON) == CEParticle::PARTICLE_NEUTRAL_BOSON) {
		return true; // nothing more to compare here
	}

	// Not zero: Psi*cP, d*Psi, PsiBar*c, dP*PsiBar
	// is_initial --> cP or dP
	// part_PK >= 0 --> c or cP
	bool couple_to_conj = CEUtil::LXOR((part_PK >= 0), is_initial);
	bool res = CEUtil::LEQU(couple_to_conj, isConjugated());
	return res;
}

CExpression* COpBosonField::GetExtParticleCoupling(CEParticle* part, bool is_initial, bool bClassesSource) {
	if (!CanBeCoupledExternal(part, is_initial, bClassesSource)) return NULL;

	long pk = (isConjugated())?(-m_particle_kind):m_particle_kind;
	// Note: we should also inverse momentum sign if (part->get_PK() < 0)
	// but boson propagator/density matrix is even function on momentum

	//MARK: last change here
	if (bClassesSource) { // drop class-flag and append pk-identifier
		long pk_abs=((pk>=0)?pk:-pk);
		if ((pk_abs & CEParticle::PARTICLE_CLASS_FLAG) == CEParticle::PARTICLE_CLASS_FLAG) {
			long pkX=CLMath::abs(part->get_PK()); if(pk<0)pkX=-pkX; pk=pkX;
		}
	}

	// Note getVectorIndex() may be NULL for scalar particles
	CExpression* res = new CWaveFuncBoson(pk,  // pk>0: A(mom), pk<0: A*(mom)
		m_vertex_ID, m_point_ID,
		part->get_momentum_symbol(),
		getVectorIndex());
	copyAttributesTo(res, true);
	return res;
}
