////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// OpFermionField.cpp: implementation of the COpFermionField class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

COpFermionField::COpFermionField(long _particle_kind,
								 int _vertex_ID,
								 unsigned long _point_ID,
								 bool _is_conjugated)
: CEFieldOperator(_particle_kind, _is_conjugated, _vertex_ID, _point_ID)
{
	if (!CEParticle::is_fermion_PK(m_particle_kind)) { CEException::halt("Fermion operator for non-fermion particle"); }
}

// used for cloning
COpFermionField::COpFermionField(long _particle_kind,
								 int _vertex_ID,
								 unsigned long _point_ID,
								 bool _is_conjugated,
								 int _coupling_no,
								 bool _is_CT_coupling)
: CEFieldOperator(_particle_kind, _is_conjugated, _vertex_ID, _point_ID, _coupling_no, _is_CT_coupling) {
	if (!CEParticle::is_fermion_PK(m_particle_kind)) { CEException::halt("Fermion operator for non-fermion particle"); }
}

COpFermionField::~COpFermionField() {}

CExpression* COpFermionField::clone() const {
	CExpression* res = new COpFermionField(m_particle_kind,
		m_vertex_ID,
		m_point_ID,
		m_is_conjugated,
		m_coupling_no,
		m_is_CT_coupling);
	copyAttributesTo(res);
	return res;
}

const char* COpFermionField::CLASS_NAME = "COpFermionField";
const char* COpFermionField::ClassName() const { return CLASS_NAME; }

void COpFermionField::collectProperties(CEProperties* props, unsigned set) {
	CEFieldOperator::collectProperties(props, set);
	props->setCommutative(false);
}

unsigned long COpFermionField::props_long() {
	unsigned long res = CEFieldOperator::props_long();
	res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_COMMUTATIVE, false);
	return res;
}

CEString* COpFermionField::toString(int mode) {
	CEString* res = new CEString("\\[Psi]");

	if (m_is_conjugated) { res->insertString("\\("); res->AddString("\\&_\\)"); }

	res->AddString("\\_\\(");
	res->AddString(CDiagramPhysics::get_particle_symbolA(m_particle_kind, CDiagramPhysics::SYMBOL_STYLE_NB));
	res->AddString("\\)");

	res->insertString("\\(");
	res->AddString("\\)\\+");
	res->add_int(m_point_ID);

	if (m_coupling_no > 0) {
		res->insertString("\\(");
		res->AddString("\\)\\&");
		if (m_is_CT_coupling) {
			res->AddString("\\(");
			res->add_int(m_coupling_no);
			res->AddString("*\\)");
		} else {
			res->add_int(m_coupling_no);
		}
	}

	// sign
	if (getFlag()) res->insertString("\\(");
	else res->insertString("- \\(");
	res->AddString("\\)");
	return res;
}

bool COpFermionField::equals(CExpression *other) {
	if (!CEFieldOperator::equals(other)) return false;

	// specific comparison (PK, point ID, CN, iC & iCT are already compared)
	return true;
}

#ifdef PF_LESS_THAN
bool COpFermionField::less_than(CExpression* other) {
	if (other == NULL) return false;
	if (other->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		const char* o_cn = other->ClassName();
		// compare to COpFermionField, COpBosonField, CWaveFuncFermion, CWaveFuncBoson
		// A < Psi < S, Wf < Field
		if (o_cn == COpFermionField::CLASS_NAME) {
			CEFieldOperator* o = (CEFieldOperator*)other;
			if (!CEUtil::LEQU(m_is_conjugated, o->isConjugated())) return m_is_conjugated; // u_bar < u
			const char* sym_a = CDiagramPhysics::get_particle_symbolA(m_particle_kind, CDiagramPhysics::SYMBOL_STYLE_NB);
			const char* sym_b = CDiagramPhysics::get_particle_symbolA(o->getParticleKind(), CDiagramPhysics::SYMBOL_STYLE_NB);
			if (CEUtil::a_less_b(sym_a,sym_b)) return true;
			if (!CEUtil::equals(sym_a,sym_b)) return false;
			if (m_point_ID != o->getPointID()) return ((long)m_point_ID < o->getPointID());
			if (m_coupling_no != o->getCouplingNumber()) return (m_coupling_no<o->getCouplingNumber());
			return false;
		}
		if (o_cn == COpBosonField::CLASS_NAME) {
			bool is_Scalar_b = (((COpBosonField*)other)->getVectorIndex() == NULL);
			return (is_Scalar_b);
		}
		if (o_cn == CWaveFuncFermion::CLASS_NAME) {
			return false;
		}
		if (o_cn == CWaveFuncBoson::CLASS_NAME) {
			bool is_Scalar_b = (((CWaveFuncBoson*)other)->getVectorIndex() == NULL);
			return (is_Scalar_b);
		}
	}
	return CExpression::less_than(other);
}
#endif //PF_LESS_THAN

bool COpFermionField::MayBeCoupledWith(CEFieldOperator* fo) {
	bool res_common = CEFieldOperator::MayBeCoupledWith(fo);
	if (!res_common) return false;

	// PK is already compared
	// compare conjugation flags
	if (!CEUtil::equals(ClassName(), fo->ClassName())) { dbg->log(0, "<Error> fermion field operator with unexpeted container", fo); CEException::halt(); }
	COpFermionField* ffo = (COpFermionField*)fo;
	if (CEUtil::LEQU(isConjugated(), ffo->isConjugated())) return false;
	return true;
}

bool COpFermionField::CanBeCoupledExternal(CEParticle* part, bool is_initial, bool bCheckClassesOnly) {
	if (part == NULL) return false;
	if (isCoupled()) return false; // is coupled to another internal?

	long part_PK = part->get_PK();
	if (bCheckClassesOnly) {
		long part_PK_abs=CLMath::abs(part_PK);
		long pk=getParticleKind(); // always positive
		if ((pk & CEParticle::PARTICLE_CLASS_FLAG)==CEParticle::PARTICLE_CLASS_FLAG) {
			if (!CEParticle::FitClassPk(part_PK_abs,pk)) return false;
		} else {
			if (part_PK_abs != getParticleKind()) return false;
		}
	} else {
		if (CLMath::abs(part_PK) != getParticleKind()) return false;
	}

	// Not zero: Psi*cP, d*Psi, PsiBar*c, dP*PsiBar
	// is_initial --> cP or dP
	// part_PK >= 0 --> c or cP
	bool couple_to_conj = CEUtil::LXOR((part_PK >= 0), is_initial);
	bool res = CEUtil::LEQU(couple_to_conj, isConjugated());
	return res;
}

CExpression* COpFermionField::GetExtParticleCoupling(CEParticle* part, bool is_initial, bool bClassesSource) {
	if (!CanBeCoupledExternal(part, is_initial, bClassesSource)) return NULL;
	long pk = (isConjugated())?(-m_particle_kind):m_particle_kind;
	long pk_part=part->get_PK();
	bool is_antipart = (pk_part < 0);
	CExpression* mom = part->get_momentum_symbol();
	bool del_mom=false;
	if (is_antipart) { mom=mom->clone(); mom->inverseFlag(); del_mom=true; } 

	//MARK: last change here
	if (bClassesSource) { // drop class-flag and append pk-identifier
		long pk_abs=((pk>=0)?pk:-pk);
		if ((pk_abs & CEParticle::PARTICLE_CLASS_FLAG) == CEParticle::PARTICLE_CLASS_FLAG) {
			long pkX=CLMath::abs(pk_part); if(pk<0)pkX=-pkX; pk=pkX;
		}
	}

	CExpression* res = new CWaveFuncFermion(pk, // pk>0: u(mom), pk<0: uBar(mom)
		m_vertex_ID,
		m_point_ID,
		mom);
	if (del_mom) { delete mom; }
	copyAttributesTo(res, true);
	return res;
}
