////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// AmplCalcLib.cpp: implementation of the CAmplCalcLib class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CAmplCalcLib::dbg = new CDebug("CAmplCalcLib");
const DWORD CAmplCalcLib::A_MANDELSTAMS				= 0x02;
const DWORD CAmplCalcLib::A_NO_CROSSINGS			= 0x80;
const DWORD CAmplCalcLib::A_USE_FIERZ_TRANSFORM		= 0x10;
const DWORD CAmplCalcLib::A_SIMPLE_BASIS			= 0x04; // unpolarized & massless

#define CALC_AMPL_SCROLL_NO	10

//////////////////////////////////////////////////////////////////////
// Static entry
//////////////////////////////////////////////////////////////////////
CExpression* CAmplCalcLib::CalcAmplitude(CMatrixElement* me, DWORD settings) {
	if (me == NULL) return NULL;// throw?
	dbg->log(10, "CalcAmplitude, me=", me);

	CAmplCalcLib* proc = new CAmplCalcLib();
	proc->doCalcAmplitude(me, settings);
	CExpression* res = proc->GetMEExpr();
	if (res == NULL) { dbg->log(0, "CAmplCalcLib::CalcAmplitude() ERROR, NULL result"); }
	delete proc;

	dbg->log(20, "CalcAmplitude() finished", res);
	return res;
}

CESymbol* CAmplCalcLib::m_BasisVecSymM = NULL;
CESymbol* CAmplCalcLib::m_BasisVecSym0 = NULL;
CESymbol* CAmplCalcLib::m_BasisVecPol = NULL;
CESymbol* CAmplCalcLib::m_BasisMassSym = NULL;
CESymbol* CAmplCalcLib::GetBasisVecSym(bool _mom, bool _massive) {
	CESymbol* res = NULL;
	if(_mom) { res = (_massive?m_BasisVecSymM:m_BasisVecSym0); } 
	else { res = m_BasisVecPol; }
	//if (do_init) { if(res==NULL){InitBasisSyms();res=GetBasisVecSym(_mom,_massive,false);} }
	return res;
}
CESymbol* CAmplCalcLib::GetBasisMassSym() {
	//if (do_init) { if(m_BasisMassSym==NULL)InitBasisSyms(); }
	return m_BasisMassSym;
}

void CAmplCalcLib::InitBasisSyms() {
	if (m_BasisVecSymM==NULL) {
		CESymbol* tmp = new CESymbol("QA\\_M");
		if (CLHEPBasics::isSymbolBooked(tmp)) {
			dbg->log(0, "ERROR: Can't book symbol for AMPLITUDE BASIS, unique is used", tmp);
			delete tmp; tmp = CESymbol::getUnique();
		}
		m_BasisVecSymM = tmp;
		CEProcessor::AddToDeleteList(m_BasisVecSymM);
	}
	if (m_BasisVecSym0==NULL) {
		CESymbol* tmp = new CESymbol("QA\\_0");
		if (CLHEPBasics::isSymbolBooked(tmp)) {
			dbg->log(0, "ERROR: Can't book symbol for AMPLITUDE BASIS, unique is used", tmp);
			delete tmp; tmp = CESymbol::getUnique();
		}
		m_BasisVecSym0=tmp;
		CEProcessor::AddToDeleteList(m_BasisVecSym0);
	}
	if (m_BasisVecPol==NULL) {
		CESymbol* tmp = new CESymbol("H\\_Q");
		if (CLHEPBasics::isSymbolBooked(tmp)) {
			dbg->log(0, "ERROR: Can't book symbol for AMPLITUDE BASIS, unique is used", tmp);
			delete tmp; tmp = CESymbol::getUnique();
		}
		m_BasisVecPol=tmp;
		CEProcessor::AddToDeleteList(m_BasisVecPol);
	}
	if (m_BasisMassSym==NULL) {
		CESymbol* tmp = new CESymbol("M\\_Q");
		if (CLHEPBasics::isSymbolBooked(tmp)) {
			dbg->log(0, "ERROR: Can't book symbol for AMPLITUDE BASIS, unique is used", tmp);
			delete tmp; tmp = CESymbol::getUnique();
		}
		m_BasisMassSym=tmp;
		CEProcessor::AddToDeleteList(m_BasisMassSym);
	}
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CAmplCalcLib::CAmplCalcLib() {
	m_settings=0;
	m_me=NULL;
	m_CalcBasisVecSym=NULL; m_BasisSpinorDM=NULL; m_CalcBasisMassSym=NULL; m_CalcBasisVectorE=NULL;
	m_DMCache = new CECache();
	m_BS_rule=NULL;
	m_kin=NULL;
	m_DensCache = NULL;

	m_mass_rule=NULL;
	m_kin_simp_rules=NULL;
	m_basic_coupling_vals=NULL;
	m_mand_vars_rule=NULL;
}

CAmplCalcLib::~CAmplCalcLib() {
	m_kin=NULL;
	if (m_me != NULL) { delete m_me; m_me=NULL; }
	// m_CalcBasisVecSym is not-cloned object
	// m_CalcBasisMassSym is not-cloned object
	// m_CalcBasisVectorE is not-cloned object
	if (m_BasisSpinorDM != NULL) { delete m_BasisSpinorDM; m_BasisSpinorDM=NULL; }
	if (m_BS_rule != NULL) { delete m_BS_rule; m_BS_rule=NULL; }
	if (m_DMCache != NULL) { delete m_DMCache; m_DMCache=NULL; }
	if (m_DensCache != NULL) { delete m_DensCache; m_DensCache=NULL; }

	if (m_mass_rule != NULL) { delete m_mass_rule; m_mass_rule=NULL; }
	if (m_kin_simp_rules != NULL) { delete m_kin_simp_rules; m_kin_simp_rules=NULL; }
	if (m_basic_coupling_vals != NULL) { delete m_basic_coupling_vals; m_basic_coupling_vals=NULL; }
	if (m_mand_vars_rule != NULL) { delete m_mand_vars_rule; m_mand_vars_rule=NULL; }
}

//////////////////////////////////////////////////////////////////////
// result access
//////////////////////////////////////////////////////////////////////
// to retrieve ME result
CMatrixElement* CAmplCalcLib::ExtractME() { CMatrixElement* res=m_me; m_me=NULL; return res; }

// to retrieve Expr result
CExpression* CAmplCalcLib::GetMEExpr() {
	if (m_me != NULL) {
		CExpression* me = m_me->get_expr_val();
		if (me != NULL) return me->clone();
	}
	return NULL;
}

//////////////////////////////////////////////////////////////////////
// Implementation
//////////////////////////////////////////////////////////////////////
void CAmplCalcLib::doCalcAmplitude(CMatrixElement* me, DWORD settings) {
	///////////////////////////////////////////////// <settings arrangement>
	m_settings = settings;
	///////////////////////////////////////////////// </settings arrangement>

	///////////////////////////////////////////////// <init & check-outs>
	if (me == NULL) return;
	if (CETensor::isNDimensionSpace()) { dbg->log(0, "{Error} CalcAmplitude() call at N-dimesion space, action cancelled"); return; }
	m_kin = CEProcessor::kinematics();
	if (m_kin == NULL) { dbg->log(0, "{Error} CalcAmplitude() No kinematics installed"); return; }

	m_kin_simp_rules = m_kin->getReplacements(CKinematics::REPL_COUPLINGS_SIMPLIFY);
	bool CMS_beams = false;
	m_basic_coupling_vals = m_kin->getBasicCouplingsValues(CMS_beams);
	m_mand_vars_rule = NULL;
	if (m_settings & CAmplCalcLib::A_MANDELSTAMS) {
		m_mand_vars_rule = m_kin->getReplacements(CKinematics::REPL_MANDELSTAM_VARIABLES);
		if ((m_mand_vars_rule != NULL) && (m_mand_vars_rule->size() == 0)) { delete m_mand_vars_rule; m_mand_vars_rule = NULL; }
	}
	m_mass_rule = CLHEPBasics::get_mass_order_rule();
	dbg->log(25, "kin rules", m_kin_simp_rules);
	dbg->log(25, "couplings vals", m_basic_coupling_vals);
	dbg->log(25, "mandelstams rules", m_mand_vars_rule);
	dbg->log(25, "mass rules", m_mass_rule);

	m_CalcBasisVecSym=NULL; // m_CalcBasisVecSym is not cloned object
	m_CalcBasisMassSym=NULL; //m_CalcBasisMassSym is not cloned object
	m_CalcBasisVectorE=NULL; // m_CalcBasisVectorE is not cloned object
	if (m_BasisSpinorDM!=NULL) { delete m_BasisSpinorDM; m_BasisSpinorDM=NULL; }
	if (m_BS_rule != NULL) { delete m_BS_rule; m_BS_rule=NULL; }

	if(m_DMCache==NULL) {m_DMCache=new CECache();} else {m_DMCache->clear_all();}

	CExpression* _zero = CENumber::GetZero(); // temporary

	m_me = me->clone();
	///////////////////////////////////////////////// </init & check-outs>

	///////////////////////////////////////////////// <prepare>
	m_DensCache = new CDenCacheLib();
	m_DensCache->AddSimplRule(m_kin_simp_rules);
	if (m_mand_vars_rule != NULL) { m_DensCache->AddSimplRule(m_mand_vars_rule); }
	m_DensCache->AddSimplRule(m_basic_coupling_vals);
	m_DensCache->AddSimplRule(m_mass_rule);

	// choose basis vector symbol, mass etc.
	InitBasisSyms();
	bool _AbstractPolarizedBasis=false;
	bool _MassiveBasis=false;
	if ((m_settings & CAmplCalcLib::A_SIMPLE_BASIS)==CAmplCalcLib::A_SIMPLE_BASIS) {
		_AbstractPolarizedBasis=false;
		_MassiveBasis=false;
	} else {
		_AbstractPolarizedBasis=existMasslessFermion();
		_MassiveBasis=false;
	}
	m_CalcBasisVecSym = GetBasisVecSym(true,_MassiveBasis);
	dbg->log(35, "CalcAmplitude(): basis vector symbol involved:", m_CalcBasisVecSym);

	// basis mass
	if (_MassiveBasis) {
		m_CalcBasisMassSym = GetBasisMassSym();
		CExpression* m2 = m_CalcBasisMassSym->sqra();
		CEVector::AddToVecSquareCache(m_CalcBasisVecSym,m2); delete m2;
		dbg->log(2, "CalcAmplitude(): basis vector MASS symbol involved:", m_CalcBasisMassSym);
	} else {
		m_CalcBasisMassSym = NULL;
		CEVector::AddToVecSquareCache(m_CalcBasisVecSym,_zero);
		dbg->log(2, "CalcAmplitude(): basis vector is set massless");
	}

	// polarization
	if (_AbstractPolarizedBasis) {
		m_CalcBasisVectorE = GetBasisVecSym(false,_MassiveBasis);
		CExpression* _Minus1 = CENumber::GetUnit(); _Minus1->inverseFlag();
		CEVector::AddToVecSquareCache(m_CalcBasisVectorE,_Minus1);
		delete _Minus1;
		dbg->log(2, "CalcAmplitude(): basis vector polarization symbol involved:", m_CalcBasisVectorE);
		//TODO: check for CPProtonSF particle -- error will occur here for PROTON
		m_BasisSpinorDM = CPFermion::createDM(m_CalcBasisVecSym, m_CalcBasisMassSym, m_CalcBasisVectorE, true);
		dbg->log(2, "CalcAmplitude(): basis vector density matrix involved:", m_BasisSpinorDM);
	} else {
		m_CalcBasisVectorE = NULL;
		m_BasisSpinorDM = CPFermion::createDM(m_CalcBasisVecSym, 1, true, NULL);
		dbg->log(2, "CalcAmplitude(): basis vector density matrix involved:", m_BasisSpinorDM);
	}

	// m_BS_rule init
	// old: m_BS_rule was used to set basis vector p.p-->0
	if (m_CalcBasisVectorE != NULL) {
		// p.e-->0
		CExpression* q_E = CExpression::arrge(new CESProduct(m_CalcBasisVecSym, m_CalcBasisVectorE));
		m_BS_rule = new CEReplasement(q_E, _zero, false); //, CEReplasement::REPL_AVOID_NULL_DENOMINATORS
		delete q_E;
	}

	// append crossings (Note: may be done the after evaluation)
	if ((m_settings & CAmplCalcLib::A_NO_CROSSINGS)==0) {
		CEPtrArray* cross_infos = m_kin->getCrossingsTransformations();
		if (cross_infos != NULL) {
			if (cross_infos->size() > 0) { m_me->AppendCrossings(cross_infos); }
			delete cross_infos;
		}
	}

	// turn ME to "Sum{Prod{.uOu..uOu.}}" form
	if (!CMELib::ToNormalForm(m_me)) { dbg->log(0, "CalcAmplitude() ERROR: can't normalize ME form:", m_me); return; }
	///////////////////////////////////////////////// </prepare>
	
	///////////////////////////////////////////////// <simplify amplitude> // m-->0 etc.
	CExpression* me_expr = m_me->get_expr_val();
	CExpression* me_res = kin_reduce(me_expr,false,false,false);
	if (me_res != me_expr) { m_me->set_expr_val(me_res); delete me_res; }
	///////////////////////////////////////////////// <simplify amplitude>

	///////////////////////////////////////////////// <calc amplitude> //m_me == SUM{me}
	// 1. inline uu->Tr{}/()
	calcFermionSeqs(m_me->get_expr_val());
	dbg->log(35, "CalcAmplitude(): uOu-->Tr[]/sqrt{...} substituted", m_me);

	// 2. Involve Fierz identities //TODO: implement it

	// 3. Traces evaluation & vector indices summation in (..+.+..)^i..(..+.+..)_i
	calcWholeME();
	///////////////////////////////////////////////// <calc amplitude>

	///////////////////////////////////////////////// <arrange result>
	m_me->Arrange();
	///////////////////////////////////////////////// </arrange result>

	///////////////////////////////////////////////// <finalize>
	delete _zero;
	///////////////////////////////////////////////// </finalize>
	dbg->log(35, "CalcAmplitude(): completed", m_me);
}

/*
*	Changes \hat{u}Ou-tales to Sp{O Pi P Pj}/sqrt{ Sp{P Pi} Sp{P Pj} } in ME
*	Expressions are processed in-line
*/
void CAmplCalcLib::calcFermionSeqs(CExpression* me_expr) {

	// parse ME
	if(me_expr->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* sum_me = (CESum*)me_expr;
		long s = sum_me->size();
		long i;
		for(i=0; i<s; i++) { calcFermionSeqs(sum_me->GetAt(i)); }
	} else if(me_expr->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		calcFermionSeqsPr((CEProduct*)me_expr);
	}

	// check for no spinor left in expression
	if (CMELib::checkContainSpinors(me_expr)) { dbg->log(0, "CAmplCalcLib::calcFermionSeqs() Internal error, unexpected spinor", me_expr); CEException::halt(); }
}

void CAmplCalcLib::calcFermionSeqsPr(CEProduct* prod) {
	dbg->log(35, "calcFermionSeqsPr call", prod);
	CSequence* cpls = CLFermionSpinSum::ExtractSpinorCouplings(prod);
	long s = cpls->size();
	long i;
	for(i=0; i<s; i++) {
		CExpression* item = cpls->GetAt(i);
		if(!item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			dbg->log(0, "calcFermionSeqsPr() internal error: non-product UU-seq", item); CEException::halt();
		}

		CExpression* subst = calcSeqSubst((CEProduct*)item);
		prod->addItem(subst);
	}
	delete cpls;
	dbg->log(35, "calcFermionSeqsPr() result", prod);
}

// source may be changed, but not deleted, result is cloned
CExpression* CAmplCalcLib::calcSeqSubst(CEProduct* uu_seq) {
	dbg->log(55, "calcSeqSubst() call", uu_seq);
	long s = uu_seq->size();
	CExpression* tmp = uu_seq->GetAt(0);
	if(!tmp->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) { dbg->log(0, "calcSeqSubst() INTERNAL ERROR: Spinor expected", uu_seq); CEException::halt(); }
	CESpinor* s_left = (CESpinor*)tmp;
	if (!s_left->is_conjugated()) { dbg->log(0, "calcSeqSubst() INTERNAL ERROR: Conjugated spinor is expected", uu_seq); CEException::halt(); }
	tmp = uu_seq->GetAt(s-1);
	if((tmp == NULL) || (!tmp->getProperty(CEProperties::PROP_MASK_IS_SPINOR))) { dbg->log(0, "calcSeqSubst() INTERNAL ERROR: Spinor expected", uu_seq); CEException::halt(); }
	CESpinor* s_right = (CESpinor*)tmp;
	if (s_right->is_conjugated()) { dbg->log(0, "calcSeqSubst() INTERNAL ERROR: Non-conjugated spinor is expected", uu_seq); CEException::halt(); }
	CExpression* p_left = s_left->base();
	CExpression* p_right = s_right->base();

	CExpression* dm_left = getCachedDM(p_left);
	if (dm_left != NULL) {
		dm_left = dm_left->clone();
	} else {
		dm_left = findFerionDM(p_left);
		addDMToCache(p_left, dm_left);
	}
	dm_left = kin_reduce(dm_left,false,false,true);

	CExpression* dm_right = getCachedDM(p_right);
	if (dm_right != NULL) {
		dm_right = dm_right->clone();
	} else  {
		dm_right = findFerionDM(p_right);
		addDMToCache(p_right, dm_right);
	}
	dm_right = kin_reduce(dm_right,false,false,true);

	CExpression* res = NULL;

	// Tr{O P2 P P1}
	CEProduct* pr_tmp = new CEProduct();
	long i;
	for(i=1; i<(s-1); i++) { // add internal objects of UU-tail
		pr_tmp->addItem(uu_seq->GetAt(i)->clone());
	}
	pr_tmp->addItem(dm_right->clone());
	pr_tmp->addItem(m_BasisSpinorDM->clone());
	pr_tmp->addItem(dm_left->clone());
	res = new CFSp(pr_tmp); delete pr_tmp;

	// Sqr{ Tr{P P2} Tr(P P1) }
	pr_tmp = new CEProduct();
	CEProduct* pr_den1 = new CEProduct();
	pr_den1->addItem(m_BasisSpinorDM->clone());
	pr_den1->addItem(dm_right->clone());
	//TEMP dbg->log(0, "denominator pr_den1",pr_den1);
	//TEMP CExpression* tmpX1=new CFSp(pr_den1); dbg->log(0, "tmpX1",tmpX1); CExpression* tmpX2=tmpX1->evaluate(); if(tmpX2!=tmpX1) delete tmpX1;
	//TEMP dbg->log(0, "tmpX2",tmpX2); tmpX2=CExpression::reduce(tmpX2); dbg->log(0, "R tmpX2",tmpX2); delete tmpX2;
	pr_tmp->addItem(new CFSp(pr_den1)); delete pr_den1;

	CEProduct* pr_den2 = new CEProduct();
	pr_den2->addItem(m_BasisSpinorDM->clone());
	pr_den2->addItem(dm_left->clone());
	pr_tmp->addItem(new CFSp(pr_den2)); delete pr_den2;

	res = res->divide(CEPower::sqrt(pr_tmp));

	delete dm_left; delete dm_right;
	dbg->log(55, "calcSeqSubst() result", res);
	return res;
}

CExpression* CAmplCalcLib::findFerionDM(CExpression* p) {
	CExpression* p_abs = CLSimplifyMisc::TruncateSign(p);
	CEParticle* part = m_kin->get_particle(p_abs);
	delete p_abs;
	if (part == NULL) {
		dbg->log(0, "ERROR: ME does not fit kinematics, no particle for momentum", p_abs); CEException::halt();
	}
	if (!part->check_PK(CEParticle::PARTICLE_FERMION, true)) {
		dbg->log(0, "ERROR: ME does not fit kinematics, non-fermion particle for spinor", p_abs); CEException::halt();
	}

	if (part->get_polarization_info() == 0) {
		dbg->log(0, "ERROR: Amplitudes are not calculatable for UNPOLARIZED fermions", part); CEException::halt();
	}

	// RETRIEVE MASS SYMBOL AND POLARIZATION STATE FROM KINEMATICS
	// LET PARTICLE OBJECT DEFINE THE DENSITY MATRIX
	// Note: unpolarized final-state fermions will have "*2" in this way
	// TODO: define what to do with unpolarized fermions
	CEPtrArray* DM_vars = new CEPtrArray(); DM_vars->add(p->clone());
	CExpression* res = part->getDensityMatrix(DM_vars);
	delete DM_vars;
	if (res == NULL) { dbg->log(0, "Internal Error: No density matrix found for particle", part); CEException::halt(); }

	return res;
}

CExpression* CAmplCalcLib::getCachedDM(CExpression* p) { return m_DMCache->get_val(p); }
void CAmplCalcLib::addDMToCache(CExpression* p, CExpression* dm) { m_DMCache->put_value(p,dm); }

/*CExpression* CAmplCalcLib::getCachedDM(CExpression* p) {
	int s = m_DMCache->size();
	int i;
	for(i=0; i<s; i++) {
		CSequence* item_seq = (CSequence*)m_DMCache->GetAt(i);
		if (p->equals(item_seq->GetAt(0))) return item_seq->GetAt(1);
	}
	return NULL;
}
void CAmplCalcLib::addDMToCache(CExpression* p, CExpression* dm) {
	CSequence* item_seq = new CSequence();
	item_seq->addItem(p->clone());
	item_seq->addItem(dm->clone());
	m_DMCache->addItem(item_seq);
}*/

void CAmplCalcLib::calcWholeME() { // parses sum & processes me item-by-item
	CExpression* me_expr = m_me->get_expr_val();

	// parse ME
	if(me_expr->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* me_sum = (CESum*)me_expr;
		long s = me_sum->size();

		//	*****************************************************	<scroll>
		CEProgressWatch* pw=NULL; if(s>=CALC_AMPL_SCROLL_NO){pw=CEProcessor::newProgressWatch("Amplitude calculation (%ld items)",s);}
		//	*****************************************************	<\scroll>
		long i;
		for(i=0; i<s; i++) {
			CExpression* me_item = me_sum->GetAt(i);
			CExpression* res_item = calcSingleME(me_item);
			if (res_item != me_item) { me_sum->SetAt(i, res_item); delete res_item; }
			//	*****************************************************	<scroll>
			if (pw != NULL) pw->step(i+1, s);
			//	*****************************************************	<\scroll>
		}
		//	*****************************************************	<scroll>
		if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
		//	*****************************************************	<\scroll>
	} else {
		CExpression* me_res = calcSingleME(me_expr);
		if (me_res != me_expr) { m_me->set_expr_val(me_res); delete me_res; }
	}
}

// may change src, but not delete
CExpression* CAmplCalcLib::calcSingleME(CExpression* me) {
	dbg->log(45, "calcSingleME():", me);

	///////////////////////////////////////////////// <calc traces>
	CExpression* res=NULL;
	bool bApplyFierz=false;
	if ((m_settings & CAmplCalcLib::A_USE_FIERZ_TRANSFORM)==CAmplCalcLib::A_USE_FIERZ_TRANSFORM) bApplyFierz=true;
	if (CFSp::m_TracesBondarev) bApplyFierz=true;
	if (bApplyFierz) {
		CExpression* res_tmp = CExpression::reduce(me->clone());
		DWORD n=0;
		res_tmp = CLFiertzTransform::DoFiertzTransform(res_tmp, &n);
		dbg->log(105, n, "Fierz transform applied %ld times",res_tmp);
		if (n>0) {
			res = res_tmp->evaluate();
			if (res_tmp != res) delete res_tmp;
		} else {
			delete res_tmp;
			res = me->evaluate();
			if(res==me) { res=res->clone(); } // clone me for simplier further processing
		}
	} else {
		res = me->evaluate();
		if(res==me) { res=res->clone(); } // clone me for simplier further processing
	}
	dbg->log(55, "calcSingleME(): traces calculated:", res);
	///////////////////////////////////////////////// </calc traces>

	res = m_DensCache->InvolveDenominatorsCache(res);
	m_DensCache->SimplifyCachedDenominators();

	///////////////////////////////////////////////// <couple all the boson helicities>
	CEPtrArray* v_inds = m_me->get_external_indices_list();
	int i;
	for(i=0; i<v_inds->size(); i++) {
		CExprContainer* cart_i = (CExprContainer*)v_inds->GetAt(i);
		int part_ID = cart_i->get_info();
		CEParticle* part_i = m_kin->get_particle_by_ID(part_ID);
		if (part_i == NULL) { dbg->log(0, part_ID, "ERROR No particle for part_ID=%ld"); CEException::halt(); }
		if (!part_i->check_PK(CEParticle::PARTICLE_VECTOR_BOSON, true)) { dbg->log(0, part_ID, "ERROR ME does not fit kinematics, PID=%ld, boson expected", part_i); CEException::halt(); }
		CPAbstractBoson* part_boson = (CPAbstractBoson*)part_i;
		CExpression* indx = cart_i->get_expr();

		CExpression* pol_sym = part_boson->get_pol_vector_symbol();
		res = res->multiply(new CEVector(pol_sym, indx, false));
	}
	///////////////////////////////////////////////// </couple all the boson helicities>

	///////////////////////////////////////////////// <kin-arrange>
	res = kin_reduce(res,true,true,true);
	//int ka_setts = CLHEPBasics::KA_NO_EXPAND;
	//int ka_setts = 0;
	//if (m_settings & CAmplCalcLib::A_MANDELSTAMS) { ka_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
	//res = CLHEPBasics::KinArrange(res, ka_setts);
	//dbg->log(25, "CalcAmplitude(): kin-arranged:", res);

	res = CLFactor::factorize(res);
	//int ks_setts = CLHEPBasics::KS_FACTORIZE_DENOMINATORS;
	//if (m_settings & CAmplCalcLib::A_MANDELSTAMS) { ks_setts |= CLHEPBasics::KS_INVOLVE_MANDELSTAM_VARS; }
	//res = CLHEPBasics::KinSimplify(res, ks_setts);
	//dbg->log(25, "CalcAmplitude(): kin-simplified:", res);
	///////////////////////////////////////////////// </kin-arrange>

	res = m_DensCache->RestoreDenominators(res);
	return res;
}

/*
*	Similar to KinArrange, but includes m_BS_rule rule
*/
CExpression* CAmplCalcLib::kin_reduce(CExpression* val, bool do_expand_res, bool do_expand_src, bool del_src) {
	dbg->log(55, "kin_reduce():", val);
	//m_BS_rule; m_mass_rule; m_kin_simp_rules; m_basic_coupling_vals; m_mand_vars_rule;
	CExpression* res = val;
	if (do_expand_src) {
		// apply some rules to make reduction faster
		if (m_mass_rule != NULL) { res = m_mass_rule->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
		if (m_kin_simp_rules != NULL) { res = m_kin_simp_rules->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
		if (m_basic_coupling_vals != NULL) { res = m_basic_coupling_vals->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }

		res = CExpression::reduce(res,del_src); // couple tensor indices
		if ((!del_src)&&(res!=val)) del_src=true;
	}

	if (CEAntiSymmProd::InvolveAsymmProd(res)) {
		if (do_expand_res) { res= CExpression::reduce(res,del_src); }
		else { res= CExpression::arrge(res,del_src); }
		if ((!del_src)&&(res!=val)) del_src=true;
	}

	if (m_BS_rule != NULL) { res = m_BS_rule->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
	if (m_kin_simp_rules != NULL) { res = m_kin_simp_rules->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
	if (m_basic_coupling_vals != NULL) { res = m_basic_coupling_vals->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
	if (m_mand_vars_rule != NULL) { res = m_mand_vars_rule->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
	if (m_mass_rule != NULL) { res = m_mass_rule->ApplyTo(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }

	if (do_expand_res) { res= CExpression::reduce(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
	else { res= CExpression::arrge(res,del_src); if ((!del_src)&&(res!=val)) del_src=true; }
	dbg->log(55, "kin_reduce() result:", res);
	return res;
}

bool CAmplCalcLib::existMasslessFermion() {
	if (m_kin==NULL) CEException::halt("CAmplCalcLib::existMasslessFermion()");
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics(); if(phys==NULL) return false;
	CEProcessEnvironment* env = CEProcessor::getEnvironment();

	int s = m_kin->num_init_particles();
	int i;
	for(i=0; i<s; i++) {
		CEParticle* part = m_kin->pin(i);
		int pk = part->get_PK();
		if (CEParticle::is_fermion_PK(pk)) {
			if (env->GetMassCalcOrder(pk) == 0) return true;
			if (CEBasic::isZero(part->get_sym_mass())) return true;
		}
	}
	
	s = m_kin->num_final_particles();
	for (i=0; i<s; i++) {
		CEParticle* part = m_kin->pfin(i);
		int pk = part->get_PK();
		if (CEParticle::is_fermion_PK(pk)) {
			if (env->GetMassCalcOrder(pk) == 0) return true;
			if (CEBasic::isZero(part->get_sym_mass())) return true;
		}
	}
	return false;
}
