////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LDiagOpSeqRestore.cpp: implementation of the CLDiagOpSeqRestore class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLDiagOpSeqRestore::dbg = new CDebug("CLDiagOpSeqRestore");
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CLDiagOpSeqRestore::CLDiagOpSeqRestore(CDiagramPhysics* _phys, CKinematics* _kin) : CELibrary() {
	m_phys = _phys; m_kin = _kin; m_res_diag = NULL; m_source = NULL;
}

CLDiagOpSeqRestore::~CLDiagOpSeqRestore() { if (m_res_diag != NULL) delete m_res_diag; }

/*
*	Does not affects the source.
*/
CFeynDiagram* CLDiagOpSeqRestore::RestoreDiagram(CMEOperatorSeq* source, CDiagramPhysics* _phys, CKinematics* _kin) {
	dbg->log(20, "RestoreDiagram() call", source);
	if (source == NULL) { dbg->log(0, "Error: NULL source."); return NULL; }
	if (_phys == NULL) { dbg->log(0, "Error: No diagram physics found. Engine stopped."); return NULL; }
	if (_kin == NULL) { dbg->log(0, "Error: No kinematics found. Engine stopped."); return NULL; }
	///////////////////////////////////////////////////////////////////////////

	CLDiagOpSeqRestore* lib = new CLDiagOpSeqRestore(_phys, _kin);
	CFeynDiagram* res = lib->RestoreDiagramImpl(source);
	delete lib;
	///////////////////////////////////////////////////////////////////////////

	dbg->log(20, "RestoreDiagram() result", res);
	return res;
}

/*
*	Does not affects the source.
*/
CFeynDiagram* CLDiagOpSeqRestore::RestoreDiagramImpl(CMEOperatorSeq* source) {
	m_source = source;

	// 1. create all external lines
	m_res_diag = CFeynDiagram::create_blanc(m_phys, true, m_kin);

	// 2. create all vertices (point IDs should equal vertex number)
	CreateVerticesList();
	dbg->log(55, "vertices created", m_res_diag);

	// 3. bind external lines to vertices
	BindExtLines();
	dbg->log(55, "ext lines binded", m_res_diag);

	// 4. create internal lines for every coupling and bind to vertices
	CreateIntLines();
	dbg->log(55, "internal lines created", m_res_diag);

	// 5. attach common diagram factor
	long m = source->get_multiplicity();
	int n = source->getMaxPointID() + 1; // diagram order
	long n_fact = CLMath::factorial(n);
	//bool ferm_sign = source->getStandardOrderSign();
	bool ferm_sign = source->get_ferm_sign();
	if ((!ferm_sign) || (m != n_fact)) {
		CENumber* factor_expr = NULL;
		if(m!=n_fact){ factor_expr = CENumber::GetFraction(m,n_fact); factor_expr->Arrange(AL_INLINE); }
		else factor_expr = CENumber::GetUnit();
		if (!ferm_sign) factor_expr->inverseFlag();
		m_res_diag->set_factor_expr(factor_expr);
		delete factor_expr;
	}
	dbg->log(55, "diagram factor arranged");

	CFeynDiagram* res = m_res_diag; m_res_diag = NULL;
	return res;
}

void CLDiagOpSeqRestore::CreateVerticesList() {
	CEPtrArray* vi_arr = m_phys->getVerticesInfo();
	int vi_arr_size = vi_arr->size();
	long s = m_source->size();

	int points_num = m_source->getMaxPointID() + 1;
	long i;
	for(i=0; i<points_num; i++) {

		// find out vertex ID
		int vertex_ID = -1;
		bool vert_ID_found = false;
		long j;
		for (j=0; j<s; j++) {
			CExpression* itm_expr = m_source->GetAt(j);
			if (itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) {
				CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
				if (fo->getPointID() == i) {
					vertex_ID = fo->getVertexID(); vert_ID_found = true; break;
				}
			}
			else if (itm_expr->getProperty(CEProperties::PROP_MASK_IS_WAVE_FUNCTION)) {
				CParticleWaveFunction* wf = (CParticleWaveFunction*)itm_expr;
				if (wf->getPointID() == i) {
					vertex_ID = wf->getVertexID(); vert_ID_found = true; break;
				}
			}
		}
		if (!vert_ID_found) { dbg->log(0, i, "Error: No field operator or wave function found at point ID == %ld", m_source); CEException::halt(); }
		if ((vertex_ID < 0) || (vertex_ID > vi_arr_size)) {
			dbg->log(0, vertex_ID, "Error: Invalid vertex info ID found: %ld", m_source); CEException::halt();
		}

		// create vertex object
		CDiagramVertexInfo* vi = (CDiagramVertexInfo*)vi_arr->GetAt(vertex_ID);
		CDVertex* v = new CDVertex(vi, NULL);
		m_res_diag->add_vertex(v);
	}
}

void CLDiagOpSeqRestore::BindExtLines() {
	int s = m_res_diag->get_lines_number();
	int i;
	for(i=0; i<s; i++) {
		CDLine* l = m_res_diag->get_line(i);
		if (l->is_external()) {
			CDExtLine* el = (CDExtLine*)l;
			CEParticle* part = m_kin->get_particle_by_ID(el->get_particle_ID());
			if (part == NULL) { dbg->log(0, el->get_particle_ID(), "Error: Can't find particle (ID == %ld)", el); CEException::halt(); }
			CParticleWaveFunction* wf = m_source->GetWaveFunction(part);
			if (wf == NULL) { dbg->log(0, el->get_particle_ID(), "Error: Particle (ID == %ld) is not coupled to source sequence", m_source); CEException::halt(); }

			CDVertex* v = m_res_diag->get_vertex(wf->getPointID()); // point ID corresponds the vertex no.
			el->get_links()->add(v);
			v->get_links()->add(el); // fill vertex data in arbitrary order first
		}
	}
}

void CLDiagOpSeqRestore::CreateIntLines() {

	// 1. Attach internal lines to vertices in arbitrary order
	int s = m_source->getMaxCouplingNo();
	int i;
	for(i=1; i<=s; i++) {
		CEFieldOperator* fo1 = m_source->GetFieldOperatorCoupled(i, 0);
		CEFieldOperator* fo2 = m_source->GetFieldOperatorCoupled(i, 1);
		if ((fo1 == NULL) || (fo2 == NULL)) { dbg->log(0, i, "<Error> Can't find operator for coupling #%ld", m_source); CEException::halt(); }

		// arrange operators order (for fermions)
		if (CEParticle::is_fermion_PK(fo1->getParticleKind())) {
			if ((!CEUtil::equals(fo1->ClassName(), COpFermionField::CLASS_NAME)) || (!CEUtil::equals(fo2->ClassName(), COpFermionField::CLASS_NAME))) { dbg->log(0, "Operator types internal error"); CEException::halt(); }

			COpFermionField* ffo1 = (COpFermionField*)fo1;
			COpFermionField* ffo2 = (COpFermionField*)fo2;
			if (CEUtil::LEQU(ffo1->isConjugated(), ffo2->isConjugated())) {
				dbg->log(0, "<Error> Invalid fermion operators coupling: similar conjugation data", m_source); CEException::halt(); }

			if (ffo2->isConjugated()) {
				// swap operators
				fo1 = ffo2; fo2 = ffo1;
			}
		} else {
			// arrange operators order (for charged bosons)
			if (!CEParticle::is_neutral_bosonic_PK(fo1->getParticleKind())) {
				if ((!CEUtil::equals(fo1->ClassName(), COpBosonField::CLASS_NAME)) || (!CEUtil::equals(fo2->ClassName(), COpBosonField::CLASS_NAME))) { dbg->log(0, "Operator types internal error"); CEException::halt(); }
				
				COpBosonField* bfo1 = (COpBosonField*)fo1; COpBosonField* bfo2 = (COpBosonField*)fo2;
				if (CEUtil::LEQU(bfo1->isConjugated(), bfo2->isConjugated())) { dbg->log(0, "<Error> Invalid fermion operators coupling: similar conjugation data", m_source); CEException::halt(); }
				if (bfo2->isConjugated()) {
					// swap operators
					fo1 = bfo2; fo2 = bfo1;
				}
			}
		}

		// now fo1 is conjugated (if charged)
		// line if from Psi_bar to Psi, pk is for incoming to Psi(link 2)
		bool _is_CT_line = fo2->isCounterTermCoupling();
		CDIntLine* int_l_new = NULL;
		if (!_is_CT_line) int_l_new = CDIntLine::createIntLine(fo2->getParticleKind(), NULL);
		else int_l_new = CDIntLine::createCounterTermLine(fo2->getParticleKind(), NULL);
		m_res_diag->add_line(int_l_new);

		CDVertex* v1 = m_res_diag->get_vertex(fo1->getPointID());
		int_l_new->get_links()->add(v1);
		v1->get_links()->add(int_l_new);

		CDVertex* v2 = m_res_diag->get_vertex(fo2->getPointID());
		int_l_new->get_links()->add(v2);
		v2->get_links()->add(int_l_new);
	}
	dbg->log(55, "internal lines created (arbitrary order)", m_res_diag);

	// 2. Sort lines order according to VertexInfo structure at every vertex
	s = m_res_diag->get_vertices_number();
	for (i=0; i<s; i++) {
		CDVertex* v = m_res_diag->get_vertex(i);
		dbg->log(155, i, "vertex sort #%ld");
		bool sort_res = v->sortLinksOrder();
		if(!sort_res) { dbg->log(0, "<Error diagram restoration> Error vertex binding", m_res_diag); CEException::halt(); }
	}
}
