////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LFermionSpinSum.cpp: implementation of the CLFermionSpinSum class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLFermionSpinSum::dbg = new CDebug("CLFermionSpinSum");
CLFermionSpinSum* CLFermionSpinSum::LibInstance = NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLFermionSpinSum::CLFermionSpinSum() : CELibrary() {
	m_spinor_couplings = NULL; m_sum_couplings_result = NULL;
}

CLFermionSpinSum::~CLFermionSpinSum() {}

CEObject* CLFermionSpinSum::doTask() { return NULL; }

/*
*	Couples the (..u)...(u..)-tales.
*	Separates gell-mann objects to separate chains
*
*	Processes objects inline.
*/
void CLFermionSpinSum::sum_fermion_spins(CExpression *source)
{
	dbg->log(20, "sum spins", source);
	if (source == NULL) return;

	if (source->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		sum_spins_in_product((CEProduct*)source);
	}
	else if (source->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* sum = (CESum*)source;

		long s = sum->size();
		long i;
		for (i=0; i<s; i++) { sum_fermion_spins(sum->GetAt(i)); }
	}
	dbg->log(20, "sum spins result", source);
}

void CLFermionSpinSum::sum_spins_in_product(CEProduct *source)
{
	if (source == NULL) return;

	if (LibInstance == NULL) LibInstance = new CLFermionSpinSum();

	// process source
	LibInstance->sum_spins_task_impl(source);

	// delete lib instanse (or include it to finalization delete list)
	delete LibInstance; LibInstance = NULL;
}

// source to be changed
CSequence* CLFermionSpinSum::ExtractSpinorCouplings(CEProduct* source) {
	if (source == NULL) return NULL;

	if (LibInstance == NULL) LibInstance = new CLFermionSpinSum();

	// process source
	CSequence* res = LibInstance->extract_couplings_task_impl(source);

	// delete lib instanse (or include it to finalization delete list)
	delete LibInstance; LibInstance = NULL;

	// self-check
	if (CMELib::checkContainSpinors(source)) { dbg->log(0, "CLFermionSpinSum::ExtractSpinorCouplings() Internal error, unexpected spinor after extraction", source); CEException::halt(); }

	return res;
}

CSequence* CLFermionSpinSum::extract_couplings_task_impl(CEProduct* seq) {
	dbg->log(30, "sum_spins_task_impl", seq);
	if (seq == NULL) return NULL;
	
	init_processing();// init
	
	try {
		// 1. extract all the spinor couplings into array
		extract_spinor_couplings(seq); // fill out couplings array
	}
	catch(CEException* exc) { throw exc; }
	catch(...) { dbg->log(0, "Unexpected error caught."); CEException::halt(); }

	CSequence* res=m_spinor_couplings; m_spinor_couplings=NULL;
	finalize_processing();

	return res;
}

void CLFermionSpinSum::sum_spins_task_impl(CEProduct *seq) {
	dbg->log(30, "sum_spins_task_impl", seq);
	if (seq == NULL) return;
	
	init_processing();// init
	
	try {
		// 1. extract all the spinor couplings into array
		extract_spinor_couplings(seq); // fill out couplings array
		
		if (m_spinor_couplings->size() > 0) {
			// 2. examine couplings array
			bool sum_OK = sum_spins_in_selected_couplings();
			if (!sum_OK) {dbg->log(0, "CLFermionSpinSum fails."); CEException::halt();}
			
			// 3. add spin sum result to source product
			if (m_sum_couplings_result != NULL) {
				seq->addItemInline(m_sum_couplings_result);
				m_sum_couplings_result = NULL;
			}

			// 4. add rest spinor couplings
			int rest_sz = m_spinor_couplings->size();
			if (rest_sz > 0) {
				int i;
				for(i=0; i<rest_sz; i++) {
					seq->addItemInline(m_spinor_couplings->GetAt(i));
				}
			}

			// WHAT ABOUT U..UU..U --> 1*RES: "1" REMAINS
			// ARRANGE HERE OR CHECK ON UNITS OR LEAVE AS IT IS?
			// 5. inline arrange
			seq->Arrange(AL_INLINE);
		}
	}
	catch(CEException* exc) {
		throw exc;
	}
	catch(...) {
		dbg->log(0, "Unexpected error caught.");
		printf("Unexpected error caught in CLFermionSpinSum.\n");
		CEException::halt();
	}

	finalize_processing();
}

void CLFermionSpinSum::init_processing()
{
	dbg->log(30, "init_processing()");
	if (m_spinor_couplings != NULL) delete m_spinor_couplings;
	m_spinor_couplings = new CSequence();

	if (m_sum_couplings_result != NULL) {
		delete m_sum_couplings_result; m_sum_couplings_result = NULL;
	}
}

void CLFermionSpinSum::finalize_processing()
{
	if (m_spinor_couplings != NULL) {delete m_spinor_couplings; m_spinor_couplings = NULL;}
	if (m_sum_couplings_result != NULL) {delete m_sum_couplings_result; m_sum_couplings_result = NULL;}
}

void CLFermionSpinSum::extract_spinor_couplings(CEProduct *seq)
{
	dbg->log(30, "extract_spinor_couplings()");
	int i;
	for(i=0; i<(seq->size()-1); i++) {
		CExpression* item = seq->GetAt(i);
		if (item->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) {
			CESpinor* sp = (CESpinor*)item;

			// scan for closing spinor
			int j;
			for(j=(i+1); j<seq->size(); j++) {
				item = seq->GetAt(j);
				if (item->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) {
					CESpinor* sp2 = (CESpinor*)item;
					//extract items from i to j into m_spinor_couplings array

					CExpression* extr = seq->extract_items(i, j); i--;
					if (extr == NULL) {dbg->log(0, "Error: NULL unexpected."); return;}
					m_spinor_couplings->addItem(extr);
					break;
				}
			}
		}
	}
	dbg->log(30, "extract_spinor_couplings() processed: ", m_spinor_couplings);
}

/*
*	Extracts first spinor sequence & calls itself again
*/
bool CLFermionSpinSum::sum_spins_in_selected_couplings()
{
	dbg->log(30, "sum_spins_in_selected_couplings()");
	int s = m_spinor_couplings->size();
	if (s == 0) {
		dbg->log(30, "sum_spins_in_selected_couplings() completed:", m_sum_couplings_result);
		return true;
	}

	CESpinor* right_0 = get_bracket_spinor(0, true);
	if (right_0 == NULL) { dbg->log(0, "Unexpected error. (right,0)->NULL", m_spinor_couplings); return false; }

	int i;
	for(i=0; i<s; i++) {
		CESpinor* left_i = get_bracket_spinor(i, false);
		if (left_i == NULL) {dbg->log(0, i, "Unexpected error. (left,%ld)->NULL", m_spinor_couplings); return false;}

		// compare spinors
		CExpression* spinor_couple = right_0->couple_with_spinor(left_i, true);
		if (spinor_couple != NULL) {
			if (i == 0) {
				// remove first coupling to Sp[]-object
				CEProduct* couple_body = extract_coupling(0, true, true);
				//Nt. We may not(?) use 'right_0' or 'left_i' below!
				couple_body->addItemInline(spinor_couple);

				// filter sequence for gell-mann objects
				CExpression* seq_GM = CLFermionSpinSum::ExtractGellMannObjects(couple_body);

				// add traces of 'couple_body' and 'seq_GM' to sum-result product
				if (m_sum_couplings_result == NULL) m_sum_couplings_result = new CEProduct();
				m_sum_couplings_result->addItemInline(new CFSp(couple_body)); delete couple_body;
				if(seq_GM!=NULL){ m_sum_couplings_result->addItemInline(new CEGellMannTr(seq_GM)); delete seq_GM; }
				return sum_spins_in_selected_couplings();
			} else {
				CEProduct* couple_A = extract_coupling(0, false, true);
				// after A is extracted B moves left by 1
				CEProduct* couple_B = extract_coupling(i-1, true, false);
				//Nt. We may not(?) use 'right_0' or 'left_i' below!
				// create common new coupling
				couple_A->addItemInline(spinor_couple);
				couple_A->addItemInline(couple_B);

				// note: arranging can cause unexpected numerical factor before the coupling,
				// no arrangment is allowed here!
				//couple_A->Arrange(AL_INLINE);
				
				// the position does not matter
				m_spinor_couplings->InsertAtX(0, couple_A); //deletes couple_A
				return sum_spins_in_selected_couplings();
			}
		}
	}
	dbg->log(30, "sum_spins_in_selected_couplings() completed, BUT some of spinor tales have no pair", m_spinor_couplings);
	return true; // some of spinor tales have no pair
}

CESpinor* CLFermionSpinSum::get_bracket_spinor(int i, bool at_right)
{
	if (m_spinor_couplings == NULL) return NULL;
	CExpression* item = m_spinor_couplings->GetAt(i);
	if (item == NULL) return NULL;
	if (!(item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT))) return NULL;
	CEProduct* couple = (CEProduct*)item;

	CESpinor* res = NULL;
	int spinor_pos = (at_right)?(couple->size()-1):0;
	item = couple->GetAt(spinor_pos);
	if (item != NULL) {
		if (item->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) res = (CESpinor*)item;
	}
	return res;
}

CEProduct* CLFermionSpinSum::extract_coupling(int i, bool delete_spinor_left, bool delete_spinor_right)
{
	CExpression* item = m_spinor_couplings->GetAt(i)->clone();
	m_spinor_couplings->RemoveAt(i);
	CEProduct* res = (CEProduct*)item;
	
	if (delete_spinor_left) {
		CESpinor* r = (CESpinor*)res->GetAt(0);// try on spinor
		res->RemoveAt(0);
	}
	
	if (delete_spinor_right) {
		int last_pos = res->size()-1;
		CESpinor* l = (CESpinor*)res->GetAt(last_pos);// try on spinor
		res->RemoveAt(last_pos);
	}
	
	return res;
}

CExpression* CLFermionSpinSum::ExtractGellMannObjects(CEProduct *seq) {
	if(seq==NULL)return NULL;

	CEProduct* res = NULL;
	long s = seq->size();
	for(long i=0; i<s; i++) {
		CExpression* item = seq->GetAt(i);
		if(item==NULL)break; //error
		if(item->ClassName() == CEGellMann::CLASS_NAME) {
			item = seq->RemoveAtX(i); s--; i--; if(item==NULL){ CEException::halt("in CLFermionSpinSum::ExtractGellMannObjects"); }
			if(res==NULL)res=new CEProduct();
			res->addItem(item);
		}
	}
	if(seq->size()==0){ seq->addItem(CENumber::GetUnit()); }
	return res;
}
