////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LFiertzTransform.cpp: implementation of the CLFiertzTransform class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLFiertzTransform::dbg = new CDebug("CLFierzTransform");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CLFiertzTransform::CLFiertzTransform() {}
CLFiertzTransform::~CLFiertzTransform() {}

//////////////////////////////////////////////////////////////////////
// Static entry
//////////////////////////////////////////////////////////////////////
CExpression* CLFiertzTransform::DoFiertzTransform(CExpression* src, DWORD* nApplied) { // source must be reduced!
	if (nApplied!=NULL) (*nApplied)=0;
	if (src==NULL) return NULL;

	dbg->log(25, "DoFierzTransform() src", src);

	// try turn all Traces to normal form
	CExpression* res = NormalizeTraces(src);
	if (res != src) { delete src; src=res; }

	// Fierz transform
	res = DoFTImpl(src, nApplied);
	if (res != src) delete src;
	if (nApplied!=NULL) {
		dbg->log(25, *nApplied, "DoFierzTransform() res (appied %ld times)", res);
	} else {
		dbg->log(25, "DoFierzTransform() res", res);
	}

	return res;
}

//////////////////////////////////////////////////////////////////////
// Normalize Traces
//////////////////////////////////////////////////////////////////////
CExpression* CLFiertzTransform::NormalizeTraces(CExpression* src) {
	// process only numerator sums & no deep objects --> no rules 
	dbg->log(45, "NormalizeTraces(expr)", src);

	if (src==NULL) return NULL; if (!src->getLevel()) return src;
	long s;long i;
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* src_sum=(CESum*)src;
		s=src_sum->size();
		for(i=0;i<s;i++) {
			CExpression* item = src_sum->GetAt(i);
			CExpression* item_res = NormalizeTraces(item);
			if (item_res != item) { src_sum->SetAtX(i,item_res); }
		}
		return src;
	}

	CExpression* res = NULL;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (src->ClassName() == CFSp::CLASS_NAME) {
			res=((CFSp*)src)->ToNormalForm();
			if (res!=NULL) { CExpression* res2=CFSp::TruncateZeroTraces(res,false); if((res!=src)&&(res!=res2))delete res; res=res2; }
			if (res!=NULL) return res;
		}
		return src;
	}

	CEProduct* pr=(CEProduct*)src;
	s=pr->size();
	bool do_reduce=false;
	for(i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i); if (!item->getLevel()) continue;
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			dbg->log(0, "Internal ERROR: non-reduced source. NormalizeTraces skipped", pr); return pr;
		}
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			dbg->log(0, "Internal ERROR: non-reduced source. NormalizeTraces skipped", pr); return pr;
		}
		if (item->ClassName() == CFSp::CLASS_NAME) {
			res = ((CFSp*)item)->ToNormalForm();
			if (res!=NULL) { CExpression* res2=CFSp::TruncateZeroTraces(res,false); if((res!=item)&&(res!=res2))delete res; res=res2; }
			if ((res!=NULL) && (res != item)) { pr->SetAtX(i, res); do_reduce=true; }
		}
	}

	res = pr;
	if (do_reduce) res = CExpression::reduce(res, false);
	dbg->log(45, "NormalizeTraces(expr) result", res);
	return res;
}

//////////////////////////////////////////////////////////////////////
// Implementation
//////////////////////////////////////////////////////////////////////
CExpression* CLFiertzTransform::DoFTImpl(CExpression* src, DWORD* nApplied) { // source parser, source may be changed, but not deleted
	dbg->log(45, "DoFTImpl(expr)", src);
	if (src==NULL) return NULL;
	if (!src->getLevel()) { return src; }

	long s;long i;
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* src_sum=(CESum*)src;
		s=src_sum->size();
		for(i=0;i<s;i++) {
			CExpression* item = src_sum->GetAt(i);
			CExpression* item_res = DoFTImpl(item, nApplied);
			if (item_res != item) { src_sum->SetAtX(i,item_res); }
		}
		return src;
	}

	if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return src;

	CEProduct* pr=(CEProduct*)src;
	s=pr->size();

	// 1. Check product
	int _TrNum = 0;
	for(i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		if (!item->getLevel()) continue;
		if (item->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			dbg->log(0, "Internal ERROR: non-reduced source. DoFierzTransform skipped", pr); return pr;
		}
		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			dbg->log(0, "Internal ERROR: non-reduced source. DoFierzTransform skipped", pr); return pr;
		}
		if (item->ClassName() == CFSp::CLASS_NAME) {
			if (((CFSp*)item)->IsOfNormalForm()) {
				_TrNum++;
			} else {
				dbg->log(0, "Internal WARNING: Tr[...] of unexpected expression. Fierz transform is skipped for it", item);
			}
		}
	}
	dbg->log(75, _TrNum, "DoFTImpl() %ld normalized traces found");
	if (_TrNum <= 1) return pr;

	// 1.1. Expand Gamma-5 and 1 --> G+ and G-
	for(i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i);
		if (!item->getLevel()) continue;
		if (item->ClassName() != CFSp::CLASS_NAME) continue;
		if (!((CFSp*)item)->IsOfNormalForm()) continue;
		CExpression* sub_res = ((CFSp*)item)->ExpandLastGammaX(true); // item or SUM(Sp,Sp,Sp..)
		if (sub_res != item) { // ==item means gP or gM is last item
			if (sub_res == NULL) CEException::halt("Fierz transform failed (point #3)");
			dbg->log(55, "Last Gamma/Unit expanded: ", pr, sub_res);
			pr->SetAtX(i,sub_res);
			sub_res=CExpression::reduce(pr, false);
			CExpression* res = DoFTImpl(sub_res, nApplied);
			if ((sub_res != src) && (sub_res != res)) { delete sub_res; }
			return res;
		}
	}
	dbg->log(75, "DoFTImpl() GP/M detected");

	// 2. Couple Free Metrics to Sp[] objects, recursive
	for(i=0; i<s; i++) {
		CExpression* item = pr->GetAt(i); if (!item->getLevel()) continue;
		if (item->ClassName() == CEMetric::CLASS_NAME) {
			CEMetric* metr = (CEMetric*)item;
			bool MetricCoupled=true;
			for(long j=0; j<s; j++) {
				CExpression* item_j = pr->GetAt(j); if (!item_j->getLevel()) continue;
				if (item_j->ClassName() == CFSp::CLASS_NAME) {
					if (((CFSp*)item_j)->CoupleGammaToMetric(metr)) { MetricCoupled=true; break; }
				}
			}
			if (MetricCoupled) {
				// remove metric & start loop from first position
				pr->RemoveAt(i); s--; i=-1; // to be 0 at next loop
			}
		}
	}
	dbg->log(75, "DoFTImpl() metrics coupled");

	// To-delete now: nothing
	// Normal-formed Tr-objects have gP or gM now
	// 3. Find Sp-Sp couplings and replace by Fierz-results
	for(i=0; i<(s-1); i++) {
		CExpression* item = pr->GetAt(i); if (!item->getLevel()) continue;
		if (item->ClassName() != CFSp::CLASS_NAME) continue;
		CFSp* sp1 = (CFSp*)item;
		if (!sp1->IsOfNormalForm()) continue;
		CEPtrArray* inds = sp1->GetTIndicesList(); // to delete inds
		if (inds==NULL) continue;
		long inds_num=inds->size();
		if (inds_num==0) { delete inds; continue; }
		for(long j=i+1; j<s; j++) {
			CExpression* itemJ = pr->GetAt(j); if (!itemJ->getLevel()) continue;
			if (itemJ->ClassName() != CFSp::CLASS_NAME) continue;
			CFSp* sp2 = (CFSp*)itemJ;
			if (!sp2->IsOfNormalForm()) continue;
			for(long ind_no=0; ind_no<inds_num; ind_no++) {
				if (inds->GetAt(ind_no) != NULL) {
					CTIndex* ti = (CTIndex*)inds->GetAt(ind_no);
					int ind_no2 = sp2->FindGammaCoupleForIndex(ti->index(),ti->covariance());
					if (ind_no2>=0) {
						CExpression* sub_res = DoFTImplCore(sp1,ind_no,sp2,ind_no2);
						if (sub_res != NULL) {
							if (nApplied!=NULL) (*nApplied)++;
							delete inds;
							//(debug:)if ((nApplied!=NULL)&&(*nApplied != 1)) { delete sub_res; return pr; }
							pr->RemoveAt(j);
							pr->SetAtX(i,sub_res);
							sub_res = CExpression::reduce(pr, false);
							dbg->log(45, "Fierz transform performed", sub_res);
							CExpression* res = DoFTImpl(sub_res, nApplied);
							if ((sub_res != src) && (sub_res != res)) { delete sub_res; }
							return res;
						}
					}
				}
			}
		}
		delete inds;
	}
	return pr;
}

CExpression* CLFiertzTransform::DoFTImplCore(CFSp* sp1, int i1, CFSp* sp2, int i2) { // does not deletes input
	dbg->log(75, "DoFTImplCore()", sp1, sp2);
	dbg->log(75, i1, i2, "DoFTImplCore(i1: %ld, i2: %ld)");
	// 1. Check num of gammas: for mod-2 and unexpected objects
	// 2. Check Gamma-5/+/- are at final pos
	// Unexpected-objects & last-item G5/+/- check is already done in DoFTImpl()
	// ...if (((CFSp*)item)->IsOfNormalForm()) {...
	int size_1 = sp1->CalcGammaNum(); int size_2 = sp2->CalcGammaNum();
	dbg->log(105, size_1, size_2, "DoFTImplCore() size_1: %ld, size_2: %ld");
	if ((size_1<0) || (size_2<0)) {
		dbg->log(75, "DoFTImplCore() res: NULL");
		return NULL;
	}
	if ((size_1%2 != 0) || (size_2%2 != 0)) {
		dbg->log(75, "DoFTImplCore() res: ZERO");
		return CENumber::GetZero();
	}
	if ((size_1<2) || (size_2<2)) {
		dbg->log(75, "DoFTImplCore() res: ZERO");
		return CENumber::GetZero();
	}

	// 3. Turn i1 & i2 items to zero pos
	if (!sp1->SwapToGamma(i1)) { dbg->log(0, i1, "Internal ERROR: sp1->SwapToGamma(%ld). NULL result", sp1); return NULL; }
	if (!sp2->SwapToGamma(i2)) { dbg->log(0, i2, "Internal ERROR: sp2->SwapToGamma(%ld). NULL result", sp2); return NULL; }

	// 4. Last-gamma IDs
	int last1 = sp1->GetLastGammaXID();
	if ((last1 == 0) || (last1==INT_MIN)) { dbg->log(0, last1, "Internal ERROR: sp1->GetLastGammaXID() unexpected result %ld ", sp1); return NULL; }
	int last2 = sp2->GetLastGammaXID();
	if ((last2 == 0) || (last2==INT_MIN)) { dbg->log(0, last2, "Internal ERROR: sp2->GetLastGammaXID() unexpected result %ld ", sp2); return NULL; }

	// 5. Apply Fierz formula
	CExpression* res = DoFTFormula(sp1, (last1>0), sp2, (last2>0));
	dbg->log(75, "DoFTImplCore() result", res);
	return res;
}

CExpression* CLFiertzTransform::DoFTFormula(CFSp* sp1, bool gP1, CFSp* sp2, bool gP2) { // does not affects input
	
	// sp1: Tr[G_mu a_1...a_n G_pm], sp2: Tr[G^mu b_1...b_n G_pm]
	//dbg->log(45, gP1, "DoFTFormula: %ld", sp1); dbg->log(45, gP2, "DoFTFormula: %ld", sp2);

	CExpression* a1_n = NULL;
	CExpression* b1_n = sp2->RetrieveSrcSeq(true, true, true);

	if (CEUtil::LEQU(gP1,gP2)) {
		a1_n = sp1->RetrieveSrcSeq(false, true, true);
	} else {
		a1_n = sp1->RetrieveSrcSeq(true, true, true);
	}
	
	if ((a1_n == NULL) || (b1_n == NULL)) { dbg->log(0, "Internal ERROR in Fierz transform", sp1, sp2); return NULL; }

	CExpression* res_seq = a1_n->multiply(b1_n)->multiply(new CEGammaSpecial( (gP2?1:(-1)) ));
	CExpression* res = new CFSp(res_seq); delete res_seq;
	res = res->multiply(CENumber::GetTwo());

	if (!sp1->getFlag()) res->inverseFlag();
	if (!sp2->getFlag()) res->inverseFlag();
	//dbg->log(45, "DoFTFormula: res", res);
	return res;
}
