////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LGammaSeqSimplify.cpp: implementation of the CLGammaSeqSimplify class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLGammaSeqSimplify::dbg = new CDebug("CLGammaSeqSimplify");

CKinematics* CLGammaSeqSimplify::m_Kin=NULL;// updated at every simplify() call
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CLGammaSeqSimplify::CLGammaSeqSimplify() {}
CLGammaSeqSimplify::~CLGammaSeqSimplify() {}

/*
*	///////////////////////////////////////////////////// interface entry
*	reduces source!
*	deletes source
*/
CExpression* CLGammaSeqSimplify::simplify(CExpression *source) {
	dbg->log(11, "CLGammaSeqSimplify::simplify() source", source);
	if (source == NULL) return NULL;

	// do some simplification even for NULL kinematics
	m_Kin = CEProcessor::GetKinIfDeclared();

	source = CExpression::reduce(source);

	CExpression* res = doSimplifyGS(source);
	if (res != source) delete source;

	res = CExpression::reduce(res);
	dbg->log(21, "CLGammaSeqSimplify::simplify() result", res);
	return res;
}

///////////////////////////////////////////////////// private implementation

bool CLGammaSeqSimplify::replacePairs(CEProduct *source) {
	dbg->log(71, "replacePairs()");
	//dbg->log(171, "replacePairs()", source);

	bool res = false;
	long fmp = source->getFirstMatrixPos();
	if ((fmp >= 0) && (source->size()>(1+fmp))) {
		//dbg->log(171, fmp, "fmp=%ld");

		CExpression* expr_a = source->GetAt(fmp);
		CExpression* expr_b = NULL;
		long i;
		for (i=fmp+1; i<(source->size()); i++) {
			expr_b = source->GetAt(i);
			//dbg->log(171, i, "i=%ld, expr_b", expr_b);

			bool pair_replaced = false;
			if ((expr_a->getProperty(CEProperties::PROP_MASK_IS_TENSOR))
				&& (expr_b->getProperty(CEProperties::PROP_MASK_IS_TENSOR))) {
				CExpression* unite_res = (IETensor::GetIETensor(expr_a))->unite_with(IETensor::GetIETensor(expr_b), true);
				//dbg->log(171, "unite_res", unite_res);
				if (unite_res != NULL) {
					source->SetAt(i-1, unite_res);
					delete unite_res;
					source->RemoveAt(i);
					pair_replaced = true;
				}
			}

			if (pair_replaced) {
				res = true;
				expr_a = source->GetAt(i);
			}
			else {
				expr_a = expr_b;
			}
		}
	}
	return res;
}

bool CLGammaSeqSimplify::excangePair(CEProduct* source, long first_pos) {
	dbg->log(71, first_pos, "excangePair(%ld)", source);

	long size = source->size();
	if ((first_pos < source->getFirstMatrixPos()) || ((first_pos+1) >= size)) {
		dbg->log(0, first_pos, "ERROR in excangePair(%ld): invalid position", source);
		return false;
	}
	CExpression* expr_a = source->GetAt(first_pos);
	CExpression* expr_b = source->GetAt(first_pos+1);
	CExpression* subst = getPairSubst(expr_a, expr_b);
	if (subst != NULL) {
		source->SetAt(first_pos, subst);
		delete subst;
		source->RemoveAt(first_pos+1);

		//dbg->log(171, first_pos, "excangePair(%ld) done", source);
		return true;
	}
	return false;
}

/*
*	source is not deleted.
*/
CExpression* CLGammaSeqSimplify::doSimplifyGS(CExpression *source) {
	dbg->log(31, "doSimplifyGS()", source);

	CEProperties* props = source->getProperties();
	if (props->isSum()) {
		CESum* _sum = (CESum*)source;
		long _size = _sum->size();

		//	*****************************************************	<scroll>
		CEProgressWatch* pw = NULL;
		if (_size > 20) { pw = CEProcessor::newProgressWatch("Gamma Sequence Simplification (%ld items)", _size); }
		//	*****************************************************	<\scroll>
		long i;
		for (i=0; i<_size; i++) {
			CExpression* item = _sum->GetAt(i);
			CExpression* sub_res = doSimplifyGS(item);
			if (sub_res != item) {
				_sum->SetAt(i, sub_res);
				delete sub_res;
			}
			//	*****************************************************	<scroll>
			if (pw != NULL) pw->step(i+1, _size);
			//	*****************************************************	<\scroll>
		}
		//	*****************************************************	<scroll>
		if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
		//	*****************************************************	<\scroll>
	}
	else if (props->isProduct()) {
		CEProduct* seq = (CEProduct*)source;
		if (expandPairs(seq)) {
			dbg->log(61, "expandPairs() done, next iteration");
			//dbg->log(161, "intermadiate result (not reduced): ", seq);
			CExpression* src_int = CExpression::reduce(seq, false);
			source = doSimplifyGS(src_int);
			if ((src_int != source) && (src_int != seq)) delete src_int;
			//delete seq;
		}
	}
	delete props;
	dbg->log(61, "doSimplifyGS(): result", source);
	return source;
}

/*
*	main simplification procedure
*/
bool CLGammaSeqSimplify::expandPairs(CEProduct *source) {
	dbg->log(41, "expandPairs:", source);
	if (replacePairs(source)) return true; // try replace adjacent pairs

	dbg->log(61, "expandPairs: no pairs replaced, try to move non-commutatives");
	int fmp = source->getFirstMatrixPos();
	bool err_break = false;
	if ((fmp >= 0) && (source->size()>(1+fmp))) {
		//moving left.
		long s = source->size();
		long i;
		for (i = fmp+1; i<s; i++) {
			if (is_to_move_left(i, source)) {
				dbg->log(71, "moving item left");
				if (excangePair(source, i-1)) return true;
				//can't exchange pair stop moving left.
				dbg->log(0, i-1, "Warning (#1): cant'exchange non-commutative pair at pos %ld in", source);
				dbg->log(2, "Left: ", source->GetAt(i-1));
				dbg->log(2, "Right: ", source->GetAt(i));
				err_break = true; break;
			}
		}
		//moving right.
		if (!err_break) for (i = s-2; i>=fmp; i--) {
			if (is_to_move_right(i, source)) {
				dbg->log(71, "moving item right");
				if (excangePair(source, i)) return true;
				//stop moving right.
				dbg->log(0, i, "Warning (#2): cant'exchange non-commutative pair at pos %ld in", source);
				dbg->log(2, "Left: ", source->GetAt(i));
				dbg->log(2, "Right: ", source->GetAt(i+1));
				err_break = true; break;
			}
		}

		//replasing kinematically-dependent vectors
		//if there is nothing to move.
		// this engine is not used currently: KINEMATICS->isPrimary() returns permanent "true";
		if (m_Kin != NULL) {
			if (!err_break) for (i = fmp; i<s; i++) {
				CExpression* expr = source->GetAt(i);
				if (CEUtil::equals(expr->ClassName(), CESGammaProduct::CLASS_NAME)) {
					CExpression* base = ((CESGammaProduct*)expr)->getBaseExpression();
					if (!m_Kin->isPrimary(base)) {
						CExpression* subst_base = m_Kin->getEqual(base);
						if (subst_base != NULL) {
							CExpression* subst = new CESGammaProduct(subst_base);
							delete subst_base;
							source->SetAt(i, subst);
							delete subst;
							return true;
						}
					}
				}
			}
		}
	}
	if (err_break) dbg->log(2, "expandPairs() : Error have been occured");
	return false;
}

/*
* ???
* Gamma-5 object suspected at expr_b only
* They shold be moved to the last position of sequence.
*/
CExpression* CLGammaSeqSimplify::getPairSubst(CExpression *expr_a, CExpression *expr_b) {
	if ((expr_a->getProperty(CEProperties::PROP_MASK_IS_TENSOR)) &&
		(expr_b->getProperty(CEProperties::PROP_MASK_IS_TENSOR))) {
			return (IETensor::GetIETensor(expr_a))->exchange_subst(IETensor::GetIETensor(expr_b),true);
	}
	/*
	*	Old implementation:
	*	if (expr_a->getProperty(CEProperties::PROP_MASK_IS_TENSOR)) return (IETensor::GetIETensor(expr_a))->exchange_subst(expr_b,true);
	*	else if (expr_b->getProperty(CEProperties::PROP_MASK_IS_TENSOR)) return (IETensor::GetIETensor(expr_b))->exchange_subst(expr_a, false);
	*/
	return NULL;
}

bool CLGammaSeqSimplify::is_to_move_left(long i, CEProduct *source) {
	dbg->log(71, i, "is_to_move_left(%ld)", source);

	CExpression* expr = source->GetAt(i);
	if (expr->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) return false; // do not move spinors
	bool is_tensor = expr->getProperty(CEProperties::PROP_MASK_IS_TENSOR); // cache tensor check

	long f = source->getFirstMatrixPos();
	if (f<0) return false;
	long j;
	for (j=i-1; j>=f; j--) {
		CExpression* other = source->GetAt(j);
		if (expr->equals(other)) return true;//?

		if ((is_tensor) && (other->getProperty(CEProperties::PROP_MASK_IS_TENSOR))) {
			CExpression* temp = (IETensor::GetIETensor(expr))->unite_with(IETensor::GetIETensor(other), false);
			if (temp != NULL) {
				delete temp;
				return true;
			}
			
			// check if they can be exchanged
			temp = (IETensor::GetIETensor(expr))->exchange_subst(IETensor::GetIETensor(other), false);
			if (temp == NULL) return false;
			delete temp;
		}
	}
	return false;
}

bool CLGammaSeqSimplify::is_to_move_right(long i, CEProduct *source) {
	dbg->log(71, i, "is_to_move_right(%ld)", source);

	CExpression* expr = source->GetAt(i);
	if (expr->getProperty(CEProperties::PROP_MASK_IS_SPINOR)) return false; // do not move spinors

	// every GAMMA\\_{5/+/-} should be moved right unless it is followed by spinor
	if (CEUtil::equals(expr->ClassName(), CEGammaSpecial::CLASS_NAME)) {
		CExpression* next = source->GetAt(i+1);
		if ((next == NULL) || (next->getProperty(CEProperties::PROP_MASK_IS_SPINOR))) return false;
		return true;
	}

	long s = source->size();
	long j;
	for (j=i+1; j<s; j++) {
		CExpression* other = source->GetAt(j);
		if (expr->equals(other)) return true; // they can certainly be united
		if ((expr->getProperty(CEProperties::PROP_MASK_IS_TENSOR))
			&& (other->getProperty(CEProperties::PROP_MASK_IS_TENSOR))) {
			CExpression* temp = (IETensor::GetIETensor(expr))->unite_with(IETensor::GetIETensor(other), true);
			if (temp != NULL) {
				delete temp;
				return true;
			}

			// check if they can be exchanged
			temp = (IETensor::GetIETensor(expr))->exchange_subst(IETensor::GetIETensor(other), true);
			if (temp == NULL) return false;
			delete temp;
		}
	}
	return false;
}
