////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LHEPBasics.cpp: implementation of the CLHEPBasics class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CLHEPBasics::dbg = new CDebug("CLHEPBasics");
CSequence* CLHEPBasics::m_BookedSymsList = NULL;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CLHEPBasics::CLHEPBasics() : CELibrary() { }
CLHEPBasics::~CLHEPBasics() { }

void CLHEPBasics::Finalize() {
	if (m_BookedSymsList != NULL) { delete m_BookedSymsList; m_BookedSymsList=NULL; }
	if (m_ConstantsList != NULL) { delete m_ConstantsList; m_ConstantsList=NULL; }
}

//////////////////////////////////////////////////////////////////////
// Statics
//////////////////////////////////////////////////////////////////////
CERule* CLHEPBasics::get_mass_order_rule() {
	dbg->log(35, "get_mass_order_rule()");

	CEPtrArray* mass_order_info_arr = CEProcessor::getEnvironment()->GetMassOrderInfo();
	if (mass_order_info_arr == NULL) return NULL;// EXACT_MASS flag is set.
	long s = mass_order_info_arr->size();
	if (s == 0) return NULL;
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if (phys == NULL) return NULL;

	CERulesList* res = new CERulesList();

	long i;
	for(i=0; i<s; i++) {
		CExprContainer* cart = (CExprContainer*)mass_order_info_arr->GetAt(i);
		int part_kind = cart->get_info(0);
		int mass_order = cart->get_info(1);
		if (mass_order < 0) continue;

		if (!phys->validate_part_kind(part_kind)) {
			// Note: this may occur if one sets SetMassCalcOrder(switched off pk)
			dbg->log(0,part_kind,"ERROR: %x");
			continue;
		}
		
		CExpression* mass_sym = phys->get_mass_sym(part_kind);
		if (mass_sym == NULL) continue;
		CERule* r = NULL;
		if (mass_order == 0) {
			CENumber zero((long)0);
			r = new CEReplasement(mass_sym, &zero, false, CEReplasement::REPL_AVOID_NULL_DENOMINATORS);
		}
		else {
			r = new CEPowerReplRule(mass_sym, mass_order);
		}
		res->addRule(r);
	}
	if (res->getRulesNum() == 0) { delete res; res = NULL; }
	dbg->log(25, "Mass Order Rule composed:", res);
	return res;
}

const int CLHEPBasics::MO_EXPAND	= 0x01; // reduces or arranges result
const int CLHEPBasics::MO_NO_ZERO	= 0x02; // cancels appying if zero result
CExpression* CLHEPBasics::ApplyMassOrder(CExpression* source, int settings) {
	if (source == NULL) return NULL;
	CERule* r_mass = get_mass_order_rule();
	if (r_mass == NULL) return source;

	CExpression* src_backup = NULL;
	if (settings & MO_NO_ZERO) { src_backup = source->clone(); }

	source = r_mass->ApplyTo(source);
	delete r_mass;
	
	if (settings & MO_EXPAND) source = CExpression::reduce(source);
	else source = CExpression::arrge(source);

	if (settings & MO_NO_ZERO) {
		if (CEBasic::isZero(source)) { // cances rule
			delete source; source = src_backup; src_backup = NULL;
		}
	}

	if (src_backup != NULL) delete src_backup;
	return source;
}

/*
*	KinArrange:
*	1. Replace similar couplings.
*	2. Replace couplings by Mandelstam variables.
*	3. Reduce result.
*	4. Keep the mass according to get_mass_order_rule() rule.
*	Notes:
*	1. Large expressions are greatly reduced by applying mass rule to the source.
*/
const int CLHEPBasics::KA_MASS_EXACT				= 0x01;
const int CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS	= 0x02;
const int CLHEPBasics::KA_NO_EXPAND					= 0x04;
CExpression* CLHEPBasics::KinArrange(CExpression* source, int settings) {
	dbg->log(15, "KinArrange():", source);
	dbg->reset_periods(15);
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) { dbg->log(0, "Error: No kinematics installed."); throw(new CEException(CEException::INT_ERROR)); }

	CEReplList* r_kin = kin->getReplacements(CKinematics::REPL_COUPLINGS_SIMPLIFY);
	if ((r_kin != NULL) && (r_kin->size() == 0)) { delete r_kin; r_kin = NULL; }

	//CEReplList* r_bc = kin->getBasicCouplingsValues(false);
	//if ((r_bc != NULL) && (r_bc->size() == 0)) { delete r_bc; r_bc = NULL; }

	CEReplList* r_mand_vars = NULL;
	if (settings & KA_INVOLVE_MANDELSTAM_VARS) r_mand_vars = kin->getReplacements(CKinematics::REPL_MANDELSTAM_VARIABLES);
	if ((r_mand_vars != NULL) && (r_mand_vars->size() == 0)) { delete r_mand_vars; r_mand_vars = NULL; }

	CERule* r_mass = NULL;
	if ((settings & KA_MASS_EXACT) == 0) {
		r_mass = get_mass_order_rule();
	}

	// apply mass rule to the source
	if (r_mass != NULL) { source = r_mass->ApplyTo(source); dbg->log(75, "KinArrange() mass rule applied:", source); }

	// 1. first change similar couplings
	if (r_kin != NULL) { source = r_kin->ApplyTo(source); delete r_kin; }
	//if (r_bc != NULL) { source = r_bc->ApplyTo(source); delete r_bc; }

	// at least one reduction should be performed
	if (settings & KA_NO_EXPAND) source = CExpression::arrge(source);
	else source = CExpression::reduce(source);

	// 2. involve mandelstam variables
	if (r_mand_vars != NULL) {
		source = r_mand_vars->ApplyTo(source); delete r_mand_vars;
		if (r_mass != NULL) source = r_mass->ApplyTo(source);

		if (settings & KA_NO_EXPAND) source = CExpression::arrge(source);
		else source = CExpression::reduce(source);
	}
	
	// apply mass rule to the result
	if (r_mass != NULL) source = CExpression::arrge(r_mass->ApplyTo(source));
	
	if (r_mass != NULL) delete r_mass;
	dbg->log(15, "KinArrange() result:", source);

	dbg->logperiod(15);
	return source;
}

const int CLHEPBasics::KS_MASS_EXACT					= 0x001;
const int CLHEPBasics::KS_INVOLVE_MANDELSTAM_VARS		= 0x002;
const int CLHEPBasics::KS_NO_EXPAND						= 0x004;
const int CLHEPBasics::KS_FACTORIZE_FUNCTIONS			= 0x010;
const int CLHEPBasics::KS_FACTORIZE_DENOMINATORS		= 0x020;
const int CLHEPBasics::KS_FACTORIZE_ALL_DENOMINATORS	= 0x040;
const int CLHEPBasics::KS_FACTORIZE						= 0x070;
const int CLHEPBasics::KS_MOMENTA_SUBST					= 0x100;
const int CLHEPBasics::KS_NO_COUPLINGS_SUBST			= 0x200;
const int CLHEPBasics::KS_FAST							= 0x400; //simplify small sub-objects only

CSequence* CLHEPBasics::m_ConstantsList = NULL;
void CLHEPBasics::InitConstantsList() { // fill 'm_ConstantsList' if NULL
	if(m_ConstantsList==NULL){
		m_ConstantsList=new CSequence();
		m_ConstantsList->addItem(CVIScalarCalcLib::getUVSymbol()->clone());
		m_ConstantsList->addItem(CVIScalarCalcLib::getIRSymbol()->clone());
		CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
		if(phys!=NULL){ m_ConstantsList->addSeq(phys->CollectConstants()); }
	}
}

/*
*	Source is expected to be already expanded?
*
*	1. Factorize denominators.
*	2. Partial factorize functions (VI, ect.).
*	3. Tries to simplify sums by applying all kinematical relations.
*/
CExpression* CLHEPBasics::KinSimplify(CExpression* source, int settings) {
	dbg->log(15, settings, "KinSimplify(%ld):", source);
	dbg->reset_periods(15);

	if(source==NULL)return source;
	CKinematics* kin=CEProcessor::kinematics(); if(kin==NULL){ dbg->log(0,"<ERROR in KinSimplify()> No kinematics installed"); return source; }
	bool bFast=((settings & CLHEPBasics::KS_FAST)==CLHEPBasics::KS_FAST);

	int pf_flags = 0; // flags for partial factorization
	if ((settings & KS_FACTORIZE) && (!bFast)) {
		//asdfg: Source is expected to be already expanded?
		// source = CExpression::reduce(source); // first expand source to sum

		if(m_ConstantsList==NULL)InitConstantsList();

		pf_flags = 0;
		if(settings & KS_FACTORIZE_FUNCTIONS){ pf_flags |= CViewSimplify::PART_FACT_FUNCTIONS; }
		if(settings & KS_FACTORIZE_ALL_DENOMINATORS){ pf_flags|=(CViewSimplify::PART_FACT_DENOMINATORS|CViewSimplify::PART_FACT_ALL_DENOMINATORS); }
		else if(settings & KS_FACTORIZE_DENOMINATORS){ pf_flags|=(CViewSimplify::PART_FACT_DENOMINATORS|CViewSimplify::PART_FACT_ALL_SINGLE_DENS); }
		source = CViewSimplify::factorize_partial(source, pf_flags, m_ConstantsList);
		dbg->log(30, settings, "KinSimplify(%ld) facotrized:", source);
		dbg->logperiod(30);
	}

	// Get kinematical relations
	int all_repl_type=0;
	//CKinematics::REPL_ALL_KIN_DEPENDENCIES, REPL_ALL_MOMENTA_SUBST, REPL_ALL_MANDELSTAM_X, REPL_ALL_COUPLINGS_X
	if((settings & KS_NO_COUPLINGS_SUBST) == 0){ all_repl_type|=CKinematics::REPL_ALL_COUPLINGS_X; }
	if(settings & KS_INVOLVE_MANDELSTAM_VARS){ all_repl_type=CKinematics::REPL_ALL_MANDELSTAM_X; }//play Mandelstams only
	if(settings & KS_MOMENTA_SUBST){ all_repl_type|=CKinematics::REPL_ALL_MOMENTA_SUBST; }
	CEReplList* r_simpl_rules=kin->getReplacements(all_repl_type); if(r_simpl_rules==NULL){ dbg->log(0, "Error: no \"replace all\" rules found in kinematics."); CEException::halt(); }

	// Get 'environment' rules (p2k2->p1pk, m^2->0 etc.)
	CERulesList r_env; CERule* r=NULL;
	if((settings&KS_NO_COUPLINGS_SUBST)==0){ r=kin->getReplacements(CKinematics::REPL_COUPLINGS_SIMPLIFY); if(r!=NULL)r_env.combine_with(r); }
	if(settings&KS_INVOLVE_MANDELSTAM_VARS){ r=kin->getReplacements(CKinematics::REPL_MANDELSTAM_VARIABLES); if(r!=NULL)r_env.combine_with(r); }
	if((settings&KS_MASS_EXACT)==0){
		r=get_mass_order_rule();
		if(r!=NULL){ // pre-simplify r_simpl_rules rules by masses->0 shifts
			if ((r_simpl_rules != NULL) && ((r_simpl_rules->GetRuleKind() & CERule::RULE_COMPOUND) != 0)) {
				CERulesList* rl=(CERulesList*)r_simpl_rules; int rl_num=rl->getRulesNum();
				for(int i=0; i<rl_num; i++) {
					CERule* r_i = rl->getRule(i);
					if(CEUtil::equals(r_i->ClassName(), CEReplasement::CLASS_NAME)) {
						CEReplasement* repl_i=(CEReplasement*)r_i;
						repl_i->set_b(CExpression::reduce(r->ApplyTo(repl_i->b()->clone())),false);
					}
				}
			}
			r_env.combine_with(r);
		}
	}

	// Simplify
	source = CSubstSimplify::SimpRepl(source, r_simpl_rules, ((r_env.getRulesNum()>0)?(&r_env):NULL), 1, bFast);
	delete r_simpl_rules;

	dbg->log(30, "KinSimplify() replacements simplification result:", source);
	dbg->logperiod(30);

	if ((settings & KS_FACTORIZE) && (!bFast)) {
		//first expand source to sum
		if(settings & KS_NO_EXPAND){ source=CExpression::arrge(source); }else{ source=CExpression::reduce(source); }
		source = CViewSimplify::factorize_partial(source, pf_flags, m_ConstantsList);
	}

	dbg->log(15, "KinSimplify() result:", source);
	dbg->logperiod(15);
	return source;
}

/*
*	Avoid last vector in kinematics ==> uniform vectors set
*/
CExpression* CLHEPBasics::KinToUniformVectors(CExpression* source) {
	CKinematics* kin = CEProcessor::kinematics(); if (kin == NULL) { dbg->log(0, "<ERROR> No kinematics installed"); return source; }

	int f = kin->num_final_particles();
	CEParticle* p_last = kin->pfin(f-1);
	if (p_last == NULL) { CEException::halt("CLHEPBasics::KinToUniformVectors: NULL partincle found"); }

	CExpression* sym = p_last->get_momentum_symbol();
	CExpression* subst = kin->getEqual(sym);
	
	// change vector
	source = CEReplasement::Replace(source, sym, subst);
	delete subst;
	source = CExpression::arrge(source);
	return source;
}

/*
*	Used by CMESquareLib engine.
*	Works with ME inline.
*	NOTE: work with the specific ME only and fails with others!!!!
*	TO BE RE-IMPLEMENTED BEFORE FUTURE USE.
*
*	1. For proper implementation the "internal momenta - particle kind" information
*	should ba attached to ME structure.
*	
*	This works if the virtual momenta are the photons
*	and no more virtual photons persist in diagram.
*	
*	The another question: should we empty the ME value if no IR-divergence persist?
*	If yes all the calls of this function should be changed.
*/
void CLHEPBasics::GetIRLimit(CMatrixElement* source) {
	dbg->log(15, "GetIRLimit() source:", source);
	if (source == NULL) return;
	CEPtrArray* virts_list = source->get_virt_momenta_list();
	long s = virts_list->size();
	if (s == 0) return;

	CENumber zero((long)0);
	CExpression* me_val = source->get_expr_val()->clone();
	long i;
	for (i=0; i<s; i++) {
		CExpression* v_sym = (CExpression*)virts_list->GetAt(i);
		// set it to zero in every ME factor except the IR-singular denominator.
		
		CEPtrArray* data_arr = new CEPtrArray();
		data_arr->add(new CEReplasement(v_sym, &zero, false));
		data_arr->add(zero.clone());
		CERule* IR_rule = new CESocketRule((void*)CLHEPBasics::IRLimitRuleProc, data_arr);
		delete data_arr;
		me_val = IR_rule->ApplyTo(me_val);
		delete IR_rule;
	}
	source->set_expr_val(me_val);
	dbg->log(15, "GetIRLimit() result:", source);
}

/*
*	Internal static use by GetIRLimit() implementation.
*/
CExpression* CLHEPBasics::IRLimitRuleProc(CExpression* src, CEPtrArray* data_arr) {
	dbg->log(50, "IRLimitRuleProc():", src);
	if ((src == NULL) || (data_arr == NULL)) return src;
	bool src_lev = src->getLevel();

	CExpression* res = NULL;
	if ((src->getProperty(CEProperties::PROP_MASK_IS_SUM)) ||
		(src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT))) {

		CERule* IR_rule = new CESocketRule((void*)CLHEPBasics::IRLimitRuleProc, data_arr);
		CExpression* src_cpy = src->clone();
		res = src_cpy->applyRule(IR_rule);
		if (res != src_cpy) delete src_cpy;
		delete IR_rule;
		dbg->log(90, "IRLimitRuleProc() recursive call result:", res);
	} else {
		CERule* r = (CERule*)data_arr->GetAt(0);
		res = r->ApplyTo(src->clone());
	}

	res = CExpression::arrge(res);
	if (!src_lev) {
		CExpression* res_reduced = KinArrange(res->clone(), KA_MASS_EXACT);
		res_reduced->inverseLevel(); // inverse level back to compare result with ZERO.
		res_reduced = CExpression::arrge(res_reduced);

		CExpression* zero = (CExpression*)data_arr->GetAt(1);
		bool zero_result = (zero->equal_or_opposite(res_reduced));
		delete res_reduced;

		if (zero_result) {
			delete res;
			dbg->log(90, "IRLimitRuleProc() returns source back", src);
			return src->clone();// return cloned value for no subitems to be processed.
		}
	}

	dbg->log(90, "IRLimitRuleProc() result:", res);
	return res;
}

/*
*	x B_0(t, o, m) ==> -1, ect.
*	Destroys the source.
*/
CExpression* CLHEPBasics::SingularArrange(CExpression* source) {
	if (source == NULL) { dbg->log(0, "SingularArrange(): NULL source"); return NULL; }
	dbg->log(15, "SingularArrange():", source);

	// expand source
	source = CExpression::reduce(source);
	source = TruncateDeltaPower(source,2); //remove Eta^2 powers

	bool Ndim = CETensor::isNDimensionSpace();
	if (!Ndim) CETensor::setNDimensionSpace(true);

	CExpression* dimDeltaSym = CETensor::getDimensionDelta();
	CExpression* singSym = CVIScalarCalcLib::getUVSymbol();
	CExpression* singIRSym = CVIScalarCalcLib::getIRSymbol();
	if ((dimDeltaSym == NULL) || (singSym == NULL)) { CEException::halt(dbg, "SingularArrange(): NULL dimDeltaSym or sing-sym found"); }

	CExpression* res = SingularArrangeImpl(source, dimDeltaSym, singSym, singIRSym);
	if (res != source) delete source;

	if (!Ndim) CETensor::setNDimensionSpace(false);
	dbg->log(15, "SingularArrange() result:", res);
	return res;
}

CExpression* CLHEPBasics::GetUVTerm(CExpression* source) {
	if (source == NULL) { dbg->log(0, "GetUVTerm(): NULL source"); return NULL; }
	dbg->log(15, "GetUVTerm():", source);

	bool Ndim = CETensor::isNDimensionSpace();
	if (!Ndim) CETensor::setNDimensionSpace(true); // to access "dimDeltaSym"
	CExpression* dimDeltaSym = CETensor::getDimensionDelta();
	if (dimDeltaSym == NULL) { CEException::halt(dbg, "GetUVTerm(): NULL dimDeltaSym found"); }
	if (!Ndim) CETensor::setNDimensionSpace(false);

	// expand source
	source = source->multiply(dimDeltaSym->clone());
	source = SingularArrange(source);
	source = source->divide(dimDeltaSym->clone());
	source = CExpression::reduce(source);

	dbg->log(15, "GetUVTerm() result:", source);
	return source;
}

/*
*	replaces delta^x ==> 0 for x>=iPowMin
*	works with reduced source.
*	reduces result
*/
CExpression* CLHEPBasics::TruncateDeltaPower(CExpression* source, int iStartFromPow) {
	if ((source == NULL) || (iStartFromPow<=0)) return NULL;
	dbg->log(35, iStartFromPow, "TruncateDeltaPower(from pow = %ld):", source);

	bool Ndim = CETensor::isNDimensionSpace();
	if (!Ndim) CETensor::setNDimensionSpace(true);
	CExpression* dimDeltaSym = CETensor::getDimensionDelta();
	if (dimDeltaSym == NULL) { CEException::halt(dbg, "SingularArrange(): NULL dimDeltaSym found"); }
	CEPowerReplRule* r = new CEPowerReplRule(dimDeltaSym, iStartFromPow-1);
	CExpression* res = r->ApplyTo(source, true);
	delete r;
	if (!Ndim) CETensor::setNDimensionSpace(false);

	res = CExpression::reduce(res);
	dbg->log(15, "SingularArrange() result:", res);
	return res;
}

// does not delete input
CExpression* CLHEPBasics::SingularArrangeImpl(CExpression* src, CExpression* dimDeltaSym, CExpression* singSym, CExpression* singIRSym) {
	dbg->log(70, "SingularArrangeImpl():", src);
	CExpression* res = NULL;
	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* sum_src = (CESum*)src;
		long s = sum_src->size();
		long i;
		for(i=0; i<s; i++) {
			CExpression* item = sum_src->GetAt(i);
			CExpression* item_res = SingularArrangeImpl(item, dimDeltaSym, singSym, singIRSym);
			if ((item_res != NULL) && (item_res != item)) {
				sum_src->SetAt(i, sum_src); delete sum_src;
			}
		}
		res = sum_src;
	} else {
		CEProduct* prod_src = NULL;
		if (src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) { prod_src = (CEProduct*)src;
		} else { prod_src = new CEProduct(); prod_src->addItem(src->clone()); }
		res = SingularArrangeProduct(prod_src, dimDeltaSym, singSym, singIRSym);
		if (res != NULL) { // value changed
			if ((res != prod_src) && (prod_src != src)) delete prod_src; // delete internal
		} else { // res == NULL
			if (prod_src != src) delete prod_src; // no prod-factorization required
			res = src;
		}
	}
	dbg->log(70, "SingularArrangeImpl() result:", res);
	return res;
}

// narrow-usage function; CEBasic::LeaveSeries() should make it generally
// does not delete source
// NULL result means the source is unchanged
// no reduction is performed
CExpression* CLHEPBasics::SingularArrangeProduct(CEProduct* src, CExpression* dimDeltaSym, CExpression* singSym, CExpression* singIRSym) {
	if ((src == NULL) || (dimDeltaSym == NULL) || (singSym == NULL) || (singIRSym == NULL)) { dbg->log(0, "{Error} SingularArrangeProduct(): NULL call"); return false; }
	if (!CRContainCheck::search(src, dimDeltaSym)) return NULL;
	long s = src->size();
	long i=0;

	///// scan expression
	long pos_D = -1; long sym_spotted = 0; // delta(Eta) pos
	long pos_Xi = -1; long num_Xi = 0; // 1/delta pos
	long pos_IR = -1; long num_IR = 0; // IR-pos
	long pos_VI = -1; long num_VI = 0; // VI
	long pow_D_pos = -1; long pow_D_val = 0; long num_D_pow = 0; // {\delta}^pow
	for (i=0; i<s; i++) {
		CExpression* item = src->GetAt(i);
		if ((!item->getFlag()) || (!item->getLevel())) continue;
		if (dimDeltaSym->equals(item)) { pos_D = i; sym_spotted++; }
		if (singSym->equals(item)) { pos_Xi = i; num_Xi++; }
		if (singIRSym->equals(item)) { pos_IR = i; num_IR++; }
		if (item->getProperty(CEProperties::PROP_MASK_IS_VIRT_INTEGRAL)) { pos_VI = i; num_VI++; }
		if (CEUtil::equals(item->ClassName(), CEPower::CLASS_NAME)) {
			if (dimDeltaSym->equals(((CEPower*)item)->basis())) {
				pow_D_pos = i; num_D_pow++; pow_D_val = ((CEPower*)item)->getLongPower();
				if (pow_D_val==LONG_MIN) { dbg->log(0, "ERROR: non-integer power at singular item", src); CEException::halt(); }
				//TODO: check logic!!!! The last power is only searched, is it OK?
			}
		}
	}
	if ((num_Xi+num_VI+num_D_pow+num_IR) > 1) { dbg->log(0, "{Error} SingularArrangeProduct(): Complicated source call", src); return NULL; }
	if ((num_Xi > 0) && (pos_Xi < 0)) { dbg->log(0, "{Internal Error} SingularArrangeProduct(): Invalid Xi-data", src); return NULL; }
	if ((num_VI > 0) && (pos_VI < 0)) { dbg->log(0, "{Internal Error} SingularArrangeProduct(): Invalid VI-data", src); return NULL; }
	if ((num_IR > 0) && (pos_IR < 0)) { dbg->log(0, "{Internal Error} SingularArrangeProduct(): Invalid IR-data", src); return NULL; }
	if ((num_D_pow > 0) && (pow_D_pos < 0)) { dbg->log(0, "{Internal Error} SingularArrangeProduct(): Invalid pow-D-data", src); return NULL; }

	// powers
	if (pow_D_pos >= 0) {
		dbg->log(0, "WARNING: delta power after TruncateDeltaPower()", src);
		if ((num_D_pow != 1) || (pow_D_val < 1)) { dbg->log(0, "{Error} SingularArrangeProduct(): Complicated source call", src); return NULL; }
		if (pow_D_val == 1) { //consider as 'delta'
			pos_D = pow_D_pos; sym_spotted++;
			pow_D_val = 0; num_D_pow = 0; pow_D_pos = -1;
		} else { // pow_D_val > 1
			// set as '0'
			CENumber zero((long)0);
			src->SetAt(pow_D_pos, &zero);
			return src;
		}
	}

	if ((sym_spotted != 1) || (pos_D < 0)) { dbg->log(0, "{Error} SingularArrangeProduct(): Complicated source call", src); return NULL; }

	// no IR allowed
	if (num_IR > 0) { dbg->log(0, "{Error} SingularArrangeProduct(): Complicated source call (IR-term)", src); return NULL; }

	// 1. exchange delta*Xi ==> -2;
	if (pos_Xi >= 0) {
		CExpression* mTwo = CENumber::GetTwo(); mTwo->inverseFlag();
		CExpression* unit = CENumber::GetUnit();
		//src->GetAt(pos_D)->copyAttributesTo(unit, true);
		src->SetAt(pos_D, unit); delete unit;
		//src->GetAt(pos_Xi)->copyAttributesTo(mTwo, true);
		src->SetAt(pos_Xi, mTwo); delete mTwo;
		return src;
	}

	// 2. exchange delta*VI ==> -2*{VI-data};
	if (pos_VI >= 0) {
		CExpression* VI_expr = src->GetAt(pos_VI);
		if (!VI_expr->getProperty(CEProperties::PROP_MASK_IS_VIRT_INTEGRAL)) { CEException::halt(dbg, "No VI found at expected position"); }
		CExpression* VI_subst = ((CVirtIntegral*)VI_expr)->GetSingularTermFactor();
		if (VI_subst == NULL) {// can't appreciate VI
			dbg->log(0,"Internal WARNING: unknown 4-dimensional limit for virtual integral ",VI_expr);
			return NULL;
		}
		CExpression* mTwo=CENumber::GetTwo(); mTwo->inverseFlag(); src->SetAt(pos_D, mTwo); delete mTwo;
		src->SetAt(pos_VI, VI_subst); delete VI_subst; // note VI-flags ate included into VI_subst
		//debug
		//if(CRContainCheck::search(src,dimDeltaSym)) { dbg->log(0, "{SingularArrange Error} Dimension delta is left in result: ", src); }
		return src;
	}

	// 3. if no VI or Xi is left ==> repl delta ==> 0 (Warning: expression MUST be expanded)
	if (CRContainCheck::search(src, singSym)) { dbg->log(0, "{Error} SingularArrangeProduct(): Complicated source call", src); return NULL; }
	if (CRContainCheck::search(src, CEProperties::PROP_MASK_IS_VIRT_INTEGRAL)) { dbg->log(0, "{Error} SingularArrangeProduct(): Complicated source call", src); return NULL; }

	CENumber zero((long)0);
	src->SetAt(pos_D, &zero);
	return src;
}

/*
*	A >> B check (abs only)
*/
bool CLHEPBasics::IsAGGthanB_abs(CExpression* a, CExpression* b) {
	if ((a == NULL) || (b == NULL)) { dbg->log(0, "<ERROR> IsAGGthanB_abs(): NULL parameter"); CEException::halt(); }
	bool res = false;
	CExpression* a_cpy = a->clone(); CExpression* b_cpy = b->clone(); CERule* r = get_mass_order_rule();
	if (r != NULL) {
		a_cpy = CExpression::reduce(r->ApplyTo(a_cpy));
		b_cpy = CExpression::reduce(r->ApplyTo(b_cpy));
	}
	// dbg->log(105, "CLHEPBasics::IsAGGthanB_abs", a_cpy, b_cpy);

	if ((CEBasic::isZero(b_cpy)) && (!CEBasic::isZero(a_cpy))) res = true;
	delete a_cpy; delete b_cpy; if (r != NULL) delete r;
	return res;
}

// A << B check (abs only)
bool CLHEPBasics::IsALLthanB_abs(CExpression* a, CExpression* b) {
	if ((a == NULL) || (b == NULL)) { dbg->log(0, "<ERROR> IsALLthanB_abs(): NULL parameter"); CEException::halt(); }
	bool res = false;
	CExpression* a_cpy = a->clone(); CExpression* b_cpy = b->clone(); CERule* r = get_mass_order_rule();
	if (r != NULL) {
		a_cpy = CExpression::reduce(r->ApplyTo(a_cpy));
		b_cpy = CExpression::reduce(r->ApplyTo(b_cpy));
	}

	if ((CEBasic::isZero(a_cpy)) && (!CEBasic::isZero(b_cpy))) res = true;
	delete a_cpy; delete b_cpy; if (r != NULL) delete r;
	return res;
}

// 0 means unknown result
int CLHEPBasics::GetExprSign(CExpression* src) {
	if (src == NULL) { dbg->log(0, "<ERROR> GetExprSign(): NULL parameter"); CEException::halt(); }

	if (src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return 0;

	if (!src->getFlag()) {
		src = src->clone(); src->setFlag();
		int res = -GetExprSign(src); delete src;
		return res;
	}

	if (src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		CEProduct* pr = (CEProduct*)src;
		long s = pr->size();
		if (s == 0) return 0;
		bool res = true;
		long i;
		for (i=0; i<s; i++) {
			int sub_res = GetExprSign(pr->GetAt(i));
			if (sub_res == 0) return 0;
			if (sub_res < 0) res = !res;
		}
		return ((res)?1:(-1));
	}

	if (src->getProperty(CEProperties::PROP_MASK_IS_NUMBER)) {
		CENumber* num = (CENumber*)src;
		return num->cmpZero();
	}
	
	CKinematics* kin = CEProcessor::kinematics();
	if (kin != NULL) { int res = kin->checkExprSign(src); if (res != 0) return res; }

	return 0;
}

/////////////////////////////////////////////////////////// me2-->diffCS
// destroys source
CExpression* CLHEPBasics::GetDiffCS(CExpression* me2, bool bGeV2Pb, bool bPostArrange) {
	if (me2 == NULL) { dbg->log(0, "CLHEPBasics::GetDiffCS(NULL) call"); return NULL; }
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) { dbg->log(0, "CLHEPBasics::GetDiffCS() ERROR: No kinematiks installed"); CEException::halt(); return NULL; }
	if (kin->num_init_particles() != 2) { dbg->log(0, kin->num_init_particles(), "CLHEPBasics::GetDiffCS ERROR: cross section not available for %ld->XXX process"); CEException::halt(); return NULL; }

	CEParticle* p1 = kin->pin(0);
	CEParticle* p2 = kin->pin(1);

	// factor = 1/ 4 sqrt( (p1*p1)**2 - m1**2 m2**2 )
	CExpression* I_expr = (new CESProduct(p1->get_momentum_symbol(),p2->get_momentum_symbol()))->sqra();
	I_expr = I_expr->minus(p1->get_sym_mass()->sqra()->multiply(p2->get_sym_mass()->sqra()));
	I_expr = CENumber::GetFour()->multiply(CEPower::sqrt(I_expr));

	// simplify FLUX factor
	DWORD KA_setts=0; if (CEProcessor::m_MandelstamsEverywhere) { KA_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
	I_expr = CLHEPBasics::KinArrange(I_expr, KA_setts);
	DWORD KS_setts=0; if (CEProcessor::m_MandelstamsEverywhere) { KS_setts |= CLHEPBasics::KS_INVOLVE_MANDELSTAM_VARS; }
	I_expr = CLHEPBasics::KinSimplify(I_expr, KS_setts);//KS_INVOLVE_MANDELSTAM_VARS
	CExpression* res = me2->divide(I_expr);

	if (bGeV2Pb) {
		CDiagramPhysics* phys=CDiagramPhysics::get_curr_physics();
		if (phys == NULL) { dbg->log(0, "CLHEPBasics::GetDiffCS() ERROR: No physics declared"); CEException::halt(); return NULL; }
		CExpression* fGeV2Pb = phys->GetGeV2PbFactor();
		if (fGeV2Pb == NULL) { dbg->log(0, "CLHEPBasics::GetDiffCS() ERROR: NULL fGeV2Pb factor"); CEException::halt(); return NULL; }
		res=res->multiply(fGeV2Pb);
	}
	dbg->log(27,"Differential cross section factor applied", res);

	if(bPostArrange){ // arrange result
		res=CExpression::arrge(res); //TODO: check why it's so slow (1.5 mins for pe->peee QED square)
	}
	return res;
}

/////////////////////////////////////////////////////////// Symbol booking
bool CLHEPBasics::BookSymbol(CESymbol* sym) { // true is booked, source unaffected
	dbg->log(30, "BookSymbol call", sym);
	if (sym == NULL) return false;
	if ((!sym->getFlag()) || (!sym->getLevel())) {
		dbg->log(0, "{BookSymbol} Error, try booking complicate symbol", sym);
		return false;
	}

	// check for booked
	if (isSymbolBooked(sym)) return false;

	// book symbol
	if (m_BookedSymsList== NULL) { m_BookedSymsList=new CSequence(); }
	m_BookedSymsList->addItem(sym->clone());

	dbg->log(70, "BookSymbol: symbol booked OK", sym);
	return true;
}

bool CLHEPBasics::isSymbolBooked(CESymbol* sym) { // source unaffected
	if (sym == NULL) return true; // it may not be booked, hence it is booked
	if ((!sym->getFlag()) || (!sym->getLevel())) {
		dbg->log(0, "{BookSymbol} Error, check for complicate symbol booked");
		return true;
	}

	// check kinematics (momenta & polarizations)
	CKinematics* kin = CEProcessor::GetKinIfDeclared();
	if (kin != NULL) {
		if (kin->isSymbolUsed(sym)) return true;
	}
	//else return true; // no symbol bookng available, all are booked by future kin-set

	// check pre-defined basis vectors
	if (isBasisSymbol(sym)>0) return true;

	// check booked list
	if (m_BookedSymsList != NULL) {
		int s = m_BookedSymsList->size();
		int i;
		for(i=0; i<s; i++) { if(sym->equals(m_BookedSymsList->GetAt(i))) return true; }
	}
	return false;
}

// returns cloned result (to be deleted after use)
CESymbol* CLHEPBasics::BookSymbolOrUnique(const char* sym, const char* alt, const char* alt2) {
	CESymbol* res = NULL;
	if (sym != NULL) {
		res = new CESymbol(sym);
		if (BookSymbol(res)) return res;
		delete res;
	}

	if (alt != NULL) {
		res = new CESymbol(alt);
		if (BookSymbol(res)) return res;
		delete res;
	}

	if (alt2 != NULL) {
		res = new CESymbol(alt2);
		if (BookSymbol(res)) return res;
		delete res;
	}
	
	res = CESymbol::getUnique(); // unique symbol should not be added to booked-list
	return res;
}

/////////////////////////////////////////////////////////// Basis vector symbols info
int CLHEPBasics::isBasisSymbol(CESymbol* sym) {
	if(sym == NULL)return 0;

	// Amplitude basis (res:1)
	if (sym->equals(CAmplCalcLib::GetBasisVecSym(true,true))) return 1;
	if (sym->equals(CAmplCalcLib::GetBasisVecSym(true,false))) return 1;
	if (sym->equals(CAmplCalcLib::GetBasisVecSym(false,true))) return 1;
	if (sym->equals(CAmplCalcLib::GetBasisMassSym())) return 1;

	// Bondarev functions basis (res:2)
	for(int i=0; i<4; i++) { if (sym->equals(CEFuncBondarev::GetBondarevBaseSym(i))) return 2; }

	return 0;
}

/////////////////////////////////////////////////////////// Sofr bremsstrahlung factor (IR-limit)
CExpression* CLHEPBasics::GetSoftBremsstrahlungFactor() {
	CKinematics* kin=CEProcessor::kinematics(); CDiagramPhysics* phys=CDiagramPhysics::get_curr_physics(); if((kin==NULL)||(phys==NULL))return NULL;
	if(phys->ClassName() != CSMPhysics::CLASS_NAME) {
		dbg->log(0, "ERROR. Soft bremsstrahlung calculation is available for internal SM only.");
		dbg->log(0, "Plese wait for the next version of ALHEP.");
		return NULL;
	}
	CExpression* symE=((CSMPhysics*)phys)->getModelContsSym(CSMPhysics::CONST_E); if(symE==NULL)return NULL;
	CExpression* symESrq=symE->sqra();
	bool bNoProtonRadiation = (((((CSMPhysics*)phys)->getPhysicsID())&CSMPhysics::PHYS_PROTON_STRUCTFUNC)==CSMPhysics::PHYS_PROTON_STRUCTFUNC);

	CESum* sumRes=new CESum();
	for(int i=-kin->num_final_particles(); i<=kin->num_init_particles(); i++) {
		if(i==0)continue;
		CEParticle* partI=kin->get_particle_by_ID(i); if(partI==NULL)continue;//error
		if((bNoProtonRadiation)&&(abs(partI->get_PK())==CEParticle::PARTICLE_PROTON))continue;
		int i3c=phys->getCharge3(partI->get_PK()); if(i3c==0)continue; CENumber chrgI(i3c,3); chrgI.Arrange(AL_INLINE);
		CExpression* momI=partI->get_momentum_symbol(); if(momI==NULL){ CEException::halt("partI->get_momentum_symbol()==NULL"); }
		int iKinFlag=((i<0)?1:(-1)); //'+1' for final/'-1' for initials
		//calc. factor
		CExpression* item=symESrq->clone(); item=item->multiply(chrgI.sqra()); //if((i3c!=3)&&(i3c!=-3))item=item->multiply(chrgI.sqra());
		item=item->multiply(new CEFunction(CFFortran::CFuncRegistry::F_SOFT_BREMS_INT_1,momI));
		sumRes->addItem(item);

		for(int j=i+1; j<=kin->num_init_particles(); j++) {
			if(j==0)continue;
			CEParticle* partJ=kin->get_particle_by_ID(j); if(partJ==NULL)continue;//error
			if((bNoProtonRadiation)&&(abs(partJ->get_PK())==CEParticle::PARTICLE_PROTON))continue;
			int j3c=phys->getCharge3(partJ->get_PK()); if(j3c==0)continue; CENumber chrgJ(i3c,3); chrgJ.Arrange(AL_INLINE);
			CExpression* momJ=partJ->get_momentum_symbol(); if(momJ==NULL){ CEException::halt("partJ->get_momentum_symbol()==NULL"); }
			int jKinFlag=((j<0)?1:(-1)); //'+1' for final/'-1' for initials

			//calc. factor
			CExpression* item=symESrq->clone();
			if((iKinFlag*jKinFlag) < 0){ item=item->multiply(CENumber::GetMinusUnit()); }
			item=item->multiply(chrgI.clone()); item=item->multiply(chrgJ.clone()); //if(i3c!=3)item=item->multiply(chrgI.clone()); if(j3c!=3)item=item->multiply(chrgJ.clone());
			item=item->multiply(new CEFunction(CFFortran::CFuncRegistry::F_SOFT_BREMS_INT_2,momI,momJ))->multiply(CENumber::GetTwo());
			sumRes->addItem(item);
		}
	}

	//Arrange
	dbg->log(2, "GetSoftBremsstrahlungFactor() result (before simplification)", sumRes);
	CExpression* res=CExpression::arrge(sumRes);
	dbg->log(2, "GetSoftBremsstrahlungFactor() result", res);
	return res;
}
