////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LHEPRunProject.cpp: implementation of the CLHEPRunProject class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#define CAlHepLib_VER		0x010201
#define VM_USE_PYTHIA8		"VM_USE_PYTHIA8"
#define N_R_TERMS			2					//+gluon, +photon
#define RUN_STAT_FILE		"chStat.inf"

bool CLHEPRunProject::CLHEPRunProject_Settings::bEvtNuclonEntriesDefault	= true;

CDebug* CLHEPRunProject::dbg = new CDebug("CLHEPRunProject");

//////////////////////////////////////////////////////////////////////
const unsigned long CLHEPRunProject::flag_RUN_DEFAULT	= 0; //default flags
const unsigned long CLHEPRunProject::flag_RUN_SQUARE	= 1; //square matrix elements instead of amplitude calculation
const unsigned long CLHEPRunProject::flag_RUN_AMPLITUDE	= 2; //calculate amplitude instead of squaring ME
const unsigned long CLHEPRunProject::flag_RUN_RC		= 4; //include radiative correction (Loops + Counterterms + Bremsstrahlung)
const unsigned long CLHEPRunProject::flag_RUN_LOAD		= 8; //try to load intermadiate expressions (instead of re-calculating them)

bool CLHEPRunProject::Run(const char* title,
						  unsigned long iFlags,
						  const char* fnTeX) {
	dbg->log(7,"CLHEPRunProject::Run(). Title: ",title);

	//check 'title'
	if(!CEUtil::IsAllowedFileName(title)){ dbg->log(0,"Run: invalid project name"); return false; }

	//check 'iFlags' flags
	if(((iFlags & CLHEPRunProject::flag_RUN_SQUARE) == CLHEPRunProject::flag_RUN_SQUARE) &&
		((iFlags & CLHEPRunProject::flag_RUN_AMPLITUDE) == CLHEPRunProject::flag_RUN_AMPLITUDE))
	{
		dbg->log(0,"Invalid 'Run' parameter: use either 'RUN_AMPLITUDE', 'RUN_SQUARE' or nothing for auto-select");
		iFlags |= (CLHEPRunProject::flag_RUN_AMPLITUDE | CLHEPRunProject::flag_RUN_SQUARE);
		iFlags ^= (CLHEPRunProject::flag_RUN_AMPLITUDE | CLHEPRunProject::flag_RUN_SQUARE);
	}

	//check kinematics: no quark-containing nuclons in final-state
	//(the only form-factor-described protons are alowed in final-state)
	//...

	CLHEPRunProject lib(title,iFlags,fnTeX);
	bool bRes=lib.RunImpl();
	if(bRes){
		dbg->log(7,"CLHEPRunProject::Run() completed successfully");
	}else{
		dbg->log(0,"CLHEPRunProject::Run() FAILED!");
	}
	return bRes;
}

//
CLHEPRunProject::CLHEPRunProject(const char* title,
								 unsigned long iFlags,
								 const char* fnTeX) : CELibrary()
//, m_sTitle(title)
{
	m_sTitle.Reset(); m_sTitle.AddString(title);
	m_sTexReport.Reset(); if(fnTeX!=NULL){ m_sTexReport.AddString(fnTeX); }else{ m_sTexReport.AddString(title); m_sTexReport.AddString(".tex"); }
	m_setts.m_iFlags=iFlags;
	m_kin=NULL; m_phys=NULL;
	m_codeLang=NULL;
	m_iCurrKinConfNo=0;
}
CLHEPRunProject::~CLHEPRunProject() {
	if(m_codeLang!=NULL)delete m_codeLang;
}
//
bool CLHEPRunProject::RunImpl() {
	// 1.1. Init secondary fields in 'm_setts' and 'm_infConfig'
	//
	if((!m_setts.m_bCodeCpp) && (m_setts.m_bUseAbstractTypes)) { dbg->log(0,"WARNING: 'm_bUseAbstractTypes' flag will be ignored for NON-C++ code generation"); m_setts.m_bUseAbstractTypes=false; }
	if(!m_setts.m_bCodeCpp) { dbg->log(0,"ERROR: Fortran output not implemented for 'Run' procedure"); return false; }
	//m_setts.m_bSeparateChannelsAsProcs=true;			//default: false

	// 1.2. Basic init
	m_kin=CEProcessor::kinematics(); if(m_kin==NULL)return false;
	m_phys=CDiagramPhysics::get_curr_physics(); if(m_phys==NULL)return false;
	if((m_kin->num_init_particles()<=0) || (m_kin->num_final_particles()<=0)) { dbg->log(0,m_kin->num_init_particles(),m_kin->num_final_particles(),"ERROR: Invalid kinematics %ld -> %ld"); return false; }
	if(m_kin->num_init_particles()!=2){ dbg->log(0,m_kin->num_init_particles(),"ERROR: 'Run' procedure not implemented for %ld-particle initial state"); return false; }
	CEParticle* p1=m_kin->pin(0);
	CEParticle* p2=m_kin->pin(1);
	if((p1==NULL) || (p2==NULL)){ dbg->log(0,m_kin->num_init_particles(),"Internal ERROR: NULL particle in kinematics"); return false; }
	m_infConfig.m_bNuclonToQuarks=true;
	if(m_phys->ClassName()==CSMPhysics::CLASS_NAME){
		if(((((CSMPhysics*)m_phys)->getPhysicsID())&CSMPhysics::PHYS_PROTON_STRUCTFUNC)==CSMPhysics::PHYS_PROTON_STRUCTFUNC){
			m_infConfig.m_bNuclonToQuarks=false;
		}
	}
	if(m_codeLang==NULL)delete m_codeLang;
	m_codeLang=CCodeLanguage::CreateLanguage((m_setts.m_bCodeCpp ? CLFileManager::FT_CPP : CLFileManager::FT_F77) | (m_setts.m_bUseAbstractTypes ? CLFileManager::FT_CPP_TYPE_ABSTRACT : 0));
	if(m_codeLang==NULL)return false;
	int i;

	// 1.3. init secondary fields in 'm_infConfig'
	//
	if ((p1->check_PK(CEParticle::PARTICLE_PROTON,true)) && (p2->check_PK(CEParticle::PARTICLE_PROTON,true))) {
		//Note: p+pBar collisions are OK here
		m_infConfig.m_bAllowQuarksSwap=true;
	}else{
		m_infConfig.m_bAllowQuarksSwap=false;
	}
	m_infConfig.m_pkInitials[0]=p1->get_PK();
	m_infConfig.m_pkInitials[1]=p2->get_PK();
	//
	// Init particle titles for local use (and rename similar particles)
	m_infConfig.m_symInIDs.clear_all();
	m_infConfig.m_symFinIDs.clear_all();
	for(i=0; i<(m_kin->num_init_particles()+m_kin->num_final_particles()); i++) {
		bool bFinal=(i>=m_kin->num_init_particles()); int iNo=(bFinal?(i-m_kin->num_init_particles()):i);
		CEString* sID = new CEString(); CEParticle* p=(bFinal?(m_kin->pfin(iNo)):(m_kin->pin(iNo))); const char* symBasic=NULL; long pk=0;
		if(p!=NULL) { pk=p->get_PK(); symBasic=CDiagramPhysics::get_particle_symbolA(pk,CDiagramPhysics::SYMBOL_STYLE_TEX); }
		if(symBasic!=NULL){
			*sID<<symBasic;
			if(p->check_PK(CEParticle::PARTICLE_QUARK,true)){ if(pk<0){ *sID<<"bar"; } }
			else if(p->check_PK(CEParticle::PARTICLE_PROTON,true)){ if(pk<0){ *sID<<"bar"; } }
			else{ int iC3=m_phys->getCharge3(pk); bool bCharged=((iC3!=INT_MIN)&&(iC3!=0)); if(bCharged){ *sID<<((iC3>0)?"+":"-"); } }
		}else{ *sID<<"Unknown"; }
		bool bRename=false; //rename similar particles (e.g. it two 'tbar' spotted)
		for(int j=0; j<(bFinal?(m_infConfig.m_symFinIDs.size()):(m_infConfig.m_symInIDs.size())); j++) {
			CEString* sPrev=(CEString*)(bFinal?(m_infConfig.m_symFinIDs.GetAt(j)):(m_infConfig.m_symInIDs.GetAt(j))); if(sPrev==NULL)continue;
			if(sPrev->equals(sID)) { *sPrev<<"_"<<(j+1); bRename=true; }
		}
		if(bRename){ *sID<<"_"<<(iNo+1); }
		if(bFinal){ m_infConfig.m_symFinIDs.add(sID); }else{ m_infConfig.m_symInIDs.add(sID); }
	}

	//check kinematics
	if(!CheckKinematics()) { dbg->log(0,"Run engine is not implemented for current kinematics. Stopped."); CEException::halt(); return false; }

	//check RC availability
	if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_RC) {
		if(!m_setts.m_bCodeCpp){
			dbg->log(0,"ERROR: radiative correction is not fully implemented in Fortran code");
			dbg->log(0,"ERROR: (one may remove this check and continue on own risk)");
			return false;
		}
	}

	dbg->log(25,"<Run> 1. Initialized");

	// 2. Create project directory
	CEString sPath; if(!CVMFileUtil::GetCurrentPath(sPath)){ dbg->log(0,"ERROR: CVMFileUtil::GetCurrentPath failed"); CEException::halt(); }
	CVMFileUtil::MkDir(m_sTitle.GetBuffer()); CVMFileUtil::ChDir(m_sTitle.GetBuffer());
	if(CVMFileUtil::GetCurrentPath(m_sProjPath)){
		dbg->log(45,"<Run> Working dir changed");
		//Note: For proper reading by Mathematica path 'D:\Work\channels\AA_ll\' must be converted to 'D:\\Work\\channels\\AA_ll\\'
		//[wrong]dbg->log(45,"<Run> Working dir changed to ",m_sProjPath.GetBuffer());
	}

	// 2.1. Start files in working dir
	CESupport::saveTeX(m_sTexReport.GetBuffer(),"",CESupport::FILE_TRUNCATE,false);
	m_sProjMainFile.Reset(); m_sProjMainFile<<m_sTitle.GetBuffer()<<".cpp";

	//2.9. Reset code libraries
	// Nt: for every set of final particles we use separated instance of code library
	m_ProcCodeLibs.clear_all();
	int nKinConfigs=1;
	if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_RC) { nKinConfigs+=N_R_TERMS; } //TODO: make this more abstract!
	for(i=0; i<nKinConfigs; i++) {
		CCodeProcessAdv* _ProcCodeLib = NULL;
		if(i==0){
			_ProcCodeLib = new CCodeProcessAdv(m_sTitle.GetBuffer());
		}else{
			CEString sTmp(m_sTitle.GetBuffer()); sTmp<<"R"<<i; _ProcCodeLib = new CCodeProcessAdv(sTmp.GetBuffer());
		}
		if(_ProcCodeLib==NULL)CEException::halt();
		_ProcCodeLib->Reset();
		_ProcCodeLib->m_bUseEvtNuclonEntries = m_setts.m_bUseEvtNuclonEntries;
		_ProcCodeLib->m_beam1.pk = p1->get_PK();
		_ProcCodeLib->m_beam1.bPartonsUsed = CLHEPRunProject::CheckParticleConfigNum(p1,NULL);
		//if(_ProcCodeLib->m_beam1.bPartonsUsed) {
		//	_ProcCodeLib->m_beam1.pdfId1=4; _ProcCodeLib->m_beam1.pdfId2=46; //default: CTEQ5L from PYTHIA
		//}
		_ProcCodeLib->m_beam2.pk = p2->get_PK();
		_ProcCodeLib->m_beam2.bPartonsUsed = CLHEPRunProject::CheckParticleConfigNum(p2,NULL);
		//if(_ProcCodeLib->m_beam2.bPartonsUsed) {
		//	_ProcCodeLib->m_beam2.pdfId1=4; _ProcCodeLib->m_beam2.pdfId2=46; //default: CTEQ5L from PYTHIA
		//}
		m_ProcCodeLibs.add(_ProcCodeLib);
	}
	m_iCurrKinConfNo=0;

	// 2.99 Prepare for channel iterations, load channel statistics file
	m_infConfig.m_nSubProcsTried=0;
	m_infConfig.m_nSubRunCalls=0;
	if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) {
		if(m_infConfig.m_sSubRunStat != NULL){ CMemoryManager::mfree(m_infConfig.m_sSubRunStat); m_infConfig.m_sSubRunStat=NULL; }
		m_infConfig.m_sSubRunStat = CAbstractFileParser::LoadFileToMemory(RUN_STAT_FILE,&(m_infConfig.m_szSubRunStat),CAbstractFileParser::LOAD_AS_BINARY);
		if(m_infConfig.m_sSubRunStat == NULL) {
			dbg->log(0,"WARNING: can't read channel statistics file. 'RUN_LOAD' flag is cancelled");
			m_setts.m_iFlags ^= CLHEPRunProject::flag_RUN_LOAD; m_infConfig.m_szSubRunStat=0;
		}
		//Note: total number of combinations (m_infConfig.m_nConfigCombinations) is estimated in first 'SetNextSubProcConfig' call
	}else{ //if((m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) == 0) {
		ofstream ofsStat(RUN_STAT_FILE, ios::out|ios::trunc); if(ofsStat){ ofsStat.close(); }
	}
	dbg->log(25,"<Run> 2. Prepared for sub-process processing");

	// 3. Iterate sub-processes
	int iKinConf=0;	// no. of kin-configuration, 0 for Born&Loop, 1+... for bremsstrahlung
	int iKinConfMax=0;
	for(;;){
		if(SetNextSubProcConfig()){
			//first call: allocate statistics, check previous stat. file size...				
			//TODO: implement this
			//if(m_infConfig.m_nSubProcsTried == 1) {
			//	if(m_infConfig.m_nConfigCombinations<0)CEException::halt("Int. error: m_infConfig.m_nConfigCombinations<0");
			//	...
			//}

			//fast check for total charge conservation -- don't try create diagrams here
			if(!CheckChargeConservation())continue;

			dbg->log(45,m_infConfig.m_nSubProcsTried,"<Run> Trying sub-process no. %ld",m_kin);

			// Born [+ Loops + CT +...]
			// Note: loops MAY not be currently considered for vanishing Born process
			// E.g. gamma+gamma -> gamma+gamma process may not be evaluated
			bool bChOk=CLHEPRunProject::RunSubProcess(iKinConf);
			if(bChOk){ m_infConfig.m_nSubProcs++; }
			if((m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) == 0) {
				ofstream ofsStat(RUN_STAT_FILE, ios::out|ios::app); if(ofsStat){ ofsStat<<(bChOk?'1':'0'); ofsStat.close(); }
			}

			if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_RC) {	// Add R-terms
				// Create adjacent kinematics
				m_setts.m_iFlags ^= CLHEPRunProject::flag_RUN_RC;	//avoid RC to R-term
				ResetBremmsConfig();
				for(;;){
					if(!SetNextBremmsConfig()) { break; }
					int iCurrKinConfNoPrev=m_iCurrKinConfNo;
					m_iCurrKinConfNo = m_infConfig.m_nBremsConfId+1;	//Note: do not mix different kinematics into single class!
					if(iKinConfMax < m_iCurrKinConfNo) {
						iKinConfMax = m_iCurrKinConfNo;
						if(iKinConfMax>=nKinConfigs){ CEException::halt("Internal error: invalid number of kinematical configurations"); }
					}
					bool bChOk=CLHEPRunProject::RunSubProcess(m_iCurrKinConfNo);
					if(bChOk){ m_infConfig.m_nSubProcs++; }
					if((m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) == 0) {
						ofstream ofsStat(RUN_STAT_FILE, ios::out|ios::app); if(ofsStat){ ofsStat<<(bChOk?'1':'0'); ofsStat.close(); }
					}
					m_iCurrKinConfNo=iCurrKinConfNoPrev;				//restore value
				}
				m_setts.m_iFlags |= CLHEPRunProject::flag_RUN_RC;	//avoid RC to R-term
			}
		}else{
			break;
		}
	}
	dbg->log(15,m_infConfig.m_nSubProcs,m_infConfig.m_nSubProcsTried,"<Run> Channels accepted: %ld (Tried: %ld)");
	dbg->log(25,"<Run> 3. Sub-processes processed");

	//
	RestoreOriginalConfig(); //restore original kinematics

	// 4. Create numerical integration/generation project
	m_iCurrKinConfNo=0;
	if(!m_setts.m_bSeparateChannelsAsProcs) {
		CreateProcFile();
		if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_RC) {	// Add R-terms
			for(i=0; i<N_R_TERMS; i++) {
				m_iCurrKinConfNo++;
				if(m_iCurrKinConfNo >= m_ProcCodeLibs.size()){ CEException::halt("error in R-kinematics ordering"); break; }
				if(ProcCodeLib()->GetNSubs()<=0)continue; //no R-term with this kinematics
				CreateProcFile();
			}
		}
	}
	m_iCurrKinConfNo=0;
	CreateMainFile();

	// 4.9. Create Makefile
	CreateMakefile();
	dbg->log(25,"<Run> 4. Numerical project created");

	// 9. finalize
	CVMFileUtil::ChDir(sPath.GetBuffer());

	dbg->log(25,"<Run> 9. Procedure completed");
	return (m_infConfig.m_nSubProcs>0);
}

//
bool CLHEPRunProject::SetNextSubProcConfig() {
	//
	int nInit = m_kin->num_init_particles();
	if(m_infConfig.m_bAllowQuarksSwap && (nInit!=2)){ dbg->log(0,"Internal ERROR: m_infConfig.m_bAllowQuarksSwap && (nInit!=2) case is not implemented"); CEException::halt(); }

	if(m_infConfig.m_nSubProcsTried<=0){
		m_infConfig.m_arrSubProcConfig.ToSize(nInit); m_infConfig.m_arrSubProcConfig.FillAll(0);
		//dbg->log(145,"CLHEPRunProject::SetNextSubProcConfig() debug 1");
		//m_infConfig.m_arrOrigParticles.SetAt()

		//
		//estimate the total number of configurations (--> 'm_infConfig.m_nConfigCombinations')
		//+ copy parton-based particles (e.g. nuclons) to 'm_infConfig.m_arrOrigParticles'
		//Nt: nConfig[nuclon] MAY occur == 1 (e.g. for debug purposes)
		//Nt: here we consider initial-state multiuplicity only
		//
		m_infConfig.m_nConfigCombinations=1;
		int nConfig1=1; int nConfig2=1;
		for(int i=0; i<nInit; i++) {
			CEParticle* p=m_kin->pin(i); if(p==NULL){ CEException::halt(dbg,"pin(i)==NULL"); return false; }
			int nConfig=1; bool bParton=CheckParticleConfigNum(p,&nConfig);
			if(bParton){
				if(i==0){ m_infConfig.m_bPartonsUsed_beam1=true; }else if(i==1){ m_infConfig.m_bPartonsUsed_beam2=true; }
			}
			if(i==0)nConfig1=nConfig; if(i==1)nConfig2=nConfig;
			//dbg->log(145,i,nConfig,"CLHEPRunProject::SetNextSubProcConfig() debug 1.%ld.%ld");
			if(nConfig>1){
				//save nuclon particle & set kinematics to first parton state
				m_infConfig.m_arrOrigParticles.SetAt(i,p,false); //Nt: don't allow this array to delete particle
				m_infConfig.m_nConfigCombinations*=nConfig;
				int pID=CKinematics::PID(i,true);
				CEParticle* pSubst=CLHEPRunProject::CreateParticleConfig(pID,0,p); //'0' --> set first state
				CEParticle* pPrev=m_kin->ReplaceParticle(pID,pSubst); if((pPrev!=p)&&(pPrev!=NULL)){delete pPrev;}
			}
		}

		// correct internal flags
		if((m_infConfig.m_bAllowQuarksSwap) && (!m_infConfig.m_bPartonsUsed_beam1) && (!m_infConfig.m_bPartonsUsed_beam2)) {
			dbg->log(0,"Warning: 'm_infConfig.m_bAllowQuarksSwap' flag is set off -- no parton-based particles found");
			m_infConfig.m_bAllowQuarksSwap=false;
		}

		ProcCodeLib()->m_bAutoSwapPartons=m_infConfig.m_bAllowQuarksSwap;
		if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_RC) { //add same flag to derived processes
			for(int i=0; i<N_R_TERMS; i++) {
				if(m_ProcCodeLibs.size() >= m_iCurrKinConfNo+i){ break; } //internal error actually
				ProcCodeLib(m_iCurrKinConfNo+i)->m_bAutoSwapPartons=m_infConfig.m_bAllowQuarksSwap;
			}
		}

		//correct 'm_infConfig.m_nConfigCombinations' for 'p+p'-like processes
		if(m_infConfig.m_bAllowQuarksSwap && (nInit==2)){
			if(nConfig1==nConfig2){
				m_infConfig.m_nConfigCombinations = nConfig1*(nConfig1+1)/2;
			}else{ dbg->log(0,"Internal ERROR: (nInit==2)&&(m_infConfig.m_bAllowQuarksSwap) but (nConfig1!=nConfig2)"); CEException::halt(); }
		}
		//
		m_infConfig.m_nSubProcsTried=1;
		dbg->log(25,m_infConfig.m_nConfigCombinations,"Number of kinematic configurations to try: %ld");
		return true;
	}

	bool bConfigOk=false;
	for(int i=0; i<nInit; i++) {
		CEParticle* pOrig=(CEParticle*)m_infConfig.m_arrOrigParticles.GetAt(i); if(pOrig==NULL)continue;
		int nConfig=1; bool bParton=CheckParticleConfigNum(pOrig,&nConfig); if(!bParton)continue;
		if(nConfig<=1)continue; //the only allowed parton is already set
		int iConfig=m_infConfig.m_arrSubProcConfig.GetAt(i);
		iConfig++; bool fCarry=false;
		if(m_infConfig.m_bAllowQuarksSwap && (i<(nInit-1))) {
			if(iConfig>m_infConfig.m_arrSubProcConfig.GetAt(i+1)){ iConfig=0; fCarry=true; }
		}else{
			if(iConfig>=nConfig){ iConfig=0; fCarry=true; }
		}
		//
		m_infConfig.m_arrSubProcConfig.SetAt(i,iConfig);
		//
		int pID=CKinematics::PID(i,true);
		CEParticle* pSubst=CLHEPRunProject::CreateParticleConfig(pID,iConfig,pOrig);
		CEParticle* pPrev=m_kin->ReplaceParticle(pID,pSubst); if((pPrev!=pOrig)&&(pPrev!=NULL)){delete pPrev;}
		//
		if(!fCarry){ bConfigOk=true; break; }
	}
	if(bConfigOk){ m_infConfig.m_nSubProcsTried++; }
	return bConfigOk;
}
bool CLHEPRunProject::CheckChargeConservation() { //check current kinematics
	int i; int iC3_in=0; int iC3_fin=0;
	for(i=0;i<m_kin->num_init_particles();i++){ CEParticle* p=m_kin->pin(i); if(p!=NULL)iC3_in+=m_phys->getCharge3(p->get_PK()); }
	for(i=0;i<m_kin->num_final_particles();i++){ CEParticle* p=m_kin->pfin(i); if(p!=NULL)iC3_fin+=m_phys->getCharge3(p->get_PK()); }
	return (iC3_in==iC3_fin);
}
bool CLHEPRunProject::CheckParticleConfigNum(CEParticle* p, int* nConfig) { //'true' if particle contains partons
	bool res=false;
	if(p->check_PK(CEParticle::PARTICLE_PROTON,true)){
		if(m_infConfig.m_bNuclonToQuarks) {
			if(nConfig!=NULL) {
				//TODO: uncomment next line after gluons are implemented!
				//*nConfig=m_setts.m_nQuarksInNuclon*2+1; //Sum{q,qbar}+gluon
				*nConfig=m_setts.m_nQuarksInNuclon*2;	//debug
			}
			res=true;
		}
	}
	return res;
}
CEParticle* CLHEPRunProject::CreateParticleConfig(int pID, int iConf, CEParticle* p) {
	//dbg->log(145,pID,iConf,"CLHEPRunProject::CreateParticleConfig() PID: %ld, nConf: %ld");
	//N[generations]: m_setts.m_nQuarksInNuclon
	CEParticle* res=NULL;
	if((m_infConfig.m_bNuclonToQuarks) && (p->check_PK(CEParticle::PARTICLE_PROTON,true))) {
		int iSign=((p->get_PK()>=0)?1:-1);
		CExpression* mom=p->get_momentum_symbol();

		if(iConf == m_setts.m_nQuarksInNuclon*2) { //gluon
			//
			// TODO: implement gluons here!
			//
			//res=new CPFermion(iSign*pkQAbs,((mom!=NULL)?mom=mom->clone():NULL),m_phys->get_mass_sym(pkQAbs)->clone(),(new CESymbol("eTmp")),pID);
			//
		}else{
			int iGen=iConf;
			if(iGen >= m_setts.m_nQuarksInNuclon){ iSign*=-1; iGen-=m_setts.m_nQuarksInNuclon; }
			if(iGen >= m_setts.m_nQuarksInNuclon) { CEException::halt("Unexpected configuration in CLHEPRunProject::CreateParticleConfig"); return NULL; }
			int pkQAbs=0;
			switch(iGen){
			case 0: pkQAbs=CEParticle::PARTICLE_QUARK_U; break;
			case 1: pkQAbs=CEParticle::PARTICLE_QUARK_D; break;
			case 2: pkQAbs=CEParticle::PARTICLE_QUARK_S; break;
			case 3: pkQAbs=CEParticle::PARTICLE_QUARK_C; break;
			case 4: pkQAbs=CEParticle::PARTICLE_QUARK_B; break;
			case 5: pkQAbs=CEParticle::PARTICLE_QUARK_T; break;
			}

			if(pkQAbs==0){ CEException::halt("Unexpected configuration in CLHEPRunProject::CreateParticleConfig"); return NULL; }
			//
			if(mom!=NULL){ mom=mom->clone(); }else{ mom=new CENull(); }
			//Note: mass of initial parton is ZERO
			CExpression* mass = new CENumber(0);
			//[for massive parton:] CExpression* mass=m_phys->get_mass_sym(pkQAbs); if(mass!=NULL){ mass=mass->clone(); }else{ mass=new CENull(); }
			CEString symPolTmp("eTm"); symPolTmp.add_int((pID>=0)?pID:-pID);
			CExpression* symPol = new CESymbol(symPolTmp.GetBuffer());	// temporary symbol for polarization, never used
			//res=new CPFermion(iSign*pkQAbs,mom,mass,symPol,pID);
			res=CEParticle::create_particle(iSign*pkQAbs,pID,mom,mass,symPol,NULL);
			res->set_polarization_info(0);	// initial protons must be unpolarized (checked by 'CheckKinematics')
		}
	}
	return res;
}

void CLHEPRunProject::ResetBremmsConfig() { //clear internal counter before 'SetNextBremmsConfig' calls
	if(m_infConfig.m_nBremsConfId>=0) { //remove prev. additional particle
		CEParticle* p = m_kin->RemoveLastFinalParticle();
		if(p!=NULL){
			if(m_infConfig.m_nBremsConfId<m_infConfig.m_arrBremsParticles.size()) {
				m_infConfig.m_arrBremsParticles.SetAt(m_infConfig.m_nBremsConfId,p,false);
			} else {
				delete p;
			}
		}
	}
	m_infConfig.m_nBremsConfId=-1;
}
bool CLHEPRunProject::SetNextBremmsConfig() {
	//
	//	add next possible massless boson to kinematics
	//	clears added particles at last (ret==false) call
	//
	if(m_infConfig.m_arrBremsParticles.size() == 0){
		CESymbol* momK=new CESymbol("k");
		if((m_kin->find_momentum(momK)!=0) || (m_kin->find_polarization_vec(momK)!=0)){ delete momK; momK=CESymbol::getUnique(); }
		dbg->log(0,"[Info] Symbol for the bremsstrahlung particle momentum is selected: ",momK->get_name());
		CESymbol* polK=new CESymbol("e\\_k");
		if((m_kin->find_momentum(polK)!=0) || (m_kin->find_polarization_vec(polK)!=0)){ delete polK; polK=CESymbol::getUnique(); }
		dbg->log(0,"[Info] Symbol for the dummy polariation vector of bremsstrahlung particle: ",polK->get_name());

		// 1. Photon
		CEParticle* p=CEParticle::create_particle(CEParticle::PARTICLE_PHOTON,0,momK->clone(),NULL,polK->clone(),NULL);
		if(p==NULL) { dbg->log(0,"Internal error: can't create temporary particle"); CEException::halt(); return false; }
		p->set_polarization_info(0);
		m_infConfig.m_arrBremsParticles.add(p);

		// 2. Gluon
		p=CEParticle::create_particle(CEParticle::PARTICLE_GLUON,0,momK->clone(),NULL,polK->clone(),NULL);
		if(p==NULL) { dbg->log(0,"Internal error: can't create temporary particle"); CEException::halt(); return false; }
		p->set_polarization_info(0);
		m_infConfig.m_arrBremsParticles.add(p);

		delete momK; delete polK;
	}
	if(m_infConfig.m_arrBremsParticles.size()<=0) return false;

	bool bConfOk=false;
	if(m_infConfig.m_nBremsConfId<0)m_infConfig.m_nBremsConfId=-1;
	if(m_infConfig.m_nBremsConfId>=0) { //remove prev. additional particle
		CEParticle* p = m_kin->RemoveLastFinalParticle(); if(p==NULL)CEException::halt("Internal error: NULL temporary particle (1)");
		m_infConfig.m_arrBremsParticles.SetAt(m_infConfig.m_nBremsConfId,p,false);
	}
	m_infConfig.m_nBremsConfId++;
	if(m_infConfig.m_nBremsConfId < m_infConfig.m_arrBremsParticles.size()) { //add new additional particle
		CEParticle* p = (CEParticle*)m_infConfig.m_arrBremsParticles.GetAt(m_infConfig.m_nBremsConfId); if(p==NULL)CEException::halt("Internal error: NULL temporary particle (2)");
		m_kin->AddFinalParticle(p);
		m_infConfig.m_arrBremsParticles.SetAt(m_infConfig.m_nBremsConfId,NULL,false);
		bConfOk=true;
	}else{
		m_infConfig.m_nBremsConfId=-1;
	}

	return bConfOk;
}

void CLHEPRunProject::RestoreOriginalConfig() {
	for(int i=0; i<m_kin->num_init_particles(); i++) {
		CEParticle* pOrig=(CEParticle*)m_infConfig.m_arrOrigParticles.GetAt(i);
		if(pOrig!=NULL){
			int pID=(i+1); CEParticle* pPrev=m_kin->ReplaceParticle(pID,pOrig); if((pPrev!=pOrig)&&(pPrev!=NULL)){delete pPrev;}
		}
	}
}
bool CLHEPRunProject::CheckKinematics() { //check if procedure is implemented for current kinematics
	//initial protons must be unpolarized
	for(int i=0; i<m_kin->num_init_particles(); i++) {
		CEParticle* p = m_kin->pin(i);
		if(p==NULL) { dbg->log(0,"Internal ERROR: NULL particle in kinematics"); return false; }
		if(p->check_PK(CEParticle::PARTICLE_PROTON,true)) {
			if(p->get_polarization_info() != 0) { dbg->log(0,"ERROR: Procedure is not implemented for POLARIZED initial protons"); return false; }
		}
	}
	return true;
}

//
//	iKinConfNo: ID of kinematical configuration (used to mix processes with different kinematics, like V+R-terms)
//
bool CLHEPRunProject::RunSubProcess(int iKinConfNo) {	//returns 'true' for non-zero cross section

	// 0.9 Check channel for being pre-calculated (and appeared empty or not)
	m_infConfig.m_nSubRunCalls++;
	if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) {
		bool bInvalidStat=false;
		if(m_infConfig.m_sSubRunStat!=NULL){
			if (m_infConfig.m_nSubRunCalls <= m_infConfig.m_szSubRunStat) {
				char cFlag=*(m_infConfig.m_sSubRunStat+m_infConfig.m_nSubRunCalls-1);
				if (cFlag=='0') { return false; } //skip channel
				if (cFlag!='1') { bInvalidStat=true; }
			}else{ //No more channels calculated at previous loop ==> stop loading, append channel file
				m_setts.m_iFlags ^= CLHEPRunProject::flag_RUN_LOAD;
			}
		}else{ bInvalidStat=true; }
		if(bInvalidStat) {
			//invalid channel statistics ==> all previous calculation is probably wronng
			dbg->log(0,"ERROR: invalid channels statistics is used. Remove 'RUN_LOAD' flag and run project again");
			CEException::halt(); return false;
		}
	}

	// 1. Init
	bool bOk;
	int nInits=m_kin->num_init_particles();
	int nFinals=m_kin->num_final_particles();
	int iOrder=nInits+nFinals-2;						// tree-level order
	CEParticle* pIn1 = ((nInits>0)?m_kin->pin(0):NULL);
	CEParticle* pIn2 = ((nInits>1)?m_kin->pin(1):NULL);
	bool bCalcMESqr=true;				//square or amplitude?
	if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_SQUARE) {
		bCalcMESqr=true;
	}else if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_AMPLITUDE){
		bCalcMESqr=false;
	}else{ //auto-select
		bCalcMESqr=(nFinals<=2); //square for 2->2 process only
		//TODO: check also polarization values
	}
	char buf[256];
	int iProcID = m_infConfig.m_nSubProcs;	//this ID will be assigned to the channel if it appears non-empty
	char sIdSub[20]; sprintf(sIdSub,"Sub%03ld",iProcID);

	//TODO: mix ELW & QCD orders (?)
	//log filename
	//common tex filename

	//TODO: examine this!
	//misc commands
	// polarization setts
	//SetFermionHelicity(1, -1); SetFermionHelicity(2, 1);
	//SetPolarized(1, 0); // set unpolarized
	//SetPolarized(2, 0); // set unpolarized
	//SetPolarized(-1, 0); // set unpolarized
	//SetPolarized(-2, 0); // set unpolarized
	//SetPhotonDMBase(-2, "p\_1");
	//SetPolarized(-3, 0); // set unpolarized
	// Other polarization declarations:
	//SetPhotonHelicity(-2, 1, "p\_1", "p\_2");
	// SetFermionHelicity(-1, 0); SetFermionHelicity(-1, "h_1");

	// 1.99 Try to load matrix elements without creating diagrams
	// Note: if stat file loaded ==> we know, the channel is not empty
	CExpression* resBorn = NULL;		//me^2 or amplitude
	CExpression* resLoop = NULL;		//me4*Conj[me2], may be NULL
	bool bLoaded_res = false;
	if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) {
		sprintf(buf,"%s_Minimized.xml",sIdSub); CESerializable* serTmp = CESupport::read_file(buf, 0);
		if(CEBasic::isExpression(serTmp)) { resBorn = (CExpression*)serTmp; bLoaded_res=true; }
		else{
			//Note: we may not simply stop loading here <= stat file is longer an will be incorrect
			dbg->log(0,"ERROR: Loaded information is incorrect. Run project again without 'RUN_LOAD' option"); CEException::halt();
		}
	}
	if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) {
		if (m_setts.m_iFlags & CLHEPRunProject::flag_RUN_RC) {
			sprintf(buf,"%sV_Minimized.xml",sIdSub); CESerializable* serTmp = CESupport::read_file(buf, 0);
			if(CEBasic::isExpression(serTmp)) { resLoop = (CExpression*)serTmp; bLoaded_res=true; }
			else{
				//Note: we may not simply stop loading here <= stat file is longer an will be incorrect
				dbg->log(0,"ERROR: Loaded information is incorrect. Run project again without 'RUN_LOAD' option"); CEException::halt();
			}
		}
	}

	// 2. Compose diagrams (born, loops + brems)
	// Script: ComposeDiagrams
	CDiagramSet* diag_Born = NULL;
	CDiagramSet* diag_Loops = NULL;
	if(!bLoaded_res) {
		CMEComposeLib* libDiag = new CMEComposeLib();
		bOk = libDiag->ComposeME(iOrder); if(!bOk){ dbg->log(0, "Error creating Born-level diagrams"); CEException::halt("Unexpected error"); delete libDiag; return false; }
		diag_Born = (libDiag->GetDiagramsCreated())->clone(); delete libDiag; libDiag=NULL;

		if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_RC) {
			libDiag = new CMEComposeLib(); bOk = libDiag->ComposeME(iOrder+2); if(!bOk){ dbg->log(0, "Error creating loop diagrams"); CEException::halt("Unexpected error"); delete libDiag; return false; }
			diag_Loops = (libDiag->GetDiagramsCreated())->clone(); delete libDiag; libDiag=NULL;
		}
		//TODO: set brems kinematics + CDiagramSet* diag_Brems=NULL;
		if(diag_Born==NULL){ CEBasic::DeleteObj(diag_Born,diag_Loops); return false; } //e.g. forbidden channel
		if(diag_Born->get_diagrams_number()<=0){ CEBasic::DeleteObj(diag_Born,diag_Loops); return false; }	//e.g. forbidden channel
	}else{
		//also try load diagrams for proper TeX output file
		//if(m_setts.m_iFlags & CLHEPRunProject::flag_RUN_LOAD) {}
		if(resBorn!=NULL) {
			sprintf(buf,"%s_diagsB.xml",sIdSub); CESerializable* serTmp = CESupport::read_file(buf, 0);
			if(serTmp!=NULL) { if(CEUtil::equals(serTmp->ClassName(), CDiagramSet::CLASS_NAME)) { diag_Born=(CDiagramSet*)serTmp; }else{ delete diag_Born; } }
			//asdf
			//[debug]if(diag_Born!=NULL) { sprintf(buf,"%s_diagsB_test.xml",sIdSub); CESupport::store(diag_Born,buf,0); }
		}
		if(resLoop!=NULL) {
			sprintf(buf,"%s_diagsL.xml",sIdSub); CESerializable* serTmp = CESupport::read_file(buf, 0);
			if(serTmp!=NULL) { if(CEUtil::equals(serTmp->ClassName(), CDiagramSet::CLASS_NAME)) { diag_Loops=(CDiagramSet*)serTmp; }else{ delete diag_Born; } }
			//[debug]if(diag_Loops!=NULL) { sprintf(buf,"%s_diagsL_test.xml",sIdSub); CESupport::store(diag_Loops,buf,0); }
		}
	}

	//Nt: ToDelete: CEBasic::DeleteObj(diag_Born,diag_Loops);

	// 3. Retrieve matrix element
	// Script: RetrieveME
	CMatrixElement* me_Born = NULL;
	CMatrixElement* me_Loop = NULL;
	if(!bLoaded_res) {
		me_Born = diag_Born->RetrieveMatrixElement();
		if(me_Born==NULL){ dbg->log(0,"[ERROR] RetrieveMatrixElement(): NULL"); CEException::halt("[ERROR] RetrieveMatrixElement(): NULL"); CEBasic::DeleteObj(diag_Born,diag_Loops,me_Born); return false; }

		if(diag_Loops!=NULL){
			me_Loop = diag_Loops->RetrieveMatrixElement();
			//meL = RetrieveME(SelectDiagrams(diags,SEL_LOOPS)); //get matrix element
			//meCT = RetrieveME(SelectDiagrams(diags,SEL_CT)); //get matrix element
			if(me_Loop==NULL){ dbg->log(0,"[ERROR] RetrieveMatrixElement(Loops): NULL"); CEException::halt("[ERROR] RetrieveMatrixElement(Loops): NULL"); CEBasic::DeleteObj(diag_Born,diag_Loops,me_Born,me_Loop); return false; }
		}
	}

	//Nt: ToDelete: CEBasic::DeleteObj(diag_Born,diag_Loops,me_Born,me_Loop);

	// 4. Square matrix element / calculate amplitude / calculate loops / get bremsstrahlung
	//
	// Loops:
	//mesqr = SquareME(me, meB, SQR_DEFAULT);
	//Kin-Arrange automatic here?
	////mesqr = KinArrange(mesqr);
	//mesqr = ReduceTensorVI(mesqr);
	//mesqr = KinArrange(mesqr);
	//SaveNB(nbfile, mesqr, "VI evaluated");
	//mesqr = ConvertInvariantVI(mesqr);
	//mesqr = CalcScalarVI(mesqr);  // use pre-calculated values
	//mesqr = KinArrange(mesqr);
	//SaveNB(nbfile, mesqr, "VI scalars ");
	//mesqr = SingularArrange(mesqr);
	//SetNDimensionSpace(0);
	//mesqr = KinArrange(mesqr);
	////mesqr = KinSimplify(mesqr);
	//
	// Bremsstrahlung:
	//mesqr=GetSoftBremsstrahlung();
	//SaveNB(nbfile, mesqr, "Soft factor");
	//
	DWORD iKASetts = 0; if(CEProcessor::m_MandelstamsEverywhere){ iKASetts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
	DWORD iKSSetts = 0; if(CEProcessor::m_MandelstamsEverywhere){ iKSSetts |= CLHEPBasics::KS_INVOLVE_MANDELSTAM_VARS; }
	//
	if(!bLoaded_res) {
		if(bCalcMESqr){
			//Script: SquareME(me, SQR_DEFAULT);
			DWORD sqr_setts =	CMESquareLib::SQRME_EXPAND_AND_SPLIT |
								CMESquareLib::SQRME_PRESIMPLIFY_ME |			//mandatory (too slow otherwise)
								CMESquareLib::SQRME_CACHE_DENOMINATORS | 
								CMESquareLib::SQRME_EARLY_PHOTON_HELICITIES;	//asdf: otherwise errors?! !!!!! TODO: fix it!
			if(CEProcessor::m_MandelstamsEverywhere){ sqr_setts |= CMESquareLib::SQRME_ALLOW_MANDELSTAMS; }

			//Born amplitude squared
			resBorn = CMESquareLib::SquareME(me_Born, NULL, sqr_setts);
			//Script: KinArrange() or 'CEProcessor::m_bAutoKinArrange'
			resBorn = CLHEPBasics::KinArrange(resBorn, iKASetts);
			if(nFinals==2){
				resBorn = CLHEPBasics::KinSimplify(resBorn, iKSSetts);
			}

			//Loop*Conj[Born]
			CExpression* meSqr = NULL;
			if(me_Loop!=NULL){
				CETensor::setNDimensionSpace(true);
				resLoop = CMESquareLib::SquareME(me_Loop, me_Born, sqr_setts);	//we want 'me_Born' to be conjugated here
				resLoop = CLHEPBasics::KinArrange(resLoop, iKASetts);

				//Reduce tensor VI
				resLoop = CVirtIntegralLib::ReduceTensorVI(resLoop);
				resLoop = CLHEPBasics::KinArrange(resLoop, iKASetts);
				resLoop = CLHEPBasics::KinSimplify(resLoop, iKSSetts);

				//Convert VI to invariant form
				resLoop = CVirtIntegralLib::ConvertVI0ToInvForm(resLoop);

				//Calculate scalar VI (use pre-calculated values)
				resLoop = CVIScalarCalcLib::CalcScalarVI(resLoop);
				resLoop = CLHEPBasics::KinArrange(resLoop, iKASetts);
				//Note: unknown (complicated) scalar VIs still remain

				//\Xi*B0(t,0,m) ==> -1 ect.
				resLoop = CLHEPBasics::SingularArrange(resLoop);
				CETensor::setNDimensionSpace(false);
				resLoop = CLHEPBasics::KinArrange(resLoop, iKASetts);
				resLoop = CLHEPBasics::KinSimplify(resLoop, iKSSetts);

				//Restore renormalization constants
				//TODO: separate Loops & CT calculations by using meCT & meLoops
				CERule* r = m_phys->getRenormConstValues();
				resLoop = r->ApplyTo(resLoop);
				resLoop = CExpression::arrge(resLoop);

				//asdf

				//add R-soft

				//asdf
			}
		}else{
			// Script: ampl = CalcAmplitude(meA);
			DWORD fCalcAmplSetts = 0; if(CEProcessor::m_MandelstamsEverywhere){ fCalcAmplSetts |= CAmplCalcLib::A_MANDELSTAMS; }
			resBorn = CAmplCalcLib::CalcAmplitude(me_Born, fCalcAmplSetts);

			//Script: KinArrange() or 'CEProcessor::m_bAutoKinArrange'
			resBorn = CLHEPBasics::KinArrange(resBorn, iKASetts);
		}
	}
	//Nt: ToDelete: CEBasic::DeleteObj(diag_Born,diag_Loops,me_Born,me_Loop,resBorn,resLoop);

	// 4.1. Zero check
	if(!bLoaded_res) {
		if((resBorn==NULL) || (CEBasic::isZero(resBorn))){ CEBasic::DeleteObj(diag_Born,diag_Loops,me_Born,me_Loop,resBorn,resLoop); return false; }
		//Note: we remain 'resLoop' as '0' until saving
	}

	// 5. Simplify results
	if(!bLoaded_res) {
		//
	}
	//

	// 5.1. Zero check (again)
	if(!bLoaded_res) {
		if((resBorn==NULL) || (CEBasic::isZero(resBorn))){ CEBasic::DeleteObj(diag_Born,diag_Loops,me_Born,me_Loop,resBorn,resLoop); return false; }
		//Note: we remain 'resLoop' as '0' until saving
	}

	// 5.2. Now the sub-process is 'approved', the ID for the process is 'm_infConfig.m_nSubProcs'
	char bufB[256]; char bufV[256];
	char bufNB[256]; sprintf(bufNB,"%s.nb",sIdSub);
	if(!bLoaded_res) {
		CESupport::SaveHeadersToNB(bufNB,CESupport::FILE_TRUNCATE);

		char* tmpKin1=m_kin->toLine(STORE_WR_MATHEMATICA);
		if(tmpKin1!=NULL){
			bOk=CESupport::SaveTextToNB(bufNB,"Kinematics",0); if(!bOk)dbg->log(0,"Error writing NB file");
			bOk=CESupport::SaveTextToNB(bufNB,tmpKin1,0); if(!bOk)dbg->log(0,"Error writing NB file");
			CMemoryManager::mfree(tmpKin1);
		}
	}

	// 5.5.1. Draw diagrams
	// Script: DrawDiagrams(diags, texfile[, DD_DEFAULT|DD_DONT_SWAP_TALES]);
	int fDrawDiag =	CFeynDiagram::DRAW_CAPS_SMALL|CFeynDiagram::SET_LOOP_120|
					CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES|CFeynDiagram::DRAW_DIAGRAM_NUMBERS;
	if (m_kin->num_final_particles() > 2) fDrawDiag |= CFeynDiagram::DRAW_SWAP_TALES;
	//User options: always 'CFeynDiagram::DRAW_DONT_SWAP_TALES' or 'CFeynDiagram::DRAW_SWAP_TALES'
	//if ((fDrawDiag & CFeynDiagram::DRAW_DONT_SWAP_TALES) == CFeynDiagram::DRAW_DONT_SWAP_TALES) {
	//	if ((fDrawDiag & CFeynDiagram::DRAW_SWAP_TALES) == CFeynDiagram::DRAW_SWAP_TALES) { fDrawDiag ^= CFeynDiagram::DRAW_SWAP_TALES; }
	//}
	bool bTeXMultiFiles = false;					//parameter
	const char* _fn=NULL; int fDrawFileSetts=0;		//CESupport::FILE_TRUNCATE | CESupport::FILE_CLOSE
	//if(!bLoaded_res) {
	{
		if(bTeXMultiFiles){
			sprintf(buf,"%s.tex",sIdSub); _fn=buf; fDrawFileSetts=CESupport::FILE_TRUNCATE;
		}else{
			_fn = m_sTexReport.GetBuffer();
			CEString sMsg("Sub-process no. "); sMsg.add_int(iProcID); sMsg.AddString(": ");
			char* tmp1=m_kin->toLine(STORE_TEX); if(tmp1!=NULL){ sMsg.AddString("$"); sMsg.AddString(tmp1); sMsg.AddString("$"); CMemoryManager::mfree(tmp1); }
			CESupport::saveTeX(_fn,sMsg.GetBuffer(),fDrawFileSetts,false); fDrawFileSetts=0;
		}
		if(diag_Born!=NULL){ //Nt: it may be ==NULL is me2 is pre-loaded
			diag_Born->SetDrawingFlags(fDrawDiag);
			bool bDiagLocateOK = diag_Born->LocateDiagramsView(CLDiagLocateView::ALGORITHM_DEFAULT);
			if(bDiagLocateOK){ CESupport::saveTeX(_fn,diag_Born,fDrawFileSetts); fDrawFileSetts=0; }
			else{ CESupport::saveTeX(buf,"ERROR: can't save diagrams, diag_Born->LocateDiagramsView() failed",fDrawFileSetts); fDrawFileSetts=0; dbg->log(0, "ERROR: diag_Born->LocateDiagramsView() == false"); }
		}
		if(diag_Loops!=NULL){ //Nt: it may be ==NULL is me2 is pre-loaded
			diag_Loops->SetDrawingFlags(fDrawDiag);
			bool bDiagLocateOK = diag_Loops->LocateDiagramsView(CLDiagLocateView::ALGORITHM_DEFAULT);
			if(bDiagLocateOK){ CESupport::saveTeX(_fn,diag_Loops,fDrawFileSetts); fDrawFileSetts=0; }
			else{ CESupport::saveTeX(buf,"ERROR: can't save diagrams, diag_Loops->LocateDiagramsView() failed",fDrawFileSetts); fDrawFileSetts=0; dbg->log(0, "ERROR: diag_Loops->LocateDiagramsView() == false"); }
		}
	}

	// 5.5.2. Save diagrams to 'xml'-file
	// Script: Save("diags.xml",diags); //save to XML file
	// Cross-check: diags=Load("diags.xml"); //load from XML file
	DWORD fSaveXMLSetts=0; //use or not 'CESupport::SERIAL_VERIFY', add 'CESupport::SERIAL_EXPRESSION' for expressions
	if(!bLoaded_res) {
		if(diag_Born!=NULL){ sprintf(buf,"%s_diagsB.xml",sIdSub); CESupport::store(diag_Born,buf,fSaveXMLSetts); }
		if(diag_Loops!=NULL){ sprintf(buf,"%s_diagsL.xml",sIdSub); CESupport::store(diag_Loops,buf,fSaveXMLSetts); }
	}

	// 5.5.3. Destroy diag_XXX objects
	if(diag_Born!=NULL){ delete diag_Born; diag_Born=NULL; }
	if(diag_Loops!=NULL){ delete diag_Loops; diag_Loops=NULL; }
	//Nt: ToDelete: CEBasic::DeleteObj(me_Born,me_Loop,resBorn,resLoop);

	// 5.5.4. Save 'me_XXX' to 'nb'-files and destroy
	// Script: SaveNB(nbfile, me, "Matrix element");
	if(!bLoaded_res) {
		if(me_Born!=NULL){
			bOk = CESupport::SaveTextToNB(bufNB,"Born matrix element",0); if(!bOk)dbg->log(0,"Error writing NB file");
			bOk = CESupport::SaveExprToNB(bufNB,me_Born,0); if(!bOk)dbg->log(0,"Error writing NB file");
		}
		if(me_Loop!=NULL){
			bOk = CESupport::SaveTextToNB(bufNB,"Loop matrix element",0); if(!bOk)dbg->log(0,"Error writing NB file");
			bOk = CESupport::SaveExprToNB(bufNB,me_Loop,0); if(!bOk)dbg->log(0,"Error writing NB file");
		}
	}
	if(me_Born != NULL){ delete me_Born; me_Born=NULL; }
	if(me_Loop != NULL){ delete me_Loop; me_Loop=NULL; }
	//Nt: ToDelete: CEBasic::DeleteObj(resBorn,resLoop), ToDo: write 'bufNB' file closing header

	// 6. Minimize expressions
	// Script: mesqr=Minimize(mesqr);
	DWORD fMinimizeSetts = CViewSimplify::PART_FACT_DEFAULT;// | CViewSimplify::PART_FACT_VERIFY;
	if(!bLoaded_res) {
		resBorn = CViewSimplify::factorize_partial(resBorn, fMinimizeSetts);
		if((resBorn==NULL) || (CEBasic::isZero(resBorn))){ //NULL/Zero checks -> now it's an error
			dbg->log(0,iProcID,"Unexpected NULL/ZERO after minimize in sub-process %ld",resBorn);
			CESupport::SaveTextToNB(bufNB,"Unexpected NULL/ZERO after minimize",CESupport::FILE_CLOSE);
			CEBasic::DeleteObj(resBorn,resLoop); return false;
		}
		if(resLoop != NULL){
			resLoop = CViewSimplify::factorize_partial(resLoop,CViewSimplify::PART_FACT_DEFAULT);
		}
	}

	// 6.1. Save result
	// Script: Save("mesqr.xml", mesqr);
	if(!bLoaded_res) {
		sprintf(buf,"%s_Minimized.xml",sIdSub); CESupport::store(resBorn,buf,fSaveXMLSetts|CESupport::SERIAL_EXPRESSION);
		if(resLoop!=NULL){ sprintf(buf,"%sV_Minimized.xml",sIdSub); CESupport::store(resLoop,buf,fSaveXMLSetts|CESupport::SERIAL_EXPRESSION); }
		// Script: SaveNB(nbfile, mesqr, "Minimized");
		bOk = CESupport::SaveTextToNB(bufNB,(bCalcMESqr?"ME^2, minimized":"Amplitude, minimized"),0); if(!bOk)dbg->log(0,"Error writing NB file");
		bOk = CESupport::SaveExprToNB(bufNB,resBorn,0); if(!bOk)dbg->log(0,"Error writing NB file");
		if(resLoop!=NULL) {
			bOk = CESupport::SaveTextToNB(bufNB,(bCalcMESqr?"2Re[ME{loop}*Cjg{Me}], minimized":"E{loop}, minimized"),0); if(!bOk)dbg->log(0,"Error writing NB file");
			bOk = CESupport::SaveExprToNB(bufNB,resLoop,0); if(!bOk)dbg->log(0,"Error writing NB file");
		}
	}

	// 6.2 Check Loop term for zero (after save/load operations!)
	if(resLoop != NULL) { if(CEBasic::isZero(resLoop)) { delete resLoop; resLoop=NULL; } }

	// 7. Create code
	// Script: SetPolarized(-1, 0); SetPolarized(-2, 0); SetPolarized(-3, 0); // set unpolarized
	//
	// Script: SetParameterS(PAR_STR_FORTRAN_TEMP, "TMP1");
	if(m_setts.m_bCodeCpp){ sprintf(buf,"tmp"); }else{ sprintf(buf,"tmp%ld",iProcID); } CFFortran::SetTEMPstr(buf);
	// Script: f = NewCodeFile("res.cpp", CODE_CPP); //or CODE_F77
	sprintf(buf,((m_setts.m_bCodeCpp)?"%s.cpp":"%s.f77"),sIdSub);
	DWORD fCodeSetts = ((m_setts.m_bCodeCpp)?CLFileManager::FT_CPP:CLFileManager::FT_F77);
	if(m_setts.m_bUseAbstractTypes)fCodeSetts|=CLFileManager::FT_CPP_TYPE_ABSTRACT;
	int idCodeFile = CLFileManager::new_file(buf, fCodeSetts);
	if (idCodeFile < 0) {
		dbg->log(0,iProcID,"Error creation code file for sub-process: %ld");
		if(!bLoaded_res) {
			CESupport::SaveTextToNB(bufNB,"Error creation code file for the sub-process",CESupport::FILE_CLOSE);
		}
		CEBasic::DeleteObj(resBorn,resLoop); return false;
	}
	// Script: CreateCodeProc(f, "BHRes4", mesqr, CODE_XXX|CODE_YYY...);
	sprintf(buf,"%s",sIdSub);											//function title (currently same to 'sIdSub')
	if(resLoop != NULL) { sprintf(bufB,"%sB",sIdSub); }					//Born function title
	DWORD fCodeSettsAdv = 0;
	if(!bCalcMESqr)fCodeSettsAdv |= CLFileManager::FT_SRC_IS_AMPLITUDE;	//square amplitude
	fCodeSettsAdv |= CLFileManager::FT_CHECK_DENOMINATORS;				//check 1/0 limits
	fCodeSettsAdv |= CLFileManager::FT_REAL_8BIT;						//real values
	//fCodeSettsAdv |= CLFileManager::FT_COMPLEX_16BIT;					//complex values
	fCodeSettsAdv |= CLFileManager::FT_POWER_PARAMS;					//F(M^2) instead of F(M)
	//[no!]fCodeSettsAdv |= CLFileManager::FT_PYTHIA_VECTORS;			//use PYTHIA PUP(I,J) vectors
	//TODO:remove next line if LoopTools is included in 'vmlib.h'
	if(m_infConfig.m_bAllowQuarksSwap && (nInits==2) && (pIn1!=NULL) && (pIn2!=NULL) && (pIn1->get_PK()!=pIn2->get_PK())){
		fCodeSettsAdv |= CLFileManager::FT_ALLOW_SWAP_INITIALS;
	}
	fCodeSettsAdv |= CLFileManager::FT_NON_STATIC_IMPL;
	CLFileManager::add_to_file(idCodeFile, ((resLoop!=NULL)?bufB:buf), resBorn, fCodeSettsAdv);
	dbg->log(8,"Process implementation class created, file: ",buf);
	if(resLoop != NULL) {
		sprintf(bufV,"%sV",sIdSub);										//Loop function title (currently same to 'sIdSubL')
		fCodeSettsAdv |= CLFileManager::FT_LOOPTOOLS;					//add LoopTools header (will be added once per file)
		CLFileManager::add_to_file(idCodeFile, bufV, resLoop, fCodeSettsAdv);
		dbg->log(8,"Process V-term implementation class created, file: ",bufV);
	}

	// Add process class information
	if(resLoop!=NULL){
		ProcCodeLib()->AddSubProcess(buf,idCodeFile,bufB,bufV);
	}else{
		ProcCodeLib()->AddSubProcess(buf,idCodeFile);
	}

	if(m_setts.m_bSeparateChannelsAsProcs){
		CEString sProcClassName; sProcClassName<<"CProc"<<buf;
		ProcCodeLib()->SaveProcessClass(buf,idCodeFile,sProcClassName.GetBuffer(),false);
		m_infConfig.m_nSubProcsActual += ProcCodeLib()->GetNSubsActual();
		ProcCodeLib()->Reset();

		//
		m_infConfig.m_subProcClassHeaders.add(new CEString(CLFileManager::GetHeaderFileName(idCodeFile)));
		m_infConfig.m_subProcClassNames.add(new CEString(sProcClassName));
	}

	////////////////////////////////////////////////////////////////////////////////////////////////
	// finalize

	// save masses of initial particles
	// TODO1: not required for partons (they are massless) --> skip them
	// TODO2: remove const-include-engine from non-statics of CFFortran to allow move these lines to where-using place!
	//
	if(m_infConfig.m_fnConstInclude.length()<=0)m_infConfig.m_fnConstInclude.AddString(CLFileManager::GetConstIncludeFileName(idCodeFile));
	if(m_infConfig.m_symMassIn[0].length()<=0){
		m_infConfig.m_symMassIn[0].AddString(CLFileManager::RequestParticleMassIncluded(idCodeFile,m_infConfig.m_pkInitials[0]));
	}
	if(m_infConfig.m_symMassIn[1].length()<=0){
		m_infConfig.m_symMassIn[1].AddString(CLFileManager::RequestParticleMassIncluded(idCodeFile,m_infConfig.m_pkInitials[1]));
	}

	// Finalize file
	CLFileManager::close_file(idCodeFile);

	//Nt: ToDelete: CEBasic::DeleteObj(resBorn), ToDo: write 'bufNB' file closing header
	if(resBorn!=NULL){ delete resBorn; resBorn=NULL; }
	if(resLoop!=NULL){ delete resLoop; resLoop=NULL; }
	if(!bLoaded_res) {
		CESupport::SaveHeadersToNB(bufNB,CESupport::FILE_CLOSE); //don't touch NB file if it was not started
	}
	return true;
}

void CLHEPRunProject::CreateProcFile() {
	//TODO: iterate all kinematic configs

	//Nt: don't use 'm_kin' here! (elastic 'kin' is installed during R-term saving)

	//dbg->log(0,ProcCodeLib()->GetNSubs(),"CreateProcFile. ProcCodeLib()->GetNSubs()=%ld");
	DWORD fCodeSetts = (((m_setts.m_bCodeCpp)?CLFileManager::FT_CPP:CLFileManager::FT_F77)
		| ((m_setts.m_bUseAbstractTypes)?(CLFileManager::FT_CPP_TYPE_ABSTRACT):0));
	//

	bool bSingleSubFile = (ProcCodeLib()->GetNSubs()==1);
	if(bSingleSubFile)
		// the only inplementation class --> add to same file
	{
		int idCodeFile = CLFileManager::open_file(ProcCodeLib()->SubProc(0)->sImplFileName.GetBuffer(),fCodeSetts,true);
		if (idCodeFile >= 0) {
			CEString sProcClass; sProcClass<<"CProc"<<ProcCodeLib()->SubProc(0)->sId.GetBuffer();
			ProcCodeLib()->SaveProcessClass(ProcCodeLib()->SubProc(0)->sId.GetBuffer(),idCodeFile,sProcClass.GetBuffer(),false);
			m_infConfig.m_subProcClassHeaders.add(new CEString(CLFileManager::GetHeaderFileName(idCodeFile)));
			m_infConfig.m_subProcClassNames.add(new CEString(sProcClass));
			CLFileManager::close_file(idCodeFile);
		}else{ dbg->log(0,"Error opening code file for process: ",ProcCodeLib()->SubProc(0)->sImplFileName.GetBuffer()); }
	}
	else if(ProcCodeLib()->GetNSubs()>=1) // several implementation classes --> add to separate file
	{
		CEString idProc(ProcCodeLib()->GetProcTitle());
		CEString fnProcClass(idProc); fnProcClass<<"Proc.cpp";
		CEPtrArray sPreIncludes;
		for(int i=0; i<ProcCodeLib()->GetNSubs(); i++) {
			CEString* sHdr = &ProcCodeLib()->SubProc(i)->sImplFileNameHeader; if(sHdr->length()>0){ sPreIncludes.add(sHdr->clone()); }
		}
		int idCodeFile = CLFileManager::new_file(fnProcClass.GetBuffer(), fCodeSetts, ((sPreIncludes.size()>0)?(&sPreIncludes):NULL));
		//dbg->log(0,"CreateProcFile. file created: ",fnProcClass.GetBuffer());
		if (idCodeFile >= 0) {
			CEString sProcClass; sProcClass<<"CProc"<<idProc.GetBuffer();
			ProcCodeLib()->SaveProcessClass(idProc.GetBuffer(),idCodeFile,sProcClass.GetBuffer(),false);
			for(int i=0; i<sPreIncludes.size(); i++) {
				CEObject* s=sPreIncludes.GetAt(i);
				if(s->GetObjectKind()==CEObject::KIND_STRING){ m_infConfig.m_subProcClassHeaders.add(((CEString*)s)->clone()); }
			}
			m_infConfig.m_subProcClassHeaders.add(new CEString(CLFileManager::GetHeaderFileName(idCodeFile)));
			m_infConfig.m_subProcClassNames.add(new CEString(sProcClass));
			CLFileManager::close_file(idCodeFile);
		}else{ dbg->log(0,"Error (1) creating code file for process: ",idProc.GetBuffer()); }
	} else {
		dbg->log(0,"Error (2) creating code file for process: ",ProcCodeLib()->GetProcTitle());
		dbg->log(0,ProcCodeLib()->GetNSubs(),"Invalid number of sub-processes: %ld");
	}
	m_infConfig.m_nSubProcsActual += ProcCodeLib()->GetNSubsActual();
	ProcCodeLib()->Reset();
}

//
//	Create numerical integration/generation project
//	Note: is 'exit(-1)' is better than 'return -1' here?
//
void CLHEPRunProject::CreateMainFile() {
	//oprtions
	bool bLHEOutput=true;
	bool bPlotsE=true;
	bool bPlotsPt=true;
	bool bPlotsEta=true;
	bool bPlots = (bPlotsE || bPlotsPt || bPlotsEta);

	//
	CEString ln; CEString ln2; CEString ln3; int i; char tmpBuf[32];
	const char* varCS="cs";
	const char* varI="i"; bool bVarIDecl=false;
	int nFinals=m_kin->num_final_particles();
	int iFirstFinalPos = 2;	//id of the first final particle in event
	if(m_setts.m_bUseEvtNuclonEntries) { if(m_infConfig.m_bPartonsUsed_beam1)iFirstFinalPos++; if(m_infConfig.m_bPartonsUsed_beam2)iFirstFinalPos++; }

	ofstream ofs(m_sProjMainFile.GetBuffer()); if(!ofs){ dbg->log(0,"Can't create/open file ",m_sProjMainFile.GetBuffer()); return; }

	//Credits
	m_codeLang->StartCommentLine(&ofs); ofs<<"\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"Main entry for project '"<<m_sTitle.GetBuffer()<<"'\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"\n";
	CEString sVer; CESupport::CreateVersionString(&sVer);
	m_codeLang->StartCommentLine(&ofs); ofs<<"Created by: "<<sVer.GetBuffer()<<"\n";
	unsigned long lib_ver=CAlHepLib_VER;
	m_codeLang->StartCommentLine(&ofs); ofs<<"To be used with CAlHepLib library version: "<<((lib_ver>>16)&0xff)<<"."<<((lib_ver>>8)&0xff)<<".XX\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"Original path: "<<m_sProjPath.GetBuffer()<<"\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"\n"; ofs<<"\n";

	//Includes
	ln<<"#include <math.h>"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); //Or move it into 'vmlib.h'?
	ln<<"#include \""<<CCodeProcessImpl::FN_CPP_LIB_PATH<<CCodeProcessImpl::FN_CPP_LIB_FILE<<"\""; m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	if(m_infConfig.m_fnConstInclude.length()>0){
		ln<<"#include \""<<m_infConfig.m_fnConstInclude.GetBuffer()<<"\""; m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	}
	ofs<<"\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"Sub-processes:\n";
	int nHdrs=m_infConfig.m_subProcClassHeaders.size();
	for(i=0; i<nHdrs; i++) {
		CEString* hdrFileName = (CEString*)m_infConfig.m_subProcClassHeaders.GetAt(i);
		if(hdrFileName!=NULL) { ln<<"#include \""<<hdrFileName->GetBuffer()<<"\""; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); }
		else{ m_codeLang->StartCommentLine(&ofs,true); ofs<<"ERROR: Header file not found for sub-process #"<<i<<"\n"; }
	}
	m_codeLang->StartCommentLine(&ofs);
	ofs<<"Total "<<nHdrs<<" header file"<<((nHdrs!=1)?"s":"");
	if(nHdrs!=m_infConfig.m_nSubProcs){
		ofs<<" with "<<m_infConfig.m_subProcClassNames.size()<<" process"<<((m_infConfig.m_subProcClassNames.size()!=1)?"es":"")
			<<" and "<<m_infConfig.m_nSubProcsActual<<" sub-process"<<((m_infConfig.m_nSubProcsActual!=1)?"es":"");
	}
	ofs<<" used\n"; ofs<<"\n";

	//Entry point
	m_codeLang->StartCommentLine(&ofs); ofs<<"Main entry\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"Arguments:\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"  argv[1] - the uncertainty of integration (min and default: 0.01, less than 10^-5 is not recommended)\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"  argv[2] - number of events to generate\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"  argv[3] - random numbers subset [0..99], use whole set if not specified\n";
	ln<<"int main(int argc, char* argv[]) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent(); ofs<<"\n";

	//Check version
	m_codeLang->StartCommentLine(&ofs); ofs<<"Check library version\n";
	ln<<m_codeLang->GetOpIf()<<" (("<<CCodeProcessImpl::CN_CPP_LIB_CLASS<<"::GetVersion() >> 8) != 0x";
	char* bufTmp=(char*)CMemoryManager::mmalloc(20); if(bufTmp!=NULL){ sprintf(bufTmp, "%lx", ((unsigned long)CAlHepLib_VER>>8)); ln<<bufTmp; CMemoryManager::mfree(bufTmp); }
	ln<<" ) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"ERROR: Library version is not compartible. Use version #"<<(((unsigned long)CAlHepLib_VER>>16)&0xff)<<"."<<(((unsigned long)CAlHepLib_VER>>8)&0xff)<<".XX";
	m_codeLang->OutConsoleWriteLine(&ofs,&ln);
	ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);

	//Check linked packages
	if(m_infConfig.m_bPartonsUsed_beam1 || m_infConfig.m_bPartonsUsed_beam2) {
		ln<<"#ifndef "<<VM_USE_PYTHIA8; ofs<<ln.GetBuffer()<<"\n"; ln.Reset();
		ln<<"ERROR: PYTHIA8 must be linked. Set '"<<VM_USE_PYTHIA8<<"' flag in file '"<<CCodeProcessImpl::FN_CPP_LIB_PATH<<CCodeProcessImpl::FN_CPP_LIB_FILE<<"'";
		m_codeLang->OutConsoleWriteLine(&ofs,&ln);
		ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
		ln<<"#endif"; ofs<<ln.GetBuffer()<<"\n"; ln.Reset();
	}
	ofs<<"\n";

	//Select RND-bank
	m_codeLang->StartCommentLine(&ofs); ofs<<"Select random numbers set\n";
	ln<<m_codeLang->GetOpIf()<<"(argc>=4) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	m_codeLang->StartCommentLine(&ofs); ofs<<"Note: argv[3] must be within [0..(CAlHepLib::NUM_RND_BANKS-1)], by default [0..99]\n";
	ln<<m_codeLang->GetOpIf()<<"(!"<<CCodeProcessImpl::CN_CPP_LIB_CLASS<<"::SelectRndBank(argv[3])) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"Invalid RND bank number"; m_codeLang->OutConsoleWriteLine(&ofs,&ln);
	ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); ofs<<"\n";

	//Integration precision
	m_codeLang->StartCommentLine(&ofs); ofs<<"Precision of integration\n";
	ln<<"double fPrecision = 0.01"; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<m_codeLang->GetOpIf()<<"(argc>=2) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	m_codeLang->StartCommentLine(&ofs); ofs<<"Note: argv[1] is the uncertainty of integration (min==0.01, less than 10^-5 is not recommended)\n";
	ln<<"fPrecision = atof(argv[1]); "; ln<<m_codeLang->GetOpIf()<<"((fPrecision<=0.)||(fPrecision>0.01)) fPrecision = 0.01"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); ofs<<"\n";

	// Number of events
	m_codeLang->StartCommentLine(&ofs); ofs<<"Number of events to generate\n";
	const char* varNEvents = "NEvents"; ln<<m_codeLang->GetTypePrefixInt()<<" "<<varNEvents<<" = 100"; m_codeLang->OutCodeLine(&ofs,&ln,-1,false,"default number");
	ln<<m_codeLang->GetOpIf()<<"(argc>=3) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<varNEvents<<" = atol(argv[2]); "; ln<<m_codeLang->GetOpIf()<<"("<<varNEvents<<"<0)"<<varNEvents<<"=0"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); ofs<<"\n";

	//Init library
	m_codeLang->StartCommentLine(&ofs); ofs<<"Initialize & test library\n";
	ln<<m_codeLang->GetOpIf()<<"(!"<<CCodeProcessImpl::CN_CPP_LIB_CLASS<<"::Init()) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"ERROR: Library initialization failed."; m_codeLang->OutConsoleWriteLine(&ofs,&ln); ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	ln<<m_codeLang->GetOpIf()<<"(!"<<CCodeProcessImpl::CN_CPP_LIB_CLASS<<"::TestLibrary(true)) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"ERROR: Internal library test failed."; m_codeLang->OutConsoleWriteLine(&ofs,&ln);
	ln<<CCodeProcessImpl::CN_CPP_LIB_CLASS<<"::TestLibrary(false)"; m_codeLang->OutCodeLine(&ofs,&ln,-1,false,"repeat test in non-silent mode");
	ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); ofs<<"\n";

	//Local vars
	const char* varProc="proc";
	//const char* varProcV="procV";
	//const char* varProcR="procR";
	//asdfgh

	//Processes
	int nProcs=0;
	if(m_infConfig.m_subProcClassNames.size() <= 0){
		ln<<"ERROR. No sub-processes exist"; m_codeLang->OutConsoleWriteLine(&ofs,&ln); ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
		m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndProc(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); ofs.close(); return;
	}else if(m_infConfig.m_subProcClassNames.size() == 1){
		m_codeLang->StartCommentLine(&ofs); ofs<<"Process instance\n";
		CEString* procCN = (CEString*)(m_infConfig.m_subProcClassNames.GetAt(0)); if(procCN!=NULL){ ln<<procCN->GetBuffer(); }else{ ln<<"(!Error!)"; }
		ln<<" "<<varProc<<"(0)"; m_codeLang->OutCodeLine(&ofs,&ln); nProcs=1;
	}else{
		m_codeLang->StartCommentLine(&ofs); ofs<<"Process set\n";
		ln<<CCodeProcessImpl::CN_CHEPProcessLiteSet<<" "<<varProc; m_codeLang->OutCodeLine(&ofs,&ln);
		for(i=0;i<m_infConfig.m_subProcClassNames.size();i++){
			ln<<varProc<<".AddProcess(new ";
			CEString* procCN = (CEString*)(m_infConfig.m_subProcClassNames.GetAt(i)); if(procCN!=NULL){ ln<<procCN->GetBuffer(); }else{ ln<<"(!Error!)"; }
			ln<<"("<<i<<"))"; m_codeLang->OutCodeLine(&ofs,&ln); nProcs++;
		}
	}
	ofs<<"\n";

	//Process estimates
	m_codeLang->StartCommentLine(&ofs); ofs<<"Set expected values for cross section (if known), in pb\n";
	//
	if(ProcCodeLib()->m_bEstimatedIsStatic) {
		for(i=0;i<m_infConfig.m_subProcClassNames.size();i++){
			CEString* procCN = (CEString*)(m_infConfig.m_subProcClassNames.GetAt(i)); if(procCN==NULL)continue;
			ln<<procCN->GetBuffer()<<"::m_fCrossSectionExpected_pb = -1."; m_codeLang->OutCodeLine(&ofs,&ln);
		}
	}else{
		if(m_infConfig.m_subProcClassNames.size() == 1){
			ln<<varProc<<".m_fCrossSectionExpected_pb = -1."; m_codeLang->OutCodeLine(&ofs,&ln);
		}else{
			for(i=0;i<m_infConfig.m_subProcClassNames.size();i++) {
				ln<<varProc<<".GetProcess("<<i<<")->m_fCrossSectionExpected_pb = -1."; m_codeLang->OutCodeLine(&ofs,&ln);
			}
		}
	}
	ofs<<"\n";

	//Initial beams
	m_codeLang->StartCommentLine(&ofs); ofs<<"Initial beams\n";
	ln2<<m_phys->get_particle_symbolA(m_infConfig.m_pkInitials[0],CDiagramPhysics::SYMBOL_STYLE_TEX);
	ln<<m_codeLang->GetTypePrefixReal()<<" EIn1="; sprintf(tmpBuf,"%g",GetDefaultEnergy(m_infConfig.m_pkInitials[0])); ln<<tmpBuf;
	m_codeLang->OutCodeLine(&ofs,&ln,-1,false,ln2.GetBuffer()); ln2.Reset();
	ln2<<m_phys->get_particle_symbolA(m_infConfig.m_pkInitials[1],CDiagramPhysics::SYMBOL_STYLE_TEX);
	ln<<m_codeLang->GetTypePrefixReal()<<" EIn2="; sprintf(tmpBuf,"%g",GetDefaultEnergy(m_infConfig.m_pkInitials[1])); ln<<tmpBuf;
	m_codeLang->OutCodeLine(&ofs,&ln,-1,false,ln2.GetBuffer()); ln2.Reset();
	ln<<CCodeProcessImpl::CN_CPP_4VEC_CLASS<<" vec1(EIn1,0.,0.,";
	if(m_infConfig.m_symMassIn[0].length()>0){ln<<"sqrt(EIn1*EIn1-"<<m_infConfig.m_symMassIn[0].GetBuffer()<<"*"<<m_infConfig.m_symMassIn[0].GetBuffer()<<")";}else{ln<<"EIn1";}
	ln<<")"; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<CCodeProcessImpl::CN_CPP_4VEC_CLASS<<" vec2(EIn2,0.,0.,";
	if(m_infConfig.m_symMassIn[1].length()>0){ln<<"-sqrt(EIn2*EIn2-"<<m_infConfig.m_symMassIn[1].GetBuffer()<<"*"<<m_infConfig.m_symMassIn[1].GetBuffer()<<")";}else{ln<<"EIn2";}
	ln<<")"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->OutCodeLine(&ofs,&ln);
	ofs<<"\n";

	//Cuts
	m_codeLang->StartCommentLine(&ofs); ofs<<"Set cuts (TODO: modify them on your own needs)\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"ID of the first final particle in generated event is: "<<iFirstFinalPos<<"\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"Final particles:";
	for(i=0;i<nFinals;i++){ //View: {X+,X-} for most particles, but {q,qbar} for quarks
		ofs<<" "<<iFirstFinalPos+i<<":";
		CEString* symID=NULL; if(m_infConfig.m_symFinIDs.size()==nFinals){ symID=(CEString*)(m_infConfig.m_symFinIDs.GetAt(i)); }
		if(symID!=NULL){ ofs<<symID->GetBuffer(); }else{ ofs<<"Error"; }
	}
	ofs<<"\n";

	//asdfg: add E-cut on R-term!
	//
	//CCodeProcessImpl::CN_CHEPEvtLiteCutE
	for(i=0;i<nFinals;i++){
		CEParticle* p=m_kin->pfin(i); if(p==NULL)continue;
		if(!p->check_PK(CEParticle::PARTICLE_NEUTRINO,true)){
			m_codeLang->StartCommentLine(&ofs,true);
			ln<<varProc<<".AddCut(&"<<CCodeProcessImpl::CN_CHEPEvtLiteCutE<<"("<<iFirstFinalPos+i<<",1.))"; ofs<<ln.GetBuffer()<<";\n"; ln.Reset();
		}
	}
	//CCodeProcessImpl::CN_CHEPEvtLiteCutPt
	for(i=0;i<nFinals;i++){
		CEParticle* p=m_kin->pfin(i); if(p==NULL)continue;
		if(!p->check_PK(CEParticle::PARTICLE_NEUTRINO,true)){
			ln<<varProc<<".AddCut(&"<<CCodeProcessImpl::CN_CHEPEvtLiteCutPt<<"("<<iFirstFinalPos+i<<",1.))"; m_codeLang->OutCodeLine(&ofs,&ln);
		}
	}
	//CCodeProcessImpl::CN_CHEPEvtLiteCutEta
	for(i=0;i<nFinals;i++){
		CEParticle* p=m_kin->pfin(i); if(p==NULL)continue;
		if(!p->check_PK(CEParticle::PARTICLE_NEUTRINO,true)){
			ln<<varProc<<".AddCut(&"<<CCodeProcessImpl::CN_CHEPEvtLiteCutEta<<"("<<iFirstFinalPos+i<<",2.5))"; m_codeLang->OutCodeLine(&ofs,&ln);
		}
	}
	//CCodeProcessImpl::CN_CHEPEvtLiteCutDR
	ofs<<"\n";

	if(m_infConfig.m_bPartonsUsed_beam1 || m_infConfig.m_bPartonsUsed_beam2) {
		m_codeLang->StartCommentLine(&ofs); ofs<<"Override default PDFs (default: CTEQ5L from PYTHIA8)\n";
		if((m_infConfig.m_pkInitials[0] != m_infConfig.m_pkInitials[1])
			|| (!m_infConfig.m_bPartonsUsed_beam1) || (!m_infConfig.m_bPartonsUsed_beam2)) {
			if(m_infConfig.m_bPartonsUsed_beam1){
				m_codeLang->StartCommentLine(&ofs,true);
				ofs<<CCodeProcessImpl::CN_CPP_PDF<<"* pdf1 = "<<CCodeProcessImpl::CN_CPP_PDF<<"::Create(4, 46, "<<m_phys->GetPDGParticleCode(m_infConfig.m_pkInitials[0])<<");"; ofs<<"\n";
			}
			if(m_infConfig.m_bPartonsUsed_beam2){
				m_codeLang->StartCommentLine(&ofs,true);
				ofs<<CCodeProcessImpl::CN_CPP_PDF<<"* pdf2 = "<<CCodeProcessImpl::CN_CPP_PDF<<"::Create(4, 46, "<<m_phys->GetPDGParticleCode(m_infConfig.m_pkInitials[1])<<");"; ofs<<"\n";
			}
			m_codeLang->StartCommentLine(&ofs,true);
			ofs<<varProc<<".SetPDFs("; if(m_infConfig.m_bPartonsUsed_beam1)ofs<<"pdf1"; if(m_infConfig.m_bPartonsUsed_beam2){{if(m_infConfig.m_bPartonsUsed_beam1)ofs<<",";} ofs<<"pdf2";} ofs<<");"; ofs<<"\n";
			if(m_infConfig.m_bPartonsUsed_beam1){ m_codeLang->StartCommentLine(&ofs,true); ofs<<m_codeLang->GetOpIf()<<"(pdf1!=NULL) delete pdf1;"; ofs<<"\n"; }
			if(m_infConfig.m_bPartonsUsed_beam2){ m_codeLang->StartCommentLine(&ofs,true); ofs<<m_codeLang->GetOpIf()<<"(pdf2!=NULL) delete pdf2;"; ofs<<"\n"; }
		}else{
			//single pdf for both nuclons
			m_codeLang->StartCommentLine(&ofs,true);
			ofs<<CCodeProcessImpl::CN_CPP_PDF<<"* pdf = "<<CCodeProcessImpl::CN_CPP_PDF<<"::Create(4, 46, "<<m_phys->GetPDGParticleCode(m_infConfig.m_pkInitials[0])<<");"; ofs<<"\n";
			m_codeLang->StartCommentLine(&ofs,true);
			ofs<<m_codeLang->GetOpIf()<<"(pdf!=NULL) { "<<varProc<<".SetPDFs(pdf,pdf); delete pdf; }"; ofs<<"\n";
		}
		ofs<<"\n";
	}

	//Integrate cross section
	//TODO: iterate processes here
	const char* varProcOK = "bProcOK";
	m_codeLang->StartCommentLine(&ofs); ofs<<"Integrate cross section\n";
	ln<<m_codeLang->GetTypePrefixBool()<<" "<<varProcOK<<" = "<<varProc<<".Prepare(&vec1,&vec2,fPrecision)"; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<m_codeLang->GetOpIf()<<"(!"<<varProcOK<<") {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"Error integrating cross section"; m_codeLang->OutConsoleWriteLine(&ofs,&ln);
	ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	ln<<m_codeLang->GetTypePrefixReal()<<" "<<varCS<<"=-1."; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<m_codeLang->GetOpIf()<<"(!proc.GetCrossSection("<<varCS<<")) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"std::cout<<\"Error integrating cross section. Cross section (pb): \"; CVMUtil::Cout(&std::cout,"<<varCS<<"); std::cout<<std::endl; "; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<"return -1"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	ofs<<"\n";

	//Write table of processes here {Col1: 3, Col2: 43}
	//TODO: iterate processes (and channels?) here
	ln<<"std::cout<<\""; ln<<"---"; for(i=0;i<7;i++){ln<<"----------";} ln<<"\"<<std::endl"; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<"std::cout<<\"   Process:"; ln<<"  "; for(i=0;i<3;i++){ln<<"          ";} ln<<"Cross section (pb):\"<<std::endl"; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<"std::cout<<\""; ln2<<"1. ";
	if(m_infConfig.m_symFinIDs.size()==nFinals) {
		for(i=0;i<m_infConfig.m_symInIDs.size();i++) { ln2<<((CEString*)(m_infConfig.m_symInIDs.GetAt(i)))->GetBuffer()<<" "; }
		ln2<<"->";
		for(i=0;i<m_infConfig.m_symFinIDs.size();i++) { ln2<<" "<<((CEString*)(m_infConfig.m_symFinIDs.GetAt(i)))->GetBuffer(); }
		ln2<<" "; //mandatory space before numbers
	}else{ ln2<<"Error"; }
	ln<<ln2.GetBuffer(); for(i=0;i<(43-ln2.length());i++){ln<<" ";} ln<<"\""; ln2.Reset();
	ln<<"; "; //m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<"CVMUtil::Cout(&std::cout,"<<varCS<<"); std::cout<<std::endl"; m_codeLang->OutCodeLine(&ofs,&ln);
	//[old]ln<<"std::cout<<\"| Cross section integrated (pb): \"; CVMUtil::Cout(&std::cout,"<<varCS<<"); std::cout<<std::endl"; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<"std::cout<<\""; ln<<"---"; for(i=0;i<7;i++){ln<<"----------";} ln<<"\"<<std::endl"; m_codeLang->OutCodeLine(&ofs,&ln);
	ln<<m_codeLang->GetOpIf()<<"(("<<varCS<<"<=0.) && ("<<varNEvents<<">0)) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"Event generation cancelled due to ZERO cross section"; m_codeLang->OutConsoleWriteLine(&ofs,&ln);
	ln<<varNEvents<<" = 0"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	ofs<<"\n";

	//Generate events
	ln<<m_codeLang->GetOpIf()<<"("<<varNEvents<<"<=0) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"return 1"; m_codeLang->OutCodeLine(&ofs,&ln,-1,false,"skip event generation");
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); ofs<<"\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"Generate events\n";
	m_codeLang->StartCommentLine(&ofs); ofs<<"\n";
	const char* varIEvt = "n"; ln<<m_codeLang->GetTypePrefixInt()<<" "<<varIEvt; m_codeLang->OutCodeLine(&ofs,&ln);
	const char* varNEventsFailed = "NEventsFailed"; ln<<m_codeLang->GetTypePrefixInt()<<" "<<varNEventsFailed<<" = 0"; m_codeLang->OutCodeLine(&ofs,&ln);
	const char* varEvt = "evt"; ln<<CCodeProcessImpl::CN_CPP_EVT_CLASS<<"* "<<varEvt<<" = new "<<CCodeProcessImpl::CN_CPP_EVT_CLASS<<"()"; m_codeLang->OutCodeLine(&ofs,&ln);
	const char* varLHE = "lhe";
	if(bLHEOutput){
		m_codeLang->StartCommentLine(&ofs); ofs<<"Init LHE file\n";
		ln<<CCodeProcessImpl::CN_CHEPEvtLiteLHE<<" "<<varLHE;
		ln<<"("<<m_phys->GetPDGParticleCode(m_infConfig.m_pkInitials[0])<<","<<m_phys->GetPDGParticleCode(m_infConfig.m_pkInitials[1])<<",EIn1,EIn2,";
		ln<<"0,0,0,0,"; //pdfs
		ln<<3; //'3', !=varCS!
		ln<<","<<m_infConfig.m_subProcClassNames.size()<<")"; m_codeLang->OutCodeLine(&ofs,&ln);
		if(m_infConfig.m_subProcClassNames.size() == 1){
			ln<<varLHE<<".SetProcessInfo(&"<<varProc<<")"; m_codeLang->OutCodeLine(&ofs,&ln);
		}else{
			if(!bVarIDecl){ ln<<m_codeLang->GetTypePrefixInt()<<" "<<varI; m_codeLang->OutCodeLine(&ofs,&ln); bVarIDecl=true; }
			ln2<<varProc<<".GetProcessNum()";
			m_codeLang->AddOpForP1(&ln,varI,0,ln2.GetBuffer()); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent(); ln2.Reset();
			ln<<varLHE<<".SetProcessInfo("<<varProc<<".GetProcess("<<varI<<"))"; m_codeLang->OutCodeLine(&ofs,&ln);
			m_codeLang->DecIndent(); m_codeLang->AddOpEndFor(&ln); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
		}
	}
	//
	if(bLHEOutput){
		ln<<m_codeLang->GetOpIf()<<"(!"<<varLHE<<".StartLHEFile(\""<<m_sTitle.GetBuffer()<<".lhe\")) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
		ln<<"ERROR: Can't create output LHE file. Events will not be saved."; m_codeLang->OutConsoleWriteLine(&ofs,&ln);
		m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	}
	//
	CEString sR2R8conv1; CEString sR2R8conv2;
	if(!m_codeLang->GetTypeConversion(m_codeLang->GetDefaultRealTypeId(),CEAbstractFunction::PAR_REAL8, &sR2R8conv1,&sR2R8conv2)) {
			std::cerr<<"[Internal error] Can't find type conversion header"<<std::endl; sR2R8conv1.Reset(); sR2R8conv1<<"("; sR2R8conv2.Reset(); sR2R8conv2<<")";
	}
	if(bPlots) { m_codeLang->StartCommentLine(&ofs); ofs<<"Initialize histograms\n"; }
	if(bPlotsE) {
		ln<<CCodeProcessImpl::CN_CVMHistLite<<"* histE = new "<<CCodeProcessImpl::CN_CVMHistLite<<"["<<nFinals<<"]"; m_codeLang->OutCodeLine(&ofs,&ln);
		/* [fixed X-margins code]
		//ln<<m_codeLang->GetTypePrefix(CEAbstractFunction::PAR_REAL8)<<" EHistMax = "<<sR2R8conv1.GetBuffer()<<"(EIn1+EIn2)/2"<<sR2R8conv1.GetBuffer(); m_codeLang->OutCodeLine(&ofs,&ln);
		//for(i=0;i<nFinals;i++){
		//	CEParticle* p=m_kin->pfin(i); ln2<<"E_"<<((p!=NULL)?CDiagramPhysics::get_particle_symbolA(p->get_PK(),CDiagramPhysics::SYMBOL_STYLE_TEX):"Unknown");
		//	ln<<"histE["<<i<<"].Init(\""<<ln2.GetBuffer()<<"\",100,0.,EHistMax)"; ln2.Reset(); m_codeLang->OutCodeLine(&ofs,&ln);
		//}
		*/
		///* [auto X-margins code]
		for(i=0;i<nFinals;i++){
			CEString* symID=NULL; if(m_infConfig.m_symFinIDs.size()==nFinals){ symID=(CEString*)(m_infConfig.m_symFinIDs.GetAt(i)); }
			if(symID!=NULL){ ln2<<symID->GetBuffer(); }else{ ln2<<"Final_"<<(i+1); }
			ln<<"histE["<<i<<"].InitAutoX(\""<<ln2.GetBuffer()<<"\",100,(("<<varNEvents<<">100)?100:"<<varNEvents<<"))"; ln2.Reset(); m_codeLang->OutCodeLine(&ofs,&ln);
		}
		//*/
	}
	if(bPlotsPt) {
		ln<<CCodeProcessImpl::CN_CVMHistLite<<"* histPt = new "<<CCodeProcessImpl::CN_CVMHistLite<<"["<<nFinals<<"]"; m_codeLang->OutCodeLine(&ofs,&ln);
		for(i=0;i<nFinals;i++){
			CEString* symID=NULL; if(m_infConfig.m_symFinIDs.size()==nFinals){ symID=(CEString*)(m_infConfig.m_symFinIDs.GetAt(i)); }
			if(symID!=NULL){ ln2<<symID->GetBuffer(); }else{ ln2<<"Final_"<<(i+1); }
			ln<<"histPt["<<i<<"].InitAutoX(\""<<ln2.GetBuffer()<<"\",100,(("<<varNEvents<<">100)?100:"<<varNEvents<<"))"; ln2.Reset(); m_codeLang->OutCodeLine(&ofs,&ln);
		}
	}
	if(bPlotsEta) {
		ln<<CCodeProcessImpl::CN_CVMHistLite<<"* histEta = new "<<CCodeProcessImpl::CN_CVMHistLite<<"["<<nFinals<<"]"; m_codeLang->OutCodeLine(&ofs,&ln);
		for(i=0;i<nFinals;i++){
			CEString* symID=NULL; if(m_infConfig.m_symFinIDs.size()==nFinals){ symID=(CEString*)(m_infConfig.m_symFinIDs.GetAt(i)); }
			if(symID!=NULL){ ln2<<symID->GetBuffer(); }else{ ln2<<"Final_"<<(i+1); }
			ln<<"histEta["<<i<<"].InitAutoX(\""<<ln2.GetBuffer()<<"\",100,(("<<varNEvents<<">100)?100:"<<varNEvents<<"))"; ln2.Reset(); m_codeLang->OutCodeLine(&ofs,&ln);
		}
	}
	//
	m_codeLang->AddOpForP1(&ln,varIEvt,0,varNEvents); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<m_codeLang->GetOpIf()<<"(!"<<varProc<<".GenEvent("<<varEvt<<")) {"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true); m_codeLang->IncIndent();
	ln<<"ERROR: Event generation failed"; m_codeLang->OutConsoleWriteLine(&ofs,&ln);
	ln<<varNEventsFailed<<"++; "; ln<<m_codeLang->GetOpIf()<<"("<<varNEventsFailed<<">100)";
	ln<<"{ std::cout<<\"Too many events failed. Generation stopped.\"<<std::endl; break; }"; m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndIf(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true); ofs<<"\n";
	//
	if(bLHEOutput){
		ln<<varLHE<<".SaveEvent("<<varEvt<<")"; m_codeLang->OutCodeLine(&ofs,&ln,-1,false,"Store event to LHE file");
	}
	if(bPlots) { m_codeLang->StartCommentLine(&ofs); ofs<<"Fill histograms\n"; }
	if(bPlotsE){ for(i=0;i<nFinals;i++){ ln<<"histE["<<i<<"].Fill("<<sR2R8conv1.GetBuffer()<<varEvt<<"->m_Particles["<<iFirstFinalPos+i<<"].m_Mom.m_E"<<sR2R8conv2.GetBuffer()
		<<","<<sR2R8conv1.GetBuffer()<<varEvt<<"->wgt"<<sR2R8conv2.GetBuffer()<<")"; m_codeLang->OutCodeLine(&ofs,&ln); } }
	if(bPlotsPt){ for(i=0;i<nFinals;i++){ ln<<"histPt["<<i<<"].Fill("<<sR2R8conv1.GetBuffer()<<varEvt<<"->m_Particles["<<iFirstFinalPos+i<<"].m_Mom.GetPt()"<<sR2R8conv2.GetBuffer()
		<<","<<sR2R8conv1.GetBuffer()<<varEvt<<"->wgt"<<sR2R8conv2.GetBuffer()<<")"; m_codeLang->OutCodeLine(&ofs,&ln); } }
	if(bPlotsEta){ for(i=0;i<nFinals;i++){ ln<<"histEta["<<i<<"].Fill("<<sR2R8conv1.GetBuffer()<<varEvt<<"->m_Particles["<<iFirstFinalPos+i<<"].m_Mom.GetEta()"<<sR2R8conv2.GetBuffer()
		<<","<<sR2R8conv1.GetBuffer()<<varEvt<<"->wgt"<<sR2R8conv2.GetBuffer()<<")"; m_codeLang->OutCodeLine(&ofs,&ln); } }
	m_codeLang->DecIndent(); m_codeLang->AddOpEndFor(&ln); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	//
	if(bLHEOutput){
		ln<<varLHE<<".CloseLHEFile()"; m_codeLang->OutCodeLine(&ofs,&ln);
	}
	ln<<"delete "<<varEvt; m_codeLang->OutCodeLine(&ofs,&ln); ofs<<"\n";

	//Write generation statistics
	ln<<varProc<<".PrintGeneratorStatistics()"; m_codeLang->OutCodeLine(&ofs,&ln); ofs<<"\n";

	//Write histograms
	if(bPlots) {
		m_codeLang->StartCommentLine(&ofs); ofs<<"Save/view histograms\n";
		ln<<CCodeProcessImpl::CN_CVMOutputSettings<<"::m_bSaveOutputTime = false"; m_codeLang->OutCodeLine(&ofs,&ln);
		ln<<CCodeProcessImpl::CN_CVMHistLiteViewDef<<"::m_bPlotBins = true"; m_codeLang->OutCodeLine(&ofs,&ln);
		ln<<CCodeProcessImpl::CN_CVMHistLiteView<<" hView"; m_codeLang->OutCodeLine(&ofs,&ln);
	}
	if(bPlotsE) {
		for(i=0;i<nFinals;i++){
			CEString* symID=NULL; if(m_infConfig.m_symFinIDs.size()==nFinals){ symID=(CEString*)(m_infConfig.m_symFinIDs.GetAt(i)); }
			ln<<"hView.AddPlotToView(&histE["<<i<<"])";
			ln<<"; hView.SaveAsEPS(\"histE_"; if(symID!=NULL){ ln<<symID->GetBuffer(); }else{ ln<<(i+1); } ln<<".eps\")";
			ln<<"; hView.Clear()"; m_codeLang->OutCodeLine(&ofs,&ln);
		}
		ln<<"delete[] histE"; m_codeLang->OutCodeLine(&ofs,&ln);
	}
	if(bPlotsPt) {
		for(i=0;i<nFinals;i++){ ln<<"hView.AddPlotToView(&histPt["<<i<<"])"; m_codeLang->OutCodeLine(&ofs,&ln); }
		ln<<"hView.SaveAsEPS(\"histPt.eps\"); hView.Clear(); delete[] histPt"; m_codeLang->OutCodeLine(&ofs,&ln);
	}
	if(bPlotsEta) {
		for(i=0;i<nFinals;i++){
			CEString* symID=NULL; if(m_infConfig.m_symFinIDs.size()==nFinals){ symID=(CEString*)(m_infConfig.m_symFinIDs.GetAt(i)); }
			ln<<"hView.AddPlotToView(&histEta["<<i<<"])";
			ln<<"; hView.SaveAsEPS(\"histEta_"; if(symID!=NULL){ ln<<symID->GetBuffer(); }else{ ln<<(i+1); } ln<<".eps\")";
			ln<<"; hView.Clear()"; m_codeLang->OutCodeLine(&ofs,&ln);
		}
		ln<<"delete[] histEta"; m_codeLang->OutCodeLine(&ofs,&ln);
	}

	ln<<"return 1"; m_codeLang->OutCodeLine(&ofs,&ln);
	m_codeLang->DecIndent(); ln<<m_codeLang->GetOpEndProc(); m_codeLang->OutCodeLine(&ofs,&ln,-1,true);
	ofs.close();
}

void CLHEPRunProject::CreateMakefile() {
	/*
	*	Notes
	*	=====
	*		Possible additional features:
	*			- Script for compiling pjoject, QD and LoopTools in single procedure (see 'bhmake'/'bhmake-q' in 'vm_bh'-project);
	*			- LoopTools with QUAD-precision (see 'Makefile-q'/'bhmake-q' in 'vm_bh'-project);
	*			- Dummy 'Makefile' for 'clean' operation only + 'Makefile-Proj' for script-driven use.
	*/

	const char* sTitle=m_sTitle.GetBuffer();
	CEString sProj; sProj<<"PROJ_"<<sTitle;
	CEString sMakeScriptName; sMakeScriptName<<"make-"<<sTitle;
	CEString sMakefileName; sMakefileName<<"Makefile-"<<sTitle;
	CEString sLogo; CESupport::CreateVersionString(&sLogo); if(sLogo.length()<=0)sLogo<<"ALHEP";
	bool bUseQD = (m_setts.m_bUseAbstractTypes);
	bool bUseLT = false;
	const char* varPathQD="PATH_QD";
	const char* varCodePrec="QD_CODE_PREC";
	const char* varPathLT="PATH_LOOPTOOLS";
	const char* varLTLibs="LOOPTOOLS_LIBS";	//Additional libraries used by LoopTools (collected by LT Makefile & ported here)
	const char* varHostType="PROJ_HOSTTYPE";
	const char* varBinary="PROJ_BINARY";
	const char* varPathProj="PROJ_PATH";
	char cEol=0x0a; const char* sOffset="  ";
	std::ofstream ofs;

	//	1. Create script for compiling pjoject, QD and LoopTools in single procedure
	//
	//	Note: script is to be modified by user to adjust QD,LoopTools,ROOT... versions and locations
	//
	//	Parameters:
	//	if(bUseLT || bUseQD){
	//		#1: '-all' (recompile all dependencies: QD and LT) or '-fast' (recompile local project only)
	//	}
	//
	//	exports variables to Makefile:
	//		'varBinary'
	//		'varHostType'	= x86_64-linux, = i386-linux
	//		'varCodePrec'
	//		'varPathLT'		= ./LoopTools-2.X
	//		'varLTLibs'		= Additional libraries used by LoopTools (collected by LT Makefile & ported here)
	//		'varPathQD'		= ./qd-2.3.8
	ofs.open(sMakeScriptName.GetBuffer(), std::ios::out|std::ios::trunc|std::ios::binary);
	if(ofs){
		ofs<<"#!/bin/sh"<<cEol<<cEol;
		ofs<<"#"<<cEol; ofs<<"# Make script for "<<sTitle<<" project"<<cEol; ofs<<"# Generated by "<<sLogo.GetBuffer()<<cEol; ofs<<"#"<<cEol;
		int iParNo=0; int idParBinName=-1; int idParAll=-1; int idParPrec=-1;
		if(bUseLT || bUseQD){
			iParNo++; idParAll=iParNo;
			ofs<<"# Parameter $"<<idParAll<<":"<<cEol;
			ofs<<"#     '-all': recompile all dependencies: QD and/or LoopTools"<<cEol;
			ofs<<"#     '-fast': recompile local project only"<<cEol;
		}
		if(bUseQD){
			iParNo++; idParPrec=iParNo;
			ofs<<"# Parameter $"<<idParPrec<<":"<<cEol;
			ofs<<"#     '8': use standard 'double' variable"<<cEol;
			ofs<<"#     '16': use 'dd_real' variable from QD package"<<cEol;
			ofs<<"#     '32': use 'qd_real' variable from QD package"<<cEol;
		}
		iParNo++; idParBinName=iParNo;
		ofs<<"# Parameter $"<<idParBinName<<": binary name (default: "<<sTitle<<")"<<cEol;
		ofs<<cEol;
		
		ofs<<varPathProj<<"=pwd"<<cEol; //`pwd`
		ofs<<"echo 'Project path: '$"<<varPathProj<<cEol;
		if(bUseLT){
			ofs<<"# LoopTools path"<<cEol;
			ofs<<varPathLT<<"='./LoopTools-2.5'"<<cEol;
			ofs<<"LT_CONFIGSCRIPT='LT_configure25'"<<cEol;
			ofs<<cEol;
		}
		if(bUseQD){
			ofs<<"# QD path"<<cEol;
			ofs<<varPathQD<<"='./qd-2.3.8'"<<cEol;
			ofs<<cEol;
		}

		if(bUseLT || bUseQD){
			ofs<<"if [ \"$"<<idParAll<<"\" = \"-all\" ]"<<cEol<<"then"<<cEol; ofs<<sOffset<<"echo \"[Re-compile project and all libraries]\""<<cEol;
			ofs<<"else"<<cEol;
			ofs<<"if [ \"$"<<idParAll<<"\" = \"-fast\" ]"<<cEol<<"then"<<cEol; ofs<<sOffset<<"echo \"[Re-compiling the project only]\""<<cEol;
			ofs<<"else"<<cEol;
			ofs<<sOffset<<"echo Script error. '-all' or '-fast' parameter required"<<cEol;
			ofs<<"fi"<<cEol; ofs<<"fi"<<cEol;
			ofs<<cEol;
		}

		// Compile QD
		if(bUseQD){
			ofs<<"if [ \"$"<<idParAll<<"\" = \"-all\" ]"<<cEol<<"then"<<cEol;
			ofs<<sOffset<<"echo '......... Compile QD ........'"<<cEol;
			ofs<<sOffset<<"cd $"<<varPathQD<<"/"<<cEol;
			ofs<<sOffset<<"make clean"<<cEol;
			ofs<<sOffset<<"./configure CXX=g++ CXXFLAGS='-O2'"<<cEol;
			ofs<<sOffset<<"make"<<cEol;
			ofs<<sOffset<<"# Note: don't call 'make install'"<<cEol;
			ofs<<sOffset<<"# Note: don't call 'make clean'"<<cEol;
			ofs<<sOffset<<"# We'll use libraries generated in $"<<varPathQD<<"/src/"<<cEol;
			ofs<<sOffset<<"cd $"<<varPathProj<<"/"<<cEol;
			ofs<<sOffset<<"echo '.......... QD library path: '$"<<varPathQD<<"'/src'"<<cEol;
			ofs<<"fi"<<cEol<<cEol;
		}

		// Compile LoopTools
		if(bUseLT){
			ofs<<"cd $"<<varPathLT<<"/"<<cEol;
			ofs<<"if [ \"$"<<idParAll<<"\" = \"-all\" ]"<<cEol<<"then"<<cEol;
			ofs<<sOffset<<"echo '......... Compile LoopTools ........'"<<cEol;
			ofs<<sOffset<<"echo 'Working in dir: '`pwd`"<<cEol;
			ofs<<sOffset<<"rm -f ./configure"<<cEol;
			ofs<<sOffset<<"cp ../$LT_CONFIGSCRIPT ./configure"<<cEol;
			ofs<<sOffset<<"if [ -f ./configure ]"<<cEol;
			ofs<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<"echo "<<sMakeScriptName.GetBuffer()<<": LoopTools configure script substituted"<<cEol;
			ofs<<sOffset<<"else"<<cEol;
			ofs<<sOffset<<sOffset<<"echo "<<sMakeScriptName.GetBuffer()<<" ERROR: Cant substitute LoopTools configure script"<<cEol;
			ofs<<sOffset<<sOffset<<"echo Check if ProjectPath/'$LT_CONFIGSCRIPT' file is missing?"<<cEol;
			ofs<<sOffset<<sOffset<<"exit"<<cEol;
			ofs<<sOffset<<"fi"<<cEol;
			ofs<<sOffset<<"make clean"<<cEol;
			ofs<<sOffset<<"./configure"<<cEol;
			ofs<<"fi"<<cEol<<cEol;

			ofs<<"#.. Read '"<<varLTLibs<<"' value"<<cEol; //#.. Read 'varLTLibs' value
			ofs<<"if [ -f LDFLAGS.var ]"<<cEol;
			ofs<<"then"<<cEol;
			ofs<<sOffset<<"while read BH_LINE"<<cEol;
			ofs<<sOffset<<"do"<<cEol;
			ofs<<sOffset<<varLTLibs<<"=$BH_LINE"<<cEol;
			ofs<<sOffset<<"done < LDFLAGS.var"<<cEol;
			ofs<<"else"<<cEol;
			ofs<<sOffset<<"echo \""<<sTitle<<" project compilation ERROR.\""<<cEol;
			ofs<<sOffset<<"if [ \"$"<<idParAll<<"\" = \"-fast\" ]"<<cEol<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<"echo \"Full rebuild is required. Type: \"$0\" -all ..."<<cEol;
			ofs<<sOffset<<"fi"<<cEol;
			ofs<<sOffset<<"cd $"<<varPathProj<<"/"<<cEol;
			ofs<<sOffset<<"exit"<<cEol;
			ofs<<"fi"<<cEol;

			ofs<<"if [ \"$"<<idParAll<<"\" = \"-all\" ]"<<cEol<<"then"<<cEol;
			ofs<<sOffset<<"make"<<cEol;
			ofs<<sOffset<<"make install"<<cEol;
			ofs<<sOffset<<"make clean"<<cEol;
			ofs<<"fi"<<cEol;
			ofs<<sOffset<<"cd $"<<varPathProj<<"/"<<cEol;

			ofs<<varHostType<<"=`tcsh -cf 'echo $HOSTTYPE'`"<<cEol;
			ofs<<"echo '.......... LoopTools library path:  '$"<<varPathLT<<"'/'$"<<varHostType<<"'/bin/'"<<cEol;
			ofs<<"echo '.......... LoopTools libraries used: ' $"<<varLTLibs<<cEol;

			ofs<<"# Export variables used in Makefile"<<cEol;
			ofs<<"export "<<varHostType<<cEol;
			ofs<<"export "<<varLTLibs<<cEol;
			ofs<<"export "<<varPathLT<<cEol;
			ofs<<cEol;
		}

		//Binary name selection
		ofs<<"# Binary name: ./bin/$"<<idParBinName<<" or automatic (default: './bin/"<<sTitle<<"')"<<cEol;
		ofs<<varBinary<<"=./bin/"<<sTitle<<cEol;
		ofs<<"if [ -z $"<<idParBinName<<" ]"<<cEol;
		ofs<<"then"<<cEol;
		ofs<<sOffset<<varBinary<<"=./bin/"<<sTitle<<cEol;
		ofs<<"else"<<cEol;
		ofs<<sOffset<<varBinary<<"=./bin/\"$"<<idParBinName<<"\""<<cEol;
		ofs<<"fi"<<cEol;
		ofs<<cEol;

		if(bUseQD){
			ofs<<"# Select code precision flag: "<<CCodeProcessImpl::Flag_Real8<<"(default), "<<
				CCodeProcessImpl::Flag_Real16<<", "<<CCodeProcessImpl::Flag_Real32<<cEol;
			ofs<<"if [ \"$"<<idParPrec<<"\" = \"8\" ]"<<cEol;
			ofs<<"then"<<cEol;
			ofs<<sOffset<<"echo \"---- 8-bytes variables are used ----\""<<cEol;
			ofs<<sOffset<<varCodePrec<<"="<<CCodeProcessImpl::Flag_Real8<<cEol;
			ofs<<"else"<<cEol;
			ofs<<sOffset<<"if [ \"$"<<idParPrec<<"\" = \"16\" ]"<<cEol;
			ofs<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<"echo \"---- 16-bytes variables are used ----\""<<cEol;
			ofs<<sOffset<<sOffset<<varCodePrec<<"="<<CCodeProcessImpl::Flag_Real16<<cEol;
			ofs<<sOffset<<sOffset<<"if [ -z $"<<idParBinName<<" ]"<<cEol;
			ofs<<sOffset<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<varBinary<<"=./bin/"<<sTitle<<"16"<<cEol;
			ofs<<sOffset<<sOffset<<"fi"<<cEol;
			ofs<<sOffset<<"else"<<cEol;
			ofs<<sOffset<<sOffset<<"if [ \"$"<<idParPrec<<"\" = \"32\" ]"<<cEol;
			ofs<<sOffset<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"echo \"---- 32-bytes variables are used ----\""<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<varCodePrec<<"="<<CCodeProcessImpl::Flag_Real32<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"if [ -z $"<<idParBinName<<" ]"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<sOffset<<varBinary<<"=./bin/"<<sTitle<<"32"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"fi"<<cEol;
			ofs<<sOffset<<sOffset<<"else"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"if [ -z $"<<idParPrec<<" ]"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<sOffset<<"echo \"---- 8-bytes variables are used by default ----\""<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<sOffset<<varCodePrec<<"="<<CCodeProcessImpl::Flag_Real8<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"else"<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<sOffset<<"echo \"ERROR. Invalid option: \"$"<<idParPrec<<". Expected: 8, 16 or 32."<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<sOffset<<"echo \"---- 8-bytes variables are used by default ----\""<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<sOffset<<varCodePrec<<"="<<CCodeProcessImpl::Flag_Real8<<cEol;
			ofs<<sOffset<<sOffset<<sOffset<<"fi"<<cEol;
			ofs<<sOffset<<sOffset<<"fi"<<cEol;
			ofs<<sOffset<<"fi"<<cEol;
			ofs<<"fi"<<cEol;
			ofs<<"export "<<varCodePrec<<cEol;
			ofs<<"export "<<varBinary<<cEol;
			ofs<<cEol;
		}

		ofs<<"if [ -f $"<<varBinary<<" ]"<<cEol;
		ofs<<"then"<<cEol;
		ofs<<""<<cEol;
		ofs<<sOffset<<"echo . Removing previous binary: $"<<varBinary<<cEol;
		ofs<<sOffset<<"rm -f $"<<varBinary<<cEol;
		ofs<<"else"<<cEol;
		ofs<<sOffset<<"echo . No previous binary detected: $"<<varBinary<<cEol;
		ofs<<"fi"<<cEol;
		if(idParAll>0){
			ofs<<"if [ \"$"<<idParAll<<"\" = \"-all\" ]"<<cEol;
			ofs<<"then"<<cEol;
			ofs<<sOffset<<"make clean"<<cEol;
			ofs<<"fi"<<cEol;
		}
		ofs<<cEol;

		ofs<<"make"<<cEol; ofs<<cEol;

		ofs<<"if [ -f $"<<varBinary<<" ]"<<cEol;
		ofs<<"then"<<cEol;
		ofs<<sOffset<<"echo . Binary created: $"<<varBinary<<cEol;
		ofs<<sOffset<<"echo ................ "<<sTitle<<" project built successfully ................"<<cEol;
		if(idParAll>0){
			ofs<<sOffset<<"if [ \"$"<<idParAll<<"\" = \"-all\" ]"<<cEol;
			ofs<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<"echo \"For further recompilations type: \" $0 \"-fast\" ..."<<cEol;
			ofs<<sOffset<<"fi"<<cEol;
		}
		ofs<<"else"<<cEol;
		ofs<<sOffset<<"echo ................ "<<sTitle<<" project compilation FAILED ................"<<cEol;
		if(idParAll>0){
			ofs<<sOffset<<"if [ \"$"<<idParAll<<"\" = \"-fast\" ]"<<cEol;
			ofs<<sOffset<<"then"<<cEol;
			ofs<<sOffset<<sOffset<<"echo \"Try full compilation with: \" $0 \"-all\" ..."<<cEol;
			ofs<<sOffset<<"fi"<<cEol;
		}
		ofs<<"fi"<<cEol;
		ofs.close();
	}else{
		dbg->logIns(0,"ERROR: Can't create/open script file '%s'",sMakeScriptName.GetBuffer());
	}

	//	2. Create Makefile
	// uses external variables:
	//		'varBinary'
	//		'varHostType'	= x86_64-linux, = i386-linux
	//		'varCodePrec'
	//		'varPathLT'		= ./LoopTools-2.X
	//		'varLTLibs'		= Additional libraries used by LoopTools (collected by LT Makefile & ported here)
	//		'varPathQD'		= ./qd-2.3.8
	ofs.open(sMakefileName.GetBuffer(), std::ios::out|std::ios::trunc|std::ios::binary);
	if(ofs){
		ofs<<"#"<<cEol<<"# Makefile for "<<sTitle<<" project"<<cEol;
		ofs<<"# Generated by "<<sLogo.GetBuffer()<<cEol<<"# To be use by '"<<sMakeScriptName.GetBuffer()<<"' script only"<<cEol<<"#"<<cEol;
		//
		ofs<<sProj.GetBuffer()<<"_TARGET=$("<<varBinary<<")"<<cEol;
		ofs<<sProj.GetBuffer()<<"_PATH_SRC = src"<<cEol;
		//
		//MYPROJ_INCLUDE = -I./qd-2.3.8/include -I./$(BH_LTPATH)/$(BH_HOSTTYPE)/include
		ofs<<sProj.GetBuffer()<<"_INCLUDE ="; if(bUseQD){ ofs<<" -I$("<<varPathQD<<")/include"; }
		if(bUseLT){ ofs<<" -I$("<<varPathLT<<")/$("<<varHostType<<")/include"; } ofs<<cEol<<cEol;
		//
		//MYPROJ_SRC_DIRS = pythia jetset
		ofs<<sProj.GetBuffer()<<"_SRC_DIRS = "; ofs<<"vmlib"; ofs<<cEol<<cEol;
		//
		//#create build list
		//MYPROJ_LIST_SRC = $(MYPROJ_PATH_SRC) $(addprefix $(MYPROJ_PATH_SRC)/, $(MYPROJ_SRC_DIRS))
		ofs<<sProj.GetBuffer()<<"_LIST_SRC = $("<<sProj.GetBuffer()<<"_PATH_SRC) $(addprefix $("
			<<sProj.GetBuffer()<<"_PATH_SRC)/, $("<<sProj.GetBuffer()<<"_SRC_DIRS))"<<cEol<<cEol;

		//MYPROJ_OBJ = *.o $(wildcard $(addsuffix /*.o, src ) )
		ofs<<sProj.GetBuffer()<<"_OBJ = *.o $(wildcard $(addsuffix /*.o, $("<<sProj.GetBuffer()<<"_PATH_SRC) ) )"<<cEol<<cEol;
		//
		//MYPROJ_FLAGS_CPP_COMPILE = -O2 -I./$(MYPROJ_PATH_SRC)
		//[also]MYPROJ_FLAGS_F_COMPILE = -fugly-logint -fno-backslash  -fno-automatic -fugly-complex 
		ofs<<sProj.GetBuffer()<<"_FLAGS_CPP_COMPILE = -O2 -I./$("<<sProj.GetBuffer()<<"_PATH_SRC)"<<cEol<<cEol;
		//
		//MYPROJ_FILES_SRC_CPP = $(wildcard $(addsuffix /*.cpp, $(MYPROJ_LIST_SRC) ) )
		ofs<<sProj.GetBuffer()<<"_FILES_SRC_CPP = $(wildcard $(addsuffix /*.cpp, $("<<sProj.GetBuffer()<<"_LIST_SRC) ) )"<<cEol;
		//[also]MYPROJ_FILES_SRC_F = $(wildcard $(addsuffix /*.f, $(MYPROJ_LIST_SRC) ) )
		//[also]MYPROJ_FILES_SRC_FC = $(wildcard $(addsuffix /*.F, $(MYPROJ_LIST_SRC) ) )
		//[also]MYPROJ_FILES_SRC_FPP = $(wildcard $(addsuffix /*.fpp, $(MYPROJ_LIST_SRC) ) )
		//MYPROJ_FILES_O_TEMP_CPP = $(notdir $(MYPROJ_FILES_SRC_CPP))
		ofs<<sProj.GetBuffer()<<"_FILES_O_TEMP_CPP = $(notdir $("<<sProj.GetBuffer()<<"_FILES_SRC_CPP))"<<cEol;
		//[also]MYPROJ_FILES_O_TEMP_F = $(notdir $(MYPROJ_FILES_SRC_F))
		//[also]MYPROJ_FILES_O_TEMP_FC = $(notdir $(MYPROJ_FILES_SRC_FC))
		//[also]MYPROJ_FILES_O_TEMP_FPP = $(notdir $(MYPROJ_FILES_SRC_FPP))
		//MYPROJ_FILES_O_CPP = $(MYPROJ_FILES_O_TEMP_CPP:.cpp=.o)
		ofs<<sProj.GetBuffer()<<"_FILES_O_CPP = $("<<sProj.GetBuffer()<<"_FILES_O_TEMP_CPP:.cpp=.o)"<<cEol;
		//[also]MYPROJ_FILES_O_F = $(MYPROJ_FILES_O_TEMP_F:.f=.o)
		//[also]MYPROJ_FILES_O_FC = $(MYPROJ_FILES_O_TEMP_FC:.F=.o)
		//[also]MYPROJ_FILES_O_FPP = $(MYPROJ_FILES_O_TEMP_FPP:.fpp=.o)
		ofs<<cEol;

		//[also]#CERNLIB   := -L$(CERN)/2003/lib -lpdflib804 -lpacklib -lmathlib -lkernlib
		//MYPROJ_LIB = -L./qd-2.3.8/src -lqd -L./$(BH_LTPATH)/$(BH_HOSTTYPE)/lib -looptools
		ofs<<sProj.GetBuffer()<<"_LIB =";
		if(bUseQD){ ofs<<" -L$("<<varPathQD<<")/src -lqd"; }
		if(bUseLT){ ofs<<" -L$("<<varPathLT<<")/$("<<varHostType<<")/lib -looptools"; } ofs<<cEol;
		if(bUseLT){
			ofs<<"# Additional libraries used by LoopTools (collected by LT Makefile & ported here)"<<cEol;
			ofs<<"LT_LINKS = $("<<varLTLibs<<")"<<cEol;
		}
		if(bUseQD){
			ofs<<"# Precision definition for QD package"<<cEol;
			ofs<<sProj.GetBuffer()<<"_CODE_DEFS = -D$("<<varCodePrec<<")"<<cEol; //BHCODE_DEFS = -D$(BH_CODE_PREC)
		}
		ofs<<cEol;

		//EXE_C_COMPILE = gcc
		//EXE_LINK = g++
		ofs<<"EXE_C_COMPILE = gcc"<<cEol<<"EXE_LINK = g++"<<cEol<<cEol;

		//$(MYPROJ_TARGET): compile_all
		ofs<<sProj.GetBuffer()<<"_TARGET): compile_all"<<cEol<<cEol;

		//compile_all: $(MYPROJ_FILES_O_CPP) $(MYPROJ_FILES_O_F) $(MYPROJ_FILES_O_FPP) $(MYPROJ_FILES_O_FC)
		//	$(EXE_LINK) $^ $(MYPROJ_LIB) $(LT_LINKS) -o $(MYPROJ_TARGET)
		ofs<<"compile_all: $("<<sProj.GetBuffer()<<"_FILES_O_CPP)"<<cEol;
		ofs<<"\t$(EXE_LINK) $^ $("<<sProj.GetBuffer()<<"_LIB)"; if(bUseLT)ofs<<" $(LT_LINKS)"; ofs<<" -o $("<<sProj.GetBuffer()<<"_TARGET)"<<cEol;
		ofs<<cEol;

		//[?]VPATH := $(MYPROJ_LIST_SRC)
		//%.o: %.cpp
		//	$(EXE_C_COMPILE) -c $(BHCODE_DEFS) $(MYPROJ_FLAGS_CPP_COMPILE) $(MYPROJ_INCLUDE) $^
		ofs<<"%.o: %.cpp"<<cEol;
		ofs<<"\t$(EXE_C_COMPILE) -c"; if(bUseQD){ ofs<<" $("<<sProj.GetBuffer()<<"_CODE_DEFS)"; }
		ofs<<"$("<<sProj.GetBuffer()<<"_FLAGS_CPP_COMPILE) $("<<sProj.GetBuffer()<<"_INCLUDE) $^"<<cEol;
		//%.o: %.f
		//	g77 -c $(MYPROJ_INCLUDES_F) $(MYPROJ_FLAGS_F_COMPILE) $^
		//%.o: %.F
		//	g77 -c $(MYPROJ_INCLUDES_F) $(MYPROJ_FLAGS_F_COMPILE) $^
		//%.o: %.fpp
		//	g77 -c -ff90 $(MYPROJ_FLAGS_F_COMPILE) $^
		ofs<<cEol;

		//clean:
		//	rm -r $(MYPROJ_OBJ)
		ofs<<"clean:"<<cEol<<"\trm -r $("<<sProj.GetBuffer()<<"_OBJ)"<<cEol;

		//----
		ofs.close();
	}else{
		dbg->log(0,"ERROR: Can't create/open file: ",sMakefileName.GetBuffer());
	}

	// 3. Create Dummy 'Makefile' for 'clean' operation only	
	ofs.open("Makefile", std::ios::out|std::ios::trunc|std::ios::binary);
	if(ofs){
		ofs<<"#"<<cEol<<"# Dummy makefile for 'make clean' operation"<<cEol<<"# Generated by "<<sLogo.GetBuffer()<<cEol<<"#"<<cEol;
		ofs<<sProj.GetBuffer()<<"_PATH_SRC = src";
		ofs<<sProj.GetBuffer()<<"_OBJ = *.o $(wildcard $(addsuffix /*.o, $("<<sProj.GetBuffer()<<"_PATH_SRC) ) )"<<cEol<<cEol;
		ofs<<"clean:"<<cEol<<"\trm -r $("<<sProj.GetBuffer()<<"_OBJ)"<<cEol;
		ofs.close();
	}else{
		dbg->log(0,"Can't create/open Makefile");
	}
}

float CLHEPRunProject::GetDefaultEnergy(long pk) {
	unsigned long pkAbs=CLMath::abs(pk);
	float res=1000.; //default
	if(pkAbs == CEParticle::PARTICLE_PROTON) { res=7000.; }
	//else if(pkAb == CEParticle::PARTICLE_PROTON) { res=7000.; 
	return res;
}
