////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// MEComposeLib.cpp: implementation of the CMEComposeLib class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CMEComposeLib::dbg = new CDebug("CMEComposeLib");
const DWORD CMEComposeLib::COMPOSE_ME_NO_CROSSINGS	= 0x01;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMEComposeLib::CMEComposeLib() : CELibrary() {
	m_order = -1;
	m_ME_seq_list = NULL; //m_UnitFlags = NULL;
	
	m_ME_result = NULL;
	m_Diags_result = NULL;

	m_settings = COMPOSE_ME_NO_CROSSINGS;
	//m_settings = 0;
	m_uniform_index_list = new CEPtrArray();

	m_init_OK = true;
	m_phys = CDiagramPhysics::get_curr_physics();
	if (m_phys == NULL) { dbg->log(0, "Error: No physics installed"); m_init_OK = false; }
	m_kin = CEProcessor::kinematics();
	if (m_kin == NULL) { dbg->log(0, "Error: No kinematics installed"); m_init_OK = false; }

	m_do_involve_CT = false;
	m_bClassesInvolved = false;
}

CMEComposeLib::~CMEComposeLib() {
	if (m_ME_seq_list != NULL) delete m_ME_seq_list;
	//if (m_UnitFlags != NULL) delete m_UnitFlags;
	if (m_ME_result != NULL) delete m_ME_result;
	if (m_Diags_result != NULL) delete m_Diags_result;
	if (m_uniform_index_list != NULL) delete m_uniform_index_list;
}

/**************************************************************************
***	library usage:
***		CMEComposeLib* lib = new CMEComposeLib();
***		// lib->setCounterTermsProcessing(true); ==> counterterms (only) will be created
***		bool run_OK = lib->ComposeME(order);
***		if (!run_OK) { delete lib; throw; }
***		CMatrixElement* res_me = (lib->GetCreatedME())->clone();
***		CDiagramSet* res_diags = (lib->GetDiagramsCreated())->clone();
***		delete lib;
**************************************************************************/
CMatrixElement* CMEComposeLib::GetCreatedME() { return m_ME_result; }
CDiagramSet* CMEComposeLib::GetDiagramsCreated() { return m_Diags_result; }

/*
*	<OLD>Note: after setCounterTermsProcessing(true) is set, the counterterm diagrams will only be created.
*	<OLD>The counterterm diagrams will actualy have (order+2) order. I.e. one should call ComposeME(2) to create CT for loop corrections in 2->2 process.
*
*	Note: if physics have no CSMPhysics::PHYS_RENORM_COUNTERTERMS flag set, no coounterterm diagrams will be created.
*/
void CMEComposeLib::setCounterTermsProcessing(bool _do_involve_CT) { m_do_involve_CT = _do_involve_CT; }

bool CMEComposeLib::ComposeME(int order) { // library uses current physics & kinematics
	if (!m_init_OK) { dbg->log(2, order, "ComposeME(%ld) error: can't initialize library"); return false; }
	dbg->log(15, order, "ComposeME(%ld) engine started");
	dbg->log(15, order, "Kinematics:", m_kin);

	// init
	m_order = order;
	if ((m_phys != NULL) && (m_phys->CounterTermsDeclared())) {
		int part_num = m_kin->num_init_particles() + m_kin->num_final_particles();
		
		// note: at tree level there are (part_num-2) order diagrams
		// loop order is at least tree_level + 2
		// note: hight-order loops counterterms are not implemented ==> "==" check.
		if (m_order == part_num) m_do_involve_CT = true; // need to involve counterterms
	} else {
		if (m_do_involve_CT) dbg->log(2, "Warning: CT-flag is whitched off since no CT are supposed in physics model");
		m_do_involve_CT = false;
	}

	bool log_job = (dbg->get_level() >= 20);
	if (log_job) printf("ComposeDiagrams engine started (involve CT: %s)\n",(m_do_involve_CT?"true":"false"));

	// 1. retrieve operator sequence structure from physics
	FillMESeqList2();
	if (log_job) printf("Done: 1. retrieve operator sequence structure from physics\n");

	// 1.1 filter order
	FilterSeqOrder();
	if (log_job) printf("Done: 1.1 filter order\n");

	// 2. Wick theorem
	WickTheoremExpand();
	if (log_job) printf("Done: 2. Wick theorem\n");

	// 2.1 extract unbinded graphs
	TopologyFilter();
	if (log_job) printf("Done: 2.1 extract unbinded graphs\n");

	// 3. Sort & factorize sequences
	SortSequences();
	dbg->log(15, "Wick theorem operator sub-result (sequence arranged)", m_ME_seq_list);
	if (log_job) printf("Done: 3. Sort & factorize sequences\n");

	/*if (m_ME_seq_list->size()>2) { //TEST
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(2);
		m_ME_seq_list->RemoveAt(2,false);
		m_ME_seq_list->clear_all();
		m_ME_seq_list->add(seq);
		dbg->log(0,"m_ME_seq_list cutted", m_ME_seq_list);
	}*/

	// 4. Couple operator sequences to external particles
	// no sort may be performed after externals coupling
	// (wave functions are not supposed in sorting algorythms)
	//int lev_old = dbg->set_level(200); // DEBUG
	CoupleExternals();
	dbg->log(25, "External fields coupled", m_ME_seq_list);
	if (log_job) printf("Done: 4. Couple operator sequences to external particles\n");
	//dbg->set_level(lev_old); // DEBUG

	// 4.1. Arrange crossings to normal state
	ArrangeCrossings();
	if (log_job) printf("Done: 4.1. Arrange crossings to normal state\n");

	/*if (order>2) { //TEST
		m_ME_seq_list->RemoveAt(1,true); m_ME_seq_list->RemoveAt(1,true);
		dbg->log(25, "m_ME_seq_list is cut", m_ME_seq_list);
		CMEOperatorSeq::dbg->set_level(50);
	}*/

	//SortSequences();
	//if (log_job) printf("Done: 4.11. SortSequences\n");

	// 4.2. Unify loops direction
	UnifyLoops();
	if (log_job) printf("Done: 4.2. Unify loops direction\n");

	// 4.8. Another sort & factorize sequences
	SortSequences();
	if (log_job) printf("Done: 4.8. Sequences sorted\n");
	//dbg->log(55, "SortSequences() completed", m_ME_seq_list);

	// 4.9
	TransformStandardSpinorOrder();
	if (log_job) printf("Done: 4.9. Spinor order transformed\n");

	// 4.99. Inverse common sign for (-1) + (-2) --> 1 + 2.
	ArrangeCommonSign();
	if (log_job) printf("Done: 4.99. Common sign adjusted\n");

	// 5. Restore Feynman Diagrams / Class Diagrams
	dbg->log(25, "before diagrams restoration operator sequences", m_ME_seq_list);
	RestoreDiagrams();
	if (log_job) printf("Done: 5. Restore Feynman Diagrams%s\n",(m_bClassesInvolved?" (diagram classes)":""));

	// 5.5. Restore Feynman Diagrams From Classes
	if (m_bClassesInvolved) {
		dbg->log(25, "Diagrams classes->diagrams restoration");
		RestoreClassesToDiagrams();
		if (log_job) printf("Done: 5.5. Restore Feynman Diagrams Form Classes\n");

		CLDiagTopologyInspector::FilterSimilarDiagrams(m_Diags_result,true,false);
		if (log_job) printf("Done: 5.6. Diagrams filtered\n");
	}

	// 6. Retrieve Matrix Element from sequences & diagrams, & compare them
	// 6.1. Arrange Diagrams Momenta
	int diags_total_num = m_Diags_result->get_diagrams_number();
	bool ME_available = false;
	bool mom_OK = m_Diags_result->ArrangeMomenta();
	if (log_job) printf("Done: 6.1. Arrange Diagrams Momenta\n");

	// 6.2. Drop diagrams with ZERO matrix element
	if (mom_OK) {
	//Old: if ((CEProcessor::m_AllowClassDiagrams) && (mom_OK)) {
	//Now: non-class diagrams may also be ZERO. E.g. double-emission of SF-proton.
		//long nDropped=0;
		long nDropped=DropZeroDiagrams();
		if (log_job) printf("Done: 6.2. Drop out zero diagrams. Total %ld diagrams dropped.\n", nDropped);
	}

	// 6.3 Retrieve ME from diagrams (only).
	if (mom_OK) {
		//m_ME_result = m_Diags_result->RetrieveMatrixElement();
		//if ((m_ME_result != NULL) || (diags_total_num == 0)) ME_available = true;
	} else {
		dbg->log(0, "Error diagrams momenta arrangment.");
	}
	//if (!ME_available) dbg->log(2, "No ME are available.");

	// finalize
	delete m_ME_seq_list; m_ME_seq_list = NULL; //delete m_UnitFlags; m_UnitFlags=NULL;

	//////////////////////////////////////////////////////////////////////////////////////
	///                        Add counterterms diagrams & ME                          ///
	///	        Note: CT are currently created with common diagrams list               ///
	//////////////////////////////////////////////////////////////////////////////////////
	// check for loops & counterterms consideration
	/*if (m_phys->CounterTermsDeclared()) {
		int part_num = m_kin->num_init_particles() + m_kin->num_final_particles();

		// note: at tree level there are (part_num-2) order diagrams
		// loop order is at least tree_level + 2
		// note: hight-order loops counterterms are not implemented ==> "==" check.
		if (m_order == part_num) { // need to involve counterterms
			CMEComposeLib* ct_lib = new CMEComposeLib();
			ct_lib->setCounterTermsProcessing(true);
			bool ct_run_OK = ct_lib->ComposeME(order-2); // subtract 2 from order
			if (ct_run_OK) {
				// append resulting diagrams & ME

				CDiagramSet* ct_diags = ct_lib->GetDiagramsCreated();
				int ct_s = ct_diags->get_diagrams_number();
				int ct_i;
				for(ct_i=0; ct_i<ct_s; ct_i++) {
					m_Diags_result->add_diagram(ct_diags->get_diagram(ct_i)->clone());
				}
				
				if (ME_available) { // append counterterms to ME if one is OK
					CMatrixElement* ct_me = ct_lib->GetCreatedME();
					if (ct_me != NULL) {
						if (m_ME_result != NULL) m_ME_result->add(ct_me->clone());
						else m_ME_result = ct_me->clone();
					} else {
						if (ct_s > 0) { dbg->log(0, "No counterterms matrix element available"); }
					}
				}
			} else { dbg->log(0, "Error creating counterterms, no counterterms created"); }
			delete ct_lib;
		}
	}*/
	//////////////////////////////////////////////////////////////////////////////////////

	// declare success
	if (m_ME_result != NULL) {
		dbg->log(15, order, "ComposeME(%ld) engine finished, ME created:", m_ME_result);
	} else {
		dbg->log(15, order, "ComposeME(%ld) engine finished.");
	}
	return true;
}

CExpression* CMEComposeLib::ChangePointID(CExpression* src, int ID_old, int ID_new) {
	if (src == NULL) return NULL;
	if (ID_old == ID_new) return src;

	CEPtrArray* data_arr = new CEPtrArray();
	data_arr->add(new CENumber(ID_old));
	data_arr->add(new CENumber(ID_new));
	CERule* r = new CESocketRule((void*)CMEComposeLib::ChangePointIDRuleProc, data_arr);
	delete data_arr;

	src = r->ApplyTo(src);
	delete r;
	return src;
}

CExpression* CMEComposeLib::ChangePointIDRuleProc(CExpression* src, CEPtrArray* data_arr) {
	if (src == NULL) return src;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) return src;
	CEFieldOperator* fo = (CEFieldOperator*)src;

	int ID_old = ((CENumber*)data_arr->GetAt(0))->getLong();
	if (fo->getPointID() == ID_old) {
		int ID_new = ((CENumber*)data_arr->GetAt(1))->getLong();
		fo->setPointID(ID_new);
	}
	return fo;// src
}

/*
*	Note: no sums should be leaved in the products expressions (check it out?)
*/
void CMEComposeLib::FillMESeqList() {
	CEException::halt("Old code invocation: CMEComposeLib::FillMESeqList()");
	if (m_ME_seq_list != NULL) delete m_ME_seq_list;
	m_ME_seq_list = new CAPtrSeq();
	
	CExpression* h_expr = m_phys->getHamiltonianExpr()->clone(); // default point is 0
	if (m_do_involve_CT) { // add counterterm hamiltonian
		CExpression* h_CT = m_phys->getRenormHamiltonian();
		if (h_CT != NULL) h_expr = h_expr->add(h_CT->clone());
	}

	CExpression* op_seq = NULL;
	// add lower-length h-products for 4-vertex diagrams & CT implementation
	int curr_order;
	// TODO: leave curr_order=m_order step only if no CT are allowed
	for(curr_order=m_order; curr_order>0; curr_order--) {
		CExpression* op_seq_order = h_expr->clone();
		int i;
		for(i=1; i<curr_order; i++) {
			CExpression* h_copy = h_expr->clone();
			h_copy = ChangePointID(h_copy, 0, i);
			h_copy = CETensor::ChangeVolatileIndices(h_copy);
			op_seq_order = op_seq_order->multiply(h_copy);
		}
		if (op_seq == NULL) op_seq = op_seq_order;
		else op_seq = op_seq->add(op_seq_order);
	}

	delete h_expr;
	dbg->log(45, "H+H*H+H*H*H+...+H*H*H*...*H", op_seq);
	op_seq = CExpression::reduce(op_seq);
	dbg->log(85, "H+H*H+H*H*H+...+H*H*H*...*H reduced", op_seq);

	// convert to internal format
	CESum* src_sum = CEBasic::convertToSum(op_seq);
	long s = src_sum->size();
	long i;
	for(i=0; i<s; i++) {
		CExpression* itm = src_sum->GetAt(i);
		CEProduct* itm_prod = CEBasic::convertToProduct(itm->clone());
		CMEOperatorSeq* op_seq = new CMEOperatorSeq(itm_prod, 1, 0);
		delete itm_prod;
		m_ME_seq_list->add(op_seq);
	}
	//src_sum->copyAttributesTo(m_ME_seq_list, true);
	delete src_sum;
	dbg->log(25, "operator sequence initialized"); dbg->log(55, m_ME_seq_list);
}

/*
*	Create H+H*H+H*H*H+...+H*H*...*H sequence
*	and turn it to CMEOperatorSeq form
*/
void CMEComposeLib::FillMESeqList2() {
	if (m_ME_seq_list != NULL) delete m_ME_seq_list;
	m_ME_seq_list = new CAPtrSeq();

	CExpression* H_expr=NULL;
	CExpression* H_expr_LO=NULL; // for lower order diagrams creation
	if (CEProcessor::m_AllowClassDiagrams) {
		H_expr=m_phys->getClassesHamiltonianE();
		if (H_expr!=NULL) { H_expr=H_expr->clone(); m_bClassesInvolved=true; }
	}
	if (H_expr==NULL) {
		H_expr=m_phys->getHamiltonianExpr();
		if (H_expr!=NULL) { H_expr=H_expr->clone(); /* default point is 0 */ }
	}
	if (H_expr==NULL) { dbg->log(0,"ERROR: NULL Hamiltonian expression spotted"); CEException::halt(); }
	dbg->log(45, "Hamiltonian retrieved (no counter-terms)", H_expr);
	H_expr_LO=H_expr->clone(); if (H_expr_LO==NULL) { CEException::halt(NULL,__FILE__,__LINE__); }
	if (m_do_involve_CT) { // add counterterm hamiltonian
		CExpression* H_CT=NULL;
		if (m_bClassesInvolved) H_CT=m_phys->getClassesHamiltonianRen();
		else H_CT = m_phys->getRenormHamiltonian();
		if (H_CT != NULL) { H_expr_LO=H_expr_LO->add(H_CT->clone()); }
		dbg->log(45, "Hamiltonian retrieved (with counter-terms)", H_expr_LO);
	}
	bool log_job = (dbg->get_level() >= 20);
	if (log_job) printf("Class diagrams creation: %s\n",(m_bClassesInvolved?"true":"false"));

	// make H a straight sum of products
	CESum* H_sum = CEBasic::convertToSum(CExpression::reduce(H_expr));
	CESum* H_sum_LO=CEBasic::convertToSum(CExpression::reduce(H_expr_LO)); // for lower order diagrams creation
	long H_size = H_sum->size();
	long H_size_LO = H_sum_LO->size();
	if(m_bClassesInvolved){ dbg->log(25, H_size, H_size_LO, "Class-H retrieved & reduced (no-CT size: %ld, with-CT size: %ld)", H_sum_LO); }
	else{ dbg->log(25, H_size, H_size_LO, "H retrieved & reduced (no-CT size: %ld, with-CT size: %ld)", H_sum_LO); }
	if ((H_size<=0) && (H_size_LO<=0)) { dbg->log(0, "WARNING: ZERO-sized hamiltonian"); return; }

	int i;
	CSequence H_points; // H values at points 1,2,...n-1
	CSequence H_points_LO; // H values at points 1,2,...n-1 (actually used 1,2,...n-3)
	for(i=1; i<m_order; i++) {
		CExpression* h_copy = H_sum->clone();
		h_copy = ChangePointID(h_copy, 0, i);
		h_copy = CETensor::ChangeVolatileIndices(h_copy);
		if (!h_copy->getProperty(CEProperties::PROP_MASK_IS_SUM)) { CEException::halt("H-copy is not the sum"); }
		if (((CESum*)h_copy)->size() != H_size) { CEException::halt("H-copy is the sum of different size"); }
		H_points.addItem(h_copy);
		
		CExpression* h_copy_LO = H_sum_LO->clone();
		h_copy_LO = ChangePointID(h_copy_LO, 0, i);
		h_copy_LO = CETensor::ChangeVolatileIndices(h_copy_LO);
		if (!h_copy_LO->getProperty(CEProperties::PROP_MASK_IS_SUM)) { CEException::halt("H-copy (including CT) is not the sum"); }
		if (((CESum*)h_copy_LO)->size() != H_size_LO) { CEException::halt("H-copy (including CT) is the sum of different size"); }
		H_points_LO.addItem(h_copy_LO);
	}
	H_points.InsertAtX(0,H_sum); // insert 0-point string
	H_points_LO.InsertAtX(0,H_sum_LO); /* insert 0-point string */

	if (m_order>40) CEException::halt("Compose Digrams Order>40, engine halted");
	int h_pos[40];

	// parse all H*H*...*H products & add to m_ME_seq_list
	int curr_order=1;
	//	*****************************************************	<scroll>
	CEProgressWatch* pw = NULL;
	// calc iterations num
	//{very old:} long iterations_num = CLMath::Pow(H_size,m_order); for(i=1;i<m_order;i++) { iterations_num += CLMath::Pow(H_size,i); }
	long iterations_num=0;
	// below: (i_tmp_order<m_order) due to processing (i_tmp_order==m_order) case further
	for(int i_tmp_order=1; i_tmp_order<m_order; i_tmp_order++) { iterations_num += CLMath::CNOfK(H_size_LO+i_tmp_order-1,i_tmp_order); }
	// add (i_tmp_order==m_order) case (H_size, but not H_size_LO here)
	iterations_num += CLMath::CNOfK(H_size+m_order-1,m_order);
	long iterations_no = 0;
	long iterations_accepted = 0;
	if (iterations_num >= 10) { pw = CEProcessor::newProgressWatch("H*H*H... items composing (total %ld iterations)", iterations_num); }
	//	*****************************************************	<\scroll>
	for(; curr_order<=m_order; curr_order++) {
		int h_no=0; for(i=0;i<curr_order;i++) {h_pos[i]=0;}
		bool HH_next_exist=true;
		long order_mu = CLMath::factorial(curr_order); // order multiplicity
		for(;HH_next_exist;) {
			CExpression* op_seq_order = NULL;
			long item_mu = 1; // for aabbbcddx: 2!*3!*1!*2!*1!, (a<b<c<d<x) used to calc item factor
			int last_H_pos = -1; // to calc item_mu
			int H_pos_same_items = 0; // to calc item_mu
			for(i=0; i<curr_order; i++) {
				CESum* H_sumI=NULL;
				if (curr_order==m_order) { H_sumI=(CESum*)H_points.GetAt(i); } else { H_sumI=(CESum*)H_points_LO.GetAt(i); }
				CExpression* item = H_sumI->GetAt(h_pos[i])->clone();
				H_sumI->copyAttributesTo(item, true); // sign
				if (op_seq_order != NULL) { op_seq_order=op_seq_order->multiply(item); }
				else { op_seq_order=item; }
				
				if (last_H_pos == h_pos[i]) {
					H_pos_same_items++;
					item_mu *= H_pos_same_items;
				} else { // (re)init
					last_H_pos=h_pos[i];
					H_pos_same_items=1;
				}
			}
			if (op_seq_order == NULL) CEException::halt("CMEComposeLib: NULL op_seq_order");
			op_seq_order = CExpression::reduce(op_seq_order);
			if (op_seq_order->getProperty(CEProperties::PROP_MASK_IS_SUM)) { CEException::halt("CMEComposeLib INTERNAL ERROR: op_seq_order is sum"); }

			if (order_mu < item_mu) { dbg->log(0, "(Internal) ERROR diagram factor calculation!", op_seq_order); CEException::halt(); }
			long item_factor = order_mu/item_mu; // items factor
			if (item_factor*item_mu != order_mu) { dbg->log(0, order_mu,item_mu, "(Internal) ERROR diagram factor calculation (%ld,%ld)!", op_seq_order); CEException::halt(); }


			CEProduct* itm_prod = CEBasic::convertToProduct(op_seq_order);
			CMEOperatorSeq* op_seq = new CMEOperatorSeq(itm_prod, 1, 0);
			op_seq->Arrange(AL_INLINE); // inline possible field operators POWERS
			op_seq->set_multiplicity(item_factor);
			delete itm_prod;

			// pre-accept op_seq
			if (PreAcceptSeq(op_seq)) {
				m_ME_seq_list->add(op_seq);
				iterations_accepted++;
			} else { delete op_seq; }

			//	*****************************************************	<scroll>
			iterations_no++; if(pw!=NULL) { pw->step(iterations_no, iterations_num); }
			//	*****************************************************	<\scroll>

			// next combination check
			int pos_X=curr_order-1; // first index to increase
			bool incr_indices=true;
			for(;incr_indices;) {
				h_pos[pos_X]++;
				long H_check_size=((curr_order==m_order)?H_size:H_size_LO);
				if (h_pos[pos_X]>=H_check_size) {
					if (pos_X > 0) { pos_X--; /* h_pos[pos_X] will be corrected further */ }
					else { HH_next_exist=false; incr_indices=false; }
				} else {
					// truncate next indices to h_pos[pos_X]
					int h_val = h_pos[pos_X]; // cache pos
					for(i=pos_X+1;i<curr_order;i++) { h_pos[i]=h_val; }
					incr_indices=false;
				}
			}
		}
	}
	//	*****************************************************	<scroll>
	if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
	if (iterations_no != iterations_num) { dbg->log(0, iterations_no, iterations_num, "Warning: %ld, interation performed instead of %ld"); }
	//	*****************************************************	<\scroll>
	dbg->log(25, iterations_accepted, iterations_num, "operator sequence initialized (accepted %ld of %ld)");
	if (dbg->get_level() >= 55) {
		if (m_ME_seq_list->size()>300) { dbg->log(0, "Warning: long sequence logging. It may take several minutes. Reduce \"debug.ini\" parameters for faster operation."); }
		dbg->log(55, m_ME_seq_list);
	}
}

bool CMEComposeLib::PreAcceptSeq(CMEOperatorSeq* seq) {
	bool bAccept=true;
	if (!CheckSeqOrder(seq, true)) {
		bAccept=false;
		//dbg->log(105,bAccept,"PreAcceptSeq (1) res: %ld",seq);
	} else if (seq->is_kin_vanishing(m_kin,true,m_bClassesInvolved)) {
		bAccept=false;
		/*
		if (dbg->GetMessagesCounter()==264) {
			dbg->log(105,bAccept,"PreAcceptSeq (2) res: %ld",seq);
			CMEOperatorSeq::dbg->set_level(200);
			CEFieldOperator::dbg->set_level(200);
			seq->is_kin_vanishing(m_kin,true,m_bClassesInvolved);
			CMEOperatorSeq::dbg->set_level(5);
			CEFieldOperator::dbg->set_level(5);
		}
		dbg->log(105,bAccept,"PreAcceptSeq (2) res: %ld",seq);
		*/
	}
	//dbg->log(105,bAccept,"PreAcceptSeq res: %ld",seq);
	return bAccept;
}

/*
*	TODO: Check if this procedure is affected anything!
*
*	Note: order may raise by 2 after CT-propagators are involved
*/
void CMEComposeLib::FilterSeqOrder() {
	long s = m_ME_seq_list->size();
	long i;
	for(i=0; i<s; i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);
		if (!CheckSeqOrder(seq, true)) {
			m_ME_seq_list->RemoveAt(i, true); i--; s--;
		}
	}
}

/*
*	Old/fixed: Call ALWAYS (seq,false) until this procedure is modified!
*
*	Old/fixed: TODO: check allow_CT_gap parameter for correct use here!!!!
*	Old/fixed: allow_CT_gap means one CT may be added in future (as propagator CT)
*	Old/fixed: parameter is useless if (!m_do_involve_CT)
*/
bool CMEComposeLib::CheckSeqOrder(CMEOperatorSeq* seq, bool allow_CT_gap) {
	long order = seq->calcSeqOrder(m_phys, m_do_involve_CT);
	bool res = false;
	res = (order == m_order); // Ignore allow_CT_gap now!

	// Allow CT-propagators 'gap' if CT-propagators are implementes NOT as 2-tale vertices.
	if(allow_CT_gap && ((!m_do_involve_CT) || (m_phys->AreCTPropagatorsVertices()))) allow_CT_gap=false;
	if(allow_CT_gap && (order != m_order)) {
		if ((seq->calcCounterTermsNum(m_phys)==0) && (m_order==(order+2))) res=true;
	}

	/*//old
	if (!m_do_involve_CT) {
		res = (order == m_order);
	} else { // Note: if (allow_CT_gap) order may raise by 2 after CT-propagators are involved
		long CT_num = seq->calcCounterTermsNum(m_phys);
		if (CT_num > 1) {
			res = false;
		} else if (CT_num == 1) {
			res = (order == m_order);
		} else if (CT_num == 0) {
			if (order == m_order) {
				res = true; // allow non-CT diagrams creation
			} else {
				if (allow_CT_gap) res = (m_order == (order+2));
			}
		} else CEException::halt(dbg, "Seq Order Error");
	}
	*/
	return res;
}

void CMEComposeLib::WickTheoremExpand() {
	long log_it_num=16; if ((m_order>=4)||((m_kin!=NULL)&&(m_kin->num_final_particles()>=4))) log_it_num=4;
	long list_pos;
	//	*****************************************************	<scroll>
	long it_num = m_ME_seq_list->size(); long it_no=0;
	CEProgressWatch* pw = NULL; if (it_num >= log_it_num) { pw = CEProcessor::newProgressWatch("Wick theorem"); pw->step(0,it_num); }
	//	*****************************************************	<\scroll>

	for (list_pos=0; list_pos<m_ME_seq_list->size(); ) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(list_pos);

		// couple fields in sequence
		dbg->log(85, "Apply Wick theorem to sequence", seq);
		CEPtrArray res_arr;
		WickTheoremLoop(&res_arr, (CMEOperatorSeq*)seq->clone(), 0);
		dbg->log(85, "Wick theorem results array", &res_arr);

		// remove old sequence
		m_ME_seq_list->RemoveAt(list_pos,true);

		// insert new sequences
		long s = res_arr.size();
		long i;
		for (i=0; i<s; i++) {
			m_ME_seq_list->InsertAt(list_pos+i, ((CMEOperatorSeq*)res_arr.GetAt(i))->clone());
		}
		list_pos += s;

		//	*****************************************************	<scroll>
		if (pw != NULL) { it_no++; pw->step(it_no, it_num); }
		//	*****************************************************	<\scroll>
	}
	//	*****************************************************	<scroll>
	if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
	//	*****************************************************	<\scroll>
	dbg->log(25, "Wick theorem involved"); dbg->log(55, m_ME_seq_list);
}

/*
*	The source operator sequence is destroyed after use.
*/
void CMEComposeLib::WickTheoremLoop(CEPtrArray* res_arr, CMEOperatorSeq* seq, long first_pos) {
	long s = seq->size();
	long i;
	for (i=first_pos; i<s; i++) {
		CExpression* itm_expr = seq->GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo_A = (CEFieldOperator*)itm_expr;
		if (fo_A->isCoupled()) continue;

		long j;
		for (j=i+1; j<s; j++) {
			itm_expr = seq->GetAt(j);
			if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
			CEFieldOperator* fo_B = (CEFieldOperator*)itm_expr;
			if (fo_B->isCoupled()) continue;
			if (!fo_A->MayBeCoupledWith(fo_B)) continue;

			// do couple (in copied object) & do not couple (in source sequence)
			CMEOperatorSeq* seq_copy = (CMEOperatorSeq*)seq->clone(); // should return identical copy
#ifdef ALTA_DEBUG_CHECKS
			if (!seq->equals(seq_copy)) { dbg->log(0, "<error> not identical cloning for operator sequence", seq); CEException::halt(); }
#endif //ALTA_DEBUG_CHECKS
			CEFieldOperator* fo_A_copy = (CEFieldOperator*)seq_copy->GetAt(i);
			CEFieldOperator* fo_B_copy = (CEFieldOperator*)seq_copy->GetAt(j);
			int couple_No = seq_copy->incMaxCouplingNo();
			fo_A_copy->setCouplingNo(couple_No); fo_B_copy->setCouplingNo(couple_No);
			if (seq_copy->is_kin_vanishing(m_kin,true,m_bClassesInvolved)) { delete seq_copy; } else { WickTheoremLoop(res_arr, seq_copy, i+1); }

			// CT: make the same couple with counterterm propagator
			if (m_do_involve_CT) {
				seq_copy = (CMEOperatorSeq*)seq->clone(); // should return identical copy
#ifdef ALTA_DEBUG_CHECKS
				if (!seq->equals(seq_copy)) { dbg->log(0, "<error> not identical cloning for operator sequence", seq); CEException::halt(); }
#endif //ALTA_DEBUG_CHECKS
				CEFieldOperator* fo_A_copy = (CEFieldOperator*)seq_copy->GetAt(i);
				CEFieldOperator* fo_B_copy = (CEFieldOperator*)seq_copy->GetAt(j);
				int couple_No = seq_copy->incMaxCouplingNo();
				fo_A_copy->setCouplingNo(couple_No); fo_B_copy->setCouplingNo(couple_No);
				fo_A_copy->setCounterTermCoupling(true); fo_B_copy->setCounterTermCoupling(true);
				if(seq_copy->is_kin_vanishing(m_kin,true,m_bClassesInvolved)){ delete seq_copy; }
				else{ WickTheoremLoop(res_arr, seq_copy, i+1); }
			}
		}
	}

	// check sequence for 0 in specified kinematics
	if (seq->is_kin_vanishing(m_kin,false,m_bClassesInvolved)) { delete seq; return; }

	if (m_do_involve_CT) { // check for counterterms number
		// the only one counterterm (or no one) may occur in every operand
		if (seq->calcCounterTermsNum(m_phys) > 1) { delete seq; return; }
		if(!CheckSeqOrder(seq,false)){ delete seq; return; } //Nt: CT-propagator was(or not) involved above => no more 'gaps'
	}

	// add final sequence to results array
	InsertSeqToArray(res_arr,seq);
}

void CMEComposeLib::InsertSeqToArray(CEPtrArray* arr, CMEOperatorSeq* seq) { // deletes seq
	if ((arr==NULL)||(seq==NULL)) return;
	dbg->log(145,"InsertSeqToArray",seq);

#if defined(ALTA_DEBUG)
	DWORD log_setts_old = CMEOperatorSeq::m_log_settings;
	CMEOperatorSeq::m_log_settings = CMEOperatorSeq::ME_SEQ_SORT_COMPLETE_LOG;
	// log settings at sequance sorting allows to trace internal errors (sort(A)->B, sort(B)->A)
#endif //ALTA_DEBUG
	seq->sort_internal(m_uniform_index_list);
#if defined(ALTA_DEBUG)
	CMEOperatorSeq::m_log_settings = log_setts_old;
#endif //ALTA_DEBUG
	//dbg->log(145,"InsertSeqToArray (seq sorted)",seq);

	// sort
	long s = arr->size();
	long i;
	for (i=0; i<s; i++) {
		CMEOperatorSeq* seqI = (CMEOperatorSeq*)arr->GetAt(i);
		if (seqI->equals_topology(seq)) {	
			seqI->inc_multiplicity(seq->get_multiplicity()); delete seq; return;
		}
	}
	dbg->log(145,"InsertSeqToArray (seq arranged)",seq);
	arr->add(seq);
}

void CMEComposeLib::TopologyFilter() {
	long s = m_ME_seq_list->size();
	long i;
	for (i=0; i<s; i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);
		bool chk_res = CLDiagTopologyInspector::CheckOpSeq(seq, CLDiagTopologyInspector::NON_EULER_CIRCUIT);
		if (chk_res) { m_ME_seq_list->RemoveAt(i,true); i--; s--; }
	}
}

void CMEComposeLib::CoupleExternals() {

	long list_pos;
	for (list_pos=0; list_pos<m_ME_seq_list->size(); ) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(list_pos);

		// couple sequence to external fields
		dbg->log(85, "Couple external fields to sequence", seq);
		CEPtrArray* res_arr = new CEPtrArray();
		CEPtrArray* crossings_data = new CEPtrArray();
		CoupleExternalsLoop(res_arr, (CMEOperatorSeq*)seq->clone(), -m_kin->num_final_particles(), crossings_data);
		delete crossings_data;
		dbg->log(85, "Couple external fields results array", res_arr);

		// remove old sequence
		m_ME_seq_list->RemoveAt(list_pos,true);

		// insert new sequences
		long s = res_arr->size();
		long i;
		for (i=0; i<s; i++) {
			m_ME_seq_list->InsertAt(list_pos+i, ((CMEOperatorSeq*)res_arr->GetAt(i))->clone());
		}
		delete res_arr;
		list_pos += s;
	}
}

/*
*	_PID is Particle ID (not the PointID)
*/
void CMEComposeLib::CoupleExternalsLoop(CEPtrArray* res_arr, CMEOperatorSeq* seq, int _PID, CEPtrArray* crossings_data) {
	int is = m_kin->num_init_particles();
	//int fs = m_kin->num_final_particles();

	dbg->log(145, _PID, "CoupleExternalsLoop (PID == %ld)", seq);
	dbg->log(145, "crossing data", res_arr);

	if (_PID == 0) {
		// also clear crossings data
		crossings_data->clear_all();
		_PID++;
	}
	if (_PID > is) {
		// all the external particles ae linked now, stop recursion
		// search for uncoupled operators in sequence
		if (seq->have_uncoupled_field_op()) {
			delete seq;
		} else { // add final sequence to results array
			InsertSeqToArray(res_arr,seq); //res_arr->add(seq);
			dbg->log(145, "result array changed", res_arr);
		}
		return;
	}

	long s = seq->size();
	CEParticle* part = m_kin->get_particle_by_ID(_PID);
	int part_PK = part->get_PK();
	//bool is_fermion = part->check_PK(CEParticle::PARTICLE_FERMION, true);
	bool is_fermion = CEParticle::is_fermion_PK(part_PK);
	bool avoid_crossings = ((m_settings & COMPOSE_ME_NO_CROSSINGS) != 0);
	dbg->log(145, is_fermion, "is_fermion:%ld.");
	// couple this particle to every field operator
	if (_PID < 0) { // couple to final particle (finals are at left)
		long i=0;
		int start_pos = 0;
		if (avoid_crossings) {
			int last_pos = GetLastExtCopulePos(crossings_data, part_PK);
			if (last_pos >= 0) start_pos = last_pos + 1;

			// note: we start to couple form "start_pos" only
			// but we should pass all the fermion operators before "start_pos"
			if (is_fermion) {
				for (i=0; i<start_pos; i++) {
					CExpression* itm_expr = seq->GetAt(i);
					if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
					CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
					if (CEParticle::is_fermion_PK(fo->getParticleKind())) seq->inverse_ferm_sign();
				}
			}
		}

		//dbg->log(145, start_pos, "coupling form start_pos:%ld.", seq);
		for (i=start_pos; i<s; i++) {
			//dbg->log(145, i, (seq->get_ferm_sign())?1:(-1), "coupling to pos:%ld (curr seq sign:%ld).");
			CExpression* itm_expr = seq->GetAt(i);
			if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue; // skip item
			CEFieldOperator* fo = (CEFieldOperator*)itm_expr;

			// couple to this operator (current sequence is not affected)
			if (fo->CanBeCoupledExternal(part, false, m_bClassesInvolved)) {
				CMEOperatorSeq* seq_copy = (CMEOperatorSeq*)seq->clone(); // should return identical copy
				if (!seq->equals(seq_copy)) { dbg->log(0, "<error> not identical cloning for operator sequence", seq); CEException::halt(); }
				CEFieldOperator* fo_copy = (CEFieldOperator*)seq_copy->GetAt(i);
				CExpression* subst_expr = fo_copy->GetExtParticleCoupling(part, false, m_bClassesInvolved);
				if (subst_expr == NULL) { dbg->log(0, "<error> can't bind operator to external field -- NULL coupling", fo_copy); CEException::halt(); }
				seq_copy->SetAt(i, subst_expr); delete subst_expr;
				CEPtrArray* crossings_data_new = CEPtrArray::get_copy(crossings_data, true);
				SetLastExtCopulePos(crossings_data_new, part_PK, i);
				CoupleExternalsLoop(res_arr, seq_copy, _PID+1, crossings_data_new);
				delete crossings_data_new;
			}

			// skip operator
			if (is_fermion) {
				if (CEParticle::is_fermion_PK(fo->getParticleKind())) seq->inverse_ferm_sign();
			}
		}
	} else { // couple initial particle (finals are at right)
		long start_pos = s-1;
		long i=0;
		if (avoid_crossings) {
			long last_pos = GetLastExtCopulePos(crossings_data, part_PK);
			if (last_pos >= 0) start_pos = last_pos - 1;

			// note: we start to couple form "start_pos" only
			// but we should pass all the fermion operators before "start_pos"
			if (is_fermion) {
				for (i=s-1; i>start_pos; i--) {
					CExpression* itm_expr = seq->GetAt(i);
					if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
					CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
					if (CEParticle::is_fermion_PK(fo->getParticleKind())) seq->inverse_ferm_sign();
				}
			}
		}
		for (i=start_pos; i>=0; i--) {
			CExpression* itm_expr = seq->GetAt(i);
			if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue; // skip item
			CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
			
			// couple to this operator
			if (fo->CanBeCoupledExternal(part, true, m_bClassesInvolved)) {
				CMEOperatorSeq* seq_copy = (CMEOperatorSeq*)seq->clone(); // should return identical copy
				if (!seq->equals(seq_copy)) { dbg->log(0, "<error> not identical cloning for operator sequence", seq); CEException::halt(); }
				CEFieldOperator* fo_copy = (CEFieldOperator*)seq_copy->GetAt(i);
				CExpression* subst_expr = fo_copy->GetExtParticleCoupling(part, true, m_bClassesInvolved);
				if (subst_expr == NULL) { dbg->log(0, "<error> can't bind operator to external field -- NULL coupling", fo_copy); CEException::halt(); }
				seq_copy->SetAt(i, subst_expr); delete subst_expr;
				CEPtrArray* crossings_data_new = CEPtrArray::get_copy(crossings_data, true);
				SetLastExtCopulePos(crossings_data_new, part_PK, i);
				CoupleExternalsLoop(res_arr, seq_copy, _PID+1, crossings_data_new);
				delete crossings_data_new;
			}
			
			// skip operator
			if (is_fermion) {
				if (CEParticle::is_fermion_PK(fo->getParticleKind())) seq->inverse_ferm_sign();
			}
		}
	}
	
	// delete this sequence
	delete seq;
}

long CMEComposeLib::GetLastExtCopulePos(CEPtrArray* crossings_data, int pk) {
	if (crossings_data == NULL) { dbg->log(0, "<Internal Error> GetLastExtCopulePos(): NULL crossing data array"); CEException::halt(); }
	long s = crossings_data->size();
	long i;
	for (i=0; i<s; i++) {
		CExprContainer* item = (CExprContainer*)crossings_data->GetAt(i);
		if (item->get_info(0) == pk) return item->get_info(1);
	}
	return -1;
}

void CMEComposeLib::SetLastExtCopulePos(CEPtrArray* crossings_data, int pk, long pos) {
	if (crossings_data == NULL) { dbg->log(0, "<Internal Error> SetLastExtCopulePos(): NULL crossing data array"); CEException::halt(); }
	long s = crossings_data->size();
	long i;
	for (i=0; i<s; i++) {
		CExprContainer* item = (CExprContainer*)crossings_data->GetAt(i);
		if (item->get_info(0) == pk) { item->set_info(1, pos); return; }
	}
	crossings_data->add(new CExprContainer(NULL, pk, pos));
}

void CMEComposeLib::SortSequences() {
#if defined(ALTA_DEBUG)
	DWORD log_setts_old = CMEOperatorSeq::m_log_settings;
	CMEOperatorSeq::m_log_settings = CMEOperatorSeq::ME_SEQ_SORT_COMPLETE_LOG;
	// log settings at sequance sorting allows to trace internal errors (sort(A)->B, sort(B)->A)
#endif //ALTA_DEBUG

	// sort
	long s = m_ME_seq_list->size();
	//	*****************************************************	<scroll>
	CEProgressWatch* pw = NULL;
	if (s >= 1000) { pw = CEProcessor::newProgressWatch("Operator sequences sorting (total %ld expressions)", s); }
	//	*****************************************************	<\scroll>
	long i;
	for (i=0; i<s; i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);
		seq->sort_internal(m_uniform_index_list);

		//	*****************************************************	<scroll>
		if (pw != NULL) pw->step(i+1, s);
		//	*****************************************************	<\scroll>
	}
	//	*****************************************************	<scroll>
	if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
	//	*****************************************************	<\scroll>
	dbg->log(45, "ME operator sequences sorted", m_ME_seq_list);

	// factorize similar sequences
	for (i=0; i<m_ME_seq_list->size(); i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);

		long j;
		for (j=i+1; j<m_ME_seq_list->size(); j++) {
			CMEOperatorSeq* seq_next = (CMEOperatorSeq*)m_ME_seq_list->GetAt(j);
			if (seq->equals_topology(seq_next)) {
				seq->inc_multiplicity(seq_next->get_multiplicity());
				m_ME_seq_list->RemoveAt(j,true); j--;
			}
		}
	}
	dbg->log(45, "ME operator sequences factorized", m_ME_seq_list);

#if defined(ALTA_DEBUG)
	CMEOperatorSeq::m_log_settings = log_setts_old;
#endif //ALTA_DEBUG
}

void CMEComposeLib::TransformStandardSpinorOrder() {
	dbg->log(75, "TransformStandardSpinorOrder() call", m_ME_seq_list);
	long s = m_ME_seq_list->size();
	long i;
	for (i=0; i<s; i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);
		seq->ToStandardFOCouplingsOrder();
	}
	dbg->log(45, "TransformStandardSpinorOrder() finised", m_ME_seq_list);
}

void CMEComposeLib::ArrangeCommonSign() {
	long p_signed = 0;
	long m_signed = 0;

	long s = m_ME_seq_list->size();
	long i;
	for (i=0; i<s; i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);
		//if (seq->getStandardOrderSign()) { p_signed++; }
		if (seq->get_ferm_sign()) { p_signed++; }
		else { m_signed++; }
	}

	if (p_signed < m_signed) {
		// inverse common sign
		for (i=0; i<s; i++) {
			CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);
			seq->inverse_ferm_sign();
		}
		dbg->log(45, "Common sign inversed", m_ME_seq_list);
	}
}

// transform crossing diagrams to normal ones
void CMEComposeLib::ArrangeCrossings() {
	dbg->log(65, "ArrangeCrossings", m_ME_seq_list);
	long s = m_ME_seq_list->size();
	bool _changed = false;
	long i;
	for (i=0; i<s; i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);

		// check structure for crossings & normalize it
		bool item_changed = seq->NormalizeCrossings(m_kin);
		if (item_changed) _changed = true;
	}
	
	if (_changed) {
		dbg->log(45, "Crossings inversed", m_ME_seq_list);
	}
}

void CMEComposeLib::UnifyLoops() {
	int part_num = m_kin->num_init_particles() + m_kin->num_final_particles();
	// note: at tree level there are (part_num-2) order diagrams
	// loop order is at least tree_level + 2
	if (m_order < part_num) return; // no loops here
	dbg->log(65, "UnifyLoops", m_ME_seq_list);
	CIntArray arrChanged;

	long s = m_ME_seq_list->size();
	bool _changed = false;
	long i;
	for (i=0; i<s; i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);

		// check structure for loops & normalize it
		CEPtrArray* loops = CMEOpSeqLoopsLib::GetLoopsData(seq, m_kin);
		if (loops == NULL) continue;
		// TODO: remove next check (to 40-70 debug level) after library is debugged.
		if (loops->size() > 1) { dbg->log(0, loops->size(), "Warning: %ld loops found", loops); }

		// remove non-inversable loops
		long j;
		for (j=0; j<loops->size(); j++) {
			CMESeqFORoute* ld = (CMESeqFORoute*)loops->GetAt(j);
			if (!CMEOpSeqLoopsLib::CanBeInversed(ld)) { loops->RemoveAt(j, true); j--; }
		}
		bool close_loops = CMEOpSeqLoopsLib::HaveCommonLines(loops);
		if (close_loops) { dbg->log(0, "Warning: Common at single-PK loops found", loops); }

		// Note: if (close_loops) ==> inverse loop precedure destroys adjacent loops data
		long loops_num = loops->size();
		for (j=0; j<loops_num; j++) {
			CMESeqFORoute* ld = (CMESeqFORoute*)loops->GetAt(j);
			bool is_ok = CMEOpSeqLoopsLib::CheckLoopOrder(ld, seq, m_kin);
			if (!is_ok) { CMEOpSeqLoopsLib::InverseLoop(ld); _changed = true; arrChanged.Add(i); }
		}
	}
	
	if (_changed) {
		dbg->log(25, "Some of loops are inversed", m_ME_seq_list);
		//dbg->log(25, "Changed list:", &arrChanged);
	}
}

void CMEComposeLib::RestoreDiagrams() {
	if (m_Diags_result != NULL) {
		dbg->log(0, "Warning: previously created diagrams are truncated");
		delete m_Diags_result;
	}
	m_Diags_result = new CDiagramSet(m_phys);

	// create diagram and parse it's topoogy
	long i;
	for (i=0; i<m_ME_seq_list->size(); i++) {
		CMEOperatorSeq* seq = (CMEOperatorSeq*)m_ME_seq_list->GetAt(i);
		CFeynDiagram* diag = CLDiagOpSeqRestore::RestoreDiagram(seq, m_phys, m_kin);
		if (diag == NULL) {
			dbg->log(0, "Error: can't create diagram topology. Operator sequence is ignored.", seq);
			m_ME_seq_list->RemoveAt(i,true); i--;
			continue;
		}
		
		//int required_order = (m_do_involve_CT)?(m_order+2):m_order;
		if (!diag->is_valid(m_order)) {
			dbg->log(0, "Error: Can't create valid diagram for sequence. Item is dropped out.", seq);
			delete diag; m_ME_seq_list->RemoveAt(i,true); i--;
			continue;
		}

		// check topology and drop out specific diagrams
		bool drop_diag = CLDiagTopologyInspector::CheckDiagram(diag,
			CLDiagTopologyInspector::NON_EULER_CIRCUIT |
			CLDiagTopologyInspector::LOOPS_IN_EXTERNAL_LINES |
			CLDiagTopologyInspector::ODD_LINEAR_FERMION_LOOP);
		if (drop_diag) {
			delete diag;
			m_ME_seq_list->RemoveAt(i,true); i--;
			continue;
		}

		m_Diags_result->add_diagram(diag);
	}
	if (m_bClassesInvolved) {
		dbg->log(15, m_Diags_result->get_diagrams_number(), "Diagram classes created (total %ld diargam classes)");
	} else {
		dbg->log(15, m_Diags_result->get_diagrams_number(), "Feynman diagrams created (total %ld diargams)");
	}
	dbg->log(15, "Corresponding operator list", m_ME_seq_list);
}

void CMEComposeLib::RestoreClassesToDiagrams() {
	if (!m_bClassesInvolved) return;
	if (m_Diags_result == NULL) { dbg->log(0, "ERROR: NULL class diagrams"); return; }
	CDiagramSet* dClasses=m_Diags_result;
	m_Diags_result=new CDiagramSet(m_phys);

	long s=dClasses->get_diagrams_number(); long i;
	for(i=0; i<s; i++) {
		//if (i==0) dbg->set_level(500);
		CFeynDiagram* d=dClasses->get_diagram(i); if(d==NULL){dbg->log(0,"WARNING:NULL diagram in set",dClasses); continue; }
		RestoreClassToDiagrams(d->clone(),0);
		//if (i==0) dbg->set_level(5);
	}
	delete dClasses; dClasses=NULL;

	// re-arrange diagram signs
	// Note:  signs may be -1 if all the "+1" classes are dropped out
	// Note #2: this check is not uniform when m_ME_seq_list should be properly kept
	long p_signed = 0; long m_signed = 0;
	s=m_Diags_result->get_diagrams_number();
	for(i=0; i<s; i++) {
		CFeynDiagram* d=m_Diags_result->get_diagram(i); if(d==NULL){ continue; }
		CExpression* f=d->get_factor_expr(); if(f==NULL){ p_signed++; continue; }
		if (f->getFlag()) { p_signed++; } else { m_signed++; }
	}
	if (p_signed < m_signed) { // inverse common sign
		for(i=0; i<s; i++) {
			CFeynDiagram* d=m_Diags_result->get_diagram(i); if(d==NULL){ continue; }
			CExpression* f=d->get_factor_expr();
			if(f!=NULL) { f=f->clone(); } else { f=CENumber::GetUnit(); }
			f->inverseFlag(); if (CEBasic::isUnit(f)) { delete f; f=NULL; }
			d->set_factor_expr(f); if(f!=NULL)delete f;
		}
		dbg->log(45, "Diagrams common sign inversed");
	}

	dbg->log(25, s, "Diagrams restored form classes (total %ld diagrams)");
}

void CMEComposeLib::RestoreClassToDiagrams(CFeynDiagram* dClass, long nNextVert) {
	dbg->log(75, nNextVert, "RestoreClassToDiagrams (nNextVert: %ld)",dClass);
	if (dClass==NULL) return;
	if (nNextVert<0)nNextVert=0;
	int i;

	CEPtrArray* verts=dClass->get_vertices_list();
	int vertsNum=verts->size();
	if (nNextVert>vertsNum) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams()"); return;}
	if (nNextVert==vertsNum) {
		dbg->log(105,"diagram added:",dClass);
		m_Diags_result->add_diagram(dClass);
		return;
	}
	
	CDVertex* v = (CDVertex*)verts->GetAt(nNextVert); if(v==NULL)CEException::halt("CMEComposeLib::RestoreClassToDiagrams ERROR: NULL vertex");
	CDiagramVertexInfo* vi=v->get_vertex_info(); if(vi==NULL)CEException::halt("CMEComposeLib::RestoreClassToDiagrams ERROR: NULL vertex-info ref.");
	int s=vi->get_tales_num();
	int vID=vi->get_vertex_ID();
	long viMultiplicity=vi->CalcPartsIdentityOrder();
	//bool isCT=vi->check_vert_type(CDiagramVertexInfo::VERTEX_COUNTERTERM);

	CIntArray idAltVerts;
	CEPtrArray* arr_VIs=m_phys->getVerticesInfo();
	long varsN=m_phys->CollectVerticesForClass(vID,&idAltVerts);
	if(varsN!=idAltVerts.GetSize()) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams(), #4"); return; }

	long varI;
	for(varI=0; varI<varsN; varI++) { dbg->log(75, varI,varsN,"-- viX (%ld/%ld)",arr_VIs->GetAt(idAltVerts.GetAt(varI))); }
	for(varI=0; varI<varsN; varI++) {
		long vIdX=idAltVerts.GetAt(varI);
		if ((vIdX<0) || (vIdX>=arr_VIs->size())) { dbg->log(0, vIdX, "WARNING: invalid vertex to substitute. Unknown vertex ID (%ld)."); continue; }
		CDiagramVertexInfo* viX=(CDiagramVertexInfo*)arr_VIs->GetAt(vIdX); if(viX==NULL){ dbg->log(0, vIdX, "WARNING: invalid vertex to substitute. NULL vertex-info ref. (%ld)."); continue; }
		if (viX->get_tales_num() != s) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams(), #5"); return; }
		long viXMultiplicity=viX->CalcPartsIdentityOrder();

		/*
		* Adjust diagram factor.
		* {NNAB}->1/2{XYAB}, {NNAB}->{XXAB}, {NNN}->1/6{XYZ}, {NNMM}->1/4{XYAB}
		*/
		CExpression* mFactor=NULL; // to be applied if (viMultiplicity != viXMultiplicity) only
		if(viMultiplicity != viXMultiplicity) {
			CExpression* f=dClass->get_factor_expr();
			if (f!=NULL) { f=f->clone(); } else { f=CENumber::GetUnit(); }
			f=f->multiply(CENumber::GetFraction(viXMultiplicity,viMultiplicity));
			f=CExpression::reduce(f);
			if (CEBasic::isUnit(f)) { delete f; f=NULL; }
			mFactor=f;
		}

		CEPtrArray orders;
		vi->CollectClassFitOrders(viX, &orders); // orders of vi class-links to fit viX-specimen vertex
		if (orders.size()*viXMultiplicity != viMultiplicity) {
			dbg->log(0,viMultiplicity/viXMultiplicity,"WARNING: Class-fit orders invalid array. Size expected: %ld",&orders);
		}

		// Note: do not warry about crossing appearing here!
		// CollectClassFitOrders() will not produce orders that differ by equal-PK exchange

		for(long iOrder=0;iOrder<orders.size();iOrder++) {
			CIntArray* order = (CIntArray*)orders.GetAt(iOrder);
			if (order==NULL) { CEException::halt("Invalid viX->CollectFitOrders() result (NULL)"); continue; }
			if (order->GetSize()!=s) { CEException::halt("Invalid viX->CollectFitOrders() result (invalid size)"); continue; }
			
			CFeynDiagram* dClassCpy=dClass->clone(); if(dClassCpy==NULL) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams(), #6"); return; }
			CEPtrArray* vertsCpy=dClassCpy->get_vertices_list(); if (vertsNum!=vertsCpy->size()) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams(), #7"); return; }
			CDVertex* vCpy = (CDVertex*)vertsCpy->GetAt(nNextVert); if(vCpy==NULL) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams(), #8"); return; }

			// sort vertex links according to new order
			if (!vCpy->sortLinksOrder(order)) { dbg->log(0,"ERROR: Cant sort vertex links according to specified order", order, vCpy); CEException::halt(); }

			CEPtrArray* linesCpy=vCpy->get_links(); if(linesCpy==NULL) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams(), #9"); return; }
			if (linesCpy->size() != s) { CEException::halt("Internal ERROR at CMEComposeLib::RestoreClassToDiagrams(), #10"); return; }
			dbg->log(75, "viX",viX);

			bool fits_OK=true;
			for(i=0; i<s; i++) {	
				CDLine* lCpy = (CDLine*)linesCpy->GetAt(i);
				long pk_line = vCpy->GetLogicalPKForAttachedLine(lCpy); long pk_line_abs = CLMath::abs(pk_line);
				long pk_alt = viX->get_particle_kind(i); long pk_alt_abs = CLMath::abs(pk_alt);
				// vi->get_particle_kind(i) should NOT always equal (or *(-1)) to pk_line due to (ClassPK!=ExtLinePK)

				if ((pk_line_abs & CEParticle::PARTICLE_CLASS_FLAG)==CEParticle::PARTICLE_CLASS_FLAG) {
					// additional check
					if (!CEUtil::LEQU(CEParticle::is_neutral_bosonic_PK(pk_alt),CEParticle::is_neutral_bosonic_PK(pk_line))) { CEException::halt("Internal error class->diagrams restoration"); }
					// the sign for neutral boson will be truncated at CDIntLine::set_particle_kind()
					vCpy->SetLogicalPKForAttachedLine(lCpy,pk_alt);
				} else {
					if (pk_line != pk_alt) { // compare also absolute values for NEUTRAL BOSONS
						if ((pk_line_abs & (long)CEParticle::PARTICLE_NEUTRAL_BOSON) == (long)CEParticle::PARTICLE_NEUTRAL_BOSON) {
							if (pk_alt_abs != pk_line_abs) fits_OK=false;
						} else {
							fits_OK=false;
						}
					}
				}
				if (!fits_OK) break;
			}

			if (fits_OK) {
				vCpy->set_vertex_info(viX);
				// Adjust diagram factor as calculated before
				if(viMultiplicity != viXMultiplicity) { dClassCpy->set_factor_expr(mFactor); }
				RestoreClassToDiagrams(dClassCpy, nNextVert+1);
			} else {
				delete dClassCpy;
			}
		}
		if (mFactor!=NULL) delete mFactor;
	}
	delete dClass;
}

long CMEComposeLib::DropZeroDiagrams() {	
	m_kin->reset_volatile_variables(true);
	//	*****************************************************	<scroll>
	CEProgressWatch* pw = NULL; long d_num=m_Diags_result->get_diagrams_number();
	if (d_num>=500) { pw = CEProcessor::newProgressWatch("Dropping zero diagrams (total %ld diagrams)", d_num); }
	//	*****************************************************	<\scroll>
	long nDropped=0;
	for(long di=0; di<m_Diags_result->get_diagrams_number(); di++) {
		//	*****************************************************	<scroll>
		if(pw!=NULL) { pw->step(di+nDropped, d_num); }
		//	*****************************************************	<\scroll>
		CFeynDiagram* d=m_Diags_result->get_diagram(di); if(d==NULL){dbg->log(0,"WARNING:NULL diagram in set",m_Diags_result); continue; }
		CMatrixElement* me=d->RetrieveMatrixElement(); if(me==NULL) continue;
		CExpression* eMe=me->get_expr_val();
		if(eMe!=NULL) {
			CExpression* eMeA=CExpression::arrge(eMe); //To move out Sp[0] items
			dbg->log(95,di,"Testing diagram #%ld for zero matrix element. Me=",eMeA);
			if (CEBasic::isZero(eMeA)) { m_Diags_result->remove_diagram(di); di--; nDropped++; }
			if (eMeA!=eMe) delete eMeA;
		}
		delete me;
	}
	//	*****************************************************	<scroll>
	if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
	//	*****************************************************	<\scroll>
	m_kin->reset_volatile_variables(false);
	return nDropped;
}
