////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// MEOpSeqLoopsLib.cpp: implementation of the CMEOpSeqLoopsLib class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CMEOpSeqLoopsLib::dbg = new CDebug("CMEOpSeqLoopsLib");
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CMEOpSeqLoopsLib::CMEOpSeqLoopsLib(CMEOperatorSeq* seq, CKinematics* kin) : CELibrary() {
	if ((seq == NULL) || (kin == NULL)) CEException::halt(dbg, "NULL");
	m_seq=seq; m_kin=kin; m_loops_data = NULL; m_route_data = NULL; m_point_go_marks = NULL;
}

CMEOpSeqLoopsLib::~CMEOpSeqLoopsLib() {
	m_seq = NULL; m_kin = NULL;
	if (m_loops_data != NULL) { delete m_loops_data; m_loops_data = NULL; }
	if (m_route_data != NULL) { delete m_route_data; m_route_data = NULL; }
	if (m_point_go_marks != NULL) {
		// structure: Arr{NULL, NULL, Arr{Non-deletable objects}, ...}
		long s = m_point_go_marks->size();
		long i;
		for(i=0; i<s; i++) {
			CEPtrArray* item = (CEPtrArray*)m_point_go_marks->GetAt(i);
			if (item != NULL) item->clearall_without_delete_objects();
		}
		delete m_point_go_marks; m_point_go_marks = NULL;
	}
}

//////////////////////////////////////////////////////////////////////
// Static Functions
//////////////////////////////////////////////////////////////////////

/*
*	Result: Array{LoopData, LoopData, LoopData} or NULL
*	LoopData: CMESeqFORoute array{} -- internal library srtucture, may be volatile changed.
*/
CEPtrArray* CMEOpSeqLoopsLib::GetLoopsData(CMEOperatorSeq* seq, CKinematics* kin) {
	if ((seq == NULL) || (kin == NULL)) CEException::halt(dbg, "NULL");
	dbg->log(25, "GetLoopsData", seq);
	CMEOpSeqLoopsLib* lib = new CMEOpSeqLoopsLib(seq, kin);
	CEPtrArray* res = lib->GetLoopsData();
	delete lib;

	return res;
}

// PK should be uniform throughout the route
bool CMEOpSeqLoopsLib::CanBeInversed(CMESeqFORoute* loop_data) {
	if (loop_data == NULL) CEException::halt(dbg, "NULL");

	int pk_in = 0; // unassigned
	int pk_out = 0; // unassigned
	long s = loop_data->size();
	CEFieldOperator* fo_prev = NULL;
	long i;
	for(i=0; i<s; i++) {
		CEPtrArray* fo_pair = (CEPtrArray*)loop_data->GetAt(i); if (fo_pair->size() != 2) CEException::halt(dbg, "Size");
		CEFieldOperator* fo = (CEFieldOperator*)fo_pair->GetAt(0);
		CEFieldOperator* fo_2 = (CEFieldOperator*)fo_pair->GetAt(1);

		if (i == 0) { pk_in = fo->getParticleKind(); pk_out = fo_2->getParticleKind(); }
		else {
			if ((pk_in != fo->getParticleKind()) ||
				(pk_out != fo_2->getParticleKind())) return false;
		}
	}
	return true;
}

bool CMEOpSeqLoopsLib::HaveCommonLines(CEPtrArray* loops_data) {
	if (loops_data == NULL) CEException::halt(dbg, "NULL");
	long s = loops_data->size();
	if (s<2) return false;
	long i;
	for (i=0; i<(s-1); i++) {
		CMESeqFORoute* loop_I = (CMESeqFORoute*)loops_data->GetAt(i);
		long j;
		for (j=i+1; j<s; j++) {
			CMESeqFORoute* loop_J = (CMESeqFORoute*)loops_data->GetAt(j);
			if (loop_I->haveCommonLineWith(loop_J)) return true;
		}
	}
	return false;
}

/*
*	Note: if loop inversed destroys adjacent loops data
*	It can be fixed, by arranging adjacent loops in this procedure
*	(additional parameter CEPtrArray* other_loops is required)
*/
void CMEOpSeqLoopsLib::InverseLoop(CMESeqFORoute* loop_data) {
	if (loop_data == NULL) CEException::halt(dbg, "NULL");
	if (!CanBeInversed(loop_data)) { dbg->log(0, "Warning: try to inverse non-inversable loop"); return; }
	dbg->log(25, "InverseLoop", loop_data);

	long s = loop_data->size();
	CMESeqFORoute* loop_data_copy = loop_data->cloneRoute();
	long i;
	for(i=0; i<s; i++) {
		//CEFieldOperator* L_prev = loop_data_copy->getFO(((i>0)?(i-1):(s-1)),0);
		CEFieldOperator* L_next = loop_data_copy->getFO(((i<(s-1))?(i+1):0),0);
		CEFieldOperator* R_prev = loop_data_copy->getFO(((i>0)?(i-1):(s-1)),1);
		//CEFieldOperator* R_next = loop_data_copy->getFO(((i<(s-1))?(i+1):0),1);
		//CEFieldOperator* R_curr = loop_data_copy->getFO(i,1);
		if ((L_next == NULL) || (R_prev == NULL)) CEException::halt(dbg, "NULL operator whilein loop inversing");
		//R_next->setCouplingNo(L_prev->getCouplingNumber());
		L_next->setCouplingNo(R_prev->getCouplingNumber());
		loop_data->setPair(i, L_next, R_prev);
	}
	delete loop_data_copy;

	// note: loop_data order should olso be inversed

	dbg->log(25, "Loop inversed:", loop_data);
}

// compare coupling numbers & point IDs
bool CMEOpSeqLoopsLib::ValidateLoopData(CMESeqFORoute* loop_data) {
	if (loop_data == NULL) return false;

	long s = loop_data->size();
	if (s == 0) return false;
	int last_pID = -1;
	long i;
	for(i=0; i<s; i++) {
		CEPtrArray* fo_pair = (CEPtrArray*)loop_data->GetAt(i);
		if (fo_pair->size() != 2) return false;

		// compare coupling numbers
		CEFieldOperator* fo = (CEFieldOperator*)fo_pair->GetAt(0);
		CEFieldOperator* fo_2 = (CEFieldOperator*)fo_pair->GetAt(1);
		if ((!fo->isCoupled()) || (!fo_2->isCoupled())) return false;
		if (fo->getCouplingNumber() != fo_2->getCouplingNumber()) return false;

		if (CLMath::abs(fo->getParticleKind()) !=
			CLMath::abs(fo_2->getParticleKind())) return false;

		// compare fo-point to previous
		if (last_pID >= 0) { if (last_pID != fo->getPointID()) return false; }
		last_pID = fo_2->getPointID();
	}
	if ((last_pID >= 0) && (s>0)) { // check route circuit
		CEPtrArray* fo_pair_0 = (CEPtrArray*)loop_data->GetAt(0);
		CEFieldOperator* fo_0 = (CEFieldOperator*)fo_pair_0->GetAt(0);
		if (last_pID != fo_0->getPointID()) return false;
	}
	return true;
}

/*
*	Note: this check is quite smple & may not cover all the cases
*/
bool CMEOpSeqLoopsLib::CheckLoopOrder(CMESeqFORoute* loop_data, CMEOperatorSeq* seq, CKinematics* kin) {
	if ((loop_data == NULL) || (seq == NULL) || (kin == NULL)) CEException::halt(dbg, "NULL");
	long s = loop_data->size();
	if (s<=2) return false;

	int num_L = kin->num_init_particles();
	int num_R = kin->num_final_particles();
	int L_part_no = 0;
	int R_part_no = 0;
	for(;;) { // consider various external links
		CParticleWaveFunction* wf_L = seq->GetWaveFunction(kin->get_particle(L_part_no, true));
		CParticleWaveFunction* wf_R = seq->GetWaveFunction(kin->get_particle(R_part_no, false));
		if ((wf_L == NULL) || (wf_R == NULL)) CEException::halt(dbg, "Internal Error. {NULL Patricle}.");

		// calc route len for every FO in loop
		// find FO, closer to L & FO, closer to R
		int pos_L = -1; int pos_R = -1;
		int pos_L_val = INT_MAX; int pos_R_val = INT_MAX;
		long i;
		for(i=0; i<s; i++) {
			CEFieldOperator* fo = loop_data->getFO(i, 0);
			int l_val = seq->CalcLinksNum(wf_L, fo);
			if (l_val < pos_L_val) { pos_L = i; pos_L_val = l_val; }
			int r_val = seq->CalcLinksNum(wf_R, fo);
			if (r_val < pos_R_val) { pos_R = i; pos_R_val = r_val; }
		}

		// order criteria: loop should flows from L to R
		if (pos_R != pos_L) {
			int del_abs = CLMath::abs(pos_R-pos_L);
			bool is_order_OK = CEUtil::LEQU((del_abs*2<s), (pos_R>pos_L));
			return is_order_OK;
		}

		// get next i-... than f- particles
		L_part_no++;
		if (L_part_no >= num_L) {
			L_part_no=0;
			R_part_no++; if (R_part_no >= num_R) break;
		}
	}
	dbg->log(2, "WARNING, Can't arrange loop order in operator sequence {Loop Data, Sequence}", loop_data, seq);
	return false;
}

//////////////////////////////////////////////////////////////////////
// Internal Implementation Functions
//////////////////////////////////////////////////////////////////////
CEPtrArray* CMEOpSeqLoopsLib::GetLoopsData() {
	if (m_loops_data != NULL) { delete m_loops_data; m_loops_data = NULL; } // init internal data
	if (m_seq->have_uncoupled_field_op()) { dbg->log(0, "WARNING: unexpected uncoupled FO spotted. Loop Data are unavailable."); return NULL; }

	// get first fo-coupling
	CParticleWaveFunction* wf1 = m_seq->GetWaveFunction(m_kin->get_particle(0, true));
	if (wf1 == NULL) { dbg->log(0, "Internal Error. Loop Data are unavailable."); return NULL; }
	long start_point = wf1->getPointID();

	// route data for every point
	int s = m_seq->getMaxPointID()+1;
	if (m_route_data != NULL) delete m_route_data;
	if (m_point_go_marks != NULL) { m_point_go_marks->clearall_without_delete_objects(); delete m_point_go_marks; }
	m_route_data = new CEPtrArray(true);
	// m_point_go_marks: to mark the routed directions from the point
	// m_point_go_marks structure: Arr{NULL, NULL, Arr{Non-deletable objects}, ...}
	m_point_go_marks = new CEPtrArray(true);
	// mark all points as unrouted, let mark arrays to be created immediately
	int i;
	for (i=0; i<s; i++) { m_route_data->add(NULL); m_point_go_marks->add(new CEPtrArray()); }
	if (m_route_data->size() != s) { dbg->log(0, "Internal Error in CEPtrArray. Loop Data are unavailable."); return NULL; }
	m_route_data->SetAt(start_point, new CMESeqFORoute());

	// calc all routes and extract loops data
	CalcRoutesFromPoint(start_point, NULL);

	delete m_route_data; m_route_data = NULL;
	for(i=0; i<s; i++) { CEPtrArray* item = (CEPtrArray*)m_point_go_marks->GetAt(i);
		if (item != NULL) item->clearall_without_delete_objects(); }
	delete m_point_go_marks; m_point_go_marks = NULL;
	CEPtrArray* res = m_loops_data; m_loops_data = NULL;
	return res;
}

// calc routes from specific point (except to point, connected by 'fo_prev' FO).
void CMEOpSeqLoopsLib::CalcRoutesFromPoint(int pID, CEFieldOperator* fo_prev) {
	CMESeqFORoute* curr_route = (CMESeqFORoute*)m_route_data->GetAt(pID);
	if (curr_route == NULL) { CEException::halt("NULL Route"); }

	int prev_spotted = false; // for internal check only
	long s = m_seq->size();
	long i;
	for(i=0; i<s; i++) {
		long next_pos = m_seq->GetNextPFPosAtPoint(pID, i, 0);
		if (next_pos < 0) break;
		i = next_pos;
		
		CEParticleFunction* pf = m_seq->GetPFAtPos(next_pos);
		if (pf == fo_prev) { prev_spotted = true; continue; }
		if (!pf->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)pf;
		if (!fo->isCoupled()) continue;

		// check if this bond is already routed (from other side)
		bool skip_this_route = false;
		CEPtrArray* point_marks = (CEPtrArray*)m_point_go_marks->GetAt(pID);
		int marks_i;
		for(marks_i=0; marks_i<point_marks->size(); marks_i++) {
			if (point_marks->GetAt(marks_i) == fo) skip_this_route = true; }
		if (skip_this_route) continue;


		CEFieldOperator* fo_cpl = m_seq->GetCoupledFO(fo);
		int next_pID = fo_cpl->getPointID();
		CMESeqFORoute* next_route = curr_route->cloneRoute();
		next_route->addPair(fo, fo_cpl);
		
		CMESeqFORoute* prev_route = (CMESeqFORoute*)m_route_data->GetAt(next_pID);
		if (prev_route != NULL) { // loop found
			// mark fo_cpl as "routed from other side"
			CEPtrArray* next_point_marks = (CEPtrArray*)m_point_go_marks->GetAt(next_pID);
			next_point_marks->add(fo_cpl);

			// create loop data
			CMESeqFORoute* loop = GetRoutesLoop(prev_route, next_route);
			delete next_route;
			if (loop == NULL) { CEException::halt("NULL Loop created"); }

			// store loop data
			if (m_loops_data == NULL) m_loops_data = new CEPtrArray();
			m_loops_data->add(loop);
		} else {
			m_route_data->SetAt(next_pID, next_route);
			CalcRoutesFromPoint(next_pID, fo_cpl);
		}
	}
	if ((fo_prev != NULL) && (!prev_spotted)) dbg->log(0, "Error: interanal procedure check fails in {CalcRoutesFromPoint()}.");
}

// Note: All the routes starts from common sequence
CMESeqFORoute* CMEOpSeqLoopsLib::GetRoutesLoop(CMESeqFORoute* r1, CMESeqFORoute* r2) {
	if ((r1 == NULL) || (r2 == NULL)) CEException::halt(dbg, "GetRoutesLoop(NULL)");
	dbg->log(55, "GetRoutesLoop", r1, r2);
	long s1 = r1->size();
	long s2 = r2->size();

	// skip common sequence
	long i=0;
	for (;((i<s1) && (i<s2));i++) { if (!r1->equalPairs(i, (CEPtrArray*)r2->GetAt(i))) break; }

	// copy pairs to loop route
	CMESeqFORoute* res = new CMESeqFORoute();
	long j;
	for(j=i; j<s1; j++) { res->addPairCopyFrom(r1, j, false); }
	for (j=(s2-1); j>=i; j--) { res->addPairCopyFrom(r2, j, true); }
	dbg->log(25, "GetRoutesLoop result:", res);

	// internal checks
	if (res->size() == 0) CEException::halt(dbg, "GetRoutesLoop Internal Error: 0-sized loop created.");
	if (!ValidateLoopData(res)) { dbg->log(0, "GetRoutesLoop Internal ERROR: Non valid loop data created", res); CEException::halt(); }
	return res;
}
