////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// MEOperatorSeq.cpp: implementation of the CMEOperatorSeq class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CMEOperatorSeq::dbg = new CDebug("CMEOperatorSeq");
DWORD CMEOperatorSeq::m_log_settings = 0;
const DWORD CMEOperatorSeq::ME_SEQ_SORT_LOG				= 0x10;
const DWORD CMEOperatorSeq::ME_SEQ_SORT_ADVANCED_LOG	= 0x11;
const DWORD CMEOperatorSeq::ME_SEQ_SORT_COMPLETE_LOG	= 0x12;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMEOperatorSeq::CMEOperatorSeq(const CEProduct* _seq, int _multiplicity, int _max_coupling_no, bool _ferm_sign) : CEProduct() {
	if (_seq == NULL) { CEException::halt("CMEOperatorSeq: NULL constructor data"); }

	m_multiplicity = _multiplicity;
	m_max_coupling_no = _max_coupling_no;

	long s = _seq->size();
	long i;
	for(i=0; i<s; i++) { addItem(_seq->GetAtConst(i)->clone()); }
	_seq->copyAttributesTo(this);

	m_ferm_sign = _ferm_sign;
}

CMEOperatorSeq::~CMEOperatorSeq() {}

CExpression* CMEOperatorSeq::clone() const {
	// Note: attributes are copying in construction
	return new CMEOperatorSeq(this, m_multiplicity, m_max_coupling_no, m_ferm_sign);
}

// arrangement should be inline
// Field operators MAY NOT be turned to POWERS
// TODO: implement it via NO_POWERS property of FieldOperator objects (and it using in CEProduct)!
CExpression* CMEOperatorSeq::Arrange(int level) {
	CEProduct::Arrange(AL_INLINE);

	// Note: field operators MAY NOT be turned to POWERS, expand powers back!
	// TODO: implement it via NO_POWERS property of FieldOperator objects (and it using in CEProduct)!
	for(long i=0;i<size();i++) {
		CExpression* item = GetAt(i);
		if (!CEUtil::equals(item->ClassName(), CEPower::CLASS_NAME)) continue;
		CEPower* itm_pow = (CEPower*)item;
		if (!(itm_pow->basis())->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;

		// inline power
		CExpression* itm_prod = itm_pow->as_product();
		RemoveAt(i);
		long itm_len = inline_product(i,itm_prod); delete itm_prod;
		i += (itm_len-1);
	}

	return this;
}

// no expand may be performed
CExpression* CMEOperatorSeq::expand(bool recursive) { return this; }

bool CMEOperatorSeq::equals(CExpression* other) {
	if (other == NULL) return false;
	if (!CEUtil::equals(ClassName(), other->ClassName())) return false;
	if (!CEProduct::equals(other)) return false;
	return (get_multiplicity() == ((CMEOperatorSeq*)other)->get_multiplicity());
}

/*bool CMEOperatorSeq::less_than(CExpression* other) {
	return CEProduct::less_than(other); // do not compare muplicity
}*/

const char* CMEOperatorSeq::CLASS_NAME = "CMEOperatorSeq";
const char* CMEOperatorSeq::ClassName() const { return CLASS_NAME; }

CEString* CMEOperatorSeq::toString(int mode) {
	CEString* res = new CEString("\\([");
	res->AddString("\\&");
	if (m_ferm_sign) res->add_int(m_multiplicity);
	else { res->AddString("\\(-"); res->add_int(m_multiplicity); res->AddString("\\)"); }
	res->AddString(CEProduct::toString(mode));
	res->AddString("]\\)");
	return res;
}

int CMEOperatorSeq::get_multiplicity() { return m_multiplicity; }
void CMEOperatorSeq::inc_multiplicity(int val) { m_multiplicity += val; }
void CMEOperatorSeq::set_multiplicity(int val) { if(val>0) m_multiplicity=val; }
int CMEOperatorSeq::getMaxCouplingNo() { return m_max_coupling_no; }
int CMEOperatorSeq::incMaxCouplingNo() { m_max_coupling_no++; return m_max_coupling_no; }
bool CMEOperatorSeq::get_ferm_sign() { return m_ferm_sign; }
void CMEOperatorSeq::inverse_ferm_sign() { m_ferm_sign = !m_ferm_sign; }
bool CMEOperatorSeq::getStandardOrderSign() {
	bool res = m_ferm_sign;
	if (!calcFOCouplingsSeqSign()) { res = !res; } // sign from wrong-sorted FO-couplings
	return res;
}

CMatrixElement* CMEOperatorSeq::ConvertToME() {
	dbg->log(0, "Wraning: CMEOperatorSeq::ConvertToME() is not implemented, diagrams engine should be used");
	return NULL;
}

/*
*	checks if sequence vanishes after the kinematics is applied
*/
bool CMEOperatorSeq::is_kin_vanishing(CKinematics* kin, bool allow_uncoupled, bool classes_only) {
	dbg->log(155, "is_kin_vanishing()", this);

	// 1. couple externals in arbitrary order (work with copied object)
	CMEOperatorSeq* seq_copy = (CMEOperatorSeq*)clone();
	int is = kin->num_init_particles();
	int fs = kin->num_final_particles();
	int _PID;
	for(_PID = -fs; _PID<=is; _PID++) {
		if (_PID == 0) continue;

		long curr_size = seq_copy->size();
		bool part_is_coupled = false;
		CEParticle* part = kin->get_particle_by_ID(_PID);
		long i;
		for (i=0; i<curr_size; i++) {
			CExpression* itm_expr = seq_copy->GetAt(i);
			if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
			CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
			
			// couple to this operator
			if ( fo->CanBeCoupledExternal(part,(_PID>0),classes_only) ) {
				seq_copy->RemoveAt(i); part_is_coupled = true; break;
			}
		}
		if (!part_is_coupled) {
			//dbg->log(205, "is_kin_vanishing() : true (!part_is_coupled)", part, this);
			delete seq_copy; return true;
		}
	}
	
	// 2. search for uncoupled operators in sequence
	if (!allow_uncoupled) {
		if (seq_copy->have_uncoupled_field_op()) {
			//dbg->log(205, "is_kin_vanishing() : true (have_uncoupled_field_op)", this);
			delete seq_copy; return true;
		}
	}

	delete seq_copy;
	dbg->log(105, "is_kin_vanishing() : false", this);
	return false;
}

int CMEOperatorSeq::calcSeqOrder(CDiagramPhysics* _phys, bool search_CT) {
	int res = 0;
	CEPtrArray* _vertices_info = _phys->getVerticesInfo();

	// calc vertices order
	int s = getMaxPointID();
	int i;
	for(i=0; i<=s; i++) {
		int vert_ID = getVertexIDAtPoint(i);
		CDiagramVertexInfo* vi = (CDiagramVertexInfo*)_vertices_info->GetAt(vert_ID);
		if (vi == NULL) { dbg->log(0, vert_ID, "<ERROR> no VERTEXINFO found for vertex #%ld"); CEException::halt(); }
		res += vi->get_vertex_order();
	}

	if (search_CT) {
		// search CT propagators
		for(i=1; i<=m_max_coupling_no; i++) { // do not sort the last
			CEFieldOperator* fo_i = GetFieldOperatorCoupled(i, 0);
			if (fo_i == NULL) { dbg->log(0, "<Error> No field operators found at coupling #%ld", this); CEException::halt(); }
			if (fo_i->isCounterTermCoupling()) res += 2;
		}
	}
	return res;
}

/*
*	counter-terms are vertices & propagators
*/
int CMEOperatorSeq::calcCounterTermsNum(CDiagramPhysics* _phys) {
	int res = 0;

	CEPtrArray* _vertices_info = _phys->getVerticesInfo();
	// calc CT vertices
	int s = getMaxPointID();
	int i;
	for(i=0; i<=s; i++) {
		int vert_ID = getVertexIDAtPoint(i);
		CDiagramVertexInfo* vi = (CDiagramVertexInfo*)_vertices_info->GetAt(vert_ID);
		if (vi == NULL) { dbg->log(0, vert_ID, "<ERROR> no VERTEXINFO found for vertex #%ld (2)"); CEException::halt(); }
		if (vi->check_vert_type(CDiagramVertexInfo::VERTEX_COUNTERTERM)) res++;
	}

	// calc CT propagators
	for(i=1; i<=m_max_coupling_no; i++) { // do not sort the last
		CEFieldOperator* fo_i = GetFieldOperatorCoupled(i, 0);
		if (fo_i == NULL) { dbg->log(0, "<Error> No field operators found at coupling #%ld", this); CEException::halt(); }
		if (fo_i->isCounterTermCoupling()) res++;
	}

	dbg->log(35, res, "calcCounterTermsNum() result: %ld for sequence", this);
	return res;
}

bool CMEOperatorSeq::have_uncoupled_field_op() {
	bool res = false;
	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
		if (fo->isCoupled()) continue;
		res = true; break;
	}
	dbg->log(60, res, "have_uncoupled_field_op() result: %ld for sequence", this);
	return res;
}

void CMEOperatorSeq::sort_internal(CEPtrArray* uniform_index_list) {
	dbg->log(15, "sort_internal()", this);

	CMEOperatorSeq* old_seq = NULL;
	if (m_log_settings != 0) old_seq = (CMEOperatorSeq*)clone();

	// 1. arrange coupling numbers by particle kind number
	//SortOpCouplsNumsByPK();
	//dbg->log(25, "SortOpCouplsNumsByPK() completed", this);

	// 2. sort point IDs (by particle kind)
	SortPointIDs();
	dbg->log(25, "SortPointIDs() completed", this);

	// 2.5. sort coupling numbers according to point ID & PK
	SortOpCouplingNumbers();
	dbg->log(25, "couplings numbers sorted", this);

	// 3. Arrange spinor couplings according to point IDs
	ArrangeSpinorCouplingsByPoint();
	dbg->log(25, "ArrangeSpinorCouplingsByPoint() completed", this);

	/*
	// 2x. sort spinor sequences by coupling numbers
	SortSpinorSeqs();

	// 3x. rename point ID according to spinors order
	ArrangePointIDs();
	*/
	
	// 4. rename (unify) vector indices according to point IDs order
	UnifyVectorIndices(uniform_index_list);
	//dbg->log(2, "UnifyVectorIndices() completed", this, uniform_index_list);
	dbg->log(25, "UnifyVectorIndices() completed", this);

	// arrange
	Arrange(AL_INLINE);
	dbg->log(15, "sort_internal() completed", this);

	// self-check procedure
	if (m_log_settings != 0) {
		if (!equals(old_seq)) {
			// log non-stable sorting only
			int lev_old_a = dbg->set_level(10); // do not log this sorting
			CMEOperatorSeq* seq2 = (CMEOperatorSeq*)clone();
			seq2->sort_internal(uniform_index_list);
			dbg->set_level(lev_old_a);

			if (!equals(seq2)) {
				if (m_log_settings == ME_SEQ_SORT_ADVANCED_LOG) {
					int lev_old = dbg->set_level(100);
					dbg->log(5, "sort_internal() non-stable sorting: source:", old_seq);
					dbg->log(5, "sort_internal() result", this);
					dbg->log(5, "sort_internal() result #2", seq2);
					dbg->set_level(lev_old);
				}
				else if (m_log_settings == ME_SEQ_SORT_COMPLETE_LOG) {
					int lev_old = dbg->set_level(100); DWORD _log_setts_old = m_log_settings; m_log_settings = 0;
					dbg->log(5, "sort_internal() non-stable sorting: source:", old_seq);
					dbg->log(5, "sort_internal() result", this);
					dbg->log(5, "sort_internal() result #2", seq2);
					dbg->log(5, "first sorting reconstruction");
					old_seq->sort_internal(uniform_index_list); delete old_seq;
					dbg->log(5, "further sorting reconstruction");
					old_seq = (CMEOperatorSeq*)clone();
					old_seq->sort_internal(uniform_index_list);
					m_log_settings = _log_setts_old; dbg->set_level(lev_old);
				}
			}
			delete seq2;
		}
		delete old_seq;
	}
}

// same to equals() except the multiplicity comparison
bool CMEOperatorSeq::equals_topology(CMEOperatorSeq* other) {
	if (other == NULL) return false;
	if (!CEUtil::equals(ClassName(), other->ClassName())) return false;
	return CEProduct::equals(other);
}

void CMEOperatorSeq::SortOpCouplsNumsByPK() {
	int i;
	for(i=1; i<m_max_coupling_no; i++) { // do not sort the last
		int min_PK = GetOpCouplePK(i);
		int min_couple = i;

		int j;
		for(j=i+1; j<=m_max_coupling_no; j++) {
			int j_PK = GetOpCouplePK(j);
			if (j_PK < min_PK) { min_couple=j; min_PK=j_PK; }
		}

		if (min_couple != i) ExchangeOpCouplingNumbers(i, min_couple);
	}
}

// -1 if no couple found (operator's PK is always positive)
int CMEOperatorSeq::GetOpCouplePK(int couple_no) {
	int res = -1;
	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
		if (fo->getCouplingNumber() == couple_no) {
			res = fo->getParticleKind();
			break;
		}
	}
	return res;
}

// op_no == 0 or 1
CEFieldOperator* CMEOperatorSeq::GetFieldOperatorCoupled(int couple_no, int op_no) {
	if ((op_no < 0) || (op_no > 1)) { dbg->log(0, op_no, "<Error> Invalid operator number parameter: %ld"); CEException::halt(); }

	int s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
		if (fo->getCouplingNumber() == couple_no) {
			if (op_no > 0) op_no--;
			else return fo;
		}
	}
	return NULL;
}

int CMEOperatorSeq::getVertexIDAtPoint(int pt) {
	// find first operator/wave function at point
	int pf_pos = GetNextPFPosAtPoint(pt, 0, 0);
	if (pf_pos < 0) {
		dbg->log(0, pt, "error: no FO found at point %ld", this);
		//dbg->set_level(1000); pf_pos = GetNextPFPosAtPoint(pt, 0, 0);
		CEException::halt();
	}
	CEParticleFunction* pf = (CEParticleFunction*)GetPFAtPos(pf_pos);
	return pf->getVertexID();
}

void CMEOperatorSeq::SortPointIDs() {
	long s = getMaxPointID();
	dbg->log(35, s, "SortPointIDs(), max point ID: %ld");
	long i;
	for(i=0; i<s; i++) { // do not sort the last
		m_curr_sort_point = i;
		int min_ID = i;
		long j;
		for(j=i+1; j<=s; j++) {
			int cmp_points = ComparePointData(min_ID, j);
			dbg->log(55, cmp_points, "points comparison result: %ld");
			if (cmp_points < 0) { min_ID=j; }
		}
		if (min_ID != i) ExchangePointIDs(i, min_ID);
	}
}

/*
*	1 means a should be before b
*	sorting reasons: PK, operators number at point
*	<?> + vertex ID
*	<?> + coupling of particle with the lowest PK
*/
int CMEOperatorSeq::ComparePointData(int a, int b) {
	dbg->log(65, a, b, "ComparePointData(%ld, %ld)", this);
	if (a == b) return 0;

	//int res = CompareFieldOperators(GetFermFOAtPoint(a, true), GetFermFOAtPoint(b, true));
	//if (res != 0) return res;
	//res = CompareFieldOperators(GetFermFOAtPoint(a, false), GetFermFOAtPoint(b, false));
	//if (res != 0) return res;
	//int boson_num_a = GetNumFOAtPoint(a, CEParticle::PARTICLE_VECTOR_BOSON);
	//int boson_num_b = GetNumFOAtPoint(b, CEParticle::PARTICLE_VECTOR_BOSON);
	//if (boson_num_a != boson_num_b) return CLMath::step(boson_num_a, boson_num_b);
	//if (boson_num_a == 1) {
	//	res = CompareFieldOperators(
	//		GetFOAtPos(GetNextFOPosAtPoint(a, 0, CEParticle::PARTICLE_VECTOR_BOSON)),
	//		GetFOAtPos(GetNextFOPosAtPoint(b, 0, CEParticle::PARTICLE_VECTOR_BOSON))
	//		);
	//	if (res != 0) return res;
	//}

	// Note: similar Vert-IDs may have different particles PK set!
	// this is due to CLASS-vertex coupled to fixed external field

	//1. compare external fields coupled from the lowest PK
	long pk_last = -CEParticle::PARTICLE_KIND_MAX;
	for(;;) { // every pk
		// find out lowest pk (>pk_last) at a-point
		long pk_curr_a = FindNextMinPKAtPoint(pk_last, a, true, true);
		long pk_curr_b = FindNextMinPKAtPoint(pk_last, b, true, true);
		if ((pk_curr_a==0)&&(pk_curr_b==0)) break;
		if(pk_curr_a==0)return -1;
		if(pk_curr_b==0)return 1;
		if(pk_curr_a != pk_curr_b)return CLMath::step(pk_curr_a, pk_curr_b);
		long pk_curr=pk_curr_a;

		// compare pk data
		CEParticleFunction* f_a = NULL; CEParticleFunction* f_b = NULL;
		for(;;) {
			f_a = FindNextMinFunctionPointPK(f_a, a, pk_curr, true); // get min pf (but greater than f_a) for pk
			f_b = FindNextMinFunctionPointPK(f_b, b, pk_curr, true); // get min pf (but greater than f_b) for pk
			if ((f_a == NULL) || (f_b == NULL)) {
				if ((f_a != NULL) || (f_b != NULL)) { if (f_a!=NULL) return 1; if (f_b!=NULL) return -1; }
				break; // both NULLs: no more functions with PK == pk_curr
			}
			int cmp_res = ComparePointFunctions(f_a, f_b); if(cmp_res!=0)return cmp_res;
		}
		pk_last = pk_curr;
	}

	//2. compare fields operators from the lowest PK
	pk_last = -CEParticle::PARTICLE_KIND_MAX;
	for(;;) { // every pk
		// find out lowest pk (>pk_last) at a-point
		long pk_curr_a = FindNextMinPKAtPoint(pk_last, a, true);
		long pk_curr_b = FindNextMinPKAtPoint(pk_last, b, true);
		if ((pk_curr_a==0) && (pk_curr_b==0)) break;
		if (pk_curr_a==0) return -1;
		if (pk_curr_b==0) return 1;
		if (pk_curr_a != pk_curr_b) return CLMath::step(pk_curr_a, pk_curr_b);
		long pk_curr=pk_curr_a;

		// compare pk data
		CEParticleFunction* f_a = NULL; CEParticleFunction* f_b = NULL;
		for(;;) {
			f_a = FindNextMinFunctionPointPK(f_a, a, pk_curr); // get min pf (but greater than f_a) for pk
			f_b = FindNextMinFunctionPointPK(f_b, b, pk_curr); // get min pf (but greater than f_b) for pk
			if ((f_a == NULL) || (f_b == NULL)) {
				if ((f_a != NULL) || (f_b != NULL)) { if(f_a!=NULL)return 1; if(f_b!=NULL)return -1; }
				break; // both NULLs: no more functions with PK == pk_curr
			}
			int cmp_res = ComparePointFunctions(f_a, f_b);
			if (cmp_res != 0) return cmp_res;
		}

		pk_last = pk_curr;
	}

	// compare vertex IDs
	// Note: this check is last to unify Class-NonClass diagrams structure
	int vIDa = getVertexIDAtPoint(a);
	int vIDb = getVertexIDAtPoint(b);
	if (vIDa != vIDb) return CLMath::step(vIDa, vIDb);

	dbg->log(15, a, b, "<Warning> Can't arrange operator sequence (equal points are: %ld, %ld)", this);
	return 0;
}

int CMEOperatorSeq::ComparePointFunctions(CEParticleFunction* f_a, CEParticleFunction* f_b) {
	dbg->log(85, "ComparePointFunctions() (a,b)", f_a, f_b);
	if (f_a == f_b) return 0;
	if ((f_a == NULL) || (f_b == NULL)) { dbg->log(0, "<Internal Error> ComparePointFunctions: NULL parameter"); CEException::halt(); }

	bool a_is_FO = f_a->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR);
	bool b_is_FO = f_b->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR);
	if ((!a_is_FO) && (b_is_FO)) return 1;
	if ((a_is_FO) && (!b_is_FO)) return -1;

	if (a_is_FO) {
		// both are FO
		// compare field operators
		CEFieldOperator* fo_a = (CEFieldOperator*)f_a;
		CEFieldOperator* fo_b = (CEFieldOperator*)f_b;
		if ((fo_a->isCoupled()) && (fo_b->isCoupled())) {
			// compare other link points if they are already sorted
			int pt_a = GetCoupleOtherOpPointID(fo_a);
			int pt_b = GetCoupleOtherOpPointID(fo_b);

			// compare coupled points if they are fixed (at least one of them)
			if ((pt_a < m_curr_sort_point) || (pt_b < m_curr_sort_point)) {
				if (pt_a != pt_b) return CLMath::step(pt_a, pt_b);
			}

			bool a_self = (pt_a == fo_a->getPointID());
			bool b_self = (pt_b == fo_b->getPointID());
			if (a_self) { // self coupling is prior
				if (!b_self) return 1;
			} else {
				if (b_self) return -1;
			}

			if ((!a_self) && (!b_self)) {
				// compare other points Vertex-IDs
				int vIDa2 = getVertexIDAtPoint(pt_a);
				int vIDb2 = getVertexIDAtPoint(pt_b);
				if (vIDa2 != vIDb2) return CLMath::step(vIDa2, vIDb2);
			}
			return 0;
		}

		if (fo_a->isCoupled()) return -1;
		if (fo_b->isCoupled()) return 1;
	} else {
		// both are WF

		// compare particle ID
		CParticleWaveFunction* wf_a = (CParticleWaveFunction*)f_a;
		CParticleWaveFunction* wf_b = (CParticleWaveFunction*)f_b;
		bool mom_cmp = (wf_a->getMomentumSym()->CExpression::less_than(wf_b->getMomentumSym()));
		return ((mom_cmp)?1:(-1));
	}

	// compare PK (useless???)
	// long pk_a = f_a->getParticleKind(); long pk_b = f_b->getParticleKind();
	// if (pk_a != pk_b) { return CLMath::step(pk_a, pk_b); }
	return 0;
}

// note: 0 result means no pk are found
int CMEOperatorSeq::FindNextMinPKAtPoint(long pk_prev, int pID, bool fermions_first, bool wf_only) {
	int res = 0;

	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (itm_expr==NULL) continue;
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) continue;
		if (wf_only) { if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_WAVE_FUNCTION)) continue; }
		CEParticleFunction* pf = (CEParticleFunction*)itm_expr;
		if (pID != pf->getPointID()) continue; // skip operator
		long pk = pf->getLogicalPK();

		if (res == 0) { // compare to prev
			int cmp_res = CEParticle::compare_part_kind(pk_prev, pk, fermions_first);
			if (cmp_res > 0) res = pk;
		} else { // compare to res
			int cmp_res = CEParticle::compare_part_kind(res, pk, fermions_first);
			if (cmp_res < 0) {
				cmp_res = CEParticle::compare_part_kind(pk_prev, pk, fermions_first);
				if (cmp_res > 0) {
					// note: res>prev, pk<res, but pk may also be pk<=prev
					res = pk;
				}
			}
		}
	}

	return res;
}

CEParticleFunction* CMEOperatorSeq::FindNextMinFunctionPointPK(CEParticleFunction* f_prev, int pID, long pk, bool wf_only) {
	bool prev_spotted = false;
	CEParticleFunction* res = NULL;

	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (itm_expr==NULL) continue;
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) continue;
		if (wf_only) { if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_WAVE_FUNCTION)) continue; }
		CEParticleFunction* pf = (CEParticleFunction*)itm_expr;
		if ((pk != pf->getLogicalPK()) || (pID != pf->getPointID())) continue; // skip operator
		if (pf == f_prev) { prev_spotted = true; continue; }

		if (res == NULL) { // compare to prev
			if (f_prev == NULL) {
				res = pf;
			} else {
				int cmp_res = ComparePointFunctions(f_prev, pf);
				if ((cmp_res > 0) || ((cmp_res == 0) && (prev_spotted))) res = pf;
			}
		} else { // compare to res
			int cmp_res = ComparePointFunctions(res, pf);
			if (cmp_res < 0) {
				if (f_prev == NULL) {
					res = pf;
				} else {
					cmp_res = ComparePointFunctions(f_prev, pf);
					if ((cmp_res > 0) || ((cmp_res == 0) && (prev_spotted))) {
						// note: res>prev, pf<res, but pf may also be pf<prev
						res = pf;
					}
				}
			}
		}
	}

	return res;
}

/*int CMEOperatorSeq::CompareFieldOperators(CEFieldOperator* fo_a, CEFieldOperator* fo_b) {
	dbg->log(85, "CompareFieldOperators() (a)", fo_a);
	dbg->log(85, "CompareFieldOperators() (b)", fo_b);

	if (fo_a == fo_b) return 0;
	if (fo_a == NULL) return 1;
	if (fo_b == NULL) return -1;
	
	if ((fo_a->isCoupled()) && (fo_b->isCoupled())) {
		// compare other link points if they are already sorted
		int pt_a = GetCoupleOtherOpPointID(fo_a);
		int pt_b = GetCoupleOtherOpPointID(fo_b);
		if ((pt_a < m_curr_sort_point) || (pt_b < m_curr_sort_point)) {
			if (pt_a != pt_b) return CLMath::step(pt_a, pt_b);
		}
		if (pt_a == fo_a->getPointID()) {
			if (pt_b != fo_b->getPointID()) return 1;
		} else if (pt_b == fo_b->getPointID()) return -1;

		return 0; // do not compare coupling numbers
		//int cn_a = fo_a->getCouplingNumber();
		//int cn_b = fo_b->getCouplingNumber();
		//if (cn_a != cn_b) return CLMath::step(cn_a, cn_b);
	}
	if (fo_a->isCoupled()) return -1;
	if (fo_b->isCoupled()) return 1;

	long pk_a = fo_a->getParticleKind();
	long pk_b = fo_b->getParticleKind();
	if (pk_a != pk_b) return CLMath::step(pk_a, pk_b);

	return 0;
}*/

CEFieldOperator* CMEOperatorSeq::GetCoupledFO(CEFieldOperator* fo) {
	if ((fo == NULL) || (!fo->isCoupled())) { dbg->log(0, "GetCoupledFO() invalid parameter:", fo); CEException::halt(); }
	int cn = fo->getCouplingNumber();
	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* itm_fo = (CEFieldOperator*)itm_expr;
		if ((cn == itm_fo->getCouplingNumber()) && (itm_fo != fo)) return itm_fo;
	}
	dbg->log(0, "<Error> Can't find operator coupled to", fo, this); CEException::halt();
	return NULL;
}

int CMEOperatorSeq::GetCoupleOtherOpPointID(CEFieldOperator* fo) {
	CEFieldOperator* fo_cpl = GetCoupledFO(fo);
	if (fo_cpl == NULL) { dbg->log(0, "GetCoupleOtherOpPos() <Error> Can't find operator coupled to:", fo); CEException::halt(); }
	return fo_cpl->getPointID();
}

/*int CMEOperatorSeq::GetOpPos(CEFieldOperator* fo) {
	if (fo == NULL) { dbg->log(0, "<Error> GetCoupleOtherOpPos(NULL) call"); CEException::halt(); }
	int s = size();
	int i;
	for(i=0; i<s; i++) {
		if (GetAt(i) == fo) return i;
	}
	dbg->log(0, "<Error> Can't find operator... ", fo); dbg->log(0, "... in sequence", this); CEException::halt();
	return -1;
}*/

void CMEOperatorSeq::SortOpCouplingNumbers() {
	int i;
	for(i=1; i<m_max_coupling_no; i++) { // do not sort the last
		int min_couple_no = i;

		int j;
		for(j=i+1; j<=m_max_coupling_no; j++) {
			int cmp_couple_nums = CompareOpCouplingNumbers(min_couple_no, j);
			dbg->log(75, cmp_couple_nums, "coupling numbers cmp result: %ld");
			if (cmp_couple_nums < 0) { min_couple_no=j; }
		}
		
		if (min_couple_no != i) ExchangeOpCouplingNumbers(i, min_couple_no);
	}
}

// '1' means a is 'less' than b (right order)
int CMEOperatorSeq::CompareOpCouplingNumbers(int a, int b) {
	dbg->log(75, a, b, "coupling numbers comparison (%ld, %ld)", this);

	// compare point IDs
	CEFieldOperator* fo_a1 = GetFieldOperatorCoupled(a, 0);
	CEFieldOperator* fo_a2 = GetFieldOperatorCoupled(a, 1);
	CEFieldOperator* fo_b1 = GetFieldOperatorCoupled(b, 0);
	CEFieldOperator* fo_b2 = GetFieldOperatorCoupled(b, 1);
	if ((fo_a1 == NULL) || (fo_a2 == NULL) || (fo_b1 == NULL) || (fo_b2 == NULL)) {
		dbg->log(0, "<Error> comparison couplings #%ld & #%ld. No coupled operators found"); CEException::halt();
	}

	int pt_a1 = fo_a1->getPointID(); int pt_a2 = fo_a2->getPointID();
	int pt_b1 = fo_b1->getPointID(); int pt_b2 = fo_b2->getPointID();
	int min_a = CLMath::lmin(pt_a1, pt_a2); int min_b = CLMath::lmin(pt_b1, pt_b2);
	if (min_a != min_b) return CLMath::step(min_a, min_b);
	int max_a = CLMath::lmax(pt_a1, pt_a2); int max_b = CLMath::lmax(pt_b1, pt_b2);
	if (max_a != max_b) return CLMath::step(max_a, max_b);
	// points are equal here

	CEFieldOperator* low_pt_fo_a = (pt_a1 <= pt_a2)?fo_a1:fo_a2;
	CEFieldOperator* low_pt_fo_b = (pt_b1 <= pt_b2)?fo_b1:fo_b2;
	bool is_ferm_a = CEUtil::equals(low_pt_fo_a->ClassName(), COpFermionField::CLASS_NAME);
	bool is_ferm_b = CEUtil::equals(low_pt_fo_b->ClassName(), COpFermionField::CLASS_NAME);
	if (is_ferm_a && is_ferm_b) {
		// for fermion operators
		// loop spotted: f^1..f^2 * f^2..f^1
		COpFermionField* lpt_ffo_a = (COpFermionField*)low_pt_fo_a;
		COpFermionField* lpt_ffo_b = (COpFermionField*)low_pt_fo_b;
		if (CEUtil::LEQU(lpt_ffo_a->isConjugated(), lpt_ffo_b->isConjugated())) { dbg->log(0, "Internal Error in CompareOpCouplingNumbers(), invalid operator structure", this); CEException::halt(); }
		if (lpt_ffo_a->isConjugated()) return 1;
		return -1;
	}

	// compare pk absolute
	long pk_a = low_pt_fo_a->getParticleKind();
	long pk_b = low_pt_fo_b->getParticleKind();
	if (pk_a != pk_b) return CLMath::step(pk_a, pk_b);

	// pk sign compare
	if (!CEUtil::LEQU(low_pt_fo_a->isConjugated(), low_pt_fo_b->isConjugated())) {
		if (low_pt_fo_a->isConjugated()) return 1;
		return -1;
	}

	// do not compare vector indices! Theq will be unified further
	// 
	/*
	bool is_boson_a = CEUtil::equals(low_pt_fo_a->ClassName(), COpBosonField::CLASS_NAME);
	bool is_boson_b = CEUtil::equals(low_pt_fo_b->ClassName(), COpBosonField::CLASS_NAME);
	if (is_boson_a && is_boson_b) {
		COpBosonField* lpt_bfo_a = (COpBosonField*)low_pt_fo_a;
		COpBosonField* lpt_bfo_b = (COpBosonField*)low_pt_fo_b;
		CTIndex* lpt_vi_a = lpt_bfo_a->getVectorIndex();
		CTIndex* lpt_vi_b = lpt_bfo_b->getVectorIndex();
		CExpression* lpt_ind_a=(lpt_vi_a!=NULL)?lpt_vi_a->index():NULL;
		CExpression* lpt_ind_b=(lpt_vi_b!=NULL)?lpt_vi_b->index():NULL;
		if ((lpt_ind_a!=NULL)&&(lpt_ind_b!=NULL)) {
			if(lpt_ind_a->equals(lpt_ind_b)) {
				if (CEUtil::LEQU(lpt_vi_a->covariance(), lpt_vi_b->covariance())) { dbg->log(0, "Internal Error in CompareOpCouplingNumbers(), invalid operator structure", this); CEException::halt(); }
				if (lpt_vi_a->covariance()) return 1;
				return -1;
			} else {
				if (lpt_ind_a->less_than(lpt_ind_b)) return 1;
				return -1;
			}
		}
	}
	*/

	dbg->log(35, a, b, "Note: unsortable CN pair: %ld, %ld", this);
	// example X--AA--X self energy loop
	// unified after unifying indices
	return 0;
}

void CMEOperatorSeq::ExchangeOpCouplingNumbers(int a, int b) {
	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
		if (!fo->isCoupled()) continue;
		int coupl_no = fo->getCouplingNumber();
		if (coupl_no == a) { fo->setCouplingNo(b); }
		else if (coupl_no == b) { fo->setCouplingNo(a); }
	}
}

void CMEOperatorSeq::ArrangeSpinorCouplingsByPoint() { // sort by pointIDs
	dbg->log(65, "ArrangeSpinorCouplingsByPoint()");
	long s = GetSpinorCouplingsNum();
	long i;
	for (i=0; i<(s-1); i++) {
		int min_pt_ID  = GetFermPFNoX(i, true)->getPointID();
		int min_seq = i;

		long j;
		for(j=i+1; j<s; j++) {
			int j_pt_ID  = GetFermPFNoX(j, true)->getPointID();
			if (j_pt_ID < min_pt_ID) {
				min_pt_ID = j_pt_ID; min_seq = j;
			}
		}
		if (min_seq != i) MoveSpinorCoupling(min_seq, i);
	}
}

/*void CMEOperatorSeq::SortSpinorSeqs() { // sort by coupling numbers
	int s = GetSpinorCouplingsNum();
	int i;
	for(i=0; i<s; i++) {

		int min_seq = i;
		int j;
		for(j=i+1; j<s; j++) {
			if (!CheckSpinorCouplsOrder(min_seq, j)) min_seq = j;
		}
		
		if (min_seq != i) MoveSpinorCoupling(min_seq, i);
	}
}*/

// note: spinors may be both FO and WF but not the usual spinors
int CMEOperatorSeq::GetSpinorCouplingsNum() {
	int res = 0;
	long s = size();
	bool seq_started = false;
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (CEUtil::equals(itm_expr->ClassName(), COpFermionField::CLASS_NAME)) {
			COpFermionField* fo = (COpFermionField*)itm_expr;
			if (CEUtil::LEQU(seq_started, fo->isConjugated())) { dbg->log(0, i, "<ERROR> invalid spinor structure (coupling no. %ld)", this); CEException::halt(); }
			seq_started = !seq_started;
			if (seq_started) res++;
		}
		else if (CEUtil::equals(itm_expr->ClassName(), CWaveFuncFermion::CLASS_NAME)) {
			CWaveFuncFermion* wf = (CWaveFuncFermion*)itm_expr;
			if (CEUtil::LEQU(seq_started, (wf->getParticleKind() < 0))) { dbg->log(0, i, "<ERROR> invalid spinor structure (coupling no. %ld)", this); CEException::halt(); }
			seq_started = !seq_started;
			if (seq_started) res++;
		}
		else if (CEUtil::equals(itm_expr->ClassName(), CESpinor::CLASS_NAME)) {
			dbg->log(0, i, "<ERROR> Unexpected plain spinor spotted (coupling no. %ld)", this); CEException::halt();
		}
	}
	if (seq_started) { dbg->log(0, "<ERROR> invalid spinor structure", this); CEException::halt(); }
	return res;
}

CEParticleFunction* CMEOperatorSeq::GetFermPFNoX(int sp_seq_no, bool get_start_op) {
	//dbg->log(85, sp_seq_no, get_start_op, "GetFermPFNoX(%ld, %ld)");
	int pos = GetSpinorCouplingPos(sp_seq_no, get_start_op);
	if (pos < 0) {
		dbg->log(0, sp_seq_no, "<ERROR> can't find requested operator sequence (%ld)", this); CEException::halt();
	}
	CEParticleFunction* res = GetPFAtPos(pos);

	// check for fermion PK
	if (!CEParticle::is_fermion_PK(res->getParticleKind())) {
		dbg->log(0, "<Internal Error> GetFermPFNoX(): fermion function expected", res); CEException::halt();
	}
	dbg->log(95, sp_seq_no, get_start_op, "GetFermPFNoX(%ld, %ld) result:", res);
	return res;
}

// usage warning: spinors may be both FO, WF but not the usual spinors
/*COpFermionField* CMEOperatorSeq::GetFermFieldOp(int sp_seq_no, bool get_start_op) {
	int pos = GetSpinorCouplingPos(sp_seq_no, get_start_op);
	if (pos < 0) {
		dbg->log(0, sp_seq_no, "<ERROR> can't find requested operator sequence (%ld)", this); CEException::halt();
	}

	CExpression* fo_expr = GetAt(pos);
	if (!CEUtil::equals(fo_expr->ClassName(), COpFermionField::CLASS_NAME)) {
		dbg->log(0, "<Internal Error> fermion operator expected", fo_expr); CEException::halt();
	}
	return (COpFermionField*)fo_expr;
}*/

/*COpFermionField* CMEOperatorSeq::GetFermFOAtPoint(int pointID, bool get_start_op) {
	int s = size();
	int i;
	for(i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
		if (fo->getPointID() != pointID) continue;
		if (!CEParticle::is_fermion_PK(fo->getParticleKind())) continue;
		COpFermionField* ffo = (COpFermionField*)fo;
		if (CEUtil::LEQU(ffo->isConjugated(), get_start_op)) return ffo;
	}
	return NULL;
}*/

// PK == 0 means any PK, the abs(PK) is only compared
long CMEOperatorSeq::GetNextPFPosAtPoint(int pointID, long start_search_pos, long PK_abs) {
	long s = size();
	//dbg->log(105, start_search_pos, PK_abs, "GetNextPFPosAtPoint (pos: %ld, pk: %ld)", this);
	//CMEComposeLib::dbg->log(105, start_search_pos, PK_abs, "GetNextPFPosAtPoint (pos: %ld, pk: %ld)", this);
	long i;
	for (i=start_search_pos; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		//dbg->log(105, "item: ", itm_expr);
		//CMEComposeLib::dbg->log(105, "item: ", itm_expr);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) continue;
		CEParticleFunction* pf = (CEParticleFunction*)itm_expr;
		if (pf->getPointID() != pointID) continue;
		//dbg->log(105, "point ID fitted");
		//CMEComposeLib::dbg->log(105, "point ID fitted");

		if (PK_abs == 0) return i;
		long pf_PK_abs = CLMath::abs(pf->getParticleKind());
		if ((pf_PK_abs & PK_abs) == PK_abs) return i;
	}
	return -1;
}

// PK == 0 means any PK, the abs(PK) is only compared
/*int CMEOperatorSeq::GetNextFOPosAtPoint(int pointID, int start_search_pos, long PK) {
	int s = size();
	long PK_abs = CLMath::abs(PK);
	int i;
	for(i=start_search_pos; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
		if (fo->getPointID() != pointID) continue;

		if (PK == 0) return i;
		if ((fo->getParticleKind() & PK_abs) == PK_abs) {
			if (PK > 0) return i;
			//else { // check conjugated fermion opearator
			//	if (CEParticle::is_fermion_PK(fo->getParticleKind()) {
			//		if (((COpFermionField*)fo)->isConjugated()) return i;
			//	}
		}
	}
	return -1;
}*/

// never returns NULL, throws for wrong type
CEParticleFunction* CMEOperatorSeq::GetPFAtPos(int pos) {
	if ((pos <0) || (pos >= size())) { dbg->log(0, pos, "GetPFAtPos(%ld) wrong position"); CEException::halt(); }
	CExpression* itm_expr = GetAt(pos);
	if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) {
		dbg->log(0, pos, "GetPFAtPos(%ld) particle function expected", this); CEException::halt();
	}
	return (CEParticleFunction*)itm_expr;
}

// never returns NULL, throws for wrong type
/*CEFieldOperator* CMEOperatorSeq::GetFOAtPos(int pos) {
	if ((pos <0) || (pos >= size())) { dbg->log(0, pos, "GetFOAtPos(%ld) wrong position"); CEException::halt(); }
	CExpression* itm_expr = GetAt(pos);
	if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) {
		dbg->log(0, pos, "GetFOAtPos(%ld) field operator expected", this); CEException::halt();
	}
	return (CEFieldOperator*)itm_expr;
}*/

/*int CMEOperatorSeq::GetNumFOAtPoint(int pointID, long PK) {
	int res = 0;
	int s = size();
	long PK_abs = CLMath::abs(PK);
	int i;
	for(i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) continue;
		CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
		if (fo->getPointID() != pointID) continue;

		if (PK == 0) { res++; continue; }
		if ((fo->getParticleKind() & PK_abs) == PK_abs) {
			if (PK > 0) { res++; continue; }
			else {
				// check conjugated fermion opearator
				if (CEParticle::is_fermion_PK(fo->getParticleKind()) {
					if (((COpFermionField*)fo)->isConjugated()) { res++; continue; }
				}
			}
		}
	}
	return res;
}*/

// searches for the WF in sequence, that describes specific particle
CParticleWaveFunction* CMEOperatorSeq::GetWaveFunction(CEParticle* part) {
	if (part == NULL) return NULL;
	CExpression* mom = part->get_momentum_symbol();
	if (mom == NULL) { dbg->log(0, "Error: NULL particle's momentum", part); CEException::halt(); }

	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_WAVE_FUNCTION)) continue;
		CParticleWaveFunction* wf = (CParticleWaveFunction*)itm_expr;
		if (mom->equal_or_opposite(wf->getMomentumSym())) return wf;
	}
	return NULL;
}

/*
*	order rule
*	"true" result also means equality ==> do not exchange items if order is "true"
*/
/*bool CMEOperatorSeq::CheckSpinorCouplsOrder(int a, int b) {
	if (a == b) return true;
	COpFermionField* a_op = GetFermFieldOp(a, true);
	COpFermionField* b_op = GetFermFieldOp(b, true);

	if ((!a_op->isCoupled()) && (!b_op->isCoupled())) {
		if ( a_op->getParticleKind() != b_op->getParticleKind() ) {
			return (a_op->getParticleKind() <= b_op->getParticleKind());
		}

		// compare final operators if initials are equal
		a_op = GetFermFieldOp(a, false);
		b_op = GetFermFieldOp(b, false);
	}

	if ((a_op->isCoupled()) && (b_op->isCoupled())) {
		return (a_op->getCouplingNumber() <= b_op->getCouplingNumber());
	}

	if (a_op->isCoupled()) {
		return false;
	}

	if (b_op->isCoupled()) {
		return true;
	}

	if (a_op->getParticleKind() != b_op->getParticleKind()) {
		return (a_op->getParticleKind() <= b_op->getParticleKind());
	}

	// compare bosonic operators with same point IDs
	return true;
}*/

// negative result means no coupling have been found
int CMEOperatorSeq::GetSpinorCouplingPos(int i, bool get_start_pos) {
	//dbg->log(95, i, get_start_pos, "GetSpinorCouplingPos(%ld, %ld) call");
	int res = -1;
	int couple_no = -1;
	long s = size();
	bool seq_started = false;
	long j;
	for (j=0; j<s; j++) {
		CExpression* itm_expr = GetAt(j);
		//dbg->log(95, j, "item(%ld)", itm_expr);
		if (itm_expr->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) {
			CEParticleFunction* pf = (CEParticleFunction*)itm_expr;
			long pk = pf->getLogicalPK();
			if (CEParticle::is_fermion_PK(pk)) {
				if (CEUtil::LEQU(seq_started, (pk<0))) { dbg->log(0, j, "<ERROR> invalid spinor structure (pos: %ld)", this); CEException::halt(); }
				seq_started = !seq_started;
				if (seq_started) couple_no++;
				if ( (couple_no == i) && (CEUtil::LEQU(get_start_pos, seq_started)) ) {
					res=j; break;
				}
			}
		}
		else if (CEUtil::equals(itm_expr->ClassName(), CESpinor::CLASS_NAME)) {
			dbg->log(0, "<INTERNAL ERROR> {GetSpinorCouplingPos} Unexpected plain spinor object spotted", this); CEException::halt();
			//CESpinor* sp = (CESpinor*)itm_expr;
			//if (CEUtil::LEQU(seq_started, sp->is_conjugated())) { dbg->log(0, "<ERROR> invalid spinor structure", this); CEException::halt(); }
			//seq_started = !seq_started;
			//if (seq_started) couple_no++;
			//if ( (couple_no == i) && (CEUtil::LEQU(get_start_pos, seq_started)) ) { res=j; break; }
		}
	}
	if (couple_no > i) { dbg->log(0, "<INTERNAL ERROR> searched coupling is skipped", this); CEException::halt(); }
	if ((res<0) && (seq_started)) { dbg->log(0, "<ERROR> invalid spinor structure", this); CEException::halt(); }
	dbg->log(85, i, get_start_pos, res, "GetSpinorCouplingPos(%ld, %ld) result: %ld");
	return res;
}

/*
*	positions are logical (numbers of spinor coupling)
*	Note: fermion exchange sign should be taken into account
*/
void CMEOperatorSeq::MoveSpinorCoupling(int i, int new_pos) {
	if (i == new_pos) return;
	int c_num = GetSpinorCouplingsNum();
	if ((i<0) || (i>=c_num) || (new_pos<0) || (new_pos>=c_num)) { dbg->log(0, i, "MoveSpinorCoupling(%ld, ...): invalid parameters", this); CEException::halt(); }

	int i_start_pos = GetSpinorCouplingPos(i, true); // absolute operator pos
	int i_end_pos = GetSpinorCouplingPos(i, false); // absolute operator pos
	if ((i_start_pos < 0) || (i_end_pos < 0)) { dbg->log(0, i, "MoveSpinorCoupling(%ld, ...) internal error", this); CEException::halt(); }
	int num_FO_in_cpl = GetFermOperatorsNumBetween(i_start_pos, i_end_pos+1); // absolute positions here

	// extract coupling as sequence
	CSequence sp_seq;
	long s = size();
	long i_pos;
	for (i_pos=i_start_pos; i_pos<=i_end_pos; i_pos++) {
		CExpression* itm_expr = GetAt(i_start_pos);
		sp_seq.addItem(itm_expr->clone());
		RemoveAt(i_start_pos);
	}

	// found new position for coupling
	int to_pos = -1;
	if (new_pos == (c_num-1)) {
		to_pos = GetSpinorCouplingPos(new_pos-1, false) + 1;// absolute operator pos
	} else {
		to_pos = GetSpinorCouplingPos(new_pos, true);// absolute operator pos
	}
	if (to_pos < 0) { dbg->log(0, i, "MoveSpinorCoupling(%ld, ...) internal error, can't localize target position", this); CEException::halt(); }

	// calculate number of fermions skipped (if required)
	bool bReverseSign=false;
	if (!CLMath::is_even(num_FO_in_cpl)) {
		int num_FO_skipped = GetFermOperatorsNumBetween(i_start_pos, to_pos); // positions are swapped if (i_start_pos>to_pos)
		if (!CLMath::is_even(num_FO_skipped)) { bReverseSign=true; }
	}

	// inline sequence
	bool do_append = (to_pos >= size());
	long seq_len = sp_seq.size();
	long seq_pos;
	for (seq_pos=0; seq_pos<seq_len; seq_pos++) {
		CExpression* itm_expr = sp_seq.GetAt(seq_pos);
		if (do_append) addItem(itm_expr->clone());
		else InsertAt(to_pos+seq_pos, itm_expr);
	}

	// inverse common sign if odd FO-replacements performed
	if (bReverseSign) { inverse_ferm_sign(); }
}

/*
*	Calculates operators num from position BEFORE pos_a to position BEFORE pos_b.
*/
int CMEOperatorSeq::GetFermOperatorsNumBetween(int pos_a, int pos_b) {
	long s = size();
	// note: pos_x may be == s;
	if ((pos_a < 0) || (pos_b < 0) || (pos_a > s) || (pos_b > s)) { CEException::halt(dbg, "<INTERNAL ERROR> GetFermOperatorsNumBetween: invalid parameters"); }
	if (pos_b < pos_a) { int pos_tmp=pos_a; pos_a=pos_b; pos_b=pos_tmp; } // swap parameters

	int res = 0;
	long i;
	for (i=pos_a; i<pos_b; i++) {
		CExpression* itm_expr = GetAt(i);
		if (itm_expr->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) {
			CEFieldOperator* fo = (CEFieldOperator*)itm_expr;
			if (CEParticle::is_fermion_PK(fo->getParticleKind())) res++;
		}
	}
	return res;
}

/*void CMEOperatorSeq::ArrangePointIDs() {
	int s = GetSpinorCouplingsNum();
	int i;
	for(i=0; i<s; i++) {
		COpFermionField* fo = GetFermFieldOp(i, true);

		// point ID should be (i+1)
		int fo_point = fo->getPointID();
		if (fo_point != (i+1)) {
			if (fo_point <= i) { // internal error
				dbg->log(0, "ArrangePointIDs() internal error", this); CEException::halt();
			}
			ExchangePointIDs(fo_point, i+1);
		}
	}
}*/

void CMEOperatorSeq::ExchangePointIDs(int a, int b) {
	dbg->log(55, a, b, "ExchangePointIDs(%ld, %ld)");
	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) continue;
		CEParticleFunction* pf = (CEParticleFunction*)itm_expr;
		int pt_ID = pf->getPointID();
		if (pt_ID == a) { pf->setPointID(b); }
		else if (pt_ID == b) { pf->setPointID(a); }
	}
}

/*
*	uniform_index_list: list of previously used indices.
*/
void CMEOperatorSeq::UnifyVectorIndices(CEPtrArray* uniform_index_list) {
	if (uniform_index_list == NULL) {
		dbg->log(0, "<Warning> No uniform indices list avaliable, sort is incompletable"); return;
	}
	Arrange(AL_INLINE); // arrange

	long s = size();
	int ind_no = 0;
	int max_point_num = getMaxPointID();
	//CMEComposeLib::dbg->log(105, s,max_point_num,"UnifyVectorIndices(s: %ld, max_point_num: %ld)");
	int point_no;
	for(point_no = 0; point_no <= max_point_num; point_no++) {
		long i;
		for(i=0; i<s; i++) {
			long next_pos = GetNextPFPosAtPoint(point_no, i, CEParticle::PARTICLE_VECTOR_BOSON);
			if (next_pos < 0) break;
			i = next_pos;

			CEParticleFunction* pf = GetPFAtPos(next_pos);
			CExpression* ind = NULL;
			if (pf->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) {
				CEFieldOperator* fo = (CEFieldOperator*)pf;
				if (!CEUtil::equals(fo->ClassName(), COpBosonField::CLASS_NAME)) { dbg->log(0, next_pos, "UnifyVectorIndices() error: bosonic operator is expected at pos %ld", this); CEException::halt(); }
				if (!((COpBosonField*)fo)->isVectorBoson()) { dbg->log(0, next_pos, "UnifyVectorIndices() error: VECTOR bosonic operator is expected at pos %ld", this); CEException::halt(); }
				ind = ((COpBosonField*)fo)->getVectorIndex()->index();
			} else if (pf->getProperty(CEProperties::PROP_MASK_IS_WAVE_FUNCTION)) {
				CParticleWaveFunction* wf = (CParticleWaveFunction*)pf;
				if (!CEUtil::equals(wf->ClassName(), CWaveFuncBoson::CLASS_NAME)) { dbg->log(0, next_pos, "UnifyVectorIndices() error: bosonic wave function is expected at pos %ld", this); CEException::halt(); }
				if (!((CWaveFuncBoson*)wf)->isVectorBoson()) { dbg->log(0, next_pos, "UnifyVectorIndices() error: VECTOR bosonic wave function is expected at pos %ld", this); CEException::halt(); }
				ind = ((CWaveFuncBoson*)wf)->getVectorIndex()->index();
			}
			if (ind == NULL) continue; // nothing to unify

			// set index to unified
			for(; uniform_index_list->size() <= ind_no ;) { uniform_index_list->add(CESymbol::getUnique()); }
			CExpression* ind_alt = (CExpression*)uniform_index_list->GetAt(ind_no);
			ind_no++;

			if (ind->equals(ind_alt)) continue; // it may occur if sorting is repeated
			if (CRContainCheck::search(this, ind_alt)) { // involve temporary index for "ind_alt"
				CExpression* ind_unique = CESymbol::getUnique();
				CEReplasement r_tmp(ind_alt, ind_unique); delete ind_unique;
				CExpression* r_res = r_tmp.ApplyTo(this, false); if (r_res != this) { CEException::halt("Internal error in UnifyVectorIndices()"); }
			}
			CEReplasement r(ind, ind_alt);
			CExpression* r_res = r.ApplyTo(this, false); if (r_res != this) { CEException::halt("Internal error in UnifyVectorIndices() #2"); }
		}
	}
}

int CMEOperatorSeq::getMaxPointID() {
	int res = -1;
	long s = size();
	long i;
	for (i=0; i<s; i++) {
		CExpression* itm_expr = GetAt(i);
		if (!itm_expr->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) continue;
		int fo_point = ((CEParticleFunction*)itm_expr)->getPointID();
		if (fo_point > res) res = fo_point;
	}
	return res;
}

/*
*	Transforms crossed diagrams (sequence order) to normal ones.
*	Used for coupled SEQ.
*	Takes care of the sign for every crossing transformation.
*	Returns true is SEQ is changed.
*	TODO: may be greatly fastered (if avoid extra WF searching)
*/
bool CMEOperatorSeq::NormalizeCrossings(CKinematics* kin) {
	if (kin == NULL) { dbg->log(0, "NormalizeCrossings(NULL)"); return false; }
	int in_num = kin->num_init_particles();
	int f_num = kin->num_final_particles();
	if ((in_num < 1) || (f_num < 1)) { dbg->log(0, "NormalizeCrossings: Error in kinematics", kin); return false; }
	bool CR_found = false;

	// initial particles
	CParticleWaveFunction* wf_L = NULL;
	CParticleWaveFunction* wf_R = NULL;
	int part_N = f_num;
	int part_no1 = 0;
	int part_no2 = part_N-1;
	long cnt_chk;
	for(cnt_chk=0; cnt_chk<1000; cnt_chk++) { // counter to escape ethrnal loops in error cases
		if ((part_no1>=part_N)||(part_no2<0)||(part_no1>part_no2)) break; // no more particle combinations
		if (part_no1 == part_no2)  {
			wf_L = GetWaveFunction(kin->pfin(part_no1)); wf_R = NULL;
		} else {
			wf_L = GetWaveFunction(kin->pfin(part_no1)); wf_R = GetWaveFunction(kin->pfin(part_no2));
		}
		if (wf_L == NULL) { dbg->log(0, "<WARNING> NormalizeCrossings: NULL left route"); break; }

		int sub_res = NormCrossings(true, wf_L, wf_R, kin);
		if (sub_res < 0) break;
		else if (sub_res > 0) {
			CR_found = true; //part_no1=0; part_no2=part_N-1; // restore route particles
		} else { // 0 (undefined result)
			part_no2--; if (part_no2==part_no1) { part_no1++; part_no2=part_N-1; }
		}
	}

	// final particles
	part_N=in_num;
	part_no1 = 0;
	part_no2 = part_N-1;
	for(cnt_chk=0; cnt_chk<1000; cnt_chk++) { // counter to escape ethrnal loops in error cases
		if ((part_no1>=part_N)||(part_no2<0)||(part_no1>part_no2)) break; // no more particle combinations
		if (part_no1 == part_no2)  {
			wf_L = GetWaveFunction(kin->pin(part_no1)); wf_R = NULL;
		} else {
			wf_L = GetWaveFunction(kin->pin(part_no1)); wf_R = GetWaveFunction(kin->pin(part_no2));
		}
		if (wf_L == NULL) { dbg->log(0, "<WARNING> NormalizeCrossings: NULL left route"); break; }

		int sub_res = NormCrossings(false, wf_L, wf_R, kin);
		if (sub_res < 0) break;
		else if (sub_res > 0) {
			CR_found = true; //part_no1=0; part_no2=part_N-1; // restore route particles
		} else { // 0 (undefined result)
			part_no2--; if (part_no2==part_no1) { part_no1++; part_no2=part_N-1; }
		}
	}

	return CR_found;
}

/*
*	connection should be closer to pf_route_close, and far from pf_route_far
*	result: 1 -- crossing found (routes also may be crossed), -1 -- no crossing is needed, 0 -- undefined case
*	in undefined case the route operators should be changed
*/
int CMEOperatorSeq::NormCrossings(bool in_inits,
								   CParticleWaveFunction* pf_route_close,
								   CParticleWaveFunction* pf_route_far,
								   CKinematics* kin) {
	
	int s = (in_inits)?(kin->num_init_particles()):(kin->num_final_particles());
	if (s < 2) return -1;
	bool undef_res = false;
	long PK_cl = ((pf_route_close!=NULL)?(pf_route_close->getLogicalPK()):0);
	long PK_far = ((pf_route_far!=NULL)?(pf_route_far->getLogicalPK()):0);

	int i;
	for (i=0; i<(s-1); i++) {
		CEParticle* pA = kin->get_particle(i, in_inits); if (pA == NULL) { dbg->log(0, "NormalizeCrossings: Error in kinematics", kin); return -1; }
		long PK_A = pA->get_PK();
		CParticleWaveFunction* wf_A = NULL;
		int route_A_cl = -1;
		int route_A_far = -1;

		int j;
		for (j=i+1; j<s; j++) {
			CEParticle* pB = kin->get_particle(j, in_inits); if (pB == NULL) { dbg->log(0, "NormalizeCrossings: Error in kinematics", kin); return -1; }
			long PK_B = pB->get_PK();
			if (PK_A != PK_B) continue; // consider similar PK only

			if (wf_A == NULL) { // fill A-data if required
				wf_A = GetWaveFunction(pA);
				if (wf_A == NULL) return -1; // e.g. uncoupled sequence
				route_A_cl = CalcLinksNum(pf_route_close, wf_A);
				route_A_far = (pf_route_far != NULL)?(CalcLinksNum(pf_route_far, wf_A)):0;
				if (route_A_cl <= 0) { return -1; } // unbinded diagram
			}
			
			// fill B-data
			CParticleWaveFunction* wf_B = GetWaveFunction(pB);
			if (wf_B == NULL) continue; // e.g. uncoupled sequence
			int route_B_cl = CalcLinksNum(pf_route_close, wf_B);
			int route_B_far = (pf_route_far != NULL)?(CalcLinksNum(pf_route_far, wf_B)):0;
			if (route_B_cl <= 0) { return -1; } // unexpected unbinded diagram

			//printf("route_A_cl: %ld, route_B_cl:%ld, route_A_far:%ld, route_B_far:%ld\n", route_A_cl, route_B_cl, route_A_far, route_B_far);
			bool exch_CR=false;
			bool exch_Routes=false;
			int route_cmp_val = (route_A_cl-route_B_cl) - (route_A_far-route_B_far);
			if (route_cmp_val > 0) {
				exch_CR=true;
			}

			if ((in_inits) && (route_A_cl>route_B_far) && (route_A_cl<route_B_cl) && (route_A_far>route_B_far) && (PK_cl==PK_far) && (PK_cl != 0)) {
				exch_CR=true; exch_Routes=true;
			}

			if (exch_Routes) {
				ExchangeCrossingParticles(PK_cl, pf_route_close, pf_route_far); // exchange ?????
			}

			if (exch_CR) { // crossing found
				//ERROR occurs here if PK_A != wf_A->LogicalPK() == wf_B->LogicalPK()
				//TODO: fix it correctly!
				//Old: ExchangeCrossingParticles(PK_A, wf_A, wf_B); // exchange wf_A & wf_B
				ExchangeCrossingParticles(wf_A->getLogicalPK(), wf_A, wf_B); // exchange wf_A & wf_B
				return 1;
			} else if (route_cmp_val == 0) undef_res = true;
		}
	}
	return (undef_res)?0:(-1);
}

/*
*	returns number of vertices between functions (0 for equals, -1 for unbinded operators, INT_MIN for error)
*/
int CMEOperatorSeq::CalcLinksNum(CEParticleFunction* f_a, CEParticleFunction* f_b) {
	if((f_a == NULL) || (f_b == NULL)) return INT_MIN;
	if(f_a == f_b) return 0;

	// calc bond length
	int s = getMaxPointID()+1;
	int* _vals = (int*)CMemoryManager::mmalloc(s*sizeof(int));
	int i;
	for(i=0; i<s; i++) { _vals[i]=-1; } // mark all as unrouted

	// mark points, connected to f_a; try all functions, connected to points; loop
	CalcLinksNumProc(_vals, 0, f_a, -1);

	int b_pID = f_b->getPointID();
	int res = _vals[b_pID];

	//////////////////////////////////////////////////////////////////
	//dbg->log(0, b_pID, res, "CalcLinksNum(%ld) res: %ld");
	//for (i=0; i<s; i++) printf("%ld-", _vals[i]);

	CMemoryManager::mfree(_vals);

	if (res < 0) { dbg->log(0, "CalcLinksNum() WARNING: unbinded diagram"); }
	return res;
}

/*
*	increases step
*	try points, connected to f_a
*	'last_point' point is unaffected
*/
void CMEOperatorSeq::CalcLinksNumProc(int* _vals, int step_no, CEParticleFunction* f, int last_point) {
	step_no++;

	int pID = f->getPointID();
	if (pID != last_point) {
		CalcLinksNumProc(_vals, step_no, pID, f);
	}
	
	// the same for other linked point
	if (f->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR)) {
		CEFieldOperator* fo = (CEFieldOperator*)f;
		if (fo->isCoupled()) {
			CEFieldOperator* fo_cpl = GetCoupledFO(fo);
			pID = fo_cpl->getPointID();
			if (pID != last_point) CalcLinksNumProc(_vals, step_no, pID, fo_cpl);
		}
	}
}

/*
*	mark point
*	try all functions, connected to points (ecxept of f_last)
*/
void CMEOperatorSeq::CalcLinksNumProc(int* _vals, int step_no, int pID, CEParticleFunction* f_last) {
	int old_val = _vals[pID];
	if ((old_val >= 0) && (old_val <= step_no)) return;
	_vals[pID] = step_no;
	
	long s = size();
	long i;
	for(i=0; i<s; i++) {
		long next_pos = GetNextPFPosAtPoint(pID, i, 0);
		if (next_pos < 0) break;
		i = next_pos;
		
		CEParticleFunction* pf = GetPFAtPos(next_pos);
		if (pf != f_last) CalcLinksNumProc(_vals, step_no, pf, pID);
	}
}

void CMEOperatorSeq::ExchangeCrossingParticles(long pk, CParticleWaveFunction* wf_A, CParticleWaveFunction* wf_B) {
	//dbg->log(2, "ExchangeCrossingParticles", this, wf_A, wf_B);

	//checks
	if ((wf_A==NULL) || (wf_B==NULL)) { CEException::halt("CMEOperatorSeq::ExchangeCrossingParticles(NULL)"); }
	if ((pk != wf_A->getLogicalPK()) || (pk != wf_B->getLogicalPK())) {
		dbg->log(0,"Internal error in CMEOperatorSeq::ExchangeCrossingParticles: wrong PK info");
		dbg->log(0,pk,wf_A->getLogicalPK(),wf_B->getLogicalPK(),"pk: %x, wf_A->getLogicalPK(): %x, wf_B->getLogicalPK(): %x",this);
		CEException::halt();
	}

	// exchange particle identifiers
	CExpression* A_mom = wf_A->getMomentumSym();
	CExpression* B_mom_cpy = wf_B->getMomentumSym()->clone();
	wf_B->setMomentumSym(A_mom);
	wf_A->setMomentumSym(B_mom_cpy); delete B_mom_cpy;

	// flip diagram sign if fermions were crossed
	if (CEParticle::is_fermion_PK(pk)) inverse_ferm_sign();

	dbg->log(35, pk, "Particles exchanged (pk: %ld)");
}

// additional sign factor (when diagram is retrieved)
bool CMEOperatorSeq::calcFOCouplingsSeqSign() {
	CMEOperatorSeq* cpy = (CMEOperatorSeq*)clone();
	cpy->ToStandardFOCouplingsOrder();
	bool res = CEUtil::LEQU(get_ferm_sign(), cpy->get_ferm_sign());
	delete cpy;
	return res;
}

/*
*	Standard Order is:
*	FF(1) F(1)F(2) F(2)F
*	F(3)F(1) F(1)F(2) F(2)F(3)
*	Note: can be fastered by precalculation of couplings data
*/
void CMEOperatorSeq::ToStandardFOCouplingsOrder() {
	dbg->log(65, "ToStandardFOCouplingsOrder()", this);
	if (have_uncoupled_field_op()) { CEException::halt(dbg, "<ERROR in ToStandardFOCouplingsOrder>: not all the FOs are coupled ==> can't determine standard order."); }

	int s = GetSpinorCouplingsNum();
	int i;
	for(i=0; i<(s-1); i++) {
		// coupling I
		int pos_L = GetSpinorCouplingPos(i, true);
		int pos_R = GetSpinorCouplingPos(i, false);
		if ((pos_L < 0) || (pos_R < 0)) CEException::halt(dbg, "<Internal ERROR> {ToStandardFOCouplingsOrder}");
		CEParticleFunction* pf_L = GetPFAtPos(pos_L);
		CEParticleFunction* pf_R = GetPFAtPos(pos_R);
		bool is_FO_L = pf_L->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR);
		bool is_FO_R = pf_R->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR);
		if ((!is_FO_L) && (!is_FO_R)) continue;
		//dbg->log(85, i, "Couple %ld; L", pf_L);
		//dbg->log(85, "R:", pf_R);
		//dbg->log(85, is_FO_L, is_FO_R, "is_FO_L: %ld, is_FO_R: %ld");

		int move_pos = -1;
		bool seq_OK = false; // AB*BC*CD*DA -- sequence
		int last_CN = 0;
		int j;
		for(j=s-1; j>i; j--) {
			// coupling J
			int pos_JL = GetSpinorCouplingPos(j, true);
			int pos_JR = GetSpinorCouplingPos(j, false);
			if ((pos_JL < 0) || (pos_JR < 0)) CEException::halt(dbg, "<Internal ERROR> {ToStandardFOCouplingsOrder}");
			CEParticleFunction* pf_JL = GetPFAtPos(pos_JL);
			CEParticleFunction* pf_JR = GetPFAtPos(pos_JR);
			bool is_FO_JL = pf_JL->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR);
			bool is_FO_JR = pf_JR->getProperty(CEProperties::PROP_MASK_IS_FIELD_OPERATOR);
			//dbg->log(85, j, is_FO_JL, is_FO_JR, "j: %ld, is_FO_JL: %ld, is_FO_JR: %ld");

			// check out seq_OK
			if (seq_OK) {
				if ((!is_FO_JR) || (!is_FO_JL)) {
					seq_OK = false;
					if (move_pos < 0) CEException::halt(dbg, "Int Error: (move_pos < 0) while seq_OK");
					break;
				} else if (last_CN != ((CEFieldOperator*)pf_JR)->getCouplingNumber()) {
					seq_OK = false;
					if (move_pos < 0) CEException::halt(dbg, "Int Error: (move_pos < 0) while seq_OK");
					break;
				} else {
					last_CN = ((CEFieldOperator*)pf_JL)->getCouplingNumber();
				}
			}

			//if ((!is_FO_JL) && (!is_FO_JR)) continue;
			if (move_pos < 0) { // find out move pos
				if ((is_FO_L) && (is_FO_JR)) {
					if (((CEFieldOperator*)pf_L)->getCouplingNumber() == ((CEFieldOperator*)pf_JR)->getCouplingNumber()) {
						move_pos = j;
						if (is_FO_JL) { seq_OK = true; last_CN = ((CEFieldOperator*)pf_JL)->getCouplingNumber(); }
					}
				} else if ((is_FO_R) && (is_FO_JL)) {
					if (((CEFieldOperator*)pf_R)->getCouplingNumber() == ((CEFieldOperator*)pf_JL)->getCouplingNumber()) {
						move_pos = j-1;
					}
				}
			}

			if ((!seq_OK) && (move_pos>=0)) break;
			//if (((!is_FO_L) || (L_move_pos >= 0)) && ((!is_FO_R) || (R_move_pos >= 0))) break;
		}

		// last check for seq_OK
		if (seq_OK) {
			if ((!is_FO_R) || (!is_FO_L)) {
				seq_OK = false;
			} else if (last_CN != ((CEFieldOperator*)pf_R)->getCouplingNumber()) {
				seq_OK = false;
			}
		}

		//dbg->log(85, move_pos, seq_OK, "move pos: %ld, seq_OK: %ld");
		// seq_OK is the check for AB*BC*CA ==> BC*CA*AB (recursive) move
		if ((seq_OK) || (move_pos < 0)) continue;
		//int move_pos = (L_move_pos >= 0)?L_move_pos:R_move_pos;
		if (move_pos == i) continue;
		if (move_pos < i) CEException::halt(dbg, "INTERNAL ERROR, wrong FO-couple move position");
		
		MoveSpinorCoupling(i, move_pos);
		dbg->log(65, i, move_pos, "%ld --> %ld moved", this);
		//i--;
		i = -1; // reconsider all the previous for XA*AB...BX ==> XA...AB*BX case
	}
	dbg->log(45, "ToStandardFOCouplingsOrder() finished", this);
}
