////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// MESquareLib.cpp: implementation of the CMESquareLib class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CMESquareLib::dbg = new CDebug("CMESquareLib");

const DWORD CMESquareLib::SQRME_DEFAULT						= 0x00;
const DWORD CMESquareLib::SQRME_SIMPLE						= 0x00;
const DWORD CMESquareLib::SQRME_SPLIT_SUM					= 0x01;
//const DWORD CMESquareLib::SQRME_EXPAND_ME					= 0x02;
const DWORD CMESquareLib::SQRME_EXPAND_AND_SPLIT			= 0x03;
const DWORD CMESquareLib::SQRME_PRESIMPLIFY_ME				= 0x04;
const DWORD CMESquareLib::SQRME_CACHE_DENOMINATORS			= 0x08;
const DWORD CMESquareLib::SQRME_CMS_BEAMS					= 0x10;
const DWORD CMESquareLib::SQRME_SKIP_VI_CALC				= 0x20;
const DWORD CMESquareLib::SQRME_NO_CROSSINGS_FOR_ME1		= 0x40;
const DWORD CMESquareLib::SQRME_NO_CROSSINGS_FOR_ME2		= 0x80;
const DWORD CMESquareLib::SQRME_ALLOW_MANDELSTAMS			= 0x000100;
const DWORD CMESquareLib::SQRME_BERENDS_HELICITIES			= 0x000200;
const DWORD CMESquareLib::SQRME_BERENDS_HELICITIES_PRECISE	= 0x000400;
const DWORD CMESquareLib::SQRME_IR_LIMIT					= 0x000800;
const DWORD CMESquareLib::SQRME_EARLY_PHOTON_HELICITIES		= 0x001000;
const DWORD CMESquareLib::SQRME_LATE_PHOTON_HELICITIES		= 0x002000;
const DWORD CMESquareLib::SQRME_FORCE_4DIM					= 0x004000;
const DWORD CMESquareLib::SQRME_4DIM_BOSON_POLARIZATION		= 0x008000; // unused
const DWORD CMESquareLib::SQRME_NO_POSTARRANGE				= 0x010000;
const DWORD CMESquareLib::SQRME_NO_SUB_ARRANGE				= 0x020000;
const DWORD CMESquareLib::SQRME_SAVE_MEMORY					= 0x040000;

/*
*	Photon polarization involving notes.
*
*	Photon polarization (if declared) is involved after the ME*cjgME is calculated.
*	The dimension is set before the polarization involved.
*	It is considered as most proper way (the abstract density matrices may be involved in this way).
*
*	The other ways to involve photon helicity:
*	SQRME_BERENDS_HELICITIES:
*		tries to use berends gamma*e->(..) formula.
*		if ME have non-"gamma*e" photon polarization, option is ignored
*		'SQRME_EARLY_PHOTON_HELICITIES' is set automatically for non-"gamma*e" cases
*	SQRME_BERENDS_HELICITIES_PRECISE:
*		tries to use precise gamma*e->(..) formula.
*		the SQRME_BERENDS_HELICITIES option is automatically set (for internal use).
*		if ME have non-"gamma*e" photon polarization, option is ignored
*		'SQRME_EARLY_PHOTON_HELICITIES' is set automatically for non-"gamma*e" cases
*	SQRME_EARLY_PHOTON_HELICITIES:
*		involve photon polarization vector at the matrix element reduction stage.
*		I.e. before squarion procedure, while the dimension is N-dim (for loops).
*		(SQRME_BERENDS_HELICITIES(_X) options works as usual: first try to involve gamma*e form).
*	SQRME_LATE_PHOTON_HELICITIES:
*		Photon helicities (if declared) will be involved into result expression.
*/

//////////////////////////////////////////////////////////////////////
// Static Entry
// Matrix Elements are not affected
//////////////////////////////////////////////////////////////////////

CExpression* CMESquareLib::SquareME(CMatrixElement* me_a, CMatrixElement* me_b, DWORD settings) {
	if (me_a == NULL) return NULL;// throw?

	dbg->log(10, "SquareME");
	dbg->log(20, "SquareME (a=)", me_a);
	dbg->log(20, "SquareME (b=)", me_b);

	///////////////////////////////////////////////// <settings arrangement>
	// check out settings
	if (settings & SQRME_BERENDS_HELICITIES_PRECISE) {
		if ((settings & SQRME_BERENDS_HELICITIES) == 0) settings |= SQRME_BERENDS_HELICITIES;
	}
	if (settings & SQRME_BERENDS_HELICITIES) {
		if ((settings & SQRME_EARLY_PHOTON_HELICITIES) == 0) settings |= SQRME_EARLY_PHOTON_HELICITIES;
	}
	if ( ((settings & SQRME_BERENDS_HELICITIES) == 0) &&
		((settings & SQRME_EARLY_PHOTON_HELICITIES) == 0) ) {
		//settings |= SQRME_LATE_PHOTON_HELICITIES;
	}
	///////////////////////////////////////////////// </settings arrangement>

	// store settings
	bool N_dim = CETensor::isNDimensionSpace();	//may be forsed to N-dim during calculation
	int sp_setts = CFSp::m_Settings;
	CFSp::m_Settings |= CFSp::SP_CHECK_SUCCESS;

	CMESquareLib* proc = new CMESquareLib(settings);
	CExpression* res = proc->run_task(me_a, me_b);
	delete proc;

	// restore settings
	CETensor::setNDimensionSpace(N_dim);
	CFSp::m_Settings = sp_setts;

	dbg->log(15, "SquareME() finished", res);
	return res;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMESquareLib::CMESquareLib(DWORD settings) : CELibrary()
{
	m_settings = settings;// SQRME_DEFAULT
	m_me_a = NULL;
	m_me_b = NULL;
	m_ME_SQUARE = false;
	m_NDimCalc = false;
	m_bRestoreProtonStrFunc = false;
	m_loops_no = -1;

	//m_a = NULL;
	//m_b = NULL;
	//m_inds_a = NULL;
	//m_inds_b = NULL;

	m_kin = CEProcessor::kinematics();
	m_vector_parts_DMs = new CEPtrArray(); m_vector_PostProcDMs = new CEPtrArray();
	m_mass_rule = NULL;
	m_mand_vars_rule = NULL;

	// kinematic rules init (ones per calculation for time saving)
	m_kin_simp_rules = (m_kin != NULL)?(m_kin->getReplacements(CKinematics::REPL_COUPLINGS_SIMPLIFY)):NULL;
	bool CMS_beams = ((m_settings & SQRME_CMS_BEAMS) != 0);
	m_basic_coupling_vals = (m_kin != NULL)?(m_kin->getBasicCouplingsValues(CMS_beams)):NULL;

	if ((m_settings & SQRME_ALLOW_MANDELSTAMS) && (m_kin != NULL)) {
		m_mand_vars_rule = m_kin->getReplacements(CKinematics::REPL_MANDELSTAM_VARIABLES);
		if ((m_mand_vars_rule != NULL) && (m_mand_vars_rule->size() == 0)) {
			delete m_mand_vars_rule; m_mand_vars_rule = NULL;
		}
	}

	dbg->log(25, "kin_simp_rules", m_kin_simp_rules);
	dbg->log(25, "mand_vars_rule", m_mand_vars_rule);

	m_denominators_cache = NULL;
}

CMESquareLib::~CMESquareLib()
{
	if ((m_me_a != m_me_b) && (m_me_b != NULL)) { delete m_me_b; m_me_b = NULL; }
	if (m_me_a != NULL) { delete m_me_a; m_me_a = NULL; }

	delete m_vector_parts_DMs; delete m_vector_PostProcDMs;

	if (m_mass_rule != NULL) delete m_mass_rule;
	if (m_kin_simp_rules != NULL) delete m_kin_simp_rules;
	if (m_mand_vars_rule != NULL) delete m_mand_vars_rule;
	if (m_basic_coupling_vals != NULL) delete m_basic_coupling_vals;

	if (m_denominators_cache != NULL) delete m_denominators_cache;
}

CExpression* CMESquareLib::run_task(CMatrixElement* me_a, CMatrixElement* me_b) {
	if ((me_a == NULL) || (m_kin == NULL)) return NULL;

	////////////////////////////////////////////////////////////	<init>
	int i;

	// clone input parameters
	m_me_a = me_a->clone();
	if ((me_b == NULL) || (me_b == me_a)) { m_me_b = m_me_a; m_ME_SQUARE = true; }
	else { m_me_b = me_b->clone(); m_ME_SQUARE = false; }
	m_loops_no = m_me_a->get_virt_momenta_list()->size();
	if (m_me_b != m_me_a) { m_loops_no += m_me_b->get_virt_momenta_list()->size(); }
	else { m_loops_no*=2; }

	// dimension definition
	m_NDimCalc = (m_loops_no > 0); // set N-dimension space for loop consideration
	if (m_settings & CMESquareLib::SQRME_FORCE_4DIM) m_NDimCalc = false;
	CETensor::setNDimensionSpace(m_NDimCalc);

	// calc up to finite mass degree (init)
	m_mass_rule = CLHEPBasics::get_mass_order_rule();

	// check if proton structure functions should be reconstrucred
	for(i=0;i<m_kin->num_init_particles();i++){
		if(m_kin->pin(i)->check_PK(CEParticle::PARTICLE_PROTON,true)) { m_bRestoreProtonStrFunc=true; m_arrProtonID.Add(CKinematics::PID(i,true)); }
	}
	for(i=0;i<m_kin->num_final_particles();i++){
		if(m_kin->pfin(i)->check_PK(CEParticle::PARTICLE_PROTON,true)) { m_bRestoreProtonStrFunc=true; m_arrProtonID.Add(CKinematics::PID(i,false)); }
	}
	for(i=0;i<m_arrProtonID.GetSize();i++) m_SFTmpIndices.addItem(CESymbol::getUnique()); // temporary indices
	if(m_bRestoreProtonStrFunc) {
		dbg->log(49,"m_bRestoreProtonStrFunc=true");
	}
	////////////////////////////////////////////////////////////	</init>

	/////////////////////////////////////////////////// <Matrix Elements Preprocessing>
	// arrange ME structure (inline products)
	if (!CMELib::ToNormalForm(m_me_a)) { dbg->log(0, "CMESquareLib::run_task() WARNING: can't normalize ME:", m_me_a); }
	if (m_me_b != m_me_a) { if (!CMELib::ToNormalForm(m_me_b)) { dbg->log(0, "CMESquareLib::run_task() WARNING: can't normalize ME:", m_me_b); } }

	if (m_settings & SQRME_IR_LIMIT) {
		CLHEPBasics::GetIRLimit(m_me_a); dbg->log(20, "IR limit of ME#1:", m_me_a);
		if (m_me_b != m_me_a) { CLHEPBasics::GetIRLimit(m_me_b); dbg->log(20, "IR limit of ME#2:", m_me_b); }
	}

	// if ME are equal but used separately (as different objects, e.g. Square(me, me) call).
	if (m_me_b != m_me_a) {
		if (m_me_a->equals(m_me_b)) {
			m_me_b->ChangeVolatileIndices(false);
			dbg->log(20, "Indices changed at ME#2:", m_me_b);
		}
	}

	// add all the crossings to matrix elements.
	AppendCrossings();
	/////////////////////////////////////////////////// </Matrix Elements Preprocessing>

	dbg->log(20, "Start library processing.");
	dbg->logperiod(10);
	if (m_settings & SQRME_CACHE_DENOMINATORS) init_denominators_cache();

	/////////////////////////////////////////////	1. Simplify/arrange matrix elements
	bool res_OK = PrepareMatrixElementsExprs();
	if (!res_OK) { dbg->log(0, "Error coupling ME (PrepareMatrixElementsExprs fails)."); throw(new CEException(CEException::INT_ERROR)); }

	// denominators simplification
	if (m_settings & SQRME_CACHE_DENOMINATORS) simplify_cached_denominators();
	// change vector indices and create density matrices
	res_OK = prepareVectorPolarizationData();
	if (!res_OK) { dbg->log(0, "Error coupling ME (VectorPolarizationData fails)."); throw(new CEException(CEException::INT_ERROR)); }

	dbg->logperiod(10, "Matrix elements simplified.");
	dbg->log(20, "ME #1:", m_me_a);
	dbg->log(20, "ME #2:", m_me_b);
	/////////////////////////////////////////////	2. Square
	CExpression* res = NULL;
	if (m_settings & SQRME_SPLIT_SUM) {
		res = coupleMESplitting();
	} else {
		res = coupleExprs(m_me_a->get_expr_val(), m_me_b->get_expr_val());
	}

	if ((m_settings & SQRME_EARLY_PHOTON_HELICITIES) == 0) {
		// 4-dim is needed // TODO: check if N-dim affects result here
		if (m_NDimCalc) CETensor::setNDimensionSpace(false);
		// involve post-proc density matrices (to be coupled at 4-dim)
		res = InvolvePostProcDMs(res);
		if (m_NDimCalc) CETensor::setNDimensionSpace(m_NDimCalc);
	}

	// reduce expression before restoring denominators
	// (already reduced)
	//res = CExpression::reduce(res);

	dbg->logperiod(55, "Matrix element squared (no denominators).");
	if (m_settings & SQRME_CACHE_DENOMINATORS) res = restore_denominators(res);

	/////////////////////////////////////////////	3. Calculate virtual integrals
	if ((m_settings & SQRME_SKIP_VI_CALC) == 0) {
		dbg->log(55, "Squaring result (before VI restoring):", res);
		res = virt_integrate(res);
	}

	//dbg->log(15, "CMESquareLib processing result:", res); //same to final ouptut output in CMESquareLib::SquareME(){}
	return res;
}

/*
*	Append crossings parts to matrix elements source.
*/
void CMESquareLib::AppendCrossings() {
	// extra check to skip this procedure (may be omitted).
	if ( (m_settings & SQRME_NO_CROSSINGS_FOR_ME1) && 
		((m_settings & SQRME_NO_CROSSINGS_FOR_ME2) || (m_me_b == m_me_a)) ) {
		dbg->log(21, "AppendCrossings() skipped.");
		return;
	}

	dbg->log(20, "AppendCrossings().");
	CEPtrArray* cross_infos = m_kin->getCrossingsTransformations();
	dbg->log(21, "AppendCrossings() crossing info retrieved:", cross_infos);
	if (cross_infos == NULL) return;
	int s = cross_infos->size();
	if (s == 0) { delete cross_infos; return; }

	if ((m_settings & SQRME_NO_CROSSINGS_FOR_ME1) == 0) {
		m_me_a->AppendCrossings(cross_infos);
		dbg->log(21, "AppendCrossings() for ME#1 completed:", m_me_a);
	}
	if (m_me_b != m_me_a) { // m_ME_SQUARE
		// the same for ME#2
		if ((m_settings & SQRME_NO_CROSSINGS_FOR_ME2) == 0) {
			m_me_b->AppendCrossings(cross_infos);
			dbg->log(21, "AppendCrossings() for ME#2 completed", m_me_b);
		}
	}

	delete cross_infos;
}

/*
*	Copy ME expressions to "m_a" and "m_b".
*	Ivolve Berends-like helicity vectors.
*	Simplification of matrix elements.
*	Use possible equivelence of matrix elements for maximal simplification performance.
*/
bool CMESquareLib::PrepareMatrixElementsExprs() {
	// ME objects should be still equal here in squaring case
	if ((m_ME_SQUARE) && (m_me_b != m_me_a)) { dbg->log(0, "Error: different MEs in squaring procedure."); throw(new CEException(CEException::INT_ERROR)); }
	dbg->log(20, m_ME_SQUARE, "PrepareMatrixElementsExprs(), (ME are equal == %d)");

	CExpression* me_expr = m_me_a->get_expr_val()->clone();
	me_expr = PrepareME(me_expr,
		m_me_a->get_external_indices_list(),
		m_me_a->get_virt_momenta_list()->size());
	m_me_a->set_expr_val(me_expr); delete me_expr;

	if (!m_ME_SQUARE) {
		me_expr = m_me_b->get_expr_val()->clone();
		me_expr = PrepareME(me_expr,
			m_me_b->get_external_indices_list(),
			m_me_b->get_virt_momenta_list()->size());
		m_me_b->set_expr_val(me_expr); delete me_expr;
	} else {
		if (m_me_b != m_me_a) { dbg->log(0, "Warning: unexpected cloned ME2"); delete m_me_b; }
		m_me_b = m_me_a->clone(); // ME will now be cloned.

		// change volatile indices in m_me_b.
		m_me_b->ChangeVolatileIndices(false);
		//m_me_b->ChangeVolatileIndices(true);
	}
	return true;
}

/*
*	Simplifies matrix elements.
*	Ivolve Berends-like helicity vectors.
*	Multiplies ME to the helicity vectors
*	and extracts vector index from further proceedings.
*/
CExpression* CMESquareLib::PrepareME(CExpression* me_expr, CEPtrArray* inds, int loops_no) {
	dbg->log(21, "PrepareME() :", me_expr);
	bool to_be_simplified = true;

	// Involve BERENDS helicities in separated cycle
	// because no simplification may be performed
	// until all the helicities are inserted.
	if (m_settings & SQRME_BERENDS_HELICITIES) {
		bool calc_berends_precise = ((m_settings & SQRME_BERENDS_HELICITIES_PRECISE) != 0);
		int i;
		for(i=0; i<inds->size(); i++) {
			CExprContainer* cart = (CExprContainer*)inds->GetAt(i);
			int part_ID = cart->get_info();
			CEParticle* part = m_kin->get_particle_by_ID(part_ID);
			if ((part == NULL) || (!part->check_PK(CEParticle::PARTICLE_PHOTON, true))) continue;

			CExpression* h_ber_val = ((CPPhoton*)part)->getBerendsHelicity(calc_berends_precise);
			// if (h_ber_val == NULL) ignore SQRME_BERENDS_HELICITIES setting
			// and try to involve usual helicity vector
			if (h_ber_val == NULL) continue;
			dbg->log(21, "PrepareME() gamma-coupled (Berends form) helicity vector found: ", h_ber_val);

			// extract denominator from helicity expression
			h_ber_val = involve_denominators_cache(h_ber_val, true);

			// replace GAMMA\\_{index} with h_ber_val.
			CExpression* G_ind = new CEGamma(cart->get_expr(), true);
			CEReplasement* r = new CEReplasement(G_ind, h_ber_val); delete G_ind; delete h_ber_val;
			CExpression* me_copy = r->ApplyTo(me_expr->clone()); delete r;

			// search if current vector index is still contained in ME expression.
			if (CRContainCheck::search(me_copy, cart->get_expr())) {
				// The BERENDS helicities can't be involved in this matrix element.
				// Because vactor index does not occur in GAMMA\\_{index}-objects only.
				// It may (should) occur for gamma-WW and ZWW vertices
				dbg->log(0, part_ID, "Note: The BERENDS-like helicities can't be involved for particle no. %ld.", me_expr);
				dbg->log(0, "Helicitiy value will be involved in the vector form.");
				//dbg->log(0, "PrepareME() WARNING: Index failed ", cart->get_expr());
				delete me_copy;
				continue;
			}
			delete me_expr;
			me_expr = CExpression::arrge(me_copy);
			
			// do not involve polarization objects
			// (e.g. helicity vectors & density martices) for this index
			inds->RemoveAt(i, true); i--;
			dbg->log(41, "PrepareME() helicity vector involved (res ME): ", me_expr);
		}

		// involve non-'gamma*e' polarizations at early stage (here)
		if ((m_settings & SQRME_EARLY_PHOTON_HELICITIES) == 0) m_settings |= SQRME_EARLY_PHOTON_HELICITIES;
	}

	// now the simplification is allowed
	// TODO: check if this extra reduction can hurry calculation
	// Perhaps (CLGammaSeqSimplify::simplify()) can also be doubled here
	me_expr = CExpression::reduce(me_expr);
	if (m_mass_rule != NULL) { me_expr = CExpression::arrge(m_mass_rule->ApplyTo(me_expr)); }
	dbg->log(46, "PrepareME() me_expr pre-reduced: ", me_expr);

	// involve helicity vectors
	if (m_settings & SQRME_EARLY_PHOTON_HELICITIES) {
		int i;
		for(i=0; i<inds->size(); i++) {
			CExprContainer* cart = (CExprContainer*)inds->GetAt(i);
			int part_ID = cart->get_info();
			CEParticle* part = m_kin->get_particle_by_ID(part_ID);
			if ((part == NULL) || (!part->check_PK(CEParticle::PARTICLE_VECTOR_BOSON, true))) continue;
			CPAbstractBoson* part_boson = (CPAbstractBoson*)part;
			
			CExpression* h_val = part_boson->getHelicityVector(cart->get_expr());
			if (h_val != NULL) {
				dbg->log(21, "PrepareME() helicity vector found: ", h_val);
				
				// extract denominator from helicity expression
				h_val = involve_denominators_cache(h_val, true);
				
				me_expr = CExpression::arrge(me_expr->multiply(h_val));
				inds->RemoveAt(i, true); i--; // do not use density martix for this index

				// perform simplification after every the helicity vector multiplication
				if (m_settings & SQRME_PRESIMPLIFY_ME) me_expr = CLGammaSeqSimplify::simplify(me_expr);

				to_be_simplified = false; // to avoid extra simplification
				dbg->log(41, "PrepareME() helicity vector involved (res ME): ", me_expr);
			} // else common DM should be used
		}
	}

	// presimplify ME (speed profit)
	if ((to_be_simplified) && (m_settings & SQRME_PRESIMPLIFY_ME)) {
		//int lev_old = CLGammaSeqSimplify::dbg->set_level(100);//////// temp
		me_expr = CLGammaSeqSimplify::simplify(me_expr);
		//CLGammaSeqSimplify::dbg->set_level(lev_old);//////// temp
	}

	// similar couplings unify
	if (m_kin_simp_rules != NULL) {
		me_expr = CExpression::arrge(m_kin_simp_rules->ApplyTo(me_expr));
	}

	// involve mandelstam variables
	if ((m_settings & SQRME_ALLOW_MANDELSTAMS) && (m_mand_vars_rule != NULL)) {
		me_expr = CExpression::arrge(m_mand_vars_rule->ApplyTo(me_expr));
	}

	// mass simplify
	if (m_mass_rule != NULL) {
		me_expr = CExpression::arrge(m_mass_rule->ApplyTo(me_expr));
	}

	// addition
	// multiply *i for every loop in ME (will be canceled in VI-restoration)
	if (loops_no > 0) {
		int l_i;
		for(l_i=0; l_i<loops_no; l_i++) { me_expr = me_expr->multiply(CENumber::GetImUnit()); }
	}

	if ((m_settings & SQRME_EXPAND_AND_SPLIT) == SQRME_EXPAND_AND_SPLIT) {
		// Performance Note: if (m_settings & SQRME_PRESIMPLIFY_ME) ME is already expanded
		me_expr = CExpression::reduce(me_expr);
	}

	dbg->log(21, "PrepareME() (result)", me_expr);
	return me_expr;
}

/*
*	Changes similar vector indices in second ME (old: no similar indices should be here),
*	fills out "m_vector_parts_DMs" & "m_vector_PostProcDMs" data structures.
*	if (m_loops_no == 0) ==> all DMs to m_vector_parts_DMs.
*/
bool CMESquareLib::prepareVectorPolarizationData() {
	dbg->log(20, "prepareVectorPolarizationData()");
	// Note: both arrays should be sorted and should contain data for similar particles
	CEPtrArray* inds_a = m_me_a->get_external_indices_list();
	CEPtrArray* inds_b = m_me_b->get_external_indices_list();

	if (inds_b->size() != inds_a->size()) return false;
	CEReplList* rl = new CEReplList(); // list of ext indices to be changed
	// Note: this is probably unneeded (vect indices are now changed earlier)
	int i;
	for(i=0; i<inds_a->size(); i++) {
		CExprContainer* cart_a = (CExprContainer*)inds_a->GetAt(i);
		CExprContainer* cart_b = (CExprContainer*)inds_b->GetAt(i);

		// check data
		int part_ID = cart_a->get_info();
		if (cart_b->get_info() != part_ID) return false; // Error
		CEParticle* part = m_kin->get_particle_by_ID(part_ID);
		if (part == NULL) return false; // Error
		if (!part->check_PK(CEParticle::PARTICLE_VECTOR_BOSON, true)) return false; // Error
		CPAbstractBoson* part_boson = (CPAbstractBoson*)part;

		// check index identity
		CExpression* a = cart_a->get_expr();
		CExpression* b = cart_b->get_expr(); bool do_del_b = false;
		if (a->equals(b)) {
			b = CESymbol::getUnique(); do_del_b = true;
			rl->add(a, b); // 'a' means old 'b' here
			cart_b->set_expr(b);
		}

		// for any helicity data:
		// N-dim ==> to m_vector_PostProcDMs, 4-dim ==> to m_vector_parts_DMs
		bool to_PostProcDMs = ((m_settings & SQRME_LATE_PHOTON_HELICITIES) != 0); // (m_NDimCalc)
		CExpression* dm = NULL;
		//bool h_allowed = ((m_settings & SQRME_LATE_PHOTON_HELICITIES) != 0);
		bool h_allowed = true;

		if (h_allowed) { // search for helicity
			CExpression* h_val_a = part_boson->getHelicityVector(a);
			if (h_val_a != NULL) {
				dbg->log(21, "PrepareME() helicity vector found: ", h_val_a);
				CExpression* h_val_b = part_boson->getHelicityVector(b);
				if (h_val_b == NULL) CEException::halt(dbg, "NULL b-vector");
				dbg->log(21, "PrepareME() helicity vector found: ", h_val_b);
				h_val_a = involve_denominators_cache(h_val_a, true); // extract denominator from helicity expression
				h_val_b = involve_denominators_cache(h_val_b, true); // extract denominator from helicity expression
				dm = h_val_a->multiply(h_val_b);

				// Note???: DO NOT simplify dm while in N-dim (or 4-dim if required) calculation
				if(!m_NDimCalc) { // TODO: check what's wrong with N-dim here
					dm = CExpression::reduce(dm);
				}
			}
		}

		if (dm == NULL) { // no helicity data are available
			CEPtrArray* dm_pars = new CEPtrArray();
			dm_pars->add(a->clone());
			dm_pars->add((do_del_b)?(b):(b->clone()));
			// get density matrix of particle 'part_ID' with indices {a, b}
			dm = part_boson->getDensityMatrix(dm_pars);
			delete dm_pars;
		}

		if (dm != NULL) {
			dm = involve_denominators_cache(dm, true); // extract denominator if exists

			if (to_PostProcDMs) {
				// dm will be arranged in post-processing
				m_vector_PostProcDMs->add(dm);
			} else {
				// reduce DM (?)
				//dm = CExpression::reduce(dm);
				m_vector_parts_DMs->add(dm);
			}
		}
		else return false; // Error
	}

	if (rl->size() > 0) {
		dbg->log(0, "Warning: non-unque indices spotted (old code involved to fix)", rl);
		CExpression* b_new = rl->ApplyTo(m_me_b->get_expr_val()->clone());
		m_me_b->set_expr_val(b_new);
		delete b_new;
	}
	delete rl;
	dbg->log(20, "prepareVectorPolarizationData() : OK, density matrices lists (in-line, post-proc):", m_vector_parts_DMs, m_vector_PostProcDMs);
	return true;
}

/*
*	Involve photon polarization vectors into (ME*cjgME)_{a, b} expression.
*/
CExpression* CMESquareLib::InvolvePostProcDMs(CExpression* expr) {
	if (m_settings & SQRME_EARLY_PHOTON_HELICITIES) return expr; // helicities are involved

	// multiply to post-proc density matrices
	int dm_num = m_vector_PostProcDMs->size();
	int dm_i;
	for(dm_i=0; dm_i<dm_num; dm_i++) {
		dbg->log(0, "Vector particle density matrix involving (post-processing)");
		CExpression* dm = (CExpression*)m_vector_PostProcDMs->GetAt(dm_i);
		//dm = CExpression::reduce(dm->clone());
		dm = CExpression::arrge(dm->clone());
		dbg->log(27, "post-proc DM reduced", dm);
		expr = expr->multiply(dm);
		expr = CExpression::reduce(expr);
		dbg->log(0, "Vector particle density matrix involving: done");
	}
	
	if (dm_num > 0) dbg->log(56, "post-proc DMs involved:", expr);
	return expr;
}

/*
*	Note: the SQUARING procedure can be fastered as follows
*	(A+B)^2->AA+2AB+BB instead of AA+AB+BA+BB
*	Warning: this may be done IF NO ARRANGING HAVE BEEN PERFORMRD
*	in matrix elements since thay have been separated.
*	That means any arranging here should swich off the m_ME_SQUARE flag.
*/
CExpression* CMESquareLib::coupleMESplitting() {
	dbg->log(20, "coupleMESplitting()");
	CExpression* me_a = m_me_a->get_expr_val();
	CExpression* me_b = m_me_b->get_expr_val();
	
	CESum* sum_a = NULL; long s_a=1;
	if (me_a->getProperty(CEProperties::PROP_MASK_IS_SUM)) { sum_a = (CESum*)me_a; s_a=sum_a->size(); }
	CESum* sum_b = NULL; long s_b=1;
	if (me_b->getProperty(CEProperties::PROP_MASK_IS_SUM)) { sum_b = (CESum*)me_b; s_b=sum_b->size(); }

	//	*****************************************************	<scroll>
	CEProgressWatch* pw = NULL;
	if (s_a*s_b > 10) {
		char* proc_title = (char*)CMemoryManager::mmalloc(100);
		if (m_ME_SQUARE) sprintf(proc_title, "ME Squaring (%ld items)", s_a);
		else sprintf(proc_title, "Coupling ME(%ld items) to conjg[ME(%ld items)]", s_a, s_b);
		pw = CEProcessor::newProgressWatch(proc_title);
		CMemoryManager::mfree(proc_title);
	}
	//	*****************************************************	<\scroll>
	if ((m_ME_SQUARE) && (s_a != s_b)) {// see notes above.
		dbg->log(0, "WARNING: illegal arranging have been performed in SQUARE procedure. SQUARE flag is swiched off.");
		m_ME_SQUARE = false;
	}

	// calc result
	CESum* res_sum = new CESum();
	bool _SubArrange=((m_settings & SQRME_NO_SUB_ARRANGE)==0); // arrange by every N(1000) items
	bool _SubMinimize=((m_settings & SQRME_SAVE_MEMORY)==SQRME_SAVE_MEMORY); // minimize by every N(1000) items --> reduce memory
	CESum* sub_sum = (_SubArrange)?(new CESum()):NULL; long sub_items=0; long sub_max=1000;

	long i_a;
	long i_b=0;
	for(i_a=0; i_a<s_a; i_a++) {
		CExpression* expr_a = NULL;
		if (sum_a != NULL) expr_a = sum_a->GetAt(i_a);
		else expr_a = me_a;

		// add fast squaring here
		if (m_ME_SQUARE) {
			for(i_b=0; i_b<=i_a; i_b++) {
				CExpression* expr_b = NULL;
				if (sum_b != NULL) expr_b = sum_b->GetAt(i_b);
				else expr_b = me_b;

				CExpression* couple_res = coupleExprs(expr_a, expr_b);
				if (i_b != i_a) { couple_res = couple_res->multiply(CENumber::GetTwo()); }

				if (_SubArrange) {
					sub_sum->addItem(couple_res);
					if (couple_res->getProperty(CEProperties::PROP_MASK_IS_SUM)) { sub_items+=((CESum*)couple_res)->size(); }
					else { sub_items+=1; }
					if (sub_items >= sub_max) {
						couple_res=CExpression::reduce(sub_sum); sub_sum=new CESum(); sub_items=0;
						if (_SubMinimize) { couple_res=CViewSimplify::factorize_partial(couple_res,CViewSimplify::PART_FACT_DEFAULT); }
						res_sum->addItem(couple_res);
					}
				} else {
					res_sum->addItem(couple_res);
				}

				//	*****************************************************	<scroll>
				if (pw != NULL) pw->step((i_a*(i_a+1))/2+i_b+1, s_a*(s_a+1)/2);
				//	*****************************************************	<\scroll>
			}
		}
		else {
			for(i_b=0; i_b<s_b; i_b++) {
				CExpression* expr_b = NULL;
				if (sum_b != NULL) expr_b = sum_b->GetAt(i_b);
				else expr_b = me_b;
				CExpression* couple_res = coupleExprs(expr_a, expr_b);

				if (_SubArrange) {
					sub_sum->addItem(couple_res);
					if (couple_res->getProperty(CEProperties::PROP_MASK_IS_SUM)) { sub_items+=((CESum*)couple_res)->size(); }
					else { sub_items+=1; }
					if (sub_items >= sub_max) {
						couple_res=CExpression::reduce(sub_sum); sub_sum=new CESum(); sub_items=0;
						if (_SubMinimize) { couple_res=CViewSimplify::factorize_partial(couple_res,CViewSimplify::PART_FACT_DEFAULT); }
						res_sum->addItem(couple_res);
					}
				} else {
					res_sum->addItem(couple_res);
				}

				//	*****************************************************	<scroll>
				if (pw != NULL) pw->step(i_a*s_b+i_b+1, s_a*s_b);
				//	*****************************************************	<\scroll>
			}
		}
	}
	if (_SubArrange) {
		if (sub_sum->size() > 0) {
			CExpression* couple_res=CExpression::reduce(sub_sum); sub_sum=NULL;
			if (_SubMinimize) { couple_res=CViewSimplify::factorize_partial(couple_res,CViewSimplify::PART_FACT_DEFAULT); }
			res_sum->addItem(couple_res);
		}
	}
	if (sub_sum != NULL) { delete sub_sum; sub_sum=NULL; }
	//	*****************************************************	<scroll>
	if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
	//	*****************************************************	<\scroll>

	if ((sum_a != NULL) && (!sum_a->getFlag())) res_sum->inverseFlag();
	if ((sum_b != NULL) && (!sum_b->getFlag())) res_sum->inverseFlag();

	dbg->log(41, "coupleMESplitting() internal result (not reduced): ", res_sum);
	CExpression* res = res_sum;
	bool skip_arrange = ((m_settings & SQRME_NO_POSTARRANGE) == SQRME_NO_POSTARRANGE);
	if (_SubMinimize) skip_arrange=true; // no reason to reduce in memory saving mode
	if (m_loops_no>0) skip_arrange=false; // Force post-reduce for VI calc
	if (!skip_arrange) {
		res = CExpression::reduce(res); // reduce later
		dbg->log(80, "coupleMESplitting() internal result (reduced): ", res);
	}
	return res;
}

CExpression* CMESquareLib::coupleExprs(CExpression* a, CExpression* b) {
	dbg->log(46, "coupleExprs() a:", a);
	dbg->log(46, "coupleExprs() b:", b);
	//dbg->logperiod(35, "CMESquareLib::coupleExprs()"); //TEMP

	CExpression* a_x = a->clone();
	CExpression* b_x = CEBasic::Congugate(b,false); if(b_x==b)b_x=b_x->clone();

	dbg->log(146, "coupleExprs() b*:", b_x);

	// if no preliminary simplification involved
	if ((m_settings & SQRME_PRESIMPLIFY_ME) == 0) {
		a_x = CLGammaSeqSimplify::simplify(a_x);
		b_x = CLGammaSeqSimplify::simplify(b_x);
	}

	// reconstruct proton structure functions
	//TODO: think if there is a reason to extract this during the Pre-Simplify-ME stage
	CExpression* _FactorSF=NULL;
	if(m_bRestoreProtonStrFunc) {
		CSequence seqME; seqME.addItem(a_x); seqME.addItem(b_x); CSequence seqFact;
		if (CPProtonSF::ReconstructSF(&seqME, &seqFact, &m_arrProtonID, &m_SFTmpIndices)) {
			_FactorSF=CENumber::GetUnit(); for(int i=0; i<seqFact.size(); i++) { _FactorSF=_FactorSF->multiply(seqFact.GetAt(i)); }
			seqFact.ClearWODelObjs();
		}
		a_x=seqME.GetAt(0); b_x=seqME.GetAt(1); seqME.ClearWODelObjs();
	}

	CExpression* res = a_x->multiply(b_x); if(_FactorSF!=NULL)res=res->multiply(_FactorSF);

	// calc up to fixed mass degree
	if (m_mass_rule != NULL) res = m_mass_rule->ApplyTo(res);
	//dbg->logperiod(55, "CMESquareLib::coupleExprs() #00"); //TEMP
	//CESum::dbg->set_level(100); //TEMP
	//CEProduct::dbg->set_level(100); //TEMP
	res = CExpression::reduce(res);
	//dbg->logperiod(55, "CMESquareLib::coupleExprs() #01"); //TEMP
	//CEException::halt(); //TEMP

	if (m_mass_rule != NULL) res = CExpression::arrge(m_mass_rule->ApplyTo(res));
	dbg->log(56, "ME1*cjg(ME2) reduced:", res);

	// sum (average) fermion spin
	CLFermionSpinSum::sum_fermion_spins(res);
	dbg->log(56, "ME1*cjg(ME2) fermion spins summarized:", res);

	// evaluate spurs
	CExpression* res2 = res->evaluate();
	if (res2 != res) { delete res; res = res2; }
	if (m_mass_rule != NULL) res = m_mass_rule->ApplyTo(res);

	res = CExpression::reduce(res);

	//TEMP//TODO: remove next to ALL_REAL option
	//res = CLSimplifyMisc::get_Re(res, true);

	if (m_mass_rule != NULL) res = CExpression::arrge(m_mass_rule->ApplyTo(res));
	dbg->log(56, "ME1*cjg(ME2) spurs evaluated:", res);
	//dbg->logperiod(55, "CMESquareLib::coupleExprs() #1"); //TEMP

	// multiply to vector particles density matrices
	int dm_num = m_vector_parts_DMs->size();
	int dm_i=0;
	for(dm_i=0; dm_i<dm_num; dm_i++) {
		CExpression* dm = (CExpression*)m_vector_parts_DMs->GetAt(dm_i);
		res = res->multiply(dm->clone());
		res = CExpression::reduce(res);
		if (m_mass_rule != NULL) res = CExpression::arrge(m_mass_rule->ApplyTo(res));
		dbg->log(56, dm_i, "ME1*cjg(ME2) *DMs (%ld):", res);
	}
	dbg->log(56, "ME1*cjg(ME2) evaluated *DMs:", res);
	
	// Re(ME1*ME2^{*})
	// Note: Virtual integrals may also contain Im part ==> TODO: get Re() after VI are restored
	// Note: Abstract polarization vectors may also be complex, check it/ add ALL_REAL option
	res = CLSimplifyMisc::get_Re(res, true);
	//dbg->logperiod(55, "CMESquareLib::coupleExprs() #2"); //TEMP

	// simplifying (normalizing) result
	res = kinematic_reduce(res);

	//dbg->logperiod(55, "CMESquareLib::coupleExprs() #3"); //TEMP

	if ((m_settings & SQRME_ALLOW_MANDELSTAMS) && (m_mand_vars_rule != NULL)) {
		res = CExpression::arrge(m_mand_vars_rule->ApplyTo(res));
	}

	//dbg->logperiod(55, "CMESquareLib::coupleExprs() #4"); //TEMP

	if (m_mass_rule != NULL) res = CExpression::arrge(m_mass_rule->ApplyTo(res));
	dbg->log(76, "ME1*cjg(ME2)*DMs kinematic relations involved:", res);

	// change E*p*p*p*p --> E_pppp in non-loop processing
	if (m_loops_no <= 0) {
		if (CEAntiSymmProd::InvolveAsymmProd(res)) { res = CExpression::reduce(res); }
	}

	// further simplification
	// helicities in Berends notation
	/*if (CEProcessor::getEnvironment()->m_photon_helisity_notation == CPPhoton::PH_BERENDS_NOTATION) {
		// Substitute photon helicities
		CEMap* map = data->get_node(CPPhoton::dm_data_node);
		if (map != NULL) {
			//dbg->log(10, "to multiply map", map);
			map->reset_iterator();
			bool mult_conj =  false;
			for (;map->has_next();) {
				CENamedItem* item = (CENamedItem*)map->next();
				CExpression* mult_it = (CExpression*)(item->getObject());
				if (mult_conj) {
					mult_it = CEBasic::Congugate(mult_it->clone(),true);
					mult_it = CExpression::reduce(mult_it);
					mult_conj = false;
				} else {
					mult_it = mult_it->clone();
					mult_conj = true;
				}
				dbg->log(30, "to multiply item", mult_it);
				res = res->multiply(mult_it);
				//res = CExpression::reduce(res);
				res = CHEPLib::repl_similar_kin(res);
				if (rule_mx != NULL) {res->applyRule(rule_mx);res = CExpression::reduce(res);}
				//dbg->log(30, "reduced", res);
				dbg->logperiod(30);
			}
		}
	}*/

	//dbg->logperiod(35, "CMESquareLib::coupleExprs() done"); //TEMP
	dbg->log(46, "ME1*cjg(ME2)*DMs sub-result:", res);
	return res;
}

/*
*	Extracts denominators from matrix elements ("m_me_a" and "m_me_b")
*	"m_a" and "m_b" are supposed to be initializad later.
*/
void CMESquareLib::init_denominators_cache() {
	dbg->log(31, "init_denominators_cache()");
	if (m_denominators_cache != NULL) { delete m_denominators_cache; m_denominators_cache=NULL; } // Error ?

	m_denominators_cache = new CEPtrArray();
	CExpression* me = NULL; CRReplaceDenominators* r = NULL; CExpression* me_new = NULL;
	
	me = m_me_a->get_expr_val();
	// No mass rule may be involed in ME denominator in general case
	// Mass may be omitted if only no virtual integration will follows
	bool allow_mass_simp = (m_me_a->get_virt_momenta_list()->size() == 0);
	int mass_policy_flag = ((allow_mass_simp)?D_CACHE_ALLOW_MASS_SIMPLIFICATION:0);

	r = new CRReplaceDenominators(CRReplaceDenominators::REPL_DENS_DEFAULT, mass_policy_flag);
	me_new = r->ApplyTo(me, false);
	if (me_new != me) { m_me_a->set_expr_val(me_new); delete me_new; }

	if (m_me_b != m_me_a) {
		// perform extracting denominators in the same rule
		me = m_me_b->get_expr_val();
		allow_mass_simp = (m_me_b->get_virt_momenta_list()->size() == 0);
		int mass_policy_new = ((allow_mass_simp)?D_CACHE_ALLOW_MASS_SIMPLIFICATION:0);
		if (mass_policy_new != mass_policy_flag) r->SetDensFlags(mass_policy_new);

		me_new = r->ApplyTo(me, false);
		if (me_new != me) { m_me_b->set_expr_val(me_new); delete me_new; }
	}

	m_denominators_cache->add_array(r->getReplacedValues()->clone());
	delete r;
	dbg->log(20, "denominators cache created:", m_denominators_cache);
}

const DWORD CMESquareLib::D_CACHE_ALLOW_MASS_SIMPLIFICATION	= 0x01; // flag in denominators cache structure
CExpression* CMESquareLib::involve_denominators_cache(CExpression* val, bool allow_mass_simp) {
	dbg->log(31, "involve_denominators_cache()", val);
	if (val == NULL) return NULL;
	if (m_denominators_cache == NULL) m_denominators_cache = new CEPtrArray();

	int mass_policy_flag = ((allow_mass_simp)?D_CACHE_ALLOW_MASS_SIMPLIFICATION:0);
	CRReplaceDenominators* r = new CRReplaceDenominators(
		CRReplaceDenominators::REPL_DENS_DEFAULT, m_denominators_cache, mass_policy_flag);
	CExpression* res = r->ApplyTo(val, true);
	CEPtrArray* new_dens_arr = r->getReplacedValues();
	if (new_dens_arr->size() != m_denominators_cache->size()) {
		// update denominators cache array
		delete m_denominators_cache; m_denominators_cache = new_dens_arr->clone();
		dbg->log(20, "denominators cache updated:", m_denominators_cache);
	}
	delete r;
	dbg->log(36, "expression denominator cached: ", res);
	return res;
}

void CMESquareLib::simplify_cached_denominators() {
	dbg->log(31, "simplify_cached_denominators()");
	//[debug]dbg->log(71, "dens cache before simplification", m_denominators_cache);
	
	long s = m_denominators_cache->size();
	long i;
	for(i=0; i<s; i++) {
		CExprContainer* cart = (CExprContainer*)m_denominators_cache->GetAt(i);
		CExpression* d = cart->get_expr();
		bool mass_simp_allowed = ((cart->get_info() & D_CACHE_ALLOW_MASS_SIMPLIFICATION) == D_CACHE_ALLOW_MASS_SIMPLIFICATION);

		// 1. simplify
		CExpression* d_new = CExpression::reduce(d, false);
		if (d_new != d) { cart->set_expr(d_new); delete d_new; }

		// 2. involve kinematics identities
		CExpression* d_simp = kinematic_reduce(cart->get_expr()->clone());

		// involve mandelstam variables
		// no mandelstams here ==> errors in VICompoeLib
		//if ((m_settings & SQRME_ALLOW_MANDELSTAMS) && (m_mand_vars_rule != NULL)) {
		//	d_simp = CExpression::arrge(m_mand_vars_rule->ApplyTo(d_simp));
		//}

		// 3. apply mass rule
		if ((mass_simp_allowed) && (m_mass_rule != NULL)) {
			//dbg->log(36, "m_mass_rule:", m_mass_rule);
			d_simp = CExpression::arrge(m_mass_rule->ApplyTo(d_simp));
		}

		cart->set_expr(d_simp); delete d_simp;
	}

	dbg->log(36, "denominators simplified:", m_denominators_cache);
	//CEException::halt();
}

CExpression* CMESquareLib::restore_denominators(CExpression* val) {
	dbg->log(31, "restore_denominators()");
	if ((m_denominators_cache == NULL) || (m_denominators_cache->size() == 0)) return val;
	//[debug]dbg->log(31, "m_denominators_cache", m_denominators_cache);

	long s = m_denominators_cache->size();
	CEReplList* r = new CEReplList();
	long i;
	for(i=0; i<s; i++) {
		CExprContainer* cart = (CExprContainer*)m_denominators_cache->GetAt(i);
		r->add(cart->get_info_expr(), cart->get_expr());
	}
	//[debug]dbg->log(90, "restore_denominators() internal expression", val);
	//[debug]dbg->log(31, "restore_denominators() internal restore rule", r);
	val = r->ApplyTo(val);
	//[debug]dbg->log(70, "restore_denominators() internal result", val);
	delete r;

	// reduce/arrange here? // Nt: reduction is MUCH FASTER BEFORE restoring denominators
	bool skip_arrange = ((m_settings & SQRME_NO_POSTARRANGE) == SQRME_NO_POSTARRANGE);
	if (!skip_arrange) {
		//[debug]dbg->log(31, "Temp 1", val);
		//val = CExpression::reduce(val);
		val = CExpression::arrge(val);
		//[debug]dbg->log(31, "Temp 2", val);
	}
	dbg->logperiod(10, "Denominators restored.");
	return val;
}

/*
*	May not involve mass rules.
*/
CExpression* CMESquareLib::kinematic_reduce(CExpression* val) {
	dbg->log(61, "kinematic_reduce()", val);
	bool do_reduce = false;

	// rule takes care about kinematically-similar couplings
	if (m_kin_simp_rules != NULL) { // else error?
		//[debug]dbg->log(91, "using rules", m_kin_simp_rules);
		val = m_kin_simp_rules->ApplyTo(val);
		do_reduce = true;
	}

	//[debug]dbg->log(91, "kinematic_reduce() sub-result (step 1):", val);
	
	// set to zero e*p couplings and e_i*p_j for collinear consideration
	if (m_basic_coupling_vals != NULL) {
		//[debug]dbg->log(91, "using rules", m_basic_coupling_vals);
		val = m_basic_coupling_vals->ApplyTo(val);
		do_reduce = true;
	}

	if (do_reduce) val = CExpression::reduce(val);
	dbg->log(61, "kinematic_reduce() result:", val);
	return val;
}

CExpression* CMESquareLib::virt_integrate(CExpression* val) {
	CEPtrArray* virts_1 = m_me_a->get_virt_momenta_list();
	long s_1 = virts_1->size();
	CEPtrArray* virts_2 = m_me_b->get_virt_momenta_list();
	long s_2 = virts_2->size();
	if ((s_1+s_2) == 0) return val; // no virtual momentum exists

	dbg->log(25, "virt_integrate()");
	dbg->log(46, "virt_integrate() source:", val);
	dbg->log(25, s_1+s_2, "virt_integrate() total %ld virtual momenta.");
	if ((s_1+s_2) > 1) {
		dbg->log(0, "Error: no multiple virtual integration implemented. Integration canceled.");
		return val; // no virtual momentum can be integrated
	}

	CExpression* v_sym;
	if (s_1 == 1) v_sym = (CExpression*)virts_1->GetAt(0);
	else v_sym = (CExpression*)virts_2->GetAt(0);

	CExpression* res = CVIComposeLib::composeVI(val, v_sym);
	if (res == NULL) {
		dbg->log(0, "Error composing virtual integrals. Integration canceled.");
		return val;
	}
	delete val;

	// Arrange factor (*i is already multiplied in ME preparation procedure)
	CExpression* pi = CEBasic::GetPiSym(); // not cloned
	CExpression* vi_fact = CENumber::GetUnit()->divide((CENumber::GetNumber(16))->multiply(pi->sqra()));
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if ((phys != NULL) && (CEUtil::equals(phys->ClassName(), CSMPhysics::CLASS_NAME))) {
		vi_fact = vi_fact->multiply(((CSMPhysics*)phys)->getModelContsUnitFactor());
		vi_fact = CExpression::reduce(vi_fact);
	}
	dbg->log(95, "VI-factor created(tmp):", vi_fact); // TODO:decrease dbg level
	res = vi_fact->multiply(res);
	dbg->log(46, "virt_integrate() result:", res);
	return res;
}
