////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// VIComposeLib.cpp: implementation of the CVIComposeLib class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

// max number of virtual momentum-containing denominators allowed in single product
// engine throws internal error if number is exceeded
#define MAX_ALLOWED_DENS_NUM	4

CDebug* CVIComposeLib::dbg = new CDebug("CVIComposeLib");
CDebug* CVIComposeLib::dbg_parsing = new CDebug("CVIComposeLibParsing");

//////////////////////////////////////////////////////////////////////
// Static Entries
// Input parameters are not affected.
// NULL result means error.
//////////////////////////////////////////////////////////////////////
CExpression* CVIComposeLib::composeVI(CExpression* source, CExpression* virt_sym) {
	dbg->log(10, "composeVI()", source);
	dbg->log(20, "composeVI() virt symbol:", virt_sym);
	CExpression* res = NULL;

	if ((source != NULL) && (virt_sym != NULL)) {
		CVIComposeLib* lib = new CVIComposeLib(source, virt_sym);
		res = lib->do_compose();
		delete lib;
	} else {
		printf("Error: composeVI() NULL parameters.\n");
		dbg->log(0, "Error: composeVI() NULL parameters.");
	}

	dbg->log(30, "composeVI() result:", res);
	return res;
}

/*
*	The sources are unaffected.
*/
CExpression* CVIComposeLib::composeVI(CVIDensData* dens_data, CExpression* num, CExpression* virt_sym) {
	dbg->log(10, "composeVI() dens_data: ", dens_data);
	dbg->log(20, "composeVI() num:", num);
	dbg->log(20, "composeVI() virt symbol:", virt_sym);
	CExpression* res = NULL;

	if ((dens_data != NULL) && (num != NULL) && (virt_sym != NULL)) {
		// Library "source" is not used here.
		CVIComposeLib* lib = new CVIComposeLib(NULL, virt_sym);
		res = lib->composeVI_impl((CVIDensData*)dens_data->clone(), num->clone(), false);
		delete lib;
	} else {
		printf("Error: composeVI(DD, num) NULL parameters.\n");
		dbg->log(0, "Error: composeVI(DD, num) NULL parameters.");
	}

	dbg->log(30, "composeVI() result:", res);
	return res;
}

/*
*	Does not affect input parameters.
*/
CVirtIntegral* CVIComposeLib::createVISeq(CSequence *dens_data, CSequence *_couplings, CEPtrArray* v_indices) {
	dbg->log(30, "creating VI", dens_data);
	if (dens_data == NULL) { //Error
		printf("Error: CVIComposeLib::createVirtIntegral(NULL) call.\n");
		dbg->log(0, "Error: createVirtIntegral(NULL) call.");
		return NULL;
	}

	CVIDensData* dd = new CVIDensData(dens_data);
	CVirtIntegral* res = createVI(dd, _couplings, v_indices);
	delete dd;
	return res;
}

/*
*	Does not affect input parameters.
*/
CVirtIntegral* CVIComposeLib::createVI(CVIDensData* dens_data, CSequence* _couplings, CEPtrArray* v_indices) {
	if (dens_data == NULL) { dbg->log(0, "Error: createVI(NULL) call."); return NULL; }

	CVirtIntegral* res = NULL;
	bool NoCouplings = (_couplings != NULL)?(_couplings->size()==0):(true);
	bool NoIndices = (v_indices != NULL)?(v_indices->size()==0):(true);
	if (NoCouplings && NoIndices) { //CVirtInt0
		res = new CVirtInt0(dens_data);
	}
	else { //CVirtIntC
		res = new CVirtIntC(dens_data, _couplings, v_indices);
	}
	dbg->log(20, "VI created", res);
	return res;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CVIComposeLib::CVIComposeLib(CExpression* source, CExpression* virt_sym) : CELibrary() {
	m_source = (source!=NULL)?(source->clone()):NULL;
	if (virt_sym != NULL) {
		m_virt_sym = virt_sym->clone();
		if (!m_virt_sym->getFlag()) m_virt_sym->setFlag();
		if (!m_virt_sym->getLevel()) m_virt_sym->setLevel(true);
	} else {
		m_virt_sym = NULL;
	}

	m_DENSDATA_cache = new CEPtrArray();

	m_error_flag = false;
	//m_P_curr = NULL;

	// cache neutrino mass symbols
	m_NeutrinoMassSyms = NULL;
	// add neutrino masses to m_NeutrinoMassSyms (if they are marked as 0-order).
	CEPtrArray* mass_order_info_arr = CEProcessor::getEnvironment()->GetMassOrderInfo();
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if ((mass_order_info_arr != NULL) && (phys != NULL)) {
		int info_s = mass_order_info_arr->size();
		m_NeutrinoMassSyms = new CSequence();
		int info_i;
		for(info_i=0; info_i<info_s; info_i++) {
			CExprContainer* cart = (CExprContainer*)mass_order_info_arr->GetAt(info_i);
			int part_kind = cart->get_info(0);
			int mass_order = cart->get_info(1);
			if (((part_kind & CEParticle::PARTICLE_NEUTRINO)==CEParticle::PARTICLE_NEUTRINO) && (mass_order == 0)) {
				CExpression* mass_sym = phys->get_mass_sym(part_kind);
				//dbg->log(0, part_kind, "Neutrino (%ld) mass to zero", mass_sym);
				if (mass_sym != NULL) m_NeutrinoMassSyms->addItem(mass_sym->clone());
			}
		}
	}
}

CVIComposeLib::~CVIComposeLib() {
	if (m_source != NULL) delete m_source;
	if (m_virt_sym != NULL) delete m_virt_sym;

	delete m_DENSDATA_cache;
	//if (m_P_curr != NULL) delete m_P_curr;
}

//////////////////////////////////////////////////////////////////////
// Lib entry
//////////////////////////////////////////////////////////////////////
CExpression* CVIComposeLib::do_compose() {
	dbg->log(40, "do_compose().");
	if ((m_source == NULL) || (m_virt_sym == NULL)) { dbg->log(0, "ERROR: Null parameters."); return NULL; }
	
	CExpression* res = parse_expr(m_source);

	dbg->log(50, "do_compose() result:", res);
	return res;
}

//////////////////////////////////////////////////////////////////////
// Sum splitting procedure
// Does not affect source
//////////////////////////////////////////////////////////////////////
CExpression* CVIComposeLib::parse_expr(CExpression* source) {
	dbg->log(50, "parse_expr()", source);
	CExpression* res = NULL;
	if (source->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* sum = (CESum*)source;
		long s = sum->size();
		if (s == 0) return NULL;

		//	*****************************************************	<scroll>
		CEProgressWatch* pw = NULL;
		if (s > 50) pw = CEProcessor::newProgressWatch(
			"CVIComposeLib: denominators parsing (%ld items in sum).", s);
		//	*****************************************************	<\scroll>

		CESum* sum_res = new CESum();
		long i;
		for (i=0; i<s; i++) {
			if (pw != NULL) pw->step(i+1, s);

			CExpression* sub_res = parse_expr(sum->GetAt(i));
			if ((sub_res != NULL) && (!m_error_flag)) sum_res->addItem(sub_res);
			else {
				if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
				dbg->log(0, "Error. parse_expr() fails on", sum->GetAt(i));
				// one item fail means the whole procedure fail
				delete sum_res;
				return NULL;
			}
		}
		if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
		sum->copyAttributesTo(sum_res, true);
		res = sum_res;
	}
	else {
		// cast any other source to product
		bool del_prod = false;
		CEProduct* pr_source = NULL;
		if (source->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			pr_source = (CEProduct*)source;
		} else {
			// this may be (k^2-m^2)^{-1}
			pr_source = new CEProduct();
			pr_source->addItem(source->clone());
			del_prod = true;
		}

		res = parse_product(pr_source);
		if ((del_prod) && (res != pr_source)) delete pr_source;
	}
	dbg->log(90, "parse_expr() result", res);
	return res;
}

//////////////////////////////////////////////////////////////////////
// Product parsing implementation
// Does not affect source
//////////////////////////////////////////////////////////////////////
CExpression* CVIComposeLib::parse_product(CEProduct* prod_source) {
	dbg->log(50, "parse_product()", prod_source);
	dbg_parsing->log(10, "parse_product()", prod_source);

	CEProduct* prod_val = (CEProduct*)prod_source->clone(); // clone source

	// 0.1. check source for Zero
	if (CEBasic::isZeroProd(prod_val)) { dbg->log(40, "Zero VI"); return prod_val; }

	// 1. extract denominators to DENSDATA structures
	CVIDensData* dens_data = extract_denominators_data(prod_val);
	if ((m_error_flag) || (dens_data == NULL)) { delete prod_val; return NULL; }
	dbg->log(40, "dens_data", dens_data);
	dbg_parsing->log(20, "dens_data extracted", dens_data);
	if (dens_data->size() > MAX_ALLOWED_DENS_NUM) {
		dbg->log(0, "ERROR: too many denominators extracted, DENSDATA", dens_data);
		dbg->log(2, "in source", prod_source);
		dbg->log(2, "expression after denominators are extracted", prod_val);
		CEException::halt();
	}

	CExpression* rest = CExpression::reduce(prod_val);
	dbg->log(60, "rest", rest);

	CExpression* res = NULL;
	if (dens_data->size() > 0) {
		res = composeVI_impl(dens_data, rest, false);
	} else {
		if (CEBasic::isZero(rest)) {
			dbg->log(0, "Warning: 0-VI spotted", prod_source);
		} else {
			dbg->log(0, "Error: 0-sized DENSDATA restored, rest factor is", rest);
			CEException::halt(); // error: all the items in sum should be integrated
			delete dens_data;
			res = rest;
		}
	}
	dbg_parsing->log(10, "parse_product() result", res);
	return res;
}

//////////////////////////////////////////////////////////////////////
// Denominators Extraction & Parsing Procedures
//////////////////////////////////////////////////////////////////////
/*
* returns sequence of DENSDATA structures
*/
CVIDensData* CVIComposeLib::extract_denominators_data(CEProduct* prod_source) {
	dbg->log(40, "extract_denominators_data()", prod_source);
	CSequence* seq_res = new CSequence();
	long i;
	for (i=0; i<prod_source->size(); i++) {
		CExpression* item = prod_source->GetAt(i);
		if (!item->getLevel()) {
			CSequence* den_data = parse_denominator(item);
			if (den_data != NULL) {
				seq_res->addItem(den_data);
				prod_source->RemoveAt(i);
				i--;
			}
			
			if (m_error_flag) { delete seq_res; return NULL; }
		}
	}
	CVIDensData* dd_res = new CVIDensData(seq_res);
	delete seq_res;
	return dd_res;
}

/*
*	returns DENSDATA structure or NULL.
*	Note: denominators may be reduced.
*	(den_expr->getLevel()) == false here.
*/
CSequence* CVIComposeLib::parse_denominator(CExpression* den_expr) {
	dbg->log(45, "parse_denominator()", den_expr);
	if (!CRContainCheck::search(den_expr, m_virt_sym)) return NULL;

	// search in DENSDATA cache
	int DD_s = m_DENSDATA_cache->size();
	int DD_j;
	for(DD_j=0; DD_j<DD_s; DD_j++) {
		CExprContainer* cart = (CExprContainer*)m_DENSDATA_cache->GetAt(DD_j);
		if (den_expr->equals(cart->get_info_expr())) {
			return (CSequence*)cart->get_expr()->clone();
		}
	}

	///////////////////////////////////////////
	// inverse level before parsing and arrange
	CExpression* den = den_expr->clone();
	if (den->getLevel()) { dbg->log(0, "Internal ERROR: parse_denominator(expr with TRUE level) call.", den); delete den; return NULL; }
	den->inverseLevel();
	den = CExpression::reduce(den);
	if (!den->getLevel()) {
		// possible unarranged denominator
		dbg->log(0, "Warning: extra LEVEL flag appeared(?) after reducing denominator.", den_expr);
		dbg->log(0, "Arranged denominator (first level flag has dropped):", den); delete den; return NULL;
	}
	///////////////////////////////////////////

	if (!den->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* den_stub = new CESum();
		den_stub->addItem(den);
		den = den_stub;
	}
	
	CESum* den_sum = (CESum*)den;
	long s = den_sum->size();
	// den_sum is here k^k+2k^p+2k^q+Const == (k+P)^2+(Const-P^2)
	// we extract all the scalar couplings containing k, find out the 'P' momentum
	// and extract P^2 from the rest of the sum to find M^2.

	CExpression* p = NULL;
	CENumber nTwo((long)2);
	//CExpression* tmp_vec_index = new CESymbol("TmpIndex");
	bool kk_found = false;
	bool comm_sign = true;
	long i;
	for (i=0; i<s; i++) {
		CExpression* item = den_sum->GetAt(i);
		if (!CRContainCheck::search(item, m_virt_sym)) continue;

		// 1. check if item is k^2
		// 2. check if item is 2*p^k
		// No constant before k^2 is allowed
		// No other constant except 2 (or -2) before p^k is allowed

		if (CEUtil::equals(item->ClassName(), CESProduct::CLASS_NAME)) {
			CExpression* c_m = get_coupled_mom((CESProduct*)item);
			if (c_m == NULL) continue;
			
			if (c_m->equals(m_virt_sym)) {
				if (kk_found) {
					//Error: unsimplified denominator.
					dbg->log(0, "Error: unsimplified denominator.");
					decl_error(p, den_sum); return NULL;
				}
				if (!item->getFlag()) comm_sign = false;
				kk_found = true;
			} else {
				// unexpected p^k coupling
				dbg->log(0, "Error: unexpected p^k coupling.");
				decl_error(p, den_sum); return NULL;
			}

			den_sum->RemoveAt(i); i--;
			continue;
		}

		if (item->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
			CEProduct* item_prod = (CEProduct*)item;
			bool k_item_found = false;
			CExpression* p_add_sym = NULL;

			// check if item is 2*p^k
			long j;
			for(j=0; j<item_prod->size(); j++) {
				CExpression* item_2 = item_prod->GetAt(j);
				if (CEUtil::equals(item_2->ClassName(), CESProduct::CLASS_NAME)) {
					CExpression* c_m = get_coupled_mom((CESProduct*)item_2);
					if (c_m == NULL) continue;
					if (k_item_found) {
						//Error: illegal denominator.
						dbg->log(0, "Error: illegal denominator: <k*p_i>*<k*p_j>.");
						decl_error(p, den_sum, p_add_sym); return NULL;
					}
					k_item_found = true;

					if (c_m->equals(m_virt_sym)) {
						//Error: unsimplified denominator.
						dbg->log(0, "Error: illegal or unsimplified denominator: Const*k^2.");
						decl_error(p, den_sum, p_add_sym); return NULL;
					} else {
						p_add_sym = c_m->clone();
						if (!item_2->getFlag()) p_add_sym->inverseFlag();
					}
					
					item_prod->RemoveAt(j); j--;
					continue;
				}
			}
			if (!k_item_found) {
				dbg->log(0, "Error: 'k' symbol should exist but not found in product:", item_prod);
				decl_error(p, den_sum, p_add_sym); return NULL;
			}

			CExpression* rest_factor = CExpression::reduce(item_prod->clone());
			// if rest contains virt symbol ==> ERROR
			if (CRContainCheck::search(rest_factor, m_virt_sym)) {
				//Error: illegal denominator.
				dbg->log(0, "Error: 'k' symbol should be already extracted from product:", rest_factor);
				delete rest_factor;
				decl_error(p, den_sum, p_add_sym); return NULL;
			}
			
			if (p_add_sym == NULL) {
				dbg->log(0, i, "Error: NULL symbol extracted from p*k coupling in item %ld of sum:", den_sum);
				decl_error(p, den_sum, p_add_sym); return NULL;
			}

			// check if res factor is '2' or '-2'
			if (!rest_factor->getFlag()) { rest_factor->inverseFlag(); p_add_sym->inverseFlag(); }

			if (!nTwo.equals(rest_factor)) {
				dbg->log(0, "Error: illegal denominator: <p*k>*(Const != 2), Const:", rest_factor);
				decl_error(p, den_sum, p_add_sym); return NULL;
			}
			delete rest_factor;
			
			// add p_add_sym to 'p' vector
			if (p == NULL) p = p_add_sym;
			else p = p->add(p_add_sym);

			den_sum->RemoveAt(i); i--;
			continue;
		}

		// Error: some other item in denominator contains virtual symbol
		// Error: illegal denominator.
		dbg->log(0, "Error: illegal denominator due to item:", item);
		decl_error(p, den_sum); return NULL;
	}

	if (!kk_found) {
		// Error: no k^2 item found.
		dbg->log(0, "Error: No k^2 item found in denominator.", den_sum);
		decl_error(p, den_sum); return NULL;
	}

	CExpression* sum_res = CExpression::reduce(den_sum); // den_sum is no longer used
	if (!comm_sign) {
		sum_res->inverseFlag();
		if (p != NULL) p->inverseFlag();
	}

	// calc 'm2'
	if (p != NULL) {
		p = CEProcessor::kinematics()->get_simpliest(CExpression::reduce(p));

		// subtract <p*p> from denominator expression
		CExpression* p2 = new CESProduct(p, p);
		p2 = CExpression::reduce(p2);

		/* // old implementation
		CExpression* tmp_vec_index_1 = new CESymbol("TmpIndex1");
		CExpression* tmp_vec_index_2 = new CESymbol("TmpIndex2");
		CExpression* p2 = (new CEVector(p, tmp_vec_index_1, true))->
			multiply(new CEVector(p, tmp_vec_index_2, true))->
			multiply(new CEMetric(tmp_vec_index_1, tmp_vec_index_2, false, false));
		p2 = CExpression::reduce(p2);
		if ((CRContainCheck::search(p2, tmp_vec_index_1)) ||
			(CRContainCheck::search(p2, tmp_vec_index_2))) { dbg->log(0, "<p*p> creation internal error"); CEException::halt(); }
		delete tmp_vec_index_1; delete tmp_vec_index_2;*/

		sum_res = CExpression::reduce(sum_res->minus(p2));
	}
	CExpression* m2 = sum_res;

	// kin-simplify
	//dbg->log(45, "parse_denominator before simplification m2", m2);
	m2 = CLHEPBasics::KinArrange(m2, CLHEPBasics::KA_MASS_EXACT);
	m2 = CLHEPBasics::KinSimplify(m2, CLHEPBasics::KS_MASS_EXACT);
	//dbg->log(45, "parse_denominator m2 simplified", m2);

	// calc 'm'
	CExpression* m = NULL;
	bool is_zero_mass = false;
	if (CEBasic::isZero(m2)) { m = CENumber::GetZero(); is_zero_mass = true; }
	else {
		if (CEUtil::equals(m2->ClassName(), CEPower::CLASS_NAME)) {
			if (m2->getFlag()) {
				// Error: should be (k+p)^2-m^2
				dbg->log(0, "Error: illegal denominator (k+p)^2+m^2 found ('-' sign was expected).");
				decl_error(p, m2); return NULL;
			}
			m = ((CEPower*)m2)->basis()->clone();
		}
		else {
			// Error: unexpected mass construction
			dbg->log(0, "Error: can't resolve mass in denominator, the mass-like expression is:", m2);
			decl_error(p, m2); return NULL;
		}
	}
	delete m2;
	
	////////////////// neutrino mass ==> to zero (if neutrino is marked as 0-order)
	if ((!is_zero_mass) && (m_NeutrinoMassSyms != NULL)) {
		int neu_s = m_NeutrinoMassSyms->size();
		int neu_i;
		for(neu_i=0; neu_i<neu_s; neu_i++) {
			if (m->equals(m_NeutrinoMassSyms->GetAt(neu_i))) {
				delete m; m = CENumber::GetZero(); is_zero_mass = true; break;
			}
		}
	}

	if (p == NULL) p = CENumber::GetZero();
	CSequence* res = new CSequence();
	res->addItem(p);
	res->addItem(m);

	dbg->log(45, "parse_denominator result", res);

	// store to DENSDATA cache
	m_DENSDATA_cache->add(new CExprContainer(res, 0, 0, den_expr));
	return res;
}

/*
*	returns symbol for momenta coupled to virtual one
*	(or NULL if no virtual monenta involved in coupling)
*	scalar product supposed to be reduced
*/
CExpression* CVIComposeLib::get_coupled_mom(CESProduct* val) {
	if (!CRContainCheck::search(val, m_virt_sym)) return NULL;

	CExpression* val_a = val->get_a();
	CExpression* val_b = val->get_b();
	if (val_a->equals(m_virt_sym)) return val_b;
	if (val_b->equals(m_virt_sym)) return val_a;

	// Error (perhaps, coupling is not reduced).
	m_error_flag = true;
	return NULL;
}

CExpression* CVIComposeLib::decl_error(CExpression* expr_to_delete_1,
								   CExpression* expr_to_delete_2,
								   CExpression* expr_to_delete_3,
								   CExpression* expr_to_delete_4)
{
	if (expr_to_delete_1 != NULL) delete expr_to_delete_1;
	if (expr_to_delete_2 != NULL) delete expr_to_delete_2;
	if (expr_to_delete_3 != NULL) delete expr_to_delete_3;
	if (expr_to_delete_4 != NULL) delete expr_to_delete_4;

	m_error_flag = true;
	dbg->log(0, "ERROR parsing denominator.");
	return NULL;
}

/*
*	Composes VI with specified DENSDATA and numerator.
*	Main call should be with "skip_DD_arrange = false",
*	"skip_DD_arrange = true" is used for faster reqursive calls.
*	destroys input objects
*	
*/
CExpression* CVIComposeLib::composeVI_impl(CVIDensData* dens_data,
										   CExpression* num,
										   bool skip_DD_arrange) {
	dbg->log(50, "composeVI_impl(DDATA, NUM)");
	if (!skip_DD_arrange) {
		CExpression* num_new = CVirtIntegralLib::normalize_integral(dens_data, num, m_virt_sym);
		if (num_new == NULL) return decl_error(dens_data, num);
		// original "num" was deleted during normalizing if no error have occured
		num = num_new;
	}

	// Note: num is reduced here
	long i=0;
	if (num->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* num_sum = (CESum*)num;
		long s = num_sum->size();
		CESum* res_sum = new CESum();
		for (i=0; i<s; i++) {
			CExpression* sub_res = composeVI_impl((CVIDensData*)dens_data->clone(),
				num_sum->GetAt(i)->clone(), true);
			if ((m_error_flag) || (sub_res == NULL)) {
				dbg->log(0, "Error. composeVI_impl(sub-items) results in NULL.");
				return decl_error(dens_data, num, res_sum);
			}
			res_sum->addItem(sub_res);
		}
		num_sum->copyAttributesTo(res_sum, true);
		delete num_sum;
		delete dens_data;
		dbg->log(50, "composeVI_impl(DDATA, Sum) result:", res_sum);
		return res_sum;
	}

	CEProduct* num_prod;
	if (num->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		num_prod = (CEProduct*)num;
	} else {
		num_prod = new CEProduct();
		num_prod->addItem(num);
		num_prod->Arrange(AL_INLINE);
	}
	dbg->log(40, "integrating product:", num_prod);
	dbg->log(75, "dens_data:", dens_data);

	// Extract k^2 form numerator
	for (i=0; i<num_prod->size(); i++) {
		CExpression* item = num_prod->GetAt(i);
		if (!item->getFlag()) { dbg->log(0, "Error: unaranged expression parsing.", num_prod); return decl_error(num_prod, dens_data); }

		if (!CEUtil::equals(item->ClassName(), CESProduct::CLASS_NAME)) continue;
		CExpression* cpl = get_coupled_mom((CESProduct*)item);
		if (cpl == NULL) continue;

		if (m_virt_sym->equals(cpl)) { //expanding k^2/(k^2-m^2) = 1+m^2/(k^2-m^2)
			CSequence* seq0 = (CSequence*)dens_data->GetAt(0);
			if (!CEBasic::isZero(seq0->GetAt(0))) return decl_error(num_prod, dens_data);
			CExpression* m = seq0->GetAt(1);
			
			CExpression* sub_res = NULL;
			if (!CEBasic::isZero(m)) {
				CEProduct* pr_2 = (CEProduct*)num_prod->clone();
				CExpression* m2 = m->sqra();
				pr_2->SetAt(i, m2);// replace k^2 with m^2
				delete m2;
				sub_res = composeVI_impl((CVIDensData*)dens_data->clone(), pr_2, true);
			}
			num_prod->RemoveAt(i);
			dens_data->RemoveAt(0);
			
			CExpression* res = composeVI_impl(dens_data, num_prod, false);
			if (res == NULL) return decl_error();
			if (sub_res != NULL) res = res->add(sub_res);
			return res;
		}

		if (!cpl->getFlag()) {
			// Error: unaranged expression parsing
			dbg->log(0, "Error: unaranged expression parsing.", num_prod); throw(new CEException(CEException::INT_ERROR));
		}
	}

	// no k^2 found in numerator
	// only p.k factors (and powers) and k^i vectors remain.
	CEPtrArray* v_indices = new CEPtrArray();
	CSequence* couplings = new CSequence();
	for (i=0; i<num_prod->size(); i++) {
		CExpression* item = num_prod->GetAt(i);
		if (CEUtil::equals(item->ClassName(), CEVector::CLASS_NAME)) {
			CEVector* vec = (CEVector*)item;
			CExpression* v_base = vec->base();
			if (!v_base->getFlag()) { dbg->log(0, "Error: unaranged expression parsing.", num_prod); delete v_indices; return decl_error(num_prod, dens_data, couplings); }
			if (!m_virt_sym->equals(v_base)) continue;
			v_indices->add(CTIndex::CreateVectorIndex(vec->index(), vec->covariance()));
			num_prod->RemoveAt(i); i--;
			continue;
		}
		if (CEUtil::equals(item->ClassName(), CESProduct::CLASS_NAME)) {
			CExpression* cpl = get_coupled_mom((CESProduct*)item);
			if (cpl == NULL) continue;

			// the SProduct should be already arranged and no sign may occur at "cpl" object.
			couplings->addItem(cpl->clone());
			num_prod->RemoveAt(i);
			i--;
			continue;
		}
		if (CEUtil::equals(item->ClassName(), CEPower::CLASS_NAME)) {
			CEPower* item_pow = (CEPower*)item;
			CExpression* base = item_pow->basis();
			if (CEUtil::equals(base->ClassName(), CESProduct::CLASS_NAME)) {
				CExpression* cpl = get_coupled_mom((CESProduct*)base);
				if (cpl == NULL) continue;
				CExpression* power_expanded = item_pow->as_product();
				num_prod->inline_product(i+1, power_expanded);
				delete power_expanded;
				num_prod->RemoveAt(i);
				i--;
			}
			continue;
		}
	}

	if (v_indices->size() == 0) { delete v_indices; v_indices=NULL; }
	CVirtIntegral* vi_new = createVI(dens_data, couplings, v_indices);
	delete dens_data; delete couplings;
	if (v_indices != NULL) delete v_indices;
	if (vi_new != NULL) num_prod->addItem(vi_new);
	else { //Error
		dbg->log(0, "Error: (CVirtIntegral*) vi_new == NULL");
		return decl_error();
	}
	CExpression* res =  CExpression::reduce(num_prod); // for arranging rest factor
	dbg->log(50, "composeVI_impl(DDATA, Sum) result:", res);
	return res;
}
