////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// VIScalarCalcLib.cpp: implementation of the CVIScalarCalcLib class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CVIScalarCalcLib::dbg = new CDebug("CVIScalarCalcLib");
CECache* CVIScalarCalcLib::m_scalar_VI_cache = NULL;
bool CVIScalarCalcLib::m_init_OK = false;

CExpression* CVIScalarCalcLib::m_XI0_symbol = NULL; // UV-divergent parameter
CExpression* CVIScalarCalcLib::m_IR_symbol = NULL; // IR-divergent parameter
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CVIScalarCalcLib::CVIScalarCalcLib() : CELibrary() {}
CVIScalarCalcLib::~CVIScalarCalcLib() {}

// returns not-cloned value (never NULL)
CExpression* CVIScalarCalcLib::getUVSymbol() {
	if (!m_init_OK) init();
	if (m_XI0_symbol == NULL) CEException::halt(dbg, "NULL UV symbol");
	return m_XI0_symbol;
}

// returns not-cloned value (never NULL)
CExpression* CVIScalarCalcLib::getIRSymbol() {
	if (!m_init_OK) init();
	if (m_IR_symbol == NULL) CEException::halt(dbg, "NULL IR symbol");
	return m_IR_symbol;
}

void CVIScalarCalcLib::init() {
	// init cache
	if (m_scalar_VI_cache != NULL) { delete m_scalar_VI_cache; m_scalar_VI_cache = NULL; }
	m_scalar_VI_cache = new CECache();

	if (m_XI0_symbol == NULL) m_XI0_symbol = new CESymbol("\\[Chi]");
	if (m_IR_symbol == NULL) m_IR_symbol = new CESymbol("X\\+\\(IR\\)");
	m_init_OK = true;
}

void CVIScalarCalcLib::finalize() {
	if (m_scalar_VI_cache != NULL) {
		dbg->log(10, "Scalar VI cache", m_scalar_VI_cache);
		delete m_scalar_VI_cache; m_scalar_VI_cache = NULL;
	}
	if (m_XI0_symbol != NULL) { delete m_XI0_symbol; m_XI0_symbol = NULL; }
	if (m_IR_symbol != NULL) { delete m_IR_symbol; m_IR_symbol = NULL; }
	m_init_OK = false;
}

CECache* CVIScalarCalcLib::GetScalarVICache() { return m_scalar_VI_cache; }

//////////////////////////////////////////////////////////////////////

CExpression* CVIScalarCalcLib::CalcScalarVI(CExpression* src) {
	dbg->log(25, "CalcScalarVI()", src);
	if (src == NULL) return NULL;

	if (!m_init_OK) init();

	//////////////////////////////////////////
	CERule* r = new CESocketRule((void*)CVIScalarCalcLib::CalcScalarVI_ruleImpl);
	src = r->ApplyTo(src);
	delete r;
	//////////////////////////////////////////

	dbg->log(25, "CalcScalarVI() result", src);
	return src;
}

// parse expression for virtual integrals, data_arr is unused
CExpression* CVIScalarCalcLib::CalcScalarVI_ruleImpl(CExpression* src, CEPtrArray* data_arr) {
	if (src == NULL) return src;
	if (!(CEUtil::equals(src->ClassName(), CVirtInt0::CLASS_NAME))) return src;
	dbg->log(45, "CalcScalarVI_ruleImpl()", src);
	
	CExpression* res = CalcVI((CVirtInt0*)src);

	dbg->log(45, "CalcScalarVI_ruleImpl() result", res);
	return res;
}

// apply flags & cache results
CExpression* CVIScalarCalcLib::CalcVI(CVirtInt0* src) {
	if (src == NULL) return NULL;

	if ((!src->getFlag()) || (!src->getLevel())) {
		CVirtInt0* src_truncated = (CVirtInt0*)src->clone();
		src_truncated->clear_all_attributes();
		CExpression* res = CalcVI(src_truncated);
		if (res != src_truncated) delete src_truncated;
		if (res != NULL) src->copyAttributesTo(res, true);
		return res;
	}

	if (!m_init_OK) init(); // extra (useless) check

	// get data from bank
	CExprContainer* val_cart = m_scalar_VI_cache->get_val_container(src);
	if (val_cart != NULL) {
		CExpression* res = val_cart->get_info_expr();
		if (res != NULL) res = res->clone(); // clone result
		dbg->log(75, "cached value found", res);
		return res;
	}

	// calculate integral & cache result
	CExpression* res = GetScalarVIValue(src);
	
	m_scalar_VI_cache->put_value(src, res);
	return res;
}

/*
*	Table values.
*	Note: no sidn/level flag may appear at source expression.
*/
CExpression* CVIScalarCalcLib::GetScalarVIValue(CVirtInt0* src) {
	if (src == NULL) return NULL;

	CVIDensData* dd = src->get_dens_data();
	dbg->log(75, "GetScalarVIValue, dd:", dd);
	CExpression* res = NULL;
	switch(dd->getDensNumber()) {
	case 1:
		res = GetScalarA(dd);
		break;
	case 2:
		res = GetScalarB(dd);
		break;
	case 3:
		res = GetScalarC(dd);
		break;
	case 4:
		res = GetScalarD(dd);
		break;
	default:
		dbg->log(0, "Warning: unexpected scalar integral evaluation", src);
	}
	return res;
}

/*
*	Note: \xi == \alpha-1; \alpha == \xi+1
*
*	Implemented values:
*	A(m) = m^2(x+1-ln(m^2/mu^2))
*	A(0) = 0
*/
CExpression* CVIScalarCalcLib::GetScalarA(CVIDensData* dd) {
	CExpression* m = dd->mass(0);
	CExpression* res = NULL;

	if (!CEBasic::isZero(m)) {
		CExpression* mDR = CDiagramPhysics::get_curr_physics()->get_DR_mass_sym();
		res = m_XI0_symbol->clone()->add(CENumber::GetUnit());
		if (!mDR->equals(m)) {
			CExpression* mass_fr = m->sqra()->divide(mDR->sqra());
			res = res->minus(new CFLogN(mass_fr));
			delete mass_fr;
		}
		res = m->sqra()->multiply(res);
	} else { // zero mass
		res = CENumber::GetZero();
	}
	return res;
}

/*
*	Note: \xi == \alpha-1; \alpha == \xi+1
*
*	Implemented values:
*	B(0,m,m) = x-ln(m^2/mu^2)
*	B(m^2,0,m) = x-ln(m^2/mu^2)+2
*	B(r,0,m) {|r|>>m^2} = x+2-ln(|r|/mu^2)
*	B(m^2,0,M) {m<<M} = x+1-ln(M^2/mu^2)
*	B(r,0,M) {r<0} = x+2-ln(M^2/mu^2)-(1-M^2/r)ln(1-r/M^2)
*	Note: for r>0 result is same if r<M^2, eslwhere result is COMPLEX
*	[Error? Misprint?] B(r,m,m) {r != 0} = x+2-ln(|r|/mu^2)
*	B(r,m,m) {|r|>>m^2} = x+2-ln(|r|/mu^2)
*	B(s,M,M) = x+2+B_sM(s,M)-function, SUBSTITUTION IS CANCELED NOW
*	B(r,M,m) {M>>m, |r|>>m^2, r<0} = x+2-ln(M^2/mu^2)-(1-M^2/r)ln(1-r/M^2)
*	B(m^2,M,m) {M>>m) = x+1-ln(M^2/mu^2)
*	B(r,m1,m2) {|r|>>m1^2, |r|>>m2^2} = x+2-ln(|r|/mu^2)
*/
CExpression* CVIScalarCalcLib::GetScalarB(CVIDensData* dd) {
	if (!dd->isInvariant()) {
		dd = (CVIDensData*)dd->clone(); dd->convertInvariant();
		CExpression* res = GetScalarB(dd); delete dd;
		return res;
	}

	CExpression* m1 = dd->mass(0);
	CExpression* m2 = dd->mass(1);
	CExpression* r = dd->getInvValue(0);
	//CExpression* mDR = CDiagramPhysics::get_curr_physics()->get_DR_mass_sym();
	CExpression* res = NULL;

	//////////////////////////////////////////////////////////////// results table
	// B(0,X,X)
	if (CEBasic::isZero(r)) {
		// B(0,m,m)
		if (m1->equals(m2)) {
			res = GetXiVal(m1);
		}
		return res;
	}
	// B(X,0,X)
	if (CEBasic::isZero(m1)) {
		// B(X,0,0)
		if (CEBasic::isZero(m2)) {
			// r != 0 here
			res = GetXiValR(r)->add(CENumber::GetTwo());
			return res;
		}
		
		CExpression* m2_sqr = CExpression::arrge(m2->sqra());
		// B(m^2,0,m)
		if (r->equals(m2_sqr)) {
			res = GetXiVal(m2)->add(CENumber::GetTwo());
			delete m2_sqr; return res;
		}

		// B(t,0,m), B(s,0,m)
		if (CLHEPBasics::IsAGGthanB_abs(r, m2_sqr)) {
			res = GetXiValR(r)->add(CENumber::GetTwo());
			delete m2_sqr; return res;
		}

		// B(m^2,0,M), M^2>>m^2
		//if ((CLHEPBasics::GetExprSign(r) > 0) && (CLHEPBasics::IsAGGthanB_abs(m2_sqr, r))) {
		if (CLHEPBasics::IsAGGthanB_abs(m2_sqr, r)) { // TODO: check also (CLHEPBasics::GetExprSign(r) > 0)
			res = GetXiVal(m2)->add(CENumber::GetUnit());
			delete m2_sqr; return res;
		}

		// B(X,0,M)
		if (CLHEPBasics::GetExprSign(r) < 0) { // r < 0, B(t,0,M)
			res = GetXiVal(m2)->add(CENumber::GetTwo());
			
			CExpression* tmp1 = (CENumber::GetUnit())->minus(m2->sqra()->divide(r->clone()));
			CExpression* tmp2 = (CENumber::GetUnit())->minus(r->clone()->divide(m2->sqra()));
			res = res->minus(tmp1->multiply(new CFLogN(tmp2))); delete tmp2;
			delete m2_sqr; return res;
		} else if (CLHEPBasics::GetExprSign(r) > 0) { // B(s,0,M)



		} // undefined sgin elsewhere

		delete m2_sqr; return res;
	}

	// B(X,X,0)
	if (CEBasic::isZero(m2)) {
		// note: normalized B-integral should have {0,m} denominator mass order
		dbg->log(0, "Not normalized B0 integral spotted", dd);
		return res;
	}

	// B(X,m,m)
	if (m1->equals(m2)) {
		CExpression* m2_sqr = CExpression::arrge(m2->sqra());

		// B(t,m,m), B(s,m,m)
		if (CLHEPBasics::IsAGGthanB_abs(r, m2_sqr)) {
			res = GetXiValR(r)->add(CENumber::GetTwo());
		} else {
			/*
			CExpression* tmp_zero = CENumber::GetZero();
			if (CLHEPBasics::IsAGGthanB_abs(r, tmp_zero)) {
				res = GetXiVal(m1)->add(CENumber::GetTwo());
				// & add abstract behaviour function, see p.5, red book
				CEPtrArray* tmp_data_arr = new CEPtrArray();
				tmp_data_arr->add(r->clone());
				tmp_data_arr->add(m1->clone());
				res = res->add(new CEFunction("B\\_\\(sM\\)",tmp_data_arr));
				delete tmp_data_arr;
			}
			delete tmp_zero;
			*/
		}
		delete m2_sqr; return res;
	}

	// B(r,M,m) or B(r,m,M): M>>m
	CExpression* gg_mass = NULL; // mass that >> that other (ll_mass)
	CExpression* ll_mass = NULL;
	if (CLHEPBasics::IsAGGthanB_abs(m1, m2)) {
		gg_mass = m1; ll_mass = m2;
	} else if (CLHEPBasics::IsAGGthanB_abs(m2, m1)) {
		gg_mass = m2; ll_mass = m1;
	}

	if ((gg_mass != NULL) && (ll_mass != NULL)) {
		CExpression* mll_sqr = CExpression::arrge(ll_mass->sqra());
		//dbg->log(2, "gg_mass", gg_mass); dbg->log(2, "ll_mass", ll_mass); dbg->log(2, "mll_sqr", mll_sqr); dbg->log(2, "r", r);

		// B(r,M,m): |r|>>m^2
		if (CLHEPBasics::IsAGGthanB_abs(r, mll_sqr)) {
			// B(t,M,m)
			if (CLHEPBasics::GetExprSign(r) < 0) { // r < 0
				res = GetXiVal(gg_mass)->add(CENumber::GetTwo());

				CExpression* tmp1 = (CENumber::GetUnit())->minus(gg_mass->sqra()->divide(r->clone()));
				CExpression* tmp2 = (CENumber::GetUnit())->minus(r->clone()->divide(gg_mass->sqra()));
				res = res->minus(tmp1->multiply(new CFLogN(tmp2))); delete tmp2;
			}
			delete mll_sqr; return res;
		}

		// B(m^2,M,m)
		if (r->equals(mll_sqr)) {
			res = GetXiVal(gg_mass)->add(CENumber::GetUnit());
			delete mll_sqr; return res;
		}
		delete mll_sqr; return res;
	}

	CExpression* m1_sqr = CExpression::arrge(m1->sqra());
	CExpression* m2_sqr = CExpression::arrge(m2->sqra());
	if ((CLHEPBasics::IsAGGthanB_abs(r,m1_sqr)) && (CLHEPBasics::IsAGGthanB_abs(r,m2_sqr))) {
		// B(r,m1,m2) {|r|>>m1^2, |r|>>m2^2}
		if (CLHEPBasics::GetExprSign(r) < 0) { // r < 0
			res = GetXiValR(r)->add(CENumber::GetTwo());
			delete m1_sqr; delete m2_sqr; return res;
		}
	}
	delete m1_sqr; delete m2_sqr;

	// indefined result
	return res;
}

CExpression* CVIScalarCalcLib::GetScalarC(CVIDensData* dd) {
	if (!dd->isInvariant()) {
		dd = (CVIDensData*)dd->clone(); dd->convertInvariant();
		CExpression* res = GetScalarC(dd); delete dd;
		return res;
	}

	CExpression* m1 = dd->mass(0);
	CExpression* m2 = dd->mass(1);
	CExpression* m3 = dd->mass(2);
	CExpression* r1 = dd->getInvValue(0);
	CExpression* r2 = dd->getInvValue(1);
	CExpression* r3 = dd->getInvValue(2);
	CExpression* mDR = CDiagramPhysics::get_curr_physics()->get_DR_mass_sym();
	CExpression* res = NULL;

	/////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////
	return res;
}

CExpression* CVIScalarCalcLib::GetScalarD(CVIDensData* dd) {
	if (!dd->isInvariant()) {
		dd = (CVIDensData*)dd->clone(); dd->convertInvariant();
		CExpression* res = GetScalarD(dd); delete dd;
		return res;
	}

	CExpression* m1 = dd->mass(0);
	CExpression* m2 = dd->mass(1);
	CExpression* m3 = dd->mass(2);
	CExpression* m4 = dd->mass(3);
	CExpression* r1 = dd->getInvValue(0);
	CExpression* r2 = dd->getInvValue(1);
	CExpression* r3 = dd->getInvValue(2);
	CExpression* rr1 = dd->getInvValue(3);
	CExpression* rr2 = dd->getInvValue(4);
	CExpression* rr3 = dd->getInvValue(4);
	CExpression* mDR = CDiagramPhysics::get_curr_physics()->get_DR_mass_sym();
	CExpression* res = NULL;

	/////////////////////////////////////////////////////////////////////////////
	if ((CEBasic::isZero(m1)) && (!CEBasic::isZero(m2)) && (m2->equals(m3)) && (m2->equals(m4))) {
		CExpression* mSqr = m2->sqra();
		if ((r1->equals(mSqr)) && (r2->equals(mSqr)) &&
			(!CEBasic::isZero(r3)) && (!CEBasic::isZero(rr1)) &&
			(CLHEPBasics::IsAGGthanB_abs(r3, mSqr)) && (CLHEPBasics::IsAGGthanB_abs(rr1, mSqr))) {
				CExpression* tHlf = CExpression::reduce(r3->clone()->divide(CENumber::GetTwo()));
			if ((tHlf->equals(rr2)) && (tHlf->equals(rr3)) &&
				(CLHEPBasics::GetExprSign(r3) < 0) && (CLHEPBasics::GetExprSign(rr1) < 0)) {
				CExpression* s_inv = rr1->clone(); s_inv->inverseFlag();
				s_inv = CExpression::reduce(s_inv->multiply(CENumber::GetTwo()));
				res = GetDChiralFormula(m2, r3, s_inv);
				delete s_inv;
			}
			delete tHlf;
		}
		delete mSqr;
	}
	/////////////////////////////////////////////////////////////////////////////
	return res;
}

// returns Xi value with DR-mass contribution like ln(m_SQR/DR_mass_SQR)
CExpression* CVIScalarCalcLib::GetXiVal(CExpression* m) {
	if (m == NULL) CEException::halt(dbg, "CVIScalarCalcLib::GetXiVal(NULL)");

	CExpression* res = m_XI0_symbol->clone();
	CExpression* mDR = CDiagramPhysics::get_curr_physics()->get_DR_mass_sym();
	if (!m->equals(mDR)) {
		CExpression* fr = (m->sqra())->divide(mDR->sqra());
		res = res->minus(new CFLogN(fr)); delete fr;
	}
	return res;
}

// returns Xi value with DR-mass contribution like ln(r/DR_mass_SQR)
CExpression* CVIScalarCalcLib::GetXiValR(CExpression* r) {
	if (r == NULL) CEException::halt(dbg, "CVIScalarCalcLib::GetXiValR(NULL)");

	CExpression* res = m_XI0_symbol->clone();
	CExpression* mDR = CDiagramPhysics::get_curr_physics()->get_DR_mass_sym();

	CExpression* r_cpy = r->clone();
	// TODO: truncate sing at r_cpy here		!!!!!!!!!!!
	CExpression* fr = r_cpy->divide(mDR->sqra());
	res = res->minus(new CFLogN(fr)); delete fr;
	return res;
}

// returns decomposition for D(m^2, m^2, t, -s/2, t/2, t/2, 0, m, m, m) integral
CExpression* CVIScalarCalcLib::GetDChiralFormula(CExpression* m, CExpression* t, CExpression* s) {
	CExpression* m2 = m->sqra();
	CExpression* sd2 = CExpression::reduce(s->clone()->divide(CENumber::GetTwo()));
	CExpression* msd2 = sd2->clone(); msd2->inverseFlag();
	CExpression* td2 = CExpression::reduce(t->clone()->divide(CENumber::GetTwo()));
	CSequence* dd_seq = NULL;

	// 1. C0(m^2, m^2, -s/2, 0, m, m) (\lambda-divergent)
	dd_seq = new CSequence();
	dd_seq->addItem(m2->clone());
	dd_seq->addItem(m2->clone());
	dd_seq->addItem(msd2->clone());
	dd_seq->addItem(CENumber::GetZero());
	dd_seq->addItem(m->clone());
	dd_seq->addItem(m->clone());
	CVIDensData* dd = new CVIDensData(dd_seq, 3);
	delete dd_seq;
	CExpression* C_lambda = CVirtInt0::createInvVI(dd);
	delete dd;

	// 2. C0(0, s, s/2, m, m, m) (\lambda-divergent)
	// (the same to) C0(0, 0, -s/2, m, m, m) (\lambda-divergent)
	dd_seq = new CSequence();
	dd_seq->addItem(CENumber::GetZero());
	dd_seq->addItem(s->clone());
	dd_seq->addItem(sd2->clone());
	dd_seq->addItem(m->clone());
	dd_seq->addItem(m->clone());
	dd_seq->addItem(m->clone());
	dd = new CVIDensData(dd_seq, 3);
	delete dd_seq;
	CExpression* C_s = CVirtInt0::createInvVI(dd);
	delete dd;

	// 1. C0(m^2, t, t/2, 0, m, m) (\lambda-divergent)
	dd_seq = new CSequence();
	dd_seq->addItem(m2->clone());
	dd_seq->addItem(t->clone());
	dd_seq->addItem(td2->clone());
	dd_seq->addItem(CENumber::GetZero());
	dd_seq->addItem(m->clone());
	dd_seq->addItem(m->clone());
	dd = new CVIDensData(dd_seq, 3);
	delete dd_seq;
	CExpression* C_t = CVirtInt0::createInvVI(dd);
	delete dd;

	CExpression* res = (C_lambda->add(C_s))->divide(t->clone());
	res = res->add(C_t->multiply(CENumber::GetTwo())->divide(s->clone()));
	CExpression* l2ts_fr = t->clone()->divide(s->clone()); l2ts_fr->inverseFlag();
	CExpression* l2ts = new CFLogN(l2ts_fr); delete l2ts_fr;
	res = res->minus(l2ts->divide(s->clone()->multiply(t->clone())));

	delete m2;
	delete sd2;
	delete msd2;
	delete td2;

	return res;
}
