////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// VirtIntC.cpp: implementation of the CVirtIntC class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const char* CVirtIntC::CLASS_NAME		= "CVirtIntC";
const char* CVirtIntC::CLASS_TAG_NAME	= "viC";
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CVirtIntC::CVirtIntC(const CVIDensData* ddata, CSequence* _couplings, CEPtrArray* _indices)
:CVirtIntegral(ddata)//,IETensor() // toString() & equals() will be overridden
{
	if (_couplings != NULL) {
		m_couplings = (CSequence*)_couplings->clone();
	} else {
		m_couplings = new CSequence();
	}

	//m_base = NULL;
	if (_indices != NULL) {
		if (_indices->size() > 0) m_indices=_indices->clone();
		else m_indices=NULL;
	} else {
		m_indices = NULL; //new CEPtrArray();
	}

	if ((m_couplings->size() == 0) && (m_indices == NULL)) {
		CVirtIntegral::dbg->log(0, "Error: CVirtIntC(DD, NULL or 0-size couplings & NULL indices)."); CEException::halt();
	}
}

CVirtIntC::~CVirtIntC() {
	delete m_couplings;
	if (m_indices != NULL) { delete m_indices; m_indices=NULL; }
	//if (m_base != NULL) { delete m_base; m_base=NULL; }
}

CExpression* CVirtIntC::clone() const {
	CExpression* res = new CVirtIntC(get_dens_data_const(), m_couplings, m_indices);
	copyAttributesTo(res);
	return res;
}

CSequence* CVirtIntC::GetCouplings() { return m_couplings; }

CEString* CVirtIntC::toString(int mode) {
	CEString* res = NULL;
	if (getFlag()) res = new CEString("\\(");
	else res = new CEString("- \\(");

	int r = rank();
	if (r>0) res->AddString("\\(");

	switch (get_dens_data()->size())
	{
	case 1: res->AddString("A"); break;
	case 2: res->AddString("B"); break;
	case 3: res->AddString("C"); break;
	case 4: res->AddString("D"); break;
	case 5: res->AddString("E"); break;
	case 6: res->AddString("F"); break;
	default:
		res->AddString("X");
	}

	res->AddString("\\+");
	res->AddString(m_couplings->toString(mode));
	res->AddString("\\)");

	if (r > 0) {
		bool last_covariance;
		int i;
		for(i=0; i<r; i++) {
			if (i==0) {
				last_covariance = covariance(1);
				res->AddString((last_covariance)?"\\_":"\\&");
				res->AddString("\\(");
				res->AddString(index(1)->toString(mode));
				res->AddString("\\)");
			}
			else {
				bool cov = covariance(i+1);
				if (cov == last_covariance) {
					long ins_pos = res->length()-2;//insert before bracket closed
					res->insertString(index(i+1)->toString(mode), ins_pos);
					res->insertString("\\ ", ins_pos);
				} else {
					res->insertString("\\(", 0);
					res->AddString("\\)");
					res->AddString((cov)?"\\_":"\\&");
					res->AddString("\\(");
					res->AddString(index(i+1)->toString(mode));
					res->AddString("\\)");
					last_covariance = cov;
				}
			}
		}
		res->AddString("\\)");
	}

	//res->AddString("[");
	res->AddString("\\&");
	res->AddString(get_dens_data()->toString(mode));
	//res->AddString("]");
	return res;
}

CExpression* CVirtIntC::applyRule(CEAbstractRule* rule) {
	CExpression* dd_apply = CVirtIntegral::applyRule(rule);
	if(dd_apply != this) CEException::halt("CVirtIntC::applyRule internal ERROR");

	CERule* r = (CERule*)rule;
	if (m_couplings != NULL) {
		CExpression* cpl_new = r->ApplyTo(m_couplings, false);
		if ((cpl_new != m_couplings) && (cpl_new != NULL)) {
			if (CEUtil::equals(cpl_new->ClassName(), CSequence::CLASS_NAME)) {
				delete m_couplings; m_couplings=(CSequence*)cpl_new;
			} else CVirtIntegral::dbg->log(0, "CVirtIntC::applyRule ERROR: rule->ApplyTo(m_couplings) != CSequence");
		}
	}

	int rnk=rank();
	if (rnk>0) {
		int i;
		for(i=1; i<=rnk; i++) {
			CExpression* item = index(i);
			CExpression* item_repl = r->ApplyTo(item, false);
			if ((item_repl != item) && (item_repl != NULL)) {
				setIndex(i, item_repl);
				delete item_repl;
			}
		}
	}
	return this;
}

CExpression* CVirtIntC::evaluate(int level) {
	CVirtIntegral::dbg->log(25, "evaluate", this);
	CExpression* res = CVirtIntegralLib::get_stored(this);
	if (res != NULL) {
		CVirtIntegral::dbg->log(25, "found stored value", res);
		return res;
	}
	
	// A_x*p^x --> to Zero
	int c_num = m_couplings->size();
	int d_num = get_dens_data()->getDensNumber();
	if ((d_num == 1) && (c_num == 1)) {
		res = CENumber::GetZero();
	}

	if (res == NULL) {
		res = tensor_reduction();
	}

	if (res != NULL) {
		res = CLHEPBasics::KinArrange(res);
		CVirtIntegral::dbg->log(25, "tensor reduction performed", res);
		CVirtIntegralLib::store_vi(this, res);
		CVirtIntegral::dbg->log(45, "stored");
		return res;
	}
	return this;
}

CExpression* CVirtIntC::expand(bool recursive) {
#if defined(D0_ONLY_CALC)
	if (get_dens_data()->size() < 4) return CENumber::GetZero();
#endif //D0_ONLY_CALC

	CVirtIntegral::dbg->log(30, "expanding", this);
	//mormalizing should be here!
	CExpression* dens_normd = normalize_denominators();
	if (dens_normd != NULL) {
		CVirtIntegral::dbg->log(35, "expanded (denominators normalized)", dens_normd);
		return dens_normd;
	}
	CVirtIntegral::dbg->log(35, "expanded", this);
	return this;
}

CExpression* CVirtIntC::Arrange(int level) {
#if defined(D0_ONLY_CALC)
	if ((level == AL_EXTRACT) && (get_dens_data()->size() < 4)) return CENumber::GetZero();
#endif //D0_ONLY_CALC

	get_dens_data()->sort();

	int cpl_no = m_couplings->size();
	for(int cpl_i=0; cpl_i<cpl_no; cpl_i++) {
		CExpression* item = m_couplings->GetAt(cpl_i);
		if (!item->getFlag()) { item->inverseFlag(); inverseFlag(); }
	}

	return this;
}

bool CVirtIntC::contain_vector(void * obj) {
	if (CVirtIntegral::contain_vector(obj)) return true;
	if (m_couplings != NULL) {
		if (m_couplings == obj) return true;
		if (m_couplings->contain_vector(obj)) return true;
	}

	if (m_indices != NULL) {
		if (m_indices == obj) return true;
		if (m_indices->contain_vector(obj)) return true;
		/*
		int s = rank();
		int i;
		for(i=0; i<s; i++) {
			CTIndex* ind = (CTIndex*)m_indices->GetAt(i);
		if ((ind == obj) || (ind->contain_vector(obj))) return true;
		*/
	}
	return false;
}

const char* CVirtIntC::ClassName() const { return CLASS_NAME; }
const char* CVirtIntC::get_TAG_Name() { return CLASS_TAG_NAME; }

long CVirtIntC::store_stream(ofstream *p_ofs, DWORD settings) {
	long res = CESupport::store_stream(get_dens_data(), p_ofs, settings);
	res += CESupport::store_stream(m_couplings, p_ofs, settings);

	if (m_indices != NULL) {
		int r = m_indices->size();
		CENumber* num = new CENumber(r);
		res += CESupport::store_stream(num, p_ofs, settings);
		delete num;
		for (int i=0; i<r; i++) {
			CTIndex* ti = (CTIndex*)m_indices->GetAt(i);
			CExpression* ind = ti->index();
			int i_cov = (ti->covariance())?1:0;
			num = new CENumber(i_cov);
			res += CESupport::store_stream(ind, p_ofs, settings);
			res += CESupport::store_stream(num, p_ofs, settings);
			delete num;
		}
	}
	return res;
}

CESerializable* CVirtIntC::restore_from_buffer(const char* data, long pos, CEMap* params, DWORD settings) {
	// load DENSDATA
	long skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { CVirtIntegral::dbg->log(0, "Error restoring CVirtIntC : Can't load DENSDATA."); return NULL; }
	CESerializable* dd = CESupport::read_buffer(data, pos, settings);
	if ((dd == NULL) || (!CEUtil::equals(dd->ClassName(), CVIDensData::CLASS_NAME))) {
		CVirtIntegral::dbg->log(0, "Error loading DENSDATA for CVirtIntC:", dd);
		if (dd != NULL) delete dd;
		return NULL;
	}
	CVIDensData* viDD = (CVIDensData*)dd;

	// load couplings
	pos = skip_pos;
	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos <= 0) { CVirtIntegral::dbg->log(0, "Error restoring CVirtIntC : Can't load couplings sequence."); return NULL; }
	CESerializable* coupl_void = CESupport::read_buffer(data, pos, settings);
	if ((coupl_void == NULL) || (!CEUtil::equals(coupl_void->ClassName(), CSequence::CLASS_NAME))) {
		CVirtIntegral::dbg->log(0, "Error loading coupling sequence for CVirtIntC.");
		delete viDD;
		if (coupl_void != NULL) delete coupl_void;
		return NULL;
	}
	CSequence* coupl_seq = (CSequence*)coupl_void;

	// check up sequence size
	if (coupl_seq->size() == 0) {
		CVirtIntegral::dbg->log(0, "Error: empty couplings settings loaded for CVirtIntC.");
		delete viDD; delete coupl_seq; return NULL;
	}
	pos = skip_pos;

	long r=0;
	if (!CESupport::readXMLLong(data, &pos, &r, settings, CLASS_NAME, true)) {
		//CVirtIntegral::dbg->log(0, "WARNING restoring CVirtIntC : OLD notation spotted");
		r=0;
	}

	CEPtrArray* _indices = NULL;
	if (r>0) {
		_indices = new CEPtrArray();
		CExpression* ind = NULL;
		long cov=0;
		for (int i=0; i<r; i++) {
			ind = CESupport::readXMLExpr(data, &pos, settings, CLASS_NAME);
			if (ind == NULL) { CVirtIntegral::dbg->log(0, "Error restoring indices."); delete viDD; delete coupl_seq; delete _indices; return NULL;}
			if (!CESupport::readXMLLong(data, &pos, &cov, settings, CLASS_NAME)) {
				CVirtIntegral::dbg->log(0, "Error restoring indices."); delete viDD; delete coupl_seq; delete _indices; return NULL;
			}
			_indices->add(CTIndex::CreateVectorIndex(ind, (cov>0)));
			delete ind;
		}
	}

	skip_pos = CESupport::skip_item_in_buffer(data, pos, settings);
	if (skip_pos > 0) { CVirtIntegral::dbg->log(0, "Error restoring CVirtIntC: Extra data found."); delete viDD; delete coupl_seq; return NULL; }

	CVirtIntC* res = new CVirtIntC(viDD, coupl_seq, _indices);
	delete viDD; delete coupl_seq;
	if (_indices != NULL) delete _indices;
	return res;
}

bool CVirtIntC::equals(CExpression* other) {
	if (!CEUtil::equals(ClassName(), other->ClassName())) return false;
	if (!attributes_equals_to(other)) return false;

	CVirtIntC* o = (CVirtIntC*)other;

	if (!get_dens_data()->equals(o->get_dens_data())) return false;
	if (!m_couplings->equals(o->GetCouplings())) return false;

	if (m_indices != NULL) {
		int r = rank();
		if (r != o->rank()) return false;
		for(int i=1; i<=r; i++) {
			CTIndex* ti1=GetTIndex(i); CTIndex* ti2=o->GetTIndex(i);
			if(ti1!=NULL){ if(!ti1->equalsIndex(ti2))return false; }else{ if(ti2!=NULL)return false; }
		}
	} else {
		if (o->rank() > 0) return false;
	}

	return true;
}

///////////////////////////////////////// TENSOR implementation <BEGIN>
void CVirtIntC::collectProperties(CEProperties* props, unsigned set) {
	CVirtIntegral::collectProperties(props, set);
	if (rank() > 0) props->setTensor();
}

unsigned long CVirtIntC::props_long() {
	unsigned long res = CVirtIntegral::props_long();
	if (rank() > 0) { res = CEProperties::set_property_to(res, CEProperties::PROP_MASK_IS_TENSOR, true); }
	return res;
}

CExpression* CVirtIntC::index(int i) {
	CTIndex* tindex = ((m_indices!=NULL)? (CTIndex*)m_indices->GetAt(i-1) : NULL);
	return ((tindex != NULL) ? tindex->index() : NULL);
}
const CExpression* CVirtIntC::GetIndex(int i) const {
	const CTIndex* tindex = ((m_indices!=NULL) ? (const CTIndex*)m_indices->GetAtConst(i-1) : NULL);
	return ((tindex != NULL) ? tindex->GetIndex() : NULL);
}
CTIndex* CVirtIntC::GetTIndex(int i) { return ((m_indices!=NULL)?(CTIndex*)m_indices->GetAt(i-1) : NULL); }
const CTIndex* CVirtIntC::GetTIndexConst(int i) const { return ((m_indices!=NULL)?(const CTIndex*)m_indices->GetAtConst(i-1) : NULL); }

bool CVirtIntC::covariance(int i) const {
	if (m_indices == NULL) { dbg->log(0, "CVirtIntC::covariance() Invalid request, NULL indices list"); return false; }
	const CTIndex* tindex = (const CTIndex*)m_indices->GetAtConst(i-1);
	if (tindex != NULL) return tindex->covariance();
	dbg->log(0, "CVirtIntC::covariance() Invalid request");
	return false;
}

bool CVirtIntC::setIndex(int i, CExpression* index) {
	if (m_indices == NULL) { dbg->log(0, "CVirtIntC::setIndex() Invalid request, NULL indices list", this); return false; }
	CTIndex* tindex = (CTIndex*)m_indices->GetAt(i-1);
	if (tindex != NULL) return tindex->setIndex(index);
	dbg->log(0, "CVirtIntC::setIndex() Invalid request", this);
	return false;
}

bool CVirtIntC::liftIndex(int i, bool to_covariance) {
	if (m_indices == NULL) { dbg->log(0, "CVirtIntC::liftIndex() Invalid request, NULL indices list", this); return false; }
	CTIndex* tindex = (CTIndex*)m_indices->GetAt(i-1);
	if (tindex != NULL) return tindex->liftIndex(to_covariance);
	dbg->log(0, "CVirtIntC::liftIndex() Invalid request", this);
	return false;
}

bool CVirtIntC::changeIndex(int i, CExpression *index, bool covariance) {
	if (m_indices == NULL) { dbg->log(0, "CVirtIntC::changeIndex() Invalid request, NULL indices list", this); return false; }
	CTIndex* tindex = (CTIndex*)m_indices->GetAt(i-1);
	if (tindex != NULL) return tindex->changeIndex(index, covariance);
	dbg->log(0, "CVirtIntC::changeIndex() Invalid request, NULL indices list", this);
	return false;
}

CExpression* CVirtIntC::ThisExpr() { return (CExpression*)this; }
void* CVirtIntC::GetInterface(long interface_ID) {
	if (interface_ID == CConst::dwIETensor) return (IETensor*)this;
	return CExpression::GetInterface(interface_ID);
}

CExpression* CVirtIntC::base() { return NULL; }
const CExpression* CVirtIntC::GetBase() const { return NULL; }
CExpression* CVirtIntC::unite_with(IETensor *other, bool other_is_at_right, bool bCalledByOther) { return NULL; }
int CVirtIntC::rank() const { return (m_indices != NULL)?(m_indices->size()):0; }

CExpression* CVirtIntC::exchange_subst(IETensor* other, bool other_is_at_right, bool bCalledByOther) {
	CExpression* res = NULL;
	// VI-C is commutative, hence:
	if(other_is_at_right){ res=other->ThisExpr()->clone()->multiply(clone()); }else{ res = clone()->multiply(other->ThisExpr()->clone()); }
	return res;
}

bool CVirtIntC::setBase(CExpression *_new) { dbg->log(0, "WARNING: CVirtIntC::setBase() call", this, _new); return false; } // nothing to do here
///////////////////////////////////////// TENSOR implementation <END>

/*
*	Note: All items in m_couplings are supposed to be positive.
*/
CExpression* CVirtIntC::tensor_reduction() {
	CVIDensData* dd = get_dens_data();

	// TODO: remove next check if no promblems occured
	bool pre_arrange = false; // signed coplings are originated from VI1,2-evaluation engine (right part of coeff. matrix)
	for(int cpl_i=0; cpl_i<m_couplings->size(); cpl_i++) { if (!m_couplings->GetAt(cpl_i)->getFlag()) { pre_arrange=true; break; } }
	if (pre_arrange) Arrange(AL_INLINE);

	//1. No k^2 may appear in numerator

	bool bExpandBasisOnly=true;
	CExpression* res=NULL;

	if(!bExpandBasisOnly){
		//2. pk/(p,k) reduction
		res=tens_reduce_similar();

		//3. Combined pk processing
		if (res == NULL) res = tens_reduce_combines();
		if (res != NULL) {
			CVirtIntegral::dbg->log(25, "first-step tensor reduction", res);
			//normalize
			res = CExpression::reduce(res);
			//further reduce
			CExpression* res_evl = res->evaluate();
			if (res_evl != res) delete res;
			return res_evl;
		}
	}

	res = expand_to_vi_basis();
	if (res != NULL) {
		CVirtIntegral::dbg->log(25, "second-step tensor reduction", res);
		res = CExpression::reduce(res);
		//further reduce
		CExpression* res_evl = res->evaluate();
		if (res_evl != res) delete res;
		return res_evl;
	}

	return NULL;
}

/*
*	pk/(p,k) reduction
*	?????	Note: All items in m_couplings are supposed to be positive.
*	DensData supposed to be normalized (i.e. p_0 == 0)
*/
CExpression* CVirtIntC::tens_reduce_similar()
{
	CVirtIntegral::dbg->log(35, "TR: tens_reduce_similar()", this);
	CVIDensData* dd = get_dens_data();
	int c_s = m_couplings->size();
	int i;
	for (i=0; i<c_s; i++) {
		CExpression* mom = m_couplings->GetAt(i);
		int s = dd->size();
		int j;
		for (j=1; j<s; j++) {
			CExpression* dmom = dd->momentum(j);
			if (mom->equals(dmom)) {
				return tens_reduce_similar(i, j, true);
			}

			if (mom->is_opposite_to(dmom)) {
				return tens_reduce_similar(i, j, false);
			}
		}
	}
	CVirtIntegral::dbg->log(35, "TR: no similar pairs found", this);
	return NULL;
}

/*
*	Replaces previously found pair
*	j -- pos in denominator, i -- in numerator
*/
CExpression* CVirtIntC::tens_reduce_similar(int i, int j, bool sign)
{
	CExpression* res = NULL;
	CVIDensData* dd = get_dens_data();
	//int c_s = m_couplings->size();
	CExpression* mom = dd->momentum(j);
	if (!CEBasic::isZero(dd->momentum(0))) { CVirtIntegral::dbg->log(0, "<ERROR> CVirtIntC::tens_reduce_similar: first denominator not reduced (k2-m2)"); CEException::halt(); }

	CESum* res_sum = new CESum();
	CExpression* temp = CEVector::square(mom);
	temp = CExpression::reduce(dd->mass(j)->sqra()->minus(dd->mass(0)->sqra())->minus(temp));

	CSequence* seq_new = (CSequence*)m_couplings->clone();
	seq_new->RemoveAt(i);

	CVirtIntegral* vi = NULL;
	if (!CEBasic::isZero(temp)) {
		vi = CVIComposeLib::createVI(dd, seq_new, m_indices);
		if (vi != NULL) res_sum->addItem(temp->multiply(vi));
		else { CEException::halt(CVirtIntegral::dbg, "Error: vi = NULL (1)"); }
	} else { delete temp; }
	
	CVIDensData* dd_new = (CVIDensData*)dd->clone();
	dd_new->RemoveAt(0);
	vi = CVIComposeLib::createVI(dd_new, seq_new, m_indices); delete dd_new;
	if (vi != NULL) {
		vi->inverseFlag();
		res_sum->addItem(vi);
	} else { CEException::halt(CVirtIntegral::dbg, "Error: vi = NULL (2)"); }
	
	dd_new = (CVIDensData*)dd->clone();
	dd_new->RemoveAt(j);
	vi = CVIComposeLib::createVI(dd_new, seq_new, m_indices); delete dd_new;
	if (vi != NULL) res_sum->addItem(vi);
	else { CEException::halt(CVirtIntegral::dbg, "Error: vi = NULL (3)"); }
	res = res_sum;
	
	res = res->multiply(CENumber::GetHalf());
	copyAttributesTo(res, true);
	if (!sign) res->inverseFlag();
	delete seq_new;
	return res;
}

/*
*	returns new expression except of old VI-C.
*	NULL if not changed
*/
CExpression* CVirtIntC::normalize_denominators()
{
	CVirtIntegral::dbg->log(20, "normalize_denominators()", this);
	CVIDensData* dd = get_dens_data();
	dd->sort();

	//No q->q+x shifts for LoopTools-compartible mode
	if(CVirtIntegral::PAR_LOOPTOOLS_COMPARTIBLE)return NULL;

	CExpression* p_del = dd->get_norm_p_del();
	if (p_del==NULL) return NULL;

	int c_s = m_couplings->size();
	int i_s = (m_indices != NULL)?(m_indices->size()):0;
	int vi_s = c_s + i_s;
	if (vi_s == 0) { CVirtIntegral::dbg->log(0, "ERROR: VI-C with 0-size couplings & indices", this); CEException::halt(); }
	// vi_s MAY NOT be 0
	if (vi_s > 2) {
		CVirtIntegral::dbg->log(10, "Warning: denominators normalizing not implemented for m_couplings->size()>2", m_couplings);
		return NULL;
	}
	p_del = p_del->clone();
	dd = (CVIDensData*)dd->clone();
	dd->change_var_minus_pdel(p_del);

	CVirtIntegral::dbg->log(35, "normalize_denominators(): dd new", dd);
	CVirtIntegral::dbg->log(35, "normalize_denominators() p_del", p_del);

	CExpression* res = NULL;
	//binomials should be here
	CVirtIntegral* vi = NULL;
	if (vi_s == 1) {
		res = CVIComposeLib::createVI(dd, m_couplings, m_indices);
		if (res == NULL) { CEException::halt(); }
		vi = CVIComposeLib::createVI(dd, NULL, NULL);
		if (vi == NULL) { CEException::halt(); }
		res = res->minus( GetExpandCpl(p_del, 0)->multiply(vi) );
	}
	else if (vi_s == 2) {
		res = CVIComposeLib::createVI(dd, m_couplings, m_indices);
		if (res == NULL) { CEException::halt(); }

		CSequence* seq_new = (CSequence*)m_couplings->clone();
		CEPtrArray* ind_new=((m_indices!=NULL)?(m_indices->clone()):NULL);
		XCplRemoveAt(seq_new, ind_new, 0);
		vi = CVIComposeLib::createVI(dd, seq_new, ind_new); delete seq_new; if(ind_new!=NULL){ delete ind_new; }
		if (vi == NULL) { CEException::halt(); }
		res = res->minus( GetExpandCpl(p_del, 0)->multiply(vi) );

		seq_new = (CSequence*)m_couplings->clone();
		ind_new=((m_indices!=NULL)?(m_indices->clone()):NULL);
		XCplRemoveAt(seq_new, ind_new, 1);
		vi = CVIComposeLib::createVI(dd, seq_new, ind_new); delete seq_new; if(ind_new!=NULL){ delete ind_new; }
		if (vi == NULL) { CEException::halt(); }
		res = res->minus( GetExpandCpl(p_del, 1)->multiply(vi) );

		vi = CVIComposeLib::createVI(dd, NULL, NULL);
		if (vi == NULL) { CEException::halt(); }
		res = res->add( GetExpandCpl(p_del, 0)->multiply(GetExpandCpl(p_del, 1))->multiply(vi));
	}
	else {
		CVirtIntegral::dbg->log(0, vi_s, "Error: unexpected 'vi_s': %ld", this);
		delete dd;
		delete p_del;
		return NULL;
	}
	delete p_del; delete dd;
	copyAttributesTo(res, true);

	CVirtIntegral::dbg->log(20, "normalize_denominators() composed", res);
	res = CExpression::reduce(res);
	return res;
}

CExpression* CVirtIntC::tens_reduce_combines()
{
	CVirtIntegral::dbg->log(35, "TR: tens_reduce_combines()", this);
	int c_s = m_couplings->size();
	if (c_s == 0) return NULL;
	int i;
	for (i=0; i<c_s; i++) {
		//p_0 == 0 (normalized) ==> start from 1'st momentum in denominator
		CExpression* res = tens_reduce_combines(1, i, NULL);
		if (res != NULL) return res;
	}
	CVirtIntegral::dbg->log(35, "TR: no combines found", this);
	return NULL;
}

CExpression* CVirtIntC::tens_reduce_combines(int d_pos, int n_pos, CExpression* sum)
{
	CVIDensData* dd = get_dens_data();
	int s = dd->size();

	if (d_pos >= s) return NULL;

	CExpression* p = dd->momentum(d_pos);
	CExpression* s_plus = NULL;
	CExpression* s_minus = NULL;
	if (sum != NULL) {
		s_plus = sum->clone()->add(p->clone());
		s_minus = sum->clone()->minus(p->clone());

		s_plus = CEProcessor::kinematics()->get_simpliest(CExpression::reduce(s_plus));
		s_minus = CEProcessor::kinematics()->get_simpliest(CExpression::reduce(s_minus));
	}
	else {
		s_plus = p->clone();
		s_minus = p->clone();
		s_minus->inverseFlag();
	}
	
	CExpression* num_mom = m_couplings->GetAt(n_pos);
	//CVirtIntegral::dbg->log(45, "TR: p", p);
	//CVirtIntegral::dbg->log(45, "TR: s_plus", s_plus);
	//CVirtIntegral::dbg->log(45, "TR: s_minus", s_minus);
	//CVirtIntegral::dbg->log(45, "TR: s_zero", sum);
	//CVirtIntegral::dbg->log(45, "TR: num_mom", num_mom);
	if (num_mom->equals(s_plus)) {
		delete s_plus;
		delete s_minus;
		return tens_reduce_similar(n_pos, d_pos, true);
	}
	
	if (num_mom->equals(s_minus)) {
		delete s_plus;
		delete s_minus;
		return tens_reduce_similar(n_pos, d_pos, false);
	}

	//CVirtIntegral::dbg->log(45, "no equals");

	CExpression* res = tens_reduce_combines(d_pos+1, n_pos, s_plus);
	if (res != NULL) {
		res = tens_reduce_similar(n_pos, d_pos, true)->add(res);
	}
	else {
		res = tens_reduce_combines(d_pos+1, n_pos, s_minus);
		if (res != NULL) {
			res = tens_reduce_similar(n_pos, d_pos, false)->add(res);
		}
		else res = tens_reduce_combines(d_pos+1, n_pos, sum);
	}

	delete s_plus;
	delete s_minus;
	return res;
}

CExpression* CVirtIntC::expand_to_vi_basis()
{
	CVirtIntegral::dbg->log(30, "CVirtIntC::expand_to_vi_basis()", this);
	int c_s = m_couplings->size();
	int i_s = rank();
	int vi_s = c_s+i_s;
	if (vi_s <= 0) return NULL;

	CVIDensData* dd = get_dens_data();
	int s = dd->getDensNumber();
	int i;
	bool expand_OK = false;
	CESum* res = new CESum();
	switch (vi_s)
	{
	case 1: // X_i
		for (i=1; i<s; i++) {
			CVirtIntegral* vi = new CVirtInt1(dd, 1, i);
			res->addItem(GetExpandCpl(dd->momentum(i), 0)->multiply(vi));
			expand_OK = true;
		}
		break;
	
	case 2: // X_ij
		{
			switch(s) {
			case 2:
				{
					CExpression* p1 = dd->momentum(1);
					CVirtIntegral* vi = new CVirtInt1(dd, 2, 1);
					res->addItem(GetExpandCpl(p1, 0)->multiply(GetExpandCpl(p1, 1))->multiply(vi));
					vi = new CVirtInt1(dd, 2, 2);
					res->addItem(GetExpandCpl(0, 1)->multiply(vi));
					expand_OK = true;
				}
				break;
			case 3:
				{
					CExpression* p1 = dd->momentum(1);
					CExpression* p2 = dd->momentum(2);
					
					CVirtIntegral* vi = new CVirtInt1(dd, 2, 1);
					res->addItem(GetExpandCpl(p1, 0)->multiply(GetExpandCpl(p1, 1))->multiply(vi));
					vi = new CVirtInt1(dd, 2, 2);
					res->addItem(GetExpandCpl(p2, 0)->multiply(GetExpandCpl(p2, 1))->multiply(vi));
					vi = new CVirtInt1(dd, 2, 3);
					CExpression* fact = GetExpandCpl(p1, 0)->multiply(GetExpandCpl(p2, 1));
					fact = fact->add(GetExpandCpl(p2, 0)->multiply(GetExpandCpl(p1, 1)));
					res->addItem(fact->multiply(vi));
					vi = new CVirtInt1(dd, 2, 4);
					res->addItem(GetExpandCpl(0, 1)->multiply(vi));
					expand_OK = true;
				}
				break;
			case 4: //{D_11,D_12,D_13,D_22,D_23,D_33,D_00}
				{
					CExpression* p1=dd->momentum(1); CExpression* p2=dd->momentum(2); CExpression* p3=dd->momentum(3);
					res->addItem( GetExpandCpl(p1,0)->multiply(GetExpandCpl(p1,1))->multiply(new CVirtInt1(dd,2,1)) );
					res->addItem( ((GetExpandCpl(p1,0)->multiply(GetExpandCpl(p2,1)))
						->add((GetExpandCpl(p2,0)->multiply(GetExpandCpl(p1,1)))))->multiply(new CVirtInt1(dd,2,2)) );
					res->addItem( ((GetExpandCpl(p1,0)->multiply(GetExpandCpl(p3,1)))
						->add((GetExpandCpl(p3,0)->multiply(GetExpandCpl(p1,1)))))->multiply(new CVirtInt1(dd,2,3)) );
					res->addItem( GetExpandCpl(p2,0)->multiply(GetExpandCpl(p2,1))->multiply(new CVirtInt1(dd,2,4)) );
					res->addItem( ((GetExpandCpl(p2,0)->multiply(GetExpandCpl(p3,1)))
						->add((GetExpandCpl(p3,0)->multiply(GetExpandCpl(p2,1)))))->multiply(new CVirtInt1(dd,2,5)) );
					res->addItem( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p3,1))->multiply(new CVirtInt1(dd,2,6)) );
					res->addItem(GetExpandCpl(0,1)->multiply(new CVirtInt1(dd,2,7)));
					expand_OK = true;
				}
				break;
			}
		}
		break;

	case 3: // X_ijk
		//CExpression* q1 = m_couplings->GetAt(0);
		//CExpression* q2 = m_couplings->GetAt(1);
		//CExpression* q3 = m_couplings->GetAt(2);

		switch(s) {
		case 3:
			{
				CExpression* p1=dd->momentum(1); CExpression* p2=dd->momentum(2);

				res->addItem(
					GetExpandCpl(0, 1)->multiply(GetExpandCpl(p1, 2))
					->add( GetExpandCpl(1, 2)->multiply(GetExpandCpl(p1, 0)) )
					->add( GetExpandCpl(0, 2)->multiply(GetExpandCpl(p1, 1)) )
					->multiply(new CVirtInt1(dd, 3, 1)) // C_001
					);

				res->addItem(
					GetExpandCpl(0, 1)->multiply(GetExpandCpl(p2, 2))
					->add( GetExpandCpl(1, 2)->multiply(GetExpandCpl(p2, 0)) )
					->add( GetExpandCpl(0, 2)->multiply(GetExpandCpl(p2, 1)) )
					->multiply(new CVirtInt1(dd, 3, 2)) // C_002
					);

				res->addItem(
					GetExpandCpl(p1, 0)->multiply(GetExpandCpl(p1, 1))->multiply(GetExpandCpl(p1, 2))
					->multiply(new CVirtInt1(dd, 3, 3)) // C_111
					);

				res->addItem(
					GetExpandCpl(p2, 0)->multiply(GetExpandCpl(p2, 1))->multiply(GetExpandCpl(p2, 2))
					->multiply(new CVirtInt1(dd, 3, 4)) // C_222
					);

				res->addItem(
					GetExpandCpl(p1, 0)->multiply(GetExpandCpl(p1, 1))->multiply(GetExpandCpl(p2, 2))
					->add( GetExpandCpl(p1, 0)->multiply(GetExpandCpl(p2, 1))->multiply(GetExpandCpl(p1, 2)) )
					->add( GetExpandCpl(p2, 0)->multiply(GetExpandCpl(p1, 1))->multiply(GetExpandCpl(p1, 2)) )
					->multiply(new CVirtInt1(dd, 3, 5)) // C_112
					);

				res->addItem(
					GetExpandCpl(p2, 0)->multiply(GetExpandCpl(p2, 1))->multiply(GetExpandCpl(p1, 2))
					->add( GetExpandCpl(p2, 0)->multiply(GetExpandCpl(p1, 1))->multiply(GetExpandCpl(p2, 2)) )
					->add( GetExpandCpl(p1, 0)->multiply(GetExpandCpl(p2, 1))->multiply(GetExpandCpl(p2, 2)) )
					->multiply(new CVirtInt1(dd, 3, 6)) // C_122
					);
				expand_OK = true;
			}
			break;
		case 4:
			{
				//1:001; 2:002; 3:003; 4:111; 5:222; 6:333; 7:112; 8:113; 9:122; 10:123; 11:133; 12:223; 13:233
				for(i=1; i<=3; i++){
					CExpression* pI=dd->momentum(i);
					res->addItem(
						GetExpandCpl(0,1)->multiply(GetExpandCpl(pI,2))
						->add(GetExpandCpl(1,2)->multiply(GetExpandCpl(pI,0)))
						->add(GetExpandCpl(0,2)->multiply(GetExpandCpl(pI,1)))
						->multiply(new CVirtInt1(dd,3,i)) // D_001, D_002, D_003
					);
				}
				for(i=1; i<=3; i++){
					CExpression* pI=dd->momentum(i);
					res->addItem(
						GetExpandCpl(pI,0)->multiply(GetExpandCpl(pI,1))->multiply(GetExpandCpl(pI,2))
						->multiply(new CVirtInt1(dd,3,i+3)) // D_111, D_222, D_333
					);
				}
				CExpression* p1=dd->momentum(1); CExpression* p2=dd->momentum(2); CExpression* p3=dd->momentum(3);
				res->addItem(
					GetExpandCpl(p1,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p2,2))
					->add( GetExpandCpl(p1,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p1,2)) )
					->add( GetExpandCpl(p2,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p1,2)) )
					->multiply(new CVirtInt1(dd,3,7)) //D_112
					);
				res->addItem(
					GetExpandCpl(p1,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p3,2))
					->add( GetExpandCpl(p1,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p1,2)) )
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p1,2)) )
					->multiply(new CVirtInt1(dd,3,8)) //D_113
					);
				res->addItem(
					GetExpandCpl(p1,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p2,2))
					->add( GetExpandCpl(p2,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p2,2)) )
					->add( GetExpandCpl(p2,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p1,2)) )
					->multiply(new CVirtInt1(dd,3,9)) //D_122
					);
				res->addItem(
					GetExpandCpl(p1,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p3,2))
					->add( GetExpandCpl(p1,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p2,2)) )
					->add( GetExpandCpl(p2,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p3,2)) )
					->add( GetExpandCpl(p2,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p1,2)) )
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p2,2)) )
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p1,2)) )
					->multiply(new CVirtInt1(dd,3,10)) //D_123
					);
				res->addItem(
					GetExpandCpl(p1,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p3,2))
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p1,1))->multiply(GetExpandCpl(p3,2)) )
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p1,2)) )
					->multiply(new CVirtInt1(dd,3,11)) //D_133
					);
				res->addItem(
					GetExpandCpl(p2,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p3,2))
					->add( GetExpandCpl(p2,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p2,2)) )
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p2,2)) )
					->multiply(new CVirtInt1(dd,3,12)) //D_223
					);
				res->addItem(
					GetExpandCpl(p2,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p3,2))
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p2,1))->multiply(GetExpandCpl(p3,2)) )
					->add( GetExpandCpl(p3,0)->multiply(GetExpandCpl(p3,1))->multiply(GetExpandCpl(p2,2)) )
					->multiply(new CVirtInt1(dd,3,13)) //D_233
					);
				expand_OK = true;
			}
			break;
		}
		break;
	}

	if (!expand_OK) {
		CVirtIntegral::dbg->log(0, "ERROR: vector basis decomposition is not available for this inegral", this);
		delete res; res=NULL;
		CEException::halt();
	}

	//CVirtIntegral::dbg->log(5, "CVirtIntC::expand_to_vi_basis() result", res);

	if (res != NULL) { copyAttributesTo(res, true); }
	return res;
}

CExpression* CVirtIntC::GetExpandCpl(CExpression* base, int i) {
	int c_s = m_couplings->size();
	if (i < c_s) {
		return new CESProduct(base, m_couplings->GetAt(i));
	}

	if (m_indices==NULL) { dbg->log(0, i, "CVirtIntC::GetExpandCpl(%ld): (NULL m_indices)", this); return NULL; }
	i -= c_s;
	
	CTIndex* ti = (CTIndex*)m_indices->GetAt(i);
	if (ti == NULL) { CEException::halt("CVirtIntC::GetExpandCpl"); }
	return new CEVector(base, ti->index(), ti->covariance());
}

CExpression* CVirtIntC::GetExpandCpl(int i, int j) {
	int c_s = m_couplings->size();

	if (i < c_s) {
		if (j < c_s) {
			return new CESProduct(m_couplings->GetAt(i), m_couplings->GetAt(j));
		}
		if (m_indices==NULL) { dbg->log(0, j, "CVirtIntC::GetExpandCpl(%ld): (NULL m_indices)", this); return NULL; }
		j-=c_s;
		CTIndex* ti_j = (CTIndex*)m_indices->GetAt(j);
		if (ti_j == NULL) { CEException::halt("CVirtIntC::GetExpandCpl"); }
		return new CEVector(m_couplings->GetAt(i), ti_j->index(), ti_j->covariance());
	}

	if (m_indices==NULL) { dbg->log(0, i, "CVirtIntC::GetExpandCpl(%ld): (NULL m_indices)", this); return NULL; }
	i-=c_s;
	CTIndex* ti_i = (CTIndex*)m_indices->GetAt(i);
	if (ti_i == NULL) { CEException::halt("CVirtIntC::GetExpandCpl"); }

	if (j < c_s) {
		return new CEVector(m_couplings->GetAt(j), ti_i->index(), ti_i->covariance());
	}
	j-=c_s;
	CTIndex* ti_j = (CTIndex*)m_indices->GetAt(j);
	if (ti_j == NULL) { CEException::halt("CVirtIntC::GetExpandCpl"); }
	return new CEMetric(ti_i->index(), ti_j->index(), ti_i->covariance(), ti_j->covariance());
}

void CVirtIntC::XCplRemoveAt(CSequence* _couplings, CEPtrArray* _indices, int i) {
	int c_s = _couplings->size();
	if (i < c_s) { _couplings->RemoveAt(i); return; }

	i-=c_s;
	if ((_indices == NULL) || (i >= _indices->size())) {
		dbg->log(0, i, "CVirtIntC::XCplRemoveAt(WRONG ARGUMENTS): %ld");
		dbg->log(2, "CVirtIntC::XCplRemoveAt(WRONG ARGUMENTS), other arguments:", _couplings, _indices);
		CEException::halt();
	}
	_indices->RemoveAt(i, true);
}
