////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LHEPRunProject.h: interface for the CLHEPBasics class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(HEADER_LHEPRUNPROJECT_H__INCLUDED_)
#define HEADER_LHEPRUNPROJECT_H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

/*
*	Global run procedure
*/
class CLHEPRunProject : public CELibrary {
	CEString m_sTitle;
	CEString m_sProjPath;		//path to directory with 'Makefile', not to the current one
	CEString m_sProjMainFile;	//filename of the 'main'-containing file
	CEString m_sTexReport;
	CKinematics* m_kin;
	CDiagramPhysics* m_phys;
	CCodeLanguage* m_codeLang;
	CEPtrArray m_ProcCodeLibs;	//array of CCodeProcessAdv instances
	int m_iCurrKinConfNo;			//number of active configuration (derived parameter: 0, or m_infConfig.m_nBremsConfId+1)
	CCodeProcessAdv* ProcCodeLib() { return ProcCodeLib(m_iCurrKinConfNo); }
	CCodeProcessAdv* ProcCodeLib(int i) { return (CCodeProcessAdv*)(m_ProcCodeLibs.GetAt(i)); }

public:
	class CLHEPRunProject_Settings {
	public:
		//default values
		static bool bEvtNuclonEntriesDefault;	//default for 'setts.m_bUseEvtNuclonEntries'
	public:
		//basic parameters
		unsigned long m_iFlags;
		int m_nQuarksInNuclon;				//number of quarks to consider in initial nuclon
		bool m_bUseEvtNuclonEntries;		//def:bEvtNuclonEntriesDefault, add 'nuclon' lines before partons in event (if applicable)

		//other parameters, propably derived from the basics
		bool m_bCodeCpp;					//otherwise: f77
		bool m_bUseAbstractTypes;			//e.g. use 'tReal' instead of 'double'
		bool m_bSeparateChannelsAsProcs;	//def:false, integrate/generate parton channels separatelly (as processes)

		//
		CLHEPRunProject_Settings() {
			m_iFlags=0; m_bCodeCpp=true; m_nQuarksInNuclon=4;
			m_bUseEvtNuclonEntries=CLHEPRunProject_Settings::bEvtNuclonEntriesDefault;
			m_bUseAbstractTypes=true; m_bSeparateChannelsAsProcs=false;
		}
	};
private:
	CLHEPRunProject_Settings m_setts;

private:
	// Sub-processes information
	class CLHEPRunProject_ConfigInfo {
	public:
		long m_nSubProcs;
		long m_nSubProcsTried;
		long m_nSubRunCalls;			//may be different from 'm_nSubProcsTried', TODO: unify them?
		char* m_sSubRunStat;			//list of valid/empty channels, "0010010"
		long m_szSubRunStat;
		long m_nBremsConfId;			//-1 if no tried
		long m_nSubProcsActual;
		CIntArray m_arrSubProcConfig;
		CEPtrArray m_arrOrigParticles;
		CEPtrArray m_arrBremsParticles;	//particles to be added in bremsstrahlung kinematics (photon, gluon etc.)
		CEPtrArray m_subProcClassNames;
		CEPtrArray m_subProcClassHeaders;
		//
		CEString m_fnConstInclude;
		long m_pkInitials[2];			//e.g. protons
		CEString m_symMassIn[2];		//as in include file
		CEPtrArray m_symInIDs;			//arr{CEString*}, initial particles as 't','tbar','W+','e-','e+' etc.
		CEPtrArray m_symFinIDs;			//arr{CEString*}, final particles as 't','tbar','W+','e-','e+' etc.
		//
		bool m_bNuclonToQuarks;
		bool m_bAllowQuarksSwap;		//generate same procedure for 'q1+q2->...' and 'q2+q1->...' sub-processes
		//
		long m_nConfigCombinations;		//{x3} for every nuclon, etc.
		bool m_bPartonsUsed_beam1;		//internal flag, may be 'true' even with 'm_nConfigCombinations==1'
		bool m_bPartonsUsed_beam2;		//internal flag, may be 'true' even with 'm_nConfigCombinations==1'
		//
		CLHEPRunProject_ConfigInfo() {
			m_nSubProcs=0; m_nSubProcsTried=0; m_nBremsConfId=-1; m_nSubProcsActual=0; m_arrOrigParticles.SetAllowNULLs(true); m_arrBremsParticles.SetAllowNULLs(true);
			m_nSubRunCalls=0; m_sSubRunStat=NULL; m_szSubRunStat=0;
			m_nConfigCombinations=1;
			m_bNuclonToQuarks=false;	//default
			m_bAllowQuarksSwap=false;	//default
			m_bPartonsUsed_beam1=false;	//internal flag
			m_bPartonsUsed_beam2=false;	//internal flag
			m_subProcClassHeaders.SetAllowNULLs(true);
			m_pkInitials[0]=0; m_pkInitials[1]=0;
		}
		~CLHEPRunProject_ConfigInfo() {
			m_arrOrigParticles.clearall_without_delete_objects();
			if(m_sSubRunStat!=NULL){ CMemoryManager::mfree(m_sSubRunStat); m_sSubRunStat=NULL; m_szSubRunStat=0; }
			//Nt: m_arrBremsParticles owns its content
		}
	};
private:
	CLHEPRunProject_ConfigInfo m_infConfig;

public:
	static CDebug* dbg;

public:
	static const unsigned long flag_RUN_DEFAULT;	//default flags
	static const unsigned long flag_RUN_SQUARE;		//square matrix elements instead of amplitude calculation
	static const unsigned long flag_RUN_AMPLITUDE;	//calculate amplitude instead of squaring ME
	static const unsigned long flag_RUN_RC;			//include radiative correction (Loops + Counterterms + Bremsstrahlung)
	static const unsigned long flag_RUN_LOAD;		//try to load intermadiate expressions

public:
	static bool Run(const char* title,
					unsigned long iFlags,
					const char* fnTeX=NULL
					);

private:
	CLHEPRunProject(const char* title,
					unsigned long iFlags,
					const char* fnTeX	//if NULL => set to 'title.tex'
					);
public:
	virtual ~CLHEPRunProject();

private:
	bool RunImpl();
	bool SetNextSubProcConfig();		//must be followed by 'RestoreOriginalConf'
	bool CheckChargeConservation();		//check current kinematics
	bool CheckParticleConfigNum(CEParticle* p, int* nConfig);	//'true' if particle contains partons
	CEParticle* CreateParticleConfig(int pID, int iConf, CEParticle* p);
	void ResetBremmsConfig();			//clear internal counter before 'SetNextBremmsConfig' calls
	bool SetNextBremmsConfig();			//add next possible massless boson to kinematics (or remove any at the end of loop)
	void RestoreOriginalConfig();
	bool CheckKinematics();				//check if procedure is implemented for current kinematics
	//
	//	iKinConfNo: ID of kinematical configuration (used to mix processes with different kinematics, like V+R-terms)
	bool RunSubProcess(int iKinConfNo);	//returns 'true' for non-zero cross section
	//
	void CreateProcFile();
	void CreateMainFile();
	void CreateMakefile();

private:
	float GetDefaultEnergy(long pk);
};

#endif // !defined(HEADER_LHEPRUNPROJECT_H__INCLUDED_)
