////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// MatrixElement.cpp: implementation of the CMatrixElement class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const char* CMatrixElement::CLASS_NAME = "CMatrixElement";
CDebug* CMatrixElement::dbg = new CDebug("CMatrixElement");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMatrixElement::CMatrixElement(CExpression* expr_val,
							   CEPtrArray* list_virt_momenta,
							   CEPtrArray* external_vector_indices)
: CESerializable()
{
	m_expr_val = NULL;
	if (expr_val != NULL) m_expr_val = expr_val->clone();
	m_virt_momenta_list = CEPtrArray::get_copy(list_virt_momenta, true);
	m_external_vector_indices = CEPtrArray::get_copy(external_vector_indices, true);

	// sort m_external_vector_indices;
	// sort with minimal operations for previously sorted array
	int s = m_external_vector_indices->size();
	int i_prev = INT_MIN;
	int i;
	for(i=0; i<s; i++) {
		int i_val = ((CExprContainer*)m_external_vector_indices->GetAt(i))->get_info();
		if (i_val < i_prev) {
			int j=i-1;
			for(;;j--) {
				if (j<=0) break;
				// compare to previous before 'j'
				int j_prev = ((CExprContainer*)m_external_vector_indices->GetAt(j-1))->get_info();
				if (i_val >= j_prev) break;
			}
			m_external_vector_indices->MoveFromTo(i, j);
		}
		else { i_prev = i_val; }
	}
}

CMatrixElement::~CMatrixElement()
{
	if (m_expr_val != NULL) delete m_expr_val;
	if (m_virt_momenta_list != NULL) delete m_virt_momenta_list;
	if (m_external_vector_indices != NULL) delete m_external_vector_indices;
}

CExpression* CMatrixElement::get_expr_val() { return m_expr_val; }
void CMatrixElement::set_expr_val(CExpression* _val) {
	if (m_expr_val != _val) {
		if (m_expr_val != NULL) delete m_expr_val;
		m_expr_val = (_val == NULL)?NULL:(_val->clone());
	}
}

CEPtrArray* CMatrixElement::get_virt_momenta_list() { return m_virt_momenta_list; }
CEPtrArray* CMatrixElement::get_external_indices_list() { return m_external_vector_indices; }

const char* CMatrixElement::ClassName() const { return CLASS_NAME; }
CEObject* CMatrixElement::copy() const { return clone(); }
CMatrixElement* CMatrixElement::clone() const {
	return new CMatrixElement(m_expr_val, m_virt_momenta_list, m_external_vector_indices);
}

bool CMatrixElement::contain_vector(void * obj) {
	if ((m_expr_val == obj) ||
		(m_virt_momenta_list == obj) ||
		(m_external_vector_indices == obj)) return true;
	if (m_expr_val != NULL) { if (m_expr_val->contain_vector(obj)) return true; }
	if (m_virt_momenta_list != NULL) { if (m_virt_momenta_list->contain_vector(obj)) return true; }
	if (m_external_vector_indices != NULL) { if (m_external_vector_indices->contain_vector(obj)) return true; }
	return false;
}

bool CMatrixElement::obj_equals(CEObject* other) {
	if (other == NULL) return false;
	if (other == this) return true;
	if (!CEUtil::equals(ClassName(), other->ClassName())) return false;
	return equals((CMatrixElement*)other);
}

bool CMatrixElement::equals(CMatrixElement* other) {
	if (other == NULL) return false;
	if (other == this) return true;

	if ( !(get_virt_momenta_list()->obj_equals(other->get_virt_momenta_list())) ) return false;
	if ( !(get_external_indices_list()->obj_equals(other->get_external_indices_list())) ) return false;
	
	CExpression* a = get_expr_val();
	CExpression* b = other->get_expr_val();
	if (a != b) {
		if (a == NULL) return false; // b != NULL
		if (!a->equals(b)) return false;
	}
	return true;
}

void CMatrixElement::Arrange() {
	m_expr_val = CExpression::arrge(m_expr_val);
}

CExpression* CMatrixElement::get_ext_vector_index(int part_ID) {
	int s = m_external_vector_indices->size();
	int i;
	for(i=0; i<s; i++) {
		CExprContainer* cart = (CExprContainer*)m_external_vector_indices->GetAt(i);
		int i_val = cart->get_info();
		if (i_val == part_ID) return cart->get_expr();
		if (i_val > part_ID) break; // array is sorted
	}
	return NULL; // no vector index for part_ID particle found
}

void CMatrixElement::AppendCrossings(CEPtrArray* cross_infos) {
	dbg->log(20, "AppendCrossings()", cross_infos);
	dbg->log(40, "AppendCrossings() ME before appending:", this);
	if (cross_infos == NULL) return;
	int s = cross_infos->size();
	if (s == 0) return;
	
	CExpression* expr_new = get_expr_val()->clone();
	int i;
	for (i=0; i<s; i++) {
		CExpression* cross_expr = expr_new->clone();

		CEPtrArray* info = (CEPtrArray*)cross_infos->GetAt(i);
		CExprContainer* PID_cart = (CExprContainer*)info->GetAt(0);
		CExpression* v_ind_1 = get_ext_vector_index(PID_cart->get_info(0));
		if (v_ind_1 != NULL) {
			CExpression* v_ind_2 = get_ext_vector_index(PID_cart->get_info(1));
			if (v_ind_2 == NULL) { dbg->log("ERROR: invalid crossing data or matrix element structure."); throw(new CEException(CEException::INT_ERROR)); }
			CEReplasement* repl_inds = new CEReplasement(v_ind_1, v_ind_2, true);
			cross_expr = repl_inds->ApplyTo(cross_expr);
			delete repl_inds;
		}

		CERule* r = (CERule*)info->GetAt(1);
		cross_expr = r->ApplyTo(cross_expr);

		int part_PK = PID_cart->get_info(2);
		if (CEParticle::is_fermion_PK(part_PK)) {
			expr_new = expr_new->minus(cross_expr);
		} else {
			expr_new = expr_new->add(cross_expr);
		}
	}
	
	set_expr_val(expr_new);
	delete expr_new;
	dbg->log(40, "AppendCrossings() result", this);
}

CEString* CMatrixElement::toString(int mode) {
	CEString* res = new CEString("{");
	if (m_expr_val != NULL) res->AddString(m_expr_val->toString(mode));
	else res->AddString("NULL");

	// add virt momenta
	res->AddString(", {");
	int s = m_virt_momenta_list->size();
	int i;
	for(i=0; i<s; i++) {
		res->AddString(((CESerializable*)m_virt_momenta_list->GetAt(i))->toString(mode));
		if (i < (s-1)) res->AddString(", ");
	}

	// add ext indices info
	res->AddString("}, {");
	s = m_external_vector_indices->size();
	for(i=0; i<s; i++) {
		res->AddString(((CESerializable*)m_external_vector_indices->GetAt(i))->toString(mode));
		if (i < (s-1)) res->AddString(", ");
	}
	res->AddString("}}");
	return res;
}

/*
*	other will be deleted
*	returns false if non-compartible ME are summed
*	NOTE! all the volatile variables are supposed to either EQUAL or UNIQUE
*	In other cases the errors will occur.
*/
bool CMatrixElement::addME(CMatrixElement* other) {
	if (other == NULL) return false;
	CExpression* o_val = other->get_expr_val();
	if (o_val == NULL) { delete other; return false; }

	// set volatile variables in other same to this ME
	CEReplList rl;
	CEPtrArray* _virts = get_virt_momenta_list();
	CEPtrArray* o_virts = other->get_virt_momenta_list();
	int s = _virts->size();
	int o_s = o_virts->size();

	if (s!=o_s) {
		dbg->log(0, "ERROR: Tries to sum up INCOMPARTIBLE matrix elements.");
		dbg->log(0, "The number of loops is different.");
		dbg->log(0, s, o_s, "%ld- and %ld-loop(s) integrals may not be summed up.");
		dbg->log(0, "Loop diagrams must be splitted from Born/counter-term ones!");
		dbg->log(0, "Matrix element log follows.", this, other);
		delete other;
		return false;
	}
	int i; int j;
	// Note: old code allowed various loops at single ME, like born+loop
	for(i=0; i<s; i++) { //for(i=0; ((i<s) && (i<o_s)); i++) {
		CExpression* a = (CExpression*)_virts->GetAt(i);
		CExpression* b = (CExpression*)o_virts->GetAt(i);
		if (!a->equals(b)) rl.add(b, a);
	}
	// if (s < o_s) { for(i=s; i<o_s; i++) { _virts->add(((CExpression*)o_virts->GetAt(i))->clone()); } }
	
	CEPtrArray* _inds = get_external_indices_list();
	CEPtrArray* o_inds = other->get_external_indices_list();
	int s1 = _inds->size(); int s2 = o_inds->size();
	if (s1 != s2) {
		dbg->log(0, "ERROR: Tries to sum up INCOMPARTIBLE matrix elements.");
		dbg->log(0, "The external vector indices lists is different.");
		dbg->log(0, "Please, check if diagrams are created at the same time");
		dbg->log(0, "Matrix element log follows.", this, other);
		delete other;
		CEException::halt(NULL,__FILE__,__LINE__);
		return false;
	}
	for(i=0; i<s1; i++) {
		CExprContainer* cart_a = (CExprContainer*)_inds->GetAt(i);
		CExpression* a=((cart_a!=NULL)?cart_a->get_expr():NULL); if (a==NULL) CEException::halt(NULL,__FILE__,__LINE__);
		for(j=0; j<s1; j++) {
			CExprContainer* cart_b = (CExprContainer*)o_inds->GetAt(i);
			CExpression* b=((cart_b!=NULL)?cart_b->get_expr():NULL); if (b==NULL) CEException::halt(NULL,__FILE__,__LINE__);
			if (!a->equals(b)) {
				//rl.add(b, a);
				dbg->log(0, "ERROR: Tries to sum up INCOMPARTIBLE matrix elements.");
				dbg->log(0, "The external vector indices lists is different.");
				dbg->log(0, "Please, check if diagrams are created at the same time");
				dbg->log(0, "Matrix element log follows.", this, other);
				delete other;
				CEException::halt(NULL,__FILE__,__LINE__);
				return false;
			}
		}
	}

	o_val = o_val->clone();
	o_val = rl.ApplyTo(o_val, true);

	// add other ME
	if (m_expr_val != NULL) m_expr_val = m_expr_val->add(o_val);
	else m_expr_val = o_val;

	delete other;
	return true;
}

/*
*	call(true) means do not change external vector indices (they will be changed later).
*/
void CMatrixElement::ChangeVolatileIndices(bool do_not_change_ext_vec_indices) {
	dbg->log(20, do_not_change_ext_vec_indices, "ChangeVolatileIndices(%ld)", this);
	if (m_expr_val == NULL) return; // nothing to change

	CEPtrArray* protected_syms = NULL;
	if (do_not_change_ext_vec_indices) { // protect external vector indices
		protected_syms = new CEPtrArray();
		int _i;
		for (_i=0; _i<m_external_vector_indices->size(); _i++) {
			CExprContainer* cart = (CExprContainer*)m_external_vector_indices->GetAt(_i);
			protected_syms->add(cart->get_expr()->clone());
		}
	}
	
	CRCollectSymbols* r_coll = new CRCollectSymbols(
		CRCollectSymbols::COLLECT_UNIQUES | CRCollectSymbols::COLLECT_TENSOR_INDICES_ONLY,
		protected_syms);
	if (protected_syms != NULL) delete protected_syms;
	
	m_expr_val = r_coll->ApplyTo(m_expr_val); // nothing changes in 'm_expr_val' here
	CEPtrArray* unique_indices = r_coll->getCollectedSymbols();
	
	// add virtual momenta symbols to volatiles list
	int _iv;
	for(_iv=0; _iv<m_virt_momenta_list->size(); _iv++) {
		unique_indices->add(((CExpression*)m_virt_momenta_list->GetAt(_iv))->clone());
	}
	
	int s = unique_indices->size();
	if (s == 0) { delete r_coll; return; } // nothing to change

	// create unique substitutes
	CEReplList* r_list = new CEReplList();
	int i;
	for (i=0; i<s; i++) {
		CExpression* unique_new = CESymbol::getUnique();
		r_list->add((CExpression*)unique_indices->GetAt(i), unique_new);
		delete unique_new;
	}
	delete r_coll; // unique_indices is no longer required
	
	m_expr_val = r_list->ApplyTo(m_expr_val);
	// apply same rule to ME data
	if (!do_not_change_ext_vec_indices) {
		int _i;
		for (_i=0; _i<m_external_vector_indices->size(); _i++) {
			CExprContainer* cart = (CExprContainer*)m_external_vector_indices->GetAt(_i);
			CExpression* expr_new = r_list->ApplyTo(cart->get_expr()->clone());
			cart->set_expr(expr_new);
			delete expr_new;
		}
	}
	for(_iv=0; _iv<m_virt_momenta_list->size(); _iv++) {
		CExpression* expr_new = r_list->ApplyTo(
			((CExpression*)m_virt_momenta_list->GetAt(_iv))->clone());
		m_virt_momenta_list->SetAt(_iv, expr_new, true);
	}

	delete r_list;

	// volatile indices in "m_b" are now changed
	dbg->log(40, "ChangeVolatileIndices() result", this);
}
