////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// PFermion.cpp: implementation of the CPFermion class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Color indices container
// Used to uniform color indices in various sub-processes & avoid extra unique symbols
//////////////////////////////////////////////////////////////////////
CPFermion::CPFermionColorIndexContainer::CPFermionColorIndexContainerC::CPFermionColorIndexContainerC(int _idKin, CESymbol* _ClrIndex)
: CEObject() { m_idKin=_idKin; m_ClrIndex=_ClrIndex; }
CPFermion::CPFermionColorIndexContainer::CPFermionColorIndexContainerC::~CPFermionColorIndexContainerC()
{ if(m_ClrIndex!=NULL)delete m_ClrIndex; }
CPFermion::CPFermionColorIndexContainer::CPFermionColorIndexContainer() : CEObject() {}
CPFermion::CPFermionColorIndexContainer::~CPFermionColorIndexContainer() {}
CESymbol* CPFermion::CPFermionColorIndexContainer::GetColorIndex(int kinID){ //returns non-cloned value, NULL if error
	for(int i=0;i<m_info.size();i++){
		CPFermionColorIndexContainerC* cart=(CPFermionColorIndexContainerC*)(m_info.GetAt(i)); if(cart==NULL)return NULL;
		if(cart->IdKin()==kinID)return cart->ClrIndex();
	}
	CESymbol* res=CESymbol::getUnique(); m_info.add(new CPFermionColorIndexContainerC(kinID,res)); return res;
}
CPFermion::CPFermionColorIndexContainer* CPFermion::m_ColorIndexCache = NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CPFermion::CPFermion(long _fermion_PK,
					 CExpression* id,
					 CExpression* sym_mass,
					 CExpression* _pol_vec_sym,
					 int _ID_kin,
					 CEValue* _mass_val)
					 :CEParticle(id, sym_mass, _pol_vec_sym, _ID_kin, _mass_val)
{
	// _fermion_PK < 0 means fermion to be antiparticle
	m_fermion_PK = _fermion_PK;

	m_helicity_value = 0; // 0 value means no integer helicity to be involved, but does not mean unpolarized fermion
	m_helicity_symbol = NULL; // NULL does not mean unpolarized fermion
	if (check_PK(CEParticle::PARTICLE_NEUTRINO, true)) { m_helicity_value = (m_fermion_PK > 0)?(-1):1; }

	m_eColorIndex=NULL; //NULL if no color is applied
	if(check_PK(CEParticle::PARTICLE_QUARK,true)){
		if(CPFermion::m_ColorIndexCache==NULL){
			CPFermion::m_ColorIndexCache=new CPFermion::CPFermionColorIndexContainer();
			CEProcessor::AddToDeleteList(CPFermion::m_ColorIndexCache);
		}
		if(CPFermion::m_ColorIndexCache!=NULL){ m_eColorIndex=CPFermion::m_ColorIndexCache->GetColorIndex(_ID_kin); }
		if(m_eColorIndex==NULL)m_eColorIndex=CESymbol::getUnique();
	}
	m_init_OK = true;
}

CPFermion::~CPFermion() {
	if (m_helicity_symbol != NULL) delete m_helicity_symbol;
}

long CPFermion::get_PK() { return m_fermion_PK; }
CExpression* CPFermion::GetColorIndex() { return m_eColorIndex; }	//NULL if no color is applied
CExpression* CPFermion::get_helicity_symbol() { return m_helicity_symbol; }
int CPFermion::get_helicity_value() { return m_helicity_value; }

/*
*	vars->GetAt(0) should be particle momentum with appropriate sign
*/
CExpression* CPFermion::getDensityMatrix(CEPtrArray* vars) {

	//
	CEObject* base_obj = vars->GetAt(0);
	if (!CEBasic::isExpression(base_obj)) { dbg->log(0, "Error: CPFermion::getDensityMatrix() invalid parameter", base_obj); CEException::halt(); }
	CExpression* base = ((CExpression*)base_obj)->clone();
	bool base_sign = base->getFlag();
	if (!base_sign) base->setFlag();
	if (!base->equals(get_momentum_symbol())) {
		dbg->log(0, "CPFermion::getDensityMatrix() Error: momentum parameter does not feet the particle data", base_obj); dbg->log(0, "CPFermion::getDensityMatrix() Particle is", this);
		delete base; CEException::halt(); return NULL;
	}
	CExpression* m = get_sym_mass();
	if (m == NULL) { dbg->log(0, "CPFermion::getDensityMatrix() Error: No mass symbol found for fermion", this); delete base; CEException::halt(); return NULL; }
	int _pol_info = get_polarization_info();

	// TODO: check if ( base_sign EQUALS to (m_fermion_PK>=0) )

	CExpression* e=NULL; CExpression* eVecInd=NULL; //{pol_sym,NULL} or {get_pol_vec_predifined(),get_pol_vec_predef_index()}
	if((_pol_info!=0) && (m_helicity_value==0) && (m_helicity_symbol == NULL)) {
		e=get_pol_vec_predifined(); if(e!=NULL){ eVecInd=get_pol_vec_predef_index(); }
		if(eVecInd==NULL){ e=get_pol_vector_symbol(); }
	}
	CExpression* clrIndex1=NULL; CExpression* clrIndex2=NULL;

	CExpression* res = CreateDensityMatrixImpl(base, m, base_sign,
		(_pol_info!=0),
		e, eVecInd, m_helicity_symbol, m_helicity_value, (m_fermion_PK>=0),
		is_initial(),
		clrIndex1, clrIndex2);
	delete base;

	//dbg->log(2, "getDensityMatrix result", res, m_helicity_symbol);
	return res;
}

// m==NULL means massless case
// e==NULL means unpolarized
// f_charge: true for fermion, false for anti-fermion
CExpression* CPFermion::createDM(CExpression* p, CExpression* m, CExpression* e, bool f_charge) {
	//[old]return createDMimpl(p, m, (e != NULL), e, NULL, 0, f_charge, true);
	if (p==NULL) { dbg->log(0, "Error: CPFermion::createDM() NULL parameter"); CEException::halt(); return NULL; }
	CExpression* pAbs=p; bool bMomSign=pAbs->getFlag();
	if(!bMomSign){ pAbs=pAbs->clone(); pAbs->inverseFlag(); if(!pAbs->getFlag())CEException::halt("CPFermion::createDM internal error (non-equivelent cloning)"); }
	CExpression* res = CreateDensityMatrixImpl(pAbs,m,bMomSign,(e!=NULL),e,NULL,NULL,0,f_charge,true,NULL,NULL);
	if(!bMomSign){ delete pAbs; }
	return res;
}

// m==NULL means massless case
CExpression* CPFermion::createDM(CExpression* p, int h, bool f_charge, CExpression* m) {
	//[old]return createDMimpl(p, m, (h != 0), NULL, NULL, h, f_charge, true);
	if (p==NULL) { dbg->log(0, "Error: CPFermion::createDM() NULL parameter"); CEException::halt(); return NULL; }
	CExpression* pAbs=p; bool bMomSign=pAbs->getFlag();
	if(!bMomSign){ pAbs=pAbs->clone(); pAbs->inverseFlag(); if(!pAbs->getFlag())CEException::halt("CPFermion::createDM internal error (non-equivelent cloning)"); }
	CExpression* res = CreateDensityMatrixImpl(pAbs,m,bMomSign,(h!=0),NULL,NULL,NULL,h,f_charge,true,NULL,NULL);
	if(!bMomSign){ delete pAbs; }
	return res;
}

// common case
CExpression* CPFermion::CreateDensityMatrixImpl(CExpression* pAbs, CExpression* m, bool bMomSign,
											bool is_polarized,
											CExpression* e,				//set polarization using \hat{eSym}
											CExpression* eVecIndex,		//if !=NULL: set polarization using gamma^{index} * e [=={...}_{index}]
											CExpression* h_sym, int h_val,
											bool bPKge0,				//true(fermion), false(antifermion)
											bool bAverageSpins,			//otherwise sum spins for unpolarized particle
											CExpression* clrIndex1, CExpression* clrIndex2
																		//set NULL for non-quark particle
																		)
{
	//params tests
	if (pAbs == NULL) { dbg->log(0, "Error: CPFermion::CreateDensityMatrixImpl() NULL parameter"); CEException::halt(); return NULL; }

	// TODO: check if "bMomSign" should be EQUALS to "bPKge0"(m_fermion_PK >= 0) flag
	if (!CEUtil::LEQU(bMomSign, bPKge0)) {
		dbg->log(0, bMomSign, bPKge0, "[CPFermion::CreateDensityMatrixImpl] WARNING: base_sign == %ld, bPKge0 == %ld");
	}

	// 1. unpolarized (p,m) stub
	// P(p)  ==> \hat{p}+m
	// P(-p) ==> \hat{p}-m
	CExpression* PMstub=new CESGammaProduct(pAbs);
	if(m!=NULL){ CExpression* mCpyX=m->clone(); if(!bMomSign)mCpyX->inverseFlag(); PMstub=PMstub->add(mCpyX); }

	// 2. polarization factor
	CExpression* res = NULL;
	if (is_polarized) { // polarized
		if (h_val != 0) {
			// integer helicity value
			// ferm(+/-)	: (\hat{p}+m)(\gamma_{-/+})
			// a-ferm(+/-)	: (\hat{p}+/-m)(\gamma_{+/-})
			
			int g_value = ((h_val>0)?(-1):(+1));
			if (!bPKge0) g_value = -g_value;
			res = PMstub->multiply( new CEGammaSpecial(g_value));
		} else if (h_sym != NULL) {
			// abstract helicity symbol
			// ferm(h)		: 1/2*(\hat{p}+m)(1-\gamma_5*h)
			// a-ferm(h)	: 1/2*(\hat{p}-m)(1+\gamma_5*h)
			
			CExpression* g5h = (new CEGammaSpecial(0))->multiply(h_sym->clone());
			if (bPKge0) g5h->inverseFlag();
			res = (CENumber::GetHalf())->multiply(PMstub)->multiply((CENumber::GetUnit())->add(g5h));
		} else {
			// common case
			// (\hat{p}+m)(1+\gamma_5*\hat{e})
			// ferm(e)		: 1/2*(\hat{p}+m)(1-\gamma_5*\hat{e})
			// a-ferm(e)	: 1/2*(\hat{p}-m)(1+\gamma_5*\hat{e})

			if(e==NULL){ dbg->log(0, "Error: CPFermion::CreateDensityMatrixImpl(e==NULL)"); delete PMstub; CEException::halt(); return NULL; }
			CExpression* e_gamma=NULL;
			if(eVecIndex!=NULL){
				//'e' is a vector like [...]_{eVecIndex} ==> use "gamma^{eVecIndex} * e-vector"
				e_gamma = (new CEGamma(eVecIndex,true))->multiply(e->clone());
			}else{
				//'e' is a symbol ==> use "\hat{e}"
				e_gamma=new CESGammaProduct(e);
			}
			CExpression* g5e=(new CEGammaSpecial(0))->multiply(e_gamma);
			if(bPKge0)g5e->inverseFlag();
			res = (CENumber::GetHalf())->multiply(PMstub)->multiply((CENumber::GetUnit())->add(g5e));

		}
	} else { // unpolarized
		// initial (average spins)	: 1/2*(\hat{p}+m)
		// final (sum spins)		: (\hat{p}+m)
		if (bAverageSpins) {
			res = (CENumber::GetHalf())->multiply(PMstub);
		} else {
			res = PMstub;
		}
	}
	return res;
}

/* [old]
CExpression* CPFermion::createDMimpl(CExpression* p,
									 CExpression* m, bool is_polarized,
									 CExpression* e, CExpression* h_sym, int h_val,
									 bool is_particle, bool _average) {
	
	// param tests
	if (p == NULL) { dbg->log(0, "Error: CPFermion::createDMimpl() NULL parameter"); CEException::halt(); }
	
	bool base_sign = p->getFlag();

	// TODO: check if "base_sign" should be EQUALS to "is_particle"(m_fermion_PK >= 0) flag
	if (!CEUtil::LEQU(base_sign, is_particle)) {
		dbg->log(0, base_sign, is_particle, "CPFermion::createDMimpl {WARNING} base_sign: %ld, is_particle: %ld");
	}

	// 1. unpolarized (p,m) stub
	// P(p)  ==> \hat{p}+m
	// P(-p) ==> \hat{p}-m
	CExpression* PMstub = NULL;
	if (base_sign) {
		PMstub = new CESGammaProduct(p);
		if (m != NULL) PMstub=PMstub->add(m->clone());
	} else {
		CExpression* base = p->clone(); // clone for sign truncating
		base->inverseFlag();
		if (!base->getFlag()) { CEException::halt("CPFermion::createDMimpl internal error (non-equivelent cloning)"); }
		PMstub = new CESGammaProduct(base);
		delete base;
		if (m != NULL) PMstub=PMstub->minus(m->clone());
	}
	
	CExpression* res = NULL;
	// 2. polarization factor
	if (is_polarized) { // polarized
		if (h_val != 0) {
			// integer helicity value
			// ferm(+/-)	: (\hat{p}+m)(\gamma_{-/+})
			// a-ferm(+/-)	: (\hat{p}+/-m)(\gamma_{+/-})
			
			int g_value = ((h_val>0)?(-1):(+1));
			if (!is_particle) g_value = -g_value;
			res = PMstub->multiply( new CEGammaSpecial(g_value));
		} else if (h_sym != NULL) {
			// abstract helicity symbol
			// ferm(h)		: 1/2*(\hat{p}+m)(1-\gamma_5*h)
			// a-ferm(h)	: 1/2*(\hat{p}-m)(1+\gamma_5*h)
			
			CExpression* g5h = (new CEGammaSpecial(0))->multiply(h_sym->clone());
			if (is_particle) g5h->inverseFlag();
			res = (CENumber::GetHalf())->multiply(PMstub)->multiply((CENumber::GetUnit())->add(g5h));
		} else {
			// common case
			// (\hat{p}+m)(1+\gamma_5*\hat{e})
			// ferm(e)		: 1/2*(\hat{p}+m)(1-\gamma_5*\hat{e})
			// a-ferm(e)	: 1/2*(\hat{p}-m)(1+\gamma_5*\hat{e})
			
			if (e == NULL) { dbg->log(0, "Error: CPFermion::createDMimpl(NULL polarization vector)"); delete PMstub; CEException::halt(); }
			CExpression* g5e = (new CEGammaSpecial(0))->multiply(new CESGammaProduct(e));
			if (is_particle) g5e->inverseFlag();
			res = (CENumber::GetHalf())->multiply(PMstub)->multiply((CENumber::GetUnit())->add(g5e));
		}
	} else { // unpolarized
		// initial (average)      : 1/2*(\hat{p}+m)
		// final (sum helicities) : (\hat{p}+m)
		if (_average) {
			res = (CENumber::GetHalf())->multiply(PMstub);
		} else {
			res = PMstub;
		}
	}

	return res;
}
*/

void CPFermion::setHelicityData(int _val) {
	if (check_PK(CEParticle::PARTICLE_NEUTRINO, true)) {
		if (_val != m_helicity_value) {
			if (m_fermion_PK > 0) dbg->log(0, _val, "Error: setting %ld helicity for neutrino is forbidden.", this);
			else dbg->log(0, _val, "Error: setting %ld helicity for antineutrino is forbidden.", this);
		}
		//m_helicity_value = (m_fermion_PK > 0)?(-1):1;
		dbg->log(0, m_helicity_value, "Helicity is set to %ld.");
		return;
	}

	// clear old data
	m_helicity_value = 0;
	if (m_helicity_symbol != NULL) delete m_helicity_symbol; m_helicity_symbol = NULL;

	// "_val" == 0 means clear helicity data
	if (_val != 0) {
		m_helicity_value = (_val>0)?1:(-1);
		
		// also set particle as 'polarized'
		if (get_polarization_info() == 0) CEParticle::set_polarization_info(1);
	}
}

void CPFermion::setHelicityData(CExpression* _val) {
	if (check_PK(CEParticle::PARTICLE_NEUTRINO, true)) {
		if (_val != NULL) {
			if (m_fermion_PK > 0) dbg->log(0, "Error: setting of abstract helicity for neutrino is forbidden.", _val, this);
			else dbg->log(0, "Error: setting of abstract helicity for (anti)neutrino is forbidden.", _val, this);
		} else {
			dbg->log(0, "Error: clearing helicity data for neutrino is forbidden.", this);
		}
		//m_helicity_value = (m_fermion_PK > 0)?(-1):1;
		dbg->log(0, m_helicity_value, "Helicity is set to %ld.");
		return;
	}

	// clear old data
	m_helicity_value = 0;
	if (m_helicity_symbol != NULL) delete m_helicity_symbol; m_helicity_symbol = NULL;

	// "_val" == NULL clears helicity data
	if (_val != NULL) {
		m_helicity_symbol = _val->clone();

		// also set particle as 'polarized'
		if (get_polarization_info() == 0) CEParticle::set_polarization_info(1);
	}
}

///////////////////////////////////////////////////////// CPProtonSF - proton for Structure functions approach
CExpression* CPProtonSF::getDensityMatrix(CEPtrArray* vars) {
	return CENumber::GetUnit();
}
CPProtonSF::CPProtonSF(long _fermion_PK, //always=CEParticle::PARTICLE_PROTON
					   CExpression* id, CExpression* sym_mass, CExpression* _pol_vec_sym,
					   int _ID_kin, CEValue* _mass_val)
					   : CPFermion(_fermion_PK, id, sym_mass, _pol_vec_sym, _ID_kin, _mass_val)
{
}
CPProtonSF::~CPProtonSF() {}
//TODO: cache hadronic tensor object (and return it after Q replacing)
CExpression* CPProtonSF::CreateHadronicTensor(CExpression* Q, CExpression* ind1, CExpression* ind2, bool cov1, bool cov2) {
	CExpression* p=get_momentum_symbol(); CExpression* m=get_sym_mass();
	int iPol = get_polarization_info(); CExpression* e=((iPol!=0)?(get_pol_vector_symbol()):NULL);
	if((Q==NULL) || (m==NULL) || (p==NULL) || ((iPol!=0) && (e==NULL)))return NULL;
	CEPower M2(m,(long)2);
	CExpression* Q2=new CESProduct(Q,Q); Q2->inverseFlag(); Q2=CExpression::arrge(Q2);
	CExpression* tauP=Q2->clone()->divide(CENumber::GetFour())->divide(M2.clone());
	CEFunction symGm("G\\_m"); symGm.add_param(Q2); symGm.add_param(&M2);
	CEFunction symGe("G\\_e"); symGe.add_param(Q2); symGe.add_param(&M2);
	//to delete: Q2,tauP

	if(m_TmpIndices.size() < 2) { m_TmpIndices.addItem(CESymbol::getUnique()); m_TmpIndices.addItem(CESymbol::getUnique()); }
	CExpression* i3=m_TmpIndices.GetAt(0); CExpression* i4=m_TmpIndices.GetAt(1);

	// F_1..4
	CSequence seqFn;
	bool bUseSFAsG=false;
	if(bUseSFAsG){
		seqFn.addItem( //F1
			CENumber::GetFour()->multiply(tauP->clone())->multiply(M2.clone())->multiply(symGm.sqra()));
		seqFn.addItem( //F2
			CENumber::GetFour()->multiply(M2.clone())->multiply(
			(symGe.sqra()->add(tauP->clone()->multiply(symGm.sqra())))->divide(CENumber::GetUnit()->add(tauP->clone()))));
		if (iPol!=0) {
			seqFn.addItem( //F3
				CENumber::GetNumber(-2)->multiply(M2.clone())->multiply(symGe.clone())->multiply(symGm.clone()));
			seqFn.addItem( //F4
				CENumber::GetMinusUnit()->multiply(M2.clone())->multiply(symGm.clone())->multiply(
				(symGe.clone()->minus(symGm.clone()))->divide(CENumber::GetUnit()->add(tauP->clone()))));
		}
	} else {
		CEFunction* _F1=new CEFunction("F1"); _F1->add_param(Q2); _F1->add_param(&M2); seqFn.addItem(_F1);
		CEFunction* _F2=new CEFunction("F2"); _F2->add_param(Q2); _F2->add_param(&M2); seqFn.addItem(_F2);
		if (iPol!=0) {
			CEFunction* _F3=new CEFunction("F3"); _F3->add_param(Q2); _F3->add_param(&M2); seqFn.addItem(_F3);
			CEFunction* _F4=new CEFunction("F4"); _F4->add_param(Q2); _F4->add_param(&M2); seqFn.addItem(_F4);
		}
	}

	bool bTransverse=false;
	CExpression* res=CENumber::GetZero(); CExpression* wF=NULL; CExpression* _VecPart=NULL;

	// w1*F_1
	_VecPart=new CEMetric(ind1,ind2,cov1,cov2);
	if(bTransverse){ _VecPart=_VecPart->plus( (new CEVector(Q,ind1,cov1))->multiply(new CEVector(Q,ind2,cov2))->divide(Q2->clone()) ); }
	wF=(_VecPart)->multiply(seqFn.GetAt(0)->clone()); wF->inverseFlag();
	res=res->add(wF);

	// w2*F_2
	if(bTransverse){
		_VecPart=(
			(new CEVector(p,ind1,cov1))->plus( (new CEVector(Q,ind1,cov1))->multiply(new CESProduct(p,Q))->divide(Q2->clone()) )
			)->multiply(
			(new CEVector(p,ind2,cov2))->plus( (new CEVector(Q,ind2,cov2))->multiply(new CESProduct(p,Q))->divide(Q2->clone()) )
			);
	} else {
		_VecPart=(new CEVector(p,ind1,cov1))->multiply(new CEVector(p,ind2,cov2));
	}
	wF=_VecPart->divide(M2.clone())->multiply(seqFn.GetAt(1)->clone());
	res=res->add(wF);

	if (iPol!=0) {
		// w3*F_3
		wF=CENumber::GetImUnit()->multiply(new CEAntisymmetric(ind1,cov1,ind2,cov2,i3,true,i4,true))
			->multiply(new CEVector(Q,i3,false))->multiply(new CEVector(e,i4,false))
			->divide(m->clone())->multiply(seqFn.GetAt(2)->clone()); wF->inverseFlag();
		res=res->add(wF);

		// w4*F_4
		wF=CENumber::GetImUnit()->multiply(new CEAntisymmetric(ind1,cov1,ind2,cov2,i3,true,i4,true))
			->multiply(new CEVector(Q,i3,false))->multiply(new CEVector(p,i4,false))->multiply(new CESProduct(e,Q))
			->divide(new CEPower(m,(long)3))->multiply(seqFn.GetAt(3)->clone());
		res=res->add(wF);
	}

	res=CExpression::arrge(res);
	dbg->log(45,"CPProtonSF::CreateHadronicTensor() result: ", res);
	delete Q2; delete tauP;
	return res;
}
// statics
#define CPProtonSF_VEC_TITLE	"PROTON_SF"
CExpression* CPProtonSF::GetGammaVertexSym(CExpression* momQ,CExpression* mom1,CExpression* mom2) {
	CEFunction* sfVec=new CEFunction(CPProtonSF_VEC_TITLE); sfVec->add_param(momQ); sfVec->add_param(mom1); sfVec->add_param(mom2); return sfVec;
}
//
//	ReconstructSF: for all protons Proton{p}^a...ProtonConj{p}^b --> \Sum_i{F{p}_i^ab}
//	Note: pA_{\mu}p-vertex gives 'Proton{p}_{\mu}' objects, that must be coupled
//
//	me: input & output arr{me#1, me#2}
//	factor: output
//	protonIDs: input, used to cache IDs collection
//	tmpIndices: input
//
bool CPProtonSF::ReconstructSF(CSequence* me, CSequence* factor, CIntArray* protonIDs, CSequence* tmpIndices) {
	CKinematics* kin=CEProcessor::kinematics();
	if((me==NULL) || (factor==NULL) || (protonIDs==NULL) || (tmpIndices==NULL) || (kin==NULL)) return false;
	if(me->size() != 2)return false;
	int n=protonIDs->GetSize(); int i;
	if (n!=tmpIndices->size()) { CEException::halt("CPProtonSF::ReconstructSF: invalid call"); }
	if (n%2 != 0) { dbg->log(0,n,"CPProtonSF ERROR. Unexpected number of protons spotted: %ld"); return false; }
	int nD2=n/2;

	//data: indices, SF-found, next index to use from 'indices' {0+ for me1, nD2+ for me2}
	CEPtrArray data; data.add(tmpIndices->clone()); data.add(new CSequence()); data.add(CENumber::GetZero());
	CESocketRule r((void*)CPProtonSF::ExtractSFVectorRuleImpl,&data,
		CESocketRule::SR_APPLY_TO_SUBITEMS_FIRST
		| CESocketRule::SR_APPLY_TO_CHANGED_SOURCE //(vecSF.vecSF)-->extractedSF+vecSF-->extractedSF+extractedSF
		| CESocketRule::USE_EXTERNAL_DATA_ARRAY);
	//apply to me#1
	CExpression* me1=me->GetAt(0); CExpression* me1x=r.ApplyTo(me1,false); if((me1x!=me1)&&(me1x!=NULL)){ me->SetAt(0,me1x); delete me1x; }
	CSequence* _SFFound1 =(CSequence*)data.ExtractAt(1);
	data.InsertAt(1,new CSequence()); data.SetAt(2,CENumber::GetNumber(nD2),true);
	//apply to me#2
	CExpression* me2=me->GetAt(1); CExpression* me2x=r.ApplyTo(me2,false); if((me2x!=me2)&&(me2x!=NULL)){ me->SetAt(1,me2x); delete me2x; }
	CSequence* _SFFound2=(CSequence*)data.ExtractAt(1);

	//TODO: Test SF-vector extraction by (res*SF==init)?
	//TODO: Or modify CPProtonSF::ExtractSFVectorRuleImpl to check all the factors
	if((_SFFound1->size()!=nD2)||(_SFFound2->size()!=nD2)){ dbg->log(0,"CPProtonSF ERROR. Can't reconstruct proton structure functions (0)",_SFFound1,_SFFound2); delete _SFFound1; delete _SFFound2; return false; }

	//dbg->log(0,"m_SFFound1,_SFFound2",_SFFound1,_SFFound2);
	// Select pairs from data for every protonIDs & replace by Proton-SF
	bool bFitOK=true;
	for(i=0; i<nD2; i++) {
		CEFunction* f1=(CEFunction*)_SFFound1->GetAt(i);
		int i2=-1; int j;
		for(j=0; j<nD2; j++) { if(f1->equals(_SFFound2->GetAt(j))) { i2=j; break; } }
		if(i2<0){bFitOK=false; dbg->log(0,"i2<0"); break;}
		CPProtonSF* part=NULL;
		CExpression* p=f1->get_param(2); if(p==NULL){bFitOK=false; dbg->log(0,"p==NULL");  break;}
		for(j=0; j<n; j++) {
			CEParticle* partX=kin->get_particle_by_ID(protonIDs->GetAt(j)); if(partX==NULL){bFitOK=false;break;}
			if(p->equals(partX->get_momentum_symbol())) { part=(CPProtonSF*)partX; break; }
		}
		if(part==NULL) {bFitOK=false; dbg->log(0,"part==NULL");  break;}
		CExpression* f=part->CreateHadronicTensor(f1->get_param(0),tmpIndices->GetAt(i),tmpIndices->GetAt(nD2+i2),true,true);
		if(f==NULL){bFitOK=false; dbg->log(0,"f==NULL");  break;}
		factor->addItem(f);
	}
	//TODO: check if every item of protonIDs is used && used once
	if(!bFitOK){ dbg->log(0,"CPProtonSF ERROR. Can't reconstruct proton structure functions",_SFFound1,_SFFound2); delete _SFFound1; delete _SFFound2; return false; }
	delete _SFFound1; delete _SFFound2;

	//dbg->log(0,"CPProtonSF::ReconstructSF after extraction",me,factor);
	//CEException::halt();
	return true;
}
CExpression* CPProtonSF::ExtractSFVectorRuleImpl(CExpression* src, CEPtrArray* data_arr) {
	//check for Proton_{\mu}
	if (src->ClassName() == CEVector::CLASS_NAME) {
		CExpression* eBase=((CEVector*)src)->base();
		if(eBase==NULL)return src; if (eBase->ClassName()!=CEFunction::CLASS_NAME) return src;
		CEFunction* fBase=(CEFunction*)eBase; if (!(CEUtil::equals(fBase->getTitle(), CPProtonSF_VEC_TITLE))) return src;
		if(!src->getLevel()) { dbg->log(0, "ERROR: proton structure function sub-part spotted in denominator.", src); CEException::halt(); return src; }
		CExpression* _Ind=CPProtonSF::AddSFVecAndGetSubstIndex(fBase,data_arr); if(_Ind==NULL)CEException::halt("Internal error in CPProtonSF::ExtractSFVectorRuleImpl");
		CExpression* res=new CEMetric(_Ind, ((CEVector*)src)->index(), false, ((CEVector*)src)->covariance()); src->copyAttributesTo(res,true);
		return res;
	}
	//check for \hat{Proton}
	if (src->ClassName() == CESGammaProduct::CLASS_NAME) {
		CExpression* eBase=((CESGammaProduct*)src)->getBaseExpression();
		if(eBase==NULL)return src; if (eBase->ClassName()!=CEFunction::CLASS_NAME) return src;
		CEFunction* fBase=(CEFunction*)eBase; if (!(CEUtil::equals(fBase->getTitle(), CPProtonSF_VEC_TITLE))) return src;
		if(!src->getLevel()) { dbg->log(0, "ERROR: proton structure function sub-part spotted in denominator.", src); CEException::halt(); return src; }
		CExpression* _Ind=CPProtonSF::AddSFVecAndGetSubstIndex(fBase,data_arr); if(_Ind==NULL)CEException::halt("Internal error in CPProtonSF::ExtractSFVectorRuleImpl");
		CExpression* res=new CEGamma(_Ind, false); src->copyAttributesTo(res,true);
		return res;
	}
	//check for Proton.vec
	if (src->ClassName() == CESProduct::CLASS_NAME) {
		CExpression* eBase1=((CESProduct*)src)->get_a(); CExpression* eBase2=((CESProduct*)src)->get_b();
		if ((eBase1!=NULL) && (eBase1->ClassName()==CEFunction::CLASS_NAME)) {
			CEFunction* fBase=(CEFunction*)eBase1;
			if (CEUtil::equals(fBase->getTitle(), CPProtonSF_VEC_TITLE)) {
				if(!src->getLevel()) { dbg->log(0, "ERROR: proton structure function sub-part spotted in denominator.", src); CEException::halt(); return src; }
				CExpression* _Ind=CPProtonSF::AddSFVecAndGetSubstIndex(fBase,data_arr); if(_Ind==NULL)CEException::halt("Internal error in CPProtonSF::ExtractSFVectorRuleImpl");
				CExpression* res=new CEVector(eBase2,_Ind,false); src->copyAttributesTo(res,true);
				return res; // second base will also be checked due to SR_APPLY_TO_CHANGED_SOURCE mode
			}
		}
		if ((eBase2!=NULL) && (eBase2->ClassName()==CEFunction::CLASS_NAME)) {
			CEFunction* fBase=(CEFunction*)eBase2;
			if (CEUtil::equals(fBase->getTitle(), CPProtonSF_VEC_TITLE)) {
				if(!src->getLevel()) { dbg->log(0, "ERROR: proton structure function sub-part spotted in denominator.", src); CEException::halt(); return src; }
				CExpression* _Ind=CPProtonSF::AddSFVecAndGetSubstIndex(fBase,data_arr); if(_Ind==NULL)CEException::halt("Internal error in CPProtonSF::ExtractSFVectorRuleImpl");
				CExpression* res=new CEVector(eBase1,_Ind,false); src->copyAttributesTo(res,true);
				return res; // second base will also be checked due to SR_APPLY_TO_CHANGED_SOURCE mode
			}
		}
		return src;
	}
	return src;
}

CExpression* CPProtonSF::AddSFVecAndGetSubstIndex(CExpression* sfVecFunc, CEPtrArray* data_arr) {
	if(data_arr==NULL)return NULL;
	int iNo=-1;
	CSequence* _SF=(CSequence*)data_arr->GetAt(1); CSequence* _Inds=(CSequence*)data_arr->GetAt(0);
	int nIndShift=((CENumber*)(data_arr->GetAt(2)))->getLong();
	for(int i=0; i<_SF->size(); i++) { if(sfVecFunc->equals(_SF->GetAt(i))) { iNo=i; break; } }
	if(iNo<0) { _SF->addItem(sfVecFunc->clone()); iNo=_SF->size()-1; }
	iNo+=nIndShift;
	if(iNo>_Inds->size()) { dbg->log(0,iNo,"Internal error in CPProtonSF::AddSFVecAndGetSubstIndex (iNo=%ld)",_Inds); CEException::halt(); return NULL; }
	CExpression* _Ind=_Inds->GetAt(iNo);
	return _Ind;
}
