////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// WaveFuncBoson.cpp: implementation of the CWaveFuncBoson class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWaveFuncBoson::CWaveFuncBoson(long _particle_kind,
							   int _vertex_ID,
							   unsigned long _point_ID,
							   CExpression* _momentum_sym,
							   CTIndex* _vector_index)
: CParticleWaveFunction(_particle_kind, _vertex_ID, _point_ID, _momentum_sym)
{
	// Note: NULL vector index means scalar particle wave function

	if (!CEParticle::is_bosonic_PK(m_particle_kind)) { CEException::halt("Bosonic operator construction for non-bosonic field"); }

	if (_vector_index != NULL) {
		m_vector_index = _vector_index->clone();
	} else { // Error
		m_vector_index = NULL;
		if (CEParticle::is_vec_bosonic_PK(m_particle_kind)) CEException::halt("NULL vector index for vector-bosonic field operator");
	}
}

CWaveFuncBoson::~CWaveFuncBoson() {
	if (m_vector_index != NULL) delete m_vector_index;
}

bool CWaveFuncBoson::contain_vector(void * obj) {
	if (CParticleWaveFunction::contain_vector(obj)) return true;
	if (m_vector_index == obj) return true;
	if (m_vector_index != NULL) { if (m_vector_index->contain_vector(obj)) return true; }
	return false;
}

const char* CWaveFuncBoson::CLASS_NAME = "CWaveFuncBoson";
const char* CWaveFuncBoson::ClassName() const { return CLASS_NAME; }
CTIndex* CWaveFuncBoson::getVectorIndex() { return m_vector_index; }
bool CWaveFuncBoson::isVectorBoson() { return (m_vector_index != NULL); }

CEString* CWaveFuncBoson::toString(int mode) {
	CEString* res = NULL;
	// base symbol
	if (m_vector_index != NULL) {
		res = new CEString("A");
	} else {
		res = new CEString("S");
	}

	res->AddString("\\_\\(");
	if (m_momentum_sym != NULL) { res->AddString(m_momentum_sym->toString(mode)); }
	else { res->AddString("?"); }
	res->AddString("\\)");
	
	res->insertString("\\(");
	res->AddString("\\)\\+");
	res->add_int(m_point_ID);
	
	// vector index
	if (m_vector_index != NULL) {
		bool cov = m_vector_index->covariance();
		CExpression* ind = m_vector_index->index();
		if (ind == NULL) CEException::halt("NULL expression for vector index");

		res->insertString("\\(", 0);
		res->AddString("\\)");
		res->AddString((cov)?"\\_":"\\&");
		res->AddString("\\(");
		res->AddString(ind->toString(mode));
		res->AddString("\\)");
	}

	// sign
	if (getFlag()) res->insertString("\\(");
	else res->insertString("- \\(");
	res->AddString("\\)");
	return res;
}

CExpression* CWaveFuncBoson::clone() const {
	CExpression* res = new CWaveFuncBoson(m_particle_kind,
		m_vertex_ID,
		m_point_ID,
		m_momentum_sym,
		m_vector_index);
	copyAttributesTo(res);
	return res;
}

bool CWaveFuncBoson::equals(CExpression *other) {
	if (!CParticleWaveFunction::equals(other)) return false;
	CWaveFuncBoson* wf_other = (CWaveFuncBoson*)other;

	// compare index
	CTIndex* ind1 = getVectorIndex();
	CTIndex* ind2 = wf_other->getVectorIndex();

	if ((ind1 == NULL) && (ind2 == NULL)) return true;
	if ((ind1 == NULL) || (ind2 == NULL)) return false;
	bool res = ind1->equalsIndex(ind2);
	return res;
}

#ifdef PF_LESS_THAN
bool CWaveFuncBoson::less_than(CExpression* other) {
	if (other == NULL) return false;
	if (other->getProperty(CEProperties::PROP_MASK_IS_PARTICLE_FUNCTION)) {
		if (!CEUtil::LEQU(getFlag(), other->getFlag())) return other->getFlag(); // compare sign: "-"sym is < than "sym"
		const char* o_cn = other->ClassName();
		// compare to COpFermionField, COpBosonField, CWaveFuncFermion, CWaveFuncBoson
		// A < Psi < S, Wf < Field
		if (o_cn == COpFermionField::CLASS_NAME) {
			bool is_Scalar = (getVectorIndex() == NULL);
			return (!is_Scalar);
		}
		if (o_cn == COpBosonField::CLASS_NAME) {
			COpBosonField* o = (COpBosonField*)other;
			CTIndex* i_b = o->getVectorIndex();
			if (!CEUtil::LEQU((m_vector_index==NULL), (i_b==NULL))) return (m_vector_index != NULL); // A<S
			return true;
		}
		if (o_cn == CWaveFuncFermion::CLASS_NAME) {
			bool is_Scalar = (getVectorIndex() == NULL);
			return (!is_Scalar);
		}
		if (o_cn == CWaveFuncBoson::CLASS_NAME) {
			CWaveFuncBoson* o = (CWaveFuncBoson*)other;
			CTIndex* i_b = o->getVectorIndex();
			if (!CEUtil::LEQU((m_vector_index==NULL), (i_b==NULL))) return (m_vector_index != NULL); // A<S
			CExpression* m_b = o->getMomentumSym();
			if (!CEUtil::LEQU((m_momentum_sym==NULL), (m_b==NULL))) return (m_momentum_sym != NULL);
			if (m_momentum_sym != NULL) { if (m_momentum_sym->less_than(m_b)) return true; if (!m_momentum_sym->equals(m_b)) return false; }
			if (m_point_ID != o->getPointID()) return ((long)m_point_ID < o->getPointID());
			if ((m_vector_index != NULL) && (i_b != NULL)) {
				if (!CEUtil::LEQU(m_vector_index->covariance(), i_b->covariance())) return (i_b->covariance()); // \\& < \\_
				CExpression* ia = m_vector_index->index(); CExpression* ib = i_b->index();
				if (ia->less_than(ib)) return true;
				if (!ia->equals(ib)) return false;
			}
			const char* sym_a = CDiagramPhysics::get_particle_symbolA(m_particle_kind, CDiagramPhysics::SYMBOL_STYLE_NB);
			const char* sym_b = CDiagramPhysics::get_particle_symbolA(o->getParticleKind(), CDiagramPhysics::SYMBOL_STYLE_NB);
			if (CEUtil::a_less_b(sym_a,sym_b)) return true;
			//if (!CEUtil::equals(sym_a,sym_b)) return false;
			return false;
		}
	}
	return CExpression::less_than(other);
}
#endif //PF_LESS_THAN

CExpression* CWaveFuncBoson::applyRule(CEAbstractRule* rule) {
	CERule* r = (CERule*)rule;
	
	if (m_vector_index != NULL) {
		CExpression* item = m_vector_index->index();
		CExpression* item_repl = r->ApplyTo(item, false);
		if ((item_repl != item) && (item_repl != NULL)) {
			m_vector_index->setIndex(item_repl);
			delete item_repl;
		}
	}
	return CParticleWaveFunction::applyRule(rule);
}
