////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EHistogram.cpp: implementation of the CEHistogram class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDebug* CEHistogram::dbg = new CDebug("CEHistogram");
const unsigned CEHistogram::G_PLOT_HISTO	= 0x01;
const unsigned CEHistogram::G_LOG_AXIS_Y	= 0x02;

CHBank::CHBank(unsigned _size, double l, double r):CEObject(true)
{
	m_size = _size;
	m_values = new double[_size];
	m_margines = new double[_size+1];
	m_cell_ivents = new unsigned long[_size];
	m_average = 0;
	m_num_ivents = 0;
	m_num_outmargins = 0;
	m_is_normalized = false;

	//clearing values
	unsigned i;
	for (i = 0; i<_size; i++) {
		m_values[i] = 0;
		m_cell_ivents[i] = 0;
	}

	//setting margines
	double step = (r-l)/_size;
	double margin = l;
	for (i = 0; i<_size; i++) {
		m_margines[i] = margin;
		margin += step;
	}
	m_margines[_size] = r;
}

CHBank::~CHBank()
{
	delete[] m_values;
	delete[] m_margines;
	delete[] m_cell_ivents;
}

const char* CHBank::CLASS_NAME = "CHBank";
const char* CHBank::ClassName() const { return CLASS_NAME; }

void CHBank::add_value(double val, double pow) {
	if (m_is_normalized) {
		CEHistogram::dbg->log(1, "Bank is already normalized");
		m_num_outmargins++;
		return;
	}
	m_average += pow;
	if (val < m_margines[0]) {
		m_num_outmargins++;
		return;
	}
	unsigned i;
	for (i = 0; i<m_size; i++) {
		if (val < m_margines[i+1]) {
			m_values[i] += pow;
			m_cell_ivents[i]++;
			m_num_ivents++;
			return;
		}
	}
	m_num_outmargins++;
}

void CHBank::normalize() {
	if (m_is_normalized) {
		CEHistogram::dbg->log(1, "Bank is already normalized");
		return;
	}
	unsigned i;
	for (i = 0; i<m_size; i++) {
		if (m_cell_ivents[i] != 0) m_values[i] /= m_cell_ivents[i];
	}
	m_average /= m_num_ivents;
	m_is_normalized = true;
}

bool CHBank::is_normalized() {
	return m_is_normalized;
}

void CHBank::add_values_from(CHBank* other) {
	if (!is_normalized()) normalize();
	if (other == NULL) return;
	if (!other->is_normalized()) other->normalize();
	if (size() != other->size()) {
		CEHistogram::dbg->log(1, "Different sizes");
		return;
	}
	unsigned i;
	for (i = 0; i<m_size; i++) {
		m_values[i] += other->value(i);
	}
	m_average += other->average();
}

CHBank* CHBank::create_frame() {
	return new CHBank(m_size, m_margines[0], m_margines[m_size]);
}

unsigned CHBank::size() {
	return m_size;
}

double* CHBank::margines() {
	return m_margines;
}

double CHBank::max_val()
{
	double res = 0;
	unsigned i;
	for (i = 0; i<m_size; i++) {
		if (res < m_values[i]) res = m_values[i];
	}
	return res;
}

double CHBank::average() {
	if (!is_normalized()) normalize();
	return m_average;
}

double CHBank::value(unsigned i) {
	return m_values[i];
}

unsigned long CHBank::num_ivents() {
	return m_num_ivents;
}

unsigned long CHBank::num_outmargins() {
	return m_num_outmargins;
}

CEHistogram::CEHistogram(unsigned bar_num, double l, double r) : CESerializable()
{
	m_active_bank = 0;
	m_a_values = new CEPtrArray();
	//m_max_height = 0;
	//m_num_ivents = 0;
	m_aspect_ratio = 0.61803;
	//m_bar_num = bar_num;
	//m_values = (long*)CMemoryManager::mmalloc(m_bar_num*sizeof(long));
	/*double* vals = new double[m_bar_num];
	unsigned i;
	for (i = 0; i<m_bar_num; i++) {
		//*(m_values + i*sizeof(long)) = 0;
		//m_values[i] = i;
		//m_values[i] = rand();
		vals[i] = 0;
	}
	m_a_values->Add(vals);*/
	/*m_margines = new double[m_bar_num+1];
	//m_margines = (double*)CMemoryManager::mmalloc((m_bar_num+1)*sizeof(double));
	
	set_borders(l, r);
	//log_values();
	m_num_outmargins = 0;*/

	CHBank* b0 = new CHBank(bar_num, l, r);
	m_a_values->add(b0);
}

CEHistogram::~CEHistogram()
{
	//m_a_values->clear();
	delete m_a_values;
	//delete[] m_margines;
	//CMemoryManager::mfree(m_values);
	//CMemoryManager::mfree(m_margines);
}

CEString* CEHistogram::toString(int mode)
{
	return to_postscript(m_active_bank);
}

/*void CEHistogram::set_borders(double l, double r)
{
	//m_left_border = l;
	//m_right_border = r;
	double step = (r-l)/m_bar_num;
	double margin = l;
	//double* p_margin_value = m_margines;
	unsigned i;
	for (i = 0; i<m_bar_num; i++) {
		m_margines[i] = margin;
		//*p_margin_value = margin;
		margin += step;
		//p_margin_value += sizeof(double);
	}
	m_margines[m_bar_num] = r;
	//*p_margin_value = r;
}*/

void CEHistogram::add_value(double val, double pow, int bank)
{
	get_bank(bank, true)->add_value(val, pow);
}

const char* CEHistogram::ClassName() const
{
	return "CEHistogram";
}

void CEHistogram::visualize(int bank)
{
	//CEProcessor::kinematics()->log_vectors();
	dbg->log(10, bank, "visualizing bank #%d");
	int ab_old = m_active_bank;
	m_active_bank = bank;
	CESupport::SaveGraphics(this);
	//CEProcessor::kinematics()->log_vectors();
	//dbg->log(10, bank, "graphics saved");
	char* data = (char*)CMemoryManager::mmalloc(200);
	sprintf(data, "Ivents: %d, average: %g, c/s: %g pb, max: %g, outmargins: %d.",
		get_bank(bank)->num_ivents(),
		get_bank(bank)->average(),
		get_bank(bank)->average()*0.175,
		get_bank(bank)->max_val(),
		get_bank(bank)->num_outmargins());
	//CEProcessor::kinematics()->log_vectors();
	CESupport::Save(data);
	CMemoryManager::mfree(data);
	m_active_bank = ab_old;
	//CEProcessor::kinematics()->log_vectors();
}

void CEHistogram::add_val(CEString *s, double val)
{
	char* data = (char*)CMemoryManager::mmalloc(20);
	sprintf(data, "%g ", val);
	s->AddString(data);
	CMemoryManager::mfree(data);
}

/*double CEHistogram::max_val()
{
	double res = 0;
	//long* p_value = m_values;
	unsigned i;
	for (i = 0; i<m_bar_num; i++) {
		if (res < m_values[i]) res = m_values[i];
		//p_value += sizeof(long);
	}
	return res;
}*/

/*void CEHistogram::log_values()
{
	unsigned i;
	for (i = 0; i<m_bar_num; i++) {
		printf("%f, %f, ", m_values[i], m_margines[i]);
	}
	printf("\n");
}*/

void CEHistogram::add_val2(CEString *s, double val)
{
	char* data = (char*)CMemoryManager::mmalloc(20);
	sprintf(data, "%.2g", val);
	s->AddString(data);
	CMemoryManager::mfree(data);
}

CEString* CEHistogram::to_postscript(int bank)
{
	dbg->log(20, bank, "Generating output for bank# %d");
	CEString* res = new CEString("%!\n%%Creator: Vovka\n%%AspectRatio: ");
	add_val(res, m_aspect_ratio);
	res->AddString("\nMathPictureStart\n/Mabs {\nMgmatrix idtransform\nMtmatrix dtransform\n} bind def\n/Mabsadd { Mabs\n3 -1 roll add\n3 1 roll add\nexch } bind def");
	res->AddString("\n%% Graphics\n%%IncludeResource: font Courier\n%%IncludeFont: Courier\n/Courier findfont 10  scalefont  setfont");
	res->AddString("\n% Scaling calculations");
	res->AddString("\n0.461538 0.153846 0.0147151 0.0255915 [");
	res->AddString("\n[.15385 .00222 -9 -9 ]");
	res->AddString("\n[.15385 .00222 9 0 ]");
	res->AddString("\n[.30769 .00222 -9 -9 ]");
	res->AddString("\n[.30769 .00222 9 0 ]");
	res->AddString("\n[.46154 .00222 -6 -9 ]");
	res->AddString("\n[.46154 .00222 6 0 ]");
	res->AddString("\n[.61538 .00222 -6 -9 ]");
	res->AddString("\n[.61538 .00222 6 0 ]");
	res->AddString("\n[.76923 .00222 -6 -9 ]");
	res->AddString("\n[.76923 .00222 6 0 ]");
	res->AddString("\n[.92308 .00222 -6 -9 ]");
	res->AddString("\n[.92308 .00222 6 0 ]");
	res->AddString("\n[-0.0125 .14267 -6 -4.5 ]");
	res->AddString("\n[-0.0125 .14267 0 4.5 ]");
	res->AddString("\n[-0.0125 .27063 -12 -4.5 ]");
	res->AddString("\n[-0.0125 .27063 0 4.5 ]");
	res->AddString("\n[-0.0125 .39859 -12 -4.5 ]");
	res->AddString("\n[-0.0125 .39859 0 4.5 ]");
	res->AddString("\n[-0.0125 .52654 -12 -4.5 ]");
	res->AddString("\n[-0.0125 .52654 0 4.5 ]");
	res->AddString("\n[ 0 0 0 0 ]");
	res->AddString("\n[ 1 .61803 0 0 ]");
	res->AddString("\n] MathScale");
	res->AddString("\n% Start of Graphics");
	res->AddString("\n1 setlinecap\n1 setlinejoin\nnewpath");
	res->AddString("\n0 g");
	res->AddString("\n.25 Mabswid");

	double bottom = m_aspect_ratio/40;
	double left = 0.05;
	double height = 0.95*(m_aspect_ratio-bottom);
	
	//Axis
	res->AddString("\n");
	add_val(res, left);
	add_val(res, bottom/2);
	res->AddString("m\n");
	add_val(res, left);
	add_val(res, m_aspect_ratio);
	res->AddString("L\ns");

	res->AddString("\n");
	add_val(res, left/2);
	add_val(res, bottom);
	res->AddString("m\n1 ");
	add_val(res, bottom);
	res->AddString("L\ns");

	//double mv = max_value_in_banks();

	CHBank* b_curr = get_bank(bank);
	double mv = b_curr->max_val();

	//Axis subscriptions
	add_subscriptions(res, bottom, height, left, mv, bank);

	res->AddString("\nclosepath");
	res->AddString("\nclip");
	res->AddString("\nnewpath");
	res->AddString("\n0 g");
	res->AddString("\n.5 Mabswid\n");

	/*CHBank* b_curr = get_bank(0);
	unsigned i;
	for (i=0;((i<2) && (b_curr != NULL));) {
		if (b_curr->is_normalized()) {
			add_histo(res, bottom, height, left, mv, b_curr);
		}
		i++;
		b_curr = get_bank(i);
	}*/

	if (b_curr->is_normalized())
		add_graph(res, bottom, height, left, mv, b_curr);
		//add_histo(res, bottom, height, left, mv, b_curr);

	/*CHBank* b = get_bank(bank);
	unsigned m_bar_num = b->size();
	double mv = b->max_val();

	//double y_curr = bottom;
	//long* p_value = m_values;
	double x_curr = left;
	double x_step = (.98-left)/m_bar_num;
	add_val(res, x_curr);
	add_val(res, bottom);
	res->AddString("m\n");
	unsigned i;
	for (i = 0; i<m_bar_num; i++) {
		add_val(res, x_curr);
		double h_curr = bottom+height*b->value(i)/mv;
		add_val(res, h_curr);
		res->AddString("L\n");

		x_curr += x_step;
		add_val(res, x_curr);
		add_val(res, h_curr);
		res->AddString("L\n");
		//add_val(res, x_curr);
		//add_val(res, bottom);
		//res->AddString("L\ns\n");
		
		//res->AddString("\n0 .01472 L");
		//p_value += sizeof(long);
	}
	add_val(res, x_curr);
	add_val(res, bottom);
	res->AddString("L\ns\n");*/

	res->AddString("s\n");

	res->AddString("% End of Graphics");
	res->AddString("\nMathPictureEnd\n\\\n");
	dbg->log(20, bank, "Output generated for bank# %d");
	//return new CEString("No histogram serialization implemented.");
	return res;
}

void CEHistogram::add_subscriptions(CEString *res,
									double bottom,
									double height,
									double left,
									double mv,
									int bank)
{
	if (res == NULL) {
		dbg->log(1, bank, "Error generating subscriptions for bank# %d: NULL output string");
		return;
	}
	dbg->log(30, bank, "Generating h-subscriptions for bank# %d");
	//0x
	int num_vals = 5;
	CHBank* b = get_bank(bank);
	unsigned m_bar_num = b->size();
	double l = b->margines()[0];
	double r = b->margines()[m_bar_num];
	double val_max = fabs(l);
	int i;
	for(i=1; i<num_vals; i++) {
		double curr_val;
		if (i==num_vals) curr_val = fabs(r);
		else curr_val = fabs(l+(r-l)*i/(num_vals-1));
		if (curr_val>val_max) val_max = curr_val;
	}
	int ox_pow = CLMath::get_power(val_max);
	double x_pos = left;
	double x_step = (.98-left)/(num_vals-1);
	dbg->log(10, (int)(val_max*100), "val_max = %d");
	dbg->log(10, (int)(ox_pow), "ox_pow = %d");
	dbg->log(10, (int)(x_pos*100), "x_pos = %d");
	dbg->log(10, (int)(x_step*100), "x_step = %d");
	double f_ox = (ox_pow == 1)?1:(1/pow((double)10, ox_pow));
	dbg->log(10, (int)(f_ox*100), "f_ox = %d");
	double x_curr_val = l*f_ox;
	double x_disp_step = ((r-l)/(num_vals-1))*f_ox;
	for (i=0; i<num_vals; i++) {
		//double curr_val;
		//if (i==num_vals) curr_val = r*f_ox;
		//else curr_val = (l+(r-l)*i/(num_vals-1))*f_ox;
		res->AddString("[(");
		add_val2(res, x_curr_val);
		res->AddString(")] ");
		/*if (i == (num_vals-1)) {
			add_val(res, 0.99);
		}
		else {
			add_val(res, 0.01+((double)i)*0.95/(num_vals-1));
		}*/
		add_val(res, x_pos + 0.01);
		res->AddString(".005 0 1 Mshowa\n");
		
		add_val(res, x_pos);
		add_val(res, bottom*.7);
		res->AddString("m\n");
		add_val(res, x_pos);
		add_val(res, bottom*1.1);
		res->AddString("L\ns");

		x_pos += x_step;
		x_curr_val += x_disp_step;
	}
	if (ox_pow != 0) {
		res->AddString("[(*10^");
		add_val2(res, ox_pow);
		res->AddString(")] ");
		add_val(res, 0.98);
		add_val(res, bottom*1.2);
		res->AddString("0 1 Mshowa\n");
	}

	dbg->log(30, bank, "Generating v-subscriptions for bank# %d");
	dbg->log(10, (int)(100*mv), "mv = %d");
	dbg->log(10, (int)(100*height), "height = %d");
	//0y
	num_vals = 5;
	//CHBank* b = get_bank(bank);
	val_max = mv;//b->max_val();
	int oy_pow = CLMath::get_power(val_max);
	dbg->log(10, (int)(oy_pow*100), "oy_pow = %d");
	//val_max /= (10^(oy_pow));
	val_max /= pow((double)10, oy_pow);
	dbg->log(10, (int)(val_max*100), "val_max = %d");
	int val = (int)(val_max*num_vals*10);
	dbg->log(10, (int)(val*100), "val = %d");
	double disp_mv = ((double)val)/(num_vals*10);
	dbg->log(10, (int)(disp_mv*100), "disp_mv = %d");
	double h_step = height*disp_mv/(val_max*(num_vals-1));
	double h_pos = bottom;
	double disp_step = disp_mv/(num_vals-1);
	double curr_val = 0;
	for (i=0; i<num_vals; i++) {
		res->AddString("[(");
		add_val2(res, curr_val);
		res->AddString(")] ");

		//height
		add_val(res, left/2);
		add_val(res, h_pos + 0.02);
		res->AddString("0 1 Mshowa\n");

		add_val(res, left*.8);
		add_val(res, h_pos);
		res->AddString("m\n");
		add_val(res, left);
		add_val(res, h_pos);
		res->AddString("L\ns");

		h_pos += h_step;
		curr_val += disp_step;
	}
	if (oy_pow != 0) {
		res->AddString("[(");
		add_val2(res, oy_pow);
		res->AddString(")] ");
		add_val(res, left*3/2);
		add_val(res, height+.03);
		res->AddString("0 1 Mshowa\n");
	}
}

/*double* CEHistogram::values(int bank, bool create_bank)
{
	int s = m_a_values->GetSize();
	if (bank < s) return (double*)m_a_values->GetAt(bank);
	if (!create_bank) return NULL;
	double* vals;
	int i;
	for(i=s; i<=bank; i++) {
		vals = new double[m_bar_num];
		unsigned j;
		for (j = 0; j<m_bar_num; j++) {
			vals[j] = 0;
		}
		m_a_values->Add(vals);
	}
	return vals;
}*/

CHBank* CEHistogram::get_bank(int num, bool create_bank)
{
	int s = m_a_values->size();
	if (num < s) return (CHBank*)m_a_values->GetAt(num);
	if (!create_bank) return NULL;
	CHBank* bank0 = get_bank(0, false);
	CHBank* bank;
	int i;
	for(i=s; i<=num; i++) {
		bank = bank0->create_frame();
		m_a_values->add(bank);
	}
	return bank;
}

void CEHistogram::add_histo(CEString *res, double bottom, double height, double left, double mv, CHBank* bank)
{
	dbg->log(20, "Visualizing histo data");
	unsigned m_bar_num = bank->size();
	double x_curr = left;
	double x_step = (.98-left)/m_bar_num;
	add_val(res, x_curr);
	add_val(res, bottom);
	res->AddString("m\n");
	unsigned i;
	for (i = 0; i<m_bar_num; i++) {
		add_val(res, x_curr);
		double h_curr = bottom+height*bank->value(i)/mv;
		add_val(res, h_curr);
		res->AddString("L\n");

		x_curr += x_step;
		add_val(res, x_curr);
		add_val(res, h_curr);
		res->AddString("L\n");
	}
	add_val(res, x_curr);
	add_val(res, bottom);
	res->AddString("L\n");
}

void CEHistogram::add_graph(CEString *res, double bottom, double height, double left, double mv, CHBank* bank)
{
	dbg->log(20, "Visualizing histo data");
	unsigned m_bar_num = bank->size();
	double x_step = (.98-left)/m_bar_num;
	double x_curr = left + x_step/2;
	double y_curr = bottom+height*bank->value(0)/mv;
	add_val(res, x_curr);
	add_val(res, y_curr);
	res->AddString("m\n");
	unsigned i;
	for (i = 1; i<m_bar_num; i++) {
		x_curr += x_step;
		add_val(res, x_curr);
		y_curr = bottom+height*bank->value(i)/mv;
		add_val(res, y_curr);
		res->AddString("L\n");
	}
}

double CEHistogram::max_value_in_banks()
{
	double res = 0;
	unsigned i;
	for (i=0;;i++) {
		CHBank* b = get_bank(i);
		if (b == NULL) {
			if (res == 0) dbg->log(1, "Warning: no valid and normalized data found");
			return res;
		}
		if (b->is_normalized()) {
			double val = b->max_val();
			if (val > res) res = val;
		}
	}
}

void CEHistogram::add_bank(int bank, int dest_bank)
{
	CHBank* b = get_bank(bank);
	CHBank* b_dest = get_bank(dest_bank);
	if ((b == NULL) || (b_dest == NULL)) return;
	b_dest->add_values_from(b);
}

void CEHistogram::normalize_all()
{
	CHBank* b_curr = get_bank(0);
	unsigned i;
	for (i=0;(b_curr != NULL);) {
		if (!b_curr->is_normalized()) b_curr->normalize();
		i++;
		b_curr = get_bank(i);
	}
}
