////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EMathEquRule.cpp: implementation of the CEMathEquRule class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CRuleCondition::CRuleCondition() : CEObject(true) {}
const char* CRuleCondition::CLASS_NAME = "CRuleCondition";
const char* CRuleCondition::ClassName() const { return CLASS_NAME; }
bool CRuleCondition::contain_vector(void * obj) { return false; }
CEObject* CRuleCondition::copy() const { return clone(); }

CRuleConditionEquZero::CRuleConditionEquZero(bool bEquZero, CExpression* var) : CRuleCondition() {
	m_bEquZero=bEquZero; m_var=((var!=NULL)?var->clone():NULL);
}
CRuleConditionEquZero::~CRuleConditionEquZero() { if (m_var!=NULL) delete m_var; }
const char* CRuleConditionEquZero::CLASS_NAME = "CRuleConditionEquZero";
const char* CRuleConditionEquZero::ClassName() const { return CLASS_NAME; }
bool CRuleConditionEquZero::contain_vector(void * obj) {
	if (obj==NULL) return false;
	if (m_var!=NULL) { if ((obj==m_var) || (m_var->contain_vector(obj))) return true; }
	return false;
}
char* CRuleConditionEquZero::toLine(int mode) {
	CEString res("(");
	if (m_var!=NULL) res.AddString(m_var->toString(mode)); else res.AddString("NULL");
	if (m_bEquZero) { res.AddString("==0"); } else { res.AddString("!=0"); }
	res.AddString(")");
	return CEUtil::copyString(res.getData());
}
CRuleCondition* CRuleConditionEquZero::clone() const { return new CRuleConditionEquZero(m_bEquZero,m_var); }
bool CRuleConditionEquZero::FitsCondition(CERule* r) {
	if (r==NULL) return false;
	if (m_var==NULL) return false;
	CExpression* var_cpy=CExpression::arrge(r->ApplyTo(m_var->clone()));
	bool res=false;
	if (CEBasic::isZero(var_cpy)) { if (m_bEquZero) res=true; } else { if (!m_bEquZero) res=true; }
	delete var_cpy;
	//if (!res) { CERule::dbg->log(2,"CRuleConditionEquZero::FitsCondition FALSE",this,r); }
	return res;
}

const int CRuleConditionAndOr::COND_AND	=1;
const int CRuleConditionAndOr::COND_OR	=2;
CRuleConditionAndOr::CRuleConditionAndOr(int type) : CRuleCondition() { m_type=type; }
CRuleConditionAndOr::~CRuleConditionAndOr() {}
void CRuleConditionAndOr::AddCondition(const CRuleCondition* c) { m_Conditions.add(c->clone()); }
int CRuleConditionAndOr::GetType() { return m_type; }
const char* CRuleConditionAndOr::CLASS_NAME = "CRuleConditionAndOr";
const char* CRuleConditionAndOr::ClassName() const { return CLASS_NAME; }
bool CRuleConditionAndOr::contain_vector(void * obj) {
	if (obj==NULL) return false;
	if ((obj==&m_Conditions) || (m_Conditions.contain_vector(obj))) return true;
	return false;
}
char* CRuleConditionAndOr::toLine(int mode) {
	CEString res("(");
	long s=m_Conditions.size();
	for(long i=0; i<s; i++) {
		if (i>0) {
			if (m_type == CRuleConditionAndOr::COND_AND) { res.AddString(" && "); }
			else if (m_type == CRuleConditionAndOr::COND_OR) { res.AddString(" || "); }
			else { res.AddString(" ?? "); }
		}
		CRuleCondition* c=(CRuleCondition*)m_Conditions.GetAt(i);
		if (c != NULL) {
			char* tmp=c->toLine(mode); res.AddString(tmp); CMemoryManager::mfree(tmp);
		} else {
			res.AddString("NULL");
		}
	}
	res.AddString(")");
	return CEUtil::copyString(res.getData());
}
CRuleCondition* CRuleConditionAndOr::clone() const {
	CRuleConditionAndOr* res = new CRuleConditionAndOr(m_type);
	long s=m_Conditions.size();
	for(long i=0; i<s; i++) { res->AddCondition((const CRuleCondition*)m_Conditions.GetAtConst(i)); }
	return res;
}
bool CRuleConditionAndOr::FitsCondition(CERule* r) {
	if (r==NULL) return false;
	long s=m_Conditions.size();
	bool res=false;
	if (m_type == CRuleConditionAndOr::COND_AND) { res=true; }
	else if (m_type == CRuleConditionAndOr::COND_OR) { res=false; }
	else return false;
	for(long i=0; i<s; i++) {
		CRuleCondition* c=(CRuleCondition*)m_Conditions.GetAt(i); if(c==NULL) { CEException::halt(NULL,__FILE__,__LINE__); continue; }
		bool bFits=c->FitsCondition(r);
		if (m_type == CRuleConditionAndOr::COND_AND) {
			if (!bFits) { res=false; break; }
		} else if (m_type == CRuleConditionAndOr::COND_OR) {
			if (bFits) { res=true; break; }
		} else { res=false; break; }
	}
	//if (!res) { CERule::dbg->log(2,"CRuleConditionAndOr::FitsCondition FALSE",this,r); }
	return res;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
/*
*	CENull-symbols mean 'any unused value', originates from "___"-parameters.
*/
CEMathEquRule::CEMathEquRule(CExpression* left, CExpression* val) : CEObject(true) {
	m_Left = ((left!=NULL)?left->clone():NULL);
	m_Value = ((val!=NULL)?val->clone():NULL);

	if (m_Left != NULL) {
		if (!m_Left->getFlag()) { m_Left->inverseFlag(); if(m_Value!=NULL)m_Value->inverseFlag(); }
		if (!m_Left->getLevel()) { m_Left->inverseLevel(); if(m_Value!=NULL)m_Value->inverseLevel(); }

		m_Left=CEMathEquRule::ConvertTemplateLeft(m_Left, &m_TempSyms);
	}
	m_Conditions=NULL;
}

CEMathEquRule::~CEMathEquRule() {
	if (m_Left!=NULL) delete m_Left;
	if (m_Value!=NULL) delete m_Value;
	if (m_Conditions!=NULL) delete m_Conditions;
}

void CEMathEquRule::AddCondition(const CRuleCondition* c) {
	if (c==NULL) return;
	if (m_Conditions==NULL) { m_Conditions=new CRuleConditionAndOr(CRuleConditionAndOr::COND_AND); }
	m_Conditions->AddCondition(c);
}

bool CEMathEquRule::IsSimilarTo(CExpression* left) { // equals()
	if (left == m_Left) return true; // NULL==NULL-check
	if ((left == NULL) || (m_Left==NULL)) return false;
	if (!left->getFlag()) { left->inverseFlag(); bool res=IsSimilarTo(left); left->inverseFlag(); return res; }
	if (!left->getLevel()) { left->inverseLevel(); bool res=IsSimilarTo(left); left->inverseLevel(); return res; }

	left=left->clone();

	CSequence _TempSyms;
	left=CEMathEquRule::ConvertTemplateLeft(left, &_TempSyms);

	bool res=false;
	if ((left->equals(m_Left)) && (_TempSyms.equals(&m_TempSyms))) res=true;
	delete left;
	return res;
}

CExpression* CEMathEquRule::ConvertTemplateLeft(CExpression* left, CSequence* _TempSyms) {
	if ((left==NULL) || (_TempSyms==NULL)) return left;
	long s; long i; CEReplList repl;

	CRCollectSymbols r(0,NULL,CESymbol::CLASS_NAME);
	r.ApplyTo(left,false);
	CEPtrArray* syms_list = r.getCollectedSymbols();
	if (syms_list!=NULL) {
		s=syms_list->size();
		for(i=0; i<s; i++) {
			CExpression* itemE=(CExpression*)syms_list->GetAt(i);
			if ((itemE==NULL) || (itemE->ClassName() != CESymbol::CLASS_NAME)) { CEException::halt("INTERNAL ERROR in CEMathEquRule::CEMathEquRule"); continue; }
			CESymbol* item = (CESymbol*)itemE;
			if (!item->getFlag()) item->setFlag();
			if (!item->getLevel()) item->setLevel(true);
			
			const char* s_sym=item->get_name(); if(s_sym==NULL) continue;
			int s_symL=CEUtil::length(s_sym);
			
			int lastSpace=s_symL-1;
			for (;lastSpace>=0;lastSpace--) { if (*(s_sym+lastSpace) != '_') break; }
			if (lastSpace == s_symL-1) continue;
			
			if (lastSpace>=0) {
				char* s_sub=CEUtil::copyStringTo(s_sym,lastSpace);
				if (s_sub==NULL) { CEException::halt("INTERNAL ERROR in CEMathEquRule::CEMathEquRule. Memalloc failed."); continue; }
				CESymbol* sub = new CESymbol(s_sub);
				repl.add(item, sub, false);
				_TempSyms->addItem(sub);
				CMemoryManager::mfree(s_sub);
			} else {
				CENull sub;
				repl.add(item, &sub, false);
			}
		}
	}
	
	// check to no repl-symbols occur in rule before
	// i.e.: Rule[a_,a] should declare warning
	for (i=0; i<_TempSyms->size(); i++) {
		if (CRContainCheck::search(left, _TempSyms->GetAt(i))) {
			CERule::dbg->log(0, "Math-Rule WARNING: template parameter appear also as non-template", left);
			CERule::dbg->log(0, "Math-Rule WARNING: suspected parameter treated as template further", _TempSyms->GetAt(i));
		}
	}
	
	left=repl.ApplyTo(left);
	return left;
}

const char* CEMathEquRule::CLASS_NAME = "CEMathEquRule";
const char* CEMathEquRule::ClassName() const { return CLASS_NAME; }

bool CEMathEquRule::contain_vector(void * obj) {
	if (obj==NULL) return false;
	if (m_Left!=NULL) { if((m_Left==obj) || (m_Left->contain_vector(obj))) return true; }
	if (m_Value!=NULL) { if((m_Value==obj) || (m_Value->contain_vector(obj))) return true; }
	if((&m_TempSyms==obj) || (m_TempSyms.contain_vector(obj))) return true;
	if (m_Conditions!=NULL) { if((m_Conditions==obj) || (m_Conditions->contain_vector(obj))) return true; }
	return false;
}

char* CEMathEquRule::toLine(int mode) {
	CEString res("{");
	if (m_Left!=NULL) res.AddString(m_Left->toString(mode)); else res.AddString("NULL");
	res.AddString(":=");
	if (m_Value!=NULL) res.AddString(m_Value->toString(mode)); else res.AddString("NULL");
	char* tmp=NULL;
	res.AddString("; TempSyms: "); tmp=m_TempSyms.toLine(mode); if(tmp!=NULL){ res.AddString(tmp); CMemoryManager::mfree(tmp); }
	if (m_Conditions != NULL) {
		res.AddString("; Cond: "); tmp=m_Conditions->toLine(mode); if(tmp!=NULL){ res.AddString(tmp); CMemoryManager::mfree(tmp); }
	}
	res.AddString("}");
	return CEUtil::copyString(res.getData());
}

bool CEMathEquRule::IsMathTemplateLeft(CExpression* left) {
	//CEBasic::dbg->log(2,"CEMathEquRule::IsMathTemplateLeft",left);
	if (left==NULL) return false;
	CRCollectSymbols r(0,NULL,CESymbol::CLASS_NAME);
	r.ApplyTo(left,false);
	CEPtrArray* syms_list = r.getCollectedSymbols();
	if (syms_list==NULL) return false;

	bool res=false;
	long s=syms_list->size();
	for(long i=0; i<s; i++) {
		CExpression* itemE=(CExpression*)syms_list->GetAt(i);
		if ((itemE==NULL) || (itemE->ClassName() != CESymbol::CLASS_NAME)) { CEException::halt("INTERNAL ERROR in CEMathEquRule::IsMathTemplateLeft"); continue; }
		const char* s_sym=((CESymbol*)itemE)->get_name();
		if(s_sym==NULL) continue;
		int s_symL=CEUtil::length(s_sym); if (s_symL==0) continue;
		if (*(s_sym+s_symL-1) == '_') { res=true; break; }
	}
	//CEBasic::dbg->log(2, res, "CEMathEquRule::IsMathTemplateLeft res: %ld", syms_list); //CEException::halt();
	return res;
}

bool CEMathEquRule::IsApplicableTo(CExpression* src) { // unaffects source
	if (src==NULL) return false;
	//CEBasic::dbg->log(2,"CEMathEquRule::IsApplicableTo",src,m_Left);
	CSequence repls;
	long i; long s=m_TempSyms.size();
	for(i=0; i<s; i++) { repls.addItem(new CENull()); }
	bool res = FitsLeft(m_Left, src, &repls);
	if ((res) && (m_Conditions != NULL)) { // also check conditions
		CEReplList r;
		for(i=0; i<s; i++) {
			CExpression* tmpl=m_TempSyms.GetAt(i); CExpression* subst=repls.GetAt(i);
			if (subst->ClassName() != CENull::CLASS_NAME) { r.add(tmpl,subst,false); }
		}
		res = m_Conditions->FitsCondition(&r);
	}
	//CEBasic::dbg->log(2,res,"CEMathEquRule::IsApplicableTo result: %ld",&repls);
	return res;
}

/*
*	deletes or unaffects source
*	returns src or cloned result
*/
CExpression* CEMathEquRule::ApplyTo(CExpression* src, bool del_src) {
	if (src==NULL) return src;

	// 1. Check src fits a template
	CSequence repls;
	long i; long s=m_TempSyms.size();
	for(i=0; i<s; i++) { repls.addItem(new CENull()); }
	if (!FitsLeft(m_Left, src, &repls)) return src;

	// 2. Create repl rule
	CEReplList r;
	for(i=0; i<s; i++) {
		CExpression* tmpl=m_TempSyms.GetAt(i);
		CExpression* subst=repls.GetAt(i);
		if (subst->ClassName() != CENull::CLASS_NAME) { r.add(tmpl,subst,false); }
	}

	//bool bLog=false;
	// 3. Check additional conditions
	if (m_Conditions != NULL) { // also check conditions
		if (!m_Conditions->FitsCondition(&r)) return src;
		//bLog=true;
	}

	//if (bLog) { CERule::dbg->log(2,"********** this:",this); CERule::dbg->log(2,"src:",src); }
	if (del_src) delete src;
	CExpression* res = m_Value;
	if (res != NULL) res=r.ApplyTo(res->clone());
	//if (bLog) CERule::dbg->log(2,"res:",res);
	return res;
}

/*
*	Todo: implememt it as common engine
*	Now works for: CESymbol, CSequence, CEFunction, CESum, CEPower, CEProduct, CENumber
*/
bool CEMathEquRule::FitsLeft(CExpression* left, CExpression* src, CSequence* repls) {
	if ((left==NULL) || (src==NULL)) return false;
	long i; long s;

	if (left->ClassName() == CSequence::CLASS_NAME) {
		if (!left->attributes_equals_to(src)) return false;
		if (src->ClassName() != CSequence::CLASS_NAME) return false;
		s=((CSequence*)left)->size();
		if (s != ((CSequence*)src)->size()) return false;
		for (i=0; i<s; i++) {
			if (!FitsLeft(((CSequence*)left)->GetAt(i), ((CSequence*)src)->GetAt(i), repls)) return false;
		}
		return true;
	}

	if (left->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		if (!left->attributes_equals_to(src)) return false;
		if (!src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return false;
		s=((CESum*)left)->size();
		if (s != ((CESum*)src)->size()) return false;
		for (i=0; i<s; i++) {
			if (!FitsLeft(((CESum*)left)->GetAt(i), ((CESum*)src)->GetAt(i), repls)) return false;
		}
		return true;
	}

	if (left->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		if (!left->attributes_equals_to(src)) return false;
		if (!src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) return false;
		s=((CEProduct*)left)->size();
		if (s != ((CEProduct*)src)->size()) return false;
		for (i=0; i<s; i++) {
			if (!FitsLeft(((CEProduct*)left)->GetAt(i), ((CEProduct*)src)->GetAt(i), repls)) return false;
		}
		return true;
	}

	if (left->ClassName() == CEFunction::CLASS_NAME) {
		if (!left->attributes_equals_to(src)) return false;
		if (src->ClassName() != CEFunction::CLASS_NAME) return false;
		if (!CEUtil::equals( ((CEFunction*)left)->getTitle(), ((CEFunction*)src)->getTitle() )) return false;
		s=((CEFunction*)left)->params_num();
		if (s != ((CEFunction*)src)->params_num()) return false;
		for (i=0; i<s; i++) {
			if (!FitsLeft(((CEFunction*)left)->get_param(i), ((CEFunction*)src)->get_param(i), repls)) return false;
		}
		return true;
	}

	if (left->ClassName() == CEPower::CLASS_NAME) {
		if (!left->attributes_equals_to(src)) return false;
		if (src->ClassName() != CEPower::CLASS_NAME) return false;
		if (!FitsLeft(((CEPower*)left)->basis(), ((CEPower*)src)->basis(), repls)) return false;
		if (!FitsLeft(((CEPower*)left)->power(), ((CEPower*)src)->power(), repls)) return false;
		return true;
	}

	// compare symbols
	if (left->equals(src)) return true;

	if (left->ClassName() == CENull::CLASS_NAME) return true; // any subst

	// check template
	if (left->ClassName() != CESymbol::CLASS_NAME) return false;

	// allowed sign at both left and src: a_ -> -b, 1/a -> 1/b
	if (!left->getFlag()) {
		left=left->clone(); left->inverseFlag(); if (!left->getFlag()) CEException::halt("CEMathEquRule::FitsLeft: unexpected sign");
		src=src->clone(); src->inverseFlag();
		bool res=FitsLeft(left, src, repls);
		delete left; delete src;
		return res;
	}
	if (!left->getLevel()) {
		left=left->clone(); left->inverseLevel(); if (!left->getLevel()) CEException::halt("CEMathEquRule::FitsLeft: unexpected denominator");
		src=src->clone(); src->inverseLevel();
		bool res=FitsLeft(left, src, repls);
		delete left; delete src;
		return res;
	}

	s=m_TempSyms.size(); long templI=-1;
	for(i=0; i<s; i++) { if (left->equals(m_TempSyms.GetAt(i))) { templI=i; break; } }
	if (templI<0) return false;

	CExpression* prev_subst = repls->GetAt(templI);
	if (prev_subst->ClassName() != CENull::CLASS_NAME) {
		if (!src->equals(prev_subst)) return false;
		return true;
	}
	repls->SetAt(templI,src);
	return true;
}
