////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// ENumVector.cpp: implementation of the CENumVector class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDebug* CENumVector::dbg = new CDebug("CENumVector");

#ifdef USE_PLAIN_ARRAY_FORSEQ // biuld sequence-expressions on CAPtrArray* basis
CENumVector::CENumVector(CEPtrArray* pPtrArray, CEValue* _mass):CSequence(pPtrArray) {
	m_mass = _mass;
	//if (_mass != NULL) m_mass = _mass->clone();
}
#else
CENumVector::CENumVector(CAPtrSeq* pPtrArray, CEValue* _mass):CSequence(pPtrArray) {
	m_mass = _mass;
	//if (_mass != NULL) m_mass = _mass->clone();
}
#endif //USE_PLAIN_ARRAY_FORSEQ

CENumVector::CENumVector(CEValue* _mass)
						 :CSequence()
{
	m_mass = _mass;
	//if (_mass != NULL) m_mass = _mass->clone();
}

CENumVector::~CENumVector()
{
	if (m_mass != NULL) delete m_mass;
}

CExpression* CENumVector::clone() const
{
	CENumVector* res = new CENumVector(
		((m_mass == NULL)?m_mass:(CEValue*)m_mass->clone()));
	int s = size();
	int i;
	for (i=0; i<s; i++) {
		res->addItem(GetAtConst(i)->clone());
	}
	copyAttributesTo(res);
	return res;
}

void CENumVector::addvector(CENumVector * other, bool delete_other)
{
	int s = size();
	bool sign = (getFlag() == other->getFlag());
	int i;
	for (i=0; i<s; i++) {
		CExpression* item = GetAt(i);
		CExpression* o_item = other->GetAt(i)->clone();
		if (!sign) o_item->inverseFlag();
		CExpression* temp = item->add(o_item);
		if (temp != item) {
			SetAt(i, temp);
			delete temp;
		}
	}
	if (m_mass != NULL) {
		delete m_mass;
		m_mass = NULL;
	}
	if (delete_other) delete other;
}

void CENumVector::substvector(CENumVector* other, bool delete_other) {
	other->inverseFlag();
	addvector(other, delete_other);
	if (!delete_other) other->inverseFlag();
}

/*
* The metric supposed to be diag{1,-1,-1,-1}.
*/
CExpression* CENumVector::couple(CENumVector * other, bool delete_other)
{
	bool sign = (getFlag() == other->getFlag());
	CExpression* item = GetAt(0)->clone();
	CExpression* o_item = other->GetAt(0)->clone();
	CExpression* res = item->multiply(o_item);
	int s = size();
	int i;
	for (i=1; i<s; i++) {
		item = GetAt(i)->clone();
		o_item = other->GetAt(i)->clone();
		CExpression* temp = item->multiply(o_item);
		temp->inverseFlag();
		res = res->add(temp);
	}
	if (!sign) res->inverseFlag();
	if (delete_other) delete other;
	return res;
}

CExpression* CENumVector::squarevector()
{
	if (m_mass != NULL) return m_mass->clone()->multiply(m_mass->clone());
	return couple(this, false);
}

CEValue* CENumVector::mass()
{
	CEValue* res;
	if (m_mass != NULL)
		res = (CEValue*)m_mass->clone();
	else {
		CEValue* sq = (CEValue*)squarevector();
		res = sq->Sqrtx();
	}
	return res;
}

CENumVector* CENumVector::num_stub()
{
	CENumVector* res = new CENumVector();
	int s = size();
	int i;
	for (i=0; i<s; i++) {
		res->addItem(GetAt(i)->clone());
	}
	copyAttributesTo(res);
	return res;
}

CEValue* CENumVector::square()
{
	CExpression* res = squarevector();
	if (res->getProperty(CEProperties::PROP_MASK_IS_NUMERIC_VALUE))
		return (CEValue*)res;
	dbg->log(1, "Error: Vector square is not numeric.");
	return NULL;
}

void CENumVector::copy_values_from(CENumVector * other)
{
	if (other == NULL) return;
	clear_all();
	int s = other->size();
	int i;
	for (i=0; i<s; i++) {
		addItem(other->GetAt(i)->clone());
	}
	other->copyAttributesTo(this);
}

CEString* CENumVector::toString(int mode)
{
	CEString* res  = new CEString("\\(");
	if (!getFlag()) res->AddString("\\((- 1)\\) ");
	res->AddString("{");
	int s = size();
	int i;
	for (i=0; i<s; i++) {
		if (i != 0) res->AddString(", ");
		res->AddString(GetAt(i)->toString(mode));
	}
	res->AddString("}\\ {");
	if (m_mass != NULL) res->AddString(m_mass->toString(mode));
	res->AddString("}");
	res->AddString("\\)");
	return res;
}

double CENumVector::angle_cos(CENumVector *other)
{
	int s = size();
	if ((other == NULL) || (s == 0) || (other->size() != s)) {
		dbg->log(1, "Error calculating angle (1)");
		return 0;
	}
	CExpression* res = new CERealValue(0);
	CExpression* t1 = new CERealValue(0);
	CExpression* t2 = new CERealValue(0);
	int i;
	for(i=1; i<s; i++) {
		CExpression* a = GetAt(i);
		CExpression* b = other->GetAt(i);
		if ((a->getProperty(CEProperties::PROP_MASK_IS_NUMERIC_VALUE)) &
			(b->getProperty(CEProperties::PROP_MASK_IS_NUMERIC_VALUE))) {
			//CEValue* a = (CEValue*)expr_a;
			//CEValue* b = (CEValue*)expr_b;
			res->add(a->clone()->multiply(b->clone()));
			t1->add(a->clone()->multiply(a->clone()));
			t2->add(b->clone()->multiply(b->clone()));
		}
		else {
			delete res;
			delete t1;
			delete t2;
			dbg->log(1, "Error calculating angle (2)");
			return 0;
		}
	}
	res = res->divide(((CEValue*)t1)->Sqrtx()->multiply(((CEValue*)t2)->Sqrtx()));
	double cos_val = ((CEValue*)res)->to_double();
	delete res;
	if ((cos_val>1) || (cos_val<-1)) {
		dbg->log(1, "Error calculating angle (3)");
	}
	return cos_val;
}


/*
*      ex-SUBROUTINE LOR0K(R,RS,K,M)
*
*   OPEH PEOPAOBAHE 4-BEKTOPA.
*   R(4)  - 4-BEKTOP B CXOHO CCTEME OTCETA
*   RS(4) - TOT E BEKTOP B HO CCTEME OTCETA
*   K(4)  - BEKTOP-APAMETP PEOPAOBAH,
*	      MEET CMC 4-MCA ACT
*	      C MACCO	M :
*
*	  OCECTBETC PEOPAOBAHE BEKTOPA
*	  R  CCTEM, E TA ACTA OKOTC,
*	  B CCTEM, E OH PABEH K.
*
*	Warning: no precision checkout performed.
*/
void CENumVector::lor0k(CENumVector * k)
{
	if (k->is_rest()) return;
	k->inverse_direction();
	lork0(k);
	k->inverse_direction();
	/*
	REAL EPS/Z3B800000/
	REAL R(4),RS(4),K(4),M
	RS(4)=(R(1)*K(1)+R(2)*K(2)+R(3)*K(3)+R(4)*K(4))/M
	S=(R(4)+RS(4)) / (K(4)+M)
	QEPS=2.*EPS*RS(4)
	DO 1 J=1,3
	Q=R(J)+K(J)*S
	IF(Q.LT.QEPS .AND. Q.GT.-QEPS)Q=0.
	1 RS(J)=Q
	RETURN
	END
	*/
}


/*
*     ex-SUBROUTINE LORK0(R,RS,K,M)
*
*   OPEH PEOPAOBAHE 4-BEKTOPA.
*   R(4)  - 4-BEKTOP B CXOHO CCTEME OTCETA
*   RS(4) - TOT E BEKTOP B HO CCTEME OTCETA
*   K(4 ) - BEKTOP-APAMETP PEOPAOBAH,
*	      MEET CMC 4-MCA ACT
*	      C MACCO	M :
*
*	  OCECTBETC PEOPAOBAHE BEKTOPA
*	  R  CCTEM, E MC ACT PABEH
*	  K,  B CCTEM, E OHA OKOTC.
*
*	Warning: no precision checkout performed.
*/
void CENumVector::lork0(CENumVector * k)
{
	int s = size();
	if ((k == NULL) || (s != 4) || (k->size() != s)) {
		dbg->log(1, "Error. Loretz transformation: invalid data.");
		//delete k;
		return;
	}
	CEValue* m = k->mass();
	CExpression* E_old = GetAt(0)->clone();
	CExpression* E_new = couple(k, false)->divide(m->clone());
	SetAt(0, E_new);
	CExpression* k0 = k->GetAt(0)->clone();
	CExpression* sf = E_old->add(E_new)->divide(m->add(k0));
	int i;
	for (i=1; i<s; i++) {
		CEValue* item = (CEValue*)GetAt(i);
		CEValue* temp = (CEValue*)k->GetAt(i)->clone()->multiply(sf->clone());
		temp->inverseFlag();
		item->add(temp);
	}
	delete sf;
	//delete k;
	/*
	REAL R(4),RS(4),K(4),M
	RS(4)=(R(4)*K(4)-R(3)*K(3)-R(2)*K(2)-R(1)*K(1))/M
	QEPS=2.*EPS*RS(4)
	S=(R(4)+RS(4)) / (K(4)+M)
	DO 1 J=1,3
	Q=R(J)-K(J)*S
	IF(Q.LT.QEPS .AND. Q.GT.-QEPS)Q=0.
	1 RS(J)=Q
	RETURN
	END
	*/
}

void CENumVector::inverse_direction()
{
	int s = size();
	int i;
	for (i=1; i<s; i++) {
		GetAt(i)->inverseFlag();
	}
}

bool CENumVector::is_rest() {
	int s = size();
	int i;
	for (i=1; i<s; i++) {
		if(!CEBasic::isZero(GetAt(i))) return false;
	}
	return true;
}

void CENumVector::set_mass(CEValue * m)
{
	if (m_mass != m) {
		if (m_mass != NULL) delete m_mass;
		m_mass = m;
	}
}
