////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// EParticle.cpp: implementation of the CEParticle class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const long CEParticle::PARTICLE_FERMION				= 0x200000; // flag
const long CEParticle::PARTICLE_LEPTON				= 0x210000; // flag
const long CEParticle::PARTICLE_ELECTRON			= 0x210001;
const long CEParticle::PARTICLE_MUON				= 0x210002;
const long CEParticle::PARTICLE_TAU_LEPTON			= 0x210003;

const long CEParticle::PARTICLE_NEUTRINO			= 0x220000; // flag
const long CEParticle::PARTICLE_NEUTRINO_ELECTRON	= 0x220001;
const long CEParticle::PARTICLE_NEUTRINO_MUON		= 0x220002;
const long CEParticle::PARTICLE_NEUTRINO_TAU		= 0x220003;

const long CEParticle::PARTICLE_QUARK				= 0x240000; // flag or abstract quark
const long CEParticle::PARTICLE_QUARK_CONJUGETED	= 0x2c0000; // flag or abstract quark ("conjugated": d,s,b quarks)
const long CEParticle::PARTICLE_QUARK_U				= 0x240001; // u, I3=+1/2, Q=2/3
const long CEParticle::PARTICLE_QUARK_C				= 0x240002; // c, I3=+1/2, Q=2/3
const long CEParticle::PARTICLE_QUARK_T				= 0x240003; // t, I3=+1/2, Q=2/3
const long CEParticle::PARTICLE_QUARK_D				= 0x2c0001; // d, I3=-1/2, Q=-1/3, "conjugated"
const long CEParticle::PARTICLE_QUARK_S				= 0x2c0002; // s, I3=-1/2, Q=-1/3, "conjugated"
const long CEParticle::PARTICLE_QUARK_B				= 0x2c0003; // b, I3=-1/2, Q=-1/3, "conjugated"

const long CEParticle::PARTICLE_PROTON				= 0x200101; // Abstract id used for srtucture functions approach
const long CEParticle::PARTICLE_NEUTRON				= 0x200102; // Abstract id used for srtucture functions approach

const long CEParticle::PARTICLE_MODEL_FERMION		= 0xa00000; // flag

const long CEParticle::PARTICLE_BOSON				= 0x100000; // flag
const long CEParticle::PARTICLE_NEUTRAL_BOSON		= 0x110000;
const long CEParticle::PARTICLE_VECTOR_BOSON		= 0x140000; // flag
const long CEParticle::PARTICLE_NEUTRAL_VEC_BOSON	= 0x150000; // NEUTRAL | VECTOR
const long CEParticle::PARTICLE_PHOTON				= 0x150001;
const long CEParticle::PARTICLE_Z_BOSON				= 0x150002;
const long CEParticle::PARTICLE_W_BOSON				= 0x140008;
const long CEParticle::PARTICLE_GLUON				= 0x150010;

const long CEParticle::PARTICLE_SCALAR				= 0x120000; // flag
const long CEParticle::PARTICLE_HIGGS				= 0x130001;
const long CEParticle::PARTICLE_SCALAR_PHI			= 0x120002;
const long CEParticle::PARTICLE_SCALAR_CHI			= 0x130003;
const long CEParticle::PARTICLE_FPGHOST				= 0x1a0000; // flag, TODO: make it FERMIONS
const long CEParticle::PARTICLE_FPGHOST_UA			= 0x1a0001;
const long CEParticle::PARTICLE_FPGHOST_UZ			= 0x1a0002;
const long CEParticle::PARTICLE_FPGHOST_UWp			= 0x1a0003;
const long CEParticle::PARTICLE_FPGHOST_UWm			= 0x1a0004;

const long CEParticle::PARTICLE_MODEL_BOSON			= 0x900000; // flag

const long CEParticle::PARTICLE_CLASS_FLAG			= 0x400000; // "abstract" particle, to be substituted by real one
const long CEParticle::PARTICLE_KIND_MAX			= 0xffffff;

//const long CEParticle::PARTICLE_ENUM_MASK	= 0x00ffff;

/*
* to compare if 'usual' particle fits 'class', as "UsualPK & FitFlagPk | PARTICLE_CLASS_FLAG== ClassPK"
*/
long CEParticle::GetClassFitFlagPk(long pkClass) {
	if (pkClass<0) pkClass=-pkClass;
	long pkFitFlag=0;
	if ((pkClass & CEParticle::PARTICLE_BOSON) == CEParticle::PARTICLE_BOSON) {
		pkFitFlag=CEParticle::PARTICLE_SCALAR|CEParticle::PARTICLE_NEUTRAL_BOSON|CEParticle::PARTICLE_VECTOR_BOSON|CEParticle::PARTICLE_FPGHOST;
	} else if ((pkClass & CEParticle::PARTICLE_FERMION) == CEParticle::PARTICLE_FERMION) {
		pkFitFlag=CEParticle::PARTICLE_FERMION;
	}
	return pkFitFlag;
}

bool CEParticle::FitClassPk(long pk, long pkClass) {
	if (pk<0) pk=-pk;
	if (pkClass<0) pkClass=-pkClass;
	return ((pk & GetClassFitFlagPk(pkClass) | PARTICLE_CLASS_FLAG) == pkClass);
}

CDebug* CEParticle::dbg = new CDebug("CEParticle");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEParticle::CEParticle(CExpression* id,
					   CExpression* sym_mass,
					   CExpression* _pol_vec_sym,
					   int _ID_kin,
					   CEValue* _mass)
{
	m_init_OK = false;
	m_ID_in_kin = _ID_kin;

	m_id = id;
	m_sym_mass = sym_mass;
	m_mass_value = _mass;
	m_pol_vec_sym = _pol_vec_sym;

	m_pol_vec_sym_cjg = NULL;
	if (m_pol_vec_sym != NULL) {
		// create conjugated symbol
		CEString* str_pol_sym = m_pol_vec_sym->toString();
		char* C_str = CEUtil::concat("C", str_pol_sym->getData());
		delete str_pol_sym;
		m_pol_vec_sym_cjg = new CESymbol(C_str);
		CMemoryManager::mfree(C_str);
	}

	m_stub = NULL;
	if (_mass != NULL) m_stub = new CENumVector((CEValue*)_mass->clone());
	m_polarization_info = 1; // default is polarized case

	m_pol_vec_predefined = NULL; m_pol_vec_predef_index = NULL;
	dbg->log(10, "New particle created", m_id);
}

CEParticle::~CEParticle()
{
	if (m_id != NULL) delete m_id;
	if (m_sym_mass != NULL) delete m_sym_mass;
	if (m_pol_vec_sym != NULL) delete m_pol_vec_sym;
	if (m_pol_vec_sym_cjg != NULL) delete m_pol_vec_sym_cjg;
	if (m_mass_value != NULL) delete m_mass_value;
	if (m_stub != NULL) delete m_stub;

	if (m_pol_vec_predefined != NULL) delete m_pol_vec_predefined;
	if (m_pol_vec_predef_index != NULL) delete m_pol_vec_predef_index;
}

bool CEParticle::is_init_OK() { return m_init_OK; }
const char* CEParticle::ClassName() const { return "<Particle>"; }
CEString* CEParticle::toString(int mode)
{
	CEString* res  = new CEString("\\({");
	res->add_int(m_ID_in_kin);
	res->AddString(", ");
	if (m_id != NULL) res->AddString(m_id->toString(mode));
	else res->AddString("NULL");
	res->AddString(", ");
	if (m_sym_mass != NULL) res->AddString(m_sym_mass->toString(mode));
	if (m_stub != NULL) {
		res->AddString(", values:");
		res->AddString(m_stub->toString(mode));
	} else {
		res->AddString(", no values");
	}
	if (m_pol_vec_predefined != NULL) {
		res->AddString(", polarization:");
		res->AddString(m_pol_vec_predefined->toString(mode));
		res->AddString(", index:");
		res->AddString(m_pol_vec_predef_index->toString(mode));
	}
	res->AddString("}\\)");
	return res;
}

CENumVector* CEParticle::get_stub() { return m_stub; }
CEValue* CEParticle::mass() { return m_mass_value; }
CExpression* CEParticle::get_momentum_symbol() { return m_id; }
CExpression* CEParticle::get_sym_mass() { return m_sym_mass; }
CExpression* CEParticle::get_pol_vector_symbol() { return m_pol_vec_sym; }
CExpression* CEParticle::get_pol_vector_symbol_cjg() { return m_pol_vec_sym_cjg; }
CExpression* CEParticle::get_pol_vec_predifined() { return m_pol_vec_predefined; }
CExpression* CEParticle::get_pol_vec_predef_index() { return m_pol_vec_predef_index; }
int CEParticle::get_polarization_info() { return m_polarization_info; }

void CEParticle::set_sym_mass(CExpression* _m_sym) { // no checks for symbol usage inside
	if (_m_sym == NULL) { dbg->log(0,"ERROR: set_sym_mass(NULL) call"); return; }
	if (m_sym_mass != NULL) delete m_sym_mass;
	m_sym_mass=_m_sym->clone();
}

void CEParticle::set_polarization_info(int _pol_val) {
	if ((_pol_val == 0) || (_pol_val == 1)) {
		m_polarization_info=_pol_val;
	}
}

int CEParticle::GetPolarizationsNumber() { // 3 for massive boson partice, 2 for massless & fermion partices
	CEProcessEnvironment* env = CEProcessor::getEnvironment();
	long pk = get_PK();
	if (CEParticle::is_fermion_PK(pk)) {
		return 2;
	} else if (CEParticle::is_bosonic_PK(pk)) {
		if (env->GetMassCalcOrder(pk) == 0) return 2;
		CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
		// Note: actual mass symbol may differs from phys->get_mass_sym(pk)
		// But get_sym_mass() may be 0 for chiral limit (may it????)
		// massless particle should not have non-zero mass get_sym_mass()
		CExpression* mass_sym = phys->get_mass_sym(pk);
		if (CEBasic::isZero(mass_sym)) return 2;
	}
	return 3;
}

void CEParticle::predefine_pol_vector(CExpression* vec, CExpression* ind, bool ind_cov) {
	if ((vec == NULL) || (ind == NULL)) { dbg->log(0, "Error: CEParticle::predefine_pol_vector(NULL)"); return; }

	CExpression* vec_new = vec->clone();

	// exchange index to unique
	CExpression* ind_new = CESymbol::getUnique();
	vec_new = CEReplasement::Replace(vec_new, ind, ind_new);
	if (CRContainCheck::search(vec_new, ind)) { dbg->log(0, "Error in CEParticle::predefine_pol_vector(). old index left after exchange"); CEException::halt(); }

	if (m_pol_vec_predefined != NULL) delete m_pol_vec_predefined;
	if (m_pol_vec_predef_index != NULL) delete m_pol_vec_predef_index;

	if (ind_cov) { // move index down
		CESymbol* u = CESymbol::getUnique();
		vec_new = vec_new->multiply(new CEMetric(ind_new, u, !ind_cov, false));
		vec_new = CExpression::arrge(vec_new);
		delete ind_new;
		ind_new=u;
		ind_cov = false;
	}

	m_pol_vec_predefined = vec_new;
	m_pol_vec_predef_index = ind_new;
	dbg->log(0, "Predefined polarization vector is set", m_pol_vec_predefined, m_pol_vec_predef_index);
}

/*
*	vec_sym should be  "+"-signed.
*/
CENumVector* CEParticle::get_vector_stub(CExpression* vec_sym) {
	if (vec_sym == NULL) return NULL;
	if (vec_sym->equals(get_momentum_symbol())) return get_stub();
	return NULL;
}

/*
*	compares pk to particle's pk
*	if (check_flag_only == true) the flags part of pk is only checked out.
*/
bool CEParticle::check_PK(long pk, bool check_flag_only) {
	long kind = get_PK();

	if (check_flag_only) { // check flags
		long kind_abs = (kind>=0)?(kind):(-kind);
		long pk_abs = (pk>=0)?(pk):(-pk);
		//long pk_flags = pk_abs & (!PARTICLE_ENUM_MASK);
		if ((kind_abs & pk_abs) != pk_abs) return false;
		return true;
	}

	// check equality
	if (kind != pk)  return false;
	return true;
}

/*
*	ID of particle in kinematics
*/
int CEParticle::get_ID() { return m_ID_in_kin; }
void CEParticle::set_ID(int pID) { m_ID_in_kin=pID; } // ID of particle in kinematics (e.g. if changed)
bool CEParticle::is_initial() { return (get_ID() > 0); }

///////////////////////////////////////////////////////////////////////////////////////////////////
//          STATICS
///////////////////////////////////////////////////////////////////////////////////////////////////
// {-1 for b<a (wrong order)}
int CEParticle::compare_part_kind(long a, long b, bool fermions_first) {
	int res = 0;
	if (fermions_first) {
		bool a_is_ferm = CEParticle::is_fermion_PK(a);
		bool b_is_ferm = CEParticle::is_fermion_PK(b);

		if ((a_is_ferm) && (!b_is_ferm)) { res = 1; }
		else if ((!a_is_ferm) && (b_is_ferm)) { res = -1; }
		else { res = CLMath::step(a, b); }

	} else {
		res = CLMath::step(a, b);
	}
	return res;
}

bool CEParticle::is_fermion_PK(long pk) {
	if (pk < 0) pk = -pk;
	return ((pk & CEParticle::PARTICLE_FERMION) == CEParticle::PARTICLE_FERMION);
}

bool CEParticle::is_bosonic_PK(long pk) {
	if (pk < 0) pk = -pk;
	return ((pk & CEParticle::PARTICLE_BOSON) == CEParticle::PARTICLE_BOSON);
}

bool CEParticle::is_vec_bosonic_PK(long pk) {
	if (pk < 0) pk = -pk;
	return ((pk & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
}

bool CEParticle::is_scalar_PK(long pk) {
	if (pk < 0) pk = -pk;
	return ((pk & CEParticle::PARTICLE_SCALAR) == CEParticle::PARTICLE_SCALAR);
}

bool CEParticle::is_neutral_bosonic_PK(long pk) {
	if (pk < 0) pk = -pk;
	return ((pk & CEParticle::PARTICLE_NEUTRAL_BOSON) == CEParticle::PARTICLE_NEUTRAL_BOSON);
}

bool CEParticle::is_class_PK(long pk) {
	if (pk < 0) pk = -pk;
	return ((pk & CEParticle::PARTICLE_CLASS_FLAG) == CEParticle::PARTICLE_CLASS_FLAG);
}

CEParticle* CEParticle::create_particle(long _kind,
										int _ID_kin,
										CExpression* _momentum_sym,
										CExpression* _mass_sym,
										CExpression* _pol_vec_sym,
										CEPtrArray *params)
{
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics(); if(phys==NULL)CEException::halt("CEParticle::create_particle: NO PHYSICS INSTALLED.");
	dbg->log(18,_kind,"Create particle. id=%ld");

	// check if particle can be created
	if ((!phys->validate_part_kind(_kind)) || (_momentum_sym == NULL) || (_pol_vec_sym == NULL)) {
		if (_momentum_sym != NULL) delete _momentum_sym;
		if (_mass_sym != NULL) delete _mass_sym;
		if (_pol_vec_sym != NULL) delete _pol_vec_sym;
		dbg->log(0,"CEParticle::create_particle ERROR: invalid PK, NULL momentum or polarization symbol spotted");
		return NULL; // error
	}

	// truncate sign at NEUTRAL BOSON particles
	if ((_kind < 0) &&
		((CLMath::abs(_kind) & CEParticle::PARTICLE_NEUTRAL_BOSON) == CEParticle::PARTICLE_NEUTRAL_BOSON)
		) { _kind = -_kind; }

	long kind_abs = CLMath::abs(_kind);
	CEParticle* res = NULL;
	if ((kind_abs & CEParticle::PARTICLE_LEPTON) == CEParticle::PARTICLE_LEPTON) {
		res = new CPFermion(_kind, _momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, NULL);
	} else if ((kind_abs & CEParticle::PARTICLE_NEUTRINO) == CEParticle::PARTICLE_NEUTRINO) {
		// let abstract fermion class to describe neutrino
		res = new CPFermion(_kind, _momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, NULL);
	} else if ((kind_abs & CEParticle::PARTICLE_QUARK) == CEParticle::PARTICLE_QUARK) {
		// let abstract fermion class to describe quarks
		res = new CPFermion(_kind, _momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, NULL);
	} else if (_kind == PARTICLE_PHOTON) {
		if (_mass_sym != NULL) delete _mass_sym; // no massive photons are implemented
		res = new CPPhoton(_momentum_sym, _pol_vec_sym, _ID_kin, params);
	} else if (_kind == PARTICLE_Z_BOSON) {
		res = new CPBosonZ(_momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, params);
	} else if (kind_abs == PARTICLE_W_BOSON) {
		res = new CPBosonW(_kind, _momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, params);
	} else if (CEParticle::is_bosonic_PK(kind_abs)) {
		res = new CPBoson(_kind, _momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, params);
	} if ((kind_abs & CEParticle::PARTICLE_PROTON) == CEParticle::PARTICLE_PROTON) {
		if(kind_abs<0){ dbg->log(0,"Unexpected particle. Anti-Proton is not allowed in current version."); CEException::halt(); }
		res = new CPProtonSF(_kind, _momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, NULL);
		//TODO: PARTICLE_NEUTRON
	} else if (CEParticle::is_fermion_PK(kind_abs)) {
		res = new CPFermion(_kind, _momentum_sym, _mass_sym, _pol_vec_sym, _ID_kin, NULL);
	}

	if (res != NULL) {
		if (!res->is_init_OK()) {
			dbg->log(0, _kind, "Warning: Particle init error (id == %ld)");
			//delete res; res = NULL;
		}
	} else {
		dbg->log(0, _kind, "[ERROR] Can't create particle (id == %ld). It's probably forbidden.");
	}
	dbg->log(8,"Particle created: ",res);
	return res;
}
