////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// Kinematics.cpp: implementation of the CKinematics class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDebug* CKinematics::dbg = new CDebug("CKinematics");
CEData* CKinematics::data = NULL;

int CKinematics::REPL_SIMILAR_COUPLINGS		= 0x01;
int CKinematics::REPL_MANDELSTAM_VARIABLES	= 0x02;
int CKinematics::REPL_ALL_MANDELSTAM_X		= 0x04;
int CKinematics::REPL_ALL_MOMENTA_SUBST		= 0x08;
int CKinematics::REPL_ALL_COUPLINGS_X		= 0x10;
int CKinematics::REPL_BASIS_COUPLINGS		= 0x20;
int CKinematics::REPL_COUPLINGS_SIMPLIFY	= 0x21; // REPL_SIMILAR_COUPLINGS|REPL_BASIS_COUPLINGS
int CKinematics::REPL_ALL_KIN_DEPENDENCIES	= 0x1D; //(0x10 | 0x08 | 0x04 | 0x01)

CKinematics::CKinematics()
{
	m_init_particles = new CEPtrArray();
	m_final_particles = new CEPtrArray();
	//InitVectors();
	m_Collinear_Simpl_list_added = false;

	m_BasisVecs = NULL;
}

/*CKinematics::CKinematics(CEPtrArray* arr_init_particles,
						 CEPtrArray* arr_final_particles)
{
	m_init_particles = arr_init_particles;
	m_final_particles = arr_final_particles;
	if (m_init_particles == NULL) m_init_particles = new CEPtrArray();
	if (m_final_particles == NULL) m_final_particles = new CEPtrArray();
}*/

CKinematics::~CKinematics()
{
	dbg->log(20, "~CKinematics()");
	//log_vectors();
	//m_init_particles->clear();
	delete m_init_particles;
	dbg->log(35, "m_init_particles deleted");
	//m_final_particles->clear();
	delete m_final_particles;
	if(m_BasisVecs != NULL) { delete m_BasisVecs; m_BasisVecs =NULL; }
	dbg->log(35, "m_final_particles deleted");
}

void CKinematics::DeclareBasisVectors() { // init symbols for basis
	if(m_BasisVecs != NULL) {
		if (m_BasisVecs->size() == 4) return; // already inited
		delete m_BasisVecs; m_BasisVecs=NULL; // reinit
	}
	m_BasisVecs = new CSequence();
	bool log_visible = false;

	CESymbol* vec_sym = new CESymbol("\\[Eta]\\+0");
	if (isSymbolUsed(vec_sym)) {
		dbg->log(0, "ERROR: can't use default symbol for vector basis, UNIQUE is involved", vec_sym);
		delete vec_sym; vec_sym=CESymbol::getUnique(); log_visible=true;
	}
	m_BasisVecs->addItem(vec_sym);

	vec_sym = new CESymbol("\\[Eta]\\+1");
	if (isSymbolUsed(vec_sym)) {
		dbg->log(0, "ERROR: can't use default symbol for vector basis, UNIQUE is involved", vec_sym);
		delete vec_sym; vec_sym=CESymbol::getUnique(); log_visible=true;
	}
	m_BasisVecs->addItem(vec_sym);

	vec_sym = new CESymbol("\\[Eta]\\+2");
	if (isSymbolUsed(vec_sym)) {
		dbg->log(0, "ERROR: can't use default symbol for vector basis, UNIQUE is involved", vec_sym);
		delete vec_sym; vec_sym=CESymbol::getUnique(); log_visible=true;
	}
	m_BasisVecs->addItem(vec_sym);

	vec_sym = new CESymbol("\\[Eta]\\+3");
	if (isSymbolUsed(vec_sym)) {
		dbg->log(0, "ERROR: can't use default symbol for vector basis, UNIQUE is involved", vec_sym);
		delete vec_sym; vec_sym=CESymbol::getUnique(); log_visible=true;
	}
	m_BasisVecs->addItem(vec_sym);

	if (log_visible) { dbg->log(0, "Vector Basis Composed (WITH ERRORS)", m_BasisVecs); }
	else { dbg->log(2, "Vector Basis Composed", m_BasisVecs); }
}

CExpression* CKinematics::getEqual(CExpression *expr)
{
	if (expr == NULL) return NULL;
	if (expr->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CESum* sum = (CESum*)expr;

		long s = sum->size();
		if (s == 0) return NULL;

		CESum* res = (CESum*)sum->clone();
		long i;
		for (i=0; i<s; i++) {
			CExpression* item = res->GetAt(i);
			CExpression* subst = getEqual(item);
			if ((subst != NULL) && (subst != item)) {
				res->SetAt(i, subst);
				delete subst;
				return CExpression::reduce(res);
			}
		}
		delete res;
		return NULL;
	}

	int fm = find_momentum(expr);
	if (fm == 0) return NULL;
	CExpression* res = get_kin_equal(fm);
	if (!expr->getFlag()) res->inverseFlag();
	return res;
}

/*
*	particle_num is the valid particle ID, returned by find_momentum()
*/
CESum* CKinematics::get_kin_equal(int particle_num) {
	
	CESum* res = new CESum();
	int s = num_init_particles();
	int i;
	for(i=0; i<s; i++) {
		CExpression* mom = pin(i)->get_momentum_symbol();
		if ((particle_num-1) != i) {
			CExpression* mom_copy = mom->clone();
			if (particle_num > 0) mom_copy->inverseFlag();
			res->addItem(mom_copy);
		}
	}
	
	s = num_final_particles();
	for (i=0; i<s; i++) {
		CExpression* mom = pfin(i)->get_momentum_symbol();
		if ((-particle_num-1) != i) {
			CExpression* mom_copy = mom->clone();
			if (particle_num < 0) mom_copy->inverseFlag();
			res->addItem(mom_copy);
		}
	}
	return res;
}

CExpression* CKinematics::get_momentum(int particle_num) {
	if (particle_num == 0) return NULL;
	if (particle_num > 0) { // from initials
		int s = num_init_particles();
		int i = particle_num-1;
		if (i >= s) return NULL;

		return pin(i)->get_momentum_symbol();
	}
	else { // from finals
		int s = num_final_particles();
		int i = -particle_num-1;
		if (i >= s) return NULL;

		return pfin(i)->get_momentum_symbol();
	}
	return NULL;
}

bool CKinematics::isPrimary(CExpression *expr)
{
	return true;
}

CEReplList* CKinematics::getReplacements(int type)
{
	//return NULL;
	CEReplList* res = new CEReplList();

	if (type & CKinematics::REPL_BASIS_COUPLINGS) {
		if (m_BasisVecs != NULL) {
			int s = m_BasisVecs->size();
			int i;
			int j;
			CExpression* z = CENumber::GetZero();
			for(i=0; i<(s-1); i++) {
				CExpression* ei = m_BasisVecs->GetAt(i); 
				if (ei == NULL) continue;
				for(j=i+1; j<s; j++) {
					CExpression* ej = m_BasisVecs->GetAt(i); 
					if (ej == NULL) continue;

					CExpression* a = CExpression::reduce(new CESProduct(ei, ej));
					res->add(a, z);
					delete a;
				}
			}
			delete z;
		}
	}

	if (type & CKinematics::REPL_ALL_MOMENTA_SUBST) {
		
		int s = num_init_particles();
		int i;
		for (i=0; i<s; i++) {
			CExpression* mom = pin(i)->get_momentum_symbol();
			CExpression* mom_alt = get_kin_equal(i+1);
			res->add(mom, mom_alt, false);
			delete mom_alt;
		}

		s = num_final_particles();
		for (i=0; i<s; i++) {
			CExpression* mom = pfin(i)->get_momentum_symbol();
			CExpression* mom_alt = get_kin_equal(-i-1);
			res->add(mom, mom_alt, false);
			delete mom_alt;
		}
	}

	if (type & CKinematics::REPL_ALL_COUPLINGS_X) {
		// for every coupling create a substitute.
		// every p_i*p_j coupling should produce 2 different representations:
		// p_i*(p_x+p_y-p_z) and (p_a+p_b-p_c)*p_j.
		int is = num_init_particles();
		int fs = num_final_particles();
		int i;
		for(i=-fs; i<=is; i++) { // for each particle in arbitrary order
			if (i==0) continue;
			CExpression* pI = get_particle_by_ID(i)->get_momentum_symbol();
			CExpression* momI_alt = get_kin_equal(i);
			int j;
			for(j=-fs; j<=is; j++) { // for each other particle in arbitrary order
				if ((j==0) || (j==i)) continue;
				CExpression* pJ = get_particle_by_ID(j)->get_momentum_symbol();

				// p_i*p_j -- arrange momenta order
				CExpression* a = CExpression::reduce(new CESProduct(pI, pJ));

				// substitute expresion (p_a+p_b-p_c)*p_j, expanded and arranged.
				CExpression* b = CExpression::reduce(new CESProduct(momI_alt, pJ));
				res->add(a, b);
				delete a; delete b;
			}
			delete momI_alt;
		}
	}
	return res;
}

bool CKinematics::isSymbolUsed(CESymbol* sym) {
	if (sym == NULL) return false;

	// check basis vectors
	if (m_BasisVecs != NULL) {
		if (GetBasisVecNo(sym) >= 0) return true;
	}

	// check kinematics (momenta & polarizations)
	int s = num_init_particles();
	int i;
	for (i=0; i<s; i++) {
		CEParticle* part = pin(i);
		CExpression* cmp = part->get_momentum_symbol();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
		cmp = part->get_pol_vector_symbol();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
		cmp = part->get_pol_vector_symbol_cjg();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
		cmp = part->get_sym_mass();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
	}
	
	s = num_final_particles();
	for (i=0; i<s; i++) {
		CEParticle* part = pfin(i);
		CExpression* cmp = part->get_momentum_symbol();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
		cmp = part->get_pol_vector_symbol();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
		cmp = part->get_pol_vector_symbol_cjg();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
		cmp = part->get_sym_mass();
		if ((cmp != NULL) && (sym->equals(cmp))) return true;
	}

	return false;
}

int CKinematics::checkExprSign(CExpression* expr) { return 0; }

/*
*	returns array of crossings info.
*	format: array { cross#1, cross#2, ... }
*	cross info: array {container{NULL, PID#1, PID#2, NULL}, rule}
*	crossing rules includes: momenta and polarization vector replacements
*	spare vector indices should be replaced manually (using PID numbers
*	and Matrix Elements supplementary info)
*/
CEPtrArray* CKinematics::getCrossingsTransformations() {
	dbg->log(20, "getCrossingsTransformations()");

	CEPtrArray* res = new CEPtrArray();
	int is = m_init_particles->size();
	int fs = m_final_particles->size();

	// scan final particles
	int i;
	for(i=0; i<(fs-1); i++) {
		CEParticle* part1 = pfin(i);
		int PK1 = part1->get_PK();
		int j;
		for(j=i+1; j<fs; j++) {
			CEParticle* part2 = pfin(j);
			int PK2 = part2->get_PK();
			if (PK1 == PK2) {
				CEPtrArray* info = new CEPtrArray();
				info->add(new CExprContainer(NULL, part1->get_ID(), part2->get_ID(), PK1));
				CEReplList* r = new CEReplList();
				r->add(part1->get_momentum_symbol(), part2->get_momentum_symbol(), true);
				r->add(part1->get_pol_vector_symbol(), part2->get_pol_vector_symbol(), true);
				info->add(r);
				res->add(info);
			}
		}
	}

	// involve initial crossings for scattering process
	bool involve_initial_crossings = (is == 2);

	if (false) { // alternating (currently unused)
		// if final particles crossings number == particles num/2
		// do not involve initial particles crossings (they are already involved (?))
		involve_initial_crossings = ((2*(res->size()) < fs) && (is == 2));
	}

	if (involve_initial_crossings) {
		// scan initial particles
		for (i=0; i<(is-1); i++) {
			CEParticle* part1 = pin(i);
			int PK1 = part1->get_PK();
			int j;
			for(j=i+1; j<is; j++) {
				CEParticle* part2 = pin(j);
				int PK2 = part2->get_PK();
				if (PK1 == PK2) {
					CEPtrArray* info = new CEPtrArray();
					info->add(new CExprContainer(NULL, part1->get_ID(), part2->get_ID(), PK1));
					CEReplList* r = new CEReplList();
					r->add(part1->get_momentum_symbol(), part2->get_momentum_symbol(), true);
					r->add(part1->get_pol_vector_symbol(), part2->get_pol_vector_symbol(), true);
					info->add(r);
					res->add(info);
				}
			}
		}
	}

	if (res->size() == 0) { delete res; res = NULL; }
	dbg->log(30, "getCrossingsTransformations() result:", res);
	return res;
}

/*
*	returns rule for set bsaic vector couplings to its values:
*		e*p --> zero, e*e-->-1, e*Ce-->1, etc.
*	for (CM_beams == true) also returns relations for
*		e_i*p_j, e_i*e_j.
*/
CEReplList* CKinematics::getBasicCouplingsValues(bool CM_beams) {
	dbg->log(20, CM_beams, "getBasicCouplingsValues(c.m.s. beams == %ld)");
	CEReplList* r = new CEReplList();

	CENumber zero((long)0);
	CExpression* munit = CENumber::GetUnit(); munit->inverseFlag();

	int is = num_init_particles();
	int fs = num_final_particles();
	int i;
	for(i=-fs; i<=is; i++) { // for each particle in arbitrary order
		if (i==0) continue;
		CEParticle* part = get_particle_by_ID(i);

		CExpression* p = part->get_momentum_symbol();
		CExpression* e = part->get_pol_vector_symbol();
		CExpression* ce = part->get_pol_vector_symbol_cjg();
		if ((e==NULL) || (ce == NULL)) break;
		CExpression* a = NULL;

		a = CExpression::reduce(new CESProduct(p, e));
		r->add(a, &zero); delete a;
		
		a = CExpression::reduce(new CESProduct(p, ce));
		r->add(a, &zero); delete a;
		
		a = CExpression::reduce(new CESProduct(e, e));
		r->add(a, &zero); delete a;
		
		a = CExpression::reduce(new CESProduct(ce, ce));
		r->add(a, &zero); delete a;
		
		a = CExpression::reduce(new CESProduct(e, ce));
		r->add(a, munit); delete a;
	}
	delete munit;

	if ((CM_beams) && (is == 2)) {
		// if consider beams in c.m.s. (initial beams are collinear)
		// the relations between e,ce, and p of different particles are also available
		
		// for photon vector polarizations only
		// Suppose initial momenta collinear
		CEParticle* _part1_abs = pin(0);
		CEParticle* _part2_abs = pin(1);
		
		int i;
		for(i=0; i<2; i++) {
			CEParticle* part1 = NULL;
			CEParticle* part2 = NULL;
			if (i == 0) {	//	1
				part1 = _part1_abs;
				part2 = _part2_abs;
			} else {			//	2
				part1 = _part2_abs;
				part2 = _part1_abs;
			}
			//	couple polarizations of #1 to momenta of #2

			if (part1->check_PK(CEParticle::PARTICLE_PHOTON, true)) {
				CExpression* e1 = part1->get_pol_vector_symbol();
				CExpression* ce1 = part1->get_pol_vector_symbol_cjg();
				CExpression* p2 = part2->get_momentum_symbol();
				if ((e1 != NULL) && (ce1 != NULL)) {
					CExpression* a = NULL;
					
					a = CExpression::reduce(new CESProduct(p2, e1));
					r->add(a, &zero); delete a;
					
					a = CExpression::reduce(new CESProduct(p2, ce1));
					r->add(a, &zero); delete a;
				}
			}
		}
	}

	dbg->log(30, CM_beams, "getBasicCouplingsValues(c.m.s. beams == %ld) result:", r);
	return r;
}

void CKinematics::init()
{
	if (data == NULL) data = CEProcessor::getDataStorage();
}

/*
*	To be called before matrix element retrieving
*	reset vector indices
*	flase means indices to be cleared
*/
void CKinematics::reset_volatile_variables(bool do_create) {
	int s = m_init_particles->size();
	int i;
	for (i=0; i<s; i++) {
		CEParticle* part = pin(i);
		if (part->check_PK(CEParticle::PARTICLE_VECTOR_BOSON, true)) {
			((CPAbstractBoson*)part)->reset_vector_index(do_create);
		}
	}
	
	s = m_final_particles->size();
	for (i=0; i<s; i++) {
		CEParticle* part = pfin(i);
		if (part->check_PK(CEParticle::PARTICLE_VECTOR_BOSON, true)) {
			((CPAbstractBoson*)part)->reset_vector_index(do_create);
		}
	}
}

/*void CKinematics::InitVectors()
{

}*/

//int CKinematics::num_final_particles() { return m_final_particles->size(); }
//int CKinematics::num_init_particles() { return m_init_particles->size(); }

CENumVector* CKinematics::get_vector_by_ID(int id)
{
	if (id == 0) return NULL;
	CENumVector* res = NULL;
	if (id > 0) res = vector(id, true);
	else res = vector((-id), false);
	return res;
}

CENumVector* CKinematics::vector(int i, bool from_initial)
{
	CENumVector* res = NULL;
	CEParticle* part = get_particle(i, from_initial);
	if (part != NULL) res = part->get_stub();
	/*if (from_initial) {
		if ((i<num_init_particles()) && (i>=0)) {
			res = ((CEParticle*)m_init_particles->GetAt(i))->get_stub();
		}
		if (res == NULL)
			dbg->log(1, i, "Warning: no such initial vector in kinematic: %d");
	}
	else {
		if ((i<num_final_particles()) && (i>=0))
			res = ((CEParticle*)m_final_particles->GetAt(i))->get_stub();
		if (res == NULL)
			dbg->log(1, i, "Warning: no such final-state vector in kinematic: %d");
	}*/
	return res;
}

CENumVector* CKinematics::get_sum_initials()
{
	CENumVector* res = vector(0, true)->num_stub();
	int s = num_init_particles();
	int i;
	for(i=1; i<s; i++) {
		res->addvector(vector(i, true), false);
	}
	return res;
}

CEValue* CKinematics::mass(int i, bool from_initial)
{
	return vector(i, from_initial)->mass();
}

CENumVector* CKinematics::vfin(int i)
{
	return vector(i, false);
}

CENumVector* CKinematics::vin(int i)
{
	return vector(i, true);
}

void CKinematics::fill_data_storage()
{

}

void CKinematics::log_vectors()
{
	dbg->log(20, "logging vectors");
	int s = m_init_particles->size();
	int i;
	for (i=0; i<s; i++) {
		dbg->log(30, "init vector", (CESerializable*)m_init_particles->GetAt(i));
	}
	s = m_final_particles->size();
	for (i=0; i<s; i++) {
		dbg->log(30, "final vector", (CESerializable*)m_final_particles->GetAt(i));
	}
	dbg->log(20, "vectors logging complete");
}

//CEParticle* CKinematics::pin(int i0) { return get_particle(i0, true); }
//CEParticle* CKinematics::pfin(int i0) { return get_particle(i0, false); }

//substitute particle function, used for {p->quarks} substitutions, clears all pre-inited secondary variables
//Important: this function MUST call 'CEProcessor::OnKinematicsChanged()'!
CEParticle* CKinematics::ReplaceParticle(int pId, CEParticle* part) { //returns previous, that is NOT deleted
	if((pId==0)||(part==NULL))return NULL;
	CEParticle* prev = NULL;
	if (pId > 0) {	//initial, no=pId-1
		int no=pId-1;
		if((no<num_init_particles()) && (no>=0)){
			prev = (CEParticle*)m_init_particles->GetAt(no);
			if(prev!=NULL){ m_init_particles->SetAt(no,part,false); }else{ CEException::halt(dbg,"ReplaceParticle: NULL particle spotted"); }
		}
	}else{	//final, no=(-pId-1)
		int no=(-pId-1);
		if((no<num_final_particles()) && (no>=0)){
			prev = (CEParticle*)m_final_particles->GetAt(no);
			if(prev!=NULL){ m_final_particles->SetAt(no,part,false); }else{ CEException::halt(dbg,"ReplaceParticle: NULL particle spotted"); }
		}
	}

	if (prev!=NULL) { ClearVolatileVariables(); } //clear secondary variables (elsewhere nothing was replaced)
	CEProcessor::OnKinematicsChanged(); //clear kinematics-dependent caches
	return prev;
}

//	used e.g. to create R-kinematics for current process
//	clears all pre-inited secondary variables
//	Important: this function MUST call 'CEProcessor::OnKinematicsChanged()'!
int CKinematics::AddFinalParticle(CEParticle* part) { //returns PID, gets ownership of 'part'
	if(part==NULL)return 0;
	int pId = -num_final_particles()-1; //final ID = -no-1
	part->set_ID(pId); m_final_particles->add(part);

	ClearVolatileVariables(); //clear secondary variables
	CEProcessor::OnKinematicsChanged(); //clear kinematics-dependent caches
	return pId;
}

//	used e.g. to return to original kinematics after R-kinematics was set
//	Note: every particle contains pID ==> be careful with removing them!
//	clears all pre-inited secondary variables
//	Important: this function MUST call 'CEProcessor::OnKinematicsChanged()'!
CEParticle* CKinematics::RemoveLastFinalParticle() { //returns previous, that is NOT deleted
	int no=num_final_particles()-1;
	CEParticle* prev = (CEParticle*)m_final_particles->ExtractAt(num_final_particles()-1);
	if(prev==NULL){ CEException::halt(dbg,"RemoveParticle: NULL particle spotted"); }
	if (prev!=NULL) { ClearVolatileVariables(); } //clear secondary variables (elsewhere nothing was replaced)
	CEProcessor::OnKinematicsChanged(); //clear kinematics-dependent caches
	return prev;
}

//clear seconrdary variables (e.g. if kinematics is changed)
void CKinematics::ClearVolatileVariables() {
	if(m_BasisVecs!=NULL) { CSequence* tmp=m_BasisVecs; m_BasisVecs=NULL; delete tmp; }
	m_Collinear_Simpl_list_added=false;
}

/*
*	ID > 0 means initial particle
*/
CEParticle* CKinematics::get_particle_by_ID(int id) {
	if (id == 0) return NULL;
	CEParticle* res = NULL;
	if (id > 0) res = get_particle(id-1, true);
	else res = get_particle((-id-1), false);
	return res;
}

int CKinematics::get_particle_ID(CEParticle* part) {
	if (part == NULL) return 0;
	int s = m_init_particles->size();
	int i;
	for (i=0; i<s; i++) {
		if (part == pin(i)) return (i+1);
	}
	s = m_final_particles->size();
	for (i=0; i<s; i++) {
		if (part == pfin(i)) return (-i-1);
	}
	return 0;
}

CEParticle* CKinematics::get_particle(int i0, bool from_initial)
{
	CEParticle* res = NULL;
	if (from_initial) {
		if ((i0<num_init_particles()) && (i0>=0)) {
			res = (CEParticle*)m_init_particles->GetAt(i0);
		}
		if (res == NULL)
			dbg->log(0, i0, from_initial, "Warning: no such initial particle in kinematic: %d");
	}
	else {
		if ((i0<num_final_particles()) && (i0>=0))
			res = (CEParticle*)m_final_particles->GetAt(i0);
		if (res == NULL)
			dbg->log(0, i0, "Warning: no such final-state particle in kinematic: %d");
	}
	return res;
}

CEParticle* CKinematics::get_particle(CExpression* momentum) {
	if (momentum == NULL) return NULL;
	int s = m_init_particles->size();
	int i;
	for (i=0; i<s; i++) {
		CEParticle* part = pin(i);
		if (momentum->equals(part->get_momentum_symbol())) return part;
	}
	s = m_final_particles->size();
	for (i=0; i<s; i++) {
		CEParticle* part = pfin(i);
		if (momentum->equals(part->get_momentum_symbol())) return part;
	}
	return NULL;
}

CExpression* CKinematics::get_mass(CExpression * momentum) {
	if (momentum == NULL) return NULL;
	CEParticle* part = get_particle(momentum);
	if (part == NULL) return NULL;
	return part->get_sym_mass();
}

 // returns cloned value or NULL
CExpression* CKinematics::GetVectorSquare(CExpression* vec_base) {
	if (vec_base == NULL) return NULL;

	// first search basis vectors
	if (m_BasisVecs != NULL) {
		if (GetBasisVecNo(vec_base) >= 0) {
			CExpression* res = CENumber::GetUnit();
			res->inverseFlag();
			return res;
		}
	}

	// search particle
	CEParticle* part = get_particle(vec_base);
	if (part == NULL) return NULL;
	CExpression* m = part->get_sym_mass();
	if (m == NULL) return NULL;
	return new CEPower(m, (long)2);
}

int CKinematics::GetBasisVecNo(CExpression* vec_base) { // -1 if not found
	if ((m_BasisVecs == NULL) || (vec_base == NULL)) return -1;
	int i;
	for(i=0; i<m_BasisVecs->size(); i++) {
		if (vec_base->equals(m_BasisVecs->GetAt(i))) return i;
	}
	return -1;
}

CSequence* CKinematics::GetBasis() {
	if (m_BasisVecs == NULL) DeclareBasisVectors();
	return m_BasisVecs;
}


/*
*	TODO: make this procedure more correct, + add momenta.
*/
long CKinematics::GetHashCode() {
	long res=0;
	int nNums[]={3,5,7,11,13,17,19,23,29,31,37,39,41,43,47}; //15 numbers
	int nNumsMax=15;
	int nNumI=0; int nNumCurr=0;
	int i;
	for (i=0; i<m_init_particles->size(); i++) {
		CEParticle* part = pin(i); //part->get_PK();part->get_momentum_symbol()
		nNumCurr=nNums[nNumI]; nNumI++; if(nNumI>=nNumsMax)nNumI=0;
		res+=nNumCurr*part->get_PK();
	}
	for (i=0; i<m_final_particles->size(); i++) {
		CEParticle* part = pfin(i); //part->get_PK();part->get_momentum_symbol()
		nNumCurr=nNums[nNumI]; nNumI++; if(nNumI>=nNumsMax)nNumI=0;
		res+=nNumCurr*part->get_PK();
	}
	return res;
}

CExpression* CKinematics::get_simpliest(CExpression * expr)
{
	//expr = CExpression::reduce(expr);

	if (expr->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		CExpression* res = getEqual(expr);
		if (res == NULL) return expr;

		// Note: reduction should be included into getEqual() <TODO>
		res = CExpression::reduce(res);

		if (!res->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
			if (res != expr) delete expr;
			return res;
		}

		if (((CESum*)expr)->size() > ((CESum*)res)->size()) {
			if (res != expr) delete expr;
			return res;
		}
		else {
			delete res;
			return expr;
		}
	}
	return expr;
}

/*
* 0: Can't find momentum.
* >0: is initial
* <0: is final
*/
int CKinematics::find_momentum(CExpression * expr)
{
	if (expr == NULL) return 0;
	if (!expr->getFlag()) {
		CExpression* temp = expr->clone();
		temp->setFlag();
		int res = find_momentum(temp);
		delete temp;
		return res;
	}

	int s = m_init_particles->size();
	int i;
	for (i=0; i<s; i++) {
		if (expr->equals(pin(i)->get_momentum_symbol()))
			return i+1;
	}
	
	s = m_final_particles->size();
	for (i=0; i<s; i++) {
		if (expr->equals(pfin(i)->get_momentum_symbol()))
			return -i-1;
	}
	return 0;
}

int CKinematics::find_polarization_vec(CExpression* expr) { // check also CJG vector
	if (expr == NULL) return 0;
	if (!expr->getFlag()) { CExpression* temp = expr->clone(); temp->setFlag(); int res = find_polarization_vec(temp); delete temp; return res; }
	int s = m_init_particles->size(); int i;
	for (i=0; i<s; i++) { if (expr->equals(pin(i)->get_pol_vector_symbol())) return i+1; }
	s = m_final_particles->size();
	for (i=0; i<s; i++) { if (expr->equals(pfin(i)->get_pol_vector_symbol())) return -i-1; }
	return 0;
}

const char* CKinematics::CLASS_NAME = "CKinematics";
const char* CKinematics::ClassName() const { return CLASS_NAME; }
bool CKinematics::contain_vector(void * obj) {
	if ((m_init_particles == obj) || (m_final_particles == obj)) return true;
	if (m_init_particles != NULL) { if (m_init_particles->contain_vector(obj)) return true; }
	if (m_final_particles != NULL) { if (m_final_particles->contain_vector(obj)) return true; }
	return false;
}

char* CKinematics::toLine(int mode) {
	return toLineAdv(true, true, false, mode);
}
char* CKinematics::toLineAdv(bool bInitials, bool bFinals, bool bIDs, int mode) {
	CEString res;
	CDiagramPhysics* phys=CDiagramPhysics::get_curr_physics();
	int DPh_StmKind=((mode==STORE_WR_MATHEMATICA)?CDiagramPhysics::SYMBOL_STYLE_NB:CDiagramPhysics::SYMBOL_STYLE_TEX);
	int s = m_init_particles->size(); int i;
	if(bInitials){
		for (i=0; i<s; i++) {
			if(i>0){ if(bIDs){res.AddString(", ");}else{res.AddString("+");} }
			if(bIDs){ res<<i<<"."; }
			if(phys!=NULL) {
				CEParticle* _p = pin(i); if(_p==NULL){ res.AddString("NULL"); continue; }
				//View: {X^+,X^-} for most particles, but {q,qbar} for quarks
				if(_p->check_PK(CEParticle::PARTICLE_QUARK,true)){
					if(_p->get_PK()<0){ if (mode==STORE_WR_MATHEMATICA){ res.AddString("\\("); }else{ res.AddString("\\bar{"); } }
					res.AddString( phys->get_particle_symbol(_p->get_PK(),DPh_StmKind) );
					if(_p->get_PK()<0){ if(mode==STORE_WR_MATHEMATICA){ res.AddString("\\&_\\)"); }else{ res.AddString("}"); } }
				}else{
					int iC3=phys->getCharge3(_p->get_PK()); bool biC3OK=((iC3!=INT_MIN)&&(iC3!=0));
					if (biC3OK) { if (mode==STORE_WR_MATHEMATICA) res.AddString("\\("); }
					res.AddString( phys->get_particle_symbol(_p->get_PK(),DPh_StmKind) );
					if (biC3OK) { if (mode==STORE_WR_MATHEMATICA) { res.AddString((iC3>0)?"\\^\\(+\\)\\)":"\\^\\(-\\)\\)"); }else { if(iC3>0) res.AddString("^+"); else res.AddString("^-"); } }
				}
			} else {
				res.AddString("#"); res.add_int(pin(i)->get_PK());
			}
			res.AddString("(");
			res.AddString(pin(i)->get_momentum_symbol()->toString(mode));
			res.AddString(",");
			res.AddString(pin(i)->get_pol_vector_symbol()->toString(mode));
			res.AddString(")");
		}
	}
	if(bInitials && bFinals){
		if(mode==STORE_TEX){ res.AddString("\\to "); }else{ res.AddString("-->"); }
	}
	s = m_final_particles->size();
	if(bFinals){
		for (i=0; i<s; i++) {
			if(i>0){ if(bIDs){res.AddString(", ");}else{res.AddString("+");} }
			if(bIDs){ res<<(i+m_init_particles->size())<<"."; }
			if(phys!=NULL) {
				CEParticle* _p = pfin(i); if(_p==NULL){ res.AddString("NULL"); continue; }
				//View: {X^+,X^-} for most particles, but {q,qbar} for quarks
				if(_p->check_PK(CEParticle::PARTICLE_QUARK,true)){
					if(_p->get_PK()<0){ if (mode==STORE_WR_MATHEMATICA){ res.AddString("\\("); }else{ res.AddString("\\bar{"); } }
					res.AddString( phys->get_particle_symbol(_p->get_PK(),DPh_StmKind) );
					if(_p->get_PK()<0){ if(mode==STORE_WR_MATHEMATICA){ res.AddString("\\&_\\)"); }else{ res.AddString("}"); } }
				}else{
					int iC3=phys->getCharge3(_p->get_PK()); bool biC3OK=((iC3!=INT_MIN)&&(iC3!=0));
					if (biC3OK) { if (mode==STORE_WR_MATHEMATICA) res.AddString("\\("); }
					res.AddString( phys->get_particle_symbol(_p->get_PK(),DPh_StmKind) );
					if (biC3OK) { if (mode==STORE_WR_MATHEMATICA) { res.AddString((iC3>0)?"\\^\\(+\\)\\)":"\\^\\(-\\)\\)"); }else { if(iC3>0) res.AddString("^+"); else res.AddString("^-"); } }
				}
			} else {
				res.AddString("#"); res.add_int(pfin(i)->get_PK());
			}
			res.AddString("(");
			res.AddString(pfin(i)->get_momentum_symbol()->toString(mode));
			res.AddString(",");
			res.AddString(pfin(i)->get_pol_vector_symbol()->toString(mode));
			res.AddString(")");
		}
	}
	return CEUtil::copyString(res.getData());
}

void CKinematics::add_inv_repl_to_list(CEParticle* p1,
									   CEParticle* p2,
									   bool sum_sqr,
									   bool plus,
									   CExpression* inv_sym,
									   CEReplList* list)
{
	if ((p1 == NULL) || (p2 == NULL) || (inv_sym == NULL) || (list == NULL)) return;

	CExpression* a = CExpression::reduce(
		new CESProduct(p1->get_momentum_symbol(), p2->get_momentum_symbol()));
	CExpression* b = inv_sym->clone();
	if (sum_sqr) b=b->minus(p1->get_sym_mass()->sqra())->minus(p2->get_sym_mass()->sqra());
	if (!plus) b->inverseFlag();
	b = CExpression::reduce(b->multiply(data->gete(CConst::DV_HALF)->clone()));
	list->add(a, b, false);
	delete a;
	delete b;
}

CENumVector* CKinematics::get_vector_stub(CExpression *h_sym)
{
	if (h_sym == NULL) return NULL;
	
	//	the stub for "-"-signed vector should be "-"-signed.
	//	but we return the native stub of particle without  cloning
	//	so we do not process the "-"-signed requests
	/*if (!h_sym->getFlag()) {
		CExpression* temp = h_sym->clone();
		temp->setFlag();
		CENumVector* res = get_helisity_stub(temp)
		//	???	res = (CENumVector*)res->clone(); res->inverseFlag();
		delete temp;
		return res;
	}*/

	CENumVector* res = NULL;
	int s = m_init_particles->size();
	int i;
	for (i=0; i<s; i++) {
		res = pin(i)->get_vector_stub(h_sym);
		if (res != NULL) return res;
	}
	s = m_final_particles->size();
	for (i=0; i<s; i++) {
		res = pfin(i)->get_vector_stub(h_sym);
		if (res != NULL) return res;
	}
	return NULL;
}

/*
*	Initial momenta are supposed to be collinear ==>
*	Addittional polarization relations are available.
*	(couplings of polarization #1 to momenta #2 == 0) (Always?)
*	Calling this function leads to non-covariant results.
*/
void CKinematics::add_collinear_simpl_list()
{
	dbg->log(0, "Error: old code involved : add_collinear_simpl_list().");
	throw(new CEException("Old code usage"));

	if (m_Collinear_Simpl_list_added) return;

	/////////////////////////////////////	for photon vector polarizations only
	int s = m_init_particles->size();
	if (s != 2) return;	//	decay mode have no additional relations

	//	if no vector notation involved ==> return
	if (CEProcessor::getEnvironment()->m_photon_helisity_notation != CPPhoton::PH_VECTOR_NOTATION)
		return;

	//	Suppose initial momenta collinear
	CEParticle* part1 = pin(0);
	CEParticle* part2 = pin(1);
	CExpression* p1 = part1->get_momentum_symbol();
	CExpression* p2 = part2->get_momentum_symbol();

	CEObject* rep_data = CEProcessor::getDataStorage()->get(CPPhoton::PH_REPL_DATA_ID);
	CEReplList* reps = NULL;
	if (rep_data == NULL) {
		reps = new CEReplList();
		CEProcessor::getDataStorage()->put(CPPhoton::PH_REPL_DATA_ID, reps);
	}
	else reps = (CEReplList*)rep_data;

	int i;
	for (i=0; i<2; i++) {
		CEParticle* part = NULL;
		CExpression* p = NULL;
		if (i == 0) {	//	1
			part = part1;
			p = p2;		//	couple polarizations of #1 to momenta of #2
		}
		else {			//	2
			part = part2;
			p = p1;		//	couple polarizations of #2 to momenta of #1
		}
		
		//////////////////////////////////////////////////////////////////////////////////
		//	//TODO: To be restored !!!!
		//////////////////////////////////////////////////////////////////////////////////
		/*if (part->get_PK() == CEParticle::PARTICLE_PHOTON) {
			CPPhoton* ph = (CPPhoton*)part;
			CESymbol* pol_vec = ph->get_pol_vec_symbol();
			CESymbol* pol_vec_cjg = ph->get_pol_vec_cjg_symbol();
			if ((pol_vec != NULL) && (pol_vec_cjg != NULL) && (p != NULL)) {
				CENumber zero((long)0);
				CExpression* a = new CESProduct(p, pol_vec);
				reps->add(a, &zero);
				delete a;
				
				a = new CESProduct(pol_vec, p);
				reps->add(a, &zero);
				delete a;

				a = new CESProduct(p, pol_vec_cjg);
				reps->add(a, &zero);
				delete a;
				
				a = new CESProduct(pol_vec_cjg, p);
				reps->add(a, &zero);
				delete a;
			}
			else {
				//	else polarizations are still not initialized?
				dbg->log(0, "Error: photon polarization symbols are not initialized.");
				return;
			}
		}*/
	}

	m_Collinear_Simpl_list_added = true;
}
