////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// SubstSimplify.cpp: implementation of the CSubstSimplify class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CSubstSimplify::dbg = new CDebug("CSubstSimplify");
int CSubstSimplify::m_SIMP_SUM_LOOP_LEVEL = 1; // Note: multi-sequence simplification is destroyed
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSubstSimplify::CSubstSimplify() { }
CSubstSimplify::~CSubstSimplify() { }

CExpression* CSubstSimplify::simp(CExpression *source, CEReplList *repls, bool do_del_repls)
{
	dbg->log(20, "simp", source);
	dbg->log(20, "simp (repl)", repls);

	CExpression* res = simp_sums(source, repls, true);

	if (do_del_repls) delete repls;
	dbg->log(20, "simp (result)", res);
	return res;
}

CExpression* CSubstSimplify::simp_kin(CExpression *source, bool u_rev)
{
	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) return source;	// No kinematics installed.

	CEReplList* repls = kin->getReplacements(CKinematics::REPL_ALL_KIN_DEPENDENCIES);
	if (repls == NULL) {	//	Error in kinematics
		printf("Error: kin->getReplacements(ALL) ==> NULL");
		return source;
	}

	int s = repls->size();
	if (s == 0) {
		delete repls;
		return source;
	}

	if (u_rev) {
		int i;
		for (i=0; i<s; i++) {
			CEReplasement* r = repls->get(i);
			//r->set_a(CHEPLib::to_ultrarev(r->a()->clone()), false);
			r->set_b(CHEPLib::to_ultrarev(r->b()->clone()), false);
		}
	}

	//CExpression* res = CHEPLib::repl_kin_invariants(source, false);
	//CExpression* res = CHEPLib::repl_similar_kin(source, true);
	CExpression* res = source;
	res = simp(res, repls, true);
	/*if (u_rev) {
		res = CHEPLib::to_ultrarev(res);
	}*/
	return res;
}

/*
*	Recurcively works with sums.
*/
CExpression* CSubstSimplify::simp_sums(CExpression *source, CEReplList *repls, bool del_src)
{
	if (source->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		CEProduct* pr = (CEProduct*)source;
		long i;
		for(i=0; i<pr->size(); i++) {
			CExpression* item = pr->GetAt(i);
			CExpression* item_res = simp_sums(item, repls, false);
			if ((item_res != item) && (item_res != NULL)) {
				pr->SetAt(i, item_res);
				delete item_res;
			}
		}
		return source;
	}

	if (!source->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		return source;
	}

	//	Inline all the sum sub-items.
	CExpression* res = CExpression::reduce(source, del_src);
	if (res != source) del_src = true;

	////////////////////////////////
	int s = repls->size();
	int i;
	for (i=0; i<s; i++) {
		CEReplasement* r = repls->get(i);
		CExpression* cpy = res->clone();
		cpy = r->ApplyTo(cpy);
		cpy = CExpression::reduce(cpy);
		if (is_a_simplier_b(cpy, res)) {
			if (del_src) delete res;
			res = cpy;
			del_src = true;
		}
		else delete cpy;
	}
	////////////////////////////////
	return res;
}





///////////////////////////////////////////////////////////////////////////////////////////////////

/*
*	Try apply rules to every item (sum) in source.
*	source should be previously reduced/arranged or factorized.
*
*	The 'loop_level' parameter indicates the variety of orders the rules will be applied.
*	'loop_level <= 1' means all the rules will be applied consecutively.
*	It is the default fast simplification.
*	'loop_level == r_simp_set->size()' means
*	the rules will be applied in all the possible orders.
*	Note: the duration of this task grows up as (r_simp_set->size())! (factorial!!!).
*
*	bFast: simplify small objects only
*/
CExpression* CSubstSimplify::SimpRepl(CExpression* source, CEReplList* r_simp_set, CERule* r_env, int loop_level, bool bFast) {
	if ((r_simp_set == NULL) || (source == NULL)) return source;
	dbg->log(20, "SimpRepl()", source);
	dbg->log(25, "SimpRepl() simp rules:", r_simp_set);
	dbg->log(25, "SimpRepl() environment rules:", r_env);
	int loop_lev_old = m_SIMP_SUM_LOOP_LEVEL;
	m_SIMP_SUM_LOOP_LEVEL=loop_level;
	int iFlags=(bFast?1:0);

	CEPtrArray* data_arr = new CEPtrArray();
	//data_arr #0: iFlags
	data_arr->add(new CENumber(iFlags));
	//data_arr #1: r_simp_set
	data_arr->add(r_simp_set);
	//data_arr #2: r_env
	if(r_env!=NULL)data_arr->add(r_env);
	CERule* r = new CESocketRule((void*)CSubstSimplify::SimpSumRuleProc, data_arr,
		CESocketRule::SR_APPLY_TO_SUBITEMS_FIRST);
	data_arr->clearall_without_delete_objects();
	delete data_arr;

	source = r->ApplyTo(source);

	delete r;
	dbg->log(25, "SimpRepl() result:", source);
	m_SIMP_SUM_LOOP_LEVEL=loop_lev_old;
	return source;
}

/*
*	<TODO>: applying rules to every item in sum.
*	Modified item should be reduced, but the whole sum -- arranged only.
*/
CExpression* CSubstSimplify::SimpSumRuleProc(CExpression* src, CEPtrArray* data_arr) {
	if ((src == NULL) || (data_arr == NULL)) return src;
	if (!src->getProperty(CEProperties::PROP_MASK_IS_SUM)) return src;
	int iFlags=((CENumber*)data_arr->GetAt(0))->getLong();
	if((iFlags&1)==1){ //TODO: use constants here
		//'fast' mode
		if(CRCollectSymbols::CountObjects(src) > 40)return src;
	}

	dbg->log(40, "SimpSumRuleProc()", src);
	CEReplList* r_simp_set = (CEReplList*)data_arr->GetAt(1);
	CERule* r_env = (data_arr->size()>=3)?((CERule*)data_arr->GetAt(2)):NULL;
	int s = r_simp_set->size();

	// Trying start replacements in different sequences is useless if no evironment rules involved.
	if (r_env == NULL) m_SIMP_SUM_LOOP_LEVEL=1;

	int* rules_used_info = (int*)CMemoryManager::mmalloc(s*sizeof(int));
	bool src_changed = false;
	for(;;) {
		int j;
		for (j=0; j<s; j++) rules_used_info[j] = 0; // set no rules are used
		CExpression* src_simple = SimpSumRuleProcImpl(src, r_simp_set, r_env, rules_used_info);
		if (src_simple != src) {
			dbg->log(55, "SimpSumRuleProc() simplified:", src_simple);
			if (src_changed) delete src;
			else src_changed = true;
			src = src_simple;
			// make next simplification attempt.

			//////////////////////////////////// no simplification repeatition required!
			break;
		}
		else break; // stop simplification;
	}
	CMemoryManager::mfree(rules_used_info);

	// if result is not the sum --> run this task for it's sub-items.
	if (src_changed) {
		if (!src->getProperty(CEProperties::PROP_MASK_IS_SUM))  {
			src = SimpRepl(src, r_simp_set, r_env);
		}
	}

	dbg->log(45, "SimpSumRuleProc() result:", src);
	return src;
}

/*
*	Applies one of the simplification rules to source and calls itself for next rule.
*	Returns the simpliest found form of source.
*	Source is supposed to be the sum.
*	Does no delete source.
*	"rules_used_info" contain '1' if rule is already used and '0' if not.
*/
CExpression* CSubstSimplify::SimpSumRuleProcImpl(CExpression* src, CEReplList* r_simp_set, CERule* r_env, int* rules_used_info, int loop_cnt) {
	dbg->log(80, "SimpSumRuleProcImpl():", src);

	CExpression* res = src; // the simpliest of all the forms.

	//int loop_cnt = 0;
	int s = r_simp_set->size();
	int i;

	/*long sum_i=0; long sum_s=0;
	if (res->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
		sum_s=((CESum*)res)->size();
	} else {
		s=0;
	}*/

	for (i=0; i<s; i++) {
		if (rules_used_info[i] != 0) continue; // rule #i is already applied;
		dbg->log(90, i, "SimpSumRuleProcImpl() #%ld");
		CEReplasement* r = r_simp_set->get(i);

		// apply simplification replacement.
		CExpression* cpy=NULL;
		/*if (sum_i < sum_s) {
			// try aplly to every item in sum
			cpy = res->clone();
			cpy = r->ApplyTo(res->clone()); // start form source every time
			cpy = CExpression::reduce(cpy);
		} else {
			// (???) & to whole sum afterwards
			cpy = r->ApplyTo(res->clone()); // start form source every time
			cpy = CExpression::reduce(cpy);
		}*/
		DWORD r_flags = r->GetOperationFlags();
		r->SetOperationFlags(r_flags | CEReplasement::REPL_NOT_IN_SUBSUMS);
		r->ResetTimesApplied();
		cpy = r->ApplyTo(res->clone());
		r->SetOperationFlags(r_flags); // restore flags
		if (r->GetTimesApplied() == 0) { delete cpy; continue; }

		cpy = CExpression::reduce(cpy);

		// apply environment rules
		if (r_env != NULL) { cpy = CExpression::reduce(r_env->ApplyTo(cpy)); }

		// try to factorize result.
		CExpression* tmp = cpy->clone();
		CExpression* f_res = CLFactor::factorize(tmp, CLFactor::DONOT_FACT_FRACTIONS);
		if (f_res != NULL) {
			if (f_res != tmp) { // note: tmp is already deleted
				if (!f_res->getProperty(CEProperties::PROP_MASK_IS_SUM)) {
					delete cpy;
					if (res != src) delete res; // may it occur?
					res = f_res;
					break; // stop rules applying
				}
			}
			delete f_res;
		}

		// compare new simplified value to previous one.
		// compare sums (by number of items).
		if (is_a_simplier_b(cpy, res)) { if(res!=src) {delete res;} res=cpy; } else { delete cpy; }

		// Or if the factorized expression is already found.
		if (!res->getProperty(CEProperties::PROP_MASK_IS_SUM)) break;
	}

	// old: (...) check & uncomment next to allow multi-sequence simplification
	// to be placed after the rule#i was apllied
	/*if (loop_cnt < m_SIMP_SUM_LOOP_LEVEL) {
		// apply all the replacements again
		// rules_called_data structure may not be copied if flag is unset after use.
		rules_used_info[i] = 1; // mark this rule as 'used'
		CExpression* res_new = CSubstSimplify::SimpSumRuleProcImpl(cpy, r_simp_set, r_env, rules_used_info, loop_cnt+1);
		rules_used_info[i] = 0; // unset mark.
		if (res_new != cpy) { delete cpy; cpy=res_new; }
	}*/

	dbg->log(85, "SimpSumRuleProcImpl() result:", res);
	return res;
}

bool CSubstSimplify::is_a_simplier_b(CExpression *a, CExpression *b) {
	long s_a = 0;
	long s_b = 0;

	if (a->getProperty(CEProperties::PROP_MASK_IS_SUM))
		s_a = ((CESum*)a)->size();
	if (b->getProperty(CEProperties::PROP_MASK_IS_SUM))
		s_b = ((CESum*)b)->size();

	return (s_a < s_b);
}
