////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// AbstractFileParser.cpp: implementation of the CAbstractFileParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const int CAbstractFileParser::SYM_UNDEFINED		= 0x00;
const int CAbstractFileParser::SYM_SPACE			= 0x01;
const int CAbstractFileParser::SYM_LINE_BREAK		= 0x02;
const int CAbstractFileParser::SYM_END_OF_BUFFER	= 0x03;//INT_MIN
const int CAbstractFileParser::SYM_TABULATOR		= 0x04;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAbstractFileParser::CAbstractFileParser(const char* _source, long len) : CEUtil()
{
	m_bDontSkipLnBreaksIfNoSpacesFollows=false;
	m_source = _source;
	m_buffer_length = len;

	m_curr_pos = 0;
	m_curr_line_num = 1;
	m_curr_line_pos = 1;
	m_operand_types_data = new CEPtrArray();

	//m_operand_types_data->add(CParserSymbolsList::create_list(
	//	CAbstractFileParser::SYM_SPACE, " ", CParserSymbolsList::READ_BY_STRING));
	//m_operand_types_data->add(CParserSymbolsList::create_list(
	//	CAbstractFileParser::SYM_TABULATOR, "\t", CParserSymbolsList::READ_BY_STRING));

	m_last_item_type = SYM_UNDEFINED;
	m_last_sym_pool = NULL;
	m_curr_sym_pool = NULL;
}

CAbstractFileParser::~CAbstractFileParser() {
	if (m_operand_types_data != NULL) delete m_operand_types_data;
}

char* CAbstractFileParser::LoadFileToMemory(const char *fn) {
	return LoadFileToMemory(fn, NULL, 0);
}

const int CAbstractFileParser::LOAD_AS_BINARY			= 0x01;
const int CAbstractFileParser::LOAD_FORCE_CREATE		= 0x02;
const int CAbstractFileParser::LOAD_FULL_LINE_BREAK		= 0x04;
/*
*	used buffer size is filesize-lines_num(linebreak is two symbol long)+1('\0' symbol)
*/
char* CAbstractFileParser::LoadFileToMemory(const char *fn, long* len, int load_opts)
{
	bool force_create = ((load_opts & LOAD_FORCE_CREATE) == LOAD_FORCE_CREATE);
	bool load_as_text = ((load_opts & LOAD_AS_BINARY) == 0);
	bool line_break_full = ((load_opts & LOAD_FULL_LINE_BREAK) == LOAD_FULL_LINE_BREAK);

	// TODO: implement nocreate option here
	//if (!force_create) { open_flags = open_flags|ios::nocreate; }
#ifdef ALHEP_WIN
	// Note: linux-based reading is similar for filebuf::text and filebuf::binary modes
	ifstream ifs(fn, (load_as_text)?(ios::in):(ios::in|ios::binary));
#else
	ifstream ifs(fn, ios::in);
#endif // ALHEP_WIN
	if (!ifs) {printf("<file \"%s\" not found.>\n", fn);return NULL;}

	//0x3fffff;	//  4Mb buffer
	//0xffffff;	//  16Mb buffer
	//0xffff;	// 64kb buffer (for tests)

	long buf_size = 0x100; //256b
	long buf_size_increment = buf_size; //256b
	long buf_size_max_allowed = 0x4000000; //64Mb
	char* buf = (char*)CMemoryManager::mmalloc(buf_size);
	//char* buf_curr = buf;
	long buf_pos = 0;
	long lines_read = 1;	// zero-size file also contains one (empty) line.

	bool loop = true;
	for(;loop;) {
		
		int c=ifs.get();
		if (c != EOF) {
			if (c == '\n') {
				lines_read++;
				if (line_break_full) { // add extra 0x0d symbol (required in some of Windows utilites)
					*(buf+buf_pos*sizeof(char)) = (char)(0x0d); buf_pos++;
				}
			}

			if (c == 0x0d) {
				if (!line_break_full) { // do not add 0x0d symbol if spotted
					continue;
				}
			}

			// add symbol
			*(buf+buf_pos*sizeof(char)) = (char)c;
			buf_pos++;

		}
		else {
			loop = false;
		}

		// we should leave the memory for '\0' symbol and for extra 2 symbols for (line_break_full) possible line break
		if (buf_pos >= (buf_size-2)) {
			long buf_new_size = buf_size + buf_size_increment;
			if (buf_new_size > buf_size_max_allowed) {
				printf("<Error: file %s is too long.>\n", fn);
				loop = false;
			}
			else {
				// realloc buffer
				char* buf_new = CMemoryManager::mReallocChar(buf, buf_new_size);
				if (buf_new != NULL) {
					buf = buf_new;
					buf_size = buf_new_size;
					buf_size_increment = buf_size; // all the increments will double the buffer size
				} else {
					// we don't know whether 'buf' is still valid
					CMemoryManager::mfree(buf);
					printf("<Error memore allocation of size %ld.>\n", buf_new_size);
					buf = NULL;
					loop = false;
				}
			}
		}
	}
	ifs.close();

	if (buf != NULL) {
		*(buf+buf_pos*sizeof(char)) = '\0';
		printf("<file \"%s\" is loaded to memory, size: %ld, %ld line(s).>\n", fn, buf_pos+1, lines_read);
	}
	if (len != NULL) { *len = buf_pos; }
	return buf;
}

char* CAbstractFileParser::get_next_item() {
	skip_spaces();
	
	// search for item end
	int curr_type = get_curr_sym_type();
	m_last_item_type = curr_type;
	if (curr_type == SYM_END_OF_BUFFER) return NULL;

	long start_pos = m_curr_pos;
	int read_mode = get_curr_sym_read_mode();
	if (read_mode == CParserSymbolsList::READ_BY_STRING) {
		for(;;) {
			skip_symbol();
			if (get_curr_sym_type() != curr_type) break;
		}
	}
	else {// (read_mode == CParserSymbolsList::READ_BY_SYMBOL)
		skip_symbol();
	}
	
	char* res = copyStringFromTo(m_source, start_pos, m_curr_pos-1);
	return res;
}

int CAbstractFileParser::get_last_item_type() {
	return m_last_item_type;
}

char* CAbstractFileParser::read_string_until(int break_symbols,
											 int error_symbols,
											 bool skip_BS)
{
	long start_pos = m_curr_pos;
	for(;;) {
		int curr_type = get_curr_sym_type();
		if (curr_type == break_symbols) break;
		if (curr_type == error_symbols) return NULL;
		if (curr_type == SYM_END_OF_BUFFER) return NULL;
		skip_symbol();
	}

	char* res = NULL;
	if (start_pos == m_curr_pos) res = CEUtil::copyString("");
	else res = copyStringFromTo(m_source, start_pos, m_curr_pos-1);
	if (skip_BS) skip_symbol();
	return res;
}

char* CAbstractFileParser::read_string_untilC(char sym, int error_symbols, bool skip_BS) {
	long start_pos = m_curr_pos;
	for(;;) {
		int curr_type = get_curr_sym_type();
		char c = peak_next_sym();
		if (c == sym) break;
		if (curr_type == error_symbols) return NULL;
		if (curr_type == SYM_END_OF_BUFFER) return NULL;
		skip_symbol();
	}

	char* res = NULL;
	if (start_pos == m_curr_pos) res = CEUtil::copyString("");
	else res = copyStringFromTo(m_source, start_pos, m_curr_pos-1);
	if (skip_BS) skip_symbol();
	return res;
}

char CAbstractFileParser::peak_sym(int offset) {
	// current symbol is 'next' (at m_curr_pos)
	return *(m_source + m_curr_pos + offset);
}

char CAbstractFileParser::peak_next_sym() {return peak_sym(0);}
char CAbstractFileParser::peak_prev_sym() {return peak_sym(-2);}

void CAbstractFileParser::skip_spaces() {
	bool loop = true;
	for (;;) {
		char c = *(m_source + m_curr_pos);
		if (c == '\0') break;//data buffer end exceeded

		//if ((c != ' ') && (c != 0x0d) && (c != '\t') && (c != '\n')) return; // stop skipping
		//m_bDontSkipLnBreaksIfNoSpacesFollows

		bool skip_it=false;
		if (m_bDontSkipLnBreaksIfNoSpacesFollows) { // for Mathematica-files line breaks
			skip_it = ((c==' ') || (c==0x0d) || (c=='\t'));
			if ((!skip_it) && (c=='\n')) {
				for(int iNext=1;;iNext++) {
					char cNext = *(m_source+m_curr_pos+iNext);
					if ((cNext==' ')||(cNext=='\t')) { skip_it=true; break; }

					if (cNext==0x0d) {
						char cNext2 = *(m_source+m_curr_pos+iNext+1);
						if (cNext2=='\n') {
							skip_symbol(); skip_symbol(); iNext-=2;
							continue;
						}
					}
					if (cNext=='\n') {
						skip_symbol(); iNext--;
						continue;
					}
					break;
				}

				/*
				for(int iNext=1;;iNext++) {
					char cNext = *(m_source+m_curr_pos+iNext);
					if ((cNext==' ')||(cNext=='\t')) { skip_it=true; break; }
					if (cNext==0x0d) continue;
					if (cNext=='\n') continue;
					break;
				}
				*/
			}
		} else { // usual
			skip_it = ((c==' ') || (c==0x0d) || (c=='\t') || (c=='\n'));
		}

		if (!skip_it) return; // stop skipping
		skip_symbol(); // skip next
	}
}

char CAbstractFileParser::skip_symbol() {
	char sym_curr = *(m_source + m_curr_pos);
	
	if (m_buffer_length >= 0) {
		if (m_curr_pos >= m_buffer_length) return sym_curr; // nothing to skip
	} else {
		if (sym_curr == '\0') { return sym_curr; } // nothing to skip
	}

	if (sym_curr == '\n') {
		m_curr_line_num++; m_curr_line_pos = 1;
	} else {
		m_curr_line_pos++;
	}
	m_curr_pos++;
	m_last_sym_pool = m_curr_sym_pool; m_curr_sym_pool = NULL; // NULL mean undefined
	return sym_curr;
}

int CAbstractFileParser::get_curr_sym_type() {
	if (m_curr_sym_pool != NULL) return m_curr_sym_pool->get_symbols_type();

	if ((m_buffer_length >= 0) && (m_curr_pos >= m_buffer_length)) return CAbstractFileParser::SYM_END_OF_BUFFER;

	// Note: '\n' == 0x0a, Win-notation is {0x0d,0x0a}
	// 0x0d should be treated as SPACE therefore
	// 0x0d may occur if file was loaded in binary mode or with LOAD_FULL_LINE_BREAK option
	char sym = *(m_source + m_curr_pos);

	if (sym == ' ') return CAbstractFileParser::SYM_SPACE;
	if (sym == 0x0d) return CAbstractFileParser::SYM_SPACE;
	if ((m_buffer_length < 0) && (sym == '\0')) return CAbstractFileParser::SYM_END_OF_BUFFER;
	if (sym == '\n') return CAbstractFileParser::SYM_LINE_BREAK;
	if (sym == '\t') return CAbstractFileParser::SYM_TABULATOR;

	// first check for fitting previous type (as posterior symbol)
	if (m_last_sym_pool != NULL) {
		if (m_last_sym_pool->contain_symbol(sym, true)) {
			m_curr_sym_pool = m_last_sym_pool;
			return m_last_sym_pool->get_symbols_type();
		}
	}

	// search for symbol pool
	int s = m_operand_types_data->size();
	int i;
	for(i=0; i<s; i++) {
		CParserSymbolsList* sym_list = (CParserSymbolsList*)m_operand_types_data->GetAt(i);
		if (sym_list->contain_symbol(sym)) {
			m_curr_sym_pool = sym_list;
			return sym_list->get_symbols_type();
		}
	}
	return SYM_UNDEFINED;
}

int CAbstractFileParser::get_curr_sym_read_mode() {
	if (m_curr_sym_pool != NULL) return m_curr_sym_pool->get_read_mode();

	char sym = *(m_source + m_curr_pos);

	if (sym == '\n') return CParserSymbolsList::READ_BY_SYMBOL;
	if (sym == '\0') return CParserSymbolsList::READ_BY_SYMBOL;

	int s = m_operand_types_data->size();
	int i;
	for(i=0; i<s; i++) {
		CParserSymbolsList* sym_list = (CParserSymbolsList*)m_operand_types_data->GetAt(i);
		if (sym_list->contain_symbol(sym)) {
			m_curr_sym_pool = sym_list;
			return sym_list->get_read_mode();
		}
	}
	return CParserSymbolsList::READ_BY_STRING; //mode for undefined values
}

int CAbstractFileParser::get_curr_line_num() { return m_curr_line_num; }
int CAbstractFileParser::get_curr_line_pos() { return m_curr_line_pos; }

void CAbstractFileParser::skip_N_symbols(int N) {
	if (m_buffer_length < 0) {
		for(long i=0; i<N; i++) { if (skip_symbol() == SYM_END_OF_BUFFER) {break;} }
		return;
	}

	// else fast skip
	if (N <= 0) return;
	if (m_curr_pos+N >= m_buffer_length) { N=(int)(m_buffer_length-m_curr_pos); }

	// Note: m_curr_line_pos counter is not keeped
	m_curr_pos += N;
	m_last_sym_pool = NULL; m_curr_sym_pool = NULL; // NULL mean undefined
}
