////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// FileParserAdv.cpp: implementation of the CFileParserAdv class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CFileParserAdv::dbg = new CDebug("CFileParserAdv");

const char* CFileParserAdv::STR_EOF_OPERAND		= "EOF";
const char* CFileParserAdv::STR_INVALID_OPERAND	= "ERROR";

// note: 0x01 - 0x07 are system symbols (space, tabulator, EOLine, EOF)
const int CFileParserAdv::OP_INVALID		= 0x08;
const int CFileParserAdv::OP_NO_OPERAND		= 0x09;
const int CFileParserAdv::OP_END_OF_LINE	= 0x0f;
const int CFileParserAdv::OP_SLASH			= 0x0e;
const int CFileParserAdv::OP_COLON			= 0x0d;

// 1Xh are symbolic operands
const int CFileParserAdv::OP_OPERAND_ANY	= 0x10;
const int CFileParserAdv::OP_VARIABLE		= 0x11;
const int CFileParserAdv::OP_FUNCTION_CALL	= 0x12;

// 2Xh are operations
const int CFileParserAdv::OP_OPERATION_ANY	= 0x20;
const int CFileParserAdv::OP_OPERATION_LEFT	= 0x21;
const int CFileParserAdv::OP_OPERATION_RIGHT	= 0x22;

// 4Xh are brackets
const int CFileParserAdv::OP_BRACKET_ANY		= 0x40;
const int CFileParserAdv::OP_BRACKET_LEFT		= 0x41;
const int CFileParserAdv::OP_BRACKET_RIGHT		= 0x42;
const int CFileParserAdv::OP_BRACKET_LEFT_2		= 0x43;
const int CFileParserAdv::OP_BRACKET_RIGHT_2	= 0x44;

// 8Xh are values
const int CFileParserAdv::OP_VALUE_ANY		= 0x80;
const int CFileParserAdv::OP_VALUE_NUMBER	= 0x81;
const int CFileParserAdv::OP_VALUE_STRING	= 0x82;

const int CFileParserAdv::OP_SPECIAL_1		= 0x1001;

// comments notation
int CFileParserAdv::COMMENTS_C		= 0x01; // comments: /*...*/, //...
int CFileParserAdv::COMMENTS_MMOD	= 0x02; // comments: (*...*)

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CFileParserAdv::CFileParserAdv(const char* _source) : CAbstractFileParser(_source) {
	m_operands_list = new CEPtrArray();
	m_list_stack = new CEPtrStack();
	m_push_back_stack = NULL;
	m_next_operand = NULL;
}

CFileParserAdv::~CFileParserAdv() {
	if (m_operands_list != NULL) delete m_operands_list;
	if (m_list_stack != NULL) delete m_list_stack;
	if (m_push_back_stack != NULL) delete m_push_back_stack;
}

const char* CFileParserAdv::get_next_operand(bool skip_next) {
	if (m_next_operand == NULL) read_next_operand();
	const char* res = m_next_operand->getId();
	if (skip_next) skip_next_operand();
	return res;
}

int CFileParserAdv::get_next_op_type() {
	if (m_next_operand == NULL) read_next_operand();
	long type_res = m_next_operand->get_int();
	return type_res;
}

bool CFileParserAdv::skip_next_operand(bool add_to_list) {
	// store current operand in operand list
	if (add_to_list) m_operands_list->add(m_next_operand);
	else {if (m_next_operand != NULL) delete m_next_operand;}
	// set current operand NULL
	m_next_operand = NULL;
	return true;
}

void CFileParserAdv::insert_operand_as_current(const char* c_op, int type) { // before current
	if (m_push_back_stack == NULL) m_push_back_stack = new CEPtrStack();

	// store current operand to m_push_back_stack
	if (m_next_operand != NULL) m_push_back_stack->push(m_next_operand);

	m_next_operand = new CENamedItem(type, CEUtil::copyString(c_op), false);
}

void CFileParserAdv::push_back_last_operand() {
	if (m_push_back_stack == NULL) m_push_back_stack = new CEPtrStack();

	// store current operand to m_push_back_stack
	if (m_next_operand != NULL) m_push_back_stack->push(m_next_operand);
	
	// set last operand as current
	if ((m_operands_list == NULL) || (m_operands_list->size() == 0)) { // error?
		dbg->log(0, "WARNING <push_back_last_operand() Internal Error>: Nothing to push back.");
		return;
	}

	long s = m_operands_list->size();
	m_next_operand = (CENamedItem*)m_operands_list->GetAt(s-1);
	m_operands_list->RemoveAt(s-1, false);
}

int CFileParserAdv::get_operation_offset(int type, bool from_start, int last_found_pos) {
	dbg->log(99, last_found_pos, from_start, "get_operation_offset: last %ld, from start: %ld");
	long s = m_operands_list->size();
	long i;
	long res=INT_MIN;
	if (from_start) {
		if (last_found_pos<0) { // incl. INT_MIN
			i=0;
		} else {
			i=last_found_pos+1;
		}
		
		for (; i<s; i++) { if (((CENamedItem*)m_operands_list->GetAt(i))->get_int() == type) { res=i; break; } }
		if ((res == INT_MIN) && (m_next_operand != NULL)) { if (m_next_operand->get_int() == type) res=s; }
	} else {
		if (last_found_pos==INT_MIN) {
			i=s;
		} else {
			i=s+last_found_pos-1;
		}
		
		if (i==s) {
			if (m_next_operand != NULL) { if (m_next_operand->get_int() == type) res=0; }
			i=s-1;
		}
		if (res == INT_MIN) {
			for (; i>=0; i--) { if (((CENamedItem*)m_operands_list->GetAt(i))->get_int() == type) { res=(i-s); break; } }
		}
	}

	dbg->log(99, res, "get_operation_offset result: %ld");
	return res; //return ((from_start)?(-1):(INT_MIN));
}

void CFileParserAdv::LogReadOperandsList(int level) {
	if (dbg->get_level()<=level) return;
	CEString _S("{");
	long s = m_operands_list->size();
	long i=0;
	for (; i<s; i++) {
		if (i>0) _S.AddString(" ");
		CENamedItem* item = (CENamedItem*)m_operands_list->GetAt(i);
		if (item == NULL) { _S.AddString("NULL"); continue; }
		_S.AddString("(x"); _S.add_int(item->get_int(),true); _S.AddString(")"); _S.AddString(item->getId());
	}

	if (m_next_operand != NULL) {
		if (i>0) _S.AddString(" ");
		_S.AddString("(x"); _S.add_int(m_next_operand->get_int(),true); _S.AddString(")"); _S.AddString(m_next_operand->getId());
	}
	_S.AddString("}");
	dbg->log(level, "Read list: ", _S.getData());
}

CENamedItem* CFileParserAdv::get_operand_cart(int offset, bool from_start) {
	long s = m_operands_list->size();
	if (!from_start) offset = s + offset;
	dbg->log(40, offset, "get_operand_cart no. %ld");

	// check offset
	if (offset == s) return m_next_operand;
	if ((offset > s) || (offset < 0)) return NULL;

	// return value from previous operands list
	CENamedItem* res = (CENamedItem*)m_operands_list->GetAt(offset);
	return res;
}

int CFileParserAdv::get_operands_num() { return m_operands_list->size(); }

const char* CFileParserAdv::get_operand(int offset, bool from_start) {
	CENamedItem* op_cart = get_operand_cart(offset, from_start);
	if (op_cart == NULL) return NULL;
	return op_cart->getId();
}

int CFileParserAdv::get_operand_type(int offset, bool from_start) {
	CENamedItem* op_cart = get_operand_cart(offset, from_start);
	if (op_cart == NULL) return CFileParserAdv::OP_NO_OPERAND;
	return op_cart->get_int();
}

void CFileParserAdv::start_new_operands_list() {
	m_list_stack->push(m_operands_list);
	m_operands_list = new CEPtrArray();
}

void CFileParserAdv::finish_operands_list() {
	if (m_operands_list != NULL) delete m_operands_list;
	m_operands_list = (CEPtrArray*)m_list_stack->pop();
}

bool CFileParserAdv::ReadPushedBackOperand() {
	if (m_next_operand != NULL) skip_next_operand(true); // unused check

	if (m_push_back_stack != NULL) {
		CEObject* val = m_push_back_stack->pop();
		if (val != NULL) {
			m_next_operand = (CENamedItem*)val;
			dbg->log(70, "ReadPushedBackOperand()", m_next_operand);
			return true;
		}
	}
	dbg->log(170, "no pushed back operands found");
	return false;
}

// may call itself recursively
bool CFileParserAdv::read_next_operand() {
	if (ReadPushedBackOperand()) return true; // fisrt read data from push-back stack
	if (m_next_operand != NULL) { dbg->log(0, "Internal ERROR at CFileParserAdv::read_next_operand(): m_next_operand != NULL", m_next_operand); CEException::halt(); }

	// read data from straem
	char* op_id = get_next_item();
	int op_type = get_last_item_type();
	dbg->log(20, op_type, "next operand read (type == 0x%lx):");

	////////////////////////////////////////////////////////////////////
	// preliminary type parsing BEGIN
	////////////////////////////////////////////////////////////////////
	if (op_type == CAbstractFileParser::SYM_END_OF_BUFFER) {
		if (op_id == NULL) op_id = CEUtil::copyString(STR_EOF_OPERAND);
	}
	if (op_id == NULL) return false;

	if (op_type == CAbstractFileParser::SYM_LINE_BREAK) {
		// may occur if only CAbstractFileParser::m_bDontSkipLnBreaksIfNoSpacesFollows=true;

		// ignore line-break if comments follows
		bool bIgnoreLineBreak=false;
		if (m_CommentsType==COMMENTS_C) {
			if ((peak_sym(0)== '/') && ((peak_sym(1)== '/') || (peak_sym(1)== '*'))) bIgnoreLineBreak=true;
		} else if (m_CommentsType==COMMENTS_MMOD) {
			if ((peak_sym(0)== '(') && (peak_sym(1)== '*')) bIgnoreLineBreak=true;
		}

		if (bIgnoreLineBreak) {
			return read_next_operand();
		} else {
			op_type=OP_END_OF_LINE;
		}
	}
	
	if (op_type & CFileParserAdv::OP_VALUE_ANY) {
		// numbers
		if (op_type == CFileParserAdv::OP_VALUE_NUMBER) {
			// nothing to do, copmpiller will do everything
		}
		// strings
		else if (op_type == CFileParserAdv::OP_VALUE_STRING) {
			// read the whole string and place it instead of op_id;
			// escape-sequences should be taken into acount ... if there two \"-s together(???)
			int id_len = CEUtil::length(op_id);
			switch(id_len) {
			case 1: {
				char* str = read_string_until(CFileParserAdv::OP_VALUE_STRING, CAbstractFileParser::SYM_END_OF_BUFFER, true);
				if (str != NULL) {
					CMemoryManager::mfree(op_id);
					op_id = str;
				} else {// Error
					op_type = OP_INVALID;
				}
					} break;
			case 2:
				// the empty string read
				CMemoryManager::mfree(op_id);
				op_id = CEUtil::copyString("");
				break;
			default:
				op_type = OP_INVALID; // Error
			}
		}
	} else {
		bool _SlashToDivide=(op_type == CFileParserAdv::OP_SLASH);

		// possible skip comments
		if (m_CommentsType==COMMENTS_C) {
			if (op_type == CFileParserAdv::OP_SLASH) {
				char c_next = peak_next_sym();
				if (c_next == '/') { // skip till end of line
					_SlashToDivide=false;
					char* str = read_string_until(CAbstractFileParser::SYM_LINE_BREAK, CAbstractFileParser::SYM_END_OF_BUFFER, true);
					dbg->log(20, "comments line skiped:"); dbg->log(20, str);
					CMemoryManager::mfree(op_id); op_id=NULL; CMemoryManager::mfree(str);
					return read_next_operand(); // further read
				}
				else if (c_next == '*') { // skip till "*/"-string
					_SlashToDivide=false;
					for(;;) {
						char* str = read_string_until(CFileParserAdv::OP_SLASH, CAbstractFileParser::SYM_END_OF_BUFFER, true);
						if (str == NULL) { // EOF or internal error
							op_type = OP_INVALID; break;
						}
						CMemoryManager::mfree(str);
						char c_prev = peak_prev_sym();
						if (c_prev == '*') {
							dbg->log(20, "comments skiped");
							CMemoryManager::mfree(op_id); op_id=NULL;
							return read_next_operand(); // further read
						}
					}
				}
			}
		} else if (m_CommentsType==COMMENTS_MMOD) {
			//if (op_type == CFileParserAdv::OP_BRACKET_LEFT) {
			if (*op_id == '(') {
				char c_next = peak_next_sym();
				if (c_next == '*') { // skip till "*)"-string
					for(;;) {
						//char* str = read_string_until(CFileParserAdv::OP_BRACKET_RIGHT, CAbstractFileParser::SYM_END_OF_BUFFER, true);
						char* str = read_string_untilC(')', CAbstractFileParser::SYM_END_OF_BUFFER, true);
						if (str == NULL) { op_type = OP_INVALID; break; } // EOF or internal error
						CMemoryManager::mfree(str);
						char c_prev = peak_prev_sym();
						if (c_prev == '*') {
							dbg->log(20, "comments skiped");
							CMemoryManager::mfree(op_id); op_id=NULL;
							return read_next_operand(); // further read
						}
					}
				}
			}  
		}
		
		if ((_SlashToDivide) && (op_type == CFileParserAdv::OP_SLASH)) { // 'op_type' may be already changed
			// return divide operation
			op_type = OP_OPERATION_RIGHT; //op_type = OP_INVALID;
		}
	}
	////////////////////////////////////////////////////////////////////
	// preliminary type parsing END
	////////////////////////////////////////////////////////////////////

	CENamedItem* itemNew = new CENamedItem(op_type, op_id, false);
	bool bAccept = OnNewOperandRead(itemNew);
	if (bAccept) {
		m_next_operand = itemNew;
		if (*(itemNew->getId()) == '\n') { dbg->log(20, "EndOfLine"); } else dbg->log(20, "Read operand: ", itemNew->getId());
	} else {
		if (*(itemNew->getId()) == '\n') { dbg->log(20, "Dropped operand: EndOfLine"); } else dbg->log(20, "Dropped operand: ", itemNew->getId());
		delete itemNew;
		return read_next_operand();
	}
	return true;
}

/*
*	Used to modify read value in childs, called in read_next_operand().
*	Returns 'false' to drop operand and call read_next_operand() again.
*/
bool CFileParserAdv::OnNewOperandRead(CENamedItem* item) { return true; }
