////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// LFileManager.cpp: implementation of the CLFileManager class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////// CLFileManager constants
//
const DWORD CLFileManager::FT_F90							= 0x000001;
const DWORD CLFileManager::FT_F77							= 0x000002;
const DWORD CLFileManager::FT_CPP							= 0x000004;
const DWORD CLFileManager::FT_POWER_PARAMS					= 0x000008;
const DWORD CLFileManager::FT_REAL_8BIT						= 0x000010;
const DWORD CLFileManager::FT_COMPLEX_16BIT					= 0x000020;
const DWORD CLFileManager::FT_DONOT_CREATE_4VECTOR_ENTRY	= 0x000040;
const DWORD CLFileManager::FT_DONOT_SPLIT_FUNC				= 0x000080;
const DWORD CLFileManager::FT_SEPARATE_VIRT_INTEGRALS		= 0x000100;
const DWORD CLFileManager::FT_NO_COMMON_PARAMETERS			= 0x000200;
const DWORD CLFileManager::FT_NO_COMMON_ARRAYS				= 0x000400;
const DWORD CLFileManager::FT_LOOPTOOLS						= 0x000800;
const DWORD CLFileManager::FT_DONOT_INVOLVE_CONSTANTS		= 0x001000;
const DWORD CLFileManager::FT_CHECK_DENOMINATORS			= 0x002000;
const DWORD CLFileManager::FT_PYTHIA_VECTORS				= 0x004000;
const DWORD CLFileManager::FT_SRC_IS_AMPLITUDE				= 0x008000;
const DWORD CLFileManager::FT_SRC_IS_DIFFCS					= 0x010000;
const DWORD CLFileManager::FT_SRC_SAVE_AS_IS				= 0x010000; //same to FT_SRC_IS_DIFFCS
const DWORD CLFileManager::FT_NO_INCLUDES					= 0x020000; // do not separate constants to include file
const DWORD CLFileManager::FT_FULL_INCLUDE					= 0x040000; // create include file with all possible constants. WARNING: no names check!
const DWORD CLFileManager::FT_CPP_TYPE_ABSTRACT				= 0x080000; // (C++ only) abstract 'tReal'/'tComplex' types instead of 'double'/'complex'
const DWORD CLFileManager::FT_FORCE_REAL					= 0x100000; //ignore imaginary parts of every function/variable
const DWORD CLFileManager::FT_ALLOW_SWAP_INITIALS			= 0x200000; //add initial particles swap mechanism
const DWORD CLFileManager::FT_NON_STATIC_IMPL				= 0x400000;	// (C++ only) (default: false) create non-static functions

//////////////////////////////////////////////////// CCodeLanguage
//

// Constants
const char* CCodeLanguage::TYPE_REAL8_F90			= "REAL*8";
const char* CCodeLanguage::TYPE_COMPLEX16_F90		= "COMPLEX*16";
const char* CCodeLanguage::TYPE_REAL8_F77			= "double precision";
const char* CCodeLanguage::TYPE_COMPLEX16_F77		= "double complex";
const char* CCodeLanguage::TYPE_INTEGER				= "INTEGER";
const char* CCodeLanguage::TYPE_REAL8_CPP			= "double";
const char* CCodeLanguage::TYPE_COMPLEX16_CPP		= "complex";
const char* CCodeLanguage::TYPE_INTEGER_CPP			= "long";
const char* CCodeLanguage::TYPE_REAL8_CPP_ABS		= "tReal";		//used if CLFileManager::FT_CPP_TYPE_ABSTRACT
const char* CCodeLanguage::TYPE_COMPLEX16_CPP_ABS	= "tComplex";	//used if CLFileManager::FT_CPP_TYPE_ABSTRACT
const char* CCodeLanguage::TYPE_4VEC_CPP			= "CVM4Vec";
const char* CCodeLanguage::TYPE_BOOL_CPP			= "bool";

//
const char* CCodeLanguage::CPP_POW_DEF_SYM			= "powX";
const char* CCodeLanguage::CPP_REAL8_CONV_ADV		= "DBLtR";
const char* CCodeLanguage::CPP_CMPLX_C16_CONV		= "CMPLtC";
const char* CCodeLanguage::CPP_CMPL16_R8_CONV		= "Re";
const char* CCodeLanguage::CPP_CMPLX_RX_CONV		= "Re";

//
// TODO: add line-breaking after m_LINEBREAK_SYMS symbols
// for F77 code line breaking, the break is suspected BEFORE 'pos' position
const char* CCodeLanguage::m_LINEBREAK_SYMS = " +-*=/(),";
CDebug* CCodeLanguage::dbg = new CDebug("CCodeLanguage");

//
CCodeLanguage* CCodeLanguage::CreateLanguage(DWORD conventions) {
	//truncated hi-level settings
	DWORD convs_lang = conventions & (CLFileManager::FT_F90 | CLFileManager::FT_F77 | CLFileManager::FT_CPP);
	CCodeLanguage* res = new CCodeLanguage(convs_lang);
	if(((conventions & CLFileManager::FT_CPP_TYPE_ABSTRACT)==CLFileManager::FT_CPP_TYPE_ABSTRACT) && (res!=NULL)) {
		res->SetAbstractAvailable(true);
	}
	return res;
}

//
CCodeLanguage::CCodeLanguage(DWORD conventions) : CEObject(true) {
	m_conventions = conventions;
	m_IndentSize=0;
	m_bIsAbstractAvailable = false;
}
CCodeLanguage::~CCodeLanguage() {}
char* CCodeLanguage::toLine(int mode) { return CEUtil::copyString("CCodeLanguage"); }
const char* CCodeLanguage::CLASS_NAME = "CCodeLanguage";
const char* CCodeLanguage::ClassName() const { return CLASS_NAME; }
bool CCodeLanguage::contain_vector(void * obj) { return false; /*no non-static vectors*/ }

////////////////////////////////////////////////////////////////////// code-write procedures
//
void CCodeLanguage::StartCommentLine(ofstream* p_ofs, bool isHiddenCode) {
	//Note: 'hidden' is a line to be uncommented by user in future

	if(p_ofs==NULL)return;
	if (m_conventions & CLFileManager::FT_F77) {
		*p_ofs << "C";
		if(isHiddenCode){ *p_ofs << "     "; if (m_IndentSize>0) { for(int i=0;i<m_IndentSize;i++) { *p_ofs << " "; } } }else{ *p_ofs << "."; }
	} else if (m_conventions & CLFileManager::FT_F90) {
		*p_ofs << "!";
		if(isHiddenCode){ if (m_IndentSize>0) { for(int i=0;i<m_IndentSize;i++) { *p_ofs << " "; } } }else{ *p_ofs << "."; }
	} else if (m_conventions & CLFileManager::FT_CPP) {
		if(isHiddenCode){
			if (m_IndentSize>0) { for(int i=0;i<m_IndentSize;i++) { *p_ofs << "\t"; } }
			*p_ofs << "//";
		}else{
			*p_ofs << "//.";
		}
	} else  {
		*p_ofs << "C";
		if(isHiddenCode){ *p_ofs << "     "; if(m_IndentSize>0){ for(int i=0;i<m_IndentSize;i++) { *p_ofs << "\t"; } } }else{ *p_ofs << "."; }
	}
}

void CCodeLanguage::OutCodeLine(ofstream* p_ofs, CEString* c_line, int _label, bool bTruncateLine, const char* comment) { // also resets 'c_line'
	OutCodeLine(p_ofs, c_line->getData(), _label, bTruncateLine, comment); c_line->Reset();
}

void CCodeLanguage::OutCodeLine(ofstream* p_ofs, const char* _line, int _label, bool bTruncateLine, const char* comment) {
	//Notes:
	//	this procedure breaks lines for f77/f90 code,
	//	line may be breaked BEFORE 'm_LINEBREAK_SYMS' symbols only (not between "N.N")
	//	'comment' is added after line in C++ mode and a separated line otherwise

	if((comment!=NULL) && ((m_conventions & CLFileManager::FT_CPP) == 0)) { StartCommentLine(p_ofs); *p_ofs<<comment<<"\n"; comment=NULL; }

	if(p_ofs==NULL)return;
	long l = CEUtil::length(_line); if (l<=0) return;
	long len_max = GetCodeLengthMax(); if(len_max<=10) { dbg->log(0, len_max, "Internal ERROR: MAX-CODE-LINE is too small (%ld). IGNORED"); len_max=64; }
	bool _started=false;
	long pos = 0; // == num-of-chars-saved
	for(;(l-pos)>len_max;) {	//Nt: case (l<=len_max) is processed below
		// 1. search for m_LINEBREAK_SYMS-symbol near the #pos+len_max position
		long pos_end = pos+len_max;
		for(;;pos_end--) { // backward search
			if (pos_end <= pos) { pos_end=-1; break; } // not found
			if (IsLineBreakAllowed(_line, pos_end, l)) break;
		}
		if (pos_end < 0) {
			dbg->log(0, "WARNING. Maximum code length exceded. Can't find how to break code line. CHECK THE CODE!");
			for(pos_end=pos+len_max+1;;pos_end++) { // forward search
				if (pos_end >= l) { pos_end=-1; break; } // not found
				if (IsLineBreakAllowed(_line, pos_end, l)) break;
			}
		}
		if (pos_end<0)  break; // stop splitting, out rest

		// 2. write line upto found symbol, do not write #pos_end symbol here
		if(_started) {StartContinuationLine(p_ofs);} else {StartCodeLine(p_ofs,_label); _started=true;}
		for (;pos<pos_end;) { *p_ofs << (char)(*(_line+pos)); pos++; }
		//[old & wrong!]if(m_conventions & CLFileManager::FT_CPP)*p_ofs << ";";
		*p_ofs << "\n";
	}

	// write rest
	if(_started) {StartContinuationLine(p_ofs);} else {StartCodeLine(p_ofs,_label); _started=true;}
	*p_ofs << (char*)(_line+pos);
	if(!bTruncateLine) { if(m_conventions & CLFileManager::FT_CPP)*p_ofs << ";"; }
	if((comment!=NULL) & ((m_conventions & CLFileManager::FT_CPP)==CLFileManager::FT_CPP)) { *p_ofs << " // " << comment; }
	*p_ofs << "\n";
}

void CCodeLanguage::OutConsoleWriteLine(ofstream* p_ofs, CEString* c_line, bool bAddEOL) { //don't need quotes in 'c_line', resets 'c_line'
	if(m_conventions & CLFileManager::FT_CPP){
		c_line->insertString("std::cout<<\""); c_line->AddString("\"");
		if(bAddEOL) c_line->AddString("<<std::endl");
	} else{ c_line->insertString("print *, <<\""); c_line->AddString("\""); }
	OutCodeLine(p_ofs, c_line);
}

////////////////////////////////////////////////////////////////////// type functions
//
int CCodeLanguage::GetDefaultTypeId(DWORD type_convs) {				//uses 'type_convs' to select default type
	int type=CEAbstractFunction::PAR_DEFAULT;
	bool bAbstractVars = ((IsAbstractAvailable()) &&
		((type_convs & CLFileManager::FT_CPP_TYPE_ABSTRACT)==CLFileManager::FT_CPP_TYPE_ABSTRACT));
	if((type_convs & CLFileManager::FT_REAL_8BIT)==CLFileManager::FT_REAL_8BIT) {
		type=(bAbstractVars ? CEAbstractFunction::PAR_REAL_ABSTRACT:CEAbstractFunction::PAR_REAL8);
	}else if(type_convs & CLFileManager::FT_COMPLEX_16BIT){
		type=(bAbstractVars ? CEAbstractFunction::PAR_COMPLEX_ABSTRACT:CEAbstractFunction::PAR_COMPLEX);
	}
	return type;
}
int CCodeLanguage::GetDefaultRealTypeId() {							//returns 'real8' or 'real*X'
	return (IsAbstractAvailable()?CEAbstractFunction::PAR_REAL_ABSTRACT:CEAbstractFunction::PAR_REAL8);
}
const char* CCodeLanguage::GetTypePrefix(int type) { return GetTypePrefixImpl(type); }
const char* CCodeLanguage::GetTypePrefixAdv(DWORD type_convs) { return GetTypePrefixImpl(GetDefaultTypeId(type_convs)); }
const char* CCodeLanguage::GetTypePrefixReal() { return GetTypePrefixImpl(CEAbstractFunction::PAR_REAL_ABSTRACT); }
const char* CCodeLanguage::GetTypePrefixComplex() { return GetTypePrefixImpl(CEAbstractFunction::PAR_COMPLEX_ABSTRACT); }
const char* CCodeLanguage::GetTypePrefixInt() { return GetTypePrefixImpl(CEAbstractFunction::PAR_INTEGER); }
const char* CCodeLanguage::GetTypePrefixBool() {
	bool bCpp = ((m_conventions & CLFileManager::FT_CPP)==CLFileManager::FT_CPP);
	return ((bCpp)?TYPE_BOOL_CPP:(GetTypePrefixInt()));
}
const char* CCodeLanguage::GetTypePrefixImpl(int type) { //cancels 'abstract' type if not allowed by 'IsAbstractAvailable()'

	bool bF77 = ((m_conventions & CLFileManager::FT_F77)==CLFileManager::FT_F77);
	bool bF90 = ((m_conventions & CLFileManager::FT_F90)==CLFileManager::FT_F90);
	bool bCpp = ((m_conventions & CLFileManager::FT_CPP)==CLFileManager::FT_CPP);

	//Cancel 'abstract' type if forbidden ('Real*N'-->'real*8', 'Complex*N'-->'complex*16')
	bool bAbstractAllowed = IsAbstractAvailable();
	if(type==CEAbstractFunction::PAR_REAL_ABSTRACT){ if((!bCpp) || (!bAbstractAllowed))type=CEAbstractFunction::PAR_REAL8; }
	if(type==CEAbstractFunction::PAR_COMPLEX_ABSTRACT){ if((!bCpp) || (!bAbstractAllowed))type=CEAbstractFunction::PAR_COMPLEX; }

	const char* res="UNKNOWN";
	int iLangId=0; if(bCpp){iLangId=1;}else if(bF77){iLangId=2;}else if(bF90){iLangId=3;} //1:C++, 2:f77, 3:f90
	if((type==CEAbstractFunction::PAR_DEFAULT) || (iLangId==0))return res; //'unknown'

	switch(iLangId){
	case 1: //C++
		if (type==CEAbstractFunction::PAR_REAL8) res=TYPE_REAL8_CPP;
		else if (type==CEAbstractFunction::PAR_REAL_ABSTRACT) res=TYPE_REAL8_CPP_ABS;
		else if (type==CEAbstractFunction::PAR_COMPLEX) res=TYPE_COMPLEX16_CPP;
		else if (type==CEAbstractFunction::PAR_COMPLEX_ABSTRACT) res=TYPE_COMPLEX16_CPP_ABS;
		else if (type==CEAbstractFunction::PAR_INTEGER) res=TYPE_INTEGER_CPP;
		else if (type==CEAbstractFunction::PAR_4VEC_REAL) res=TYPE_4VEC_CPP;
		break;
	case 2: //f77
		if ((type==CEAbstractFunction::PAR_REAL8) || (type==CEAbstractFunction::PAR_REAL_ABSTRACT)) res=TYPE_REAL8_F77;
		else if ((type==CEAbstractFunction::PAR_COMPLEX) || (type==CEAbstractFunction::PAR_COMPLEX_ABSTRACT)) res=TYPE_COMPLEX16_F77;
		else if (type==CEAbstractFunction::PAR_INTEGER) res=TYPE_INTEGER;
		break;
	case 3: //f90
		if ((type==CEAbstractFunction::PAR_REAL8) || (type==CEAbstractFunction::PAR_REAL_ABSTRACT)) res=TYPE_REAL8_F90;
		else if ((type==CEAbstractFunction::PAR_COMPLEX) || (type==CEAbstractFunction::PAR_COMPLEX_ABSTRACT)) res=TYPE_COMPLEX16_F90;
		else if (type==CEAbstractFunction::PAR_INTEGER) res=TYPE_INTEGER;
		break;
	}
	return res;
}

/*
*	Mode (!bForce) is used for assignment:
*		tTo a=b;
*	Mode (bForce) is used for explicit conversion, e.g.:
*		tTo bSqr=(tTo)b*(tTo)b; that may differ from '(tTo)(b*b)'.
*		F77FUNC(..,DBLE(1),..)
*/
bool CCodeLanguage::GetTypeConversion(int tFrom, int tTo, CEString* sBefore, CEString* sAfter,
									  DWORD type_convs,
									  CExpression* val,
									  bool bForce) {
	if((sBefore==NULL) || (sAfter==NULL)) return false;
	if(sBefore->length()>0)sBefore->Reset(); if(sAfter->length()>0)sAfter->Reset();

	bool bF77 = ((m_conventions & CLFileManager::FT_F77)==CLFileManager::FT_F77);
	bool bF90 = ((m_conventions & CLFileManager::FT_F90)==CLFileManager::FT_F90);
	bool bCpp = ((m_conventions & CLFileManager::FT_CPP)==CLFileManager::FT_CPP);

	if(val!=NULL){
		//TODO: check if this check is still required!
		//Old: Note: Every number is stored as REAL (like 1D0) or COMPLEX ==> turn to INTERGER or REQUIRED type manually!
		//Old: Set appropriate 'tFrom' type for numeric value
		if(val->ClassName()==CENumber::CLASS_NAME) {
			//Old: if(((CENumber*)val)->IsReal()){ tFrom=CEAbstractFunction::PAR_REAL8; } //real number
			//Old: else{ tFrom=CEAbstractFunction::PAR_COMPLEX; } //complex number
		}
	}

	//Cancel 'abstract' type if forbidden ('Real*N'-->'real*8', 'Complex*N'-->'complex*16')
	bool bAbsForbidden= ((!bCpp) || (!IsAbstractAvailable()));
	if(bAbsForbidden) {
		if(tFrom==CEAbstractFunction::PAR_REAL_ABSTRACT)tFrom=CEAbstractFunction::PAR_REAL8;
		if(tFrom==CEAbstractFunction::PAR_COMPLEX_ABSTRACT)tFrom=CEAbstractFunction::PAR_COMPLEX;
		if(tTo==CEAbstractFunction::PAR_REAL_ABSTRACT)tTo=CEAbstractFunction::PAR_REAL8;
		if(tTo==CEAbstractFunction::PAR_COMPLEX_ABSTRACT)tTo=CEAbstractFunction::PAR_COMPLEX;
	}

	if(tTo==CEAbstractFunction::PAR_DEFAULT)tTo=GetDefaultTypeId(type_convs);
	if(tFrom==CEAbstractFunction::PAR_DEFAULT)tFrom=GetDefaultTypeId(type_convs);
	if(tFrom==tTo)return true; //check after possible changes in 'tFrom' and 'tTo'

	//Cache long checks
	bool bFromReal8=(tFrom==CEAbstractFunction::PAR_REAL8);
	bool bFromRealN=(tFrom==CEAbstractFunction::PAR_REAL_ABSTRACT);
	bool bFromCmpl16=(tFrom==CEAbstractFunction::PAR_COMPLEX);
	bool bFromCmplN=(tFrom==CEAbstractFunction::PAR_COMPLEX_ABSTRACT);
	bool bFromInt=(tFrom==CEAbstractFunction::PAR_INTEGER);
	bool bFrom4Vec=(tFrom==CEAbstractFunction::PAR_4VEC_REAL);
	//bool bFromDef=(tFrom==CEAbstractFunction::PAR_DEFAULT); //No more defaults here

	bool res=false;
	if(bCpp){ //C++
		if (tTo==CEAbstractFunction::PAR_REAL8) {
			if(bFromInt){ res=true; if(bForce){*sBefore<<"("<<TYPE_REAL8_CPP<<")("; *sAfter<<")";} }
			else if(bFromCmpl16){ *sBefore<<CPP_CMPL16_R8_CONV<<"("; *sAfter<<")"; res=true; }
			else if(bFromRealN){ *sBefore<<CPP_REAL8_CONV_ADV<<"("; *sAfter<<")"; res=true; }
			else if(bFromCmplN){ //cmplX->xmpl16->real8
				*sBefore<<CPP_CMPL16_R8_CONV<<"("<<CPP_CMPLX_C16_CONV<<"("; *sAfter<<"))"; res=true;
			}
		} else if (tTo==CEAbstractFunction::PAR_REAL_ABSTRACT) {
			if(bFromInt){ res=true; if(bForce){ *sBefore<<"("<<TYPE_REAL8_CPP_ABS<<")("; *sAfter<<")"; } }
			//Note: real*8 force cast is required because (realN)(a*a) != (realN)a*(realN)a
			//Note: but that cast should be NOT here!
			else if(bFromReal8){ res=true; if(bForce){ *sBefore<<"("<<TYPE_REAL8_CPP_ABS<<")("; *sAfter<<")"; } }
			else if(bFromCmpl16){
				//[cmpl16->real8->realX] bForce: "(realX)((real8)c)", else: "(real8)c"
				if(bForce){ *sBefore<<"("<<TYPE_REAL8_CPP_ABS<<")("<<CPP_CMPL16_R8_CONV<<"("; *sAfter<<"))"; }
				else{ *sBefore<<CPP_CMPL16_R8_CONV<<"("; *sAfter<<")"; }
				res=true;
			}
			else if(bFromCmplN){ *sBefore<<CPP_CMPLX_RX_CONV<<"("; *sAfter<<")"; res=true; }
		} else if (tTo==CEAbstractFunction::PAR_COMPLEX) {
			if(bFromInt||bFromReal8){ res=true; if(bForce){ *sBefore<<"("<<TYPE_COMPLEX16_CPP<<")("; *sAfter<<")"; } }
			else if(bFromRealN){ *sBefore<<CPP_REAL8_CONV_ADV<<"("; *sAfter<<")"; res=true; }//cast to real*8
			else if(bFromCmplN){ *sBefore<<CPP_CMPLX_C16_CONV<<"("; *sAfter<<")"; res=true; }
		} else if (tTo==CEAbstractFunction::PAR_COMPLEX_ABSTRACT) {
			if(bFromInt){ res=true; if(bForce){ *sBefore<<"("<<TYPE_COMPLEX16_CPP_ABS<<")("; *sAfter<<")"; } }
			//Note: real8/cmpl16 force cast is required because (cmplN)(a*a) != (cmplN)a*(cmplN)a
			else if(bFromReal8||bFromCmpl16){ *sBefore<<"("<<TYPE_COMPLEX16_CPP_ABS<<")("; *sAfter<<")"; res=true; }
			else if(bFromRealN){ res=true; if(bForce){ *sBefore<<"("<<TYPE_COMPLEX16_CPP_ABS<<")("; *sAfter<<")"; } }
			//TODO: check/implement 'bFromRealN' conversion after tComplex selection
		} else if (tTo==CEAbstractFunction::PAR_INTEGER) {
			//Note: Every number is stored as REAL (like 1D0) or COMPLEX (and tForm is real or complex here)
			if(bFromReal8||bFromCmpl16){ *sBefore<<"("<<TYPE_INTEGER_CPP<<")("; *sAfter<<")"; res=true; }
			else if(bFromRealN){ *sBefore<<"("<<TYPE_INTEGER_CPP<<")("<<CPP_REAL8_CONV_ADV<<"("; *sAfter<<"))"; res=true; }
			else if(bFromCmplN){ *sBefore<<"("<<TYPE_INTEGER_CPP<<")("<<CPP_CMPL16_R8_CONV<<"("<<CPP_CMPLX_C16_CONV<<"("; *sAfter<<")))"; res=true; }
		} else if (tTo==CEAbstractFunction::PAR_4VEC_REAL) {
			//(val==0) might be casted here
		}
	} else if(bF77 || bF90) {
		//asdf: insert 'bForce' flag here
		//
		//Note: In Fortran we MUST convert parameters to required type (no type-length check is performed at function call)
		bool bFromOK=(bFromReal8 || bFromRealN || bFromCmpl16 || bFromCmplN || bFromInt);
		if ((tTo==CEAbstractFunction::PAR_REAL8) || (tTo==CEAbstractFunction::PAR_REAL_ABSTRACT)) {
			//Note: if(bFromInt)res=true; <-- Wrong! 'FUNC(DBLE(0))' is required
			if(bFromOK){ *sBefore<<"DBLE("; *sAfter<<")"; res=true; }
		} else if ((tTo==CEAbstractFunction::PAR_COMPLEX) || (tTo==CEAbstractFunction::PAR_COMPLEX_ABSTRACT)) {
			if(bFromOK){ sBefore->AddString("DCMPLX("); sAfter->AddString(")"); res=true; }
		} else if (tTo==CEAbstractFunction::PAR_INTEGER) {
			if(bFromOK){ sBefore->AddString("INT("); sAfter->AddString(")"); res=true; }
		}
	}
	return res;
}
/*
*	GetOperatingType: type of operations with variable
*	Example: real8 may be kept/passed as real8, but MUST be operated as 'realX'
*/
int CCodeLanguage::GetOperatingType(int type) { // real8==>realX, cmpl16==>cmplX
	//[old]if(type_convs==0) { type_convs = (IsAbstractAvailable()?CLFileManager::FT_CPP_TYPE_ABSTRACT:0); }
	//[old]if ((type_convs & CLFileManager::FT_CPP_TYPE_ABSTRACT)==CLFileManager::FT_CPP_TYPE_ABSTRACT) {
	if (IsAbstractAvailable()) {
		if(type==CEAbstractFunction::PAR_REAL8)type=CEAbstractFunction::PAR_REAL_ABSTRACT;
		else if(type==CEAbstractFunction::PAR_COMPLEX)type=CEAbstractFunction::PAR_COMPLEX_ABSTRACT;
	}
	//[old]if(type==CEAbstractFunction::PAR_DEFAULT)type=GetDefaultTypeId(type_convs);
	if(type==CEAbstractFunction::PAR_DEFAULT){ dbg->log(0,"WARNING: GetOperatingType(CEAbstractFunction::PAR_DEFAULT)"); }
	return type;
}

////////////////////////////////////////////////////////////////////// operations
//
const char* CCodeLanguage::GetOpIf() { return ((m_conventions & CLFileManager::FT_CPP)?"if":"IF"); }
const char* CCodeLanguage::GetOpEqu() { return ((m_conventions & CLFileManager::FT_CPP)?"==":".EQ."); }
const char* CCodeLanguage::GetOpThen() { return ((m_conventions & CLFileManager::FT_CPP)?"{":"THEN"); }
const char* CCodeLanguage::GetOpElse() { return ((m_conventions & CLFileManager::FT_CPP)?"} else {":"ELSE"); }
const char* CCodeLanguage::GetOpEndIf() { return ((m_conventions & CLFileManager::FT_CPP)?"}":"ENDIF"); }
const char* CCodeLanguage::GetOpEndProc() { return ((m_conventions & CLFileManager::FT_CPP)?"}":"END"); }
void CCodeLanguage::AddOpFor(CEString* ln, const char* iStr, int iMin, int iMax, int iLabel, bool bDeclareI) {
	if(ln==NULL)return;
	if(m_conventions & CLFileManager::FT_CPP){
		*ln<<"for ("; if(bDeclareI){*ln<<GetTypePrefixInt()<<" ";} *ln<<iStr<<"="<<iMin<<"; "<<iStr<<"<="<<iMax<<"; "<<iStr<<"++) {";
	} else {
		*ln<<"DO "; if(iLabel>=0){*ln<<iLabel<<" ";} *ln<<iStr<<" ="<<iMin<<","<<iMax;
	}
}
void CCodeLanguage::AddOpForP1(CEString* ln, const char* iStr, int iMin, int iMaxP1, int iLabel, bool bDeclareI) {
	if(ln==NULL)return;
	if(m_conventions & CLFileManager::FT_CPP){
		*ln<<"for ("; if(bDeclareI){*ln<<GetTypePrefixInt()<<" ";} *ln<<iStr<<"="<<iMin<<"; "<<iStr<<"<"<<iMaxP1<<"; "<<iStr<<"++) {";
	} else {
		*ln<<"DO "; if(iLabel>=0){*ln<<iLabel<<" ";} *ln<<iStr<<" ="<<iMin<<","<<(iMaxP1-1);
	}
}
void CCodeLanguage::AddOpForP1(CEString* ln, const char* iStr, int iMin, const char* sMaxP1, int iLabel, bool bDeclareI) {
	if(ln==NULL)return;
	if(m_conventions & CLFileManager::FT_CPP){
		*ln<<"for ("; if(bDeclareI){*ln<<GetTypePrefixInt()<<" ";} *ln<<iStr<<"="<<iMin<<"; "<<iStr<<"<"<<sMaxP1<<"; "<<iStr<<"++) {";
	} else {
		*ln<<"DO "; if(iLabel>=0){*ln<<iLabel<<" ";} *ln<<iStr<<" ="<<iMin<<","<<sMaxP1<<"-1";
	}
}
void CCodeLanguage::AddOpEndFor(CEString* ln) {
	if(ln==NULL)return;
	if(m_conventions & CLFileManager::FT_CPP){ ln->AddString("}"); }else{ ln->AddString("CONTINUE"); }
}
void CCodeLanguage::AddFuncCallName(CEString* ln, const char* name) {
	// C++ function call if "FName()" to separate from return variable name
	if((m_conventions & CLFileManager::FT_CPP)==CLFileManager::FT_CPP) {*ln<<"F";} *ln<<name;
}

//protected functions
#define MAX_F77_LINE_LEN	72
int CCodeLanguage::GetCodeLengthMax() { // LINE-length - GAP - indent
	int res;
	if (m_conventions & CLFileManager::FT_F77) { res=MAX_F77_LINE_LEN-6-m_IndentSize; }
	else if(m_conventions & CLFileManager::FT_CPP) { res=0xffff; } //No line-breaking
	else { res=256-m_IndentSize; }
	return res;
}
bool CCodeLanguage::IsLineBreakAllowed(const char* _line, long pos, long len) {
	if ((pos < 0) || (pos >= len)) return false;
	char c = *(_line+pos);
	char c0 = (pos>0)?(*(_line+pos-1)):0;
	if (CEUtil::findSymbol(m_LINEBREAK_SYMS,c) < 0) return false;
	if ((c != ' ') && (c == c0)) return false; // **-like operations
	return true;
}
void CCodeLanguage::StartCodeLine(ofstream* p_ofs, int _label) { // 6-sym gap for F77
	if(p_ofs==NULL)return;
	if (m_conventions & CLFileManager::FT_F77) {
		if (_label >= 0) {
			*p_ofs << " ";
			int s = CLMath::get_digits_num(_label);
			if (s > 4) {
				dbg->log(0,_label,"ERROR: Label %ld is too big."); *p_ofs << "XXXX ";
			} else {
				for(int i=0; i<(4-s); i++) { *p_ofs << " "; }
				*p_ofs << _label << " ";
			}
		} else {
			*p_ofs << "      ";
		}
	}
	if (m_IndentSize>0) {
		if (m_conventions & CLFileManager::FT_CPP) { for(int i=0;i<m_IndentSize;i++) { *p_ofs << "\t"; } }
		else { for(int i=0;i<m_IndentSize;i++) { *p_ofs << " "; } }
	}
}
void CCodeLanguage::StartContinuationLine(ofstream* p_ofs) { // 5-sym gap for F77 + "&"
	if (m_conventions & CLFileManager::FT_F77) {
		*p_ofs << "     &";
	} if (m_conventions & CLFileManager::FT_F90) {
		*p_ofs << "&";
	} else if (m_conventions & CLFileManager::FT_CPP) {
		*p_ofs << "???";	//this case should not occur
	} else {
		*p_ofs << "&";
	}
	if (m_IndentSize>0) {
		if (m_conventions & CLFileManager::FT_CPP) { for(int i=0;i<m_IndentSize;i++) { *p_ofs << "\t"; } }
		else { for(int i=0;i<m_IndentSize;i++) { *p_ofs << " "; } }
	}
}

//////////////////////////////////////////////////// CLFileManager
//
CDebug* CLFileManager::dbg = new CDebug("CLFileManager");
CEPtrArray* CLFileManager::m_files = NULL;
int CLFileManager::open_file(const char* file_name, DWORD conventions, bool bAppend, CEPtrArray* sPreIncludes) {
	if(file_name==NULL)file_name="default.out";

	//TODO: save 'pre-includes' here?

	//Auto-select language
	if ((conventions & (FT_F90|FT_F77|FT_CPP)) == 0) {
		int iDotLast=-1; for(;;){int i=CEUtil::findSymbol(file_name,'.',((iDotLast>=0)?(iDotLast+1):0)); if(i>=0){ iDotLast=i; }else break; }
		int fLang=0;
		if(iDotLast>=0){
			const char* s=file_name+iDotLast+1;
			if(CEUtil::equals(s,"cpp"))fLang=FT_CPP; else if(CEUtil::equals(s,"CPP"))fLang=FT_CPP;
			else if(CEUtil::equals(s,"c"))fLang=FT_CPP; else if(CEUtil::equals(s,"C"))fLang=FT_CPP;
			else if(CEUtil::equals(s,"cc"))fLang=FT_CPP; else if(CEUtil::equals(s,"CC"))fLang=FT_CPP;
			else if(CEUtil::equals(s,"c++"))fLang=FT_CPP; else if(CEUtil::equals(s,"C++"))fLang=FT_CPP;
			// f90
			else if(CEUtil::equals(s,"f90"))fLang=FT_F90; else if(CEUtil::equals(s,"F90"))fLang=FT_F90;
			// f77 elsewhere
		}
		if(fLang==0)fLang=FT_F77; //default
		dbg->log(0,"Code language selected: ",((fLang==FT_F77)?"F77":((fLang==FT_CPP)?"C++":((fLang==FT_F90)?"F90":"Unknown"))));
		conventions |= fLang;
	}

	if (m_files == NULL) m_files = new CEPtrArray(true);	//allow NULLs
	//select next file ID by searching for first 'NULL' entry in 'm_files' (previously deleted file)
	int next_ID=-1;
	for(int i=0;i<m_files->size();i++){ CFFortran* file = (CFFortran*)m_files->get(i); if(file==NULL){ next_ID=i; break; } }
	if(next_ID<0){ next_ID=m_files->size(); m_files->add(NULL); }

	bool create_OK = false;
	// parsing conventions & create header
	if ((conventions & FT_F90) || (conventions & FT_F77)) {
		CFFortran* file = new CFFortran(file_name, conventions, bAppend, sPreIncludes);
		m_files->SetAt(next_ID,file,true); //previous is actually NULL
		create_OK=true;
	} else if (conventions & CLFileManager::FT_CPP) {
		CFFortran* file = new CFFortran(file_name, conventions, bAppend, sPreIncludes);
		m_files->SetAt(next_ID,file,true); //previous is actually NULL
		create_OK=true;
	}
	if(create_OK)dbg->log(20, (bAppend?"File opened: ":"New file created: "), file_name);

	// additional init for PHYSICS CONSTANTS (startup init does not process PHYSICS calls)
	CFFortran::init();

	return (create_OK)?next_ID:(-1);
}

CCodeLanguage* CLFileManager::GetCodeLanguageRef(int file_id) {
	if ((file_id>=0) && (file_id<(m_files->size()))) {
		CFFortran* file = (CFFortran*)m_files->get(file_id); if(file!=NULL)return file->GetCodeLanguage();
	}
	return NULL;
}
const char* CLFileManager::GetFileName(int file_id) {
	if ((file_id>=0) && (file_id<(m_files->size()))) {
		CFFortran* file = (CFFortran*)m_files->get(file_id); if(file!=NULL)return file->GetFileName();
	}
	return NULL;
}
const char* CLFileManager::GetHeaderFileName(int file_id) {
	if ((file_id>=0) && (file_id<(m_files->size()))) {
		CFFortran* file = (CFFortran*)m_files->get(file_id); if(file!=NULL)return file->GetHeaderFileName();
	}
	return NULL;
}
const char* CLFileManager::GetConstIncludeFileName(int file_id) {
	if ((file_id>=0) && (file_id<(m_files->size()))) {
		CFFortran* file = (CFFortran*)m_files->get(file_id); if(file!=NULL)return file->GetConstIncludeFileName();
	}
	return NULL;
}
const char* CLFileManager::RequestParticleMassIncluded(int file_id, long pk) {
	if ((file_id>=0) && (file_id<(m_files->size()))) {
		CFFortran* file = (CFFortran*)m_files->get(file_id); if(file!=NULL)return file->RequestParticleMassIncluded(pk);
	}
	return NULL;
}

void CLFileManager::add_to_file(int file_id,
								const char* func_name,
								CExpression* expr,
								DWORD conventions)
{
	if ((file_id<0) || (file_id >= (m_files->size()))) return;
	CFFortran* file = (CFFortran*)m_files->get(file_id);

	dbg->log(20, file_id, "Add expression to file #%ld:", file);
	if ((file == NULL) || (expr == NULL)) CEException::halt(dbg, "Error function creating: NULL file or expression object.");

	//Copy expression - it may be modified during saving
	CExpression* expr_cpy = expr->clone();
	file->add_function(func_name, expr_cpy, conventions);
	delete expr_cpy;
}
void CLFileManager::close_file(int file_id) { //not required -- releases a bit of memory only
	if ((file_id<0) || (file_id >= (m_files->size()))) return;
	m_files->SetAt(file_id,NULL,true);
}
void CLFileManager::finalize() {
	if (m_files != NULL) { delete m_files; m_files = NULL; }
}
CLFileManager::CLFileManager() {}
CLFileManager::~CLFileManager() {}
//
