////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// GeneralKinematics.cpp: implementation of the CGeneralKinematics class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CGeneralKinematics::CGeneralKinematics()
{

}

CGeneralKinematics::~CGeneralKinematics()
{

}


/*
*	Used to be used for detection of kin-dependent momenta
*	'primary' was independent
*/
bool CGeneralKinematics::isPrimary(CExpression *expr)
{
	return true;
}

/*
*	Compose the kinematically equal construction
*	(Implies the 4-momenta conservation law)
*	Examples (for i1+i2->f1+f2+f3 kinematics):
*	getEqual(i1+i2) ==> f1+f2+f3
*	getEqual(i1) ==> f1+f2+f3-i2
*	getEqual(f3-i2) ==> i1-f1-f2
*/
CExpression* CGeneralKinematics::getEqual(CExpression *expr)
{
	return CKinematics::getEqual(expr);
}

/*
*	why it still virtual?
*	should it be implemented?
*/
void CGeneralKinematics::InitVectors(double energy)
{
	printf("Error kinmatics reinitialization.");
}

/*
*	used for numerical calculations
*	see analogous implementation in Elastic kinematics class
*/
void CGeneralKinematics::fill_data_storage()
{
	// to be implemented
}

/*
*	used for numerical calculations
*	see analogous implementation in Elastic kinematics class
*/
void CGeneralKinematics::set_CMS_energy(double energy)
{
	// to be implemented
}

/*
*	see analogous implementation in Elastic kinematics class
*/
CEReplList* CGeneralKinematics::getReplacements(int type)
{
	dbg->log(15, type, "CGeneralKinematics::getReplacements (%ld)");
	CEReplList* res = CKinematics::getReplacements(type);
	
	/////////////////////////////////////////////////////////////////////////////////////
	// rules for 2->2 process
	if ((num_final_particles() == 2) && (num_init_particles() == 2)) {
		
		//temporary data
		CExpression* half = data->gete(CConst::DV_HALF);
		CExpression* a = NULL;
		CExpression* b = NULL;
		CExpression* temp = NULL;
		
		CExpression* mi0 = pin(0)->get_sym_mass();
		CExpression* mi1 = pin(1)->get_sym_mass();
		CExpression* mf0 = pfin(0)->get_sym_mass();
		CExpression* mf1 = pfin(1)->get_sym_mass();
		
		CExpression* pi0 = pin(0)->get_momentum_symbol();
		CExpression* pi1 = pin(1)->get_momentum_symbol();
		CExpression* pf0 = pfin(0)->get_momentum_symbol();
		CExpression* pf1 = pfin(1)->get_momentum_symbol();

		////////////////////////////////////////////////////////////// REPL_SIMILAR_COUPLINGS
		if (type & CKinematics::REPL_SIMILAR_COUPLINGS) {
			a = CExpression::reduce(new CESProduct(pf0, pf1));
			b = new CESProduct(pi0, pi1);
			temp = new CEPower(mi0, (long)2);
			temp = temp->add(new CEPower(mi1, (long)2));
			temp = temp->minus(new CEPower(mf0, (long)2));
			temp = temp->minus(new CEPower(mf1, (long)2));
			temp = temp->multiply(half->clone());
			b = CExpression::reduce(b->add(temp));
			res->add(a, b, false);
			delete a;
			delete b;

			a = CExpression::reduce(new CESProduct(pf1, pi1));
			b = new CESProduct(pf0, pi0);
			temp = new CEPower(mf1, (long)2);
			temp = temp->add(new CEPower(mi1, (long)2));
			temp = temp->minus(new CEPower(mf0, (long)2));
			temp = temp->minus(new CEPower(mi0, (long)2));
			temp = temp->multiply(half->clone());
			b = CExpression::reduce(b->add(temp));
			res->add(a, b, false);
			delete a;
			delete b;
			
			a = CExpression::reduce(new CESProduct(pf1, pi0));
			b = new CESProduct(pf0, pi1);
			temp = new CEPower(mf1, (long)2);
			temp = temp->add(new CEPower(mi0, (long)2));
			temp = temp->minus(new CEPower(mf0, (long)2));
			temp = temp->minus(new CEPower(mi1, (long)2));
			temp = temp->multiply(half->clone());
			b = CExpression::reduce(b->add(temp));
			res->add(a, b, false);
			delete a;
			delete b;
		}

		////////////////////////////////////////////////////////////// REPL_MANDELSTAM_VARIABLES
		if (type & CKinematics::REPL_MANDELSTAM_VARIABLES) {
			dbg->log(15, "creating MANDELSTAM replacements");
			CExpression* t = new CESymbol("t");
			CExpression* u = new CESymbol("u");
			CExpression* s = new CESymbol("s");
			
			//CExpression* s = CExpression::reduce(
			//	mi0->sqra()->add(mi1->sqra())->add(mf0->sqra())->add(mf1->sqra())
			//	->minus(t->clone())->minus(u->clone()));
			
			CKinematics::add_inv_repl_to_list(pin(0), pin(1), true, true, s, res);
			CKinematics::add_inv_repl_to_list(pfin(0), pfin(1), true, true, s, res);
			CKinematics::add_inv_repl_to_list(pin(0), pfin(0), true, false, t, res);
			CKinematics::add_inv_repl_to_list(pin(1), pfin(1), true, false, t, res);
			CKinematics::add_inv_repl_to_list(pfin(0), pin(1), true, false, u, res);
			CKinematics::add_inv_repl_to_list(pin(0), pfin(1), true, false, u, res);
			
			delete s; delete t; delete u;
			dbg->log(15, "MANDELSTAM replacements created");
		}
		
		////////////////////////////////////////////////////////////// REPL_ALL_MANDELSTAM_X
		if (type & CKinematics::REPL_ALL_MANDELSTAM_X) {
			CExpression* t = new CESymbol("t");
			CExpression* u = new CESymbol("u");
			CExpression* s = new CESymbol("s");

			CExpression* b = CExpression::reduce(
				mi0->sqra()->add(mi1->sqra())->add(mf0->sqra())->add(mf1->sqra())
				->minus(t->clone())->minus(u->clone()));
			res->add(s, b, false);
			delete b;
			
			b = CExpression::reduce(
				mi0->sqra()->add(mi1->sqra())->add(mf0->sqra())->add(mf1->sqra())
				->minus(s->clone())->minus(u->clone()));
			res->add(t, b, false);
			delete b;

			b = CExpression::reduce(
				mi0->sqra()->add(mi1->sqra())->add(mf0->sqra())->add(mf1->sqra())
				->minus(s->clone())->minus(t->clone()));
			res->add(u, b, false);
			delete b;
			
			delete s; delete t; delete u;
		}
	}

	dbg->log(45, type, "CGeneralKinematics: replacements created: ", res);
	return res;
}

/*
*	returns 0 is result is unknown
*	expr is supposed to be "+"-signed.
*/
int CGeneralKinematics::checkExprSign(CExpression* expr) {
	int res = CKinematics::checkExprSign(expr);
	if (res != 0) return res;
	
	if (expr == NULL) return 0;
	if (!expr->getFlag()) {
		CExpression* expr_abs = expr->clone(); expr_abs->setFlag();
		res = -checkExprSign(expr_abs); delete expr_abs;
		return res;
	}

	CExpression* s = new CESymbol("s"); if (expr->equals(s)) res = 1; delete s;
	if (res != 0) return res;
	CExpression* t = new CESymbol("t"); if (expr->equals(t)) res = -1; delete t;
	if (res != 0) return res;
	CExpression* u = new CESymbol("u"); if (expr->equals(u)) res = -1; delete u;
	if (res != 0) return res;

	return 0;
}

/*
*	to be run from the script processor
*	CEPtrArray *args contatins the CScriptValue-s
*
*	returns 1 if init OK
*	returns (-1)*(number of invalid argument) if init failed due to invalid argument
*	returns 0 if init failed
*
*	arguments order:
*	* (int) number of initials
*	* (int) number of finals
*	* for every particle:
*		(int) particle kind
*		(str) title for the momentum
*		// <old> ////////(str or int) title (or 0) for particle's mass
*		(str) title for the polarization vector
*/
int CGeneralKinematics::init_kin(CEPtrArray *args)
{
	if (args == NULL) return 0;
	int args_size = args->size();
	if (args_size <= 2) return 0;

	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if (phys == NULL) {
		dbg->log(0, "Initialization error: no physics have been installed yet.");
		return 0;
	}

	int arg_no = 0;
	CScriptValue* sv = NULL; // temporary

	// (int) number of initials
	sv = (CScriptValue*)args->GetAt(arg_no); arg_no++;
	if (!sv->check_value_type(CScriptValue::VAL_INT)) return (-arg_no);
	int in_num = sv->get_int_value();
	if ((in_num < 2) || (in_num > 2)) return (-arg_no);
	
	// (int) number of finals
	sv = (CScriptValue*)args->GetAt(arg_no); arg_no++;
	if (!sv->check_value_type(CScriptValue::VAL_INT)) return (-arg_no);
	int fin_num = sv->get_int_value();
	if ((fin_num < 2) || (fin_num > 20)) return (-arg_no);
	
	// for every particle
	// process initial & final particles in single cycle
	// i<0 means initial particle, i>=0 means final
	int i;
	for(i=-in_num; i<fin_num; i++) {
		bool from_initial = (i<0);

		// (int) particle kind
		if (arg_no >= args_size) return (-arg_no-1);
		sv = (CScriptValue*)args->GetAt(arg_no); arg_no++;
		if (!sv->check_value_type(CScriptValue::VAL_INT)) return (-arg_no);
		long part_kind = sv->get_int_value();
		//if (part_kind >= 10000) part_kind = (10000-part_kind);
		//if ((part_kind < 0) || (!phys->validate_part_kind(part_kind))) return (-arg_no);
		if (!phys->validate_part_kind(part_kind)) {
			dbg->log(0, part_kind, part_kind, "Error: Unknown particle kind: x%lx, %ld");
			return (-arg_no);
		}

		// (str) title for the momentum
		if (arg_no >= args_size) return (-arg_no-1);
		sv = (CScriptValue*)args->GetAt(arg_no); arg_no++;
		if (!sv->check_value_type(CScriptValue::VAL_STRING)) return (-arg_no);
		CEString* mom_str = sv->get_str_value();
		if (mom_str == NULL) return (-arg_no);
		CExpression* mom_sym = new CESymbol(mom_str->getData());

		// retrieve mass symbol from physics model
		CExpression* mass_sym = phys->get_mass_sym(CLMath::abs(part_kind));
		if (mass_sym != NULL) {
			mass_sym = mass_sym->clone();
		} else {
			// error: no mass symbol found
			dbg->log(0, part_kind, "Init error: no mass symbol found for particle %ld.");
			return 0;
		}
		/* <old>
		// (str or int) title (or 0) for particle's mass
		if (arg_no >= args_size) return (-arg_no-1);
		sv = (CScriptValue*)args->GetAt(arg_no); arg_no++;
		if (!sv->check_value_type(CScriptValue::VAL_INT | CScriptValue::VAL_STRING)) return (-arg_no);
		CExpression* mass_sym = NULL;
		if (sv->check_value_type(CScriptValue::VAL_INT)) {
			int mass_int = sv->get_int_value();
			if (mass_int != 0) return (-arg_no);
			mass_sym = CENumber::GetZero();
		} else { // string
			CEString* mass_str = sv->get_str_value();
			if (mass_str == NULL) return (-arg_no);
			mass_sym = new CESymbol(mass_str->getData());
		}
		*/
		
		// (str) title for the polarization vector
		if (arg_no >= args_size) return (-arg_no-1);
		sv = (CScriptValue*)args->GetAt(arg_no); arg_no++;
		if (!sv->check_value_type(CScriptValue::VAL_STRING)) return (-arg_no);
		CEString* pol_str = sv->get_str_value();
		if (pol_str == NULL) return (-arg_no);
		CExpression* _pol_vector_sym = new CESymbol(pol_str->getData());

		int part_ID = (i<0)?(in_num+1+i):(-1-i); // id < 0 for final particles, id==0 means error.
		CEPtrArray* params = new CEPtrArray(); // empty parameters array
		CEParticle* part = CEParticle::create_particle(part_kind, part_ID, mom_sym, mass_sym, _pol_vector_sym, params);
		delete params;

		if (part == NULL) {
			dbg->log(20, "Error creating particle (possible wrong arguments).");
			return (-arg_no);
		}
		
		if (from_initial) {
			dbg->log(20, "Initial particle created:", part);
			m_init_particles->add(part);
		}
		else {
			dbg->log(20, "Final-state particle created:", part);
			m_final_particles->add(part);
		}
	}

	//ph1->set_helicity(1, f11, f21);
	//ph2->set_helicity(-1, f11, f21);
	//ph1->create_polarization_vectors_symbols();
	//ph1->fill_pol_vector_stubs();
	//ph2->create_polarization_vectors_symbols();
	//ph2->fill_pol_vector_stubs();

	dbg->log(20, "Kinematic initialized.");
	return 1;
}
