////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// ScriptFuncInfo.cpp: implementation of the CScriptFuncInfo class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <stdarg.h>

const char* CScriptFuncInfo::CLASS_NAME = "CScriptFuncInfo";
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CScriptFuncInfo::CScriptFuncInfo(const char* f_title, void* p_func, int args_num, int min_args_number, int* args_type)
: CESerializable()
{
	if (p_func != NULL) {
		//m_pFunction = (CEPtrArray *(__cdecl *)(CEPtrArray *))p_func;
		m_pFunction = (CEPtrArray *(*)(CEPtrArray *))p_func;
	} else {
		m_pFunction = NULL;
	}

	m_args_num = args_num;
	m_args_type = args_type;
	m_args_min_number = min_args_number;

	if (f_title != NULL) {
		m_title = CEUtil::copyString(f_title);
	} else {
		m_title = CEUtil::copyString("???");
	}
}

CScriptFuncInfo::~CScriptFuncInfo()
{
	if (m_args_type != NULL) CMemoryManager::mfree(m_args_type);
	if (m_title != NULL) CMemoryManager::mfree(m_title);
}

/*
*	Note: NULL function pointer means the fuinction is system.
*	CScriptFuncInfo object is only used to check variables type.
*/
CScriptFuncInfo* CScriptFuncInfo::createFuncInfo(const char* f_title, void* p_func, int args_num, int min_args_number, ...) {
	va_list vl;
	va_start(vl, min_args_number);
	if (args_num < 0) return NULL;

	int* args_type = NULL;
	if (args_num > 0) {

		args_type = (int*)CMemoryManager::mmalloc(args_num*sizeof(int));
		int i;
		for(i=0; i<args_num; i++) {
			args_type[i] = va_arg(vl, int);
		}
	}

	va_end(vl);

	if ((min_args_number < 0) || (min_args_number > args_num)) min_args_number = args_num;
	return new CScriptFuncInfo(f_title, p_func, args_num, min_args_number, args_type);
}

CEString* CScriptFuncInfo::toString(int mode) {
	CEString* res = new CEString("<CScriptFuncInfo: ");
	res->AddString(m_title);
	if (isSystemFunction()) res->AddString(" {SYSTEM FUNCTION}");
	res->AddString(">");
	return res;
}

const char* CScriptFuncInfo::ClassName() const { return CLASS_NAME; }
CEObject* CScriptFuncInfo::copy() const { return clone(); }
CScriptFuncInfo* CScriptFuncInfo::clone() const {

	int* args_type_copy = NULL;
	if (m_args_num > 0) {
		args_type_copy = (int*)CMemoryManager::mmalloc(m_args_num*sizeof(int));
		int i;
		for(i=0; i<m_args_num; i++) { args_type_copy[i] = m_args_type[i]; }
	}
	return new CScriptFuncInfo(m_title, (void*)m_pFunction, m_args_num, m_args_min_number, args_type_copy);
}

/*
*	positive arguments means number of invalid argument
*	negative means invalid total number of arguments
*	0 means OK
*/
int CScriptFuncInfo::check_args(CEPtrArray* args) {
	if ((m_args_type == NULL) && (m_args_num > 0)) {
		// internal error
		return INT_MIN;
	}

	long s = args->size();
	long i;
	for(i=0; i<s; i++) {
		if (i >= m_args_num) {
			// Extra agrument: OK
			// Note! m_args_num is the number of arguments to check (size of m_args_type array)
			// Function may take more parameters and check their type itself.
			break;
		}

		int agr_type = m_args_type[i];
		CScriptValue* val = (CScriptValue*)args->GetAt(i);

		if (!val->check_value(agr_type, true)) return i+1;
	}

	if (s < m_args_min_number) return -s-1;
	return 0;
}

bool CScriptFuncInfo::isSystemFunction() { return (m_pFunction == NULL); }
const char* CScriptFuncInfo::GetFuncTitle() { return m_title; }

CEPtrArray* CScriptFuncInfo::invoke_function(CEPtrArray* args, bool do_check_args) {
	if (isSystemFunction()) { CScriptProcessor::dbg->log(0, "Error: {CScriptFuncInfo} invalid system function invokation", this); return FReturnError(args, 0); }

	if (do_check_args) {
		int inv_arg_no = check_args(args);
		if (inv_arg_no != 0) return FReturnError(args, inv_arg_no);
	}
	CEPtrArray* res_array = (*m_pFunction)(args);
	return res_array;
}

const unsigned CScriptFuncInfo::ERROR_HALT_FLAG	= 0x1000;
/*
*	Clears data array & returns error data
*	code is:
*	positive arguments means number of invalid argument
*	negative means invalid total number of arguments
*	0 means unexpected error
*	0x10XX mean no error halt (XX is the halt code) (ERROR_HALT_FLAG | halt_code)
*/
CEPtrArray* CScriptFuncInfo::FReturnError(CEPtrArray* arg_list, int code) {
	delete arg_list;
	CEPtrArray* res = new CEPtrArray();
	res->add(CScriptValue::newUndefinedValue());
	res->add(CScriptValue::newValue(code));
	return res;
}

CEPtrArray* CScriptFuncInfo::FReturnInvalidParam(int parNo, CEPtrArray* arg_list) {
	if(arg_list!=NULL) delete arg_list;
	CEPtrArray* res = new CEPtrArray();
	res->add(CScriptValue::newUndefinedValue());
	res->add(CScriptValue::newValue(parNo));
	return res;
}
