////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// SriptFunctionLib.cpp: implementation of the CScriptFunctionLib class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CScriptFunctionLib::dbg = new CDebug("CScriptFunctionLib");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CScriptFunctionLib::CScriptFunctionLib(){}
CScriptFunctionLib::~CScriptFunctionLib(){}

//////////////////////////////////////////////////////////////////////
CExpression* CScriptFunctionLib::GetExpressionCpy(CScriptValue* val) { //val: str, int or expr
	if(val==NULL)return NULL;
	CExpression* res=NULL;
	if (val->check_value(CScriptValue::VAL_STRING, false)) { res=new CESymbol((val->get_str_value())->getData()); }
	else if (val->check_value(CScriptValue::VAL_INT, false)) { res=CENumber::GetNumber(val->get_int_value()); }
	else if (val->check_value(CScriptValue::VAL_EXPRESSION, false)) { res=val->get_expr_value(); if(res!=NULL)res=res->clone(); }
	return res;
}

//////////////////////////////////////////////////////////////////////
CEPtrArray* CScriptFunctionLib::SetKinematics(CEPtrArray *arg_list)
{
	if (arg_list == NULL) {dbg->log(0, "NULL argument list"); return NULL;}
	CEPtrArray* res = new CEPtrArray();

	CGeneralKinematics* kin = new CGeneralKinematics();
	int err_code = kin->init_kin(arg_list);
	if (err_code > 0) { //init OK
		CEProcessor::setKinematics(kin,true);
	} else {
		if (err_code < 0) dbg->log(0, -err_code, "SetKinematics: invalid argument (#%ld).");
		else dbg->log(0, "SetKinematics: error creating.");
		res->add(CScriptValue::newUndefinedValue());
		res->add(CScriptValue::newValue(-err_code));
	}

	delete arg_list;
	return res;
}

CEPtrArray* CScriptFunctionLib::SFuncExpand(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	if ((arg_num<1) || (arg_num>2)) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);

	CScriptValue* val = (CScriptValue*)arg_list->GetAt(0);
	if (!val->check_value(CScriptValue::VAL_EXPRESSION, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CExpression* expr_source = val->get_expr_value();

	int _flags=0;
	if (arg_num > 1) { _flags = ((CScriptValue*)arg_list->GetAt(1))->get_int_value(); }

	CExpression* expr_res=NULL;
	if ((_flags & CLSimplifyMisc::MISC_EXPAND_SAVE_MEMORY)==CLSimplifyMisc::MISC_EXPAND_SAVE_MEMORY) {
		expr_res = CLSimplifyMisc::ExpandSaveMemory(expr_source->clone());
	} else {
		expr_res = CExpression::reduce(expr_source->clone());
	}

	if (expr_res == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);
	arg_list->SetAt(0, CScriptValue::newValue(expr_res, true), true);
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncArrange(CEPtrArray* arg_list)
{
	int arg_num = arg_list->size();
	if (arg_num != 1) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);

	CScriptValue* val = (CScriptValue*)arg_list->GetAt(0);
	if (!val->check_value(CScriptValue::VAL_EXPRESSION, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CExpression* expr_source = val->get_expr_value();

	CExpression* expr_res = CExpression::arrge(expr_source->clone());
	if (expr_res == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);
	
	arg_list->SetAt(0, CScriptValue::newValue(expr_res, true), true);
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncLoadModel(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnInvalidParam(1,arg_list);
	CEString* str0 = val0->get_str_value();
	const char* _ModFileName=((str0!=NULL)?str0->getData():NULL);
	if (_ModFileName==NULL) return CScriptFuncInfo::FReturnInvalidParam(1,arg_list);

	const char* _GenFileName=NULL;
	if (arg_num > 1) {
		CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1);
		if (!val1->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnInvalidParam(2,arg_list);
		CEString* str1 = val1->get_str_value(); if(str1!=NULL)_GenFileName=str1->getData();
	}
	//if (_GenFileName==NULL) { _GenFileName="Lorentz.gen"; dbg->logIns(4, "No *.GEN file is specified. Presumed: %s", _GenFileName); }
	bool installedOK = CSModelLoader::InstallModelFile(_GenFileName, _ModFileName);
	if (installedOK) {
		// log out correspondence rules
		// note: correspondence rules logging requires physics to be already installed
		CDiagramPhysics* curr_phys = CDiagramPhysics::get_curr_physics();
		if (curr_phys!=NULL) {
			CDiagramPhysics::dbg->log(7, "New hamiltonian expression", curr_phys->getHamiltonianExpr());
			CDiagramPhysics::dbg->log(7, "New correspondence rules", curr_phys->getCorrespondenceRules());
		} else {
			dbg->log(0, "INTERNAL ERROR: NULL physics after LoadModel() processing");
			return CScriptFuncInfo::FReturnError(arg_list, 0);
		}
	} else {
		dbg->log(0, "Error reading model file");
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncGetSoftBremsstrahlung(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();

	CExpression* csBorn=NULL; bool del_csBorn=false;
	if (arg_num > 0) {
		CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
		if (val0->check_value(CScriptValue::VAL_EXPRESSION, true)) { csBorn=val0->get_expr_value();
		} else if (val0->check_value(CScriptValue::VAL_INT, true)) { csBorn=new CENumber(val0->get_int_value()); del_csBorn=true; }
		else return CScriptFuncInfo::FReturnError(arg_list, 1);
	}

	CExpression* res = CLHEPBasics::GetSoftBremsstrahlungFactor();
	if((res != NULL) && (csBorn!=NULL)){ res=res->multiply(csBorn->clone()); }
	if(del_csBorn){ delete csBorn; csBorn=NULL; }

	if(res==NULL)return CScriptFuncInfo::FReturnError(arg_list, 0);
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

/*
*	1: (str) workspace title
*	2: (int) 'RUN_XXX' flags
*/
CEPtrArray* CScriptFunctionLib::SFuncRun(CEPtrArray* arg_list) {
	int arg_num = arg_list->size(); //pre-checked: 0(str)
	CEString* str0 = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	const char* sTitle = ((str0!=NULL)?(str0->getData()):NULL);
	//check project title [Nt: the check will be repeated in 'CLHEPRunProject::Run()']
	if(!CEUtil::IsAllowedFileName(sTitle)){ dbg->log(0,"Run: invalid project name"); return CScriptFuncInfo::FReturnInvalidParam(1,arg_list); }

	long iRunFlags=0;
	if(arg_num>=2){
		iRunFlags = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		//check flags here [Nt: the check will be repeated in 'CLHEPRunProject::Run()']
	}

	const char* sTexFile=NULL;
	if(arg_num>=3){
		CEString* str3 = ((CScriptValue*)arg_list->GetAt(2))->get_str_value();
		if(str3!=NULL){ sTexFile=str3->getData(); }
	}

	bool bRes=CLHEPRunProject::Run(sTitle,iRunFlags,sTexFile);
	if(!bRes){ dbg->log(0,"Run: failed to create project: ",sTitle); return CScriptFuncInfo::FReturnInvalidParam(0,arg_list); }
	arg_list->clear_all(); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncLoadExprFromFile(CEPtrArray* arg_list)
{
	int arg_num = arg_list->size();
	if (arg_num != 1) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);

	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CEString* str0 = val0->get_str_value();
	
	CESerializable* read_val = CESupport::read_file(str0->getData(), 0);
	if (read_val == NULL) {
		dbg->log(0, "Error reading from file");
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	CScriptValue* res_val = NULL;
	if (CEBasic::isExpression(read_val)) {
		res_val = CScriptValue::newValue((CExpression*)read_val, true);
	} else if (CEUtil::equals(read_val->ClassName(), CMatrixElement::CLASS_NAME)) {
		res_val = CScriptValue::newObjectValue(CScriptValue::VAL_MATRIX_ELEMENT, read_val, true);
	} else if (CEUtil::equals(read_val->ClassName(), CDiagramSet::CLASS_NAME)) {
		res_val = CScriptValue::newValue((CDiagramSet*)read_val, true);
	}

	arg_list->SetAt(0, res_val, true);
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncStoreValueToFile(CEPtrArray* arg_list)
{
	int arg_num = arg_list->size();
	if (arg_num != 2) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);

	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CEString* str0 = val0->get_str_value();

	CESerializable* data_val = NULL; bool del_data_val=false;
	DWORD serial_settings = CESupport::SERIAL_VERIFY;
	//CESupport::SERIAL_EXPRESSION | CESupport::SERIAL_VERIFY
	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1);
	if (val1->check_value(CScriptValue::VAL_EXPRESSION, true)) {
		data_val = val1->get_expr_value();
		serial_settings |= CESupport::SERIAL_EXPRESSION;
	} else if (val1->check_value(CScriptValue::VAL_MATRIX_ELEMENT, true)) {
		dbg->log(0, "ERROR: SaveXML() operation is NOT IMPLEMENTED for MATRIX ELEMENT type");
		arg_list->RemoveAt(0, true);
		arg_list->RemoveAt(0, true);
		return arg_list;
		//data_val = (CESerializable*)val1->get_obj_value();
	} else if (val1->check_value(CScriptValue::VAL_DIAGRAM_SET, true)) {
		data_val = (CESerializable*)val1->get_obj_value();
	} else if (val1->check_value(CScriptValue::VAL_INT, true)) {
		data_val = new CENumber(val1->get_int_value()); del_data_val=true;
	} else return CScriptFuncInfo::FReturnError(arg_list, 2);

	CESupport::store(data_val, str0->getData(), serial_settings);
	if (del_data_val) delete data_val;
	arg_list->RemoveAt(0, true);
	arg_list->RemoveAt(0, true);
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetParameter(CEPtrArray* arg_list) {
	CEString* str0 = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	int int1 = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();

	int res = CEProcessor::setGlobalParameter(str0->getData(), int1);

	if (res < 0) {
		if (res == -1) return CScriptFuncInfo::FReturnError(arg_list, 1);
		if (res == -2) return CScriptFuncInfo::FReturnError(arg_list, 2);
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetParameterS(CEPtrArray* arg_list) {
	CEString* str0 = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	CEString* str1 = ((CScriptValue*)arg_list->GetAt(1))->get_str_value();

	int res = CEProcessor::setGlobalParameter(str0->getData(), str1->getData());

	if (res < 0) {
		if (res == -1) return CScriptFuncInfo::FReturnError(arg_list, 1);
		if (res == -2) return CScriptFuncInfo::FReturnError(arg_list, 2);
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetModelParameter(CEPtrArray* arg_list) { // used by LoadModel() function
	CEString* str0 = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();

	CExpression* data_val=NULL; bool del_data_val=false;
	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1);
	if (val1->check_value(CScriptValue::VAL_EXPRESSION, true)) {
		data_val = val1->get_expr_value();
	} else if (val1->check_value(CScriptValue::VAL_INT, true)) {
		data_val = new CENumber(val1->get_int_value()); del_data_val=true;
	} else return CScriptFuncInfo::FReturnError(arg_list, 2);
	if (data_val==NULL) return CScriptFuncInfo::FReturnError(arg_list, 2);

	CSModelLoader::setModelParameter(str0->getData(), data_val);
	if (del_data_val) delete data_val;
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncStoreValueToNB(CEPtrArray* arg_list)
{
	int arg_num = arg_list->size();
	if ((arg_num < 2) || (arg_num > 4)) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);

	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CEString* str0 = val0->get_str_value();

	CESerializable* data_val=NULL; bool del_data_val=false;
	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1);
	if (val1->check_value(CScriptValue::VAL_EXPRESSION, true)) {
		data_val = val1->get_expr_value();
	} else if (val1->check_value(CScriptValue::VAL_MATRIX_ELEMENT, true)) {
		data_val = (CESerializable*)val1->get_obj_value();
	} else if (val1->check_value(CScriptValue::VAL_INT, true)) {
		data_val = new CENumber(val1->get_int_value()); del_data_val=true;
	} else return CScriptFuncInfo::FReturnError(arg_list, 2);

	int next_arg = 2;
	CEString* title_str = NULL;
	if (arg_num > next_arg) {
		CScriptValue* title_val = (CScriptValue*)arg_list->GetAt(next_arg);
		if (title_val->check_value(CScriptValue::VAL_STRING, true)) {
			next_arg++; // skip argument
			title_str = title_val->get_str_value();
		}
	}

	int _flags = 0; // append by default
	if (arg_num > next_arg) {
		CScriptValue* flags_val = (CScriptValue*)arg_list->GetAt(next_arg);
		if (!flags_val->check_value(CScriptValue::VAL_INT, true)) return CScriptFuncInfo::FReturnError(arg_list, next_arg+1);
		_flags = flags_val->get_int_value();
	}

	bool ok1 = true;
	if (title_str != NULL) {
		int title_flags = _flags;
		if (title_flags & CESupport::FILE_CLOSE) title_flags ^= CESupport::FILE_CLOSE;
		ok1 = CESupport::SaveTextToNB(str0->getData(), title_str->getData(), title_flags);
		if (_flags & CESupport::FILE_TRUNCATE) _flags ^= CESupport::FILE_TRUNCATE;
	}
	bool ok2 = CESupport::SaveExprToNB(str0->getData(), data_val, _flags);
	if (del_data_val) delete data_val;
	bool ok = ((ok1) && (ok2));
	if (!ok) return CScriptFuncInfo::FReturnError(arg_list, 0);
	
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncStoreValueToM(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();

	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CEString* str0 = val0->get_str_value();

	CESerializable* data_val=NULL; bool del_data_val=false;
	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1);
	if (val1->check_value(CScriptValue::VAL_EXPRESSION, true)) {
		data_val = val1->get_expr_value();
	} else if (val1->check_value(CScriptValue::VAL_MATRIX_ELEMENT, true)) {
		data_val = (CESerializable*)val1->get_obj_value();
	} else if (val1->check_value(CScriptValue::VAL_INT, true)) {
		data_val = new CENumber(val1->get_int_value()); del_data_val=true;
	} else return CScriptFuncInfo::FReturnError(arg_list, 2);

	bool ok = CESupport::SaveExprToM(str0->getData(), data_val);
	if (del_data_val) delete data_val;
	if (!ok) return CScriptFuncInfo::FReturnError(arg_list, 0);
	arg_list->clear_all(); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncStoreTextToNB(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	//printf("***%ld***",arg_num);
	if ((arg_num < 2) || (arg_num > 3)) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);

	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CEString* str0 = val0->get_str_value();
	//printf("***X2***");

	int next_arg = 1;
	CEString* title_str = NULL;
	if (arg_num > next_arg) {
		CScriptValue* title_val = (CScriptValue*)arg_list->GetAt(next_arg);
		if (title_val->check_value(CScriptValue::VAL_STRING, true)) {
			next_arg++; // skip argument
			title_str = title_val->get_str_value();
		}
	}

	int _flags = 0; //CESupport::FILE_TRUNCATE | CESupport::FILE_CLOSE;
	if (arg_num > next_arg) {
		CScriptValue* flags_val = (CScriptValue*)arg_list->GetAt(next_arg);
		if (!flags_val->check_value(CScriptValue::VAL_INT, true)) return CScriptFuncInfo::FReturnError(arg_list, next_arg+1);
		_flags = flags_val->get_int_value();
	}

	//printf("***X3***");
	bool ok = false;
	if (title_str != NULL) {
		ok = CESupport::SaveTextToNB(str0->getData(), title_str->getData(), _flags);
	} else {
		ok = CESupport::SaveHeadersToNB(str0->getData(), _flags);
	}

	if (!ok) return CScriptFuncInfo::FReturnError(arg_list, 0);
	//printf("***X4***");
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncViewFeynmanRules(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();

	CEString* fn = NULL;
	if (arg_num > 0) fn = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	if (fn != NULL) { if (CEUtil::equals(fn->getData(), CDebug::GetDebugFileName())) fn=NULL; }

	int _flags = 0; // append by default
	if (arg_num > 1) {
		_flags = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		if ((_flags & (CESupport::FILE_CLOSE|CESupport::FILE_TRUNCATE)) != _flags) {
			dbg->log(0, _flags, CESupport::FILE_CLOSE|CESupport::FILE_TRUNCATE, "Script WARNING: (ViewFeynmanRules) invalid flag (%ld, alowed: %ld), parameter ignored");
			_flags=0;
		}
		if ((fn == NULL) && (_flags != 0)) {
			dbg->log(0, _flags, "Script WARNING: (ViewFeynmanRules) invalid flag (%ld, 0 alowed for debug.nb file output), parameter ignored");
			_flags=0;
		}
	}

	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if (phys != NULL) {
		CDiagCorrRules* rules = phys->getCorrespondenceRules();
		if (fn != NULL) {
			bool save_ok = CESupport::SaveExprToNB(fn->getData(), rules, _flags);
			if (!save_ok) dbg->log(0, "(ViewFeynmanRules)Error saving NB file");
		} else {
			int l_old = dbg->set_level(15); dbg->log(7, "ViewFeynmanRules()", rules); dbg->set_level(l_old);
		}
	} else {
		dbg->log(0, "ViewFeynmanRules() Error: No physics installed");
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncViewTensorVITable(CEPtrArray* arg_list) {
	// Same to DumpDataStorage now, dumping all the objects in DS
	int arg_num = arg_list->size();
	CEString* fn = NULL;
	if (arg_num > 0) fn = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	if (fn != NULL) { if (CEUtil::equals(fn->getData(), CDebug::GetDebugFileName())) fn=NULL; }

	int _flags = 0; // append by default
	if (arg_num > 1) {
		_flags = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		if ((_flags & (CESupport::FILE_CLOSE|CESupport::FILE_TRUNCATE)) != _flags) {
			dbg->log(0, _flags, CESupport::FILE_CLOSE|CESupport::FILE_TRUNCATE, "Script WARNING: (ViewTensorVITable) invalid flag (%ld, alowed: %ld), parameter ignored");
			_flags=0;
		}
		if ((fn == NULL) && (_flags != 0)) {
			dbg->log(0, _flags, "Script WARNING: (ViewTensorVITable) invalid flag (%ld, 0 alowed for debug.nb file output), parameter ignored");
			_flags=0;
		}
	}

	CEData* data = CEProcessor::getDataStorage();
	if (data != NULL) {
		if (fn != NULL) {
			bool save_ok = CESupport::SaveExprToNB(fn->getData(), data, _flags);
			if (!save_ok) dbg->log(0, "(ViewTensorVITable) Error saving NB file");
		} else {
			int l_old = dbg->set_level(15); dbg->log(7, "Tensor VI cache (some other data included):", data); dbg->set_level(l_old);
		}
	} else {
		dbg->log(0, "ViewTensorVITable() Error: No data storage are available");
	}

	arg_list->clear_all();
	return arg_list;

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncViewScalarVICache(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CEString* fn = NULL;
	if (arg_num > 0) fn = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	if (fn != NULL) { if (CEUtil::equals(fn->getData(), CDebug::GetDebugFileName())) fn=NULL; }

	int _flags = 0; // append by default
	if (arg_num > 1) {
		_flags = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		if ((_flags & (CESupport::FILE_CLOSE|CESupport::FILE_TRUNCATE)) != _flags) { dbg->log(0, _flags, CESupport::FILE_CLOSE|CESupport::FILE_TRUNCATE, "Script WARNING: (ViewScalarVICache) invalid flag (%ld, alowed: %ld), parameter ignored"); _flags=0; }
		if ((fn == NULL) && (_flags != 0)) { dbg->log(0, _flags, "Script WARNING: (ViewScalarVICache) invalid flag (%ld, 0 alowed for debug.nb file output), parameter ignored"); _flags=0; }
	}

	CECache* vi_cache = CVIScalarCalcLib::GetScalarVICache();
	if (vi_cache != NULL) {
		if (fn != NULL) {
			bool save_ok = CESupport::SaveExprToNB(fn->getData(), vi_cache, _flags);
			if (!save_ok) dbg->log(0, "(ViewScalarVICache) Error saving NB file");
		} else {
			int l_old = dbg->set_level(15); dbg->log(7, "Scalar VI cache:", vi_cache); dbg->set_level(l_old);
		}
	} else {
		dbg->log(0, "ViewScalarVICache() Error: No scalar integrals cache are available");
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncViewConstants(CEPtrArray* arg_list) {

	CDiagramPhysics* phys=CDiagramPhysics::get_curr_physics();
	CEPtrArray* vals = ((phys!=NULL)?phys->CollectConstantValues():NULL);
	if (vals != NULL) {
		int l_old=dbg->set_level(15); dbg->log(7,"Model constants list:",vals); dbg->set_level(l_old);
		delete vals; vals=NULL;
	} else { dbg->log(0, "ViewConstants() Error: No constants list available"); }

	arg_list->clear_all();
	return arg_list;
}

/*
*	Note: this is deprecated function, referred as 'ComposeDiagramsB' in script language
*/
CEPtrArray* CScriptFunctionLib::SFuncComposeDiagrams(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	if ((arg_num != 1) && (arg_num != 2)) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);

	int int_val = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	int _alg_setts = 0; // library default settings will be used
	if (arg_num > 1) {
		_alg_setts = (DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		// possible check settings value
	}

	CDiagramSet* diags = CFeynDiagramFactory::ComposeAndFilterDiagrams(int_val, _alg_setts);

	if (diags == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);
	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(diags, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncDrawDiagrams(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	if ((arg_num < 2) || (arg_num > 4)) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);
	
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_DIAGRAM_SET, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CDiagramSet* diags = (CDiagramSet*)val0->get_obj_value();

	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1);
	if (!val1->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnError(arg_list, 2);
	CEString* fn = val1->get_str_value();

	 // defaults
	int draw_flags = CFeynDiagram::DRAW_CAPS_SMALL|CFeynDiagram::SET_LOOP_120|CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES|CFeynDiagram::DRAW_DIAGRAM_NUMBERS;
	if (arg_num > 2) draw_flags = (DWORD)((CScriptValue*)arg_list->GetAt(2))->get_int_value();

	int _setts = 0; //CESupport::FILE_TRUNCATE;
	if (arg_num > 3) {
		_setts = (DWORD)((CScriptValue*)arg_list->GetAt(3))->get_int_value();
		if ((_setts != CESupport::FILE_TRUNCATE) && (_setts != 0)) return CScriptFuncInfo::FReturnError(arg_list, 3);
	}

	CKinematics* kin = CEProcessor::kinematics();
	if (kin != NULL) { if (kin->num_final_particles()>2) draw_flags |= CFeynDiagram::DRAW_SWAP_TALES; }
	if ((draw_flags & CFeynDiagram::DRAW_DONT_SWAP_TALES) == CFeynDiagram::DRAW_DONT_SWAP_TALES) {
		if ((draw_flags & CFeynDiagram::DRAW_SWAP_TALES) == CFeynDiagram::DRAW_SWAP_TALES) { draw_flags ^= CFeynDiagram::DRAW_SWAP_TALES; }
	}

	int old_draw_flags = diags->GetDrawingFlags();
	diags->SetDrawingFlags(draw_flags);

	bool locate_OK = diags->LocateDiagramsView(CLDiagLocateView::ALGORITHM_DEFAULT);
	if (!locate_OK) { dbg->log(0, "ERROR: locate_OK = FALSE\n"); }
	CESupport::saveTeX(fn->getData(), diags, _setts);

	diags->SetDrawingFlags(old_draw_flags); // restore old flags (system defaults)
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncStoreTextToTeX(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	if ((arg_num < 2) || (arg_num > 3)) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);
	
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_STRING, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CEString* fn_str = val0->get_str_value();

	int next_arg = 1;
	CEString* text_str = NULL;
	if (arg_num > next_arg) {
		CScriptValue* title_val = (CScriptValue*)arg_list->GetAt(next_arg);
		if (title_val->check_value(CScriptValue::VAL_STRING, true)) {
			next_arg++; // skip argument
			text_str = title_val->get_str_value();
		}
	}

	int _flags = 0; //CESupport::FILE_TRUNCATE | CESupport::FILE_CLOSE;
	if (arg_num > next_arg) {
		CScriptValue* flags_val = (CScriptValue*)arg_list->GetAt(next_arg);
		if (!flags_val->check_value(CScriptValue::VAL_INT, true)) return CScriptFuncInfo::FReturnError(arg_list, next_arg+1);
		_flags = flags_val->get_int_value();
	}

	bool ok = false;
	if (text_str != NULL) {
		ok = CESupport::saveTeX(fn_str->getData(), text_str->getData(), _flags);
	} else {
		ok = CESupport::saveTeX(fn_str->getData(), "", _flags);
	}

	if (!ok) return CScriptFuncInfo::FReturnError(arg_list, 0);
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncRetrieveME(CEPtrArray* arg_list)
{
	int arg_num = arg_list->size();
	if (arg_num != 1) return CScriptFuncInfo::FReturnError(arg_list, arg_num);
	
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	if (!val0->check_value(CScriptValue::VAL_DIAGRAM_SET, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CDiagramSet* diags = (CDiagramSet*)val0->get_obj_value();

	CMatrixElement* me = diags->RetrieveMatrixElement();
	if (me == NULL) {
		printf("RetrieveMatrixElement(): NULL\n");
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}
	
	arg_list->SetAt(0, CScriptValue::newObjectValue(CScriptValue::VAL_MATRIX_ELEMENT, me, true), true);
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncComposeDiagsWick(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	if (arg_num != 1) return CScriptFuncInfo::FReturnError(arg_list, arg_num);
	int order = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();

	CMEComposeLib* lib = new CMEComposeLib();
	bool run_OK = lib->ComposeME(order);
	if (!run_OK) {
		dbg->log(0, "Error creating diagrams"); delete lib; return CScriptFuncInfo::FReturnError(arg_list, 0);
	}
	
	CDiagramSet* diags = (lib->GetDiagramsCreated())->clone();
	delete lib;

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(diags, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncCreateME(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	if (arg_num != 1) return CScriptFuncInfo::FReturnError(arg_list, arg_num);
	int order = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	
	// create ME & diagrams from Wick operators decomposition
	CMEComposeLib* lib = new CMEComposeLib();
	bool run_OK = lib->ComposeME(order);
	if (!run_OK) {
		dbg->log(0, "Error ME creating");
		delete lib;
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	CMatrixElement* me = (lib->GetCreatedME())->clone();
	CDiagramSet* diags = (lib->GetDiagramsCreated())->clone();
	delete lib;

	arg_list->clear_all();
	arg_list->add(CScriptValue::newObjectValue(CScriptValue::VAL_MATRIX_ELEMENT, me, true));
	arg_list->add(CScriptValue::newUndefinedValue()); // no errors declared
	arg_list->add(CScriptValue::newValue(diags, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSquareME(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();

	CMatrixElement* me_a = (CMatrixElement*)((CScriptValue*)arg_list->GetAt(0))->get_obj_value();
	CMatrixElement* me_b = NULL;

	DWORD sqr_default = CMESquareLib::SQRME_EXPAND_AND_SPLIT |
		CMESquareLib::SQRME_PRESIMPLIFY_ME |
		CMESquareLib::SQRME_CACHE_DENOMINATORS;
	DWORD sqr_setts = sqr_default;

	if (arg_num > 1) {
		CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1);
		if (val1->check_value(CScriptValue::VAL_MATRIX_ELEMENT, false)) {
			// 2-nd argument may be valid NULL object (means squaring of me#1)
			me_b = (CMatrixElement*)val1->get_obj_value();

			if (arg_num > 2) sqr_setts = (DWORD)((CScriptValue*)arg_list->GetAt(2))->get_int_value();
		}
		else if (val1->check_value(CScriptValue::VAL_INT, false)) {
			sqr_setts = (DWORD)val1->get_int_value();

			sqr_setts |= sqr_default; //|SQR_DEFAULT //TODO: allow to switch off defaults
		}
		else return CScriptFuncInfo::FReturnError(arg_list, 2);
	}

	if (CEProcessor::m_MandelstamsEverywhere) { sqr_setts |= CMESquareLib::SQRME_ALLOW_MANDELSTAMS; }
	
	// couple me_a to me_b or to itself
	CExpression* res = CMESquareLib::SquareME(me_a, me_b, sqr_setts);

	if (CEProcessor::m_bAutoKinArrange && (CEProcessor::kinematics()!=NULL)) {
		DWORD KA_setts = 0;
		if (CEProcessor::m_MandelstamsEverywhere) { KA_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
		res = CLHEPBasics::KinArrange(res, KA_setts);
		if (res == NULL) { dbg->log(0, "ERROR: Auto-KinArrange() failed."); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newObjectValue(CScriptValue::VAL_EXPRESSION, res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncCalcAmplitude(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();

	CMatrixElement* me = (CMatrixElement*)((CScriptValue*)arg_list->GetAt(0))->get_obj_value();
	DWORD setts = 0;
	if (arg_num > 1) { setts = (DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value(); }

	if (CEProcessor::m_MandelstamsEverywhere) { setts |= CAmplCalcLib::A_MANDELSTAMS; }

	// calc amplitude
	CExpression* res = CAmplCalcLib::CalcAmplitude(me, setts);

	if (CEProcessor::m_bAutoKinArrange && (CEProcessor::kinematics()!=NULL)) {
		DWORD KA_setts = 0;
		if (CEProcessor::m_MandelstamsEverywhere) { KA_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
		res = CLHEPBasics::KinArrange(res, KA_setts);
		if (res == NULL) { dbg->log(0, "ERROR: Auto-KinArrange() failed."); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newObjectValue(CScriptValue::VAL_EXPRESSION, res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncToDiffCS(CEPtrArray* arg_list) {
	CExpression* me2 = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* res = CLHEPBasics::GetDiffCS(me2->clone(),true,true);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newObjectValue(CScriptValue::VAL_EXPRESSION, res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetMassCalcOrder(CEPtrArray* arg_list) {
	int part_kind = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	int mass_order = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();

	if (part_kind < 0) part_kind = -part_kind; // truncate sign
	//if (part_kind >= 10000) part_kind -= 10000;
	if (mass_order >= 100) mass_order = -1;
	CEProcessor::getEnvironment()->SetMassCalcOrder(part_kind, mass_order);

	arg_list->clear_all();
	return arg_list;
}

const int CScriptFunctionLib::SelectDiagrams_LOOPS	= -1;
const int CScriptFunctionLib::SelectDiagrams_CT		= -2;
CEPtrArray* CScriptFunctionLib::SFuncSelectDiagrams(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CDiagramSet* diags = (CDiagramSet*)((CScriptValue*)arg_list->GetAt(0))->get_obj_value();
	int diags_num = diags->get_diagrams_number();
	CDiagramSet* diags_res = diags->getEmptyFrame();

	// Parameters: Number, String("1..4"), SpecialCommand
	// check for special case
	bool bSpecValue=false; int iSpecVal=0;
	if (arg_num>=2) {
		CScriptValue* val_1 = (CScriptValue*)arg_list->GetAt(1);
		if (val_1->check_value(CScriptValue::VAL_INT, true)) { iSpecVal=val_1->get_int_value(); if(iSpecVal<0)bSpecValue=true; }
	}

	if (bSpecValue) {
		if ((iSpecVal!=CScriptFunctionLib::SelectDiagrams_LOOPS) && (iSpecVal!=CScriptFunctionLib::SelectDiagrams_CT)) return CScriptFuncInfo::FReturnError(arg_list, 2);
		if (arg_num>2) { dbg->log(0,"SelectDiagrams(): Too many parameters for special call. Extra parameters are ignored."); }
		long i;
		for(i=0; i<diags_num; i++) {
			CFeynDiagram* d = diags->get_diagram(i); if(d==NULL)continue;
			if (iSpecVal==CScriptFunctionLib::SelectDiagrams_LOOPS) {
				if (d->GetLoopsNumber()==1) diags_res->add_diagram(d->clone());
			} else if (iSpecVal==CScriptFunctionLib::SelectDiagrams_CT) {
				if (d->GetCTNumber()>0) diags_res->add_diagram(d->clone());
			}
		}
	} else { // iSpecVal may be undefined here
		long i;
		for(i=1; i<arg_num; i++) {
			CScriptValue* val_i = (CScriptValue*)arg_list->GetAt(i);
			if (val_i->check_value(CScriptValue::VAL_INT, true)) {
				int diag_no = val_i->get_int_value();
				if ((diag_no < 0) || (diag_no >= diags_num)) return CScriptFuncInfo::FReturnError(arg_list, 1+i);
				CFeynDiagram* d = diags->get_diagram(diag_no); if(d!=NULL)diags_res->add_diagram(d->clone());
			} else if (val_i->check_value(CScriptValue::VAL_STRING, true)) {
				CEString* sDiagsI = val_i->get_str_value(); if (sDiagsI==NULL) return CScriptFuncInfo::FReturnError(arg_list, 1+i);
				const char* buf=sDiagsI->getData(); if (buf==NULL) return CScriptFuncInfo::FReturnError(arg_list, 1+i);
				long bufL=CEUtil::length(buf);
				long iRange = CEUtil::findSubString(buf,"Range("); // Allowed value: "Range(N,M)"
				if (iRange==0) {
					long iComm=CEUtil::findSubString(buf,","); if(CEUtil::findSubString(buf+iComm+1,",")>0) return CScriptFuncInfo::FReturnError(arg_list, 1+i);
					long iEnd=CEUtil::findSubString(buf,")");
					long i11=6; long i12=iComm-1; // first/last number pos
					long i21=iComm+1; long i22=iEnd-1;
					if ((i11>i21) || (i21>i22)) CScriptFuncInfo::FReturnError(arg_list, 1+i);
					long iDiagFirst=-1; long iDiagLast=-1;
					char* num1=CEUtil::copySubString(buf+i11,i12-i11+1); char* num2=CEUtil::copySubString(buf+i21,i22-i21+1);
					bool bNumsOK=(CEUtil::atolSafe(num1,&iDiagFirst)&&(CEUtil::atolSafe(num2,&iDiagLast)));
					if (num1!=NULL) CMemoryManager::mfree(num1); if (num2!=NULL) CMemoryManager::mfree(num2);
					if (!bNumsOK) return CScriptFuncInfo::FReturnError(arg_list, 1+i);
					if ((iDiagFirst<0) || (iDiagLast<0)) return CScriptFuncInfo::FReturnError(arg_list, 1+i);

					for(long iTmp=iDiagFirst; iTmp<=iDiagLast; iTmp++) { // Let's allow (iDiagLast >= diags_num)
						if (iTmp>=diags_num) break;
						CFeynDiagram* d = diags->get_diagram(iTmp); if(d!=NULL)diags_res->add_diagram(d->clone());
					}
				} else {
					return CScriptFuncInfo::FReturnError(arg_list, 1+i);
				}
			} else {
				return CScriptFuncInfo::FReturnError(arg_list, 1+i);
			}
		}
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(diags_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncCollectDiffDiagrams(CEPtrArray* arg_list) {
	CDiagramSet* diags1=(CDiagramSet*)((CScriptValue*)arg_list->GetAt(0))->get_obj_value(); if(diags1==NULL)return CScriptFuncInfo::FReturnError(arg_list,1);
	CDiagramSet* diags2=(CDiagramSet*)((CScriptValue*)arg_list->GetAt(1))->get_obj_value(); if(diags2==NULL)return CScriptFuncInfo::FReturnError(arg_list,2);

	CLDiagTopologyInspector::DropSimilarDiagrams(diags1, diags2, NULL);

	long n1=diags1->get_diagrams_number(); long n2=diags2->get_diagrams_number();
	CDiagramSet* diags_res = diags1->clone(); diags_res->add_diagrams(diags2);
	dbg->log(0, diags_res->get_diagrams_number(), "CollectDiffDiagrams result: total %ld different diagrams");
	if(n1>0)dbg->log(0, 0, n1-1, "Diagrams ## %ld-%ld are from the first set.");
	if(n2>0)dbg->log(0, n1, n1+n2-1, "Diagrams ## %ld-%ld are from the second set.");

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(diags_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncCollectEqualDiagrams(CEPtrArray* arg_list) {
	CDiagramSet* diags1=(CDiagramSet*)((CScriptValue*)arg_list->GetAt(0))->get_obj_value(); if(diags1==NULL)return CScriptFuncInfo::FReturnError(arg_list,1);
	CDiagramSet* diags2=(CDiagramSet*)((CScriptValue*)arg_list->GetAt(1))->get_obj_value(); if(diags2==NULL)return CScriptFuncInfo::FReturnError(arg_list,2);

	CDiagramSet* diags_res = diags1->getEmptyFrame();
	CLDiagTopologyInspector::DropSimilarDiagrams(diags1, diags2, diags_res);
	
	dbg->log(0, diags_res->get_diagrams_number(), "CollectDiffDiagrams result: total %ld similar diagrams");

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(diags_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetPhotonHelicity(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value(); //if (part_ID >= 10000) part_ID = -(part_ID-10000);
	CEParticle* part = CEProcessor::kinematics()->get_particle_by_ID(part_ID);
	if (part == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	// note: photon particle's pk is always '+'-signed.
	if (!part->check_PK(CEParticle::PARTICLE_PHOTON, false)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CPPhoton* ph_part = (CPPhoton*)part;

	int h_val = ((CScriptValue*)arg_list->GetAt(1))->get_int_value(); //if (h_val >= 10000) h_val = -(h_val-10000);
	CExpression* base_1 = NULL;
	CExpression* base_2 = NULL;
	if (h_val != 0) {
		if (arg_num > 3) {
			CEString* str_base_1 = ((CScriptValue*)arg_list->GetAt(2))->get_str_value();
			CEString* str_base_2 = ((CScriptValue*)arg_list->GetAt(3))->get_str_value();
			base_1 = new CESymbol(str_base_1->getData());
			base_2 = new CESymbol(str_base_2->getData());
		} else { dbg->log(0, "SetPhotonHelicity(): basis vectors expected"); return CScriptFuncInfo::FReturnError(arg_list, -arg_num); }
	}

	ph_part->setHelicityData(h_val, base_1, base_2);

	if (base_1 != NULL) delete base_1;
	if (base_2 != NULL) delete base_2;
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetPhotonDMBase(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value(); //if (part_ID >= 10000) part_ID = -(part_ID-10000);
	CEParticle* part = CEProcessor::kinematics()->get_particle_by_ID(part_ID);
	if (part == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	// note: photon particle's pk is always '+'-signed.
	if (!part->check_PK(CEParticle::PARTICLE_PHOTON, false)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CPPhoton* ph_part = (CPPhoton*)part;

	CExpression* DM_base = NULL;
	if (arg_num > 1) {
		CEString* str_base = ((CScriptValue*)arg_list->GetAt(1))->get_str_value();
		DM_base = new CESymbol(str_base->getData());
	}

	ph_part->setUnpolarizedDMBase(DM_base);
	if (DM_base != NULL) delete DM_base;

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetStoksParams(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value(); //if (part_ID >= 10000) part_ID = -(part_ID-10000);
	CEParticle* part = CEProcessor::kinematics()->get_particle_by_ID(part_ID);
	if (part == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	// note: photon particle's pk is always '+'-signed.
	if (!part->check_PK(CEParticle::PARTICLE_PHOTON, false)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CPPhoton* ph_part = (CPPhoton*)part;

	//int _pol_info_arg = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	//bool use_Stoks = (_pol_info_arg != 0);
	bool use_Stoks = true;

	CExpression* base_1 = NULL;
	CExpression* base_2 = NULL;
	if (use_Stoks) {
		if (arg_num > 2) {
			CEString* str_base_1 = ((CScriptValue*)arg_list->GetAt(1))->get_str_value();
			CEString* str_base_2 = ((CScriptValue*)arg_list->GetAt(2))->get_str_value();
			base_1 = new CESymbol(str_base_1->getData());
			base_2 = new CESymbol(str_base_2->getData());
		} else { dbg->log(0, "SetStoksParams(): basis vectors expected"); return CScriptFuncInfo::FReturnError(arg_list, -arg_num); }
	}

	ph_part->setStoksParamsUse(use_Stoks, base_1, base_2);

	if (base_1 != NULL) delete base_1;
	if (base_2 != NULL) delete base_2;

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetFermionHelicity(CEPtrArray* arg_list) {
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	//if (part_ID >= 10000) part_ID = -(part_ID-10000);
	CEParticle* part = CEProcessor::kinematics()->get_particle_by_ID(part_ID);
	if (part == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	// check particle kind
	// note: photon particle's pk is always '+'-signed.
	if (!part->check_PK(CEParticle::PARTICLE_FERMION, true)) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CPFermion* part_ferm = (CPFermion*)part;

	CScriptValue* val1=(CScriptValue*)arg_list->GetAt(1);
	CESymbol* h_sym=NULL; int h_val=INT_MIN;
	if (val1->check_value(CScriptValue::VAL_STRING, true)) { h_sym=new CESymbol((val1->get_str_value())->getData()); }
	else if (val1->check_value(CScriptValue::VAL_INT, false)) { h_val = val1->get_int_value(); }
	else { return CScriptFuncInfo::FReturnError(arg_list, 2); }
	//if (h_val >= 10000) h_val = -(h_val-10000);

	if (h_sym!=NULL) { part_ferm->setHelicityData(h_sym); delete h_sym; }
	else { part_ferm->setHelicityData(h_val); }

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetPolarized(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	//if (part_ID >= 10000) part_ID = -(part_ID-10000);
	CEParticle* part = CEProcessor::kinematics()->get_particle_by_ID(part_ID);
	if (part == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	int _pol_info = 1;
	if (arg_num > 1) {
		int _pol_info_arg = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		if (_pol_info_arg == 0) _pol_info = 0;
	}
	
	part->set_polarization_info(_pol_info);

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetMassSym(CEPtrArray* arg_list) { // for pID
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	CEParticle* part = CEProcessor::kinematics()->get_particle_by_ID(part_ID);
	if (part == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1); CESymbol* sym1 = NULL;
	if (val1->check_value(CScriptValue::VAL_STRING, false)) { sym1 = new CESymbol((val1->get_str_value())->getData()); }
	else if (val1->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		CExpression* tmp = val1->get_expr_value();
		if (CEUtil::equals(CESymbol::CLASS_NAME, tmp->ClassName())) { sym1 = (CESymbol*)(tmp->clone()); }
		else { dbg->log(0, "SFuncSetMassSym(): parameter 1 is not SYMBOLIC expression or STRING"); }
	}
	if (sym1 == NULL) return CScriptFuncInfo::FReturnError(arg_list, 2);

	//TODO: include symbol check for usage
	part->set_sym_mass(sym1);

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetHelicityExpr(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	CEParticle* part = CEProcessor::kinematics()->get_particle_by_ID(part_ID);
	if (part == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	CExpression* vec = ((CScriptValue*)arg_list->GetAt(1))->get_expr_value();
	CExpression* ind = ((CScriptValue*)arg_list->GetAt(2))->get_expr_value();
	if (!CEUtil::equals(CESymbol::CLASS_NAME, ind->ClassName())) { dbg->log(0, "SetHelicityExpr(pID, vec, SYMBOLIC index expression required, int)"); return CScriptFuncInfo::FReturnError(arg_list, 1); }
	int ind_cov = ((CScriptValue*)arg_list->GetAt(3))->get_int_value();

	if (CEParticle::is_fermion_PK(part->get_PK())) {
		part->predefine_pol_vector(vec, ind, ind_cov>0);
	} else {
		dbg->log(0, "ERROR: Function SetHelicityExpr() is implemented for FERMIONS only, polarization expressions are not set");
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetBondarevBaseSyms(CEPtrArray* arg_list) {
	CExpression* tmp=NULL;
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0); CESymbol* sym0 = NULL;
	if (val0->check_value(CScriptValue::VAL_STRING, false)) { sym0 = new CESymbol((val0->get_str_value())->getData()); }
	else if (val0->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		tmp = val0->get_expr_value();
		if (CEUtil::equals(CESymbol::CLASS_NAME, tmp->ClassName())) { sym0 = (CESymbol*)(tmp->clone()); }
		else { dbg->log(0, "SFuncSetBondarevBaseSyms(): parameter is not SYMBOLIC expression or STRING"); }
	}
	if (sym0 == NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);

	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(1); CESymbol* sym1 = NULL;
	if (val1->check_value(CScriptValue::VAL_STRING, false)) { sym1 = new CESymbol((val1->get_str_value())->getData()); }
	else if (val1->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		tmp = val1->get_expr_value();
		if (CEUtil::equals(CESymbol::CLASS_NAME, tmp->ClassName())) { sym1 = (CESymbol*)(tmp->clone()); }
		else { dbg->log(0, "SFuncSetBondarevBaseSyms(): parameter is not SYMBOLIC expression or STRING"); }
	}
	if (sym1 == NULL) return CScriptFuncInfo::FReturnError(arg_list, 2);

	CScriptValue* val2 = (CScriptValue*)arg_list->GetAt(0); CESymbol* sym2 = NULL;
	if (val2->check_value(CScriptValue::VAL_STRING, false)) { sym2 = new CESymbol((val2->get_str_value())->getData()); }
	else if (val2->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		tmp = val2->get_expr_value();
		if (CEUtil::equals(CESymbol::CLASS_NAME, tmp->ClassName())) { sym2 = (CESymbol*)(tmp->clone()); }
		else { dbg->log(0, "SFuncSetBondarevBaseSyms(): parameter is not SYMBOLIC expression or STRING"); }
	}
	if (sym2 == NULL) return CScriptFuncInfo::FReturnError(arg_list, 3);

	CScriptValue* val3 = (CScriptValue*)arg_list->GetAt(0); CESymbol* sym3 = NULL;
	if (val3->check_value(CScriptValue::VAL_STRING, false)) { sym3 = new CESymbol((val3->get_str_value())->getData()); }
	else if (val3->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		tmp = val3->get_expr_value();
		if (CEUtil::equals(CESymbol::CLASS_NAME, tmp->ClassName())) { sym3 = (CESymbol*)(tmp->clone()); }
		else { dbg->log(0, "SFuncSetBondarevBaseSyms(): parameter is not SYMBOLIC expression or STRING"); }
	}
	if (sym3 == NULL) return CScriptFuncInfo::FReturnError(arg_list, 4);

	CEFuncBondarev::SetBondarevBaseSyms(sym0,sym1,sym2,sym3);
	delete sym0; delete sym1; delete sym2; delete sym3;

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncNewFortranFile(CEPtrArray* arg_list) {
	dbg->log(0, "------ ERROR: deprecated function call  ------");
	dbg->log(0, "'NewFortranFile' function is DEPRECATED.");
	dbg->log(0, "Use 'NewCodeFile' procedure.");
	return CScriptFunctionLib::SFuncNewCodeFile(arg_list);
}

CEPtrArray* CScriptFunctionLib::SFuncCreateFortranProc(CEPtrArray* arg_list) {
	dbg->log(0, "------ ERROR: deprecated function call  ------");
	dbg->log(0, "'CreateFortranProc' function is DEPRECATED.");
	dbg->log(0, "Use 'CreateCodeProc' procedure.");
	return CScriptFunctionLib::SFuncCreateCodeProc(arg_list);
}

CEPtrArray* CScriptFunctionLib::SFuncNewCodeFile(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	
	CEString* str_file_name = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	DWORD _setts = 0; //Auto-define language by default
	if (arg_num > 1) {
		_setts = (DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		//Auto-define language by default //if ((_setts & (CLFileManager::FT_F90 | CLFileManager::FT_F77 | CLFileManager::FT_CPP)) == 0) { _setts = CLFileManager::FT_F77; }
	}

	// create fortran file
	int file_ID = CLFileManager::new_file(str_file_name->getData(), _setts);
	if (file_ID < 0) { dbg->log(0, "Error creation file"); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	
	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(file_ID));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncCreateCodeProc(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	
	int file_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	CEString* proc_name = ((CScriptValue*)arg_list->GetAt(1))->get_str_value();
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(2))->get_expr_value();
	DWORD _setts = CLFileManager::FT_REAL_8BIT; //CLFileManager::FT_COMPLEX_16BIT
	// FT_REAL_8BIT			= 0x08;
	// FT_COMPLEX_16BIT		= 0x10;
	if (arg_num > 3) {
		_setts = (DWORD)((CScriptValue*)arg_list->GetAt(3))->get_int_value();
		//if ((_setts != CLFileManager::FT_REAL_8BIT) && (_setts != CLFileManager::FT_COMPLEX_16BIT)) {
		//	return CScriptFuncInfo::FReturnError(arg_list, 4);
		//}
	}

	// add fortran procedure code to file
	CLFileManager::add_to_file(file_ID, proc_name->getData(), expr_val, _setts);

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncMinimize(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	DWORD _setts = CViewSimplify::PART_FACT_DEFAULT;// | CViewSimplify::PART_FACT_VERIFY;
	if (arg_num > 1) {
		_setts = (DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	}

	CExpression* res_expr = CViewSimplify::factorize_partial(expr_val->clone(), _setts);
	if (res_expr == NULL) { // error
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}
	
	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res_expr, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncFactor(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	//DWORD _setts = CLFactor::FACT_VERIFY;//CLFactor::DONOT_FACT_FRACTIONS;
	DWORD _setts = CLFactor::DONOT_FACT_FRACTIONS; //| CLFactor::FACT_VERIFY
	if (arg_num > 1) {
		_setts = (DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	}

	CExpression* expr_res = CLFactor::factorize(expr_val->clone(), _setts);
	if (expr_res == NULL) { // error
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(expr_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncEquals(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CExpression* expr_1 = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	CExpression* expr_2 = ((CScriptValue*)arg_list->GetAt(1))->get_expr_value();
	DWORD _setts = 0;
	if (arg_num > 2) { _setts = (DWORD)((CScriptValue*)arg_list->GetAt(2))->get_int_value(); }

	if (expr_1->equals(expr_2)) { printf("Equals: expressions are EQUAL\n"); }
	else {
		if (_setts == 0) { printf("Equals: expressions are NOT EQUAL\n"); }
		else {
			// advanced comparison
			expr_1 = CExpression::reduce(expr_1->clone());
			expr_2 = CExpression::reduce(expr_2->clone());
			if (expr_1->equals(expr_2)) printf("Equals: expressions become EQUAL after expanding\n");
			else printf("Equals: expressions are NOT EQUAL\n");
			delete expr_1; delete expr_2;
		}
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncEvaluate(CEPtrArray* arg_list) {
	/*
	dbg->log(0, "------ ERROR: deprecated function call  ------");
	dbg->log(0, "'Evaluate' function is DEPRECATED.");
	dbg->log(0, "Use 'ReduceTensorVI' procedure.");
	dbg->log(0, "=== Automatically redirected to ReduceTensorVI() ===");
	return SFuncReduceTensorVI(arg_list);
	*/
	///* //Old:
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	expr_val=expr_val->clone();
	CExpression* res = expr_val->evaluate(); if(res!=expr_val)delete expr_val;
	if (res == NULL) { dbg->log(0, "ERROR: evaluate() result == NULL."); return CScriptFuncInfo::FReturnError(arg_list, 0); }

	if (CEProcessor::m_bAutoKinArrange && (CEProcessor::kinematics()!=NULL)) {
		DWORD KA_setts = 0; if (CEProcessor::m_MandelstamsEverywhere) { KA_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
		res = CLHEPBasics::KinArrange(res, KA_setts); if(res==NULL){ dbg->log(0, "ERROR: Auto-KinArrange() failed."); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	}
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
	//*/
}

CEPtrArray* CScriptFunctionLib::SFuncKinArrange(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	DWORD _setts = 0;
	if (arg_num > 1) _setts = (DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	if (CEProcessor::m_MandelstamsEverywhere) { _setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }

	CExpression* expr_res = CLHEPBasics::KinArrange(expr_val->clone(), _setts);
	if (expr_res == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(expr_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncKinSimplify(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	//Default settings
	DWORD _setts = CLHEPBasics::KS_FACTORIZE_FUNCTIONS|CLHEPBasics::KS_FACTORIZE_DENOMINATORS;
	//Other setts: CLHEPBasics::KS_FACTORIZE; CLHEPBasics::KS_FACTORIZE_ALL_DENOMINATORS;

	if(arg_num>1)_setts=(DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	if (CEProcessor::m_MandelstamsEverywhere) { _setts |= CLHEPBasics::KS_INVOLVE_MANDELSTAM_VARS; }
	_setts |= CLHEPBasics::KS_FAST; //the only difference to 'SFuncKinFullSimplify'

	CExpression* expr_res = CLHEPBasics::KinSimplify(expr_val->clone(), _setts);
	if (expr_res == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(expr_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncKinFullSimplify(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	//Default settings
	DWORD _setts = CLHEPBasics::KS_FACTORIZE_FUNCTIONS|CLHEPBasics::KS_FACTORIZE_DENOMINATORS;
	//Other setts: CLHEPBasics::KS_FACTORIZE; CLHEPBasics::KS_FACTORIZE_ALL_DENOMINATORS;

	if (arg_num > 1) _setts = (DWORD)((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	if (CEProcessor::m_MandelstamsEverywhere) { _setts |= CLHEPBasics::KS_INVOLVE_MANDELSTAM_VARS; }

	CExpression* expr_res = CLHEPBasics::KinSimplify(expr_val->clone(), _setts);
	if (expr_res == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(expr_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSimplify(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	CExpression* val = ((CScriptValue*)arg_list->GetAt(1))->get_expr_value();
	CExpression* subst = ((CScriptValue*)arg_list->GetAt(2))->get_expr_value();
	int iFlags=0; if(arg_num>3)iFlags=(DWORD)((CScriptValue*)arg_list->GetAt(3))->get_int_value();

	bool bFast = (iFlags>0);
	CEReplList r(val,subst);
	CExpression* res = CSubstSimplify::SimpRepl(src->clone(),&r,NULL,1,bFast);

	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSingularArrange(CEPtrArray* arg_list) {
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	
	CExpression* res = CLHEPBasics::SingularArrange(expr_val->clone());
	if (res == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);
	if (CEProcessor::m_bAutoKinArrange && (CEProcessor::kinematics()!=NULL)) {
		DWORD KA_setts = 0;
		if (CEProcessor::m_MandelstamsEverywhere) { KA_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
		res = CLHEPBasics::KinArrange(res, KA_setts);
		if (res == NULL) { dbg->log(0, "ERROR: Auto-KinArrange() failed."); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	}
	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncGetUVTerm(CEPtrArray* arg_list) {
	CExpression* expr_val = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	
	CExpression* expr_res = CLHEPBasics::GetUVTerm(expr_val->clone());
	if (expr_res == NULL) return CScriptFuncInfo::FReturnError(arg_list, 0);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(expr_res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncRange(CEPtrArray* arg_list) {
	long i1 = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	long i2 = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();

	CEString* sRes=new CEString("Range("); sRes->add_int(i1); sRes->AddString(","); sRes->add_int(i2); sRes->AddString(")");
	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(sRes, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetDiagramPhysics(CEPtrArray* arg_list) {
	long phys_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	if ((phys_ID & (CSMPhysics::PHYS_SM | CSMPhysics::PHYS_QCD)) == 0)
		return CScriptFuncInfo::FReturnError(arg_list, 1);

	CDiagramPhysics* phys_old = CDiagramPhysics::get_curr_physics();
	if ((phys_old == NULL) || (!CEUtil::equals(CSMPhysics::CLASS_NAME,phys_old->ClassName())) || (phys_old->getPhysicsID() != phys_ID)) {
		CDiagramPhysics* phys_new = new CSMPhysics(phys_ID);
		CDiagramPhysics::install_diagram_physics(phys_new, true);
		
		// log out correspondence rules
		// note: correspondence rules logging requires physics to be already installed
		CDiagramPhysics::dbg->log(7, "New hamiltonian expression", phys_new->getHamiltonianExpr());
		CDiagramPhysics::dbg->log(7, "New correspondence rules", phys_new->getCorrespondenceRules());
	} else {
		CDiagramPhysics::dbg->log(0, "{SetDiagramPhysics:} Diagram physics remains old (no relevant changes are found)");
		CDiagramPhysics::dbg->log(7, "Hamiltonian expression is", phys_old->getHamiltonianExpr());
		CDiagramPhysics::dbg->log(7, "Correspondence rules are", phys_old->getCorrespondenceRules());
	}

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetVirtIntegralValues(CEPtrArray* arg_list) {
	CVirtIntegralLib::fill_data_storage();
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetDRMassPK(CEPtrArray* arg_list) {
	int pk = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if(phys==NULL) { dbg->log(0,"<ERROR> No physics installed"); return CScriptFuncInfo::FReturnError(arg_list, 0); }

	if (pk != 0) {
		// validate PK
		if (!phys->validate_part_kind(pk)) { dbg->log(0, pk, "<ERROR> Invalid PK value: %ld"); return CScriptFuncInfo::FReturnError(arg_list, 1); }
	}

	phys->set_DR_mass_PK(pk);

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSetNDimensionSpace(CEPtrArray* arg_list) {
	int par = 1;
	int arg_num = arg_list->size();
	if (arg_num > 0) par = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();

	if (par != 0) CETensor::setNDimensionSpace(true);
	else CETensor::setNDimensionSpace(false);

	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncReduceTensorVI(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* res = CVirtIntegralLib::ReduceTensorVI(src->clone());
	if(res==NULL){ dbg->log(0, "ERROR: SFuncReduceTensorVI() result == NULL."); return CScriptFuncInfo::FReturnError(arg_list, 0); }

	if (CEProcessor::m_bAutoKinArrange && (CEProcessor::kinematics()!=NULL)) {
		DWORD KA_setts=0; if(CEProcessor::m_MandelstamsEverywhere){ KA_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
		res = CLHEPBasics::KinArrange(res, KA_setts);
		if(res==NULL){ dbg->log(0, "ERROR: Auto-KinArrange() failed."); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	}
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncConvertInvariantVI(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* res = CVirtIntegralLib::ConvertVI0ToInvForm(src->clone());
	if (res == NULL) { // error
		dbg->log(0, "ERROR: SFuncConvertInvariantVI() result == NULL.");
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncCalcScalarVI(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* res = CVIScalarCalcLib::CalcScalarVI(src->clone());
	if (res == NULL) { // error
		dbg->log(0, "ERROR: CalcScalarVI() result == NULL.");
		return CScriptFuncInfo::FReturnError(arg_list, 0);
	}

	if (CEProcessor::m_bAutoKinArrange && (CEProcessor::kinematics()!=NULL)) {
		DWORD KA_setts = 0;
		if (CEProcessor::m_MandelstamsEverywhere) { KA_setts |= CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS; }
		res = CLHEPBasics::KinArrange(res, KA_setts);
		if (res == NULL) { dbg->log(0, "ERROR: Auto-KinArrange() failed."); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncRestoreRenormConstants(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	
	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if (phys == NULL) { dbg->log(0, "<ERROR> No physics installed"); return CScriptFuncInfo::FReturnError(arg_list, 0); }

	CERule* r = phys->getRenormConstValues();
	CExpression* res = r->ApplyTo(src->clone());
	delete r;
	res = CExpression::arrge(res);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSubstVecZero(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	CEString* str = ((CScriptValue*)arg_list->GetAt(1))->get_str_value();

	CExpression* res = src->clone();
	CExpression* sym = new CESymbol(str->getData());
	CExpression* zero = CENumber::GetZero();
	res = CEReplasement::Replace(res, sym, zero);
	delete sym; delete zero;

	res = CExpression::arrge(res);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSubstKinVector(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	CEString* str = ((CScriptValue*)arg_list->GetAt(1))->get_str_value();

	CKinematics* kin = CEProcessor::kinematics();
	if (kin == NULL) { dbg->log(0, "<ERROR> No kinematics installed"); return CScriptFuncInfo::FReturnError(arg_list, 0); }

	CExpression* res = src->clone();
	CExpression* sym = new CESymbol(str->getData());
	CExpression* subst = kin->getEqual(sym);
	res = CEReplasement::Replace(res, sym, subst);
	delete sym; delete subst;
	res = CExpression::arrge(res);

	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncGetMassSymPK(CEPtrArray* arg_list) {
	int pk = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();

	CDiagramPhysics* phys = CDiagramPhysics::get_curr_physics();
	if (phys == NULL) { dbg->log(0, "GetMassSym() ERROR: No physics installed"); return CScriptFuncInfo::FReturnError(arg_list, 0); }
	
	CExpression* mass_sym = phys->get_mass_sym(pk);
	CExpression* res = mass_sym->clone();

	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

//macros
CEPtrArray* CScriptFunctionLib::SFuncInitial(CEPtrArray* arg_list) {
	int part_No = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	int part_ID = part_No+1;
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(part_ID)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncFinal(CEPtrArray* arg_list) {
	int part_No = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	int part_ID = -(part_No+1);
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(part_ID)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSym(CEPtrArray* arg_list) {
	CEString* str = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	CExpression* res = new CESymbol(str->getData());
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncNum(CEPtrArray* arg_list) {
	int arg_num=arg_list->size();
	int iNum=((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	int iDen=0; if(arg_num>1){ iDen=((CScriptValue*)arg_list->GetAt(1))->get_int_value(); if(iDen==0)return CScriptFuncInfo::FReturnError(arg_list,2); }
	CENumber* res=((iDen!=0)?(new CENumber(iNum,iDen)):new CENumber(iNum));
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncImU(CEPtrArray* arg_list) {
	CExpression* res = CENumber::GetImUnit();
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncInfinity(CEPtrArray* arg_list) {
	CExpression* res = new CEInfinity();
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncGetUnique(CEPtrArray* arg_list) {
	CExpression* res = CESymbol::getUnique();
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncVec(CEPtrArray* arg_list) {
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	CExpression* base = NULL;
	if (val0->check_value(CScriptValue::VAL_STRING, false)) {
		base = new CESymbol((val0->get_str_value())->getData());
	} else if (val0->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		CExpression* val0_expr = val0->get_expr_value();
		if(val0_expr!=NULL){
			//Old: if (CEUtil::equals(CESymbol::CLASS_NAME, val0_expr->ClassName())) { base = (CESymbol*)(val0_expr->clone()); }
			base=val0_expr->clone();
		}
	}

	//Old: if (base == NULL) { dbg->log(0, "Vec(symbolic expression OR string,...) expected"); return CScriptFuncInfo::FReturnError(arg_list, 1); }
	if(base==NULL){ dbg->log(0, "Vec(expression OR string,...) expected"); return CScriptFuncInfo::FReturnError(arg_list, 1); }

	CExpression* val1_expr = ((CScriptValue*)arg_list->GetAt(1))->get_expr_value();
	if (!CEUtil::equals(CESymbol::CLASS_NAME, val1_expr->ClassName())) { dbg->log(0, "Vec(..., symbol expression,...) expected"); return CScriptFuncInfo::FReturnError(arg_list, 2); }
	CESymbol* ind = (CESymbol*)val1_expr;

	int cov_val = ((CScriptValue*)arg_list->GetAt(2))->get_int_value();
	bool cov = (cov_val>0);

	CExpression* res = new CEVector(base, ind, cov);
	delete base;

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncEps(CEPtrArray* arg_list) {
	CExpression* val0_expr = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	if (!CEUtil::equals(CESymbol::CLASS_NAME, val0_expr->ClassName())) { dbg->log(0, "Esp(SYMBOL expressions required)"); return CScriptFuncInfo::FReturnError(arg_list, 1); }
	int cov0 = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	CExpression* val1_expr = ((CScriptValue*)arg_list->GetAt(2))->get_expr_value();
	if (!CEUtil::equals(CESymbol::CLASS_NAME, val1_expr->ClassName())) { dbg->log(0, "Esp(SYMBOL expressions required)"); return CScriptFuncInfo::FReturnError(arg_list, 3); }
	int cov1 = ((CScriptValue*)arg_list->GetAt(3))->get_int_value();
	CExpression* val2_expr = ((CScriptValue*)arg_list->GetAt(4))->get_expr_value();
	if (!CEUtil::equals(CESymbol::CLASS_NAME, val2_expr->ClassName())) { dbg->log(0, "Esp(SYMBOL expressions required)"); return CScriptFuncInfo::FReturnError(arg_list, 5); }
	int cov2 = ((CScriptValue*)arg_list->GetAt(5))->get_int_value();
	CExpression* val3_expr = ((CScriptValue*)arg_list->GetAt(6))->get_expr_value();
	if (!CEUtil::equals(CESymbol::CLASS_NAME, val3_expr->ClassName())) { dbg->log(0, "Esp(SYMBOL expressions required)"); return CScriptFuncInfo::FReturnError(arg_list, 7); }
	int cov3 = ((CScriptValue*)arg_list->GetAt(7))->get_int_value();

	CExpression* res = new CEAntisymmetric(val0_expr,(cov0>0),val1_expr,(cov1>0),val2_expr,(cov2>0),val3_expr,(cov3>0));

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSqrt(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* res = CEPower::sqrt(src->clone());

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSqr(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* res = src->sqra();

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSpr(CEPtrArray* arg_list) {
	CExpression* src1 = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	CExpression* src2 = ((CScriptValue*)arg_list->GetAt(1))->get_expr_value();

	CExpression* res = new CESProduct(src1, src2);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncPower(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	int pow = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();

	CExpression* res = new CEPower(src, (long)pow);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncReplace(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	CScriptValue* val1=(CScriptValue*)arg_list->GetAt(1);
	CScriptValue* val2=(CScriptValue*)arg_list->GetAt(2);

	CExpression* exprOld = NULL; //cloned
	if (val1->check_value(CScriptValue::VAL_STRING, false)) { exprOld = new CESymbol((val1->get_str_value())->getData()); }
	else if (val1->check_value(CScriptValue::VAL_EXPRESSION, false)) { exprOld=val1->get_expr_value(); if(exprOld!=NULL)exprOld=exprOld->clone(); }
	if(exprOld==NULL)return CScriptFuncInfo::FReturnError(arg_list, 2);
	CExpression* exprNew = NULL; //cloned
	if (val2->check_value(CScriptValue::VAL_STRING, false)) { exprNew = new CESymbol((val2->get_str_value())->getData()); }
	else if (val2->check_value(CScriptValue::VAL_INT, false)) { exprNew = CENumber::GetNumber(val2->get_int_value()); }
	else if (val2->check_value(CScriptValue::VAL_EXPRESSION, false)) { exprNew=val2->get_expr_value(); if(exprNew!=NULL)exprNew=exprNew->clone(); }
	if(exprNew==NULL){ delete exprOld; return CScriptFuncInfo::FReturnError(arg_list, 3); }

	CEReplasement r(exprOld,exprNew,false); delete exprOld; delete exprNew;
	CExpression* res = r.ApplyTo(src->clone(),true);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncSubstVec(CEPtrArray* arg_list) {
	CExpression* src = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* vec_base = ((CScriptValue*)arg_list->GetAt(1))->get_expr_value();
	if (!CEUtil::equals(CESymbol::CLASS_NAME, vec_base->ClassName())) { dbg->log(0, "SFuncSubstVec(expr, SYMBOL expressions required)"); return CScriptFuncInfo::FReturnError(arg_list, 2); }

	CExpression* vec_new = ((CScriptValue*)arg_list->GetAt(2))->get_expr_value();
	CExpression* ind_new = ((CScriptValue*)arg_list->GetAt(3))->get_expr_value();

	if (!CEUtil::equals(CESymbol::CLASS_NAME, ind_new->ClassName())) { dbg->log(0, "SFuncSubstVec(expr, sym, expr, SYMBOLIC index expected)"); return CScriptFuncInfo::FReturnError(arg_list, 4); }

	CExpression* res = src->clone();
	dbg->log(0, "ERROR: CScriptFunctionLib::SFuncSubstVec: Not implemented yet");

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncFrac(CEPtrArray* arg_list) {
	CScriptValue* val1 = (CScriptValue*)arg_list->GetAt(0);
	CScriptValue* val2 = (CScriptValue*)arg_list->GetAt(1);
	CExpression* expr1 = NULL; bool expr1_cloned=false;
	if (val1->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		expr1 = val1->get_expr_value();
	} else if (val1->check_value(CScriptValue::VAL_INT, false)) {
		expr1 = new CENumber(val1->get_int_value()); expr1_cloned=true;
	}
	if (expr1==NULL) return CScriptFuncInfo::FReturnError(arg_list, 1);
	CExpression* expr2 = NULL; bool expr2_cloned=false;
	if (val2->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		expr2 = val2->get_expr_value();
	} else if (val2->check_value(CScriptValue::VAL_INT, false)) {
		expr2 = new CENumber(val2->get_int_value()); expr2_cloned=true;
	}
	if (expr2==NULL) return CScriptFuncInfo::FReturnError(arg_list, 2); 

	if (!expr1_cloned) expr1=expr1->clone();
	if (!expr2_cloned) expr2=expr2->clone();
	CExpression* res = expr1->divide(expr2);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncGamma(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	CESymbol* sym = NULL;
	int gamma_5=0; bool is_gamma_5=false;
	if (val0->check_value(CScriptValue::VAL_STRING, false)) {
		sym = new CESymbol((val0->get_str_value())->getData());
	} else if (val0->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		CExpression* val0_expr = val0->get_expr_value();
		if (CEUtil::equals(CESymbol::CLASS_NAME, val0_expr->ClassName())) {
			sym = (CESymbol*)(val0_expr->clone());
		}
	} else if (val0->check_value(CScriptValue::VAL_INT, false)) {
		gamma_5 = (val0->get_int_value());
		is_gamma_5=true;
	}
	if ((sym == NULL)&&(!is_gamma_5)) { dbg->log(0, "SFuncGamma(SYMBOL,STRING or INT,...) object expected"); return CScriptFuncInfo::FReturnError(arg_list, 1); }

	bool cov=true;
	if(arg_num>1) {
		if (is_gamma_5) { dbg->log(0, "Gamma(int) Unexpected SECOND parameter for Gamma-5 creation"); return CScriptFuncInfo::FReturnError(arg_list, 2); }
		int i_cov = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
		cov = (i_cov>0);
	}

	CExpression* res = NULL;
	if (is_gamma_5) {
		if (gamma_5>0) gamma_5=1; // truncate higher numbers
		if (gamma_5<0) gamma_5=-1; // truncate higher numbers
		res = new CEGammaSpecial(gamma_5);
	} else { // sym != NULL
		if (arg_num>1) { // gamma-vector
			res = new CEGamma(sym,cov);
		} else { // gamma-coupling
			res = new CESGammaProduct(sym);
		}
	}
	delete sym;

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncGellMann(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	CScriptValue* val0 = (CScriptValue*)arg_list->GetAt(0);
	CESymbol* sym = NULL;
	if (val0->check_value(CScriptValue::VAL_STRING, false)) {
		sym = new CESymbol((val0->get_str_value())->getData());
	} else if (val0->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		CExpression* val0_expr = val0->get_expr_value();
		if (CEUtil::equals(CESymbol::CLASS_NAME, val0_expr->ClassName())) {
			sym = (CESymbol*)(val0_expr->clone());
		}
	}
	if (sym == NULL) { dbg->log(0, "SFuncGellMann(SYMBOL or STRING) object expected"); return CScriptFuncInfo::FReturnError(arg_list, 1); }

	CExpression* res = new CEGellMann(sym); delete sym;

	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res, true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncToAsymmProd(CEPtrArray* arg_list) {
	CExpression* expr1 = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	CExpression* res = expr1->clone();
	if (CEAntiSymmProd::InvolveAsymmProd(res)) {
		res= CExpression::arrge(res);
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncTr(CEPtrArray* arg_list) {
	CExpression* expr1 = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	/*
	// 1. Gell-Mann trace
	bool bDelExpr1=false; bool bNeedArrange=false;
	if(CEGellMannTr::IsGellMannOperator(expr1)) {
		CExpression* tmp = CEGellMannTr::TraceFactorize(expr1,true,false);
		if(tmp!=NULL){ expr1=tmp; bDelExpr1=true; bNeedArrange=true; }else{ CEException::halt("IError in CScriptFunctionLib::SFuncTr"); }
	}

	// 2. Gamma trace
	CExpression* res = new CFSp(expr1); if(bDelExpr1){ delete expr1; }

	// 3. Arrange if required (e.g. to separate Sp[Tr[]] --> Sp[]*Tr[])
	if(bNeedArrange){
		//res = CExpression::arrge(res);
		res = CExpression::reduce(res);
	}
	*/

	CExpression* res = new CFSp(expr1);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncTrGellMann(CEPtrArray* arg_list) {
	CExpression* expr1 = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();

	// Gell-Mann trace
	//if(CEGellMannTr::IsGellMannOperator(expr1)) {
	CExpression* res = CEGellMannTr::TraceFactorize(expr1,true,false);
	if(res==NULL){ CEException::halt("IError in CScriptFunctionLib::SFuncTr"); }
	if(res==expr1)res=res->clone();
	res=CExpression::arrge(res);

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncFiertzTransform(CEPtrArray* arg_list) {
	CExpression* expr1 = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	
	CExpression* res = CLFiertzTransform::DoFiertzTransform(expr1->clone());

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncDumpDataStorage(CEPtrArray* arg_list) {
	CEProcessor::getDataStorage()->store_node();
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncLog(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	// consider first argument only
	if (arg_num != 1) return CScriptFuncInfo::FReturnError(arg_list, -arg_num);
	CScriptValue* val = (CScriptValue*)arg_list->GetAt(0);
	CEString* log_str = new CEString("<Log> ");
	CEPtrArray arrObjs;
	bool val_logged = false;

	// check for string
	if (val->check_value(CScriptValue::VAL_STRING, false)) {
		log_str->AddString("str: ");
		if (val->check_value(CScriptValue::VAL_STRING, true)) {
			log_str->AddString(val->get_str_value()->clone());
			arrObjs.add(val->get_str_value()->clone());
		} else {
			log_str->AddString("NULL");
		}
		val_logged = true;
	}

	// check for integer
	if (val->check_value(CScriptValue::VAL_INT, false)) {
		if (val_logged) log_str->AddString("   ");
		log_str->AddString("int: ");
		log_str->add_int(val->get_int_value(),true);
		val_logged = true;
	}

	// check for expression
	if (val->check_value(CScriptValue::VAL_EXPRESSION, false)) {
		if (val_logged) log_str->AddString("   ");
		log_str->AddString("expr: ");
		if (val->check_value(CScriptValue::VAL_EXPRESSION, true)) {
			log_str->AddString(val->get_expr_value()->ClassName());
			arrObjs.add(val->get_expr_value()->clone());
		} else {
			log_str->AddString("NULL");
		}
		val_logged = true;
	}

	// check for matrix element
	if (val->check_value(CScriptValue::VAL_MATRIX_ELEMENT, false)) {
		if (val_logged) log_str->AddString("   ");
		if (val->check_value(CScriptValue::VAL_MATRIX_ELEMENT, true)) {
			log_str->AddString("me");
			arrObjs.add(val->get_obj_value()->copy());
		} else {
			log_str->AddString("me: NULL");
		}
		val_logged = true;
	}

	// check for diagram set
	if (val->check_value(CScriptValue::VAL_DIAGRAM_SET, false)) {
		if (val_logged) log_str->AddString("   ");
		log_str->AddString("diags: ");
		if (val->check_value(CScriptValue::VAL_DIAGRAM_SET, true)) {
			CDiagramSet* diags_val = (CDiagramSet*)val->get_obj_value();
			log_str->add_int(diags_val->get_diagrams_number());
			log_str->AddString(" diagrams");
			dbg->log(2, "Log():", diags_val);
			arrObjs.add(diags_val->copy());
		} else {
			log_str->AddString("NULL");
		}
		val_logged = true;
	}

	// check for undefined value
	if (val->check_value(CScriptValue::VAL_UNDEFINED, false)) {
		if (val_logged) log_str->AddString("   ");
		log_str->AddString("undefined value");
		val_logged = true;
	}

	if (!val_logged) {
		log_str->AddString("Unknown variable type. (Error?)");
	}

	//printf(log_str->getData()); printf("\n");
	dbg->log(0,log_str->getData(),((arrObjs.size()>0)?&arrObjs:NULL));

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(log_str, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncViewParticleData(CEPtrArray* arg_list) {
	int part_ID = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
	//if (part_ID >= 10000) part_ID = -(part_ID-10000);
	CKinematics* kin = CEProcessor::kinematics(); if (kin == NULL) CEException::halt(dbg, "No kinematics found");
	CEParticle* part = kin->get_particle_by_ID(part_ID);
	
	CEString* log_str = new CEString("{ID="); log_str->add_int(part_ID); log_str->AddString("}: ");
	if (part != NULL) {
		int pk = part->get_PK();
		log_str->AddString(CDiagramPhysics::get_particle_symbolA(pk, CDiagramPhysics::SYMBOL_STYLE_NB));
		log_str->AddString(", ");
		log_str->AddString(CEBasic::ObjToString(part->get_momentum_symbol(), STORE_PLAIN_TEXT));
		log_str->AddString(", ");
		log_str->AddString(CEBasic::ObjToString(part->get_pol_vector_symbol(), STORE_PLAIN_TEXT));
		log_str->AddString(", ");
		log_str->AddString(CEBasic::ObjToString(part->get_sym_mass(), STORE_PLAIN_TEXT));
		log_str->AddString(", ");
		log_str->AddString((part->get_polarization_info() != 0)?("polarized"):("unpolarized"));

		if (CEParticle::is_fermion_PK(pk)) {
			CPFermion* ferm = (CPFermion*)part;
			log_str->AddString(", ");
			log_str->AddString(CEBasic::ObjToString(ferm->get_helicity_symbol(), STORE_PLAIN_TEXT));
			log_str->AddString(", ");
			log_str->add_int(ferm->get_helicity_value());
		} else if (part->check_PK(CEParticle::PARTICLE_PHOTON, true)) {
			CPPhoton* phot = (CPPhoton*)part;
			CSequence* _stoks = phot->GetStoksParamsDeclared();
			if (_stoks != NULL) {
				log_str->AddString(", Stoks: ");
				log_str->AddString(CEBasic::ObjToString(_stoks, STORE_PLAIN_TEXT));
			} else {
				log_str->AddString(", ");
				log_str->add_int(phot->get_helicity_value());
				log_str->AddString(", ");
				log_str->AddString(CEBasic::ObjToString(phot->get_helicity_base(1), STORE_PLAIN_TEXT));
				log_str->AddString(", ");
				log_str->AddString(CEBasic::ObjToString(phot->get_helicity_base(2), STORE_PLAIN_TEXT));
			}
		}
	} else {
		log_str->AddString(": No particle exist");
	}

	printf(log_str->getData()); printf("\n");
	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(log_str, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncHalt(CEPtrArray* arg_list) {
	int par = 0;
	if (arg_list->size() > 0) {
		par = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
		if ((par < 0) || (par >= (int)CScriptFuncInfo::ERROR_HALT_FLAG)) {
			return CScriptFuncInfo::FReturnError(arg_list, 1); // invalid argument
		}
	}
	return CScriptFuncInfo::FReturnError(arg_list, CScriptFuncInfo::ERROR_HALT_FLAG | par);
}

CEPtrArray* CScriptFunctionLib::SFuncTimer(CEPtrArray* arg_list) {
	CScriptProcessor::dbg->logperiod(0, "Time elapsed: ");
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncUseCharBank(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();
	bool use_CM = true;
	int par = 1; // CLSymbolDataBank::NO_MAP_FOR_UNIQUES;
	if (arg_num > 0) {
		par = ((CScriptValue*)arg_list->GetAt(0))->get_int_value();
		if (par == 0) use_CM = false;
	}

	if (use_CM) { CLSymbolDataBank::init(par); }
	else { CLSymbolDataBank::finalize(); }
	arg_list->clear_all();
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncChangeDebugLevel(CEPtrArray* arg_list) {
	CEString* obj_title = ((CScriptValue*)arg_list->GetAt(0))->get_str_value();
	int lev_new = ((CScriptValue*)arg_list->GetAt(1))->get_int_value();
	if (lev_new <= 0) lev_new = 1;

	int lev_old = 1;
	CDebug* dbg_obj = CDebug::GetDebugObject(obj_title->getData());
	if (dbg_obj != NULL) {
		lev_old = dbg_obj->set_level(lev_new);
	} else {
		dbg->log(0, "Error: Debug object not found for ID", obj_title);
		return CScriptFuncInfo::FReturnError(arg_list, 1);
	}

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(lev_old));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncDbgExtractVISet(CEPtrArray* arg_list) {
	CExpression* expr = ((CScriptValue*)arg_list->GetAt(0))->get_expr_value();
	dbg->log(0,"DbgExtractVISet() call");
	CESum* res = new CESum();
	CRCollectSymbols r0(0,NULL,CVirtInt0::CLASS_NAME); r0.ApplyTo(expr,false); CEPtrArray* vi0list=r0.getCollectedSymbols();
	if(vi0list!=NULL){
		dbg->log(0,vi0list->size(),"DbgExtractVISet: total %ld VI-0 integrals.",vi0list);
		for(int i=0; i<vi0list->size(); i++){ res->addItem(((CExpression*)vi0list->GetAt(i))->clone()); }
	}
	CRCollectSymbols r1(0,NULL,CVirtInt1::CLASS_NAME); r1.ApplyTo(expr,false); CEPtrArray* vi1list=r1.getCollectedSymbols();
	if(vi1list!=NULL){
		dbg->log(0,vi1list->size(),"DbgExtractVISet: total %ld VI-1 integrals.",vi1list);
		for(int i=0; i<vi1list->size(); i++){ res->addItem(((CExpression*)vi1list->GetAt(i))->clone()); }
	}
	CRCollectSymbols rC(0,NULL,CVirtIntC::CLASS_NAME); rC.ApplyTo(expr,false); CEPtrArray* viClist=rC.getCollectedSymbols();
	if(viClist!=NULL){
		dbg->log(0,viClist->size(),"DbgExtractVISet: total %ld VI-C integrals.",viClist);
		for(int i=0; i<viClist->size(); i++){ res->addItem(((CExpression*)viClist->GetAt(i))->clone()); }
	}
	arg_list->clear_all(); arg_list->add(CScriptValue::newValue(res,true)); return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncTestCalc(CEPtrArray* arg_list) {
	int arg_num = arg_list->size();

	CExpression* a = (arg_num>0)?((CScriptValue*)arg_list->GetAt(0))->get_expr_value():NULL;
	CExpression* b = (arg_num>1)?((CScriptValue*)arg_list->GetAt(1))->get_expr_value():NULL;

	CExpression* res = CTestCalculation::ScriptTestCommand(a, b);
	if ((res != NULL) && ((res == a) || (res == b))) res = res->clone();

	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}

CEPtrArray* CScriptFunctionLib::SFuncTestCalcB(CEPtrArray* arg_list) {
	CExpression* res = CTestCalculation::ScriptTestCommandB();
	arg_list->clear_all();
	arg_list->add(CScriptValue::newValue(res, true));
	return arg_list;
}
