////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// SriptProcessor.cpp: implementation of the CScriptProcessor class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
//#include <stdexcpt.h>

CDebug* CScriptProcessor::dbg = new CDebug("CScriptProcessor");
CEMap* CScriptProcessor::m_functions_register = NULL;
bool CScriptProcessor::m_init_OK = false;
//CEMap* CScriptProcessor::m_map_operations = NULL;

const DWORD CScriptProcessor::LINE_BRAKE_EOF			= 0x01;
const DWORD CScriptProcessor::LINE_BRAKE_RIGHT_BRACE	= 0x02;
const DWORD CScriptProcessor::LINE_BRAKE_PRIORITY		= 0x04;
const DWORD CScriptProcessor::LINE_BRAKE_ANY_FLAG		= 0xff;
//const DWORD CScriptProcessor::LINE_BRAKE_CURRENT_SKIPED	= 0x100;

CScriptProcessor* CScriptProcessor::m_WorkingProc = NULL;

CEMap* CScriptProcessor::m_PreDefinedConstants=NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CScriptProcessor::CScriptProcessor(CScriptParser* _parser) : CEAbstractSession()
{
	m_parser = _parser;
	m_map_locals = CEMap::createNewMap();

	m_line_brake_flags = 0;
	m_whole_value = NULL;
	m_curr_operand_value = NULL;
	m_last_func_values = NULL;
	m_line_priority = 0;
}

CScriptProcessor::~CScriptProcessor()
{
	delete m_map_locals;
	if (m_last_func_values != NULL) delete m_last_func_values;
}

//	***************************************************************
//	***                    static entries                       ***
//	***************************************************************
CScriptValue* CScriptProcessor::run_file_script(const char *file_name)
{
	printf("Script processor started (file: %s).\n", file_name);
	dbg->log(8, "reading commands from:"); dbg->log(8, file_name);

	// update function register (INIT)
	if (!m_init_OK) {init();}

	char* file_buffer = CAbstractFileParser::LoadFileToMemory(file_name);
	if (file_buffer == NULL) throw new CEException("Error reading file");
	CScriptParser* parser = new CScriptParser(file_buffer);
	CScriptValue* res = NULL;

	try {
		res = CScriptProcessor::run_script(parser);
	}
	catch(CEException exc) {
		bool is_error = exc.isErrorBreak();
		if (is_error) { printf("<Error> "); } else { printf("<Break> "); } printf(exc.getMessage()->getData());
		printf(" {Script line: %ld, Pos: %ld}\n", parser->get_curr_line_num(), parser->get_curr_line_pos()-1);
		if (!is_error) { // proper finalization for the non-error exit
			if (m_WorkingProc != NULL) { delete m_WorkingProc; m_WorkingProc=NULL; }
			delete parser; CMemoryManager::mfree(file_buffer);
		}
		throw exc; // Note: we throw on Halt() exit for closing all the sub-function calls.
	}
	catch(CEException* exc) {
		bool is_error = exc->isErrorBreak();
		if (is_error) { printf("<Error> "); } else { printf("<Break> "); } printf(exc->getMessage()->getData());
		printf(" {Script line: %ld, Pos: %ld}\n", parser->get_curr_line_num(), parser->get_curr_line_pos()-1);
		if (!is_error) { // proper finalization for the non-error exit
			if (m_WorkingProc != NULL) { delete m_WorkingProc; m_WorkingProc=NULL; }
			delete parser; CMemoryManager::mfree(file_buffer);
		}
		throw exc; // Note: we throw on Halt() exit for closing all the sub-function calls.
	}
	catch(...) {
		printf("Stopped with undefined exception\n");
		printf("Line: %ld, Pos: %ld.\n", parser->get_curr_line_num(), parser->get_curr_line_pos()-1);
		throw new CEException("Unknown script halt (internal error)");
	}

	delete parser;
	CMemoryManager::mfree(file_buffer);
	return res;
}

CScriptValue* CScriptProcessor::run_script(CScriptParser *prs)
{
	if (m_WorkingProc != NULL) { delete m_WorkingProc; m_WorkingProc=NULL; }
	CScriptProcessor* sp = new CScriptProcessor(prs);
	m_WorkingProc = sp;
	CScriptValue* res = sp->run_script_processor();
	m_WorkingProc = NULL;
	delete sp;
	return res;
}

//	***************************************************************
//	***                  script processing                      ***
//	***************************************************************
CScriptValue* CScriptProcessor::run_script_processor()
{
	// set predefined variables
	init_variables();

	// main cycle start
	for(;;) {
		// parse next line
		CScriptValue* line_val = run_script_line();
		if (line_val != NULL) delete line_val;
		if (m_line_brake_flags & LINE_BRAKE_EOF) break;

		// any other line break means error
		if (m_line_brake_flags != 0) {
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected syntax");
		}
	}
	// main cycle end
	return NULL;
}

/*
*	m_whole_value is the clone of string value (or NULL) -- to be deleted if useless
*
*	<old!> NOTE: m_whole_value refers the used object -- do not delete it
*	<old!> returns non-cloned object: clone before use!
*
*	Note: _whole_val is used to continue parsing of half-processed line.
*	In particular, if first operand in string is already processed,
*	one can decide to start the new string processor to continue parsing
*	_whole_val is deleted after use.
*	Used by for priority-tree calls.
*/
CScriptValue* CScriptProcessor::run_script_line(int line_priority, CScriptValue* _whole_val)
{
	dbg->log(20, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos(), "run_script_line(line %ld, pos: %ld)");
	//////////////////////////////////////////////////////// store environment BEGIN
	CScriptValue* old_whole_val = m_whole_value; m_whole_value = _whole_val;
	CScriptValue* old_curr_op_val = m_curr_operand_value; m_curr_operand_value = NULL;
	int old_LP = m_line_priority; m_line_priority = line_priority;
	m_parser->start_new_operands_list();
	m_line_brake_flags = 0; // volatile, to be checked after line is parsed
	//////////////////////////////////////////////////////// store environment END

	// reading all the operands in the string
	bool operand_loop = true;
	int op_cnt;
	for(op_cnt = 0;operand_loop; op_cnt++) {
		int op_type = m_parser->get_next_op_type();
		dbg->log(30, op_type, "reading next operand (type == %ld)");

		// check up type
		// 1. OP_INVALID: invalid operand
		if (op_type == CFileParserAdv::OP_INVALID) declare_script_error(CScriptErrors::ILLEGAL_OPERATION, -1, "Invalid operand spotted");

		// 2. OP_NO_OPERAND: how can occur?
		if (op_type == CFileParserAdv::OP_NO_OPERAND) declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "No next operand found");

		// 3. OP_END_OF_LINE
		if (op_type == CFileParserAdv::OP_END_OF_LINE) {
			int prev_type = m_parser->get_operand_type(-1);
			if (prev_type & CFileParserAdv::OP_OPERATION_ANY) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected END_OF_LINE");
			// do not skip line delimiter for priority sub-string parsing
			if (m_line_priority == 0) m_parser->skip_next_operand();
			break;
		}

		// 4.1. OP_BRACKET_RIGHT: bracket closing (end of substring)
		if ((op_type == CFileParserAdv::OP_BRACKET_RIGHT) || (op_type == CFileParserAdv::OP_BRACKET_RIGHT_2)) {
			m_line_brake_flags = m_line_brake_flags | LINE_BRAKE_RIGHT_BRACE;
			int prev_type = m_parser->get_operand_type(-1);
			if (prev_type & CFileParserAdv::OP_OPERATION_ANY) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected \")\" symbol");
			// do not skip line delimiter for priority sub-string parsing
			if (m_line_priority == 0) m_parser->skip_next_operand();
			break;
		}

		// 4.2. SYM_END_OF_BUFFER: end of file exceed
		if (op_type == CAbstractFileParser::SYM_END_OF_BUFFER) {
			dbg->log(20, "EOF encounted");
			m_line_brake_flags = m_line_brake_flags | LINE_BRAKE_EOF;
			if (op_cnt > 0) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected EOF (\";\" missed?)");
			// do not skip line delimiter for priority sub-string parsing
			if (m_line_priority == 0) m_parser->skip_next_operand();
			break;
		}

		// next processing requires operand string
		const char* c_op = m_parser->get_next_operand(false);
		dbg->log(30, c_op);
		if (c_op == NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);

		// 4.3. Lower priority spotted
		if ((m_line_priority > 0) && (getOperandPriority(c_op, op_type) <= m_line_priority)) {
			dbg->log(50, "Lower priority spotted");
			m_line_brake_flags = m_line_brake_flags | LINE_BRAKE_PRIORITY;
			int prev_type = m_parser->get_operand_type(-1);
			if (prev_type & CFileParserAdv::OP_OPERATION_ANY) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Unexpected lower priority spotted");
			break; // do not skip this operand
		}

		// all the cases above breakes the line parsing
		// all the operands are added to line operands list.

		// skip current operand
		m_parser->skip_next_operand(true); // 'c_op' is deleted if not added to skipped-list

		// 6. operations
		if ((op_type == CFileParserAdv::OP_OPERATION_LEFT) || (op_type == CFileParserAdv::OP_OPERATION_RIGHT)) {
			do_operation_involve(c_op, op_type); continue;
		}

		// self-test
		if (m_curr_operand_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);

		// Further TARGET: calculate the value of current operand & update string value

		// 7. left brace
		if (op_type == CFileParserAdv::OP_BRACKET_LEFT) {
			CScriptValue* brace_val = run_script_line();
			if ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0) {
				declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unxpected symbol spotted: \"(\"");
			} else if (brace_val == NULL) {
				declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "NULL object in brackets");
			}
			clear_line_break_flag(); // clear end-of-string flag

			m_curr_operand_value = brace_val;
		}
		// 8. VALUES
		else if (op_type & CFileParserAdv::OP_VALUE_ANY) {
			if (op_type  ==  CFileParserAdv::OP_VALUE_NUMBER) {
				long op_lval = atol(c_op);
				// check up conversion
				if (op_lval == 0) {
					bool op_num_err = false;
					int op_num_len = CEUtil::length(c_op);
					if (op_num_len <= 0) op_num_err = true;
					int op_num_cnt;
					for(op_num_cnt=0; op_num_cnt<op_num_len; op_num_cnt++) {
						if (*(c_op+op_num_cnt) != '0') op_num_err = true;
					}
					if (op_num_err) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Invalid number format");
				}
				m_curr_operand_value = CScriptValue::newValue(op_lval);
			}
			else if (op_type  ==  CFileParserAdv::OP_VALUE_STRING) {
				CEString* op_sval = new CEString(c_op);
				m_curr_operand_value = CScriptValue::newValue(op_sval, true);
			}
			else {
				// unknown value
				declare_script_error(CScriptErrors::UNDEFINED_OBJECT, -1, c_op);
			}
		}
		else {
			// supposed: op_type  == CFileParserAdv::OP_VARIABLE
			if (!(op_type == CFileParserAdv::OP_VARIABLE)) {
				// error: unexpected operand type
				declare_script_error(CScriptErrors::INTERNAL_ERROR);
			}
			
			// 9. Functions
			bool func_invoked = do_function_cast(c_op, op_type);
			
			// 10. Variables
			if (!func_invoked) do_variable_involve(c_op, op_type);
		}

		// m_curr_operand_value should NOT be NULL here

		// 11. Update value of the string (perform right-handed operations)
		update_string_value();
		if(m_curr_operand_value!=NULL)declare_script_error(CScriptErrors::INTERNAL_ERROR); // m_curr_operand_value should BE NULL here

		// appending current operand to line operands storage (?)
		/*if ((m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED) == 0) {
			m_parser->skip_next_operand();
			m_line_brake_flags = m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED;
		}*/
	}

	// cycle breaks before last m_parser->skip_next_operand() invoked
	// m_parser->skip_next_operand();
	//if ((m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED) == 0) {
	//	m_parser->skip_next_operand();
	//	m_line_brake_flags = m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED;
	//}

	// left operation executing
	lefthand_operation_execute();

	// return value composing
	CScriptValue* res = m_whole_value;
	
	//////////////////////////////////////////////////////// restore environment BEGIN
	m_curr_operand_value = old_curr_op_val;
	m_whole_value = old_whole_val;
	m_line_priority = old_LP;
	m_parser->finish_operands_list();
	//////////////////////////////////////////////////////// restore environment END
	return res;
}

void CScriptProcessor::clear_line_break_flag() {
	m_line_brake_flags = (m_line_brake_flags | LINE_BRAKE_ANY_FLAG) ^ LINE_BRAKE_ANY_FLAG;
}

//	***************************************************************
//	***                 operands processing                     ***
//	***************************************************************
void CScriptProcessor::do_operation_involve(const char* c_op, int op_type) {
	dbg->log(40, "do_operation_involve()");
	// check up operation existance
	if (op_type == CFileParserAdv::OP_OPERATION_LEFT) {
		// truncate whole value for left operation
		if (m_whole_value != NULL) {delete m_whole_value; m_whole_value = NULL;}
	}

	// check up previous operand: right-hand operation may not be followed by another operation
	int prev_offs = -2;
	int prev_type = m_parser->get_operand_type(prev_offs);
	if (prev_type == CFileParserAdv::OP_OPERATION_RIGHT) {
		declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Rigth-hand operation followed by another one");
	}

	
	// <old> check up left operand for binary operations. (?)
	// NOTE: operation processing is keeped in update_string_value() function
	// after the next operand is read out.
}

bool CScriptProcessor::do_function_cast(const char *c_op, int op_type)
{
	dbg->log(40, "do_function_cast()");
	if (!((op_type == CFileParserAdv::OP_FUNCTION_CALL) ||
		(op_type == CFileParserAdv::OP_VARIABLE))) {
		return false;
	}

	// find function in system functions
	// find function in function register
	CScriptFuncInfo* func_info = (CScriptFuncInfo*)m_functions_register->get(c_op);
	if (func_info == NULL) {// no function found
		if (op_type == CFileParserAdv::OP_FUNCTION_CALL) declare_script_error(CScriptErrors::UNKNOWN_FUNCTION);
		return false;
	}

	// function title check (may be omitted)
	if (!CEUtil::equals(c_op, func_info->GetFuncTitle())) {
		dbg->log(0, "Warning: function title differs from Function Info data (Internal Error)");
		dbg->log(0, c_op);
		dbg->log(0, "Function Info Object is", func_info);
	}

	bool SYS_func_invoke = func_info->isSystemFunction();

	CEPtrArray* arguments_list = new CEPtrArray();
	//m_parser->skip_next_operand();
	//m_line_brake_flags = m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED;
	int next_type = m_parser->get_next_op_type();
	if (next_type == CFileParserAdv::OP_BRACKET_LEFT) {
		// fill out arguments_list
		m_parser->skip_next_operand(false);
		int args_cnt = 0;
		for (;;) {
			// if no more lines exist --> break;
			CScriptValue* arg = run_script_line();

			// no only breaks allowed are 0 (; or ,) and LINE_BRAKE_RIGHT_BRACE
			if ((m_line_brake_flags & LINE_BRAKE_EOF) || (m_line_brake_flags & LINE_BRAKE_PRIORITY)) {
				declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "\")\" or \",\" are expected");
			}

			if (arg != NULL) {
				dbg->log(20, "arg", arg);
				arguments_list->add(arg);
				args_cnt++;
			} else {
				// this may be the void  Function() call.
				// elsewhere --> error.
				if ((args_cnt > 0) || ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0)) declare_script_error(CScriptErrors::NULL_OBJECT);
			}

			if (m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) {
				clear_line_break_flag();
				break;
			}
		}
	}
	else {
		// check up errors
		//if (next_type != CFileParserAdv::OP_END_OF_LINE) {
		//	declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Unexpected operand after function call");
		//}
		//m_parser->skip_next_operand();
	}
	// else try to ivoke parameterless function

	CEPtrArray* res_array = NULL;
	if (!SYS_func_invoke) {
		// clear previous function result
		if (m_last_func_values != NULL) { delete m_last_func_values; m_last_func_values = NULL; }
		// invoke
		res_array = invoke_function(func_info, arguments_list); // Warning: arguments_list is deleted therein, we must delete res_array
	} else {
		// invoke system function
		res_array = invoke_system_function(func_info, arguments_list); // Warning: arguments_list is deleted therein, we must delete res_array
	}

	CScriptValue* res = NULL;
	if (res_array != NULL) {
		int res_size = res_array->size();
		if (res_size > 0) {
			res = ((CScriptValue*)res_array->GetAt(0))->clone();

			if (res_size > 1) {
				// Error if res#2 is integer.
				// undefined (non-integer) value means no error have occured.
				CScriptValue* err_val = (CScriptValue*)res_array->GetAt(1);
				if (err_val->check_value_type(CScriptValue::VAL_INT)) {

					// parse error code
					int err_code = err_val->get_int_value();
					if ((err_code > 0) && (err_code & CScriptFuncInfo::ERROR_HALT_FLAG)) { // halt operation
						dbg->log(0, "Script processing halted");
						int halt_param = err_code - CScriptFuncInfo::ERROR_HALT_FLAG;
						declare_script_error(CScriptErrors::HALT_COMMAND, halt_param);
					}

					if (err_code > 0) {
						//c_op
						CEString sTmp("ERROR. "); sTmp.AddString(c_op); sTmp.AddString("() invalid argument #"); sTmp.add_int(err_code); dbg->log(0, sTmp.getData());
						declare_script_error(CScriptErrors::PARAMETER_INVALID, err_code, c_op);
					}
					else if (err_code < 0) {
						dbg->log(0, -err_code-1, "Invalid number of arguments: %ld");
						declare_script_error(CScriptErrors::PARAMETERS_NUMBER_INVALID, -err_code-1, c_op);
					}
					else dbg->log(0, "Error function execute.");
					declare_script_error(CScriptErrors::INTERNAL_ERROR);
				}
			}
		}

		// store function result array for non-system function
		if (!SYS_func_invoke) {
			if (m_last_func_values != NULL) { dbg->log(0, "Warning: old function array != NULL"); }
			m_last_func_values = res_array;
		} else {
			delete res_array;
		}
	}
	if (res == NULL) res = CScriptValue::newUndefinedValue();

	// result
	if (m_curr_operand_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
	m_curr_operand_value = res;

	dbg->log(40, "function invoked");
	return true;
}

/*
*	op_type is unused (should be CFileParserAdv::OP_VARIABLE)
*/
void CScriptProcessor::do_variable_involve(const char *c_op, int op_type)
{
	dbg->log(40, "do_variable_involve()");
	if (m_curr_operand_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR,-1,"m_curr_operand_value != NULL at CScriptProcessor::do_variable_involve()");
	// look for op in local variables list
	CScriptValue* var = (CScriptValue*)m_map_locals->get(c_op);
	if (var != NULL) {
		m_curr_operand_value = var->clone();
		dbg->log(105, "do_variable_involve(): m_curr_operand_value created", m_curr_operand_value);
		return;
	}

	// Next item check-up
	int next_type = m_parser->get_next_op_type(); const char* next_op = m_parser->get_next_operand(false);

	// * Is it an array element? Array may be ARR[i,i...][i,i...]...[i,..]
	// Note: array elements are allowed as "Predefined constants" only.
	CEPtrArray index_full; bool is_array=false;
	for(;;) {
		if (next_type != CFileParserAdv::OP_BRACKET_LEFT_2) break;
		is_array=true; CEPtrArray index_list; int index_cnt=0; m_parser->skip_next_operand(false);
		for (;;) {
			CScriptValue* ind = run_script_line();
			// the breaks allowed are LINE_BRAKE_RIGHT_BRACE and EOL,EOF (at index!=0)
			if ( ((m_line_brake_flags & LINE_BRAKE_EOF)==LINE_BRAKE_EOF) || ((m_line_brake_flags & LINE_BRAKE_PRIORITY)==LINE_BRAKE_PRIORITY) ) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "\"]\" or \",\" are expected"); }
			if (ind != NULL) { index_list.add(ind); index_cnt++; dbg->log(85, "index", ind); }
			else {
				// this may be the void  array[] call.
				if ((index_cnt > 0) || ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0)) declare_script_error(CScriptErrors::NULL_OBJECT, -1 , "Empty array index");
			}
			if (m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) { clear_line_break_flag(); break; }
		}
		int s = index_list.size(); CScriptValue* inds=NULL;
		if (s==1) {
			inds = ((CScriptValue*)index_list.GetAt(0))->clone();
		} else { // 0,2,3..
			CScriptValueArray* indsA=new CScriptValueArray();
			for(int i=0; i<s; i++) { indsA->addToArray((CScriptValue*)index_list.GetAt(i),false); }
			inds=indsA;
		}
		index_full.add(inds);
		// Nt: right bracked should already be skipped
		next_type = m_parser->get_next_op_type(); next_op = m_parser->get_next_operand(false);
	}
	if (is_array) {
		CScriptValue* inds=NULL; int s=index_full.size();
		if (s==1) {
			inds = ((CScriptValue*)index_full.GetAt(0))->clone();
		} else { // 0,2,3..
			CScriptValueArray* indsA=new CScriptValueArray();
			for(int i=0; i<s; i++) { indsA->addToArray((CScriptValue*)index_full.GetAt(i),false); }
			inds=indsA;
		}
		CScriptValue* base = CScriptValue::newValue(new CESymbol(c_op),true);
		//CScriptValueArrayElement res = new CScriptValueArrayElement(base, inds);
		CScriptValueArrayElement arr_elt(base, inds); delete base; delete inds;
		CExpression* arr_expr = arr_elt.TurnToExprCloned(); if(arr_expr==NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR, -1 , "Can't create array element value");
		// Nt: arr_expr is a CEFunction-class object
		dbg->log(85, "Array element spotted. Involved as CEFunction. Trying to resolve as predefined constant", arr_expr);
		//m_curr_operand_value = CScriptValue::newValue(arr_expr,true);
		CScriptValue* predef = CScriptProcessor::getPreDefinedConstant(arr_expr);
		if (predef!=NULL) {
			m_curr_operand_value=predef->clone();
			//dbg->log(0, m_curr_operand_value->get_int_value(),m_curr_operand_value->get_int_value(), "Predefined constant resolved: x%lx, %ld", arr_expr);
			dbg->log(85, "Predefined constant resolved:", m_curr_operand_value);
			delete arr_expr;
			return;
		}
		
		dbg->log(0, "Can't resolve array element value. No predefined constant fits", arr_expr);
		declare_script_error(CScriptErrors::INTERNAL_ERROR, -1 , "Array element is allowed for predifined values only");
		delete arr_expr;
		return;
	}

	//////////////////////////////
	// unknown variable


	// Old:
	// * Unknown variable may only occur at start of line, check it up
	// * lookup for previous operations
	// * NOTE: current operand becomes '-1' after CScriptParser::skip_next_operand() invoked
	/*
	int offs = -2;//((m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED)?(-2):(-1));
	int prev_type = m_parser->get_operand_type(offs);
	if (!((prev_type == CFileParserAdv::OP_NO_OPERAND) || (prev_type == CFileParserAdv::OP_END_OF_LINE))) {
		dbg->log(2, prev_type, "Error: prev type == %ld."); declare_script_error(CScriptErrors::UNDEFINED_OBJECT, -1, c_op);
	}
	// add new local variable
	// Old: value is undefined. UNDEFINED_EXPRESSION should be instance of unique class for safer work
	CScriptValue* varValue = CScriptValue::newUndefinedValue();
	m_map_locals->put(c_op, varValue->clone()); m_curr_operand_value=varValue;
	*/

	// add new local variable
	// Now: value is a symbol equal to variable notation
	CScriptValue* varValue = CScriptValue::newValue(new CESymbol(c_op),false);
	m_map_locals->put(c_op, varValue->clone()); m_curr_operand_value=varValue;
}

//	***************************************************************
//	***                 perform operations                      ***
//	***************************************************************
void CScriptProcessor::update_string_value()
{
	dbg->log(40, "update_string_value()");
	// executs last operation
	if (m_curr_operand_value == NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
	
	// lookup previous operations
	// NOTE: current operand becomes '-1' after CScriptParser::skip_next_operand() invoked
	//dbg->log(40, m_parser->get_operand_type(-1), "-1 = %ld");
	int offs = -2;//((m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED)?(-2):(-1));
	int prev_type = m_parser->get_operand_type(offs);
	dbg->log(40, prev_type, "prev_type = %ld");
	if ((prev_type == CFileParserAdv::OP_NO_OPERAND) || (prev_type == CFileParserAdv::OP_END_OF_LINE)) {
		if (m_whole_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
		m_whole_value = m_curr_operand_value; m_curr_operand_value = NULL;
		return;
	}

	if (prev_type == CFileParserAdv::OP_OPERATION_LEFT) {
		// left-hand operations are executed after the whole line value calculated
		// ('='-operation is left-hand operation)
		if (m_whole_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
		m_whole_value = m_curr_operand_value; m_curr_operand_value = NULL;
		return;
	}

	if (prev_type == CFileParserAdv::OP_OPERATION_RIGHT) {
		const char* op_operation = m_parser->get_operand(offs);

		// check next operation priority
		// read next operand and push it back to stream
		int next_type = m_parser->get_next_op_type();
		// bool push_back_next = true; // no push back required -- we have skipped nothing
		if (next_type == CFileParserAdv::OP_OPERATION_RIGHT) {
			const char* next_c_op = m_parser->get_next_operand(false);
			int op_prior = getOperandPriority(op_operation, prev_type);
			if (op_prior < getOperandPriority(next_c_op, next_type)) {
				// push back next operand
				// no push back required -- we have skipped nothing
				//m_parser->push_back_last_operand(); push_back_next = false;

				// start new string processor using current value as first & update current value
				CScriptValue* sub_val = run_script_line(op_prior, m_curr_operand_value); // m_curr_operand_value will be deleted
				// no matter how was the line breaked
				if (sub_val == NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "NULL result for sub-string");
				clear_line_break_flag(); // clear end-of-string flag
				dbg->log(90, "sub-value", sub_val);
				m_curr_operand_value = sub_val;
			}
		} else if (next_type == CFileParserAdv::OP_OPERATION_LEFT) {
			// error: how can left-hand operation follow the right-handed?
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Left-handed operation follows the after right one");
		}

		// push back next operand
		// no push back required -- we have skipped nothing
		//if (push_back_next) m_parser->push_back_last_operand();

		// check if there was a first operand
		// bool first_op_exist = (m_parser->get_operand_type(offs-1) == CFileParserAdv::OP_VARIABLE);
		// if it exists, there should be a legal WHOLE_VALUE
		// if it does not exist, there should be no WHOLE_VALUE
		// CHECK SKIPPED: the WHLOLE_VALUE shows if there was the previous operand.

		// if previous or current values are UNDEFINED declare error.
		if ( ((m_whole_value != NULL) &&
			(m_whole_value->value_type() & CScriptValue::VAL_UNDEFINED)) ||
			(m_curr_operand_value->value_type() & CScriptValue::VAL_UNDEFINED) ) {
			declare_script_error(CScriptErrors::UNDEFINED_OBJECT);
		}
		
		int op_len = CEUtil::length(op_operation);
		if (op_len != 1) { declare_script_error(CScriptErrors::ILLEGAL_OPERATION); m_curr_operand_value = NULL; return; }
		
		// all the operations heredown are available for integer values only
		bool is_int_first = ((m_whole_value != NULL) &&
			((m_whole_value->value_type() & CScriptValue::VAL_INT) != 0));
		bool is_int_second = ((m_curr_operand_value->value_type() & CScriptValue::VAL_INT) != 0);

		// old int-check
		//if ( ((m_whole_value != NULL) &&
		//	((m_whole_value->value_type() & CScriptValue::VAL_INT) == 0)) ||
		//	((m_curr_operand_value->value_type() & CScriptValue::VAL_INT) == 0) ) {
		//	declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
		//}

		dbg->log(105, "common right-handed operation point");
		dbg->log(175, "letf (m_whole_value)", m_whole_value);
		dbg->log(175, "reight (current)",m_curr_operand_value);
		/*
		*	Common policy: we need to update m_whole_value
		*	delete m_curr_operand_value and set m_curr_operand_value = NULL;
		*/
		if (((m_whole_value == NULL) || is_int_first) && (is_int_second)) {
			// perform integer operation
			int value_res = INT_MIN;
			if (m_whole_value != NULL) {
				value_res = int_operation_binary(op_operation,
					m_whole_value->get_int_value(),
					m_curr_operand_value->get_int_value());
				delete m_whole_value;  m_whole_value = NULL;
			} else {
				value_res = int_operation_unitary(op_operation,
					m_curr_operand_value->get_int_value());
			}
			delete m_curr_operand_value; m_curr_operand_value = NULL;
			if (value_res == INT_MIN) { dbg->log(2, "Error: INT_MIN integer value (No operation performed?)."); declare_script_error(CScriptErrors::INTERNAL_ERROR);}
			m_whole_value = CScriptValue::newValue(value_res);
			return;
		}

		bool is_expr_first = ((m_whole_value != NULL) &&
			((m_whole_value->value_type() & CScriptValue::VAL_EXPRESSION) != 0));
		bool is_expr_second = ((m_curr_operand_value->value_type() & CScriptValue::VAL_EXPRESSION) != 0);

		if ((is_expr_first) || (is_expr_second)) {
			// perform expression operation
			CExpression* value_res = NULL;
			// define operands
			CExpression* _expr_curr = NULL;
			bool del_expr_curr = false;
			if (is_expr_second) _expr_curr = m_curr_operand_value->get_expr_value();
			else if (is_int_second) { _expr_curr = new CENumber(m_curr_operand_value->get_int_value()); del_expr_curr = true; }
			else { dbg->log(0, "Error: invalid right-handed operation.", m_curr_operand_value); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
			if (m_whole_value != NULL) {
				// define operands
				CExpression* _expr_comm = NULL;
				bool del_expr_comm = false;
				if (is_expr_first) _expr_comm = m_whole_value->get_expr_value();
				else if (is_int_first) { _expr_comm = new CENumber(m_whole_value->get_int_value()); del_expr_comm = true; }
				else { dbg->log(0, "Error: invalid right-handed operation.", m_whole_value); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
				value_res = expr_operation_binary(op_operation, _expr_comm, _expr_curr);
				if (del_expr_comm) delete _expr_comm;
				delete m_whole_value; m_whole_value = NULL;
			} else {
				value_res = expr_operation_unitary(op_operation, _expr_curr);
			}
			if (del_expr_curr) delete _expr_curr;
			delete m_curr_operand_value; m_curr_operand_value = NULL;
			if (value_res == NULL) { dbg->log(0, "Error: NULL expression value (No operation performed?)."); declare_script_error(CScriptErrors::INTERNAL_ERROR);}
			m_whole_value = CScriptValue::newValue(value_res, true);
			return;
		}

		bool is_str_first = ((m_whole_value != NULL) &&
			((m_whole_value->value_type() & CScriptValue::VAL_STRING) != 0));
		bool is_str_second = ((m_curr_operand_value->value_type() & CScriptValue::VAL_STRING) != 0);

		// string + string
		if (((m_whole_value == NULL) || is_str_first) && (is_str_second)) {
			// perform string operation
			CEString* value_res = NULL;
			if (!is_str_second) { dbg->log(0, "Error: invalid right-handed string operation.", m_curr_operand_value); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
			CEString* _str_curr = m_curr_operand_value->get_str_value();
			if (m_whole_value != NULL) {
				if (!is_str_first) { dbg->log(0, "Error: invalid right-handed string operation.", m_whole_value); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
				CEString* _str_comm = m_whole_value->get_str_value();
				value_res = str_operation_binary(op_operation, _str_comm, _str_curr);
				delete m_whole_value; m_whole_value = NULL;
			} else {
				value_res = str_operation_unitary(op_operation, _str_curr);
			}
			delete m_curr_operand_value; m_curr_operand_value = NULL;
			if (value_res == NULL) { dbg->log(0, "Error: NULL expression value (No operation performed?)."); declare_script_error(CScriptErrors::INTERNAL_ERROR);}
			m_whole_value = CScriptValue::newValue(value_res, true);
			return;
		}

		// int+string and string+int operations
		if (((is_str_first) && (is_int_second)) || ((is_int_first) && (is_str_second))) {
			CEString* value_res = NULL;
			if (is_str_first) {
				value_res=strint_operation_binary(op_operation, m_whole_value->get_str_value(), m_curr_operand_value->get_int_value());
			} else { // is_str_second
				value_res=intstr_operation_binary(op_operation, m_whole_value->get_int_value(), m_curr_operand_value->get_str_value());
			}
			delete m_whole_value; m_whole_value = NULL;
			delete m_curr_operand_value; m_curr_operand_value = NULL;
			if (value_res == NULL) { dbg->log(0, "Error: NULL operation result (No operation performed?)."); declare_script_error(CScriptErrors::INTERNAL_ERROR);}
			m_whole_value = CScriptValue::newValue(value_res, true);
			return;
		}
	}
	delete m_curr_operand_value; m_curr_operand_value = NULL;
	declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Can't calculate operand value");
}

void CScriptProcessor::lefthand_operation_execute()
{
	dbg->log(40, "lefthand_operation_execute()");
	int op_offset = m_parser->get_left_operation_offset(true);
	if (op_offset < 0) return;
	if (op_offset != 1) declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	
	const char* op_sym = m_parser->get_operand(1, true);
	
	if (CEUtil::equals("=", op_sym)) {
		int first_op_type = m_parser->get_operand_type(0, true);
		if (!(first_op_type == CFileParserAdv::OP_VARIABLE)) declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
		const char* first_op = m_parser->get_operand(0, true);
		
		// '=' operation
		if (m_whole_value == NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
		m_map_locals->put(first_op, m_whole_value);
		m_whole_value = NULL;
	}
	else declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
}

/******************************************************************
*	Implementation of right-hand numerical operations.
******************************************************************/
int CScriptProcessor::int_operation_binary(const char* c_op, int _whole, int _current) {
	int res = INT_MIN;

	if (*c_op == '+') {
		res = _whole + _current;
	} else if (*c_op == '-') {
		res = _whole - _current;
	} else if (*c_op == '|') {
		res = _whole | _current;
	} else if (*c_op == '*') {
		res = _whole * _current;
	} else if (*c_op == '/') {
		res = _whole / _current;
	} else {
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	}

	return res;
}

int CScriptProcessor::int_operation_unitary(const char* c_op, int _current) {
	int res = INT_MIN;

	if (*c_op == '+') {
		res = _current;
	} else if (*c_op == '-') {
		res = -_current;
	} else if (*c_op == '|') {
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	} else {
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	}

	return res;
}

// does not affect input expressions
CExpression* CScriptProcessor::expr_operation_binary(const char* c_op, CExpression* _whole, CExpression* _current) {
	CExpression* res = NULL;
	if ((_whole == NULL) || (_current == NULL)) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }

	if (*c_op == '+') {
		res = _whole->clone()->add(_current->clone());
	} else if (*c_op == '-') {
		res = _whole->clone()->minus(_current->clone());
	} else if (*c_op == '*') {
		res = _whole->clone()->multiply(_current->clone());
	} else if (*c_op == '/') {
		res = _whole->clone()->divide(_current->clone());
	} else {
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	}

	return res;
}

// does not affect input expression
CExpression* CScriptProcessor::expr_operation_unitary(const char* c_op, CExpression* _current) {
	CExpression* res = NULL;
	if (_current == NULL) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }

	if (*c_op == '+') {
		res = _current->clone();
	} else if (*c_op == '-') {
		res = _current->clone(); res->inverseFlag();
	} else {
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	}

	return res;
}

CEString* CScriptProcessor::str_operation_binary(const char* c_op, CEString* _whole, CEString* _current) { // does not affect input
	CEString* res = NULL;
	if ((_whole == NULL) || (_current == NULL)) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }

	if (*c_op == '+') {
		res = _whole->clone();
		res->AddString(_current, false);
	} else {
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	}

	return res;
}

CEString* CScriptProcessor::str_operation_unitary(const char* c_op, CEString* _current) { // does not affect input
	CEString* res = NULL;
	if (_current == NULL) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
	
	declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
	return res;
}

CEString* CScriptProcessor::strint_operation_binary(const char* c_op, CEString* _str, int _val) { // does not affect input
	CEString* res = NULL;
	if (_str == NULL) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
	if (*c_op == '+') { res=_str->clone(); res->add_int(_val); }
	else { declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
	return res;
}

CEString* CScriptProcessor::intstr_operation_binary(const char* c_op, int _val, CEString* _str) { // does not affect input
	CEString* res = NULL;
	if (_str == NULL) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
	if (*c_op == '+') { res=new CEString(""); res->add_int(_val); res->AddString(_str, false); }
	else { declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
	return res;
}

/******************************************************************
*	Operations priorities.
******************************************************************/
int CScriptProcessor::getOperandPriority(const char* c_op, int op_type) {
	int res = 100;

	if (op_type == CFileParserAdv::OP_OPERATION_RIGHT) {
		if ((*c_op == '+') || (*c_op == '-')) {
			res = 1;
		} else if (*c_op == '|') {
			res = 2;
		} else if ((*c_op == '*') || (*c_op == '/')) {
			res = 3;
		} else {
			dbg->log(0, "Error: right-hand operation with undeclared priority:");
			dbg->log(0, c_op);
		}
	}

	return res;
}

/*****************************************************************/

//	***************************************************************
//	***               init\finalize processor                   ***
//	***************************************************************
bool CScriptProcessor::init()
{
	if (m_init_OK) return true;
	if (m_PreDefinedConstants==NULL) m_PreDefinedConstants=CEMap::createNewMap();

	// program timer init
	CScriptProcessor::dbg->reset_periods(0);

	if (m_functions_register == NULL) m_functions_register = CEMap::createNewMap();
	//if (m_map_operations == NULL) m_map_operations = CEMap::createNewMap();

	m_functions_register->put("SetKinematics", CScriptFuncInfo::createFuncInfo("SetKinematics",
		(void*)CScriptFunctionLib::SetKinematics, 2, 2, CScriptValue::VAL_INT, CScriptValue::VAL_INT));
	m_functions_register->put("Expand", CScriptFuncInfo::createFuncInfo("Expand",
		(void*)CScriptFunctionLib::SFuncExpand, 2, 1, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("Arrange", CScriptFuncInfo::createFuncInfo("Arrange",
		(void*)CScriptFunctionLib::SFuncArrange, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Load", CScriptFuncInfo::createFuncInfo("Load",
		(void*)CScriptFunctionLib::SFuncLoadExprFromFile, 1, 1, CScriptValue::VAL_STRING));
	m_functions_register->put("Save", CScriptFuncInfo::createFuncInfo("Save",
		(void*)CScriptFunctionLib::SFuncStoreValueToFile, 2, 2,
		CScriptValue::VAL_STRING,
		CScriptValue::VAL_EXPRESSION | CScriptValue::VAL_MATRIX_ELEMENT | CScriptValue::VAL_DIAGRAM_SET | CScriptValue::VAL_INT));
	m_functions_register->put("SaveNB", CScriptFuncInfo::createFuncInfo("SaveNB",
		(void*)CScriptFunctionLib::SFuncStoreValueToNB, 4, 2,
		CScriptValue::VAL_STRING,
		CScriptValue::VAL_EXPRESSION | CScriptValue::VAL_MATRIX_ELEMENT | CScriptValue::VAL_INT,
		CScriptValue::VAL_STRING | CScriptValue::VAL_INT,
		CScriptValue::VAL_INT));
	m_functions_register->put("SaveM", CScriptFuncInfo::createFuncInfo("SaveM",
		(void*)CScriptFunctionLib::SFuncStoreValueToM, 2, 2,
		CScriptValue::VAL_STRING, CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_MATRIX_ELEMENT|CScriptValue::VAL_INT));
	m_functions_register->put("MarkNB", CScriptFuncInfo::createFuncInfo("MarkNB",
		(void*)CScriptFunctionLib::SFuncStoreTextToNB, 3, 2,
		CScriptValue::VAL_STRING,
		CScriptValue::VAL_STRING | CScriptValue::VAL_INT,
		CScriptValue::VAL_INT));
	m_functions_register->put("MarkTeX", CScriptFuncInfo::createFuncInfo("MarkTeX",
		(void*)CScriptFunctionLib::SFuncStoreTextToTeX, 3, 2,
		CScriptValue::VAL_STRING,
		CScriptValue::VAL_STRING | CScriptValue::VAL_INT,
		CScriptValue::VAL_INT));

	m_functions_register->put("ComposeDiagramsB", CScriptFuncInfo::createFuncInfo("ComposeDiagramsB",
		(void*)CScriptFunctionLib::SFuncComposeDiagrams, 2, 1,
		CScriptValue::VAL_INT, CScriptValue::VAL_INT));
	m_functions_register->put("DrawDiagrams", CScriptFuncInfo::createFuncInfo("DrawDiagrams",
		(void*)CScriptFunctionLib::SFuncDrawDiagrams, 4, 2,
		CScriptValue::VAL_DIAGRAM_SET,
		CScriptValue::VAL_STRING,
		CScriptValue::VAL_INT, CScriptValue::VAL_INT));
	m_functions_register->put("RetrieveME", CScriptFuncInfo::createFuncInfo("RetrieveME",
		(void*)CScriptFunctionLib::SFuncRetrieveME, 1, 1, CScriptValue::VAL_DIAGRAM_SET));
	m_functions_register->put("ComposeDiagrams", CScriptFuncInfo::createFuncInfo("ComposeDiagrams",
		(void*)CScriptFunctionLib::SFuncComposeDiagsWick, 1, 1, CScriptValue::VAL_INT));
	m_functions_register->put("CreateME", CScriptFuncInfo::createFuncInfo("CreateME",
		(void*)CScriptFunctionLib::SFuncCreateME, 1, 1, CScriptValue::VAL_INT));
	m_functions_register->put("SquareME", CScriptFuncInfo::createFuncInfo("SquareME",
		(void*)CScriptFunctionLib::SFuncSquareME, 3, 1,
		CScriptValue::VAL_MATRIX_ELEMENT,
		CScriptValue::VAL_MATRIX_ELEMENT | CScriptValue::VAL_INT,
		CScriptValue::VAL_INT));
	m_functions_register->put("CalcAmplitude", CScriptFuncInfo::createFuncInfo("CalcAmplitude",
		(void*)CScriptFunctionLib::SFuncCalcAmplitude, 2, 1,
		CScriptValue::VAL_MATRIX_ELEMENT, CScriptValue::VAL_INT));
	m_functions_register->put("ToDiffCS", CScriptFuncInfo::createFuncInfo("ToDiffCS",
		(void*)CScriptFunctionLib::SFuncToDiffCS, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("LoadModel", CScriptFuncInfo::createFuncInfo("LoadModel",
		(void*)CScriptFunctionLib::SFuncLoadModel, 2, 1, CScriptValue::VAL_STRING, CScriptValue::VAL_STRING));
	m_functions_register->put("GetSoftBremsstrahlung", CScriptFuncInfo::createFuncInfo("GetSoftBremsstrahlung",
		(void*)CScriptFunctionLib::SFuncGetSoftBremsstrahlung, 1, 0, CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_INT));
	m_functions_register->put("Run", CScriptFuncInfo::createFuncInfo("Run",
		(void*)CScriptFunctionLib::SFuncRun, 3, 1, CScriptValue::VAL_STRING, CScriptValue::VAL_INT, CScriptValue::VAL_STRING));

	m_functions_register->put("SetParameter", CScriptFuncInfo::createFuncInfo("SetParameter",
		(void*)CScriptFunctionLib::SFuncSetParameter, 2, 2, CScriptValue::VAL_STRING, CScriptValue::VAL_INT));
	m_functions_register->put("SetParameterS", CScriptFuncInfo::createFuncInfo("SetParameterS",
		(void*)CScriptFunctionLib::SFuncSetParameterS, 2, 2, CScriptValue::VAL_STRING, CScriptValue::VAL_STRING));
	m_functions_register->put("SetModelParameter", CScriptFuncInfo::createFuncInfo("SetModelParameter",
		(void*)CScriptFunctionLib::SFuncSetModelParameter, 2, 2, CScriptValue::VAL_STRING, CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_INT));
	m_functions_register->put("SetMassCalcOrder", CScriptFuncInfo::createFuncInfo("SetMassCalcOrder",
		(void*)CScriptFunctionLib::SFuncSetMassCalcOrder, 2, 2,
		CScriptValue::VAL_INT, CScriptValue::VAL_INT));
	m_functions_register->put("SelectDiagrams", CScriptFuncInfo::createFuncInfo("SelectDiagrams",
		(void*)CScriptFunctionLib::SFuncSelectDiagrams, 2, 2,
		CScriptValue::VAL_DIAGRAM_SET, CScriptValue::VAL_INT|CScriptValue::VAL_STRING)); /*String may be: Range(1,N)*/
	m_functions_register->put("CollectDiffDiagrams", CScriptFuncInfo::createFuncInfo("CollectDiffDiagrams",
		(void*)CScriptFunctionLib::SFuncCollectDiffDiagrams, 2, 2, CScriptValue::VAL_DIAGRAM_SET, CScriptValue::VAL_DIAGRAM_SET));
	m_functions_register->put("CollectEqualDiagrams", CScriptFuncInfo::createFuncInfo("CollectEqualDiagrams",
		(void*)CScriptFunctionLib::SFuncCollectEqualDiagrams, 2, 2, CScriptValue::VAL_DIAGRAM_SET, CScriptValue::VAL_DIAGRAM_SET));
	m_functions_register->put("SetPhotonHelicity", CScriptFuncInfo::createFuncInfo("SetPhotonHelicity",
		(void*)CScriptFunctionLib::SFuncSetPhotonHelicity, 4, 2,
		CScriptValue::VAL_INT, CScriptValue::VAL_INT,
		CScriptValue::VAL_STRING, CScriptValue::VAL_STRING));
	m_functions_register->put("SetPhotonDMBase", CScriptFuncInfo::createFuncInfo("SetPhotonDMBase",
		(void*)CScriptFunctionLib::SFuncSetPhotonDMBase, 2, 1,
		CScriptValue::VAL_INT, CScriptValue::VAL_STRING));
	m_functions_register->put("SetStoksParams", CScriptFuncInfo::createFuncInfo("SetStoksParams",
		(void*)CScriptFunctionLib::SFuncSetStoksParams, 3, 3,
		CScriptValue::VAL_INT, CScriptValue::VAL_STRING, CScriptValue::VAL_STRING));
	m_functions_register->put("SetFermionHelicity", CScriptFuncInfo::createFuncInfo("SetFermionHelicity",
		(void*)CScriptFunctionLib::SFuncSetFermionHelicity, 2, 2,
		CScriptValue::VAL_INT, CScriptValue::VAL_INT | CScriptValue::VAL_STRING));
	m_functions_register->put("SetPolarized", CScriptFuncInfo::createFuncInfo("SetPolarized",
		(void*)CScriptFunctionLib::SFuncSetPolarized, 2, 1,
		CScriptValue::VAL_INT, CScriptValue::VAL_INT));
	m_functions_register->put("SetMassSym", CScriptFuncInfo::createFuncInfo("SetMassSym",
		(void*)CScriptFunctionLib::SFuncSetMassSym, 2, 2, CScriptValue::VAL_INT, CScriptValue::VAL_STRING|CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("SetHelicityExpr", CScriptFuncInfo::createFuncInfo("SetHelicityExpr",
		(void*)CScriptFunctionLib::SFuncSetHelicityExpr, 4, 4,
		CScriptValue::VAL_INT,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("SetBondarevBaseSyms", CScriptFuncInfo::createFuncInfo("SetBondarevBaseSyms",
		(void*)CScriptFunctionLib::SFuncSetBondarevBaseSyms, 4, 4,
		CScriptValue::VAL_STRING|CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_STRING|CScriptValue::VAL_EXPRESSION,
		CScriptValue::VAL_STRING|CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_STRING|CScriptValue::VAL_EXPRESSION));

	// NewFortranFile: deprecated
	m_functions_register->put("NewFortranFile", CScriptFuncInfo::createFuncInfo("NewFortranFile",
		(void*)CScriptFunctionLib::SFuncNewFortranFile, 2, 1, CScriptValue::VAL_STRING, CScriptValue::VAL_INT));
	// CreateFortranProc: deprecated
	m_functions_register->put("CreateFortranProc", CScriptFuncInfo::createFuncInfo("CreateFortranProc",
		(void*)CScriptFunctionLib::SFuncCreateFortranProc, 4, 3, CScriptValue::VAL_INT, CScriptValue::VAL_STRING,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("NewCodeFile", CScriptFuncInfo::createFuncInfo("NewCodeFile",
		(void*)CScriptFunctionLib::SFuncNewCodeFile, 2, 1, CScriptValue::VAL_STRING, CScriptValue::VAL_INT));
	m_functions_register->put("CreateCodeProc", CScriptFuncInfo::createFuncInfo("CreateCodeProc",
		(void*)CScriptFunctionLib::SFuncCreateCodeProc, 4, 3, CScriptValue::VAL_INT, CScriptValue::VAL_STRING,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("Minimize", CScriptFuncInfo::createFuncInfo("Minimize",
		(void*)CScriptFunctionLib::SFuncMinimize, 2, 1,
		CScriptValue::VAL_EXPRESSION,
		CScriptValue::VAL_INT));
	m_functions_register->put("Factor", CScriptFuncInfo::createFuncInfo("Factor",
		(void*)CScriptFunctionLib::SFuncFactor, 2, 1,
		CScriptValue::VAL_EXPRESSION,
		CScriptValue::VAL_INT));
	m_functions_register->put("Equals", CScriptFuncInfo::createFuncInfo("Equals",
		(void*)CScriptFunctionLib::SFuncEquals, 3, 2,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("Evaluate", CScriptFuncInfo::createFuncInfo("Evaluate",
		(void*)CScriptFunctionLib::SFuncEvaluate, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("KinArrange", CScriptFuncInfo::createFuncInfo("KinArrange",
		(void*)CScriptFunctionLib::SFuncKinArrange, 2, 1,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("KinSimplify", CScriptFuncInfo::createFuncInfo("KinSimplify",
		(void*)CScriptFunctionLib::SFuncKinSimplify, 2, 1,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("KinFullSimplify", CScriptFuncInfo::createFuncInfo("KinFullSimplify",
		(void*)CScriptFunctionLib::SFuncKinFullSimplify, 2, 1,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("SingularArrange", CScriptFuncInfo::createFuncInfo("SingularArrange",
		(void*)CScriptFunctionLib::SFuncSingularArrange, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Simplify", CScriptFuncInfo::createFuncInfo("Simplify",
		(void*)CScriptFunctionLib::SFuncSimplify, 4, 3,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));

	m_functions_register->put("SetDiagramPhysics", CScriptFuncInfo::createFuncInfo("SetDiagramPhysics",
		(void*)CScriptFunctionLib::SFuncSetDiagramPhysics, 1, 1, CScriptValue::VAL_INT));
	m_functions_register->put("SetDRMassPK", CScriptFuncInfo::createFuncInfo("SetDRMassPK",
		(void*)CScriptFunctionLib::SFuncSetDRMassPK, 1, 1, CScriptValue::VAL_INT));
	m_functions_register->put("SetNDimensionSpace", CScriptFuncInfo::createFuncInfo("SetNDimensionSpace",
		(void*)CScriptFunctionLib::SFuncSetNDimensionSpace, 1, 0, CScriptValue::VAL_INT));
	m_functions_register->put("ReduceTensorVI", CScriptFuncInfo::createFuncInfo("ReduceTensorVI",
		(void*)CScriptFunctionLib::SFuncReduceTensorVI, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("ConvertInvariantVI", CScriptFuncInfo::createFuncInfo("ConvertInvariantVI",
		(void*)CScriptFunctionLib::SFuncConvertInvariantVI, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("CalcScalarVI", CScriptFuncInfo::createFuncInfo("CalcScalarVI",
		(void*)CScriptFunctionLib::SFuncCalcScalarVI, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("RestoreRenormConstants", CScriptFuncInfo::createFuncInfo("RestoreRenormConstants",
		(void*)CScriptFunctionLib::SFuncRestoreRenormConstants, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("GetMassSymPK", CScriptFuncInfo::createFuncInfo("GetMassSymPK",
		(void*)CScriptFunctionLib::SFuncGetMassSymPK, 1, 1, CScriptValue::VAL_INT));

	m_functions_register->put("SetVirtIntegralValues", CScriptFuncInfo::createFuncInfo("SetVirtIntegralValues",
		(void*)CScriptFunctionLib::SFuncSetVirtIntegralValues, 0, 0));
	m_functions_register->put("DumpDataStorage", CScriptFuncInfo::createFuncInfo("DumpDataStorage",
		(void*)CScriptFunctionLib::SFuncDumpDataStorage, 0, 0));

	m_functions_register->put("SubstVecZero", CScriptFuncInfo::createFuncInfo("SubstVecZero",
		(void*)CScriptFunctionLib::SFuncSubstVecZero, 2, 2, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_STRING));
	m_functions_register->put("SubstKinVector", CScriptFuncInfo::createFuncInfo("SubstKinVector",
		(void*)CScriptFunctionLib::SFuncSubstKinVector, 2, 2, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_STRING));
	m_functions_register->put("GetUVTerm", CScriptFuncInfo::createFuncInfo("GetUVTerm",
		(void*)CScriptFunctionLib::SFuncGetUVTerm, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Range", CScriptFuncInfo::createFuncInfo("Range",
		(void*)CScriptFunctionLib::SFuncRange, 2, 2, CScriptValue::VAL_INT, CScriptValue::VAL_INT));

	m_functions_register->put("Sym", CScriptFuncInfo::createFuncInfo("Sym",
		(void*)CScriptFunctionLib::SFuncSym, 1, 1, CScriptValue::VAL_STRING));
	m_functions_register->put("Num", CScriptFuncInfo::createFuncInfo("Num",
		(void*)CScriptFunctionLib::SFuncNum, 2, 1, CScriptValue::VAL_INT, CScriptValue::VAL_INT));
	m_functions_register->put("ImU", CScriptFuncInfo::createFuncInfo("ImU",
		(void*)CScriptFunctionLib::SFuncImU, 0, 0));
	m_functions_register->put("Infinity", CScriptFuncInfo::createFuncInfo("Infinity", (void*)CScriptFunctionLib::SFuncInfinity, 0, 0));
	m_functions_register->put("GetUnique", CScriptFuncInfo::createFuncInfo("GetUnique",
		(void*)CScriptFunctionLib::SFuncGetUnique, 0, 0));
	m_functions_register->put("Vec", CScriptFuncInfo::createFuncInfo("Vec",
		(void*)CScriptFunctionLib::SFuncVec, 3, 3,
		CScriptValue::VAL_STRING|CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("Eps", CScriptFuncInfo::createFuncInfo("Eps",
		(void*)CScriptFunctionLib::SFuncEps, 8, 8,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("Sqrt", CScriptFuncInfo::createFuncInfo("Sqrt",
		(void*)CScriptFunctionLib::SFuncSqrt, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Sqr", CScriptFuncInfo::createFuncInfo("Sqr",
		(void*)CScriptFunctionLib::SFuncSqr, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Spr", CScriptFuncInfo::createFuncInfo("Spr",
		(void*)CScriptFunctionLib::SFuncSpr, 2, 2, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Power", CScriptFuncInfo::createFuncInfo("Power",
		(void*)CScriptFunctionLib::SFuncPower, 2, 2, CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_INT));
	m_functions_register->put("Replace", CScriptFuncInfo::createFuncInfo("Replace",
		(void*)CScriptFunctionLib::SFuncReplace, 3, 3, CScriptValue::VAL_EXPRESSION,
		CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_STRING, CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_STRING|CScriptValue::VAL_INT));
	m_functions_register->put("SubstVec", CScriptFuncInfo::createFuncInfo("SubstVec",
		(void*)CScriptFunctionLib::SFuncSubstVec, 4, 4,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Frac", CScriptFuncInfo::createFuncInfo("Frac",
		(void*)CScriptFunctionLib::SFuncFrac, 2, 2,
		CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_INT, CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_INT));
	m_functions_register->put("Gamma", CScriptFuncInfo::createFuncInfo("Gamma",
		(void*)CScriptFunctionLib::SFuncGamma, 2, 1,
		CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_STRING|CScriptValue::VAL_INT, CScriptValue::VAL_INT));
	m_functions_register->put("GellMann", CScriptFuncInfo::createFuncInfo("GellMann",
		(void*)CScriptFunctionLib::SFuncGellMann, 1, 1, CScriptValue::VAL_EXPRESSION|CScriptValue::VAL_STRING));
	m_functions_register->put("ToAsymmProd", CScriptFuncInfo::createFuncInfo("ToAsymmProd",
		(void*)CScriptFunctionLib::SFuncToAsymmProd, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("Tr", CScriptFuncInfo::createFuncInfo("Tr",
		(void*)CScriptFunctionLib::SFuncTr, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("TrGellMann", CScriptFuncInfo::createFuncInfo("TrGellMann",
		(void*)CScriptFunctionLib::SFuncTrGellMann, 1, 1, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("FierzTransform", CScriptFuncInfo::createFuncInfo("FierzTransform",
		(void*)CScriptFunctionLib::SFuncFiertzTransform, 1, 1, CScriptValue::VAL_EXPRESSION));
	// remain old for some time
	m_functions_register->put("FiertzTransform", CScriptFuncInfo::createFuncInfo("FiertzTransform",
		(void*)CScriptFunctionLib::SFuncFiertzTransform, 1, 1, CScriptValue::VAL_EXPRESSION));

	// macros
	m_functions_register->put("Initial", CScriptFuncInfo::createFuncInfo("Initial",
		(void*)CScriptFunctionLib::SFuncInitial, 1, 1, CScriptValue::VAL_INT));
	m_functions_register->put("Final", CScriptFuncInfo::createFuncInfo("Final",
		(void*)CScriptFunctionLib::SFuncFinal, 1, 1, CScriptValue::VAL_INT));

	// view functions
	m_functions_register->put("ViewFeynmanRules", CScriptFuncInfo::createFuncInfo("ViewFeynmanRules",
		(void*)CScriptFunctionLib::SFuncViewFeynmanRules, 2, 0, CScriptValue::VAL_STRING, CScriptValue::VAL_INT));
	m_functions_register->put("ViewTensorVITable", CScriptFuncInfo::createFuncInfo("ViewTensorVITable",
		(void*)CScriptFunctionLib::SFuncViewTensorVITable, 2, 0, CScriptValue::VAL_STRING, CScriptValue::VAL_INT));
	m_functions_register->put("ViewScalarVICache", CScriptFuncInfo::createFuncInfo("ViewScalarVICache",
		(void*)CScriptFunctionLib::SFuncViewScalarVICache, 2, 0, CScriptValue::VAL_STRING, CScriptValue::VAL_INT));
	m_functions_register->put("ViewConstants", CScriptFuncInfo::createFuncInfo("ViewConstants",
		(void*)CScriptFunctionLib::SFuncViewConstants, 0, 0));

	m_functions_register->put("Log", CScriptFuncInfo::createFuncInfo("Log",
		(void*)CScriptFunctionLib::SFuncLog, 0, 0));
	m_functions_register->put("ViewParticleData", CScriptFuncInfo::createFuncInfo("ViewParticleData",
		(void*)CScriptFunctionLib::SFuncViewParticleData, 1, 1, CScriptValue::VAL_INT));
	m_functions_register->put("Halt", CScriptFuncInfo::createFuncInfo("Halt",
		(void*)CScriptFunctionLib::SFuncHalt, 1, 0, CScriptValue::VAL_INT));
	m_functions_register->put("Timer", CScriptFuncInfo::createFuncInfo("Timer",
		(void*)CScriptFunctionLib::SFuncTimer, 0, 0));
	m_functions_register->put("UseCharBank", CScriptFuncInfo::createFuncInfo("UseCharBank",
		(void*)CScriptFunctionLib::SFuncUseCharBank, 1, 0, CScriptValue::VAL_INT));
	m_functions_register->put("ChangeDebugLevel", CScriptFuncInfo::createFuncInfo("ChangeDebugLevel",
		(void*)CScriptFunctionLib::SFuncChangeDebugLevel, 2, 2,
		CScriptValue::VAL_STRING, CScriptValue::VAL_INT));

	//debug
	m_functions_register->put("DbgExtractVISet", CScriptFuncInfo::createFuncInfo("DbgExtractVISet",
		(void*)CScriptFunctionLib::SFuncDbgExtractVISet, 1, 1, CScriptValue::VAL_EXPRESSION));

	//test functions
	m_functions_register->put("TestCalc", CScriptFuncInfo::createFuncInfo("TestCalc",
		(void*)CScriptFunctionLib::SFuncTestCalc, 2, 0,
		CScriptValue::VAL_EXPRESSION, CScriptValue::VAL_EXPRESSION));
	m_functions_register->put("TestCalcB", CScriptFuncInfo::createFuncInfo("TestCalcB",
		(void*)CScriptFunctionLib::SFuncTestCalcB, 0, 0));

	// system functions
	m_functions_register->put("GetLastFuncValue", CScriptFuncInfo::createFuncInfo("GetLastFuncValue",
		NULL, 1, 1, CScriptValue::VAL_INT));

	m_init_OK = true;
	dbg->log(10, "CScriptProcessor initialized successfully.");
	return true;
}

void CScriptProcessor::addPreDefinedConstant(CExpression* src, long value) {
	if (m_PreDefinedConstants==NULL) { dbg->log(0,"INTERNAL ERROR: Can't add predefined script constant. Script Processor is not inited"); return; }
	if (src==NULL) return;
	CEString* s=src->toString();
	if(s==NULL) { dbg->log(0,"INTERNAL ERROR: Can't add predefined script constant. src->toString()==NULL"); return; }
	m_PreDefinedConstants->put(s->getData(), CScriptValue::newValue(value));
	// test map
	//dbg->log(0,value,getPreDefinedConstant(src)->get_int_value(),"addPreDefinedConstant: x%lx -> x%lx", src);
	delete s;
}

CScriptValue* CScriptProcessor::getPreDefinedConstant(CScriptValue* src) {
	if (m_PreDefinedConstants==NULL) { dbg->log(0,"INTERNAL ERROR: Can't find predefined script constant. Script Processor is not inited"); return NULL; }
	if (src==NULL)return NULL;
	CScriptValue* res=NULL;
	CExpression* srcE=src->get_expr_value();
	if (srcE!=NULL) res=(CScriptValue*)m_PreDefinedConstants->get(srcE);
	if (res == NULL) {
		CEString* srcS=src->get_str_value();
		if (srcS!=NULL) res=(CScriptValue*)m_PreDefinedConstants->get(srcS->getData());
	}
	return res;
}

CScriptValue* CScriptProcessor::getPreDefinedConstant(CExpression* src) {
	if (m_PreDefinedConstants==NULL) { dbg->log(0,"INTERNAL ERROR: Can't find predefined script constant. Script Processor is not inited"); return NULL; }
	if (src==NULL)return NULL;
	return (CScriptValue*)m_PreDefinedConstants->get(src);
}

/*
*	Predefined variables initialization
*/
void CScriptProcessor::init_variables() {
	if (m_map_locals == NULL) return;

	m_map_locals->put("ELECTRON", CScriptValue::newValue(CEParticle::PARTICLE_ELECTRON));
	m_map_locals->put("MUON", CScriptValue::newValue(CEParticle::PARTICLE_MUON));
	m_map_locals->put("TAULEPTON", CScriptValue::newValue(CEParticle::PARTICLE_TAU_LEPTON));
	m_map_locals->put("PHOTON", CScriptValue::newValue(CEParticle::PARTICLE_PHOTON));
	//m_map_locals->put("ZBOZON", CScriptValue::newValue(CEParticle::PARTICLE_Z_BOSON));
	m_map_locals->put("ZBOSON", CScriptValue::newValue(CEParticle::PARTICLE_Z_BOSON));
	//m_map_locals->put("WBOZON", CScriptValue::newValue(CEParticle::PARTICLE_W_BOSON));
	m_map_locals->put("WBOSON", CScriptValue::newValue(CEParticle::PARTICLE_W_BOSON));
	m_map_locals->put("NEUTRINO_ELECTRON", CScriptValue::newValue(CEParticle::PARTICLE_NEUTRINO_ELECTRON));
	m_map_locals->put("NEUTRINO_MUON", CScriptValue::newValue(CEParticle::PARTICLE_NEUTRINO_MUON));
	m_map_locals->put("NEUTRINO_TAU", CScriptValue::newValue(CEParticle::PARTICLE_NEUTRINO_TAU));
	m_map_locals->put("QUARK_D", CScriptValue::newValue(CEParticle::PARTICLE_QUARK_D));
	m_map_locals->put("QUARK_U", CScriptValue::newValue(CEParticle::PARTICLE_QUARK_U));
	m_map_locals->put("QUARK_S", CScriptValue::newValue(CEParticle::PARTICLE_QUARK_S));
	m_map_locals->put("QUARK_C", CScriptValue::newValue(CEParticle::PARTICLE_QUARK_C));
	m_map_locals->put("QUARK_B", CScriptValue::newValue(CEParticle::PARTICLE_QUARK_B));
	m_map_locals->put("QUARK_T", CScriptValue::newValue(CEParticle::PARTICLE_QUARK_T));
	m_map_locals->put("PROTON", CScriptValue::newValue(CEParticle::PARTICLE_PROTON));

	m_map_locals->put("FILE_START", CScriptValue::newValue(CESupport::FILE_TRUNCATE));
	m_map_locals->put("FILE_CLOSE", CScriptValue::newValue(CESupport::FILE_CLOSE));

	m_map_locals->put("DD_SMALL", CScriptValue::newValue(CFeynDiagram::DRAW_CAPS_SMALL|CFeynDiagram::SET_LOOP_120|CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES|CFeynDiagram::DRAW_DIAGRAM_NUMBERS));
	m_map_locals->put("DD_DEFAULT", CScriptValue::newValue(CFeynDiagram::DRAW_CAPS_SMALL|CFeynDiagram::SET_LOOP_120|CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES|CFeynDiagram::DRAW_DIAGRAM_NUMBERS));
	m_map_locals->put("DD_LARGE", CScriptValue::newValue(CFeynDiagram::SET_LOOP_120|CFeynDiagram::DRAW_CAPS_PARTICLE_TITLES));
	m_map_locals->put("DD_MOMENTA", CScriptValue::newValue(CFeynDiagram::DRAW_CAPS_MOMENTA));
	m_map_locals->put("DD_NUMBERS", CScriptValue::newValue(CFeynDiagram::DRAW_DIAGRAM_NUMBERS));
	m_map_locals->put("DD_SWAP_TALES", CScriptValue::newValue(CFeynDiagram::DRAW_SWAP_TALES));
	m_map_locals->put("DD_DONT_SWAP_TALES", CScriptValue::newValue(CFeynDiagram::DRAW_DONT_SWAP_TALES));

	// SQUARE options
	m_map_locals->put("SQR_DEFAULT", CScriptValue::newValue(
		CMESquareLib::SQRME_EXPAND_AND_SPLIT |
		CMESquareLib::SQRME_PRESIMPLIFY_ME |
		CMESquareLib::SQRME_CACHE_DENOMINATORS));
	m_map_locals->put("SQR_CMS", CScriptValue::newValue(CMESquareLib::SQRME_CMS_BEAMS));
	m_map_locals->put("SQR_NO_CROSSING_1", CScriptValue::newValue(CMESquareLib::SQRME_NO_CROSSINGS_FOR_ME1));
	m_map_locals->put("SQR_NO_CROSSING_2", CScriptValue::newValue(CMESquareLib::SQRME_NO_CROSSINGS_FOR_ME2));
	m_map_locals->put("SQR_MANDELSTAMS", CScriptValue::newValue(CMESquareLib::SQRME_ALLOW_MANDELSTAMS));
	m_map_locals->put("SQR_PH_GAMMA_CHIRAL", CScriptValue::newValue(CMESquareLib::SQRME_BERENDS_HELICITIES));
	m_map_locals->put("SQR_PH_BERENDS", CScriptValue::newValue(CMESquareLib::SQRME_BERENDS_HELICITIES));
	m_map_locals->put("SQR_PH_GAMMA_PRECISE", CScriptValue::newValue(CMESquareLib::SQRME_BERENDS_HELICITIES_PRECISE));
	m_map_locals->put("SQR_IR_ONLY", CScriptValue::newValue(CMESquareLib::SQRME_IR_LIMIT));
	m_map_locals->put("SQR_PH_EARLY", CScriptValue::newValue(CMESquareLib::SQRME_EARLY_PHOTON_HELICITIES));
	m_map_locals->put("SQR_PH_LATE", CScriptValue::newValue(CMESquareLib::SQRME_LATE_PHOTON_HELICITIES));
	m_map_locals->put("SQR_FORCE_4DIM", CScriptValue::newValue(CMESquareLib::SQRME_FORCE_4DIM));
	m_map_locals->put("SQR_NO_POSTARRANGE", CScriptValue::newValue(CMESquareLib::SQRME_NO_POSTARRANGE));
	m_map_locals->put("SQR_NO_SUB_ARRANGE", CScriptValue::newValue(CMESquareLib::SQRME_NO_SUB_ARRANGE));
	m_map_locals->put("SQR_SAVE_MEMORY", CScriptValue::newValue(CMESquareLib::SQRME_SAVE_MEMORY));

	// for CalcAmpl
	m_map_locals->put("CA_MANDELSTAMS", CScriptValue::newValue(CAmplCalcLib::A_MANDELSTAMS));
	m_map_locals->put("CA_NO_CROSSINGS", CScriptValue::newValue(CAmplCalcLib::A_NO_CROSSINGS));
	m_map_locals->put("CA_USE_FIERZ_TRANSFORM", CScriptValue::newValue(CAmplCalcLib::A_USE_FIERZ_TRANSFORM));
	m_map_locals->put("CA_SIMPLE_BASIS", CScriptValue::newValue(CAmplCalcLib::A_SIMPLE_BASIS));

	m_map_locals->put("EXPAND_SAVE_MEMORY", CScriptValue::newValue(CLSimplifyMisc::MISC_EXPAND_SAVE_MEMORY));

	//////////////// physics parameters
	m_map_locals->put("PHYS_QED", CScriptValue::newValue(CSMPhysics::PHYS_PURE_QED));
	m_map_locals->put("PHYS_QCD", CScriptValue::newValue(CSMPhysics::PHYS_QCD));
	m_map_locals->put("PHYS_Z", CScriptValue::newValue(CSMPhysics::PHYS_Z_VERTICES));
	m_map_locals->put("PHYS_W", CScriptValue::newValue(CSMPhysics::PHYS_W_VERTICES));
	m_map_locals->put("PHYS_ELW", CScriptValue::newValue(CSMPhysics::PHYS_ELW_FEYN)); // Feyn-SM with no rare vertices
	m_map_locals->put("PHYS_EONLY", CScriptValue::newValue(CSMPhysics::PHYS_EONLY));
	m_map_locals->put("PHYS_NOQUARKS", CScriptValue::newValue(CSMPhysics::PHYS_NOQUARKS));
	m_map_locals->put("PHYS_MU", CScriptValue::newValue(CSMPhysics::PHYS_MUON));
	m_map_locals->put("PHYS_TAU", CScriptValue::newValue(CSMPhysics::PHYS_TAU_LEPTON));
	m_map_locals->put("PHYS_SCALARS", CScriptValue::newValue(CSMPhysics::PHYS_SCALARS));
	m_map_locals->put("PHYS_GHOSTS", CScriptValue::newValue(CSMPhysics::PHYS_GHOSTS));
	m_map_locals->put("PHYS_QUARKS", CScriptValue::newValue(CSMPhysics::PHYS_QUARKS_3GEN));
	m_map_locals->put("PHYS_QUARKS_1GEN", CScriptValue::newValue(CSMPhysics::PHYS_QUARKS_1GEN));
	m_map_locals->put("PHYS_QUARKS_2GEN", CScriptValue::newValue(CSMPhysics::PHYS_QUARKS_2GEN));
	m_map_locals->put("PHYS_QUARKS_3GEN", CScriptValue::newValue(CSMPhysics::PHYS_QUARKS_3GEN));
	m_map_locals->put("PHYS_GAUGE_UNITARY", CScriptValue::newValue(CSMPhysics::PHYS_UNITARY_GAUGE));
	m_map_locals->put("PHYS_CT", CScriptValue::newValue(CSMPhysics::PHYS_RENORM_COUNTERTERMS));
	m_map_locals->put("PHYS_NO_QEDCT", CScriptValue::newValue(CSMPhysics::PHYS_NOQED_COUNTERTERMS));
	m_map_locals->put("PHYS_ANOMALOUS_CONST", CScriptValue::newValue(CSMPhysics::PHYS_ANOMALOUS_CONST));
	m_map_locals->put("PHYS_4BOSONS_ANOMALOUS", CScriptValue::newValue(CSMPhysics::PHYS_ANOMALOUS_CONST));
	m_map_locals->put("PHYS_ANOMALOUS_CONST_NOLCH", CScriptValue::newValue(CSMPhysics::PHYS_ANOMALOUS_CONST_NOLCH));
	m_map_locals->put("PHYS_ANOMALOUS_CONST_ONLY", CScriptValue::newValue(CSMPhysics::PHYS_ANOMALOUS_CONST_ONLY));
	//m_map_locals->put("PHYS_RARE_VERICES", CScriptValue::newValue(CSMPhysics::PHYS_RARE2_VERICES));
	//m_map_locals->put("PHYS_RARE_3", CScriptValue::newValue(CSMPhysics::PHYS_RARE3_VERICES));
	m_map_locals->put("PHYS_RARE_VERICES", CScriptValue::newValue(CSMPhysics::PHYS_RARE3_VERICES));
	m_map_locals->put("PHYS_SM", CScriptValue::newValue(CSMPhysics::PHYS_SM));
	m_map_locals->put("PHYS_NO_RARE_VERTICES", CScriptValue::newValue(CSMPhysics::PHYS_NO_RARE_VERTICES));
	m_map_locals->put("PHYS_Q1GEN_ONLY", CScriptValue::newValue(CSMPhysics::PHYS_Q1GEN_ONLY));
	m_map_locals->put("PHYS_Q2GEN_ONLY", CScriptValue::newValue(CSMPhysics::PHYS_Q2GEN_ONLY));
	m_map_locals->put("PHYS_PROTON_STRUCTFUNC", CScriptValue::newValue(CSMPhysics::PHYS_PROTON_STRUCTFUNC));

	// SetParameter params
	m_map_locals->put("PAR_MASSIVE_VEC_MATRIX_FORM", CScriptValue::newValue(CEProcessor::GPAR_MASSIVE_VECTOR_MATRIX));
	m_map_locals->put("PAR_EXAMINE_LESS_THEN", CScriptValue::newValue(CEProcessor::GPAR_EXAMINE_LESS_THEN));
	m_map_locals->put("PAR_EXAMINE_EQUALS", CScriptValue::newValue(CEProcessor::GPAR_EXAMINE_EQUALS));
	m_map_locals->put("PAR_MANDELSTAMS", CScriptValue::newValue(CEProcessor::GPAR_MANDELSTAMS));
	m_map_locals->put("PAR_SCREEN_SIZE", CScriptValue::newValue(CEProcessor::GPAR_SCREEN_SIZE));
	m_map_locals->put("PAR_TRACES_BONDEREV", CScriptValue::newValue(CEProcessor::GPAR_TRACES_BONDEREV));
	m_map_locals->put("PAR_TRACES_BONDEREV_F3EXPAND", CScriptValue::newValue(CEProcessor::GPAR_TRACES_BONDEREV_F3EXPAND));
	//"PAR_STR_FORTRAN_TEMP": deprecated!
	m_map_locals->put("PAR_STR_FORTRAN_TEMP", CScriptValue::newValue(CEProcessor::GPAR_STR_FORTRAN_TEMP));
	//"PAR_STR_CODE_TEMP"
	m_map_locals->put("PAR_STR_CODE_TEMP", CScriptValue::newValue(CEProcessor::GPAR_STR_CODE_TEMP));
	m_map_locals->put("PAR_ALLOW_CLASS_DIAGRAMS", CScriptValue::newValue(CEProcessor::GPAR_ALLOW_CLASS_DIAGRAMS));
	m_map_locals->put("PAR_DROP_ZERO_DIAGRAMS", CScriptValue::newValue(CEProcessor::GPAR_DROP_ZERO_DIAGRAMS));
	m_map_locals->put("PAR_AUTO_KINARRANGE", CScriptValue::newValue(CEProcessor::GPAR_AUTO_KINARRANGE));
	m_map_locals->put("PAR_CODE_GEN_FAST", CScriptValue::newValue(CEProcessor::GPAR_CODE_GEN_FAST));
	m_map_locals->put("PAR_EVALUATE_COMPLICATED_VI", CScriptValue::newValue(CEProcessor::GPAR_EVALUATE_COMPLICATED_VI));
	m_map_locals->put("PAR_LOOPTOOLS_COMPARTIBLE", CScriptValue::newValue(CEProcessor::GPAR_LOOPTOOLS_COMPARTIBLE));
	m_map_locals->put("PAR_RUN_EVT_NUCLON_ENTRIES", CScriptValue::newValue(CEProcessor::GPAR_RUN_EVT_NUCLON_ENTRIES));

	// KIN-ARRANGE parameters
	m_map_locals->put("KA_MASS_EXACT", CScriptValue::newValue(CLHEPBasics::KA_MASS_EXACT));
	m_map_locals->put("KA_MANDELSTAMS", CScriptValue::newValue(CLHEPBasics::KA_INVOLVE_MANDELSTAM_VARS));
	m_map_locals->put("KA_NO_EXPAND", CScriptValue::newValue(CLHEPBasics::KA_NO_EXPAND));

	// KIN-SIMPLIFY parameters
	m_map_locals->put("KS_MASS_EXACT", CScriptValue::newValue(CLHEPBasics::KS_MASS_EXACT));
	m_map_locals->put("KS_MANDELSTAMS", CScriptValue::newValue(CLHEPBasics::KS_INVOLVE_MANDELSTAM_VARS));
	m_map_locals->put("KS_NO_EXPAND", CScriptValue::newValue(CLHEPBasics::KS_NO_EXPAND));
	m_map_locals->put("KS_FACTORIZE_FUNCTIONS", CScriptValue::newValue(CLHEPBasics::KS_FACTORIZE_FUNCTIONS));
	m_map_locals->put("KS_FACTORIZE_DENS", CScriptValue::newValue(CLHEPBasics::KS_FACTORIZE_DENOMINATORS));
	m_map_locals->put("KS_FACTORIZE_DEFAULT", CScriptValue::newValue(CLHEPBasics::KS_FACTORIZE_FUNCTIONS|CLHEPBasics::KS_FACTORIZE_DENOMINATORS));
	m_map_locals->put("KS_FACTORIZE_ALL_DENS", CScriptValue::newValue(CLHEPBasics::KS_FACTORIZE_ALL_DENOMINATORS));

	// for MINIMIZE
	m_map_locals->put("MIN_VERIFY", CScriptValue::newValue(CViewSimplify::PART_FACT_VERIFY));
	m_map_locals->put("MIN_NUMBERS", CScriptValue::newValue(CViewSimplify::PART_FACT_NUMBERS));
	m_map_locals->put("MIN_FUNCTIONS", CScriptValue::newValue(CViewSimplify::PART_FACT_FUNCTIONS));
	m_map_locals->put("MIN_DENS", CScriptValue::newValue(CViewSimplify::PART_FACT_DENOMINATORS));
	m_map_locals->put("MIN_NUMERATORS", CScriptValue::newValue(CViewSimplify::PART_FACT_NUMERATORS));
	m_map_locals->put("MIN_ALL_DENOMINATORS", CScriptValue::newValue(CViewSimplify::PART_FACT_ALL_DENOMINATORS));
	m_map_locals->put("MIN_ALL_SINGLE_DENS", CScriptValue::newValue(CViewSimplify::PART_FACT_ALL_SINGLE_DENS));
	m_map_locals->put("MIN_DEFAULT", CScriptValue::newValue(CViewSimplify::PART_FACT_DEFAULT));

	// for Run
	m_map_locals->put("RUN_DEFAULT", CScriptValue::newValue(CLHEPRunProject::flag_RUN_DEFAULT));
	m_map_locals->put("RUN_AMPLITUDE", CScriptValue::newValue(CLHEPRunProject::flag_RUN_AMPLITUDE));
	m_map_locals->put("RUN_SQUARE", CScriptValue::newValue(CLHEPRunProject::flag_RUN_SQUARE));
	m_map_locals->put("RUN_RC", CScriptValue::newValue(CLHEPRunProject::flag_RUN_RC));
	m_map_locals->put("RUN_LOAD", CScriptValue::newValue(CLHEPRunProject::flag_RUN_LOAD));

	// for FACTOR
	m_map_locals->put("FACT_NO_NUMBERS", CScriptValue::newValue(CLFactor::DONOT_FACT_NUMBERS));
	m_map_locals->put("FACT_NO_DENS", CScriptValue::newValue(CLFactor::DONOT_FACT_FRACTIONS));
	m_map_locals->put("FACT_ALL_DENS", CScriptValue::newValue(CLFactor::FACT_ALL_DENOMINATORS));
	m_map_locals->put("FACT_ALL_SINGLE_DENS", CScriptValue::newValue(CLFactor::FACT_ALL_SINGLE_DENS));
	m_map_locals->put("FACT_VERIFY", CScriptValue::newValue(CLFactor::FACT_VERIFY));
	m_map_locals->put("FACT_NO_EXPAND", CScriptValue::newValue(CLFactor::NO_EXPAND));

	m_map_locals->put("MIN_NUMBERS", CScriptValue::newValue(CViewSimplify::PART_FACT_NUMBERS));
	m_map_locals->put("MIN_FUNCTIONS", CScriptValue::newValue(CViewSimplify::PART_FACT_FUNCTIONS));
	m_map_locals->put("MIN_DENS", CScriptValue::newValue(CViewSimplify::PART_FACT_DENOMINATORS));
	m_map_locals->put("MIN_NUMERATORS", CScriptValue::newValue(CViewSimplify::PART_FACT_NUMERATORS));
	m_map_locals->put("MIN_ALL_DENOMINATORS", CScriptValue::newValue(CViewSimplify::PART_FACT_ALL_DENOMINATORS));
	m_map_locals->put("MIN_ALL_SINGLE_DENS", CScriptValue::newValue(CViewSimplify::PART_FACT_ALL_SINGLE_DENS));
	m_map_locals->put("MIN_DEFAULT", CScriptValue::newValue(CViewSimplify::PART_FACT_DEFAULT));

	// FORTRAN settings
	m_map_locals->put("CODE_F90", CScriptValue::newValue(CLFileManager::FT_F90));
	m_map_locals->put("CODE_F77", CScriptValue::newValue(CLFileManager::FT_F77));
	m_map_locals->put("CODE_CPP", CScriptValue::newValue(CLFileManager::FT_CPP));

	m_map_locals->put("CODE_REAL8", CScriptValue::newValue(CLFileManager::FT_REAL_8BIT));
	m_map_locals->put("CODE_COMPLEX16", CScriptValue::newValue(CLFileManager::FT_COMPLEX_16BIT));
	m_map_locals->put("CODE_NO_4VEC_ENTRY", CScriptValue::newValue(CLFileManager::FT_DONOT_CREATE_4VECTOR_ENTRY));
	m_map_locals->put("CODE_SEPARATE_VI", CScriptValue::newValue(CLFileManager::FT_SEPARATE_VIRT_INTEGRALS));
	m_map_locals->put("CODE_NO_SPLIT", CScriptValue::newValue(CLFileManager::FT_DONOT_SPLIT_FUNC));
	m_map_locals->put("CODE_NO_COMMON", CScriptValue::newValue(CLFileManager::FT_NO_COMMON_PARAMETERS));
	m_map_locals->put("CODE_NO_COMMON_ARRAYS", CScriptValue::newValue(CLFileManager::FT_NO_COMMON_ARRAYS));
	m_map_locals->put("CODE_LOOPTOOLS", CScriptValue::newValue(CLFileManager::FT_LOOPTOOLS));
	m_map_locals->put("CODE_NO_CONSTANTS", CScriptValue::newValue(CLFileManager::FT_DONOT_INVOLVE_CONSTANTS));
	m_map_locals->put("CODE_POWER_PARAMS", CScriptValue::newValue(CLFileManager::FT_POWER_PARAMS));
	m_map_locals->put("CODE_CHECK_DENOMINATORS", CScriptValue::newValue(CLFileManager::FT_CHECK_DENOMINATORS));
	m_map_locals->put("CODE_PYTHIA_VECTORS", CScriptValue::newValue(CLFileManager::FT_PYTHIA_VECTORS));
	m_map_locals->put("CODE_IS_AMPLITUDE", CScriptValue::newValue(CLFileManager::FT_SRC_IS_AMPLITUDE));
	m_map_locals->put("CODE_IS_DIFF_CS", CScriptValue::newValue(CLFileManager::FT_SRC_IS_DIFFCS));
	m_map_locals->put("CODE_SAVE_AS_IS", CScriptValue::newValue(CLFileManager::FT_SRC_SAVE_AS_IS));//same to CODE_IS_DIFF_CS currently
	m_map_locals->put("CODE_NO_INCLUDES", CScriptValue::newValue(CLFileManager::FT_NO_INCLUDES));
	m_map_locals->put("CODE_COMPLETE_INCLUDE", CScriptValue::newValue(CLFileManager::FT_FULL_INCLUDE));
	m_map_locals->put("CODE_TYPE_ABSTRACT", CScriptValue::newValue(CLFileManager::FT_CPP_TYPE_ABSTRACT));
	m_map_locals->put("CODE_FORCE_REAL", CScriptValue::newValue(CLFileManager::FT_FORCE_REAL));
	m_map_locals->put("CODE_ALLOW_SWAP_INITIALS", CScriptValue::newValue(CLFileManager::FT_ALLOW_SWAP_INITIALS));

	m_map_locals->put("SEL_LOOPS", CScriptValue::newValue(CScriptFunctionLib::SelectDiagrams_LOOPS));
	m_map_locals->put("SEL_CT", CScriptValue::newValue(CScriptFunctionLib::SelectDiagrams_CT));
}

void CScriptProcessor::finalize()
{
	if (m_functions_register != NULL) delete m_functions_register;
	//if (m_map_operations != NULL) delete m_map_operations;
	dbg->log(25,"m_PreDefinedConstants",m_PreDefinedConstants);
	if (m_PreDefinedConstants!=NULL) { delete m_PreDefinedConstants; m_PreDefinedConstants=NULL; }

	m_init_OK = false;
}

//	***************************************************************
//	***               script error processing                   ***
//	***************************************************************
void CScriptProcessor::declare_script_error(int err_code, int i_param, const char* c_param)
{
	if (err_code != CScriptErrors::HALT_COMMAND) {
		CEString* line = new CEString(">>> ");
		// get current operands line
		int s = m_parser->get_operands_num();
		for (int i=0; i<s; i++) { if(i>0)line->AddString(" "); line->AddString(m_parser->get_operand(i,true)); }
		line->AddString(" <<< ERROR");
		dbg->log(0, line->getData());
		delete line;
	}

	CScriptException* scr_exc = new CScriptException(err_code,
		m_parser->get_curr_line_num(),
		m_parser->get_curr_line_pos()-1,
		i_param,
		c_param);
	throw scr_exc;
}

// Warning: args will be deleted therein, delete result array after call!
CEPtrArray* CScriptProcessor::invoke_function(CScriptFuncInfo *func_info, CEPtrArray *args)
{
	const char *f_title = func_info->GetFuncTitle();

	// log parameter list
	if (dbg->is_log_keeped(20)) {
		int a_s = args->size();
		dbg->log(20, "function call:");
		dbg->log(20, f_title);
		dbg->log(22, a_s, "%ld parameters");
		int i;
		for(i=0; i<a_s; i++) {
			dbg->log(22, (i+1), "%ld", args->GetAt(i));
		}
		dbg->log(22, "parameter list end.");
		dbg->log(0, ".....> ", f_title); // to keep console output at high debug level
	} else {
		dbg->log(0, ".....> ", f_title);
	}
	
	CEPtrArray* res_array = func_info->invoke_function(args, true); //Warning: args will be deleted therein

	// log parameter list
	if (dbg->is_log_keeped(20)) {
		int r_s = res_array->size();
		dbg->log(20, "function call result");
		dbg->log(22, r_s, "%ld result item(s)");
		int i;
		for(i=0; i<r_s; i++) {
			dbg->log(22, (i+1), "%ld", res_array->GetAt(i));
		}
		dbg->log(22, "results list end.");
	}

	return res_array;
}

// Warning: args will be deleted therein, delete result array after call!
CEPtrArray* CScriptProcessor::invoke_system_function(CScriptFuncInfo *func_info, CEPtrArray* args) {
	const char *f_title = func_info->GetFuncTitle();
	int arg_num = args->size();
	CScriptValue* res = NULL;

	if (CEUtil::equals("GetLastFuncValue", f_title)) {
		// returns undefined value if no value can be found

		int par = ((CScriptValue*)args->GetAt(0))->get_int_value();
		if (par > 0) par++; // the '1' pos is the error code, skip it

		if ((m_last_func_values == NULL) || (par >= m_last_func_values->size())) {
			// invalid call
			dbg->log(0, ((par>0)?(par-1):par), "GetLastFuncValue() WARNING: can't find requested value");
			//return CScriptFuncInfo::FReturnError(args, 1); // positive arguments means number of invalid argument
		}
		else {
			res = ((CScriptValue*)m_last_func_values->GetAt(par))->clone();
		}
	} else {
		dbg->log(0, "Error: unimplemented system function call", func_info);
		declare_script_error(CScriptErrors::INTERNAL_ERROR);
	}
	
	if (res == NULL) res = CScriptValue::newUndefinedValue();

	args->clear_all();
	args->add(res);
	return args;
}
