////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// SMPhysics.cpp: implementation of the CSMPhysics class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

const DWORD CSMPhysics::PHYS_PURE_QED				= 0x0000001;
const DWORD CSMPhysics::PHYS_Z_VERTICES				= 0x0000002;
const DWORD CSMPhysics::PHYS_W_VERTICES				= 0x0000004;
const DWORD CSMPhysics::PHYS_MUON					= 0x0000008;
const DWORD CSMPhysics::PHYS_TAU_LEPTON				= 0x0000010;
const DWORD CSMPhysics::PHYS_SCALARS				= 0x0000020; // including Higgs
const DWORD CSMPhysics::PHYS_GHOSTS					= 0x0000040;
const DWORD CSMPhysics::PHYS_EONLY					= 0x0000080;
const DWORD CSMPhysics::PHYS_QUARKS_1GEN			= 0x0000100; // u,d
const DWORD CSMPhysics::PHYS_QUARKS_2GEN			= 0x0000300; // u,d,s,c
const DWORD CSMPhysics::PHYS_QUARKS_3GEN			= 0x0000700; // u,d,s,c,b,t
const DWORD CSMPhysics::PHYS_NOQUARKS				= 0x0000800;
const DWORD CSMPhysics::PHYS_UNITARY_GAUGE			= 0x0001000;
const DWORD CSMPhysics::PHYS_RENORM_COUNTERTERMS	= 0x0002000;
const DWORD CSMPhysics::PHYS_NOQED_COUNTERTERMS		= 0x0004000;
const DWORD CSMPhysics::PHYS_ANOMALOUS_CONST		= 0x0008000;
const DWORD CSMPhysics::PHYS_ANOMALOUS_CONST_NOLCH	= 0x0010000;
const DWORD CSMPhysics::PHYS_ANOMALOUS_CONST_ONLY	= 0x0020000;
const DWORD CSMPhysics::PHYS_RARE2_VERICES			= 0x0040000;
const DWORD CSMPhysics::PHYS_RARE3_VERICES			= 0x00c0000; // vertices with 3 or more scalars
const DWORD CSMPhysics::PHYS_NO_RARE_VERTICES		= 0x0100000; // cancels PHYS_RARE3_VERICES (but leaves PHYS_RARE2_VERICES flag)
const DWORD CSMPhysics::PHYS_Q1GEN_ONLY				= 0x0200000; // 1 generation of quarks only (cancels PHYS_QUARKS_3GEN, leaves PHYS_QUARKS_1GEN)
const DWORD CSMPhysics::PHYS_Q2GEN_ONLY				= 0x0400000; // 2 generations of quarks only (cancels PHYS_QUARKS_3GEN, sets PHYS_QUARKS_2GEN)
const DWORD CSMPhysics::PHYS_PROTON_STRUCTFUNC		= 0x0800000; // including proton & structure functions
const DWORD CSMPhysics::PHYS_QCD					= 0x1000000;

//const DWORD CSMPhysics::PHYS_NO_RARE3_VERTICES		=;
//const DWORD CSMPhysics::PHYS_NO_RARE2_VERTICES		=;

// Standard Model in 't Hooft-Feynman gauge;
const DWORD CSMPhysics::PHYS_ELW_FEYN = PHYS_PURE_QED|PHYS_Z_VERTICES|PHYS_W_VERTICES|
	PHYS_MUON|PHYS_TAU_LEPTON|PHYS_SCALARS|PHYS_GHOSTS|PHYS_QUARKS_3GEN|PHYS_RARE2_VERICES|PHYS_RARE3_VERICES;

//Standard Model in Unitary gauge;
const DWORD CSMPhysics::PHYS_SM = PHYS_PURE_QED|PHYS_Z_VERTICES|PHYS_W_VERTICES|
	PHYS_MUON|PHYS_TAU_LEPTON|PHYS_SCALARS|PHYS_GHOSTS|PHYS_QUARKS_3GEN|PHYS_RARE2_VERICES|PHYS_RARE3_VERICES|PHYS_UNITARY_GAUGE;

CSMPhysics::CSMPhysClassCplInfo::CSMPhysClassCplInfo(int _pkConfigID, bool _bIsCt, int iVecSymmFlags) : CEObject(true) {
	m_pkConfigID=_pkConfigID; m_bIsCt=_bIsCt;
	m_partPKs.ToSize(0);
	m_iVecSymmFlags = iVecSymmFlags; // 0, 1 (1-2), 2(3-4) or 1|2
}

void CSMPhysics::CSMPhysClassCplInfo::SetClassCouplingFor(long pk1,long pk2,long pk3,long pk4) {
	m_partPKs.ToSize(0);
	if ((pk1==0) || (pk2==0)) CEException::halt("CSMPhysics::CSMPhysClassCplInfo::SetClassCouplingFor invalid params");

	m_partPKs.Add(pk1);
	m_partPKs.Add(pk2);
	if (pk3!=0) { m_partPKs.Add(pk3); if(pk4!=0){m_partPKs.Add(pk4);} }
	
	// adjust PKs (truncate particle number and append CLASS-flag)
	for(int i=0; i<m_partPKs.GetSize(); i++) {
		long pk=m_partPKs.GetAt(i);
		long pk_abs=((pk<0)?(-pk):pk);
		pk_abs = (CEParticle::GetClassFitFlagPk(pk_abs) & pk_abs);
		pk_abs |= CEParticle::PARTICLE_CLASS_FLAG;
		pk=((pk<0)?(-pk_abs):pk_abs);
		m_partPKs.SetAt(i,pk);
	}
}

CSMPhysics::CSMPhysClassCplInfo::CSMPhysClassCplInfo(class CSMPhysClassCplInfo& inf) : CEObject(true) {
	m_pkConfigID=inf.m_pkConfigID; m_bIsCt=inf.m_bIsCt; m_iVecSymmFlags=inf.m_iVecSymmFlags;
	long s = inf.m_partPKs.GetSize();
	for (long i=0; i<s; i++) { m_partPKs.Add(inf.m_partPKs.GetAt(i)); }
}

CSMPhysics::CSMPhysClassCplInfo::~CSMPhysClassCplInfo() {}

bool CSMPhysics::CSMPhysClassCplInfo::contain_vector(void * obj) {
	if (obj==NULL) return false;
	if ((obj == &m_partPKs) || (m_partPKs.contain_vector(obj))) return true;
	return false;
}

const char* CSMPhysics::CSMPhysClassCplInfo::CLASS_NAME = "CSMPhysics::CSMPhysClassCplInfo";
const char* CSMPhysics::CSMPhysClassCplInfo::ClassName() const { return CSMPhysics::CSMPhysClassCplInfo::CLASS_NAME; }
char* CSMPhysics::CSMPhysClassCplInfo::toLine(int mode) {
	CEString res;
	if (m_bIsCt) res.AddString("{SM_ClassCT: "); else res.AddString("{SM_ClassCpl: ");
	char* line=m_partPKs.toLine(mode); res.AddString(line); CMemoryManager::mfree(line);
	if (m_iVecSymmFlags!=0) { res.AddString(" ("); res.add_int(m_iVecSymmFlags); res.AddString(")"); }
	res.AddString("}");
	char* resS=CEUtil::copyString(res.getData());
	//if (resS==NULL) CEException::halt("111");
	return resS;
}

bool CSMPhysics::CSMPhysClassCplInfo::equals(CSMPhysClassCplInfo* other) {
	if (other==NULL) return false;
	if (m_pkConfigID != other->m_pkConfigID) return false;
	if (!CEUtil::LEQU(m_bIsCt,other->m_bIsCt)) return false;
	if (m_iVecSymmFlags != other->m_iVecSymmFlags) return false;
	long s=m_partPKs.GetSize(); if(s!=other->m_partPKs.GetSize()) return false;
	for (long i=0; i<s; i++) {
		if (m_partPKs.GetAt(i) != other->m_partPKs.GetAt(i)) return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CSMPhysics::CSMPhysics(DWORD phys_ID) : CDiagramPhysics()
{
	dbg->log(25, "Standard Model Physics creation.");
	m_H_vect_index=NULL; m_H_vect_index_2=NULL; m_H_vect_index_3=NULL; m_H_vect_index_4=NULL;

	// init masses
	CExpression* sym = NULL;
	//sym = new CESymbol("m\\_f"); set_mass_sym(CEParticle::PARTICLE_FERMION, sym); delete sym;
	sym = new CESymbol("m\\_\\[Nu]"); set_mass_sym(CEParticle::PARTICLE_NEUTRINO, sym); delete sym;
	//sym = CENumber::GetZero(); set_mass_sym(CEParticle::PARTICLE_NEUTRINO, sym); delete sym;

	sym = CENumber::GetZero();
	set_mass_sym(CEParticle::PARTICLE_PHOTON, sym); set_mass_sym(CEParticle::PARTICLE_FPGHOST_UA, sym);
	set_mass_sym(CEParticle::PARTICLE_GLUON, sym);
	delete sym;

	/*
	*	Note: Charge(W)==Charge(Phi)==-1
	*/
	sym = new CESymbol("m\\_W");
	set_mass_sym(CEParticle::PARTICLE_W_BOSON, sym); set_mass_sym(CEParticle::PARTICLE_SCALAR_PHI, sym);
	set_mass_sym(CEParticle::PARTICLE_FPGHOST_UWp, sym); set_mass_sym(CEParticle::PARTICLE_FPGHOST_UWm, sym);
	delete sym;

	sym = new CESymbol("m\\_Z");
	set_mass_sym(CEParticle::PARTICLE_Z_BOSON, sym); set_mass_sym(CEParticle::PARTICLE_SCALAR_CHI, sym); set_mass_sym(CEParticle::PARTICLE_FPGHOST_UZ, sym);
	delete sym;

	m_SM_corr_rules = NULL;
	m_Hamiltonian_expr = NULL;
	m_Renorm_Hamiltonian = NULL;
	m_HamiltonianReplsRule = NULL;
	m_model_consts = NULL;
	m_renorm_consts = NULL;
	m_eClassesHamiltonian = NULL;
	m_eClassesHamiltonianR = NULL;
	m_bClassVerticesComplete=true;

	m_phys_ID = phys_ID;
	if ((m_phys_ID & (PHYS_SM | PHYS_QCD)) == 0) m_phys_ID = PHYS_PURE_QED | phys_ID; // preserve rest flags
}

const char* CSMPhysics::CLASS_NAME = "CSMPhysics";
const char* CSMPhysics::ClassName() const { return CLASS_NAME; }

void CSMPhysics::init_physics() {
	dbg->log(35, "Standard Model Physics initialization.");

	//set 3-generations if no NO-HIGHER-GENERATIONS flags are used in 'PHYS_QCD' model
	if (((m_phys_ID & PHYS_QCD) == PHYS_QCD) &&
		((m_phys_ID & (PHYS_NOQUARKS|PHYS_Q1GEN_ONLY|PHYS_Q2GEN_ONLY)) == 0)) {
		m_phys_ID|=PHYS_QUARKS_3GEN;
	}

	// check NO-XXX physics options
	if ((m_phys_ID & PHYS_EONLY) == PHYS_EONLY) { DWORD noFlag=PHYS_MUON|PHYS_TAU_LEPTON; m_phys_ID|=noFlag; m_phys_ID^=noFlag; }
	if ((m_phys_ID & PHYS_NO_RARE_VERTICES) == PHYS_NO_RARE_VERTICES) {
		if ((m_phys_ID & PHYS_RARE3_VERICES) == PHYS_RARE3_VERICES) {
			m_phys_ID|=PHYS_RARE3_VERICES; m_phys_ID^=PHYS_RARE3_VERICES; // clear PHYS_RARE3_VERICES flag
			m_phys_ID|=PHYS_RARE2_VERICES; // restore PHYS_RARE2_VERICES flag
		}
	}
	if ((m_phys_ID & PHYS_NOQUARKS) == PHYS_NOQUARKS) { m_phys_ID|=PHYS_QUARKS_3GEN; m_phys_ID^=PHYS_QUARKS_3GEN; }
	if ((m_phys_ID & PHYS_Q1GEN_ONLY) == PHYS_Q1GEN_ONLY) {
		m_phys_ID|=PHYS_QUARKS_3GEN; m_phys_ID^=PHYS_QUARKS_3GEN; // clear PHYS_QUARKS_3GEN flag
		m_phys_ID|=PHYS_QUARKS_1GEN; // restore/set PHYS_QUARKS_1GEN flag
	}
	if ((m_phys_ID & PHYS_Q2GEN_ONLY) == PHYS_Q2GEN_ONLY) {
		m_phys_ID|=PHYS_QUARKS_3GEN; m_phys_ID^=PHYS_QUARKS_3GEN; // clear PHYS_QUARKS_3GEN flag
		m_phys_ID |= PHYS_QUARKS_2GEN; // restore/set PHYS_QUARKS_2GEN flag
	}

	// particles data init
	CEProcessor::getEnvironment()->ClearMassCalcOrders();
	CEProcessor::getEnvironment()->SetMassCalcOrder(CEParticle::PARTICLE_NEUTRINO_ELECTRON, 0);
	CEProcessor::getEnvironment()->SetMassCalcOrder(CEParticle::PARTICLE_NEUTRINO_MUON, 0);
	CEProcessor::getEnvironment()->SetMassCalcOrder(CEParticle::PARTICLE_NEUTRINO_TAU, 0);

	// internal constants init
	initModelConstsList();
	initRenormConstsList();

	init_corr_rules();
	//init_H_data();
	bool clOK=InitCouplingsClasses();
	if (!clOK) {
		if (m_bClassVerticesComplete) {
			dbg->log(0, "Warning: class diagrams not available. InitCouplingsClasses() failed.");
			m_bClassVerticesComplete=false;
		}
	}

	dbg->log(15, m_bClassVerticesComplete, "Standard Model Physics initialized (Classes ok: %ld).");
}

CSMPhysics::~CSMPhysics() {
	if (m_SM_corr_rules != NULL) delete m_SM_corr_rules;
	if (m_Hamiltonian_expr != NULL) delete m_Hamiltonian_expr;
	if (m_Renorm_Hamiltonian != NULL) delete m_Renorm_Hamiltonian;
	if (m_HamiltonianReplsRule != NULL) delete m_HamiltonianReplsRule;
	if (m_renorm_consts != NULL) delete m_renorm_consts;
	if (m_model_consts != NULL) delete m_model_consts;
	if (m_eClassesHamiltonian != NULL) delete m_eClassesHamiltonian;
	if (m_eClassesHamiltonianR != NULL) delete m_eClassesHamiltonianR;

	if (m_H_vect_index != NULL) delete m_H_vect_index;
	if (m_H_vect_index_2 != NULL) delete m_H_vect_index_2;
	if (m_H_vect_index_3 != NULL) delete m_H_vect_index_3;
	if (m_H_vect_index_4 != NULL) delete m_H_vect_index_4;
}

const int CSMPhysics::CONST_E		= 1;
const int CSMPhysics::CONST_CW		= 2;
const int CSMPhysics::CONST_SW		= 3;
const int CSMPhysics::CONST_CTGW	= 4;
const int CSMPhysics::CONST_ALPHA	= 5;
const int CSMPhysics::CONST_GP		= 11;
const int CSMPhysics::CONST_GM		= 12;
const int CSMPhysics::CONST_NEU		= 13;
const int CSMPhysics::CONST_QU		= 21; // -1/3 {u,c,t}
const int CSMPhysics::CONST_QD		= 22; // 2/3 {d,s,b}
const int CSMPhysics::CONST_GU_P	= 23;
const int CSMPhysics::CONST_GU_M	= 24;
const int CSMPhysics::CONST_GD_P	= 25;
const int CSMPhysics::CONST_GD_M	= 26;
const int CSMPhysics::CONST_G_S		= 27;

void CSMPhysics::initModelConstsList() {
	dbg->log(45, "Standard Model constants list initialization.");

	if (m_model_consts != NULL) { delete m_model_consts; m_model_consts = NULL; }
	m_model_consts = new CEPtrArray();
	
	involveModelConst(CONST_E, new CESymbol("e"));
	involveModelConst(CONST_ALPHA, new CESymbol("\\[Alpha]"));
	involveModelConst(CONST_CW, new CESymbol("c\\+\\(W\\)"));
	involveModelConst(CONST_SW, new CESymbol("s\\+\\(W\\)"));
	involveModelConst(CONST_CTGW, new CESymbol("ctg\\+\\(W\\)"));
	involveModelConst(CONST_GP, new CESymbol("\\(g\\+p\\)"));
	involveModelConst(CONST_GM, new CESymbol("\\(g\\+m\\)"));
	involveModelConst(CONST_NEU, new CESymbol("\\(g\\+W\\)")); // = e/(sinW*sqrt(2))
	involveModelConst(CONST_QU, new CESymbol("Q\\_u"));
	involveModelConst(CONST_QD, new CESymbol("Q\\_d"));
	involveModelConst(CONST_GU_P, new CESymbol("Gu\\_P"));
	involveModelConst(CONST_GU_M, new CESymbol("Gu\\_M"));
	involveModelConst(CONST_GD_P, new CESymbol("Gd\\_P"));
	involveModelConst(CONST_GD_M, new CESymbol("Gd\\_M"));
	involveModelConst(CONST_G_S, new CESymbol("g\\_s"));
}

void CSMPhysics::involveModelConst(int id, CExpression* val) { // deletes value
	m_model_consts->add(new CExprContainer(val, id)); delete val; }

CExpression* CSMPhysics::getModelContsSym(int id) {
	if (m_model_consts == NULL) initModelConstsList();
	int s = m_model_consts->size();
	int i;
	for(i=0; i<s; i++) {
		CExprContainer* itm = (CExprContainer*)m_model_consts->GetAt(i);
		if (itm->get_info() == id) return itm->get_expr();
	}
	dbg->log(0, id, "<INTERNAL ERROR> can't find model constant symbol (ID == %ld)");
	CEException::halt();
	return NULL;
}

// returns \alpha*4*pi/e^2 = 1
CExpression* CSMPhysics::getModelContsUnitFactor() {
	CExpression* pi = CEBasic::GetPiSym(); // not cloned
	CExpression* res = getModelContsSym(CSMPhysics::CONST_ALPHA)->clone();
	res = res->multiply(CENumber::GetFour())->multiply(pi->clone());
	res = res->divide(getModelContsSym(CSMPhysics::CONST_E)->sqra());
	return res;
}

const int CSMPhysics::DZ_AA			= 1;
const int CSMPhysics::DZ_ZA			= 2;
const int CSMPhysics::DZ_AZ			= 3;
const int CSMPhysics::DZ_ZZ			= 4;
const int CSMPhysics::DZ_E			= 5;
const int CSMPhysics::DZ_F_EL		= 10;
const int CSMPhysics::DZ_F_MU		= 11;
const int CSMPhysics::DZ_F_TAU		= 12;
const int CSMPhysics::DZ_F_EL_P		= 13;
const int CSMPhysics::DZ_F_EL_M		= 14;
const int CSMPhysics::DZ_F_MU_P		= 15;
const int CSMPhysics::DZ_F_MU_M		= 16;
const int CSMPhysics::DZ_F_TAU_P	= 17;
const int CSMPhysics::DZ_F_TAU_M	= 18;
const int CSMPhysics::DM_F_EL		= 20;
const int CSMPhysics::DM_F_MU		= 21;
const int CSMPhysics::DM_F_TAU		= 22;
const int CSMPhysics::DM_F_EL_P		= 23;
const int CSMPhysics::DM_F_EL_M		= 24;
const int CSMPhysics::DM_F_MU_P		= 25;
const int CSMPhysics::DM_F_MU_M		= 26;
const int CSMPhysics::DM_F_TAU_P	= 27;
const int CSMPhysics::DM_F_TAU_M	= 28;
void CSMPhysics::initRenormConstsList() {
	dbg->log(45, "Renormalization symbols list initialization.");

	if (m_renorm_consts != NULL) { delete m_renorm_consts; m_renorm_consts = NULL; }
	m_renorm_consts = new CEPtrArray();

	involveRenormConst(DZ_AA, new CESymbol("\\[Delta]Z\\+\\(AA\\)"));
	involveRenormConst(DZ_ZA, new CESymbol("\\[Delta]Z\\+\\(ZA\\)"));
	involveRenormConst(DZ_AZ, new CESymbol("\\[Delta]Z\\+\\(AZ\\)"));
	involveRenormConst(DZ_ZZ, new CESymbol("\\[Delta]Z\\+\\(ZZ\\)"));

	involveRenormConst(DZ_E, new CESymbol("\\[Delta]Z\\+\\(e\\)"));
	involveRenormConst(DZ_F_EL, new CESymbol("\\[Delta]F\\+\\(e\\)"));
	involveRenormConst(DZ_F_MU, new CESymbol("\\[Delta]F\\+\\(\\[Mu]\\)"));
	involveRenormConst(DZ_F_TAU, new CESymbol("\\[Delta]F\\+\\(\\[Tau]\\)"));

	involveRenormConst(DZ_F_EL_P, new CESymbol("\\[Delta]FP\\+\\(e\\)"));
	involveRenormConst(DZ_F_EL_M, new CESymbol("\\[Delta]FM\\+\\(e\\)"));
	involveRenormConst(DZ_F_MU_P, new CESymbol("\\[Delta]FP\\+\\(\\[Mu]\\)"));
	involveRenormConst(DZ_F_MU_M, new CESymbol("\\[Delta]FM\\+\\(\\[Mu]\\)"));
	involveRenormConst(DZ_F_TAU_P, new CESymbol("\\[Delta]FP\\+\\(\\[Tau]\\)"));
	involveRenormConst(DZ_F_TAU_M, new CESymbol("\\[Delta]FM\\+\\(\\[Tau]\\)"));

	involveRenormConst(DM_F_EL, new CESymbol("\\[Delta]m\\+\\(e\\)"));
	involveRenormConst(DM_F_MU, new CESymbol("\\[Delta]m\\+\\(\\[Mu]\\)"));
	involveRenormConst(DM_F_TAU, new CESymbol("\\[Delta]m\\+\\(\\[Tau]\\)"));

	involveRenormConst(DM_F_EL_P, new CESymbol("\\[Delta]mP\\+\\(e\\)"));
	involveRenormConst(DM_F_EL_M, new CESymbol("\\[Delta]mM\\+\\(e\\)"));
	involveRenormConst(DM_F_MU_P, new CESymbol("\\[Delta]mP\\+\\(\\[Mu]\\)"));
	involveRenormConst(DM_F_MU_M, new CESymbol("\\[Delta]mM\\+\\(\\[Mu]\\)"));
	involveRenormConst(DM_F_TAU_P, new CESymbol("\\[Delta]mP\\+\\(\\[Tau]\\)"));
	involveRenormConst(DM_F_TAU_M, new CESymbol("\\[Delta]mM\\+\\(\\[Tau]\\)"));
}

void CSMPhysics::involveRenormConst(int id, CExpression* val) { // deletes value
	m_renorm_consts->add(new CExprContainer(val, id)); delete val; }

CExpression* CSMPhysics::getRenormContsSym(int id) {
	if (m_renorm_consts == NULL) initRenormConstsList();
	int s = m_renorm_consts->size();
	int i;
	for(i=0; i<s; i++) {
		CExprContainer* itm = (CExprContainer*)m_renorm_consts->GetAt(i);
		if (itm->get_info() == id) return itm->get_expr();
	}
	dbg->log(0, id, "<ERROR> can't find renormalization constant symbol (ID == %ld)");
	CEException::halt();
	return NULL;
}

CDiagCorrRules* CSMPhysics::getCorrespondenceRules() { return m_SM_corr_rules; }
DWORD CSMPhysics::getPhysicsID() { return m_phys_ID; }
bool CSMPhysics::CounterTermsDeclared() { return ((getPhysicsID() & CSMPhysics::PHYS_RENORM_COUNTERTERMS)==CSMPhysics::PHYS_RENORM_COUNTERTERMS); }

const char* CSMPhysics::get_particle_symbol(long _kind, int sym_style) {
	long pk_abs = CLMath::abs(_kind);
	const char* res = NULL;

	if ((pk_abs & CEParticle::PARTICLE_CLASS_FLAG) == CEParticle::PARTICLE_CLASS_FLAG) {
		res=GetClassPKTitle(pk_abs);
		return res;
	}

	if (sym_style == CDiagramPhysics::SYMBOL_STYLE_TEX) {
		if (pk_abs == CEParticle::PARTICLE_FERMION) res = "f";
		else if (pk_abs == CEParticle::PARTICLE_LEPTON) res = "l";
		else if (pk_abs == CEParticle::PARTICLE_ELECTRON) res = "e";
		else if (pk_abs == CEParticle::PARTICLE_MUON) res = "\\mu";
		else if (pk_abs == CEParticle::PARTICLE_TAU_LEPTON) res = "\\tau";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO) res = "\\nu";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_ELECTRON) res = "\\nu_{e}";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_MUON) res = "\\nu_{\\mu}";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_TAU) res = "\\nu_{\\tau}";
		else if (pk_abs == CEParticle::PARTICLE_QUARK) res = "q";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_CONJUGETED) res = "\bar{q}";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_U) res = "u";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_D) res = "d";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_S) res = "s";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_C) res = "c";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_B) res = "b";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_T) res = "t";
		else if (pk_abs == CEParticle::PARTICLE_VECTOR_BOSON) res = "V";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRAL_BOSON) res = "Neutral";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRAL_VEC_BOSON) res = "V_{0}";
		else if (pk_abs == CEParticle::PARTICLE_PHOTON) res = "\\gamma";
		else if (pk_abs == CEParticle::PARTICLE_Z_BOSON) res = "Z";
		else if (pk_abs == CEParticle::PARTICLE_W_BOSON) res = "W";
		else if (pk_abs == CEParticle::PARTICLE_SCALAR) res = "Scalar";
		else if (pk_abs == CEParticle::PARTICLE_HIGGS) res = "H";
		else if (pk_abs == CEParticle::PARTICLE_SCALAR_PHI) res = "\\phi";
		else if (pk_abs == CEParticle::PARTICLE_SCALAR_CHI) res = "\\chi";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UA) res = "G_{\\gamma}";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UZ) res = "G_z";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWp) res = "G^{+}_{W}";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWm) res = "G^{-}_{W}";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST) res = "u^x";
		else if (pk_abs == CEParticle::PARTICLE_BOSON) res = "Boson";
		else if (pk_abs == CEParticle::PARTICLE_PROTON) res = "p";
		else if (pk_abs == CEParticle::PARTICLE_GLUON) res = "g";
	} else if (sym_style == CDiagramPhysics::SYMBOL_STYLE_NB) {
		if (pk_abs == CEParticle::PARTICLE_FERMION) res = "f";
		else if (pk_abs == CEParticle::PARTICLE_LEPTON) res = "l";
		else if (pk_abs == CEParticle::PARTICLE_ELECTRON) res = "e";
		else if (pk_abs == CEParticle::PARTICLE_MUON) res = "\\[Mu]";
		else if (pk_abs == CEParticle::PARTICLE_TAU_LEPTON) res = "\\[Tau]";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO) res = "\\[Nu]";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_ELECTRON) res = "\\[Nu]\\_e";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_MUON) res = "\\[Nu]\\_\\[Mu]";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_TAU) res = "\\[Nu]\\_\\[Tau]";
		else if (pk_abs == CEParticle::PARTICLE_QUARK) res = "q";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_CONJUGETED) res = "\\(q\\&_\\)";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_U) res = "u";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_D) res = "d";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_S) res = "s";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_C) res = "c";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_B) res = "b";
		else if (pk_abs == CEParticle::PARTICLE_QUARK_T) res = "t";
		else if (pk_abs == CEParticle::PARTICLE_VECTOR_BOSON) res = "V";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRAL_BOSON) res = "Neutral";
		else if (pk_abs == CEParticle::PARTICLE_NEUTRAL_VEC_BOSON) res = "V\\_0";
		else if (pk_abs == CEParticle::PARTICLE_PHOTON) res = "\\[Gamma]";
		else if (pk_abs == CEParticle::PARTICLE_Z_BOSON) res = "Z";
		else if (pk_abs == CEParticle::PARTICLE_W_BOSON) res = "W";
		else if (pk_abs == CEParticle::PARTICLE_SCALAR) res = "Scalar";
		else if (pk_abs == CEParticle::PARTICLE_HIGGS) res = "H";
		else if (pk_abs == CEParticle::PARTICLE_SCALAR_PHI) res = "\\[Phi]";
		else if (pk_abs == CEParticle::PARTICLE_SCALAR_CHI) res = "\\[Chi]";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UA) res = "\\(G\\&A\\)";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UZ) res = "\\(G\\&Z\\)";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWp) res = "\\(G\\&Wp\\)";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWm) res = "\\(G\\&Wm\\)";
		else if (pk_abs == CEParticle::PARTICLE_FPGHOST) res = "\\(u\\&X\\)";
		else if (pk_abs == CEParticle::PARTICLE_BOSON) res = "Boson";
		else if (pk_abs == CEParticle::PARTICLE_PROTON) res = "p";
		else if (pk_abs == CEParticle::PARTICLE_GLUON) res = "g";
	} else {
		dbg->log(0, sym_style, "get_particle_symbol() invalid style parameter: %ld");
		throw(new CEException(CEException::INT_ERROR));
	}
	return res;
}

int CSMPhysics::getCharge3(long pk) { // charge * 3 (INT_MIN for unknown)
	long pk_abs = ((pk>=0)?pk:-pk);
	int iCharge3=INT_MIN;

	if (pk_abs == CEParticle::PARTICLE_ELECTRON) iCharge3=-3;
	else if (pk_abs == CEParticle::PARTICLE_MUON) iCharge3=-3;
	else if (pk_abs == CEParticle::PARTICLE_TAU_LEPTON) iCharge3=-3;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_ELECTRON) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_MUON) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_TAU) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_U) iCharge3=2; // +2/3 {u,c,t}
	else if (pk_abs == CEParticle::PARTICLE_QUARK_D) iCharge3=-1; // -1/3 {d,s,b}
	else if (pk_abs == CEParticle::PARTICLE_QUARK_S) iCharge3=-1;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_C) iCharge3=2;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_B) iCharge3=-1;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_T) iCharge3=2;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRAL_BOSON) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_PHOTON) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_Z_BOSON) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_W_BOSON) iCharge3=-3;
	else if (pk_abs == CEParticle::PARTICLE_HIGGS) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_SCALAR_PHI) iCharge3=-3;
	else if (pk_abs == CEParticle::PARTICLE_SCALAR_CHI) iCharge3=0;
	else if (pk_abs == CEParticle::PARTICLE_PROTON) iCharge3=3;
	else if (pk_abs == CEParticle::PARTICLE_GLUON) iCharge3=0;

	if ((pk < 0) && (iCharge3 != INT_MIN)) iCharge3=-iCharge3;
	return iCharge3;
}

bool CSMPhysics::validate_part_kind(long _kind) { // is it particle PK (real,ghost,...) or flag (neutral,scalar,...)
	long pk_abs = CLMath::abs(_kind);
	bool res = false;
	if (pk_abs == CEParticle::PARTICLE_FERMION) res = false;
	else if (pk_abs == CEParticle::PARTICLE_LEPTON) res = false;
	else if (pk_abs == CEParticle::PARTICLE_ELECTRON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_MUON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_TAU_LEPTON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO) res = false;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_ELECTRON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_MUON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRINO_TAU) res = true;
	else if (pk_abs == CEParticle::PARTICLE_QUARK) res = false;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_CONJUGETED) res = false;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_U) res = true;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_D) res = true;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_S) res = true;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_C) res = true;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_B) res = true;
	else if (pk_abs == CEParticle::PARTICLE_QUARK_T) res = true;
	else if (pk_abs == CEParticle::PARTICLE_VECTOR_BOSON) res = false;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRAL_BOSON) res = false;
	else if (pk_abs == CEParticle::PARTICLE_NEUTRAL_VEC_BOSON) res = false;
	else if (pk_abs == CEParticle::PARTICLE_PHOTON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_Z_BOSON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_W_BOSON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_SCALAR) res = false;
	else if (pk_abs == CEParticle::PARTICLE_HIGGS) res = true;
	else if (pk_abs == CEParticle::PARTICLE_SCALAR_PHI) res = true;
	else if (pk_abs == CEParticle::PARTICLE_SCALAR_CHI) res = true;
	else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UA) res = true;
	else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UZ) res = true;
	else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWp) res = true;
	else if (pk_abs == CEParticle::PARTICLE_FPGHOST_UWm) res = true;
	else if (pk_abs == CEParticle::PARTICLE_FPGHOST) res = false;
	else if (pk_abs == CEParticle::PARTICLE_BOSON) res = false;
	else if (pk_abs == CEParticle::PARTICLE_PROTON) res = true;
	else if (pk_abs == CEParticle::PARTICLE_GLUON) res = true;
	return res;
}

CExprContainer* CSMPhysics::getMassSymbolCart(long pk_abs) {
	CExprContainer* res = NULL;
	long s = m_particle_mass_symbols.size();
	long i;
	for(i=0; i<s; i++) {
		CExprContainer* item = (CExprContainer*)m_particle_mass_symbols.GetAt(i);
		if (item->get_info() == pk_abs) {
			res = item;
			break;
		}
	}
	return res;
}

// returns not cloned value
CExpression* CSMPhysics::get_mass_sym(long pk) {
	long pk_abs = (pk>=0)?pk:(-pk);

	CExprContainer* cart = getMassSymbolCart(pk_abs);
	if (cart != NULL) return cart->get_expr();

	// create default symbol for particle mass
	const char* part_sym = get_particle_symbol(pk_abs, CDiagramPhysics::SYMBOL_STYLE_NB);
	char* data = (char*)CMemoryManager::mmalloc(32);
	if (part_sym != NULL) sprintf(data, "m\\_%s", part_sym);
	else sprintf(data, "m\\_%ld", pk_abs);
	CESymbol* def_sym = new CESymbol(data);
	CMemoryManager::mfree(data);
	set_mass_sym(pk_abs, def_sym); delete def_sym;

	// find out symbol set (we need non-cloned value)
	cart = getMassSymbolCart(pk_abs);
	if (cart != NULL) return cart->get_expr();
	
	dbg->log(0, "Error: No mass symbol found after creation one.");
	throw(new CEException(CEException::INT_ERROR));
	//return NULL;
}

void CSMPhysics::set_mass_sym(long pk, CExpression* val) {
	long pk_abs = (pk>=0)?pk:(-pk);

	CExprContainer* cart = getMassSymbolCart(pk_abs);
	if (cart != NULL) {
		cart->set_expr(val);
	} else {
		cart = new CExprContainer(val, pk_abs);
		m_particle_mass_symbols.add(cart);
	}
}

/*
*	Written to avoid complex initialization in class constructor.
*/
/*void CSMPhysics::initPhysics() {
	init_corr_rules();
}*/

/*
*	THESE DATA SHOULD BE GENERATED FROM INTERACTION LAGRANGIAN
*/
void CSMPhysics::init_corr_rules() {
	dbg->log(35, "Standard Model Physics CORRESPONDENCE RULES initialization.");

	//////////////////////////////////////////////////////////////	prepare BEGIN
	// Numbers cache (to be cloned before use)
	CENumber unit((long)1);
	CENumber _Zero((long)0);
	
	m_SM_corr_rules = new CDiagCorrRules();
	CExpression* p = m_SM_corr_rules->get_internal_symbol(CDiagCorrRules::INTERNAL_SYM_MOMENTUM);
	CExpression* m = m_SM_corr_rules->get_internal_symbol(CDiagCorrRules::INTERNAL_SYM_MASS);
	CExpression* i1 = m_SM_corr_rules->get_internal_symbol(CDiagCorrRules::INTERNAL_SYM_VECTOR_INDEX);
	CExpression* i2 = m_SM_corr_rules->get_internal_symbol(CDiagCorrRules::INTERNAL_SYM_VECTOR_INDEX_2);
	CExpression* i3 = CESymbol::getUnique();
	CExpression* i4 = CESymbol::getUnique();
	CExpression* i_gm1 = CESymbol::getUnique();
	CExpression* i_gm2 = CESymbol::getUnique();
	CExpression* i_gm3 = CESymbol::getUnique();
	CExpression* i_gm4 = CESymbol::getUnique();

	CExpression* p1 = new CESymbol("Qtmp\\_1");
	CExpression* p2 = new CESymbol("Qtmp\\_2");
	CExpression* p3 = new CESymbol("Qtmp\\_3");
	CExpression* p4 = new CESymbol("Qtmp\\_4");

	CExpression* i_run1 = CESymbol::getUnique();
	CExpression* i_run2 = CESymbol::getUnique();

	CExpression* Im = CENumber::GetImUnit();
	CExpression* mIm = Im->clone(); mIm->inverseFlag();
	CExpression* mU = CENumber::GetMinusUnit();

	// quark/ conjugeted quark charge factors
	CExpression* Qq = getModelContsSym(CONST_QU);
	CExpression* Qqcjg = getModelContsSym(CONST_QD);
	CExpression* QgP = getModelContsSym(CONST_GU_P);
	CExpression* QgM = getModelContsSym(CONST_GU_M);
	CExpression* QgPcjg = getModelContsSym(CONST_GD_P);
	CExpression* QgMcjg = getModelContsSym(CONST_GD_M);
	CExpression* Vqqc = new CESymbol("Vud");

	// model constants
	CExpression* e = getModelContsSym(CONST_E);
	CExpression* sW = getModelContsSym(CONST_SW);
	CExpression* cW = getModelContsSym(CONST_CW);
	CExpression* gs = getModelContsSym(CONST_G_S);
	CExpression* mH = get_mass_sym(CEParticle::PARTICLE_HIGGS);
	CExpression* mW = get_mass_sym(CEParticle::PARTICLE_W_BOSON);
	CExpression* mE = get_mass_sym(CEParticle::PARTICLE_ELECTRON);
	CExpression* mMu = get_mass_sym(CEParticle::PARTICLE_MUON);
	CExpression* mTau = get_mass_sym(CEParticle::PARTICLE_TAU_LEPTON);
	CExpression* mQd = get_mass_sym(CEParticle::PARTICLE_QUARK_D);
	CExpression* mQu = get_mass_sym(CEParticle::PARTICLE_QUARK_U);
	CExpression* mQs = get_mass_sym(CEParticle::PARTICLE_QUARK_S);
	CExpression* mQc = get_mass_sym(CEParticle::PARTICLE_QUARK_C);
	CExpression* mQb = get_mass_sym(CEParticle::PARTICLE_QUARK_B);
	CExpression* mQt = get_mass_sym(CEParticle::PARTICLE_QUARK_T);

	// vertices
	CEPtrArray* vert_info_list = getVerticesInfo();
	int next_vert_no = vert_info_list->size();

	CExpression* val = NULL;
	CSequence* syms_seq = NULL;
	bool RARE2_ALLOWED = true; //((m_phys_ID & PHYS_RARE2_VERICES) == PHYS_RARE2_VERICES);
	bool RARE3_ALLOWED = ((m_phys_ID & PHYS_RARE3_VERICES) == PHYS_RARE3_VERICES);
	CExpression* val_tmp = NULL; // for templates

	bool FeynGauge = true;
	if ((m_phys_ID & PHYS_UNITARY_GAUGE) == PHYS_UNITARY_GAUGE) {
		FeynGauge = false;
		if ((m_phys_ID & PHYS_GHOSTS) == PHYS_GHOSTS) {
			m_phys_ID = m_phys_ID^PHYS_GHOSTS; // dbg->log(0, "WARNING: GHOST vertices not initialized at UNITARY GAUGE");
		}
	}
	bool bPhysCT = ((m_phys_ID&PHYS_RENORM_COUNTERTERMS)==PHYS_RENORM_COUNTERTERMS);
	
	if (m_Hamiltonian_expr != NULL) { delete m_Hamiltonian_expr; m_Hamiltonian_expr = NULL; }
	if (m_Renorm_Hamiltonian != NULL) { delete m_Renorm_Hamiltonian; m_Renorm_Hamiltonian = NULL; }
	if (m_HamiltonianReplsRule != NULL) { delete m_HamiltonianReplsRule; m_HamiltonianReplsRule = NULL; }
	CERulesList* ham_r = new CERulesList();
	//////////////////////////////////////////////////////////////	prepare END

	dbg->log(7, m_phys_ID, "Initializing physics with ID: %x");

	///// TODO: check! why it was different PK order in addHamExprVFF() and addVertexInfo()

	////////////////////////////////////////////////////////////// 1. QED
	if (m_H_vect_index!=NULL) { delete m_H_vect_index; m_H_vect_index=NULL; }
	if (m_H_vect_index_2!=NULL) { delete m_H_vect_index_2; m_H_vect_index_2=NULL; }
	if (m_H_vect_index_3!=NULL) { delete m_H_vect_index_3; m_H_vect_index_3=NULL; }
	if (m_H_vect_index_4!=NULL) { delete m_H_vect_index_4; m_H_vect_index_4=NULL; }
	m_H_vect_index = CESymbol::getUnique();
	m_H_vect_index_2 = CESymbol::getUnique();
	m_H_vect_index_3 = CESymbol::getUnique();
	m_H_vect_index_4 = CESymbol::getUnique();

	//Note: we may use 'ham_vec_QED' for any FFV-vertices, but only once for every particle set
	CESymbol* ham_tmp1 = CESymbol::getUnique(false);
	CEVector* ham_vec_QED = new CEVector(ham_tmp1, m_H_vect_index, false); delete ham_tmp1;
	CESymbol* tmp_R = CESymbol::getUnique(false);
	CEVector* hvec_RN_QED = new CEVector(tmp_R, m_H_vect_index, false); delete tmp_R;

	CExpression* ham_vec_QED_subst = new CEGamma(m_H_vect_index, false);
	ham_r->addRule(createVecReplRule(ham_vec_QED, ham_vec_QED_subst)); delete ham_vec_QED_subst;

	CSequence _SymsSeq;

	dbg->log(75, "SM init step 1");

	//if (CESymbol::m_UseCharMap) CEException::halt();
	// electron vertex
	if (m_phys_ID & PHYS_PURE_QED) {
		addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_ELECTRON,CEParticle::PARTICLE_ELECTRON,false);
		addCorrRuleAFF(next_vert_no, i1); next_vert_no++;
		if(bPhysCT){ addHamExprVFF(next_vert_no,hvec_RN_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_ELECTRON,CEParticle::PARTICLE_ELECTRON,true); addCorrRuleGFF_CT(next_vert_no, i1); next_vert_no++; }
	}

	if (m_phys_ID & (PHYS_PURE_QED|PHYS_MUON)) { // muon vertex
		addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_MUON,CEParticle::PARTICLE_MUON,false);
		addCorrRuleAFF(next_vert_no, i1); next_vert_no++;
		if(bPhysCT){
			CEException::halt("Internal error in moun-leptons counter-terms implementation. Use model files.");
			//asdfg: Error here in 'addCorrRuleGFF_CT': it uses getRenormContsSym(DZ_E) instead of 'DZ_MU'
			addHamExprVFF(next_vert_no,hvec_RN_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_MUON,CEParticle::PARTICLE_MUON,true); addCorrRuleGFF_CT(next_vert_no, i1); next_vert_no++;
		}
	}

	if (m_phys_ID & (PHYS_PURE_QED|PHYS_TAU_LEPTON)) { // tau lepton vertex
		addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_TAU_LEPTON,CEParticle::PARTICLE_TAU_LEPTON,false);
		addCorrRuleAFF(next_vert_no, i1); next_vert_no++;
		if(bPhysCT){
			CEException::halt("Internal error in tau-leptons counter-terms implementation. Use model files.");
			//asdfg: Error here in 'addCorrRuleGFF_CT': it uses getRenormContsSym(DZ_E) instead of 'DZ_TAU'
			addHamExprVFF(next_vert_no,hvec_RN_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_TAU_LEPTON,CEParticle::PARTICLE_TAU_LEPTON,true); addCorrRuleGFF_CT(next_vert_no, i1); next_vert_no++;
		}
	}
	dbg->log(75, "SM init step 2");

	if (((m_phys_ID & PHYS_PURE_QED) == PHYS_PURE_QED) &&
		((m_phys_ID & PHYS_QUARKS_1GEN) == PHYS_QUARKS_1GEN)) {
		// Denner: -ieQg (here), Cheng,Lee: +ieQg
		//asdfg: check sign
		val = mIm->clone()->multiply(e->clone())->multiply(Qq->clone())->multiply(new CEGamma(i1, true));
		syms_seq = new CSequence(); syms_seq->addItem(i1->clone());
		addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_U,false);
		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++;
		if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
			addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_C,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++;
		}
		if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
			addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_T,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++;
		}
		delete val; val=NULL;

		// Denner: -ieQg (here), Cheng,Lee: +ieQg
		//asdfg: check sign
		val = mIm->clone()->multiply(e->clone())->multiply(Qqcjg->clone())->multiply(new CEGamma(i1, true));
		addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_QUARK_D,CEParticle::PARTICLE_QUARK_D,false);
		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++;
		if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
			addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_QUARK_S,CEParticle::PARTICLE_QUARK_S,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++;
		}
		if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
			addHamExprVFF(next_vert_no,ham_vec_QED,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_QUARK_B,CEParticle::PARTICLE_QUARK_B,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++;
		}
		delete val; val=NULL; delete syms_seq; syms_seq=NULL;
	}

	if (((m_phys_ID & PHYS_PURE_QED) == PHYS_PURE_QED) &&
		((m_phys_ID & PHYS_PROTON_STRUCTFUNC) == PHYS_PROTON_STRUCTFUNC)) {
		_SymsSeq.clear_all(); _SymsSeq.addItem(p1->clone()); _SymsSeq.addItem(i1->clone()); _SymsSeq.addItem(p2->clone()); _SymsSeq.addItem(p3->clone());
		CExpression* sfVec=CPProtonSF::GetGammaVertexSym(p1,p2,p3);
		val = Im->clone()->multiply(e->clone())->multiply(new CEVector(sfVec,i1,true));
		delete sfVec;
		CESymbol* tmp1=CESymbol::getUnique(false); CEVector* hVecX=new CEVector(tmp1,m_H_vect_index,false); delete tmp1;
		addHamExprVFF(next_vert_no,hVecX,m_H_vect_index,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_PROTON,CEParticle::PARTICLE_PROTON,false);
		m_SM_corr_rules->addVertexCorrRule(next_vert_no,val,&_SymsSeq,true,true); next_vert_no++;
		delete hVecX; delete val; val=NULL;
	}
	////////////////////////////////////////////////////////////// </QED>

	////////////////////////////////////////////////////////////// <QCD>
	if (m_phys_ID & PHYS_QCD) {
		dbg->log(75, "SM init step 2_qcd");
		//
		//key symbols: G[vector_index,t_index]
		_SymsSeq.clear_all(); _SymsSeq.addItem(i1->clone()); _SymsSeq.addItem(i_gm1->clone());
		//
		val = Im->clone()->multiply(gs->clone())->multiply(new CEGamma(i1,true))->multiply(new CEGellMann(i_gm1));
		//
		CESymbol* tmp1=CESymbol::getUnique(false); CEVector* hVecGQQ = new CEVector(tmp1,m_H_vect_index,false); delete tmp1;
		//guu
		addHamExprVFF(next_vert_no,hVecGQQ,m_H_vect_index,CEParticle::PARTICLE_GLUON,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_U,false);
		m_SM_corr_rules->addVertexCorrRule(next_vert_no,val,&_SymsSeq,false,false); next_vert_no++;
		//gdd
		addHamExprVFF(next_vert_no,hVecGQQ,m_H_vect_index,CEParticle::PARTICLE_GLUON,-CEParticle::PARTICLE_QUARK_D,CEParticle::PARTICLE_QUARK_D,false);
		m_SM_corr_rules->addVertexCorrRule(next_vert_no,val,&_SymsSeq,false,false); next_vert_no++;

		if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
			//gcc, gss
			addHamExprVFF(next_vert_no,hVecGQQ,m_H_vect_index,CEParticle::PARTICLE_GLUON,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_C,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no,val,&_SymsSeq,false,false); next_vert_no++;
			addHamExprVFF(next_vert_no,hVecGQQ,m_H_vect_index,CEParticle::PARTICLE_GLUON,-CEParticle::PARTICLE_QUARK_S,CEParticle::PARTICLE_QUARK_S,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no,val,&_SymsSeq,false,false); next_vert_no++;
		}
		if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
			//gtt, gbb
			addHamExprVFF(next_vert_no,hVecGQQ,m_H_vect_index,CEParticle::PARTICLE_GLUON,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_T,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no,val,&_SymsSeq,false,false); next_vert_no++;
			addHamExprVFF(next_vert_no,hVecGQQ,m_H_vect_index,CEParticle::PARTICLE_GLUON,-CEParticle::PARTICLE_QUARK_B,CEParticle::PARTICLE_QUARK_B,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no,val,&_SymsSeq,false,false); next_vert_no++;
		}
		delete hVecGQQ; delete val; val=NULL;

		//ggg
		//key symbols: G[{vector_index,t_index}x3]
		_SymsSeq.clear_all();
		_SymsSeq.addItem(i1->clone()); _SymsSeq.addItem(i_gm1->clone());
		_SymsSeq.addItem(i2->clone()); _SymsSeq.addItem(i_gm2->clone());
		_SymsSeq.addItem(i3->clone()); _SymsSeq.addItem(i_gm3->clone());
		//TODO: implement this! //asdfg

		//gggg
		//key symbols: G[{vector_index,t_index}x4]
		_SymsSeq.clear_all();
		_SymsSeq.addItem(i1->clone()); _SymsSeq.addItem(i_gm1->clone());
		_SymsSeq.addItem(i2->clone()); _SymsSeq.addItem(i_gm2->clone());
		_SymsSeq.addItem(i3->clone()); _SymsSeq.addItem(i_gm3->clone());
		_SymsSeq.addItem(i4->clone()); _SymsSeq.addItem(i_gm4->clone());
		//TODO: implement this! //asdfg
	}
	////////////////////////////////////////////////////////////// </QCD>

	// local deletes
	delete ham_vec_QED;
	delete hvec_RN_QED;

	dbg->log(75, "SM init step 3");
	////////////////////////////////////////////////////////////// Weak
	CESymbol* ham_tmp2 = CESymbol::getUnique(false);
	CEVector* ham_vec_Z = new CEVector(ham_tmp2, m_H_vect_index, false); delete ham_tmp2;

	// 2. Z-boson vertex info
	if (m_phys_ID & PHYS_Z_VERTICES) {
		// Zff
		addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
			CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_ELECTRON, CEParticle::PARTICLE_ELECTRON, false);
		
		// Zll vertex: +i*g_mu*(CgP*g(+)+CgM*g(-))  (see Denner'93)
		// CgP = -e sinW/cosW
		// CgM = e/(2 sinW cosW)-e sinW/cosW
		CExpression* const_gP = getModelContsSym(CONST_GP); CExpression* const_gM = getModelContsSym(CONST_GM);
		val = Im->clone()->multiply(new CEGamma(i1, true))
			->multiply( const_gP->clone()->multiply(new CEGammaSpecial(1))->add(const_gM->clone()->multiply(new CEGammaSpecial(-1))) );
		_SymsSeq.clear_all(); _SymsSeq.addItem(i1->clone());
		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
		next_vert_no++;

		if (m_phys_ID & PHYS_MUON) {
			addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
				CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_MUON, CEParticle::PARTICLE_MUON, false);
			
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
			next_vert_no++;
		}

		if (m_phys_ID & PHYS_TAU_LEPTON) {
			addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
				CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_TAU_LEPTON, CEParticle::PARTICLE_TAU_LEPTON, false);
			
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
			next_vert_no++;
		}
		delete val; val=NULL;
		
		// Znunu vertex: +i*e*g_mu*g(-)/(2*cW*sW)
		addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
			CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_NEUTRINO_ELECTRON, CEParticle::PARTICLE_NEUTRINO_ELECTRON, false);
		val = Im->clone()->multiply(e->clone())->multiply(new CEGamma(i1, true))->multiply(new CEGammaSpecial(-1))
			->divide(CENumber::GetTwo()->multiply(sW->clone())->multiply(cW->clone()));
		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq); next_vert_no++;

		if (m_phys_ID & PHYS_MUON) {
			addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
				CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_NEUTRINO_MUON, CEParticle::PARTICLE_NEUTRINO_MUON, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq); next_vert_no++;
		}

		if (m_phys_ID & PHYS_TAU_LEPTON) {
			addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
				CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_NEUTRINO_TAU, CEParticle::PARTICLE_NEUTRINO_TAU, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq); next_vert_no++;
		}
		delete val; val=NULL;

		dbg->log(75, "SM init step 3a");

		if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
			val = Im->clone()->multiply(new CEGamma(i1, true))->multiply(
				QgP->clone()->multiply(new CEGammaSpecial(1))->add(QgM->clone()->multiply(new CEGammaSpecial(-1))) );

			addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
				CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_QUARK_U, CEParticle::PARTICLE_QUARK_U, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
			next_vert_no++;

			if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
				addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
					CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_QUARK_C, CEParticle::PARTICLE_QUARK_C, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
				next_vert_no++;
			}
			if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
				addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
					CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_QUARK_T, CEParticle::PARTICLE_QUARK_T, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
				next_vert_no++;
			}
			delete val; val=NULL;

			val = Im->clone()->multiply(new CEGamma(i1, true))->multiply(
				QgPcjg->clone()->multiply(new CEGammaSpecial(1))->add(QgMcjg->clone()->multiply(new CEGammaSpecial(-1))) );

			addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
				CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_QUARK_D, CEParticle::PARTICLE_QUARK_D, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
			next_vert_no++;

			if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
				addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
					CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_QUARK_S, CEParticle::PARTICLE_QUARK_S, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
				next_vert_no++;
			}
			if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
				addHamExprVFF(next_vert_no, ham_vec_Z, m_H_vect_index,
					CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_QUARK_B, CEParticle::PARTICLE_QUARK_B, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, &_SymsSeq);
				next_vert_no++;
			}
			delete val; val=NULL;
		}
	}
	///////////////////////////////////////////////////////////
	delete ham_vec_Z;
	///////////////////////////////////////////////////////////

	dbg->log(75, "SM init step 4");
	// 4. W-boson vertex info
	// TODO: check out W to be W^{+} !!!
	CEVector* ham_vec_W = CEVector::CreateUnique(m_H_vect_index, false, false);

	if (m_phys_ID & PHYS_W_VERTICES) {

		// We\nu vertex
		addHamExprVFF(next_vert_no, ham_vec_W, m_H_vect_index,
			-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_ELECTRON, -CEParticle::PARTICLE_NEUTRINO_ELECTRON, false);
		// We\nu vertex: i*g*CgW*g(-)
		// CgW = e/( sinW sqrt(2) )
		// g(-) in our g_5-notation -- TODO: check it again
		val = Im->clone()->multiply(getModelContsSym(CONST_NEU)->clone())->multiply(new CEGamma(i1, true))->multiply(new CEGammaSpecial(-1));

		syms_seq = new CSequence(); syms_seq->addItem(i1->clone());
		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
		next_vert_no++;

		if (m_phys_ID & PHYS_MUON) {
			addHamExprVFF(next_vert_no, ham_vec_W, m_H_vect_index,
				-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_MUON, -CEParticle::PARTICLE_NEUTRINO_MUON, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
		}
		if (m_phys_ID & PHYS_TAU_LEPTON) {
			addHamExprVFF(next_vert_no, ham_vec_W, m_H_vect_index,
				-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_TAU_LEPTON, -CEParticle::PARTICLE_NEUTRINO_TAU, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
		}
		delete val; val=NULL;

		dbg->log(75, "SM init step 4a");

		if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
			val_tmp = Im->clone()->multiply(getModelContsSym(CONST_NEU)->clone())->multiply(new CEGamma(i1, true))->multiply(new CEGammaSpecial(-1));
			val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateDU());
			addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_D,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;

			if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateSU());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateDC());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateSC());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateBU());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateBC());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateDT());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateST());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateBT());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			}
			delete val_tmp;
		}

		dbg->log(75, "SM init step 4b");

		// -{W-boson} vertex
		addHamExprVFF(next_vert_no, ham_vec_W, m_H_vect_index,
			CEParticle::PARTICLE_W_BOSON, -CEParticle::PARTICLE_ELECTRON, CEParticle::PARTICLE_NEUTRINO_ELECTRON, false);
		// i*g*CgW*g(-) // g(-) in our g_5-notation (the same to +W-vertex)
		val = Im->clone()->multiply(getModelContsSym(CONST_NEU)->clone())->multiply(new CEGamma(i1, true))->multiply(new CEGammaSpecial(-1));

		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
		next_vert_no++;

		if (m_phys_ID & PHYS_MUON) {
			addHamExprVFF(next_vert_no, ham_vec_W, m_H_vect_index,
				CEParticle::PARTICLE_W_BOSON, -CEParticle::PARTICLE_MUON, CEParticle::PARTICLE_NEUTRINO_MUON, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
		}

		if (m_phys_ID & PHYS_TAU_LEPTON) {
			addHamExprVFF(next_vert_no, ham_vec_W, m_H_vect_index,
				CEParticle::PARTICLE_W_BOSON, -CEParticle::PARTICLE_TAU_LEPTON, CEParticle::PARTICLE_NEUTRINO_TAU, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
		}
		delete val;val=NULL;

		dbg->log(75, "SM init step 4c");

		if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
			// the same notation, Vqq* == Vqq
			val_tmp = Im->clone()->multiply(getModelContsSym(CONST_NEU)->clone())->multiply(new CEGamma(i1, true))->multiply(new CEGammaSpecial(-1));

			val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateDU());
			addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_U,-CEParticle::PARTICLE_QUARK_D,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateSU());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_U,-CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateDC());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_C,-CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateSC());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_C,-CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateBU());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_U,-CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateBC());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_C,-CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateDT());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_T,-CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateST());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_T,-CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(CQuarkMixMatrix::CreateBT());
				addHamExprVFF(next_vert_no,ham_vec_W,m_H_vect_index,CEParticle::PARTICLE_W_BOSON,CEParticle::PARTICLE_QUARK_T,-CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			}
			delete val_tmp;
		}
		delete syms_seq; syms_seq=NULL;

		dbg->log(75, "SM init step 4d");

		// W-\gamma vertex
		CESymbol* ham_AWW_tmp = CESymbol::getUnique();
		CExpression* AWW_core = new CERank3Tensor(ham_AWW_tmp,
			m_H_vect_index, false, m_H_vect_index_2, false, m_H_vect_index_3, false);
		delete ham_AWW_tmp;

		// vertex: A_i, W_i1, -W_i2
		Add3BosonVertex(next_vert_no, AWW_core, m_H_vect_index, m_H_vect_index_2, m_H_vect_index_3,
			CEParticle::PARTICLE_PHOTON, -CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, false);

		// -ieC (g_ab (2-1)_c + g_bc (3-2)_1 + g_ca (1-3)_b)
		// TODO: check out this!

		CExpression* vec_val = (
			(new CEMetric(i1, i2, true, true))->multiply(
			(new CEVector(p2, i3, true))->minus(new CEVector(p1, i3, true)))
			)->add(
			(new CEMetric(i2, i3, true, true))->multiply(
			(new CEVector(p3, i1, true))->minus(new CEVector(p2, i1, true)))
			)->add(
			(new CEMetric(i3, i1, true, true))->multiply(
			(new CEVector(p1, i2, true))->minus(new CEVector(p3, i2, true)))
			);
//		CExpression* vec_val =
//			(
//			(new CEMetric(i1, i2, true, true))->multiply(
//			(new CEVector(p2, i3, true))->minus(new CEVector(p1, i3, true)))
//			)->add
//			(
//			(new CEMetric(i2, i3, true, true))->multiply(
//			(new CEVector(p3, i1, true))->minus(new CEVector(p2, i1, true)))
//			)->add
//			(
//			(new CEMetric(i3, i1, true, true))->multiply(
//			(new CEVector(p1, i2, true))->minus(new CEVector(p3, i2, true)))
//			);
		vec_val = mIm->clone()->multiply(e->clone())->multiply(vec_val);
		val = vec_val->clone(); // * 1 without counterterms
//		//val->inverseFlag(); //For sign testing

		dbg->log(75, "SM init step 4e");

		syms_seq = new CSequence();
		syms_seq->addItem(p1->clone());
		syms_seq->addItem(i1->clone());
		syms_seq->addItem(p2->clone());
		syms_seq->addItem(i2->clone());
		syms_seq->addItem(p3->clone());
		syms_seq->addItem(i3->clone());

		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true);
		m_SM_corr_rules->m_X_rule_no = next_vert_no; // temp
		next_vert_no++;
		delete val;val=NULL;

		// WWZ vertex
		if (m_phys_ID & PHYS_Z_VERTICES) {
			CExpression* ZWW_base = CESymbol::getUnique();
			IETensor::GetIETensor(AWW_core)->setBase(ZWW_base); delete ZWW_base;
			Add3BosonVertex(next_vert_no, AWW_core, m_H_vect_index, m_H_vect_index_2, m_H_vect_index_3,
				CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, false);
			
			// -ieC (g_ab (2-1)_c + g_bc (3-2)_1 + g_ca (1-3)_b)
			// C==-ctgW 
			// NOTATION: Denner: -ctgW, Renton ctgW, Gross: ctgW, Chang-Lee: -ctgW (but -1 on every Z-vertex)
			// NOTATION: Denner (TODO: check it!)

			val = vec_val->clone()->multiply(getModelContsSym(CONST_CTGW)->clone())->multiply(mU->clone());
//			//val->inverseFlag(); //For sign testing
			
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true);
			next_vert_no++;
			delete val;val=NULL;
		}
		delete vec_val; delete syms_seq; syms_seq=NULL;
		////////////////////////////////
		delete AWW_core;
		////////////////////////////////

		dbg->log(75, "SM init step 4f");

		// WWXX - vertices
		// correspondence: ie^2C(2 g_ab g_cd - g_ac g_bd - g_ad g_bc)
		CExpression* vec4_val = (CENumber::GetTwo())
			->multiply(new CEMetric(i1, i2, true, true))->multiply(new CEMetric(i3, i4, true, true))
			->minus(
			(new CEMetric(i1, i3, true, true))->multiply(new CEMetric(i2, i4, true, true))
			)->minus(
			(new CEMetric(i1, i4, true, true))->multiply(new CEMetric(i2, i3, true, true))
			);
		vec4_val = Im->clone()->multiply(e->sqra())->multiply(vec4_val);

		CESymbol* ham_WWAA_base = CESymbol::getUnique();
		CExpression* WWXX_core = new CERank4Tensor(ham_WWAA_base,
			m_H_vect_index, false, m_H_vect_index_2, false, m_H_vect_index_3, false, m_H_vect_index_4, false);
		CExpression* WWAA_core = new CERank4Tensor(ham_WWAA_base,
			m_H_vect_index, false, m_H_vect_index_2, false, m_H_vect_index_3, false, m_H_vect_index_4, false, 0, 1); // symmetric by 3&4
		CExpression* WWmWmW_core = new CERank4Tensor(ham_WWAA_base,
			m_H_vect_index, false, m_H_vect_index_2, false, m_H_vect_index_3, false, m_H_vect_index_4, false, 1, 1); // symmetric by 1&2 and 3&4
		delete ham_WWAA_base;

		dbg->log(75, "SM init step 4g");

		// WWAA (W_i, -W_i1, A_i2, A_i3)
		val = vec4_val->clone()->multiply(mU->clone()); // * C==(-1) without counterterms
		bool flag_use_momenta = false;
		if (m_phys_ID & PHYS_ANOMALOUS_CONST) {
			// add WWAA-anomalus term (adding to feynman rule only) (from Igor)
			// order: W_1,-W_2,A_3,A_4
			// (e^2 / 8 Lambda^2) * {
			// 4 a_0 g^{12}((34)g^{34} - 3^4 4^3) + a_c (3^1 4^2+3^2 4^1) g^{34}
			// + (34) (g^{31} g^{42} + g^{41} g^{32}) - 3^4 (4^2 g^{31}+4^1 g^{32}) - 4^3 (3^2 g^{41}+3^1 g^{42})
			// + a'_0 g^{12} 3_x 4_y e^{3x4y}   )
			CExpression* a0 = new CESymbol("a\\_0");
			CExpression* aC = new CESymbol("a\\_c");
			CExpression* ahat0 = new CESymbol("ah\\_0");
			CExpression* Lam = new CESymbol("\\[CapitalLambda]");

			//new CEVector(p,i,true)
			//new CEMetric(i,i,true,true)
			//new CESProduct(p,p)
			//()->multiply()->add(()->multiply())
			CExpression* anom_val = (CENumber::GetFour())->multiply(a0->clone())->multiply(new CEMetric(i1, i2, true, true))->multiply(
				(new CESProduct(p3, p4))->multiply(new CEMetric(i3, i4, true, true))->minus((new CEVector(p3, i4, true))->multiply(new CEVector(p4, i3, true)))
				);
			anom_val=anom_val->add(
				aC->clone()->multiply(
				(new CEVector(p3, i1, true))->multiply(new CEVector(p4, i2, true))->add((new CEVector(p3, i2, true))->multiply(new CEVector(p4, i1, true)))
				)->multiply(new CEMetric(i3, i4, true, true))
				);
			anom_val=anom_val->add( aC->clone()->multiply(new CESProduct(p3,p4))->multiply(
				(new CEMetric(i3,i1,true,true))->multiply(new CEMetric(i4,i2,true,true))->add((new CEMetric(i4,i1,true,true))->multiply(new CEMetric(i3,i2,true,true)))
				));
			anom_val=anom_val->minus( aC->clone()->multiply(new CEVector(p3,i4,true))->multiply(
				(new CEVector(p4,i2,true))->multiply(new CEMetric(i3,i1,true,true))->add((new CEVector(p4,i1,true))->multiply(new CEMetric(i3,i2,true,true)))
				));
			anom_val=anom_val->minus( aC->clone()->multiply(new CEVector(p4,i3,true))->multiply(
				(new CEVector(p3,i2,true))->multiply(new CEMetric(i4,i1,true,true))->add((new CEVector(p3,i1,true))->multiply(new CEMetric(i4,i2,true,true)))
				));

			if ((m_phys_ID & PHYS_ANOMALOUS_CONST_NOLCH) == 0) {
				anom_val=anom_val->add( ahat0->clone()->multiply(new CEMetric(i1,i2,true,true))
					->multiply(new CEVector(p3,i_run1,false))->multiply(new CEVector(p4,i_run2,false))
					->multiply(new CEAntisymmetric(i3, true, i_run1, true, i4, true, i_run2, true)) );
			}

			// TODO: check if there required Im and what sign should be before it.
			anom_val=Im->clone()->multiply(e->sqra())->divide((CENumber::GetNumber(8))->multiply(Lam->sqra()))->multiply(anom_val);
			delete a0; delete aC; delete ahat0; delete Lam;
			
			// TODO: adjust logic or remove next line and uncomment the second one
			// we clear the SM AAWW vertex and replace it with anomalous contribution
			// the SM is hardly-separatable from (SM+anomal) ==> it should be calculated separately (by setting another physics)
			if (m_phys_ID & PHYS_ANOMALOUS_CONST_ONLY) {
				delete val; val = anom_val;
			} else {
				val = val->add(anom_val);
			}
			flag_use_momenta = true;
		}

		dbg->log(75, "SM init step 4h");

		// TODO: remove next after crossings problem fixing
		val = val->divide(CENumber::GetTwo());

		syms_seq = new CSequence();
		if (flag_use_momenta) {
			syms_seq->addItem(p1->clone());
			syms_seq->addItem(i1->clone());
			syms_seq->addItem(p2->clone());
			syms_seq->addItem(i2->clone());
			syms_seq->addItem(p3->clone());
			syms_seq->addItem(i3->clone());
			syms_seq->addItem(p4->clone());
			syms_seq->addItem(i4->clone());
		} else {
			syms_seq->addItem(i1->clone());
			syms_seq->addItem(i2->clone());
			syms_seq->addItem(i3->clone());
			syms_seq->addItem(i4->clone());
		}
		Add4BosonVertex(next_vert_no, WWAA_core, m_H_vect_index, m_H_vect_index_2, m_H_vect_index_3, m_H_vect_index_4,
			-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, CEParticle::PARTICLE_PHOTON, false);
		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, flag_use_momenta);
		next_vert_no++;
		delete val;val=NULL;

		dbg->log(75, "SM init step 4i");

		if (m_phys_ID & PHYS_Z_VERTICES) {
			// WWAZ
			// set new base expression
			CExpression* WWAZ_base = CESymbol::getUnique();
			IETensor::GetIETensor(WWXX_core)->setBase(WWAZ_base); delete WWAZ_base;
			Add4BosonVertex(next_vert_no, WWXX_core,
				m_H_vect_index, m_H_vect_index_2, m_H_vect_index_3, m_H_vect_index_4,
				-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, CEParticle::PARTICLE_Z_BOSON, false);
			// Correspondence: * (+ctgW) without counterterms
			// C==ctgW (Denner: ctgW, Renton -ctgW, Gross: -ctgW)
			// (Chang-Lee: -ctgW, but -1 on every Z-vertex)
			// NOTATION: Denner (TODO: check it!)
			val = vec4_val->clone()->multiply(getModelContsSym(CONST_CTGW)->clone());//->multiply(mU->clone());

			if (m_phys_ID & PHYS_ANOMALOUS_CONST) {
				// add WWAZ-anomalus term (adding to feynman rule only) (from Igor)
				// order: W_1,-W_2,A_3,Z_4
				// -(e^2 / 16 Lambda^2 cosW) * {
				//	a_n ( -(34) (g^31 g^24 - g^32 g^14) - (31) (g^32 g^14 - g^34 g^12) - (23) (g^34 g^12 - g^13 g^42)
				//		+ 4^3 (3^1 g^42 - 3^2 g^41) + 1^3 (3^2 g^41 - 3^4 g^12) + 2^3 (3^4 g^12 - 3^1 g^24)
				//		- g^34 (3^2 1^1 - 3^1 2^2) - g^31 (3^4 2^2 - 3^2 4^4) - g^32 (3^1 4^4 - 3^4 1^1) // can this line produce non-zero result?
				//		)
				//	+ahat_n 3_x (
				//		(3+4)^4 e^123x + (3+1)^1 e^243x + (3+2)^2 e^413x
				//		- (4-1)_y g^41 e^y23x - (1-2)_y g^12 e^y43x - (2-4)_y g^24 e^y13x
				//		)
				
				CExpression* an = new CESymbol("a\\_n");
				CExpression* ahat_n = new CESymbol("ahat\\_n");
				CExpression* Lam = new CESymbol("\\[CapitalLambda]");

				//TEMPLATES: new CEVector(p,i,true) //new CEMetric(i,i,true,true) //new CESProduct(p,p) //()->multiply()->add(()->multiply())
				CExpression* tmp_a = (new CESProduct(p3,p4))->multiply(
					(new CEMetric(i3,i1,true,true))->multiply(new CEMetric(i2,i4,true,true))->minus((new CEMetric(i2,i3,true,true))->multiply(new CEMetric(i1,i4,true,true)))
					);
				tmp_a->inverseFlag();
				tmp_a=tmp_a->minus( (new CESProduct(p3,p1))->multiply(
					(new CEMetric(i3,i2,true,true))->multiply(new CEMetric(i1,i4,true,true))->minus((new CEMetric(i3,i4,true,true))->multiply(new CEMetric(i1,i2,true,true)))
					) );
				tmp_a=tmp_a->minus( (new CESProduct(p2,p3))->multiply(
					(new CEMetric(i3,i4,true,true))->multiply(new CEMetric(i1,i2,true,true))->minus((new CEMetric(i1,i3,true,true))->multiply(new CEMetric(i4,i2,true,true)))
					) );
				tmp_a=tmp_a->add( (new CEVector(p4,i3,true))->multiply(
					(new CEVector(p3,i1,true))->multiply(new CEMetric(i4,i2,true,true))->minus((new CEVector(p3,i2,true))->multiply(new CEMetric(i4,i1,true,true)))
					) );
				tmp_a=tmp_a->add( (new CEVector(p1,i3,true))->multiply(
					(new CEVector(p3,i2,true))->multiply(new CEMetric(i4,i1,true,true))->minus((new CEVector(p3,i4,true))->multiply(new CEMetric(i1,i2,true,true)))
					) );
				tmp_a=tmp_a->add( (new CEVector(p2,i3,true))->multiply(
					(new CEVector(p3,i4,true))->multiply(new CEMetric(i1,i2,true,true))->minus((new CEVector(p3,i1,true))->multiply(new CEMetric(i2,i4,true,true)))
					) );
				tmp_a=tmp_a->minus( (new CEMetric(i3,i4,true,true))->multiply(
					(new CEVector(p3,i2,true))->multiply(new CEVector(p1,i1,true))->minus((new CEVector(p3,i1,true))->multiply(new CEVector(p2,i2,true)))
					) );
				tmp_a=tmp_a->minus( (new CEMetric(i3,i1,true,true))->multiply(
					(new CEVector(p3,i4,true))->multiply(new CEVector(p2,i2,true))->minus((new CEVector(p3,i2,true))->multiply(new CEVector(p4,i4,true)))
					) );
				tmp_a=tmp_a->minus( (new CEMetric(i3,i2,true,true))->multiply(
					(new CEVector(p3,i1,true))->multiply(new CEVector(p4,i4,true))->minus((new CEVector(p3,i4,true))->multiply(new CEVector(p1,i1,true)))
					) );
				CExpression* anom_val = an->clone()->multiply(tmp_a);

				//TEMP:
				//delete anom_val; anom_val=CENumber::GetZero();


				if ((m_phys_ID & PHYS_ANOMALOUS_CONST_NOLCH) == 0) {
					tmp_a= ((new CEVector(p3,i4,true))->add(new CEVector(p4,i4,true)))->multiply(new CEAntisymmetric(i1, true, i2, true, i3, true, i_run1, true));
					tmp_a=tmp_a->add( ((new CEVector(p3,i1,true))->add(new CEVector(p1,i1,true)))->multiply(new CEAntisymmetric(i2, true, i4, true, i3, true, i_run1, true)) );
					tmp_a=tmp_a->add( ((new CEVector(p3,i2,true))->add(new CEVector(p2,i2,true)))->multiply(new CEAntisymmetric(i4, true, i1, true, i3, true, i_run1, true)) );
					tmp_a=tmp_a->minus( ((new CEVector(p4,i_run2,false))->minus(new CEVector(p1,i_run2,false)))->multiply(new CEMetric(i4,i1,true,true))
						->multiply(new CEAntisymmetric(i_run2, true, i2, true, i3, true, i_run1, true) ) );
					tmp_a=tmp_a->minus( ((new CEVector(p1,i_run2,false))->minus(new CEVector(p2,i_run2,false)))->multiply(new CEMetric(i1,i2,true,true))
						->multiply(new CEAntisymmetric(i_run2, true, i4, true, i3, true, i_run1, true) ) );
					tmp_a=tmp_a->minus( ((new CEVector(p2,i_run2,false))->minus(new CEVector(p4,i_run2,false)))->multiply(new CEMetric(i2,i4,true,true))
						->multiply(new CEAntisymmetric(i_run2, true, i1, true, i3, true, i_run1, true) ) );

					tmp_a = ahat_n->clone()->multiply(new CEVector(p3,i_run1,false))->multiply(tmp_a);
					anom_val = anom_val->add(tmp_a);
				}


				// TODO: check if there required Im and what sign should be before it.
				tmp_a = e->sqra()->divide( (CENumber::GetNumber(16))->multiply(Lam->sqra())->multiply(getModelContsSym(CONST_CW)->clone()) );
				tmp_a->inverseFlag();
				anom_val=tmp_a->multiply(anom_val);

				delete an; delete ahat_n; delete Lam;
			
				// TODO: adjust logic or remove next line and uncomment the second one
				// we clear the SM AAWW vertex and replace it with anomalous contribution
				// the SM is hardly-separatable from (SM+anomal) ==> it should be calculated separately (by setting another physics)
				if (m_phys_ID & PHYS_ANOMALOUS_CONST_ONLY) {
					delete val; val = anom_val;
				} else {
					val = val->add(anom_val);
				}
			}

			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, flag_use_momenta);
			next_vert_no++;
			delete val;val=NULL;

			dbg->log(75, "SM init step 4j");
			
			// WWZZ
			// set new base expression
			CExpression* WWZZ_base = CESymbol::getUnique();
			IETensor::GetIETensor(WWAA_core)->setBase(WWZZ_base); delete WWZZ_base;
			Add4BosonVertex(next_vert_no, WWAA_core,
				m_H_vect_index, m_H_vect_index_2, m_H_vect_index_3, m_H_vect_index_4,
				-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_Z_BOSON, false);
			// Correspondence: * (-ctgW^2) without counterterms
			val = vec4_val->clone()->multiply(mU->clone())->multiply(getModelContsSym(CONST_CTGW)->sqra()); // C==-ctgW^2

			// TODO: remove next after crossings problem fixing
			val = val->divide(CENumber::GetTwo());

			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, flag_use_momenta);
			next_vert_no++;
			delete val;val=NULL;
		}

		dbg->log(75, "SM init step 4k");

		// WWWW
		CExpression* WWWW_base = CESymbol::getUnique();
		IETensor::GetIETensor(WWmWmW_core)->setBase(WWWW_base); delete WWWW_base;
		Add4BosonVertex(next_vert_no, WWmWmW_core,
			m_H_vect_index, m_H_vect_index_2, m_H_vect_index_3, m_H_vect_index_4,
			-CEParticle::PARTICLE_W_BOSON, -CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, false);
		// Correspondence: * (sW^(-2)) without counterterms
		val = vec4_val->clone()->divide(getModelContsSym(CONST_SW)->sqra()); // C==1/sW^2

		// TODO: remove next after crossings problem fixing
		val = val->divide(CENumber::GetFour());

		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, flag_use_momenta);
		next_vert_no++;
		delete val;val=NULL;

		//////////////////////////////////////////////////////////
		delete vec4_val;
		delete syms_seq; syms_seq=NULL;
		delete WWXX_core; delete WWAA_core; delete WWmWmW_core;
		//////////////////////////////////////////////////////////
	} //if (m_phys_ID & PHYS_W_VERTICES)
	///////////////////////////////////////////////////////////
	delete ham_vec_W;
	///////////////////////////////////////////////////////////

	dbg->log(75, "SM init step 5");

	////////////////////////////////////////////////////////////// Higgs & Scalars
	if (m_phys_ID & PHYS_SCALARS) {

		// SSS-couplings <BEGIN>
		val_tmp = mIm->clone()->multiply(e->clone());
		val_tmp = val_tmp->multiply(mH->sqra()->divide( CENumber::GetTwo()->multiply(sW->clone())->multiply(mW->clone()) ));
		syms_seq = new CSequence();
		if (RARE3_ALLOWED) {
			// HHH
			val = val_tmp->clone()->multiply(CENumber::GetThree());
			val = val->divide(CENumber::GetNumber(6)); // TODO: remove this after crossings problem fixing
			Add3BosonVertex(next_vert_no, NULL, NULL, NULL, NULL,
				CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
			delete val;val=NULL;

			if (FeynGauge) {
				// HChiChi
				val = val_tmp->clone();
				val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
				Add3BosonVertex(next_vert_no, NULL, NULL, NULL, NULL,
					CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;

				// HFiFi
				val = val_tmp->clone();
				Add3BosonVertex(next_vert_no, NULL, NULL, NULL, NULL,
					CEParticle::PARTICLE_HIGGS, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;
			}
		}
		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// SSS-couplings <END>

		dbg->log(75, "SM init step 5a");

		CESymbol* VSX_tmp = CESymbol::getUnique();
		CExpression* VSX_core = new CEVector(VSX_tmp, m_H_vect_index, false);
		delete VSX_tmp;

		// VSS-couplings <BEGIN>
		val_tmp = Im->clone()->multiply(e->clone());
		CExpression* p2m3 = p2->clone()->minus(p3->clone());
		val_tmp = val_tmp->multiply(new CEVector(p2m3, i1, true));
		delete p2m3;
		syms_seq = new CSequence();
		syms_seq->addItem(p1->clone()); syms_seq->addItem(i1->clone());
		syms_seq->addItem(p2->clone()); // S
		syms_seq->addItem(p3->clone()); // S
		if ((RARE2_ALLOWED) && (FeynGauge)) {
			// AChiH
			// Zero + CT

			if (m_phys_ID & PHYS_Z_VERTICES) { // ZChiH
				val = val_tmp->clone()->multiply(mIm->clone())->divide(
					CENumber::GetTwo()->multiply(sW->clone())->multiply(cW->clone()));
				Add3BosonVertex(next_vert_no, VSX_core, m_H_vect_index, NULL, NULL,
					CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_HIGGS, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);
				next_vert_no++;
				delete val;val=NULL;
			}

			// AFi{-Fi}
			val = val_tmp->clone()->multiply(mU->clone());
			Add3BosonVertex(next_vert_no, VSX_core, m_H_vect_index, NULL, NULL,
				CEParticle::PARTICLE_PHOTON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);
			next_vert_no++;
			delete val;val=NULL;

			if (m_phys_ID & PHYS_Z_VERTICES) { // ZFi{-Fi}
				val = val_tmp->clone()->multiply(mU->clone())->multiply(sW->sqra()->minus(cW->sqra()))
					->divide(CENumber::GetTwo()->multiply(sW->clone())->multiply(cW->clone()));
				Add3BosonVertex(next_vert_no, VSX_core, m_H_vect_index, NULL, NULL,
					CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);
				next_vert_no++;
				delete val;val=NULL;
			}
			if (m_phys_ID & PHYS_W_VERTICES) {
				// W{-Fi}H and inversion
				val = val_tmp->clone()->multiply(mU->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
				Add3BosonVertex(next_vert_no, VSX_core, m_H_vect_index, NULL, NULL,
					-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_HIGGS, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);
				next_vert_no++;
				delete val;val=NULL;
				val = val_tmp->clone()->divide(CENumber::GetTwo()->multiply(sW->clone()));
				Add3BosonVertex(next_vert_no, VSX_core, m_H_vect_index, NULL, NULL,
					CEParticle::PARTICLE_W_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_HIGGS, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);
				next_vert_no++;
				delete val;val=NULL;
				// W{-Fi}Chi and inversion
				val = val_tmp->clone()->multiply(mIm->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
				Add3BosonVertex(next_vert_no, VSX_core, m_H_vect_index, NULL, NULL,
					-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_CHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);
				next_vert_no++;
				Add3BosonVertex(next_vert_no, VSX_core, m_H_vect_index, NULL, NULL,
					CEParticle::PARTICLE_W_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_CHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);
				next_vert_no++;
				delete val;val=NULL;
			}
		}
		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// VSS-couplings <END>
		dbg->log(75, "SM init step 5b");

		CESymbol* VV_tmp = CESymbol::getUnique();
		// Note: VVS/VVSS_core-matrix is considered SYMMETRIC ALWAYS! (also in classes-create procedure)
		CExpression* VVX_core = new CEMatrix(VV_tmp, m_H_vect_index, m_H_vect_index_2, false, false, 1);
		delete VV_tmp;

		// VVS-couplings <BEGIN>
		val_tmp = Im->clone()->multiply(e->clone())->multiply(new CEMetric(i1, i2, true, true));
		val_tmp = val_tmp->multiply(mW->clone());
		syms_seq = new CSequence();
		syms_seq->addItem(i1->clone()); syms_seq->addItem(i2->clone());

		if (m_phys_ID & PHYS_W_VERTICES) { // W{-W}H
			val = val_tmp->clone()->divide(sW->clone());
			Add3BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL,
				-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_HIGGS, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
			delete val;val=NULL;
		}
		if (m_phys_ID & PHYS_Z_VERTICES) {
			// ZZH
			val = val_tmp->clone()->divide( sW->clone()->multiply(cW->sqra()) );
			val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
			Add3BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL,
				CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_HIGGS, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
			delete val;val=NULL;
			// ZAH
			// Zero + CT
		}

		if (FeynGauge) {
			if (m_phys_ID & PHYS_W_VERTICES) { //{-W}AFi and inversion
				Add3BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL,
					CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, -CEParticle::PARTICLE_SCALAR_PHI, false);
				val = val_tmp->clone()->multiply(mU->clone());
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				Add3BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL,
					-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_W_VERTICES) && (m_phys_ID & PHYS_Z_VERTICES)) { //{-W}ZFi and inversion
				Add3BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL,
					CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, false);
				val = val_tmp->clone()->multiply(mU->clone())->multiply(sW->clone())->divide(cW->clone());
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				Add3BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL,
					-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;
			}
		}

		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// VVS-couplings <END>
		dbg->log(75, "SM init step 5c");

		// SSSS-couplings <BEGIN>
		val_tmp = mIm->clone()->multiply(e->sqra());
		val_tmp = val_tmp->multiply(mH->sqra()->divide( CENumber::GetFour()->multiply(sW->sqra())->multiply(mH->sqra()) ));
		syms_seq = new CSequence();
		if (RARE3_ALLOWED) {
			// HHHH
			val = val_tmp->clone()->multiply(CENumber::GetThree());
			val = val->divide(CENumber::GetNumber(24)); // TODO: remove this after crossings problem fixing
			Add4BosonVertex(next_vert_no, NULL, NULL, NULL, NULL, NULL,
				CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
			next_vert_no++;
			delete val;val=NULL;

			if (FeynGauge) {
				// HHChiChi
				val = val_tmp->clone();
				val = val->divide(CENumber::GetFour()); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, NULL, NULL, NULL, NULL, NULL,
					CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;

				// HHFiFi
				val = val_tmp->clone();
				val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, NULL, NULL, NULL, NULL, NULL,
					CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;

				// ChiChiChiChi
				val = val_tmp->clone()->multiply(CENumber::GetThree());
				val = val->divide(CENumber::GetNumber(24)); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, NULL, NULL, NULL, NULL, NULL,
					CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;

				// ChiChiFiFi
				val = val_tmp->clone();
				val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, NULL, NULL, NULL, NULL, NULL,
					CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;

				// FiFiFiFi
				val = val_tmp->clone()->multiply(CENumber::GetTwo());
				val = val->divide(CENumber::GetFour()); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, NULL, NULL, NULL, NULL, NULL,
					-CEParticle::PARTICLE_SCALAR_PHI, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;
			}
		}
		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// SSSS-couplings <END>
		dbg->log(75, "SM init step 5d");

		// VVSS-couplings <BEGIN>
		val_tmp = Im->clone()->multiply(e->sqra())->multiply(new CEMetric(i1, i2, true, true));
		syms_seq = new CSequence();
		syms_seq->addItem(i1->clone()); syms_seq->addItem(i2->clone());
		if (RARE2_ALLOWED) {
			if (m_phys_ID & PHYS_W_VERTICES) {
				// WWHH
				val = val_tmp->clone()->divide(CENumber::GetTwo()->multiply(sW->sqra()));
				val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
					-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;

				if (FeynGauge) {
					// WWChiChi
					val = val_tmp->clone()->divide(CENumber::GetTwo()->multiply(sW->sqra()));
					val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
					// WWFiFi
					val = val_tmp->clone()->divide(CENumber::GetTwo()->multiply(sW->sqra()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_W_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
				}
			}
			if (m_phys_ID & PHYS_Z_VERTICES) {
				if (FeynGauge) {
					// ZZFiFi
					val = val_tmp->clone();
					val = val->multiply( (sW->sqra()->minus(cW->sqra()))->sqra() )
						->divide( CENumber::GetTwo()->multiply(sW->sqra())->multiply(cW->sqra()) );
					val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
					// ZAFiFi
					val = val_tmp->clone();
					val = val->multiply(sW->sqra()->minus(cW->sqra()))->divide(sW->clone()->multiply(cW->clone()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_PHOTON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
				}
			}
			if (FeynGauge) {
				// AAFiFi
				val = val_tmp->clone()->multiply(CENumber::GetTwo());
				val = val->divide(CENumber::GetTwo()); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
					CEParticle::PARTICLE_PHOTON, CEParticle::PARTICLE_PHOTON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_PHI, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;
			}
			if (m_phys_ID & PHYS_Z_VERTICES) {
				// ZZHH
				val = val_tmp->clone()->divide(CENumber::GetTwo()->multiply(sW->sqra())->multiply(cW->sqra()));
				val = val->divide(CENumber::GetFour()); // TODO: remove this after crossings problem fixing
				Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
					CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_HIGGS, CEParticle::PARTICLE_HIGGS, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
				next_vert_no++;
				delete val;val=NULL;
				// ZAHH
				// Zero + CT
				if (FeynGauge) {
					// ZZChiChi
					val = val_tmp->clone()->divide(CENumber::GetTwo()->multiply(sW->sqra())->multiply(cW->sqra()));
					val = val->divide(CENumber::GetFour()); // TODO: remove this after crossings problem fixing
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_SCALAR_CHI, CEParticle::PARTICLE_SCALAR_CHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
					// ZAChiChi
					// Zero + CT
				}
			}
			if ((m_phys_ID & PHYS_W_VERTICES) && (m_phys_ID & PHYS_Z_VERTICES)) {
				if (FeynGauge) {
					// WZ{-Fi}H and inversion
					val = val_tmp->clone()->multiply(mU->clone())->divide(CENumber::GetTwo()->multiply(cW->clone()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_HIGGS, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_HIGGS, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
					// WZ{-Fi}Chi and inversion
					val = val_tmp->clone()->multiply(mIm->clone())->divide(CENumber::GetTwo()->multiply(cW->clone()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_Z_BOSON, CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_CHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
					val = val_tmp->clone()->multiply(Im->clone())->divide(CENumber::GetTwo()->multiply(cW->clone()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_Z_BOSON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_CHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
				}
			}
			if (m_phys_ID & PHYS_W_VERTICES) {
				if (FeynGauge) {
					// WA{-Fi}H and inversion
					val = val_tmp->clone()->multiply(mU->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_HIGGS, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_HIGGS, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
					// WA{-Fi}Chi and inversion
					val = val_tmp->clone()->multiply(mIm->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						-CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_CHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
					val = val_tmp->clone()->multiply(Im->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
					Add4BosonVertex(next_vert_no, VVX_core, m_H_vect_index, m_H_vect_index_2, NULL, NULL,
						CEParticle::PARTICLE_W_BOSON, CEParticle::PARTICLE_PHOTON, -CEParticle::PARTICLE_SCALAR_PHI, CEParticle::PARTICLE_SCALAR_CHI, false);
					m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);
					next_vert_no++;
					delete val;val=NULL;
				}
			}
		}
		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// VVSS-couplings <END>
		dbg->log(75, "SM init step 5e");

		// SFF-couplings <BEGIN>
		val_tmp = Im->clone()->multiply(e->clone())->divide(sW->clone()->multiply(mW->clone()));
		CExpression* FiFF_tmp =  Im->clone()->multiply(getModelContsSym(CONST_NEU)->clone())->divide(mW->clone());
		syms_seq = new CSequence();
		// Hff, neutrinos means massless (no HNu{-Nu} vertex)
		val = val_tmp->clone()->multiply(mU->clone())->multiply(mE->clone())->divide(CENumber::GetTwo());
		addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_ELECTRON,CEParticle::PARTICLE_ELECTRON,false);
		m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
		if (m_phys_ID & PHYS_MUON) {
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mMu->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_MUON,CEParticle::PARTICLE_MUON,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
		}
		if (m_phys_ID & PHYS_TAU_LEPTON) {
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mTau->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_TAU_LEPTON,CEParticle::PARTICLE_TAU_LEPTON,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
		}
		if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mQd->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_QUARK_D,CEParticle::PARTICLE_QUARK_D,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mQu->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_U,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
		}
		if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mQs->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_QUARK_S,CEParticle::PARTICLE_QUARK_S,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mQc->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_C,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
		}
		if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mQb->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_QUARK_B,CEParticle::PARTICLE_QUARK_B,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			val = val_tmp->clone()->multiply(mU->clone())->multiply(mQt->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_T,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
		}

		if (FeynGauge) {
			// Chiff, neutrinos means massless (no ChiNu{-Nu} vertex)
			val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(mIm->clone())->multiply(mE->clone())->divide(CENumber::GetTwo());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_ELECTRON,CEParticle::PARTICLE_ELECTRON,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			if (m_phys_ID & PHYS_MUON) {
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(mIm->clone())->multiply(mMu->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_MUON,CEParticle::PARTICLE_MUON,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if (m_phys_ID & PHYS_TAU_LEPTON) {
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(mIm->clone())->multiply(mTau->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_TAU_LEPTON,CEParticle::PARTICLE_TAU_LEPTON,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(Im->clone())->multiply(mQu->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_U,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(mIm->clone())->multiply(mQd->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_QUARK_D,CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(Im->clone())->multiply(mQs->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_QUARK_S,CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(mIm->clone())->multiply(mQc->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_C,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(Im->clone())->multiply(mQb->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_QUARK_B,CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val = val_tmp->clone()->multiply(new CEGammaSpecial(0))->multiply(mIm->clone())->multiply(mQt->clone())->divide(CENumber::GetTwo());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_T,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}

			// Fi{-u_i}d_j
			if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateDU());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQu->clone())->minus((new CEGammaSpecial(1))->multiply(mQd->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateSU());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQu->clone())->minus((new CEGammaSpecial(1))->multiply(mQs->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateDC());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQc->clone())->minus((new CEGammaSpecial(1))->multiply(mQd->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateSC());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQc->clone())->minus((new CEGammaSpecial(1))->multiply(mQs->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateBU());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQu->clone())->minus((new CEGammaSpecial(1))->multiply(mQb->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_U,CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateBC());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQc->clone())->minus((new CEGammaSpecial(1))->multiply(mQb->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_C,CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateDT());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQt->clone())->minus((new CEGammaSpecial(1))->multiply(mQd->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateST());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQt->clone())->minus((new CEGammaSpecial(1))->multiply(mQs->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateBT());
				val=val->multiply((new CEGammaSpecial(-1))->multiply(mQt->clone())->minus((new CEGammaSpecial(1))->multiply(mQb->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_QUARK_T,CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}

			// {-Fi}u_i{-d_j}
			if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateDU());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQu->clone())->minus((new CEGammaSpecial(-1))->multiply(mQd->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_U,-CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateSU());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQu->clone())->minus((new CEGammaSpecial(-1))->multiply(mQs->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_U,-CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateDC());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQc->clone())->minus((new CEGammaSpecial(-1))->multiply(mQd->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_C,-CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateSC());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQc->clone())->minus((new CEGammaSpecial(-1))->multiply(mQs->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_C,-CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateBU());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQu->clone())->minus((new CEGammaSpecial(-1))->multiply(mQb->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_U,-CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateBC());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQc->clone())->minus((new CEGammaSpecial(-1))->multiply(mQb->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_C,-CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateDT());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQt->clone())->minus((new CEGammaSpecial(-1))->multiply(mQd->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_T,-CEParticle::PARTICLE_QUARK_D,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateST());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQt->clone())->minus((new CEGammaSpecial(-1))->multiply(mQs->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_T,-CEParticle::PARTICLE_QUARK_S,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
				val=FiFF_tmp->clone()->multiply(CQuarkMixMatrix::CreateBT());
				val=val->multiply((new CEGammaSpecial(1))->multiply(mQt->clone())->minus((new CEGammaSpecial(-1))->multiply(mQb->clone())));
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_QUARK_T,-CEParticle::PARTICLE_QUARK_B,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}

			// Fi{-Nu}l
			val = FiFF_tmp->clone()->multiply(new CEGammaSpecial(1))->multiply(mU->clone())->multiply(mE->clone());
			addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_ELECTRON,-CEParticle::PARTICLE_NEUTRINO_ELECTRON,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			if (m_phys_ID & PHYS_MUON) {
				val = FiFF_tmp->clone()->multiply(new CEGammaSpecial(1))->multiply(mU->clone())->multiply(mMu->clone());
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_MUON,-CEParticle::PARTICLE_NEUTRINO_MUON,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if (m_phys_ID & PHYS_TAU_LEPTON) {
				val = FiFF_tmp->clone()->multiply(new CEGammaSpecial(1))->multiply(mU->clone())->multiply(mTau->clone());
				addHamExprVFF(next_vert_no,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,CEParticle::PARTICLE_TAU_LEPTON,-CEParticle::PARTICLE_NEUTRINO_TAU,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}

			// {-Fi}{-l}Nu
			val = FiFF_tmp->clone()->multiply(new CEGammaSpecial(-1))->multiply(mU->clone())->multiply(mE->clone());
			addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_ELECTRON,CEParticle::PARTICLE_NEUTRINO_ELECTRON,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			if (m_phys_ID & PHYS_MUON) {
				val = FiFF_tmp->clone()->multiply(new CEGammaSpecial(-1))->multiply(mU->clone())->multiply(mMu->clone());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_MUON,CEParticle::PARTICLE_NEUTRINO_MUON,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
			if (m_phys_ID & PHYS_TAU_LEPTON) {
				val = FiFF_tmp->clone()->multiply(new CEGammaSpecial(-1))->multiply(mU->clone())->multiply(mTau->clone());
				addHamExprVFF(next_vert_no,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_TAU_LEPTON,CEParticle::PARTICLE_NEUTRINO_TAU,false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq); next_vert_no++; delete val;val=NULL;
			}
		}
		delete FiFF_tmp;
		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// SFF-couplings <END>
		delete VVX_core;
		delete VSX_core;
	}

	dbg->log(75, "SM init step 6");
	////////////////////////////////////////////////////////////// Ghosts
	// TODO: change Ghosts to be FERMIONS
	if (m_phys_ID & PHYS_GHOSTS) {
		CESymbol* VSX_tmp = CESymbol::getUnique();
		CExpression* VSX_core = new CEVector(VSX_tmp, m_H_vect_index, false);
		delete VSX_tmp;
		// V{-G}G-couplings <BEGIN>
		// TODO: Chack sing of p2 (momentum seems to be outgoing in Denner'93)
		val_tmp = Im->clone()->multiply(e->clone())->multiply(new CEVector(p2, i1, true));
		syms_seq = new CSequence();
		syms_seq->addItem(p1->clone()); syms_seq->addItem(i1->clone()); //V
		syms_seq->addItem(p2->clone()); //G
		syms_seq->addItem(p3->clone()); //G
		if (RARE2_ALLOWED) {
			// AGWpGWp and AGWmGWm
			val = val_tmp->clone();
			Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UWp, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
			val = val_tmp->clone()->multiply(mU->clone());
			Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_PHOTON,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UWm, false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;

			if (m_phys_ID & PHYS_Z_VERTICES) { // ZGWpGWp and ZGWmGWm
				val = val_tmp->clone()->multiply(mU->clone())->multiply(cW->clone())->divide(sW->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_Z_BOSON,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UWp, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(cW->clone())->divide(sW->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_Z_BOSON,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UWm, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
			}

			if ((m_phys_ID & PHYS_W_VERTICES) && (m_phys_ID & PHYS_Z_VERTICES)) {
				// W{-GWp}Gz and {-W}{-GWm}Gz
				val = val_tmp->clone()->multiply(cW->clone())->divide(sW->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UZ, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(mU->clone())->multiply(cW->clone())->divide(sW->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UZ, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
				// W{-Gz}GWm and {-W}{-Gz}GWp
				val = val_tmp->clone()->multiply(mU->clone())->multiply(cW->clone())->divide(sW->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UZ,CEParticle::PARTICLE_FPGHOST_UWm, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(cW->clone())->divide(sW->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UZ,CEParticle::PARTICLE_FPGHOST_UWp, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
			}

			if (m_phys_ID & PHYS_W_VERTICES) {
				// W{-GWp}Ga and {-W}{-GWm}Ga
				val = val_tmp->clone()->multiply(mU->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UA, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone();
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UA, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
				// W{-Ga}GWm and {-W}{-Ga}GWp
				val = val_tmp->clone();
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,-CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UA,CEParticle::PARTICLE_FPGHOST_UWm, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
				val = val_tmp->clone()->multiply(mU->clone());
				Add3BosonVertex(next_vert_no,VSX_core,m_H_vect_index,NULL,NULL,CEParticle::PARTICLE_W_BOSON,-CEParticle::PARTICLE_FPGHOST_UA,CEParticle::PARTICLE_FPGHOST_UWp, false);
				m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq, true,true);next_vert_no++;delete val;val=NULL;
			}
		}
		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// VGG-couplings <END>
		delete VSX_core;
		dbg->log(75, "SM init step 6a");

		// SGG-couplings <BEGIN>
		val_tmp = Im->clone()->multiply(e->clone());
		syms_seq = new CSequence();
		if ((RARE3_ALLOWED) && (m_phys_ID & PHYS_SCALARS) && (FeynGauge)) {
			// H{-Gz}Gz
			val=val_tmp->clone()->multiply(mU->clone())->multiply(mW->clone())
				->divide( sW->clone()->multiply(cW->sqra())->multiply(CENumber::GetTwo()) );
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_FPGHOST_UZ,CEParticle::PARTICLE_FPGHOST_UZ,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;

			// H{-GWp}GWp and H{-GWm}GWm
			val=val_tmp->clone()->multiply(mU->clone())->multiply(mW->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UWp,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_HIGGS,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UWm,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;
			delete val;val=NULL;
			
			// Chi{-GWp}GWp and Chi{-GWm}GWm
			val=val_tmp->clone()->multiply(mIm->clone())->multiply(mW->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UWp,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			val=val_tmp->clone()->multiply(Im->clone())->multiply(mW->clone())->divide(CENumber::GetTwo()->multiply(sW->clone()));
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_SCALAR_CHI,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UWm,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;

			// Fi{-Gz}GWm and {-Fi}{-Gz}GWp
			val=val_tmp->clone()->multiply(mW->clone())->divide(CENumber::GetTwo()->multiply(sW->clone())->multiply(cW->clone()));
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_FPGHOST_UZ,CEParticle::PARTICLE_FPGHOST_UWm,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			val=val_tmp->clone()->multiply(mW->clone())->divide(CENumber::GetTwo()->multiply(sW->clone())->multiply(cW->clone()));
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_FPGHOST_UZ,CEParticle::PARTICLE_FPGHOST_UWp,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			
			// Fi{-GWp}Gz and {-Fi}{-GWm}Gz
			val=val_tmp->clone()->multiply(mW->clone())->multiply(sW->sqra()->minus(cW->sqra()))->divide(CENumber::GetTwo()->multiply(sW->clone())->multiply(cW->clone()));
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UZ,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			val=val_tmp->clone()->multiply(mW->clone())->multiply(sW->sqra()->minus(cW->sqra()))->divide(CENumber::GetTwo()->multiply(sW->clone())->multiply(cW->clone()));
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UZ,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			
			// Fi{-Ga}GWm and {-Fi}{-Ga}GWp
			// TODO: check out, why it's not in Denner'93 ?????

			// Fi{-GWp}Ga and {-Fi}{-GWm}Ga
			val=val_tmp->clone()->multiply(mW->clone());
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,-CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_FPGHOST_UWp,CEParticle::PARTICLE_FPGHOST_UA,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
			val=val_tmp->clone()->multiply(mW->clone());
			Add3BosonVertex(next_vert_no,NULL,NULL,NULL,NULL,CEParticle::PARTICLE_SCALAR_PHI,-CEParticle::PARTICLE_FPGHOST_UWm,CEParticle::PARTICLE_FPGHOST_UA,false);
			m_SM_corr_rules->addVertexCorrRule(next_vert_no, val, syms_seq);next_vert_no++;delete val;val=NULL;
		}
		delete val_tmp; delete syms_seq; syms_seq=NULL;
		// SGG-couplings <END>
	}

	////////////////////////////////////////////////////////
	dbg->log(75, "SM init step 7");

	// particle EXTERNAL lines
	// fermions
	val = new CESpinor(p, false);
	syms_seq = new CSequence(); syms_seq->addItem(p->clone());
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_ELECTRON, val, syms_seq);
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_NEUTRINO_ELECTRON, val, syms_seq);
	if (m_phys_ID & PHYS_MUON) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_MUON, val, syms_seq);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_NEUTRINO_MUON, val, syms_seq);
	}
	if (m_phys_ID & PHYS_TAU_LEPTON) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_TAU_LEPTON, val, syms_seq);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_NEUTRINO_TAU, val, syms_seq);
	}
	if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_D, val, syms_seq);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_U, val, syms_seq);

		if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_S, val, syms_seq);
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_C, val, syms_seq);
		}
		if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_B, val, syms_seq);
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_T, val, syms_seq);
		}
	}
	delete val;val=NULL;

	// conjugated particles
	val = new CESpinor(p, true);
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_ELECTRON, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_NEUTRINO_ELECTRON, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
	if (m_phys_ID & PHYS_MUON) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_MUON, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_NEUTRINO_MUON, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
	}
	if (m_phys_ID & PHYS_TAU_LEPTON) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_TAU_LEPTON, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_NEUTRINO_TAU, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
	}
	if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_D, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_U, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);

		if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_S, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_C, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
		}
		if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_B, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_QUARK_T, val, syms_seq, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
		}
	}
	delete val;val=NULL;
	delete syms_seq; syms_seq=NULL;

	// Proton described by proton tensor
	if ((m_phys_ID & PHYS_PROTON_STRUCTFUNC) == PHYS_PROTON_STRUCTFUNC) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_PROTON, &unit, NULL);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_PROTON, &unit, NULL, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
	}

	// bosons
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_PHOTON, &unit, NULL);
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_Z_BOSON, &unit, NULL);
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_W_BOSON, &unit, NULL);
	m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_W_BOSON, &unit, NULL, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);

	////////////////////////////////////////////////////////////// Higgs
	if (m_phys_ID & PHYS_SCALARS) {
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_HIGGS, &unit, NULL);
		m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_HIGGS, &unit, NULL, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);

		if (FeynGauge) {
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_SCALAR_CHI, &unit, NULL);
			m_SM_corr_rules->addExtLineCorrRule(CEParticle::PARTICLE_SCALAR_CHI, &unit, NULL, CDiagCorrRules::CORR_RULE_FLAG_CONJUGATED);
		}
	}

	dbg->log(75, "SM init step 8");

	// particle INTERNAL lines
	// fermions
	val = Im->clone()->multiply((new CESGammaProduct(p))->add(m->clone()))->divide((new CESProduct(p, p))->minus(m->sqra()));
	syms_seq = new CSequence(); syms_seq->addItem(p->clone()); syms_seq->addItem(m->clone());
	m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_ELECTRON, val, syms_seq);
	//if (m_phys_ID & PHYS_W_VERTICES) {
	m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_NEUTRINO_ELECTRON, val, syms_seq);
	//}
	if (m_phys_ID & PHYS_MUON) {
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_MUON, val, syms_seq);
		//if (m_phys_ID & PHYS_W_VERTICES) {
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_NEUTRINO_MUON, val, syms_seq);
		//}
	}
	if (m_phys_ID & PHYS_TAU_LEPTON) {
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_TAU_LEPTON, val, syms_seq);
		//if (m_phys_ID & PHYS_W_VERTICES) {
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_NEUTRINO_TAU, val, syms_seq);
		//}
	}
	if ((m_phys_ID & PHYS_QUARKS_1GEN)==PHYS_QUARKS_1GEN) {
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_QUARK_D, val, syms_seq);
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_QUARK_U, val, syms_seq);

		if ((m_phys_ID & PHYS_QUARKS_2GEN) == PHYS_QUARKS_2GEN) {
			m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_QUARK_S, val, syms_seq);
			m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_QUARK_C, val, syms_seq);
		}
		if ((m_phys_ID & PHYS_QUARKS_3GEN) == PHYS_QUARKS_3GEN) {
			m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_QUARK_B, val, syms_seq);
			m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_QUARK_T, val, syms_seq);
		}
	}
	// Proton described by proton tensor: no double interation => ZERO internal line
	if ((m_phys_ID & PHYS_PROTON_STRUCTFUNC) == PHYS_PROTON_STRUCTFUNC) {
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_PROTON, &_Zero, syms_seq);
	}
	
	dbg->log(75, "SM init step 8a");
	// val is now fermion propagator
	if (bPhysCT) {
		//Note: adjust AreCTPropagatorsVertices() after moving next lines to 2-tale vertices
		// TODO: check counterterms usage (+ possible make it as 2-tale vertex)

		// Note: e^2 is moved into constant
		// CT = G*i{ (p-m)*(d_f + d_fP*gP + d_fM*gM) - (d_m + d_mP*gP + d_mM*gM) }*G
		CExpression* ferm_CT = GetFermCTDeclaration(CEParticle::PARTICLE_ELECTRON, p, m);
		ferm_CT = (val->clone())->multiply(ferm_CT)->multiply(val->clone());
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_ELECTRON, ferm_CT, syms_seq, CDiagCorrRules::CORR_RULE_COUNTER_TERM);
		delete ferm_CT;

		if (m_phys_ID & PHYS_MUON) {
			ferm_CT = GetFermCTDeclaration(CEParticle::PARTICLE_MUON, p, m);
			ferm_CT = (val->clone())->multiply(ferm_CT)->multiply(val->clone());
			m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_MUON, ferm_CT, syms_seq, CDiagCorrRules::CORR_RULE_COUNTER_TERM);
			delete ferm_CT;
		}
		
		if (m_phys_ID & PHYS_TAU_LEPTON) {
			ferm_CT = GetFermCTDeclaration(CEParticle::PARTICLE_TAU_LEPTON, p, m);
			ferm_CT = (val->clone())->multiply(ferm_CT)->multiply(val->clone());
			m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_TAU_LEPTON, ferm_CT, syms_seq, CDiagCorrRules::CORR_RULE_COUNTER_TERM);
			delete ferm_CT;
		}

		//Required 0-counterterm value to avoid internal errors during diagrams generation
		if ((m_phys_ID & PHYS_PROTON_STRUCTFUNC) == PHYS_PROTON_STRUCTFUNC) {
			m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_PROTON, &_Zero, syms_seq, CDiagCorrRules::CORR_RULE_COUNTER_TERM);
		}
	}
	delete val; val=NULL; delete syms_seq; syms_seq=NULL;

	dbg->log(75, "SM init step 8b");
	// photons
	val = mIm->clone()->multiply(new CEMetric(i1, i2, false, false))->divide(new CESProduct(p, p));
	syms_seq = new CSequence();
	syms_seq->addItem(p->clone()); syms_seq->addItem(m->clone());
	syms_seq->addItem(i1->clone()); syms_seq->addItem(i2->clone());
	m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_PHOTON, val, syms_seq);

	if (bPhysCT) {
		//Note: adjust AreCTPropagatorsVertices() after moving next lines to 2-tale vertices
		// TODO: check it
		CExpression* val2=val->clone(); val2=CEReplasement::Replace(val2,i1,i3); val2=CEReplasement::Replace(val2,i2,i4);
		CExpression* prop_CT =val->clone()
// (-i)
			->multiply(mIm->clone())->multiply(
// (+i)
//Incorrect! See also SM.mod and loadmodel problem			->multiply(Im->clone())->multiply(
// non-transverse:
			new CEMetric(i2, i3, true, true))->multiply(new CESProduct(p, p)
// transverse:
//			(new CEMetric(i2,i3,true,true))->multiply(new CESProduct(p,p))->minus((new CEVector(p,i2,true))->multiply(new CEVector(p,i3,true)))
			)->multiply(getRenormContsSym(DZ_AA)->clone())
			->multiply(val2);
		syms_seq->clear_all();
		syms_seq->addItem(p->clone()); syms_seq->addItem(m->clone());
		syms_seq->addItem(i1->clone()); syms_seq->addItem(i4->clone());
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_PHOTON, prop_CT, syms_seq, CDiagCorrRules::CORR_RULE_COUNTER_TERM);
	}
	delete val; val=NULL; delete syms_seq; syms_seq=NULL;

	// bosons
	//if (m_phys_ID & PHYS_Z_VERTICES) {
	if (true) {
		CExpression* A_prop_numerator = new CEMetric(i1, i2, false, false);
		if (m_phys_ID & PHYS_UNITARY_GAUGE) {
			A_prop_numerator = A_prop_numerator->minus((new CEVector(p, i1, false))->multiply(new CEVector(p, i2, false))->divide(new CEPower(m, (long)2)));
		}
		val = mIm->clone()->multiply(A_prop_numerator)->divide((new CESProduct(p, p))->minus(new CEPower(m, (long)2)));
		syms_seq = new CSequence();
		syms_seq->addItem(p->clone()); syms_seq->addItem(m->clone());
		syms_seq->addItem(i1->clone()); syms_seq->addItem(i2->clone());
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_Z_BOSON, val, syms_seq);
		delete val; val=NULL; delete syms_seq; syms_seq=NULL;
	}

	//if (m_phys_ID & PHYS_W_VERTICES) {
	if (true) {
		CExpression* A_prop_numerator = (new CEMetric(i1, i2, false, false));
		if (m_phys_ID & PHYS_UNITARY_GAUGE) {
			A_prop_numerator = A_prop_numerator->minus((new CEVector(p, i1, false))->multiply(new CEVector(p, i2, false))->divide(new CEPower(m, (long)2)));
			//TEMP:
			//A_prop_numerator = (new CEVector(p, i1, false))->multiply(new CEVector(p, i2, false))->divide(new CEPower(m, (long)2));
		}
		val = mIm->clone()->multiply(A_prop_numerator)->divide((new CESProduct(p, p))->minus(new CEPower(m, (long)2)));
		syms_seq = new CSequence();
		syms_seq->addItem(p->clone()); syms_seq->addItem(m->clone());
		syms_seq->addItem(i1->clone()); syms_seq->addItem(i2->clone());
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_W_BOSON, val, syms_seq);
		delete val; val=NULL; delete syms_seq; syms_seq=NULL;
	}

	//gluon
	if ((m_phys_ID & PHYS_QCD) == PHYS_QCD) {
		CTIndex* tindA=CTIndex::CreateMatrixIndex(i_gm1,8); CTIndex* tindB=CTIndex::CreateMatrixIndex(i_gm2,8);
		val = mIm->clone()->multiply(new CEMetric(i1,i2,false,false))->multiply(new CEMetric(tindA,tindB))->divide(new CESProduct(p, p));
		if(tindA!=NULL)delete tindA; if(tindB!=NULL)delete tindB;
		_SymsSeq.clear_all();
		_SymsSeq.addItem(p->clone()); _SymsSeq.addItem(m->clone());
		_SymsSeq.addItem(i1->clone()); _SymsSeq.addItem(i_gm1->clone());
		_SymsSeq.addItem(i2->clone()); _SymsSeq.addItem(i_gm2->clone());
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_GLUON, val, &_SymsSeq);
		delete val; val=NULL;
	}

	////////////////////////////////////////////////////////////// Higgs
	if (m_phys_ID & PHYS_SCALARS) { // Higgs & scalars
		val = Im->clone()->divide((new CESProduct(p, p))->minus(new CEPower(m, (long)2)));
		syms_seq = new CSequence(); syms_seq->addItem(p->clone()); syms_seq->addItem(m->clone());
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_HIGGS, val, syms_seq);

		if (m_phys_ID & PHYS_UNITARY_GAUGE) { delete val; val=CENumber::GetZero(); } // TODO: check this
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_SCALAR_PHI, val, syms_seq);
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_SCALAR_CHI, val, syms_seq); // same (mass differs only)
		delete val; val=NULL; delete syms_seq; syms_seq=NULL;
	}

	////////////////////////////////////////////////////////////// Ghosts
	if (m_phys_ID & PHYS_GHOSTS) {
		val = Im->clone()->divide((new CESProduct(p, p))->minus(new CEPower(m, (long)2)));
		if (m_phys_ID & PHYS_UNITARY_GAUGE) { delete val; val=CENumber::GetZero(); } // TODO: check this
		syms_seq = new CSequence(); syms_seq->addItem(p->clone()); syms_seq->addItem(m->clone());
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_FPGHOST_UA, val, syms_seq);
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_FPGHOST_UZ, val, syms_seq); // same (mass differs only)
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_FPGHOST_UWp, val, syms_seq); // same (mass differs only)
		m_SM_corr_rules->addIntLineCorrRule(CEParticle::PARTICLE_FPGHOST_UWm, val, syms_seq); // same (mass differs only)
		delete val; val=NULL; delete syms_seq; syms_seq=NULL;
	}

	dbg->log(75, "SM init step 9");

	//////////////////////////////////////////////////////////////////// end section
	m_HamiltonianReplsRule = ham_r;
	delete i3; delete i4;
	delete p1; delete p2; delete p3; delete p4;
	delete Im; delete mIm; delete mU;
	delete i_run1; delete i_run2;
	delete Vqqc;
	////////////////////////////////////////////////////////////////////

	dbg->log(25, "Hamiltonian initialized", m_Hamiltonian_expr);
	dbg->log(75, "Hamiltonian replacements are", m_HamiltonianReplsRule);
	if (dbg->is_log_keeped(50)) { // check hamiltinian rule
		CExpression* h_actual = m_Hamiltonian_expr->clone(); h_actual = m_HamiltonianReplsRule->ApplyTo(h_actual);
		dbg->log(20, "Actual hamiltonian is", h_actual); delete h_actual;
	}
	
	// note: correspondence rules logging requires physics to be already installed
	//dbg->log(70, "Correspondence rules", m_SM_corr_rules);
}

/*
*	CT-declaration formuila
*	CT = i{ (p-m)*(d_f + d_fP*gP + d_fM*gM) - (d_m + d_mP*gP + d_mM*gM) }
*/
CExpression* CSMPhysics::GetFermCTDeclaration(int pk, CExpression* p, CExpression* m) {
	CExpression* dF_term = NULL; CExpression* dM_term = NULL;

	if ((m_phys_ID & PHYS_NOQED_COUNTERTERMS) == 0) {
		CExpression* dZ = NULL; CExpression* dM = NULL;
		if (pk == CEParticle::PARTICLE_ELECTRON) {
			dZ = getRenormContsSym(DZ_F_EL); dM = getRenormContsSym(DM_F_EL);
		} else if (pk == CEParticle::PARTICLE_MUON) {
			dZ = getRenormContsSym(DZ_F_MU); dM = getRenormContsSym(DM_F_MU);
		} else if (pk == CEParticle::PARTICLE_TAU_LEPTON) {
			dZ = getRenormContsSym(DZ_F_TAU); dM = getRenormContsSym(DM_F_TAU);
		} else { CEException::halt(dbg, "invalid PK"); }
		if ((dZ == NULL) || (dM == NULL)) { CEException::halt(dbg, "NULL dZ || dM"); }
		
		dF_term = dZ->clone();
		dM_term = dM->clone();
	}
	
	if (m_phys_ID & (PHYS_Z_VERTICES | PHYS_W_VERTICES)) {
		CExpression* dZ_P = NULL; CExpression* dM_P = NULL;
		CExpression* dZ_M = NULL; CExpression* dM_M = NULL;

		if (pk == CEParticle::PARTICLE_ELECTRON) {
			dZ_P = getRenormContsSym(DZ_F_EL_P); dM_P = getRenormContsSym(DM_F_EL_P);
			dZ_M = getRenormContsSym(DZ_F_EL_M); dM_M = getRenormContsSym(DM_F_EL_M);
		} else if (pk == CEParticle::PARTICLE_MUON) {
			dZ_P = getRenormContsSym(DZ_F_MU_P); dM_P = getRenormContsSym(DM_F_MU_P);
			dZ_M = getRenormContsSym(DZ_F_MU_M); dM_M = getRenormContsSym(DM_F_MU_M);
		} else if (pk == CEParticle::PARTICLE_TAU_LEPTON) {
			dZ_P = getRenormContsSym(DZ_F_TAU_P); dM_P = getRenormContsSym(DM_F_TAU_P);
			dZ_M = getRenormContsSym(DZ_F_TAU_M); dM_M = getRenormContsSym(DM_F_TAU_M);
		} else { CEException::halt(dbg, "invalid PK"); }
		if ((dZ_P == NULL) || (dM_P == NULL) ||
			(dZ_M == NULL) || (dM_M == NULL)) { CEException::halt(dbg, "NULL dZ || dM (P/M)"); }
		
		CExpression* dF5 = (new CEGammaSpecial(1))->multiply(dZ_P->clone());
		dF5 = dF5->add((new CEGammaSpecial(-1))->multiply(dZ_M->clone()));
		
		CExpression* dM5 = (new CEGammaSpecial(1))->multiply(dM_P->clone());
		dM5 = dM5->add((new CEGammaSpecial(-1))->multiply(dM_M->clone()));
		
		if (dF_term != NULL) { dF_term = dF_term->add(dF5); dM_term = dM_term->add(dM5); }
		else { dF_term = dF5; dM_term = dM5; }
	}
	
	if (dF_term == NULL) dF_term = CENumber::GetZero();
	if (dM_term == NULL) dM_term = CENumber::GetZero();
	
	CExpression* ferm_CT = ((new CESGammaProduct(p))->minus(m->clone()))->multiply(dF_term);
	ferm_CT = ferm_CT->minus(dM_term);
	ferm_CT = ferm_CT->multiply(CENumber::GetImUnit());
	dbg->log(25, pk, "ferm_CT declaration created (pk == %ld)", ferm_CT);
	return ferm_CT;
}

void CSMPhysics::addVertexInfo(int v_num, int pk1, int pk2, int pk3, bool is_renorm_CT) {
	CEPtrArray* vert_info_list = getVerticesInfo();
	if (vert_info_list->size() != v_num) { // check number
		dbg->log(0, v_num, vert_info_list->size(), "ERROR vertices enumeration: number %ld at vetices list size %ld"); CEException::halt();
	}

	CDiagramVertexInfo* vi = NULL;
	if (!is_renorm_CT) {
		vi = new CDiagramVertexInfo(v_num, pk1, pk2, pk3);
	} else {
		vi = CDiagramVertexInfo::newCountertermVertex(v_num, pk1, pk2, pk3);
	}
	dbg->log(55, v_num, "Vertex Info (no. %ld): ", vi);
	// total charge check (self-check)
	if ((getCharge3(pk1)!=INT_MIN) && (getCharge3(pk2)!=INT_MIN) && (getCharge3(pk3)!=INT_MIN)) {
		if (getCharge3(pk1)+getCharge3(pk2)+getCharge3(pk3) != 0) {
			dbg->log(0,"Warning: Charge-check failed for vertex",vi);
			dbg->log(0,getCharge3(pk1),getCharge3(pk2),getCharge3(pk3),"Charges*3: %ld, %ld, %ld");
		}
	}
	vert_info_list->add(vi);
}

void CSMPhysics::addVertex4Info(int v_num, int pk1, int pk2, int pk3, int pk4, bool is_renorm_CT) {
	CEPtrArray* vert_info_list = getVerticesInfo();
	if (vert_info_list->size() != v_num) { // check number
		dbg->log(0, v_num, vert_info_list->size(), "ERROR vertices enumeration: number %ld at vetices list size %ld"); CEException::halt();
	}

	CDiagramVertexInfo* vi = NULL;
	if (!is_renorm_CT) {
		vi = new CDiagramVertexInfo(v_num, pk1, pk2, pk3, pk4);
	} else {
		dbg->log(0, "ERROR: counterterms are not implemented for VVVV-vertex");
		//vi = CDiagramVertexInfo::newCountertermVertex(v_num, pk1, pk2, pk3);
	}
	dbg->log(55, v_num, "VVVV-vertex Info (no. %ld): ", vi);

	// total charge check (self-check)
	if ((getCharge3(pk1)!=INT_MIN) && (getCharge3(pk2)!=INT_MIN) && (getCharge3(pk3)!=INT_MIN) && (getCharge3(pk4)!=INT_MIN)) {
		if (getCharge3(pk1)+getCharge3(pk2)+getCharge3(pk3)+getCharge3(pk4) != 0) { dbg->log(0,"Warning: Charge-check failed for vertex",vi); }
	}
	vert_info_list->add(vi);
}

/*
*	vec_core is "gamma_{\mu}"-like expression.
*	NULL vector index means SCALAR particle
*	NULL vector core means UNIT factor (for SCALAR particles)
*/
void CSMPhysics::addHamExprVFF(int v_num,
							   CExpression* vec_core, CExpression* vect_index,
							   int pk_a, int pk_f1, int pk_f2, bool is_renorm_CT) {
	bool bIsClassCpl=false;
	if ((CLMath::abs(pk_a) & CEParticle::PARTICLE_CLASS_FLAG) && (CLMath::abs(pk_f1) & CEParticle::PARTICLE_CLASS_FLAG) && (CLMath::abs(pk_f2) & CEParticle::PARTICLE_CLASS_FLAG)) {
		bIsClassCpl=true;
		if (!m_bClassVerticesComplete) return;
	} else if ((CLMath::abs(pk_a) & CEParticle::PARTICLE_CLASS_FLAG) || (CLMath::abs(pk_f1) & CEParticle::PARTICLE_CLASS_FLAG) || (CLMath::abs(pk_f2) & CEParticle::PARTICLE_CLASS_FLAG)) {
		if (m_bClassVerticesComplete) {
			m_bClassVerticesComplete=false;
			CEString tmp("Warning: class diagrams not available. internal error in addHamExprVFF. PKs: ");
			tmp.add_int(pk_a,true); tmp.AddString(","); tmp.add_int(pk_f1,true); tmp.AddString(","); tmp.add_int(pk_f2,true);
			dbg->log(0, tmp.getData());
		}
	}

	// first fermion PK should be < 0, second > 0.
	if (pk_f1 > 0) {
		if (pk_f2 > 0) { dbg->log(0, pk_f1, pk_f2, "addHamExprVFF (pk1: %ld, pk2: %ld) invalid call"); CEException::halt(); }
		int pk_tmp = pk_f2;
		pk_f2 = pk_f1;
		pk_f1 = pk_tmp;
	} else if (pk_f2 < 0) { dbg->log(0, pk_f1, pk_f2, "addHamExprVFF (pk1: %ld, pk2: %ld) invalid call"); CEException::halt(); }

	// vector index check
	bool bVec_a=((CLMath::abs(pk_a) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	if (vect_index == NULL) { if(bVec_a){ dbg->log(0, v_num, "addHamExprVFF error. Vector index missed for vertex no %ld."); CEException::halt(); } }
	else { if(!bVec_a){ dbg->log(0, v_num, "addHamExprVFF error. Unexpected vector index for vertex no %ld."); CEException::halt(); } }

	CExpression* ham_val = new COpFermionField(pk_f1, v_num);
	if (vec_core != NULL) { ham_val = ham_val->multiply(vec_core->clone()); }
	ham_val = ham_val->multiply(new COpBosonField(pk_a, vect_index, true, v_num));
	ham_val = ham_val->multiply(new COpFermionField(pk_f2, v_num));

	if (bIsClassCpl) {
		if (!is_renorm_CT) {
			if (m_eClassesHamiltonian==NULL) m_eClassesHamiltonian=ham_val; else m_eClassesHamiltonian=m_eClassesHamiltonian->add(ham_val);
		} else {
			if (m_eClassesHamiltonianR==NULL) m_eClassesHamiltonianR=ham_val; else m_eClassesHamiltonianR=m_eClassesHamiltonianR->add(ham_val);
		}
		m_idClassVertices.SetAt(v_num,-1); // to be re-filled by caller afterwards
	} else {
		if (!is_renorm_CT) {
			if (m_Hamiltonian_expr==NULL) m_Hamiltonian_expr=ham_val; else m_Hamiltonian_expr=m_Hamiltonian_expr->add(ham_val);
		} else {
			if (m_Renorm_Hamiltonian==NULL) m_Renorm_Hamiltonian=ham_val; else m_Renorm_Hamiltonian=m_Renorm_Hamiltonian->add(ham_val);
		}
		
		long nClPos=InvolveVertexClass(pk_a, pk_f1, pk_f2, 0, is_renorm_CT, v_num, 1);
		m_idClassVertices.SetAt(v_num,nClPos);
	}
	addVertexInfo(v_num, pk_a, pk_f1, pk_f2, is_renorm_CT);
}

// Note vect_index_vX == NULL means SCALAR particle here
// NULL vec_core means UNIT (no vector indices, all SCALAR particles)
void CSMPhysics::Add3BosonVertex(int v_num, CExpression* vec_core,
							   CExpression* vect_index_a,
							   CExpression* vect_index_v1, CExpression* vect_index_v2,
							   int pk_a, int pk_v1, int pk_v2, bool is_renorm_CT) {
	bool bIsClassCpl=false;
	if ((CLMath::abs(pk_a) & CEParticle::PARTICLE_CLASS_FLAG) && (CLMath::abs(pk_v1) & CEParticle::PARTICLE_CLASS_FLAG) && (CLMath::abs(pk_v2) & CEParticle::PARTICLE_CLASS_FLAG)) {
		bIsClassCpl=true;
		if (!m_bClassVerticesComplete) return;
	} else if ((CLMath::abs(pk_a) & CEParticle::PARTICLE_CLASS_FLAG) || (CLMath::abs(pk_v1) & CEParticle::PARTICLE_CLASS_FLAG) || (CLMath::abs(pk_v2) & CEParticle::PARTICLE_CLASS_FLAG)) {
		if (m_bClassVerticesComplete) {
			m_bClassVerticesComplete=false;
			CEString tmp("Warning: class diagrams not available. internal error in Add3BosonVertex. PKs: ");
			tmp.add_int(pk_a,true); tmp.AddString(","); tmp.add_int(pk_v1,true); tmp.AddString(","); tmp.add_int(pk_v2,true);
			dbg->log(0, tmp.getData());
		}
	}

	// vector indices check
	bool bVecIndOK=true;
	bool bVec_a=((CLMath::abs(pk_a) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	bool bVec_v1=((CLMath::abs(pk_v1) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	bool bVec_v2=((CLMath::abs(pk_v2) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	if (vect_index_a == NULL) { if(bVec_a){ dbg->log(0, v_num, "Add3BosonVertex error. Vector index missed for vertex no %ld."); bVecIndOK=false; } }
	else { if(!bVec_a){ dbg->log(0, v_num, "Internal error. Add3BosonVertex vector index for vertex no %ld."); bVecIndOK=false; } }
	if (vect_index_v1 == NULL) { if(bVec_v1){ dbg->log(0, v_num, "Add3BosonVertex error. Vector index missed for vertex no %ld."); bVecIndOK=false; } }
	else { if(!bVec_v1){ dbg->log(0, v_num, "Add3BosonVertex error. Unexpected vector index for vertex no %ld."); bVecIndOK=false; } }
	if (vect_index_v2 == NULL) { if(bVec_v2){ dbg->log(0, v_num, "Add3BosonVertex error. Vector index missed for vertex no %ld."); bVecIndOK=false; } }
	else { if(!bVec_v2){ dbg->log(0, v_num, "Add3BosonVertex error. Unexpected vector index for vertex no %ld."); bVecIndOK=false; } }
	if (!bVecIndOK) { CEException::halt(); }

	CExpression* ham_val = new COpBosonField(pk_a, vect_index_a, true, v_num);
	ham_val = ham_val->multiply(new COpBosonField(pk_v1, vect_index_v1, true, v_num));
	ham_val = ham_val->multiply(new COpBosonField(pk_v2, vect_index_v2, true, v_num));
	if (vec_core != NULL) {
		ham_val = ham_val->multiply(vec_core->clone());
	}

	if (bIsClassCpl) {
		if (!is_renorm_CT) {
			if (m_eClassesHamiltonian==NULL) m_eClassesHamiltonian=ham_val; else m_eClassesHamiltonian=m_eClassesHamiltonian->add(ham_val);
		} else {
			if (m_eClassesHamiltonianR==NULL) m_eClassesHamiltonianR=ham_val; else m_eClassesHamiltonianR=m_eClassesHamiltonianR->add(ham_val);
		}
		m_idClassVertices.SetAt(v_num,-1);
	} else {
		if (!is_renorm_CT) {
			if (m_Hamiltonian_expr==NULL) m_Hamiltonian_expr=ham_val; else m_Hamiltonian_expr=m_Hamiltonian_expr->add(ham_val);
		} else {
			if (m_Renorm_Hamiltonian==NULL) m_Renorm_Hamiltonian=ham_val; else m_Renorm_Hamiltonian=m_Renorm_Hamiltonian->add(ham_val);
		}
		
		long nClPos = InvolveVertexClass(pk_a, pk_v1, pk_v2, 0, is_renorm_CT, v_num, 2);
		m_idClassVertices.SetAt(v_num,nClPos);
	}
	addVertexInfo(v_num, pk_a, pk_v1, pk_v2, is_renorm_CT);
}

// Note vect_index_vX == NULL means SCALAR particle here
// NULL vec_core means UNIT (no vector indices, all SCALAR particles)
void CSMPhysics::Add4BosonVertex(int v_num, CExpression* vec_core,
								CExpression* vect_index_v1,
								CExpression* vect_index_v2,
								CExpression* vect_index_v3,
								CExpression* vect_index_v4,
								int pk_v1, int pk_v2,
								int pk_v3, int pk_v4,
								bool is_renorm_CT) {
	bool bIsClassCpl=false;
	if ((CLMath::abs(pk_v1)&CEParticle::PARTICLE_CLASS_FLAG)&&(CLMath::abs(pk_v2)&CEParticle::PARTICLE_CLASS_FLAG)&&(CLMath::abs(pk_v3)&CEParticle::PARTICLE_CLASS_FLAG)&&(CLMath::abs(pk_v4)&CEParticle::PARTICLE_CLASS_FLAG)) {
		bIsClassCpl=true;
		if (!m_bClassVerticesComplete) return;
	} else if ((CLMath::abs(pk_v1)&CEParticle::PARTICLE_CLASS_FLAG)||(CLMath::abs(pk_v2)&CEParticle::PARTICLE_CLASS_FLAG)||(CLMath::abs(pk_v3)&CEParticle::PARTICLE_CLASS_FLAG)||(CLMath::abs(pk_v4)&CEParticle::PARTICLE_CLASS_FLAG)) {
		if (m_bClassVerticesComplete) {
			m_bClassVerticesComplete=false; dbg->log(0, "Warning: class diagrams not available. internal error in Add4BosonVertex.");
		}
	}

	// vector indices check
	bool bVec_v1=((CLMath::abs(pk_v1) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	bool bVec_v2=((CLMath::abs(pk_v2) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	bool bVec_v3=((CLMath::abs(pk_v3) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	bool bVec_v4=((CLMath::abs(pk_v4) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	if (vect_index_v1 == NULL) { if(bVec_v1){ dbg->log(0, v_num, pk_v1, "Add4BosonVertex error. Vector index missed for vertex no %ld. pk: %x"); CEException::halt(); } }
	else { if(!bVec_v1){ dbg->log(0, v_num, "Add4BosonVertex error. Unexpected vector index for vertex no %ld."); CEException::halt(); } }
	if (vect_index_v2 == NULL) { if(bVec_v2){ dbg->log(0, v_num, pk_v2,"Add4BosonVertex error. Vector index missed for vertex no %ld. pk: %x"); CEException::halt(); } }
	else { if(!bVec_v2){ dbg->log(0, v_num, "Add4BosonVertex error. Unexpected vector index for vertex no %ld."); CEException::halt(); } }
	if (vect_index_v3 == NULL) { if(bVec_v3){ dbg->log(0, v_num, pk_v3, "Add4BosonVertex error. Vector index missed for vertex no %ld. pk: %x"); CEException::halt(); } }
	else { if(!bVec_v3){ dbg->log(0, v_num, "Add4BosonVertex error. Unexpected vector index for vertex no %ld."); CEException::halt(); } }
	if (vect_index_v4 == NULL) { if(bVec_v4){ dbg->log(0, v_num, pk_v4, "Add4BosonVertex error. Vector index missed for vertex no %ld. pk: %x"); CEException::halt(); } }
	else { if(!bVec_v4){ dbg->log(0, v_num, "Add4BosonVertex error. Unexpected vector index for vertex no %ld."); CEException::halt(); } }

	CExpression* ham_val = new COpBosonField(pk_v1, vect_index_v1, true, v_num);
	ham_val = ham_val->multiply(new COpBosonField(pk_v2, vect_index_v2, true, v_num));
	ham_val = ham_val->multiply(new COpBosonField(pk_v3, vect_index_v3, true, v_num));
	ham_val = ham_val->multiply(new COpBosonField(pk_v4, vect_index_v4, true, v_num));
	if (vec_core != NULL) {
		ham_val = ham_val->multiply(vec_core->clone());
	}

	if (bIsClassCpl) {
		if (!is_renorm_CT) {
			if (m_eClassesHamiltonian==NULL) m_eClassesHamiltonian=ham_val; else m_eClassesHamiltonian=m_eClassesHamiltonian->add(ham_val);
		} else {
			if (m_eClassesHamiltonianR==NULL) m_eClassesHamiltonianR=ham_val; else m_eClassesHamiltonianR=m_eClassesHamiltonianR->add(ham_val);
		}
		m_idClassVertices.SetAt(v_num,-1);
	} else {
		if (!is_renorm_CT) {
			if (m_Hamiltonian_expr==NULL) m_Hamiltonian_expr=ham_val; else m_Hamiltonian_expr=m_Hamiltonian_expr->add(ham_val);
		} else {
			if (m_Renorm_Hamiltonian==NULL) m_Renorm_Hamiltonian=ham_val; else m_Renorm_Hamiltonian=m_Renorm_Hamiltonian->add(ham_val);
		}

		int iVecSymmFlags=0;
		if ((vect_index_v1!=NULL) && (vect_index_v2!=NULL) && (pk_v1==pk_v2)) { iVecSymmFlags|=1; }
		if ((vect_index_v3!=NULL) && (vect_index_v4!=NULL) && (pk_v3==pk_v4)) { iVecSymmFlags|=2; }
		long nClPos=InvolveVertexClass(pk_v1, pk_v2, pk_v3, pk_v4, is_renorm_CT, v_num, 3, iVecSymmFlags);
		m_idClassVertices.SetAt(v_num,nClPos);
	}
	addVertex4Info(v_num, pk_v1, pk_v2, pk_v3, pk_v4, is_renorm_CT);
}

// Unused function now
void CSMPhysics::addHamExprASV(int v_num, CExpression* vec_core,
							   CExpression* vect_index_a, CExpression* vect_index_v,
							   int pk_a, int pk_s, int pk_v, bool is_renorm_CT) {
	bool bIsClassCpl=false;
	if ((CLMath::abs(pk_a) & CEParticle::PARTICLE_CLASS_FLAG) && (CLMath::abs(pk_s) & CEParticle::PARTICLE_CLASS_FLAG) && (CLMath::abs(pk_v) & CEParticle::PARTICLE_CLASS_FLAG)) {
		bIsClassCpl=true;
		if (!m_bClassVerticesComplete) return;
	} else if ((CLMath::abs(pk_a) & CEParticle::PARTICLE_CLASS_FLAG) || (CLMath::abs(pk_s) & CEParticle::PARTICLE_CLASS_FLAG) || (CLMath::abs(pk_v) & CEParticle::PARTICLE_CLASS_FLAG)) {
		if (m_bClassVerticesComplete) {
			m_bClassVerticesComplete=false; dbg->log(0, "Warning: class diagrams not available. internal error in addHamExprASV.");
		}
	}

	// vector indices check
	bool bVec_a=((CLMath::abs(pk_a) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	bool bVec_v=((CLMath::abs(pk_v) & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON);
	if (vect_index_a == NULL) { if(bVec_a){ dbg->log(0, v_num, "Internal error. Vector index missed for vertex no %ld."); CEException::halt(); } }
	else { if(!bVec_a){ dbg->log(0, v_num, "Internal error. Unexpected vector index for vertex no %ld."); CEException::halt(); } }
	if (vect_index_v == NULL) { if(bVec_v){ dbg->log(0, v_num, "Internal error. Vector index missed for vertex no %ld."); CEException::halt(); } }
	else { if(!bVec_v){ dbg->log(0, v_num, "Internal error. Unexpected vector index for vertex no %ld."); CEException::halt(); } }

	CExpression* ham_val = new COpBosonField(pk_a, vect_index_a, true, v_num);
	ham_val = ham_val->multiply(new COpBosonField(pk_s, NULL, true, v_num));
	ham_val = ham_val->multiply(new COpBosonField(pk_v, vect_index_v, true, v_num));
	ham_val = ham_val->multiply(vec_core->clone());

	if (bIsClassCpl) {
		if (!is_renorm_CT) {
			if (m_eClassesHamiltonian==NULL) m_eClassesHamiltonian=ham_val; else m_eClassesHamiltonian=m_eClassesHamiltonian->add(ham_val);
		} else {
			if (m_eClassesHamiltonianR==NULL) m_eClassesHamiltonianR=ham_val; else m_eClassesHamiltonianR=m_eClassesHamiltonianR->add(ham_val);
		}
		m_idClassVertices.SetAt(v_num,-1);
	} else {
		if (!is_renorm_CT) {
			if (m_Hamiltonian_expr==NULL) m_Hamiltonian_expr=ham_val; else m_Hamiltonian_expr=m_Hamiltonian_expr->add(ham_val);
		} else {
			if (m_Renorm_Hamiltonian==NULL) m_Renorm_Hamiltonian=ham_val; else m_Renorm_Hamiltonian=m_Renorm_Hamiltonian->add(ham_val);
		}

		long nClPos=InvolveVertexClass(pk_a, pk_s, pk_v, 0, is_renorm_CT, v_num, 4);
		m_idClassVertices.SetAt(v_num,nClPos);
	}
}

// Denner:    +ieg (this)
// Cheng,Lee: -ieg
void CSMPhysics::addCorrRuleAFF(int v_num, CExpression* ind) {
	//Old: [x(-1)] CExpression* mIm = CENumber::GetImUnit(); mIm->inverseFlag();
	CExpression* Im = CENumber::GetImUnit();
	CExpression* val = Im->multiply(getModelContsSym(CONST_E)->clone())->multiply(new CEGamma(ind, true));
	CSequence* syms_seq = new CSequence();
	syms_seq->addItem(ind->clone());
	m_SM_corr_rules->addVertexCorrRule(v_num, val, syms_seq);
	delete val;
	delete syms_seq;
}

// Denner:    +ieg(dZe+1/2dZaa+dF)
void CSMPhysics::addCorrRuleGFF_CT(int v_num, CExpression* ind) {
	//Old: [x(-1)] CExpression* mIm = CENumber::GetImUnit(); mIm->inverseFlag();
	CExpression* Im = CENumber::GetImUnit();
	CExpression* val = Im->multiply(getModelContsSym(CONST_E)->clone())->multiply(new CEGamma(ind, true));

	// counterterm value
	CExpression* ct = getRenormContsSym(DZ_E)->clone();
	ct = ct->add(getRenormContsSym(DZ_F_EL)->clone());
	ct = ct->add(CENumber::GetHalf()->multiply(getRenormContsSym(DZ_AA)->clone()));
	//	*e^2:
	//ct = (new CEPower(getModelContsSym(CONST_E), (long)2))->multiply(ct);

	val = val->multiply(ct);

	CSequence* syms_seq = new CSequence();
	syms_seq->addItem(ind->clone());
	m_SM_corr_rules->addVertexCorrRule(v_num, val, syms_seq);
	delete val;
	delete syms_seq;
}

void CSMPhysics::init_H_data() {
	dbg->log(0, "Warning: unused (old) code invokation");
}

CExpression* CSMPhysics::getHamiltonianExpr() { return m_Hamiltonian_expr; }
CExpression* CSMPhysics::getRenormHamiltonian() { return m_Renorm_Hamiltonian; }
CERule* CSMPhysics::getHReplRule() { return m_HamiltonianReplsRule; }

CERule* CSMPhysics::createVecReplRule(CEVector* vec, CExpression* subst_expr) {
	if ((vec == NULL) || (subst_expr == NULL)) {
		dbg->log(0, "createVecReplRule: invalid parameters", vec); CEException::halt();
	}
	CEPtrArray* data_arr = new CEPtrArray();
	data_arr->add(vec);
	data_arr->add(subst_expr);
	CERule* r = new CESocketRule((void*)CSMPhysics::ReplVectorsRuleProc, data_arr);
	data_arr->clearall_without_delete_objects();
	delete data_arr;
	return r;
}

/*
*	used with CESocketRule
*	replaces vectors with arbitrary indices
*	data array:
*		0: vector with fixed index
*		1: substitute expression with the same index
*/
CExpression* CSMPhysics::ReplVectorsRuleProc(CExpression* src, CEPtrArray* data_arr) {
	if (src == NULL) return src;
	if (!CEUtil::equals(src->ClassName(), CEVector::CLASS_NAME)) return src;
	CEVector* src_vec = (CEVector*)src;

	if ((data_arr == NULL) || (data_arr->size() != 2)) {
		dbg->log(0, "ReplVectorsRuleProc: invalid data", data_arr); CEException::halt();
	}
	CEVector* val1 = (CEVector*)data_arr->GetAt(0);
	CExpression* val2 = (CExpression*)data_arr->GetAt(1);

	CExpression* vec_base = src_vec->base();
	if (!vec_base->equals(val1->base())) return src;
	CExpression* res = val2->clone();

	CExpression* ind_src = src_vec->index();
	CExpression* ind_int = val1->index();
	if (!ind_src->equals(ind_int)) { // change index in subst expression
		CERule* repl_ind_rule = new CEReplasement(ind_int, ind_src);
		res = repl_ind_rule->ApplyTo(res);
		delete repl_ind_rule;
	}

	src->copyAttributesTo(res, true);
	val1->copyAttributesTo(res, true); // val1 should not have any attributes, but...
	return res;
}

CSequence* CSMPhysics::CollectConstants() {
	CSequence* res = new CSequence();
	res->addItem(getModelContsSym(CONST_E)->clone());
	res->addItem(getModelContsSym(CONST_ALPHA)->clone());
	res->addItem(getModelContsSym(CONST_CTGW)->clone());
	res->addItem(getModelContsSym(CONST_CW)->clone());
	res->addItem(getModelContsSym(CONST_SW)->clone());
	res->addItem(getModelContsSym(CONST_GP)->clone());
	res->addItem(getModelContsSym(CONST_GM)->clone());
	res->addItem(getModelContsSym(CONST_NEU)->clone());
	res->addItem(getModelContsSym(CONST_QU)->clone());
	res->addItem(getModelContsSym(CONST_QD)->clone());
	res->addItem(getModelContsSym(CONST_GU_P)->clone());
	res->addItem(getModelContsSym(CONST_GU_M)->clone());
	res->addItem(getModelContsSym(CONST_GD_P)->clone());
	res->addItem(getModelContsSym(CONST_GD_M)->clone());
	res->addItem(getModelContsSym(CONST_G_S)->clone());
	res->addItem(get_mass_sym(CEParticle::PARTICLE_ELECTRON)->clone());
	res->addItem(get_mass_sym(CEParticle::PARTICLE_MUON)->clone());
	res->addItem(get_mass_sym(CEParticle::PARTICLE_TAU_LEPTON)->clone());
	res->addItem(get_mass_sym(CEParticle::PARTICLE_W_BOSON)->clone());
	res->addItem(get_mass_sym(CEParticle::PARTICLE_Z_BOSON)->clone());
	res->addItem(get_mass_sym(CEParticle::PARTICLE_PROTON)->clone());
	res->addItem(CEBasic::GetPiSym()->clone());
	res->addItem(new CESymbol(CDiagramPhysics::m_SymGeV2PbConst));
	res->addItem(new CESymbol("DEGREE"));
	return res;
}

CEPtrArray* CSMPhysics::CollectConstantValues() {
	//TODO: fill also CollectConstants() list
	CEPtrArray* res = new CEPtrArray();

	//	gP = -e*(-1)*SinW/CosW
	//	gM = e*(-1)/(2.*SinW*CosW) - e*(-1)*SinW/CosW
	//	GuP = -e*Qu*SinW/CosW
	//	GuM = -e*Qu*SinW/CosW + e/(2.*SinW*CosW)
	//	GdP = -e*Qd*SinW/CosW
	//	GdM = -e*Qd*SinW/CosW + e*(-1)/(2.*SinW*CosW)
	//	gW = e/sW/sqrt(2)

	//Note: Don't worry about compiler warning 'taking address of temporary' here!
	res->add(new CExprContainer(getModelContsSym(CONST_E),&CERealValue(0.30286)));
	res->add(new CExprContainer(getModelContsSym(CONST_ALPHA),&CERealValue(1./137.035999679)));	//PDG'09
	res->add(new CExprContainer(getModelContsSym(CONST_CTGW),&CERealValue(1.87399)));
	res->add(new CExprContainer(getModelContsSym(CONST_CW),&CERealValue(0.882248)));
	res->add(new CExprContainer(getModelContsSym(CONST_SW),&CERealValue(0.4707849)));
	res->add(new CExprContainer(getModelContsSym(CONST_GP),&CERealValue(0.161612)));
	res->add(new CExprContainer(getModelContsSym(CONST_GM),&CERealValue(-0.202973)));
	res->add(new CExprContainer(getModelContsSym(CONST_NEU),&CERealValue(0.454888)));			// TODO: check mantissa!
	res->add(new CExprContainer(getModelContsSym(CONST_G_S),&CERealValue(0.11847)));			// PDG'10

	// quarks
	res->add(new CExprContainer(getModelContsSym(CONST_QU),&CERealValue((double)2./(double)3.)));
	res->add(new CExprContainer(getModelContsSym(CONST_QD),&CERealValue(-(double)1./(double)3.)));
	res->add(new CExprContainer(getModelContsSym(CONST_GU_P),&CERealValue(-0.107741)));
	res->add(new CExprContainer(getModelContsSym(CONST_GU_M),&CERealValue(0.256844)));
	res->add(new CExprContainer(getModelContsSym(CONST_GD_P),&CERealValue(0.0538707)));
	res->add(new CExprContainer(getModelContsSym(CONST_GD_M),&CERealValue(-0.310714)));

	// masses
	//http://pdg.lbl.gov/2007/reviews/contents_sports.html#constantsetc
	//5.510998918e-4
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_ELECTRON),&CERealValue(5.109989e-4)));	//PDG'09
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_MUON),&CERealValue(0.105658389)));
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_TAU_LEPTON),&CERealValue(1.77699)));
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_W_BOSON),&CERealValue(80.399)));			//80.45(PYTHIA), 80.403(PDG'09), 80.399(PDG'10)
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_Z_BOSON),&CERealValue(91.1876)));			//PDG'10: 91.1876
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_PROTON),&CERealValue(0.938272013)));		//PDG'09
	//
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_QUARK_U),&CERealValue(2.34e-3)));			//PDG'10: 2.34 +/- 0.19 MeV
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_QUARK_D),&CERealValue(4.81e-3)));			//PDG'10: 4.81 +/- 0.14 MeV
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_QUARK_S),&CERealValue(0.1006)));			//PDG'10: 100.6 MeV
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_QUARK_C),&CERealValue(1.271)));			//PDG'10: 1.271 +/- 0.012 GeV
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_QUARK_B),&CERealValue(4.19)));			//PDG'10 //TODO: check!!!!
	res->add(new CExprContainer(get_mass_sym(CEParticle::PARTICLE_QUARK_T),&CERealValue(172.0)));			//PDG'10: 172.0 +/- 0.9 +/- 1.3 GeV

	// pi
	double _pi= 3.1415926535897932385; //=acos((double)-1), pi=3.1415926535897932384626433832795029
	// Note: 8-bytes 'pi' is 3.1415926535897931
	res->add(new CExprContainer(CEBasic::GetPiSym(),&CERealValue(_pi)));

	// additional constants (not used directly)
	// Nt: CERealValue(3.8937966E+8) ==> 3.89379659999999999999E+8
	res->add(new CExprContainer(&CESymbol(CDiagramPhysics::m_SymGeV2PbConst),&CERealValue(3.8937966E+8)));
	res->add(new CExprContainer(&CESymbol("DEGREE"),&CERealValue(_pi/180)));
	return res;
}

CERule* CSMPhysics::getRenormConstValues() {
	CERulesList* res = new CERulesList();
	CExpression* val = NULL;
	CExpression* pi = CEBasic::GetPiSym(); // not cloned
	CExpression* pi2 = pi->sqra(); // cloned
	CExpression* a0 = CVirtInt0::get_a0m_cpy(); // cloned
	CExpression* x0 = CVIScalarCalcLib::getUVSymbol(); // not cloned
	CExpression* xIR = CVIScalarCalcLib::getIRSymbol(); // not cloned
	CExpression* alpha = getModelContsSym(CSMPhysics::CONST_ALPHA); // not cloned
	CExpression* mDR = get_DR_mass_sym(); // not cloned
	CExpression* mE = get_mass_sym(CEParticle::PARTICLE_ELECTRON); // not cloned
	CExpression* frac_tmp = mE->sqra()->divide(mDR->sqra()); // cloned
	CExpression* lnEM = new CFLogN(frac_tmp); // cloned
	delete frac_tmp;
	////////////////////////////////////////////////////////////////////////////

	val = new CESymbol("DZ\\+\\(AA\\)"); // abstract value (temporary)
	//val = a0->clone()->add(lnEM->clone());
	//val = val->divide(CENumber::GetNumber(3));
	res->addRule(new CEReplasement(getRenormContsSym(DZ_AA), val));

	val = val->divide(CENumber::GetTwo());
	val->inverseFlag();
	res->addRule(new CEReplasement(getRenormContsSym(DZ_E), val)); delete val; val=NULL;

	val = new CESymbol("DM\\+\\(e\\)"); // abstract value (temporary)
	//val = (a0->clone()->add(lnEM->clone()))->multiply(CENumber::GetNumber(3));
	//val = val->minus(CENumber::GetFour());
	val = val->multiply(mE->clone())->divide(CENumber::GetFour());
	res->addRule(new CEReplasement(getRenormContsSym(DM_F_EL), val)); delete val; val=NULL;

	val = x0->clone()->minus(lnEM->clone()->multiply(CENumber::GetNumber(3)));
	val = val->add(CENumber::GetFour());
	val = val->add(xIR->clone()->multiply(CENumber::GetTwo()));
	val = val->multiply(alpha->clone())->divide(CENumber::GetFour())->divide(pi->clone());
	val->inverseFlag();
	res->addRule(new CEReplasement(getRenormContsSym(DZ_F_EL), val)); delete val; val=NULL;

	////////////////////////////////////////////////////////////////////////////
	delete pi2; delete a0; delete lnEM;

	/*
	involveRenormConst(DZ_AA, new CESymbol("\\[Delta]Z\\+\\(AA\\)"));
	involveRenormConst(DZ_ZA, new CESymbol("\\[Delta]Z\\+\\(ZA\\)"));
	involveRenormConst(DZ_AZ, new CESymbol("\\[Delta]Z\\+\\(AZ\\)"));
	involveRenormConst(DZ_ZZ, new CESymbol("\\[Delta]Z\\+\\(ZZ\\)"));

	involveRenormConst(DZ_E, new CESymbol("\\[Delta]Z\\+\\(e\\)"));
	involveRenormConst(DZ_F_EL, new CESymbol("\\[Delta]F\\+\\(e\\)"));
	involveRenormConst(DZ_F_MU, new CESymbol("\\[Delta]F\\+\\(\\[mu]\\)"));
	involveRenormConst(DZ_F_TAU, new CESymbol("\\[Delta]F\\+\\(\\[tau]\\)"));

	involveRenormConst(DM_F_EL, new CESymbol("\\[Delta]m\\+\\(e\\)"));
	involveRenormConst(DM_F_MU, new CESymbol("\\[Delta]m\\+\\(\\[mu]\\)"));
	involveRenormConst(DM_F_TAU, new CESymbol("\\[Delta]m\\+\\(\\[tau]\\)"));
	*/
	dbg->log(2, "renorm constants", res);
	return res;
}

CExpression* CSMPhysics::getClassesHamiltonianE() { return (m_bClassVerticesComplete?m_eClassesHamiltonian:NULL); } // NULL if not available
CExpression* CSMPhysics::getClassesHamiltonianRen() { return (m_bClassVerticesComplete?m_eClassesHamiltonianR:NULL); } // NULL if not available

long CSMPhysics::CollectVerticesForClass(long vnumClassVertex, CIntArray* idAltVerts) {
	if ((vnumClassVertex<0) || (idAltVerts==NULL)) return 0;
	CEPtrArray* arr_VIs=getVerticesInfo(); if (arr_VIs==NULL) return 0;
	long i=0; long idClassVertex=-1;
	long s=m_idClassVertices.GetSize(); if(s!=arr_VIs->size()) CEException::halt("Internal ERROR at CSMPhysics::CollectVerticesForClass()");
	if (vnumClassVertex>=s) { dbg->log(0,vnumClassVertex,s,"ERROR: CSMPhysics::CollectVerticesForClass(%ld,..) invalid call. max: %ld"); return 0; }
	CDiagramVertexInfo* viClassV=(CDiagramVertexInfo*)arr_VIs->GetAt(vnumClassVertex); if(viClassV==NULL) { dbg->log(0,"WARNING: NULL record at getVerticesInfo()"); return 0; }
	bool isCT=viClassV->check_vert_type(CDiagramVertexInfo::VERTEX_COUNTERTERM);

	// 1. find class-info object position (at m_VertexClasses array)
	idClassVertex=m_idClassVertices.GetAt(vnumClassVertex);
	if (idClassVertex<0) { // Note: next lines replace previous if m_idClassVertices.GetAt(class_vertex)==-1
		s=m_vnumClassVertices.GetSize(); for(i=0;i<s;i++) { if (m_vnumClassVertices.GetAt(i)==vnumClassVertex) { idClassVertex=i; break; } }
		if (idClassVertex<0) {
			s=m_vnCTClassVertices.GetSize();  for(i=0;i<s;i++) { if (m_vnCTClassVertices.GetAt(i)==vnumClassVertex) { idClassVertex=i; isCT=true; break; } }
		}
	}

	long res=0;
	if (idClassVertex>=0) {
		for(i=0;i<s;i++) {
			if (i>=arr_VIs->size()) { dbg->log(0,s,arr_VIs->size(),"WARNING: m_idClassVertices does not meet to getVerticesInfo() (%ld,%ld). It may lead to incomplete set of diagrams."); break; }
			if ((m_idClassVertices.GetAt(i)==idClassVertex) && (i != vnumClassVertex)) {
				CDiagramVertexInfo* vi=(CDiagramVertexInfo*)arr_VIs->GetAt(i); if(vi==NULL) { dbg->log(0,"WARNING: NULL record at getVerticesInfo()"); continue; }
				bool isViCT=vi->check_vert_type(CDiagramVertexInfo::VERTEX_COUNTERTERM);
				if (CEUtil::LEQU(isCT,isViCT)) { idAltVerts->Add(i); res++; }
			}
		}
	}
	return res;
}

bool CSMPhysics::AreCTPropagatorsVertices() {
	//Currently: if CT are declared, CT-propagators are involved NOT as 2-vertices.
	return ((m_phys_ID & PHYS_RENORM_COUNTERTERMS) == 0);
}

bool CSMPhysics::InitCouplingsClasses() {
	if (!m_bClassVerticesComplete) return false;
	if ((m_H_vect_index==NULL)||(m_H_vect_index_2==NULL)||(m_H_vect_index_3==NULL)||(m_H_vect_index_4==NULL)) {
		dbg->log(0, "Warning: class diagrams not available. internal error in InitCouplingsClasses #1.");
		m_bClassVerticesComplete=false; return false;
	}
	if (m_eClassesHamiltonian!=NULL) { delete m_eClassesHamiltonian; m_eClassesHamiltonian=NULL;}
	if (m_eClassesHamiltonianR!=NULL) { delete m_eClassesHamiltonianR; m_eClassesHamiltonianR=NULL;}
	long nClRulesCreated=0;
	int s=m_VertexClasses.size();
	CEPtrArray* vert_info_list = getVerticesInfo(); if(vert_info_list==NULL){
		dbg->log(0, "Warning: class diagrams not available. internal error in InitCouplingsClasses #2.");
		m_bClassVerticesComplete=false; return false;
	}
	for(int i=0; i<s; i++) {
		m_vnumClassVertices.SetAt(i,-1); m_vnCTClassVertices.SetAt(i,-1); // before any return
		CSMPhysClassCplInfo* inf = (CSMPhysClassCplInfo*)m_VertexClasses.GetAt(i);
		if(inf==NULL){ dbg->log(0, "Warning: class diagrams not available. internal error in InitCouplingsClasses #3."); m_bClassVerticesComplete=false; break; }
		int partN = inf->m_partPKs.GetSize();
		if ((partN<2) || (partN>4)) {
			m_bClassVerticesComplete=false; dbg->log(0, partN, "ERROR. Correspondence classes rule can't be installed. Vertex with %ld tales is not supported."); break;
		}
		/*
		*	m_VertexClasses: arr of CSMPhysClassCplInfo, used to fill m_eClassesHamiltonianX
		*	  pos corresponding CLASS-vertex info in m_VertexClasses (or -1)
		*	m_idClassVertices[0,getVerticesInfo()->size()]: position of corresponding CLASS-vertex info in m_VertexClasses, -1 for CLASS vertex
		*	m_vnumClassVertices[0,m_VertexClasses.size()]: reverse for m_idClassVertices for class-vertices (or -1 if vertex is CT-only)
		*	  vnum for every item of m_VertexClasses (or -1 if vertex is CT-only)
		*	m_vnCTClassVertices[0,m_VertexClasses.size()]: reverse for m_idClassVertices for class-vertices (or -1 if vertex is non-CT-only)
		*	  vnCT for every item of m_VertexClasses (or -1 if vertex is non-CT-only)
		*/
		long pk0=inf->m_partPKs.GetAt(0); long pk1=inf->m_partPKs.GetAt(1);
		long pk2=((partN>2)?inf->m_partPKs.GetAt(2):0); long pk3=((partN>3)?inf->m_partPKs.GetAt(3):0);
		bool bVec0=((CLMath::abs(pk0) & CEParticle::PARTICLE_VECTOR_BOSON)==CEParticle::PARTICLE_VECTOR_BOSON);
		bool bVec1=((CLMath::abs(pk1) & CEParticle::PARTICLE_VECTOR_BOSON)==CEParticle::PARTICLE_VECTOR_BOSON);
		bool bVec2=((CLMath::abs(pk2) & CEParticle::PARTICLE_VECTOR_BOSON)==CEParticle::PARTICLE_VECTOR_BOSON);
		bool bVec3=((CLMath::abs(pk3) & CEParticle::PARTICLE_VECTOR_BOSON)==CEParticle::PARTICLE_VECTOR_BOSON);

		long v_num=vert_info_list->size();
		CExpression* vec_core=NULL;
		bool bAdded=false;
		switch(inf->m_pkConfigID) {
		case 1:
			if (partN==3) {
				if (bVec0) {
					CESymbol* ham_base = CESymbol::getUnique(false);
					vec_core = new CEVector(ham_base, m_H_vect_index, false); delete ham_base;
					addHamExprVFF(v_num,vec_core,m_H_vect_index,pk0,pk1,pk2,inf->m_bIsCt);
					bAdded=true;
				} else {
					addHamExprVFF(v_num,NULL,NULL,pk0,pk1,pk2,inf->m_bIsCt);
					bAdded=true;
				}
			}
			break;
		case 2:
			if (partN==3) {
				if (bVec0) {
					if (bVec1) {
						if (bVec2) {
							CESymbol* ham_base = CESymbol::getUnique();
							vec_core = new CERank3Tensor(ham_base, m_H_vect_index, false, m_H_vect_index_2, false, m_H_vect_index_3, false); delete ham_base;
							Add3BosonVertex(v_num,vec_core,m_H_vect_index,m_H_vect_index_2,m_H_vect_index_3,pk0,pk1,pk2,inf->m_bIsCt);
							bAdded=true;
						} else {
							CESymbol* ham_base = CESymbol::getUnique();
							// Note: VVS/VVSS-matrix is considered SYMMETRIC ALWAYS! (also in couplings-create procedure)
							vec_core = new CEMatrix(ham_base, m_H_vect_index, m_H_vect_index_2, false, false, 1); delete ham_base;
							Add3BosonVertex(v_num,vec_core,m_H_vect_index,m_H_vect_index_2,NULL,pk0,pk1,pk2,inf->m_bIsCt);
							bAdded=true;
						}
					} else if (!bVec2) {
						CESymbol* ham_base = CESymbol::getUnique();
						vec_core = new CEVector(ham_base, m_H_vect_index, false); delete ham_base;
						Add3BosonVertex(v_num,vec_core,m_H_vect_index,NULL,NULL,pk0,pk1,pk2,inf->m_bIsCt);
						bAdded=true;
					}
				} else if ((!bVec1) && (!bVec2)) {
					Add3BosonVertex(v_num,NULL,NULL,NULL,NULL,pk0,pk1,pk2,inf->m_bIsCt);
					bAdded=true;
				}
			}
			break;
		case 3:
			if (partN==4) {
				if ((bVec0) && (bVec1)) {
					if ((bVec2) && (bVec3)) {
						int iSymm12=0; if (((inf->m_iVecSymmFlags)&1)==1) iSymm12=1;
						int iSymm34=0; if (((inf->m_iVecSymmFlags)&2)==2) iSymm34=1;
						CESymbol* ham_base = CESymbol::getUnique();
						vec_core=new CERank4Tensor(ham_base,m_H_vect_index,false,m_H_vect_index_2,false,m_H_vect_index_3,false,m_H_vect_index_4,false,iSymm12,iSymm34);
						delete ham_base;
						Add4BosonVertex(v_num,vec_core,m_H_vect_index,m_H_vect_index_2,m_H_vect_index_3,m_H_vect_index_4,pk0,pk1,pk2,pk3,inf->m_bIsCt);
						bAdded=true;
					} else if ((!bVec2) && (!bVec3)) {
						CESymbol* ham_base = CESymbol::getUnique();
						// Note: VVS/VVSS-matrix is considered SYMMETRIC ALWAYS! (also in couplings-create procedure)
						vec_core = new CEMatrix(ham_base, m_H_vect_index, m_H_vect_index_2, false, false, 1); delete ham_base;
						Add4BosonVertex(v_num,vec_core,m_H_vect_index,m_H_vect_index_2,NULL,NULL,pk0,pk1,pk2,pk3,inf->m_bIsCt);
						bAdded=true;
					}
				} else if ((!bVec0) && (!bVec1) && (!bVec2) && (!bVec3)) {
					Add4BosonVertex(v_num,NULL,NULL,NULL,NULL,NULL,pk0,pk1,pk2,pk3,inf->m_bIsCt);
					bAdded=true;
				}
			}
			break;
		case 4: // unused now
			if (partN==3) { // bVec1==false always here
				vec_core=NULL; // fill vec_core here
				if (vec_core!=NULL) {
					addHamExprASV(v_num,vec_core,m_H_vect_index,m_H_vect_index_2,pk0,pk1,pk2,inf->m_bIsCt);
					bAdded=true;
				}
			}
			break;
		}
		if (vec_core!=NULL) delete vec_core;
		if (bAdded) {
			m_idClassVertices.SetAt(v_num,i);
			if (inf->m_bIsCt) { m_vnCTClassVertices.SetAt(i,v_num); } else { m_vnumClassVertices.SetAt(i,v_num); }
			nClRulesCreated++;
		}
		if (!bAdded) {
			dbg->log(0, "Warning: class diagrams not available. internal error in InitCouplingsClasses #4.");
			m_bClassVerticesComplete=false;
		}
		if (!m_bClassVerticesComplete) break;
	}
	if (!m_bClassVerticesComplete) { dbg->log(2, nClRulesCreated, "Can't create class-vertex rules"); return false; }
	dbg->log(2, nClRulesCreated, "Total %ld class-vertex rule(s) created");
	dbg->log(55, "Classes Hamiltonian",getClassesHamiltonianE());
	dbg->log(55, "Classes Hamiltonian (renorm)",getClassesHamiltonianRen());
	return true;
}

long CSMPhysics::InvolveVertexClass(long pk1, long pk2, long pk3, long pk4, bool bIsCT, long nVClassNo, int pkConfID, int iVecSymmFlags) {
	if (!m_bClassVerticesComplete) return -1;
	CSMPhysClassCplInfo inf(pkConfID,bIsCT,iVecSymmFlags);
	inf.SetClassCouplingFor(pk1,pk2,pk3,pk4);
	if (inf.m_partPKs.GetSize() < 2) {
		dbg->log(0, "Warning: class diagrams not available. internal error in InvolveVertexClass().");
		m_bClassVerticesComplete=false; return -1;
	}
	return InvolveVertexClass(&inf);
}

long CSMPhysics::InvolveVertexClass(CSMPhysClassCplInfo* inf) { // inf is unaffected
	if (inf==NULL) return -1;
	long id=-1;
	long i; long s=m_VertexClasses.size();
	for(i=0; i<s; i++) {
		CSMPhysClassCplInfo* infI = (CSMPhysClassCplInfo*)m_VertexClasses.GetAt(i); if(infI==NULL)continue;
		if (infI->equals(inf)) { id=i; break; }
	}
	if (id<0) {
		m_VertexClasses.add(new CSMPhysClassCplInfo(*inf));
		id=m_VertexClasses.size()-1;
	}
	return id;
}
