////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// SModelLoader.cpp: implementation of the CSModelLoader class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

CDebug* CSModelLoader::dbg = new CDebug("CSModelLoader");

const DWORD CSModelLoader::LINE_BRAKE_EOF			= 0x01;
const DWORD CSModelLoader::LINE_BRAKE_RIGHT_BRACE	= 0x02;
const DWORD CSModelLoader::LINE_BRAKE_PRIORITY		= 0x04;
const DWORD CSModelLoader::LINE_BRAKE_ANY_FLAG		= 0xff;

#define POL_VEC_NAME	"PolarizationVector"

#define TMPFUNC_VarIfCondition	"TMP_VarIfCondition"
#define TMPFUNC_LineIfCondition	"TMP_LineIfCondition"
#define TMPFUNC_ConditionEqu0	"TMP_ConditionEqu0"
#define TMPFUNC_ConditionNEqu0	"TMP_ConditionNEqu0"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CSModelLoader::CSModelLoader(const char* _GenBuf, const char* _ModelBuf) : CEAbstractSession()
{
	m_GenBuf=_GenBuf;
	m_ModelBuf=_ModelBuf;
	m_ParserGen=new CMathFileParser(m_GenBuf);
	m_ParserModel=new CMathFileParser(m_ModelBuf);
	m_bGenericPropagatorsFilling=false;
	m_bGenericCouplingsFilling=false;
	m_bFormatParameterFilling=false;

	m_bCouplingMatricesFilling=false;
	m_bClassesDescriptionFilling=false;
	m_bDontSkipRBrace=false;

	m_phys=NULL;
}

CSModelLoader::~CSModelLoader() {
	if (m_ParserGen!=NULL) delete m_ParserGen;
	if (m_ParserModel!=NULL) delete m_ParserModel;
	if (m_phys!=NULL) delete m_phys;
}

const char* CSModelLoader::CLASS_NAME = "CSModelLoader";
const char* CSModelLoader::ClassName() const { return CLASS_NAME; }

//////////////////////////////////////////////////////////////////////
// Entry
//////////////////////////////////////////////////////////////////////
bool CSModelLoader::InstallModelFile(const char* _GenFile, const char* _ModelFile) {
	if (_ModelFile==NULL) return false;
	if (_GenFile==NULL) {
		_GenFile="Lorentz.gen";
		dbg->logIns(4, "No *.GEN file is specified. Presumed: %s", _GenFile);
	}
	
	dbg->logIns(0, "Reading generic model file: %s", _GenFile);
	char* _GenBuf = CAbstractFileParser::LoadFileToMemory(_GenFile);
	if (_GenBuf == NULL) {
		dbg->logIns(0, "Generic model file loading FAILED (%s)", _GenFile);
		return false;
	}

	dbg->logIns(0, "Reading model file: %s", _ModelFile);
	char* _ModelBuf = CAbstractFileParser::LoadFileToMemory(_ModelFile);
	if (_ModelBuf == NULL) {
		dbg->logIns(0, "Model file loading FAILED (%s)", _ModelFile);
		CMemoryManager::mfree(_GenBuf);
		return false;
	}

	CSModelLoader* lib = LoadModel(_GenBuf, _ModelBuf);
	if (lib!=NULL) { lib->m_GenBuf=NULL; lib->m_GenBuf=NULL; }
	CMemoryManager::mfree(_ModelBuf);
	CMemoryManager::mfree(_GenBuf);

	// install physics model
	bool intsOK=false;
	if (lib!=NULL) {
		intsOK=lib->DoInitPhysics();
		if ((intsOK)&&(lib->m_phys != NULL)) {
			CDiagramPhysics::install_diagram_physics(lib->m_phys, true);
			lib->m_phys=NULL;
			// log out correspondence rules
			// note: correspondence rules logging requires physics to be already installed
			//CDiagramPhysics::dbg->log(7, "New hamiltonian expression", phys_new->getHamiltonianExpr());
			//CDiagramPhysics::dbg->log(7, "New correspondence rules", phys_new->getCorrespondenceRules());
		} else {
			if (intsOK)intsOK=false;
			CDiagramPhysics::dbg->log(0, "Error physics initialization. No model installed.");
		}
		delete lib;
	}
	return intsOK;
}

CSModelLoader* CSModelLoader::LoadModel(const char* _GenBuf, const char* _ModelBuf) {
	if ((_ModelBuf==NULL) || (_GenBuf==NULL)) return NULL;
	
	CSModelLoader* lib = new CSModelLoader(_GenBuf,_ModelBuf);
	if (!lib->DoLoadModel()) {
		delete lib; return NULL;
	}
	return lib;
}

CEPtrArray* CSModelLoader::m_PreDefValues = NULL; // array of {(expr)par, (expr)value}, deleted by CEProcessor::AddToDeleteList() engine
void CSModelLoader::setModelParameter(const char* par, CExpression* val_const) {
	if ((par==NULL) || (val_const==NULL)) return;
	if (m_PreDefValues==NULL) { m_PreDefValues=new CEPtrArray(); CEProcessor::AddToDeleteList(m_PreDefValues); }
	CExpression* ePar = NULL;

	int s=CEUtil::length(par); int i;
	int pos_braL=CEUtil::findSymbol(par,'[',(long)0);
	int pos_braL2=-1; if(pos_braL>=0)pos_braL2=CEUtil::findSymbol(par,'[',(long)(pos_braL+1));
	int pos_braR=-1; if(pos_braL>=0)pos_braR=CEUtil::findSymbol(par,']',(long)(pos_braL+1));
	if ((pos_braL>0) && (pos_braR>pos_braL) && (pos_braR==s-1) && (pos_braL2<0)) {
		if ((pos_braR-1)>=(pos_braL+1)) {
			char* tmp1=CEUtil::copyStringFromTo(par,0,pos_braL-1); //(char*)CMemoryManager::mmalloc(s+2);
			char* tmp2=CEUtil::copyStringFromTo(par,pos_braL+1,pos_braR-1);

			ePar=new CEFunction(tmp1,new CESymbol(tmp2));
			CMemoryManager::mfree(tmp1); CMemoryManager::mfree(tmp2);
		} else {
			char* tmp1=CEUtil::copyStringFromTo(par,0,pos_braL-1); //(char*)CMemoryManager::mmalloc(s+2);
			ePar=new CEFunction(tmp1);
			CMemoryManager::mfree(tmp1);
		}
	} else {
		if (pos_braL2>=0) {
			dbg->logIns(0,"WARNING: can't resolve parameter string %s: too many brackets. Parameter ignored.", par);
		} else if ((pos_braL>0) && (pos_braR<=pos_braL)) {
			dbg->logIns(0,"WARNING: can't resolve parameter string %s: can't fing right bracket. Parameter ignored.", par);
		} else if ((pos_braL>0) && (pos_braR>pos_braL) && (pos_braR!=s-1)) {
			dbg->logIns(0,"WARNING: can't resolve parameter string %s: right bracket is allowed at end of operand only. Parameter ignored.", par);
		} else {
			if (pos_braL>0) { dbg->logIns(0,"WARNING: can't resolve parameter string %s: unknown error. Parameter ignored.", par); }
		}
	}
	
	if (ePar==NULL) ePar=new CESymbol(par);
	CExpression* val=val_const->clone();
	if (!ePar->getFlag()) { ePar->inverseFlag(); val->inverseFlag(); }

	s=m_PreDefValues->size();
	bool par_found=false;
	for (i=0;i<s;i++) {
		CEPtrArray* cartI=(CEPtrArray*)m_PreDefValues->GetAt(i); if(cartI==NULL)continue;
		CExpression* parI=(CExpression*)cartI->GetAt(0); if(parI==NULL)continue;
		if (ePar->equals(parI)) { cartI->SetAt(1,val,true); par_found=true; val=NULL; break; }
	}
	if (!par_found) {
		CEPtrArray* cart = new CEPtrArray();
		cart->add(ePar); ePar=NULL; cart->add(val); val=NULL;
		m_PreDefValues->add(cart);
	}
	if(ePar!=NULL)delete ePar; if(val!=NULL)delete val;
}

//////////////////////////////////////////////////////////////////////
// Implementation
//////////////////////////////////////////////////////////////////////
bool CSModelLoader::DoLoadModel() {
	if (m_PreDefValues!=NULL) { dbg->log(15, "Predefined constants:",m_PreDefValues); }

	bool res=false;
	try{
		dbg->log(2, "Parsing GENERIC MODEL file");
		for(;;) {
			// parse next line
			m_parser=m_ParserGen; m_line_priority=0;
			CScriptValue* line_val = DoLoadModelLine(); if (line_val!=NULL) delete line_val;
			if ((m_line_brake_flags & LINE_BRAKE_EOF) == LINE_BRAKE_EOF) break;
			// any other line break means error
			if (m_line_brake_flags != 0) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected syntax. EndOfFile is not reached"); }
		}
		res=true;
	} catch(CEException* exc) {
		dbg->log(0, "<Error loading GENERIC MODEL file> ", exc->getMessage()->getData());
		const char* _fn=exc->getFileName(); long lnum=exc->getLineNumber();
		if (_fn!=NULL) {
			CEString tmp("{Source file halted:"); tmp.AddString(_fn); if(lnum>=0){tmp.AddString(", line: ");tmp.add_int(lnum);} tmp.AddString("}");
			dbg->log(0, exc->getMessage()->getData());
		}
		dbg->log(0, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos()-1, " <Error GENERIC MODEL file position:>{Line: %ld, Pos: %ld}");
		delete exc;
	} catch(...) {
		dbg->log(0, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos()-1, " <Error loading GENERIC MODEL file> {Line: %ld, Pos: %ld}");
	}
	if(!res)return false;

	res=false;
	try{
		dbg->log(2, "Parsing MODEL file");
		for(;;) {
			// parse next line
			m_parser=m_ParserModel; m_line_priority=0;
			CScriptValue* line_val = DoLoadModelLine(); if (line_val!=NULL) delete line_val;
			if ((m_line_brake_flags & LINE_BRAKE_EOF) == LINE_BRAKE_EOF) break;
			// any other line break means error
			if (m_line_brake_flags != 0) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected syntax. EndOfFile is not reached"); }

		}
		res=true;
	} catch(CEException* exc) {
		dbg->log(0, "<Error loading MODEL file> ", exc->getMessage()->getData());
		const char* _fn=exc->getFileName(); long lnum=exc->getLineNumber();
		if (_fn!=NULL) {
			CEString tmp("{Source file halted:"); tmp.AddString(_fn); if(lnum>=0){tmp.AddString(", line: ");tmp.add_int(lnum);} tmp.AddString("}");
			dbg->log(0, exc->getMessage()->getData());
		}
		dbg->log(0, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos()-1, " <Error MODEL file position:>{Line: %ld, Pos: %ld}");
		delete exc;
	} catch(...) {
		dbg->log(0, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos()-1, " <Error loading MODEL file> {Line: %ld, Pos: %ld}");
	}
	if (!res)return false;

	// log read information
	dbg->log(15, "m_KinIndices", &m_KinIndices);
	dbg->log(15, "m_GenericPropagators:", &m_GenericPropagators);
	dbg->log(15, "m_GenericCouplings", &m_GenericCouplings);
	dbg->log(15, "m_ClassesDescription", &m_ClassesDescription);
	dbg->log(15, "m_CouplingMatrices", &m_CouplingMatrices);
	dbg->log(15, "m_Variables", &m_Variables);
	dbg->log(15, "m_MathEquRules", &m_MathEquRules);
	return res;
}

bool CSModelLoader::DoInitPhysics() { // m_phys is inited here
	m_phys = new CPhysModelFA();

	if (!InitParticleList()) {
		dbg->log(0, "ERROR particles list initialization");
		if (m_MsgLastError.length()>0) dbg->log(0, "Last Error:", m_MsgLastError.getData());
		CEException::halt("ERROR particles list initialization");
		return false;
	}

	if (!InitPropagatorsList()) {
		dbg->log(0, "ERROR propagators list initialization");
		if (m_MsgLastError.length()>0) dbg->log(0, "Last Error:", m_MsgLastError.getData());
		CEException::halt("ERROR propagators list initialization");
		return false;
	}

	if (!InitCouplingsList()) {
		dbg->log(0, "ERROR couplings list initialization");
		if (m_MsgLastError.length()>0) dbg->log(0, "Last Error:", m_MsgLastError.getData());
		CEException::halt("ERROR couplings list initialization");
		return false;
	}

	if (!InitCouplingsClasses()) {
		dbg->log(0, "WARNING: couplings CLASSES list initialization failed. No fast diagrams can be generated.");
	}

	// Log vertices (incl. classes-vertices)
	if (dbg->get_level()>=25) {
		CEPtrArray* vert_info_list = m_phys->getVerticesInfo(); if(vert_info_list==NULL){ dbg->log(0,"ERROR:m_phys->getVerticesInfo()==NULL."); CEException::halt(); }
		long s=vert_info_list->size(); long i;
		CEString sLog("\\<\\-------- Vertices log -----------");
		sLog.AddString("\n"); //CEBasic::appendLineBreak(&sLog, STORE_WR_MATHEMATICA);
		for(i=0;i<s;i++) {
			CDiagramVertexInfo* vi=(CDiagramVertexInfo*)vert_info_list->GetAt(i); if(vi==NULL) { dbg->log(0,"WARNING: NULL record at getVerticesInfo()"); continue; }
			sLog.AddString("{#"); sLog.add_int(i); sLog.AddString("} ");
			//sLog.AddString(vi->toString(STORE_WR_MATHEMATICA));
			if (vi->check_vert_type(CDiagramVertexInfo::VERTEX_COUNTERTERM)) sLog.AddString("<CT> ");
			int vi_s = vi->get_tales_num();
			for(int vi_i=0; vi_i<vi_s; vi_i++) {
				long pk = vi->get_particle_kind(vi_i);
				const char* vi_pk_sym=m_phys->get_particle_symbol(pk,CDiagramPhysics::SYMBOL_STYLE_NB);
				if (CEUtil::equals(vi_pk_sym,"Null")) { sLog.add_int(pk,true); } else { if(pk<0)sLog.AddString("-"); sLog.AddString(vi_pk_sym); }
				if ((vi_i+1) < vi_s) sLog.AddString(", ");
			}
			long vnumClass=-1; long vnumClassCT=-1;
			if (m_phys->GetClassVertexID(i, &vnumClass, &vnumClassCT)) {
				sLog.AddString(" (class:"); sLog.add_int(vnumClass); sLog.AddString(", class-CT:"); sLog.add_int(vnumClassCT); sLog.AddString(")");
			}
			sLog.AddString("\n"); //CEBasic::appendLineBreak(&sLog, STORE_WR_MATHEMATICA);
		}
		sLog.AddString("-------- Vertices log {END} -----\\\n\\>");
		dbg->log(25,sLog.getData());
		//dbg->log(25,"m_VertexClasses",&m_phys->m_VertexClasses);
	}

	//dbg->log(15, "m_phys->m_ParticleInfos", &m_phys->m_ParticleInfos);
	m_phys->m_IsInited=true;
	return true;
}

void CSModelLoader::declare_script_error(int err_code, int i_param, const char* c_param) {
	CScriptException* scr_exc = new CScriptException(err_code,
		m_parser->get_curr_line_num(),
		m_parser->get_curr_line_pos()-1,
		i_param,
		c_param);

	dbg->log(2, "**** Internal structures log ****");
	dbg->log(2, "m_KinIndices", &m_KinIndices);
	dbg->log(2, "m_GenericPropagators", &m_GenericPropagators);
	dbg->log(2, "m_GenericCouplings", &m_GenericCouplings);
	dbg->log(2, "m_ClassesDescription", &m_ClassesDescription);
	dbg->log(2, "m_CouplingMatrices", &m_CouplingMatrices);
	throw scr_exc;
}

#ifdef LOADMODEL_DROP_ZERO_PROPS //Do not use this mechanism! Inf vertex * 1/Inf propagator is allowed! TODO: fix it!
bool CSModelLoader::bIsZeroPropagatorPK(long pk) { // particles with ZERO propagator -> to drop out adjacent vertices
	long pk_abs=(pk>=0)?pk:-pk;
	long s=m_ZeroPropagatorPKs.GetSize();
	for(long i=0; i<s; i++) { if (m_ZeroPropagatorPKs.GetAt(i)==pk_abs) return true; }
	return false;
}
#endif //LOADMODEL_DROP_ZERO_PROPS

/*
*	m_whole_value is the clone of string value (or NULL) -- to be deleted if useless
*	Note: _whole_val is used to continue parsing of half-processed lines.
*	Used for priority-tree calls.
*/
CScriptValue* CSModelLoader::DoLoadModelLine(int line_priority, CScriptValue* _whole_val)
{
	dbg->log(55, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos(), "DoLoadModelLine(line %ld, pos %ld)");
	//////////////////////////////////////////////////////// store environment BEGIN
	CScriptValue* old_whole_val = m_whole_value; m_whole_value = _whole_val;
	CEPtrArray* _LeftValuesOld=m_LeftValues; m_LeftValues=new CEPtrArray();
	CScriptValue* old_curr_op_val = m_curr_operand_value; m_curr_operand_value = NULL;
	int old_LP = m_line_priority; m_line_priority = line_priority;
	m_parser->start_new_operands_list();
	m_line_brake_flags = 0; // volatile, to be checked after line is parsed
	//////////////////////////////////////////////////////// store environment END

	// reading all the operands in the string
	bool operand_loop = true;
	int op_cnt;
	for(op_cnt = 0;operand_loop; op_cnt++) {
		int op_type = m_parser->get_next_op_type();
		dbg->log(85, op_type, "reading next operand (type == 0x%lx)");

		// check up type
		// 1. OP_INVALID, OP_NO_OPERAND: error
		if (op_type == CFileParserAdv::OP_INVALID) declare_script_error(CScriptErrors::ILLEGAL_OPERATION, -1, "Invalid operand spotted");
		if (op_type == CFileParserAdv::OP_NO_OPERAND) declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "No next operand found");

		// 3. OP_END_OF_LINE
		if (op_type == CFileParserAdv::OP_END_OF_LINE) {
			int prev_type = m_parser->get_operand_type(-1);
			if (prev_type & CFileParserAdv::OP_OPERATION_ANY) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected END_OF_LINE");
			// do not skip line delimiter for priority sub-string parsing
			if (m_line_priority == 0) m_parser->skip_next_operand();
			break;
		}

		// next processing requires operand string
		const char* c_op = m_parser->get_next_operand(false); dbg->log(85, c_op); if(c_op==NULL)declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "NULL operand read");
		//if (op_type == CFileParserAdv::OP_OPERATION_RIGHT) { dbg->log(2,"1 ", c_op); m_parser->LogReadOperandsList(2); }

		// 4.0.9 check for empty operand (*)
		if ((op_type == CFileParserAdv::OP_VARIABLE) ||
			((op_type & CFileParserAdv::OP_VALUE_ANY)==CFileParserAdv::OP_VALUE_ANY) ||
			(*c_op == ')') ||
			(*c_op == ']')) {
			m_parser->skip_next_operand(true);
			int next_type = m_parser->get_next_op_type();
			const char* op_next = m_parser->get_next_operand(false);
			if ((next_type == CFileParserAdv::OP_VARIABLE) ||
				((next_type & CFileParserAdv::OP_VALUE_ANY)==CFileParserAdv::OP_VALUE_ANY) ||
				(*op_next == '(')) {
				dbg->log(85, "Empty operand changed to \"*\""); //m_parser->LogReadOperandsList(105);
				m_parser->insert_operand_as_current("*",CFileParserAdv::OP_OPERATION_RIGHT); // add "fictional" operand
			}
			m_parser->push_back_last_operand();
			int op_type_old=op_type; const char* c_op_old=c_op;
			op_type = m_parser->get_next_op_type(); c_op = m_parser->get_next_operand(false);
			if ((op_type_old!=op_type) || (c_op_old!=c_op)) {
				declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "Pushed-back operand does not meet original");
			}
		}
		//if (op_type == CFileParserAdv::OP_OPERATION_RIGHT) { dbg->log(2,"2: ",c_op); m_parser->LogReadOperandsList(2); }

		// 4.1. OP_BRACKET_RIGHT: bracket closing (end of substring)
		if (op_type == CFileParserAdv::OP_BRACKET_RIGHT) {
			m_line_brake_flags = m_line_brake_flags | LINE_BRAKE_RIGHT_BRACE;
			int prev_type = m_parser->get_operand_type(-1);
			if (prev_type & CFileParserAdv::OP_OPERATION_ANY) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected \")\" symbol");
			// Note:
			//	1. skip "]"-bracket for global line (m_line_priority == 0)
			//	2. do not skip line delimiter for priority sub-string parsing
			if ((m_line_priority == 0) && (!m_bDontSkipRBrace)) m_parser->skip_next_operand();
			break;
		}
		//if (op_type == CFileParserAdv::OP_OPERATION_RIGHT) { dbg->log(2,"3: ",c_op); m_parser->LogReadOperandsList(2); }

		// 4.2. SYM_END_OF_BUFFER: end of file exceed
		if (op_type == CAbstractFileParser::SYM_END_OF_BUFFER) {
			m_line_brake_flags = m_line_brake_flags | LINE_BRAKE_EOF; dbg->log(85, "EOF encounted");
			// if (op_cnt > 0) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unexpected EOF (\";\" missed?)");
			break;
		}

		// 4.3. Lower priority spotted
		if ((m_line_priority > 0) && (getOperandPriority(c_op, op_type) <= m_line_priority)) {
			m_line_brake_flags = m_line_brake_flags | LINE_BRAKE_PRIORITY; dbg->log(85, "Lower priority spotted");
			int prev_type = m_parser->get_operand_type(-1);
			if (prev_type & CFileParserAdv::OP_OPERATION_ANY) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "something like \"/+\""); // "a/+"
			break; // do not skip this operand
		}

		// 5.5 skip-symbols
		/*
		if (op_type == CFileParserAdv::OP_COLON) {
			m_parser->skip_next_operand(false); // ignore it
			int next_type = m_parser->get_next_op_type();
			if (next_type == CFileParserAdv::OP_OPERATION_LEFT) continue;
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "unexpected colon");
		}
		*/

		// all the cases above breakes the line parsing
		// all the operands are added to line operands list.

		// test
		//if(m_parser->get_curr_line_num() == 154) { dbg->log(2, "c_op == ", c_op); m_parser->LogReadOperandsList(2); }

		// 5.6.2 empty operand (*)
		// Note: This check is spotted many times
		if ((op_type == CFileParserAdv::OP_VARIABLE)
			|| ((op_type & CFileParserAdv::OP_VALUE_ANY)==CFileParserAdv::OP_VALUE_ANY)
			|| (*c_op == '(')) {
			int prev_type = m_parser->get_operand_type(-1);
			const char* c_op_prev=m_parser->get_operand(-1);
			if (c_op_prev!=NULL) {
				if ((prev_type == CFileParserAdv::OP_VARIABLE) ||
					((prev_type & CFileParserAdv::OP_VALUE_ANY)==CFileParserAdv::OP_VALUE_ANY) ||
					(*c_op_prev == ')') || (*c_op_prev == ']')) {
					// add "fictional" operand
					m_parser->insert_operand_as_current("*",CFileParserAdv::OP_OPERATION_RIGHT);
					c_op = m_parser->get_next_operand(false); op_type = m_parser->get_next_op_type();
				}
			}
		}

		// skip current operand
		m_parser->skip_next_operand(true); // 'c_op' is deleted if not added to skipped-list

		// 5.9.1 empty operand --> (*)
		// Note: This check is spotted many times
		/*
		if ((op_type == CFileParserAdv::OP_VARIABLE) ||
			((op_type & CFileParserAdv::OP_VALUE_ANY)==CFileParserAdv::OP_VALUE_ANY)) {
			int next_type = m_parser->get_next_op_type();
			char* op_next = m_parser->get_next_operand(false);
			if ((next_type == CFileParserAdv::OP_VARIABLE) ||
				((next_type & CFileParserAdv::OP_VALUE_ANY)==CFileParserAdv::OP_VALUE_ANY) ||
				(*op_next == '(')) {
				m_parser->insert_operand_as_current("*",CFileParserAdv::OP_OPERATION_RIGHT); // add "fictional" operand
			}
		}
		*/

		// 5.9.2 { "->" to comma }
		if ((op_type == CFileParserAdv::OP_OPERATION_RIGHT) && (*c_op == '-')) {
			const char* c_op = m_parser->get_next_operand(false);
			if (*c_op == '>') {
				dbg->log(85, " changing \"->\" to end-of=substring");
				//m_parser->LogReadOperandsList(95);
				m_parser->push_back_last_operand();
				m_parser->skip_next_operand(false); // remove "-"
				c_op = m_parser->get_next_operand(false);
				if (*c_op != '>') declare_script_error(CScriptErrors::INTERNAL_ERROR); // error!
				m_parser->skip_next_operand(false); // remove ">"
				m_parser->insert_operand_as_current(",",CFileParserAdv::OP_END_OF_LINE);
				m_parser->LogReadOperandsList(95);
				continue;
			}
		}

		// unused self-test
		if (m_curr_operand_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);

		// 6. operations
		// if (op_type == CFileParserAdv::OP_OPERATION_RIGHT) { dbg->log(2,"4: ",c_op); m_parser->LogReadOperandsList(2); }
		if ((op_type == CFileParserAdv::OP_OPERATION_LEFT) || (op_type == CFileParserAdv::OP_OPERATION_RIGHT)) {
			do_operation_involve(c_op, op_type); continue;
		}

		// Further TARGET: calculate the value of current operand & update string value
		// 7. left brace
		if (op_type == CFileParserAdv::OP_BRACKET_LEFT) {
			CScriptValue* brace_val=NULL;
			if (*c_op=='{') {
				int nBracVals=0;
				CScriptValueArray* arr=new CScriptValueArray();
				for (;;) {
					CScriptValue* subVal = DoLoadModelLine();
					if ((m_line_brake_flags & LINE_BRAKE_EOF) || (m_line_brake_flags & LINE_BRAKE_PRIORITY)) {declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "\"}\" or \",\" are expected");}
					if (subVal != NULL) {
						nBracVals++; dbg->log(85, "read array element", subVal);
						arr->addToArray(subVal,true);
					} else {
						// this may be the void {} object, but not {a,NULL}
						if ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0) {
							declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Right bracket expected");
						}
						if (nBracVals > 0) {
							// {object,object,NULL}
							//Note: this could be if "}" is first symbol on the next line --> No errors
							//declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "{...,NULL}");
						}
					}
					if (m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) {
						if (m_bDontSkipRBrace) { // skip right bracket manually
							if ((m_parser->get_next_op_type() & CFileParserAdv::OP_BRACKET_RIGHT)==CFileParserAdv::OP_BRACKET_RIGHT) { m_parser->skip_next_operand(false); }
							else { dbg->log(0,"INTERNAL ERROR: right bracket missed after function call"); CEException::halt(); }
						}
						clear_line_break_flag(); break;
					}
				}
				brace_val=arr;
				dbg->log(85, "array value read", brace_val);
			} else {
				brace_val = DoLoadModelLine();
				if ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0) {
					dbg->log(0,"ERROR: Bracket is not closed! Right braket expected for: ", c_op);
					m_parser->LogReadOperandsList(2);
					declare_script_error(CScriptErrors::STATEMENT_ILLEGAL);
				} else if (brace_val == NULL) {
					declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "NULL object in brackets");
				}
				if (m_bDontSkipRBrace) { // skip right bracket manually
					if ((m_parser->get_next_op_type() & CFileParserAdv::OP_BRACKET_RIGHT)==CFileParserAdv::OP_BRACKET_RIGHT) { m_parser->skip_next_operand(false); }
					else { dbg->log(0,"INTERNAL ERROR: right bracket missed after function call"); CEException::halt(); }
				}
				clear_line_break_flag(); // clear end-of-string flag
			}
			m_curr_operand_value = brace_val;
		}
		// 8. VALUES
		else if (op_type & CFileParserAdv::OP_VALUE_ANY) {
			if (op_type  ==  CFileParserAdv::OP_VALUE_NUMBER) {
				long op_lval=0;
				if (!CEUtil::atolSafe(c_op,&op_lval)) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "atol failed");
				m_curr_operand_value = CScriptValue::newValue(op_lval);
			}
			else if (op_type  ==  CFileParserAdv::OP_VALUE_STRING) {
				CEString* op_sval = new CEString(c_op);
				m_curr_operand_value = CScriptValue::newValue(op_sval, true);
			}
			else {
				// unknown value
				declare_script_error(CScriptErrors::UNDEFINED_OBJECT);
			}
		}
		// 9. Variables and functions
		else if (op_type == CFileParserAdv::OP_VARIABLE) {
			// 9.1. Functions
			//bool func_invoked = do_function_cast(c_op, op_type);
			// 9.2. Variables
			//if (!func_invoked) do_variable_involve(c_op, op_type);
			do_variable_involve(c_op, op_type);
		}
		// 10.9 Unknown
		else {
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Unknown object");
		}
		// m_curr_operand_value should NOT be NULL here

		// 11. Update value of the string (perform right-handed operations)
		update_string_value();
		if(m_curr_operand_value!=NULL)declare_script_error(CScriptErrors::INTERNAL_ERROR); // m_curr_operand_value should BE NULL here

		// appending current operand to line operands storage (?)
		/*if ((m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED) == 0) {
			m_parser->skip_next_operand();
			m_line_brake_flags = m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED;
		}*/
	}

	// cycle breaks before last m_parser->skip_next_operand() invoked
	//if((m_line_brake_flags&LINE_BRAKE_CURRENT_SKIPED)==0) { m_parser->skip_next_operand(); m_line_brake_flags|=LINE_BRAKE_CURRENT_SKIPED; }

	// left operation executing
	lefthand_operation_execute();

	// return value composing
	CScriptValue* res = m_whole_value; m_whole_value=NULL;
	
	//////////////////////////////////////////////////////// restore environment BEGIN
	if (m_LeftValues!=NULL) { delete m_LeftValues; m_LeftValues=NULL; }
	if (m_curr_operand_value!=NULL) { delete m_curr_operand_value; m_curr_operand_value=NULL; } // error/warning if was (m_curr_operand_value!=NULL)
	m_curr_operand_value = old_curr_op_val; m_whole_value = old_whole_val; m_line_priority = old_LP; m_LeftValues=_LeftValuesOld;
	m_parser->finish_operands_list();
	//////////////////////////////////////////////////////// restore environment END
	return res;
}

void CSModelLoader::clear_line_break_flag() { m_line_brake_flags = (m_line_brake_flags | LINE_BRAKE_ANY_FLAG) ^ LINE_BRAKE_ANY_FLAG; }

/*
*	Note: current operand {c_op, op_type} is already skipped.
*	m_curr_operand_value == NULL at start
*	m_whole_value != NULL at start (not obligatory, e.g. for unitaty operations)
*/
void CSModelLoader::do_operation_involve(const char* c_op, int op_type) {
	dbg->log(85, "do_operation_involve()");
	// check up operation existance
	if (op_type == CFileParserAdv::OP_OPERATION_LEFT) { // "="
		// truncate whole value for left operation
		if (m_LeftValues==NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR,-1,"NULL left oparands array");
		/*if (m_left_value!=NULL) {
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"More than one left operand in string!");
			delete m_left_value; m_left_value=NULL;
		}*/
		if (m_whole_value == NULL) {
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Invalid pos of left operand!");
		}

		// special flags
		CESymbol* lValSym=m_whole_value->GetValidExprSymbol();
		if (lValSym!=NULL) {
			const char* sLName=lValSym->get_name();
			if (CEUtil::equals(sLName, "M$GenericPropagators")) m_bGenericPropagatorsFilling=true;
			else if (CEUtil::equals(sLName, "M$GenericCouplings")) m_bGenericCouplingsFilling=true;
			else if (CEUtil::equals(sLName, "M$ClassesDescription")) m_bClassesDescriptionFilling=true;
			else if (CEUtil::equals(sLName, "M$CouplingMatrices")) m_bCouplingMatricesFilling=true;
		}

		m_LeftValues->add(m_whole_value); m_whole_value=NULL;
		//if (m_whole_value != NULL) {delete m_whole_value; m_whole_value = NULL;}
	}

	if (op_type == CFileParserAdv::OP_OPERATION_RIGHT) {
		// check up previous operand: right-hand operation may not be followed by another operation
		int prev_offs = -2; // previous
		int prev_type = m_parser->get_operand_type(prev_offs);
		if (prev_type == CFileParserAdv::OP_OPERATION_RIGHT) {
			/*
			*	This may not be second ot 2 right-handed opertions.
			*/
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Two right-handed operations in a raw");
		}

		int next_type = m_parser->get_next_op_type(); const char* next_op = m_parser->get_next_operand(false);
		if (next_type == CFileParserAdv::OP_OPERATION_RIGHT) {
			/*
			*	Two right right-handed opertions, like "a==-1"
			*	Note: the following line "a == -b == c" after {-b} evaluation becomes "a == stub ==".
			*/
			int op_prior = getOperandPriority(c_op, op_type);
			if (op_prior < getOperandPriority(next_op, next_type)) { // check next operation priority
				// start new string processor using NULL value as first & update current value
				//dbg->set_level(500);
				//m_parser->LogReadOperandsList(2);
				CScriptValue* sub_val = DoLoadModelLine(op_prior); // m_curr_operand_value will be deleted
				// no matter how was the line breaked
				if (sub_val == NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "NULL result for sub-string");
				clear_line_break_flag(); // clear end-of-string flag
				dbg->log(85, "inline string sub-value", sub_val);

				// insert & skip stub into operands list
				m_parser->insert_operand_as_current("stub",CFileParserAdv::OP_VARIABLE);
				m_parser->skip_next_operand();
				//m_parser->LogReadOperandsList(2);
				//dbg->set_level(5);
				m_curr_operand_value = sub_val;
				update_string_value();
				dbg->log(85, "whole value after sub-line insertion", m_whole_value);
			} else {
				dbg->log(0,"Two right-handed operations in a raw. The second operation must have higher priority!");
				declare_script_error(CScriptErrors::STATEMENT_ILLEGAL);
			}
		}
	}
	// NOTE: operation processing is keeped in update_string_value() function after the next operand is read out.
}

// op_type is unused (should be CFileParserAdv::OP_VARIABLE)
void CSModelLoader::do_variable_involve(const char *c_op, int op_type) {
	dbg->log(85, "do_variable_involve(), ", c_op);

	// Next item check-up
	int next_type = m_parser->get_next_op_type(); const char* next_op = m_parser->get_next_operand(false);

	// *. Is it a inline-function? (like Metric[a,b]->g_ab)
	// Nt: for Inline[a,b] defined, do not treat Inline as inline-object.
	int N=getFuncSpecInfo(c_op); // -1
	int N2=getFuncSpecInfo2(c_op); // -1
	if (m_bFormatParameterFilling) N=-1;
	if ((N>=0) && ((N==0) || (next_type == CFileParserAdv::OP_BRACKET_LEFT))) {
		dbg->log(85, "inline object: ", c_op);
		CEPtrArray index_list; int index_cnt = 0;
		if (next_type == CFileParserAdv::OP_BRACKET_LEFT) {
			m_parser->skip_next_operand(false);
			bool bneedEquConv=bNeedEqualsConvert(c_op);
			bool bneedEquConvOld=m_parser->m_bUseRHand2Equs; bool bDontSkipRBraceOld=m_bDontSkipRBrace;
			if (bneedEquConv) { m_parser->m_bUseRHand2Equs=bneedEquConv; m_bDontSkipRBrace=true; }
			bool bBracketClosed=false;
			for (;;) {
				CScriptValue* ind = DoLoadModelLine();
				if ((m_line_brake_flags & LINE_BRAKE_EOF) || (m_line_brake_flags & LINE_BRAKE_PRIORITY)) {declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "\"]\" or \",\" are expected");}
				if (ind != NULL) {
					index_list.add(ind); index_cnt++; dbg->log(85, "inline function parameter", ind);
				} else {
					// this may be the void  array[] call.
					if ((index_cnt > 0) || ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0)) {
						dbg->log(0,"Unexpected NULL object: [..., NULL]",&index_list);
						m_parser->LogReadOperandsList(2);
						declare_script_error(CScriptErrors::NULL_OBJECT);
					}
				}
				if (m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) {
					if (m_bDontSkipRBrace) { // skip right bracket manually
						if ((m_parser->get_next_op_type() & CFileParserAdv::OP_BRACKET_RIGHT)==CFileParserAdv::OP_BRACKET_RIGHT) { m_parser->skip_next_operand(false); }
						else { dbg->log(0,"INTERNAL ERROR: right bracket missed after function call"); CEException::halt(); }
					}
					clear_line_break_flag(); bBracketClosed=true; break;
				}
			}
			if (!bBracketClosed) {
				dbg->log(0,"ERROR: right bracket expected after function call"); declare_script_error(CScriptErrors::STATEMENT_ILLEGAL);
			}
			/*
			*	Note the following case:
			*	If[...]==...
			*	is proceeded correctly due to m_bDontSkipRBrace=true flag.
			*	The "==" symbol should not be read as right-handed here
			*/
			if (bneedEquConv) {
				// restore flags
				m_parser->m_bUseRHand2Equs=bneedEquConvOld; m_bDontSkipRBrace=bDontSkipRBraceOld;
			}
		}
		bool bFitsPattern=true;
		if ((N!=INT_MAX) && (index_cnt!=N)) {
			dbg->log(2,"WARNING. Unexpected number of arguments. Function:",c_op);
			dbg->log(2,N,index_cnt,"Arguments required: %ld, Arguments read: %ld",&index_list);
			m_parser->LogReadOperandsList(95);
			//declare_script_error(CScriptErrors::STATEMENT_ILLEGAL);
			bFitsPattern=false;
		}

		// to expressions
		CSequence params;
		int i;
		for(i=0; i<index_cnt; i++) {
			CScriptValue* ind=(CScriptValue*)index_list.GetAt(i);
			CExpression* expr=NULL; bool del_expr=false;
			if (ind->check_value(CScriptValue::VAL_EXPRESSION,true)) { expr=ind->get_expr_value(); }
			if (expr==NULL) { expr=ind->TurnToExprCloned(); del_expr=true; }
			if (expr==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Expression parameter expected"); }
			params.addItem((del_expr?expr:expr->clone()));
		}

		// Next item check-up
		next_type = m_parser->get_next_op_type(); next_op = m_parser->get_next_operand(false);
		CSequence params2; CSequence* pParams2=NULL;
		if ((N2>=0) && (N2!=INT_MIN)) { //if ((N2>=0) && ((N2==0) || (next_type == CFileParserAdv::OP_BRACKET_LEFT))) {
			CEPtrArray index_list; int index_cnt = 0;
			if (next_type == CFileParserAdv::OP_BRACKET_LEFT) {
				m_parser->skip_next_operand(false);
				for (;;) {
					CScriptValue* ind = DoLoadModelLine();
					if ((m_line_brake_flags & LINE_BRAKE_EOF) || (m_line_brake_flags & LINE_BRAKE_PRIORITY)) {declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "\"]\" or \",\" are expected");}
					if (ind != NULL) {
						index_list.add(ind); index_cnt++; dbg->log(85, "inline object parameter #2", ind);
					} else {
						// this may be the void  array[] call.
						if ((index_cnt > 0) || ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0)) declare_script_error(CScriptErrors::NULL_OBJECT, -1, "[..., NULL]");
					}
					if (m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) {
						if (m_bDontSkipRBrace) { // skip right bracket manually
							if ((m_parser->get_next_op_type() & CFileParserAdv::OP_BRACKET_RIGHT)==CFileParserAdv::OP_BRACKET_RIGHT) { m_parser->skip_next_operand(false); }
							else { dbg->log(0,"INTERNAL ERROR: right bracket missed after function call"); CEException::halt(); }
						}
						clear_line_break_flag(); break;
					}
				}
			}
			if ((N2!=INT_MAX) && (index_cnt!=N2)) {
				if (bFitsPattern) {
					dbg->log(2,"WARNING. Unexpected number of arguments [at second brackets]. Function:",c_op);
					dbg->log(2,N2,index_cnt,"Arguments required at second brackets: %ld, Arguments read: %ld",&index_list);
					//declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Unexpected number of arguments");
					bFitsPattern=false;
				}
			}
			// to expressions-2
			for(i=0; i<index_cnt; i++) {
				CScriptValue* ind=(CScriptValue*)index_list.GetAt(i); CExpression* expr=NULL; bool del_expr=false;
				if (ind->check_value(CScriptValue::VAL_EXPRESSION,true)) { expr=ind->get_expr_value(); }
				if (expr==NULL) { expr=ind->TurnToExprCloned(); del_expr=true; }
				if (expr==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Expression parameter expected"); }
				params2.addItem((del_expr?expr:expr->clone()));
			}
			pParams2=&params2;
		}

		if (bFitsPattern) {
			CScriptValue* res = getFuncSpecValue(c_op,&params,pParams2); if(res==NULL)declare_script_error(CScriptErrors::INTERNAL_ERROR);
			m_curr_operand_value = res;
			dbg->log(85, "inline object created");
		} else {
			CEFunction* f_res=new CEFunction(c_op,bIsFuncCommutative(c_op));
			if ((params.size()>0) || (params2.size()>0)) f_res->add_param(&params);
			if (params2.size()>0) f_res->add_param(&params2);
			dbg->log(2, "Can't create inline object. Abstract function is involved",f_res);
			CScriptValue* res = CScriptValue::newValue(f_res,true);
			m_curr_operand_value = res;
		}
		//m_parser->LogReadOperandsList(2); CEException::halt(); // test
		return;
	}

	// 1. Is it a function?
	/*
	if ((next_type == CFileParserAdv::OP_BRACKET_LEFT) && (*next_op == '[')) {
		CEPtrArray* index_list = new CEPtrArray();
		m_parser->skip_next_operand(false);
		int index_cnt = 0;
		for (;;) {
			CScriptValue* ind = DoLoadModelLine();
			if ((m_line_brake_flags & LINE_BRAKE_EOF) || (m_line_brake_flags & LINE_BRAKE_PRIORITY)) {declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "\"]\" or \",\" are expected");}
			if (ind != NULL) {
				index_list->add(ind); index_cnt++; dbg->log(50, "index", ind);
			} else {
				// this may be the void  array[] call.
				if ((index_cnt > 0) || ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0)) declare_script_error(CScriptErrors::NULL_OBJECT);
			}

			if (m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) {
				if (m_bDontSkipRBrace) { // skip right bracket manually
					if ((m_parser->get_next_op_type() & CFileParserAdv::OP_BRACKET_RIGHT)==CFileParserAdv::OP_BRACKET_RIGHT) { m_parser->skip_next_operand(false); }
					else { dbg->log(0,"INTERNAL ERROR: right bracket missed after function call"); CEException::halt(); }
				}
				clear_line_break_flag(); break;
			}
		}

		//index_list
		CScriptValue* res = CScriptValue::newArray();
		m_curr_operand_value = res;
		dbg->log(40, "array item involved");
		return;
	}*/

	// special Format[params]-->do not use inline objects for params
	bool bFormatFlagOld=m_bFormatParameterFilling; if (CEUtil::equals(c_op,"Format")) { m_bFormatParameterFilling=true; }

	// *. Is it array element?
	dbg->log(185, "do_variable_involve() #2, ", c_op);
	CEPtrArray index_full; bool is_array=false;
	for(;;) {
		if ((next_type == CFileParserAdv::OP_BRACKET_LEFT) && (*next_op == '[')) {
			is_array=true;
			CEPtrArray index_list;
			m_parser->skip_next_operand(false);
			int index_cnt = 0;
			for (;;) {
				CScriptValue* ind = DoLoadModelLine();
				if (ind!=NULL) dbg->log(85, "Array index read:", ind); else dbg->log(85, "Array index read: NULL");
				if (((m_line_brake_flags & LINE_BRAKE_EOF)==LINE_BRAKE_EOF) || ((m_line_brake_flags & LINE_BRAKE_PRIORITY)==LINE_BRAKE_PRIORITY)) {declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "\"]\" or \",\" are expected");}
				if (ind != NULL) {
					index_list.add(ind); index_cnt++;
				} else {
					// this may be the void  array[] call.
					if ((index_cnt > 0) || ((m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) == 0)) declare_script_error(CScriptErrors::NULL_OBJECT, -1 , "{a,NULL} array");
				}
				if (m_line_brake_flags & LINE_BRAKE_RIGHT_BRACE) {
					if (m_bDontSkipRBrace) { // skip right bracket manually
						if ((m_parser->get_next_op_type() & CFileParserAdv::OP_BRACKET_RIGHT)==CFileParserAdv::OP_BRACKET_RIGHT) { m_parser->skip_next_operand(false); }
						else { dbg->log(0,"INTERNAL ERROR: right bracket missed after function call"); CEException::halt(); }
					}
					clear_line_break_flag(); break;
				}
			}
			int s = index_list.size();
			CScriptValue* inds = NULL;

			if (s==1) {
				inds = ((CScriptValue*)index_list.GetAt(0))->clone();
			} else { // 0,2,3..
				CScriptValueArray* indsA=new CScriptValueArray();
				for(int i=0; i<s; i++) { indsA->addToArray((CScriptValue*)index_list.GetAt(i),false); }
				inds=indsA;
			}
			index_full.add(inds);

			// Nt: right bracked should already be skipped
			next_type = m_parser->get_next_op_type(); next_op = m_parser->get_next_operand(false);
		} else {
			break;
		}
	}
	m_bFormatParameterFilling=bFormatFlagOld;
	
	if (is_array) {
		CScriptValue* inds = NULL;
		int s = index_full.size();
		if (s==1) {
			inds = ((CScriptValue*)index_full.GetAt(0))->clone();
		} else { // 0,2,3..
			CScriptValueArray* indsA=new CScriptValueArray();
			for(int i=0; i<s; i++) { indsA->addToArray((CScriptValue*)index_full.GetAt(i),false); }
			inds=indsA;
		}
		CScriptValue* base = CScriptValue::newValue(new CESymbol(c_op),true);
		CScriptValueArrayElement* res = new CScriptValueArrayElement(base, inds);
		delete base; delete inds;
		m_curr_operand_value = res;
		dbg->log(85, "array item involved", m_curr_operand_value);
		// return; // Do not return: variable restoring follows
	}

	if (m_curr_operand_value == NULL) {
		// Usual variable
		m_curr_operand_value = CScriptValue::newValue(new CESymbol(c_op),true);
		// return; // Do not return: variable restoring follows
	}
	
	// look for op in local variables list
	CExpression* currOpExpr = m_curr_operand_value->TurnToExprCloned();
	if (currOpExpr != NULL) {
		// 1. try expr->{expr or non-expr} variables
		CScriptValue* val = GetVariableValue(currOpExpr);
		if (val != NULL) {
			if (!val->check_value_type(CScriptValue::VAL_UNDEFINED)) {
				delete m_curr_operand_value;
				m_curr_operand_value = val->clone(); dbg->log(85, "variable value restored", m_curr_operand_value);
				delete currOpExpr;
				// continue substitution
				currOpExpr=m_curr_operand_value->TurnToExprCloned(); if(currOpExpr==NULL)return;
				// return;
			}
		}

		// 2. try expr->expr variables (TODO: remove this repeation) and rules
		CExpression* valE = GetVarRuleValueECloned(currOpExpr);
		if (valE != NULL) {
			delete m_curr_operand_value;
			m_curr_operand_value = CScriptValue::newValue(valE, true); dbg->log(85, "Rule/variable value restored", m_curr_operand_value);
			delete currOpExpr;
			// continue substitution
			currOpExpr=m_curr_operand_value->TurnToExprCloned(); if(currOpExpr==NULL)return;
			//return;
		}

		// 3. try pre-defined constants (m_PreDefValues)
		if (m_PreDefValues!=NULL) {
			bool inv_flag = (!currOpExpr->getFlag()); if(inv_flag)currOpExpr->inverseFlag();
			CExpression* val_PDV = NULL;
			long s_PDV=m_PreDefValues->size();
			for(long i_PDV=0;i_PDV<s_PDV;i_PDV++) {
				CEPtrArray* cart=(CEPtrArray*)m_PreDefValues->GetAt(i_PDV); if(cart==NULL)continue;
				CExpression* par_PDVi=(CExpression*)cart->GetAt(0); CExpression* val_PDVi=(CExpression*)cart->GetAt(1); if((par_PDVi==NULL)||(val_PDVi==NULL))continue;
				if (currOpExpr->equals(par_PDVi)) { val_PDV=val_PDVi->clone(); break; }
			}
			if(inv_flag)currOpExpr->inverseFlag();

			if (val_PDV != NULL) {
				if(inv_flag)val_PDV->inverseFlag();
				delete m_curr_operand_value;
				m_curr_operand_value = CScriptValue::newValue(val_PDV, true); dbg->log(85, "Pre-defined constant value restored", m_curr_operand_value);
				delete currOpExpr;
				// continue substitution
				currOpExpr=m_curr_operand_value->TurnToExprCloned(); if(currOpExpr==NULL)return;
				//return;
			}
		}
		delete currOpExpr;
	}
}

//	***************************************************************
//	***                 perform operations                      ***
//	***************************************************************
void CSModelLoader::update_string_value() {
	dbg->log(85, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos(), "update_string_value(%ld : %ld)"); m_parser->LogReadOperandsList(155);
	// executs last operation
	if (m_curr_operand_value == NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
	
	// lookup previous operations
	// NOTE: current operand becomes '-1' after CScriptParser::skip_next_operand() invoked
	//dbg->log(40, m_parser->get_operand_type(-1), "-1 = %ld");
	int offs = -2; // previous //((m_line_brake_flags & LINE_BRAKE_CURRENT_SKIPED)?(-2):(-1));
	int prev_type = m_parser->get_operand_type(offs);
	const char* op_operation = m_parser->get_operand(offs);

	dbg->log(85, prev_type, "prev_type = 0x%lx");
	if ((prev_type == CFileParserAdv::OP_NO_OPERAND) || (prev_type == CFileParserAdv::OP_END_OF_LINE)) {
		if (m_whole_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR,-1,"Unexpected previous operand");
		m_whole_value = m_curr_operand_value; m_curr_operand_value=NULL;
		return;
	}

	if (prev_type == CFileParserAdv::OP_OPERATION_LEFT) {
		// left-hand operations are executed after the whole line value calculated ('='-operation is left-hand operation)
		if (m_whole_value != NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
		m_whole_value = m_curr_operand_value; m_curr_operand_value = NULL;
		return;
	}

	if (prev_type == CFileParserAdv::OP_OPERATION_RIGHT) {

		// 5.6.2 empty operand --> (*)
		int next_type = m_parser->get_next_op_type();
		const char* op_next = m_parser->get_next_operand(false);
		if ((next_type == CFileParserAdv::OP_VARIABLE) ||
			((next_type & CFileParserAdv::OP_VALUE_ANY)==CFileParserAdv::OP_VALUE_ANY) ||
			(*op_next == '(')) {
			m_parser->insert_operand_as_current("*",CFileParserAdv::OP_OPERATION_RIGHT); // add "fictional" operand
			int next_type = m_parser->get_next_op_type();
			op_next = m_parser->get_next_operand(false);
		}

		// check next operation priority
		// read next operand and push it back to stream
		if (next_type == CFileParserAdv::OP_OPERATION_RIGHT) {
			const char* next_c_op = m_parser->get_next_operand(false);
			int op_prior = getOperandPriority(op_operation, prev_type);
			if (op_prior < getOperandPriority(next_c_op, next_type)) {
				// start new string processor using current value as first & update current value
				CScriptValue* sub_val = DoLoadModelLine(op_prior, m_curr_operand_value); // m_curr_operand_value will be deleted
				// no matter how was the line breaked
				if (sub_val == NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "NULL result for sub-string");
				clear_line_break_flag(); // clear end-of-string flag
				m_curr_operand_value = sub_val;
				dbg->log(85, "sub-value", sub_val);
			}
		}
		/*
		else if (next_type == CFileParserAdv::OP_OPERATION_LEFT) {
			// examples "x+y=1;", "-x=1;"
			CEString sLog(""); sLog.AddString(op_operation); sLog.AddString(" "); sLog.AddString(op_next);
			dbg->log(0,"ERROR: Left-handed operation follows the after right one. Oprands: ",sLog.getData());
			m_parser->LogReadOperandsList(2);
			declare_script_error(CScriptErrors::STATEMENT_ILLEGAL);
		}
		*/

		// if previous or current values are UNDEFINED declare error.
		if ( ((m_whole_value != NULL) && (m_whole_value->value_type() & CScriptValue::VAL_UNDEFINED)) || (m_curr_operand_value->value_type() & CScriptValue::VAL_UNDEFINED) ) {
			declare_script_error(CScriptErrors::UNDEFINED_OBJECT);
		}

		int op_len = CEUtil::length(op_operation);
		// Note: operation may by long: "===", "=!=" etc.
		//if (op_len != 1) { declare_script_error(CScriptErrors::ILLEGAL_OPERATION,-1,"Unsupported operation"); return; }

		bool is_str_first = ((m_whole_value != NULL) && ((m_whole_value->value_type() & CScriptValue::VAL_STRING) != 0));
		bool is_str_second = ((m_curr_operand_value->value_type() & CScriptValue::VAL_STRING) != 0);
		if ((is_str_first) || (is_str_second)) declare_script_error(CScriptErrors::UNDEFINED_OBJECT);

		bool is_int_first = ((m_whole_value != NULL) && ((m_whole_value->value_type() & CScriptValue::VAL_INT) != 0));
		bool is_int_second = ((m_curr_operand_value->value_type() & CScriptValue::VAL_INT) != 0);
		//if ((is_int_first) || (is_int_second)) declare_script_error(CScriptErrors::UNDEFINED_OBJECT);

		bool is_expr_first = ((m_whole_value != NULL) && (m_whole_value->check_value_type(CScriptValue::VAL_EXPRESSION)));
		bool is_expr_second = (m_curr_operand_value->check_value_type(CScriptValue::VAL_EXPRESSION));
		//if ((!is_expr_first) || (!is_expr_second)) declare_script_error(CScriptErrors::UNDEFINED_OBJECT);

		dbg->log(95, "Left: ", m_whole_value);
		dbg->log(95, "Right: ", m_curr_operand_value);

		/*
		*	Common policy: we need to update m_whole_value
		*	delete m_curr_operand_value and set m_curr_operand_value = NULL;
		*/
		// Expr * Expr
		CExpression* eFirst = NULL; bool del_eFirst=false;
		if (is_expr_first) {
			eFirst=m_whole_value->get_expr_value();
		} else {
			eFirst=((m_whole_value!=NULL)?m_whole_value->TurnToExprCloned():NULL);
			if (eFirst!=NULL) del_eFirst=true;
		}
		CExpression* eSecond = NULL; bool del_eSecond=false;
		if (is_expr_second) {
			eSecond=m_curr_operand_value->get_expr_value();
		} else {
			eSecond=((m_curr_operand_value!=NULL)?m_curr_operand_value->TurnToExprCloned():NULL);
			if (eSecond!=NULL) del_eSecond=true;
		}

		if ((eFirst != NULL) || (eSecond != NULL)) {
			CExpression* value_res = NULL;
			if (eSecond==NULL) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "OPERAND = NULL");
			CExpression* _expr_curr = eSecond;
			if (eFirst != NULL) {
				CExpression* _expr_comm = eFirst;
				value_res = expr_operation_binary(op_operation, _expr_comm, _expr_curr);
				delete m_whole_value; m_whole_value = NULL;
			} else {
				if (m_whole_value!=NULL) {
					dbg->log(0, "Unexpected left operand", m_whole_value); declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Unexpected left operand"); return;
				}
				value_res = expr_operation_unitary(op_operation, _expr_curr);
			}
			delete m_curr_operand_value; m_curr_operand_value = NULL;
			if (value_res == NULL) { dbg->log(0, "Error: NULL expression value (No operation performed?)."); declare_script_error(CScriptErrors::INTERNAL_ERROR);}
			m_whole_value = CScriptValue::newValue(value_res, true);

			if (del_eSecond && (eSecond!=NULL)) delete eSecond;
			if (del_eFirst && (eFirst!=NULL)) delete eFirst;
			return;
		}
		if (del_eSecond && (eSecond!=NULL)) delete eSecond;
		if (del_eFirst && (eFirst!=NULL)) delete eFirst;
		
		dbg->log(2, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos(), "WARNING: Unexpected operation spotted (line: %ld, pos: %ld)");
	}

	dbg->log(145,m_parser->get_operand_type(-2),m_parser->get_operand_type(-1),"WARNING: Unexpexted string syntax, -2:0x%lx, -1:0x%lx");
	dbg->log(145,"m_parser->get_operand(-2): ", (const char*)((m_parser->get_operand(-2)!=NULL)?m_parser->get_operand(-2):NULL));
	dbg->log(145,"m_parser->get_operand(-1): ", (const char*)((m_parser->get_operand(-1)!=NULL)?m_parser->get_operand(-1):NULL));
	dbg->log(145,"m_curr_operand_value: ", m_curr_operand_value);
	dbg->log(145,"m_whole_value: ", m_whole_value);
	m_parser->LogReadOperandsList(2);
	if (m_whole_value != NULL) { dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"WARNING: Part of line is ignored (line: %ld, pos: %ld)", m_whole_value); delete m_whole_value; m_whole_value=NULL; }
	delete m_curr_operand_value; m_curr_operand_value = NULL;
	//declare_script_error(CScriptErrors::STATEMENT_ILLEGAL);
	return;
}

void CSModelLoader::lefthand_operation_execute() {
	dbg->log(85, "lefthand_operation_execute()");
	m_parser->LogReadOperandsList(195);
	int op_offset=INT_MIN; int nLO=0; int sLv=((m_LeftValues!=NULL)?m_LeftValues->size():0);
	for(;;) {
		op_offset = m_parser->get_operation_offset(CFileParserAdv::OP_OPERATION_LEFT, false, op_offset);
		if (op_offset == INT_MIN) break;
		nLO++;
		if (sLv<nLO) { dbg->log(0, "ERROR: NULL operand before left operation"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
		if (m_whole_value == NULL) { dbg->log(0, "ERRRO: NULL operand after the left operantion"); declare_script_error(CScriptErrors::INTERNAL_ERROR); }

		const char* op_sym = m_parser->get_operand(op_offset, false);
		if (CEUtil::equals("=", op_sym) || CEUtil::equals("==", op_sym)) { // '=' operation
			CScriptValue* lv=(CScriptValue*)m_LeftValues->GetAt(sLv-nLO);
			dbg->log(85, nLO, "Left and right values (#%ld)", lv); dbg->log(85, m_whole_value);
			bool sysFilledOk=sysEquOpExec(lv, m_whole_value);
			if (!sysFilledOk) {
				dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"WARNING: EQU operation failed: don't know how to peform it (line %ld, pos %ld). Left operand:",lv);
			}
		} else if (CEUtil::equals(op_sym,"/:")) { // '/:' operation, truncate all the left operands
			CScriptValue* lv=(CScriptValue*)m_LeftValues->GetAt(sLv-nLO);
			if (sLv>nLO) {
				dbg->log(0,m_parser->get_curr_line_num(),"WARNING: \"/:\" spotted after left-handed operation. Part of line is skipped (line %ld).");
			} else {
				dbg->log(2,m_parser->get_curr_line_num(),"Notice: \"/:\" operation is skipped (line %ld). Left operand:",lv);
			}
			break;
		}
		else {
			dbg->log(0,op_offset,"\"=\" operation expected at pos: %ld");
			declare_script_error(CScriptErrors::ILLEGAL_OPERATION);
		}
	}
	dbg->log(185, "lefthand_operation_execute(): Done");
}

bool CSModelLoader::sysEquOpExec(CScriptValue* left, CScriptValue* right) {
	dbg->log(185, "sysEquOpExec()");
	bool bRIsArray= (right->value_type() == CScriptValue::VAL_ARRAY);
	CScriptValueArray* rArr=NULL; int rArrSize=0;
	if(bRIsArray){if(right->ClassName()!=CScriptValueArray::CLASS_NAME)return false; rArr=(CScriptValueArray*)right; if (rArr!=NULL)rArrSize=rArr->getArraySize(); }

	int left_type = left->value_type();
	if (left_type == CScriptValue::VAL_ARRAY_ELEMENT) {
		if(left->ClassName() != CScriptValueArrayElement::CLASS_NAME) return false; //error?
		CScriptValueArrayElement* l= (CScriptValueArrayElement*)left;
		CScriptValue* l_base=l->GetTitle(); if (l_base==NULL) return false;
		CScriptValue* l_inds=l->GetIndices(); if (l_inds==NULL) return false;
		if (!l_base->check_value_type(CScriptValue::VAL_EXPRESSION)) return false;
		bool bIndsArr = (l_inds->value_type() == CScriptValue::VAL_ARRAY);
		CScriptValueArray* iArr=NULL; int iArrSize=0;

		if(bIndsArr) { if(l_inds->ClassName()!=CScriptValueArray::CLASS_NAME)return false; iArr=(CScriptValueArray*)l_inds; if(iArr!=NULL)iArrSize=iArr->getArraySize(); }
		CESymbol* lBaseSym=l_base->GetValidExprSymbol(); if (lBaseSym==NULL)return false;
		const char* sLName=lBaseSym->get_name();

		if (CEUtil::equals(sLName, "KinematicIndices")) {
			if (bIndsArr) return false;
			bool bIndsSym = false;
			CExpression* iExpr=NULL;
			if (l_inds->check_value(CScriptValue::VAL_EXPRESSION,true)) { iExpr=l_inds->get_expr_value(); } if(iExpr==NULL)return false;
			if (iExpr->ClassName() != CESymbol::CLASS_NAME) return false;
			//const char* iName= ((CESymbol*)iExpr)->get_name();
			if (rArr==NULL) return false;
			CEPtrArray* rec=new CEPtrArray(); rec->add(iExpr->clone());
			for (int rI=0; rI<rArrSize; rI++) {
				CScriptValue* rItm=rArr->getArrayItem(rI); if(rItm==NULL)declare_script_error(CScriptErrors::INTERNAL_ERROR);
				CESymbol* iItmSym=rItm->GetValidExprSymbol();
				if (iItmSym==NULL) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"Symbol expected");
				rec->add(iItmSym->clone());
			}
			m_KinIndices.add(rec);
			return true;
		}

		if (CEUtil::equals(sLName, "IndexRange")) {
			//IndexRange[ Index[Colour] ] = NoUnfold[Range[3]] 
			if (l_inds==NULL) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Parameter required");
			CExpression* indsE=l_inds->TurnToExprCloned();
			CExpression* rightE=right->TurnToExprCloned();
			if ((indsE!=NULL) && (rightE!=NULL)) {
				CSequence* rec = new CSequence();
				rec->addItem(indsE); rec->addItem(rightE);
				m_IndexRanges.addItem(rec); // arr of CSequence { Index[Generation],{1,2,3}}
			} else {
				if (indsE!=NULL) delete indsE;
				if (rightE!=NULL) delete rightE;
				dbg->log(0, m_parser->get_curr_line_num(), m_parser->get_curr_line_pos()-1, " <Unexpected syntax:>{Line: %ld, Pos: %ld}");
			}
			return true;
		}

		if (CEUtil::equals(sLName, "Attributes")) {
			dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"Attributes[] setting is ignored. The default attributes are used. (line: %ld, pos: %ld)"); return true;
		}

		if (CEUtil::equals(sLName, "Format")) {
			dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"Format[] setting is ignored. The default formatting is used. (line: %ld, pos: %ld)"); return true;
		}

		if (CEUtil::equals(sLName, "FourVector")) {
			dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"FourVector[] operations are ignored. The defaults are used. (line: %ld, pos: %ld)"); return true;
		}

		if (CEUtil::equals(sLName, "SpinorType")) {
			dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"SpinorType[] settings are ignored. The defaults are used. (line: %ld, pos: %ld)"); return true;
		}

		if (CEUtil::equals(sLName, "AnalyticalPropagator")) {
			//AnalyticalPropagator[External][ s1 F[j1, mom] ] == NonCommutative[ SpinorType[j1][-mom, Mass[F[j1]]] ]
			//AnalyticalPropagator[Internal][ s1 F[j1, mom] ] == NonCommutative[ DiracSlash[-mom] + Mass[F[j1]] ] *I PropagatorDenominator[mom, Mass[F[j1]]]
			if ((iArr==NULL)||(iArrSize!=2)) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalPropagator parameters invalid");
			CScriptValue* ExtInt=iArr->getArrayItem(0); if (ExtInt==NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
			CScriptValue* Par2=iArr->getArrayItem(1); if (Par2==NULL) declare_script_error(CScriptErrors::INTERNAL_ERROR);
			CESymbol* symExtInt=ExtInt->GetValidExprSymbol(); if(symExtInt==NULL)declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalPropagator parameters invalid");
			//if (CEUtil::equals(symExtInt->get_name(), "External")) { dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"AnalyticalPropagator[External] is ignored. The default one is used. (line: %ld, pos: %ld)"); return true; }
			if (CEUtil::equals(symExtInt->get_name(), "External")) {
				if (m_bGenericPropagatorsFilling) {
					CExpression* par2E=Par2->TurnToExprCloned();
					if (par2E==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Expression parameter expected"); return false; }
					CExpression* rightE=right->TurnToExprCloned();
					if (rightE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Expression value expected"); return false; }
					
					CSequence* rec = new CSequence(); rec->addItem(CENumber::GetZero()); rec->addItem(par2E); rec->addItem(rightE);
					m_GenericPropagators.addItem(rec);
					dbg->log(85, "External generic propagator read");
				} else {
					dbg->log(0, "WARNING. Propagator value setting outside M$GenericPropagators{...}. Value ignored.");
				}
			} else if (CEUtil::equals(symExtInt->get_name(), "Internal")) {
				if (m_bGenericPropagatorsFilling) {
					CExpression* par2E=Par2->TurnToExprCloned();
					if (par2E==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Expression parameter expected"); return false; }
					CExpression* rightE=right->TurnToExprCloned();
					if (rightE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Expression value expected"); return false; }
					
					CSequence* rec = new CSequence(); rec->addItem(CENumber::GetUnit()); rec->addItem(par2E); rec->addItem(rightE);
					m_GenericPropagators.addItem(rec);
					dbg->log(85, "Internal generic propagator read");
				} else {
					dbg->log(0, "WARNING. Propagator value setting outside M$GenericPropagators{...}. Value ignored.");
				}
			} else {
				declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalPropagator parameters invalid"); return false;
			}
			return true;
		}

		if (CEUtil::equals(sLName, "AnalyticalCoupling")) {
			// AnalyticalCoupling[ s1 V[j1, mom1, {li1}], s2 V[j2, mom2, {li2}] ] == G[1][s1 V[j1], s2 V[j2]] .
			// { MetricTensor[li1, li2] ScalarProduct[mom1, mom2],MetricTensor[li1, li2],FourVector[mom1, li2] FourVector[mom2, li1] },

			if (l_inds==NULL) declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling parameters required");

			if (m_bGenericCouplingsFilling) {
				CPhysModelFA::CFAModelGenericCpl* rec = new CPhysModelFA::CFAModelGenericCpl();
				int nFields=0;
				bool fieldsOK=false;
				if (iArr!=NULL) {
					for(int iPar=0; iPar<iArrSize; iPar++) {
						CScriptValue* par=iArr->getArrayItem(iPar); if (par==NULL) { declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "NULL parameter"); delete rec; return false; }
						CExpression* parE=par->TurnToExprCloned();
						if (parE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: Expression parameter expected"); delete rec; return false; }
						fieldsOK = AddFieldInfo(rec, parE); if (!fieldsOK) break;
						nFields++;
					}
				} else {
					CExpression* parE=l_inds->TurnToExprCloned();
					if (parE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: Expression parameter expected"); delete rec; return false; }
					fieldsOK = AddFieldInfo(rec, parE);
					nFields++;
				}
				if (rec->m_partInfos.size() != nFields) { declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "rec->m_partInfos.size() != nFields"); delete rec; return false; }
				if (!fieldsOK) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: Can't resolve fields identificators"); delete rec; return false; }
				CExpression* rightE=right->TurnToExprCloned(); if (rightE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: Expression value expected"); delete rec; return false; }
				if (!rightE->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: Product G[x].ME value expected"); delete rec; return false; }

				CEProduct* prodVal=(CEProduct*)rightE; // Note: product may not be arranged before here!!!
				int prodValS = prodVal->size(); if(prodValS==0){ declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: Product G[x].ME value expected (2)"); delete rec; delete rightE; return false; }
				CExpression* fGE = prodVal->extract_items(0,0);
				if (fGE->ClassName() != CEFunction::CLASS_NAME) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: G[x]-function not found"); delete fGE; delete rec; delete rightE; return false; }
				CEFunction* fG = (CEFunction*)fGE;
				if (!CEUtil::equals(fG->getTitle(),"G")) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: G[x]-function not found (2)"); delete fG; delete rec; delete rightE; return false; }
				CExpression* fG_0=NULL;
				CExpression* fG_1=NULL;
				if (fG->params_num() == 2) {
					fG_0=fG->get_param(0);
					fG_1=fG->get_param(1);
				} else if (fG->params_num() == 1) {
					CExpression* fG_X=fG->get_param(0);
					if (fG_X->ClassName() != CSequence::CLASS_NAME) {
						dbg->log(0, "AnalyticalCoupling: G[x]-function with invalid set of parameters", fG);
						declare_script_error(CScriptErrors::STATEMENT_ILLEGAL); delete fG; delete rec; delete rightE; return false;
					}
					if (((CSequence*)fG_X)->size() != 2) {
						dbg->log(0, "AnalyticalCoupling: G[x]-function with invalid set of parameters", fG);
						declare_script_error(CScriptErrors::STATEMENT_ILLEGAL); delete fG; delete rec; delete rightE; return false;
					}
					fG_0=((CSequence*)fG_X)->GetAt(0);
					fG_1=((CSequence*)fG_X)->GetAt(1);
				} else {
					dbg->log(0, "AnalyticalCoupling: G[x]-function with invalid number of parameters (2 awaited)", fG);
					declare_script_error(CScriptErrors::STATEMENT_ILLEGAL); delete fG; delete rec; delete rightE; return false;
				}
				if (fG_0==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: G[x]-function with invalid parameters (2)"); delete fG; delete rec; delete rightE; return false; }
				if (fG_0->ClassName() != CENumber::CLASS_NAME) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: G[x]-function with invalid parameters (3)"); delete fG; delete rec; delete rightE; return false; }
				int iCjg0 = ((CENumber*)fG_0)->getLong();
				if ((iCjg0 != 1) && (iCjg0 != -1)) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: G[x]-function first parameter should be +1/-1"); delete fG; delete rec; delete rightE; return false; }
				rec->m_bMomsSymmetric=(iCjg0>0);

				// fields check
				if (fG_1->ClassName() != CSequence::CLASS_NAME) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: G[x]-function with invalid field array"); delete fG; delete rec; delete rightE; return false; }
				CSequence* seqFields1=(CSequence*)fG_1;
				if (seqFields1->size() != nFields) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: fields data in G[x]-function does not meet previously declared"); delete fG; delete rec; delete rightE; return false; }
				for (int iField=0; iField<nFields; iField++) {
					CExpression* fItem=seqFields1->GetAt(iField); if(fItem==NULL) { declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "seqFields1->GetAt(iField)==NULL"); delete fG; delete rec; delete rightE; return false; }
					CPhysModelFA::CFAGenericCplPartInfo* iRec = (CPhysModelFA::CFAGenericCplPartInfo*)rec->m_partInfos.GetAt(iField); if(iRec==NULL) { declare_script_error(CScriptErrors::INTERNAL_ERROR, -1, "rec->m_partInfos.GetAt(iField)==NULL"); delete fG; delete rec; delete rightE; return false; }
					bool fitOK=false;
					if (iRec->m_SignSym != NULL) {
						if (fItem->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
							if (((CEProduct*)fItem)->size() == 1) { // compare to  "+/- F[]"
								CExpression* fItem1=((CEProduct*)fItem)->GetAt(0);
								if (fItem1!=NULL) {
									CENumber sNum(1); if(!fItem1->getFlag())sNum.inverseFlag();
									if ((iRec->m_SignSym->equals(&sNum)) && (fItem1->ClassName()==CEFunction::CLASS_NAME)) {
										if (CEUtil::equals(iRec->m_PartTitle.getData(), ((CEFunction*)fItem1)->getTitle())) {
											if ((iRec->m_PartNumber!= NULL) && (((CEFunction*)fItem1)->params_num()>=1)) {
												if (iRec->m_PartNumber->equals( ((CEFunction*)fItem1)->get_param(0) )) {
													fitOK=true;
												}
											}
										}
									}
								}
							} else if (((CEProduct*)fItem)->size() == 2) { // compare to "s F[]"
								CExpression* fItem0=((CEProduct*)fItem)->GetAt(0);
								CExpression* fItem1=((CEProduct*)fItem)->GetAt(1);
								if ((fItem0!=NULL) && (fItem1!=NULL)) {
									if ((iRec->m_SignSym->equals(fItem0)) && (fItem1->ClassName()==CEFunction::CLASS_NAME)) {
										if (CEUtil::equals(iRec->m_PartTitle.getData(), ((CEFunction*)fItem1)->getTitle())) {
											if ((iRec->m_PartNumber!= NULL) && (((CEFunction*)fItem1)->params_num()>=1)) {
												if (iRec->m_PartNumber->equals( ((CEFunction*)fItem1)->get_param(0) )) {
													fitOK=true;
												}
											}
										}
									}
								}
							}
						}
					} else {
						if (fItem->ClassName()==CEFunction::CLASS_NAME) {
							if (CEUtil::equals(iRec->m_PartTitle.getData(), ((CEFunction*)fItem)->getTitle())) {
								if ((iRec->m_PartNumber!= NULL) && (((CEFunction*)fItem)->params_num()>=1)) {
									if (iRec->m_PartNumber->equals( ((CEFunction*)fItem)->get_param(0) )) {
										fitOK=true;
									}
								}
							}
						}
					}
					if (!fitOK) {
						dbg->log(0,"AnalyticalCoupling: fields data in G[x]-function does not meet previously declared!",fItem,iRec);
						declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "AnalyticalCoupling: fields data in G[x]-function does not meet previously declared!"); delete fG; delete rec; delete rightE; return false;
					}
				}
				delete fGE;

				if (rec->m_Value!=NULL) { delete rec->m_Value; rec->m_Value=NULL; }
				rec->m_Value=CExpression::reduce(prodVal);
				dbg->log(85, "Generic coupling read", rec);
				m_GenericCouplings.add(rec);
			} else {
				dbg->log(0, "WARNING. Generic coupling value setting outside M$GenericCouplings{...}. Value ignored.");
			}
			return true;
		}

		if (m_bClassesDescriptionFilling) {
			//sLName
			// V[3]={a->b,c->d,...}
			CExpression* rightE=right->TurnToExprCloned();
			if (rightE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Class Description: Expression value expected"); return false; }
			CExpression* indsE=l_inds->TurnToExprCloned();
			if (indsE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Class Description: Expression (integer) parameter expected"); delete rightE; return false; }
			if (indsE->ClassName() != CENumber::CLASS_NAME) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "Class Description: Integer parameter expected"); delete rightE; delete indsE; return false; }
			CExpression* left_2=new CEFunction(sLName,indsE); delete indsE;
			CSequence* rec = new CSequence(); rec->addItem(left_2); rec->addItem(rightE);
			m_ClassesDescription.addItem(rec);
			dbg->log(85, "Class description read");
			return true;
		}
		
		if (m_bCouplingMatricesFilling && CEUtil::equals(sLName, "C")) {
			// C[...]=...
			CExpression* rightE=right->TurnToExprCloned();
			if (rightE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "CouplingMatrix: Expression value expected"); return false; }
			CExpression* indsE=l_inds->TurnToExprCloned();
			if (indsE==NULL) { declare_script_error(CScriptErrors::STATEMENT_ILLEGAL, -1, "CouplingMatrix: Expression parameter expected"); delete rightE; return false; }

			// Arrange infinities
			rightE=CExpression::arrge(rightE); rightE=CRArrangeInfinity::ArrageInfinity(rightE);
			if (CEBasic::isZeroSeq(rightE)) {
				CSequence* rec = new CSequence(); rec->addItem(indsE); rec->addItem(rightE); // fill rec to log only
				dbg->log(2, "CouplingMatrix equals ZERO and ignored",rec);
				delete rec;
			} else {
				CSequence* rec = new CSequence(); rec->addItem(indsE); rec->addItem(rightE);
				m_CouplingMatrices.addItem(rec);
				dbg->log(85, "CouplingMatrix read");
			}
			return true;
		}
		//return false; // no return -> involve variable further
	} else {
		CESymbol* lValSym=left->GetValidExprSymbol();
		if (lValSym!=NULL) {
			const char* sLName=lValSym->get_name();
			if (CEUtil::equals(sLName, "M$GenericPropagators")) { m_bGenericPropagatorsFilling=false; return true; }
			if (CEUtil::equals(sLName, "M$GenericCouplings")) { m_bGenericCouplingsFilling=false; return true; }
			if (CEUtil::equals(sLName, "M$ClassesDescription")) { m_bClassesDescriptionFilling=false; return true; }
			if (CEUtil::equals(sLName, "M$CouplingMatrices")) { m_bCouplingMatricesFilling=false; return true; }
		}
		//return false; // no return -> involve variable further
	}

	// involve variable
	CExpression* leftE=left->TurnToExprCloned();
	SetVariableValue(leftE, right);
	dbg->log(85, "variable/rule involved", leftE);
	delete leftE;
	return true;
}

// src: s1 V[j1, mom1, {li1}]
bool CSModelLoader::AddFieldInfo(CPhysModelFA::CFAModelGenericCpl* inf, CExpression* src) {
	if ((inf == NULL) || (src==NULL)) return false;
	CExpression* _Sign=NULL;
	CExpression* _Field=NULL;
	if (src->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
		CEProduct* prodParE=(CEProduct*)src;
		if (prodParE->size() != 2) {
			dbg->log(0, prodParE->size(), "ERROR resolving field identifier: %ld-items product spotted"); return false;
		}
		_Sign=prodParE->GetAt(0);
		_Field=prodParE->GetAt(1);
	}

	if (_Sign!=NULL) {
		if (_Sign->ClassName() != CESymbol::CLASS_NAME) {
			dbg->log(0, "ERROR resolving field identifier: unexpected sign symbol", src); return false;
		}
	}
	if (_Field==NULL) { dbg->log(0, "ERROR resolving field identifier: NULL field object", src); return false; }
	if (_Field->ClassName() != CEFunction::CLASS_NAME) { dbg->log(0, "ERROR resolving field identifier: unexpected field object (Function awaited)", src); return false; }
	CEFunction* fField = (CEFunction*)_Field;

	CPhysModelFA::CFAGenericCplPartInfo* rec = new CPhysModelFA::CFAGenericCplPartInfo();
	if (!fField->getFlag()) {
		if (_Sign != NULL) { dbg->log(0, "ERROR resolving field identifier: ambiguous sign info (s * (-1))", src); return false; }
		_Sign=CENumber::GetUnit(); _Sign->inverseFlag(); // -1
	}
	if (_Sign!=NULL) rec->m_SignSym=_Sign->clone();
	rec->m_PartTitle.AddString(fField->getTitle());

	int parNum = fField->params_num();
	CExpression* par0=NULL; CExpression* par1=NULL; CExpression* par2=NULL;
	if (parNum>=2) {
		par0 = fField->get_param(0); if(par0==NULL) { CEException::halt("NULL function parameter #0"); return false; }
		par1 = fField->get_param(1); if(par1==NULL) { CEException::halt("NULL function parameter #1"); return false; }
		if (parNum>=3) {
			par2 = fField->get_param(2); if(par2==NULL) { CEException::halt("NULL function parameter #2"); return false; }
		}
	} else if (parNum==1) {
		CExpression* parOnly=fField->get_param(0);
		if (parOnly->ClassName() == CSequence::CLASS_NAME) {
			CSequence* parSeq=(CSequence*)parOnly;
			if (parSeq->size() >= 1) {
				par0=parSeq->GetAt(0); if(par0==NULL) { CEException::halt("NULL sequence item #0"); return false; }
			}
			if (parSeq->size() >= 2) {
				par1=parSeq->GetAt(1); if(par1==NULL) { CEException::halt("NULL sequence item #1"); return false; }
			}
			if (parSeq->size() >= 3) {
				par2=parSeq->GetAt(2); if(par2==NULL) { CEException::halt("NULL sequence item #2"); return false; }
			}
		} else {
			par0=parOnly;
		}
	}

	if (par0!=NULL) rec->m_PartNumber=par0->clone();
	if (par1!=NULL) rec->m_MomSym=par1->clone();

	if (par2 != NULL) {
		if (par2->ClassName() == CSequence::CLASS_NAME) {
			for(int i=0; i<((CSequence*)par2)->size(); i++) {
				CExpression* ind = ((CSequence*)par2)->GetAt(i); if(ind==NULL) { CEException::halt("NULL vector index"); return false; }
				rec->m_VecIndices.addItem(ind->clone());
			}
		} else {
			rec->m_VecIndices.addItem(par2->clone());
		}
	}
	inf->m_partInfos.add(rec);
	return true;
}

// does not affect input expressions
CExpression* CSModelLoader::expr_operation_binary(const char* c_op, CExpression* _whole, CExpression* _current) {
	CExpression* res = NULL;
	if ((_whole == NULL) || (_current == NULL)) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }

	if (*c_op == '+') {
		res = _whole->clone()->add(_current->clone());
	} else if (*c_op == '-') {
		res = _whole->clone()->minus(_current->clone());
	} else if ((*c_op == '*') || (*c_op == '.')) {
//	} else if (*c_op == '*') {
		res = _whole->clone()->multiply(_current->clone());
	} else if (CEUtil::equals(c_op,"/")) { // compare string due to possible "/;" ("/:" is LEFT-HADNED now)
		res = _whole->clone()->divide(_current->clone());
	} else if (*c_op == '^') {
		res = new CEPower(_whole,_current);
	} else if (CEUtil::equals(c_op,"===")) {
		CExpression* par=_whole->clone()->minus(_current->clone()); res=new CEFunction(TMPFUNC_ConditionEqu0,par); delete par;
	} else if (CEUtil::equals(c_op,"=!=")) {
		CExpression* par=_whole->clone()->minus(_current->clone()); res=new CEFunction(TMPFUNC_ConditionNEqu0,par); delete par;
		//CEPtrArray pars; pars.add(_whole->clone()); pars.add(_current->clone()); res=new CEFunction(TMPFUNC_ConditionNEqu0,&pars);
	} else if (CEUtil::equals(c_op,":")) {
		CEPtrArray pars; pars.add(_whole->clone()); pars.add(_current->clone()); res=new CEFunction(TMPFUNC_VarIfCondition,&pars);
	} else if (CEUtil::equals(c_op,"|")) {
		bool resOK=false;
		if (_whole->ClassName()==CEFunction::CLASS_NAME) {
			CEFunction* f_whole=(CEFunction*)_whole;
			if (CEUtil::equals(f_whole->getTitle(),TMPFUNC_VarIfCondition)) {
				if ((f_whole->getFlag()) && (f_whole->getLevel())) {
					res=f_whole->clone();
					if (res->ClassName()!=CEFunction::CLASS_NAME) CEException::halt();
					((CEFunction*)res)->add_param(_current); // add condition
					resOK=true;
				} else {
					declare_script_error(CScriptErrors::ILLEGAL_OPERATION,-1,"Condition operations (\"_:_|_\") with unexpected sign/level flag spotted");
				}
			}
		}
		if (!resOK) { dbg->log(0,"Operation \"|\" is used in unexpected manner",_whole,_current); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }
	} else if (CEUtil::equals(c_op,"/;")) {
		CEPtrArray pars; pars.add(_whole->clone()); pars.add(_current->clone()); res=new CEFunction(TMPFUNC_LineIfCondition,&pars);
	} else if (CEUtil::equals(c_op,"==")) {
		CExpression* par=_whole->clone()->minus(_current->clone()); res=new CEBooleanEquZero(par); delete par;
	} else if (CEUtil::equals(c_op,"||")) {
		CEBooleanAndOr* eBoolOr=new CEBooleanAndOr(CEBoolean::BOOL_OR); eBoolOr->addOperand(_whole); eBoolOr->addOperand(_current); res=eBoolOr;
	} else if (CEUtil::equals(c_op,"&&")) {
		CEBooleanAndOr* eBoolOr=new CEBooleanAndOr(CEBoolean::BOOL_AND); eBoolOr->addOperand(_whole); eBoolOr->addOperand(_current); res=eBoolOr;
	}

	// "/:" is now LEFT-HANDED
	//else if (CEUtil::equals(c_op,"/:")) { res = _current->clone(); } // truncate the operand before

	if (res==NULL) { // else {
		dbg->log(0, "Unexpected operation symbol: ", c_op);
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION,-1,"Unexpected operation symbol");
	}

	return res;
}

// does not affect input expression
CExpression* CSModelLoader::expr_operation_unitary(const char* c_op, CExpression* _current) {
	CExpression* res = NULL;
	if (_current == NULL) { dbg->log(0, "Error: NULL operand"); declare_script_error(CScriptErrors::ILLEGAL_OPERATION); }

	if (*c_op == '+') {
		res = _current->clone();
	} else if (*c_op == '-') {
		res = _current->clone(); res->inverseFlag();
	} else {
		dbg->log(0, "Unexpected operation symbol: ", c_op);
		m_parser->LogReadOperandsList(95);
		declare_script_error(CScriptErrors::ILLEGAL_OPERATION,-1,"Unexpected operation symbol");
	}

	return res;
}

int CSModelLoader::getOperandPriority(const char* c_op, int op_type) {
	int res = 100;
	if (op_type == CFileParserAdv::OP_OPERATION_RIGHT) {
		if ((*c_op == '+') || (*c_op == '-')) {
			res = 4;
		} else if ((*c_op == '*') || (CEUtil::equals(c_op,"/")) || (*c_op == '.')) { // compare string for "/" due to possible "/;" and "/:"
//		} else if ((*c_op == '*') || (CEUtil::equals(c_op,"/"))) {
			res = 5;
		} else if (*c_op == '^') {
			res = 6;
		} else if ((CEUtil::equals(c_op,":")) || (CEUtil::equals(c_op,"|"))) { // '|' is inline OR operation
			res = 7;
		} else if (CEUtil::equals(c_op,"||") || CEUtil::equals(c_op,"&&")) { // '||' is function-based OR operation
			res = 2;
		} else if (CEUtil::equals(c_op,"===") || CEUtil::equals(c_op,"=!=")) {
			res = 3;
		} else if (CEUtil::equals(c_op,"==")) { // same to ===
			res = 3;
		} else if (CEUtil::equals(c_op,"/;")) {
			res = 1;
		}
		//else if (CEUtil::equals(c_op,"/:")) { res = 8; }
		else { dbg->log(0, "Error: right-hand operation with undeclared priority:"); dbg->log(0, c_op); }
	}
	return res;
}

int CSModelLoader::getFuncSpecInfo(const char* c_op) { // params num, -1 for unknown, INT_MAX for any
	if (CEUtil::equals(c_op,"MetricTensor")) return 2;
	if (CEUtil::equals(c_op,"ScalarProduct")) return 2;
	if (CEUtil::equals(c_op,"FourVector")) return 2;
	if (CEUtil::equals(c_op,"NonCommutative")) return INT_MAX;
	if (CEUtil::equals(c_op,"PropagatorDenominator")) return 2;
	if (CEUtil::equals(c_op,"SpinorType")) return 1;
	if (CEUtil::equals(c_op,"DiracSlash")) return 1;
	if (CEUtil::equals(c_op,"DiracMatrix")) return 1;
	if (CEUtil::equals(c_op,"ChiralityProjector")) return 1;
	if (CEUtil::equals(c_op,"Expand")) return 1;
	if (CEUtil::equals(c_op,"Sqrt")) return 1;
	if (CEUtil::equals(c_op,"Range")) return 1;
	if (CEUtil::equals(c_op,"NoUnfold")) return 1;
	if (CEUtil::equals(c_op,"I")) return 0;
	if (CEUtil::equals(c_op,"If")) return 3;
	return -1;
}

int CSModelLoader::getFuncSpecInfo2(const char* c_op) { // second-[] params num, , INT_MAX for any, INT_MIN if unused
	if (CEUtil::equals(c_op,"SpinorType")) return 2;
	if (CEUtil::equals(c_op,"If")) return INT_MAX;
	return INT_MIN;
}

bool CSModelLoader::bNeedEqualsConvert(const char* c_op) { // do not treat == as = in parameters
	if (CEUtil::equals(c_op,"If")) return true;
	return false;
}

bool CSModelLoader::bIsFuncCommutative(const char* c_op) { // do not treat == as = in parameters
	if (CEUtil::equals(c_op,"SpinorType")) return false;
	if (CEUtil::equals(c_op,"DiracSlash")) return false;
	if (CEUtil::equals(c_op,"ChiralityProjector")) return false;
	return true;
}

CScriptValue* CSModelLoader::getFuncSpecValue(const char* c_op, CSequence* params, CSequence* params2) { // params are unaffected
	dbg->log(95, "getFuncSpecValue: ", c_op);
	dbg->log(95, "params: ", params);
	if (params==NULL) return NULL;
	long s=params->size();
	int sReq=getFuncSpecInfo(c_op); int sReq2=getFuncSpecInfo2(c_op);
	if ((sReq != INT_MAX) && (sReq != s)) { declare_script_error(CScriptErrors::ILLEGAL_OPERATION, -1, "Invalid number of parameters"); return NULL; }
	if ((sReq2>0) && (sReq2!=INT_MAX) && (sReq2!=INT_MIN)) {
		if (params2==NULL) { declare_script_error(CScriptErrors::ILLEGAL_OPERATION, -1, "Invalid set (#2) of parameters"); return NULL; }
		if (params2->size()!=sReq2) { declare_script_error(CScriptErrors::ILLEGAL_OPERATION, -1, "Invalid set (#2) of parameters"); return NULL; }
	}
	
	if (CEUtil::equals(c_op,"MetricTensor")) {
		return CScriptValue::newValue(new CEMetric(params->GetAt(0), params->GetAt(1), false, false),true);
	}
	if (CEUtil::equals(c_op,"ScalarProduct")) {
		return CScriptValue::newValue(new CESProduct(params->GetAt(0), params->GetAt(1)),true);
	}
	if (CEUtil::equals(c_op,"FourVector")) {
		return CScriptValue::newValue(new CEVector(params->GetAt(0), params->GetAt(1), false),true);
	}
	if (CEUtil::equals(c_op,"PropagatorDenominator")) {
		CExpression* res = CENumber::GetUnit();
		res=res->divide(
			(new CESProduct(params->GetAt(0),params->GetAt(0)))->minus(params->GetAt(1)->sqra())
			);
		return CScriptValue::newValue(res,true);
	}
	if (CEUtil::equals(c_op,"NonCommutative")) {
		CExpression* res=NULL;
		for(long i=0; i<s; i++) {
			CExpression* item = params->GetAt(i);
			if(item==NULL) { declare_script_error(CScriptErrors::ILLEGAL_OPERATION, -1, "NULL inline parameter"); return NULL; }
			if (res==NULL) { res=item->clone(); } else { res=res->multiply(item->clone()); }
		}
		if (res==NULL) res=CENumber::GetUnit();
		return CScriptValue::newValue(res,true);
	}
	if (CEUtil::equals(c_op,"DiracSlash")) {
		return CScriptValue::newValue(new CESGammaProduct(params->GetAt(0)),true);
	}
	if (CEUtil::equals(c_op,"DiracMatrix")) {
		return CScriptValue::newValue(new CEGamma(params->GetAt(0),false),true);
	}
	if (CEUtil::equals(c_op,"ChiralityProjector")) {
		CExpression* valE=params->GetAt(0);
		CExpression* resE=NULL;
		if (valE->ClassName()==CENumber::CLASS_NAME) {
			int val=((CENumber*)valE)->getLong();
			resE=new CEGammaSpecial(val);
		} else {
			resE=new CEFunction("ChiralityProjector",valE,false);
			//declare_script_error(CScriptErrors::STATEMENT_ILLEGAL,-1,"ChiralityProjector: Number parameter expected");
		}
		return CScriptValue::newValue(resE,true);
	}
	if (CEUtil::equals(c_op,"Expand")) {
		return CScriptValue::newValue(CExpression::reduce(params->GetAt(0)->clone()),true);
	}
	if (CEUtil::equals(c_op,"Sqrt")) {
		return CScriptValue::newValue(CEPower::sqrt(params->GetAt(0)->clone()),true);
	}
	if (CEUtil::equals(c_op,"Range")) {
		if (params->GetAt(0)->ClassName() == CENumber::CLASS_NAME) {
			long n=((CENumber*)params->GetAt(0))->getLong();
			CSequence* res=new CSequence();
			for(long i=1; i<=n; i++) { res->addItem(CENumber::GetUnit()); }
			return CScriptValue::newValue(res,true);
		} else {
			CExpression* res = new CEFunction(c_op,params->GetAt(0));
			return CScriptValue::newValue(res,true);
		}
	}
	if (CEUtil::equals(c_op, "NoUnfold")) {
		CExpression* res = params->GetAt(0)->clone();
		dbg->log(2,m_parser->get_curr_line_num(),m_parser->get_curr_line_pos(),"Warning{line: %ld, pos: %ld}: NoUnfold[] operation is ignored. Result:",res);
		return CScriptValue::newValue(res,true);
	}
	if (CEUtil::equals(c_op,"I")) {
		return CScriptValue::newValue(CENumber::GetImUnit(),true);
	}
	if (CEUtil::equals(c_op,"SpinorType")) {
		return CScriptValue::newValue(new CESpinor(params2->GetAt(0),false),true);
	}
	if (CEUtil::equals(c_op,"If"))  {
		CFuncIf* eRes = new CFuncIf(params->GetAt(0),params->GetAt(1),params->GetAt(2));
		/*
		*	If: Extra indices will be appended to result:
		*	If[1,a,b,c,d]->a[c,d]; throwing error if a is not a symbol
		*/
		if (params2->size()>0) { for(long i=0;i<params2->size();i++) { eRes->add_param(params2->GetAt(i)); } }
		return CScriptValue::newValue(eRes,true);
	}
	return NULL;
}

bool CSModelLoader::InitParticleList() {
	if (m_phys==NULL) return false;
	CSequence* ranges = &m_IndexRanges; int rangeS=ranges->size();
	CSequence* desc = &m_ClassesDescription; int descS=desc->size();

	// check
	bool initOK=true;
	for(int rangeI=0; rangeI<rangeS; rangeI++) {
		if (!initOK) break;
		CSequence* rangeItem = (CSequence*)ranges->GetAt(rangeI); if (rangeItem==NULL) { dbg->log(0,"m_loader->m_IndexRanges->get()->NULL"); initOK=false; break; }
		CExpression* rangeArrE = rangeItem->GetAt(1); CSequence rangeArr;
		if (rangeArrE->ClassName() == CSequence::CLASS_NAME) { rangeArr.copy_from((CSequence*)rangeArrE); }
		else if (rangeArrE->ClassName() == CENumber::CLASS_NAME) { rangeArr.addItem(rangeArrE->clone()); }
		else {
			dbg->log(0, rangeI, "ERROR: IndexRange[ Index[...] ] (%ld) = should be number or array Range[N]", ranges);
			initOK=false;
			break;
		}
		for(int indexI=0; indexI<rangeArr.size(); indexI++) {
			int indx=-1; bool index_used=false;
			if (rangeArr.GetAt(indexI)->ClassName() == CENumber::CLASS_NAME) { indx = ((CENumber*)rangeArr.GetAt(indexI))->getLong(); }
			if (indx<0) { dbg->log(0,"Integer positive range expected"); initOK=false; break; }
		}
	}
	if (!initOK) return false;

	for(int descI=0; descI<descS; descI++) {
		if (!initOK)break;
		CPhysModelFA::CFAPatricleClassInfo* inf = new CPhysModelFA::CFAPatricleClassInfo();
		CSequence sqsIndices;

		CExpression* descItemE=desc->GetAt(descI);
		if (descItemE==NULL) { dbg->log(0, "Internal ERROR: NULL particle class description"); continue; }
		if (descItemE->ClassName() != CSequence::CLASS_NAME) { dbg->log(0, "Internal ERROR: particle class description is not a CSequence",descItemE); continue; }
		if (((CSequence*)descItemE)->size() < 2) { dbg->log(0, "ERROR: particle class description is not a CSequence of size 2",descItemE); continue; }
		if (((CSequence*)descItemE)->size() > 2) { dbg->log(2, ((CSequence*)descItemE)->size(), "WARNING: particle class description is CSequence of size %ld. Extra parameters are ignored. ",descItemE); }
		CExpression* faPClassID = ((CSequence*)descItemE)->GetAt(0);
		CExpression* descSeqParsE = ((CSequence*)descItemE)->GetAt(1);
		if (faPClassID==NULL) { dbg->log(0, "Internal ERROR: NULL particle class ID"); continue; }
		if (descSeqParsE==NULL) { dbg->log(0, "Internal ERROR: NULL particle class parameters"); continue; }
		if (descSeqParsE->ClassName() != CSequence::CLASS_NAME) { dbg->log(0, "Internal ERROR: particle class parameters is not a CSequence",descSeqParsE); continue; }
		CSequence* descSeq = (CSequence*)descSeqParsE;

		// class name
		if (faPClassID->ClassName() == CEFunction::CLASS_NAME) {
			inf->m_idClassName.Reset(); inf->m_idClassName.AddString(((CEFunction*)faPClassID)->getTitle());
			CExpression* par0=((CEFunction*)faPClassID)->get_param(0);
			inf->m_idNumber.Reset();
			if (par0!=NULL) {
				if (par0->ClassName() == CENumber::CLASS_NAME) { // extra check for now
					inf->m_idNumber.AddString(par0->toString(STORE_TEX));
				}
			}
		} else {
			inf->m_idClassName.Reset(); inf->m_idClassName.AddString(faPClassID->toString(STORE_TEX));
			inf->m_idNumber.Reset();
		}
		if (!initOK)break;

		// params
		dbg->log(75, "descSeq: ", descSeq);
		int iSelfConj=0;
		initOK=ParseDescSequence(descSeq, &inf->m_QuantumNumbers, &sqsIndices, &inf->m_PropagatorLabel, &inf->m_SymMassOrig, &inf->m_iPropagatorArrow, &inf->m_iPropagatorType, &iSelfConj);
		inf->m_SelfConjugated=(iSelfConj>0);

		if(!initOK) { delete inf; break; }
		dbg->log(105,"sqsIndices: ", &sqsIndices);

		if (sqsIndices.size()>0) {
			int s=sqsIndices.size();
			CIntArray rangesPos; rangesPos.ToSize(s); rangesPos.FillAll(0);
			CIntArray rangesMax; rangesMax.ToSize(s); rangesMax.FillAll(-1);
			CIntArray rangesNos; rangesNos.ToSize(s); rangesNos.FillAll(-1);
			for(int i=0; i<s; i++) {
				int rangeNo=-1; int rangeMax=-1;
				for(int rangeI=0; rangeI<rangeS; rangeI++) {
					CSequence* rangeItem = (CSequence*)ranges->GetAt(rangeI);
					if (sqsIndices.GetAt(i)->equals(rangeItem->GetAt(0))) {
						rangeNo=rangeI;
						CExpression* rangeArrE = rangeItem->GetAt(1);
						if (rangeArrE->ClassName() == CSequence::CLASS_NAME) {
							rangeMax=((CSequence*)rangeArrE)->size();
						} else if (rangeArrE->ClassName() == CENumber::CLASS_NAME) {
							rangeMax=1;
						}
						break;
					}
				}
				rangesNos.SetAt(i,rangeNo); rangesMax.SetAt(i,rangeMax);
			}
			dbg->log(105,"rangesMax: ", &rangesMax);

			for(;;) {
				dbg->log(105,"rangesPos: ", &rangesPos);
				CPhysModelFA::CFAPatricleClassInfo* infI = new CPhysModelFA::CFAPatricleClassInfo(*inf);

				int i;
				for(i=0; i<s; i++) {
					int val=-1;
					CSequence* rangeItem = (CSequence*)ranges->GetAt(rangesNos.GetAt(i)); if (rangeItem==NULL) { initOK=false; break; }
					CExpression* rangeArrE = rangeItem->GetAt(1);
					CENumber* replNum=NULL;
					if (rangeArrE->ClassName() == CSequence::CLASS_NAME) {
						rangeArrE=((CSequence*)rangeArrE)->GetAt(rangesPos.GetAt(i));
						if (rangeArrE==NULL) { initOK=false; break; }
						if (rangeArrE->ClassName() == CENumber::CLASS_NAME) {
							replNum=(CENumber*)rangeArrE; val=replNum->getLong();
						} else { initOK=false; break; }
					} else if ((i==0)&&(rangeArrE->ClassName() == CENumber::CLASS_NAME)) {
						replNum=((CENumber*)rangeArrE); val=replNum->getLong();
					} else { initOK=false; break; }
					if (val<0) { initOK=false; break; }

					infI->m_Indices.Add(val);

					// index exchange
					if (replNum!=NULL) {
						CEReplasement r(sqsIndices.GetAt(i),replNum,false);
						if (infI->m_SymMassOrig!=NULL) infI->m_SymMassOrig=r.ApplyTo(infI->m_SymMassOrig);
						if (infI->m_SymMass!=NULL) infI->m_SymMass=r.ApplyTo(infI->m_SymMass);
						if (infI->m_PropagatorLabel!=NULL) infI->m_PropagatorLabel=r.ApplyTo(infI->m_PropagatorLabel);
						if (infI->m_PropagatorLabelM!=NULL) infI->m_PropagatorLabelM=r.ApplyTo(infI->m_PropagatorLabelM);
						//dbg->log(2,"inf->m_PropagatorLabel rule:",&r);
					}
				}
				if (initOK) {
					// adjust masses -- add indices
					// adjust masses according to MQU[i]-like variables
					// OLD (what is was about?): { Note: mass rules appear already applied here }
					if (infI->m_SymMassOrig!=NULL) {
						CExpression* mOld = infI->m_SymMassOrig;
						//dbg->log(2,"Mass adjust", mOld);
						CExpression* mAdj = GetVarRuleValueECloned(mOld);
						if ((mAdj==NULL) && (mOld->ClassName()==CESymbol::CLASS_NAME)) {
							CEFunction f(((CESymbol*)mOld)->get_name());
							for(i=0; i<infI->m_Indices.GetSize(); i++) {
								CENumber nPar(infI->m_Indices.GetAt(i));
								f.add_param(&nPar);
								mAdj=GetVarRuleValueECloned(&f); if (mAdj!=NULL)break;
							}
						}
						//dbg->log(2,"Mass adjust result", mAdj);
						if (mAdj!=NULL) { infI->m_SymMassOrig = mAdj; delete mOld; }
					}

					// evaluate ComposedChar[] function and fill inf->m_PropagatorLabelM
					if (infI->m_PropagatorLabel != NULL) {
						CExpression* lbl=infI->m_PropagatorLabel;
						infI->m_PropagatorLabel=CSModelLoader::EvaluateComposedChar(lbl,STORE_TEX,false);
						if (infI->m_PropagatorLabelM != NULL) { delete infI->m_PropagatorLabelM; infI->m_PropagatorLabelM=NULL; }
						if (lbl==infI->m_PropagatorLabel) lbl=lbl->clone();
						infI->m_PropagatorLabelM=CSModelLoader::EvaluateComposedChar(lbl,STORE_WR_MATHEMATICA,true);
					}

					if(initOK){
						initOK=InitParticleInfo(infI);
						if (!initOK) { dbg->log(0, "ERROR initialization particle info (PK or internal symbol)", infI); }
					}
				}

				if (initOK) {
					m_phys->m_ParticleInfos.add(infI);
				} else {
					delete infI; break;
				}

				int rPos=s-1;
				for(; rPos>=0; ) {
					int tmp=rangesPos.GetAt(rPos)+1;
					if (tmp<rangesMax.GetAt(rPos)) {
						rangesPos.SetAt(rPos,tmp); break;
					} else {
						rangesPos.SetAt(rPos,0); rPos--;
					}
				}
				if (rPos<0) break;
			}
			delete inf;
		} else {
			// add inf without index changing
			// evaluate ComposedChar[] function and fill inf->m_PropagatorLabelM
			if (inf->m_PropagatorLabel != NULL) {
				CExpression* lbl=inf->m_PropagatorLabel;
				inf->m_PropagatorLabel=CSModelLoader::EvaluateComposedChar(lbl,STORE_TEX,false);
				if (inf->m_PropagatorLabelM != NULL) { delete inf->m_PropagatorLabelM; inf->m_PropagatorLabelM=NULL; }
				if (lbl==inf->m_PropagatorLabel) lbl=lbl->clone();
				inf->m_PropagatorLabelM=CSModelLoader::EvaluateComposedChar(lbl,STORE_WR_MATHEMATICA,true);
			}

			if(initOK){
				initOK=InitParticleInfo(inf);
				if (!initOK) { dbg->log(0, "ERROR initialization particle info (PK or internal symbol)", inf); }
			}
			if(initOK) { m_phys->m_ParticleInfos.add(inf); } else { delete inf; break; }
		}
	}

	if (initOK) {
		dbg->log(15, "Particles list inited:", &m_phys->m_ParticleInfos);
	}
	return initOK;
}

bool CSModelLoader::InitPropagatorsList() {
	if(m_phys==NULL)return false;
	long i; long nRulesCreated=0;
	//m_GenericPropagators
	bool propsOK=true;
	long partN=m_phys->GetParticleClassesNum();
	for(long partI=0;partI<partN;partI++) {
		if(!propsOK)break;
		CPhysModelFA::CFAPatricleClassInfo* inf = m_phys->GetParticleClassInfo(partI); if(inf==NULL)continue;
		long idNumber=-1; if (!CEUtil::atolSafe(inf->m_idNumber.getData(),&idNumber)) idNumber=-1;
		CEReplList r;
		bool bExtFound=false; bool bIntFound=false;

		// find appropriate generic propagator
		for(i=0;i<m_GenericPropagators.size();i++) {
			if(!propsOK)break;
			CSequence* rec = (CSequence*)m_GenericPropagators.GetAt(i); if(rec==NULL)continue;
			if(rec->size() != 3) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure"); propsOK=false; break; }
			CExpression* eiFlag=rec->GetAt(0);
			CExpression* par0=rec->GetAt(1);
			CExpression* val=rec->GetAt(2);
			if ((eiFlag==NULL)||(par0==NULL)||(val==NULL)) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure (unexpected NULL)"); propsOK=false; break; }
			bool isExt=CEBasic::isZero(eiFlag);
			if (isExt) { if(bExtFound)continue; } else { if(bIntFound)continue; }

			CExpression* par=NULL;
			if (par0->getProperty(CEProperties::PROP_MASK_IS_PRODUCT)) {
				if (((CEProduct*)par0)->size() != 2) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure"); propsOK=false; break; }
				CExpression* signItm=((CEProduct*)par0)->GetAt(0); if (signItm==NULL) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure (unexpected NULL)"); propsOK=false; break; }
				if (signItm->ClassName() != CESymbol::CLASS_NAME) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Sign symbol is expected: "); m_MsgLastError.AddString(signItm->toString()); propsOK=false; break; }
				par=((CEProduct*)par0)->GetAt(1);
				if (par==NULL) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure (unexpected NULL)"); propsOK=false; break; }
			} else {
				par=par0;
			}
			if (par->ClassName() != CEFunction::CLASS_NAME) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Particle function is expected: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
			CEFunction* fPart=(CEFunction*)par;
			const char* rec_pID=fPart->getTitle(); if(rec_pID==NULL) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Particle function is expected: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
			if (!CEUtil::equals(rec_pID,inf->m_idClassName.getData())) continue;
			int fPartS=fPart->params_num();
			if (fPartS<1) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Particle function invalid: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
			CExpression* ind0=NULL; CExpression* ind1=NULL; CExpression* ind2=NULL; CExpression* ind3=NULL;
			if (fPartS>=2) {
				ind0=fPart->get_param(0); ind1=fPart->get_param(1);
				if (fPartS>=3) ind2=fPart->get_param(2);
				if (fPartS>=4) ind3=fPart->get_param(3);
				if ((fPartS>4) || (ind0==NULL) || (ind1==NULL)) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Particle function invalid: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
			} else if (fPartS==1) {
				CExpression* indX=fPart->get_param(0); if(indX==NULL) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Particle function invalid: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
				if (indX->ClassName()==CSequence::CLASS_NAME) {
					if (((CSequence*)indX)->size() > 0) ind0=((CSequence*)indX)->GetAt(0);
					if (((CSequence*)indX)->size() > 1) ind1=((CSequence*)indX)->GetAt(1);
					if (((CSequence*)indX)->size() > 2) ind2=((CSequence*)indX)->GetAt(2);
					if (((CSequence*)indX)->size() > 3) ind3=((CSequence*)indX)->GetAt(3);
					if ((((CSequence*)indX)->size()>4) || (ind0==NULL) || (ind1==NULL)) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Particle function invalid: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
				} else {
					ind0=indX;
					m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Particle function invalid: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break;
				}
			}
			if (!propsOK) break;
			// Now: ind0:id(!=NULL), ind1:mom(!=NULL), ind2:vInd1, ind3:vInd2

			if (idNumber>=0) { CENumber nmbI(idNumber); r.add(ind0,&nmbI,false); }
			else { dbg->log(0,"WARNING: Non-NULL particle symbol but no index value are specified at particle class",inf); }

			CExpression* _Mom=NULL; CExpression* _vInd1=NULL; CExpression* _vInd2=NULL; CExpression* _Value=NULL;
			_Mom=ind1; _vInd1=ind2; _vInd2=ind3; _Value=val;

			if ((isExt) && (_vInd2!=NULL)) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Second unexpected index at external propagator: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }

			if (_vInd1!=NULL) {
				if (_vInd1->ClassName()==CSequence::CLASS_NAME) {
					if (((CSequence*)_vInd1)->size() != 1) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Invalid particle function index: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
					_vInd1=((CSequence*)_vInd1)->GetAt(0);
					if (_vInd1==NULL) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Invalid particle function index: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
				}
			}
			if (_vInd2!=NULL) {
				if (_vInd2->ClassName()==CSequence::CLASS_NAME) {
					if (((CSequence*)_vInd2)->size() != 1) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Invalid particle function index: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
					_vInd2=((CSequence*)_vInd2)->GetAt(0);
					if (_vInd2==NULL) { m_MsgLastError.AddString("Invalid INTERNAL generic propagator structure. Invalid particle function index: "); m_MsgLastError.AddString(par->toString()); propsOK=false; break; }
				}
			}
			_Value=r.ApplyTo(_Value->clone());

			// evaluate functions
			CEPtrArray rProcs_DA; rProcs_DA.add(this); rProcs_DA.add(m_phys);
			CESocketRule rProcs((void*)CSModelLoader::EvaluateCouplRProcs, &rProcs_DA,
				CESocketRule::USE_EXTERNAL_DATA_ARRAY|CESocketRule::SR_APPLY_TO_CHANGED_SOURCE|CESocketRule::SR_APPLY_TO_SUBITEMS_FIRST);
			_Value = rProcs.ApplyTo(_Value);
			//val=CExpression::reduce(val);
			CESocketRule rVars((void*)CSModelLoader::EvaluateSubstRules, &rProcs_DA, CESocketRule::USE_EXTERNAL_DATA_ARRAY);
			_Value = rVars.ApplyTo(_Value); // evaluate variables/repl rules
			rProcs_DA.clearall_without_delete_objects();
			//val=CExpression::arrge(val);

			// Exchange momentum at fermionic tale/propagator. TODO: remove this after fermionic momentum/sign problem fixing!
			if (CEParticle::is_fermion_PK(inf->m_PK)) {
				if (!_Mom->getFlag()) { dbg->log(0, "ERROR: signed momentum symbol is not allowed",inf); CEException::halt(); }
				CExpression* mMom=_Mom->clone(); mMom->inverseFlag();
				CEReplasement r(_Mom,mMom); delete mMom;
				_Value=r.ApplyTo(_Value);
			}

			// Arrange value
			_Value=CExpression::arrge(_Value);

			// Append rule even if it is zero
			long inf_pk=inf->m_PK; long inf_pk_abs=((inf_pk>=0)?inf_pk:-inf_pk); // is pk_abs always same to pk?
			if (isExt) {
				if ((inf_pk_abs & CEParticle::PARTICLE_VECTOR_BOSON) == CEParticle::PARTICLE_VECTOR_BOSON) {
					// Parameters to check:
					//		1. rec_pID=="V"
					//		2. number (numeric or symbol)
					//		3. _Mom
					//		4. _vInd1
					CExpression* _NToCheck=ind0; CENumber nmbI((long)idNumber); if(idNumber>=0)_NToCheck=&nmbI;
					_Value=ExtractPolVector(_Value,rec_pID,_NToCheck,_Mom,_vInd1);
					if (_Value==NULL) {
						dbg->log(0,"ERROR: Invalid vector boson external propagator.");
						dbg->log(0,POL_VEC_NAME,"[,,] object is either absent or invalid.");
						m_MsgLastError.AddString("Invalid vector boson external propagator.");
						dbg->log(0,"Propagator data attached.",rec,rec->GetAt(2));
						propsOK=false;
						break;
					}
				}
				propsOK=m_phys->addExtLineCorr(inf_pk,_Value,_Mom,inf->m_SymMass,_vInd1);
				if(propsOK)bExtFound=true;
				else { m_MsgLastError.AddString(m_phys->m_LastErrorMsg.getData()); }
			} else {
				// Arrange infinities
				_Value=CRArrangeInfinity::ArrageInfinity(_Value);

#ifdef LOADMODEL_DROP_ZERO_PROPS // Do not use this mechanism! Inf vertex * 1/Inf propagator is allowed! TODO: fix it!
				_Value=CRArrangeInfinity::ArrageInfinityToUnit(_Value);
				if (CEBasic::isZeroProd(_Value)) { m_ZeroPropagatorPKs.Add(inf_pk_abs); }
#endif //LOADMODEL_DROP_ZERO_PROPS

				// add propagator (and zero as well)
				propsOK=m_phys->addIntLineCorr(inf_pk,_Value,_Mom,inf->m_SymMass,_vInd1,_vInd2);
				if(propsOK)bIntFound=true;
				else { m_MsgLastError.AddString(m_phys->m_LastErrorMsg.getData()); }
			}
			delete _Value;
			if(!propsOK)break;
			nRulesCreated++;
			if (bIntFound && bExtFound) break;
		}
		if ((!bIntFound) || (!bExtFound)) {
			// Note: NO-EXTERNAL is OK for pure virtual particle

			if ((bExtFound) && (!bIntFound)) {
				dbg->log(0,"WARNING: Can't find INTERNAL generic propagator for particle (EXTERNAL is found).",inf);
			}

			//m_MsgLastError.AddString(inf->m_idClassName.getData()); m_MsgLastError.AddString("["); m_MsgLastError.AddString(inf->m_idNumber.getData()); m_MsgLastError.AddString("]");
			//propsOK=false; break;
		}
	}
	dbg->log(2, nRulesCreated, partN, "Total %ld propagator rule(s) created (total particles: %ld)");
	return propsOK;
}

//STORE_WR_MATHEMATICA,STORE_TEX
CExpression* CSModelLoader::EvaluateComposedChar(CExpression* src, int mode, bool del_src) {
	if (src==NULL) return NULL;
	CExpression* res=NULL;
	CEPtrArray data; data.add(CENumber::GetNumber(mode));
	CESocketRule r((void*)CSModelLoader::EvaluateComposedCharRProc,&data);
	if (del_src) {
		res = r.ApplyTo(src);
	} else {
		res = r.ApplyTo(src->clone());
	}
	return res;
}

CExpression* CSModelLoader::EvaluateComposedCharRProc(CExpression* src, CEPtrArray* data_arr) {
	if (src == NULL) return src;
	if (src->ClassName()!=CEFunction::CLASS_NAME) return src;
	CEFunction* f=(CEFunction*)src;
	if (!CEUtil::equals( f->getTitle(), "ComposedChar")) return src;

	int mode=0; CENumber* modeE=(CENumber*)data_arr->GetAt(0); if(modeE!=NULL)mode=modeE->getLong();
	int s=f->params_num();
	CEString str;
	// mode: STORE_WR_MATHEMATICA-> \\(G\\&A\\), STORE_TEX-> G^{+}_{W}
	if ((mode==STORE_WR_MATHEMATICA)&&(s>0)) str.AddString("\\(");
	int cnt=0;
	for(int i=0; i<s; i++) {
		cnt++;
		CExpression* item=f->get_param(i); if(item==NULL)continue; if(CENull::IsNull(item))continue;
		if (item->ClassName() == CESymbol::CLASS_NAME) { if(CEUtil::equals(((CESymbol*)item)->get_name(),"Null"))continue; }
		if (item->ClassName() == CSequence::CLASS_NAME) {
			for(int j=0; j<((CSequence*)item)->size(); j++) {
				if (j>0) cnt++;
				CExpression* item2=((CSequence*)item)->GetAt(j); if(item2==NULL)continue; if(CENull::IsNull(item2))continue;
				if (item2->ClassName() == CESymbol::CLASS_NAME) { if(CEUtil::equals(((CESymbol*)item2)->get_name(),"Null"))continue; }

				if (mode==STORE_WR_MATHEMATICA) {
					if (cnt==2) { str.AddString("\\_"); } else if (cnt==3) { str.AddString("\\^"); }
				} else if (mode==STORE_TEX) {
					if (cnt==2) { str.AddString("_"); } else if (cnt==3) { str.AddString("^"); }
				} // else add nothing -- create solid symbol

				CEString* s_item2 = item2->toString(mode); if(s_item2==NULL)s_item2=new CEString();
				if (mode==STORE_WR_MATHEMATICA) {
					const char* sbst=CConst::GetValidSymMSubst(s_item2->getData()); if(sbst!=NULL){s_item2->Reset();s_item2->AddString(sbst);}
				} else if (mode==STORE_TEX) {
					const char* sbst=CConst::GetValidSymTexSubst(s_item2->getData()); if(sbst!=NULL){s_item2->Reset();s_item2->AddString(sbst);}
				}
				str.AddString(s_item2);
			}
		} else {
			if (mode==STORE_WR_MATHEMATICA) {
				if (cnt==2) { str.AddString("\\_"); } else if (cnt==3) { str.AddString("\\^"); }
			} else if (mode==STORE_TEX) {
				if (cnt==2) { str.AddString("_"); } else if (cnt==3) { str.AddString("^"); }
			} // else add nothing -- create solid symbol

			CEString* s_item = item->toString(mode); if(s_item==NULL)s_item=new CEString();
			if (mode==STORE_WR_MATHEMATICA) {
				const char* sbst=CConst::GetValidSymMSubst(s_item->getData()); if(sbst!=NULL){s_item->Reset();s_item->AddString(sbst);}
			} else if (mode==STORE_TEX) {
				const char* sbst=CConst::GetValidSymTexSubst(s_item->getData()); if(sbst!=NULL){s_item->Reset();s_item->AddString(sbst);}
			}
			str.AddString(s_item);
		}
	}
	if ((mode==STORE_WR_MATHEMATICA)&&(s>0)) str.AddString("\\)");
	CExpression* res=new CESymbol(str.getData());
	src->copyAttributesTo(res,true);
	return res;
}

CExpression* CSModelLoader::EvaluateIndexDeltaRProc(CExpression* src, CEPtrArray* data_arr) {
	if (src == NULL) return src;
	if (src->ClassName()!=CEFunction::CLASS_NAME) return src;
	CEFunction* f=(CEFunction*)src;
	if (!CEUtil::equals( f->getTitle(), "IndexDelta")) return src;

	int s=f->params_num();
	CExpression* par0=NULL; CExpression* par1=NULL;
	if (s==2) {
		par0=f->get_param(0);
		par1=f->get_param(1);
	} else if (s==1) {
		CExpression* par=f->get_param(0); if(par==NULL)return src;
		if (par->ClassName()!=CSequence::CLASS_NAME) return src;
		if (((CSequence*)par)->size() != 2) return src;
		par0=((CSequence*)par)->GetAt(0);
		par1=((CSequence*)par)->GetAt(1);
	} else {
		return src;
	}
	if ((par0==NULL) || (par1==NULL)) return src;
	
	CExpression* res=NULL;
	if (par0->equals(par1)) {
		res=CENumber::GetUnit();
	} else {
		res=CENumber::GetZero();
	}
	src->copyAttributesTo(res,true);
	return res;
}

/*
*	var should be "+"-signed and leveled, does not affects input
*/
void CSModelLoader::SetVariableValue(CExpression* var, CScriptValue* val) {
	dbg->log(185, "SetVariableValue()");
	if (var==NULL) return;
	if ((!var->getFlag()) || (!var->getLevel())){
		if (!var->getLevel()) {
			dbg->log(0, "Invalid SetVariableValue() call: trying to  involve non-simple variable (check sign and level)", var, val); CEException::halt();
		}
		// allow -x=something, but ignore it!
		dbg->log(2, "WARNING: trying to involve non-simple variable (-Var=Value). Statement ignored", var);
		return;
	}

	CExpression* valE=((val!=NULL)?val->TurnToExprCloned():NULL);

	// template is either left or right!
	CEPtrArray arrExtrConditions;
	CESocketRule r((void*)CSModelLoader::ExtractConditionsRProc,&arrExtrConditions,CESocketRule::SR_APPLY_TO_SUBITEMS_FIRST|CESocketRule::USE_EXTERNAL_DATA_ARRAY);
	var=r.ApplyTo(var->clone()); if(var==NULL) CEException::halt("Internal ERROR",__FILE__,__LINE__);
	//dbg->log(185, "SetVariableValue(): arrExtrConditions",&arrExtrConditions);

	bool bIsTemplateLeft=CEMathEquRule::IsMathTemplateLeft(var);
	//dbg->log(185, bIsTemplateLeft, "SetVariableValue(): bIsTemplateLeft: %ld");

	CExpression* eLineCondition=NULL;
	if (valE!=NULL) {
		if (valE->ClassName()==CEFunction::CLASS_NAME) {
			CEFunction* f_val = (CEFunction*)valE;
			if (CEUtil::equals(f_val->getTitle(),TMPFUNC_LineIfCondition)) { //TMPFUNC_VarIfCondition
				if (!f_val->getFlag()) { dbg->log(0, "Error: line with condition have unexpected sign: ", val); CEException::halt(); delete valE; delete var; return; }
				if (!f_val->getLevel()) { dbg->log(0, "Error: line with condition is unexpected denominator: ", val); CEException::halt(); delete valE; delete var; return; }
				if (f_val->params_num() != 2) { dbg->log(0, "Error: line with condition have not supported form (2 params are expected): ", val); CEException::halt(); delete valE; delete var; return; } 
				valE=f_val->get_param(0); if(valE==NULL)CEException::halt(NULL,__FILE__,__LINE__);
				valE=valE->clone();
				eLineCondition=f_val->get_param(1); if(eLineCondition==NULL)CEException::halt(NULL,__FILE__,__LINE__);
				eLineCondition=eLineCondition->clone();
				delete f_val;
			}
		}
	}
	//dbg->log(185, "SetVariableValue(): eLineCondition",eLineCondition);

	if ((bIsTemplateLeft) || (eLineCondition!=NULL) || (arrExtrConditions.size()>0)) {
		long s=m_MathEquRules.size();
		CEMathEquRule* mRuleNew=NULL; bool ruleOK=true;
		if (val!=NULL) {
			if (valE != NULL) {
				mRuleNew=new CEMathEquRule(var, valE);
				if (eLineCondition != NULL) {
					CRuleConditionEquZero* cnd=NULL;
					if (eLineCondition->ClassName()==CEFunction::CLASS_NAME) {
						CEFunction* fLineCondition = (CEFunction*)eLineCondition; int fLC_parN=fLineCondition->params_num();
						if ((fLC_parN==1) && (CEUtil::equals(fLineCondition->getTitle(),TMPFUNC_ConditionEqu0))) {
							cnd=new CRuleConditionEquZero(true,fLineCondition->get_param(0));
						} else if ((fLC_parN==1) && (CEUtil::equals(fLineCondition->getTitle(),TMPFUNC_ConditionNEqu0))) {
							cnd=new CRuleConditionEquZero(false,fLineCondition->get_param(0));
						}
						//else { dbg->log(0, "ERROR: line condition is not a valid/supported boolean expression. Condition IGNORED!", eLineCondition); }
					}
					if (cnd==NULL) {
						cnd=new CRuleConditionEquZero(true,eLineCondition);
					}
					if (cnd!=NULL) mRuleNew->AddCondition(cnd);
				}

				for (int iC=0;iC<arrExtrConditions.size();) {
					CRuleCondition* cnd=(CRuleCondition*)arrExtrConditions.ExtractAt(0);
					if (cnd!=NULL) mRuleNew->AddCondition(cnd);
				}
				//dbg->log(0,"condition-dependent rule created: ",mRuleNew);
			} else {
				ruleOK=false; dbg->log(0,"ERROR: Non-expression rule result. Rule ignored.", var, val);
			}
		}
		if (ruleOK) {
			bool rFound=false;
			for(long i=0; i<s; i++) {
				CEMathEquRule* mRule = (CEMathEquRule*)m_MathEquRules.GetAt(i); if(mRule==NULL)continue;
				if (!mRule->IsSimilarTo(var)) continue;
				if (mRuleNew != NULL) {
					m_MathEquRules.SetAt(i,mRuleNew,true); mRuleNew=NULL;
				} else {
					m_MathEquRules.RemoveAt(i,true);
				}
				rFound=true; break;
			}
			if ((!rFound) && (mRuleNew!=NULL)) { m_MathEquRules.add(mRuleNew); mRuleNew=NULL; }
		}
		if (mRuleNew!=NULL) delete mRuleNew; // unused check
	} else {
		long s=m_Variables.size();
		bool bVarUsed=false;
		for(long i=0; i<s; i++) {
			CEPtrArray* item = (CEPtrArray*)m_Variables.GetAt(i); if(item==NULL)continue;
			if (item->size()<1) continue;
			if (var->equals((CExpression*)item->GetAt(0))) {
				if (item->size()<2) {
					if (val!=NULL) item->add(val->clone());
				} else {
					if (val!=NULL) { item->SetAt(1,val->clone()); } else { item->RemoveAt(1,true); }
				}
				bVarUsed=true; break;
			}
		}
		if (!bVarUsed) {
			CEPtrArray* item = new CEPtrArray();
			item->add(var->clone()); if(val!=NULL)item->add(val->clone());
			m_Variables.add(item);
		}
	}
	if (valE!=NULL) delete valE;
	if (var!=NULL) delete var;
	if (eLineCondition!=NULL) delete eLineCondition;
}

CExpression* CSModelLoader::ExtractConditionsRProc(CExpression* src, CEPtrArray* data_arr) {
	if ((src == NULL) || (data_arr==NULL)) return src;
	if (src->ClassName()!=CEFunction::CLASS_NAME) return src;
	CEFunction* f=(CEFunction*)src;
	int s=f->params_num();

	if (CEUtil::equals(f->getTitle(),TMPFUNC_VarIfCondition)) {
		if (s<=1) { dbg->log(0,"ERROR: Condition-dependent value spotted with NO condition.", f); CEException::halt(); return src; }
		CExpression* res=f->get_param(0); if(res==NULL)CEException::halt("Internal ERROR",__FILE__,__LINE__);
		res=res->clone();
		f->copyAttributesTo(res,true);

		// add (res-parI)==0 conditions
		CRuleCondition* cnd=NULL;
		if (s==2) {
			CExpression* cEqu0=f->get_param(1); if(cEqu0==NULL)CEException::halt("Internal ERROR",__FILE__,__LINE__);
			cEqu0=res->clone()->minus(cEqu0->clone());
			cnd=new CRuleConditionEquZero(true,cEqu0); delete cEqu0;
		} else { // s>2
			CRuleConditionAndOr* cndAnd=new CRuleConditionAndOr(CRuleConditionAndOr::COND_OR);
			for (int i=1;i<s;i++) {
				CExpression* cEquI=f->get_param(i); if(cEquI==NULL)CEException::halt("Internal ERROR",__FILE__,__LINE__);
				cEquI=res->clone()->minus(cEquI->clone());
				cndAnd->AddCondition(new CRuleConditionEquZero(true,cEquI)); delete cEquI;
			}
			cnd=cndAnd;
		}
		if (cnd!=NULL) data_arr->add(cnd);

		// make symbol template if required ("a" --> "a_")
		if (res->ClassName()==CESymbol::CLASS_NAME) {
			const char* s_sym = ((CESymbol*)res)->get_name();
			if (s_sym!=NULL) {
				bool bIsTemplate=false;
				int s_symL=CEUtil::length(s_sym); if(s_symL>0) { if(*(s_sym+s_symL-1)=='_')bIsTemplate=true; }
				if(!bIsTemplate) {
					CEString s_symT(s_sym); s_symT.AddString("_"); CESymbol* resT=new CESymbol(s_symT.getData());
					res->copyAttributesTo(resT,true); delete res; res=resT;
				}
			}
		} else {
			dbg->log(0,"WARNING: condition rule for non-symbolic operand",src);
		}
		return res;
	}
	return src;
}

CScriptValue* CSModelLoader::GetVariableValue(CExpression* var) { // var should be "+"-signed and leveled
	if (var==NULL) return NULL;
	if ((!var->getFlag()) || (!var->getLevel())){
		dbg->log(0, "INTERNAL ERROR: trying to  resolve non-simple variable (check sign and level)", var); CEException::halt(); return NULL;
	}
	long s=m_Variables.size();
	for(long i=0; i<s; i++) {
		CEPtrArray* item = (CEPtrArray*)m_Variables.GetAt(i); if(item==NULL)continue;
		if (item->size()<1) continue;
		if (var->equals((CExpression*)item->GetAt(0))) {
			if (item->size()<2) return NULL;
			return (CScriptValue*)item->GetAt(1);
		}
	}
	return NULL;
}

CExpression* CSModelLoader::GetMathRuleValueCloned(CExpression* var) { // any signed input
	if (var==NULL) return NULL;
	if (!var->getFlag()) {
		var->inverseFlag(); CExpression* res=GetMathRuleValueCloned(var); var->inverseFlag(); if (res!=NULL)res->inverseFlag(); return res;
	}
	if (!var->getLevel()) {
		var->inverseLevel(); CExpression* res=GetMathRuleValueCloned(var); var->inverseLevel(); if (res!=NULL) res=CEBasic::inverseLevelAdv(res); return res;
	}

	//bool bLog = ((var->ClassName() == CEFunction::CLASS_NAME) && CEUtil::equals(((CEFunction*)var)->getTitle(),"Mass"));
	//if (bLog) dbg->log(2,"MASS-------------------",var);
	long s=m_MathEquRules.size();
	for(long i=0; i<s; i++) {
		CEMathEquRule* mRule = (CEMathEquRule*)m_MathEquRules.GetAt(i); if(mRule==NULL)continue;
		//if (bLog) dbg->log(2,"mRule-------------------",mRule);
		if (!mRule->IsApplicableTo(var)) continue;
		CExpression* res=mRule->ApplyTo(var->clone());
		//if (bLog) dbg->log(2,"res-------------------",res);
		return res;
	}
	return NULL;
}

CExpression* CSModelLoader::GetVarRuleValueECloned(CExpression* var) { // any signed input
	if (var==NULL) return NULL;
	if (!var->getFlag()) {
		var->inverseFlag(); CExpression* res=GetVarRuleValueECloned(var); var->inverseFlag(); if (res!=NULL)res->inverseFlag(); return res;
	}
	if (!var->getLevel()) {
		var->inverseLevel(); CExpression* res=GetVarRuleValueECloned(var); var->inverseLevel(); if (res!=NULL) res=CEBasic::inverseLevelAdv(res); return res;
	}
	CExpression* res = GetMathRuleValueCloned(var);
	if (res==NULL) {
		CScriptValue* sval = GetVariableValue(var);
		if (sval!=NULL) res=sval->TurnToExprCloned();
	}
	if (res != NULL) {
		// try again
		CExpression* res2 = GetVarRuleValueECloned(res);
		if ((res2 != NULL) && (res2 != res)) {
			delete res; res=res2;
		}
	}
	return res;
}

/*
*	data_arr->0: CSModelLoader*
*	data_arr->1: CPhysModelFA*
*	IndexSum
*/
CExpression* CSModelLoader::EvaluateCouplRProcs(CExpression* src, CEPtrArray* data_arr) {
	if (src == NULL) return src;
	if (src->ClassName()!=CEFunction::CLASS_NAME) return src;
	CEFunction* f=(CEFunction*)src;
	int s=f->params_num();
	CExpression* par0=NULL; CExpression* par1=NULL; CExpression* par2=NULL;

	if (data_arr==NULL) { dbg->log(0, "INTERNAL ERROR: CSModelLoader::EvaluateCouplRProcs() wrong usage. NULL data array"); return src; }
	if (data_arr->size()<2) { dbg->log(0, "INTERNAL ERROR: CSModelLoader::EvaluateCouplRProcs() wrong usage. Invalid data array"); return src; }
	CSModelLoader* ldr=(CSModelLoader*)data_arr->GetAt(0);
	CPhysModelFA* phy=(CPhysModelFA*)data_arr->GetAt(1);
	if ((ldr==NULL) || (phy==NULL))return src;

	if (CEUtil::equals( f->getTitle(), "IndexSum")) {
		if (s==2) {
			par0=f->get_param(0); par1=f->get_param(1);
		} else if (s==1) {
			CExpression* par=f->get_param(0); if(par==NULL)return src;
			if (par->ClassName()!=CSequence::CLASS_NAME) return src;
			if (((CSequence*)par)->size() != 2) return src;
			par0=((CSequence*)par)->GetAt(0); par1=((CSequence*)par)->GetAt(1);
		} else { return src; }
		
		if ((par0==NULL) || (par1==NULL)) return src;
		if (par1->ClassName()!=CSequence::CLASS_NAME) return src;
		if (((CSequence*)par1)->size()!=2) return src;
		CExpression* sym = ((CSequence*)par1)->GetAt(0); CExpression* max = ((CSequence*)par1)->GetAt(1);
		if ((sym==NULL) || (max==NULL)) return src;
		long nMax=-1;
		if (max->ClassName() == CENumber::CLASS_NAME) nMax=((CENumber*)max)->getLong();
		if (nMax<0) return src;
		
		CESum* res=new CESum();
		//dbg->log(2, nMax, "nMax: %ld src: ", src);
		for (int i=1; i<=nMax; i++) {
			CENumber nVal(i);
			CEReplasement rVal(sym, &nVal, false);
			CExpression* item = par0->clone();
			item = rVal.ApplyTo(item);
			res->addItem(item);
		}
		//dbg->log(2, "res ", res); CEException::halt();
		src->copyAttributesTo(res,true);
		return res;
	}

	if (CEUtil::equals( f->getTitle(), "OddQ")) {
		if (s==1) {
			par0=f->get_param(0); if (par0==NULL) return src;
		} else return src;
		if (par0->ClassName()!=CENumber::CLASS_NAME) return src;
		CENumber* nPar= (CENumber*)par0;
		if (!nPar->is_integer()) return src;
		bool bRes=((nPar->getLong() % 2)==1);
		CExpression* res = (bRes?CENumber::GetUnit():CENumber::GetZero()); src->copyAttributesTo(res,true);
		return res;
	}

	if (CEUtil::equals( f->getTitle(), "If")) {
		//dbg->log(0,"If:",f);
		if (s==3) {
			par0=f->get_param(0); par1=f->get_param(1); par2=f->get_param(2);
		} else if (s==1) {
			CExpression* par=f->get_param(0); if(par==NULL)return src;
			if (par->ClassName()!=CSequence::CLASS_NAME) return src;
			if (((CSequence*)par)->size() != 3) return src;
			par0=((CSequence*)par)->GetAt(0); par1=((CSequence*)par)->GetAt(1); par2=((CSequence*)par)->GetAt(2);
		} else return src;
		if ((par0==NULL)||(par1==NULL)||(par2==NULL)) return src;
		if (par0->ClassName()!=CENumber::CLASS_NAME) return src;
		CENumber* nPar= (CENumber*)par0;
		if (!nPar->is_integer()) return src;
		bool bSwitch1=(nPar->getLong()>0);
		CExpression* res = (bSwitch1?par1->clone():par2->clone()); src->copyAttributesTo(res,true);
		//dbg->log(0,"If: res",res);
		return res;
	}

	if (CEUtil::equals( f->getTitle(), "ChiralityProjector")) {
		if (s==1) {
			par0=f->get_param(0); if (par0==NULL) return src;
		} else return src;
		if (par0->ClassName()!=CENumber::CLASS_NAME) return src;
		CENumber* nPar= (CENumber*)par0;
		if (!nPar->is_integer()) return src;
		int iPar=nPar->getLong();
		CExpression* res = new CEGammaSpecial(iPar);
		src->copyAttributesTo(res,true);
		return res;
	}

	if (CEUtil::equals( f->getTitle(), "AddHC")) {
		if (s==1) {
			CExpression* par=f->get_param(0); if(par==NULL)return src;
			if (par->ClassName()==CSequence::CLASS_NAME) {
				if (((CSequence*)par)->size() == 1) par0=((CSequence*)par)->GetAt(0);
			} else {
				par0=par;
			}
		}
		if (par0==NULL) { dbg->log(2,"WARNING. AddHC: non-supported parameters set",src); return src; }
		CExpression* par0c=CEBasic::Congugate(par0,false); if(par0c==par0)par0c=par0c->clone();
		CExpression* res = (par0->clone()->add(par0c))->divide(CENumber::GetTwo());
		res=CExpression::arrge(res);
		//dbg->log(2,"AddHC src,res:",src,res);
		return res;
	}

	if (CEUtil::equals( f->getTitle(), "CKM")) {
		if (s==2) {
			par0=f->get_param(0); par1=f->get_param(1);
		} else if (s==1) {
			CExpression* par=f->get_param(0); if(par==NULL)return src;
			if (par->ClassName()!=CSequence::CLASS_NAME) return src;
			if (((CSequence*)par)->size() != 2) return src;
			par0=((CSequence*)par)->GetAt(0); par1=((CSequence*)par)->GetAt(1);
		} else return src;
		if ((par0==NULL)||(par1==NULL)) return src;
		if ((par0->ClassName()!=CENumber::CLASS_NAME) || (par1->ClassName()!=CENumber::CLASS_NAME)) return src;
		CENumber* nPar0= (CENumber*)par0; CENumber* nPar1= (CENumber*)par1;
		if ((!nPar0->is_integer()) || (!nPar1->is_integer())) return src;
		long n0=nPar0->getLong(); long n1=nPar1->getLong();
		long pkQuark0=-1; long pkQuark1=-1;
		long cnum = phy->GetParticleClassesNum(); long ci;
		for(ci=0;ci<cnum;ci++) { // non-conjugted quark
			CPhysModelFA::CFAPatricleClassInfo* inf=phy->GetParticleClassInfo(ci); if(inf==NULL)continue; if(inf->m_PK==0)continue;
			if ((inf->m_PK & (CEParticle::PARTICLE_QUARK|CEParticle::PARTICLE_QUARK_CONJUGETED)) == CEParticle::PARTICLE_QUARK) {
				if ((inf->m_Indices.GetSize()>0) && (inf->m_Indices.GetAt(0) == n0)) { pkQuark0=inf->m_PK; break; }
			}
		}
		for(ci=0;ci<cnum;ci++) { // conjugted quark
			CPhysModelFA::CFAPatricleClassInfo* inf=phy->GetParticleClassInfo(ci); if(inf==NULL)continue; if(inf->m_PK==0)continue;
			if ((inf->m_PK & CEParticle::PARTICLE_QUARK_CONJUGETED) == CEParticle::PARTICLE_QUARK_CONJUGETED) {
				if ((inf->m_Indices.GetSize()>0) && (inf->m_Indices.GetAt(0) == n1)) { pkQuark1=inf->m_PK; break; }
			}
		}
		if ((pkQuark0<0) || (pkQuark1<0)) return src;
		//dbg->log(0,"CKM:",f);
		CExpression* res = new CQuarkMixMatrix(pkQuark0, pkQuark1); src->copyAttributesTo(res,true);
		//dbg->log(0,"CKM res:",res);
		return res;
	}

	if (CEUtil::equals( f->getTitle(), "Mass")) {
		if (s!=1) return src;
		CExpression* par=f->get_param(0);
		if (par->ClassName()!=CEFunction::CLASS_NAME) return src;
		CEFunction* fPart=(CEFunction*)par;
		const char* cName=fPart->getTitle(); if(cName==NULL)return src;
		bool dataOK=true;
		int fPartS=fPart->params_num();
		if (fPartS==2) {
			par0=fPart->get_param(0); par1=fPart->get_param(1);
			if ((par0==NULL)||(par1==NULL))dataOK=false;
		} else if (fPartS==1) {
			CExpression* parX=fPart->get_param(0); if(par==NULL)return src;
			if (parX->ClassName()==CSequence::CLASS_NAME) {
				if (((CSequence*)parX)->size() > 0) {
					par0=((CSequence*)parX)->GetAt(0); if(par0==NULL)dataOK=false;
				}
				if (((CSequence*)parX)->size() > 1) {
					par1=((CSequence*)parX)->GetAt(1); if(par1==NULL)dataOK=false;
				}
			} else {
				par0=parX;
			}
		}
		if (!dataOK) return src;

		char* cNumber=NULL;
		if (par0!=NULL) { cNumber=par0->toLine(); } else { cNumber=CEUtil::copyString(""); }

		CIntArray _Indices;
		if ((dataOK) && (par1 != NULL)) {
			if (par1->ClassName() == CSequence::CLASS_NAME) {
				int indsN=((CSequence*)par1)->size();
				for(int indsI=0; indsI<indsN; indsI++) {
					if (!dataOK) break;
					CExpression* indsItem=((CSequence*)par1)->GetAt(indsI); if(indsItem==NULL) { dataOK=false; break; }
					if (indsItem->ClassName() == CENumber::CLASS_NAME) {
						_Indices.Add(((CENumber*)indsItem)->getLong());
					} else {
						dataOK=false; break;
					}
				}
			} else if (par1->ClassName() == CENumber::CLASS_NAME) {
				_Indices.Add(((CENumber*)par1)->getLong());
			} else {
				dataOK=false;
			}
		}
		CExpression* res=NULL;
		if (dataOK) {
			CPhysModelFA::CFAPatricleClassInfo* inf = phy->GetParticleClassInfo(cName, cNumber, &_Indices, true);
			if (inf!=NULL) {
				res=inf->m_SymMass;
				if (res==NULL) res=inf->m_SymMassOrig;
				if (res!=NULL) {
					res=res->clone();
					src->copyAttributesTo(res,true);
				}
				//if (res==NULL) { dbg->log(2, "src, inf", src, inf, res); CEException::halt(); }
			}
			//else { dbg->log(2, "inf==NULL", src); CEException::halt(); }
		}
		//else { dbg->log(2, "!dataOK", src); CEException::halt(); }

		if(res==NULL) { dbg->log(45, "Mass function: mass not found: ", src); }
		CMemoryManager::mfree(cNumber);
		return res;
	}

	//FermionCharge
	return src;
}

CExpression* CSModelLoader::EvaluateSubstRules(CExpression* src, CEPtrArray* data_arr) { // m_Variables & m_MathEquRules
	if (src == NULL) return src;
	if (data_arr==NULL) { dbg->log(0, "INTERNAL ERROR: CSModelLoader::EvaluateSubstRules() wrong usage. NULL data array"); return src; }
	if (data_arr->size()<1) { dbg->log(0, "INTERNAL ERROR: CSModelLoader::EvaluateSubstRules() wrong usage. Invalid data array"); return src; }
	CSModelLoader* ldr=(CSModelLoader*)data_arr->GetAt(0);
	if (ldr==NULL) return src;
	CExpression* res = ldr->GetVarRuleValueECloned(src);
	if (res == NULL) res=src;
	return res;
}

/*
*	Called after InitCouplingsList() is completed. Uses class couplings from m_phys->m_VertexClasses, stored at InitCouplingsList().
*/
bool CSModelLoader::InitCouplingsClasses() {
	if (m_phys==NULL) return false;
	long nClRulesCreated=0;
	bool init_cpls_OK=true;
	int cplS=m_phys->m_VertexClasses.size();
	for(int cplI=0; cplI<cplS; cplI++) {
		if (!init_cpls_OK) break;
		CPhysModelFA::CFAModelGenericCpl* genI = (CPhysModelFA::CFAModelGenericCpl*)m_phys->m_VertexClasses.GetAt(cplI); if(genI==NULL) { CEException::halt("NULL m_phys->m_VertexClasses item"); return false; }
		dbg->log(45, "class coupling -> class vertex", genI);
		int partN = genI->m_partInfos.size();

		CExpression* val = genI->m_Value; if (val==NULL) { CEException::halt("m_phys->m_VertexClasses::m_Value==NULL"); return false; }
		val=val->clone(); bool valOK=true; // val may be: v, {v,v}
		dbg->log(145, "class coupling pre-value", val);
		if (val->ClassName() == CSequence::CLASS_NAME) { // summarize sequence {v1,v2}->v1+v2
			CESum* val2=new CESum(); for(long i2=0; i2<((CSequence*)val)->size(); i2++) { val2->addItem(((CSequence*)val)->GetAt(i2)->clone()); }
			val->copyAttributesTo(val2,true); delete val; val=val2;
		}
		if (CEBasic::isZeroSeq(val)) valOK=false;
		if ((partN<2) || (partN>4)) { valOK=false; dbg->log(0, partN, "ERROR. Correspondence classes rule can't be installed. Vertex with %ld tales is not supported."); }
		if (!valOK) { dbg->log(2,"Class coupling dropped (invalid value): ", genI); if(val!=NULL)delete val; init_cpls_OK=false; continue; }

		CExpression* vCorr=val; CExpression* vCorrCT=val;
		if (partN<=2) vCorr=NULL; // no 2-tale vertices
		int pk1=0; int pk2=0; int pk3=0; int pk4=0;
		CExpression* vInd1=NULL; CExpression* vInd2=NULL; CExpression* vInd3=NULL; CExpression* vInd4=NULL;
		CExpression* pPart1=NULL; CExpression* pPart2=NULL; CExpression* pPart3=NULL; CExpression* pPart4=NULL;

		int i;
		for(i=0; i<partN; i++) {
			CPhysModelFA::CFAGenericCplPartInfo* genInf=(CPhysModelFA::CFAGenericCplPartInfo*)genI->m_partInfos.GetAt(i); if(genInf==NULL) { CEException::halt("NULL m_phys->m_VertexClasses.m_partInfos item"); return false; }
			/* <CFAGenericCplPartInfo:>
			CExpression* m_SignSym; // si or -1
			CEString m_PartTitle; // F,V
			CExpression* m_PartNumber; // symbol -> j1
			CExpression* m_MomSym; // momi
			CSequence m_VecIndices; // without 'covariance' info */
			CExpression* vInd=NULL;
			if (genInf->m_VecIndices.size()>0) {
				if (genInf->m_VecIndices.size()>1) { dbg->log(0, "ERROR: Tensor particles are not implemented.",genInf); CEException::halt(); continue; }
				vInd=genInf->m_VecIndices.GetAt(0);
			}
			long pkI=0;
			const char* pTitle=genInf->m_PartTitle.getData();
			if (CEUtil::equals(pTitle,"F")) { pkI=CEParticle::PARTICLE_FERMION; }
			else if (CEUtil::equals(pTitle,"U")) { pkI=CEParticle::PARTICLE_FPGHOST; }
			else if (CEUtil::equals(pTitle,"S")) { pkI=CEParticle::PARTICLE_SCALAR; }
			else if (CEUtil::equals(pTitle,"SV")) { pkI=CEParticle::PARTICLE_SCALAR; }
			else if (CEUtil::equals(pTitle,"V")) { pkI=CEParticle::PARTICLE_VECTOR_BOSON; }
			else { dbg->log(0, "ERROR: Unexpected field class: ",pTitle); init_cpls_OK=false; break; }
			if (genInf->m_bIsNeutral) { pkI|=CEParticle::PARTICLE_NEUTRAL_BOSON; }
			pkI|=CEParticle::PARTICLE_CLASS_FLAG;

			CExpression* _Sign=genInf->m_SignSym;
			if(_Sign!=NULL) { // else mean "+1"
				if (_Sign->ClassName()==CENumber::CLASS_NAME) {
					long iSign=((CENumber*)_Sign)->getLong();
					if ((iSign!=1) && (iSign!=-1)) { dbg->log(0, cplI, i, "ERROR: Numeric sign expected at m_phys->m_VertexClasses.GetAt(%ld)->m_partInfos.GetAt(%ld)->m_SignSym."); CEException::halt(); }
					if ((iSign<0) && ((pkI&CEParticle::PARTICLE_NEUTRAL_BOSON)!=CEParticle::PARTICLE_NEUTRAL_BOSON)) pkI=-pkI;
				} else {
					dbg->log(0, cplI, i, "ERROR: Numeric sign expected at m_phys->m_VertexClasses.GetAt(%ld)->m_partInfos.GetAt(%ld)->m_SignSym."); CEException::halt();
				}
			}
			switch(i) {
			case 0: pPart1=genInf->m_MomSym; vInd1=vInd; pk1=pkI; break;
			case 1: pPart2=genInf->m_MomSym; vInd2=vInd; pk2=pkI; break;
			case 2: pPart3=genInf->m_MomSym; vInd3=vInd; pk3=pkI; break;
			case 3: pPart4=genInf->m_MomSym; vInd4=vInd; pk4=pkI; break;
			}
		}
		
		bool addedOK=false;
		if (init_cpls_OK) {
			addedOK=m_phys->addVertexCorr(pk1, pk2, pk3, pk4, vCorr, vCorrCT, pPart1, vInd1, pPart2, vInd2, pPart3, vInd3, pPart4, vInd4, NULL, cplI);
		}
		if (addedOK) {
			dbg->log(25,"Class rule added", genI); nClRulesCreated++;
		} else {
			if (m_phys->m_LastErrorMsg.length() > 0) {
				dbg->log(0,"Class rule addition failed. ERROR: ", m_phys->m_LastErrorMsg.getData());
			} else {
				dbg->log(0,"Class rule addition failed. ERROR UNKNOWN.");
			}
			dbg->log(2,"Class coupling tried was:", genI);
			init_cpls_OK=false; // break & retrun false
		}
		if (val!=NULL) delete val;
	}
	dbg->log(2, nClRulesCreated, "Total %ld class-vertex rule(s) created");
	if (init_cpls_OK) {
		dbg->log(55, "Classes Hamiltonian",m_phys->getClassesHamiltonianE());
		dbg->log(55, "Classes Hamiltonian (renorm)",m_phys->getClassesHamiltonianRen());
	}
	return init_cpls_OK;
}

bool CSModelLoader::InitCouplingsList() {
	long nRulesCreated=0;
	bool init_cpls_OK=true;
	int cplS=m_CouplingMatrices.size();
	//	*****************************************************	<scroll>
	CEProgressWatch* pw = NULL;
	if (cplS>5) { pw = CEProcessor::newProgressWatch("Couplings conversion (total %ld couplings)", cplS); }
	//	*****************************************************	<\scroll>
	for(int cplI=0; cplI<cplS; cplI++) {
		if (!init_cpls_OK) break;
		//	*****************************************************	<scroll>
		if (pw != NULL) pw->step(cplI, cplS);
		//	*****************************************************	<\scroll>
		// extract particles info
		CExpression* cplItemE=m_CouplingMatrices.GetAt(cplI); if (cplItemE==NULL) { CEException::halt("NULL m_CouplingMatrices item"); continue; }
		if (cplItemE->ClassName()!=CSequence::CLASS_NAME) { dbg->log(0, "ERROR: invalid coupling structure. Coupling ignored", cplItemE); continue; }
		CSequence* cpl=(CSequence*)cplItemE;
		if (cpl->size() != 2) { dbg->log(0, "ERROR: invalid coupling structure. Coupling ignored", cplItemE); continue; }

		CExpression* partInfoE=cpl->GetAt(0); if (partInfoE==NULL) { CEException::halt("ERROR: partInfoE==NULL"); continue; }
		if (partInfoE->ClassName()!=CSequence::CLASS_NAME) { dbg->log(0, "ERROR: invalid coupling structure (particle info is not seq). Coupling ignored", cplItemE); continue; }
		CSequence* partInfo=(CSequence*)partInfoE;
		CExpression* cplValueE=cpl->GetAt(1); if (cplValueE==NULL) { CEException::halt("ERROR: cplValueE==NULL"); continue; }

		int partN=partInfo->size();
		CEPtrArray arr_idCNames;
		CIntArray arr_idPSign; // signs before particle
		CEPtrArray arr_idNums; // array of CEString*
		CSequence seq_idNums; // seq of CENumbers* but same to arr_idNums
		CSequence arr_idIndices; // seq of CSequence*
		int i;
		for(i=0; i<partN; i++) {
			CExpression* partID = partInfo->GetAt(i);
			CEString* sIdCNm=new CEString(); CEString* sIdNum=new CEString(); CSequence* inds = new CSequence();
			CENumber* _nIdNum=NULL;
			int _Sign=1; if (!partID->getFlag()) _Sign=-1;
			if (partID->ClassName() == CEFunction::CLASS_NAME) {
				sIdCNm->AddString(((CEFunction*)partID)->getTitle());
				CExpression* par0=NULL;
				CExpression* par1=NULL;
				if (((CEFunction*)partID)->params_num()==1) {
					CExpression* parOnly=((CEFunction*)partID)->get_param(0);
					if (parOnly!=NULL) {
						if (parOnly->ClassName() == CSequence::CLASS_NAME) {
							if (((CSequence*)parOnly)->size()>=1) par0=((CSequence*)parOnly)->GetAt(0);
							if (((CSequence*)parOnly)->size()>=2) par1=((CSequence*)parOnly)->GetAt(1);
						} else {
							par0=parOnly;
						}
					}
				} else {
					if (((CEFunction*)partID)->params_num()>=2) {
						par0=((CEFunction*)partID)->get_param(0); par1=((CEFunction*)partID)->get_param(1);
					}
				}
				
				if (par0!=NULL) {
					if (par0->ClassName() == CENumber::CLASS_NAME) { // extra check for now
						_nIdNum=(CENumber*)par0;
						sIdNum->AddString(par0->toString(STORE_TEX));
					}
				}

				if (par1!=NULL) {
					if (par1->ClassName() == CSequence::CLASS_NAME) {
						inds->copy_from((CSequence*)par1);
					} else {
						inds->addItem(par1->clone());
					}
				}
			}
			arr_idCNames.add(sIdCNm); arr_idNums.add(sIdNum); arr_idIndices.addItem(inds);
			if (_nIdNum != NULL) { seq_idNums.addItem(_nIdNum->clone()); } else { seq_idNums.addItem(new CENull()); }
			arr_idPSign.Add(_Sign);
		}
		
		dbg->log(95, "arr_idCNames, arr_idNums, arr_idIndices", &arr_idCNames, &arr_idNums, &arr_idIndices);
		dbg->log(95, "seq_idNums, arr_idPSign", &seq_idNums, &arr_idPSign);
		
		CEPtrArray partInfos; CEPtrArray partInfos0; CIntArray partInfosPrev; CIntArray partInfosPrev0;
		for(i=0;i<partN;i++) {
			int tmp=-1;
			CPhysModelFA::CFAPatricleClassInfo* inf = m_phys->GetNextParticleClassInfo(((CEString*)arr_idCNames.GetAt(i))->getData(), ((CEString*)arr_idNums.GetAt(i))->getData(), &tmp);
			if (inf == NULL) {
				dbg->log(0, i, "ERROR: particle #%ld not found", cpl);
				partInfos.clearall_without_delete_objects(); partInfos0.clearall_without_delete_objects(); CEException::halt();
			}
			partInfos.add(inf); partInfos0.add(inf);
			partInfosPrev.Add(tmp); partInfosPrev0.Add(tmp);
		}

		// partInfos: current particles list, partInfos0: list with initial index values
		dbg->log(45, "partInfos-initial", &partInfos0);
		dbg->log(45, "partInfos index positions initial", &partInfosPrev0); 

		if (arr_idCNames.size() != partN) CEException::halt("INTERNAL ERROR arr_idCNames.size()");
		if (arr_idNums.size() != partN) CEException::halt("INTERNAL ERROR arr_idNums.size()");
		if (seq_idNums.size() != partN) CEException::halt("INTERNAL ERROR seq_idNums.size()");
		if (arr_idPSign.GetSize() != partN) CEException::halt("INTERNAL ERROR arr_idPSign.GetSize()");
		bool bFirstLoop=true;
		long iIndConfigTried=0; long iIndConfigFilled=0;
		for(;;) {
			if (!init_cpls_OK) break;
			// select next particles set
			if (!bFirstLoop) {
				for(i=partN-1; i>=0; ) {
					int tmp=partInfosPrev.GetAt(i);
					CPhysModelFA::CFAPatricleClassInfo* inf = m_phys->GetNextParticleClassInfo(((CEString*)arr_idCNames.GetAt(i))->getData(), ((CEString*)arr_idNums.GetAt(i))->getData(), &tmp);
					if (inf != NULL) {
						partInfos.SetAt(i,inf,false);
						partInfosPrev.SetAt(i,tmp);
						break;
					} else {
						partInfos.SetAt(i,partInfos0.GetAt(i),false);
						partInfosPrev.SetAt(i,partInfosPrev0.GetAt(i));
						i--;
					}
				}
				if(i<0) break;
			} else { bFirstLoop=false; }
			iIndConfigTried++;
			dbg->log(45, iIndConfigTried, "partInfos current (%ld):", &partInfos);
			dbg->log(45, "partInfos index positions current:", &partInfosPrev);

			// check quantum-number-indices and create index repl rule
			bool dropRec=false;
			CEReplList r;
			for(i=0; i<partN; i++) {
				if (dropRec) break;
				CPhysModelFA::CFAPatricleClassInfo* inf = (CPhysModelFA::CFAPatricleClassInfo*)partInfos.GetAt(i); if(inf==NULL) { CEException::halt("NULL particle info array item"); continue; }
				CSequence* idInds=(CSequence*)arr_idIndices.GetAt(i);
				int indS=inf->m_Indices.GetSize();
				if (indS != idInds->size()) { dbg->log(0, "ERROR: unexpected num of indices", cpl); CEException::halt(); break; }
				for(int indI=0; indI<indS; indI++) {
					CExpression* indE = idInds->GetAt(indI); if(indE==NULL) { CEException::halt("INTERNAL ERROR: NULL index"); continue; }
					int valI=inf->m_Indices.GetAt(indI);
					if (indE->ClassName()==CENumber::CLASS_NAME) {
						dbg->log(0,"ERROR1",inf,idInds); CEException::halt();
						if (valI != ((CENumber*)indE)->getLong()) { dropRec=true; break; }
					} else {
						CENumber nmbI(valI); r.add(indE,&nmbI,false);
					}
				}
			}
			if(dropRec) {
				dbg->log(45,"Index configuration doesn't meet to particle predefined indices"); continue;
			}

			CExpression* val = cplValueE->clone();
			val = r.ApplyTo(val);
			//dbg->log(15, "coupling pre-value 0", val);
			val = CExpression::reduce(val); // not-required reduce
			//dbg->log(15, "coupling pre-value 1", val);
			CESocketRule rID((void*)CSModelLoader::EvaluateIndexDeltaRProc); val = rID.ApplyTo(val); // evaluate Delta-functions

			//dbg->log(15, "coupling pre-value 2", val);
			val = CExpression::reduce(val); // and reduce again (required)
			//dbg->log(95, "coupling pre-value", val);

			// check for zero couplings: 0, {0,0}, {{0,0},{0}}
			if (CEBasic::isZeroSeq(val)) {
				dbg->log(45,"Zero coupling value dropped out",val); delete val; continue; // drop coupling
			}

			// found generic coupling
			CPhysModelFA::CFAModelGenericCpl* genCpl=NULL;
			CEReplList r_genPartsNum;
			for(i=0; i<m_GenericCouplings.size(); i++) {
				CPhysModelFA::CFAModelGenericCpl* genI=(CPhysModelFA::CFAModelGenericCpl*)m_GenericCouplings.GetAt(i); if(genI==NULL)continue;
				if (partN != genI->m_partInfos.size()) continue;
				bool fitsOK=true;
				int partI=0;
				for(partI=0; partI<partN; partI++) {
					if (!fitsOK) break;
					CEString* idCN_I=(CEString*)arr_idCNames.GetAt(partI); if(idCN_I==NULL) { CEException::halt("idCN_i==NULL"); fitsOK=false; break; }
					CExpression* idNM_I=seq_idNums.GetAt(partI); if(idNM_I==NULL) { CEException::halt("idNM_I==NULL"); fitsOK=false; break; }
					CPhysModelFA::CFAGenericCplPartInfo* genInf=(CPhysModelFA::CFAGenericCplPartInfo*)genI->m_partInfos.GetAt(partI); if (genInf==NULL) { fitsOK=false; break; }
					if (!CEUtil::equals(genInf->m_PartTitle.getData(), idCN_I->getData())) { fitsOK=false; break; }
					if ((genInf->m_PartNumber != NULL) && (!CENull::IsNull(idNM_I))) {
						// genInf->m_PartNumber: j1, idNM_I: 1,2,3,...
					} else {
						if ((genInf->m_PartNumber != NULL) || (!CENull::IsNull(idNM_I))) { fitsOK=false; break; }
					}

					// signs
					long iSign = arr_idPSign.GetAt(i); if ((iSign!=1) && (iSign!=-1)) { dbg->log(0,iSign,"ERROR: arr_idPSign.GetAt() enexpected sign value: %ld"); CEException::halt(); fitsOK=false; break; }
					CExpression* _Sign = genInf->m_SignSym;
					if(_Sign!=NULL) {
						if (_Sign->ClassName() == CENumber::CLASS_NAME) {
							if (((CENumber*)_Sign)->getLong() != iSign) { fitsOK=false; break; }
						} else {
							if (_Sign->ClassName() != CESymbol::CLASS_NAME) { dbg->log(0,"ERROR: Unexpected generic coupling SIGN object",_Sign,genInf); CEException::halt(); fitsOK=false; break; }
						}
					} else {
						if (iSign != 1) { fitsOK=false; break; }
					}
				}
				if (!fitsOK) { continue; }

				// fill repl-list
				for(partI=0; partI<partN; partI++) {
					CExpression* idNM_I=seq_idNums.GetAt(partI); if(idNM_I==NULL) { CEException::halt("idNM_I==NULL"); break; }
					CPhysModelFA::CFAGenericCplPartInfo* genInf=(CPhysModelFA::CFAGenericCplPartInfo*)genI->m_partInfos.GetAt(partI); if(genInf==NULL)break;
					if ((genInf->m_PartNumber != NULL) && (!CENull::IsNull(idNM_I))) {
						r_genPartsNum.add(genInf->m_PartNumber,idNM_I,false);
					}
				}
				genCpl=genI; break;
			}
			//dbg->log(15, "genCpl", genCpl);
			if (genCpl==NULL) { dbg->log(0, "ERROR: no generic coupling found", cpl); CEException::halt(); break; }
			CExpression* valG=genCpl->m_Value;
			if (valG==NULL) { dbg->log(0,"NULL generic coupling value found",cpl); CEException::halt(); break; }

			// val may be: v, {v,v}, {{v,v},v,{v,v}}
			// gen may be: v, {v,v}
			//if (valG->ClassName() != CSequence::CLASS_NAME) { dbg->log(0,"Coupling value does not meet to generic coupling",cpl,valG); valOK=false; CEException::halt(); break; }
			dbg->log(45, "coupling pre-value", val);
			dbg->log(45, "coupling generic", valG);
			bool valOK=true;
			bool bLogValueIfDropped=true; // to skip logging of non-error cases
			if (val->ClassName() == CSequence::CLASS_NAME) {
				if (valG->ClassName() != CSequence::CLASS_NAME) {
					val= val->multiply(valG->clone());
				} else {
					int valS=((CSequence*)val)->size();
					if (valS != ((CSequence*)valG)->size()) { dbg->log(0,"Coupling value does not meet to generic coupling",cpl,val,valG); valOK=false; CEException::halt(); break; }
					for(int valI=0; valI<valS; valI++) {
						if (!valOK) break;
						CExpression* val_2 = ((CSequence*)val)->GetAt(valI); CExpression* valG_2 = ((CSequence*)valG)->GetAt(valI);
						if ((val_2==NULL) || (valG_2==NULL)) { dbg->log(0, "INTERNAL ERROR: NULL sequence element"); valOK=false; break; }
						if (valG_2->ClassName() == CSequence::CLASS_NAME) { dbg->log(0,"Unexpected generic coupling structure",valG,cpl); valOK=false; CEException::halt(); break; }
						val_2=val_2->clone()->multiply(valG_2->clone());
						((CSequence*)val)->SetAtX(valI,val_2);
					}
				}
				if (!valOK) { dbg->log(0,"Coupling value does not meet to generic coupling",cpl,val,valG); CEException::halt(); break; }
			} else {
				if (valG->ClassName() == CSequence::CLASS_NAME) { dbg->log(0,"Coupling value does not meet to generic coupling",cpl,val,valG); valOK=false; CEException::halt(); break; }
				val=val->multiply(valG->clone());
			}

			// adjust value rules
			CEPtrArray rProcs_DA; rProcs_DA.add(this); rProcs_DA.add(m_phys);
			// evaluate variables/repl rules
			CESocketRule rVars((void*)CSModelLoader::EvaluateSubstRules, &rProcs_DA, CESocketRule::USE_EXTERNAL_DATA_ARRAY);
			// evaluate functions (Delta,Mass and others)
			CESocketRule rProcs((void*)CSModelLoader::EvaluateCouplRProcs, &rProcs_DA,
				CESocketRule::USE_EXTERNAL_DATA_ARRAY|CESocketRule::SR_APPLY_TO_CHANGED_SOURCE|CESocketRule::SR_APPLY_TO_SUBITEMS_FIRST);

			val = CExpression::reduce(rProcs.ApplyTo(val));
			val = CExpression::reduce(rProcs.ApplyTo(val)); // repeat for Mass[F[3,{gn}]]. TODO: avoid this extra call
			val = CExpression::arrge(rVars.ApplyTo(val));

			rProcs_DA.clearall_without_delete_objects();
			if (val==NULL) { dbg->log(2,"INTERNAL ERROR: NULL value"); CEException::halt(); continue; }
			if (CEBasic::isZeroSeq(val)) { // drop zero coupling (another check for vCorr, vCorrCT will be further)
				dbg->log(45,"Zero value dropped out"); valOK=false;
			}

			// val may be: v, {v,v}, {{v,v},{v,v},...}
			// Summarize {{v,v},{v2,v2},...}->{{v+v2,v+v2}}
			if ((valOK) && (val->ClassName() == CSequence::CLASS_NAME)) {
				CSequence* valSec=(CSequence*)val; int valSecS=valSec->size();
				//dbg->log(10,"valSec",valSec);
				int itmSize=-1;
				for(i=0; i<valSecS; i++) {
					CExpression* valItem=valSec->GetAt(i); if (valItem==NULL) { valOK=false;itmSize=-1; break; }
					if (valItem->ClassName() != CSequence::CLASS_NAME) { itmSize=-1; break; }
					int valItemS=((CSequence*)valItem)->size();
					if (itmSize >= 0) { if(itmSize!=valItemS){valOK=false;itmSize=-1;break;} } else { itmSize=valItemS; }
				}
				//dbg->log(10,itmSize,"itmSize: %ld");
				if ((valOK)&&(itmSize>0)) {
					CSequence* val_resummed=new CSequence();
					for(int itmI=0; itmI<itmSize; itmI++) {
						if (!valOK) break;
						CExpression* valI=NULL;
						for(i=0; i<valSecS; i++) {
							if (!valOK) break;
							CSequence* valItemSec=(CSequence*)valSec->GetAt(i);
							CExpression* valItemSecI=valItemSec->GetAt(itmI); if (valItemSecI==NULL) { valOK=false; break; }
							valItemSecI=valItemSecI->clone();
							valItemSec->copyAttributesTo(valItemSecI,true);
							if (valI!=NULL) { valI=valI->add(valItemSecI); } else { valI=valItemSecI; }
						}
						if (valI!=NULL) { val_resummed->addItem(valI); } else { valOK=false; }
					}
					if (valOK) {
						val->copyAttributesTo(val_resummed,true);
						delete val; val=val_resummed;
					} else {
						dbg->log(0, "ERROR: re-summation of vertex value failed", val, val_resummed);
						delete val_resummed;
					}
				} else if ((valOK)&&(itmSize==0)) {
					dbg->log(0, "WARNING: 0-sized sequence vertex value. Zero involved.");
					delete val; val=CENumber::GetZero();
				}
				//dbg->log(10,"val new:",val);
			}

			if (valOK) {
				// Arrange infinities
				if (val!=NULL) { val=CExpression::arrge(val); val=CRArrangeInfinity::ArrageInfinity(val); }
			}

			CExpression* vCorr=NULL; CExpression* vCorrCT=NULL;
			if (valOK) {
				bool valStructOK=true;
				if (val->ClassName()==CSequence::CLASS_NAME) {
					// Note: two-loop counter-terms are ignored
					if (((CSequence*)val)->size() >= 2) {
						vCorr=((CSequence*)val)->GetAt(0); vCorrCT=((CSequence*)val)->GetAt(1);
					} else if (((CSequence*)val)->size() == 1) {
						CExpression* val0=((CSequence*)val)->GetAt(0);
						if (val0->ClassName()==CSequence::CLASS_NAME) {
							if (((CSequence*)val0)->size() == 2) {
								vCorr=((CSequence*)val0)->GetAt(0); vCorrCT=((CSequence*)val0)->GetAt(1);
							} else if (((CSequence*)val0)->size() == 1) {
								CExpression* val0=((CSequence*)val)->GetAt(0);
							} else { valStructOK=false; }
						} else {
							vCorr=val0;
						}
					} else { valStructOK=false; }
				} else {
					vCorr=val;
				}
				if (!valStructOK) { dbg->log(0, "ERROR: invalid correspondence value structure spotted", val, cpl); CEException::halt(); }
			}

			// Note: vCorr & vCorrCT are NOT-cloned values. Do not modify them!
			if (CEBasic::isZeroSeq(vCorr)) vCorr=NULL;
			if (CEBasic::isZeroSeq(vCorrCT)) vCorrCT=NULL;
			if (valOK && ((vCorr==NULL)&&(vCorrCT==NULL))) { valOK=false; /*drop zero coupling*/ }

			if ((vCorr!=NULL) && (partN<=2)) { // no 2-tale vertices
				dbg->log(0,partN,"WARNING: non-counterterm vertex with %ld tales. Vertex IGNORED.",cpl);
				valOK=false; vCorr=NULL;
			}

			if ((partN<2) || (partN>4)) {
				dbg->log(0, partN, "ERROR. Correspondence rule created, but can't be installed. Vertex with %ld tales is not supported.");
				valOK=false;
			}

#ifdef LOADMODEL_DROP_ZERO_PROPS // Do not use this mechanism! Inf vertex * 1/Inf propagator is allowed! TODO: fix it!
			if (valOK) {
				bool bHasZeroPropParticle=false; // check for zero-propagator particles (Unitary gauge)
				for(i=0; i<partN; i++) {
					CPhysModelFA::CFAPatricleClassInfo* inf = (CPhysModelFA::CFAPatricleClassInfo*)partInfos.GetAt(i); if(inf==NULL) { CEException::halt(NULL,__FILE__,__LINE__); continue; }
					long pkI=inf->m_PK; if(arr_idPSign.GetAt(i)<=0)pkI=-pkI;
					if (CSModelLoader::bIsZeroPropagatorPK(pkI)) { bHasZeroPropParticle=true; break; }
				}
				if (bHasZeroPropParticle) { valOK=false; bLogValueIfDropped=false; }
			}
#endif //LOADMODEL_DROP_ZERO_PROPS

			/*
			// Debug addition: drop out all couplings except of e-gamma
			if (valOK) {
				bool bDropCpl=false; // check for zero-propagator particles (Unitary gauge)
				for(i=0; i<partN; i++) {
					CPhysModelFA::CFAPatricleClassInfo* inf = (CPhysModelFA::CFAPatricleClassInfo*)partInfos.GetAt(i); if(inf==NULL) { bDropCpl=true; break; }
					long pkI=inf->m_PK; if(pkI<0)pkI=-pkI; if ((pkI != CEParticle::PARTICLE_ELECTRON) && (pkI != CEParticle::PARTICLE_PHOTON)) { bDropCpl=true; }
				}
				if (bDropCpl) { valOK=false; bLogValueIfDropped=false; }
			}
			*/

			// add correspondence rule
			if (valOK) {
				// CPhysModelFA::addVertexCorr() parameters
				bool del_vals=false;
				int pk1=0; int pk2=0; int pk3=0; int pk4=0;
				CExpression* vInd1=NULL; CExpression* vInd2=NULL; CExpression* vInd3=NULL; CExpression* vInd4=NULL;
				CExpression* pPart1=NULL; CExpression* pPart2=NULL; CExpression* pPart3=NULL; CExpression* pPart4=NULL;

				for(i=0; i<partN; i++) {
					CPhysModelFA::CFAGenericCplPartInfo* genInf=(CPhysModelFA::CFAGenericCplPartInfo*)genCpl->m_partInfos.GetAt(i); if(genInf==NULL){ CEException::halt("unexpected NULL in CSModelLoader::InitCouplingsList"); continue; }
					CExpression* vInd=NULL;
					if (genInf->m_VecIndices.size()>0) {
						if (genInf->m_VecIndices.size()>1) { dbg->log(0, "ERROR: Tensor particles are not implemented.",genInf); CEException::halt(); continue; }
						vInd=genInf->m_VecIndices.GetAt(0);
					}
					CPhysModelFA::CFAPatricleClassInfo* inf = (CPhysModelFA::CFAPatricleClassInfo*)partInfos.GetAt(i); if(inf==NULL) { CEException::halt("NULL particle info array item"); continue; }
					long pkI=inf->m_PK;
					if (arr_idPSign.GetAt(i)<=0) pkI=-pkI;

					// push down all the vector indices
					if (vInd!=NULL) {
						if (!del_vals) { if(vCorr!=NULL)vCorr=vCorr->clone(); if(vCorrCT!=NULL)vCorrCT=vCorrCT->clone(); del_vals=true; }

						if (vCorr!=NULL) {
							CExpression* vCorrNew=CEBasic::InverseTensorIndex(vCorr,vInd,true);
							if (vCorrNew!=NULL) {
								delete vCorr; vCorr=vCorrNew;
							} else { dbg->log(0,"Warning: can't arrange vector index to be covariant", vCorr, vInd); }
						}
						if (vCorrCT!=NULL) {
							CExpression* vCorrCTNew=CEBasic::InverseTensorIndex(vCorrCT,vInd,true);
							if (vCorrCTNew!=NULL) {
								delete vCorrCT; vCorrCT=vCorrCTNew;
							} else { dbg->log(0,"Warning: can't arrange vector index to be covariant", vCorrCT, vInd); }
						}
					}

					switch(i) {
					case 0: pPart1=genInf->m_MomSym; vInd1=vInd; pk1=pkI; break;
					case 1: pPart2=genInf->m_MomSym; vInd2=vInd; pk2=pkI; break;
					case 2: pPart3=genInf->m_MomSym; vInd3=vInd; pk3=pkI; break;
					case 3: pPart4=genInf->m_MomSym; vInd4=vInd; pk4=pkI; break;
					}
				}

				// arrange genCpl signs (arrange possibility is already checked!)
				CPhysModelFA::CFAModelGenericCpl* genCplCpy=new CPhysModelFA::CFAModelGenericCpl(*genCpl);
				for(i=0; i<partN; i++) {
					CPhysModelFA::CFAGenericCplPartInfo* genInf=(CPhysModelFA::CFAGenericCplPartInfo*)genCplCpy->m_partInfos.GetAt(i);
					if (genInf==NULL) { dbg->log(0, i, "ERROR: genCplCpy->m_partInfos.GetAt(%ld)==NULL",genCplCpy); delete genCplCpy; genCplCpy=NULL; break; }
					long iSign = arr_idPSign.GetAt(i); if ((iSign!=1) && (iSign!=-1)) { dbg->log(0,iSign,"ERROR: arr_idPSign.GetAt() enexpected sign value: %ld (#2)"); CEException::halt(); }
					CExpression* _SignOld=genInf->m_SignSym;
					genInf->m_SignSym=CENumber::GetUnit(); if(iSign<0)genInf->m_SignSym->inverseFlag();
					if (_SignOld!=NULL) delete _SignOld;
				}

				// Similar-lines factor, 1/6 for HHH, 1/4 for W+W-W+W-
				// Factor is suppressed by diagram factor & additional crossings.
				// TODO: remove it after crossing problem fixing
				long iSimFactor=1;
				if (partN==2) { iSimFactor=CDiagramVertexInfo::CalcPartsIdentityOrder(pk1,pk2); }
				else if (partN==3) { iSimFactor=CDiagramVertexInfo::CalcPartsIdentityOrder(pk1,pk2,pk3); }
				else if (partN==4) { iSimFactor=CDiagramVertexInfo::CalcPartsIdentityOrder(pk1,pk2,pk3,pk4); }
				else if (partN!=1) { dbg->log(0,partN,"ERROR: unexpected number of tales: %ld"); CEException::halt(); }
				if (iSimFactor!=1) {
					CENumber nFact((long)iSimFactor);
					if (!del_vals) { if(vCorr!=NULL)vCorr=vCorr->clone(); if(vCorrCT!=NULL)vCorrCT=vCorrCT->clone(); del_vals=true; }
					if (vCorr!=NULL) { vCorr=CExpression::arrge(vCorr->divide(nFact.clone())); }
					if (vCorrCT!=NULL) { vCorrCT=CExpression::arrge(vCorrCT->divide(nFact.clone())); }
				}

				bool addedOK = m_phys->addVertexCorr(pk1, pk2, pk3, pk4, vCorr, vCorrCT,
					pPart1, vInd1, pPart2, vInd2, pPart3, vInd3, pPart4, vInd4,
					genCplCpy);
				if (genCplCpy!=NULL) delete genCplCpy;
				if (del_vals) { if(vCorr!=NULL)delete vCorr; if(vCorrCT!=NULL)delete vCorrCT; }

				if (addedOK) {
					dbg->log(25,"Correspondence rule: ", partInfo, val);
					nRulesCreated++; iIndConfigFilled++;
				} else {
					if (m_phys->m_LastErrorMsg.length() > 0) {
						dbg->log(0,"Correspondence rule addition failed. ERROR: ", m_phys->m_LastErrorMsg.getData());
					} else {
						dbg->log(0,"Correspondence rule addition failed. ERROR UNKNOWN.");
					}
					dbg->log(2,"Correspondence rule: ", partInfo, val);
					init_cpls_OK=false; // break & retrun false
				}
			} else {
				if (bLogValueIfDropped) dbg->log(2,"Value dropped: ", val);
			}
			delete val;
		}
		partInfos.clearall_without_delete_objects(); partInfos0.clearall_without_delete_objects();
		dbg->log(45, iIndConfigTried, iIndConfigFilled, "Index configurations tried: %ld, filled: %ld");
	}
	//	*****************************************************	<scroll>
	if (pw != NULL) CEProcessor::dismissProgressWatch(pw);
	//	*****************************************************	<\scroll>
	dbg->log(2, nRulesCreated, "Total %ld vertex rule(s) created");
	return init_cpls_OK;
}

CSequence* CSModelLoader::FindClassDescriptionSeq(const char* idName, int idNumber) {
	CSequence* res=NULL;
	//dbg->log(2, idNumber, "CSModelLoader::FindClassDescriptionSeq %ld");
	//dbg->log(2, "CSModelLoader::FindClassDescriptionSeq: ", idName);
	for(int i=0; i<m_ClassesDescription.size(); i++) {
		CExpression* descItemE=m_ClassesDescription.GetAt(i); if(descItemE==NULL){ dbg->log(0, "Internal ERROR: NULL particle class description"); continue; }
		if (descItemE->ClassName() != CSequence::CLASS_NAME) { dbg->log(0, "Internal ERROR: particle class description is not a CSequence",descItemE); continue; }
		if (((CSequence*)descItemE)->size() != 2) { dbg->log(0, "ERROR: particle class description is not a CSequence of size 2",descItemE); continue; }
		CExpression* faPClassID = ((CSequence*)descItemE)->GetAt(0);
		if (faPClassID==NULL) { dbg->log(0, "Internal ERROR: NULL particle class ID"); continue; }
		// class name
		if (faPClassID->ClassName() == CEFunction::CLASS_NAME) {
			if (!CEUtil::equals( ((CEFunction*)faPClassID)->getTitle(),idName) )continue;
			CExpression* par0=NULL;
			if (((CEFunction*)faPClassID)->params_num() > 0) par0=((CEFunction*)faPClassID)->get_param(0);
			if (par0!=NULL) {
				if (par0->ClassName() != CENumber::CLASS_NAME) {dbg->log(0, "ERROR: non-numeric field parameter", descItemE); continue; } // extra check for now
				if (((CENumber*)par0)->getLong() != idNumber) continue;
			} else {
				if (idNumber != -1) continue;
			}
		} else {
			CEString stmp; stmp.AddString(faPClassID->toString(STORE_TEX));
			if (!CEUtil::equals( stmp.getData(),idName) )continue;
			if (idNumber != -1) continue;
		}
		CExpression* descSeqParsE = ((CSequence*)descItemE)->GetAt(1);
		if (descSeqParsE==NULL) { dbg->log(0, "Internal ERROR: NULL particle class parameters"); break; }
		if (descSeqParsE->ClassName() != CSequence::CLASS_NAME) { dbg->log(0, "Internal ERROR: particle class parameters is not a CSequence",descSeqParsE); break; }
		res=(CSequence*)descSeqParsE;
		break;
	}
	//dbg->log(2, "CSModelLoader::FindClassDescriptionSeq: res", res);
	return res;
}

// deletes previous (*eMassOrig) and (*ePropLabel) values if changed
bool CSModelLoader::ParseDescSequence(CSequence* descSeq, CSequence* sqQNumbers, CSequence* sqIndices, CExpression** ePropLabel, CExpression** eMassOrig, int* iPropArrow, int* iPropType, int* iSelfConj) {
	if (descSeq==NULL) return false;
	bool parseOK=true;
	long descSeqS=descSeq->size(); long descSeqI=0;
	for(;descSeqI<descSeqS;descSeqI++) {
		if(!parseOK) break;
		CExpression* descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
		if (CEBasic::equalsToSymbol(descItem,"QuantumNumbers")) {
			descSeqI++; if (descSeqI>=descSeqS)break; descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
			if (sqQNumbers!=NULL) {
				if (descItem->ClassName() == CSequence::CLASS_NAME) {
					sqQNumbers->copy_from((CSequence*)descItem);
				} else {
					sqQNumbers->addItem(descItem->clone());
				}
			}
		} else if (CEBasic::equalsToSymbol(descItem,"PropagatorArrow")) {
			descSeqI++; if (descSeqI>=descSeqS)break; descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
			if (iPropArrow!=NULL) {
				if (CEBasic::equalsToSymbol(descItem,"Forward")) {
					*iPropArrow=1;
				} else if (CEBasic::equalsToSymbol(descItem,"None")) {
					*iPropArrow=0;
				} else {
					dbg->log(0,"Unknown PropagatorArrow value",descItem); *iPropArrow=0;
				}
			}
		} else if (CEBasic::equalsToSymbol(descItem,"PropagatorType")) {
			descSeqI++; if (descSeqI>=descSeqS)break; descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
			CSequence itemArr;
			if (descItem->ClassName() == CSequence::CLASS_NAME) { itemArr.copy_from((CSequence*)descItem); }
			else { itemArr.addItem(descItem->clone()); }
			if (iPropType!=NULL) {
				int tmpPropType=0x00;
				for (int i=0; i<itemArr.size(); i++) {
					if (CEBasic::equalsToSymbol(itemArr.GetAt(i),"Straight")) tmpPropType|=CLDrawDiagView::LINE_FERMION;
					else if (CEBasic::equalsToSymbol(itemArr.GetAt(i),"Sine")) tmpPropType|=CLDrawDiagView::LINE_BOSON;
					else if (CEBasic::equalsToSymbol(itemArr.GetAt(i),"ScalarDash")) tmpPropType|=CLDrawDiagView::LINE_SCALAR;
					else if (CEBasic::equalsToSymbol(itemArr.GetAt(i),"GhostDash")) tmpPropType|=CLDrawDiagView::LINE_SCALAR;
				}
				// if (tmpPropType == 0x00) tmpPropType=0x01; // 0 will calls default selection procedure
				*iPropType=tmpPropType;
			}
		} else if (CEBasic::equalsToSymbol(descItem,"PropagatorLabel")) {
			descSeqI++; if (descSeqI>=descSeqS)break; descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
			if (ePropLabel != NULL) {
				if (*ePropLabel!=NULL) { delete *ePropLabel; *ePropLabel=NULL; }
				//(*ePropLabel)->AddString(descItem->toString(STORE_TEX));
				//(*ePropLabel)->AddString(descItem->toString(STORE_WR_MATHEMATICA));
				(*ePropLabel)=descItem->clone();
			}
		} else if (CEBasic::equalsToSymbol(descItem,"Indices")) {
			descSeqI++; if (descSeqI>=descSeqS)break; descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
			if (sqIndices != NULL) {
				if (descItem->ClassName() == CSequence::CLASS_NAME) {
					sqIndices->copy_from((CSequence*)descItem);
				} else {
					sqIndices->addItem(descItem->clone());
				}
			}
		} else if (CEBasic::equalsToSymbol(descItem,"SelfConjugate")) {
			descSeqI++; if (descSeqI>=descSeqS)break; descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
			if (iSelfConj!=NULL) {
				//dbg->log(2,"CEBasic::equalsToSymbol(descItem,SelfConjugate)",descItem);
				if (CEBasic::equalsToSymbol(descItem,"False")) {
					*iSelfConj=0;
				} else if (CEBasic::equalsToSymbol(descItem,"True")) {
					*iSelfConj=1;
				} else {
					dbg->log(0,"Unknown SelfConjugate value",descItem); parseOK=false; break;
				}
			}
		} else if (CEBasic::equalsToSymbol(descItem,"Mass")) {
			descSeqI++; if (descSeqI>=descSeqS)break; descItem=descSeq->GetAt(descSeqI); if (descItem==NULL) continue;
			if (eMassOrig!=NULL) {
				if (*eMassOrig!=NULL) { delete *eMassOrig; *eMassOrig=NULL; }
				if ((descItem->ClassName() == CESymbol::CLASS_NAME) || (descItem->ClassName() == CENumber::CLASS_NAME)) {
					*eMassOrig=descItem->clone();
				} else {
					*eMassOrig=new CESymbol("MASSUNKNOWN");
					dbg->log(0,"ERROR: Mass should be symbol or number.",descItem); parseOK=false; break;
				}
			}
		} else {
			dbg->log(2,"WARNING: Class description parameter ignored: ",descItem);
		}
	}
	return parseOK;
}

bool CSModelLoader::InitParticleInfo(CPhysModelFA::CFAPatricleClassInfo* inf) { // init pk and mass symbol
	dbg->log(95,"InitParticleInfo orig",inf);
	//dbg->log(2,"InitParticleInfo orig",inf);
	if (inf == NULL) return false;
	if (m_phys == NULL) return false;
	int i;

	// data sources:
	//inf->m_idClassName
	//inf->m_QuantumNumbers
	//m_KinIndices; // (CESymbol*):{{F}{V,Lorentz},{S},{SV,Lorentz},{U}}
	//m_ClassesDescription
	const char* idName=inf->m_idClassName.getData();
	const char* idNumberS = inf->m_idNumber.getData();
	long idNumber=-1; if (!CEUtil::atolSafe(idNumberS,&idNumber)) idNumber=-1;
	int indsNum=inf->m_Indices.GetSize();
	int indGenetation=-1; // -1 means unknown
	int indColor=-1; // -1 means unknown
	// find class template
	CSequence* seqClassDescription=FindClassDescriptionSeq(idName,idNumber); if (seqClassDescription==NULL) { dbg->log(0,"ERROR: Class description for particle is not found",inf); return false; }
	CSequence descQNumbers; CSequence descIndices; CExpression* propLabel=NULL; CESymbol* propLabelS=NULL;
	bool descOK=ParseDescSequence(seqClassDescription, &descQNumbers, &descIndices, &propLabel, NULL, NULL, NULL, NULL);
	if (indsNum != descIndices.size()) { dbg->log(0, "ERROR: particle indices don't meet the class description ones", inf, &descIndices); return false; }
	for (i=0;i<descIndices.size();i++) {
		CExpression* descIndI=descIndices.GetAt(i);
		if (descIndI->ClassName() != CEFunction::CLASS_NAME) continue;
		if (!CEUtil::equals(((CEFunction*)descIndI)->getTitle(),"Index")) continue;
		if (((CEFunction*)descIndI)->params_num()!=1) continue;
		CExpression* descIndIPar0=((CEFunction*)descIndI)->get_param(0);
		if (descIndIPar0->ClassName()!=CESymbol::CLASS_NAME) continue;
		if (CEUtil::equals( ((CESymbol*)descIndIPar0)->get_name(), "Generation")) { indGenetation=inf->m_Indices.GetAt(i); }
		else if (CEUtil::equals( ((CESymbol*)descIndIPar0)->get_name(), "Colour")) { indColor=inf->m_Indices.GetAt(i); }
	}
	if (propLabel!=NULL) { if (propLabel->ClassName()==CESymbol::CLASS_NAME) propLabelS=(CESymbol*)propLabel; }
	if (inf->m_SymMassOrig==NULL) { dbg->log(0,"ERROR: no mass for particle is available", inf); return false; }

	long prePK=0;
	//long prePKFlags=0;

	bool bIsFermion=false; // eslewhere is boson
	bool bIsVector=false; // eslewhere is scalar/ghost
	bool bIsNeutral=false; // eslewhere is charged
	bool bIsFermionLepton=false;
	bool bIsFermionQuark=false;
	bool bIsQuarkConjugated=false; // {d,s,b}, eslewhere is {u,c,t}
	bool bIsGhost=false;

	CESymbol tmpChargeSym("Charge"); // temporary

	// TODO: make more smart selection here
	if (CEUtil::equals(idName,"F")) bIsFermion=true;

	bool kiInfoFound=false;
	for (i=0; i<m_KinIndices.size(); i++) {
		CEPtrArray* rec=(CEPtrArray*)m_KinIndices.GetAt(i);
		if(rec==NULL)continue; if(rec->size()<1)continue; // errors
		const char* xName= ((CESymbol*)rec->GetAt(0))->get_name();
		if (CEUtil::equals(idName,xName)) {
			if (rec->size()==1) {
				bIsVector=false;
			} else if (rec->size()>=2) {
				const char* xIndex = ((CESymbol*)rec->GetAt(1))->get_name();
				if (CEUtil::equals(xIndex,"Lorentz")) { bIsVector=true; }
				if ((rec->size()>2) || (!bIsVector)) { dbg->log(0, "WARNING: unexpected KinIndices[] structure:",rec); }
			}
			kiInfoFound=true; break;
		}
	}
	if (!kiInfoFound) { dbg->log(0, "ERROR: No KinIndices[] structure found for particle",inf); return false; }

	bIsNeutral=inf->m_SelfConjugated;

	int iCharge3=0;
	for (i=0;i<inf->m_QuantumNumbers.size();i++) {
		CExpression* qnItem=inf->m_QuantumNumbers.GetAt(i); if(qnItem==NULL)continue;
		if (qnItem->ClassName() == CESymbol::CLASS_NAME) {
			const char* qnInd=((CESymbol*)qnItem)->get_name();
			if (CEUtil::equals(qnInd,"LeptonNumber")) { bIsFermionLepton=true; continue; }
			if (CEUtil::equals(qnInd,"GhostNumber")) { bIsGhost=true; continue; }
		}
		// charge * 3
		CExpression* tmpCharge=CExpression::reduce(qnItem->clone()->divide(tmpChargeSym.clone())->multiply(CENumber::GetThree()));
		if (tmpCharge->ClassName() == CENumber::CLASS_NAME) { iCharge3=((CENumber*)tmpCharge)->getLong(); }
		delete tmpCharge;
	}
	if (bIsNeutral && (iCharge3 != 0)) { dbg->log(0,"ERROR: Self-congugated particle with non-zero charge.",inf,seqClassDescription); return false; }

	if (bIsFermion) {
		if (idNumber>=0) {
			if (idNumber==1) {
				if ((inf->m_QuantumNumbers.size()==1) && (bIsFermionLepton) && (iCharge3==0) && CEBasic::isZero(inf->m_SymMassOrig)) {
					if (indGenetation==1) prePK=CEParticle::PARTICLE_NEUTRINO_ELECTRON;
					else if (indGenetation==2) prePK=CEParticle::PARTICLE_NEUTRINO_MUON;
					else if (indGenetation==3) prePK=CEParticle::PARTICLE_NEUTRINO_TAU;
					//else if (indGenetation<0) prePKFlags|=CEParticle::PARTICLE_NEUTRINO;
				}
			} else if (idNumber==2) {
				if ((inf->m_QuantumNumbers.size()==2) && (iCharge3==-3) && (bIsFermionLepton)) {
					if (indGenetation==1) prePK=CEParticle::PARTICLE_ELECTRON;
					else if (indGenetation==2) prePK=CEParticle::PARTICLE_MUON;
					else if (indGenetation==3) prePK=CEParticle::PARTICLE_TAU_LEPTON;
					//else if (indGenetation<0) prePKFlags|=CEParticle::PARTICLE_LEPTON;
				}
			} else if (idNumber==3) {
				if ((inf->m_QuantumNumbers.size()==1) && (!bIsFermionLepton) && (iCharge3==2) && (indGenetation>0) && (indColor>0)) {
					if (indGenetation==1) prePK=CEParticle::PARTICLE_QUARK_U;
					else if (indGenetation==2) prePK=CEParticle::PARTICLE_QUARK_C;
					else if (indGenetation==3) prePK=CEParticle::PARTICLE_QUARK_T;
				}
			} else if (idNumber==4) {
				if ((inf->m_QuantumNumbers.size()==1) && (!bIsFermionLepton) && (iCharge3==-1) && (indGenetation>0) && (indColor>0)) {
					if (indGenetation==1) prePK=CEParticle::PARTICLE_QUARK_D;
					else if (indGenetation==2) prePK=CEParticle::PARTICLE_QUARK_S;
					else if (indGenetation==3) prePK=CEParticle::PARTICLE_QUARK_B;
				}
			}
		}
	} else {
		if ((CEUtil::equals(idName,"V")) && (idNumber>=0) && (propLabelS!=NULL)) {
			if (idNumber==1) {
				if ((indsNum==0) && (inf->m_QuantumNumbers.size()==0) && (bIsVector) && (bIsNeutral) && CEBasic::isZero(inf->m_SymMassOrig)) {
					prePK=CEParticle::PARTICLE_PHOTON;
				}
			} else if (idNumber==2) {
				if ((indsNum==0) && (inf->m_QuantumNumbers.size()==0) && (bIsVector) && (bIsNeutral) && (CEUtil::equals(propLabelS->get_name(),"Z"))) {
					prePK=CEParticle::PARTICLE_Z_BOSON;
				}
			} else if (idNumber==3) {
				if ((indsNum==0) && (inf->m_QuantumNumbers.size()==1) && (bIsVector) && (CEUtil::equals(propLabelS->get_name(),"W"))) {
					if (iCharge3==-3) prePK=CEParticle::PARTICLE_W_BOSON;
					else if (iCharge3==3) {
						// prePK=-CEParticle::PARTICLE_W_BOSON;
						prePK=0; // unknown particle
					}
				}
			}
		}
	}
	if (prePK < 0) { dbg->log(0,prePK,"WARNING: negative default PK presumed for particle (%ld). Default value is ignored."); prePK=0; }
	//dbg->log(2,prePK,"InitParticleInfo prePK: 0x%lx");

	//prePK=0; // truncate particle auto-identification
	if (prePK!=0) {
		inf->m_PK=prePK;
		if (inf->m_SymMass != NULL) { dbg->log(0, "ERROR: !NULL mass spotted"); CEException::halt(); } //TODO: remove it
		inf->m_SymMass=inf->m_SymMassOrig->clone();
	} else {
		long pk=m_phys->CreateNewExtraParticleID();
		if (bIsFermion) {
			pk|=CEParticle::PARTICLE_MODEL_FERMION;
			if (bIsFermionQuark) {
				pk|=CEParticle::PARTICLE_QUARK;
				if (bIsQuarkConjugated) pk|=CEParticle::PARTICLE_QUARK_CONJUGETED;
			} else {
				if (bIsFermionLepton) pk|=CEParticle::PARTICLE_LEPTON;
			}
		} else {
			pk|=CEParticle::PARTICLE_MODEL_BOSON;
			if (bIsVector) pk|=CEParticle::PARTICLE_VECTOR_BOSON;
			else pk|=CEParticle::PARTICLE_SCALAR;
			if (bIsGhost) pk|=CEParticle::PARTICLE_FPGHOST;
			if (bIsNeutral) pk|=CEParticle::PARTICLE_NEUTRAL_BOSON;
		}
		inf->m_PK=pk;
		if (inf->m_SymMass != NULL) { dbg->log(0, "ERROR: !NULL mass spotted"); CEException::halt(); } //TODO: remove it
		inf->m_SymMass=inf->m_SymMassOrig->clone();
	}

	// init m_sPropagatorLabelNb,m_sPropagatorLabelTeX
	if (inf->m_PropagatorLabelM != NULL) {
		inf->m_sPropagatorLabelNb.Reset(); inf->m_sPropagatorLabelNb.AddString(inf->m_PropagatorLabelM->toString(STORE_WR_MATHEMATICA));
		const char* sbst=CConst::GetValidSymMSubst(inf->m_sPropagatorLabelNb.getData());
		if (sbst!=NULL) { inf->m_sPropagatorLabelNb.Reset(); inf->m_sPropagatorLabelNb.AddString(sbst); }
	}
	if (inf->m_PropagatorLabel != NULL) {
		inf->m_sPropagatorLabelTex.Reset(); inf->m_sPropagatorLabelTex.AddString(inf->m_PropagatorLabel->toString(STORE_TEX));
		const char* sbst=CConst::GetValidSymTexSubst(inf->m_sPropagatorLabelTex.getData());
		if (sbst!=NULL) { inf->m_sPropagatorLabelTex.Reset(); inf->m_sPropagatorLabelTex.AddString(sbst); }
	}
	
	dbg->log(95,"InitParticleInfo res:",inf);
	//dbg->log(2,"InitParticleInfo res:",inf);
	return true;
	//return false;
}

/*
*	Used to adjust external propagator of vector bosons.
*	deletes source, check parameters, NULL if failed
*	format: POL_VEC_NAME[V[j1], mom, li2]
*/
CExpression* CSModelLoader::ExtractPolVector(CExpression* ext_prop,
											 const char* vec_id, CExpression* num, CExpression* mom, CExpression* ind) {
	if ((ext_prop==NULL)||(vec_id==NULL)||(num==NULL)||(mom==NULL)||(ind==NULL)) { CEException::halt("Internal error. CSModelLoader::ExtractPolVector(): invalid call."); if(ext_prop!=NULL)delete ext_prop; return false; }
	if (ext_prop->ClassName() == CEFunction::CLASS_NAME) {
		CEFunction* f=(CEFunction*)ext_prop;
		if (CEUtil::equals(f->getTitle(),POL_VEC_NAME)) {
			if (ComparePolVecParams(f,vec_id,num,mom,ind)) {
				CExpression* res=CENumber::GetUnit(); f->copyAttributesTo(res,true); return res;
			}
		}
	} else if (ext_prop->ClassName() == CEProduct::CLASS_NAME) {
		CEProduct* pr=(CEProduct*)ext_prop;
		pr->inline_subproducts(true); // additional arrange
		long s=pr->size(); long i;
		bool pol_vec_found=false; bool extract_OK=false;
		for(i=0; i<s; i++) {
			CExpression* item=pr->GetAt(i);
			if (item->ClassName() == CEFunction::CLASS_NAME) {
				CEFunction* f=(CEFunction*)item;
				if (CEUtil::equals(f->getTitle(),POL_VEC_NAME)) {
					if (ComparePolVecParams(f,vec_id,num,mom,ind)) {
						if (pol_vec_found) { extract_OK=false; break; }
						CENumber nSubRes((long)1); f->copyAttributesTo(&nSubRes,true); pr->SetAt(i,&nSubRes);
						pol_vec_found=true; extract_OK=true;
					}
				}
			}
		}
		if(extract_OK) return CExpression::arrge(pr);
	}
	delete ext_prop;
	return NULL;
}

/*
*	not affects source, func_name is already checked (POL_VEC_NAME)
*	format: POL_VEC_NAME[V[j1], mom, li2]
*/
bool CSModelLoader::ComparePolVecParams(CEFunction* pol_vec,
									  const char* vec_id, CExpression* num, CExpression* mom, CExpression* ind) {
	CExpression* i1=NULL; CExpression* i2=NULL; CExpression* i3=NULL;
	if (pol_vec->params_num() == 1) {
		CExpression* iX=pol_vec->get_param(0);
		if (iX!=NULL) {
			if (iX->ClassName()==CSequence::CLASS_NAME) {
				CSequence* secX=(CSequence*)iX; if(secX->size()==3){ i1=secX->GetAt(0); i2=secX->GetAt(1); i3=secX->GetAt(2); }
			}
		}
	} else if (pol_vec->params_num() == 3) { i1=pol_vec->get_param(0); i2=pol_vec->get_param(1); i3=pol_vec->get_param(2); }
	if ((i1!=NULL) && (i2!=NULL) && (i3!=NULL)) {
		if (i1->ClassName()==CEFunction::CLASS_NAME) {
			CEFunction* f1 =(CEFunction*)i1;
			if ((CEUtil::equals(f1->getTitle(),vec_id)) && (f1->params_num()==1)) {
				CExpression* numX=f1->get_param(0);
				if (numX!=NULL) {
					if ((numX->equals(num)) && (i2->equals(mom)) && (i3->equals(ind))) {
						return true;
					}
				}
			}
		}
	}
	return false;
}
